/*###############################################################

	AppleEventUtilities.c
	
	Useful (read: indispensable) routines when programming
	AppleScript into applications.  Several routines in this
	file are derived from Apples sample code, and have the
	appropriate copyrights.
	
	MacTelnet
		 1998-2003 by Kevin Grant.
		 2001-2003 by Ian Anderson.
		 1986-1994 University of Illinois Board of Trustees
		(see About box for full list of U of I contributors).
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.
	
	You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <AppleEvents.h>
#include <Folders.h>
#include <MacErrors.h>
#include <OSA.h>
#include <PLStringFuncs.h>
#include <Processes.h>
#include <Resources.h>

// resource includes
#include "StringResources.h"

// library includes
#include "DataAccessLib.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "Console.h"
#include "FileUtilities.h"
#include "TelnetErrors.h"



//
// public methods
//

/*!
Adds a result to a reply record.  Usually,
this is what you do in response to a given
Apple Event.  If you provide a non-zero
error code, the result is added to the
reply record as an Apple Event error number
or error message, depending on the data.

(3.0)
*/
OSStatus
AppleEventUtilities_AddResultToReply	(AEDesc*	inResultAppleEventPtr,
										 AEDesc*	inoutReplyAppleEventPtr,
										 OSStatus	inError)
{
	OSStatus		result = inError;
	
	
	// the reply must exist, and there must be a result to put in it
	if ((inoutReplyAppleEventPtr->descriptorType != typeNull) &&
		(inResultAppleEventPtr->descriptorType != typeNull))
	{
		if (inError == noErr)
		{
			// a normal result
			result = AEPutParamDesc(inoutReplyAppleEventPtr, keyDirectObject, inResultAppleEventPtr);
		}
		else
		{
			// abnormal result; maybe error feedback is being given (check possible feedback types)
			switch (inResultAppleEventPtr->descriptorType)
			{
				case typeSInt16: // error code
					(OSStatus)AEPutParamDesc(inoutReplyAppleEventPtr, keyErrorNumber,
												inResultAppleEventPtr);
					break;
				
				case typeChar: // error message
					(OSStatus)AEPutParamDesc(inoutReplyAppleEventPtr, keyErrorString,
												inResultAppleEventPtr);
					break;
				
				default:
					result = errAETypeError;
					break;
			}
		}
	}
	
	return result;
}// AddResultToReply


/*!
Use this Carbon-compliant routine to access data
in an Apple Event descriptor structure.  You may
not directly access the data handle, because this
approach is not supported under Carbon.  This
routine behaves exactly like the Carbon routine
AEGetDescData(), except it works on pre-Carbon
systems.

Use AEGetDescDataSize() beforehand if you wish to
allocate a precise amount of storage before using
this routine.

WARNING:	You must allocate the space yourself.
			Calling this routine will not provide
			you with a pointer to existing data;
			you must create a sufficiently large
			buffer, and then use this routine to
			copy descriptor data into your new
			buffer.

(3.0)
*/
OSStatus
AppleEventUtilities_CopyDescriptorData		(AEDesc const*	inDescriptorPtr,
											 void*			outDataPtr,
											 Size 			inDataPtrMaximumSize,
											 Size*			outActualSizePtr)
{
	return AppleEventUtilities_CopyDescriptorDataAs(inDescriptorPtr->descriptorType,
													inDescriptorPtr, outDataPtr, inDataPtrMaximumSize,
													outActualSizePtr);
}// CopyDescriptorData


/*!
Use this Carbon-compliant routine to access data
in an Apple Event descriptor structure and coerce
it (if possible) to the specified type.

Use AEGetDescDataSize() beforehand if you wish to
allocate a precise amount of storage before using
this routine.

WARNING:	You must allocate the space yourself.
			Calling this routine will not provide
			you with a pointer to existing data;
			you must create a sufficiently large
			buffer, and then use this routine to
			copy descriptor data into your new
			buffer.

(3.0)
*/
OSStatus
AppleEventUtilities_CopyDescriptorDataAs	(DescType		inDesiredType,
											 AEDesc const*	inDescriptorPtr,
											 void*			outDataPtr,
											 Size 			inDataPtrMaximumSize,
											 Size*			outActualSizePtr)
{
	OSStatus		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("copying descriptor data");
	if ((inDescriptorPtr == null) || (outDataPtr == null)) result = memPCErr;
	else
	{
		AEDesc*		descPtr = (AEDesc*)Memory_NewPtr(sizeof(AEDesc));
		
		
		if (descPtr != null)
		{
			descPtr->descriptorType = typeNull;
			descPtr->dataHandle = null;
			if (inDesiredType != inDescriptorPtr->descriptorType)
			{
				//Console_WriteLine("source and destination types differ; coercing...");
				result = AECoerceDesc(inDescriptorPtr, inDesiredType, descPtr);
			}
			else
			{
				//Console_WriteLine("source and destination types identical; duplicating...");
				result = AEDuplicateDesc(inDescriptorPtr, descPtr);
			}
		}
		else result = memFullErr;
		
		if (result == noErr)
		{
		#if TARGET_API_MAC_OS8
			Size		byteCountToCopy = AEGetDescDataSize(descPtr);
			
			
			// copy the data, reporting an error (but not aborting the copy) if theres not enough room
			Console_WriteLine("determining how much to copy");
			if (byteCountToCopy > inDataPtrMaximumSize)
			{
				Console_WriteLine("too little space to copy completely!");
				byteCountToCopy = inDataPtrMaximumSize;
				result = memSCErr;
			}
			BlockMoveData(*(descPtr->dataHandle), outDataPtr, byteCountToCopy);
		#else
			// under Carbon, there is a new API for this purpose
			result = AEGetDescData(descPtr, outDataPtr, inDataPtrMaximumSize);
		#endif
			if (outActualSizePtr != null) *outActualSizePtr = AEGetDescDataSize(descPtr);
		}
		if (descPtr != null) Memory_DisposePtr((Ptr*)&descPtr);
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// CopyDescriptorDataAs


/*!
Resolves the specified descriptor if it is an object
specifier (returning the resolved token), or otherwise
duplicates the given descriptor, returning the
duplicate.  You typically use this in all event handlers,
to ensure that you have a token descriptor and not a
reference to an object.

(3.0)
*/
OSStatus
AppleEventUtilities_DuplicateOrResolveObject	(AEDesc const*	inDescriptor,
												 AEDesc*		outDuplicate)
{
	OSStatus		result = noErr;
	
	
	if (inDescriptor->descriptorType == typeObjectSpecifier)
	{
		Console_WriteLine("given an object reference");
		result = AEResolve(inDescriptor, kAEIDoMinimum, outDuplicate);
	}
	else result = AEDuplicateDesc(inDescriptor, outDuplicate);
	
	return result;
}// DuplicateOrResolveObject


/*!
Connects to the default scripting component and
executes the script defined by the specified data.

You typically obtain data by reading from resources
of type "kOSAGenericScriptingComponentSubtype" in
compiled script files.

Specify "false" for "inNotifyUser" if you want to
have control over errors that a script might
encounter (in either case, the error code of any
error is returned).  If "true" is specified, this
routine automatically displays an error alert with
the proper message if any errors occur.

(3.0)
*/
OSStatus
AppleEventUtilities_ExecuteScript	(Handle		inScriptData,
									 Boolean	UNUSED_ARGUMENT(inNotifyUser))
{
	OSStatus				result = noErr;
	ComponentInstance		scriptingComponent = OpenDefaultComponent
													(kOSAComponentType,
													kOSAGenericScriptingComponentSubtype);
	
	
	if (scriptingComponent != null)
	{
		AEDesc		scriptDataDescriptor;
		OSAID		scriptID = 0,
					resultID = 0;
		
		
		(OSStatus)AppleEventUtilities_InitAEDesc(&scriptDataDescriptor);
		result = AECreateDesc(typeOSAGenericStorage, *inScriptData, GetHandleSize(inScriptData),
								&scriptDataDescriptor);
		result = OSALoad(scriptingComponent, &scriptDataDescriptor,
						kOSAModeNull, &scriptID);
		if (result == noErr)
		{
			result = OSAExecute(scriptingComponent, scriptID, kOSANullScript,
								kOSAModeNull, &resultID);
			(OSStatus)OSADispose(scriptingComponent, scriptID);
			(OSStatus)OSADispose(scriptingComponent, resultID);
			if (result == errOSAScriptError)
			{
				// the script reported an error - display it to the user
				Str255		helpText,
							message,
							title;
				AEDesc		errorMessageDescriptor,
							errorNumberDescriptor,
							titleDescriptor;
				
				
				(OSStatus)AppleEventUtilities_InitAEDesc(&errorMessageDescriptor);
				(OSStatus)AppleEventUtilities_InitAEDesc(&errorNumberDescriptor);
				(OSStatus)AppleEventUtilities_InitAEDesc(&titleDescriptor);
				
				// get the error number (as the function result), and display it in the help text
				if (OSAScriptError(scriptingComponent, kOSAErrorNumber,
									typeSInt16, &errorNumberDescriptor) == noErr)
				{
					Str31		numberString;
					
					
					AppleEventUtilities_CopyDescriptorData(&errorNumberDescriptor,
															&result, sizeof(result),
															null/* actual size storage */);
					(OSStatus)AEDisposeDesc(&errorNumberDescriptor);
					NumToString(result, numberString);
					GetIndString(helpText, rStringsNoteAlerts, siScriptErrorHelpText);
					StringUtilities_PMetaSetDefaults();
					StringUtilities_PBuild(helpText, numberString, EMPTY_PSTRING,
											EMPTY_PSTRING, EMPTY_PSTRING);
				}
				
				// use the exact error message from the script, if provided
				if (OSAScriptError(scriptingComponent, kOSAErrorMessage,
										cStringClass, &errorMessageDescriptor) == noErr)
				{
					AppleEventUtilities_CopyDescriptorData(&errorMessageDescriptor,
															message + 1, sizeof(message) - 1,
															null/* actual size storage */);
					message[0] = (char)AEGetDescDataSize(&errorMessageDescriptor);
					(OSErr)AEDisposeDesc(&errorMessageDescriptor);
				}
				
				// get the alert title
				GetIndString(title, rStringsAlertTitles, siAlertTitleScriptError);
				
				// display an alert for important errors
				if ((result != noErr) && (result != userCanceledErr))
				{
					TelnetAlertParameterBlock	parameters;
					
					
					Memory_WriteZero((Ptr)&parameters, sizeof(parameters));
					parameters.dialog.type = kAlertStopAlert;
					parameters.flags.alertSound = true;
					parameters.text.large = message;
					parameters.text.small = helpText;
					parameters.text.title = title;
					(SInt16)Errors_DisplayParameterizedAlert(&parameters);
				}
			}
		}
		(OSStatus)CloseComponent(scriptingComponent);
	}
	else result = errOSACantOpenComponent;
	
	return result;
}// ExecuteScript


/*!
Runs a script from a file.  In Mac OS 8.x and 9.x,
this file would be expected to contain a resource
of type "kOSAGenericScriptingComponentSubtype",
from which the necessary script data is read.

If the specified file refers to an alias, this
routine automatically resolves it before trying
to execute it as a script.

Specify "false" for "inNotifyUser" if you want to
have control over errors that a script might
encounter (in either case, the error code of any
error is returned).  If "true" is specified, this
routine automatically displays an error alert with
the proper message if any errors occur.

(3.0)
*/
OSStatus
AppleEventUtilities_ExecuteScriptFile	(FSSpec const*	inFile,
										 Boolean		inNotifyUser)
{
	OSStatus	result = noErr;
	FSSpec		file;
	
	
	// if this file is in fact an alias, ensure the original file is used for reading resources
	(OSStatus)FSMakeFSSpec(inFile->vRefNum, inFile->parID, inFile->name, &file);
	result = Navigation_EnsureOriginalFile(&file);
	
	if (result == noErr)
	{
		SInt16		scriptResourceFile = -1;
		
		
		scriptResourceFile = FSpOpenResFile(&file, fsRdPerm);
		result = ResError();
		if (result == noErr)
		{
			Handle		scriptData = null;
			SInt16		oldResFile = CurResFile();
			
			
			UseResFile(scriptResourceFile);
			if (Count1Resources(kOSAScriptResourceType) <= 0) result = resNotFound;
			else
			{
				// get first available script
				scriptData = Get1IndResource(kOSAScriptResourceType, 1);
				result = ResError();
			}
			if (result == noErr)
			{
				if (scriptData == null) result = resNotFound;
				else
				{
					DetachResource(scriptData);
					result = AppleEventUtilities_ExecuteScript(scriptData, inNotifyUser);
					
					// free memory occupied by the resource data
					Memory_ReleaseDetachedResource(&scriptData);
				}
			}
			UseResFile(oldResFile);
		}
		else
		{
			// failed...could probably tell the user something useful here
		}
	}
	
	return result;
}// ExecuteScriptFile


OSStatus
FindAppOnVolume 	(OSType		sig,
					 short		vRefNum,
					 FSSpec*	file)
{
	DTPBRec			pb;
	OSStatus		err = noErr;
	short			ioDTRefNum = 0,
					i = 0;
	FileInfo		fInfo;
	FSSpec			candidate;
	unsigned long	lastModDateTime = 0L,
					maxLastModDateTime = 0L;
	
	
	Memory_WriteZero((Ptr)&pb, sizeof(DTPBRec));
	pb.ioCompletion = null;
	pb.ioVRefNum = vRefNum;
	pb.ioNamePtr = null;
	err = PBDTGetPath(&pb);
	if (err != noErr) return err;
	ioDTRefNum = pb.ioDTRefNum;

	Memory_WriteZero((Ptr)&pb, sizeof(DTPBRec));
	pb.ioCompletion = null;
	pb.ioIndex = 0;
	pb.ioFileCreator = sig;
	pb.ioNamePtr = file->name;
	pb.ioDTRefNum = ioDTRefNum;
	err = PBDTGetAPPLSync(&pb);
	
	if (err == fnfErr || err == paramErr) return afpItemNotFound;
	if (err != noErr) return err;

	file->vRefNum = vRefNum;
	file->parID = pb.ioAPPLParID;
	
	err = FSpGetFInfo(file, (FInfo*)&fInfo);
	if (err == noErr) return noErr;
	
	i = 1;
	maxLastModDateTime = 0;
	for (;;)
	{
		Memory_WriteZero((Ptr)&pb, sizeof(DTPBRec)); 
		pb.ioCompletion = null;
		pb.ioIndex = i;
		pb.ioFileCreator = sig;
		pb.ioNamePtr = candidate.name;
		pb.ioDTRefNum = ioDTRefNum;
		err = PBDTGetAPPLSync(&pb);
		if (err != noErr) break;
		candidate.vRefNum = vRefNum;
		candidate.parID = pb.ioAPPLParID;
		Console_WriteLine("getlastmod");
		err = FileUtilities_GetLastModDateTime(file, &lastModDateTime);
		if (err == noErr)
		{
			if (lastModDateTime > maxLastModDateTime)
			{
				maxLastModDateTime = lastModDateTime;
				*file = candidate;
			}
		}
		Console_WriteValue("getlastmod error:", err);
		i++;
	}
	
	return maxLastModDateTime > 0 ? noErr : afpItemNotFound;
}// FindAppOnVolume


/*!
Locates an application based on its creator
signature.  If the application is running, the
given process serial number structure will be
filled in and "running" will be set to "true".
The location of the file on disk is also
provided.  If any problems occur, an error
code is returned; otherwise, "noErr" is returned.

(2.6)
*/
OSStatus
FindAppFromSig		(OSType					sig,
					 FSSpec*				fSpec,
					 Boolean*				running,
					 ProcessSerialNumber*	psn)
{
	OSStatus	result = noErr;
	short		sysVRefNum = 0,
				vRefNum = 0,
				index = 0;
	Boolean		hasDesktopDB = false;
	long		junk = 0L;
	
	
	Console_BeginFunction();
	if (running != null)
	{
		Console_WriteLine("checking to see if the application is running");
		result = FindRunningAppBySignature(sig, fSpec, psn);
		*running = true;
		if (result == noErr) return noErr;
		*running = false;
		if (result != procNotFound) return result;
	}
	result = FindFolder(kOnSystemDisk, kSystemFolderType, false, &sysVRefNum, &junk);
	if (result == noErr)
	{
		Console_WriteLine("searching volumes");
		vRefNum = sysVRefNum;
		index = 0;
		for (;;)
		{
			if (index == 0 || vRefNum != sysVRefNum)
			{
				result = FileUtilities_VolHasDesktopDB(vRefNum, &hasDesktopDB);
				if (result != noErr) return result;
				if (hasDesktopDB)
				{
					Console_WriteLine("found desktop database, searching");
					result = FindAppOnVolume(sig, vRefNum, fSpec);
					if (result != afpItemNotFound)
					{	
						Console_WriteValue("result", result);
						return result;
					}
				}
			}
			index++;
			result = FileUtilities_GetIndVolume(index, &vRefNum);
			Console_WriteValue("result", result);
			if (result == nsvErr)
			{
				Console_WriteValue("result", result);
				return fnfErr;
			}
			if (result != noErr) break;
		}
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// FindAppFromSig


/*----------------------------------------------------------------------------
	FindRunningAppBySignature 
	
	Find a running app given its signature.
	
	Entry:	sig = signature of app.
			
	Exit:	function result = error code.
				= procNotFound if not running.
			*fSpec = file spec of app. 
			*psn = process serial number of running app.
----------------------------------------------------------------------------*/

OSStatus FindRunningAppBySignature (OSType sig, FSSpec *fSpec,
	ProcessSerialNumber *psn)
{
	OSStatus err = noErr;
	ProcessInfoRec info;
	
	psn->highLongOfPSN = 0;
	psn->lowLongOfPSN  = kNoProcess;
	while (true) {
		err = GetNextProcess(psn);
		if (err != noErr) return err;
		info.processInfoLength = sizeof(info);
		info.processName = null;
		info.processAppSpec = fSpec;
		err = GetProcessInformation(psn, &info);
		if (err != noErr) return err;
		if (info.processSignature == sig) return noErr;
	}
}


/*!
Initializes an "AEDesc" data structure.  Since
Carbon may eventually make Apple Event structures
opaque, consistent use of this routine is
recommended to ease future development efforts.

You typically replace this...

	AEDesc	desc = { typeNull, null };

...with this:

	AEDesc	desc;
	(OSStatus)AppleEventUtilities_InitAEDesc(&desc);

(3.0)
*/
OSStatus
AppleEventUtilities_InitAEDesc		(AEDesc*	inoutDescriptorPtr)
{
	OSStatus		result = noErr;
	
	
	if (inoutDescriptorPtr == null) result = memPCErr;
	else
	{
		inoutDescriptorPtr->descriptorType = typeNull;
		inoutDescriptorPtr->dataHandle = null;
	}
	return result;
}// InitAEDesc

/*----------------------------------------------------------------------------
	LaunchAppWithEventAndString 
	
	Launch an application with an initial event with a string parameter.
	
	Entry:	running = true if application is running, in which case
				it is sent the odoc event.
			appSpec = file spec of application.
			*psn = process serial number of app if it is running.
			eventClass = event class.
			eventID = event id.
			keyword = parameter keyword (keyDirectObject if string is the
				direct object).
			str = the string parameter for the event.
			launchFileFlags = file flags.
			launchControlFlags = control flags.
			
	Exit:	function result = error code.
----------------------------------------------------------------------------*/

OSStatus
LaunchAppWithEventAndString		(Boolean				running,
								 FSSpec*				appSpec,
								 ProcessSerialNumber*	psn,
								 DescType				eventClass,
								 DescType				eventID,
								 DescType				keyword,
								 char*					str,
								 Size					stringBufferSize,
								 unsigned short			launchFileFlags,
								 unsigned short			launchControlFlags)
{
	ProcessSerialNumber		thePSN;
	AEDesc					target = {0, null},
							stringDesc = {0, null},
							launchDesc = {0, null};
	AppleEvent				theEvent = {0, null},
							theReply = {0, null};
	Boolean					autoParamValue = false;
	OSStatus				result = noErr;
	
	
	Console_BeginFunction();
	
	// make a copy of the specified application parameters; since
	// Apple Events are sent to "application" Apple Event objects,
	// a descriptor must be created that reflects these parameters
	if (running) thePSN = *psn;
	result = AECreateDesc(typeProcessSerialNumber, &thePSN, sizeof(thePSN), &target); 
	if (result == noErr)
	{
		// create an event that will be passed to the application; if
		// successful, generate a string parameter for the event, and
		// finally if that succeeds, launch the application (or just
		// activate it if its running) and send the event
		Console_WriteLine("constructing Apple Event for application");
		result = AECreateAppleEvent(eventClass, eventID, &target, kAutoGenerateReturnID,
									kAnyTransactionID, &theEvent);
		if (result == noErr)
		{
			result = AECreateDesc(typeChar, str, stringBufferSize, &stringDesc);
			if (result == noErr)
			{
				result = AEPutParamDesc(&theEvent, keyword, &stringDesc);
				if (result == noErr)
				{
					if (running)
					{
						Console_WriteLine("application is running");
						result = AESend(&theEvent, &theReply, kAENoReply, kAENormalPriority,
										kNoTimeOut, null, null);
						if (result == noErr)
						{
							if ((launchControlFlags & launchDontSwitch) == 0)
							{
								// switch-on-launch was specified, so activate the application
								result = SetFrontProcess(psn);
							}
						}
					}
					else
					{
						Console_WriteLine("application is not running");
						result = AECoerceDesc(&theEvent, typeAppParameters, &launchDesc);
						if (result == noErr)
						{
							// launch the application
							LaunchParamBlockRec		application;
							Handle					appParamsHandle = null;
							
							
							appParamsHandle = Memory_NewHandle(sizeof(AppParameters));
							if (appParamsHandle != null)
							{
								Size		actualSize = 0L;
								
								
								result = AppleEventUtilities_CopyDescriptorData
											(&launchDesc, *appParamsHandle,
											GetHandleSize(appParamsHandle), &actualSize);
							}
							else result = memFullErr;
							
							if (result == noErr)
							{
								application.launchAppSpec = appSpec;
								application.launchAppParameters = (AppParametersPtr)*appParamsHandle;
								application.launchBlockID = extendedBlock;
								application.launchEPBLength = extendedBlockLen;
								application.launchFileFlags = launchFileFlags;
								application.launchControlFlags = launchControlFlags;
								result = LaunchApplication(&application);
								if ((result == noErr) && FlagManager_Test(kFlagAppearance1_1API))
								{
									(OSStatus)PlayThemeSound(kThemeSoundLaunchApp);
								}
							}
						}
					}
				}
			}
		}
	}
	
	if (target.dataHandle != null) AEDisposeDesc(&target);
	if (stringDesc.dataHandle != null) AEDisposeDesc(&stringDesc);
	if (launchDesc.dataHandle != null) AEDisposeDesc(&launchDesc);
	if (theEvent.dataHandle != null) AEDisposeDesc(&theEvent);
	if (theReply.dataHandle != null) AEDisposeDesc(&theReply);
	
	Console_EndFunction();
	
	return result;
}// LaunchAppWithEventAndString


/*!
This method determines if the specified event is
a valid reply event.  Specifically, an invalid
reply is one with "null" data and a type of
"typeNull".

(3.0)
*/
Boolean
AppleEventUtilities_IsValidReply		(AppleEventPtr		inAppleEventPtr)
{
	Boolean		result = false;
	
	
	if (inAppleEventPtr != null) result = ((inAppleEventPtr->descriptorType != typeNull) ||
											(inAppleEventPtr->dataHandle != null));
	return result;
}// IsValidReply


/*!
To determine if an Apple Event object exists,
use this method.

(3.0)
*/
pascal OSErr
AppleEventUtilities_ObjectExists	(AppleEvent const*	inAppleEventPtr,
									 AppleEvent*		outReplyAppleEventPtr, 
									 SInt32				UNUSED_ARGUMENT(inData))
{
	AEDesc		directObject,
				directToken,
				replyDesc;
	OSErr		result = noErr;
	
	
	(OSStatus)AppleEventUtilities_InitAEDesc(&directObject);
	(OSStatus)AppleEventUtilities_InitAEDesc(&directToken);
	(OSStatus)AppleEventUtilities_InitAEDesc(&replyDesc);
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: exists event");
	
	(OSStatus)AEGetParamDesc(inAppleEventPtr, keyDirectObject, typeWildCard, &directObject);
	result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
	if (result == noErr)
	{
		Boolean       exists = false;
		
		
		if (directObject.descriptorType == typeNull) exists = true; // the application always exists!
		else
		{
			result = AEResolve(&directObject, kAEIDoMinimum, &directToken);
			if (result == noErr)
			{
				Console_WriteLine("successfully resolved");
				if (directToken.descriptorType == typeAEList)
				{
					long		itemCount = 0L;
					
					
					// for lists, objects exist if the list is not empty
					Console_WriteLine("given a list of objects; any non-empty list should exist");
					result = AECountItems(&directToken, &itemCount);
					if (result == noErr)
					{
						if (itemCount) exists = true;
						else exists = false;
					}
				}
				else exists = true;
			}
			else exists = false;
		}
		
		if (result == noErr)
		{
			result = AECreateDesc(typeBoolean, (Ptr)&exists, sizeof(exists), &replyDesc);
			if (result == noErr)
			{
				// return whether or not the object exists
				result = STATIC_CAST(AppleEventUtilities_AddResultToReply
										(&replyDesc, outReplyAppleEventPtr, result), OSErr);
			}
		}
	}
	
	if (directObject.dataHandle) AEDisposeDesc(&directObject);
	if (directToken.dataHandle) AEDisposeToken(&directToken);
	if (replyDesc.dataHandle) AEDisposeDesc(&replyDesc);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// ObjectExists


/*!
This method determines if the required events for the
specified Apple Event were received properly.  If so,
"noErr" is returned; otherwise, "errAEEventNotHandled"
is returned.

Bug: (?)	It appears as though any event that states
			it has no direct object will cause this
			routine to report a parameter error.  I
			cant figure out why this would be; it
			neither appears to be in this code, nor in
			the user terminology resource.  Any ideas?

(3.0)
*/
OSErr
AppleEventUtilities_RequiredParametersError		(AppleEvent const*	inAppleEventPtr)
{
	DescType	returnedType = '????';
	Size		actualSize = 0;
	OSErr		result = noErr;
	
	
	result = AEGetAttributePtr(inAppleEventPtr, keyMissedKeywordAttr,
								typeWildCard, &returnedType, null/* data pointer */,
								0/* maximum size */, &actualSize);
	if (result == errAEDescNotFound)
	{
		// got all the required parameters, so the event is fine
		result = noErr;
	}
	else if (result == noErr)
	{
		// missed keyword attribute exists, so a required parameter was not provided; this is an error
		result = errAEParamMissed;
	}
	
	return result;
}// RequiredParametersError