/*###############################################################

	ConnectionAE.c
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <string.h>

// Mac includes
#include <AEDataModel.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <MacErrors.h>
#include <PLStringFuncs.h>
#include <Processes.h>

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "ConnectionAE.h"
#include "ConnectionData.h"
#include "Connections.h"
#include "Console.h"
#include "GenesisAE.h"
#include "InstallAE.h"
#include "ObjectClassesAE.h"
#include "Parse.h"
#include "Preferences.h"
#include "ResourceFileManager.h"
#include "Terminology.h"

// internal methods
static ConnectionSetup**		nameToConnectionSetup	(StringPtr			inString,
														 StringPtr			inWindowName,
														 Boolean			inUseDefaultTerminal);
static void						sendCR					(ConnectionDataPtr	inConnectionDataPtr,
														 Boolean const		inEcho);
static OSErr					sendData				(AEDesc const*		inSendWhat,
														 AEDesc const*		inToWhichConnection,
														 Boolean const		inEcho,
														 Boolean const		inNewLine);



//
// public methods
//

/*!
Handles the following event types:

kTelnetEventClass
-----------------
	kTelnetEventIDConnectionOpen

(3.0)
*/
pascal OSErr
ConnectionAE_HandleConnectionOpen	(const AppleEvent*		inAppleEventPtr,
									 AppleEvent*			outReplyAppleEventPtr,
									 UInt32					UNUSED_ARGUMENT(inData))
{
	DescType			returnedType = typeWildCard;
	Str255				hostText,
						titleText;
	Size				actualSize = 0L;
	OSErr				result = noErr;
	OSErr				error = noErr; // generic error with no lasting meaning
	ConnectionSetup**	theConnectionDataHandle = null;
	
	
	// set up host parameter: this is a required parameter
	{
		error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyConnectionOpenHost,
								kTelnetParmTypeConnectionOpenHost,
								&returnedType, hostText + 1,
								sizeof(hostText) - 1 * sizeof(unsigned char), &actualSize);
		hostText[0] = (unsigned char)(actualSize / sizeof(unsigned char));
		if (error != noErr) PLstrcpy(hostText, EMPTY_PSTRING);
	}
	
	// set up port-number parameter: defaults to the telnet port if the parameter is not specified
	{
		SInt16		port = 23; // the default telnet port
		
		
		error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyConnectionOpenPortNumber,
								kTelnetParmTypeConnectionOpenPortNumber,
								&returnedType, &port, sizeof(SInt16), &actualSize);
		if (error == noErr)
		{
			Str31		portString;
			SInt32		portAsLong = port;
			
			
			NumToString(portAsLong, portString);
			PLstrcat(hostText, "\p:");
			PLstrcat(hostText, portString);
		}
	}
	
	// set up window title parameter: defaults to nothing if the parameter is not specified
	{
		error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyConnectionOpenWindowTitle,
								kTelnetParmTypeConnectionOpenWindowTitle,
								&returnedType, titleText + 1,
								sizeof(titleText) - 1 * sizeof(unsigned char), &actualSize);
		titleText[0] = (unsigned char)(actualSize / sizeof(unsigned char));
		if (error != noErr) PLstrcpy(titleText, EMPTY_PSTRING);
	}
	
	theConnectionDataHandle = nameToConnectionSetup(hostText, titleText, true/* use default terminal */);
	if (theConnectionDataHandle != null)
	{
		HLock((Handle)theConnectionDataHandle);
		HLock((Handle)(**theConnectionDataHandle).session);
		
		// set up connection method parameter: defaults to telnet if the parameter is not specified
		{
			FourCharCode		method = kTelnetEnumeratedClassConnectionMethodTelnet;
			
			
			error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyConnectionOpenProtocol,
									kTelnetParmTypeConnectionOpenProtocol,
									&returnedType, &method, sizeof(FourCharCode), &actualSize);
			
			// translate the information - incomplete
			switch (method)
			{
				case kTelnetEnumeratedClassConnectionMethodTelnet:
					(**theConnectionDataHandle).ftpState = 0;
					break;
				
				case kTelnetEnumeratedClassConnectionMethodFTP:
					(**theConnectionDataHandle).ftpState = 1;
					(**(**theConnectionDataHandle).session).halfdup = 1;
					break;
				
				default:
					// ???
					break;
			}
		}
		
		// set up open timeout parameter: defaults to the preferences value if the parameter is not specified
		{
			UInt32		timeout = 0L; // the default timeout in seconds
			Size		actualSize = 0L;
			
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TimeoutOpen,
										sizeof(timeout), &timeout, &actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				timeout = 10; // assume a 10-second timeout, if preference cant be found
			}
			error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyConnectionOpenTimeoutInSeconds,
									kTelnetParmTypeConnectionOpenTimeoutInSeconds,
									&returnedType, &timeout, sizeof(UInt32), &actualSize);
			(**theConnectionDataHandle).openTimeout = timeout;
		}
		
		(**(**theConnectionDataHandle).session).authenticate = false;
		(**(**theConnectionDataHandle).session).encrypt = false;
		
		HUnlock((Handle)(**theConnectionDataHandle).session);
		HUnlock((Handle)theConnectionDataHandle);
		
		// temp - somehow, this call produces an error, because this event specifies it has no direct object?
		//result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
		if (result == noErr)
		{
			// make the connection
			(Boolean)Connections_CreateOne(theConnectionDataHandle);
			
			// check to see if the reply is valid - if so, return a connection record
			if (AppleEventUtilities_IsValidReply(outReplyAppleEventPtr))
			{
				// return the new connection - incomplete
				SInt32		junk = 0L;
				
				
				error = AEPutParamPtr(outReplyAppleEventPtr, keyDirectObject, kTelnetClassConnection, &junk/* wrong, temp. */,
										   sizeof(SInt32/* wrong, temp. */));
			}
		}
		
		Alert_ReportOSErr(result);
	}
	else
	{
		result = memFullErr;
	}
	
	return result;
}// HandleConnectionOpen


/*!
Handles the following event types:

kTelnetEventClass
-----------------
	kTelnetEventIDDataSend

(3.0)
*/
pascal OSErr
ConnectionAE_HandleSend		(const AppleEvent*		inAppleEventPtr,
							 AppleEvent*			UNUSED_ARGUMENT(outReplyAppleEventPtr),
							 UInt32					UNUSED_ARGUMENT(inData))
{
	AEDesc        	directObject = { typeNull, null },
					connectionObject = { typeNull, null };
	Size			actualSize = 0L;
	OSErr         	result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: send event");
	
	result = AEGetParamDesc(inAppleEventPtr, keyDirectObject, typeWildCard, &directObject);
	
	// set up parameters
	if (result == noErr)
	{
		DescType		returnedType = typeWildCard;
		
		
		// get required parameters first
		result = AEGetParamDesc(inAppleEventPtr, kTelnetParameterKeyDataSendToConnection,
								typeObjectSpecifier, &connectionObject);
		if (result == noErr)
		{
			Boolean			echo = false,
							newLine = false;
			
			
			// get echo parameter; defaults to false if not specified
			(OSErr)AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyDataSendWithLocalEcho,
									typeBoolean, &returnedType, &echo, sizeof(echo), &actualSize);
			
			// get newline parameter; defaults to false if not specified
			(OSErr)AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyDataSendWithNewLine,
									typeBoolean, &returnedType, &newLine, sizeof(newLine), &actualSize);
			
			// now execute the command, if no required parameters are missing
			result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
			if (result == noErr) result = sendData(&directObject, &connectionObject, echo, newLine);
		}
	}
	
	if (directObject.dataHandle) AEDisposeDesc(&directObject);
	if (connectionObject.dataHandle) AEDisposeDesc(&connectionObject);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// HandleSend


//
// internal methods
//

/*!
This method recognizes the following input strings: "xxxx yyyy"
and "xxxx:yyyy".

If "xxxx" matches a session name, that session record is used.
Otherwise, the default session record is used with "xxxx" as
the DNS hostname.   The "yyyy", if provided, is converted to a
number.  If it is a valid number, it is used as the port to
connect to.

The second argument specifies whether to get the terminal stuff
from the session preferences.  If not, the terminal is set to
null and should be set up properly later.

The proxy is set to null.

A new connection handle is returned.  Dispose of it using the
Memory_DisposeHandle() routine.

WARNING:	Do not pass this routine a blank string.

(2.6)
*/
static ConnectionSetup**
nameToConnectionSetup	(StringPtr		inString,
						 StringPtr		inWindowName,
						 Boolean		inUseDefaultTerminal)
{
	ConnectionSetup**	result = null;
	
	
	result = (ConnectionSetup**)Memory_NewHandleCritical(sizeof(ConnectionSetup));
	if (result != null)
	{
		SessionPrefs**		sessionHandle = null;
		SInt16				portRequested = 0;
		Boolean				foundPort = false,
							wasAlias = false;
		
		
		foundPort = Connections_ProcessHostNameString(inString, &portRequested);
		
		ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
		sessionHandle = (SessionPrefs**)Get1NamedResource(SESSIONPREFS_RESTYPE, inString);
		if (sessionHandle == null) 
		{
			// connect to host xxxx w/default session
			sessionHandle = Preferences_GetDefaultSession();
			DetachResource((Handle)sessionHandle);
			HLock((Handle)sessionHandle);
			BlockMove(inString, (**sessionHandle).hostname, 64/* bytes */);
		}
		else 
		{	
			DetachResource((Handle)sessionHandle);
			HLock((Handle)sessionHandle);
			wasAlias = true;
		}
		
		(**result).session = sessionHandle;
		
		ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
		if ((inUseDefaultTerminal)||(wasAlias))
		{
			TerminalPrefs**		terminalHandle = null;
			
			
			terminalHandle = (TerminalPrefs**)Get1NamedResource(TERMINALPREFS_RESTYPE, (**sessionHandle).TerminalEmulation);
			if (terminalHandle == null) terminalHandle = Preferences_GetDefaultTerminal();
			DetachResource((Handle) terminalHandle);
			(**result).terminal = terminalHandle;
		}
		else (**result).terminal = null;
		
		// swap out space so we can make the new window CCP
		{
			SInt32		junk = 0L;
			
			
			MaxMem(&junk);
		}
		HUnlock((Handle)sessionHandle);
		
		{
			UInt16		length = StrLength(inWindowName);
			
			
			if (length > 0) BlockMove(inWindowName, (**result).windowName, (length > 63) ? 64 : (length + 1));
			else ((**result).windowName)[0] = 0;
		}
		(**result).ftpState = ((foundPort) && (portRequested == 21)) ? 1 : 0;
		(**sessionHandle).ip_address = 0;
		
		if (foundPort) (**sessionHandle).port = portRequested;
		
		// set up proxy
		(**result).proxy = null;
	}
	
	return result;
}// nameToConnectionSetup


/*!
Sends a carriage return sequence.  Depending upon the
user preference of carriage return handling, this may
send CR-LF or CR-null.

(3.0)
*/
static void
sendCR	(ConnectionDataPtr		inConnectionDataPtr,
		 Boolean const			inEcho)
{
	// if CR map is on, send CR-null instead of CR-LF
	Network_SendFlushData(inConnectionDataPtr->port,
							(inConnectionDataPtr->crmap) ? "\015\0" : "\015\012", 2);
	if (inEcho) Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)"\012\015", 2);
}// sendCR


/*!
Sends the specified data to the specified connection.
This routine recognizes string data, lists of strings,
and enumerations indicating special characters.

The reply is undefined.

(3.0)
*/
static OSErr
sendData	(AEDesc const*		inSendWhat,
			 AEDesc const*		inToWhichConnection,
			 Boolean const		inEcho,
			 Boolean const		inNewLine)
{
	AEDesc		tokenDescriptor = { typeNull, null },
				connectionDescriptor = { typeNull, null };
	OSErr		result = noErr;
	
	
	result = AppleEventUtilities_DuplicateOrResolveObject(inSendWhat, &tokenDescriptor);
	if (result == noErr)
	{
		result = AppleEventUtilities_DuplicateOrResolveObject(inToWhichConnection, &connectionDescriptor);
		Console_WriteValueFourChars("connection object type", connectionDescriptor.descriptorType);
	}
	if (result == noErr)
	{
		Console_WriteLine("successfully resolved");
		Console_WriteValueFourChars("to type", tokenDescriptor.descriptorType);
		switch (tokenDescriptor.descriptorType)
		{
			case typeAEList:
				{
					AEDesc		listItemDescriptor = { typeNull, null },
								resultDescriptor = { typeNull, null };
					DescType	returnedType;
					long		index = 0L;
					
					
					result = AECountItems(&tokenDescriptor, &index);
					for (; index > 0; index--)
					{
						result = AEGetNthDesc(&tokenDescriptor, index, typeWildCard,
												&returnedType, &listItemDescriptor);
						
						// handle the request by adding the resultant data to the end of the list
						if (result == noErr) result = sendData(&listItemDescriptor, inToWhichConnection, inEcho, inNewLine);
						if (result == noErr) result = AEPutDesc(&tokenDescriptor/* list */, index,
																&resultDescriptor/* item to add */);
						
						if (listItemDescriptor.dataHandle != null) AEDisposeDesc(&listItemDescriptor);
						if (resultDescriptor.dataHandle != null) AEDisposeDesc(&resultDescriptor);
					}
				}
				break;
			
			case typeEnumerated:
				Console_WriteLine("special character");
				{
					FourCharCode	specialCharacter = kTelnetEnumeratedClassSpecialCharacterCarriageReturn;
					Size			actualSize = 0L;
					OSErr			error = noErr;
					
					
					error = AppleEventUtilities_CopyDescriptorData(&tokenDescriptor, &specialCharacter,
																	sizeof(specialCharacter), &actualSize);
					if (error == noErr)
					{
						TelnetToken			token;
						ConnectionDataPtr	connectionDataPtr = null;
						
						
						if (GenesisAE_CreateTokenFromObjectSpecifier(&connectionDescriptor, &token) == noErr)
						{
							connectionDataPtr = token.as.object.data.connection.connectionDataPtr;
						}
						else Console_WriteLine("failed to get token data");
						
						if (connectionDataPtr != null) switch (specialCharacter)
						{
							case kTelnetEnumeratedClassSpecialCharacterAreYouThere:
								Network_SendFlushData(connectionDataPtr->port, "\377\366", 2); // IAC AYT
								break;
							
							case kTelnetEnumeratedClassSpecialCharacterAbortOutput:
								Network_SendFlushData(connectionDataPtr->port, "\377\365", 2); // IAC AO
								connectionDataPtr->timing = 1; // set emulate to TMwait
								Network_SendData(connectionDataPtr->port, "\377\375\006", 3); // send TM
								break;
							
							case kTelnetEnumeratedClassSpecialCharacterBreak:
								Network_SendFlushData(connectionDataPtr->port, "\377\363", 2); // IAC BRK
								break;
							
							case kTelnetEnumeratedClassSpecialCharacterCarriageReturn:
								sendCR(connectionDataPtr, inEcho);
								break;
							
							case kTelnetEnumeratedClassSpecialCharacterInterruptProcess:
								Network_SendFlushData(connectionDataPtr->port, "\377\364", 2); // IAC IP
								connectionDataPtr->timing = 1; // set emulate to TMwait
								Network_SendData(connectionDataPtr->port, "\377\375\006", 3); // send TM
								break;
							
							case kTelnetEnumeratedClassSpecialCharacterEraseCharacter:
								Network_SendFlushData(connectionDataPtr->port, "\377\367", 2); // IAC EC
								break;
							
							case kTelnetEnumeratedClassSpecialCharacterEraseLine:
								Network_SendFlushData(connectionDataPtr->port, "\377\370", 2); // IAC EL
								break;
							
							case kTelnetEnumeratedClassSpecialCharacterEndOfFile:
								Network_SendFlushData(connectionDataPtr->port, "\377\354", 2); // IAC EOF
								Network_SendFlushData(connectionDataPtr->port, "\377\375\006", 3); // send TM
								break;
							
							case kTelnetEnumeratedClassSpecialCharacterLineFeed:
								Network_SendFlushData(connectionDataPtr->port, "\012", 1); // LF
								break;
							
							case kTelnetEnumeratedClassSpecialCharacterSync:
								Network_SendFlush(connectionDataPtr->port);
								Network_MarkNextTransmissionAsUrgent();
								Network_SendFlushData(connectionDataPtr->port, "\377\362", 2); // IAC SYNC
								break;
							
							default:
								// ???
								break;
						}
					}
				}
				break;
			
			default:
				Console_WriteLine("string");
				{
					Ptr			string = null;
					
					
					string = Memory_NewPtr(AEGetDescDataSize(&tokenDescriptor));
					if (string != null)
					{
						Size		actualSize = 0L;
						OSErr		error = noErr;
						
						
						error = AppleEventUtilities_CopyDescriptorDataAs(cStringClass, &tokenDescriptor, string,
																			GetPtrSize(string), &actualSize);
						if (error == noErr)
						{
							TelnetToken			token;
							ConnectionDataPtr	connectionDataPtr = null;
							
							
							Console_WriteLine("request to send this line:");
							Console_WriteLine(string);
							if (GenesisAE_CreateTokenFromObjectSpecifier(&connectionDescriptor, &token) == noErr)
							{
								connectionDataPtr = token.as.object.data.connection.connectionDataPtr;
								if (connectionDataPtr != null)
								{
									Parse_TypeString(connectionDataPtr, string, GetPtrSize(string),
														false/* send to recording scripts */);
									if (inNewLine) sendCR(connectionDataPtr, inEcho);
								}
							}
							else Console_WriteLine("failed to get token data");
						}
					}
					else result = memFullErr;
				}
				break;
		}
	}
	
	if (tokenDescriptor.dataHandle != null) AEDisposeToken(&tokenDescriptor);
	
	return result;
}// sendData