/*###############################################################

	DataAccessAE.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <string.h>

// Mac includes
#include <AEDataModel.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <MacErrors.h>
#include <Processes.h>
#include <Resources.h>

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "BasicTypesAE.h"
#include "Console.h"
#include "DataAccessAE.h"
#include "GetPropertyAE.h"
#include "InstallAE.h"
#include "ObjectClassesAE.h"
#include "RequiredAE.h"
#include "ResourceFileManager.h"
#include "Terminology.h"



// internal methods
static OSErr	getData							(AEDesc*			inFromWhichObject,
												 DescType const		inGetDataOfThisType,
												 AEDesc*			outDataDescriptor,
												 Boolean*			outUsedDesiredDataType);
static OSErr	getDataSize						(AEDesc*			inFromWhichObject,
												 DescType const		inGetDataOfThisType,
												 AEDesc*			outDataDescriptor,
												 Boolean*			outUsedDesiredDataType);
static OSErr	setData							(AEDesc const*		inForWhichObject,
												 AEDesc*			inDataDescriptor);



//
// public methods
//

/*!
Handles the following event types:

kAECoreSuite
------------
	kAEGetData

(3.0)
*/
pascal OSErr
DataAccessAE_HandleGetData		(AppleEvent const*	inAppleEventPtr,
								 AppleEvent*		outReplyAppleEventPtr,
								 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = errAEEventNotHandled;
	AEDesc		directObjectDescriptor = { typeNull, null },
				resultDescriptor = { typeNull, null };
	
	
	// get the object whose data is to be returned
	result = AEGetParamDesc(inAppleEventPtr, keyDirectObject, typeWildCard, &directObjectDescriptor);
	if (result == noErr)
	{
		DescType	returnedType,
					requestedType;
		Size		actualSize = 0L;
		
		
		// try to return the data in a form that is specifically requested
		(OSErr)AEGetParamPtr(inAppleEventPtr, kTelnetPartialCoreParameterKeyGetDataAs,
								kTelnetPartialCoreParmTypeGetDataAs, &returnedType,
								(Ptr)&requestedType, sizeof(requestedType), &actualSize);
		result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
		if (result == noErr)
		{
			// get the data
			AEDesc		booleanDesc = { typeNull, null };
			Boolean		usedDesiredDataType = false; // was the type given in the "as" parameter respected?
			
			
			result = getData(&directObjectDescriptor, requestedType, &resultDescriptor, &usedDesiredDataType);
			if (BasicTypesAE_CreateBooleanDesc(usedDesiredDataType, &booleanDesc) == noErr)
			{
				// AppleScript 1.5: add a "keyAppHandledCoercion" to the reply
				(OSErr)AEPutParamDesc(&resultDescriptor, keyAppHandledCoercion, &booleanDesc);
			}
			result = AppleEventUtilities_AddResultToReply(&resultDescriptor, outReplyAppleEventPtr, result);
		}
	}
	
	if (directObjectDescriptor.dataHandle != null) AEDisposeDesc(&directObjectDescriptor);
	if (resultDescriptor.dataHandle != null) AEDisposeDesc(&resultDescriptor);
	
	return result;
}// HandleGetData


/*!
Handles the following event types:

kAECoreSuite
------------
	kAEGetDataSize

(3.0)
*/
pascal OSErr
DataAccessAE_HandleGetDataSize	(AppleEvent const*	inAppleEventPtr,
								 AppleEvent*		outReplyAppleEventPtr,
								 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = errAEEventNotHandled;
	AEDesc		directObjectDescriptor = { typeNull, null },
				resultDescriptor = { typeNull, null };
	
	
	// get the object whose data size is to be returned
	result = AEGetParamDesc(inAppleEventPtr, keyDirectObject, typeWildCard, &directObjectDescriptor);
	if (result == noErr)
	{
		DescType	returnedType,
					requestedType;
		Size		actualSize = 0L;
		
		
		// try to return the data in a form that is specifically requested
		(OSErr)AEGetParamPtr(inAppleEventPtr, kTelnetPartialCoreParameterKeyGetDataSizeAs,
								kTelnetPartialCoreParmTypeGetDataSizeAs, &returnedType,
								(Ptr)&requestedType, sizeof(requestedType), &actualSize);
		result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
		if (result == noErr)
		{
			// get the data size
			Boolean		usedDesiredDataType = false; // was the type given in the "as" parameter respected?
			
			
			result = getDataSize(&directObjectDescriptor, requestedType, &resultDescriptor, &usedDesiredDataType);
			result = AppleEventUtilities_AddResultToReply(&resultDescriptor, outReplyAppleEventPtr, result);
		}
	}
	
	if (directObjectDescriptor.dataHandle != null) AEDisposeDesc(&directObjectDescriptor);
	if (resultDescriptor.dataHandle != null) AEDisposeDesc(&resultDescriptor);
	
	return result;
}// HandleGetDataSize


/*!
Handles the following event types:

kAECoreSuite
------------
	kAESetData

(3.0)
*/
pascal OSErr
DataAccessAE_HandleSetData		(AppleEvent const*	inAppleEventPtr,
								 AppleEvent*		UNUSED_ARGUMENT(outReplyAppleEventPtr),
								 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = errAEEventNotHandled;
	AEDesc		directObjectDescriptor = { typeNull, null },
				dataDescriptor = { typeNull, null };
	
	
	// get the object whose data is to be modified
	result = AEGetParamDesc(inAppleEventPtr, keyDirectObject, typeWildCard, &directObjectDescriptor);
	if (result == noErr)
	{
		// obtain the data that should be used to make a modification to the object
		(OSErr)AEGetParamDesc(inAppleEventPtr, kTelnetPartialCoreParameterKeySetDataTo,
								kTelnetPartialCoreParmTypeSetDataTo, &dataDescriptor);
		result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
		if (result == noErr)
		{
			// set the data
			result = setData(&directObjectDescriptor, &dataDescriptor);
		}
	}
	
	if (directObjectDescriptor.dataHandle != null) AEDisposeDesc(&directObjectDescriptor);
	if (dataDescriptor.dataHandle != null) AEDisposeDesc(&dataDescriptor);
	
	return result;
}// HandleSetData


//
// internal methods
//

/*!
This routine indirectly handles all get data
Apple Events.

When a script asks for data, AppleScript will
figure out what is being said (for example,
"window 1" means a window referenced by
absolute index) and invoke an object accessor
to construct a token (see "GetObjectAE.c").
The token that is constructed is INTERNAL to
MacTelnet, and is passed in "inFromWhichObject".
This routine uses the token to determine the
nature of the request, and therefore how to
handle it.  Most data is attached to a high
level object as a property, so usually the
token will be a property token.

(3.0)
*/
static OSErr
getData		(AEDesc*			inFromWhichObject,
			 DescType const		inGetDataOfThisType,
			 AEDesc*			outDataDescriptor,
			 Boolean*			outUsedDesiredDataType)
{
	AEDesc		tokenDescriptor = { typeNull, null };
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: get data event");
	Console_WriteValueFourChars("desired type", inGetDataOfThisType);
	
	if (inFromWhichObject->descriptorType == typeObjectSpecifier)
	{
		// all thats known is that this is an object; resolve it into a known object type
		// using a previously-installed OSL object accessor (see "InstallAE.c" for details)
		Console_WriteLine("given an object reference");
		result = AEResolve(inFromWhichObject, kAEIDoMinimum, &tokenDescriptor);
	}
	else result = AEDuplicateDesc(inFromWhichObject, &tokenDescriptor);
	
	if (result == noErr)
	{
		Console_WriteLine("successfully resolved");
		Console_WriteValueFourChars("to type", tokenDescriptor.descriptorType);
		switch (tokenDescriptor.descriptorType)
		{
			case typeAEList:
				{
					long		index = 0L;
					
					
					result = AECountItems(&tokenDescriptor, &index);
					if (result == noErr)
					{
						AEDesc		listItemDescriptor = { typeNull, null },
									resultDescriptor = { typeNull, null };
						DescType	returnedType;
						
						
						for (; index > 0; index--)
						{
							result = AEGetNthDesc(&tokenDescriptor, index, typeWildCard,
													&returnedType, &listItemDescriptor);
							
							// handle the request by adding the resultant data to the end of the list
							if (result == noErr) result = getData(&listItemDescriptor,
																	inGetDataOfThisType, &resultDescriptor,
																	outUsedDesiredDataType);
							if (result == noErr) result = AEPutDesc(&tokenDescriptor/* list */, index,
																	&resultDescriptor/* item to add */);
							
							if (listItemDescriptor.dataHandle != null) AEDisposeDesc(&listItemDescriptor);
							if (resultDescriptor.dataHandle != null) AEDisposeDesc(&resultDescriptor);
						}
						result = AEDuplicateDesc(&tokenDescriptor, outDataDescriptor);
					}
				}
				break;
			
			default:
				result = GetPropertyAE_DataFromClassProperty(&tokenDescriptor, inGetDataOfThisType,
																outDataDescriptor, outUsedDesiredDataType);
				break;
		}
	}
	
	if (tokenDescriptor.dataHandle != null) AEDisposeToken(&tokenDescriptor);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// getData


/*!
This routine indirectly handles all get data size
Apple Events.

(3.0)
*/
static OSErr
getDataSize		(AEDesc*			inFromWhichObject,
				 DescType const		inGetDataOfThisType,
				 AEDesc*			outDataDescriptor,
				 Boolean*			outUsedDesiredDataType)
{
	AEDesc			sizeInfoDescriptor = { typeNull, null };
	OSErr			result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: get data size event");
	Console_WriteValueFourChars("desired type", inGetDataOfThisType);
	
	if (inFromWhichObject->descriptorType == typeObjectSpecifier)
	{
		// all thats known is that this is an object; resolve it into a known object type
		// using a previously-installed OSL object accessor (see "InstallAE.c" for details)
		Console_WriteLine("given an object reference");
		result = AEResolve(inFromWhichObject, kAEIDoMinimum, &sizeInfoDescriptor);
	}
	else result = AEDuplicateDesc(inFromWhichObject, &sizeInfoDescriptor);
	
	if (result == noErr)
	{
		Console_WriteLine("successfully resolved");
		
		if (sizeInfoDescriptor.descriptorType == typeAEList)
		{
			// create a list where each corresponding item is the size of the data in the other list item
			result = AECreateList(null, 0, false/* is a record */, outDataDescriptor);
			if (result == noErr)
			{
				AEDesc		listItemDescriptor = { typeNull, null },
							resultDescriptor = { typeNull, null };
				DescType	returnedType;
				long		index = 0L;
				
				
				result = AECountItems(&sizeInfoDescriptor, &index);
				for (; index > 0; index--)
				{
					result = AEGetNthDesc(&sizeInfoDescriptor, index, typeWildCard,
											&returnedType, &listItemDescriptor);
					
					// handle the request by adding the resultant data to the end of the list
					if (result == noErr) result = getDataSize(&listItemDescriptor,
															inGetDataOfThisType, &resultDescriptor,
															outUsedDesiredDataType);
					if (result == noErr) result = AEPutDesc(&sizeInfoDescriptor/* list */, index,
															&resultDescriptor/* item to add */);
					
					if (listItemDescriptor.dataHandle != null) AEDisposeDesc(&listItemDescriptor);
					if (resultDescriptor.dataHandle != null) AEDisposeDesc(&resultDescriptor);
				}
				result = AEDuplicateDesc(&sizeInfoDescriptor, outDataDescriptor);
			}
		}
		else // an internal MacTelnet token type!
		{
			result = GetPropertyAE_DataSizeFromClassProperty(&sizeInfoDescriptor, outDataDescriptor,
																outUsedDesiredDataType);
		}
	}
	
	if (sizeInfoDescriptor.dataHandle != null) AEDisposeToken(&sizeInfoDescriptor);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// getDataSize


/*!
This routine indirectly handles all set data
Apple Events.

When a script supplies data, AppleScript will
figure out what is being said (for example,
"window 1" means a window referenced by
absolute index) and invoke an object accessor
to construct a token (see "GetObjectAE.c").
The token that is constructed is INTERNAL to
MacTelnet, and is passed in "inForWhichObject".
This routine uses the token to determine the
nature of the request, and therefore how to
handle it.  Most data is attached to a high
level object as a property, so usually the
token will be a property token.

(3.0)
*/
static OSErr
setData		(AEDesc const*		inForWhichObject,
			 AEDesc*			inDataDescriptor)
{
	AEDesc		tokenDescriptor = { typeNull, null };
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: set data event");
	
	if (inForWhichObject->descriptorType == typeObjectSpecifier)
	{
		// all thats known is that this is an object; resolve it into a known object type
		// using a previously-installed OSL object accessor (see "InstallAE.c" for details)
		Console_WriteLine("given an object reference");
		result = AEResolve(inForWhichObject, kAEIDoMinimum, &tokenDescriptor);
	}
	else result = AEDuplicateDesc(inForWhichObject, &tokenDescriptor);
	
	if (result == noErr)
	{
		Console_WriteLine("successfully resolved");
		Console_WriteValueFourChars("to type", tokenDescriptor.descriptorType);
		switch (tokenDescriptor.descriptorType)
		{
			case typeAEList:
				{
					long		index = 0L;
					
					
					result = AECountItems(&tokenDescriptor, &index);
					if (result == noErr)
					{
						AEDesc		listItemDescriptor = { typeNull, null };
						DescType	returnedType;
						
						
						for (; index > 0; index--)
						{
							result = AEGetNthDesc(&tokenDescriptor, index, typeWildCard,
													&returnedType, &listItemDescriptor);
							
							// handle the request by adding the resultant data to the end of the list
							if (result == noErr) result = setData(&listItemDescriptor, inDataDescriptor);
							
							if (listItemDescriptor.dataHandle != null) AEDisposeDesc(&listItemDescriptor);
						}
					}
				}
				break;
			
			default:
				result = GetPropertyAE_DataToClassProperty(&tokenDescriptor, inDataDescriptor);
				break;
		}
	}
	
	if (tokenDescriptor.dataHandle != null) AEDisposeToken(&tokenDescriptor);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// setData