/*###############################################################

	GenesisAE.c
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <string.h>

// Mac includes
#include <AEDataModel.h>
#include <AEObjects.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <MacErrors.h>
#include <Processes.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "Console.h"
#include "GenesisAE.h"
#include "GetElementAE.h"
#include "InstallAE.h"
#include "ObjectClassesAE.h"
#include "RequiredAE.h"
#include "TelnetDescriptors.h"
#include "TelnetMacros.h"
#include "Terminology.h"



// internal methods
static OSErr	evolve			(TelnetTokenPtr		inoutClassStructurePtr,
								 DescType*			outFinalClassTypeFromTerminologyOrNull);
static OSErr	evolveWindow	(TelnetTokenPtr		inoutClassStructurePtr,
								 DescType*			outFinalClassTypeFromTerminologyOrNull);



//
// public methods
//

/*!
Creates any kind of MacTelnet scripting class.
The "Terminology.h" file, also used for MacTelnets
Apple Event Terminology Extension resource, contains
constants for all of MacTelnets class types.

IMPORTANT:	Before calling this routine, fill in the
			fields of the appropriate part of the
			"object" member of the "as" union, in
			your token data structure.  Provide any
			information you require, except for the
			"flags" and "as.object.eventClass", which
			are provided automatically by this method.

NOTE:	Classes are automatically evolved; unless
		there is no way to derive a child class data
		from the clues of its parent, you should only
		provide data for the least refined version of
		a class (for example, describe a "window",
		not a "clipboard window").  This helps to
		reduce redundant code, and makes it possible
		for the object accessors to be generic.

If any problems occur when creating the requested
class, an error is returned and the class should be
considered invalid with empty data.  Otherwise, you
must call the AEDisposeDesc() routine to destroy the
data in the descriptor when you are done with it.

(3.0)
*/
OSErr
GenesisAE_CreateTokenFromObjectData		(DescType const			inClassTypeFromTerminology,
										 TelnetTokenPtr			inoutClassStructurePtr,
										 AEDesc*				inoutNewClassAEDescPtr,
										 DescType*				outFinalClassTypeFromTerminologyOrNull)
{
	OSErr		result = errAENoSuchObject;
	
	
	if (inoutClassStructurePtr != null)
	{
		// define the data that identifies the token as being an object token
		inoutClassStructurePtr->flags |= kTelnetTokenFlagsIsObject;
		inoutClassStructurePtr->as.object.eventClass = inClassTypeFromTerminology;
		
		// the rest of the data for the primitive type should already be in
		// the structure; now examine the data, and determine if the object
		// data can be expanded to deserve a more refined class
		evolve(inoutClassStructurePtr, outFinalClassTypeFromTerminologyOrNull);
		
		// put the data (evolved or not) into a new token descriptor
		result = AECreateDesc(kTelnetHiddenClassTelnetToken, (Ptr)inoutClassStructurePtr,
								sizeof(TelnetToken), inoutNewClassAEDescPtr);
	}
	
	return result;
}// CreateTokenFromObjectData


/*!
Coerces an object into an internal token, if necessary.
Use this inside an accessor function to get access to
the internal data for an objects container.

(3.0)
*/
OSErr
GenesisAE_CreateTokenFromObjectSpecifier	(AEDesc const*		inFromWhat,
											 TelnetTokenPtr		outTokenDataPtr)
{
	OSErr		result = noErr;
	
	
	if ((inFromWhat != null) && (outTokenDataPtr != null))
	{
		AEDesc		containerDescriptor = { typeNull, null };
		
		
		// the class owning the requested property is provided explicitly (coercion probably isnt needed)
		result = AECoerceDesc(inFromWhat, kTelnetHiddenClassTelnetToken, &containerDescriptor);
		if (result == noErr)
		{
			result = AppleEventUtilities_CopyDescriptorData(&containerDescriptor, outTokenDataPtr,
															sizeof(TelnetToken), null/* actual size storage */);
			if (result != noErr)
			{
				if (result == memSCErr)
				{
					Console_WriteLine("WARNING - not enough buffer space was given, truncating...");
					result = noErr;
				}
				else Console_WriteLine("attempt to get container type failed");
			}
			// ??? necessary ???
			//AEDisposeDesc(&containerDescriptor);
		}
		else Console_WriteLine("tokenizing failed");
	}
	else result = memPCErr;
	
	return result;
}// CreateTokenFromObjectSpecifier


/*!
Determines whether the first class type could be an
instance of the second class.  For example, if the
minimum acceptable class type is a "window" (which is
"kTelnetClassWindow"), and the unknown type turns out
to be a "kTelnetClassTerminalWindow", this routine
returns "true" because a terminal window could be
turned into a window.

This function defines the inheritance trees that
MacTelnet supports for its Apple Event objects.  Be
sure to keep "ObjectClassesAE.h" in sync.

IMPORTANT:	It would be nice if the user terminology
			agreed with this function!  Be careful
			about making changes to inheritance code,
			and always ensure the scripting interface
			reflects changes or additions.

(3.0)
*/
Boolean
GenesisAE_FirstClassIs		(DescType const		inUnknownClassTypeFromTerminology,
							 DescType const		inMinimumClassTypeFromTerminology)
{
	Boolean		result = false;
	
	
	result = (inUnknownClassTypeFromTerminology == inMinimumClassTypeFromTerminology);
	unless (result) switch (inMinimumClassTypeFromTerminology)
	{
		case kTelnetClassWindow:
			// any kind of window class could be downgraded to a simple window
			result = ((inUnknownClassTypeFromTerminology == kTelnetClassClipboardWindow) ||
						(inUnknownClassTypeFromTerminology == kTelnetClassTerminalWindow));
			break;
		
		default:
			// cant be converted
			break;
	}
	return result;
}// FirstClassIs


//
// internal methods
//

/*!
Object accessors create primitive types; this routine
can be used to evolve an object token.  This is done
by looking at the data for an object and determining,
first, if its class has any subclass types, and second,
if the data in the superclass is sufficient to create
a refined version of the token.

For example, it is possible to use data from a simple
"window" class to determine the "terminal window" data
for it, provided the window being referenced really is
a terminal window.

If a token can be evolved, it is evolved further, more
data is added to the token and its type is changed.
Otherwise, the token is not modified.

(3.0)
*/
static OSErr
evolve		(TelnetTokenPtr		inoutClassStructurePtr,
			 DescType*			outFinalClassTypeFromTerminologyOrNull)
{
	OSErr		result = noErr;
	DescType	descType = inoutClassStructurePtr->as.object.eventClass;
	
	
	// is the specified type of token even possible to evolve, first of all?
	switch (descType)
	{
		case kTelnetClassWindow:
			result = evolveWindow(inoutClassStructurePtr, outFinalClassTypeFromTerminologyOrNull);
			break;
		
		default:
			// the specified token cant be evolved any further
			break;
	}
	if (outFinalClassTypeFromTerminologyOrNull != null) *outFinalClassTypeFromTerminologyOrNull = descType;
	return result;
}// evolve


/*!
Object accessors create primitive types; this routine
can be used to evolve an object token.  This is done
by looking at the data for an object and determining,
first, if its class has any subclass types, and second,
if the data in the superclass is sufficient to create
a refined version of the token.

For example, it is possible to use data from a simple
"window" class to determine the "terminal window" data
for it, provided the window being referenced really is
a terminal window.

If a token can be evolved, it is evolved further, more
data is added to the token and its type is changed.
Otherwise, the token is not modified.

(3.0)
*/
static OSErr
evolveWindow		(TelnetTokenPtr		inoutClassStructurePtr,
					 DescType*			outFinalClassTypeFromTerminologyOrNull)
{
	OSErr		result = noErr;
	DescType	descType = inoutClassStructurePtr->as.object.eventClass;
	WindowRef	window = inoutClassStructurePtr->as.object.data.window.ref;
	
	
	Console_BeginFunction();
	Console_WriteLine("seeing if the window can be refined");
	
	if (descType == kTelnetClassWindow)
	{
		WindowInfoRef			windowInfo = WindowInfo_GetFromWindow(window);
		WindowInfoDescriptor	windowDescriptor = kInvalidWindowInfoDescriptor;
		
		
		windowDescriptor = WindowInfo_GetWindowDescriptor(windowInfo);
		switch (windowDescriptor)
		{
			case kTelnetWindowDescriptorAnyTerminal:
				// evolve to "terminal window"
				if (ScreenManager_IsWindowAScreen(window))
				{
					Console_WriteLine("promoting to a terminal window");
					inoutClassStructurePtr->as.object.data.terminalWindow.realScreenPtr =
						ScreenManager_GetWindowRealScreenData(window);
					inoutClassStructurePtr->as.object.eventClass = kTelnetClassTerminalWindow;
				}
				break;
			
			case kTelnetWindowDescriptorClipboard:
				// evolve to "clipboard window"
				Console_WriteLine("promoting to a clipboard window");
				inoutClassStructurePtr->as.object.eventClass = kTelnetClassClipboardWindow;
				break;
			
			default:
				// not evolveable
				break;
		}
	}
	if (outFinalClassTypeFromTerminologyOrNull != null) *outFinalClassTypeFromTerminologyOrNull = descType;
	
	Console_EndFunction();
	
	return result;
}// evolveWindow