/*###############################################################

	GetElementAE.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <string.h>

// Mac includes
#include <AEDataModel.h>
#include <AEObjects.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <MacErrors.h>
#include <Processes.h>

// resource includes
#include "StringResources.h"

// library includes
#include "InterfaceLib.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "Console.h"
#include "GenesisAE.h"
#include "GetElementAE.h"
#include "InstallAE.h"
#include "ObjectClassesAE.h"
#include "RequiredAE.h"
#include "TelnetDescriptors.h"
#include "TelnetMacros.h"
#include "Terminology.h"



// internal methods
static OSErr	createObjectFromWindow				(DescType const		inDesiredClass,
													 WindowRef			inWindow,
													 AEDesc*			outObjectOfDesiredClass);
static OSErr	getMacroByAbsolutePosition			(AEDesc const*		inReference,
													 AEDesc const*		inFromWhat,
													 AEDesc*			outObjectOfDesiredClass);
static OSErr	getMacroSetByAbsolutePosition		(AEDesc const*		inReference,
													 AEDesc*			outObjectOfDesiredClass);
static OSErr	getWindowByAbsolutePosition			(DescType const		inDesiredClass,
													 AEDesc const*		inReference,
													 AEDesc*			outObjectOfDesiredClass);
static OSErr	getWindowByName						(DescType const		inDesiredClass,
													 AEDesc const*		inReference,
													 AEDesc*			outObjectOfDesiredClass);
static OSErr	getWindowByRelativePosition			(DescType const		inDesiredClass,
													 AEDesc const*		inRelativeToWhat,
													 AEDesc const*		inReference,
													 AEDesc*			outObjectOfDesiredClass);
static SInt16	getWindowCount						(void);
static Boolean	isClassObtainableFromWindow			(DescType const		inDesiredClass);



//
// public methods
//

/*!
Handles the following OSL object access:

	cApplication <- typeNull

(3.0)
*/
pascal OSErr
GetElementAE_ApplicationFromNull	(DescType			inDesiredClass,
									 AEDesc const*		inFromWhat,
									 DescType			inFromWhatClass,
									 DescType			inFormOfReference,
									 AEDesc const*		inReference,
									 AEDesc*			outObjectOfDesiredClass,
									 long				inData)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: requested application");
	
	if (inDesiredClass != cApplication)
	{
		Console_WriteLine("failed, desired type is not an application");
		result = errAEWrongDataType;
	}
	
	// create the appropriate object
	if (result == noErr) switch (inFormOfReference)
	{
		default:
			result = GetElementAE_ApplicationImplicit(outObjectOfDesiredClass);
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// ApplicationFromNull


/*!
To obtain the application class for MacTelnet
itself, use this method.

(3.0)
*/
OSErr
GetElementAE_ApplicationImplicit		(AEDesc*		outObjectOfDesiredClass)
{
	OSErr			result = noErr;
	TelnetToken		token;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: application implicitly");
	token.flags = 0;
	token.genericPointer = null;
	token.as.object.data.application.process.highLongOfPSN = 0;
	token.as.object.data.application.process.lowLongOfPSN = kCurrentProcess;
	result = GenesisAE_CreateTokenFromObjectData(kTelnetClassApplication, &token, outObjectOfDesiredClass,
													null/* new type */);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// ApplicationImplicit


/*!
Handles the following OSL object access:

	kTelnetClassMacro <- kTelnetClassMacroSet

References may only be by absolute position.

(3.0)
*/
pascal OSErr
GetElementAE_MacroFromMacroSet	(DescType			inDesiredClass,
								 AEDesc const*		inFromWhat,
								 DescType			inFromWhatClass,
								 DescType			inFormOfReference,
								 AEDesc const*		inReference,
								 AEDesc*			outObjectOfDesiredClass,
								 long				inData)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: requested macro");
	if (inDesiredClass != kTelnetClassMacro)
	{
		result = errAEWrongDataType;
	}
	else if (inFromWhatClass != kTelnetClassMacroSet)
	{
		result = errAENoSuchObject;
	}
	
	// create the appropriate object
	if (result == noErr) switch (inFormOfReference)
	{
		case formAbsolutePosition:
			result = getMacroByAbsolutePosition(inReference, inFromWhat, outObjectOfDesiredClass);
			break;
		
		case formTest:
			// handled automatically by installed count and compare object callbacks
			break;
		
		default:
			result = errAEBadTestKey;
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// MacroFromMacroSet


/*!
Handles the following OSL object access:

	kTelnetClassMacroSet <- typeNull

References may be by name, or absolute
position.

(3.0)
*/
pascal OSErr
GetElementAE_MacroSetFromNull	(DescType			inDesiredClass,
								 AEDesc const*		inFromWhat,
								 DescType			inFromWhatClass,
								 DescType			inFormOfReference,
								 AEDesc const*		inReference,
								 AEDesc*			outObjectOfDesiredClass,
								 long				inData)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: requested macro set");
	if (inDesiredClass != kTelnetClassMacroSet)
	{
		result = errAEWrongDataType;
	}
	else if ((inFromWhatClass != typeNull) && (inFromWhatClass != kTelnetClassApplication))
	{
		result = errAENoSuchObject;
	}
	
	// create the appropriate object
	if (result == noErr) switch (inFormOfReference)
	{
		case formAbsolutePosition:
			result = getMacroSetByAbsolutePosition(inReference, outObjectOfDesiredClass);
			break;
		
		case formName:
			//result = getMacroSetByName(inReference, outObjectOfDesiredClass);
			result = errAEBadTestKey;
			break;
		
		case formTest:
			// handled automatically by installed count and compare object callbacks
			break;
		
		default:
			result = errAEBadTestKey;
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// MacroSetFromNull


/*!
Handles the following OSL object accesses:

	cWindow <- typeNull
	kTelnetClassClipboardWindow <- typeNull
	kTelnetClassConnection <- typeNull
	kTelnetClassSession <- typeNull
	kTelnetClassTerminalWindow <- typeNull

References may be by name, or absolute or
relative position.

(3.0)
*/
pascal OSErr
GetElementAE_WindowFromNull		(DescType			inDesiredClass,
								 AEDesc const*		inFromWhat,
								 DescType			inFromWhatClass,
								 DescType			inFormOfReference,
								 AEDesc const*		inReference,
								 AEDesc*			outObjectOfDesiredClass,
								 long				inData)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteValueFourChars("AppleScript: requested window-related class", inDesiredClass);
	if (!isClassObtainableFromWindow(inDesiredClass))
	{
		result = errAEWrongDataType;
	}
	else if ((inFromWhatClass != typeNull) && (inFromWhatClass != kTelnetClassApplication))
	{
		result = errAENoSuchObject;
	}
	
	// create the appropriate object
	if (result == noErr) switch (inFormOfReference)
	{
		case formAbsolutePosition:
			result = getWindowByAbsolutePosition(inDesiredClass, inReference, outObjectOfDesiredClass);
			break;
		
		case formRelativePosition:
			result = getWindowByRelativePosition(inDesiredClass, inFromWhat, inReference,
													outObjectOfDesiredClass);
			break;
		
		case formName:
			result = getWindowByName(inDesiredClass, inReference, outObjectOfDesiredClass);
			break;
		
		case formTest:
			// handled automatically by installed count and compare object callbacks
			break;
		
		default:
			result = errAEBadTestKey;
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// WindowFromNull


//
// internal methods
//

/*!
Creates a token for the specified type of object,
using window data.  This will only work if you
provide a desired class that can be derived from
a window (such as "kTelnetClassWindow", or
"kTelnetClassConnection").  Since not all windows
represent all objects, obviously you must also
provide a particular window that can indeed be
used to create the desired class - for example,
passing the Preferences window while requesting
a terminal window class wont work!

(3.0)
*/
static OSErr
createObjectFromWindow		(DescType const		inDesiredClass,
							 WindowRef			inWindow,
							 AEDesc*			outObjectOfDesiredClass)
{
	OSErr		result = noErr;
	
	
	if (isClassObtainableFromWindow(inDesiredClass))
	{
		WindowInfoRef			windowInfo = WindowInfo_GetFromWindow(inWindow);
		WindowInfoDescriptor	windowDescriptor = kInvalidWindowInfoDescriptor;
		TelnetToken				token;
		
		
		windowDescriptor = WindowInfo_GetWindowDescriptor(windowInfo);
		token.flags = 0;
		token.genericPointer = null;
		switch (inDesiredClass)
		{
			case kTelnetClassClipboardWindow:
				if (windowDescriptor == kTelnetWindowDescriptorClipboard)
				{
					Console_WriteLine("apparent request: the clipboard window");
					token.as.object.data.clipboardWindow.windowClass.ref = inWindow;
				}
				else result = errAENoSuchObject;
				break;
			
			case kTelnetClassConnection:
				if (ScreenManager_IsWindowAScreen(inWindow))
				{
					Console_WriteLine("apparent request: a connection");
					token.as.object.data.connection.connectionDataPtr =
						ScreenManager_GetWindowConnectionData(inWindow);
				}
				else result = errAENoSuchObject;
				break;
			
			case kTelnetClassSession:
				if (ScreenManager_IsWindowAScreen(inWindow))
				{
					Console_WriteLine("apparent request: a session");
					token.as.object.data.session.connectionPtr =
						ScreenManager_GetWindowConnectionData(inWindow);
					token.as.object.data.session.realScreenPtr =
						ScreenManager_GetWindowRealScreenData(inWindow);
				}
				else result = errAENoSuchObject;
				break;
			
			case kTelnetClassTerminalWindow:
				if (ScreenManager_IsWindowAScreen(inWindow))
				{
					Console_WriteLine("apparent request: a terminal window");
					token.as.object.data.terminalWindow.windowClass.ref = inWindow;
					token.as.object.data.terminalWindow.realScreenPtr =
						ScreenManager_GetWindowRealScreenData(inWindow);
				}
				else result = errAENoSuchObject;
				break;
			
			case kTelnetClassWindow:
				Console_WriteLine("apparent request: a window");
				token.as.object.data.window.ref = inWindow;
				break;
			
			default:
				result = errAEWrongDataType;
				break;
		}
		if (result == noErr)
		{
			Console_WriteLine("constructing token");
			result = GenesisAE_CreateTokenFromObjectData(inDesiredClass, &token,
															outObjectOfDesiredClass, null/* new type */);
		}
	}
	else result = errAEWrongDataType;
	
	return result;
}// createObjectFromWindow


/*!
To obtain a macro class using "formAbsolutePosition"
selection data as a reference (that is, you know the
index of a macro in a macro set), use this method.
The index can be an integer (positive to go forward,
negative to go in reverse), or one of the following
constants:

	kAEFirst		the first macro
	kAELast			the last macro
	kAEMiddle		the middle macro
	kAEAny			the first macro
	kAEAll			a list of macros

MacTelnet currently identifies macros in the same
order as they appear in the macro editor.

The container type is expected to be a macro set.

(3.0)
*/
static OSErr
getMacroByAbsolutePosition	(AEDesc const*		inReference,
							 AEDesc const*		inFromWhat,
							 AEDesc*			outObjectOfDesiredClass)
{
	OSErr			result = noErr;
	TelnetToken		macroSetToken;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: macro by absolute position");
	
	result = GenesisAE_CreateTokenFromObjectSpecifier(inFromWhat, &macroSetToken);
	
	if (result == noErr)
	{
		Boolean			handled = false,	// recognized request?
						isNotAList = true;	// is the resultant object a single set, not a list of sets?
		TelnetToken		token;
		SInt16			macroSetNumber = 0;	// used for convenience
		
		
		macroSetNumber = macroSetToken.as.object.data.macroSet.number;
		Console_WriteValue("requested macro from macro set number", macroSetNumber);
#if 0
		token.flags = 0;
		token.genericPointer = null;
		if (inReference->descriptorType == typeSInt32)
		{
			SInt32		index = 0L;
			
			
			Console_WriteLine("using a number");
			result = AppleEventUtilities_CopyDescriptorDataAs(typeSInt32, inReference, &index, sizeof(SInt32),
																null/* actual size storage */);
			if (result == noErr)
			{
				MacroSet		macroSet = null;
				
				
				handled = true;
				
				Console_WriteValue("index", index);
				
				// now decide what macro set is being specified; negative indices mean search in reverse
				if (index > 0)
				{
					if (index > MACRO_SET_COUNT)
					{
						Console_WriteLine("WARNING: unexpectedly large index, bailing out...");
						result = errAEIllegalIndex;
					}
					else
					{
						UInt16		oldSetNumber = Macros_GetActiveSetNumber();
						
						
						Macros_SetActiveSetNumber(index);
						token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
						Macros_SetActiveSetNumber(oldSetNumber);
					}
				}
				else
				{
					Console_WriteLine("(negative value - unimplemented)");
					result = errAEIllegalIndex;
				}
			}
			else Console_WriteLine("could not obtain that number");
		}
		else if (inReference->descriptorType == typeAbsoluteOrdinal)
		{
			DescType	ordinal = kAEFirst;
			
			
			Console_WriteLine("using an absolute ordinal");
			result = AppleEventUtilities_CopyDescriptorData(inReference, &ordinal, sizeof(DescType),
															null/* actual size storage */);
			if (result == noErr)
			{
				handled = true;
				
				Console_WriteValueFourChars("ordinal", ordinal);
				
				switch (ordinal)
				{
					case kAEFirst:
					case kAEAny:
						Console_WriteLine("(first or any macro in a set)");
						{
							UInt16		oldSetNumber = Macros_GetActiveSetNumber();
							
							
							Macros_SetActiveSetNumber(1);
							token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
							Macros_SetActiveSetNumber(oldSetNumber);
						}
						break;
					
					case kAELast:
						Console_WriteLine("(last macro in a set)");
						{
							UInt16		oldSetNumber = Macros_GetActiveSetNumber();
							
							
							Macros_SetActiveSetNumber(MACRO_SET_COUNT);
							token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
							Macros_SetActiveSetNumber(oldSetNumber);
						}
						break;
					
					case kAEMiddle:
						Console_WriteLine("(middle macro in a set)");
						{
							UInt16		oldSetNumber = Macros_GetActiveSetNumber();
							
							
							Macros_SetActiveSetNumber(1 + INTEGER_HALVED(MACRO_SET_COUNT));
							token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
							Macros_SetActiveSetNumber(oldSetNumber);
						}
						break;
					
					case kAEAll:
						Console_WriteLine("(all macros in a set)");
						isNotAList = false;
						{
							AEDesc		macroSetDesc = { typeNull, null };
							
							
							Console_WriteLine("creating a list");
							result = AECreateList(null/* factoring */, 0/* size */,
													false/* is a record */, outObjectOfDesiredClass);
							if (result == noErr)
							{
								UInt16				oldSetNumber = Macros_GetActiveSetNumber();
								register SInt16		i = 0;
								
								
								for (i = 1; i <= MACRO_SET_COUNT; i++)
								{
									// set up the token
									Macros_SetActiveSetNumber(i);
									token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
									
									// add each macro set to the list
									Console_WriteLine("new macro set descriptor");
									result = GenesisAE_CreateTokenFromObjectData(kTelnetClassMacroSet,
																					&token, &macroSetDesc,
																					null/* new type */);
									Console_WriteValue("adding to list if zero:", result);
									if (result == noErr)
									{
										result = AEPutDesc(outObjectOfDesiredClass, 0/* index */,
															&macroSetDesc);
										if (result == noErr) result = AEDisposeDesc(&macroSetDesc);
									}
								}
								
								Macros_SetActiveSetNumber(oldSetNumber);
							}
						}
						break;
					
					default:
						// ???
						result = errAENoSuchObject;
						break;
				}
			}
			else Console_WriteLine("could not obtain that ordinal");
		}
		else
		{
			Console_WriteValueFourChars("unknown form of reference", inReference->descriptorType);
		}
		
		unless (handled) result = errAENoSuchObject; // no other types of reference are supported
		
		// if the object was resolved to a macro set, create the token
		if ((handled) && (result == noErr) && (isNotAList))
		{
			Console_WriteLine("constructing macro token");
			result = GenesisAE_CreateTokenFromObjectData(kTelnetClassMacro, &token, outObjectOfDesiredClass,
															null/* new type */);
		}
#endif
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// getMacroByAbsolutePosition


/*!
To obtain a macro set class using "formAbsolutePosition"
selection data as a reference (that is, you know the
index of a macro set in the macro set list), use this
method.  The index can be an integer (positive to go
forward, negative to go in reverse), or one of the
following constants:

	kAEFirst		the first macro set
	kAELast			the last macro set
	kAEMiddle		the middle macro set
	kAEAny			the first macro set
	kAEAll			a list of macro sets

MacTelnet currently identifies macro sets in the same
order as they appear in the macro editor.

(3.0)
*/
static OSErr
getMacroSetByAbsolutePosition	(AEDesc const*		inReference,
								 AEDesc*			outObjectOfDesiredClass)
{
	OSErr			result = noErr;
	Boolean			handled = false,	// recognized request?
					isNotAList = true;	// is the resultant object a single set, not a list of sets?
	TelnetToken		token;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: macro set by absolute position");
	
	token.flags = 0;
	token.genericPointer = null;
	if (inReference->descriptorType == typeSInt32)
	{
		SInt32		index = 0L;
		
		
		Console_WriteLine("using a number");
		result = AppleEventUtilities_CopyDescriptorDataAs(typeSInt32, inReference, &index, sizeof(SInt32),
															null/* actual size storage */);
		if (result == noErr)
		{
			MacroSet			macroSet = null;
			
			
			handled = true;
			
			Console_WriteValue("index", index);
			
			// now decide what macro set is being specified; negative indices mean search in reverse
			if (index > 0)
			{
				if (index > MACRO_SET_COUNT)
				{
					Console_WriteLine("unexpectedly large index, bailing out...");
					result = errAEIllegalIndex;
				}
				else
				{
					UInt16		oldSetNumber = Macros_GetActiveSetNumber();
					
					
					Macros_SetActiveSetNumber(index);
					token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
					Macros_SetActiveSetNumber(oldSetNumber);
				}
			}
			else
			{
				Console_WriteLine("(negative value - unimplemented)");
				result = errAEIllegalIndex;
			}
		}
		else Console_WriteLine("could not obtain that number");
	}
	else if (inReference->descriptorType == typeAbsoluteOrdinal)
	{
		DescType	ordinal = kAEFirst;
		
		
		Console_WriteLine("using an absolute ordinal");
		result = AppleEventUtilities_CopyDescriptorData(inReference, &ordinal, sizeof(DescType),
														null/* actual size storage */);
		if (result == noErr)
		{
			handled = true;
			
			Console_WriteValueFourChars("ordinal", ordinal);
			
			switch (ordinal)
			{
				case kAEFirst:
				case kAEAny:
					Console_WriteLine("(first or any macro set)");
					{
						UInt16		oldSetNumber = Macros_GetActiveSetNumber();
						
						
						Macros_SetActiveSetNumber(1);
						token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
						Macros_SetActiveSetNumber(oldSetNumber);
					}
					break;
				
				case kAELast:
					Console_WriteLine("(last macro set)");
					{
						UInt16		oldSetNumber = Macros_GetActiveSetNumber();
						
						
						Macros_SetActiveSetNumber(MACRO_SET_COUNT);
						token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
						Macros_SetActiveSetNumber(oldSetNumber);
					}
					break;
				
				case kAEMiddle:
					Console_WriteLine("(middle macro set)");
					{
						UInt16		oldSetNumber = Macros_GetActiveSetNumber();
						
						
						Macros_SetActiveSetNumber(1 + INTEGER_HALVED(MACRO_SET_COUNT));
						token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
						Macros_SetActiveSetNumber(oldSetNumber);
					}
					break;
				
				case kAEAll:
					Console_WriteLine("(all macro sets)");
					isNotAList = false;
					{
						AEDesc		macroSetDesc = { typeNull, null };
						
						
						Console_WriteLine("creating a list");
						result = AECreateList(null/* factoring */, 0/* size */,
												false/* is a record */, outObjectOfDesiredClass);
						if (result == noErr)
						{
							UInt16				oldSetNumber = Macros_GetActiveSetNumber();
							register SInt16		i = 0;
							
							
							for (i = 1; i <= MACRO_SET_COUNT; i++)
							{
								// set up the token
								Macros_SetActiveSetNumber(i);
								token.as.object.data.macroSet.number = Macros_GetActiveSetNumber();
								
								// add each macro set to the list
								Console_WriteLine("new macro set descriptor");
								result = GenesisAE_CreateTokenFromObjectData(kTelnetClassMacroSet,
																				&token, &macroSetDesc,
																				null/* new type */);
								Console_WriteValue("adding to list if zero:", result);
								if (result == noErr)
								{
									result = AEPutDesc(outObjectOfDesiredClass, 0/* index */,
														&macroSetDesc);
									if (result == noErr) result = AEDisposeDesc(&macroSetDesc);
								}
							}
							
							Macros_SetActiveSetNumber(oldSetNumber);
						}
					}
					break;
				
				default:
					// ???
					result = errAENoSuchObject;
					break;
			}
		}
		else Console_WriteLine("could not obtain that ordinal");
	}
	else
	{
		Console_WriteValueFourChars("unknown form of reference", inReference->descriptorType);
	}
	
	unless (handled) result = errAENoSuchObject; // no other types of reference are supported
	
	// if the object was resolved to a macro set, create the token
	if ((handled) && (result == noErr) && (isNotAList))
	{
		Console_WriteLine("constructing macro set token");
		result = GenesisAE_CreateTokenFromObjectData(kTelnetClassMacroSet, &token, outObjectOfDesiredClass,
														null/* new type */);
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// getMacroSetByAbsolutePosition


/*!
Creates a window object (or the specified class) using
"formAbsolutePosition" selection data as a reference.
If the specified class cannot be created from a window,
an error is returned.

The index can be an integer (positive to go forward,
negative to go in reverse), or one of the following
constants:

	kAEFirst		the first window
	kAELast			the last window
	kAEMiddle		the middle window
	kAEAny			the frontmost window
	kAEAll			a list of windows

MacTelnet currently identifies the frontmost window as
"window 1", and other windows in the window list order
relative to the frontmost window.

If a window is identified as having a more particular
type, it is automatically returned as the most specific
possible type (e.g. a "terminal window").

(3.0)
*/
static OSErr
getWindowByAbsolutePosition		(DescType const		inDesiredClass,
								 AEDesc const*		inReference,
								 AEDesc*			outObjectOfDesiredClass)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: window by absolute position");
	if (FrontWindow() != null)
	{
		Boolean			handled = false,	// recognized request?
						isNotAList = true;	// is the resultant object a single window, not a list of windows?
		WindowRef		finalWindow = null;	// window to use for token
		
		
		if (inReference->descriptorType == typeSInt32)
		{
			SInt32		index = 0L;
			
			
			Console_WriteLine("using a number");
			result = AppleEventUtilities_CopyDescriptorDataAs(typeSInt32, inReference, &index, sizeof(SInt32),
																null/* actual size storage */);
			if (result == noErr)
			{
				WindowPtr			window = FrontWindow();
				register SInt16		windowCount = 0,
									i = 0;
				
				
				handled = true;
				
				Console_WriteValue("index", index);
				
				// first, count the windows (so either positive or negative offsets can be easily worked)
				windowCount = getWindowCount();
				
				// now decide what window is being specified; negative indices mean search in reverse
				if (index > 0)
				{
					if (index > windowCount)
					{
						Console_WriteLine("WARNING: unexpectedly large index, bailing out...");
						result = errAEIllegalIndex;
					}
					else
					{
						for (i = 1; i < index; i++) window = GetNextWindow(window);
						finalWindow = window;
					}
				}
				else
				{
					Console_WriteLine("(negative value - unimplemented)");
					result = errAEIllegalIndex;
				}
			}
			else Console_WriteLine("could not obtain that number");
		}
		else if (inReference->descriptorType == typeAbsoluteOrdinal)
		{
			DescType	ordinal = kAEFirst;
			
			
			Console_WriteLine("using an absolute ordinal");
			result = AppleEventUtilities_CopyDescriptorData(inReference, &ordinal, sizeof(DescType),
															null/* actual size storage */);
			if (result == noErr)
			{
				handled = true;
				
				Console_WriteValueFourChars("ordinal", ordinal);
				
				switch (ordinal)
				{
					case kAEFirst:
					case kAEAny:
						Console_WriteLine("(first or any window)");
						finalWindow = FrontWindow();
						break;
					
					case kAELast:
						// return the last window in the list
						Console_WriteLine("(last window)");
						{
							WindowPtr			window = FrontWindow();
							register SInt16		i = 0;
							
							
							// inefficient, but good enough for now (it works)
							for (i = 1; i < getWindowCount(); i++) window = GetNextWindow(window);
							finalWindow = window;
						}
						break;
					
					case kAEMiddle:
						Console_WriteLine("(middle window)");
						{
							WindowPtr			window = FrontWindow();
							register SInt16		i = 0;
							
							
							// inefficient, but good enough for now (it works)
							for (i = 1; i < INTEGER_HALVED(getWindowCount()); i++) window = GetNextWindow(window);
							finalWindow = window;
						}
						break;
					
					case kAEAll:
						Console_WriteLine("(all windows)");
						isNotAList = false;
						{
							AEDesc			windowDesc = { typeNull, null };
							
							
							Console_WriteLine("creating a list");
							result = AECreateList(null/* factoring */, 0/* size */,
													false/* is a record */, outObjectOfDesiredClass);
							if (result == noErr)
							{
								WindowRef		firstWindow = null,
												window = null;
								
								
								Console_WriteLine("getting front window");
								window = firstWindow = FrontWindow();
								if (firstWindow == null) result = errAENoSuchObject;
								else do
								{
									// add each window to the list
									Console_WriteLine("new object in list");
									result = createObjectFromWindow(inDesiredClass, window, &windowDesc);
									Console_WriteValue("adding to list if zero:", result);
									if (result == noErr)
									{
										result = AEPutDesc(outObjectOfDesiredClass, 0/* index */, &windowDesc);
										if (result == noErr) result = AEDisposeDesc(&windowDesc);
									}
									Console_WriteLine("getting next window");
									window = GetNextWindow(window);
									if (window == null)
									{
										// return to beginning of list if the end is reached
										// ??? is this necessary ???
										window = firstWindow;
									}
								} while (window != firstWindow);
							}
						}
						break;
					
					default:
						// ???
						result = errAENoSuchObject;
						break;
				}
			}
			else Console_WriteLine("could not obtain that ordinal");
		}
		else
		{
			Console_WriteValueFourChars("unknown form of reference", inReference->descriptorType);
		}
		
		unless (handled) result = errAENoSuchObject; // no other types of reference are supported
		
		// if the object was resolved to a window, create the token
		if ((handled) && (result == noErr) && (isNotAList))
		{
			result = createObjectFromWindow(inDesiredClass, finalWindow, outObjectOfDesiredClass);
		}
	}
	else result = errAENoSuchObject;
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// getWindowByAbsolutePosition


/*!
Creates a window object (or the specified class) using
"formName" selection data as a reference.  If the
specified class cannot be created from a window, an
error is returned.

(3.0)
*/
static OSErr
getWindowByName			(DescType const		inDesiredClass,
						 AEDesc const*		inReference,
						 AEDesc*			outObjectOfDesiredClass)
{
	OSErr		result = noErr;
	Str255		name;
	Size		actualSize = 0L;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: window by name");
	result = AppleEventUtilities_CopyDescriptorDataAs(cStringClass, inReference, name + 1,
														255 * sizeof(unsigned char),
														&actualSize);
	
	if (result == noErr)
	{
		name[0] = actualSize / sizeof(unsigned char);
		
		Console_WriteValuePString("looking for window with name", name);
		if (FrontWindow() != null)
		{
			WindowPtr			window = FrontWindow();
			Str255				comparedName;
			SInt16				windowCount = getWindowCount();
			register SInt16		i = 0;
			
			
			for (i = 1; i <= getWindowCount(); i++)
			{
				GetWTitle(window, comparedName);
				if (EqualString(comparedName, name, false/* case sensitive */, true/* diacritics sensitive */))
				{
					result = createObjectFromWindow(inDesiredClass, window, outObjectOfDesiredClass);
					break;
				}
				window = GetNextWindow(window);
			}
		}
		else result = errAENoSuchObject;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// getWindowByName


/*!
Creates a window object (or the specified class) using
"formRelativePosition" selection data as a reference.
If the specified class cannot be created from a window,
an error is returned.

(3.0)
*/
static OSErr
getWindowByRelativePosition		(DescType const		inDesiredClass,
								 AEDesc const*		inRelativeToWhat,
								 AEDesc const*		inReference,
								 AEDesc*			outObjectOfDesiredClass)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: window by relative position");
	if (FrontWindow() != null)
	{
		DescType	relation = 0;	// kAENext or kAEPrevious
		WindowPtr	window = null;
		
		
		result = AppleEventUtilities_CopyDescriptorData(inRelativeToWhat, &window, sizeof(WindowPtr),
														null/* actual size storage */);
		if (result == noErr) result = AppleEventUtilities_CopyDescriptorData(inReference, &relation,
																				sizeof(DescType),
																				null/* actual size storage */);
		if ((inReference->descriptorType == typeEnumerated) &&
			(result == noErr)) // was a next or previous relation given?
		{
			WindowPtr		window = null;
			
			
			// set up the object based on the desired relationship
			switch (relation)
			{
				case kAENext:
					Console_WriteLine("next window");
					window = GetNextWindow(window);
					break;
				
				case kAEPrevious:
					Console_WriteLine("previous window (unimplemented)");
					result = errAENoSuchObject;
					break;
				
				default:
					// ???
					result = errAENoSuchObject;
					break;
			}
			
			if (result == noErr)
			{
				result = createObjectFromWindow(inDesiredClass, window, outObjectOfDesiredClass);
			}
		}
		else result = errAENoSuchObject;
	}
	else result = errAENoSuchObject;
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// getWindowByRelativePosition


/*!
To determine how many windows there are
in the application window list, use this
method.

(3.0)
*/
static SInt16
getWindowCount	(void)
{
	WindowPtr		window = null;
	SInt16     		result = 0;
	
	
	window = FrontWindow();
	while (window != null)
	{
		result++;
		window = GetNextWindow(window);
	}
	return result;
}// getWindowCount


/*!
Returns "true" only if the specified class can
be completely defined using window data.

(3.0)
*/
static Boolean
isClassObtainableFromWindow		(DescType const		inDesiredClass)
{
	Boolean		result = false;
	
	
	switch (inDesiredClass)
	{
		case kTelnetClassClipboardWindow:
		case kTelnetClassConnection:
		case kTelnetClassSession:
		case kTelnetClassTerminalWindow:
		case kTelnetClassWindow:
			result = true;
			break;
		
		default:
			break;
	}
	return result;
}// isClassObtainableFromWindow