/*###############################################################

	GetPropertyAE.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <string.h>

// Mac includes
#include <AEDataModel.h>
#include <AEObjects.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <Errors.h>
#include <Processes.h>
#include <Resources.h>
#include <Scrap.h>

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "BasicTypesAE.h"
#include "Clipboard.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "FolderManager.h"
#include "GenesisAE.h"
#include "GetElementAE.h"
#include "GetPropertyAE.h"
#include "InstallAE.h"
#include "MacLook.h"
#include "ObjectClassesAE.h"
#include "Preferences.h"
#include "RealScreen.h"
#include "RequiredAE.h"
#include "ResourceFileManager.h"
#include "TelnetDescriptors.h"
#include "TelnetMacros.h"
#include "Terminal.h"
#include "Terminology.h"



// constants
typedef SInt16 PropertyAccessMechanism;
enum
{
	kPropertyAccessModify = -1,	// change the data for the property
	kPropertyAccessObtain = 0,	// acquire the data from the property
	kPropertyAccessSize = 1		// determine the size of the data from the property
};

// internal methods
static OSErr	accessPropertyOfClassApplication			(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassApplicationPreferences	(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassClipboardWindow		(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassConnection				(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassDialogReply			(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassMacro					(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassMacroSet				(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassProxyServer			(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassSession				(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassTerminalWindow			(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassWindow					(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	accessPropertyOfClassWord					(TelnetTokenPtr				inTokenPtr,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inPropertyAccessMechanism);
static OSErr	propertyAccessError							(DescType					inDesiredClass,
															 DescType					inFormOfReference);
static OSErr	tokenAccess									(AEDesc const*				inTokenDescriptor,
															 AEDesc*					inoutDataDescriptor,
															 PropertyAccessMechanism	inAccessMechanism);



//
// public methods
//

/*!
To acquire an Apple Event descriptor containing
data corresponding to the specified property of
the specified class, use this method.  The token
encodes all necessary information about the
desired property and the class having the
property.  You generally use this routine to
help handle a get data Apple Event.

The value of "inDesiredDataType" is used to
coerce the resultant data into a form that the
user wants.  For example, it might be possible
to convert the actual property into some other
kind of object - the other type would be
"inDesiredDataType".

(3.0)
*/
OSErr
GetPropertyAE_DataFromClassProperty		(AEDesc const*		inTokenDescriptor,
										 DescType const		inDesiredDataType,
										 AEDesc*			outDataDescriptor,
										 Boolean*			outUsedDesiredDataType)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("Obtaining data...");
	
	result = tokenAccess(inTokenDescriptor, outDataDescriptor, kPropertyAccessObtain);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	
	return result;
}// DataFromClassProperty


/*!
To acquire an Apple Event descriptor containing
the size in bytes of data corresponding to the
specified property of the specified class, use
this method.  The token encodes all necessary
information about the desired property and the
class having the property.  You generally use
this routine to help handle a get data size
Apple Event.

(3.0)
*/
OSErr
GetPropertyAE_DataSizeFromClassProperty		(AEDesc const*		inTokenDescriptor,
											 AEDesc*			outDataDescriptor,
											 Boolean*			outUsedDesiredDataType)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("Obtaining data size...");
	
	result = tokenAccess(inTokenDescriptor, outDataDescriptor, kPropertyAccessSize);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	
	return result;
}// DataSizeFromClassProperty


/*!
To use an Apple Event descriptor containing
data to modify the specified property of the
specified class, use this method.  The token
encodes all necessary information about the
desired property and the class having the
property.  You generally use this routine to
help handle a set data Apple Event.

(3.0)
*/
OSErr
GetPropertyAE_DataToClassProperty		(AEDesc const*		inTokenDescriptor,
										 AEDesc*			inDataDescriptor)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("Modifying data...");
	
	result = tokenAccess(inTokenDescriptor, inDataDescriptor, kPropertyAccessModify);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	
	return result;
}// DataToClassProperty


/*!
Handles the following OSL property access:

	typeProperty <- ANYTHING

Thanks to the powerful architecture of the data
structures defined by MacTelnets property
management system ("GetPropertyAE.h"), ONE
function can return ANY property from ANY
MacTelnet Apple Event class!!!  This is in
sharp contrast to the approach taken by
virtually all applications, which is to use
separate functions containing mostly the
same code to handle properties for each.  I
think that my approach is a LOT easier to
maintain and debug.

(3.0)
*/
pascal OSErr
GetPropertyAE_FromObject	(DescType			inDesiredClass,
							 AEDesc const*		inFromWhat,
							 DescType			inFromWhatClass,
							 DescType			inFormOfReference,
							 AEDesc const*		inReference,
							 AEDesc*			outDesiredProperty,
							 long				inData)
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: requested property");
	Console_WriteValueFourChars("from class", inFromWhatClass);
	
	outDesiredProperty->dataHandle = null;
	result = propertyAccessError(inDesiredClass, inFormOfReference);
	if (result == noErr)
	{
		TelnetToken		token;
		TelnetObject	container;
		AEDesc			containerDescriptor = { typeNull, null },
						propertyTypeDescriptor = { typeNull, null };
		
		
		Console_WriteLine("the request seems to be in order");
		
		if (inFromWhatClass == typeProperty)
		{
			// the data is a property token; therefore, the data is obtained
			// in a different way
			AEDesc			objectDescriptor = { typeNull, null },
							propertyDescriptor = { typeNull, null },
							propertyTokenDescriptor = { typeNull, null };
			TelnetToken		propertyToken;
			
			
			Console_WriteLine("a property of a property");
			#if 0
			result = GetPropertyAE_DataFromClassProperty(inFromWhat, typeWildCard, &objectDescriptor);
			Console_WriteValue("data from class property result", result);
			Console_WriteValueFourChars("container descriptor type", containerDescriptor.descriptorType);
			AppleEventUtilities_DuplicateOrResolveObject(&objectDescriptor, &containerDescriptor);
			Console_WriteValueFourChars("resolved descriptor type", containerDescriptor.descriptorType);
			#endif
			result = AECoerceDesc(inFromWhat, kTelnetHiddenClassTelnetToken, &propertyDescriptor);
			if (result != noErr) Console_WriteLine("coercion to token class failed");
			result = AppleEventUtilities_CopyDescriptorData(&propertyDescriptor, &propertyToken,
															sizeof(propertyToken), null/* actual size storage */);
			if (result != noErr) Console_WriteLine("copy to property token failed");
			Console_WriteValue("property token flags", propertyToken.flags);
			Console_WriteValueFourChars("property token class", propertyToken.as.property.dataType);
			Console_WriteValueFourChars("property token container event class", propertyToken.as.property.container.eventClass);
			result = GenesisAE_CreateTokenFromObjectData(propertyToken.as.property.dataType, &propertyToken,
															&propertyTokenDescriptor, null/* data type */);
			if (result != noErr) Console_WriteLine("token from object data failed");
			{
				Boolean		usedDesiredDataType = false; // temp
				
				
				result = GetPropertyAE_DataFromClassProperty(&propertyTokenDescriptor, typeWildCard,
																&containerDescriptor, &usedDesiredDataType);
			}
			if (result != noErr) Console_WriteLine("data from class property failed");
		}
		else
		{
			// the class owning the requested property is provided explicitly
			// (coercion probably isnt needed); check for the special case of
			// a null container, since properties from here dont have any
			// physical container (or in some sense, the container is the
			// application itself, such as the "name", "version", etc.)
			if (inFromWhatClass == typeNull)
			{
				Console_WriteLine("an application property, cutting to the chase");
				GetElementAE_ApplicationImplicit(&containerDescriptor);
			}
			else
			{
				result = AECoerceDesc(inFromWhat, kTelnetHiddenClassTelnetToken, &containerDescriptor);
				if (result != noErr) Console_WriteLine("coercion to token class failed");
			}
			
			if (result == noErr)
			{
				result = AppleEventUtilities_CopyDescriptorData(&containerDescriptor, &token, sizeof(token),
																null/* actual size storage */);
			}
		}
		
		if (result == noErr)
		{
			Console_WriteValue("flags", token.flags);
			Console_WriteValueFourChars("container event class", token.as.object.eventClass);
			container = token.as.object;
			if (result != noErr)
			{
				if (result == memSCErr)
				{
					Console_WriteLine("WARNING - not enough buffer space was given, truncating...");
					result = noErr;
				}
				else Console_WriteLine("attempt to get container type failed");
			}
		}
		
		// the type of property being requested is passed by reference;
		// note that, from here onward, the "token" variable is used to
		// describe the PROPERTY, not the container
		if (result == noErr)
		{
			result = AECoerceDesc(inReference, typeType, &propertyTypeDescriptor);
			if (result != noErr) Console_WriteLine("coercion to reference failed");
		}
		if (result == noErr)
		{
			result = AppleEventUtilities_CopyDescriptorData(&propertyTypeDescriptor,
															&token.as.property.dataType,
															sizeof(token.as.property.dataType),
															null/* actual size storage */);
			if (result != noErr) Console_WriteLine("attempt to get property type failed");
		}
		
		// if successful, create a new (internal) token describing this property
		if (result == noErr)
		{
			token.flags = kTelnetTokenFlagsIsProperty;
			token.genericPointer = null;
			token.as.property.container = container;
			result = AECreateDesc(kTelnetHiddenClassTelnetToken, (Ptr)&token, sizeof(token),
									outDesiredProperty);
		}
		if (containerDescriptor.dataHandle != null) AEDisposeDesc(&containerDescriptor);
		if (propertyTypeDescriptor.dataHandle != null) AEDisposeDesc(&propertyTypeDescriptor);
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// FromObject


//
// internal methods
//

/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassApplication".

(3.0)
*/
static OSErr
accessPropertyOfClassApplication	(TelnetTokenPtr				inTokenPtr,
									 AEDesc*					inoutDataDescriptor,
									 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr				result = noErr;
	TelnetPropertyPtr	propertyPtr = &inTokenPtr->as.property;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: application property");
	
	// verify that its a property, and verify its container class type
	if (result == noErr)
	{
		result = ((inTokenPtr->flags & kTelnetTokenFlagsIsProperty) &&
					GenesisAE_FirstClassIs(propertyPtr->container.eventClass, kTelnetClassApplication)) ? noErr
				: errAEEventNotHandled;
	}
	
	// get the data
	if (result == noErr) switch (propertyPtr->dataType)
	{
		case pBestType:
		case pDefaultType:
			Console_WriteLine("default or best type");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				AEDesc		nullDesc = { typeNull, null };
				
				
				result = AEDuplicateDesc(&nullDesc, inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case pClass:
			Console_WriteLine("the class");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				DescType	classType = cApplication;
				
				
				result = AECreateDesc(typeType, (Ptr)&classType, sizeof(classType), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyActiveMacroSet:
			Console_WriteLine("active macro set");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Console_WriteLine("read");
				// UNIMPLEMENTED
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				Console_WriteLine("write");
				// UNIMPLEMENTED
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyClipboard:
			Console_WriteLine("clipboard");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				// UNIMPLEMENTED
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyFolderScriptsMenuItems:
			Console_WriteLine("user Scripts Menu Items folder");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				FSSpec		folder;
				
				
				result = FolderManager_GetFSSpec(kTelnetFolderScriptsMenuItems, &folder);
				if (result == noErr) result = FSMakeFSSpec(folder.vRefNum, folder.parID, folder.name, &folder);
				if (result == noErr) BasicTypesAE_CreateFileOrFolderDesc(&folder, inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyFolderStartupItems:
			Console_WriteLine("user Startup Items folder");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				FSSpec		folder;
				
				
				result = FolderManager_GetFSSpec(kTelnetFolderStartupItems, &folder);
				if (result == noErr) result = FSMakeFSSpec(folder.vRefNum, folder.parID, folder.name, &folder);
				if (result == noErr) BasicTypesAE_CreateFileOrFolderDesc(&folder, inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyIPAddress:
			Console_WriteLine("current IP address");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				char		string[30];
				
				
				if (Network_CurrentIPAddressToCString(string))
				{
					StringUtilities_CToPInPlace(string);
					result = AECreateDesc(typeChar, (Ptr)&string[1],
											string[0] * sizeof(unsigned char), inoutDataDescriptor);
				}
				else result = errAEEventNotHandled;
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyIsFrontmost:
			Console_WriteLine("frontmost");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean			isInFront = !FlagManager_Test(kFlagSuspended);
				
				
				result = AECreateDesc(typeBoolean, (Ptr)&isInFront, sizeof(isInFront), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyPreferences:
			Console_WriteLine("preferences");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				AEDesc			containerDesc = { typeNull, null }, // never changes, in this case
								keyDesc = { typeNull, null };
				
				
				// send a request for "preferences", which resides in a null container
				(OSErr)BasicTypesAE_CreatePropertyIDDesc(kTelnetPropertyPreferences, &keyDesc);
				result = CreateObjSpecifier(cProperty, &containerDesc,
											formPropertyID, &keyDesc, true/* dispose inputs */,
											inoutDataDescriptor);
				if (result != noErr) Console_WriteValue("error creating object specifier", result);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyProgramName:
			Console_WriteLine("name");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Str255					applicationName;
				ProcessInfoRec			processInfo;
				ProcessSerialNumber		currentProcess;
				
				
				applicationName[0] = 0;
				processInfo.processInfoLength = sizeof(processInfo);
				processInfo.processName = applicationName;
				processInfo.processAppSpec = null;
				
				GetCurrentProcess(&currentProcess);
				GetProcessInformation(&currentProcess, &processInfo);
				result = AECreateDesc(typeChar, (Ptr)&applicationName[1],
										applicationName[0] * sizeof(unsigned char), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyProgramVersion:
			Console_WriteLine("version");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				SInt16		oldResFile = CurResFile();
				Handle		versionResourceHandle = null;
				
				
				ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
				versionResourceHandle = Get1Resource('vers', 1);
				HLockHi(versionResourceHandle);
				result = AECreateDesc(typeVersion, *versionResourceHandle,
										GetHandleSize(versionResourceHandle), inoutDataDescriptor);
				HUnlock(versionResourceHandle);
				//???necessary???ReleaseResource(versionResourceHandle), versionResourceHandle = null;
				UseResFile(oldResFile);     // reset back to resource previously set
			}
			else result = errAEEventNotHandled;
			break;
		
		default:
			result = errAECantSupplyType;
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// accessPropertyOfClassApplication


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassApplicationPreferences".

(3.0)
*/
static OSErr
accessPropertyOfClassApplicationPreferences	(TelnetTokenPtr				inTokenPtr,
											 AEDesc*					inoutDataDescriptor,
											 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr				result = noErr;
	TelnetPropertyPtr	propertyPtr = &inTokenPtr->as.property;
	Boolean				canModifyProperty = true;	// change this flag when appropriate to restrict access
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: application preferences property");
	
	// verify that its a property, and verify its container class type
	if (result == noErr)
	{
		result = ((inTokenPtr->flags & kTelnetTokenFlagsIsProperty) &&
					GenesisAE_FirstClassIs(propertyPtr->container.eventClass, kTelnetClassApplicationPreferences)) ? noErr
				: errAEEventNotHandled;
	}
	
	// get the data
	if (result == noErr) switch (propertyPtr->dataType)
	{
		case kTelnetPropertyApplicationPreferencesCursorShape:
			Console_WriteLine("cursor shape");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				result = BasicTypesAE_CreateSizeDesc(sizeof(FourCharCode), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				FourCharCode	shape = kTelnetEnumeratedClassCursorShapeBlock;
				UInt16			terminalCursorType = kTerminalCursorTypeBlock;
				Size			actualSize = 0L;
				
				
				if (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TerminalCursorType,
											sizeof(terminalCursorType), &terminalCursorType,
											&actualSize) == kTelnetPrefsResultCodeSuccess)
				{
					switch (terminalCursorType)
					{
						case kTerminalCursorTypeVerticalLine:
							shape = kTelnetEnumeratedClassCursorShapeVerticalBar;
							break;
						
						case kTerminalCursorTypeThickVerticalLine:
							shape = kTelnetEnumeratedClassCursorShapeVerticalBarBold;
							break;
						
						case kTerminalCursorTypeUnderscore:
							shape = kTelnetEnumeratedClassCursorShapeUnderscore;
							break;
						
						case kTerminalCursorTypeThickUnderscore:
							shape = kTelnetEnumeratedClassCursorShapeUnderscoreBold;
							break;
						
						case kTerminalCursorTypeBlock:
						default:
							shape = kTelnetEnumeratedClassCursorShapeBlock;
							break;
					}
				}
				result = BasicTypesAE_CreateEnumerationDesc(shape, inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				FourCharCode	newShape = kTelnetEnumeratedClassCursorShapeBlock;
				
				
				// make the data match the type in the Dictionary
				result = AECoerceDesc(inoutDataDescriptor, typeEnumerated, inoutDataDescriptor);
				if (result == noErr)
				{
					Size		actualSize = 0L;
					
					
					result = AppleEventUtilities_CopyDescriptorData(inoutDataDescriptor, (Ptr)&newShape,
																	sizeof(newShape), &actualSize);
					if (result == noErr)
					{
						// update the cursor preference (translate from enumerated value)
						UInt16		terminalCursorType = kTerminalCursorTypeBlock;
						
						
						switch (newShape)
						{
							case kTelnetEnumeratedClassCursorShapeVerticalBar:
								terminalCursorType = kTerminalCursorTypeVerticalLine;
								break;
							
							case kTelnetEnumeratedClassCursorShapeVerticalBarBold:
								terminalCursorType = kTerminalCursorTypeThickVerticalLine;
								break;
							
							case kTelnetEnumeratedClassCursorShapeUnderscore:
								terminalCursorType = kTerminalCursorTypeUnderscore;
								break;
							
							case kTelnetEnumeratedClassCursorShapeUnderscoreBold:
								terminalCursorType = kTerminalCursorTypeThickUnderscore;
								break;
							
							case kTelnetEnumeratedClassCursorShapeBlock:
							default:
								terminalCursorType = kTerminalCursorTypeBlock;
								break;
						}
						Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TerminalCursorType,
											sizeof(terminalCursorType), &terminalCursorType);
					}
				}
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyApplicationPreferencesSimpleTelnet:
		case kTelnetPropertyApplicationPreferencesVisualBell:
		default:
			result = errAECantSupplyType;
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// accessPropertyOfClassApplicationPreferences


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassClipboardWindow".

(3.0)
*/
static OSErr
accessPropertyOfClassClipboardWindow	(TelnetTokenPtr				inTokenPtr,
										 AEDesc*					inoutDataDescriptor,
										 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr				result = noErr;
	TelnetPropertyPtr	propertyPtr = &inTokenPtr->as.property;
	Boolean				canModifyProperty = true;	// change this flag when appropriate to restrict access
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: clipboard window property");
	
	// extract the data for the clipboard window to which this property belongs
	//nothingyet
	
	// make ABSOLUTELY SURE the clipboard window is not screwed up (otherwise, the system could crash)
	//nothingyet
	
	// verify that its a property, and verify its container class type
	if (result == noErr)
	{
		result = ((inTokenPtr->flags & kTelnetTokenFlagsIsProperty) &&
					GenesisAE_FirstClassIs(propertyPtr->container.eventClass, kTelnetClassClipboardWindow)) ? noErr
				: errAEEventNotHandled;
	}
	
	// get the data
	if (result == noErr) switch (propertyPtr->dataType)
	{
		case kTelnetPropertyClipboardWindowContents:
			Console_WriteLine("contents");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = 0L;
				
				
			#if TARGET_API_MAC_OS8
				size = (long)InfoScrap()->scrapSize;
			#else
				{
					ScrapRef		scrap = null;
					
					
					(OSStatus)GetCurrentScrap(&scrap);
					(OSStatus)GetScrapFlavorSize(scrap, Clipboard_GetContentType(), &size);
				}
			#endif
				result = BasicTypesAE_CreateSizeDesc(size, inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				result = Clipboard_CreateContentsAEDesc(inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				result = Clipboard_AEDescToScrap(inoutDataDescriptor);
			}
			break;
		
		default:
			// this class inherits from the window class; let *it* handle any other properties
			result = accessPropertyOfClassWindow(inTokenPtr, inoutDataDescriptor, inPropertyAccessMechanism);
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// accessPropertyOfClassClipboardWindow


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassConnection".

(3.0)
*/
static OSErr
accessPropertyOfClassConnection	(TelnetTokenPtr				inTokenPtr,
								 AEDesc*					inoutDataDescriptor,
								 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr		result = noErr;
	
	
	return result;
}// accessPropertyOfClassConnection


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassDialogReply".

(3.0)
*/
static OSErr
accessPropertyOfClassDialogReply	(TelnetTokenPtr				inTokenPtr,
									 AEDesc*					inoutDataDescriptor,
									 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr		result = noErr;
	
	
	return result;
}// accessPropertyOfClassDialogReply


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassMacro".

(3.0)
*/
static OSErr
accessPropertyOfClassMacro		(TelnetTokenPtr				inTokenPtr,
								 AEDesc*					inoutDataDescriptor,
								 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr		result = noErr;
	
	
	return result;
}// accessPropertyOfClassMacro


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassMacroSet".

(3.0)
*/
static OSErr
accessPropertyOfClassMacroSet		(TelnetTokenPtr				inTokenPtr,
									 AEDesc*					inoutDataDescriptor,
									 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr				result = noErr;
	TelnetPropertyPtr	propertyPtr = &inTokenPtr->as.property;
	UInt16				macroSetNumber = null;
	Boolean				canModifyProperty = true;	// change this flag when appropriate to restrict access
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: macro set property");
	
	// access the macro set to which this property belongs
	macroSetNumber = propertyPtr->container.data.macroSet.number;
	
	// make ABSOLUTELY SURE the macro set is not screwed up (otherwise, the system could crash)
	if ((macroSetNumber < 1) || (macroSetNumber > MACRO_SET_COUNT))
	{
		Console_WriteLine("WARNING: invalid macro set for get-macro-set-property");
		result = errAEEventNotHandled;
	}
	
	// verify that its a property, and verify its container class type
	if (result == noErr)
	{
		result = ((inTokenPtr->flags & kTelnetTokenFlagsIsProperty) &&
					GenesisAE_FirstClassIs(propertyPtr->container.eventClass, kTelnetClassMacroSet)) ? noErr
				: errAEEventNotHandled;
	}
	
	// get the data
	if (result == noErr) switch (propertyPtr->dataType)
	{
		case kTelnetPropertyMacroSetKeyEquivalents:
			Console_WriteLine("key equivalents");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				result = BasicTypesAE_CreateSizeDesc(sizeof(FourCharCode), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				FourCharCode	keys = kTelnetEnumeratedClassMacroSetKeyEquivalentsCommandDigit;
				UInt16			oldActiveSetNumber = Macros_GetActiveSetNumber();
				
				
				Macros_SetActiveSetNumber(macroSetNumber);
				if (Macros_GetMode() == kTelnetMacroModeFunctionKeys)
				{
					// if the macro set uses function keys instead, change the default
					keys = kTelnetEnumeratedClassMacroSetKeyEquivalentsFunctionKeys;
				}
				Macros_SetActiveSetNumber(oldActiveSetNumber);
				result = BasicTypesAE_CreateEnumerationDesc(keys, inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				FourCharCode	newKeys = kTelnetEnumeratedClassMacroSetKeyEquivalentsCommandDigit;
				
				
				// make the data match the type in the Dictionary
				result = AECoerceDesc(inoutDataDescriptor, typeEnumerated, inoutDataDescriptor);
				if (result == noErr)
				{
					Size		actualSize = 0L;
					
					
					result = AppleEventUtilities_CopyDescriptorData(inoutDataDescriptor, (Ptr)&newKeys,
																	sizeof(newKeys), &actualSize);
					if (result == noErr)
					{
						// update the macro sets key equivalents
						UInt16			oldActiveSetNumber = Macros_GetActiveSetNumber();
						
						
						Macros_SetActiveSetNumber(macroSetNumber);
						Macros_SetMode((newKeys == kTelnetEnumeratedClassMacroSetKeyEquivalentsFunctionKeys)
										? kTelnetMacroModeFunctionKeys : kTelnetMacroModeCommandDigit);
						// rebuild macro display if necessary to show new key equivalents - INCOMPLETE
						Macros_SetActiveSetNumber(oldActiveSetNumber);
					}
				}
			}
			else result = errAEEventNotHandled;
			break;
		
		default:
			result = errAECantSupplyType;
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// accessPropertyOfClassMacroSet


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassProxyServerConfiguration".

(3.0)
*/
static OSErr
accessPropertyOfClassProxyServer	(TelnetTokenPtr				inTokenPtr,
									 AEDesc*					inoutDataDescriptor,
									 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr		result = noErr;
	
	
	return result;
}// accessPropertyOfClassProxyServer


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassSession".

(3.0)
*/
static OSErr
accessPropertyOfClassSession	(TelnetTokenPtr				inTokenPtr,
								 AEDesc*					inoutDataDescriptor,
								 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr				result = noErr;
	TelnetPropertyPtr	propertyPtr = &inTokenPtr->as.property;
	ConnectionDataPtr	connectionDataPtr = null;
	RealScreenDataPtr	realScreenPtr = null;
	Boolean				canModifyProperty = true;	// change this flag when appropriate to restrict access
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: session property");
	
	// extract the data for the session to which this property belongs
	connectionDataPtr = propertyPtr->container.data.session.connectionPtr;
	realScreenPtr = propertyPtr->container.data.session.realScreenPtr;
	
	// make ABSOLUTELY SURE the session is not screwed up (otherwise, the system could crash)
	if ((realScreenPtr == null) || (connectionDataPtr == null)/* insufficient! */)
	{
		Console_WriteLine("WARNING: invalid session for get-session-property");
		result = errAEEventNotHandled;
	}
	
	// verify that its a property, and verify its container class type
	if (result == noErr)
	{
		result = ((inTokenPtr->flags & kTelnetTokenFlagsIsProperty) &&
					GenesisAE_FirstClassIs(propertyPtr->container.eventClass, kTelnetClassSession)) ? noErr
				: errAEEventNotHandled;
	}
	
	// get the data
	if (result == noErr) switch (propertyPtr->dataType)
	{
		case kTelnetPropertySessionConnection:
			// unimplemented!
			result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertySessionWindow:
			// this will work, but note that it would be far better to use OSL here,
			// taking advantage of MacTelnets own scripting implementation to
			// generate an object reference like "window 1", instead of raw data
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				TelnetToken		token;
				
				
				token.flags = 0;
				token.genericPointer = null;
				token.as.object.data.window.ref = realScreenPtr->window;
				Console_WriteLine("constructing window token");
				result = GenesisAE_CreateTokenFromObjectData(kTelnetClassWindow, &token, inoutDataDescriptor,
																null/* new type */);
			}
			else result = errAEEventNotHandled;
			break;
		
		default:
			result = errAECantSupplyType;
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// accessPropertyOfClassSession


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassTerminalWindow".

(3.0)
*/
static OSErr
accessPropertyOfClassTerminalWindow		(TelnetTokenPtr				inTokenPtr,
										 AEDesc*					inoutDataDescriptor,
										 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr				result = noErr;
	TelnetPropertyPtr	propertyPtr = &inTokenPtr->as.property;
	RealScreenDataPtr	realScreenPtr = null;
	Boolean				canModifyProperty = true;	// change this flag when appropriate to restrict access
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: terminal window property");
	
	// extract the data for the terminal window to which this property belongs
	realScreenPtr = propertyPtr->container.data.terminalWindow.realScreenPtr;
	
	// make ABSOLUTELY SURE the terminal window is not screwed up (otherwise, the system could crash)
	if (realScreenPtr == null)
	{
		Console_WriteLine("WARNING: invalid terminal window for get-terminal-window-property");
		result = errAEEventNotHandled;
	}
	
	// verify that its a property, and verify its container class type
	if (result == noErr)
	{
		result = ((inTokenPtr->flags & kTelnetTokenFlagsIsProperty) &&
					GenesisAE_FirstClassIs(propertyPtr->container.eventClass, kTelnetClassTerminalWindow)) ? noErr
				: errAEEventNotHandled;
	}
	
	// get the data
	if (result == noErr) switch (propertyPtr->dataType)
	{
		case kTelnetPropertyTerminalWindowColumns:
			Console_WriteLine("columns");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				result = BasicTypesAE_CreateSizeDesc(sizeof(UInt32), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				TelnetScreenID		screenID = ScreenManager_GetWindowScreenID
												(propertyPtr->container.data.terminalWindow.windowClass.ref);
				
				
				result = BasicTypesAE_CreateUInt32Desc(VSgetcols(screenID), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				// make the data match the type in the Dictionary
				result = AECoerceDesc(inoutDataDescriptor, typeUInt32, inoutDataDescriptor);
				if (result == noErr)
				{
					Size		actualSize = 0L;
					UInt32		desiredColumns = 0L;
					
					
					// obtain the given column count and ensure it is in range
					result = AppleEventUtilities_CopyDescriptorData(inoutDataDescriptor, (Ptr)&desiredColumns,
																	sizeof(desiredColumns), &actualSize);
					if (result == noErr)
					{
						if ((desiredColumns >= 10) &&
							(desiredColumns <= Terminal_GetAllocatedColumnCount()))
						{
							// set the size and force the window to update
							TelnetScreenID		screenID = ScreenManager_GetWindowScreenID
														(propertyPtr->container.data.terminalWindow.windowClass.ref);
							
							
							ScreenManager_ChangeDimensionsOf(screenID, desiredColumns, VSgetlines(screenID),
																false/* record */);
						}
						else result = errAEEventNotHandled; // out of range
					}
				}
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyTerminalWindowRows:
			Console_WriteLine("row");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				result = BasicTypesAE_CreateSizeDesc(sizeof(UInt32), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				TelnetScreenID		screenID = ScreenManager_GetWindowScreenID
												(propertyPtr->container.data.terminalWindow.windowClass.ref);
				
				
				result = BasicTypesAE_CreateUInt32Desc(VSgetlines(screenID), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				// make the data match the type in the Dictionary
				result = AECoerceDesc(inoutDataDescriptor, typeUInt32, inoutDataDescriptor);
				if (result == noErr)
				{
					Size		actualSize = 0L;
					UInt32		desiredRows = 0L;
					
					
					// obtain the given column count and ensure it is in range
					result = AppleEventUtilities_CopyDescriptorData(inoutDataDescriptor, (Ptr)&desiredRows,
																	sizeof(desiredRows), &actualSize);
					if (result == noErr)
					{
						if ((desiredRows >= 10) && (desiredRows <= 200))
						{
							// set the size and force the window to update
							TelnetScreenID		screenID = ScreenManager_GetWindowScreenID
														(propertyPtr->container.data.terminalWindow.windowClass.ref);
							
							
							ScreenManager_ChangeDimensionsOf(screenID, VSgetcols(screenID), desiredRows,
																false/* record */);
						}
						else result = errAEEventNotHandled; // out of range
					}
				}
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyTerminalWindowScrollBack:
			Console_WriteLine("scrollback");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				result = BasicTypesAE_CreateSizeDesc(sizeof(UInt32), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				result = BasicTypesAE_CreateUInt32Desc(realScreenPtr->min, inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyTerminalWindowFontFamilyName:
		case kTelnetPropertyTerminalWindowFontFamilyNameBold:
			Console_WriteLine("normal or bold font");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				result = BasicTypesAE_CreateSizeDesc(StrLength(realScreenPtr->fontName) * sizeof(unsigned char),
													inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				result = AECreateDesc(cStringClass, (Ptr)(realScreenPtr->fontName + 1),
										StrLength(realScreenPtr->fontName) * sizeof(unsigned char),
										inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				// unimplemented
				result = errAEEventNotHandled;
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyTerminalWindowFontSize:
		case kTelnetPropertyTerminalWindowFontSizeBold:
			Console_WriteLine("normal or bold font size");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				result = BasicTypesAE_CreateSizeDesc(sizeof(realScreenPtr->fsiz), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				result = AECreateDesc(typeSInt16, (Ptr)&realScreenPtr->fsiz, sizeof(realScreenPtr->fsiz),
										inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				// unimplemented
				result = errAEEventNotHandled;
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyTerminalWindowFontColorText:
		case kTelnetPropertyTerminalWindowFontColorBackground:
		case kTelnetPropertyTerminalWindowFontColorBoldText:
		case kTelnetPropertyTerminalWindowFontColorBoldBackground:
		case kTelnetPropertyTerminalWindowFontColorBlinkingText:
		case kTelnetPropertyTerminalWindowFontColorBlinkingBackground:
			Console_WriteLine("color attribute");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				result = BasicTypesAE_CreateSizeDesc(sizeof(RGBColor), inoutDataDescriptor);
			}
			else if ((inPropertyAccessMechanism == kPropertyAccessObtain) ||
						(inPropertyAccessMechanism == kPropertyAccessModify))
			{
				ConnectionDataPtr		connectionDataPtr = ScreenManager_GetWindowConnectionData
																(realScreenPtr->window);
				RealScreenColorIndex	index = kRealScreenColorIndexNormalText;
				RGBColor				color;
				
				
				switch (propertyPtr->dataType)
				{
					case kTelnetPropertyTerminalWindowFontColorBlinkingBackground:
						index = kRealScreenColorIndexBlinkingBackground;
						break;
					
					case kTelnetPropertyTerminalWindowFontColorBlinkingText:
						index = kRealScreenColorIndexBlinkingText;
						break;
					
					case kTelnetPropertyTerminalWindowFontColorBoldBackground:
						index = kRealScreenColorIndexBoldBackground;
						break;
					
					case kTelnetPropertyTerminalWindowFontColorBoldText:
						index = kRealScreenColorIndexBoldText;
						break;
					
					case kTelnetPropertyTerminalWindowFontColorBackground:
						index = kRealScreenColorIndexNormalBackground;
						break;
					
					case kTelnetPropertyTerminalWindowFontColorText:
					default:
						index = kRealScreenColorIndexNormalText;
						break;
				}
				if (inPropertyAccessMechanism == kPropertyAccessModify)
				{
					// make the data match the type in the Dictionary
					result = AECoerceDesc(inoutDataDescriptor, cRGBColor, inoutDataDescriptor);
					if (result == noErr)
					{
						Size	actualSize = 0L;
						
						
						result = AppleEventUtilities_CopyDescriptorData(inoutDataDescriptor, (Ptr)&color,
																		sizeof(color), &actualSize);
						if (result == noErr)
						{
							// set the color and force the window to update
							(Boolean)RealScreen_SetColor(connectionDataPtr->vs, index, &color);
							MacLook_InvalidatePortRectangle(realScreenPtr->window);
						}
					}
				}
				else
				{
					// return the color
					(Boolean)RealScreen_GetColor(connectionDataPtr->vs, index, &color);
					result = BasicTypesAE_CreateRGBColorDesc(&color, inoutDataDescriptor);
				}
			}
			else result = errAEEventNotHandled;
			break;
		
		default:
			// this class inherits from the window class; let *it* handle any other properties
			result = accessPropertyOfClassWindow(inTokenPtr, inoutDataDescriptor, inPropertyAccessMechanism);
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// accessPropertyOfClassTerminalWindow


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassWindow".

(3.0)
*/
static OSErr
accessPropertyOfClassWindow	(TelnetTokenPtr				inTokenPtr,
							 AEDesc*					inoutDataDescriptor,
							 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr				result = noErr;
	TelnetPropertyPtr	propertyPtr = &inTokenPtr->as.property;
	WindowRef			window = null;
	WindowInfoRef		windowInfo = null;
	Boolean				canModifyProperty = true;	// change this flag when appropriate to restrict access
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: window property");
	
	// extract the window to which this property belongs
	window = propertyPtr->container.data.window.ref;
	
	// make ABSOLUTELY SURE the window data is not screwed up (otherwise, the system could crash)
	if (FlagManager_Test(kFlagCarbonWindowManager)) if (!IsValidWindowPtr(window))
	{
		Console_WriteLine("WARNING: invalid window pointer provided for get-window-property");
		result = errAEEventNotHandled;
	}
	
	// if the window is valid, get any Window Info associated with it
	if (result == noErr) windowInfo = WindowInfo_GetFromWindow(window);
	
	// verify that its a property, and verify its container class type
	if (result == noErr)
	{
		result = ((inTokenPtr->flags & kTelnetTokenFlagsIsProperty) &&
					GenesisAE_FirstClassIs(propertyPtr->container.eventClass, kTelnetClassWindow)) ? noErr
				: errAEEventNotHandled;
	}
	
	// get the data
	if (result == noErr) switch (propertyPtr->dataType)
	{
		case kTelnetPropertyWindowBounds:
		case kTelnetPropertyWindowPosition:
			Console_WriteLine("bounding rectangle or position");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = (propertyPtr->dataType == kTelnetPropertyWindowPosition)
								? sizeof(Point) : sizeof(Rect);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Rect		rect;
				Point		topLeft;
				GrafPtr		oldPort = null;
				
				
				// determine the window location in global coordinates
				GetPortBounds(GetWindowPort(window), &rect);
				SetPt(&topLeft, rect.left, rect.top);
				GetPort(&oldPort);
				SetPortWindowPort(window);
				LocalToGlobal(&topLeft);
				rect.left += topLeft.h;
				rect.top += topLeft.v;
				rect.right += topLeft.h;
				rect.bottom += topLeft.v;
				SetPort(oldPort);
				if (propertyPtr->dataType == kTelnetPropertyWindowPosition)
				{
					result = AECreateDesc(kTelnetPTypWindowPosition, (Ptr)&topLeft, sizeof(topLeft),
											inoutDataDescriptor);
				}
				else result = AECreateDesc(kTelnetPTypWindowBounds, (Ptr)&rect, sizeof(rect), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				if (propertyPtr->dataType == kTelnetPropertyWindowPosition)
				{
					canModifyProperty = true; // all windows can be moved
				}
				else if (windowInfo != null)
				{
					switch (WindowInfo_GetWindowDescriptor(windowInfo))
					{
						case kTelnetWindowDescriptorClipboard:
						case kTelnetDialogDescriptorPreferences:
							// bounding rectangles of these windows can be changed
							canModifyProperty = true;
							break;
						
						default:
							// most other windows cannot be resized
							canModifyProperty = false;
							break;
					}
				}
				else canModifyProperty = false;
				
				if (canModifyProperty)
				{
					if (propertyPtr->dataType == kTelnetPropertyWindowPosition)
					{
						Point		newPosition;
						
						
						// make the data match the type in the Dictionary
						result = AECoerceDesc(inoutDataDescriptor, kTelnetPTypWindowPosition, inoutDataDescriptor);
						if (result == noErr)
						{
							Size		actualSize = 0L;
							
							
							result = AppleEventUtilities_CopyDescriptorData(inoutDataDescriptor, (Ptr)&newPosition,
																			sizeof(newPosition), &actualSize);
							if (result == noErr)
							{
								// move the window
								MoveWindow(window, newPosition.h, newPosition.v, false/* activate */);
							}
						}
					}
					else
					{
						Rect		newContentRect;
						
						
						// make the data match the type in the Dictionary
						result = AECoerceDesc(inoutDataDescriptor, kTelnetPTypWindowBounds, inoutDataDescriptor);
						if (result == noErr)
						{
							Size		actualSize = 0L;
							
							
							result = AppleEventUtilities_CopyDescriptorData(inoutDataDescriptor, (Ptr)&newContentRect,
																			sizeof(newContentRect), &actualSize);
							if (result == noErr)
							{
								Rect		contentRect;
								SInt32		deltaX = 0,
											deltaY = 0;
								
								
								// determine if the new dimensions are acceptable (if possible)
								if (windowInfo != null)
								{
									Point		size;
									
									
									WindowInfo_GetWindowMaximumDimensions(windowInfo, &size);
									if ((newContentRect.right - newContentRect.left) > size.h)
									{
										result = errAEEventNotHandled;
									}
									if ((newContentRect.bottom - newContentRect.top) > size.v)
									{
										result = errAEEventNotHandled;
									}
									
									WindowInfo_GetWindowMinimumDimensions(windowInfo, &size);
									if ((newContentRect.right - newContentRect.left) < size.h)
									{
										result = errAEEventNotHandled;
									}
									if ((newContentRect.bottom - newContentRect.top) < size.v)
									{
										result = errAEEventNotHandled;
									}
								}
								
								// if so, continue to move and/or resize the window
								if (result == noErr)
								{
									// calculate the change in size (used for automatic control resizing)
									MacLook_GetWindowRegionBounds(window, kWindowContentRgn, &contentRect);
									deltaX = (newContentRect.right - newContentRect.left) -
												(contentRect.right - contentRect.left);
									deltaY = (newContentRect.bottom - newContentRect.top) -
												(contentRect.bottom - contentRect.top);
									
									// rearrange the window, and resize its controls if possible (and if necessary)
									MoveWindow(window, newContentRect.left, newContentRect.top, false/* activate */);
									SizeWindow(window, newContentRect.right - newContentRect.left,
												newContentRect.bottom - newContentRect.top, true/* update */);
									if (windowInfo != null)
									{
										result = WindowInfo_CallWindowResizeResponderUPP(window, deltaX, deltaY);
									}
								}
							}
						}
					}
				}
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowHasCloseBox:
			Console_WriteLine("closeable");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		flag = false;
				
				
			#if TARGET_API_MAC_OS8
				flag = GetWindowGoAwayFlag(window);
			#else
				{
					WindowAttributes	attributes;
					
					
					GetWindowAttributes(window, &attributes);
					flag = ((attributes & kWindowCloseBoxAttribute) != 0);
				}
			#endif
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowHasCollapseBox:
			Console_WriteLine("collapsible");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		flag = IsWindowCollapsable(window);
				
				
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowHasTitleBar:
			Console_WriteLine("titled");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		flag = true; // UNIMPLEMENTED - but virtually all MacTelnet windows have title bars!
				
				
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowHidden:
			Console_WriteLine("obscured");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				// this trait only applies to terminal windows
				Boolean		flag = ScreenManager_IsWindowAScreen(window);
				
				
				if (flag)
				{
					RealScreenDataPtr		realScreenPtr = ScreenManager_GetWindowRealScreenData(window);
					
					
					if (realScreenPtr != null) flag = realScreenPtr->obscured;
				}
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowIndexInWindowList:
			Console_WriteLine("index");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(SInt16);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				SInt16		index = 1; // UNIMPLEMENTED
				
				
				result = AECreateDesc(typeSInt16, (Ptr)&index, sizeof(index), inoutDataDescriptor);
				if (result == noErr)
				{
					// make the data match the type in the Dictionary
					result = AECoerceDesc(inoutDataDescriptor, kTelnetPTypWindowIndexInWindowList,
											inoutDataDescriptor);
				}
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowIsCollapsed:
			Console_WriteLine("collapsed");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		flag = IsWindowCollapsed(window);
				
				
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				result = AECoerceDesc(inoutDataDescriptor, typeBoolean, inoutDataDescriptor);
				if (result == noErr)
				{
					Size		actualSize = 0L;
					Boolean		flag = false;
					
					
					result = AppleEventUtilities_CopyDescriptorData(inoutDataDescriptor, (Ptr)&flag,
																	sizeof(flag), &actualSize);
					if (result == noErr)
					{
						// collapse or un-collapse the window (minimizes on Mac OS X)
						CollapseWindow(window, flag/* collapse / minimize */);
					}
				}
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowIsFloating:
			Console_WriteLine("floating");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		flag = false; // UNIMPLEMENTED
				
				
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowIsModal:
			Console_WriteLine("modal");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		flag = false; // UNIMPLEMENTED
				
				
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowIsResizeable:
			Console_WriteLine("resizeable");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		flag = false; // UNIMPLEMENTED
				
				
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowIsVisible:
			Console_WriteLine("visible");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		visibility = IsWindowVisible(window);
				
				
				result = AECreateDesc(typeBoolean, (Ptr)&visibility, sizeof(visibility), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowIsZoomable:
			Console_WriteLine("zoomable");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		flag = false;
				
				
			#if TARGET_API_MAC_OS8
				flag = GetWindowZoomFlag(window);
			#else
				{
					WindowAttributes	attributes;
					
					
					GetWindowAttributes(window, &attributes);
					flag = (((attributes & kWindowHorizontalZoomAttribute) != 0) ||
							((attributes & kWindowVerticalZoomAttribute) != 0));
				}
			#endif
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowIsZoomed:
			Console_WriteLine("zoomed");
			if (inPropertyAccessMechanism == kPropertyAccessSize)
			{
				Size		size = sizeof(Boolean);
				
				
				result = AECreateDesc(typeUInt32, (Ptr)&size, sizeof(size), inoutDataDescriptor);
			}
			else if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Boolean		flag = false; // UNIMPLEMENTED
				
				
				result = AECreateDesc(typeBoolean, (Ptr)&flag, sizeof(flag), inoutDataDescriptor);
			}
			else result = errAEEventNotHandled;
			break;
		
		case kTelnetPropertyWindowName:
			Console_WriteLine("name");
			if (inPropertyAccessMechanism == kPropertyAccessObtain)
			{
				Str255		windowName;
				
				
				GetWTitle(window, windowName);
				result = AECreateDesc(cStringClass, (Ptr)&windowName[1],
										windowName[0] * sizeof(unsigned char), inoutDataDescriptor);
				if (result == noErr)
				{
					// make the data match the type in the Dictionary
					result = AECoerceDesc(inoutDataDescriptor, kTelnetPTypWindowName, inoutDataDescriptor);
				}
			}
			else if (inPropertyAccessMechanism == kPropertyAccessModify)
			{
				if (windowInfo != null)
				{
					switch (WindowInfo_GetWindowDescriptor(windowInfo))
					{
						case kTelnetWindowDescriptorAnyTerminal:
						case kTelnetWindowDescriptorDebuggingConsole:
							// titles of these windows can be changed
							canModifyProperty = true;
							break;
						
						default:
							// most other window titles cannot be changed
							canModifyProperty = false;
							break;
					}
				}
				else canModifyProperty = true;
				
				if (canModifyProperty)
				{
					Str255		windowName;
					
					
					// make the data match the type in the Dictionary
					result = AECoerceDesc(inoutDataDescriptor, kTelnetPTypWindowName, inoutDataDescriptor);
					if (result == noErr)
					{
						Size	actualSize = 0L;
						
						
						result = AppleEventUtilities_CopyDescriptorData(inoutDataDescriptor, (Ptr)&windowName[1],
																		sizeof(windowName), &actualSize);
						if (result == noErr)
						{
							// set the length (its a Pascal string) and then set the title of the window
							windowName[0] = ((unsigned char)actualSize);
							SetWTitle(window, windowName);
						}
					}
				}
				else result = errAEEventNotHandled;
			}
			else result = errAEEventNotHandled;
			break;
		
		default:
			result = errAECantSupplyType;
			break;
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// accessPropertyOfClassWindow


/*!
Used to determine the data size, or acquire, or
modify, the property whose token is given.  The
token encodes the container class, as well, which
should be "kTelnetClassWord".

(3.0)
*/
static OSErr
accessPropertyOfClassWord		(TelnetTokenPtr				inTokenPtr,
								 AEDesc*					inoutDataDescriptor,
								 PropertyAccessMechanism	inPropertyAccessMechanism)
{
	OSErr		result = noErr;
	
	
	return result;
}// accessPropertyOfClassWord


/*!
To determine if a property accessor has been
invoked erroneously, given the descriptor
type parameters passed to the accessor, use
this method.

(3.0)
*/
static OSErr
propertyAccessError	(DescType		inDesiredClass,
					 DescType		inFormOfReference)
{
	OSErr		result = noErr;
	
	
	if ((inDesiredClass != cProperty) || (inFormOfReference != formPropertyID))
	{
		result = errAEWrongDataType;
	}
	
	return result;
}// propertyAccessError


/*!
This routine helps to reduce code duplication,
and handles construction of a token data type
and property access using that token.

(3.0)
*/
static OSErr
tokenAccess		(AEDesc const*				inTokenDescriptor,
				 AEDesc*					inoutDataDescriptor,
				 PropertyAccessMechanism	inAccessMechanism)
{
	OSErr			result = noErr;
	TelnetToken		token;
	
	
	Console_BeginFunction();
	Console_WriteLine("accessing data");
	result = AppleEventUtilities_CopyDescriptorData(inTokenDescriptor, &token, sizeof(token),
													null/* actual size storage */);
	if (result == noErr)
	{
		if (token.flags & kTelnetTokenFlagsIsProperty)
		{
			// request to get, set, or obtain the data size of a property of a class
			Console_WriteLine("(a property of a class)");
			Console_WriteValueFourChars("specific property", token.as.property.dataType);
			switch (token.as.property.container.eventClass)
			{
				case kTelnetClassApplication:
					result = accessPropertyOfClassApplication(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassApplicationPreferences:
					result = accessPropertyOfClassApplicationPreferences(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassClipboardWindow:
					result = accessPropertyOfClassClipboardWindow(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassConnection:
					result = accessPropertyOfClassConnection(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassDialogReply:
					result = accessPropertyOfClassDialogReply(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassMacro:
					result = accessPropertyOfClassMacro(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassMacroSet:
					result = accessPropertyOfClassMacroSet(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassProxyServerConfiguration:
					result = accessPropertyOfClassProxyServer(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassSession:
					result = accessPropertyOfClassSession(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassTerminalWindow:
					result = accessPropertyOfClassTerminalWindow(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassWindow:
					result = accessPropertyOfClassWindow(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				case kTelnetClassWord:
					result = accessPropertyOfClassWord(&token, inoutDataDescriptor, inAccessMechanism);
					break;
				
				default:
					Console_WriteValueFourChars("unrecognized type!", token.as.property.container.eventClass);
					result = errAEWrongDataType; // not a MacTelnet class
					break;
			}
		}
		else if ((token.flags & kTelnetTokenFlagsIsObject) && (inAccessMechanism == kPropertyAccessObtain))
		{
			// request to get a class itself (as opposed to one of its properties)
			Console_WriteLine("(a class itself)");
			switch (token.as.object.eventClass)
			{
				case kTelnetClassApplication:
					// taken care of automatically, because the application container is standard
					result = errAEEventNotHandled;
					break;
				
				case kTelnetClassApplicationPreferences:
					// the only recognized instance of the application preferences class is a
					// property of the null (application) container, so it is taken care of
					// automatically, elsewhere
					result = AECreateDesc(kTelnetClassApplicationPreferences,
											(Ptr)&token.as.object.data.applicationPreferences,
											sizeof(token.as.object.data.applicationPreferences), inoutDataDescriptor);
					break;
				
				case kTelnetClassClipboardWindow:
					result = AECreateDesc(kTelnetClassClipboardWindow, (Ptr)&token.as.object.data.clipboardWindow,
											sizeof(token.as.object.data.clipboardWindow), inoutDataDescriptor);
					break;
				
				case kTelnetClassConnection:
					{
						TelnetClassConnectionPtr	dataPtr = &token.as.object.data.connection;
						AEDesc						containerDesc = { typeNull, null }, // never changes, in this case
													keyDesc = { typeNull, null };
						DescType					keyForm = formAbsolutePosition;
						
						
						// TEMP - return the specifier "connection 1", which is correct only for the frontmost window!!!
						(OSErr)BasicTypesAE_CreateSInt32Desc(1, &keyDesc);
						keyForm = formAbsolutePosition;
						result = CreateObjSpecifier(kTelnetClassConnection, &containerDesc, keyForm, &keyDesc,
													true/* dispose inputs */, inoutDataDescriptor);
						if (result != noErr) Console_WriteValue("error creating object specifier", result);
					}
					break;
				
				case kTelnetClassDialogReply:
					// incomplete - there needs to be a global application property like
					// "most recent dialog reply" to handle this properly
					result = errAEEventNotHandled;
					break;
				
				case kTelnetClassMacro:
					// incomplete - needs to use token.as.object.data.macro to find macro set
					{
						TelnetClassMacroPtr		dataPtr = &token.as.object.data.macro;
						
						
						result = errAEEventNotHandled;
					}
					break;
				
				case kTelnetClassMacroSet:
					{
						TelnetClassMacroSetPtr	dataPtr = &token.as.object.data.macroSet;
						AEDesc					containerDesc = { typeNull, null }, // never changes, in this case
												keyDesc = { typeNull, null };
						
						
						// send a request for "macro set <index>", which resides in a null container
						(OSErr)BasicTypesAE_CreateSInt32Desc(dataPtr->number, &keyDesc);
						result = CreateObjSpecifier(kTelnetClassMacroSet, &containerDesc,
													formAbsolutePosition, &keyDesc, true/* dispose inputs */,
													inoutDataDescriptor);
						if (result != noErr) Console_WriteValue("error creating object specifier", result);
					}
					break;
				
				case kTelnetClassProxyServerConfiguration:
					result = AECreateDesc(kTelnetClassProxyServerConfiguration, (Ptr)&token.as.object.data.proxyServer,
											sizeof(token.as.object.data.proxyServer), inoutDataDescriptor);
					break;
				
					result = AECreateDesc(kTelnetClassSession, (Ptr)&token.as.object.data.session,
											sizeof(token.as.object.data.session), inoutDataDescriptor);
					break;
				
				case kTelnetClassSession:
					{
						TelnetClassSessionPtr	dataPtr = &token.as.object.data.session;
						AEDesc					containerDesc = { typeNull, null }, // never changes, in this case
												keyDesc = { typeNull, null };
						DescType				keyForm = formAbsolutePosition;
						
						
						// TEMP - return the specifier "session 1", which is correct only for the frontmost window!!!
						(OSErr)BasicTypesAE_CreateSInt32Desc(1, &keyDesc);
						keyForm = formAbsolutePosition;
						result = CreateObjSpecifier(kTelnetClassSession, &containerDesc, keyForm, &keyDesc,
													true/* dispose inputs */, inoutDataDescriptor);
						if (result != noErr) Console_WriteValue("error creating object specifier", result);
					}
					break;
				
				case kTelnetClassTerminalWindow:
				case kTelnetClassWindow:
					{
						TelnetClassWindowPtr	dataPtr = &token.as.object.data.window;
						AEDesc					containerDesc = { typeNull, null }, // never changes, in this case
												keyDesc = { typeNull, null };
						DescType				keyForm = formAbsolutePosition;
						
						
						if (dataPtr->ref == GetRealFrontWindow())
						{
							// frontmost window; return the specifier "window 1", which is more flexible
							(OSErr)BasicTypesAE_CreateSInt32Desc(1, &keyDesc);
							keyForm = formAbsolutePosition;
						}
						else
						{
							// return the windows name as the specifier; not as good as an index, especially
							// if multiple windows have the same name, but probably good enough in most cases
							Str255		name;
							
							
							GetWTitle(dataPtr->ref, name);
							BasicTypesAE_CreatePStringDesc(name, &keyDesc);
							keyForm = formName;
						}
						result = CreateObjSpecifier(kTelnetClassWindow, &containerDesc, keyForm, &keyDesc,
													true/* dispose inputs */, inoutDataDescriptor);
						if (result != noErr) Console_WriteValue("error creating object specifier", result);
					}
					break;
				
				case kTelnetClassWord:
					result = AECreateDesc(kTelnetClassWord, (Ptr)&token.as.object.data.word,
											sizeof(token.as.object.data.word), inoutDataDescriptor);
					break;
				
				default:
					Console_WriteLine("unrecognized type!");
					result = errAEWrongDataType; // not a MacTelnet class
					break;
			}
		}
		else
		{
			// request to get WHO KNOWS WHAT!
			Console_WriteLine("unrecognized request!");
			result = errAEWrongDataType;
		}
	}
	Console_WriteValue("result", result);
	Console_EndFunction();
	
	return result;
}// tokenAccess