/*###############################################################

	InstallAE.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <string.h>

// Mac includes
#include <AEDataModel.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <Errors.h>
#include <Processes.h>

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "CoercionsAE.h"
#include "ConnectionAE.h"
#include "Console.h"
#include "CoreSuiteAE.h"
#include "DataAccessAE.h"
#include "GenesisAE.h"
#include "GetElementAE.h"
#include "GetPropertyAE.h"
#include "HiddenAE.h"
#include "InstallAE.h"
#include "LineMode.h"
#include "ObjectClassesAE.h"
#include "RequiredAE.h"
#include "Terminology.h"
#include "TelnetDescriptors.h"
#include "TextServicesAE.h"
#include "URLAccessAE.h"



// variables
static AEAddressDesc		gSelfAddress;			// allows MacTelnet to be recordable
static ProcessSerialNumber	gSelfProcessID;			// identifies MacTelnet in terms of an operating system process
static SInt32				gRecordingCount = 0L;	// number of recordings taking place

// internal methods
static void				enforceLineMode							(ConnectionDataPtr const	inConnectionDataPtr,
																 RealScreenDataPtr const	inRealScreenDataPtr,
																 SInt32 const				inData1,
																 SInt32 const				inData2,
																 void*						inoutResultPtr);
static pascal OSErr		handleRecordingBegunEvent				(AppleEvent const*			inAppleEventPtr,
																 AppleEvent*				outReplyAppleEventPtr,
																 UInt32						inData);
static pascal OSErr		handleRecordingTerminatedEvent			(AppleEvent const*			inAppleEventPtr,
																 AppleEvent*				outReplyAppleEventPtr,
																 UInt32						inData);
static OSErr			installApplicationEventClassHandlers	(void);
static OSErr			installCoercionHandlers					(void);
static OSErr			installDataHandlers						(void);
static OSErr			installHiddenEventHandlers				(void);
static OSErr			installMiscellaneousCoreHandlers		(void);
static OSErr			installObjectAccessors					(void);
static Boolean			installRequiredHandlers					(void);
static OSErr			installTextServicesManagerHandlers		(void);
static OSErr			installURLAccessEventHandlers			(void);
static pascal OSErr		tokenComparer							(DescType					inOperator,
																 AEDesc const*				inTokenDescriptor1,
																 AEDesc const*				inTokenDescriptor2,
																 Boolean*					outAreEqual);
static pascal OSErr		tokenDestroyer							(AEDesc*					inoutTokenDescriptor);


//
// public methods
//

/*!
This method initializes the Apple Event handlers
for the Required Suite, and creates an address
descriptor so MacTelnet can send events to itself
(for recordability).

If any problems occur, false is returned and the
application should be considered impossible to
start up!

You must call AppleEvents_Done() to clean up
the memory allocations, etc. performed by a call
to AppleEvents_Init() when you are done with
Apple Events (that is, when the program ends).

(3.0)
*/
Boolean
InstallAE_Init	(void)
{
	Boolean		result = true;
	OSErr		error = noErr;
	long		gestaltValue = 0L;
	
	
	error = Gestalt(gestaltAppleEventsAttr, &gestaltValue); // see if AppleEvents are available
	{
		Boolean		haveAppleEvents = ((error == noErr) &&
										((gestaltValue >> gestaltAppleEventsPresent) & 0x0001));
		
		
		if (haveAppleEvents) result = installRequiredHandlers();
		else result = false; // no AppleScript, big problem!
	}
	
	if (result) // Got milk?  Go for cookies.
	{
		error = AEObjectInit();
		result = (error == noErr);
	}
	
	if (result) // OK so far?
	{
		// Set up a self-addressed, stamped envelope so MacTelnet can perform all
		// of its operations by making an end run through the system.  This odd
		// but powerful approach allows recording applications, such as the
		// Script Editor, to automatically write scripts based on what users do
		// in MacTelnet!
		gSelfProcessID.highLongOfPSN = 0;
		gSelfProcessID.lowLongOfPSN = kCurrentProcess; // dont use GetCurrentProcess()!
		error = AECreateDesc(typeProcessSerialNumber,
							(Ptr)&gSelfProcessID, sizeof(ProcessSerialNumber), &gSelfAddress);
		result = (error == noErr);
	}
	
	if (result) // still OK?
	{
		// install data access handlers
		if (result) error = installDataHandlers();
		result = (error == noErr);
		
		// install other types of core event handlers
		if (result) error = installMiscellaneousCoreHandlers();
		result = (error == noErr);
	}
	
	if (result) // still OK?
	{
		// install event handlers for every MacTelnet event class
		if (result) error = installApplicationEventClassHandlers();
		result = (error == noErr);
		if (result) error = installHiddenEventHandlers();
		result = (error == noErr);
		if (result) error = installURLAccessEventHandlers();
		result = (error == noErr);
	}
	
	// recording notification event handlers (not critical if their installation fails)
	(OSErr)AEInstallEventHandler(kCoreEventClass, kAENotifyStartRecording,
									NewAEEventHandlerUPP(handleRecordingBegunEvent),
									0L, false);
	(OSErr)AEInstallEventHandler(kCoreEventClass, kAENotifyStopRecording,
									NewAEEventHandlerUPP(handleRecordingTerminatedEvent),
									0L, false);
	
	// install Text Services Manager required event handlers; if this fails, pretend TSM doesnt exist
	//if (FlagManager_Test(kFlagTextServicesManager))
	//{
	//	error = installTextServicesManagerHandlers();
	//	if (error != noErr) FlagManager_Set(kFlagTextServicesManager, false);
	//}
	
	// object accessors
	error = installObjectAccessors();
	result = (error == noErr);
	
	// coercion handlers
	error = installCoercionHandlers();
	result = (error == noErr);
	
	return result;
}// Init


/*!
To clean up this module after using it,
call this method.

(3.0)
*/
void
InstallAE_Done	(void)
{
	AEDisposeDesc(&gSelfAddress);
}// Done


/*!
Creates an Apple Event using AECreateAppleEvent(), with
the most likely parameters for send to self events:
InstallAE_GetSelfAddress(), kAutoGenerateReturnID, and
kAnyTransactionID.

(3.0)
*/
OSErr
InstallAE_CreateRecordableAppleEvent	(DescType const				inEventClass,
										 DescType const				inEventID,
										 AppleEvent*				outAppleEventPtr)
{
	OSErr		result = noErr;
	
	
	result = AECreateAppleEvent(inEventClass, inEventID, InstallAE_GetSelfAddress(),
								kAutoGenerateReturnID, kAnyTransactionID, outAppleEventPtr);
	return result;
}// CreateRecordableAppleEvent


/*!
This method determines if the required events
for the specified Apple Event were received
properly.  If so, "noErr" is returned; otherwise,
"errAEEventNotHandled" is returned.

(3.0)
*/
AEAddressDesc*
InstallAE_GetSelfAddress	(void)
{
	return &gSelfAddress;
}// GetSelfAddress


//
// internal methods
//

/*!
This method, of ScreenOperationProcPtr form,
ensures that line mode is initialized and
activated for the specified connection.

(3.0)
*/
static void
enforceLineMode			(ConnectionDataPtr const	inConnectionDataPtr,
						 RealScreenDataPtr const	UNUSED_ARGUMENT(inRealScreenDataPtr),
						 SInt32 const				UNUSED_ARGUMENT(inData1),
						 SInt32 const				UNUSED_ARGUMENT(inData2),
						 void*						UNUSED_ARGUMENT(inoutResultPtr))
{
	if (inConnectionDataPtr != null)
	{
		unless (inConnectionDataPtr->lineMode.allowed)
		{
			inConnectionDataPtr->lineMode.allowed = true;
			LineMode_Init(inConnectionDataPtr);
		}
		inConnectionDataPtr->lineMode.active = true;
	}
}// enforceLineMode


/*!
Handles the following event types:

kCoreEventClass
---------------
	kAENotifyStartRecording

(3.0)
*/
static pascal OSErr
handleRecordingBegunEvent	(AppleEvent const*	inAppleEventPtr,
							 AppleEvent*		UNUSED_ARGUMENT(outReplyAppleEventPtr),
							 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr			result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: recording begun event");
	
	result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
	
	// Hmmm...is it possible for multiple recordings to be taking place at once?
	// Assume that it is, and keep track of the number of recordings in progress.
	++gRecordingCount;
	FlagManager_Set(kFlagAppleScriptRecording, true);
	
	// turn on line mode for all open connections; this is required to capture data for send recordings
	ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagAllSessions, enforceLineMode, 0L/* data 1 - unused */,
									0L/* data 2 - unused */, null/* result - unused */);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// handleRecordingBegunEvent


/*!
Handles the following event types:

kCoreEventClass
---------------
	kAENotifyStopRecording

(3.0)
*/
static pascal OSErr
handleRecordingTerminatedEvent	(AppleEvent const*	inAppleEventPtr,
								 AppleEvent*		UNUSED_ARGUMENT(outReplyAppleEventPtr),
								 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = noErr;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: recording terminated event");
	
	result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
	
	// Hmmm...is it possible for multiple recordings to be taking place at once?
	// Assume that it is, and keep track of recordings; clear the flag only when
	// all recordings have been terminated.
	--gRecordingCount;
	if (gRecordingCount <= 0) FlagManager_Set(kFlagAppleScriptRecording, false);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// handleRecordingTerminatedEvent


/*!
This routine installs Apple Event handlers for
the MacTelnet Suite.

If any handler cannot be installed, an error
is returned and some handlers may not be
installed properly.

(3.0)
*/
static OSErr
installApplicationEventClassHandlers		(void)
{
	OSErr				result = noErr;
	AEEventClass		eventClass = kTelnetEventClass;
	AEEventHandlerUPP	upp = null;
	
	
	// install event handlers for every MacTelnet event
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(AppleEvents_HandleAlert);
		result = AEInstallEventHandler(eventClass, kTelnetEventIDAlert, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(ConnectionAE_HandleConnectionOpen);
		result = AEInstallEventHandler(eventClass, kTelnetEventIDConnectionOpen, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(AppleEvents_HandleCopySelectedText);
		result = AEInstallEventHandler(eventClass, kTelnetEventIDCopyToClipboard, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(AppleEvents_HandleLaunchFind);
		result = AEInstallEventHandler(eventClass, kTelnetEventIDLaunchFind, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(ConnectionAE_HandleSend);
		result = AEInstallEventHandler(eventClass, kTelnetEventIDDataSend, upp, 0L, false);
	}
	
	return result;
}// installApplicationEventClassHandlers


/*!
This routine installs Apple Event type coercion
handlers.  MacTelnet uses coercion handlers to
convert complex data types into simpler ones,
and to resolve objects into internal tokens.

Incomplete.

(3.0)
*/
static OSErr
installCoercionHandlers		(void)
{
	OSErr					result = noErr;
	DescType				sourceType = typeNull,
							destinationType = typeNull;
	AECoercionHandlerUPP	upp = null;
	
	
	sourceType = typeObjectSpecifier;
	destinationType = kTelnetHiddenClassTelnetToken;
	upp = (AECoercionHandlerUPP)NewAECoerceDescUPP(CoercionsAE_CoerceObjectToAnotherType);
	result = AEInstallCoercionHandler(sourceType, destinationType, upp, 0L/* reference constant */,
										true/* source is a descriptor? */, false/* system type handler? */);
	
	return result;
}// installCoercionHandlers


/*!
This routine installs Apple Event handlers for
getting and setting data, such as properties
of classes.

If any handler cannot be installed, an error
is returned and some handlers may not be
installed properly.

(3.0)
*/
static OSErr
installDataHandlers		(void)
{
	OSErr				result = noErr;
	AEEventClass		eventClass = kAECoreSuite;
	AEEventHandlerUPP	upp = null;
	
	
	// install event handlers for every data event
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(DataAccessAE_HandleSetData);
		result = AEInstallEventHandler(eventClass, kAESetData, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(DataAccessAE_HandleGetData);
		result = AEInstallEventHandler(eventClass, kAEGetData, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(DataAccessAE_HandleGetDataSize);
		result = AEInstallEventHandler(eventClass, kAEGetDataSize, upp, 0L, false);
	}
	
	return result;
}// installDataHandlers


/*!
This routine installs Apple Event handlers for
the MacTelnet Suite, but in particular for those
events which are not in MacTelnets Dictionary.
These events are used to do things that are
useful for MacTelnet Help, but which are not in
general useful for scripts (such as displaying
particular dialog boxes).

If any handler cannot be installed, an error is
returned and some handlers may not be installed
properly.

(3.0)
*/
static OSErr
installHiddenEventHandlers		(void)
{
	OSErr				result = noErr;
	AEEventClass		eventClass = kTelnetEventClass;
	AEEventHandlerUPP	upp = null;
	
	
	// install event handlers for every MacTelnet event
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(HiddenAE_HandleConsoleWriteLine);
		result = AEInstallEventHandler(eventClass, kTelnetHiddenEventIDConsoleWriteLine, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(HiddenAE_HandleDisplayDialog);
		result = AEInstallEventHandler(eventClass, kTelnetHiddenEventIDDisplayDialog, upp, 0L, false);
	}
	
	return result;
}// installHiddenEventHandlers


/*!
This routine installs Apple Event handlers for
core events not related to data access.  See
installDataHandlers() for details on what is
installed for data-access-related core events.

If any handler cannot be installed, an error
is returned and some handlers may not be
installed properly.

(3.0)
*/
static OSErr
installMiscellaneousCoreHandlers	(void)
{
	OSErr				result = noErr;
	AEEventClass		eventClass = kAECoreSuite;
	AEEventHandlerUPP	upp = null;
	
	
	// install event handlers for non-data core events (see installDataHandlers() for the others)
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(AppleEventUtilities_ObjectExists);
		result = AEInstallEventHandler(eventClass, kAEDoObjectsExist, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(CoreSuiteAE_CountElements);
		result = AEInstallEventHandler(eventClass, kAECountElements, upp, 0L, false);
	}
	#if 0
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(CoreSuiteAE_CreateElement);
		result = AEInstallEventHandler(eventClass, kAECreateElement, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(CoreSuiteAE_Duplicate);
		result = AEInstallEventHandler(eventClass, kAEClone, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(CoreSuiteAE_Close);
		result = AEInstallEventHandler(eventClass, kAEClose, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(CoreSuiteAE_Delete);
		result = AEInstallEventHandler(eventClass, kAEDelete, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(CoreSuiteAE_Move);
		result = AEInstallEventHandler(eventClass, kAEMove, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(CoreSuiteAE_Save);
		result = AEInstallEventHandler(eventClass, kAESave, upp, 0L, false);
	}
	
	eventClass = kAEMiscStandards;
	
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(?);
		result = AEInstallEventHandler(eventClass, kAECut, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(?);
		result = AEInstallEventHandler(eventClass, kAECopy, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(?);
		result = AEInstallEventHandler(eventClass, kAEPaste, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(?);
		result = AEInstallEventHandler(eventClass, kAERevert, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(?);
		result = AEInstallEventHandler(eventClass, kAESelect, upp, 0L, false);
	}
	#endif
	
	return result;
}// installMiscellaneousCoreHandlers


/*!
This routine installs OSL accessors for obtaining
tokens from element references.

If any accessor cannot be installed, an error is
returned and some accessors may not be installed
properly.

(3.0)
*/
static OSErr
installObjectAccessors	(void)
{
	OSErr				result = noErr;
	DescType			obtainedClass = '????',
						referenceClass = '????';
	OSLAccessorUPP		upp = null;
	
	
	// set up callbacks
	result = AESetObjectCallbacks(NewOSLCompareUPP(tokenComparer),
									NewOSLCountUPP(CoreSuiteAE_CountElementsOfContainer),
									NewOSLDisposeTokenUPP(tokenDestroyer),	// disposes of dynamic data in tokens
									(OSLGetMarkTokenUPP)null,
									(OSLMarkUPP)null,
									(OSLAdjustMarksUPP)null,
									(OSLGetErrDescUPP)null);
	
	// install accessors for elements of objects - incomplete
	if (result == noErr)
	{
		upp = NewOSLAccessorUPP(GetElementAE_ApplicationFromNull);
		obtainedClass = cApplication;
		referenceClass = typeNull;
		result = AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewOSLAccessorUPP(GetElementAE_MacroFromMacroSet);
		obtainedClass = kTelnetClassMacro;
		referenceClass = kTelnetClassMacroSet;
		result = AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewOSLAccessorUPP(GetElementAE_MacroSetFromNull);
		obtainedClass = kTelnetClassMacroSet;
		referenceClass = typeNull;
		result = AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
	}
	if (result == noErr)
	{
		// install accessors for all window classes, and classes that can be derived from windows
		upp = NewOSLAccessorUPP(GetElementAE_WindowFromNull);
		obtainedClass = cWindow;
		referenceClass = typeNull;
		result = AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
		if (result == noErr)
		{
			obtainedClass = kTelnetClassConnection;
			result = AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
		}
		if (result == noErr)
		{
			obtainedClass = kTelnetClassSession;
			result = AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
		}
		if (result == noErr)
		{
			obtainedClass = kTelnetClassTerminalWindow;
			result = AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
		}
	}
	
	// install accessors for properties - yes, MacTelnet only uses one function to handle all properties!!!
	if (result == noErr)
	{
		upp = NewOSLAccessorUPP(GetPropertyAE_FromObject);
		obtainedClass = cProperty;
		
		// property access from basic Apple Event classes
		referenceClass = typeNull;
		result = AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
		referenceClass = typeAEList;
		(OSErr)AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
		
		// property access from MacTelnet classes
		referenceClass = kTelnetHiddenClassTelnetToken;
		(OSErr)AEInstallObjectAccessor(obtainedClass, referenceClass, upp, 0L, false);
	}
	
	return result;
}// installObjectAccessors


/*!
This routine installs Apple Event handlers for
the Required Suite, including Appearance events
and the five standard Apple Events (open and
print documents, and open, re-open and quit
application).

If ALL handlers install successfully, "true" is
returned; otherwise, "false" is returned.

(3.0)
*/
static Boolean
installRequiredHandlers		(void)
{
	Boolean			result = false;
	OSErr			error = noErr;
	AEEventClass	eventClass = '????';
	
	
	eventClass = kCoreEventClass;
	if (error == noErr) error = AEInstallEventHandler(eventClass, kAEOpenApplication,
														NewAEEventHandlerUPP(RequiredAE_HandleApplicationOpen),
														0L, false);
	if (error == noErr) error = AEInstallEventHandler(eventClass, kAEReopenApplication,
														NewAEEventHandlerUPP(RequiredAE_HandleApplicationReopen),
														0L, false);
	if (error == noErr) error = AEInstallEventHandler(eventClass, kAEOpenDocuments,
														NewAEEventHandlerUPP(RequiredAE_HandleOpenDocuments),
														0L, false);
	if (error == noErr) error = AEInstallEventHandler(eventClass, kAEPrintDocuments,
														NewAEEventHandlerUPP(RequiredAE_HandlePrintDocuments),
														0L, false);
	if (error == noErr) error = AEInstallEventHandler(eventClass, kAEQuitApplication,
														NewAEEventHandlerUPP(RequiredAE_HandleApplicationQuit),
														0L, false);
	if (error == noErr)
	{
		// ignore any errors for these
		eventClass = kAppearanceEventClass;
		(OSErr)AEInstallEventHandler(eventClass, kAEAppearanceChanged,
									NewAEEventHandlerUPP(RequiredAE_HandleAppearanceSwitch),
									0L, false);
		(OSErr)AEInstallEventHandler(eventClass, kAESystemFontChanged,
									NewAEEventHandlerUPP(RequiredAE_HandleAppearanceSwitch),
									0L, false);
		(OSErr)AEInstallEventHandler(eventClass, kAESmallSystemFontChanged,
									NewAEEventHandlerUPP(RequiredAE_HandleAppearanceSwitch),
									0L, false);
		(OSErr)AEInstallEventHandler(eventClass, kAEViewsFontChanged,
									NewAEEventHandlerUPP(RequiredAE_HandleAppearanceSwitch),
									0L, false);
	}
	
	result = (error == noErr);
	
	return result;
}// installRequiredHandlers


/*!
This routine installs Apple Event handlers for the
Text Services Manager support in MacTelnet 3.0.

If any handler cannot be installed, an error is
returned and some handlers may not be installed
properly.

(3.0)
*/
static OSErr
installTextServicesManagerHandlers		(void)
{
	OSErr				result = noErr;
	AEEventClass		eventClass = kTextServiceClass;
	AEEventHandlerUPP	upp = null;
	
	
	// install event handlers for every required TSM event
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(TextServicesAE_HandleUpdateActiveInputArea);
		result = AEInstallEventHandler(eventClass, kUpdateActiveInputArea, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(TextServicesAE_HandlePositionToOffset);
		result = AEInstallEventHandler(eventClass, kPos2Offset, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(TextServicesAE_HandleOffsetToPosition);
		result = AEInstallEventHandler(eventClass, kOffset2Pos, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(TextServicesAE_HandleShowHideInputWindow);
		result = AEInstallEventHandler(eventClass, kShowHideInputWindow, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(TextServicesAE_HandleUnicodeNotFromInputWindow);
		result = AEInstallEventHandler(eventClass, kUnicodeNotFromInputMethod, upp, 0L, false);
	}
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(TextServicesAE_HandleGetSelectedText);
		result = AEInstallEventHandler(eventClass, kGetSelectedText, upp, 0L, false);
	}
	
	return result;
}// installTextServicesManagerHandlers


/*!
This routine installs Apple Event handlers for
the URL Suite (URL Access).

If any handler cannot be installed, an error
is returned and some handlers may not be
installed properly.

(3.0)
*/
static OSErr
installURLAccessEventHandlers		(void)
{
	OSErr				result = noErr;
	AEEventClass		eventClass = kStandardURLEventClass;
	AEEventHandlerUPP	upp = null;
	
	
	// install event handlers for every supported URL event
	if (result == noErr)
	{
		upp = NewAEEventHandlerUPP(URLAccessAE_HandleUniformResourceLocator);
		result = AEInstallEventHandler(eventClass, kStandardURLEventIDGetURL, upp, 0L, false);
	}
	
	return result;
}// installURLAccessEventHandlers


/*!
This callback routine examines the specified
objects, and determines if the operator holds.
The method for doing this will obviously vary
based on the kind of objects given, and it is
possible that the comparison will fail, giving
the result "errAEEventNotHandled".

The specified operator type can be one of the
following values:

	kAEGreaterThan
	kAEGreaterThanEquals
	kAEEquals
	kAELessThan
	kAELessThanEquals
	kAEBeginsWith
	kAEEndsWith
	kAEContains

Unimplemented.

(3.0)
*/
static pascal OSErr
tokenComparer		(DescType			inOperator,
					 AEDesc const*		inObject1,
					 AEDesc const*		inObject2,
					 Boolean*			outAreEqual)
{
	OSErr		result = noErr;
	
	
	return result;
}// tokenComparer


/*!
This callback routine examines the specified
token (an Apple Event descriptor whose type is
"kTelnetHiddenClassTelnetToken"), assuming that
its data is a "TelnetToken" class (see the file
"ObjectClassesAE.h" for details).  If the class
flags indicate that the dynamic data pointer of
the token is a "Handle" or "Ptr", then the data
is disposed (releasing the memory).

(3.0)
*/
static pascal OSErr
tokenDestroyer		(AEDesc*		inoutTokenDescriptor)
{
	OSErr		result = noErr;
	
	
	if (inoutTokenDescriptor != null)
	{
		TelnetToken		token;
		
		
		if (GenesisAE_CreateTokenFromObjectSpecifier(inoutTokenDescriptor, &token) == noErr)
		{
			if (token.flags & kTelnetTokenFlagsDisposeDataPointer)
			{
				Memory_DisposePtr((Ptr*)&token.genericPointer);
			}
			else if (token.flags & kTelnetTokenFlagsDisposeDataHandle)
			{
				Memory_DisposeHandle((Handle*)&token.genericPointer);
			}
		}
	}
	
	return result;
}// tokenDestroyer