/*###############################################################

	ObjectClassesAE.h
	
	This file lists all of the object classes, and defines the
	token classes.  See "Terminology.h" for constants describing
	MacTelnets classes, all of which begin with "kTelnetClass".
	
	IMPORTANT:	Remember, a structure only needs to contain the
				data that is required in order to *find out* the
				information its OSL object representation
				contains.  For example, the data structure for
				the OSL window class only contains a Mac OS
				window pointer, because all of the properties of
				the scriptable "window" class (visibility, close
				box, etc.) can be determined using that pointer.
				
				By following this simple rule, you safeguard
				against having to update data in more than one
				place!!!  AppleScript support is supposed to
				*supplement* the application, not burden it!
				NEVER store any real data in an OSL object.
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#ifndef __OBJECTCLASSESAE__
#define __OBJECTCLASSESAE__

// Mac includes
#include <AppleEvents.h>
#include <MacTypes.h>

// MacTelnet includes
#include "ConnectionData.h"
#include "RealScreenData.h"
#include "TelnetMacros.h"



// helpers
#define INHERITANCE(x)		x		/* used for clarifying intent, nothing more */

// constants
typedef UInt32 TelnetTokenFlags;
enum
{
	kTelnetTokenFlagsIsObject = (1 << 0),				// token represents an Apple Event class instance
	kTelnetTokenFlagsIsProperty = (1 << 1),				// token represents a property of an Apple Event class
	kTelnetTokenFlagsDisposeDataPointer = (1 << 30),	// data is dynamic pointer that needs disposing
	kTelnetTokenFlagsDisposeDataHandle = (1 << 31)		// data is dynamic handle that needs disposing
};


// types
struct TelnetClassApplication
{
	// the MacTelnet-extended application class (incomplete)
	ProcessSerialNumber		process;
};
#ifndef __cplusplus
typedef struct TelnetClassApplication		TelnetClassApplication;
#endif
typedef TelnetClassApplication*				TelnetClassApplicationPtr;


struct TelnetClassApplicationPreferences
{
	// the MacTelnet-extended application preferences class (does not need any data)
	int		x;
};
#ifndef __cplusplus
typedef struct TelnetClassApplicationPreferences	TelnetClassApplicationPreferences;
#endif
typedef TelnetClassApplicationPreferences*			TelnetClassApplicationPreferencesPtr;


struct TelnetClassConnection
{
	// the MacTelnet connection class
	ConnectionDataPtr		connectionDataPtr;
};
#ifndef __cplusplus
typedef struct TelnetClassConnection		TelnetClassConnection;
#endif
typedef TelnetClassConnection*				TelnetClassConnectionPtr;


struct TelnetClassDialogReply
{
	// the MacTelnet dialog reply class
	Str31			buttonName;
	Boolean			gaveUp;
};
#ifndef __cplusplus
typedef struct TelnetClassDialogReply		TelnetClassDialogReply;
#endif
typedef TelnetClassDialogReply*				TelnetClassDialogReplyPtr;


struct TelnetClassMacro
{
	// the MacTelnet macro class
	SInt16			keyEquivalentType;
	char			string[256];
};
#ifndef __cplusplus
typedef struct TelnetClassMacro				TelnetClassMacro;
#endif
typedef TelnetClassMacro*					TelnetClassMacroPtr;


struct TelnetClassMacroSet
{
	// the MacTelnet macro set index (from 1 to MACRO_SET_COUNT)
	UInt16			number;
};
#ifndef __cplusplus
typedef struct TelnetClassMacroSet			TelnetClassMacroSet;
#endif
typedef TelnetClassMacroSet*				TelnetClassMacroSetPtr;


struct TelnetClassProxyServerConfiguration
{
	// the MacTelnet proxy server class (incomplete)
	UInt8			method;
};
#ifndef __cplusplus
typedef struct TelnetClassProxyServerConfiguration		TelnetClassProxyServerConfiguration;
#endif
typedef TelnetClassProxyServerConfiguration*			TelnetClassProxyServerConfigurationPtr;


struct TelnetClassSession
{
	// the MacTelnet session class
	ConnectionDataPtr		connectionPtr;
	RealScreenDataPtr		realScreenPtr;
};
#ifndef __cplusplus
typedef struct TelnetClassSession			TelnetClassSession;
#endif
typedef TelnetClassSession*					TelnetClassSessionPtr;


struct TelnetClassText
{
	// the MacTelnet text class
	Handle			text;
	SInt32			offset;
	UInt32			length;
};
#ifndef __cplusplus
typedef struct TelnetClassText			TelnetClassText;
#endif
typedef TelnetClassText*				TelnetClassTextPtr;


struct TelnetClassWindow
{
	// the MacTelnet-extended window class
	WindowRef			ref;
};
#ifndef __cplusplus
typedef struct TelnetClassWindow			TelnetClassWindow;
#endif
typedef TelnetClassWindow*					TelnetClassWindowPtr;


struct TelnetClassClipboardWindow
{
	// the MacTelnet clipboard window class (incomplete)
	// WARNING: property accessors will upcast this structure to a TelnetClassWindow
	INHERITANCE(TelnetClassWindow		windowClass);
};
#ifndef __cplusplus
typedef struct TelnetClassClipboardWindow	TelnetClassClipboardWindow;
#endif
typedef TelnetClassClipboardWindow*			TelnetClassClipboardWindowPtr;


struct TelnetClassTerminalWindow
{
	// the MacTelnet terminal window class
	// WARNING: property accessors will upcast this structure to a TelnetClassWindow
	INHERITANCE(TelnetClassWindow		windowClass);
	RealScreenDataPtr					realScreenPtr;
};
#ifndef __cplusplus
typedef struct TelnetClassTerminalWindow	TelnetClassTerminalWindow;
#endif
typedef TelnetClassTerminalWindow*			TelnetClassTerminalWindowPtr;


struct TelnetClassWord
{
	// the MacTelnet word class
	TelnetClassText		contents;
};
#ifndef __cplusplus
typedef struct TelnetClassWord		TelnetClassWord;
#endif
typedef TelnetClassWord*			TelnetClassWordPtr;


struct TelnetObject
{
	DescType		eventClass;		// A "kTelnetClass..." constant (see "Terminology.h" for more details).  This value
									//   determines what part of the "data" union should be used.
	
	// The following union contains an entry for every single
	// Apple Event object class supported by MacTelnet.  Each
	// object class defines data uniquely distinguishing an
	// object from others of its class.
	//
	// Since MacTelnet implements inheritance by making the
	// first part of a subclass data structure a copy of the
	// data structure of its parent class, it is possible to
	// upcast a more refined version of a class simply by
	// referencing the parent class in the union "data".  For
	// example, if "eventClass" is kTelnetClassTerminalWindow,
	// you can safely refer to either "terminalWindow" or
	// "window" in the union below, because the first element
	// of the TelnetClassTerminalWindow structure is in fact
	// a TelnetClassWindow structure.  This allows the window
	// property accessor to function without any typecasting,
	// and allows the terminal window property accessor to
	// invoke the accessor of the window class for any property
	// it does not recognize.  This also means that it is not
	// really possible for a class to have more than one
	// *immediate* superclass (so no multiple inheritance), and
	// we all know multiple inheritance is usually evil anyway,
	// so everybody should be happy!
	union
	{
		TelnetClassApplication					application;
		TelnetClassApplicationPreferences		applicationPreferences;
		TelnetClassClipboardWindow				clipboardWindow;
		TelnetClassConnection					connection;
		TelnetClassDialogReply					dialogReply;
		TelnetClassMacro						macro;
		TelnetClassMacroSet						macroSet;
		TelnetClassProxyServerConfiguration		proxyServer;
		TelnetClassSession						session;
		TelnetClassTerminalWindow				terminalWindow;
		TelnetClassText							text;
		TelnetClassWindow						window;
		TelnetClassWord							word;
		// incomplete
	} data;
};
#ifndef __cplusplus
typedef struct TelnetObject		TelnetObject;
#endif
typedef TelnetObject*			TelnetObjectPtr;


struct TelnetProperty
{
	DescType		dataType;			// A "kTelnetProperty..." constant (see "Terminology.h" for more details).
	TelnetObject	container;			// the object that contains this property
};
#ifndef __cplusplus
typedef struct TelnetProperty	TelnetProperty;
#endif
typedef TelnetProperty*			TelnetPropertyPtr;


struct TelnetToken
{
	TelnetTokenFlags	flags;			// information about this token
	
	void*				genericPointer;	// can be a pointer or a handle; if non-null, set the appropriate flag
										// ("kTelnetTokenFlagsDisposeDataPointer" or "kTelnetTokenFlagsDisposeDataHandle")
										// and AEDisposeToken() will dispose of the data for you by using the appropriate
										// Memory Manager routine (thanks to the Telnet-defined token disposal callback)
	
	// a token is either an object or a property
	union
	{
		TelnetObject		object;
		TelnetProperty		property;
	} as;
};
#ifndef __cplusplus
typedef struct TelnetToken		TelnetToken;
#endif
typedef TelnetToken*			TelnetTokenPtr;



#endif