/*###############################################################

	RequiredAE.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <string.h>

// Mac includes
#include <AEDataModel.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <Errors.h>
#include <PLStringFuncs.h>
#include <Processes.h>

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "Clipboard.h"
#include "Commands.h"
#include "ConnectionData.h"
#include "Connections.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "GenesisAE.h"
#include "InstallAE.h"
#include "ObjectClassesAE.h"
#include "PreferencesDialog.h"
#include "Sets.proto.h"
#include "RequiredAE.h"
#include "TelnetErrors.h"
#include "TelnetMacros.h"
#include "TelnetMain.h"
#include "Terminology.h"

// internal methods
OSErr		handleQuit		(DescType const		inSaveOption);



//
// public methods
//

/*!
Handles the following event types:

kAppearanceEventClass
---------------------
	kAEAppearanceChanged
	kAESystemFontChanged
	kAESmallSystemFontChanged
	kAEViewsFontChanged

(3.0)
*/
pascal OSErr
RequiredAE_HandleAppearanceSwitch	(AppleEvent const*		UNUSED_ARGUMENT(inAppleEventPtr),
									 AppleEvent*			UNUSED_ARGUMENT(outReplyAppleEventPtr),
									 UInt32					UNUSED_ARGUMENT(inData))
{
	// currently, MacTelnet does not do anything useful here; but
	// IT SHOULD recalculate window and control sizes!
	return noErr;
}// HandleAppearanceSwitch


/*!
Handles the following event types:

kCoreEventClass
---------------
	kAEOpenApplication

(3.0)
*/
pascal OSErr
RequiredAE_HandleApplicationOpen	(AppleEvent const*	UNUSED_ARGUMENT(inAppleEventPtr),
									 AppleEvent*		UNUSED_ARGUMENT(outReplyAppleEventPtr),
									 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = errAEEventNotHandled;
	
	
	// MacTelnet does nothing with this event (should it?)
	result = noErr;
	
	return result;
}// HandleApplicationOpen


/*!
Handles the following event types:

kCoreEventClass
---------------
	'pref'

(3.0)
*/
pascal OSErr
RequiredAE_HandleApplicationPreferences		(AppleEvent const*	UNUSED_ARGUMENT(inAppleEventPtr),
											 AppleEvent*		UNUSED_ARGUMENT(outReplyAppleEventPtr),
											 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = errAEEventNotHandled;
	
	
	// display the Preferences window
	PreferencesDialog_Display();
	result = noErr;
	
	return result;
}// HandleApplicationPreferences


/*!
Handles the following event types:

kCoreEventClass
---------------
	kAEReopenApplication

(3.0)
*/
pascal OSErr
RequiredAE_HandleApplicationReopen	(AppleEvent const*	UNUSED_ARGUMENT(inAppleEventPtr),
									 AppleEvent*		UNUSED_ARGUMENT(outReplyAppleEventPtr),
									 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = errAEEventNotHandled;
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: application re-opened event");
	
	// handle the case where MacTelnet has no open windows and the user
	// double-clicks its icon in the Finder
	if (FrontWindow() == null)
	{
		// no open windows - do something about it (incomplete)
	}
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// HandleApplicationReopen


/*!
Handles the following event types:

kCoreEventClass
---------------
	kAEQuitApplication

(3.0)
*/
pascal OSErr
RequiredAE_HandleApplicationQuit	(AppleEvent const*	inAppleEventPtr,
									 AppleEvent*		UNUSED_ARGUMENT(outReplyAppleEventPtr),
									 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = errAEEventNotHandled;
	AEDesc		resultDescriptor = { typeNull, null };
	
	
	result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
	if (result == noErr)
	{
		Size		actualSize = 0L;
		DescType	returnedType,
					saveOption = kAEAsk;
		
		
		// if the optional parameter was given as to whether to close
		// open windows without prompting the user, retrieve it
		result = AEGetParamPtr(inAppleEventPtr, keyAESaveOptions, typeEnumerated, &returnedType,
								(Ptr)&saveOption, sizeof(saveOption), &actualSize);
		if (result == errAEDescNotFound) result = noErr;
		if (result == noErr)
		{
			// kills all open connections (asking the user as appropriate), and if the
			// user never cancels, *flags* the main event loop to terminate cleanly
			// (never directly ExitToShell() in an Apple Event handler!)
			result = handleQuit(saveOption);
		}
	}
	
	return result;
}// HandleApplicationQuit


/*!
Handles the following event types:

kCoreEventClass
---------------
	kAEOpenDocuments

(3.0)
*/
pascal OSErr
RequiredAE_HandleOpenDocuments	(AppleEvent const*		inAppleEventPtr,
								 AppleEvent*			UNUSED_ARGUMENT(outReplyAppleEventPtr),
								 UInt32					UNUSED_ARGUMENT(inData))
{
	OSErr			result = noErr;
	FSSpec			file;
	AEDescList		docList;
	long			index = 0L,
					numberOfItemsInList = 0L;
	Size			actualSize = 0;
	AEKeyword		returnedKeyword;
	DescType		returnedType;
	FileInfo		fileInfo;
	
	
	if ((result = AEGetParamDesc(inAppleEventPtr, keyDirectObject, typeAEList, &docList)) == noErr)
	{
		// check for missing parameters
		if ((result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr)) == noErr)
		{
			// count the number of descriptor records in the list
			if ((result = AECountItems(&docList, &numberOfItemsInList)) == noErr)
			{
				UInt16		setNumber = 0; // cumulative track of how many macro sets have been imported
				UInt16		macrosImportedSuccessfully = 0;
				
				
				for (index = 1; index <= numberOfItemsInList; index++)
				{
					result = AEGetNthPtr(&docList, index, typeFSS, &returnedKeyword, &returnedType,
										(Ptr)&file, sizeof(file), &actualSize);
					if (result == noErr)
					{
						FSpGetFInfo(&file, (FInfo*)&fileInfo); // make sure its a data file
						if (fileInfo.fileCreator == 'ToyS' || fileInfo.fileType == 'osas')
						{
							// it appears to be a script; run it
							AppleEventUtilities_ExecuteScriptFile(&file, true/* notify user of errors */);
						}
						else if (fileInfo.fileCreator == kApplicationCreatorSignature &&
							fileInfo.fileType == kApplicationFileTypeSessionDescription)
						{
							// read a configuration set
							Sets_ReadFromFile(&file);
						}
						else if (//fileInfo.fileCreator == kApplicationCreatorSignature &&
								fileInfo.fileType == kApplicationFileTypeMacroSet)
						{
							// read a macro set, replacing the next set if there is more than one
							UInt16		preservedActiveSetNumber = Macros_GetActiveSetNumber();
							
							
							// no more than the maximum number of macro sets can be simultaneously imported
							if (++setNumber <= MACRO_SET_COUNT)
							{
								TelnetMacroMode		mode = kTelnetMacroModeCommandDigit;
								
								
								Macros_SetActiveSetNumber(setNumber);
								macrosImportedSuccessfully += Macros_ImportFromText
																(Macros_GetActiveSet(), &file, &mode);
								Macros_SetMode(mode);
								Macros_SetActiveSetNumber(preservedActiveSetNumber);
							}
						}
					}
				}
				
				if (macrosImportedSuccessfully)
				{
					Errors_DisplayNoteMessage(rStringsNoteAlerts, siMacroImportSuccessful, -1);
				}
				if (result == noErr) result = AEDisposeDesc(&docList);
			}
		}
	}
	
	return result;
}// HandleOpenDocuments


/*!
Handles the following event types:

kCoreEventClass
---------------
	kAEPrintDocuments

(3.0)
*/
pascal OSErr
RequiredAE_HandlePrintDocuments		(AppleEvent const*		UNUSED_ARGUMENT(inAppleEventPtr),
									 AppleEvent*			UNUSED_ARGUMENT(outReplyAppleEventPtr),
									 UInt32					UNUSED_ARGUMENT(inData))
{
	OSErr		result = errAEEventNotHandled;
	
	
	// this program doesnt print files, although this behavior should probably be corrected somehow
	result = errAEEventNotHandled;
	
	return result;
}// HandlePrintDocuments


//
// MacTelnet Suite Apple Event handlers
//

/*!
Handles the following event types:

kTelnetEventClass
-----------------
	kTelnetEventIDAlert

(3.0)
*/
pascal OSErr
AppleEvents_HandleAlert		(AppleEvent const*		inAppleEventPtr,
							 AppleEvent*			outReplyAppleEventPtr,
							 UInt32					UNUSED_ARGUMENT(inData))
{
	DescType		returnedType = typeWildCard;
	Str255			dialogText,
					helpText;
	Size			actualSize = 0L;
	OSErr			result = noErr;
	
	
	result = AEGetParamPtr(inAppleEventPtr, keyDirectObject, cString, &returnedType,
							dialogText + 1, sizeof(dialogText) - 1 * sizeof(unsigned char), &actualSize);
	dialogText[0] = (unsigned char)(actualSize / sizeof(unsigned char));
	{
		InterfaceLibAlertRef	box = null;
		
		
		box = Alert_New();
		
		if (box != null)
		{
			Str31				button1Name,
								button2Name,
								button3Name;
			OSErr				error = noErr; // generic error with no lasting meaning
			
			
			// set up network disruption parameter: defaults to false if the parameter is not specified
			{
				ModalFilterUPP		upp = null;
				Boolean				isDisruption = true;
				
				
				error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyAlertDisruptsNetwork,
										typeBoolean, &returnedType, &isDisruption, sizeof(Boolean), &actualSize);
				
				// MacTelnets custom filter procedure contains, among other things, the ability to service
				// network events.  The quick and dirty implementation of the no disruption parameter
				// is to disable the custom filter altogether.  Unfortunately, this also disables such
				// things as the ability to use Tab to switch the default button, and the ability to
				// click the default button with the space bar.  But, the custom alert filter does not
				// do anything critical, so it is safe to disable it completely.
				if (!isDisruption) upp = NewModalFilterUPP(StandardAlertEventFilter);
				else (OSErr)GetStdFilterProc(&upp);
				Alert_SetFilterProc(box, upp);
				if (!isDisruption) DisposeModalFilterUPP(upp), upp = null;
			}
			
			// set up speech parameter: defaults to false if the parameter is not specified
			{
				Boolean		isSpeech = false;
				
				
				error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyAlertSpeech,
										typeBoolean, &returnedType, &isSpeech, sizeof(Boolean), &actualSize);
				// This is currently erroneous in that it permanently sets the speech flag, affecting
				// future alerts.  This will be fixed in a future release.  Fortunately, the default
				// value is always reset for Apple Event alerts, so scripts will not know there is a
				// problem.
				Alert_SetUseSpeech(isSpeech);
			}
			
			// set up help text parameter: defaults to nothing if the parameter is not specified
			{
				error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyAlertHelpText,
										kTelnetParmTypeAlertHelpText,
										&returnedType, helpText + 1,
										sizeof(helpText) - 1 * sizeof(unsigned char), &actualSize);
				helpText[0] = (unsigned char)(actualSize / sizeof(unsigned char));
				if (error != noErr) PLstrcpy(helpText, EMPTY_PSTRING);
				Alert_SetText(box, dialogText, helpText); // sets the required, main text also
			}
			
			// set up title parameter: defaults to nothing if the parameter is not specified
			{
				Str255		titleText;
				
				
				error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyAlertTitle,
										kTelnetParmTypeAlertTitle,
										&returnedType, titleText + 1,
										sizeof(titleText) - 1 * sizeof(unsigned char), &actualSize);
				titleText[0] = (unsigned char)(actualSize / sizeof(unsigned char));
				if (error != noErr) PLstrcpy(titleText, EMPTY_PSTRING);
				Alert_SetTitle(box, titleText); // if the string is empty, the title is removed automatically
			}
			
			// set up help button parameter: defaults to false if the parameter is not specified
			{
				Boolean		isHelpButton = false;
				
				
				error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyAlertHasHelpButton,
										typeBoolean, &returnedType, &isHelpButton, sizeof(Boolean), &actualSize);
				Alert_SetHelpButton(box, isHelpButton);
			}
			
			Alert_SetParamsFor(box, kAlert_StyleOK);
			
			// set up type parameter: defaults to plain if the parameter is not specified
			{
				FourCharCode	type = kTelnetEnumeratedClassAlertIconTypeNote;
				AlertType		realType = kAlertPlainAlert;
				
				
				error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyAlertIcon, typeEnumerated,
										&returnedType, &type, sizeof(type), &actualSize);
				if (error == noErr)
				{
					switch (type)
					{
						case kTelnetEnumeratedClassAlertIconTypeNote:
							realType = kAlertNoteAlert;
							break;
						
						case kTelnetEnumeratedClassAlertIconTypeStop:
							realType = kAlertStopAlert;
							break;
						
						case kTelnetEnumeratedClassAlertIconTypeCaution:
							realType = kAlertCautionAlert;
							break;
						
						default:
							realType = kAlertPlainAlert;
							break;
					}
				}
				Alert_SetType(box, realType);
			}
			
			// set up movability parameter: defaults to false if the parameter is not specified
			{
				Boolean		isMovable = true;
				
				
				error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyAlertHasTitleBar,
										typeBoolean, &returnedType, &isMovable, sizeof(Boolean), &actualSize);
				Alert_SetMovable(box, isMovable);
			}
			
			// set up beep parameter just prior to displaying the box: defaults to false if not specified
			{
				Boolean		shouldBeep = false;
				
				
				error = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyAlertBeeps,
										typeBoolean, &returnedType, &shouldBeep, sizeof(Boolean), &actualSize);
				if (shouldBeep) BEEP;
			}
			
			// set up buttons: defaults to single OK button if not specified
			{
				AEDescList		buttonNames = { typeNull, null };
				
				
				error = AEGetParamDesc(inAppleEventPtr, kTelnetParameterKeyAlertButtons,
										typeAEList, &buttonNames);
				if (error == noErr)
				{
					SInt32		numberOfButtons = 0L;
					
					
					error = AECountItems(&buttonNames, &numberOfButtons);
					if (error == noErr)
					{
						register SInt32		i = 0L;
						AEKeyword			returnedKeyword;
						DescType			returnedType;
						Size				actualSize = 0L;
						SInt16				whichButton = kAlertStdAlertOKButton;
						StringPtr			whichString = null;
						
						
						if (numberOfButtons > 3) numberOfButtons = 3; // this is the limit, what can I say?
						for (i = 1; i <= numberOfButtons; i++)
						{
							// assign the appropriate button the text from the list,
							// in the order given; all button titles are saved, so
							// that the item hit can be returned, later, as a result
							switch (i)
							{
								case 3:
									whichButton = kAlertStdAlertOtherButton;
									whichString = button3Name;
									break;
								
								case 2:
									whichButton = kAlertStdAlertCancelButton;
									whichString = button2Name;
									break;
								
								case 1:
								default:
									whichButton = kAlertStdAlertOKButton;
									whichString = button1Name;
									break;
							}
							error = AEGetNthPtr(&buttonNames, i, cString, &returnedKeyword,
												&returnedType, (whichString + 1),
												(31) * sizeof(unsigned char), // size of button strings
												&actualSize);
							if (error == noErr)
							{
								whichString[0] = (unsigned char)(actualSize / sizeof(unsigned char));
								
								// from an AppleScript perspective, an empty string means
								// skip middle button; however, the Alert module expects
								// this to be requested by passing "null" as the string value
								if (i == 2) if (!PLstrcmp(whichString, EMPTY_PSTRING)) whichString = null;
								
								Alert_SetButtonText(box, whichButton, whichString);
							}
						}
					}
				}
			}
			
			result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
			if (result == noErr)
			{
				ModalFilterUPP		alertEventUPP = NewModalFilterUPP(StandardAlertEventFilter);
				
				
				Alert_SetFilterProc(box, alertEventUPP);			
				Alert_Display(box);
				DisposeModalFilterUPP(alertEventUPP), alertEventUPP = null;
				
				// check to see if there is a place to put the result - if so, return the button hit
				if (AppleEventUtilities_IsValidReply(outReplyAppleEventPtr))
				{
					SInt16			itemHit = Alert_ItemHit(box);
					TelnetToken		token;
					AEDesc			objectDescriptor;
					
					
					// initialize defaults first
					Memory_WriteZero(&token, sizeof(token)); // temp
					token.flags = 0;
					token.genericPointer = null;
					PLstrcpy(token.as.object.data.dialogReply.buttonName, EMPTY_PSTRING);
					token.as.object.data.dialogReply.gaveUp = false;
					
					// return "?" for the Help button, an empty string if the alert times out, or
					// the name of the button hit otherwise
					if (itemHit == kAlertStdAlertHelpButton)
					{
						PLstrcpy(token.as.object.data.dialogReply.buttonName, "\p?");
					}
					else if (itemHit == kAlertStdAlertOKButton)
					{
						PLstrcpy(token.as.object.data.dialogReply.buttonName, button1Name);
					}
					else if (itemHit == kAlertStdAlertCancelButton)
					{
						PLstrcpy(token.as.object.data.dialogReply.buttonName, button2Name);
					}
					else if (itemHit == kAlertStdAlertOtherButton)
					{
						PLstrcpy(token.as.object.data.dialogReply.buttonName, button3Name);
					}
					else
					{
						token.as.object.data.dialogReply.gaveUp = true;
					}
					error = GenesisAE_CreateTokenFromObjectData(kTelnetClassDialogReply, &token,
																&objectDescriptor, null/* new type */);
					if (error == noErr)
					{
						error = AEPutParamDesc(outReplyAppleEventPtr, keyDirectObject, &objectDescriptor);
					}
				}
			}
			
			Alert_Dispose(&box);
		}
	}
	
	return result;
}// HandleAlert


/*!
Handles the following event types:

kTelnetEventClass
-----------------
	kTelnetEventIDCopyToClipboard

(3.0)
*/
pascal OSErr
AppleEvents_HandleCopySelectedText	(AppleEvent const*	inAppleEventPtr,
									 AppleEvent*		UNUSED_ARGUMENT(outReplyAppleEventPtr),
									 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = noErr;
	AEDesc		resultDescriptor = { typeNull, null };
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: copy selected text event");
	
	// temp - somehow, this call produces an error, because this event specifies it has no direct object?
	//result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
	Console_WriteValue("required parameters error", result);
	if (result == noErr)
	{
		Size		actualSize = 0L;
		DescType	returnedType;
		Boolean		tableOption = false;
		
		
		// if the optional parameter was given as to whether to copy
		// using tabs for spaces, retrieve it
		result = AEGetParamPtr(inAppleEventPtr, kTelnetParameterKeyCopyToClipboardTable,
								typeBoolean, &returnedType, &tableOption, sizeof(tableOption), &actualSize);
		Console_WriteValue("get parameter error", result);
		if (result == errAEDescNotFound) result = noErr;
		if (result == noErr)
		{
			// copy the selected text from the frontmost window
			WindowRef		frontWindow = GetRealFrontWindow();
			
			
			if (ScreenManager_IsWindowAScreen(frontWindow))
			{
				TelnetScreenID		screenID = ScreenManager_GetWindowScreenID(frontWindow);
				
				
				Console_WriteValue("copying (apparently) for screen ID", screenID);
				Clipboard_TextToScrap(screenID, (tableOption) ? kTelnetCopyMethodTable
																: kTelnetCopyMethodStandard);
			}
			else
			{
				// error
			}
		}
	}
	
	Console_EndFunction();
	
	return result;
}// HandleCopySelectedText


/*!
Handles the following event types:

kTelnetEventClass
-----------------
	kTelnetEventIDLaunchFind

(3.0)
*/
pascal OSErr
AppleEvents_HandleLaunchFind	(AppleEvent const*		UNUSED_ARGUMENT(inAppleEventPtr),
								 AppleEvent*			UNUSED_ARGUMENT(outReplyAppleEventPtr),
								 UInt32					UNUSED_ARGUMENT(inData))
{
	return Commands_LaunchFind();
}// HandleLaunchFind


//
// internal methods
//

/*!
Handles quitting.  If any connections are active,
the user is warned that they will be killed if he
or she quits without properly closing them.

If the application shutdown occurs successfully,
"noErr" is returned; otherwise, "userCanceledErr"
is returned.

WARNING:	This routine expects to handle a
			quit application Apple Event; you
			should not use ExitToShell() within
			this routines implementation.

(3.0)
*/
OSErr
handleQuit	(DescType const		inSaveOption)
{
	register short		i = 0;
	OSErr				result = noErr;
	
	
	// if there are any live connections, confirm quitting (nice if the user accidentally hit command-Q)
	if ((inSaveOption == kAEAsk) && (Connections_GetActiveCount() > 0))
	{
		ConnectionDataPtr		screen = null;
		InterfaceLibAlertRef	box = null;
		Str255					errorString;
		ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
		
		
		GetIndString(errorString, rStringsCautionAlerts, siConnectionsAreOpenQuitAnyway);
		
		box = Alert_New();
		Alert_SetFilterProc(box, upp);
		Alert_SetHelpButton(box, false);
		Alert_SetParamsFor(box, kAlert_StyleOKCancel);
		Alert_SetText(box, errorString, EMPTY_PSTRING);
		Alert_SetType(box, kAlertCautionAlert);
		Alert_Display(box);
		
		if (Alert_ItemHit(box) == kAlertStdAlertOKButton) result = noErr;
		else result = userCanceledErr;
		
		Alert_Dispose(&box);
		DisposeModalFilterUPP(upp), upp = null;
	}
	
	if (result == noErr) Main_QuitSignal(); // only main() actually causes the application to shut down
	
	return result;
}// handleQuit