/*###############################################################

	TextServicesAE.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <string.h>

// Mac includes
#include <AEDataModel.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <MacErrors.h>
#include <Processes.h>
#include <TextServices.h>

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "InstallAE.h"
#include "Terminology.h"
#include "TextServicesAE.h"



//
// public methods
//

/**
Handles the following event types:

kTextServiceClass
-----------------
	kUpdateActiveInputArea

The document reference constant is the
window pointer of the document window.

(3.0)
*/
pascal OSErr
TextServicesAE_HandleUpdateActiveInputArea		(const AppleEvent*		inAppleEventPtr,
												 AppleEvent*			outReplyAppleEventPtr,
												 UInt32					inData)
{
	DescType	returnedType = typeWildCard;
	Size		actualSize = 0L;
	OSErr		result = errAEEventNotHandled;
	
	
	// handle the event
	{
		OSErr					error = noErr,
								resultForComponent = noErr;		// AppleEvent result
		WindowPtr				inputWindow = null;
		ComponentInstance		textServicesComponent = null;
		AEDesc					textBufferDesc = { typeChar, null },
								hiliteRangeDesc = { typeTextRangeArray, null },
								updateRangeDesc = { typeTextRangeArray, null };
		ScriptLanguageRecord	languageRecord;
		SInt32					fixLengthInfo = 0L; // length of text, or -1 (replace), or 0 (none confirmed)
		TextRange				scrollingInfo;
		OffsetArray				wordBoundaryInfo;
		Boolean					gotHighlightingInfo = false,	// got it flags for the optional parameters
								gotUpdatingInfo = false,
								gotScrollingInfo = false,
								gotWordBoundaryInfo = false;
		
		
		//
		// get required parameters
		//
		
		if (error == noErr)
		{
			// the TSM reference constant is actually a window pointer (see "RealScreenInterface.c")
			error = AEGetParamPtr(inAppleEventPtr, keyAETSMDocumentRefcon, typeSInt32,
									&returnedType, (SInt32*)&inputWindow, sizeof(inputWindow), &actualSize);
		}
		
		if (error == noErr)
		{
			// get the Text Services component that is making the request
			error = AEGetParamPtr(inAppleEventPtr, keyAEServerInstance, typeComponentInstance,
									&returnedType, &textServicesComponent, sizeof(textServicesComponent),
									&actualSize);
		}
		
		if (error == noErr)
		{
			// get the pending text (other parameters specify what to do with it); use an AppleEvent
			// descriptor so that the entire data is more likely to fit in the heap space
			error = AEGetParamDesc(inAppleEventPtr, keyAETheData, typeChar, &textBufferDesc);
		}
		
		if (error == noErr)
		{
			// get script and language information
			error = AEGetParamPtr(inAppleEventPtr, keyAEScriptTag, typeIntlWritingCode,
									&returnedType, &languageRecord, sizeof(languageRecord), &actualSize);
		}
		
		if (error == noErr)
		{
			// get fix-length info
			error = AEGetParamPtr(inAppleEventPtr, keyAEFixLength, typeSInt32,
									&returnedType, &fixLengthInfo, sizeof(fixLengthInfo), &actualSize);
		}
		
		//
		// get optional parameters
		//
		
		if (error == noErr)
		{
			// get highlighting information; use an AppleEvent descriptor so that the entire data is
			// more likely to fit in the heap space
			error = AEGetParamDesc(inAppleEventPtr, keyAEHiliteRange, typeTextRangeArray, &hiliteRangeDesc);
			gotHighlightingInfo = (error == noErr);
		}
		
		if (error == noErr)
		{
			// get updating information (the "fHiliteStyle" field of the structure is to be ignored here);
			// use an AppleEvent descriptor so that the entire data is more likely to fit in the heap space
			error = AEGetParamDesc(inAppleEventPtr, keyAEUpdateRange, typeTextRangeArray, &updateRangeDesc);
			gotUpdatingInfo = (error == noErr);
		}
		
		if (error == noErr)
		{
			// get auto-scroll information (the "fHiliteStyle" field of the structure is to be ignored here)
			error = AEGetParamPtr(inAppleEventPtr, keyAEPinRange, typeTextRange,
									&returnedType, &scrollingInfo, sizeof(scrollingInfo), &actualSize);
			gotScrollingInfo = (error == noErr);
		}
		
		if (error == noErr)
		{
			// get the physical offsets of words or clauses (useful for implementing word selection)
			error = AEGetParamPtr(inAppleEventPtr, keyAEClauseOffsets, typeOffsetArray,
									&returnedType, &wordBoundaryInfo, sizeof(wordBoundaryInfo), &actualSize);
			gotWordBoundaryInfo = (error == noErr);
		}
		
		//
		// now respond completely to the event
		//
		
		if (error == noErr)
		{
			result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
			if (result == noErr)
			{
				// check to see if there is a place to put the result - if so, return the error number
				if (AppleEventUtilities_IsValidReply(outReplyAppleEventPtr))
				{
					error = AEPutParamPtr(outReplyAppleEventPtr, keyErrorNumber, typeSInt16,
											&resultForComponent, sizeof(resultForComponent));
				}
			}
		}
		
		// if any error occurred along the way, everything else would have failed; return that error
		if (error != noErr) result = error;
		
		// clean up
		if (textBufferDesc.dataHandle != null) (OSErr)AEDisposeDesc(&textBufferDesc);
		if (hiliteRangeDesc.dataHandle != null) (OSErr)AEDisposeDesc(&hiliteRangeDesc);
		if (updateRangeDesc.dataHandle != null) (OSErr)AEDisposeDesc(&updateRangeDesc);
	}
	
	return result;
}// HandleUpdateActiveInputArea


/**
Handles the following event types:

kTextServiceClass
-----------------
	kPos2Offset

The document reference constant is the
window pointer of the document window.

(3.0)
*/
pascal OSErr
TextServicesAE_HandlePositionToOffset		(const AppleEvent*		inAppleEventPtr,
											 AppleEvent*			outReplyAppleEventPtr,
											 UInt32					inData)
{
	OSErr		result = errAEEventNotHandled;
	
	
	// unimplemented
	result = errAEEventNotHandled;
	
	return result;
}// HandlePositionToOffset


/**
Handles the following event types:

kTextServiceClass
-----------------
	kOffset2Pos

The document reference constant is the
window pointer of the document window.

(3.0)
*/
pascal OSErr
TextServicesAE_HandleOffsetToPosition		(const AppleEvent*		inAppleEventPtr,
											 AppleEvent*			outReplyAppleEventPtr,
											 UInt32					inData)
{
	OSErr		result = errAEEventNotHandled;
	
	
	// unimplemented
	result = errAEEventNotHandled;
	
	return result;
}// HandleOffsetToPosition


/**
Handles the following event types:

kTextServiceClass
-----------------
	kShowHideInputWindow

The document reference constant is the
window pointer of the document window.

(3.0)
*/
pascal OSErr
TextServicesAE_HandleShowHideInputWindow	(const AppleEvent*		inAppleEventPtr,
											 AppleEvent*			outReplyAppleEventPtr,
											 UInt32					inData)
{
	OSErr		result = errAEEventNotHandled;
	
	
	// Telnet does not manage its own floating input window, so this event is never handled
	result = errAEEventNotHandled;
	
	return result;
}// HandleShowHideInputWindow


/**
Handles the following event types:

kTextServiceClass
-----------------
	kUnicodeNotFromInputMethod

The document reference constant is the
window pointer of the document window.

(3.0)
*/
pascal OSErr
TextServicesAE_HandleUnicodeNotFromInputWindow	(const AppleEvent*		inAppleEventPtr,
												 AppleEvent*			outReplyAppleEventPtr,
												 UInt32					inData)
{
	OSErr		result = errAEEventNotHandled;
	
	
	// unimplemented
	result = errAEEventNotHandled;
	
	return result;
}// HandleUnicodeNotFromInputWindow


/**
Handles the following event types:

kTextServiceClass
-----------------
	kGetSelectedText

The document reference constant is the
window pointer of the document window.

(3.0)
*/
pascal OSErr
TextServicesAE_HandleGetSelectedText	(const AppleEvent*		inAppleEventPtr,
										 AppleEvent*			outReplyAppleEventPtr,
										 UInt32					inData)
{
	OSErr		result = errAEEventNotHandled;
	
	
	// unimplemented
	result = errAEEventNotHandled;
	
	return result;
}// HandleGetSelectedText