/*###############################################################

	URLAccessAE.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <AEDataModel.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <MacErrors.h>

// library includes
#include "DataAccessLib.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "Console.h"
#include "Terminology.h"
#include "URL.h"
#include "URLAccessAE.h"



// internal methods
static OSErr	handleURL		(AEDesc const*		inFromWhichObject,
								 AEDesc*			outResult);



//
// public methods
//

/*!
Handles the following event types:

kStandardURLEventClass
----------------------
	kStandardURLEventIDGetURL

(3.0)
*/
pascal OSErr
URLAccessAE_HandleUniformResourceLocator		(AppleEvent const*	inAppleEventPtr,
												 AppleEvent*		outReplyAppleEventPtr,
												 UInt32				UNUSED_ARGUMENT(inData))
{
	OSErr		result = noErr;
	AEDesc		directObject = { typeNull, null },
				replyDesc = { typeNull, null };
	
	
	Console_BeginFunction();
	Console_WriteLine("AppleScript: handle URL event");
	
	result = AEGetParamDesc(inAppleEventPtr, keyDirectObject, typeWildCard, &directObject);
	
	if (result == noErr)
	{
		DescType	returnedType = typeWildCard;
		
		
		// check for missing parameters
		result = AppleEventUtilities_RequiredParametersError(inAppleEventPtr);
		if (result == noErr) result = handleURL(&directObject, &replyDesc);
	}
	
	result = AppleEventUtilities_AddResultToReply(&replyDesc, outReplyAppleEventPtr, result);
	
	if (directObject.dataHandle) AEDisposeDesc(&directObject);
	
	Console_WriteValue("result", result);
	Console_EndFunction();
	return result;
}// HandleUniformResourceLocator


//
// internal methods
//

/*!
Handles the specified URL by either opening a telnet
session or launching the appropriate helper application.
If a list of URLs is given, this routine recursively
handles each URL in the list, returning a list of result
codes.  A result code of 0 is successful.

(3.0)
*/
static OSErr
handleURL	(AEDesc const*		inFromWhichObject,
			 AEDesc*			outResult)
{
	AEDesc		tokenDescriptor = { typeNull, null };
	OSErr		result = noErr;
	
	
	result = AppleEventUtilities_DuplicateOrResolveObject(inFromWhichObject, &tokenDescriptor);
	if (result == noErr)
	{
		switch (tokenDescriptor.descriptorType)
		{
			case typeAEList:
				result = AECreateList(null, 0, false/* is a record */, outResult);
				if (result == noErr)
				{
					AEDesc		listItemDescriptor = { typeNull, null },
								resultDescriptor = { typeNull, null };
					DescType	returnedType;
					long		index = 0L;
					
					
					result = AECountItems(&tokenDescriptor, &index);
					for (; index > 0; index--)
					{
						result = AEGetNthDesc(&tokenDescriptor, index, typeWildCard,
												&returnedType, &listItemDescriptor);
						
						// handle the request by adding the resultant data to the end of the list
						if (result == noErr) result = handleURL(&listItemDescriptor, &resultDescriptor);
						if (result == noErr) result = AEPutDesc(&tokenDescriptor/* list */, index,
																&resultDescriptor/* item to add */);
						
						if (listItemDescriptor.dataHandle != null) AEDisposeDesc(&listItemDescriptor);
						if (resultDescriptor.dataHandle != null) AEDisposeDesc(&resultDescriptor);
					}
					result = AEDuplicateDesc(&tokenDescriptor, outResult);
				}
				break;
			
			default:
				{
					Handle		handle = Memory_NewHandle(AEGetDescDataSize(&tokenDescriptor));
					SInt16		resultCode = 0;
					
					
					if (handle != null)
					{
						Size		actualSize = 0L;
						
						
						result = AppleEventUtilities_CopyDescriptorData(&tokenDescriptor, *handle,
																		GetHandleSize(handle), &actualSize);
						if (result == noErr) resultCode = !URL_ParseDataHandle(handle); // 0 is success
						else result = 1; // failure
						Memory_DisposeHandle(&handle);
					}
					else resultCode = 1; // failure
					result = AECreateDesc(typeSInt16, &resultCode, sizeof(resultCode), outResult);
				}
				break;
		}
	}
	
	if (tokenDescriptor.dataHandle != null) AEDisposeToken(&tokenDescriptor);
	
	return result;
}// handleURL