/*###############################################################

	EditLogInMessageDialog.c
	
	Telnet 3.0,  1998-1999 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1992 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <climits>
#else
#	include <limits.h>
#endif

// Mac includes
#include <ControlDefinitions.h>
#include <MacErrors.h>

// resource includes
#include "SpacingConstants.r"

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// Telnet includes
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "DialogResources.h"
#include "EditLogInMessageDialog.h"
#include "EventLoop.h"
#include "FolderManager.h"
#include "InterfaceLib.h"
#include "MacLook.h"
#include "Preferences.h"
#include "StringResources.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



// variables
static DialogPtr			gLogInMessageDialog = null; // the Mac OS dialog pointer
static WindowInfoRef		gLogInMessageDialogWindowInfo;

// internal methods
static pascal Boolean	editLogInMessageDialogEventFilter	(DialogPtr						inDialogPtr,
															 EventRecord*					inoutEventPtr,
															 short*							outItemIndex);
static pascal Boolean	handleNewSize						(const WindowPtr				inWindowPtr,
						 									 const SInt32					inDeltaSizeX,
															 const SInt32					inDeltaSizeY,
															 const SInt32					inData);



//
// public methods
//

/*!
This method is used to initialize the Format dialog
box.  It creates the dialog box invisibly, and uses
the specified data to set up the corresponding
controls in the dialog box.  Do not pass null for
"inSetupDataPtr".

The extended dialog resource for the preferences
dialog causes the Appearance Manager to automatically
construct an embedding hierarchy, so the dialogs
root control and container controls are supposed to
be set up once GetNewDialog() returns.

WARNING:	The pointer provided must remain valid
			until EditLogInMessageDialog_Done() is
			called.

(3.0)
*/
void
EditLogInMessageDialog_Init		(void)
{
	Cursors_DeferredUseWatch(30); // if it takes more than half a second to initialize, show the watch cursor
	
	gLogInMessageDialog = GetNewDialog(kDialogIDEditFTPLogInMessage, null, (WindowPtr)-1);
	gLogInMessageDialogWindowInfo = WindowInfo_New();
	
	if (gLogInMessageDialog != null)
	{
		ControlRef		control = null;
		Rect			contentRect,
						itemBounds;
		Handle			itemHandle = null;
		DialogItemType	itemType = 0;
		
		
		// get the size of the dialog box
		MacLook_GetWindowRegionBounds(GetDialogWindow(gLogInMessageDialog), kWindowContentRgn, &contentRect);
		
		// determine if the Help button should be disabled...
		SetUpHelpButtonDialogItem(gLogInMessageDialog, iFormatAppleGuideButton);
		
		// Telnet has a thread that takes care of this, so dont let the Dialog Manager do it
		SetDialogTracksCursor(gLogInMessageDialog, false);
		
		// adjust the OK and Cancel buttons so they are big enough for their titles
		SetDialogDefaultItem(gLogInMessageDialog, iFormatOKButton);
		SetDialogCancelItem(gLogInMessageDialog, iFormatCancelButton);
		(UInt16)Localization_AdjustDialogButtons(gLogInMessageDialog, true/* OK */, true/* Cancel */);
		Localization_AdjustHelpButtonItem(gLogInMessageDialog, iFormatAppleGuideButton);
		
		// there is no default item, because carriage returns are allowed in the text field
		SetDialogDefaultItem(gLogInMessageDialog, 0);
		
		// arrange the editable text control and the help text control
		{
			Str255		string;
			
			
			SizeDialogItem(gLogInMessageDialog, iEditFTPLogInMessageHelpText,
							contentRect.right - contentRect.left - INTEGER_DOUBLED(HSP_TEXT_AND_DIALOG), 0);
			(OSErr)GetDialogItemAsControl(gLogInMessageDialog, iEditFTPLogInMessageHelpText, &control);
			GetIndString(string, rStringsHelpText, siHelpTextEditFTPLogInMessage);
			(UInt16)Localization_SetUpMultiLineTextControl(control, string, kThemeSmallSystemFont);
			GetControlBounds(control, &itemBounds);
			
			// Although Localization_SetUpMultiLineTextControl() physically resized the
			// control, it used Control Manager routines, which means the Dialog Manager
			// has not been informed of the changes.  Therefore, it is *necessary* to
			// call SizeDialogItem() without really changing the size, for the sole
			// purpose of notifying the Dialog Manager.
			SizeDialogItem(gLogInMessageDialog, iEditFTPLogInMessageHelpText,
							itemBounds.right - itemBounds.left,
							itemBounds.bottom - itemBounds.top);
			MoveDialogItem(gLogInMessageDialog, iEditFTPLogInMessageHelpText, HSP_TEXT_AND_DIALOG,
							(contentRect.bottom - contentRect.top) - VSP_BUTTON_AND_DIALOG - BUTTON_HT -
								VSP_BUTTON_AND_CONTROL - (itemBounds.bottom - itemBounds.top));
			GetControlBounds(control, &itemBounds);
		}
		{
			ControlFontStyleRec		styleRecord;
			TerminalPrefs**			termPrefsHandle = null;
			SInt16					fontID = 0;
			
			
			// use the default terminals font for the log-in message editor, because the font is fixed-width
			termPrefsHandle = Preferences_GetDefaultTerminal();
			if (termPrefsHandle != null)
			{
				GetFNum((**termPrefsHandle).normalFont, &fontID);
				styleRecord.size = (**termPrefsHandle).fontSize;
				Memory_DisposeHandle((Handle*)&termPrefsHandle);
			}
			else
			{
				GetFNum("\pMonaco", &fontID);
				styleRecord.size = 12;
			}
			styleRecord.flags = kControlUseFontMask | kControlUseSizeMask;
			styleRecord.font = fontID;
			
			(OSErr)GetDialogItemAsControl(gLogInMessageDialog, iEditFTPLogInMessageMainTextField, &control);
			(OSErr)SetControlData(control, kControlNoPart, kControlFontStyleTag,
									sizeof(styleRecord), (Ptr)&styleRecord);
			
			// use the size of the help text control to help set up the text fields size
			(OSErr)GetDialogItemAsControl(gLogInMessageDialog, iEditFTPLogInMessageHelpText, &control);
			GetControlBounds(control, &itemBounds);
			SizeDialogItem(gLogInMessageDialog, iEditFTPLogInMessageMainTextField,
							contentRect.right - contentRect.left - INTEGER_DOUBLED(HSP_TEXT_AND_DIALOG) -
								INTEGER_DOUBLED(SPACE_EXTRA_TEXTEDIT),
							itemBounds.top - VSP_CONTROLS - VSP_TEXT_AND_DIALOG -
								INTEGER_DOUBLED(SPACE_EXTRA_TEXTEDIT));
			MoveDialogItem(gLogInMessageDialog, iEditFTPLogInMessageMainTextField,
							HSP_TEXT_AND_DIALOG + SPACE_EXTRA_TEXTEDIT,
							VSP_TEXT_AND_DIALOG + SPACE_EXTRA_TEXTEDIT);
		}
		
		// set the text field, and focus it
		{
			Handle					messageData = null;
			Size					actualSize = 0L;
			TelnetPrefsResultCode	error = kTelnetPrefsResultCodeSuccess;
			
			
			error = Preferences_GetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_ServerLogInMessage,
										sizeof(messageData), &messageData, &actualSize);
			if (error == kTelnetPrefsResultCodeSuccess)
			{
				// then a resource was found (otherwise, the user may simply never have set this preference before)
				(OSErr)GetDialogItemAsControl(gLogInMessageDialog, iEditFTPLogInMessageMainTextField, &control);
				(OSErr)SetControlData(control, kControlEditTextPart, kControlEditTextTextTag,
										GetHandleSize(messageData), *messageData);
				DrawOneControl(control);
			}
			(OSErr)SetKeyboardFocus(GetDialogWindow(gLogInMessageDialog), control, kControlFocusNextPart);
			
			Memory_DisposeHandle(&messageData);
		}
		
		// set up the Window Info information
		{
			WindowResizeResponderProcPtr	resizeProc = handleNewSize;
			
			
			WindowInfo_SetWindowDescriptor(gLogInMessageDialogWindowInfo, kTelnetDialogDescriptorEditLogInMessage);
			WindowInfo_SetWindowResizeLimits(gLogInMessageDialogWindowInfo,
													contentRect.bottom - contentRect.top/* minimum height */,
													contentRect.right - contentRect.left/* minimum width */,
													contentRect.bottom - contentRect.top/* maximum height */,
													contentRect.right - contentRect.left + 200/* maximum width */);
			WindowInfo_SetWindowResizeResponderUPP(gLogInMessageDialogWindowInfo, resizeProc, 0L);
		}
		WindowInfo_SetForDialog(gLogInMessageDialog, gLogInMessageDialogWindowInfo);
	}
}// Init


/*!
Call this method to destroy the log-in message dialog
box, and to save the users changes in the preferences
file.

(3.0)
*/
void
EditLogInMessageDialog_Done		(void)
{
	WindowInfo_Dispose(gLogInMessageDialogWindowInfo);
	DisposeDialog(gLogInMessageDialog), gLogInMessageDialog = null;
}// Done


/*!
This method displays and handles events in the
FTP Server: Edit Log-In Message dialog box.  If
the user clicks OK, "true" is returned; otherwise,
"false" is returned.

When the dialog displays itself on computers with
Appearance 1.1 or later, a zoom transition effect
is used.

(3.0)
*/
Boolean
EditLogInMessageDialog_Display	(void)
{
	Boolean		result = false;
	
	
	if (gLogInMessageDialog != null)
	{
		register DialogItemIndex	i = 0;
		DialogItemIndex				itemIndex = 0;
		ModalFilterUPP				filterProc = NewModalFilterUPP(editLogInMessageDialogEventFilter);
		
		
		// display the dialog
		TransitionDialogDisplay(gLogInMessageDialog);
		
		// handle events
		Menus_EnableEditMenu();
		itemIndex = 0;
		do
		{
			ModalDialog(filterProc, &itemIndex);
			switch (itemIndex)
			{
				case iFormatAppleGuideButton:
					{
						Str255		searchString;
						
						
						GetIndString(searchString, rStringsAppleGuideLookup, siAGIndexTermPreferences);
						AppleGuideUtilities_OpenFirstMainGuideInSearch(searchString);
					}
					break;
				
				default:
					break;
			}
		} while ((itemIndex != iFormatOKButton) && (itemIndex != iFormatCancelButton));
		DisposeModalFilterUPP(filterProc), filterProc = null;
		
		// handle OK or Cancel click
		if (itemIndex != iFormatCancelButton)
		{
			// save the new log-in message and kill the dialog
			{
				ControlRef				control = null;
				Handle					messageData = null;
				Size					dataSize = 0L;
				TelnetPrefsResultCode	error = kTelnetPrefsResultCodeSuccess;
				Boolean					isError = false;
				
				
				(OSErr)GetDialogItemAsControl(gLogInMessageDialog, iEditFTPLogInMessageMainTextField, &control);
				(OSErr)GetControlDataSize(control, kControlEditTextPart, kControlEditTextTextTag, &dataSize);
				messageData = Memory_NewHandle(dataSize);
				if (messageData != null)
				{
					(OSErr)GetControlData(control, kControlEditTextPart, kControlEditTextTextTag,
												GetHandleSize(messageData), *messageData, &dataSize);
					error = Preferences_SetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_ServerLogInMessage,
												sizeof(messageData), &messageData);
					if (error != kTelnetPrefsResultCodeSuccess) isError = true;
					Memory_DisposeHandle(&messageData);
				}
				else isError = true;
				
				if (isError) BEEP;
			}
			TransitionDialogClose(gLogInMessageDialog);
			result = true;
		}
		else
		{
			// user cancelled - close the dialog with an appropriate transition for cancelling
			TransitionDialogCloseCancel(gLogInMessageDialog);
			result = false;
		}
	}
	else Alert_ReportOSErr(memFullErr);
	
	return result;
}// Display


/*!
To set the text displayed in the log-in message
field of the visible dialog, use this method.
You might use this to handle a text drag-and-drop
into this dialog box.

(3.0)
*/
void
EditLogInMessageDialog_SetText	(void*			inDataPtr,
								 Size 			inSize)
{
	ControlRef		control = null;
	
	
	(OSErr)GetDialogItemAsControl(gLogInMessageDialog, iEditFTPLogInMessageMainTextField, &control);
	(OSErr)SetControlData(control, kControlEditTextPart, kControlEditTextTextTag, inSize, inDataPtr);
	DrawOneControl(control);
}// SetText


//
// internal methods
//

/*!
This method moves and resizes controls in response to
a resize of the Edit FTP Log-In Message movable modal
dialog box.

(3.0)
*/
static pascal Boolean
handleNewSize		(const WindowPtr		inWindowPtr,
					 const SInt32			inDeltaSizeX,
					 const SInt32			inDeltaSizeY,
					 const SInt32			UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_true;
	
	
	DialogAdjust_BeginDialogItemAdjustment(GetDialogFromWindow(inWindowPtr));
	{
		SInt32		amountH = 0L,
					amountV = 0L;
		
		
		if (inDeltaSizeX)
		{
			// controls which are moved horizontally
			if (Localization_IsLeftToRight())
			{
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iEditFTPLogInMessageOKButton, inDeltaSizeX);
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iEditFTPLogInMessageCancelButton, inDeltaSizeX);
			}
			else
			{
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iEditFTPLogInMessageAppleGuideButton, inDeltaSizeX);
			}
			
			// controls which are resized horizontally
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iEditFTPLogInMessageMainTextField, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iEditFTPLogInMessageHelpText, inDeltaSizeX);
		}
	}
	DialogAdjust_EndAdjustment(inDeltaSizeX, inDeltaSizeY); // moves and resizes controls properly
	
	return result;
}// handleNewSize


/*!
Handles special key events in the Format
dialog box.

(3.0)
*/
static pascal Boolean
editLogInMessageDialogEventFilter	(DialogPtr		inDialogPtr,
									 EventRecord*	inoutEventPtr,
									 short*			outItemIndex)
{
	Boolean		result = pascal_false,
				isButtonAndShouldFlash = false; // simplifies handling of multiple key-to-button maps
	
	
	// first, make sure help balloons show up properly for the cursor location, if Balloon Help is on
	if (IsPopUpHelpEnabled()) DisplayBalloonForDialogItemUnderMouse(inDialogPtr, inoutEventPtr->where,
																		kDialogBalloonsIDEditFTPLogInMessage);
	
	if (!result) result = (StandardDialogEventFilter(inDialogPtr, inoutEventPtr, outItemIndex));
	else if (isButtonAndShouldFlash) FlashButton(inDialogPtr, *outItemIndex);
	
	return result;
}// editLogInMessageDialogEventFilter