/*###############################################################

	AboutBox.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <limits.h>
#include <stdlib.h>

// Mac includes
#include <Appearance.h>
#include <ControlDefinitions.h>
#include <Controls.h>
#include <Fonts.h>
#include <Movies.h>
#include <QuickDraw.h>
#include <Resources.h>
#include <Script.h>
#include <TextUtils.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "GeneralResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "AboutBox.h"
#include "CoolStuff.h"
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "FolderManager.h"
#include "MacLook.h"
#include "ResourceFileManager.h"
#include "TelnetErrors.h"
#include "TelnetThreads.h"



// constants
enum
{
	kAboutBoxPageMarker = '^',		// this character in the 'TEXT' handle means new page
	kAboutBoxColumnMarker = '`',	// this character in the 'TEXT' handle means new column
	kAboutBoxColumnMaximumCount = 3	// no more than this many columns will fit
};

// types
typedef UInt32 AboutBoxState;
enum
{
	// All of these must be unique.  Using different states for
	// dissolving in and out allows different timers to be
	// attached to each effect.  So, you could make text appear
	// faster than it disappears, for example.
	kAboutBoxStateStart = 0,
	kAboutBoxStateInitialPause = 1,
	kAboutBoxStatePagePause = kAboutBoxStateInitialPause + 1,
	// The dissolve effect states must have consecutive values.
	kAboutBoxStateDissolveOutEffect1 = kAboutBoxStatePagePause + 1,
	kAboutBoxStateDissolveOutEffect2 = kAboutBoxStateDissolveOutEffect1 + 1,
	kAboutBoxStateDissolveOutEffect3 = kAboutBoxStateDissolveOutEffect1 + 2,
	kAboutBoxStateDissolveOutEffect4 = kAboutBoxStateDissolveOutEffect1 + 3,
	kAboutBoxStateDissolveOutEffect5 = kAboutBoxStateDissolveOutEffect1 + 4,
	kAboutBoxStateDissolveOutEffect6 = kAboutBoxStateDissolveOutEffect1 + 5,
	kLastAboutBoxDissolveOutState = kAboutBoxStateDissolveOutEffect6, // for convenience only
	kAboutBoxStateBlankPause = kLastAboutBoxDissolveOutState + 1,
	// The dissolve effect states must have consecutive values.
	kAboutBoxStateDissolveInEffect1 = kAboutBoxStateBlankPause + 1,
	kAboutBoxStateDissolveInEffect2 = kAboutBoxStateDissolveInEffect1 + 1,
	kAboutBoxStateDissolveInEffect3 = kAboutBoxStateDissolveInEffect1 + 2,
	kAboutBoxStateDissolveInEffect4 = kAboutBoxStateDissolveInEffect1 + 3,
	kAboutBoxStateDissolveInEffect5 = kAboutBoxStateDissolveInEffect1 + 4,
	kAboutBoxStateDissolveInEffect6 = kAboutBoxStateDissolveInEffect1 + 5,
	kLastAboutBoxDissolveInState = kAboutBoxStateDissolveInEffect6 // for convenience only
};

struct AboutBoxData
{
	WindowRef		contentWindow;
	
	struct
	{
		Movie			movie;				// music, only if the file About was found and QuickTime is installed
		SInt16			fileRefNum;			// the movie file, assumed to contain a sound track of some kind
		SInt16			volume;				// the current sound volume
		Boolean			fadingOut;			// is the volume in the process of fading down?
	} music;
	
	struct
	{
		Movie			movie;				// an animated opening, only if MacTelnetTitle.mov and QuickTime were found
		SInt16			fileRefNum;			// the movie file
	} title;
	
	AboutBoxState		state;				// how the dynamic text area gets drawn
	UInt32				time;				// time tracker; allows thread delays
	
	struct
	{
		struct
		{
			Handle			handle;			// the pages of text
			SInt32			offset;			// determines where to start reading text for the current page
			SInt32			size;			// the size of the current page, added to the offset
		} text;
		
		struct
		{
			SInt32			offset;			// determines where to start reading text for this column
			SInt32			size;			// size of this columns text, added to the page column offset
		} column[kAboutBoxColumnMaximumCount];
		
		UInt16			columnCount;		// at least one; less than or equal to "kAboutBoxColumnMaximumCount"
		UInt32			delayInTicks;		// time to wait before moving to the next page (changes); 0 means default time
	} page;
};
typedef struct AboutBoxData		AboutBoxData;
typedef AboutBoxData*			AboutBoxDataPtr;

// internal methods
static pascal void		aboutBoxDummyUserPaneIdle			(ControlRef			inControl);
static pascal void		aboutBoxPagedTextUserPaneDeviceLoop	(short				inColorDepth,
															 short				inDeviceFlags,
															 GDHandle			inTargetDevice,
															 long				inControl);
static pascal void		aboutBoxPagedTextUserPaneDraw		(ControlRef			inControl,
															 SInt16				inPart);
static pascal void		aboutBoxPagedTextUserPaneIdle		(ControlRef			inControl);
static void				drawText							(const Rect*		inTotalRectPtr,
															 AboutBoxDataPtr	inDataPtr);
static void				getNextPage							(AboutBoxDataPtr	inDataPtr);



//
// public methods
//

/*!
Displays the dialog box containing version
information and credits for this program.

In MacTelnet 3.0, suffice it to say, this
dialog is much cooler.  Users of Mac OS 8.5
and beyond get to see a cool window
transition when the dialog opens and when
it closes, not to mention a screen-sized
black backdrop with a hidden menu bar.
When this backdrop is in place, Mac OS 8.6
users (and anyone with the Draw Sprocket
extension installed) will see gamma-fading
effects!

Credits are displayed as pages, which can
be further sub-divided into columns.  A 'TEXT'
resource contains the text, where special
characters encode page and column breaks,
and custom delay-period information.

(3.0)
*/
void
AboutBox_Display	(void)
{
	DialogPtr		aboutDialog = GetNewDialog(kDialogIDAboutBox, 0L, kInFront);
	WindowRef		aboutWindow = null;
	
	
	aboutWindow = GetDialogWindow(aboutDialog);
	if (aboutDialog != null)
	{
		Boolean				menuBarIsSupposedToBeVisible = true;
		WindowRef			backdropWindow = null;
		WindowInfoRef		aboutWindowInfo = WindowInfo_New();
		ControlRef			starsControl = null,
							idleControl = null;
		AboutBoxDataPtr		dataPtr = (AboutBoxDataPtr)Memory_NewPtr(sizeof(AboutBoxData));
		
		
		// make the background black
		{
			RGBColor		black;
			
			
			black.red = black.green = black.blue = 0;
			(OSStatus)SetWindowContentColor(aboutWindow, &black);
			(OSStatus)SetThemeWindowBackground(aboutWindow, kThemeBrushBlack, true/* update */);
		}
		
		// set up the data structure, and initialize the text display
		if (dataPtr != null)
		{
			ControlRef		pagedTextPane = null;
			Rect			rect;
			OSErr			error = noErr;
			
			
			(OSErr)GetDialogItemAsControl(aboutDialog, iAboutPagedTextPane, &pagedTextPane);
			GetControlBounds(pagedTextPane, &rect);
			dataPtr->contentWindow = aboutWindow; // used by the idle procedure
			dataPtr->music.movie = null; // set up later
			dataPtr->music.fileRefNum = -1; // set up later
			dataPtr->music.volume = 0; // set up later
			dataPtr->music.fadingOut = false;
			dataPtr->title.movie = null; // set up later
			dataPtr->title.fileRefNum = -1; // set up later
			dataPtr->state = kAboutBoxStateStart;
			dataPtr->time = 0L; // re-initialized in the idle procedure, because of the state above
			dataPtr->page.text.handle = null;
			if (error == noErr)
			{
				SInt16		oldResFile = CurResFile();
				
				
				ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
				dataPtr->page.text.handle = Get1Resource('TEXT', kTextResourceIDAboutBox); // the text for the pages
				UseResFile(oldResFile);
			}
			dataPtr->page.text.offset = 0L; // start with first page
			dataPtr->page.text.size = 0L;
			dataPtr->page.columnCount = 1; // may change after analyzing the text handle
			dataPtr->page.delayInTicks = 0L;
			getNextPage(dataPtr); // get the page and column information
			
			// Set aboutBoxPagedTextUserPaneDraw() as the drawing routine for the
			// paged text pane, and install an idle procedure that handles state changes.
			{
				ControlUserPaneIdleUPP		idleUPP = NewControlUserPaneIdleUPP(aboutBoxPagedTextUserPaneIdle);
				ControlUserPaneDrawUPP		drawUPP = NewControlUserPaneDrawUPP(aboutBoxPagedTextUserPaneDraw);
				
				
				SetControlReference(pagedTextPane, REINTERPRET_CAST(dataPtr, SInt32)); // this data is required by the callback
				(OSErr)SetControlData(pagedTextPane, kControlNoPart, kControlUserPaneIdleProcTag,
										sizeof(ControlUserPaneIdleUPP), (Ptr)&idleUPP);
				(OSErr)SetControlData(pagedTextPane, kControlNoPart, kControlUserPaneDrawProcTag,
										sizeof(ControlUserPaneDrawUPP), (Ptr)&drawUPP);
			}
			WindowInfo_SetAuxiliaryDataPtr(aboutWindowInfo, pagedTextPane); // so the event filter can see this data
		}
		
		DeactivateFrontmostWindow();
		EventLoop_SetFloatersSuspended(true);
		
		// with the new window manager, display a black backdrop by creating a special window (I
		// recently found out that this works atrociously slowly under Classic on Mac OS X, however!)
		if (FlagManager_Test(kFlagCarbonWindowManager) && !FlagManager_Test(kFlagOSXClassic))
		{
			Rect		backdropBounds;
			OSStatus	error = 0L;
			RGBColor	black;
			
			
			black.red = black.green = black.blue = 0;
			
			if (FlagManager_Test(kFlagCarbonMenuManager))
			{
				menuBarIsSupposedToBeVisible = IsMenuBarVisible();
				HideMenuBar();
			}
			MacLook_GetWindowDeviceGrayRect(aboutWindow, &backdropBounds);
			error = noErr;
			InsetRect(&backdropBounds, -5, -5); // avoids potential glitches from extravagant window frames
			error = CreateNewWindow(kDocumentWindowClass, kWindowNoActivatesAttribute, &backdropBounds, &backdropWindow);
			if ((backdropWindow != null) && (error == noErr))
			{
				// create a dummy control to handle idle processing; it is attached to
				// the backdrop window so that the background music (if any) can
				// continue to fade away after the content window has been destroyed
				// (the background window lasts longer)
				idleControl = GetNewControl(kIDAnchorUserPane, backdropWindow);
				SetControlVisibility(idleControl, true/* visibility */, false/* draw */);
				{
					ControlUserPaneIdleUPP		idleUPP = NewControlUserPaneIdleUPP(aboutBoxDummyUserPaneIdle);
					
					
					SetControlReference(idleControl, REINTERPRET_CAST(dataPtr, SInt32)); // this data is required by the callback
					(OSErr)SetControlData(idleControl, kControlNoPart, kControlUserPaneIdleProcTag,
											sizeof(ControlUserPaneIdleUPP), (Ptr)&idleUPP);
				}
				
				// other setup
				error = SetWindowContentColor(backdropWindow, &black);
				if (error == noErr)
				{
					(OSStatus)TransitionWindow(backdropWindow, kWindowZoomTransitionEffect, kWindowShowTransitionAction, null);
					SelectWindow(backdropWindow);
					
					// set the gamma to zero (dark), so the about box can fade in more quickly
					CoolStuff_GammaBlack(CoolStuff_GetDefaultDrawSprocketContext()); // temp
					
					// create the coolest feature of all, the 3D star field
					{
						Rect		bounds;
						Point		center;
						
						
						MacLook_GetWindowRegionBounds(backdropWindow, kWindowContentRgn, &bounds);
						starsControl = NewControl(backdropWindow, &bounds, EMPTY_PSTRING, true/* visible */,
													kControlWantsIdle | kControlHasSpecialBackground/* feature bits */,
													0/* minimum */,
													0/* maximum */,
													kControlUserPaneProc,
													0L/* ref. con. */);
						CoolStuff_MovingStarsInit(starsControl);
						CoolStuff_MovingStarsGetCenter(starsControl, &center);
						center.h += 150; // arbitrary offset changes where the stars vanishing point is
						center.v += 100;
						CoolStuff_MovingStarsSetCenter(starsControl, center);
					}
				}
			}
		}
		
		// setting the window descriptor allows the dialog to be drawn by the event loop
		WindowInfo_SetWindowDescriptor(aboutWindowInfo, kTelnetWindowDescriptorAbout);
		WindowInfo_SetForDialog(aboutDialog, aboutWindowInfo);
		
		// look for About, and play its first sound track if its there (and obviously only if QuickTime is present);
		// while wasting time doing this, might as well look for MacTelnetTitle.mov and use it for a cool title, too...
		if ((dataPtr != null) && FlagManager_Test(kFlagQuickTime))
		{
			FSSpec		applicationFolder,
						musicMovieFile;
			OSErr		error = noErr;
			
			
			// determine the folder that looks like it contains MacTelnet (on Mac OS 8, this is
			// the applications folder; on Mac OS X, its the folder containing the application package);
			// also, start up QuickTime
			if ((FolderManager_GetFSSpec(kTelnetFolderPackageParent, &applicationFolder) == noErr) &&
				(EnterMovies() == noErr))
			{
				Str255		movieFileName;
				SInt16		movieActualID = 0;
				
				
				// create a file specification for the About file, which might be in the application folder
				GetIndString(movieFileName, rStringsImportantFileNames, siImportantFileNameAboutBoxMovie);
				error = FSMakeFSSpec(applicationFolder.vRefNum, applicationFolder.parID, movieFileName, &musicMovieFile);
				if (error == noErr)
				{
					if (error == noErr) error = OpenMovieFile(&musicMovieFile, &dataPtr->music.fileRefNum, fsCurPerm);
					if (error == noErr)
					{
						// Note: It is possible, under low memory conditions, that complex movies will fail to load here.
						error = NewMovieFromFile(&dataPtr->music.movie, dataPtr->music.fileRefNum,
													&movieActualID/* on return, the resource ID */, (StringPtr)null/* name */,
													newMovieActive | newMovieDontAskUnresolvedDataRefs,
													(Boolean*)null/* was the movie changed? */);
						if (error == noErr) if (GetMoviesError() == noErr)
						{
							Rect		offscreenRect;
							
							
							// position the movie safely offscreen, because only audio components are of interest
							GetMovieBox(dataPtr->music.movie, &offscreenRect);
							OffsetRect(&offscreenRect, -1024, -1024);
							SetMovieBox(dataPtr->music.movie, &offscreenRect);
							
							// remember volume so it can be faded out later
							dataPtr->music.volume = GetMovieVolume(dataPtr->music.movie);
							
							// make any visible components draw in a window that is owned by the About box, for safety
							SetMovieGWorld(dataPtr->music.movie, GetWindowPort(aboutWindow), (GDHandle)null);
							error = GetMoviesError();
							if (error == noErr)
							{
								// play the movie (this will play the sound, but any picture components are offscreen)
								GoToBeginningOfMovie(dataPtr->music.movie);
								StartMovie(dataPtr->music.movie);
							}
						}
					}
				}
				
				// create a file specification for the MacTelnetTitle.mov file, which might be in the application folder;
				// this movie also relies on the dummy control to handle idle events...
				GetIndString(movieFileName, rStringsImportantFileNames, siImportantFileNameAboutBoxTitleMovie);
				error = FSMakeFSSpec(applicationFolder.vRefNum, applicationFolder.parID, movieFileName, &musicMovieFile);
				if ((error == noErr) && (idleControl != null))
				{
					if (error == noErr) error = OpenMovieFile(&musicMovieFile, &dataPtr->title.fileRefNum, fsCurPerm);
					if (error == noErr)
					{
						// Note: It is possible, under low memory conditions, that complex movies will fail to load here.
						error = NewMovieFromFile(&dataPtr->title.movie, dataPtr->title.fileRefNum,
													&movieActualID/* on return, the resource ID */, (StringPtr)null/* name */,
													newMovieActive | newMovieDontAskUnresolvedDataRefs,
													(Boolean*)null/* was the movie changed? */);
						if (error == noErr) if (GetMoviesError() == noErr)
						{
							ControlRef		playerControl = null;
							Rect			movieBounds;
							
							
							// position the movie in the same place that the picture would have
							// been in, as defined by the dialog item resource, and then hide
							// the picture that is normally there (because the movie is on top!)
							GetMovieBox(dataPtr->title.movie, &movieBounds);
							{
								ControlRef		pictureControl = null;
								Rect			bounds;
								
								
								GetDialogItemAsControl(aboutDialog, iAboutPicture, &pictureControl);
								GetControlBounds(pictureControl, &bounds);
								SetControlVisibility(pictureControl, false/* visibility */, false/* draw */);
								SetRect(&movieBounds, 0, 0, movieBounds.right - movieBounds.left,
																movieBounds.bottom - movieBounds.top);
								OffsetRect(&movieBounds, bounds.left, bounds.top);
							}
							SetMovieBox(dataPtr->title.movie, &movieBounds);
							
							// align the window so it plays movies as efficiently as possible
							AlignWindow(aboutWindow, false/* activate */, null/* alignment rectangle */, null/* record */);
							
							// make any visible components draw in a window that is owned by the About box, for safety
							SetMovieGWorld(dataPtr->title.movie, GetWindowPort(aboutWindow), (GDHandle)null);
							error = GetMoviesError();
							if (error == noErr)
							{
								// play the movie
								GoToBeginningOfMovie(dataPtr->title.movie);
								StartMovie(dataPtr->title.movie);
							}
						}
					}
				}
			}
		}
		
		// display the window
		ShowWindow(aboutWindow);
		SelectWindow(aboutWindow);
		SetDialogDefaultItem(aboutDialog, 1);
		BeginUpdate(GetDialogWindow(aboutDialog));
		DrawDialog(aboutDialog);
		EndUpdate(GetDialogWindow(aboutDialog));
		if (backdropWindow != null)
		{
			ObscureCursor();
			CoolStuff_GammaControlledFadeFromBlack(CoolStuff_GetDefaultDrawSprocketContext(),
													2/* percent increase */, 2/* ticks minimum between gamma changes */,
													true/* yield to threads */);
		}
		
		// wait for a key press or mouse click
		SetDialogDefaultItem(aboutDialog, 1);
		{
			ModalFilterUPP		upp = NewModalFilterUPP(AnyKeyDefaultDialogEventFilter);
			DialogItemIndex		itemHit = 0,
								defaultItem = GetDialogDefaultItem(aboutDialog);
			
			
			Cursors_UseArrow();
			ObscureCursor();
			do
			{
				ModalDialog(upp, &itemHit);
			} while (itemHit != defaultItem);
			DisposeModalFilterUPP(upp), upp = null;
		}
		
		// fade out audio (if any); this is handled by the idle routine
		dataPtr->music.fadingOut = true;
		
		// remove the backdrop and the about box
		if (backdropWindow != null)
		{
			if (FlagManager_Test(kFlagDrawSprocket))
			{
				if (EventLoop_IsCommandKeyDown())
				{
					// the user held down the right key - play a chime sound and
					// switch out the localization settings (kind of an Easter egg)
					SInt16		oldResFile = CurResFile();
					
					
					ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
					(OSErr)Sound_PlayAsyncStereo(EASTER_EGG_SOUND_RESOURCE_TYPE, kSoundResourceIDSecretDiscovered, null/* SoundInfoRef storage */);
					Localization_Init(kLocalization_FlagReadTextRightToLeft | kLocalization_FlagReadTextBottomToTop);
					UseResFile(oldResFile);
				}
				else
				{
					// nope - just play the normal sound
					(OSErr)Sound_PlayAsyncStereo('snd ', kSoundResourceIDAboutBoxDispose, null/* SoundInfoRef storage */);
				}
				
				// the about box first disappears in a flash of white, then the stars fade out
				CoolStuff_GammaControlledFadeToWhite(CoolStuff_GetDefaultDrawSprocketContext(),
														10/* percent decrease */,
														2/* ticks minimum between gamma changes */,
														true/* yield to threads */);
				
				// brief pause when the screen is all white
				Threads_DelayMinimumTicks(30);
				
				// the user cant see the window close because the monitor gamma is set to all white!
				HideWindow(aboutWindow);
				
				// fade the gamma back to normal
				CoolStuff_GammaControlledFadeFromWhite(CoolStuff_GetDefaultDrawSprocketContext(),
														15/* percent increase */,
														3/* ticks minimum between gamma changes */,
														true/* yield to threads */);
				
				if (FlagManager_Test(kFlagThreadManager))
				{
					// brief pause when only the stars remain (it makes the transition look cooler)
					if (!Threads_ClickStopDelayTicks(90))
					{
						CoolStuff_GammaControlledFadeToBlack(CoolStuff_GetDefaultDrawSprocketContext(),
																3/* percent decrease */,
																1/* ticks minimum between gamma changes */,
																true/* yield to threads */);
					}
					else
					{
						// if the user clicks the mouse, he or she is getting impatient, so skip the fade-out
						CoolStuff_GammaBlack(CoolStuff_GetDefaultDrawSprocketContext());
					}
				}
				else
				{
					CoolStuff_GammaBlack(CoolStuff_GetDefaultDrawSprocketContext());
				}
				
				if (backdropWindow != null)
				{
					CoolStuff_MovingStarsDone(starsControl);
					DisposeControl(starsControl);
				}
				
				HideWindow(backdropWindow); // so the stars wont be visible when the gamma is restored
				
				// fade to the desktop
				CoolStuff_GammaControlledFadeFromBlack(CoolStuff_GetDefaultDrawSprocketContext(),
														7/* percent increase */,
														1/* ticks minimum between gamma changes */,
														true/* yield to threads */);
			}
			else
			{
				// without DrawSprocket, the best effect that can be done is a window transition
				(OSStatus)TransitionWindow(backdropWindow, kWindowZoomTransitionEffect,
											kWindowHideTransitionAction, null);
			}
		}
		if (FlagManager_Test(kFlagCarbonMenuManager))
		{
			if (menuBarIsSupposedToBeVisible) ShowMenuBar();
		}
		
		// destroy any movie that was playing (this automatically stops it)
		if ((dataPtr != null) && FlagManager_Test(kFlagQuickTime))
		{
			if (dataPtr->music.movie != null)
			{
				DisposeMovie(dataPtr->music.movie), dataPtr->music.movie = null;
				CloseMovieFile(dataPtr->music.fileRefNum), dataPtr->music.fileRefNum = -1;
			}
			if (dataPtr->title.movie != null)
			{
				DisposeMovie(dataPtr->title.movie), dataPtr->title.movie = null;
				CloseMovieFile(dataPtr->title.fileRefNum), dataPtr->title.fileRefNum = -1;
			}
			ExitMovies();
		}
		
		// dispose of the dialog and associated data structures
		if (backdropWindow != null) DisposeWindow(backdropWindow), backdropWindow = null;
		DisposeDialog(aboutDialog);
		if (dataPtr != null)
		{
			ReleaseResource(dataPtr->page.text.handle), dataPtr->page.text.handle = null;
			Memory_DisposePtr((Ptr*)&dataPtr);
		}
		WindowInfo_Dispose(aboutWindowInfo);
		RestoreFrontmostWindow();
		EventLoop_SetFloatersSuspended(false);
		
		// some events were handled without the Event Manager, so make sure they dont take effect in other ways
		// (like accidentally interpreting mouse clicks as clicks on the desktop and switching out of MacTelnet)
		FlushEvents(mDownMask | keyDownMask, 0/* stop nothing */);
		
		DrawMenuBar();
	}
	else Alert_ReportOSErr(ResError());
}// Display


//
// internal methods
//

/*!
A standard ControlUserPaneIdleUPP, this
method manages miscellaneous maintenance
tasks for the About box (such as fading
out the volume of QuickTime movies).

The reference constant of the control is
assumed to be of type "AboutBoxDataPtr".
That structure is modified as necessary
to reflect any change of state.

(3.0)
*/
static pascal void
aboutBoxDummyUserPaneIdle	(ControlRef		inControl)
{
	AboutBoxDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inControl), AboutBoxDataPtr);
	
	
	if (dataPtr != null)
	{
		if (FlagManager_Test(kFlagQuickTime))
		{
			// service the title movie, if present and if necessary
			if (dataPtr->title.movie != null)
			{
				if (IsMovieDone(dataPtr->title.movie))
				{
					ControlRef		pictureControl = null;
					
					
					// restore the picture once the movie (animation) finishes
					GetDialogItemAsControl(GetDialogFromWindow(dataPtr->contentWindow), iAboutPicture, &pictureControl);
					SetControlVisibility(pictureControl, true/* visibility */, true/* draw */);
					
					// delete the movie from memory, since its no longer needed (this eliminates future idle cycles, too)
					DisposeMovie(dataPtr->title.movie), dataPtr->title.movie = null;
					CloseMovieFile(dataPtr->title.fileRefNum), dataPtr->title.fileRefNum = -1;
				}
				else
				{
					MoviesTask(dataPtr->title.movie, 16/* arbitrary maximum relinquished time, in ticks */);
				}
			}
			
			// check to see if the movie stopped, and loop it, if necessary;
			// also, if the volume fade-out has begun, decrement the volume
			// (with enough idle loops, this will sound like a fade-out)
			if (dataPtr->music.movie != null)
			{
				if (IsMovieDone(dataPtr->music.movie))
				{
					GoToBeginningOfMovie(dataPtr->music.movie);
					StartMovie(dataPtr->music.movie);
				}
				
				if (dataPtr->music.fadingOut)
				{
					enum
					{
						kFadeTicks = 2 // arbitrary timer
					};
					static long		ticks = 0L;
					
					
					if ((TickCount() - ticks) > kFadeTicks)
					{
						dataPtr->music.volume -= 8; // arbitrary decrement
						if (dataPtr->music.volume < 0) dataPtr->music.volume = 0;
						else SetMovieVolume(dataPtr->music.movie, dataPtr->music.volume);
					}
					ticks = TickCount();
				}
			}
		}
	}
}// aboutBoxDummyUserPaneIdle


/*!
This method, of standard DeviceLoopDrawingUPP form,
handles drawing of the text in the middle of the
About box correctly no matter what the device
characteristics are.

(3.0)
*/
static pascal void
aboutBoxPagedTextUserPaneDeviceLoop		(short				inColorDepth,
										 short				UNUSED_ARGUMENT(inDeviceFlags),
										 GDHandle			inTargetDevice,
										 long				inControl)
{
	AboutBoxDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(REINTERPRET_CAST(inControl, ControlRef)),
													AboutBoxDataPtr);
	ColorPenState		state;
	Rect				controlRect;
	Boolean				isColorDevice = ColorUtilities_IsColorDevice(inTargetDevice);
	
	
	GetControlBounds((ControlRef)inControl, &controlRect);
	
	ColorUtilities_GetColorAndPenState(&state);
	ColorUtilities_NormalizeColorAndPen();
	
	// paint the background and set up the text color
	{
		RGBColor	background,
					foreground;
		
		
		// white foreground, black background
		background.red = background.green = background.blue = 0;
		foreground.red = foreground.green = foreground.blue = RGBCOLOR_INTENSITY_MAX;
		
		// for grayscale fading, the text color must be altered by an equally-distributed chunk of the maximum intensity
		{
			SInt32		fadeDelta = RGBCOLOR_INTENSITY_MAX / kAboutBoxDissolvePatternCountPerSet,
						intensity = 0;
			
			
			switch (dataPtr->state)
			{
				// intentional fall-through
				case kAboutBoxStateDissolveInEffect6:
				case kAboutBoxStateDissolveOutEffect1:
					intensity += fadeDelta;
				case kAboutBoxStateDissolveInEffect5:
				case kAboutBoxStateDissolveOutEffect2:
					intensity += fadeDelta;
				case kAboutBoxStateDissolveInEffect4:
				case kAboutBoxStateDissolveOutEffect3:
					intensity += fadeDelta;
				case kAboutBoxStateDissolveInEffect3:
				case kAboutBoxStateDissolveOutEffect4:
					intensity += fadeDelta;
				case kAboutBoxStateDissolveInEffect2:
				case kAboutBoxStateDissolveOutEffect5:
					intensity += fadeDelta;
				case kAboutBoxStateDissolveInEffect1:
				case kAboutBoxStateDissolveOutEffect6:
					intensity += fadeDelta;
					foreground.red = foreground.green = foreground.blue = intensity;
					break;
				
				default:
					// do nothing if the state is not a drawing transition
					break;
			}
		}
		// paint the background and set up the text color
		RGBForeColor(&foreground);
		if (FlagManager_Test(kFlagAppearance1_1API))
		{
			GrafPtr		port = null;
			
			
			GetPort(&port);
			(OSStatus)SetThemeBackground(kThemeBrushBlack, inColorDepth, isColorDevice);
		}
		else
		{
			RGBBackColor(&background);
		}
		
		// dim the colors if the window is inactive
		unless (IsControlActive((ControlRef)inControl))
		{
			RGBColor		inactiveTextColor;
			
			
			GetInactiveColors(&inactiveTextColor, null/* background color - unused */);
			RGBForeColor(&inactiveTextColor);
		}
	}
	
	if (dataPtr != null)
	{
		SInt16		patternIndex = 0; // nonzero means paint pattern
		
		
		EraseRect(&controlRect);
		if ((dataPtr->state != kAboutBoxStateBlankPause) &&
			(dataPtr->state != kAboutBoxStateDissolveInEffect1) &&
			(dataPtr->state != kAboutBoxStateDissolveOutEffect6)) // dont draw the text when its supposed to be blank!
		{
			drawText(&controlRect, dataPtr);
		}
		
		// now add effects based on the state
		switch (dataPtr->state)
		{
			case kAboutBoxStateStart:
			case kAboutBoxStateInitialPause:
			case kAboutBoxStatePagePause:
			case kAboutBoxStateBlankPause:
				// no additional work is required for these states
				break;
			
			case kAboutBoxStateDissolveOutEffect1:
			case kAboutBoxStateDissolveOutEffect2:
			case kAboutBoxStateDissolveOutEffect3:
			case kAboutBoxStateDissolveOutEffect4:
			case kAboutBoxStateDissolveOutEffect5:
			case kAboutBoxStateDissolveOutEffect6:
				// If you change "kAboutBoxDissolvePatternCountPerSet" (in
				// "GeneralResources.h"), you have to change the number of
				// states here.
				patternIndex = (dataPtr->state - kAboutBoxStateDissolveOutEffect1 + 1); // dissolve to black
				break;
			
			case kAboutBoxStateDissolveInEffect1:
			case kAboutBoxStateDissolveInEffect2:
			case kAboutBoxStateDissolveInEffect3:
			case kAboutBoxStateDissolveInEffect4:
			case kAboutBoxStateDissolveInEffect5:
			case kAboutBoxStateDissolveInEffect6:
				// If you change "kAboutBoxDissolvePatternCountPerSet" (in
				// "GeneralResources.h"), you have to change the number of
				// states here.
				patternIndex = (kLastAboutBoxDissolveInState - dataPtr->state + 1); // dissolve to the text
				break;
			
			default:
				// ??? - unknown state, this wouldnt be a good sign...
				break;
		}
		
		// set the foreground to black
		if (FlagManager_Test(kFlagAppearance1_1API))
		{
			GrafPtr		port = null;
			
			
			GetPort(&port);
			(OSStatus)SetThemePen(kThemeBrushBlack, inColorDepth, isColorDevice);
		}
		else
		{
			RGBColor		black;
			
			
			black.red = black.green = black.blue = 0;
			RGBForeColor(&black);
		}
		
		if (patternIndex > 0)
		{
			// paint a pattern
			static SInt16	resourceID = kPatternResourceIDAboutBox1;
			Pattern			pattern;
			
			
			if (patternIndex == 1)
			{
				SInt16		resourceIDOffset = 0;
				
				
				// use a random dissolving pattern; choose it only at the start of each animation *sequence*
				SetQDGlobalsRandomSeed(TickCount());
				resourceIDOffset = INTEGER_ABSOLUTE(Random()) % kAboutBoxDissolvePatternSetCount;
				resourceIDOffset = INTEGER_ABSOLUTE(resourceIDOffset);
				resourceID = kPatternResourceIDAboutBox1 + resourceIDOffset;
			}
			
			// get the proper stage of the dissolving animation from the resource file, and paint it
			GetIndPattern(&pattern, resourceID, patternIndex);
			PenMode(srcOr); // draws only the black pixels in the pattern, leaving the rest of the backdrop unaltered
			PenPat(&pattern);
			PaintRect(&controlRect);
		}
		else Threads_YieldToAnother(); // effect balancing (non-drawing states dont update as fast, so they yield less)
	}
	else
	{
		// some kind of low memory problem - just paint to black (this should never happen)
		EraseRect(&controlRect);
	}
	
	ColorUtilities_SetColorAndPenState(&state);
}// aboutBoxPagedTextUserPaneDeviceLoop


/*!
A standard ControlUserPaneDrawUPP, this
method draws in the scrolling text area
of the About box.  Exactly what gets
drawn depends on the current state of
the text pane.

The reference constant of the control is
assumed to be of type "AboutBoxDataPtr".
This structure defines state information
needed to render the control contents.

(3.0)
*/
static pascal void
aboutBoxPagedTextUserPaneDraw	(ControlRef		inControl,
								 SInt16			UNUSED_ARGUMENT(inPartCode))
{
	DeviceLoopDrawingUPP		upp = NewDeviceLoopDrawingUPP(aboutBoxPagedTextUserPaneDeviceLoop);
	
	
	DeviceLoopClipAndDrawControl(inControl, upp);
	DisposeDeviceLoopDrawingUPP(upp), upp = null;
}// aboutBoxPagedTextUserPaneDraw


/*!
A standard ControlUserPaneIdleUPP, this
method manages state changes for the
scrolling text pane of the About box.

The reference constant of the control is
assumed to be of type "AboutBoxDataPtr".
That structure is modified as necessary
to reflect any change of state.

(3.0)
*/
static pascal void
aboutBoxPagedTextUserPaneIdle	(ControlRef		inControl)
{
	AboutBoxDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inControl), AboutBoxDataPtr);
	
	
	if (dataPtr != null)
	{
		enum
		{
			kDisplayInitialDelayInTicks = 120,		// arbitrary; this plus the page delay is the starting delay
			kDisplayPageDefaultDelayInTicks = 180,	// arbitrary; used *if* the delay isnt customized in the text contents
			kDisplayBlankDelayInTicks = 50,			// arbitrary; the time a blank is displayed before new text dissolves in
			kDisplayDissolveOutDelayInTicks = 5,	// arbitrary; the time between dissolving patterns of removed text
			kDisplayDissolveInDelayInTicks = 3		// arbitrary; the time between dissolving patterns of displayed text
		};
		UInt32		nextState = dataPtr->state; // by default, dont change the state
		
		
		// The paging text pane is a state machine.  It displays pages one
		// by one, with specific delays governing animation, display of
		// pages and display of blanks between pages.
		switch (dataPtr->state)
		{
			case kAboutBoxStateStart:
				dataPtr->time = TickCount(); // initialize the timer
				nextState = kAboutBoxStateInitialPause;
				break;
			
			case kAboutBoxStateInitialPause:
				if ((TickCount() - dataPtr->time) > kDisplayInitialDelayInTicks)
				{
					nextState = kAboutBoxStatePagePause;
					dataPtr->time = TickCount();
				}
				break;
			
			case kAboutBoxStatePagePause:
				{
					UInt32		delay = dataPtr->page.delayInTicks;
					
					
					if (!delay) delay = kDisplayPageDefaultDelayInTicks; // if 0, this means use the default
					if (delay > 600) delay = 600; // arbitrary; for a bizarre, random bug where the delay somehow gets huge
					if ((TickCount() - dataPtr->time) > delay)
					{
						nextState = kAboutBoxStateDissolveOutEffect1; // begin dissolving to black
						dataPtr->time = TickCount();
					}
				}
				break;
			
			case kAboutBoxStateDissolveOutEffect1:
			case kAboutBoxStateDissolveOutEffect2:
			case kAboutBoxStateDissolveOutEffect3:
			case kAboutBoxStateDissolveOutEffect4:
			case kAboutBoxStateDissolveOutEffect5:
			case kAboutBoxStateDissolveOutEffect6:
				if ((TickCount() - dataPtr->time) > kDisplayDissolveOutDelayInTicks)
				{
					nextState++; // these states are assumed to have consecutive values (why wouldnt they?!?)
					if (dataPtr->state == kLastAboutBoxDissolveOutState) nextState = kAboutBoxStateBlankPause;
					dataPtr->time = TickCount();
				}
				break;
			
			case kAboutBoxStateBlankPause:
				if ((TickCount() - dataPtr->time) > kDisplayBlankDelayInTicks)
				{
					getNextPage(dataPtr); // now start drawing a different page of text
					nextState = kAboutBoxStateDissolveInEffect1; // initiate dissolving to the new text
					dataPtr->time = TickCount();
				}
				break;
			
			case kAboutBoxStateDissolveInEffect1:
			case kAboutBoxStateDissolveInEffect2:
			case kAboutBoxStateDissolveInEffect3:
			case kAboutBoxStateDissolveInEffect4:
			case kAboutBoxStateDissolveInEffect5:
			case kAboutBoxStateDissolveInEffect6:
				if ((TickCount() - dataPtr->time) > kDisplayDissolveInDelayInTicks)
				{
					nextState++; // these states are assumed to have consecutive values (why wouldnt they?!?)
					if (dataPtr->state == kLastAboutBoxDissolveInState) nextState = kAboutBoxStatePagePause; // cycle...
					dataPtr->time = TickCount();
				}
				break;
			
			default:
				// ??? - unknown state, this wouldnt be a good sign...
				break;
		}
		
		// update the state if necessary, and redraw the control if it changed
		{
			Boolean		updateNeeded = ((nextState != dataPtr->state) && (nextState != kAboutBoxStateInitialPause));
			
			
			dataPtr->state = nextState;
			if (updateNeeded) MacLook_InvalidateControlRectangle(inControl);
		}
	}
}// aboutBoxPagedTextUserPaneIdle


/*!
To draw the text for the current panel of the
About box in the specified area of the current
port, use this method.  The text is automatically
drawn in the appropriate number of columns, based
on the information in the structure provided.

(3.0)
*/
static void
drawText		(const Rect*			inTotalRectPtr,
				 AboutBoxDataPtr		inDataPtr)
{
	if ((inDataPtr != null) && (inTotalRectPtr != null))
	{
		register SInt16		i = 0;
		register UInt16		width = (inTotalRectPtr->right - inTotalRectPtr->left) / inDataPtr->page.columnCount;
		Rect				textRect;
		GrafPortFontState	preservedFontState;
		
		
		// preserve graphics port settings
		Localization_GetPortFontState(&preservedFontState);
		TextFont(GetScriptVariable(GetScriptManagerVariable(smKeyScript), smScriptSysFond));
		TextSize(14);
		TextFace(normal);
		HLock(inDataPtr->page.text.handle);
		for (i = 0; i < inDataPtr->page.columnCount; i++)
		{
			// draw each column of text; the fewer the columns, the wider they are
			SetRect(&textRect, inTotalRectPtr->left + i * width, inTotalRectPtr->top, 0, inTotalRectPtr->bottom);
			SetRect(&textRect, textRect.left, textRect.top, textRect.left + width, textRect.bottom);
			TETextBox((*inDataPtr->page.text.handle) + inDataPtr->page.column[i].offset,
						inDataPtr->page.column[i].size, &textRect, teCenter);
		}
		HUnlock(inDataPtr->page.text.handle);
		Localization_SetPortFontState(&preservedFontState);
	}
}// drawText


/*!
To find the offset and size of the next page, use
this method.  The pages automatically cycle, so
if the offset loops, the offset to the first page
is used.  The specified structure is updated to
point to the new page.

(3.0)
*/
static void
getNextPage		(AboutBoxDataPtr		inDataPtr)
{
	if (inDataPtr != null)
	{
		register SInt32		i = 0L;
		Ptr					textPointer = null;
		Size				textTotalSize = GetHandleSize(inDataPtr->page.text.handle);
		SInt32				newOffset = 0L,
							newSize = 0L;
		char				parsedCharacter = '\0';
		
		
		HLock(inDataPtr->page.text.handle);
		textPointer = (*inDataPtr->page.text.handle);
		
		newSize = 0L; // this is determined later in this function
		
		newOffset = inDataPtr->page.text.offset + inDataPtr->page.text.size; // start at end of current page
		if (newOffset >= textTotalSize) newOffset = 0L; // wrap to the first page if the end of the buffer is hit
		if ((textPointer + newOffset)[0] == kAboutBoxPageMarker) newOffset++; // skip any page characters
		if (inDataPtr->page.delayInTicks) newOffset++; // skip any custom delay digits
		inDataPtr->page.column[0].offset = newOffset; // this is always true
		
		inDataPtr->page.columnCount = 1; // this is true initially...
		inDataPtr->page.delayInTicks = 0L; // this is true initially...
		
		// look for columns and pages at the same time
		for (i = newOffset; i < textTotalSize; i++)
		{
			parsedCharacter = (textPointer + i)[0];
			
			if ((parsedCharacter == kAboutBoxColumnMarker) ||
				(parsedCharacter == kAboutBoxPageMarker))
			{
				SInt16		previousColumnIndex = inDataPtr->page.columnCount - 1;
				
				
				// the size of the column up to this marker is just the current position minus the previous columns offset
				inDataPtr->page.column[previousColumnIndex].size = i - inDataPtr->page.column[previousColumnIndex].offset;
				
				if (parsedCharacter == kAboutBoxPageMarker)
				{
					newSize = i++ - newOffset;
					if (i < textTotalSize)
					{
						// look for a custom delay digit (in seconds), optional, which would appear after the page break
						char	s[1];
						
						
						s[0] = (textPointer + i)[0];
						if (isdigit(s[0])) inDataPtr->page.delayInTicks = (atoi(s) << 6); // * 64 is about * 60 (good enough)
					}
					break;
				}
				else if (inDataPtr->page.columnCount < kAboutBoxColumnMaximumCount)
				{
					// the next columns offset is the character right after this one (skip the column marker character)
					inDataPtr->page.column[inDataPtr->page.columnCount++].offset = ++i;
				}
			}
		}
		
		if (i >= textTotalSize) newSize = textTotalSize - newOffset; // boundary case
		
		HUnlock(inDataPtr->page.text.handle), textPointer = null;
		inDataPtr->page.text.offset = newOffset;
		inDataPtr->page.text.size = newSize;
	}
}// getNextPage