/*###############################################################

	Alert.c
	
	This module has been created to make it possible to build
	more powerful alert windows than even the Mac OS 8 Toolbox
	allows by default.
	
	Interface Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <ASRegistry.h>
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Gestalt.h>
#include <PLStringFuncs.h>
#include <QuickDraw.h>
#include <Resources.h>
#include <Sound.h>
#include <SpeechSynthesis.h>
#include <TextUtils.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "SpacingConstants.r"



// string indices
enum
{
	rInterfaceLibAlertButtonStrings = 32000,
	BUTTON_YES = 1,
	BUTTON_NO = 2,
	BUTTON_DONTSAVE = 3,
	BUTTON_SAVE = 4,
	BUTTON_OK = 5,
	BUTTON_CANCEL = 6,
	BUTTON_QUIT = 7,
	BUTTON_CONTINUE = 8
};
enum
{
	rInterfaceLibMiscellaneousStrings = 32001,
	siInterfaceLibReportOSStatus = 1,
	siInterfaceLibReportOSStatusHelpText = 2
};

// types
struct InterfaceLibAlert
{
	AlertStdAlertParamRec	params;
	AlertType				alertType;
	StringPtr				dialogText,
							helpText;
	Str255					title;
	Str31					button1Text, // not used unless custom button text is required
							button2Text, // not used unless custom button text is required
							button3Text; // not used unless custom button text is required
	WindowRef				sheetWindow;
	SInt16					itemHit;
};
typedef struct InterfaceLibAlert InterfaceLibAlert;
typedef InterfaceLibAlert* InterfaceLibAlertPtr;

// variables
static Boolean			gNotificationIsBackgrounded = false,
						gUseSpeech = false,
						gIsInited = false;
static UInt16			gNotificationPreferences = kAlert_NotifyDisplayDiamondMark;
static short			gApplicationResourceFile = 0;
static Str255			gNotificationMessage;
static NMRecPtr			gNotificationPtr = null;



// internal prototypes
static void						newButtonString				(Str255					outStringStorage,
															 short					inButtonStringType);

static InterfaceLibAlertPtr		refAcquireLock				(InterfaceLibAlertRef	inRef);

static void						refReleaseLock				(InterfaceLibAlertRef	inRef,
															 InterfaceLibAlertPtr*	inoutPtr);

static void						setAlertVisibility			(DialogRef const		inDialog,
															 Boolean const			inIsVisible,
															 Boolean const			inAnimate);

static void						setDialogItemControlText	(DialogRef const		inDialog,
															 DialogItemIndex const	inItemIndex,
															 StringPtr const		inText);

static void						setStdFilterProcForAlert	(InterfaceLibAlertRef	inAlert);

static OSErr					standardAlert				(InterfaceLibAlertPtr	inAlert,
															 AlertType				inAlertType,
															 StringPtr				inDialogText,
															 StringPtr				inHelpText);



//
// public methods
//

/*!
This routine should be called once, before any
other Alert routine.  The corresponding method
Alert_Done() should be invoked once at the end
of the program (most likely), after the Alert
module is not needed anymore.

The Alert module needs the reference number of
your applications resource file.  You usually
call Alert_Init() at startup time.

(1.0)
*/
void
Alert_Init	(short const		inApplicationResourceFile)
{
	gApplicationResourceFile = inApplicationResourceFile;
	gIsInited = true;
}// Init


/*!
This routine should be called once, after the
Alert module is no longer needed.

(1.0)
*/
void
Alert_Done	(void)
{
	Alert_ServiceNotification();
	gIsInited = false;
}// Done


/*!
Creates a new alert window, but does not display it
(this gives you the opportunity to specify its
characteristics using other methods from this
module).  When you are finished with the alert, use
the method Alert_Dispose() to destroy it.

Generally, you call this routine to create an alert,
then call other methods to set up the attributes of
the new alert (title, help button, dialog filter
procedure, etc.).  When the alert is configured, use
Alert_Display() or a similar method to display it.

By default, all new alerts are marked as movable,
with no help button, with a single OK button and no
other buttons.

IMPORTANT:	If anything goes wrong, "null" is returned.

(1.0)
*/
InterfaceLibAlertRef
Alert_New 	(void)
{
	InterfaceLibAlertPtr		result = (InterfaceLibAlertPtr)Memory_NewPtr(sizeof(InterfaceLibAlert));
	
	
	if (result != null)
	{
		result->params.movable = true;
		result->params.helpButton = false;
		result->params.filterProc = null;
		result->params.defaultText = (StringPtr)kAlertDefaultOKText;
		result->params.cancelText = (StringPtr)null;
		result->params.otherText = (StringPtr)null;
		result->params.defaultButton = kAlertStdAlertOKButton;
		result->params.cancelButton = 0;
		result->params.position = kWindowDefaultPosition;
		PLstrcpy(result->title, EMPTY_PSTRING);
		PLstrcpy(result->button1Text, EMPTY_PSTRING);
		PLstrcpy(result->button2Text, EMPTY_PSTRING);
		PLstrcpy(result->button3Text, EMPTY_PSTRING);
		result->sheetWindow = null;
		result->itemHit = kAlertStdAlertOtherButton;
	}
	return ((InterfaceLibAlertRef)result);
}// New


/*!
When you are finished with an alert that you
created with Alert_New(), call this routine
to destroy it.  The reference is automatically
set to null.

(1.0)
*/
void
Alert_Dispose	(InterfaceLibAlertRef*		inoutAlert)
{
	InterfaceLibAlertPtr	ptr = refAcquireLock(*inoutAlert);
	
	
	if (ptr != null) DisposePtr((Ptr)ptr), ptr = null;
	*inoutAlert = null;
}// Dispose


/*!
Specifes that an alert should be rendered as a
sheet, which is a Mac OS X concept for a
window that is physically attached to another
window.

IMPORTANT:	The Alert module reserves the right
			to ignore this request; you should
			use this function when appropriate,
			but expect that the resultant alert
			may not be attached to a window.

(1.0)
*/
void
Alert_AttachToWindow	(InterfaceLibAlertRef		inAlert,
						 WindowRef					inAttachToWhichWindowOrNull)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	
	
	if (alertPtr != null)
	{
		alertPtr->sheetWindow = inAttachToWhichWindowOrNull;
		refReleaseLock(inAlert, &alertPtr);
	}
}// AttachToWindow


/*!
Use this method to display a note alert with
an OK button, the message "Checkpoint" and
the given descriptive text.  This method will
alter alert settings, so subsequently displayed
alerts will default to having no help button
and a single OK button.

(1.0)
*/
void
Alert_DebugCheckpoint	(StringPtr		inDescription)
{
	InterfaceLibAlertRef	box = null;
	
	
	box = Alert_New(); // no help button, single OK button
	setStdFilterProcForAlert(box);
	Alert_SetTitle(box, "\pDebug Checkpoint");
	Alert_SetText(box, "\pCheckpoint", inDescription);
	Alert_SetType(box, kAlertNoteAlert);
	Alert_Display(box);
	Alert_Dispose(&box);
}// DebugCheckpoint


/*!
To send an Apple Event to the current process of
the specified class, setting a "keyDirectObject"
to "typeChar" data containing the information in
the given Pascal string, use this method.  This
is useful if your application supports AppleScript
and can put this string somewhere useful for
debugging purposes (such as in a log file).  It is
also useful for debugging shared libraries, since
they cannot otherwise send information to your
applications debugging systems.

This isnt really an alert routine, but it has
similar applications to alert debugging routines,
so it is listed here.

(1.0)
*/
void
Alert_DebugSendStringToSelf		(StringPtr				inSendWhat,
								 AEEventClass const		inEventClass,
								 AEEventID const		inEventID)
{
	AppleEvent				consoleWriteEvent,
							reply;
	AEAddressDesc			currentProcessAddress;
	ProcessSerialNumber		currentProcessID;
	
	
	currentProcessID.highLongOfPSN = 0;
	currentProcessID.lowLongOfPSN = kCurrentProcess; // dont use GetCurrentProcess()!
	(OSErr)AECreateDesc(typeProcessSerialNumber,
						(Ptr)&currentProcessID, sizeof(ProcessSerialNumber), &currentProcessAddress);
	(OSErr)AECreateAppleEvent(inEventClass, inEventID, &currentProcessAddress, kAutoGenerateReturnID, kAnyTransactionID, &consoleWriteEvent);
	(OSErr)AEPutParamPtr(&consoleWriteEvent, keyDirectObject, cString, inSendWhat + 1, StrLength(inSendWhat) * sizeof(unsigned char));
	(OSErr)AESend(&consoleWriteEvent, &reply, kAENoReply | kAENeverInteract | kAEDontRecord,
					kAENormalPriority, kAEDefaultTimeout, null, null);
	AEDisposeDesc(&consoleWriteEvent);
}// DebugSendStringToSelf


/*!
To display an alert box profiling the users
system, including OS version, use this method.

(1.0)
*/
void
Alert_DebugSystemProfile		(void)
{
	Str255					versionText;
	Str255					message;
	InterfaceLibAlertRef	box = null;
	
	
	{
		long		sysv = 0L;
		UInt8		number = 0;
		Str255		string;
		
		
		(OSErr)Gestalt(gestaltSystemVersion, &sysv);
		number = Releases_GetMajorRevisionForVersion(sysv);
		NumToString((long)number, string);
		PLstrcpy(versionText, string);
		number = Releases_GetMinorRevisionForVersion(sysv);
		NumToString((long)number, string);
		PLstrcat(versionText, "\p.");
		PLstrcat(versionText, string);
		number = Releases_GetSuperminorRevisionForVersion(sysv);
		NumToString((long)number, string);
		PLstrcat(versionText, "\p.");
		PLstrcat(versionText, string);
	}
	PLstrcpy(message, "\pYour system appears to be running Mac OS ");
	PLstrcat(message, versionText);
	PLstrcat(message, "\p.");
	box = Alert_New();
	setStdFilterProcForAlert(box);
	Alert_SetParamsFor(box, kAlert_StyleOK);
	Alert_SetTitle(box, "\pSystem Profile");
	Alert_SetButtonText(box, kAlertStdAlertOKButton, "\pContinue");
	Alert_SetText(box, message, EMPTY_PSTRING);
	Alert_SetType(box, kAlertStopAlert);
	Alert_Display(box);
	Alert_Dispose(&box);
}// DebugSystemProfile


/*!
Use this method to display a note alert with
the following message:
Warning: This feature is incomplete and
contains known bugs.  Use with caution.

(1.0)
*/
void
Alert_DebugWarning	(void)
{
	InterfaceLibAlertRef	box = null;
	
	
	box = Alert_New(); // no help button, single OK button
	setStdFilterProcForAlert(box);
	Alert_SetTitle(box, "\pFeature Incomplete");
	Alert_SetText(box, "\pWarning: This feature is incomplete and contains known bugs.", "\pUse with caution.");
	Alert_SetType(box, kAlertCautionAlert);
	Alert_Display(box);
	Alert_Dispose(&box);
}// DebugWarning


/*!
Use this method to display a note alert with
an OK button, and the specified description,
with the given files information (name,
parent directory ID and volume number)
underneath.

(1.0)
*/
void
Alert_DebugWhatIsFile	(StringPtr			inDescription,
						 FSSpec const*		inFilePtr)
{
	InterfaceLibAlertRef	box = null;
	Str31					numberText;
	Str255					fileInfoText;
	
	
	PLstrcpy(fileInfoText, "\p");
	PLstrcat(fileInfoText, inFilePtr->name);
	PLstrcat(fileInfoText, "\p, vol=");
	NumToString(inFilePtr->vRefNum, numberText);
	PLstrcat(fileInfoText, numberText);
	PLstrcat(fileInfoText, "\p, parentDir=");
	NumToString(inFilePtr->parID, numberText);
	PLstrcat(fileInfoText, numberText);
	PLstrcat(fileInfoText, "\p.");
	box = Alert_New(); // no help button, single OK button
	setStdFilterProcForAlert(box);
	Alert_SetTitle(box, "\pDebug: File");
	Alert_SetText(box, inDescription, fileInfoText);
	Alert_SetType(box, kAlertNoteAlert);
	Alert_Display(box);
	Alert_Dispose(&box);
}// DebugWhatIsFile


/*!
Use this method to display a note alert with
an OK button, and the specified description,
with the given number underneath.

(1.0)
*/
void
Alert_DebugWhatIsNumber		(StringPtr		inDescription,
							 SInt32 const	inNumber)
{
	InterfaceLibAlertRef	box = null;
	Str31					numberText;
	
	
	NumToString(inNumber, numberText);
	box = Alert_New(); // no help button, single OK button
	setStdFilterProcForAlert(box);
	Alert_SetTitle(box, "\pDebug: Numerical Value");
	Alert_SetText(box, inDescription, numberText);
	Alert_SetType(box, kAlertNoteAlert);
	Alert_Display(box);
	Alert_Dispose(&box);
}// DebugWhatIsNumber


/*!
Use this method to display a note alert with
an OK button, and the specified description,
with the given coordinates shown underneath.

(1.0)
*/
void
Alert_DebugWhatIsPoint		(StringPtr		inDescription,
							 Point const	inPoint)
{
	InterfaceLibAlertRef	box = null;
	Str15					xText,
							yText;
	Str31					coordinatesText;
	
	
	NumToString(inPoint.h, xText);
	NumToString(inPoint.v, yText);
	PLstrcpy(coordinatesText, "\p(%a, %b)");
	StringUtilities_PMetaSetDefaults();
	StringUtilities_PBuild(coordinatesText, xText, yText, EMPTY_PSTRING, EMPTY_PSTRING);
	box = Alert_New(); // no help button, single OK button
	setStdFilterProcForAlert(box);
	Alert_SetTitle(box, "\pDebug: Coordinates");
	Alert_SetText(box, inDescription, coordinatesText);
	Alert_SetType(box, kAlertNoteAlert);
	Alert_Display(box);
	Alert_Dispose(&box);
}// DebugWhatIsPoint


/*!
Use this method to display a note alert with
an OK button, and the specified description,
with the given rectangle shown underneath
(in top, left, bottom, right order).

(1.0)
*/
void
Alert_DebugWhatIsRectangle	(StringPtr		inDescription,
							 Rect const*	inRectPtr)
{
	InterfaceLibAlertRef	box = null;
	Str15					tText,
							lText,
							bText,
							rText;
	Str255					rectText;
	
	
	NumToString(inRectPtr->top, tText);
	NumToString(inRectPtr->left, lText);
	NumToString(inRectPtr->bottom, bText);
	NumToString(inRectPtr->right, rText);
	PLstrcpy(rectText, "\p(T=%a, L=%b, B=%c, R=%d)");
	StringUtilities_PMetaSetDefaults();
	StringUtilities_PBuild(rectText, tText, lText, bText, rText);
	box = Alert_New(); // no help button, single OK button
	setStdFilterProcForAlert(box);
	Alert_SetTitle(box, "\pDebug: Rectangle");
	Alert_SetText(box, inDescription, rectText);
	Alert_SetType(box, kAlertNoteAlert);
	Alert_Display(box);
	Alert_Dispose(&box);
}// DebugWhatIsRectangle


/*!
Use this method to display an alert described by
an Interface Library Alert object.  For alerts
with more than one button, the item hit can be
found as a standard alert button number
(kAlertStdAlertOKButton, etc.) by invoking the
Alert_ItemHit() method.

If the program is in the background (as specified
by a call to Alert_SetIsBackgrounded()), a
notification is posted.  The alert is displayed
immediately, but the notification lets the user
know that its there.  Your application must call
Alert_ServiceNotification() whenever a resume
event is detected in your event loop, in order to
service notifications posted by the Alert module.
The Alert module assumes that it will find a
small icon ('SICN') resource with an ID number of
"kAlert_NotificationSICNResourceID" that it can
use as a Notification Manager icon in the process
menu.

Another Alert module alert box can be visible
while this method is invoked, and all displayed
alerts will still be displayed properly.  This is
better than the Mac OS default implementation - a
bug in the StandardAlert() implementation prevents
simultaneous display of alerts, because it causes
all displayed alerts to show the same alert text.

You must have called Alert_Init() properly to
display any kind of alert.  If this module has not
been initialized, "notInitErr" is returned.

(1.0)
*/
OSErr
Alert_Display	(InterfaceLibAlertRef	inAlert)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	OSErr					result = noErr;
	
	
	// only attempt to display an alert if the module was properly initialized
	if (gIsInited)
	{
		if ((gNotificationIsBackgrounded) && (gNotificationPtr == null))
		{
			gNotificationPtr = (NMRecPtr)Memory_NewPtr(sizeof(NMRec));
			if (gNotificationPtr != null)
			{
				short		oldResourceFile = CurResFile();
				
				
				gNotificationPtr->qLink = null;
				gNotificationPtr->qType = nmType;
				gNotificationPtr->nmMark = (gNotificationPreferences < kAlert_NotifyDisplayDiamondMark)
												? 0
												: diamondMark;
				UseResFile(gApplicationResourceFile);
				gNotificationPtr->nmIcon = (gNotificationPreferences < kAlert_NotifyDisplayIconAndDiamondMark)
												? null
												: GetResource('SICN', kAlert_NotificationSICNResourceID);
				gNotificationPtr->nmSound = null;
				gNotificationPtr->nmStr = (gNotificationPreferences < kAlert_NotifyAlsoDisplayAlert)
												? null
												: gNotificationMessage;
				gNotificationPtr->nmResp = null;
				gNotificationPtr->nmRefCon = 0L;
				result = NMInstall(gNotificationPtr);
				UseResFile(oldResourceFile);
			}
		}
		
		Embedding_DeactivateFrontmostWindow();
		InitCursor(), Cursors_UseArrow();
		result = standardAlert(alertPtr, alertPtr->alertType, alertPtr->dialogText, alertPtr->helpText);
		Embedding_RestoreFrontmostWindow();
		refReleaseLock(inAlert, &alertPtr);
	}
	else result = notInitErr;
	
	return result;
}// Display


/*!
Use this method to determine the number of the
button that was selected.  A value of 1, 2, or
3 indicates the respective button positions for
the default button, cancel button, and other
button.  The value 4 indicates the help button.
If the alert times out (that is, it closed
itself), 0 is returned.

(1.0)
*/
SInt16
Alert_ItemHit	(InterfaceLibAlertRef		inAlert)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	SInt16					result = kAlertStdAlertOtherButton;
	
	
	if (alertPtr != null) result = alertPtr->itemHit;
	refReleaseLock(inAlert, &alertPtr);
	return result;
}// ItemHit


/*!
Use this method to display a note alert with
an OK button, the given message and help text,
and optionally a help button.

(1.0)
*/
void
Alert_Message	(StringPtr		inDialogText,
				 StringPtr		inHelpText,
				 Boolean		inIsHelpButton)
{
	InterfaceLibAlertRef	box = null;
	short					hit = 0;
	
	
	box = Alert_New(); // no help button, single OK button
	setStdFilterProcForAlert(box);
	Alert_SetHelpButton(box, inIsHelpButton);
	Alert_SetText(box, inDialogText, inHelpText);
	Alert_SetType(box, kAlertNoteAlert);
	Alert_Display(box);
	Alert_Dispose(&box);
}// Message


/*!
Use this method to automatically report 16-bit
error results.  If the specified error is either
"noErr" or "userCanceledErr", no alert is
displayed and false is returned; otherwise,
true is returned to indicate a serious error.
The alert offers the user the option to quit
the program or continue.  If the user chooses
to quit, ExitToShell() is invoked.

(1.0)
*/
Boolean
Alert_ReportOSErr	(OSErr		inErrorCode)
{
	OSStatus		status = 0L;
	
	
	status = inErrorCode;
	return Alert_ReportOSStatus(status);
}// ReportOSErr


/*!
Use this method to automatically report error
result codes.  If the specified error is either
"noErr" or "userCanceledErr", no alert is
displayed and false is returned; otherwise,
true is returned to indicate a serious error.
The alert offers the user the option to quit
the program or continue.  If the user chooses
to quit, ExitToShell() is invoked.

(1.0)
*/
Boolean
Alert_ReportOSStatus	(OSStatus		inErrorCode)
{
	Boolean			result = true;
	
	
	if ((inErrorCode == userCanceledErr) || (inErrorCode == noErr)) result = false;
	
	if (result)
	{
		InterfaceLibAlertRef	box = null;
		Str31					quitButtonText,
								continueButtonText,
								errorCodeText;
		Str255					dialogText,
								helpText;
		Boolean					doExit = false;
		
		
		// generate dialog text
		PLstrcpy(errorCodeText, EMPTY_PSTRING);
		NumToString(inErrorCode, errorCodeText);
		GetIndString(dialogText, rInterfaceLibMiscellaneousStrings, siInterfaceLibReportOSStatus);
		StringUtilities_PMetaSetDefaults();
		StringUtilities_PBuild(dialogText, errorCodeText, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
		
		// generate help text
		GetIndString(quitButtonText, rInterfaceLibAlertButtonStrings, BUTTON_QUIT);
		GetIndString(continueButtonText, rInterfaceLibAlertButtonStrings, BUTTON_CONTINUE);
		GetIndString(helpText, rInterfaceLibMiscellaneousStrings, siInterfaceLibReportOSStatusHelpText);
		StringUtilities_PMetaSetDefaults();
		StringUtilities_PBuild(helpText, quitButtonText, continueButtonText, EMPTY_PSTRING, EMPTY_PSTRING);
		
		box = Alert_New();
		setStdFilterProcForAlert(box);
		Alert_SetHelpButton(box, false);
		Alert_SetParamsFor(box, kAlert_StyleOKCancel);
		Alert_SetButtonText(box, kAlertStdAlertOKButton, quitButtonText);
		Alert_SetButtonText(box, kAlertStdAlertCancelButton, continueButtonText);
		Alert_SetText(box, dialogText, helpText);
		Alert_SetType(box, kAlertStopAlert);
		Alert_Display(box);
		doExit = (Alert_ItemHit(box) == kAlertStdAlertOKButton);
		Alert_Dispose(&box);
		if (doExit) ExitToShell();
	}
	return result;
}// ReportOSStatus


/*!
Call this routine when a resume event occurs
to see if any notifications need servicing
(usually displays one or more alert windows
that were deferred because the program was
backgrounded).

(1.0)
*/
void
Alert_ServiceNotification	(void)
{
	if (gNotificationPtr != null)
	{
		(OSErr)NMRemove(gNotificationPtr);
		if (gNotificationPtr->nmIcon != null) ReleaseResource(gNotificationPtr->nmIcon), gNotificationPtr->nmIcon = null;
		DisposePtr((Ptr)gNotificationPtr), gNotificationPtr = null;
	}
}// ServiceNotification


/*!
Use this method to set the label of a specific
button in subsequently-displayed alerts.  This
information could be overridden by alert styles,
so you should always use Alert_SetParamsFor()
before calling this method.  The constant that
identifies the button text to change is one of
the standard alert constants below:

	kAlertStdAlertOKButton
	kAlertStdAlertCancelButton
	kAlertStdAlertOtherButton

In addition, you are allowed to specify for the
button text one of the following (as opposed to
a string):

	kAlertDefaultOKText, for kAlertStdAlertOKButton
	kAlertDefaultCancelText, for kAlertStdAlertCancelButton
	kAlertDefaultOtherText, for kAlertStdAlertOtherButton

(1.0)
*/
void
Alert_SetButtonText	(InterfaceLibAlertRef		inAlert,
					 short						inWhichButton,
					 StringPtr					inNewText)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	
	
	if (alertPtr != null)
	{
		switch (inWhichButton)
		{
			case kAlertStdAlertCancelButton:
				alertPtr->params.cancelText = inNewText;
				break;
			
			case kAlertStdAlertOtherButton:
				alertPtr->params.otherText = inNewText;
				break;
			
			case kAlertStdAlertOKButton:
			default:
				alertPtr->params.defaultText = inNewText;
				break;
		}
		refReleaseLock(inAlert, &alertPtr);
	}
}// SetButtonText


/*!
Use this method to set the dialog filter
callback routine for the alert box events
occurring in subsequently-displayed alerts.

(1.0)
*/
void
Alert_SetFilterProc	(InterfaceLibAlertRef		inAlert,
					 ModalFilterUPP				inProc)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	
	
	if (alertPtr != null)
	{
		alertPtr->params.filterProc = inProc;
		refReleaseLock(inAlert, &alertPtr);
	}
}// SetFilterProc


/*!
Use this method to specify whether a help
button appears in the alert.  Help buttons
should not appear unless help is available!
If the Contexts library or Apple Guide itself
cannot be found, this method will not accept
a parameter value of true.

(1.0)
*/
void
Alert_SetHelpButton	(InterfaceLibAlertRef		inAlert,
					 Boolean					inIsHelpButton)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	
	
	if (alertPtr != null)
	{
		alertPtr->params.helpButton = inIsHelpButton;
		if (!ContextsAvailable()) alertPtr->params.helpButton = false;
		refReleaseLock(inAlert, &alertPtr);
	}
}// SetHelpButton


/*!
Use this method whenever your application
encounters a switch event, so that the
Alert module knows whether the application
is in the background (suspended) or not.
This information is used to determine if a
notification should be posted when any
attempt is made to display an alert.

(1.0)
*/
void
Alert_SetIsBackgrounded		(Boolean		inIsApplicationSuspended)
{
	gNotificationIsBackgrounded = inIsApplicationSuspended;
}// SetIsBackgrounded


/*!
Use this method to specify whether the alert
box is movable.  Apple Human Interface
Guidelines strongly suggest movable alerts,
and there is no real reason not to use them;
however, this option is available.

(1.0)
*/
void
Alert_SetMovable	(InterfaceLibAlertRef		inAlert,
					 Boolean					inIsMovable)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	
	
	if (alertPtr != null)
	{
		alertPtr->params.movable = inIsMovable;
		refReleaseLock(inAlert, &alertPtr);
	}
}// SetMovable


/*!
If you have used Alert_SetNotificationPreferences()
to specify that a message be displayed in the event
of a background alert, you can use this method to
specify what that message should say.

(1.0)
*/
void
Alert_SetNotificationMessage		(StringPtr		inMessage)
{
	PLstrcpy(gNotificationMessage, inMessage);
}// SetNotificationMessage


/*!
Use this method to determine what (if any) background
notification steps the Alert module will take if an
attempt is made to display an alert while the program
is in the background.  The simplest preference is to
simply display the alert normally, and let the user
discover its presence on his or her own.  The next
level is to display a diamond mark in the process
menu that identifies the presence of an alert box in
the application.  Another option is to also flash an
icon (of ID "kAlert_NotificationSICNResourceID" and
type 'SICN', in the application resource file) in the
process menus title area.  The last level is to do
all of the above things, and also display a message
(specified with Alert_SetNotificationMessage()) that
tells the user your application requires his or her
attention.

(1.0)
*/
void
Alert_SetNotificationPreferences		(UInt16 const		inNotificationPreferences)
{
	gNotificationPreferences = inNotificationPreferences;
}// SetNotificationPreferences


/*!
Use this method to automatically configure
the button text, number of buttons and
default button of subsequently-displayed
alerts based on a specific style.  This
method should be called early in the alert
setup procedure, since it has the potential
to take precedence over changes made by
other alert routines (such as
Alert_SetButtonText()).

(1.0)
*/
void
Alert_SetParamsFor	(InterfaceLibAlertRef		inAlert,
					 short						inAlertStyle)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	
	
	if (alertPtr != null)
	{
		switch (inAlertStyle)
		{
			case kAlert_StyleOK:
				alertPtr->params.defaultText = (StringPtr)kAlertDefaultOKText;
				alertPtr->params.cancelText = (StringPtr)null;
				alertPtr->params.otherText = (StringPtr)null;
				alertPtr->params.defaultButton = kAlertStdAlertOKButton;
				alertPtr->params.cancelButton = 0;
				break;
			
			case kAlert_StyleCancel:
				alertPtr->params.defaultText = (StringPtr)kAlertDefaultCancelText;
				alertPtr->params.cancelText = (StringPtr)null;
				alertPtr->params.otherText = (StringPtr)null;
				alertPtr->params.defaultButton = kAlertStdAlertOKButton;
				alertPtr->params.cancelButton = 0;
				break;
			
			case kAlert_StyleOKCancel:
			case kAlert_StyleOKCancel_CancelIsDefault:
				alertPtr->params.defaultText = (StringPtr)kAlertDefaultOKText;
				alertPtr->params.cancelText = (StringPtr)kAlertDefaultCancelText;
				alertPtr->params.otherText = (StringPtr)null;
				if (inAlertStyle == kAlert_StyleOKCancel_CancelIsDefault)
				{
					alertPtr->params.defaultButton = kAlertStdAlertCancelButton;
					alertPtr->params.cancelButton = kAlertStdAlertOKButton;
				}
				else
				{
					alertPtr->params.defaultButton = kAlertStdAlertOKButton;
					alertPtr->params.cancelButton = kAlertStdAlertCancelButton;
				}
				break;
			
			case kAlert_StyleYesNo:
			case kAlert_StyleYesNo_NoIsDefault:
				newButtonString(alertPtr->button1Text, BUTTON_YES);
				newButtonString(alertPtr->button2Text, BUTTON_NO);
				alertPtr->params.defaultText = (StringPtr)alertPtr->button1Text;
				alertPtr->params.cancelText = (StringPtr)alertPtr->button2Text;
				alertPtr->params.otherText = (StringPtr)null;
				if (inAlertStyle == kAlert_StyleYesNo_NoIsDefault)
				{
					alertPtr->params.defaultButton = kAlertStdAlertCancelButton;
					alertPtr->params.cancelButton = kAlertStdAlertOKButton;
				}
				else
				{
					alertPtr->params.defaultButton = kAlertStdAlertOKButton;
					alertPtr->params.cancelButton = kAlertStdAlertCancelButton;
				}
				break;
			
			case kAlert_StyleYesNoCancel:
			case kAlert_StyleYesNoCancel_NoIsDefault:
			case kAlert_StyleYesNoCancel_CancelIsDefault:
				newButtonString(alertPtr->button1Text, BUTTON_YES);
				newButtonString(alertPtr->button2Text, BUTTON_NO);
				alertPtr->params.defaultText = (StringPtr)alertPtr->button1Text;
				alertPtr->params.cancelText = (StringPtr)alertPtr->button2Text;
				alertPtr->params.otherText = (StringPtr)kAlertDefaultCancelText;
				alertPtr->params.defaultButton = kAlertStdAlertOKButton;
				alertPtr->params.cancelButton = kAlertStdAlertOtherButton;
				if (inAlertStyle == kAlert_StyleYesNoCancel_NoIsDefault)
					alertPtr->params.defaultButton = kAlertStdAlertCancelButton;
				else if (inAlertStyle == kAlert_StyleYesNoCancel_CancelIsDefault)
					alertPtr->params.defaultButton = kAlertStdAlertOtherButton;
				else
					alertPtr->params.defaultButton = kAlertStdAlertOKButton;
				alertPtr->params.cancelButton = kAlertStdAlertOtherButton;
				break;
			
			case kAlert_StyleDontSaveCancelSave:
				newButtonString(alertPtr->button1Text, BUTTON_SAVE);
				newButtonString(alertPtr->button3Text, BUTTON_DONTSAVE);
				alertPtr->params.defaultText = (StringPtr)alertPtr->button1Text;
				alertPtr->params.cancelText = (StringPtr)kAlertDefaultCancelText;
				alertPtr->params.otherText = (StringPtr)alertPtr->button3Text;
				alertPtr->params.defaultButton = kAlertStdAlertOKButton;
				alertPtr->params.cancelButton = kAlertStdAlertCancelButton;
				break;
			
			default:
				break;
		}
		refReleaseLock(inAlert, &alertPtr);
	}
}// SetParamsFor


/*!
Use this method to set the text message
(and optional help text underneath) for
an alert.  If you do not want a text to
appear (no help text, for example), pass
an empty string.

(1.0)
*/
void
Alert_SetText	(InterfaceLibAlertRef		inAlert,
				 StringPtr					inDialogText,
				 StringPtr					inHelpText)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	
	
	if (alertPtr != null)
	{
		alertPtr->dialogText = inDialogText;
		alertPtr->helpText = inHelpText;
		refReleaseLock(inAlert, &alertPtr);
	}
}// SetText


/*!
Use this method to set the title of
an alerts window.  The title will not
show up unless an alert is also made
movable.  To remove an alerts title,
pass an empty string.

(1.0)
*/
void
Alert_SetTitle	(InterfaceLibAlertRef		inAlert,
				 StringPtr					inNewTitle)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	
	
	if (alertPtr != null)
	{
		PLstrcpy(alertPtr->title, inNewTitle);
		refReleaseLock(inAlert, &alertPtr);
	}
}// SetTitle


/*!
Use this method to set the icon for an alert.
A plain alert type has no icon.

(1.0)
*/
void
Alert_SetType	(InterfaceLibAlertRef		inAlert,
				 AlertType					inNewType)
{
	InterfaceLibAlertPtr	alertPtr = refAcquireLock(inAlert);
	
	
	if (alertPtr != null)
	{
		alertPtr->alertType = inNewType;
		refReleaseLock(inAlert, &alertPtr);
	}
}// SetType


/*!
Use this method to specify whether the alert
text of subsequently-displayed alerts should
be spoken.

(1.0)
*/
void
Alert_SetUseSpeech		(Boolean	inUseSpeech)
{
	gUseSpeech = inUseSpeech;
}// SetUseSpeech


//
// internal methods
//

/*!
Use this method to obtain a string containing
a standard button title (ex. Yes, No, Save,
etc.).

(1.0)
*/
static void
newButtonString	(Str255		outStringStorage,
				 short		inButtonStringType)
{
	GetIndString(outStringStorage, rInterfaceLibAlertButtonStrings, inButtonStringType);
}// newButtonString


/*!
To get a pointer to the internal structure
from a reference, call this routine.  When you
are finished with the pointer, unlock it using
refReleaseLock().

(1.0)
*/
static InterfaceLibAlertPtr
refAcquireLock	(InterfaceLibAlertRef	inRef)
{
	return ((InterfaceLibAlertPtr)inRef);
}// refAcquireLock


/*!
To release a lock on an instance of the internal
structure, call this routine.

(1.0)
*/
static void
refReleaseLock	(InterfaceLibAlertRef	inRef,
				 InterfaceLibAlertPtr*	inoutPtr)
{
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock


/*!
To show or hide an alert window, use this
method.  This is special because you can
animate the effect (only possible with the
Mac OS 8.5 Window Manager) by sliding the
alert to or from the menu bar.

(1.0)
*/
static void
setAlertVisibility		(DialogRef const		inDialog,
						 Boolean const			inIsVisible,
						 Boolean const			inAnimate)
{
	WindowRef		window = GetDialogWindow(inDialog);
	Boolean			haveMacOS8_5 = false,
					giveUp = false;
	long			sysv = 0L;
	
	
	if (Gestalt(gestaltSystemVersion, &sysv) == noErr)
	{
		UInt8		majorRev = Releases_GetMajorRevisionForVersion(sysv),
					minorRev = Releases_GetMinorRevisionForVersion(sysv);
		
		
		// Mac OS 8.5?
		haveMacOS8_5 = (((majorRev == 0x08) && (minorRev >= 0x05)) || (majorRev > 0x08));
	}
	else giveUp = true;
	
	// This works, and looks really cool; *however*, unfortunately, the current
	// Mac OS is too flaky about this, and sometimes chops up the animation
	// horribly (some kind of multitasking nonsense that occurs even if the
	// thread-yielding is omitted).
#if 0
	if ((inAnimate) && (haveMacOS8_5))
	{
		Rect		rect;
		
		
		if (GetWindowBounds(window, kWindowContentRgn, &rect) == noErr)
		{
			// Avoid the case where an alert might need to display centered
			// on a window thats arbitrarily far from the top of the screen,
			// since scrolling a window really far really doesnt look good.
			if (rect.top < 300)
			{
				// Use of a dynamic "delta" ensures that the animation takes
				// the same amount of time no matter how far the window has
				// to travel.  For example, if the user has a large monitor
				// and an alert shows up that is centered on a window that
				// is hundreds of pixels from the menu bar, a dynamic "delta"
				// ensures that the scroll-down effect covers more pixels
				// for each animation stage; if the user had a small monitor
				// in a similar case, the animation would then cover only a
				// few pixels per stage, by comparison.
				UInt32			junk = 0;
				ControlRef		root = null;
				SInt16			i = 0,
								j = 0,
								delta = (inIsVisible) ? INTEGER_QUARTERED(rect.top) : INTEGER_HALVED(rect.top);
				Boolean			done = false;
				
				
				if (GetRootControl(window, &root) != noErr) root = null;
				
				if (inIsVisible)
				{
					j = rect.top;
					i = -(rect.bottom - rect.top);
					MoveWindow(window, rect.left, i, false/* activate */); // arbitrary vertical starting point
					ShowWindow(window);
				}
				else
				{
					j = -(rect.bottom - rect.top);
					i = rect.top;
				}
				
				while (!done)
				{
					(OSErr)YieldToAnyThread();
					MoveWindow(window, rect.left, i, false/* activate */);
					if (root != null) if (i < delta) DrawOneControl(root);
					Delay(2, &junk); // arbitrary animation pause, in 60ths of a second
					if (inIsVisible)
					{
						done = (i > j);
						i += delta; // dynamic animation amount, in pixels, downward
					}
					else
					{
						done = (i < j);
						i -= delta; // dynamic animation amount, in pixels, upward
					}
				}
				if (!inIsVisible) HideWindow(window);
				else SelectWindow(window);
			}
			else giveUp = true;
		}
		else giveUp = true;
	}
	else
#endif
	{
		giveUp = true;
	}
	
	if (giveUp)
	{
		if (inIsVisible)
		{
			ShowWindow(window);
			SelectWindow(window);
		}
		else HideWindow(window);
	}
}// setAlertVisibility


/*!
To set the item text of a static or editable
text Appearance control in a dialog box, call
this routine.

(1.0)
*/
static void
setDialogItemControlText	(DialogRef const			inDialog,
							 DialogItemIndex const		inItemIndex,
							 StringPtr const			inText)
{
	ControlRef		control = null;
	Ptr				stringBuffer = null;
	
	
	// create a disposable copy of the given text
	stringBuffer = Memory_NewPtr(StrLength(inText)); // we skip the length byte!
	BlockMove(inText + 1, stringBuffer, GetPtrSize(stringBuffer)); // buffer has no length byte
	
	// set the control text to be the copy that was just created
	(OSErr)GetDialogItemAsControl(inDialog, inItemIndex, &control);
	(OSErr)SetControlData(control, kControlEditTextPart, kControlEditTextTextTag, GetPtrSize(stringBuffer), stringBuffer);
	
	// destroy the copy
	DisposePtr(stringBuffer), stringBuffer = null;
}// setDialogItemControlText


/*!
To call Alert_SetFilterProc() for the given
alert using the standard Mac OS dialog filter
procedure, use this method.

(1.0)
*/
static void
setStdFilterProcForAlert	(InterfaceLibAlertRef	inAlert)
{
	ModalFilterUPP		filterProc;
	
	
	(OSErr)GetStdFilterProc(&filterProc);
	Alert_SetFilterProc(inAlert, filterProc);
}// setStdFilterProcForAlert


/*!
To display a dialog box on the screen that
reflects the parameters of an Interface Library
Alert, use this method.  Any errors that occur,
preventing the dialog from displaying, will be
returned.  The item number hit is stored using
the given alert reference, and can be obtained
by calling Alert_ItemHit().

This method is invoked by Alert_Display().

(1.0)
*/
static OSErr
standardAlert	(InterfaceLibAlertPtr	inAlert,
				 AlertType				inAlertType,
				 StringPtr				inDialogText,
				 StringPtr				inHelpText)
{
	InterfaceLibAlertPtr	ptr = inAlert;
	OSErr					result = noErr;
	
	
	if (ptr != null)
	{
		enum
		{
			kTopEdgeIcon = 18		// pixels; system standard offset from the top of the alert box is 10 pixels
		};
		DialogRef			dialog = null;
		DialogItemIndex		itemHit = 0;
		ControlRef			control = null,
							iconControl = null;
		Boolean				isButton1 = true,
							isButton2 = true,
							isButton3 = true;
		
		
		dialog = GetNewDialog((ptr->params.movable) ? kInterfaceLibAlertDialogID : kInterfaceLibNonmovableAlertDialogID, null, (WindowRef)-1);
		SetDialogDefaultItem(dialog, kAlert_ItemButton1);
		SetDialogCancelItem(dialog, kAlert_ItemButton2);
		SetDialogTracksCursor(dialog, false);
		
		// if the alert should have a title, give it one
		//if (StrLength(ptr->title) > 0) SetWTitle(GetDialogWindow(dialog), ptr->title);
		if (StrLength(ptr->title) > 0) setDialogItemControlText(dialog, kAlert_ItemTitleText, ptr->title);
		
		// use a 32-bit icon for the help button, if possible
		{
			IconManagerIconRef		icon = IconManager_NewIcon();
			
			
			(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemHelpButton, &control);
			IconManager_MakeIconSuite(icon, kHelpIconResource, kSelectorAllSmallData);
			(OSErr)IconManager_SetButtonIcon(control, icon);
			IconManager_DisposeIcon(&icon);
		}
		
		// set up dialog items to reflect alert parameters
		(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemButton1, &control);
		if (ptr->params.defaultText != null)
		{
			// if the alert is configured to use the default title (OK), copy the appropriate string resource
			if (ptr->params.defaultText == (StringPtr)kAlertDefaultCancelText)
			{
				GetIndString(ptr->button1Text, rInterfaceLibAlertButtonStrings, BUTTON_OK);
				SetControlTitle(control, ptr->button1Text);
			}
			else SetControlTitle(control, ptr->params.defaultText);
		}
		else GetIndString(ptr->button1Text, rInterfaceLibAlertButtonStrings, BUTTON_OK);
		(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemButton2, &control);
		if (ptr->params.cancelText != null)
		{
			// if the alert is configured to use the default title (Cancel), copy the appropriate string resource
			if (ptr->params.cancelText == (StringPtr)kAlertDefaultCancelText)
			{
				GetIndString(ptr->button2Text, rInterfaceLibAlertButtonStrings, BUTTON_CANCEL);
				SetControlTitle(control, ptr->button2Text);
			}
			else SetControlTitle(control, ptr->params.cancelText);
		}
		else
		{
			PLstrcpy(ptr->button2Text, EMPTY_PSTRING);
			SetControlVisibility(control, false/* visible */, false/* draw */);
			isButton2 = false;
		}
		(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemButton3, &control);
		if (ptr->params.otherText != null)
		{
			// if the alert is configured to use the default title (Dont Save), copy the appropriate string resource
			if (ptr->params.otherText == (StringPtr)kAlertDefaultOtherText)
			{
				GetIndString(ptr->button3Text, rInterfaceLibAlertButtonStrings, BUTTON_DONTSAVE);
				SetControlTitle(control, ptr->button3Text);
			}
			else SetControlTitle(control, ptr->params.otherText);
		}
		else
		{
			PLstrcpy(ptr->button3Text, EMPTY_PSTRING);
			SetControlVisibility(control, false/* visible */, false/* draw */);
			isButton3 = false;
		}
		if (!ptr->params.helpButton) HideDialogItem(dialog, kAlert_ItemHelpButton);
		setDialogItemControlText(dialog, kAlert_ItemDialogText, inDialogText);
		if (inHelpText != null) setDialogItemControlText(dialog, kAlert_ItemHelpText, inHelpText);
		else HideDialogItem(dialog, kAlert_ItemHelpText);
		
		// set the alert icon properly for the alert type
		switch (inAlertType)
		{
			case kAlertStopAlert:
				iconControl = GetNewControl(kInterfaceLibControlAlertIconStop, GetDialogWindow(dialog));
				break;
			
			case kAlertNoteAlert:
				iconControl = GetNewControl(kInterfaceLibControlAlertIconNote, GetDialogWindow(dialog));
				break;
			
			case kAlertCautionAlert:
				iconControl = GetNewControl(kInterfaceLibControlAlertIconCaution, GetDialogWindow(dialog));
				break;
			
			case kAlertPlainAlert:
			default:
				iconControl = null;
				break;
		}
		MoveControl(iconControl, INTERFACELIB_ALERT_ICON_LEFT, kTopEdgeIcon);
		if (iconControl != null) AutoEmbedControl(iconControl, GetDialogWindow(dialog));
		
		// make the dialog big enough to fit the text, then display it
		{
			enum
			{
				kTopEdgeText = kTopEdgeIcon,		// system standard is to make the text the same vertical offset as the icon
				kSpacingBetweenTextItems = 10		// width in pixels vertically separating dialog text and help text
			};
			SInt16		textExpanseV = 0,
						totalExpanseV = 0, // initialized to vertical position of top edge of dialog text item
						i = 0;
			Rect		controlRect;
			
			
			totalExpanseV = kTopEdgeText;
			
			// size and arrange title text, if present
			if (StrLength(ptr->title) > 0)
			{
				(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemTitleText, &control);
				GetControlBounds(control, &controlRect);
				(OSErr)SetControlData(control, kControlEditTextPart, kControlEditTextTextTag,
										StrLength(ptr->title) * sizeof(unsigned char), ptr->title + 1);
				textExpanseV = 25; // should be fixed to properly calculate dimensions, but alas, Ive been lazy
				SizeControl(control, INTERFACELIB_ALERT_DIALOG_WD - HSP_BUTTON_AND_DIALOG - controlRect.left, textExpanseV);
				MoveControl(control, controlRect.left, totalExpanseV);
				totalExpanseV += (textExpanseV + kSpacingBetweenTextItems);
			}
			
			// size and arrange message text (always present)
			(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemDialogText, &control);
			GetControlBounds(control, &controlRect);
			textExpanseV = Localization_SetUpMultiLineTextControl(control, inDialogText, kThemeSystemFont);
			MoveControl(control, controlRect.left, totalExpanseV);
			totalExpanseV += textExpanseV;
			
			// size and arrange help text, if present
			if (inHelpText != null)
			{
				totalExpanseV += kSpacingBetweenTextItems;
				(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemHelpText, &control);
				GetControlBounds(control, &controlRect);
				textExpanseV = Localization_SetUpMultiLineTextControl(control, inHelpText, kThemeSmallSystemFont);
				MoveControl(control, controlRect.left, totalExpanseV);
				totalExpanseV += textExpanseV;
			}
			
			// make the alert window just big enough for any text and buttons being displayed
			SizeWindow(GetDialogWindow(dialog),
						INTERFACELIB_ALERT_DIALOG_WD,
						totalExpanseV + INTEGER_DOUBLED(VSP_BUTTON_AND_DIALOG) + BUTTON_HT,
						false/* update */);
			
			// for right-to-left localization only, the alert icon and text will switch places
			{
				ControlRef		control = null;
				Rect			dialogTextRect;
				
				
				(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemDialogText, &control);
				Localization_HorizontallyPlaceControls(iconControl, control);
				GetControlBounds(control, &dialogTextRect);
				if (inHelpText != null)
				{
					Rect		helpTextRect;
					
					
					// shift the title text and help text to the same horizontal position as the dialog text moved to
					(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemHelpText, &control);
					GetControlBounds(control, &helpTextRect);
					MoveControl(control, dialogTextRect.left, helpTextRect.top);
					(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemTitleText, &control);
					GetControlBounds(control, &helpTextRect);
					MoveControl(control, dialogTextRect.left, helpTextRect.top);
				}
			}
			
			// move the dialog buttons vertically to their proper place in the new dialog
			{
				DialogItemIndex		indices[2] = { kAlert_ItemButton3, 0 };
				
				
				// use this sneaky approach to correctly set up the vertical component of the third button
				(UInt16)Localization_AdjustDialogButtonArray(dialog, indices, 1);
				
				// auto-arrange the buttons in the OK and Cancel positions
				indices[0] = kAlert_ItemButton1;
				indices[1] = kAlert_ItemButton2;
				(UInt16)Localization_AdjustDialogButtonArray(dialog, indices, 2);
				
				if (isButton3)
				{
					// move and size that third button
					UInt16		x = 0,
								buttonWidth = 0;
					
					
					// first resize the button to fit its title
					buttonWidth = Localization_AutoSizeButtonItem(dialog, kAlert_ItemButton3, 0/* minimum width */);
					
					// now determine where the left edge of the button should be (locale-sensitive)
					(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemButton2, &control);
					GetControlBounds(control, &controlRect);
					if (Localization_IsLeftToRight()) x = controlRect.left - HSP_BUTTONS - buttonWidth;
					else x = controlRect.right + HSP_BUTTONS;
					
					// now move the third button appropriately
					(OSErr)GetDialogItemAsControl(dialog, kAlert_ItemButton3, &control);
					GetControlBounds(control, &controlRect);
					MoveControl(control, x, controlRect.top);
				}
			}
			
			// move the help button, if its supposed to be there
			if (ptr->params.helpButton) Localization_AdjustHelpButtonItem(dialog, kAlert_ItemHelpButton);
		}
		
		// get the rectangle where the alert box is supposed to be, and animate it into position (with Mac OS 8.5 and beyond)
		setAlertVisibility(dialog, true/* visible */, true/* animate */);
		
		if ((gUseSpeech) && (API_AVAILABLE(SpeakString)))
		{
			if (!SpeechBusySystemWide()) SpeakString(inDialogText);
		}
		
		do
		{
			ModalDialog(ptr->params.filterProc, &itemHit);
			switch (itemHit)
			{
				case kAlert_ItemButton1:
					ptr->itemHit = kAlertStdAlertOKButton;
					break;
				
				case kAlert_ItemButton2:
					ptr->itemHit = kAlertStdAlertCancelButton;
					break;
				
				case kAlert_ItemButton3:
					ptr->itemHit = kAlertStdAlertOtherButton;
					break;
				
				case kAlert_ItemHelpButton:
					ptr->itemHit = kAlertStdAlertHelpButton;
					break;
				
				default:
					ptr->itemHit = 0;
					break;
			}
		} while ((itemHit != kAlert_ItemButton1) && (itemHit != kAlert_ItemButton2) && (itemHit != kAlert_ItemButton3) &&
					(itemHit != kAlert_ItemHelpButton));
		Alert_ServiceNotification(); // just in case it wasnt removed properly
		
		setAlertVisibility(dialog, false/* visible */, true/* animate */);
		DisposeDialog(dialog);
	}
	else result = memPCErr;
	
	return result;
}// standardAlert