/*###############################################################

	Assistant.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Folders.h>
#include <Icons.h>
#include <LowMem.h>
#include <MacWindows.h>
#include <Resources.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "StringResources.h"
#include "SpacingConstants.r"

// MacTelnet includes
#include "Assistant.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "Panel.h"
#include "TelnetDescriptors.h"



// types
struct PanelData
{
	PanelRef		panel;
	Str255			title;
};
typedef struct PanelData	PanelData;
typedef PanelData*			PanelDataPtr;

// variables
static PanelRef			gCurrentPanel = null;
static WindowInfoRef	gAssistantWindowInfo = null;
static WindowRef		gAssistantWindow = null;
static ChainRef			gPanelList = null; // list of PanelData structures
static ControlRef		gControlButtonHelp = null,
						gControlBevelButtonGoBack = null,
						gControlBevelButtonGoPage = null,
						gControlBevelButtonGoForward = null,
						gControlPlacardHeader = null,
						gControlPlacardNavigation = null;

// internal methods
static PanelDataPtr		findPanelData					(PanelRef			inPanel);
static void				initializePanel					(PanelRef			inPanel);
static Boolean			isHandledBevelButtonClick		(ControlRef			inControl);
static ControlRef		newBevelButton					(PanelRef			inPanel);



//
// public methods
//

/*!
This method is used to initialize the Assistant
module.  A new window is created invisibly.

You need to install panels in the Assistant before
it can display anything useful.  When panels are
installed (via calls to Assistant_InstallPanel())
and you are ready to display the Assistant, call
Assistant_Display().

Be sure to call Assistant_Done() when you are
finished with an Assistant window.  If you need
to use a second assistant, call Assistant_Done()
to destroy the first one prior to calling this
routine, or a memory leak will occur.

(3.0)
*/
void
Assistant_Init	(StringPtr		inAssistantTitle)
{
	Cursors_UseWatch();
	
	gPanelList = Chain_New(kChainStyleSinglyLinked);
	
	// create window and associated data
	gAssistantWindow = GetNewCWindow(kWindowIDAssistant, null/* storage is not supported in Carbon */, (WindowRef)-1);
	(OSStatus)SetThemeWindowBackground(gAssistantWindow, kThemeBrushModelessDialogBackgroundActive, false/* update */);
	gAssistantWindowInfo = WindowInfo_New();
	
	if (gAssistantWindow != null)
	{
		SetWTitle(gAssistantWindow, inAssistantTitle);
		
		// create base controls
		{
			ControlRef		root = null;
			
			
			(OSErr)CreateRootControl(gAssistantWindow, &root);
		}
		gControlPlacardNavigation = GetNewControl(kIDSetupAssistantFooterPlacard, gAssistantWindow);
		gControlButtonHelp = GetNewControl(kIDAppleGuideButton, gAssistantWindow);
		Localization_AdjustHelpButtonControl(gControlButtonHelp);
		gControlBevelButtonGoBack = GetNewControl(kIDSetupAssistantBackBevelButton, gAssistantWindow);
		gControlBevelButtonGoPage = GetNewControl(kIDSetupAssistantPageBevelButton, gAssistantWindow);
		(OSErr)DeactivateControl(gControlBevelButtonGoPage); // used for display only
		gControlBevelButtonGoForward = GetNewControl(kIDSetupAssistantForwardBevelButton, gAssistantWindow);
		(OSErr)EmbedControl(gControlButtonHelp, gControlPlacardNavigation);
		(OSErr)EmbedControl(gControlBevelButtonGoBack, gControlPlacardNavigation);
		(OSErr)EmbedControl(gControlBevelButtonGoPage, gControlPlacardNavigation);
		(OSErr)EmbedControl(gControlBevelButtonGoForward, gControlPlacardNavigation);
		gControlPlacardHeader = GetNewControl(kIDSetupAssistantHeaderContentUserPane, gAssistantWindow);
		
		// determine if the Help button should be disabled...
		SetUpHelpButtonControl(gControlButtonHelp);
		
		// if necessary, rearrange the buttons for right-to-left locales
		// UNIMPLEMENTED
	}
}// Init


/*!
Disposes of an Assistant window and all associated data.
You should call this routine when you are done using an
Assistant.

(3.0)
*/
void
Assistant_Done	(void)
{
	DisposeWindow(gAssistantWindow), gAssistantWindow = null; // automatically destroys all controls
	{
		ChainLinkRef	node = null;
		PanelDataPtr	dataPtr = null;
		
		
		node = Chain_GetFirstNode(gPanelList);
		do
		{
			dataPtr = Chain_GetNodeDataPtr(gPanelList, node);
			if (dataPtr != null)
			{
				Panel_Dispose(&dataPtr->panel);
				Memory_DisposePtr((Ptr*)&dataPtr);
			}
			node = Chain_GetNextNode(gPanelList, node);
		} while (node != null);
	}
	Chain_Dispose(&gPanelList); // automatically destroys all nodes
	WindowInfo_Dispose(gAssistantWindowInfo);
}// Done


/*!
Use this method to display an Assistant window.

(3.0)
*/
void
Assistant_Display	(void)
{
	if (gAssistantWindow != null)
	{
		DialogItemIndex		hit = 0;
		
		
		// define this only for debugging; this lets you dump the dialogs
		// control embedding hierarchy to a file, before it is displayed
		#if 0
		DebugSelectControlHierarchyDumpFile(gAssistantWindow);
		#endif
		
		// display the dialog and handle events
		unless (IsWindowVisible(gAssistantWindow)) ShowWindow(gAssistantWindow);
		SelectOverRealFrontWindow(gAssistantWindow);
		Cursors_UseArrow();
		
		// set up the Window Info information
		{
			WindowInfo_SetDynamicResizing(gAssistantWindowInfo, false); // this particular window does not dynamically resize well
			WindowInfo_SetWindowDescriptor(gAssistantWindowInfo, kTelnetWindowDescriptorAssistant);
		}
		WindowInfo_SetForWindow(gAssistantWindow, gAssistantWindowInfo);
	}
	else Alert_ReportOSErr(memPCErr);
}// Display


/*!
This method responds to mouse-down events in the
Assistants window by tracking content controls
and passing events off to the displayed panel.

(3.0)
*/
void
Assistant_HandleContentClick	(EventRecord*	inoutEventPtr)
{
	SInt16				windowPart = 0;
	ControlPartCode		controlPart = 0;
	WindowRef			window = null;
	ControlRef			controlClicked = null,
						focusControl = null;
	
	
	windowPart = FindWindow(inoutEventPtr->where, &window);
	if (window != null)
	{
		GrafPtr		oldPort = null;
		Point		localMouse = inoutEventPtr->where;
		OSErr		error = noErr;
		
		
		GetPort(&oldPort);
		SetPortWindowPort(window);
		GlobalToLocal(&localMouse);
		controlClicked = FindControlUnderMouse(localMouse, window, &controlPart);
		error = GetKeyboardFocus(window, &focusControl);
		
		// if a focusable control is clicked and not focused, focus it...
		{
			UInt32		featureBits = 0L;
			
			
			(OSErr)GetControlFeatures(controlClicked, &featureBits);
			if ((featureBits & kControlGetsFocusOnClick) && ((focusControl != controlClicked) || (error != noErr)))
			{
				if (focusControl != null) Panel_Changed(gCurrentPanel, kPanel_MessageFocusLost, &focusControl);
				(OSErr)SetKeyboardFocus(window, controlClicked, kControlFocusNextPart);
				Panel_Changed(gCurrentPanel, kPanel_MessageFocusGained, &controlClicked);
			}
		}
		
		if ((controlClicked != null) && (controlPart != kControlNoPart))
		{
			Boolean			handled = false;
			
			
			// ...otherwise, handle mouse-down events in controls
			controlPart = HandleControlClick(controlClicked, localMouse, inoutEventPtr->modifiers, GetControlAction(controlClicked));
			switch (controlPart)
			{
				case kControlNoPart:
					handled = true; // do NOTHING else if the click is meaningless
					break;
				
				case kControlButtonPart:
					// help button?
					if (controlClicked == gControlButtonHelp)
					{
					#if TARGET_API_MAC_OS8
						AppleGuideUtilities_OpenFirstMainGuide();
					#else
						BEEP;
					#endif
						handled = true;
					}
					
					// this special invocation is necessary to allow bevel button clicks to be received last
					// (I will probably re-design the bevel button handler to be smarter, in the future)
					unless (handled)
					{
						handled = (Panel_Changed(gCurrentPanel, kPanel_MessageClicked, &controlClicked) == kPanel_ResponseClickHandled);
					}
					
					// if all else fails, check to see if the user clicked a category button
					unless (handled) handled = isHandledBevelButtonClick(controlClicked);
					break;
				
				default:
					break;
			}
			
			// If the Assistant window itself cant handle a click, it must have
			// been in a control owned by a panel; notify the panel of the click.
			unless (handled) Panel_Changed(gCurrentPanel, kPanel_MessageClicked, &controlClicked);
		}
		SetPort(oldPort);
	}
}// HandleContentClick


/*!
This routine responds to key-down events in items in
the Assistant window.

(3.0)
*/
Boolean
Assistant_HandleKeyDown			(EventRecord*	inoutEventPtr)
{
	Boolean			result = true;
	SInt16			keyCode = ((inoutEventPtr->message & keyCodeMask) >> 8),
					charCode = inoutEventPtr->message & charCodeMask,
					modifiers = inoutEventPtr->modifiers;
	
	
	if (charCode == 0x09)
	{
		ControlRef		focusControl = null;
		OSErr			error = noErr;
		
		
		// tab key was pressed; advance or reverse keyboard focus
		error = GetKeyboardFocus(gAssistantWindow, &focusControl);
		if (focusControl == null) error = memPCErr;
		if (error == noErr) Panel_Changed(gCurrentPanel, kPanel_MessageFocusLost, &focusControl);
		
		if (modifiers & shiftKey)
		{
			(OSErr)ReverseKeyboardFocus(gAssistantWindow);
		}
		else
		{
			(OSErr)AdvanceKeyboardFocus(gAssistantWindow);
		}
		error = GetKeyboardFocus(gAssistantWindow, &focusControl);
		if (error == noErr) Panel_Changed(gCurrentPanel, kPanel_MessageFocusGained, &focusControl);
	}
	else
	{
		ControlRef		control = null;
		
		
		if (GetKeyboardFocus(gAssistantWindow, &control) == noErr)
		{
			(SInt16)HandleControlKey(control, keyCode, charCode, modifiers);
		}
	}
	return result;
}// HandleKeyDown


/*!
Adds a panel to the Assistant window based on an
abstract Panel object.  An Assistant panel should
define a name and an icon (as recommended by the
Panel module), which is used to render the panel
when the user is at that stage of the Assistants
procedure.

The Assistant automatically adapts itself for the
panel, adding the panel to its list of panels and
firing notifications to the panel whenever necessary.
If you respond to the standard Panel interface, the
panel will function perfectly in the Assistant.

Assistant panels are ordered based on whichever is
installed first, so invoke Assistant_InstallPanel()
for panels in the order you want them to be
accessible by the user (e.g. install the introductory
panel first, the finished panel last).

(3.0)
*/
void
Assistant_InstallPanel		(PanelRef		inPanel)
{
	initializePanel(inPanel);
	
	// if this is the first panel, grab its title
	if (Chain_GetNodeCount(gPanelList) == 1)
	{
		gCurrentPanel = inPanel;
	}
}// InstallPanel


//
// internal methods
//

/*!
To acquire a pointer to the internal structure
whose panel field corresponds to the given
panel, use this method.

(3.0)
*/
static PanelDataPtr
findPanelData		(PanelRef		inPanel)
{
	ChainLinkRef	node = null;
	PanelDataPtr	result = null;
	
	
	// look for the given panel in the linked list
	node = Chain_GetFirstNode(gPanelList);
	do
	{
		result = Chain_GetNodeDataPtr(gPanelList, node);
		if (result != null) if (result->panel == inPanel) break;
		node = Chain_GetNextNode(gPanelList, node);
	} while (node != null);
	
	return result;
}// findPanelData


/*!
Initializes a panel by passing one-time property
changes to it.

(3.0)
*/
static void
initializePanel		(PanelRef		UNUSED_ARGUMENT(inPanel))
{
#if 0
	Rect			initialRect;
	Point			initialDimensions;
	ControlRef		container = null,
					placard = gControlPlacardPanels;
	
	
	GetControlBounds(placard, &initialRect);
	
	// allow the panel to create its controls by providing a pointer to the containing window
	Panel_Changed(inPanel, kPanel_MessageCreateControls, &gAssistantWindow);
	
	// Allow the panel to size and arrange its controls by setting the initial dimensions of
	// the container.  With Appearance, there is the concept of a control embedding hierarchy.
	// This works in your favor, but only if you know how to use it effectively.  The best order
	// of operations, when working with Appearance controls, is as follows:
	//	1.	Create all of your controls.
	//	2.	Embed each control in its immediate container (its OK at this point if a control is
	//		not visually contained in its embedders).
	//	3.	Set the dimensions of embedders, outside-in.  This allows you to use the dimensions
	//		of other controls (such as group boxes) to help set the size of embedded items.
	//	4.	Set the positions of controls, inside-out.  This allows you to set the position of
	//		all controls to be relative to their containers!  When you move a control that
	//		embeds other controls, the embedded controls will offset automatically.
	SetPt(&initialDimensions, initialRect.right - initialRect.left, initialRect.bottom - initialRect.top);
	Panel_GetContainerControl(inPanel, &container);
	(OSErr)EmbedControl(container, placard);
	SizeControl(container, initialDimensions.h, initialDimensions.v);
	Panel_Changed(inPanel, kPanel_MessageInitialSize, &initialDimensions);
	MoveControl(container, initialRect.left, initialRect.top); // move afterwards, so the embedded controls move too
#endif
}// initializePanel


/*!
This routine assumes the specified control is one of
the bevel buttons for an installed Assistant panel,
and tries to hide/show the appropriate panels.  You
can also use this routine to automatically select and
display a panel if no other panels are selected.

(3.0)
*/
static Boolean
isHandledBevelButtonClick	(ControlRef		UNUSED_ARGUMENT(inControl))
{
	Boolean			result = false;
	
	
#if 0
	// dont waste time if the apparently-clicked control is already selected
	unless (IsControlHilited(inControl))
	{
		register SInt16		i = 0;
		ChainLinkRef		node = Chain_GetFirstNode(gPanelList);
		PanelDataPtr		dataPtr = null,
							newPanelDataPtr = null;
		ControlRef			container = null,
							nowInvisibleContainer = null;
		
		
		do
		{
			dataPtr = Chain_GetNodeDataPtr(gPanelList, node);
			if (dataPtr != null)
			{
				Panel_GetContainerControl(dataPtr->panel, &container);
				
				// if necessary, initialize the current panel
				if (gCurrentPanel == null) gCurrentPanel = dataPtr->panel;
				
				// note which panel was displayed before (if any)
				if (gCurrentPanel == dataPtr->panel)
				{
					Boolean			visible = false;
					
					
					// this panels button is currently selected, so this is the panel that will disappear
					nowInvisibleContainer = container;
					Panel_Changed(dataPtr->panel, kPanel_MessageNewVisibility, &visible);
				}
				
				// unhighlight all buttons except the selected button, and remember
				// which panel is to be displayed next (instead of the previous one)
				if (inControl != dataPtr->bevelButton)
				{
					SetControlValue(dataPtr->bevelButton, kControlCheckBoxUncheckedValue);
				}
				else
				{
					// Then this is the category bevel button that was
					// clicked (so its panel will be displayed).  Store
					// this special data pointer.
					newPanelDataPtr = dataPtr;
				}
			}
			node = Chain_GetNextNode(gPanelList, node);
		} while (node != null);
		
		// show the new (chosen) panel
		if (newPanelDataPtr != null)
		{
			ControlRef		nowVisibleContainer = null;
			Boolean			visible = true;
			
			
			result = true;
			Panel_GetContainerControl(newPanelDataPtr->panel, &nowVisibleContainer);
			SetControlValue(newPanelDataPtr->bevelButton, kControlCheckBoxCheckedValue);
			gCurrentPanel = newPanelDataPtr->panel;
			(OSErr)ClearKeyboardFocus(gAssistantWindow);
			Panel_Changed(dataPtr->panel, kPanel_MessageNewVisibility, &visible);
			
			// swap panels
			Embedding_OffscreenSwapOverlappingControls(gAssistantWindow, nowInvisibleContainer, nowVisibleContainer);
		}
	}
#endif
	
	return result;
}// isHandledBevelButtonClick


/*!
To create a new bevel button control for the Assistant
dialog box that represents the specified panel, use this
method.  The button title and icon are acquired from the
specified panel, and the control reference constant is
set to be the specified panel reference.  The button is
returned.

(3.0)
*/
static ControlRef
newBevelButton		(PanelRef		UNUSED_ARGUMENT(inPanel))
{
	ControlRef result = null;
#if 0
	ControlRef		result = GetNewControl(kIDPrefGenericBevelButton, gAssistantWindow);
	
	
	if (result != null)
	{
		ControlRef			label = GetNewControl(kIDStaticText, gAssistantWindow),
							control = null; // generic ControlRef, used to represent the current control in a series of actions
		ChainLinkRef		node = null;
		PanelDataPtr		dataPtr = (PanelDataPtr)Memory_NewPtr(sizeof(PanelData));
		SInt16				labelHeight = 16; // I cant think of a reliable way to calculate this dynamically (yet)
		
		
		// store the panel
		dataPtr->bevelButton = result;
		dataPtr->panel = inPanel;
		node = Chain_NewNode(gPanelList, dataPtr);
		Chain_MoveNode(gPanelList, node, null/* means end of list */, kChainLinkPlacementInsertAfter);
		
		// associate the panel with the button
		SetControlReference(result, REINTERPRET_CAST(inPanel, SInt32));
		
		// make the button use the name of the panel, and set up the control font attributes for the button label
		{
			ControlFontStyleRec		style;
			Str31					title;
			
			
			control = label;
			Panel_GetName(inPanel, title);
			style.flags = kControlUseFontMask | kControlUseSizeMask | kControlUseJustMask;
			style.font = kControlFontSmallSystemFont;
			style.size = 10;
			style.just = teCenter;
			(OSErr)SetControlFontStyle(control, &style);
			(OSErr)SetControlData(control, kControlEditTextPart, kControlEditTextTextTag,
									StrLength(title) * sizeof(unsigned char), title + 1/* length byte is skipped */);
			SetControlText(result, title);
		}
		
		// make the button use the icon of the panel
		(OSErr)Panel_SetButtonIcon(result, inPanel);
		
		// embed the bevel button control in its container, and arrange the new button relative to other buttons in the same group
		{
			ControlRef		placard = gControlPlacardBevelButtons, // use a variable just in case it changes in the future (less alterations)
							buttonPane = null,
							panelPane = null;
			UInt16			numberOfButtons = Chain_GetNodeCount(gPanelList);
			UInt16			paneHeight = 0;
			
			
			// determine the height the pane should have
			{
				Rect		buttonRect;
				
				
				GetControlBounds(result, &buttonRect);
				paneHeight = buttonRect.bottom - buttonRect.top + VSP_BUTTON_AND_TEXT_LABEL + labelHeight;
			}
			
			// create an invisible embedder control that will tie together the button and its label
			buttonPane = GetNewControl(kIDAnchorUserPane, gAssistantWindow);
			(OSErr)EmbedControl(buttonPane, placard);
			(OSErr)EmbedControl(result, buttonPane);
			(OSErr)EmbedControl(label, buttonPane);
			
			// center the button and its label horizontally in the placard
			{
				Rect		paneRect,		// rectangle of area just encompassing the bevel button and its label
							placardRect,	// rectangle of placard in which the buttons sit
							buttonRect;		// rectangle of bevel button (only dimensions are important)
				
				
				GetControlBounds(placard, &placardRect);
				
				// size the controls (the bevel buttons size is defined in the resource file)
				SizeControl(buttonPane, placardRect.right - placardRect.left - INTEGER_DOUBLED(HSP_BUTTON_AND_DIALOG), paneHeight);
				GetControlBounds(buttonPane, &paneRect);
				SizeControl(label, paneRect.right - paneRect.left, labelHeight);
				
				// locations are relative to the embedders origin
				control = result;
				GetControlBounds(result, &buttonRect);
				MoveControl(control, INTEGER_HALVED(paneRect.right - paneRect.left - (buttonRect.right - buttonRect.left)), buttonRect.top);
				MoveControl(label, paneRect.left, paneRect.bottom - labelHeight);
			}
			
			// arrange the button pane, which automatically moves both the bevel button and its label
			Localization_VerticallyPlaceControls(result, label); // switches them around for bottom-to-top localizations
			MoveControl(buttonPane, HSP_BUTTON_AND_DIALOG, VSP_BUTTON_AND_DIALOG + (numberOfButtons - 1) * (paneHeight + VSP_BUTTONS));
		}
	}
#endif
	
	return result;
}// newBevelButton