/*###############################################################

	Chain.c
	
	Data Access Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdlib>
#else
#	include <stdlib.h>
#endif

// Mac includes
#include <MacErrors.h>
#include <MacMemory.h>

// library includes
#include "DataAccessLib.h"



// types
struct Chain
{
	SInt16				lockCount;			// how many locks are on the handle
	ChainDescriptor		descriptor;			// user-defined identifier, describing list content type
	ChainStyle			style;				// e.g. singly-linked or doubly-linked list?
	UInt16				nodeCount;			// this is updated if a node is deleted or added
	void*				auxiliaryDataPtr;	// user-defined data associated with the entire list
	void*				headPtr;			// LinkOneWayPtr or LinkTwoWayPtr or VectorPtr
};
#ifndef __cplusplus
typedef struct Chain	Chain;
#endif
typedef Chain*			ChainPtr;
typedef ChainPtr*		ChainHandle;

typedef void*		LinkData;		// user-defined data - the contents of a node
typedef LinkData*	LinkDataPtr;	// a pointer to node data, used for offsets in vectors

struct LinkOneWay
{
	// one-way link; used for singly-linked lists
	struct LinkOneWay*		next;				// pointer to next list node, or null
	LinkData				data;				// contents of this node
};
#ifndef __cplusplus
typedef struct LinkOneWay	LinkOneWay;
#endif
typedef LinkOneWay*			LinkOneWayPtr;

struct LinkTwoWay
{
	// two-way link; used for doubly-linked lists
	struct LinkTwoWay*		next;				// pointer to next list node, or null
	struct LinkTwoWay*		prev;				// pointer to previous list node, or null
	LinkData				data;				// contents of this node
};
#ifndef __cplusplus
typedef struct LinkTwoWay	LinkTwoWay;
#endif
typedef LinkTwoWay*			LinkTwoWayPtr;

struct Vector
{
	// the head for a vector
	UInt16					freeNodeCount;		// number of unused array cells
	LinkData				array[1];			// array; used for lists that are vectors
};
#ifndef __cplusplus
typedef struct Vector		Vector;
#endif
typedef Vector*				VectorPtr;

struct ChainArrayElement
{
	// data used in arrays made out of chains (for sorting);
	// each element contains sufficient information to allow
	// efficient sorting while making it possible to re-order
	// the chain later
	LinkData		dataPtr;
	ChainLinkRef	node;
};
#ifndef __cplusplus
typedef struct ChainArrayElement	ChainArrayElement;
#endif
typedef ChainArrayElement*			ChainArrayElementPtr;

// variables
ChainLinkCompareProcPtr		gComparisonProc = null;

// internal methods
static OSErr			buildArray					(ChainRef const			inChainRef,
													 ChainArrayElementPtr*	outArray);
static int				comparer					(void const*			inElement1,
													 void const*			inElement2);
static UInt16			decrementNodeCount			(ChainPtr				inPtr);
static UInt16			incrementNodeCount			(ChainPtr				inPtr);
static void				insertAfterNodeOneWayList	(ChainPtr				inListPtr,
												 	 LinkOneWayPtr			inNodeToInsert,
													 LinkOneWayPtr			inInsertAfter);
static void				insertAfterNodeTwoWayList	(ChainPtr				inListPtr,
												 	 LinkTwoWayPtr			inNodeToInsert,
													 LinkTwoWayPtr			inInsertAfter);
static void				insertBeforeNodeOneWayList	(ChainPtr				inListPtr,
												 	 LinkOneWayPtr			inNodeToInsert,
													 LinkOneWayPtr			inInsertBefore);
static void				insertBeforeNodeTwoWayList	(ChainPtr				inListPtr,
												 	 LinkTwoWayPtr			inNodeToInsert,
													 LinkTwoWayPtr			inInsertBefore);
static ChainPtr			refAcquireLock				(ChainRef				inRef);
static void				refReleaseLock				(ChainRef				inRef,
													 ChainPtr*				inoutPtr);
static LinkDataPtr		nodeRefToLinkDataPtr		(ChainLinkRef			inNodeRef);
static LinkOneWayPtr	nodeRefToLinkOneWayPtr		(ChainLinkRef			inNodeRef);
static LinkTwoWayPtr	nodeRefToLinkTwoWayPtr		(ChainLinkRef			inNodeRef);



//
// public methods
//

/*!
To create a new linked list, use this method.
The style of list you specify should reflect
the list's intended application: if efficiency
is more important than required storage, use a
doubly-linked list; otherwise, use a singly-
linked list.

If any problems occur creating the list, null
is returned.

(1.0)
*/
ChainRef
Chain_New		(ChainStyle const	inStyle)
{
	ChainRef		result = null;
	
	
	// the style must be valid in order for the list to be created successfully
	if ((inStyle == kChainStyleSinglyLinked) ||
		(inStyle == kChainStyleSinglyLinked))
	{
		result = (ChainRef)Memory_NewHandle(sizeof(Chain));
	}
	if (result != null)
	{
		ChainPtr	listPtr = refAcquireLock(result);
		
		
		if (listPtr != null)
		{
			// initialize the list information
			listPtr->descriptor = kChainInvalidDescriptor;
			listPtr->style = inStyle;
			listPtr->nodeCount = 0;
			listPtr->auxiliaryDataPtr = null;
			listPtr->headPtr = null;
			refReleaseLock(result, &listPtr);
		}
	}
	return result;
}// New


/*!
To completely destroy a list that you created
with Chain_New(), along with all nodes in that
list, use this method.

(1.0)
*/
void
Chain_Dispose		(ChainRef*			inoutChainRef)
{
	if (inoutChainRef != null)
	{
		Chain_DisposeAllNodes(*inoutChainRef);
		DisposeHandle((Handle)*inoutChainRef), *inoutChainRef = null;
	}
}// Dispose


/*!
To create a new linked list node, use this method.
As a convenience, you can specify a pointer to the
data that should be associated with the node (if
you don't specify it here, you need to use the
Chain_SetNodeDataPtr() method later).

The new node, although always using the same kind
of opaque reference (ChainLinkRef), will
automatically refer to the correct type of node
for the given list (for example, singly-linked or
doubly-linked).  If any problems occur creating
the list node, null is returned.

(1.0)
*/
ChainLinkRef
Chain_NewNode		(ChainRef const		inChainRef,
					 void*				inNodeDataPtr)
{
	ChainPtr		listPtr = refAcquireLock(inChainRef);
	ChainLinkRef	result = null;
	
	
	// the style must be valid in order for the node to be created successfully
	if (listPtr != null)
	{
		switch (listPtr->style)
		{
			case kChainStyleSinglyLinked:
				result = (ChainLinkRef)Memory_NewPtr(sizeof(LinkOneWay));
				if (result != null)
				{
					LinkOneWayPtr	ptr = nodeRefToLinkOneWayPtr(result);
					
					
					if (ptr != null)
					{
						// initialize the node information
						ptr->next = null;
						ptr->data = inNodeDataPtr;
					}
				}
				break;
			
			case kChainStyleDoublyLinked:
				result = (ChainLinkRef)Memory_NewPtr(sizeof(LinkTwoWay));
				if (result != null)
				{
					LinkTwoWayPtr	ptr = nodeRefToLinkTwoWayPtr(result);
					
					
					if (ptr != null)
					{
						// initialize the node information
						ptr->next = ptr->prev = null;
						ptr->data = inNodeDataPtr;
					}
				}
				break;
			
			default:
				break;
		}
		refReleaseLock(inChainRef, &listPtr);
	}
	return result;
}// NewNode


/*!
To completely destroy a list node that you created
with Chain_NewNode(), use this method.  The list
that the node is in is automatically updated so
that the absence of the node does not cause any
problems.

As a convenience, the auxiliary data pointer that
is associated with the destroyed node is returned.
Without this pointer, you would to invoke the
method Chain_GetNodeDataPtr() prior to using
Chain_DisposeNode(), to make sure you could
destroy the memory (or at least preserve its
pointer) before the node disappeared.

On output, your reference to the destroyed node is
automatically set to null.

IMPORTANT:	Do not pass null for "inChainRef",
			or the node will not be destroyed.

(1.0)
*/
void*
Chain_DisposeNode		(ChainRef const		inChainRef,
						 ChainLinkRef*		inoutChainLinkRef)
{
	void*		result = null;
	
	
	if ((inoutChainLinkRef != null) && (inChainRef != null))
	{
		result = Chain_GetNodeDataPtr(inChainRef, *inoutChainLinkRef);
		Chain_DeleteNode(inChainRef, *inoutChainLinkRef);
		DisposePtr((Ptr)*inoutChainLinkRef), *inoutChainLinkRef = null;
	}
	
	return result;
}// DisposeNode


/*!
To completely destroy the items in a list, but
leave the list intact, use this method.  You
must call Chain_Dispose() to destroy the list
itself.  Calling Chain_Dispose() will automatically
invoke Chain_DisposeAllNodes(), however.

WARNING:	If you need to dispose of memory that
			the data pointers of your nodes point
			to, call Chain_GetNodeDataPtr() on
			affected nodes before calling this
			routine.  Once this routine is called,
			none of the nodes will be accessible.

(1.0)
*/
void
Chain_DisposeAllNodes		(ChainRef const		inChainRef)
{
	ChainLinkRef	node = null;
	UInt16				i = 0;
	
	
	if (inChainRef != null) for (i = 0; i < Chain_GetNodeCount(inChainRef); i++)
	{
		node = Chain_DeleteIndexedNode(inChainRef, i);
		Chain_DisposeNode(inChainRef, &node);
	}
}// DisposeAllNodes


/*!
For debugging purposes, this method will avoid the
efficient look-up count that Chain_GetNodeCount()
obtains, and will instead traverse the entire linked
list, counting items as it goes.

(1.0)
*/
UInt16
Chain_DebugGetTraversedNodeCount	(ChainRef const		inChainRef)
{
	ChainPtr	listPtr = refAcquireLock(inChainRef);
	UInt16			result = 0;
	
	
	if (listPtr != null)
	{
		switch (listPtr->style)
		{
			case kChainStyleSinglyLinked:
				{
					LinkOneWayPtr	ptr = (LinkOneWayPtr)listPtr->headPtr;
					
					
					while (ptr != null) { ++result, ptr = ptr->next; };
				}
				break;
			
			case kChainStyleDoublyLinked:
				{
					LinkTwoWayPtr	ptr = (LinkTwoWayPtr)listPtr->headPtr;
					
					
					while (ptr != null) { ++result, ptr = ptr->next; };
				}
				break;
			
			case kChainStyleVector:
				{
					VectorPtr		ptr = (VectorPtr)listPtr->headPtr;
					
					
					result = GetPtrSize((Ptr)ptr->array) / sizeof(LinkData);
				}
				break;
			
			default:
				break;
		}
		refReleaseLock(inChainRef, &listPtr);
	}
	
	return result;
}// DebugGetTraversedNodeCount


/*!
To completely destroy the first node in a list,
but leave the remainder of the list nodes intact,
use this method.

(1.0)
*/
ChainLinkRef
Chain_DeleteFirstNode		(ChainRef const		inChainRef)
{
	return Chain_DeleteIndexedNode(inChainRef, 0);
}// DeleteFirstNode


/*!
To delete a node occupying a particular position
in a list without destroying the node, use this
method.  The deleted node (if any) is returned.
The index 0 refers to the first item in the list,
and the index (Chain_GetNodeCount() - 1) is the
last item.  Any index outside this range is
invalid and will cause null to be returned.

You can use Chain_DisposeNode() to destroy the
returned node completely.

(1.0)
*/
ChainLinkRef
Chain_DeleteIndexedNode		(ChainRef const		inChainRef,
							 UInt16 const		inZeroBasedNodeIndex)
{
	ChainLinkRef		node = Chain_GetIndexedNode(inChainRef, inZeroBasedNodeIndex),
						result = null;
	
	
	result = Chain_DeleteNode(inChainRef, node);
	
	return result;
}// DeleteIndexedNode


/*!
To delete the last node in a list, but leave the
remainder of the list nodes intact, use this method.
A reference to the deleted node is returned.  You
must call Chain_DisposeNode() to destroy the node
completely.

(1.0)
*/
ChainLinkRef
Chain_DeleteLastNode		(ChainRef const		inChainRef)
{
	return Chain_DeleteIndexedNode(inChainRef, Chain_GetNodeCount(inChainRef) - 1);
}// DeleteLastNode


/*!
To delete a node from a list without destroying
the node, use this method.  The deleted node (if
any) is returned.

You can use Chain_DisposeNode() to destroy the
returned node completely.

(1.0)
*/
ChainLinkRef
Chain_DeleteNode	(ChainRef const			inChainRef,
					 ChainLinkRef const		inNodeRef)
{
	ChainPtr		listPtr = refAcquireLock(inChainRef);
	ChainLinkRef	result = null;
	
	
	// the style must be valid in order for the list to be accessed successfully
	if (listPtr != null)
	{
		switch (listPtr->style)
		{
			case kChainStyleSinglyLinked:
				{
					LinkOneWayPtr	givenNodePtr = nodeRefToLinkOneWayPtr(inNodeRef),
									prevPtr = null,
									ptr = (LinkOneWayPtr)listPtr->headPtr;
					
					
					while ((ptr != givenNodePtr) && (ptr != null))
					{
						prevPtr = ptr;
						ptr = ptr->next;
					}
					if (ptr != null)
					{
						// if the node to delete is the first, change the list head
						if (ptr == ((LinkOneWayPtr)listPtr->headPtr))
						{
							listPtr->headPtr = ((LinkOneWayPtr)listPtr->headPtr)->next;
						}
						
						// rearrange the links around the removed item
						if (prevPtr != null) prevPtr->next = ptr->next;
						ptr->next = null;
						decrementNodeCount(listPtr);
						result = (ChainLinkRef)ptr;
					}
				}
				break;
			
			case kChainStyleDoublyLinked:
				{
					LinkTwoWayPtr	givenNodePtr = nodeRefToLinkTwoWayPtr(inNodeRef),
									ptr = (LinkTwoWayPtr)listPtr->headPtr;
					
					
					while ((ptr != givenNodePtr) && (ptr != null)) ptr = ptr->next;
					if (ptr != null)
					{
						// if the node to delete is the first, change the list head
						if (ptr == ((LinkTwoWayPtr)listPtr->headPtr))
						{
							listPtr->headPtr = ((LinkTwoWayPtr)listPtr->headPtr)->next;
						}
						
						// rearrange the links around the removed item
						if (ptr->prev != null) ptr->prev->next = ptr->next;
						if (ptr->next != null) ptr->next->prev = ptr->prev;
						ptr->next = ptr->prev = null;
						decrementNodeCount(listPtr);
						result = (ChainLinkRef)ptr;
					}
				}
				break;
			
			case kChainStyleVector:
				// just shift the array without re-allocating it
				// (but keep track of the free nodes, to prevent
				// unnecessary reallocation if nodes are added later)
				{
					LinkDataPtr		givenNodePtr = nodeRefToLinkDataPtr(inNodeRef);
					VectorPtr		ptr = (VectorPtr)listPtr->headPtr;
					
					
					if (ptr != null)
					{
						ptr->freeNodeCount++;
						//shift ptr->array - unimplemented
					}
				}
				break;
			
			default:
				break;
		}
		refReleaseLock(inChainRef, &listPtr);
	}
	
	return result;
}// DeleteNode


/*!
To acquire the auxiliary data pointer that you
associated with a linked list using the method
Chain_SetAuxiliaryDataPtr()), use this routine.

(1.0)
*/
void*
Chain_GetAuxiliaryDataPtr	(ChainRef const		inChainRef)
{
	void*		result = null;
	
	
	if (inChainRef != null)
	{
		ChainPtr		listPtr = refAcquireLock(inChainRef);
		
		
		result = listPtr->auxiliaryDataPtr;
		refReleaseLock(inChainRef, &listPtr);
	}
	
	return result;
}// GetAuxiliaryDataPtr


/*!
To determine the descriptor for a linked list,
use this method.  If you consistently use the
Chain_SetDescriptor() method on all of your
lists using constants that refer to specific
types of lists (a list of strings, for instance),
you can use this routine to help you determine
how to typecast the data of a linked list node.

(1.0)
*/
ChainDescriptor
Chain_GetDescriptor		(ChainRef const		inChainRef)
{
	ChainDescriptor		result = kChainInvalidDescriptor;
	
	
	if (inChainRef != null)
	{
		ChainPtr		listPtr = refAcquireLock(inChainRef);
		
		
		result = listPtr->descriptor;
		refReleaseLock(inChainRef, &listPtr);
	}
	
	return result;
}// GetDescriptor


/*!
To acquire a reference to the first node in a
linked list, use this method.

(1.0)
*/
ChainLinkRef
Chain_GetFirstNode		(ChainRef const		inChainRef)
{
	return Chain_GetIndexedNode(inChainRef, 0);
}// GetFirstNode


/*!
To obtain a node occupying a particular position
in a list, use this method.  The index 0 refers
to the first item in the list, and the index
(Chain_GetNodeCount() - 1) is the last item.
Any index outside this range is invalid and will
cause null to be returned.

(1.0)
*/
ChainLinkRef
Chain_GetIndexedNode	(ChainRef const		inChainRef,
						 UInt16 const		inZeroBasedNodeIndex)
{
	ChainPtr		listPtr = refAcquireLock(inChainRef);
	ChainLinkRef	result = null;
	
	
	// the style must be valid in order for the list to be accessed successfully
	if (listPtr != null)
	{
		UInt16		counter = 0;
		
		
		switch (listPtr->style)
		{
			case kChainStyleSinglyLinked:
				{
					LinkOneWayPtr	ptr = (LinkOneWayPtr)listPtr->headPtr;
					
					
					while ((counter < inZeroBasedNodeIndex) && (ptr != null))
					{
						counter++;
						ptr = ptr->next;
					}
					if (counter == inZeroBasedNodeIndex) result = (ChainLinkRef)ptr;
				}
				break;
			
			case kChainStyleDoublyLinked:
				{
					LinkTwoWayPtr	ptr = (LinkTwoWayPtr)listPtr->headPtr;
					
					
					while ((counter < inZeroBasedNodeIndex) && (ptr != null))
					{
						counter++;
						ptr = ptr->next;
					}
					if (counter == inZeroBasedNodeIndex) result = (ChainLinkRef)ptr;
				}
				break;
			
			default:
				break;
		}
		refReleaseLock(inChainRef, &listPtr);
	}
	
	return result;
}// GetIndexedNode


/*!
To acquire a reference to the first node in a
linked list, use this method.

(1.0)
*/
ChainLinkRef
Chain_GetLastNode		(ChainRef const		inChainRef)
{
	return Chain_GetIndexedNode(inChainRef, Chain_GetNodeCount(inChainRef) - 1);
}// GetLastNode


/*!
To obtain a reference to the node that follows
the specified node in the given list, use this
method.  This operation is equally efficient
no matter what the style of linked list.

If any problems occur, null is returned.

(1.0)
*/
ChainLinkRef
Chain_GetNextNode	(ChainRef const		inChainRef,
					 ChainLinkRef const	inCurrentNodeRef)
{
	ChainPtr		listPtr = refAcquireLock(inChainRef);
	ChainLinkRef	result = null;
	
	
	// the style must be valid in order for the list to be accessed successfully
	if (listPtr != null)
	{
		switch (listPtr->style)
		{
			case kChainStyleSinglyLinked:
				{
					LinkOneWayPtr	ptr = nodeRefToLinkOneWayPtr(inCurrentNodeRef);
					
					
					if (ptr != null) result = (ChainLinkRef)ptr->next;
				}
				break;
			
			case kChainStyleDoublyLinked:
				{
					LinkTwoWayPtr	ptr = nodeRefToLinkTwoWayPtr(inCurrentNodeRef);
					
					
					if (ptr != null) result = (ChainLinkRef)ptr->next;
				}
				break;
			
			default:
				break;
		}
		refReleaseLock(inChainRef, &listPtr);
	}
	return result;
}// GetNextNode


/*!
To determine in an efficient manner the number of
nodes in a linked list, use this method.  The
efficiency is the same regardless of the style of
linked list.

(1.0)
*/
UInt16
Chain_GetNodeCount		(ChainRef const		inChainRef)
{
	UInt16		result = 0;
	
	
	if (inChainRef != null)
	{
		ChainPtr		listPtr = refAcquireLock(inChainRef);
		
		
		result = listPtr->nodeCount;
		refReleaseLock(inChainRef, &listPtr);
	}
	
	return result;
}// GetNodeCount


/*!
To acquire the auxiliary data pointer that you
associated with a linked list node either at
creation time (using Chain_NewNode()) or
afterwards (with Chain_SetNodeDataPtr()),
use this method.

(1.0)
*/
void*
Chain_GetNodeDataPtr	(ChainRef const		inChainRef,
						 ChainLinkRef const	inNodeRef)
{
	void*		result = null;
	
	
	if ((inChainRef != null) && (inNodeRef != null))
	{
		switch (Chain_GetStyle(inChainRef))
		{
			case kChainStyleSinglyLinked:
				{
					LinkOneWayPtr		ptr = nodeRefToLinkOneWayPtr(inNodeRef);
					
					
					result = ptr->data;
				}
				break;
			
			case kChainStyleDoublyLinked:
				{
					LinkTwoWayPtr		ptr = nodeRefToLinkTwoWayPtr(inNodeRef);
					
					
					result = ptr->data;
				}
				break;
			
			default:
				break;
		}
	}
	
	return result;
}// GetNodeDataPtr


/*!
To obtain a reference to the node that precedes
the specified node in the given list, use this
method.  This operation is much more efficient
for a doubly-linked list than for a singly-linked
one.

If any problems occur, null is returned.

(1.0)
*/
ChainLinkRef
Chain_GetPreviousNode	(ChainRef const		inChainRef,
						 ChainLinkRef const	inCurrentNodeRef)
{
	ChainPtr		listPtr = refAcquireLock(inChainRef);
	ChainLinkRef	result = null;
	
	
	// the style must be valid in order for the list to be accessed successfully
	if (listPtr != null)
	{
		switch (listPtr->style)
		{
			case kChainStyleSinglyLinked:
				{
					LinkOneWayPtr	givenNodePtr = nodeRefToLinkOneWayPtr(inCurrentNodeRef),
									prevPtr = null,
									ptr = (LinkOneWayPtr)listPtr->headPtr;
					
					
					while ((ptr != givenNodePtr) && (ptr != null))
					{
						prevPtr = ptr;
						ptr = ptr->next;
					}
					result = (ChainLinkRef)prevPtr;
				}
				break;
			
			case kChainStyleDoublyLinked:
				{
					LinkTwoWayPtr	ptr = nodeRefToLinkTwoWayPtr(inCurrentNodeRef);
					
					
					if (ptr != null) result = (ChainLinkRef)ptr->prev;
				}
				break;
			
			default:
				break;
		}
		refReleaseLock(inChainRef, &listPtr);
	}
	return result;
}// GetPreviousNode


/*!
Under unusual circumstances, you may wish to
know whether a list is singly-linked or
doubly-linked.  You can use this method to
determine how a linked list was created.

(1.0)
*/
ChainStyle
Chain_GetStyle		(ChainRef const		inChainRef)
{
	ChainStyle		result = 0;
	
	
	if (inChainRef != null)
	{
		ChainPtr		listPtr = refAcquireLock(inChainRef);
		
		
		result = listPtr->style;
		refReleaseLock(inChainRef, &listPtr);
	}
	
	return result;
}// GetStyle


/*!
To determine if a node is the first one in a linked
list, use this method.

(1.0)
*/
Boolean
Chain_IsNodeFirst	(ChainRef const			inChainRef,
					 ChainLinkRef const		inNodeRef)
{
	return (Chain_GetFirstNode(inChainRef) == inNodeRef);
}// IsNodeFirst


/*!
To determine if a node is the last one in a linked
list, use this method.

(1.0)
*/
Boolean
Chain_IsNodeLast	(ChainRef const			inChainRef,
					 ChainLinkRef const		inNodeRef)
{
	return (Chain_GetLastNode(inChainRef) == inNodeRef);
}// IsNodeLast


/*!
To position a node in a linked list, use this
method.  The node does not have to be in a list;
if the node has no links, this method will add
them as appropriate to join the node to the list
in the specified way.  Both nodes must be in the
same list, if both are specified.  If you specify
null as the node to move, this method does nothing.

There are currently three ways to add a node to
a list: in front of an existing node, following
an existing node, or in place of an existing node.
If you pass null for the existing node, then the
node will appear at the front of the list if you
use the first placement method, at the end of the
list if you use the second placement method, or
nowhere if you use the third placement method
(effectively deleting the node from its list).

(1.0)
*/
void
Chain_MoveNode		(ChainRef const					inChainRef,
					 ChainLinkRef const				inNodeToMoveRef,
					 ChainLinkRef const				inNodeAnchorRefOrNull,
					 ChainLinkPlacement const		inPositionHow)
{
	ChainPtr		listPtr = refAcquireLock(inChainRef);
	
	
	// the style must be valid in order for the list to be accessed successfully
	if (listPtr != null)
	{
		switch (listPtr->style)
		{
			case kChainStyleSinglyLinked:
				{
					LinkOneWayPtr		ptr = nodeRefToLinkOneWayPtr(inNodeToMoveRef),
										anchorPtr = nodeRefToLinkOneWayPtr(inNodeAnchorRefOrNull);
					
					
					if (ptr != null) switch (inPositionHow)
					{
						case kChainLinkPlacementInsertAfter:
							// Delete the node from its current list position.  If the
							// node was in the list, a non-null reference (the node) is
							// returned, and the node count will have decreased (so it
							// needs to be restored later).  If no item was deleted, the
							// node count will remain the same for now, but then the
							// "moved" item will actually be added.  Therefore, as long
							// as the item being inserted is non-null, the node count
							// must be incremented, for an addition *or* for a move.
							(ChainLinkRef)Chain_DeleteNode(inChainRef, inNodeToMoveRef);
							insertAfterNodeOneWayList(listPtr, ptr, anchorPtr);
							incrementNodeCount(listPtr);
							break;
						
						case kChainLinkPlacementInsertBefore:
							// Delete the node from its current list position.  If the
							// node was in the list, a non-null reference (the node) is
							// returned, and the node count will have decreased (so it
							// needs to be restored later).  If no item was deleted, the
							// node count will remain the same for now, but then the
							// "moved" item will actually be added.  Therefore, as long
							// as the item being inserted is non-null, the node count
							// must be incremented, for an addition *or* for a move.
							(ChainLinkRef)Chain_DeleteNode(inChainRef, inNodeToMoveRef);
							insertBeforeNodeOneWayList(listPtr, ptr, anchorPtr);
							incrementNodeCount(listPtr);
							break;
						
						case kChainLinkPlacementReplace:
							{
								LinkOneWayPtr		anchorNext = anchorPtr->next;
								
								
								// Deleting nodes will destroy the next-node information.
								// Therefore, save the anchor's link first, delete the
								// two nodes, and then restore the link.
								(ChainLinkRef)Chain_DeleteNode(inChainRef, inNodeToMoveRef);
								(ChainLinkRef)Chain_DeleteNode(inChainRef, inNodeAnchorRefOrNull);
								ptr->next = anchorNext;
								if (anchorNext != null)
								{
									incrementNodeCount(listPtr); // one of the items returned...
								}
							}
							break;
						
						default:
							break;
					}
				}
				break;
			
			case kChainStyleDoublyLinked:
				{
					LinkTwoWayPtr		ptr = nodeRefToLinkTwoWayPtr(inNodeToMoveRef),
										anchorPtr = nodeRefToLinkTwoWayPtr(inNodeAnchorRefOrNull);
					
					
					if (ptr != null) switch (inPositionHow)
					{
						case kChainLinkPlacementInsertAfter:
							// Delete the node from its current list position.  If the
							// node was in the list, a non-null reference (the node) is
							// returned, and the node count will have decreased (so it
							// needs to be restored later).  If no item was deleted, the
							// node count will remain the same for now, but then the
							// "moved" item will actually be added.  Therefore, as long
							// as the item being inserted is non-null, the node count
							// must be incremented, for an addition *or* for a move.
							(ChainLinkRef)Chain_DeleteNode(inChainRef, inNodeToMoveRef);
							insertAfterNodeTwoWayList(listPtr, ptr, anchorPtr);
							incrementNodeCount(listPtr);
							break;
						
						case kChainLinkPlacementInsertBefore:
							// Delete the node from its current list position.  If the
							// node was in the list, a non-null reference (the node) is
							// returned, and the node count will have decreased (so it
							// needs to be restored later).  If no item was deleted, the
							// node count will remain the same for now, but then the
							// "moved" item will actually be added.  Therefore, as long
							// as the item being inserted is non-null, the node count
							// must be incremented, for an addition *or* for a move.
							(ChainLinkRef)Chain_DeleteNode(inChainRef, inNodeToMoveRef);
							insertBeforeNodeTwoWayList(listPtr, ptr, anchorPtr);
							incrementNodeCount(listPtr);
							break;
						
						case kChainLinkPlacementReplace:
							{
								LinkTwoWayPtr		anchorNext = anchorPtr->next,
													anchorPrev = anchorPtr->prev;
								
								
								// Deleting nodes will destroy the next/prev information.
								// Therefore, save the anchor's links first, delete the
								// two nodes, and then restore the links.
								(ChainLinkRef)Chain_DeleteNode(inChainRef, inNodeToMoveRef);
								(ChainLinkRef)Chain_DeleteNode(inChainRef, inNodeAnchorRefOrNull);
								ptr->next = anchorNext;
								ptr->prev = anchorPrev;
								if ((anchorNext != null) || (anchorPrev != null))
								{
									incrementNodeCount(listPtr); // one of the items returned...
								}
							}
							break;
						
						default:
							break;
					}
				}
				break;
			
			default:
				break;
		}
		refReleaseLock(inChainRef, &listPtr);
	}
}// MoveNode


/*!
Sorts the items in a linked list using the
quick sort algorithm.  This algorithm is
famous for being efficient in the average
case, since it applies divide-and-conquer
along with a basic insertion sort for small
arrays.

To be reasonably efficient and independent
of the list implementation, this routine
first builds an array that points to all
of the list items; then it sorts the array;
and finally, it arranges the linked list in
the proper order.  Since an array is used,
there is a chance that the routine will run
out of memory.  Check the result code - if
"noErr" is not returned, the list may not
have been sorted successfully.

There is a memory leak somewhere in this
code...I cant find it...maybe its within
the C standard library qsort() function!!!

(1.0)
*/
OSErr
Chain_QuickSort		(ChainRef const				inChainRef,
					 ChainLinkCompareProcPtr	inComparisonProc)
{
	OSErr					result = noErr;
	ChainArrayElementPtr	array = null;
	UInt16					arrayLength = Chain_GetNodeCount(inChainRef);
	SInt16					i = 0;
	
	
	// build an array, sort it, and then use the node information
	// in each array element to re-arrange linked list items; the
	// following code must be synchronized, no two threads can
	// simultaneously use qsort() because it depends on the global
	// variable "gComparisonProc" to decode a ChainArrayElement
	result = buildArray(inChainRef, &array);
	if (result == noErr)
	{
		gComparisonProc = inComparisonProc;
		qsort(array, arrayLength, sizeof(ChainArrayElement), comparer);
		for (i = arrayLength - 1; i >= 0; i--)
		{
			Chain_MoveNode(inChainRef, array[i].node, null/* anchor node */,
							kChainLinkPlacementInsertBefore);
		}
		DisposePtr((Ptr)array), array = null;
	}
	return result;
}// QuickSort


/*!
To associate an auxiliary data pointer with a
linked list as a whole, use this method.  You
can acquire this pointer later by invoking
Chain_GetAuxiliaryDataPtr().  An easy way to
associate simple identification information
with a linked list is to use a descriptor; see
the method Chain_SetDescriptor() for more
information on using descriptors.

(1.0)
*/
void
Chain_SetAuxiliaryDataPtr	(ChainRef		inChainRef,
							 void*			inAuxiliaryDataPtr)
{
	if (inChainRef != null)
	{
		ChainPtr		listPtr = refAcquireLock(inChainRef);
		
		
		listPtr->auxiliaryDataPtr = inAuxiliaryDataPtr;
		refReleaseLock(inChainRef, &listPtr);
	}
}// SetAuxiliaryDataPtr


/*!
To associate a simple identification tag with a
linked list, use this method.  You can determine
the descriptor of a linked list at a later time
by using Chain_GetDescriptor().

The most common application for a descriptor is
to identify a specific kind of linked list: for
example, a linked list of strings versus a linked
list of structures.  If you need to associate
more detailed data with a Linked List object than
a descriptor can provide, you can additionally
store an auxiliary data pointer in the object.
See Chain_SetAuxiliaryDataPtr() for details.

(1.0)
*/
void
Chain_SetDescriptor		(ChainRef					inChainRef,
						 ChainDescriptor const		inNewChainDescriptor)
{
	if (inChainRef != null)
	{
		ChainPtr		listPtr = refAcquireLock(inChainRef);
		
		
		listPtr->descriptor = inNewChainDescriptor;
		refReleaseLock(inChainRef, &listPtr);
	}
}// SetDescriptor


/*!
To specify the auxiliary data pointer to associate
with a linked list node, use this method.  You
usually only call this routine in special cases,
because you can specify this same information when
you create a node with Chain_NewNode().

(1.0)
*/
void
Chain_SetNodeDataPtr	(ChainRef const		inChainRef,
						 ChainLinkRef		inNodeRef,
						 void*				inAuxiliaryDataPtr)
{
	if ((inChainRef != null) && (inNodeRef != null))
	{
		switch (Chain_GetStyle(inChainRef))
		{
			case kChainStyleSinglyLinked:
				{
					LinkOneWayPtr		ptr = nodeRefToLinkOneWayPtr(inNodeRef);
					
					
					ptr->data = inAuxiliaryDataPtr;
				}
				break;
			
			case kChainStyleDoublyLinked:
				{
					LinkTwoWayPtr		ptr = nodeRefToLinkTwoWayPtr(inNodeRef);
					
					
					ptr->data = inAuxiliaryDataPtr;
				}
				break;
			
			default:
				break;
		}
	}
}// SetNodeDataPtr



//
// internal methods
//

/*!
Constructs a new array containing pointers to the
internal structure "ChainArrayElement" which
represent all of the items in the specified linked
list.  It is possible that this operation will
fail due to memory problems, in which case this
routine may return "memFullErr"; otherwise, "noErr"
is returned.

You must dispose of the resultant pointer yourself,
using DisposePtr().

(1.0)
*/
static OSErr
buildArray		(ChainRef const				inChainRef,
				 ChainArrayElementPtr*		outArray)
{
	ChainPtr	listPtr = refAcquireLock(inChainRef);
	OSErr			result = noErr;
	
	
	if (outArray == null) result = memPCErr;
	else if (listPtr != null)
	{
		*outArray = (ChainArrayElementPtr)Memory_NewPtr(listPtr->nodeCount * sizeof(ChainArrayElement));
		result = MemError();
		if (*outArray == null) result = memFullErr;
		else if (result == noErr)
		{
			ChainLinkRef			node = null;
			ChainArrayElementPtr	ptr = null;
			
			
			ptr = *outArray;
			node = Chain_GetFirstNode(inChainRef);
			do
			{
				// initialize each element of the array to include all information
				// necessary to reconstruct a linked list, should the order of
				// items in the array be changed
				ptr->dataPtr = Chain_GetNodeDataPtr(inChainRef, node);
				ptr->node = node;
				
				ptr++;
				node = Chain_GetNextNode(inChainRef, node);
			} while (node != null);
		}
		refReleaseLock(inChainRef, &listPtr);
	}
	return result;
}// buildArray


/*!
Compares two "ChainArrayElementPtr" values by
extracting the actual data from the structure
and then calling a user-defined routine to
compare that data.

WARNING:	You must initialize the global
			variable "gComparisonProc" first;
			this routine does not check that
			the variable is valid.

(1.0)
*/
static int
comparer	(void const*		inElement1,
			 void const*		inElement2)
{
	ChainArrayElementPtr	elem1Ptr = (ChainArrayElementPtr)inElement1,
							elem2Ptr = (ChainArrayElementPtr)inElement2;
	
	
	return InvokeChainLinkCompareProc(gComparisonProc, elem1Ptr->dataPtr, elem2Ptr->dataPtr);
}// comparer


/*!
To decrease the count of nodes in a
linked list, use this method.  The
new count is returned.

(1.0)
*/
static UInt16
decrementNodeCount		(ChainPtr		inPtr)
{
	UInt16		result = 0;
	
	
	if (inPtr != null) result = --(inPtr->nodeCount);
	return result;
}// decrementNodeCount


/*!
To increase the count of nodes in a
linked list, use this method.  The
new count is returned.

(1.0)
*/
static UInt16
incrementNodeCount		(ChainPtr		inPtr)
{
	UInt16		result = 0;
	
	
	if (inPtr != null) result = ++(inPtr->nodeCount);
	return result;
}// incrementNodeCount


/*!
To insert an item after a specific item in
a singly-linked list, or to append an item
to the end of the list, use this method.
This routine does not delete the specified
"inNodeToInsert" from its old position (if
any), nor does it adjust the node count of
the list.

To append an item to the list, pass null as
the "inInsertAfter" node.

WARNING:	Do not pass an "inNodeToInsert"
			that is null.

(1.0)
*/
static void
insertAfterNodeOneWayList	(ChainPtr			inListPtr,
							 LinkOneWayPtr		inNodeToInsert,
							 LinkOneWayPtr		inInsertAfter)
{
	if (inInsertAfter != null)
	{
		// insert normally
		inNodeToInsert->next = inInsertAfter->next;
		inInsertAfter->next = inNodeToInsert;
	}
	else
	{
		// append the node to the end of the list
		LinkOneWayPtr	currentPtr = (LinkOneWayPtr)inListPtr->headPtr;
		
		
		if (currentPtr != null) while (currentPtr->next != null) currentPtr = currentPtr->next;
		
		// If the list does not exist, modify the head; otherwise,
		// just modify the tail of the last item.
		if (currentPtr != null) currentPtr->next = inNodeToInsert;
		else inListPtr->headPtr = (void*)inNodeToInsert;
		
		inNodeToInsert->next = null;
	}
}// insertAfterNodeOneWayList


/*!
To insert an item after a specific item in
a doubly-linked list, or to append an item
to the end of the list, use this method.
This routine does not delete the specified
"inNodeToInsert" from its old position (if
any), nor does it adjust the node count of
the list.

To append an item to the list, pass null as
the "inInsertAfter" node.

WARNING:	Do not pass an "inNodeToInsert"
			that is null.

(1.0)
*/
static void
insertAfterNodeTwoWayList	(ChainPtr			inListPtr,
						 	 LinkTwoWayPtr		inNodeToInsert,
							 LinkTwoWayPtr		inInsertAfter)
{
	if (inInsertAfter != null)
	{
		// insert normally
		inNodeToInsert->prev = inInsertAfter;
		inNodeToInsert->next = inInsertAfter->next;
		if (inInsertAfter->next != null) inInsertAfter->next->prev = inNodeToInsert;
		inInsertAfter->next = inNodeToInsert;
	}
	else
	{
		// append the node to the end of the list
		LinkTwoWayPtr	currentPtr = (LinkTwoWayPtr)inListPtr->headPtr;
		
		
		if (currentPtr != null) while (currentPtr->next != null) currentPtr = currentPtr->next;
		
		// If the list does not exist, modify the head; otherwise,
		// just modify the tail of the last item.
		inNodeToInsert->prev = currentPtr;
		if (currentPtr != null) currentPtr->next = inNodeToInsert;
		else inListPtr->headPtr = (void*)inNodeToInsert;
		
		inNodeToInsert->next = null;
	}
}// insertAfterNodeTwoWayList


/*!
To insert an item before a specific item in
a singly-linked list, or to put an item at
the beginning of the list, use this method.
This routine does not delete the specified
"inNodeToInsert" from its old position (if
any), nor does it adjust the node count of
the list.

To make an item the first in the list, pass
null as the "inInsertBefore" node.

WARNING:	Do not pass an "inNodeToInsert"
			that is null.

(1.0)
*/
static void
insertBeforeNodeOneWayList	(ChainPtr			inListPtr,
						 	 LinkOneWayPtr		inNodeToInsert,
							 LinkOneWayPtr		inInsertBefore)
{
	if (inInsertBefore != null)
	{
		// insert normally
		LinkOneWayPtr	prevPtr = null,
						ptr = (LinkOneWayPtr)inListPtr->headPtr;
		
		
		while ((ptr != inInsertBefore) && (ptr != null))
		{
			prevPtr = ptr;
			ptr = ptr->next;
		}
		if (prevPtr != null)
		{
			inNodeToInsert->next = prevPtr->next;
			prevPtr->next = inNodeToInsert;
		}
	}
	else
	{
		// insert the node at the beginning of the list
		inNodeToInsert->next = (LinkOneWayPtr)inListPtr->headPtr;
		inListPtr->headPtr = (void*)inNodeToInsert;
	}
}// insertBeforeNodeOneWayList


/*!
To insert an item before a specific item in
a doubly-linked list, or to put an item at
the beginning of the list, use this method.
This routine does not delete the specified
"inNodeToInsert" from its old position (if
any), nor does it adjust the node count of
the list.

To make an item the first in the list, pass
null as the "inInsertBefore" node.

WARNING:	Do not pass an "inNodeToInsert"
			that is null.

(1.0)
*/
static void
insertBeforeNodeTwoWayList	(ChainPtr			inListPtr,
							 LinkTwoWayPtr		inNodeToInsert,
							 LinkTwoWayPtr		inInsertBefore)
{
	if (inInsertBefore != null)
	{
		// insert normally
		inNodeToInsert->next = inInsertBefore;
		inNodeToInsert->prev = inInsertBefore->prev;
		if (inInsertBefore->prev != null) inInsertBefore->prev->next = inNodeToInsert;
		inInsertBefore->prev = inNodeToInsert;
	}
	else
	{
		// insert the node at the beginning of the list
		LinkTwoWayPtr	oldHeadPtr = (LinkTwoWayPtr)inListPtr->headPtr;
		
		
		inNodeToInsert->next = oldHeadPtr;
		inNodeToInsert->prev = null;
		if (oldHeadPtr != null) oldHeadPtr->prev = inNodeToInsert;
		inListPtr->headPtr = (void*)inNodeToInsert;
	}
}// insertBeforeNodeTwoWayList


/*!
Casts the specified node reference into the
internal node data type appropriate for vector-
style lists.

WARNING:	Do not pass a node reference to this
			routine if it does not actually refer
			to a node created for a vector!

(1.0)
*/
#ifdef __cplusplus
inline
#endif
static LinkDataPtr
nodeRefToLinkDataPtr	(ChainLinkRef		inNodeRef)
{
	return ((LinkDataPtr)inNodeRef);
}// nodeRefToLinkDataPtr


/*!
Casts the specified node reference into the
internal node data type appropriate for singly-
linked lists.

WARNING:	Do not pass a node reference to this
			routine if it does not actually refer
			to a node created for a singly-linked
			list!

(1.0)
*/
#ifdef __cplusplus
inline
#endif
static LinkOneWayPtr
nodeRefToLinkOneWayPtr		(ChainLinkRef		inNodeRef)
{
	return ((LinkOneWayPtr)inNodeRef);
}// nodeRefToLinkOneWayPtr


/*!
Casts the specified node reference into the
internal node data type appropriate for doubly-
linked lists.

WARNING:	Do not pass a node reference to this
			routine if it does not actually refer
			to a node created for a doubly-linked
			list!

(1.0)
*/
#ifdef __cplusplus
inline
#endif
static LinkTwoWayPtr
nodeRefToLinkTwoWayPtr		(ChainLinkRef		inNodeRef)
{
	return ((LinkTwoWayPtr)inNodeRef);
}// nodeRefToLinkTwoWayPtr


/*!
To obtain a pointer to the internal Chain
structure, given a reference to it, use this
method.

(1.0)
*/
static ChainPtr
refAcquireLock	(ChainRef		inRef)
{
	ChainPtr		result = *((ChainHandle)inRef);
	
	
	HLock((Handle)inRef);
	result->lockCount++;
	return result;
}// refAcquireLock


/*!
To release a pointer to the internal Chain
structure, given a reference to it, use this
method.

(1.0)
*/
static void
refReleaseLock	(ChainRef		inRef,
				 ChainPtr*		inoutPtr)
{
	ChainPtr		ptr = *((ChainHandle)inRef);
	
	
	if (--ptr->lockCount <= 0) HUnlock((Handle)inRef);
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock