/*###############################################################

	Clipboard.c
	
	Handles the clipboard window in Telnet 3.0.  Much of this
	code has been derived from Apples code for SimpleText.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <Appearance.h>
#include <ControlDefinitions.h>
#include <LowMem.h>
#include <PLStringFuncs.h>
#include <QuickDraw.h>
#include <Scrap.h>
#include <Script.h>
#include <TextUtils.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "Clipboard.h"
#include "DialogUtilities.h"
#include "MacLook.h"
#include "network.proto.h"
#include "NetEvents.h"
#include "Preferences.h"
#include "RealScreen.h"
#include "rgmp.proto.h"
#include "TelnetDescriptors.h"
#include "TelnetErrors.h"
#include "TextTranslation.h"
#include "TektronixVirtualGraphics.h"



// variables
static WindowRef			gClipboard = null;
static WindowInfoRef		gClipboardWindowInfo = null;
static ControlRef			gClipboardRoot = null,
							//gClipboardScrollBarH = null,
							//gClipboardScrollBarV = null,
							gClipboardFooterPlacard = null,
							gClipboardFooterPlacardText = null,
							gClipboardContentUserPane = null,
							gClipboardWindowHeader = null,
							gClipboardWindowHeaderText = null;
static Boolean				gIsSuspended = false,
							gIsShowing = false,
							gIsDataScaled = false;
static SInt16				gScaledPercentage = 100;
static SInt16				gScrapCount = 0;
static ResType				gScrapType = '????';
static Handle				gScrapHandle = null;
static long					gCurrentOffset = 0L;
static Rect					gFullScreenClipboard;
static CQDProcsPtr   	   	gSavedProcsPtr = null;
static CQDProcs     	   	gMyColorProcs;
static QDGetPicUPP			gGetPICTDataUPP = null;

// internal methods
static pascal void	contentUserPaneDeviceLoop		(short				inColorDepth,
													 short				inDeviceFlags,
													 GDHandle			inTargetDevice,
													 long				inControl);
static pascal void	contentUserPaneDraw				(ControlRef			inControl,
													 SInt16				inPart);
static OSErr		drawPictureFromHandleAndOffset	(const Rect*		inDrawingRect,
													 Handle				inDataSourceHandle,
													 const long			inSourceOffset);
static pascal void	getPICTData						(void*				inDataPtr,
													 short				inByteCount);
static void			getValidScrapType				(long*				outScrapResult,
													 long*				outScrapOffset,
													 ResType*			outScrapType);
static PicHandle	graphicToPICT					(short				inDrawingNumber);
static pascal Boolean		handleNewSize			(const WindowRef	inWindowRef,
						 							 const SInt32		inDeltaSizeX,
													 const SInt32		inDeltaSizeY,
													 const SInt32		inData);
static void			pictureToScrap					(Handle				inPictureData);
static void			setDataTypeInformation			(void);
static void			setFullScreenRectangle			(void);
static void			setScalingInformation			(void);
static void			setUpControls					(const WindowRef	inWindowRef);
static void			textToScrap						(Handle				inTextHandle);



//
// public methods
//

/*!
This method sets up the initial configurations for
the clipboard window.  Call this method before
calling any other clipboard methods from this file.

Support for scroll bars has not been implemented
yet.  Nonetheless, the commented-out code for adding
the controls is here.

(3.0)
*/
void
Clipboard_Init	(void)
{
	WindowRef	clipboardWindow = null;
	Rect		rect;
	
	
	gGetPICTDataUPP = NewQDGetPicUPP(getPICTData);
	SetRect(&rect, 0, 0, 0, 0);
	
	// create the clipboard window
	clipboardWindow = gClipboard = GetNewCWindow(kWindowIDClipboard, null, (WindowRef)-1);
	(OSStatus)SetThemeWindowBackground(clipboardWindow, kThemeBrushFinderWindowBackground, false);
	
	// create the windows controls
	(OSErr)CreateRootControl(clipboardWindow, &gClipboardRoot);
	gClipboardContentUserPane = GetNewControl(kIDClipboardWindowContentUserPane, clipboardWindow);
	{
		ControlUserPaneDrawUPP		upp = NewControlUserPaneDrawUPP(contentUserPaneDraw);
		
		
		(OSErr)SetControlData(gClipboardContentUserPane, kControlNoPart, kControlUserPaneDrawProcTag,
							sizeof(ControlUserPaneDrawUPP), (Ptr)&upp);
	}
	//gClipboardScrollBarH = GetNewControl(kIDClipboardHorizontalScrollBar, clipboardWindow);
	//gClipboardScrollBarV = GetNewControl(kIDClipboardVerticalScrollBar, clipboardWindow);
	gClipboardFooterPlacard = NewControl(clipboardWindow, &rect, EMPTY_PSTRING, true/* visible */,
										0/* value */,
										0/* minimum */,
										0/* maximum */,
										kControlPlacardProc,
										0L/* ref. con. */);
	gClipboardFooterPlacardText = GetNewControl(kIDStaticText, clipboardWindow);
	gClipboardWindowHeader = NewControl(clipboardWindow, &rect, EMPTY_PSTRING, true/* visible */,
										0/* value */,
										0/* minimum */,
										0/* maximum */,
										kControlPlacardProc/*kControlWindowHeaderProc*/,
										0L/* ref. con. */);
	gClipboardWindowHeaderText = GetNewControl(kIDStaticText, clipboardWindow);
	//(OSErr)EmbedControl(gClipboardScrollBarV, gClipboardContentUserPane);
	//(OSErr)EmbedControl(gClipboardScrollBarH, gClipboardContentUserPane);
	(OSErr)EmbedControl(gClipboardFooterPlacardText, gClipboardFooterPlacard);
	(OSErr)EmbedControl(gClipboardWindowHeaderText, gClipboardWindowHeader);
	
	// set the font characteristics for the footer (clipboard size) text, which doesnt change
	{
		ControlFontStyleRec		styleRecord;
		Style					fontStyle = normal;
		SInt16					fontSize = 9;
		SInt16					fontID = 0;
		
		
		fontID = GetScriptVariable(GetScriptManagerVariable(smKeyScript), smScriptAppFond);
		styleRecord.flags = kControlUseFontMask | kControlUseFaceMask | kControlUseSizeMask;
		styleRecord.font = fontID;
		styleRecord.size = fontSize;
		styleRecord.style = fontStyle;
		
		(OSErr)SetControlData(gClipboardFooterPlacardText, kControlNoPart, kControlStaticTextStyleTag,
								sizeof(styleRecord), (Ptr)&styleRecord);
	}
	
	SetControlVisibility(gClipboardRoot, true/* visible */, false/* draw */);
	
	// set up the Window Info information
	gClipboardWindowInfo = WindowInfo_New();
	WindowInfo_SetWindowDescriptor(gClipboardWindowInfo, kTelnetWindowDescriptorClipboard);
	// arbitrary - minimum width and height, maximum width and height
	WindowInfo_SetWindowResizeLimits(gClipboardWindowInfo,
										90/* minimum height */,
										430/* minimum width, arbitrary */,
										SHRT_MAX/* maximum height */,
										SHRT_MAX/* maximum width */);
	WindowInfo_SetWindowResizeResponderUPP(gClipboardWindowInfo, handleNewSize, 0L);
	WindowInfo_SetForWindow(clipboardWindow, gClipboardWindowInfo);
	
	// concatenate MacTelnets name onto the end of the window title, so its obvious which applications Clipboard it is
	{
		Str255		titleString;
		
		
		GetWTitle(clipboardWindow, titleString);
		StringUtilities_PMetaSetDefaults();
		StringUtilities_PBuild(titleString, LMGetCurApName(), EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
		SetWTitle(clipboardWindow, titleString);
	}
	
	// Place the clipboard in the standard position indicated in the Human Interface Guidelines.
	// With themes, the window frame size cannot be guessed reliably.  Therefore, first, I take
	// the difference between the left edges of the window content and structure region boxes
	// and make that the left edge of the window.
	{
		Rect		screenRect,
					structureRect,
					contentRect;
		
		
		screenRect = (**(GetMainDevice())).gdRect;
		MacLook_GetWindowRegionBounds(clipboardWindow, kWindowContentRgn, &contentRect);
		MacLook_GetWindowRegionBounds(clipboardWindow, kWindowStructureRgn, &structureRect);
		{
			Rect		initialBounds;
			float		fh = 0.0,
						fw = 0.0;
			
			
			// use main screen size to calculate the initial clipboard size
			fh = 0.35 * (screenRect.bottom - screenRect.top);
			fw = 0.80 * (screenRect.right - screenRect.left);
			
			// define the default window size, and make the window that size initially
			SetRect(&initialBounds, screenRect.left + (contentRect.left - structureRect.left) + 4,
									screenRect.bottom - ((short)fh) - 4, 0, 0);
			SetRect(&initialBounds, initialBounds.left, initialBounds.top,
									initialBounds.left + (short)fw,
									initialBounds.top + (short)fh);
			SetWindowStandardState(clipboardWindow, &initialBounds);
			SizeWindow(clipboardWindow, initialBounds.right - initialBounds.left,
									initialBounds.bottom - initialBounds.top, false/* update */);
			MoveWindow(clipboardWindow, initialBounds.left, initialBounds.top, false);
			
			// define the "full screen" zoomed size
			setFullScreenRectangle();
		}
	}
	
	// arrange the controls in the clipboard window
	setUpControls(clipboardWindow);
}// Init


/*!
This method cleans up the clipboard window by
destroying data structures allocated by Clipboard_Init().

(3.0)
*/
void
Clipboard_Done	(void)
{
	DisposeWindow(gClipboard); // disposes of the clipboard window *and* all of its controls
	gClipboard = null;
	WindowInfo_Dispose(gClipboardWindowInfo);
	DisposeQDGetPicUPP(gGetPICTDataUPP), gGetPICTDataUPP = null;
}// Done


/*!
Reads an Apple Event descriptor containing data meant
to be copied to the clipboard, and attempts to copy it.
If successful, "noErr" is returned.

(3.0)
*/
OSErr
Clipboard_AEDescToScrap		(AEDesc const*		inDescPtr)
{
	OSErr		result = noErr;
	Size		dataSize = 0L;
	Handle		handle = null;
	
	
	dataSize = AEGetDescDataSize(inDescPtr);
	handle = Memory_NewHandle(dataSize);
	if (handle == null) result = memFullErr;
	else
	{
		Size		actualSize = 0L;
		
		
		// first, try coercing the data into text and copy it that way
		result = AppleEventUtilities_CopyDescriptorDataAs(typeChar, inDescPtr, *handle, GetHandleSize(handle), &actualSize);
		if (result == noErr) textToScrap(handle);
		else
		{
			// failed - try coercing the data into a picture instead, and copy it that way
			result = AppleEventUtilities_CopyDescriptorDataAs(typePict, inDescPtr, *handle, GetHandleSize(handle),
																&actualSize);
			if (result == noErr) pictureToScrap(handle);
		}
		Memory_DisposeHandle(&handle);
	}
	return result;
}// AEDescToScrap


/*!
Creates an appropriate Apple Event descriptor
containing the data currently on the clipboard
(if any).

(3.0)
*/
OSErr
Clipboard_CreateContentsAEDesc		(AEDesc*		outDescPtr)
{
	OSErr		result = noErr;
	
	
	if (outDescPtr != null)
	{
		outDescPtr->descriptorType = typeNull;
		outDescPtr->dataHandle = null;
		if (Clipboard_GetContentType() == kScrapFlavorTypeText)
		{
			Handle		handle = null;
			long		offset = 0L,		// offset from scrap handle of data of interest
						length = 0L;		// the length of what is on the scrap
			Boolean		isAnyScrap = false;	// is any text on the clipboard?
		#if TARGET_API_MAC_CARBON
			ScrapRef			currentScrap = null;
			ScrapFlavorFlags	currentScrapFlags = 0L;
		#endif
			
			
		#if TARGET_API_MAC_CARBON
			(OSStatus)GetCurrentScrap(&currentScrap);
		#endif
		
		#if TARGET_API_MAC_OS8
			isAnyScrap = (GetScrap(0L, kScrapFlavorTypeText, &offset) > 0L); // is any text on the clipboard?
		#else
			isAnyScrap = (GetScrapFlavorFlags(currentScrap, kScrapFlavorTypeText, &currentScrapFlags) == noErr);
		#endif
			
			if (isAnyScrap) // is any text on the clipboard?
			{
			#if TARGET_API_MAC_OS8
				handle = Memory_NewHandle(0L); // for characters
				length = GetScrap(handle, kScrapFlavorTypeText, &offset); // copy scrap into handle
			#else
				(OSStatus)GetScrapFlavorSize(currentScrap, kScrapFlavorTypeText, &length);
				handle = Memory_NewHandle(length); // for characters
				(OSStatus)GetScrapFlavorData(currentScrap, kScrapFlavorTypeText, &length, *handle);
			#endif
				result = AECreateDesc(typeChar, *handle, GetHandleSize(handle), outDescPtr);
			}
		}
	}
	else result = memPCErr;
	
	return result;
}// CreateContentsAEDesc


/*!
To determine what kind of data is on the
clipboard, use this method.  If the data
is something that the clipboard window
cannot render, '????' is returned.

(3.0)
*/
ResType
Clipboard_GetContentType	(void)
{
	long		scrapResult = 0L;
	long		offset = 0L;
	ResType		result = '????';
	
	
	getValidScrapType(&scrapResult, &offset, &result);
	return result;
}// GetContentType


/*!
Returns the number of data choices from the
current clipboard.  If this number changes
during the lifetime of the active program,
then the clipboard window may need updating.

(3.0)
*/
UInt32
Clipboard_GetFlavorCount	(void)
{
	UInt32		result = 0L;
	
	
#if TARGET_API_MAC_OS8
	result = InfoScrap()->scrapCount;
#else
	{
		ScrapRef		scrap = null;
		
		
		if (GetCurrentScrap(&scrap) == noErr)
		{
			(OSStatus)GetScrapFlavorCount(scrap, &result);
		}
	}
#endif
	
	return result;
}// GetFlavorCount


/*!
To obtain a rectangle in local coordinates
that describes the view area (i.e. the
part of the window showing the clipboards
contents, minus any window controls), use
this method.

(3.0)
*/
void
Clipboard_GetViewRectangle			(Rect*			outViewRectPtr)
{
	if (outViewRectPtr != null)
	{
		GetControlBounds(gClipboardContentUserPane, outViewRectPtr);
	}
}// GetViewRectangle


/*!
To obtain the Mac OS window pointer for the
clipboard window, call this method.

(3.0)
*/
WindowRef
Clipboard_GetWindow	(void)
{
	return (gClipboard);
}// GetWindow


/*!
Copies the indicated drawing to the clipboard.

(2.6)
*/
void
Clipboard_GraphicsToScrap	(short	inDrawingNumber)
{
	PicHandle	picture = null;
	
	
	picture = graphicToPICT(inDrawingNumber);
	pictureToScrap((Handle)picture);
	KillPicture(picture);
}// GraphicsToScrap


/*!
Call this routine once per event loop, to handle
periodic tasks for the clipboard.

(3.0)
*/
void
Clipboard_Idle	(void)
{
	Boolean			needsUpdating = false;
	
	
	needsUpdating = (gScrapCount != Clipboard_GetFlavorCount());
	unless (needsUpdating) needsUpdating = (gScrapType != Clipboard_GetContentType());
	
	if (needsUpdating) MacLook_InvalidatePortRectangle(gClipboard);
}// Idle


/*!
To determine if the clipboard is showing, call
this method.  This state is not exactly the
window state; specifically, if the clipboard is
suspended but set to visible, the window will be
invisible (because the application is suspended)
and yet this method will return "true".

If you need to find the definite visible state
of the clipboard window, use Clipboard_GetWindow()
to obtain the Mac OS window pointer and then use the
IsWindowVisible() system call.

(3.0)
*/
Boolean
Clipboard_IsVisible	(void)
{
	return gIsShowing;
}// IsVisible


/*!
To set the suspended state of the clipboard,
call this method.  If the clipboard is suspended,
it is collapsed, in order to remain on the screen
without showing any contents (other applications
could change the contents of the Clipboard
without Telnets knowledge, so it should not try
to render the Clipboard when in the background).

(3.0)
*/
void
Clipboard_SetSuspended	(Boolean	inIsSuspended)
{
	WindowRef		clipboard = Clipboard_GetWindow();
	
	
	gIsSuspended = inIsSuspended;
	CollapseWindow(clipboard, inIsSuspended);
}// SetSuspended


/*!
To show or hide the clipboard, call this method.

(3.0)
*/
void
Clipboard_SetVisible	(Boolean	inIsVisible)
{
	WindowRef		clipboard = Clipboard_GetWindow();
	
	
	gIsShowing = inIsVisible;
	if (inIsVisible)
	{
		ShowWindow(clipboard);
		SelectOverRealFrontWindow(clipboard);
	}
	else
	{
		HideWindow(clipboard);
	}
	
	unless (inIsVisible)
	{
#if TARGET_API_MAC_OS8
		(SInt32)UnloadScrap();
#else
		// Carbon does not require UnloadScrap() to be called
#endif
	}
}// SetVisible


/*!
Pastes text from the clipboard into the current
window.  If the clipboard does not have any text
component, this routine does nothing.

(2.6)
*/
void
Clipboard_TextFromScrap	(void)
{
	if (ScreenManager_IsWindowAScreen(GetRealFrontWindow()))
	{
		ConnectionDataPtr	connectionDataPtr = ScreenManager_GetWindowConnectionData(GetRealFrontWindow());
		TelnetScreenID		currentScreenID = 0;
		
		
		currentScreenID = connectionDataPtr->vs;
		if (connectionDataPtr->clientflags & PASTE_IN_PROGRESS)
		{
			// nothing new can be pasted until the current paste finishes
			BEEP;
		}
		else
		{
			long				offset = 0L,			// offset from scrap handle of data of interest
								length = 0L;			// the length of what is on the scrap
			Boolean				isAnyScrap = false;		// is any text on the clipboard?
		#if TARGET_API_MAC_CARBON
			ScrapRef			currentScrap = null;
			ScrapFlavorFlags	currentScrapFlags = 0L;
		#endif
			
			
		#if TARGET_API_MAC_CARBON
			(OSStatus)GetCurrentScrap(&currentScrap);
		#endif
			
			/* Flush the buffer if necessary */ //CCP fix for linemode
			if ((connectionDataPtr->ftpState != 0) &&
				(connectionDataPtr->echo) &&
				(connectionDataPtr->kblen > 0))
			{	
				Network_SendFlushData(connectionDataPtr->port, connectionDataPtr->kbbuf, connectionDataPtr->kblen);	
				connectionDataPtr->kblen = 0;
			}
			
		#if TARGET_API_MAC_OS8
			isAnyScrap = (GetScrap(0L, kScrapFlavorTypeText, &offset) > 0L); // is any text on the clipboard?
		#else
			isAnyScrap = (GetScrapFlavorFlags(currentScrap, kScrapFlavorTypeText, &currentScrapFlags) == noErr);
		#endif
			if (isAnyScrap)
			{
			#if TARGET_API_MAC_OS8
				connectionDataPtr->paste.text = Memory_NewHandle(0L); // for characters
				
				length = GetScrap(connectionDataPtr->paste.text, kScrapFlavorTypeText, &offset); // copy scrap into handle
				connectionDataPtr->paste.outLength = length;
			#else
				(OSStatus)GetScrapFlavorSize(currentScrap, kScrapFlavorTypeText, &length);
				connectionDataPtr->paste.text = Memory_NewHandle(length); // for characters
				connectionDataPtr->paste.outLength = length;
				(OSStatus)GetScrapFlavorData(currentScrap, kScrapFlavorTypeText, &connectionDataPtr->paste.outLength,
												*connectionDataPtr->paste.text);
			#endif
				HLock(connectionDataPtr->paste.text);
				connectionDataPtr->paste.nextCharPtr = *connectionDataPtr->paste.text;
				
				connectionDataPtr->clientflags |= PASTE_IN_PROGRESS;
				connectionDataPtr->paste.inCount = 0;
				connectionDataPtr->paste.outCount = 0;
				
				TextTranslation_ConvertTextFromMac((unsigned char*)connectionDataPtr->paste.nextCharPtr,
													connectionDataPtr->paste.outLength,
													connectionDataPtr->national); // translate to national characters
				
				pasteText(currentScreenID); // in "NetEvents.c"; paste over network, with echoing if needed
			}
		}
	}
	else BEEP;
}// TextFromScrap


/*!
Copies the current selection from the indicated
virtual screen.  If the selection mode is
rectangular, the text is copied line-by-line
according to the rows it crosses.

If the copy method is standard, the text is
copied to the clipboard intact.  If the table
method is used, the text is first parsed to
replace tab characters with multiple spaces,
and then the text is copied.  If inline is
specified (possibly in addition to table mode),
each line is concatenated together to form a
single line.

(3.0)
*/
void
Clipboard_TextToScrap		(TelnetScreenID			inScreenID,
							 TelnetCopyMethod		inHowToCopy)
{
	char**		text = null;			// where to store the characters
	short		tableThreshold = 0;		// zero for normal, nonzero for copy table mode
	Size		actualSize = 0L;
	
	
#if TARGET_API_MAC_OS8
	(OSStatus)ZeroScrap();
#else
	(OSStatus)ClearCurrentScrap();
#endif
	
	switch (inHowToCopy)
	{
		case kTelnetCopyMethodTable:
			// get the users one tab equals X spaces preference, if possible
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
										kTelnetPreferenceTag_CopyTableThreshold,
										sizeof(tableThreshold), &tableThreshold,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				tableThreshold = 4; // default to 4 spaces per tab if no preference can be found
			}
			break;
		
		case kTelnetCopyMethodStandard:
		default:
			tableThreshold = 0;
			break;
	}
	
	// get the highlighted characters; assume that text should be inlined unless selected in rectangular mode
	{
		RealScreenTextFlags		flags = (inHowToCopy & kTelnetCopyMethodInline) ? kRealScreenTextFlagInline : 0;
		
		
		text = RealScreen_GetSelectedTextAsNewHandle(inScreenID, tableThreshold, flags);
	}
	
	if (text == null) Errors_OutOfMemory(kInternalTelnetMemoryErrorClipboardTextToScrap);
	else if (GetHandleSize(text) > 0)
	{
		textToScrap(text);
		Memory_DisposeHandle(&text);
	}
	else BEEP; // nothing to Copy!
}// TextToScrap


/*!
Pastes text from the current selection into the
current window, leaving a copy on the clipboard
(replacing anything there).  If nothing is selected,
this routine does nothing.

Intelligently, MacTelnet automatically uses inline
copy-and-paste for this type of operation, since
virtually the only reason to use type would be to
copy and paste a command line from the scrollback
buffer.  If a command line spans multiple lines, it
would normally have new-line sequences inserted to
preserve the formatting, which is undesirable when
pasting what should be seen as a single line.

(3.0)
*/
void
Clipboard_TextType	(void)
{
	if (ScreenManager_IsWindowAScreen(GetRealFrontWindow()))
	{
		Clipboard_TextToScrap(ScreenManager_GetWindowScreenID(GetRealFrontWindow()),
								kTelnetCopyMethodStandard | kTelnetCopyMethodInline);
		Clipboard_TextFromScrap();
	}
	else BEEP;
}// TextType


/*!
To show or hide the clipboard based on whether or
not the clipboard is currently visible, call this
method.  You generally use this to handle a
"Show/Hide Clipboard" menu item.

(3.0)
*/
void
Clipboard_ToggleVisible	(void)
{
	Clipboard_SetVisible(!Clipboard_IsVisible());
}// ToggleVisible


/*!
To update the clipboard window, call this method.
This should be done when update events are
received for the clipboard window, or when the size
of the clipboard window has changed.  This method
recalculates the position and size of the controls
in the Clipboard window, and redraws the contents
of the clipboard using the absolutely current
clipboard state information.

(3.0)
*/
OSErr
Clipboard_Update	(GrafPtr		inDrawingPort)
{
	OSErr		result = noErr;
	WindowRef	clipboardWindow = gClipboard;
	GrafPtr		oldPort = null;
	
	
	GetPort(&oldPort);
	SetPort(inDrawingPort);
	
	// determine and describe what is currently on the clipboard
	setDataTypeInformation();
	
	(OSErr)SetControlVisibility(gClipboardRoot, true/* visibility */, false/* draw */);
	
	// if the data is scaled, say so; otherwise, remove the footer text completely
	setScalingInformation();
	
	// draw everything
	{
		RgnHandle		visibleRegion = Memory_NewRegion();
		
		
		if (visibleRegion != null)
		{
			GetPortVisibleRegion(inDrawingPort, visibleRegion);
			UpdateControls(clipboardWindow, visibleRegion);
			Memory_DisposeRegion(&visibleRegion);
		}
	}
	{
		Rect	drawingPortRect;
		
		
		GetPortBounds(inDrawingPort, &drawingPortRect);
		MacLook_ValidateRectangle(clipboardWindow, &drawingPortRect);
	}
	
	SetPort(oldPort);
	return result;
}// Update


/*!
To set the clipboard window size to either the user
state or the standard state, call this method with,
respectively, window parts of "inZoomIn" or "inZoomOut".
The state of modifier keys, in standard Mac OS Event
Manager format, is provided for special cases, and it
is presumed that any specified modifiers were held down
while the indicated window part was clicked.

(3.0)
*/
void
Clipboard_Zoom		(WindowPartCode		inWindowPart,
					 EventModifiers		inModifiers)
{
	GrafPtr		oldPort = null;
	
	
	GetPort(&oldPort);
	switch (inWindowPart)
	{
		case inZoomOut:
		case inZoomIn:
			// If the Option key is held down, maximum the
			// Clipboard window; otherwise, move and resize
			// the window to the appropriate boundaries.
			{
				Rect		initialRect,
							finalRect;
				
				
				GetPortBounds(GetWindowPort(gClipboard), &initialRect);
				SetPortWindowPort(gClipboard);
				EraseRect(&initialRect);
				MacLook_InvalidateRectangle(gClipboard, &initialRect);
				
				if ((inModifiers & optionKey) == optionKey)
				{
					setFullScreenRectangle();
					MoveWindow(gClipboard, gFullScreenClipboard.left, gFullScreenClipboard.top, false/* activate */);
					SizeWindow(gClipboard, gFullScreenClipboard.right - gFullScreenClipboard.left,
								gFullScreenClipboard.bottom - gFullScreenClipboard.top, true/* update */);
				}
				else ZoomWindow(gClipboard, inWindowPart, false);
				
				// resize window controls to match the new size
				GetPortBounds(GetWindowPort(gClipboard), &finalRect);
				handleNewSize(gClipboard, finalRect.right - finalRect.left - (initialRect.right - initialRect.left),
								finalRect.bottom - finalRect.top - (initialRect.bottom - initialRect.top), 0L);
			}
			break;
		
		default:
			BEEP;
			break;
	}
	SetPort(oldPort);
}// Zoom


//
// internal methods
//

/*!
This method, of standard DeviceLoopDrawingUPP form,
handles drawing of the Clipboard contents correctly
no matter what the device characteristics are.

(3.0)
*/
static pascal void
contentUserPaneDeviceLoop	(short				inColorDepth,
							 short				UNUSED_ARGUMENT(inDeviceFlags),
							 GDHandle			inTargetDevice,
							 long				inControl)
{
	SInt32			error = noErr;
	Rect			clipboardContentRect;
	RgnHandle		preservedClippingRegion = Memory_NewRegion();
	GrafPtr			currentPort = null;
	ControlRef		control = (ControlRef)inControl;
	
	
	GetControlBounds(control, &clipboardContentRect);
	GetPort(&currentPort);
	if (preservedClippingRegion != null) GetClip(preservedClippingRegion);
	
	// set up the control background and foreground colors
	(OSErr)SetUpControlBackground((ControlRef)inControl, inColorDepth/* device depth */,
									ColorUtilities_IsColorDevice(inTargetDevice));
	
	// clear out any data that was there before
	{
		ColorPenState		state;
		GrafPtr				oldPort = null;
		
		
		GetPort(&oldPort);
		SetPortWindowPort(gClipboard);
		
		ColorUtilities_GetColorAndPenState(&state);
		ColorUtilities_NormalizeColorAndPen();
		(OSStatus)SetThemeWindowBackground(gClipboard, kThemeBrushFinderWindowBackground, false);
		EraseRect(&clipboardContentRect);
		ColorUtilities_SetColorAndPenState(&state);
		SetPort(oldPort);
	}
	
	// determine what is currently on the clipboard
#if TARGET_API_MAC_OS8
	error = LoadScrap();
#else
	error = noErr;		// Carbon does not require LoadScrap() to be called
#endif
	if (error == noErr)
	{
		long		scrapResult = 0L;
		long		offset = 0L;
		ResType		validScrapType = '????';
		
		
		// figure out the scrap type and offset
		getValidScrapType(&scrapResult, &offset, &validScrapType);
		
		// remember the scrap count -- if it changes, we do an update in Clipboard_Idle()
		gScrapCount = Clipboard_GetFlavorCount();
		
		// now, draw the contents, if a supported type of data is available
		{
			Rect		clipArea = clipboardContentRect;
			Handle		scrapHandle = null;
			
			
		#if TARGET_API_MAC_OS8
			scrapHandle = InfoScrap()->scrapHandle;
		#else
			// under Carbon, the scrap handle cant be accessed; instead, a
			// handle is created by this block and the scrap is copied into it
			{
				ScrapRef	currentScrap = null;
				
				
				if (GetCurrentScrap(&currentScrap) == noErr)
				{
					Size		scrapSize = 0L;
					
					
					GetScrapFlavorSize(currentScrap, validScrapType, &scrapSize);
					scrapHandle = Memory_NewHandle(scrapSize);
					if (scrapHandle != null)
					{
						(OSStatus)GetScrapFlavorData(currentScrap, validScrapType, &scrapSize, *scrapHandle);
					}
				}
			}
		#endif
			
			// only do this if scroll bars are implemented
			//{
			//	SInt32		scrollBarSize;
			//	
			//	
			//	(OSStatus)GetThemeMetric(kThemeMetricScrollBarWidth, &scrollBarSize);
			//	clipArea.right -= scrollBarSize; // only needed if scroll bars are implemented
			//	clipArea.bottom -= scrollBarSize; // only needed if scroll bars are implemented
			//}
			ClipRect(&clipArea);
			
			// initially assume the data is not scaled
			gIsDataScaled = false;
			
			switch (validScrapType)
			{
				case kScrapFlavorTypePicture:
					drawPictureFromHandleAndOffset(&clipArea, scrapHandle, offset);
					break;
					
				case kScrapFlavorTypeText:
					{
						SInt8		oldState = '\0';
						
						
						// use the default terminals monospaced font for drawing text
						{
							TerminalPrefs**		termPrefsHandle = null;
							
							
							termPrefsHandle = Preferences_GetDefaultTerminal();
							if (termPrefsHandle != null)
							{
								TextFontByName((**termPrefsHandle).normalFont);
								TextSize((**termPrefsHandle).fontSize);
								Memory_DisposeHandle((Handle*)&termPrefsHandle);
							}
						}
						
						// use an appropriate theme color - since icon labels show up on
						// Finder window backgrounds, that color should work here
						{
							ColorPenState		state;
							
							
							ColorUtilities_GetColorAndPenState(&state);
							ColorUtilities_NormalizeColorAndPen();
							(OSStatus)SetThemeTextColor(kThemeTextColorIconLabel, inColorDepth,
														ColorUtilities_IsColorDevice(inTargetDevice));
							ColorUtilities_SetColorAndPenState(&state);
						}
						
						oldState = HGetState(scrapHandle);
						HLock(scrapHandle);
						TETextBox(*scrapHandle + offset, scrapResult, &clipArea, teJustLeft);
						HSetState(scrapHandle, oldState);
					}
					break;
				
				default:
					break;
			}
		#if TARGET_API_MAC_CARBON
			Memory_DisposeHandle(&scrapHandle);
		#endif
		}
		
		// save this for the Clipboard_Idle() routines benefit
		gScrapType = validScrapType;
		
		if (preservedClippingRegion != null) SetClip(preservedClippingRegion);
#if TARGET_API_MAC_OS8
		(SInt32)UnloadScrap();
#else
		// Carbon does not require UnloadScrap() to be called
#endif
	}
	
	Memory_DisposeRegion(&preservedClippingRegion);
}// contentUserPaneDeviceLoop


/*!
This method will draw the contents of the Clipboard
in the given controls rectangle.

(3.0)
*/
static pascal void
contentUserPaneDraw		(ControlRef		inControl,
						 SInt16			UNUSED_ARGUMENT(inPartCode))
{
	DeviceLoopDrawingUPP		upp = NewDeviceLoopDrawingUPP(contentUserPaneDeviceLoop);
	
	
	DeviceLoopClipAndDrawControl(inControl, upp);
	DisposeDeviceLoopDrawingUPP(upp), upp = null;
}// contentUserPaneDraw


/*!
This method will render a picture, given its data
and the desired boundary rectangle.  You provide
a data handle, which you can then offset from by
either zero or some other amount to reference the
correct picture data.

The picture is scaled, at its aspect ratio, to
fit the width and/or the height of the given
rectangle completely.  The global variable for
the scaling factor ("gScaledPercentage") is set
appropriately, and the "gIsDataScaled" flag is
set to true only if the picture is too large to
fit its desired rectangle.

(3.0)
*/
static OSErr
drawPictureFromHandleAndOffset	(const Rect*	inDrawingRect,
								 Handle			inDataSourceHandle,
								 const long		inSourceOffset)
{
	OSErr		result = noErr;
	PicHandle	tempPict = (PicHandle)Memory_NewHandle(sizeof(Picture));
	
	
	result = MemError();
	if (result == noErr)
	{
		Rect		drawingBounds = *inDrawingRect;
		
		
		// calculate the rectangle in which to draw, and then save the
		// picture header into the temporary handle
		{	
			PicPtr		picturePtr = null;
			UInt16		pictureHeight = 0,
						pictureWidth = 0,
						frameHeight = drawingBounds.bottom - drawingBounds.top,
						frameWidth = drawingBounds.right - drawingBounds.left;
			
			
			picturePtr = (PicPtr)((*inDataSourceHandle) + inSourceOffset);
			BlockMoveData((Ptr)picturePtr, (Ptr)*tempPict, sizeof(Picture));
			
			pictureHeight = picturePtr->picFrame.bottom - picturePtr->picFrame.top;
			pictureWidth = picturePtr->picFrame.right - picturePtr->picFrame.left;
			
			// if the picture frame will fit in the drawing area, no scaling is required
			{
				Rect		rawPictureFrame,
							rawDrawingFrame,
							junk;
				
				
				SetRect(&rawPictureFrame, 0, 0, pictureWidth, pictureHeight);
				SetRect(&rawDrawingFrame, 0, 0, frameWidth, frameHeight);
				UnionRect(&rawDrawingFrame, &rawPictureFrame, &junk);
				gIsDataScaled = (!EqualRect(&junk, &rawDrawingFrame));
			}
			
			if (gIsDataScaled)
			{
				float		wf = pictureWidth,
							hf = pictureHeight;
				float		ratio = 0.0f,
							ratioWidths = 0.0f,
							ratioHeights = 0.0f;
				
				
				ratio = wf / hf;
				wf = frameHeight;
				hf = pictureHeight;
				ratioHeights = wf / hf;
				wf = frameWidth;
				hf = pictureWidth;
				ratioWidths = wf / hf;
				
				// Make either the width or the height of the picture match the
				// maximum width of the drawing area.  This is based on which
				// dimension will fit the drawing area in such a way that the
				// limit as the other dimension approaches the maximum drawing
				// size is the maximum drawing size (that is, no clipping occurs).
				if ((pictureWidth > pictureHeight) && (ratioWidths < ratioHeights))
				{
					drawingBounds.right = drawingBounds.left + frameWidth;
					ratio = frameWidth / ratio;
					frameHeight = (UInt16)ratio;
					drawingBounds.bottom = drawingBounds.top + frameHeight;
					wf = 100 * frameHeight;
					hf = pictureHeight;
					ratio = wf / hf;
					gScaledPercentage = (UInt16)ratio;
				}
				else
				{
					drawingBounds.bottom = drawingBounds.top + frameHeight;
					ratio = ratio * frameHeight;
					frameWidth = (UInt16)ratio;
					drawingBounds.right = drawingBounds.left + frameWidth;
					wf = 100 * frameWidth;
					hf = pictureWidth;
					ratio = wf / hf;
					gScaledPercentage = (UInt16)ratio;
				}
			}
			else
			{
				// no changes are necessary
				drawingBounds.right = drawingBounds.left + pictureWidth;
				drawingBounds.bottom = drawingBounds.top + pictureHeight;
			}
		}
		
		// store into globals for the GetPicProc in preparation for the draw
		gScrapHandle = inDataSourceHandle;
		gCurrentOffset = inSourceOffset + sizeof(Picture);
		
		// install a GetPic proc
		{
			GrafPtr		currentPort = null;
			
			
			GetPort(&currentPort);
			
			SetStdCProcs(&gMyColorProcs);
			gMyColorProcs.getPicProc = gGetPICTDataUPP;
			gSavedProcsPtr = GetPortGrafProcs((CGrafPtr)currentPort);
			SetPortGrafProcs((CGrafPtr)currentPort, &gMyColorProcs);
			
			// draw the picture
			DrawPicture(tempPict, &drawingBounds);
			
			// remove the GetPic proc
			SetPortGrafProcs((CGrafPtr)currentPort, gSavedProcsPtr);
		}
		
		// update the bottom text area
		setScalingInformation();
		
		Memory_DisposeHandle(&((Handle)tempPict));
	}
	return result;
}// drawPictureFromHandleAndOffset


/*!
This is a replacement for the QuickDraw
bottleneck routine.

(3.0)
*/
static pascal void
getPICTData		(void*		inDataPtr,
				 short		inByteCount)
{ 
	long	longCount = inByteCount;
	
	
	BlockMoveData((*gScrapHandle) + gCurrentOffset, inDataPtr, longCount);
	gCurrentOffset += longCount;
}// getPICTData


/*!
This method checks the Scrap for a data type it can
render.  If none is found, this routine returns a
scrap type of '????'; otherwise, it returns the type
and GetScrap() result for the valid scrap type.

(3.0)
*/
static void
getValidScrapType	(long*		outScrapResult,
					 long*		outScrapOffset,
					 ResType*	outScrapType)
{
	register SInt16		i = 0;
	SInt32				offsetResult = 0L;
	SInt32				scrapResult = 0L;
	ResType				validScrapTypeResult = '????';
	ResType				scrapTypes[] = { kScrapFlavorTypeText, kScrapFlavorTypePicture, '????' };
#if TARGET_API_MAC_CARBON
	ScrapRef			currentScrap = null;
	ScrapFlavorFlags	currentScrapFlags = 0L;
#endif
	
	
#if TARGET_API_MAC_CARBON
	(OSStatus)GetCurrentScrap(&currentScrap);
#endif
	while (scrapTypes[i] != '????')
	{
	#if TARGET_API_MAC_OS8
		scrapResult = GetScrap(null, scrapTypes[i], &offsetResult);
		if (scrapResult > 0)
		{
			// if there is data, then a scrap of the specified type exists
			validScrapTypeResult = scrapTypes[i];
			break;
		}
	#else
		if (GetScrapFlavorFlags(currentScrap, scrapTypes[i], &currentScrapFlags) == noErr)
		{
			// if flags can be obtained with no error, then a scrap of the specified type exists
			validScrapTypeResult = scrapTypes[i];
			break;
		}
	#endif
		++i;
	}
	if (outScrapResult != null) *outScrapResult = scrapResult;
	if (outScrapOffset != null) *outScrapOffset = offsetResult;
	if (outScrapType != null) *outScrapType = validScrapTypeResult;
}// getValidScrapType


/*!
Converts the specified drawing into a standard
Mac OS picture format (PICT) and returns the
handle to the converted picture.

(2.6)
*/
static PicHandle
graphicToPICT	(short		inDrawingNumber)
{
	short		j = 0;
	PicHandle	tpic = null;
	Rect		trect;
	
	
	SetRect(&trect, 0, 0, 384, 384);
	j = VGnewwin(TEK_DEVICE_PICTURE, VGgetVS(inDrawingNumber));
	RGMPsize(&trect);
	VGzcpy(inDrawingNumber, j);
	
	tpic = OpenPicture(&trect);
	ClipRect(&trect);
	VGredraw(inDrawingNumber, j);
	ClosePicture();
	VGclose(j);
	
	return(tpic);
}// graphicToPICT


/*!
This method, which corresponds to WindowResizeResponderProcPtr
(see "InterfaceLib.h"), adjusts the contents of the specified
window based on its new dimensions.  The change between the
old dimensions and the new dimensions is given, where
negative values indicate that the size in one or both
directions has decreased.  The specified window must have
an associated WindowInfoRef, because that reference is used
to determine the exact window descriptor (and thus how to
respond to the resize).

Currently, this internal method only handles the clipboard
window.  The file "DialogAdjust.c" handles resizable dialogs.

(3.0)
*/
static pascal Boolean
handleNewSize	(const WindowRef		inWindowRef,
				 const SInt32			inDeltaSizeX,
				 const SInt32			inDeltaSizeY,
				 const SInt32			UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_false;
	
	
	DialogAdjust_BeginControlAdjustment(inWindowRef);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gClipboardWindowHeader, inDeltaSizeX);
	//DialogAdjust_AddControlAdjustment(kControlAdjustmentMoveV, gClipboardScrollBarH, inDeltaSizeY);
	//DialogAdjust_AddControlAdjustment(kControlAdjustmentMoveH, gClipboardScrollBarV, inDeltaSizeX);
	//DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gClipboardScrollBarH, inDeltaSizeX);
	//DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeV, gClipboardScrollBarV, inDeltaSizeY);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gClipboardFooterPlacard, inDeltaSizeX);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gClipboardFooterPlacardText, inDeltaSizeX);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentMoveV, gClipboardFooterPlacard, inDeltaSizeY);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gClipboardContentUserPane, inDeltaSizeX);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeV, gClipboardContentUserPane, inDeltaSizeY);
	DialogAdjust_EndAdjustment(inDeltaSizeX, inDeltaSizeY);
	
	return result;
}// handleNewSize


/*!
Copies the specified PICT (QuickDraw picture) data
to the clipboard.

(3.0)
*/
static void
pictureToScrap	(Handle		inPictureData)
{
	SInt8	hState = HGetState(inPictureData);
	
	
	HLock((Handle)inPictureData);
#if TARGET_API_MAC_OS8
	(OSStatus)ZeroScrap();
	(long)PutScrap(GetHandleSize(inPictureData), kScrapFlavorTypePicture, (Ptr)*inPictureData);
#else
	(OSStatus)ClearCurrentScrap();
	{
		ScrapRef		currentScrap = null;
		
		
		if (GetCurrentScrap(&currentScrap) == noErr)
		{
			(OSStatus)PutScrapFlavor(currentScrap, kScrapFlavorTypePicture,
										kScrapFlavorMaskNone, GetHandleSize(inPictureData),
										*inPictureData);
		}
	}
#endif
	HSetState(inPictureData, hState);
}// pictureToScrap


/*!
To set up the contents of the top text area of
the clipboard window, use this method.  The
text varies based on the presence of data on
the clipboard, as well as the kind of data and
its size.

(3.0)
*/
static void
setDataTypeInformation	(void)
{
	long		scrapResult = 0L;
	long		offset = 0L;
	ResType		validScrapType = '????';
	
	
	// figure out the scrap type and offset
	getValidScrapType(&scrapResult, &offset, &validScrapType);
	
	// create strings describing the contents
	{
		Str255		theString,
					clipboardTypeString;
		Boolean		clipboardEmpty = false;
		
		
		switch (validScrapType)
		{
			case kScrapFlavorTypePicture:
				GetIndString(clipboardTypeString, rStringsClipboardData, siClipboardTypePicture);
				break;
				
			case kScrapFlavorTypeText:
				GetIndString(clipboardTypeString, rStringsClipboardData, siClipboardTypeText);
				break;
				
			default:
				if (Clipboard_GetFlavorCount() == 0)
				{
					GetIndString(theString, rStringsClipboardData, siClipboardEmpty);
					clipboardEmpty = true;
				}
				else GetIndString(clipboardTypeString, rStringsClipboardData, siClipboardTypeUnrecognized);
				break;
		}
		
		unless (clipboardEmpty)
		{
			Str255		aboutString,
						sizeString,
						unitsString;
			
			
			GetIndString(theString, rStringsClipboardData, siClipboardContents);
			PLstrcpy(aboutString, EMPTY_PSTRING);
			// construct a size string, with appropriate units and the word about, if necessary
			{
				Size		size = 0L;
				SInt16		index = siClipboardBytes; // string index of sizes unit string (bytes, K, MB, etc.)
				
				
			#if TARGET_API_MAC_OS8
				size = (long)InfoScrap()->scrapSize;
			#else
				{
					ScrapRef		scrap = null;
					
					
					if (GetCurrentScrap(&scrap) == noErr)
					{
						(OSStatus)GetScrapFlavorSize(scrap, Clipboard_GetContentType(), &size);
					}
				}
			#endif
				if (size > INTEGER_MEGABYTES(1))
				{
					// size is approximate, so say so in the string
					GetIndString(aboutString, rStringsClipboardData, siClipboardAbout);
					size /= INTEGER_MEGABYTES(1);
					index = siClipboardMB;
				}
				else if (size > INTEGER_KILOBYTES(1))
				{
					// size is approximate, so say so in the string
					GetIndString(aboutString, rStringsClipboardData, siClipboardAbout);
					size /= INTEGER_KILOBYTES(1);
					index = siClipboardK;
				}
				else if (size == 1)
				{
					index = siClipboardByte; // use singular form "byte" when the size is exactly 1 byte
				}
				NumToString(size, sizeString);
				GetIndString(unitsString, rStringsClipboardData, index);
			}
			StringUtilities_PMetaSetDefaults();
			{
				StringPtr		first = null,
								second = null;
				Boolean			unitsBeforeQuantities = false; // TEMP - localize
				
				
				if (unitsBeforeQuantities)
				{
					first = unitsString;
					second = sizeString;
				}
				else
				{
					first = sizeString;
					second = unitsString;
				}
				StringUtilities_PBuild(theString, clipboardTypeString, aboutString, first, second);
			}
		}
		
		// figure out what font to use for the window header text, and set its size to fit
		// (this routine also sets the text contained in the control)
		(UInt16)Localization_SetUpSingleLineTextControl(gClipboardWindowHeaderText, theString, kThemeViewsFont);
	}
}// setDataTypeInformation


/*!
This routine sets up a global rectangle that
defines the full-screen zoom size for the
clipboard window.  Prior to option-zooming
the window, this routine should be called to
update the global rectangle to use the
monitor that most of the clipboard window is
within.

(3.0)
*/
static void
setFullScreenRectangle	(void)
{
	Rect			screenRect,
					structureRect,
					contentRect;
	WindowRef		clipboardWindow = gClipboard;
	float			fh = 0.0,
					fw = 0.0;
	register short	verticalExtra = 40;
	
	
	// get important rectangles
	MacLook_GetWindowDeviceGrayRect(clipboardWindow, &screenRect);
	MacLook_GetWindowRegionBounds(clipboardWindow, kWindowContentRgn, &contentRect);
	MacLook_GetWindowRegionBounds(clipboardWindow, kWindowStructureRgn, &structureRect);
	
	// define the "full screen" zoomed size
	fh = 1.0 * (screenRect.bottom - screenRect.top - INTEGER_DOUBLED(verticalExtra));
	fw = 0.70 * (screenRect.right - screenRect.left);
	SetRect(&gFullScreenClipboard,
				screenRect.left + (contentRect.left - structureRect.left) + 4, screenRect.top + verticalExtra, 0, 0);
	SetRect(&gFullScreenClipboard, gFullScreenClipboard.left, gFullScreenClipboard.top,
									gFullScreenClipboard.left + (short)fw,
									gFullScreenClipboard.top + (short)fh);
}// setFullScreenRectangle


/*!
To set up the contents of the bottom text area
of the clipboard window, use this method.  The
text is blank unless the clipboard contains a
picture that has been scaled smaller than its
actual size.

(3.0)
*/
static void
setScalingInformation	(void)
{
	// if the data is scaled, say so; otherwise, remove the footer text completely
	if (gIsDataScaled)
	{
		Str63		footerString;
		Str31		numString;
		
		
		GetIndString(footerString, rStringsClipboardData, siClipboardScalingFactor);
		NumToString((long)gScaledPercentage, numString);
		StringUtilities_PMetaSet("\p??", "\pXX", "\pYY", "\pZZ", StrLength("\p??"));
		StringUtilities_PBuild(footerString, numString, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
		(OSErr)SetControlData(gClipboardFooterPlacardText, kControlEditTextPart, kControlStaticTextTextTag,
								StrLength(footerString), (Ptr)(1 + footerString));
	}
	else
	{
		(OSErr)SetControlData(gClipboardFooterPlacardText, kControlEditTextPart, kControlStaticTextTextTag,
								StrLength(EMPTY_PSTRING), (Ptr)(1 + EMPTY_PSTRING));
	}
}// setScalingInformation


/*!
To set the initial arrangement of the controls
for the clipboard window, use this method.
Subsequent resizings will simply add or subtract
deltas from the initial control rectangles, so
tbe initial arrangement calculation only needs
to be done once.

(3.0)
*/
static void
setUpControls	(const WindowRef		inWindowRef)
{
	Rect		clipboardContentRect;
	SInt32		scrollBarSize = 0L;
	
	
	// get scroll bar height
	(OSStatus)GetThemeMetric(kThemeMetricScrollBarWidth, &scrollBarSize);
	
	// invisibly rearrange controls
	(OSErr)SetControlVisibility(gClipboardRoot, false/* visibility */, false/* draw */);
	
	// calculate the area at the top that says what is on the clipboard
	// and the area that displays the clipboard data
	{
		enum
		{
			kHeaderTextInsetH = 4, // pixels horizontally inward that the window header text is from the edges of the header
			kHeaderTextInsetV = 4 // pixels vertically inward that the window header text is from the edges of the header
		};
		Rect	topRect,
				controlRect;
		
		
		GetPortBounds(GetWindowPort(inWindowRef), &topRect);
		GetControlBounds(gClipboardWindowHeaderText, &controlRect);
		topRect.bottom = (controlRect.bottom - controlRect.top) + (2 * kHeaderTextInsetV);
		
		// resize the window header and its text control
		InsetRect(&topRect, kHeaderTextInsetH, kHeaderTextInsetV);
		MoveControl(gClipboardWindowHeaderText, topRect.left, topRect.top);
		SizeControl(gClipboardWindowHeaderText, topRect.right - topRect.left, topRect.bottom - topRect.top);
		InsetRect(&topRect, -kHeaderTextInsetH, -kHeaderTextInsetV);
		MoveControl(gClipboardWindowHeader, topRect.left - 1, topRect.top - 1);
		SizeControl(gClipboardWindowHeader, topRect.right - topRect.left + 2, topRect.bottom - topRect.top + 2);
		
		// calculate the part *not* in the top area
		clipboardContentRect = topRect;
		clipboardContentRect.top = topRect.bottom + 1;
		GetPortBounds(GetWindowPort(inWindowRef), &topRect); // "topRect" has a new meaning here...
		clipboardContentRect.bottom = topRect.bottom - scrollBarSize + 1;
	}
	
	// arrange scroll bars
	//(OSStatus)GetThemeMetric(kThemeMetricScrollBarWidth, &scrollBarSize);
	//MoveControl(gClipboardScrollBarH, clipboardContentRect.left - 1,
	//									clipboardContentRect.bottom - scrollBarSize + 1);
	//SizeControl(gClipboardScrollBarH,
	//				clipboardContentRect.right - clipboardContentRect.left - scrollBarSize + 1 + 2/* 2 * frame width */,
	//				scrollBarSize);
	//MoveControl(gClipboardScrollBarV, clipboardContentRect.right - scrollBarSize + 1,
	//										clipboardContentRect.top - 1);
	//SizeControl(gClipboardScrollBarV, scrollBarSize,
	//				clipboardContentRect.bottom - clipboardContentRect.top - scrollBarSize + 1 + 2/* 2 * frame width */);
	
	// arrange other controls
	{
		enum
		{
			kPlacardTextInsetH = 2, // pixels horizontally inward that the window footer text is from the placard edges
			kPlacardTextInsetV = 2 // pixels vertically inward that the window footer text is from the placard edges
		};
		
		
		MoveControl(gClipboardContentUserPane, clipboardContentRect.left, clipboardContentRect.top);
		SizeControl(gClipboardContentUserPane, clipboardContentRect.right - clipboardContentRect.left,
						clipboardContentRect.bottom - clipboardContentRect.top);
		MoveControl(gClipboardFooterPlacard, clipboardContentRect.left - 1, clipboardContentRect.bottom);
		SizeControl(gClipboardFooterPlacard,
					clipboardContentRect.right - clipboardContentRect.left - scrollBarSize + 1 + 2, scrollBarSize);
		InsetRect(&clipboardContentRect, kPlacardTextInsetH, kPlacardTextInsetV);
		MoveControl(gClipboardFooterPlacardText, clipboardContentRect.left,
						clipboardContentRect.bottom + 2 * kPlacardTextInsetV);
		SizeControl(gClipboardFooterPlacardText,
						clipboardContentRect.right - clipboardContentRect.left - scrollBarSize + 1 + 2,
						scrollBarSize - 2 * kPlacardTextInsetH);
		InsetRect(&clipboardContentRect, -kPlacardTextInsetH, -kPlacardTextInsetV);
	}
}// setUpControls


/*!
Copies the specified plain text data to the
clipboard.

(3.0)
*/
static void
textToScrap		(Handle		inTextHandle)
{
	Size		actualSize = 0L;
	
	
#if TARGET_API_MAC_OS8
	(OSStatus)ZeroScrap();
	(long)PutScrap(GetHandleSize(inTextHandle), kScrapFlavorTypeText, *inTextHandle);
#else
	(OSStatus)ClearCurrentScrap();
	{
		ScrapRef		currentScrap = null;
		
		
		if (GetCurrentScrap(&currentScrap) == noErr)
		{
			(OSStatus)PutScrapFlavor(currentScrap, kScrapFlavorTypeText,
										kScrapFlavorMaskNone, GetHandleSize(inTextHandle),
										*inTextHandle);
		}
	}
#endif
}// textToScrap