/*###############################################################

	ColorBox.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.
	
	You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <ControlDefinitions.h>
#include <Dialogs.h>
#include <Menus.h>
#include <Palettes.h>
#include <Resources.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "GeneralResources.h"
#include "MenuResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "ColorBox.h"
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "Preferences.h"
#include "RealScreen.h"
#include "ResourceFileManager.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



//
// types
//

/*!
A pointer to this data is attached to a controls
reference constant so that it is available within
control callback routines.
*/
struct ColorBoxData
{
	ControlUserPaneDrawUPP			uppDraw;			//!< routine specifying how to render the color box
	ControlUserPaneHitTestUPP		uppHitTest;			//!< routine that determines if a point is within an active color box
	ControlUserPaneTrackingUPP		uppTracking;		//!< routine that continuously follows a point (i.e. mouse click) in a color box
	RGBColor						displayedColor;		//!< which color (the value) the color box displays
	ColorBoxChangeNotifyProcPtr		notifyProcPtr;		//!< routine, if any, to call when the value is changed
	void const*						contextPtr;			//!< caller-defined context to pass to the notification routine
};
#ifndef __cplusplus
typedef struct ColorBoxData		ColorBoxData;
#endif
typedef ColorBoxData*			ColorBoxDataPtr;



//
// internal methods
//

static pascal void				colorBoxUserPaneDeviceLoop				(short					inColorDepth,
																		 short					inDeviceFlags,
																		 GDHandle				inTargetDevice,
																		 long					inControl);
static pascal void				colorBoxUserPaneDrawProc				(ControlRef				inControl,
																		 ControlPartCode		inPart);
static pascal ControlPartCode	colorBoxUserPaneHitTestProc				(ControlRef				inControl,
																		 Point					inLocalMouse);
static pascal ControlPartCode	colorBoxUserPaneTrackingProc			(ControlRef				inControl,
																		 Point					inGlobalMouse,
																		 ControlActionUPP		inIgnoredActionUPP);



//
// public methods
//

/*!
Assigns a user pane control the drawing, hit test, and
tracking procedures for a color box, and attaches data
to it using the control reference.

Since callbacks are installed, the control features
of the specified control must be set to allow the required
callbacks.  You generally create a control with the
special control ID "kIDColorBox" to ensure that the
control features and dimensions are set up correctly, and
then invoke this routine.

Be sure to invoke ColorBox_DetachFromUserPaneControl()
when you are finished with the color box, to dispose of
data allocated by invoking this routine.

IMPORTANT:	The specified controls reference constant
			is reserved.

(3.0)
*/
void
ColorBox_AttachToUserPaneControl	(ControlRef			inoutUserPaneControl,
									 RGBColor const*	inInitialColorOrNull)
{
	ColorBoxDataPtr		dataPtr = (ColorBoxDataPtr)Memory_NewPtr(sizeof(ColorBoxData));
	
	
	if (dataPtr != null)
	{
		// install callback procedures (the caller must dispose these later)
		dataPtr->uppDraw = NewControlUserPaneDrawUPP(colorBoxUserPaneDrawProc);
		dataPtr->uppHitTest = NewControlUserPaneHitTestUPP(colorBoxUserPaneHitTestProc);
		dataPtr->uppTracking = NewControlUserPaneTrackingUPP(colorBoxUserPaneTrackingProc);
		
		(OSStatus)SetControlData(inoutUserPaneControl, kControlEntireControl, kControlUserPaneDrawProcTag,
									sizeof(dataPtr->uppDraw), &dataPtr->uppDraw);
		(OSStatus)SetControlData(inoutUserPaneControl, kControlEntireControl, kControlUserPaneHitTestProcTag,
									sizeof(dataPtr->uppHitTest), &dataPtr->uppHitTest);
		(OSStatus)SetControlData(inoutUserPaneControl, kControlEntireControl, kControlUserPaneTrackingProcTag,
									sizeof(dataPtr->uppTracking), &dataPtr->uppTracking);
		dataPtr->notifyProcPtr = null;
		
		// initialize color
		if (inInitialColorOrNull != null) dataPtr->displayedColor = *inInitialColorOrNull;
		else dataPtr->displayedColor.red = dataPtr->displayedColor.green = dataPtr->displayedColor.blue = 0;
		
		// attach data to control
		SetControlReference(inoutUserPaneControl, REINTERPRET_CAST(dataPtr, SInt32));
	}
}// AttachToUserPaneControl


/*!
Calls ColorBox_AttachToUserPaneControl() for a dialog
items Appearance control.

(3.0)
*/
void
ColorBox_AttachToUserPaneDialogItem		(DialogRef			inDialog,
										 DialogItemIndex	inItemIndex,
										 RGBColor const*	inInitialColorOrNull)
{
	ControlRef		userPane = null;
	
	
	if (GetDialogItemAsControl(inDialog, inItemIndex, &userPane) == noErr) ColorBox_AttachToUserPaneControl(userPane, inInitialColorOrNull);
}// AttachToUserPaneDialogItem


/*!
Disposes of a user pane controls drawing, hit test, and
tracking procedures, initially installed using one of the
ColorBox_AttachToUserPane...() routines.  On input, provide
a structure filled in with UPP pointers (originally defined
by the creation call); on output, the UPPs are disposed
and the pointers in the structure are set to "null".

(3.0)
*/
void
ColorBox_DetachFromUserPaneControl		(ControlRef		inControl)
{
	ColorBoxDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inControl), ColorBoxDataPtr);
	
	
	if (dataPtr != null)
	{
		// dispose of callback data
		DisposeControlUserPaneDrawUPP(dataPtr->uppDraw), dataPtr->uppDraw = null;
		DisposeControlUserPaneHitTestUPP(dataPtr->uppHitTest), dataPtr->uppHitTest = null;
		DisposeControlUserPaneTrackingUPP(dataPtr->uppTracking), dataPtr->uppTracking = null;
		
		// dispose of color box data
		Memory_DisposePtr((Ptr*)&dataPtr);
	}
}// DetachFromUserPaneControl


/*!
Calls ColorBox_DetachFromUserPaneControl() for
a dialog items Appearance control.

(3.0)
*/
void
ColorBox_DetachFromUserPaneDialogItem	(DialogRef			inDialog,
										 DialogItemIndex	inItemIndex)
{
	ControlRef		userPane = null;
	
	
	if (GetDialogItemAsControl(inDialog, inItemIndex, &userPane) == noErr) ColorBox_DetachFromUserPaneControl(userPane);
}// DetachFromUserPaneDialogItem


/*!
Returns the color displayed by a color box.

(3.0)
*/
void
ColorBox_GetColor	(ControlRef		inControl,
					 RGBColor*		outColorPtr)
{
	ColorBoxDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inControl), ColorBoxDataPtr);
	
	
	if ((outColorPtr != null) && (dataPtr != null))
	{
		*outColorPtr = dataPtr->displayedColor;
	}
}// GetColor


/*!
Returns a new color box control, optionally displaying
the specified initial color.  This is a quick way to
create a color box; the control returned is already
the right size, has the proper control features, and
has all necessary user pane callbacks installed (in
other words, it works with no extra work on your part).
It is not necessary to invoke either of the
ColorBox_AttachToUserPane...() methods, but you should
still invoke ColorBox_DetachFromUserPaneControl() when
you are finished with this new control.

(3.0)
*/
ControlRef
ColorBox_GetNewControl	(WindowRef			inOwningWindow,
						 RGBColor const*	inInitialColorOrNull)
{
	ControlRef		result = GetNewControl(kIDColorBox, inOwningWindow);
	
	
	if (result != null) ColorBox_AttachToUserPaneControl(result, inInitialColorOrNull);
	return result;
}// GetNewControl


/*!
Changes the color displayed by a color box.

(3.0)
*/
void
ColorBox_SetColor	(ControlRef			inControl,
					 RGBColor const*	inColorPtr)
{
	ColorBoxDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inControl), ColorBoxDataPtr);
	
	
	if ((inColorPtr != null) && (dataPtr != null))
	{
		dataPtr->displayedColor = *inColorPtr;
		if (dataPtr->notifyProcPtr != null) (*(dataPtr->notifyProcPtr))(inControl, &dataPtr->displayedColor, dataPtr->contextPtr);
	}
}// SetColor


/*!
Assigns a notification procedure to a color box.  The
specified routine will automatically be called if the
color displayed by the box changes for any reason.
For modeless windows, it may be necessary to use a
notification routine, if you want to respond right away
to the users changes.

(3.0)
*/
void
ColorBox_SetColorChangeNotifyProc	(ControlRef						inControl,
									 ColorBoxChangeNotifyProcPtr	inNotifyProcPtr,
									 void const*					inContextPtr)
{
	ColorBoxDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inControl), ColorBoxDataPtr);
	
	
	if (dataPtr != null)
	{
		dataPtr->notifyProcPtr = inNotifyProcPtr;
		dataPtr->contextPtr = inContextPtr;
	}
}// SetColorChangeNotifyProc


/*!
Displays a standard Color Picker dialog box, asking the
user to choose a color.  If the user accepts the color,
this routine automatically updates the displayed color
of the color box and "true" is returned; otherwise,
"false" is returned.

You typically invoke this routine in response to a
control click in a color box control.  For 99.99% of
cases, you will not need to do anything else in response
to a control click in a color box.  Note that you can
use ColorBox_GetColor() to determine the displayed color.

(3.0)
*/
Boolean
ColorBox_UserSetColor	(ControlRef		inControl)
{
	ColorBoxDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inControl), ColorBoxDataPtr);
	Boolean				result = false;
	
	
	if (dataPtr != null)
	{
		Str255					askColorString;
		PickerMenuItemInfo		editMenuInfo;
		
		
		GetIndString(askColorString, rStringsMiscellaneous, siPickNewColor);
		Menus_FillInColorPickerMenuItemInfo(&editMenuInfo);
		DeactivateFrontmostWindow();
		result = ColorUtilities_ColorChooserDialogDisplay
								(askColorString, &dataPtr->displayedColor/* input */, &dataPtr->displayedColor/* output */,
								true/* is modal */, NewUserEventUPP(EventLoop_HandleColorPickerUpdate),
								&editMenuInfo);
		RestoreFrontmostWindow();
		if ((result) && (dataPtr->notifyProcPtr != null))
		{
			(*(dataPtr->notifyProcPtr))(inControl, &dataPtr->displayedColor, dataPtr->contextPtr);
		}
	}
	return result;
}// UserSetColor


//
// internal methods
//
#pragma mark -

/*!
This method, of standard DeviceLoopDrawingUPP form,
handles drawing of a color boxs contents correctly
no matter what the device characteristics are.

This routine, although based on the user-item-based
implementation of version 2.6, has been completely
updated in 3.0 to use Appearance APIs to make color
boxes really look like bevel buttons.  Also, these
controls use hit-test routines to use the proper
theme sounds, so youd be hard pressed to tell the
difference between the revamped color box and a
real bevel button!

(3.0)
*/
static pascal void
colorBoxUserPaneDeviceLoop	(short				inColorDepth,
							 short				UNUSED_ARGUMENT(inDeviceFlags),
							 GDHandle			inTargetDevice,
							 long				inControl)
{
	Rect				controlRect;
	ColorPenState		state;
	GrafPtr				oldPort = null;
	ControlRef			control = (ControlRef)inControl;
	
	
	GetControlBounds(control, &controlRect);
	
	ColorUtilities_GetColorAndPenState(&state);
	ColorUtilities_NormalizeColorAndPen();
	
	// set up the control background and foreground colors
	(OSErr)SetUpControlBackground(control, inColorDepth, ColorUtilities_IsColorDevice(inTargetDevice));
	
	EraseRect(&controlRect);
	if (FlagManager_Test(kFlagAppearance1_1API))
	{
		ThemeButtonDrawInfo		info;
		
		
		// given the more-advanced API, draw a real bevel button
		if (IsControlActiveAndHilited(control)) info.state = kThemeStatePressed;
		else if (IsControlActive(control)) info.state = kThemeStateActive;
		else info.state = kThemeStateInactive;
		info.value = kThemeButtonOff;
		info.adornment = kThemeAdornmentNone;
		(OSStatus)DrawThemeButton(&controlRect, kThemeSmallBevelButton, &info/* new info */, (ThemeButtonDrawInfo*)null/* old info */,
									(ThemeEraseUPP)null, (ThemeButtonDrawUPP)null, 0L/* user data */);
	}
	else
	{
		// lacking an API for drawing buttons, draw a placard instead, which in most themes looks like a bevel button
		(OSStatus)DrawThemePlacard(&controlRect, IsControlActive(control) ? kThemeStateActive : kThemeStateInactive);
	}
	InsetRect(&controlRect, 3, 3);
	
	// paint the color
	{
		RGBColor	oldForeColor,
					boxColor;
		
		
		GetForeColor(&oldForeColor);
		ColorBox_GetColor(control, &boxColor);
		RGBForeColor(&boxColor);
		PaintRect(&controlRect);
		UseLighterColors();
		FrameRect(&controlRect);
		RGBForeColor(&oldForeColor);
	}
	
	ColorUtilities_SetColorAndPenState(&state);
}// colorBoxUserPaneDeviceLoop


/*!
Draws a color box in the specified controls
rectangle, using new, Appearance-savvy user
panes and DeviceLoop().

(3.0)
*/
static pascal void
colorBoxUserPaneDrawProc	(ControlRef			inControl,
							 ControlPartCode	UNUSED_ARGUMENT(inPartCode))
{
	DeviceLoopDrawingUPP		upp = NewDeviceLoopDrawingUPP(colorBoxUserPaneDeviceLoop);
	
	
	DeviceLoopClipAndDrawControl(inControl, upp);
	DisposeDeviceLoopDrawingUPP(upp), upp = null;
}// colorBoxUserPaneDrawProc


/*!
This method, of proper ControlUserPaneHitTestUPP
form, determines if the mouse is inside a color
box button (kControlButtonPart) or not
(kControlNoPart).

(3.0)
*/
static pascal ControlPartCode
colorBoxUserPaneHitTestProc		(ControlRef		inControl,
								 Point			inLocalMouse)
{
	ControlPartCode		result = kControlNoPart;
	Rect				controlRect;
	
	
	GetControlBounds(inControl, &controlRect);
	if (PtInRect(inLocalMouse, &controlRect)) result = kControlButtonPart;
	
	return result;
}// colorBoxUserPaneHitTestProc


/*!
Determines the proper control part corresponding
to the specified mouse location, for a color box.
The only supported parts are inside the button
(kControlButtonPart) or outside (kControlNoPart).
This routine continues to loop until the mouse is
released, and returns the control part that the
mouse last intersects when released.  Theme sounds
and a new pop-up menu (after a delay) are handled.

The control action routine is ignored (but for
future reference, valid values are null (0), -1,
or a valid function pointer!).

(3.0)
*/
static pascal ControlPartCode
colorBoxUserPaneTrackingProc	(ControlRef			inControl,
								 Point				UNUSED_ARGUMENT(inLocalMouse),
								 ControlActionUPP	UNUSED_ARGUMENT(inIgnoredActionUPP))
{
	ControlPartCode		result = kControlNoPart,
						oldPart = kControlNoPart;
	Rect				controlRect;
	Boolean				firstHit = false,
						wasInside = false,
						done = false;
	UInt32				clickTime = TickCount();
	Point				localMouse;
	CGrafPtr			oldPort = null;
	GDHandle			oldDevice = null;
	
	
	GetGWorld(&oldPort, &oldDevice);
	SetPortWindowPort(GetControlOwner(inControl));
	GetControlBounds(inControl, &controlRect);
	while ((!done) && StillDown())
	{
		GetMouse(&localMouse);
		if (PtInRect(localMouse, &controlRect))
		{
			result = kControlButtonPart;
			
			if (!wasInside) // avoid playing theme sounds repeatedly
			{
				wasInside = true;
				if (FlagManager_Test(kFlagAppearance1_1API))
				{
					(OSStatus)PlayThemeSound((firstHit) ? kThemeSoundBevelPress : kThemeSoundBevelEnter);
				}
			}
			
			// if tracking lasts for more than a second, display a menu
			if ((TickCount() - clickTime) > 30/* arbitrary - how long the user must pause on mouse-down before a menu appears */)
			{
				MenuHandle		menu = GetMenu(kPopUpMenuIDColorBox);
				long			menuResult = 0L;
				Point			localPoint;
				
				
				SetPt(&localPoint, controlRect.left + 1, controlRect.top + 1);
				LocalToGlobal(&localPoint);
				
				InsertMenu(menu, kInsertHierarchicalMenu); // add to pop-up menu list
				menuResult = PopUpMenuSelect(menu, localPoint.v, localPoint.h, 1/* selected item */);
				if ((menuResult & 0xFFFF0000) != 0)
				{
					enum
					{
						kColorBoxMenuItemUseColorPicker = 1,
						//
						kColorBoxMenuItemFirstColor = 3
					};
					short		theItem = menuResult & 0xFFFF;
					
					
					switch (theItem)
					{
						case kColorBoxMenuItemUseColorPicker:
							// the use color picker item was chosen, which is just like clicking the button
							result = kControlButtonPart;
							break;
						
						default:
							// a specific color was chosen; explicitly change the color data here, instead, and bypass normal behavior
							result = kControlNoPart;
							{
								RGBColor			color;
								PaletteHandle		paletteHandle = null;
								SInt16				oldResFile = CurResFile();
								SInt16				colorIndices[] =
													{
														// fill this array with the palette color indices corresponding to items in the menu
														kANSIColorDescriptorBlack,
														kANSIColorDescriptorBlackEmphasized,
														kANSIColorDescriptorWhite,
														kANSIColorDescriptorWhiteEmphasized,
														kANSIColorDescriptorYellow,
														kANSIColorDescriptorRed,
														kANSIColorDescriptorRedEmphasized,
														kANSIColorDescriptorMagenta,
														kANSIColorDescriptorMagentaEmphasized,
														kANSIColorDescriptorBlue,
														kANSIColorDescriptorBlueEmphasized,
														kANSIColorDescriptorGreen,
														kANSIColorDescriptorGreenEmphasized,
														kANSIColorDescriptorCyan,
														kANSIColorDescriptorCyanEmphasized,
														kANSIColorDescriptorYellowEmphasized
													};
								
								
								// find the color data to change
								ColorBox_GetColor(inControl, &color);
								
								// get the standard colors
								ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
								paletteHandle = GetNewPalette(ANSICOLORS_DEFAULTID);
								
								// now figure out which color the user selected from the menu, and obey...
								GetEntryColor(paletteHandle, colorIndices[theItem - kColorBoxMenuItemFirstColor], &color);
								ColorBox_SetColor(inControl, &color);
								DisposePalette(paletteHandle), paletteHandle = null;
								UseResFile(oldResFile);
							}
							break;
					}
				}
				else
				{
					// if no menu item is selected, cancel all action on the control
					result = kControlNoPart;
				}
				DeleteMenu(kPopUpMenuIDColorBox);
				DisposeMenu(menu), menu = null;
				done = true; // stop after displaying the menu
			}
		}
		else
		{
			result = kControlNoPart;
			
			// reset the click time - the pop-up-menu-display delay
			// is accounted for only while the mouse is in the button
			clickTime = 0L;
			
			if (wasInside) // avoid playing theme sounds repeatedly
			{
				wasInside = false;
				if (FlagManager_Test(kFlagAppearance1_1API))
				{
					(OSStatus)PlayThemeSound(kThemeSoundBevelExit);
				}
			}
		}
		
		// redraw control appropriately
		if (oldPart != result) HiliteControl(inControl, result);
		oldPart = result;
	}
	SetGWorld(oldPort, oldDevice);
	HiliteControl(inControl, kControlNoPart);
	
	return result;
}// colorBoxUserPaneTrackingProc