/*###############################################################

	ColorUtilities.c
	
	Utility routines to save and restore the graphics state of
	a graphics port, and to seamlessly use the most powerful
	version of the Mac OS color picker available.
	
	Color Pen State portions of this file are from the
	OS Technologies group at Apple Computer Inc. (source: SDK
	for Appearance 1.0), and are written by Ed Voas.  Copyright
	 1997 by Apple Computer, Inc., all rights reserved.
	
	Interface Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <ColorPicker.h>
#include <Gestalt.h>
#include <PLStringFuncs.h>

// library includes
#include "InterfaceLib.h"



//
// public methods
//

/*!
This universal routine allows you to use the most
powerful version of the Mac OS color picker that
is available.  There is no support, however, for
special callback routines when using this method.

The dialog assumes the application is not ColorSync
aware, detachs the dialog body from the choices,
uses a system dialog, and is centered on the main
screen.

Since the dialog prefers to be movable, you should
(but do not have to) provide a routine that can
respond to update events for your application windows,
should the user cause a region of a background window
to require updating.  If you do not provide an update
routine, the dialog is not made movable.

The frontmost window is automatically deactivated (if
possible, that is, if it has an Appearance-based
embedding hierarchy) before the dialog appears, and
is restored to its normal state before this routine
returns.

If the user chooses a new color, true is returned;
otherwise, false is returned.

(1.0)
*/
Boolean
ColorUtilities_ColorChooserDialogDisplay	(ConstStr255Param		inPrompt,
											 const RGBColor*		inColor,
											 RGBColor*				outColor,
											 Boolean				inIsModal,
											 UserEventUPP			inUserEventProc,
											 PickerMenuItemInfo*	inEditMenuInfo)
{
	Boolean		result = false;
	UInt8		majorRev = 0,
				minorRev = 0;
	Point		dialogLocation;
	OSErr		error = noErr;
	UInt32		kColorUtilitiesColorPickerFlags =	((inUserEventProc != null) ? kColorPickerDialogIsMoveable : 0) |
													((inIsModal) ? kColorPickerDialogIsModal : 0) |
													kColorPickerCanModifyPalette |
													kColorPickerCanAnimatePalette;
	
	
	{
		long		version = 0L;
		
		
		error = Gestalt(gestaltColorPickerVersion, &version);
		majorRev = Releases_GetMajorRevisionForVersion(version);
		minorRev = Releases_GetMinorRevisionForVersion(version);
	}
	
	SetPt(&dialogLocation, 0, 0); // causes the Color Picker to get centered on the main screen
	
	// the Color Picker 2.1 dialog is the best: use it if it's available
	if (((majorRev == 0x02) && (minorRev >= 0x01)) || (majorRev > 0x02))
	{
		// Color Picker 2.1 or later
		NColorPickerInfo	info;
		
		
		info.theColor.profile = null;
		info.theColor.color.rgb.red = inColor->red;
		info.theColor.color.rgb.green = inColor->green;
		info.theColor.color.rgb.blue = inColor->blue;
		info.dstProfile = null;
		info.flags = kColorUtilitiesColorPickerFlags;
		info.placeWhere = kCenterOnMainScreen;
		info.dialogOrigin = dialogLocation;
		info.pickerType = 0L;
		info.eventProc = inUserEventProc;
		info.colorProc = null;
		info.colorProcData = 0L;
		PLstrcpy(info.prompt, inPrompt);
		info.mInfo = *inEditMenuInfo;
		info.newColorChosen = false;
		info.reserved = 0;
		
		error = NPickColor(&info);
		result = ((error == noErr) ? info.newColorChosen : false);
		if (result)
		{
			// return the color that the user selected
			outColor->red = info.theColor.color.rgb.red;
			outColor->green = info.theColor.color.rgb.green;
			outColor->blue = info.theColor.color.rgb.blue;
		}
	}
	else if (majorRev == 0x02)
	{
		// Color Picker 2.0 or earlier
		ColorPickerInfo		info;
		
		
		info.theColor.profile = null;
		info.theColor.color.rgb.red = inColor->red;
		info.theColor.color.rgb.green = inColor->green;
		info.theColor.color.rgb.blue = inColor->blue;
		info.dstProfile = null;
		info.flags = kColorUtilitiesColorPickerFlags;
		info.placeWhere = kCenterOnMainScreen;
		info.dialogOrigin = dialogLocation;
		info.pickerType = 0L;
		info.eventProc = inUserEventProc;
		info.colorProc = null;
		info.colorProcData = 0L;
		PLstrcpy(info.prompt, inPrompt);
		info.mInfo = *inEditMenuInfo;
		info.newColorChosen = false;
		info.filler = 0;
		
		error = PickColor(&info);
		result = ((error == noErr) ? info.newColorChosen : false);
		if (result)
		{
			// return the color that the user selected
			outColor->red = info.theColor.color.rgb.red;
			outColor->green = info.theColor.color.rgb.green;
			outColor->blue = info.theColor.color.rgb.blue;
		}
	}
	
	if ((majorRev < 0x02) || (error != noErr))
	{
		// no Color Picker extension, or there was a problem using the cool, new stuff?
		// Well, okay, use the GetColor() call, which works even without the Color Picker extension
		result = GetColor(dialogLocation, inPrompt, inColor, outColor);
	}
	
	return result;
}// ColorChooserDialogDisplay


//
// color pen state methods
//

/*!
To save all pen state information that might
get affected by using a theme brush under
Appearance 1.0, invoke this method.

 1997-1999 by Apple Computer, Inc.

(1.0)
*/
void
ColorUtilities_GetColorAndPenState	(ColorPenState*		outStatePtr)
{
	GrafPtr		currentPort = null;
	
	
	GetPort(&currentPort);
	
	outStatePtr->pnPixPat = null;
	outStatePtr->bkPixPat = null;
#if TARGET_API_MAC_OS8
	// these are non-color graphics port features
	outStatePtr->bkPat = currentPort->bkPat;
	outStatePtr->bkColor = currentPort->bkColor;
	outStatePtr->fgColor = currentPort->fgColor;
	outStatePtr->isColorPort = ColorUtilities_IsColorGrafPort(currentPort);
#else
	// Carbon always uses color ports
	outStatePtr->bkColor = 0;
	outStatePtr->fgColor = 0;
	outStatePtr->isColorPort = true;
#endif
	
	if (outStatePtr->isColorPort)
	{
		CGrafPtr		currentColorPort = (CGrafPtr)currentPort;
		
		
		GetForeColor(&outStatePtr->foreColor);
		GetBackColor(&outStatePtr->backColor);

	#if TARGET_API_MAC_OS8
		// If the pen pattern is not an old-style pattern,
		// copy the handle.  If it is an old style pattern,
		// GetPenState(), below, will save the right thing.
		if ((**currentColorPort->pnPixPat).patType != 0)
		{
			outStatePtr->pnPixPat = currentColorPort->pnPixPat;
		}
		
		// If the pen pattern is not an old style pattern,
		// copy the handle.  Otherwise, get the old pattern
		// into "bkPat", for restoring that way.
		if ((**currentColorPort->bkPixPat).patType != 0)
		{
			outStatePtr->bkPixPat = currentColorPort->bkPixPat;
		}
		else
		{
			outStatePtr->bkPat = *(PatPtr)(*(**(currentColorPort)->bkPixPat).patData);
		}
	#else
		outStatePtr->pnPixPat = NewPixPat();
		outStatePtr->bkPixPat = NewPixPat();
		GetPortPenPixPat(currentColorPort, outStatePtr->pnPixPat);
		GetPortBackPixPat(currentColorPort, outStatePtr->bkPixPat);
	#endif
	}
	
	GetPenState(&outStatePtr->pen);
	outStatePtr->textMode = GetPortTextMode(currentPort);
}// GetColorAndPenState


/*!
To determine the color depth of the specified port,
use this Carbon-compliant method.

(1.0)
*/
SInt16
ColorUtilities_GetCurrentDepth	(CGrafPtr		inPort)
{
	SInt16		result = 1;
	
	
#if TARGET_API_MAC_OS8
	result = (ColorUtilities_IsColorGrafPort((GrafPtr)inPort) ? (**(inPort)->portPixMap).pixelSize : 1);
#else
	result = (**GetPortPixMap(inPort)).pixelSize;
#endif
	return result;
}// GetCurrentDepth


/*!
To determine if the specified device is a Color QuickDraw,
device in a Carbon-compliant way, use this method.

(1.0)
*/
Boolean
ColorUtilities_IsColorDevice	(GDHandle		inDevice)
{
	Boolean		result = TestDeviceAttribute(inDevice, gdDevType);
	
	
	return result;
}// IsColorDevice


/*!
To determine if the specified port is a Color QuickDraw
port in a Carbon-compliant way, use this method.  Once
Mac OS 8 code is abandoned completely, this check will
become obsolete.

(1.0)
*/
Boolean
ColorUtilities_IsColorGrafPort	(GrafPtr		inPort)
{
	Boolean		result = true;
	
	
#if TARGET_API_MAC_OS8
	result = (((inPort)->portBits.rowBytes & 0xC000) == 0xC000);
#endif
	return result;
}// IsColorGrafPort


/*!
To standardize a graphics port for use with
theme brushes under Appearance 1.0 (that is,
slightly more than what PenNormal() does),
use this method.

 1997-1999 by Apple Computer, Inc.

(1.0)
*/
void
ColorUtilities_NormalizeColorAndPen	(void)
{
	RGBColor		black,
					white;
	GrafPtr			currentPort = null;
	
	
	GetPort(&currentPort);
	
	black.red = black.green = black.blue = 0x0000;
	white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
	
	RGBForeColor(&black);
	RGBBackColor(&white);
	PenNormal();
	ColorUtilities_SetWhiteBackgroundPattern();
	TextMode(srcOr);
}// NormalizeColorAndPen


/*!
To restore all pen state information that was previously
saved with ColorUtilities_GetColorAndPenState(), invoke
this method.

 1997-1999 by Apple Computer, Inc.

(1.0)
*/
void
ColorUtilities_SetColorAndPenState	(ColorPenState*		inoutStatePtr)
{
	GrafPtr		currentPort = null;
	
	
	GetPort(&currentPort);
	
	SetPenState(&inoutStatePtr->pen);
	if (ColorUtilities_IsColorGrafPort(currentPort))
	{
		RGBForeColor(&inoutStatePtr->foreColor);
		RGBBackColor(&inoutStatePtr->backColor);
		if (inoutStatePtr->pnPixPat != null) PenPixPat(inoutStatePtr->pnPixPat);
		if (inoutStatePtr->bkPixPat != null) BackPixPat(inoutStatePtr->bkPixPat);
		else BackPat(&inoutStatePtr->bkPat);
	#if TARGET_API_MAC_CARBON
		DisposePixPat(inoutStatePtr->pnPixPat);
		DisposePixPat(inoutStatePtr->bkPixPat);
	#endif
	}
	else
	{
		BackPat(&inoutStatePtr->bkPat);
		ForeColor(inoutStatePtr->fgColor);
		BackColor(inoutStatePtr->bkColor);
	}
	
	TextMode(inoutStatePtr->textMode);
}// SetColorAndPenState


//
// utility methods
//

/*!
Tells QuickDraw to use the highlighting
color instead of inversion for the next
operation.

(1.0)
*/
void
ColorUtilities_HiliteMode	(void)
{
	LMSetHiliteMode(LMGetHiliteMode() & 0x7F);
}// HiliteMode


/*!
Calls BackPat() with an all-black pattern.

(1.0)
*/
void
ColorUtilities_SetBlackBackgroundPattern	(void)
{
	Pattern			blackPat = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
	
	
	BackPat(&blackPat);
}// SetBlackBackgroundPattern


/*!
Calls PenPat() with an all-black pattern.

(1.0)
*/
void
ColorUtilities_SetBlackPenPattern	(void)
{
	Pattern			blackPat = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
	
	
	PenPat(&blackPat);
}// SetBlackPenPattern


/*!
Calls BackPat() with a checkerboard pattern.

(1.0)
*/
void
ColorUtilities_SetGrayBackgroundPattern		(void)
{
	Pattern			grayPat = { 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA };
	
	
	BackPat(&grayPat);
}// SetGrayBackgroundPattern


/*!
Calls PenPat() with a checkerboard pattern.

(1.0)
*/
void
ColorUtilities_SetGrayPenPattern	(void)
{
	Pattern			grayPat = { 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA };
	
	
	PenPat(&grayPat);
}// SetGrayPenPattern


/*!
Calls BackPat() with an all-white pattern.

(1.0)
*/
void
ColorUtilities_SetWhiteBackgroundPattern	(void)
{
	Pattern			whitePat = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	
	
	BackPat(&whitePat);
}// SetWhiteBackgroundPattern


/*!
Calls PenPat() with an all-white pattern.

(1.0)
*/
void
ColorUtilities_SetWhitePenPattern	(void)
{
	Pattern			whitePat = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	
	
	PenPat(&whitePat);
}// SetWhitePenPattern