/*###############################################################

	CommandLine.c
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <AEDataModel.h>
#include <AERegistry.h>
#include <Appearance.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <ControlDefinitions.h>
#include <MacTypes.h>
#include <PLStringFuncs.h>
#include <QuickDraw.h>
#include <TextUtils.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "AboutBox.h"
#include "CommandLine.h"
#include "DialogUtilities.h"
#include "Preferences.h"
#include "EventLoop.h"
#include "InternetConfig.h"
#include "MacLook.h"
#include "NetEvents.h"
#include "Parse.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"
#include "Terminology.h"
#include "URL.h"



// constants
enum
{
	kEditTextInsetH = 8, // pixels horizontally inward that the text field is from the edges of the window
	kEditTextInsetV = 8, // pixels vertically inward that the text field is from the edges of the window
	kHelpTextInsetH = 6, // pixels horizontally inward that the help text is from the edges of the window
	kHelpTextInsetV = 6, // pixels vertically inward that the help text is from the edges of the window
	kEditTextArrowsSpacingH = 8, // number of pixels of space between the text field and the little arrows
	kEditTextHelpTextSpacingV = 6, // number of pixels of space between the text field and the help text
	kCommandLineCommandHistorySize = 5
};

// variables
static ChainRef			gCommandLineInterpreterList = null;
static WindowRef		gCommandLineWindow = null;
static WindowInfoRef	gCommandLineWindowInfo = null;
static ControlRef		gCommandLineRoot = null,
						gCommandLinePlacard = null,
						gCommandLineHelpButton = null,
						gCommandLineArrows = null,
						gCommandLineEditText = null,
						gCommandLineHelpText = null;
static Handle			gCommandHistory[kCommandLineCommandHistorySize];
static SInt16			gCommandHistoryCurrentIndex = -1; // which history command has been displayed via up/down arrow
static Boolean			gIsShowing = false,
						gIsHelpButton = true;

// internal methods
static void								addToHistory					(Handle									inText);
static pascal ControlKeyFilterResult	commandLineKeyFilter			(ControlRef								inControl,
																		 SInt16*								inoutKeyCode,
																		 SInt16*								inoutCharCode,
																		 UInt16*								inoutModifiers);
static void								deltaWindowSize					(SInt32 const							inDeltaSizeX,
																		 SInt32 const							inDeltaSizeY);
static SInt16							getWindowHeight					(void);
static pascal Boolean					handleNewSize					(WindowRef const						inWindowRef,
						 												 SInt32 const							inDeltaSizeX,
																		 SInt32 const							inDeltaSizeY,
																		 SInt32 const							inData);
static void								initWindowSize					(void);
static Boolean							issueMessageToAllInterpreters	(CommandLineInterpreterMessage const	inMessage,
																		 UInt32 const							inArgCount, // argc
																		 TokenSetRef							inTokens, // argv
																		 void*									outDataPtr,
																		 CommandLineInterpreterProcPtr*			outProcPtrPtr);
static Boolean							parseCommandLine				(Handle									inText);
static Boolean							parseUNIX						(UInt16 const							argc,
																		 TokenSetRef							argv);
static Boolean							rotateHistory					(SInt16 const							inHowFarWhichWay);
static void								setBilineText					(StringPtr								inContents);



//
// public methods
//

/*!
This method sets up the initial configurations for
the command line windoid.  Call this method before
calling any other methods from this file.

(3.0)
*/
void
CommandLine_Init	(void)
{
	Rect	rect;
	
	
	// initialize command history
	{
		register SInt16		i = 0;
		
		
		for (i = 0; i < kCommandLineCommandHistorySize; i++) gCommandHistory[i] = null;
	}
	
	// create the command line windoid
	gCommandLineWindow = GetNewCWindow(kWindowIDCommandLine, null, (WindowRef)-1);
	
	// create the windows controls
	(OSErr)CreateRootControl(gCommandLineWindow, &gCommandLineRoot);
	gCommandLineHelpButton = GetNewControl(kIDAppleGuideButton, gCommandLineWindow);
	SetUpHelpButtonControl(gCommandLineHelpButton);
	gCommandLineArrows = GetNewControl(kIDLittleArrows, gCommandLineWindow);
	SetControlMaximum(gCommandLineArrows, kCommandLineCommandHistorySize - 1);
	SetControlMinimum(gCommandLineArrows, 0);
	SetControlValue(gCommandLineArrows, 0);
	DeactivateControl(gCommandLineArrows);
	SetRect(&rect, 0, 0, 0, 0); // not important because the controls are dynamically located and sized
	gCommandLinePlacard = NewControl(gCommandLineWindow, &rect, EMPTY_PSTRING, true/* visible */,
									0/* value */,
									0/* minimum */,
									0/* maximum */,
									kControlPlacardProc,
									0L/* ref. con. */);
	gCommandLineEditText = NewControl(gCommandLineWindow, &rect, EMPTY_PSTRING, true/* visible */,
									0/* value */,
									0/* minimum */,
									0/* maximum */,
									kControlEditTextProc,
									0L/* ref. con. */);
	if (FlagManager_Test(kFlagCarbonControlManager))
	{
		// key filters do *not* seem to work prior to Mac OS 8.5 (they cause a crash when the user types a key)
		ControlKeyFilterUPP		upp = NewControlKeyFilterUPP(commandLineKeyFilter);
		
		
		(OSErr)SetControlData(gCommandLineEditText, kControlEditTextPart, kControlEditTextKeyFilterTag,
							sizeof(ControlKeyFilterUPP), (Ptr)&upp);
	}
	gCommandLineHelpText = NewControl(gCommandLineWindow, &rect, EMPTY_PSTRING, true/* visible */,
									0/* value */,
									0/* minimum */,
									0/* maximum */,
									kControlStaticTextProc,
									0L/* ref. con. */);
	(OSErr)EmbedControl(gCommandLineArrows, gCommandLinePlacard);
	(OSErr)EmbedControl(gCommandLineHelpButton, gCommandLinePlacard);
	(OSErr)EmbedControl(gCommandLineEditText, gCommandLinePlacard);
	(OSErr)EmbedControl(gCommandLineHelpText, gCommandLinePlacard);
	
	SetControlVisibility(gCommandLineRoot, true/* visible */, false/* draw */);
	
	// Set up the Window Info information.  The window
	// resize limits are initialized, but the values are
	// unimportant because getWindowHeight() makes sure
	// they reflect the current window height.
	gCommandLineWindowInfo = WindowInfo_New();
	WindowInfo_SetWindowDescriptor(gCommandLineWindowInfo, kTelnetWindowDescriptorCommandLine);
	WindowInfo_SetWindowFloating(gCommandLineWindowInfo, true);
	WindowInfo_SetWindowResizeLimits(gCommandLineWindowInfo, 0, 0, SHRT_MAX, SHRT_MAX);
	WindowInfo_SetWindowResizeResponderUPP(gCommandLineWindowInfo, handleNewSize, 0L);
	WindowInfo_SetForWindow(gCommandLineWindow, gCommandLineWindowInfo);
	
	// use the preferred rectangle, if any; otherwise, create a default size and position
	{
		Point		deltaSize;
		
		
		SetPt(&deltaSize, 0, 0); // initially...
		deltaSize = Preferences_ArrangeWindow(gCommandLineWindow, deltaSize, kTelnetWindowPreferenceTag_CommandLine);
		
		// With themes, the window frame size cannot be guessed reliably.  Therefore, first, I take
		// the difference between the left edges of the window content and structure region boxes
		// and make that the left edge of the window.
		{
			RgnHandle	region = Memory_NewRegion();
			Rect		screenRect;
			
			
			MacLook_GetWindowDeviceGrayRect(gCommandLineWindow, &screenRect);
			if (region != null)
			{
				Rect		structureRect,
							contentRect;
				
				
				MacLook_GetWindowRegionBounds(gCommandLineWindow, kWindowStructureRgn, &structureRect);
				MacLook_GetWindowRegionBounds(gCommandLineWindow, kWindowContentRgn, &contentRect);
				{
					// if no preferences were available, create defaults
					if ((deltaSize.h == 0) && (deltaSize.v == 0))
					{
						float			fw = 0.80 * (screenRect.right - screenRect.left);
						UInt16			diffX = (contentRect.left - structureRect.left),
										diffY = (contentRect.top - structureRect.top);
						
						
						// define the default window size, and make the window that size initially
						SetRect(&contentRect, screenRect.left + diffX + 4,
												screenRect.top + GetMBarHeight() + diffY + 35/* arbitrary */, 0, 0);
						SetRect(&contentRect, contentRect.left, contentRect.top,
												contentRect.left + (SInt16)fw,
												contentRect.top + 30/* immaterial - its adjusted later */);
						deltaSize.h = contentRect.right - contentRect.left;
						deltaSize.v = contentRect.bottom - contentRect.top;
					}
					
					// set up controls - initWindowSize() requires the correct window width, however
					SizeWindow(gCommandLineWindow, contentRect.right - contentRect.left, 0, false/* update */);
					initWindowSize();
					
					// calculate proper height, and then set up the standard state and arrange the window
					contentRect.bottom = contentRect.top + getWindowHeight() - 1; // the - 1 avoids a utility window sizing bug
					SetWindowStandardState(gCommandLineWindow, &contentRect);
					SizeWindow(gCommandLineWindow, contentRect.right - contentRect.left,
													contentRect.bottom - contentRect.top,
													false/* update */);
					MoveWindow(gCommandLineWindow, contentRect.left, contentRect.top, false);
					
					// arbitrarily set the user state to be a slightly bigger window
					contentRect.right += 70;
					SetWindowUserState(gCommandLineWindow, &contentRect);
					
					// arrange controls
					(Boolean)handleNewSize(gCommandLineWindow, (SInt32)deltaSize.h, (SInt32)deltaSize.v, 0L);
				}
			}
		}
	}
	
	gCommandLineInterpreterList = Chain_New(kChainStyleSinglyLinked);
}// Init


/*!
This method cleans up the command line window
by destroying data structures allocated by
CommandLine_Init().

(3.0)
*/
void
CommandLine_Done	(void)
{
	// save window size and location in preferences
	Preferences_SaveWindowArrangement(gCommandLineWindow, kTelnetWindowPreferenceTag_CommandLine);
	
	DisposeWindow(gCommandLineWindow); // disposes of the command line window *and* all of its controls
	gCommandLineWindow = null;
	WindowInfo_Dispose(gCommandLineWindowInfo);
	Chain_Dispose(&gCommandLineInterpreterList);
	{
		register SInt16		i = 0;
		
		
		for (i = 0; i < kCommandLineCommandHistorySize; i++)
		{
			if (gCommandHistory[i] != null) Memory_DisposeHandle(&gCommandHistory[i]);
		}
	}
}// Done


/*!
To install a command interpreter for the
command line, use this method.  You cannot
remove an interpreter once it is installed.

The command line does initial filtering of a
line of text, such as checking for URLs.  If
these initial filters come up empty, the
command line traverses its list of interpreters
in the order in which they were added.  The
first interpreter whose is command yours
message response is "true" gets executed, and
the command line ceases to traverse its list.

A command interpreter should have a mixin HTML
source file for MacTelnet Help, with an index
term that is the same as the commands name.
This makes the commands help show up when any
kind of help command (including help arguments,
which are never passed to your routine) is
issued.

For information on how this function should
operate, see CommandLine.h.

(3.0)
*/
void
CommandLine_AddInterpreter		(CommandLineInterpreterProcPtr		inProcPtr)
{
	ChainLinkRef	node = Chain_NewNode(gCommandLineInterpreterList, inProcPtr);
	
	
	if (node != null) Chain_MoveNode(gCommandLineInterpreterList, node, null/* means end of list */,
										kChainLinkPlacementInsertAfter);
}// AddInterpreter


/*!
To copy a global outline of a special windoid
region into an existing regions space, use
this method.  Since the command line is a
floating window, this method is useful in
order to determine its global mouse regions.

(3.0)
*/
void
CommandLine_CopyRgn		(CommandLineRegionCode const		inRegion,
						 RgnHandle							outRegion)
{
	RgnHandle		editableTextRgn = Memory_NewRegion();
	Rect			textRect;
	
	
	GetControlBounds(gCommandLineEditText, &textRect);
	RectRgn(editableTextRgn, &textRect);
	LocalToGlobalRegion(editableTextRgn);
	
	switch (inRegion)
	{
		case kCommandLineRegionCodeWindowContent:
			(OSStatus)GetWindowRegion(gCommandLineWindow, kWindowContentRgn, outRegion);
			DiffRgn(outRegion, editableTextRgn, outRegion);
			break;
		
		case kCommandLineRegionCodeEditText:
			CopyRgn(editableTextRgn, outRegion);
			break;
		
		default:
			SetEmptyRgn(outRegion);
			break;
	}
	
	Memory_DisposeRegion(&editableTextRgn);
}// CopyRgn


/*!
To determine the text in the command line text
field, use this method.

(3.0)
*/
void
CommandLine_GetText		(Str255		outString)
{
	GetControlText(gCommandLineEditText, outString);
}// GetText


/*!
To obtain the Mac OS window pointer for the
command line windoid, call this method.

(3.0)
*/
WindowRef
CommandLine_GetWindow	(void)
{
	return ((WindowRef)gCommandLineWindow);
}// GetWindow


/*!
When a mouse-down event occurs in the content region
of the command line windoid, call this method.

(3.0)
*/
void
CommandLine_HandleContentClick		(EventRecord*		inoutEventPtr)
{
	SInt16			windowPart = 0,
					controlPart = 0;
	WindowRef		window = null;
	ControlRef		controlClicked = null,
					focusControl = null;
	
	
	windowPart = FindWindow(inoutEventPtr->where, &window);
	if (window != null)
	{
		GrafPtr		oldPort = null;
		Point		localMouse = inoutEventPtr->where;
		OSErr		error = noErr;
		
		
		GetPort(&oldPort);
		SetPortWindowPort(window);
		GlobalToLocal(&localMouse);
		controlClicked = FindControlUnderMouse(localMouse, window, &controlPart);
		error = GetKeyboardFocus(window, &focusControl);
		if ((controlPart == kControlEditTextPart) && ((focusControl == null) || (error != noErr)))
		{
			// if the text area is not focused, a click should merely focus it...
			WindowRef		frontmostNonFloatingWindow = GetRealFrontWindow();
			
			
			if (frontmostNonFloatingWindow != null) (OSErr)ClearKeyboardFocus(frontmostNonFloatingWindow);
			(OSErr)SetKeyboardFocus(gCommandLineWindow, gCommandLineEditText, kControlFocusNextPart);
		}
		else if ((controlClicked != null) && (controlPart != kControlNoPart))
		{
			// ...otherwise, handle mouse-down events in any of the command lines controls.
			// For edit text controls, HandleControlClick() automatically takes care of things
			// like double-clicking to select words, and dragging or shift-dragging to manage
			// the selection.  For other controls (the little arrows), the command line must
			// interpret what the click means from the returned part code.
			controlPart = HandleControlClick(controlClicked, localMouse, inoutEventPtr->modifiers,
											(ControlActionUPP)null);
			switch (controlPart)
			{
				case kControlButtonPart:
					// help button
					{
						Str255		searchString;
						
						
						GetIndString(searchString, rStringsAppleGuideLookup, siAGIndexTermCommandLine);
						AppleGuideUtilities_OpenFirstMainGuideInSearch(searchString);
					}
					break;
				
				case kControlUpButtonPart:
					if (!rotateHistory(+1)) { /* then its at the end */ }
					break;
				
				case kControlDownButtonPart:
					if (!rotateHistory(-1)) { /* then its at the end */ }
					break;
				
				default:
					break;
			}
		}
		SetPort(oldPort);
	}
}// HandleContentClick


/*!
When a key-down event occurs that should be directed
to the command line windoid, call this method.  If
the key event is handled, true is returned; otherwise,
false is returned.

(3.0)
*/
Boolean
CommandLine_HandleKeyDown		(EventRecord*		inoutEventPtr)
{
	Boolean		result = false;
	
	
	// fire all key events to the editable text item
	if (CommandLine_IsFocused())
	{
		SInt16		keyCode = ((inoutEventPtr->message & keyCodeMask) >> 8),
					charCode = inoutEventPtr->message & charCodeMask,
					modifiers = inoutEventPtr->modifiers;
		
		
		(SInt16)HandleControlKey(gCommandLineEditText, keyCode, charCode, modifiers);
		result = true;
	}
	
	return result;
}// HandleKeyDown


/*!
To give idle time to command line windoid
controls, use this method.

(3.0)
*/
void
CommandLine_Idle	(void)
{
	IdleControls(CommandLine_GetWindow());
}// Idle


/*!
To determine if the command line has keyboard
focus, use this method.

(3.0)
*/
Boolean
CommandLine_IsFocused	(void)
{
	Boolean			result = true;
	OSErr			error = noErr;
	ControlRef		control = null;
	
	
	error = GetKeyboardFocus(CommandLine_GetWindow(), &control);
	if ((error != noErr) || (control == null)) result = false;
	
	return result;
}// IsFocused


/*!
To determine if there is any text in the text field
of the command line windoid, use this method.

(3.0)
*/
Boolean
CommandLine_IsTextEntered	(void)
{
	Str255		text;
	
	
	CommandLine_GetText(text);
	return (StrLength(text) > 0);
}// IsTextEntered


/*!
To determine if the command line is showing, call
this method.  This state is not exactly the window
state; specifically, if the windoid is suspended
but set to visible, the window will be invisible
(because the application is suspended) and yet
this method will return "true".

If you need to find the definite visible state of
the command line window, use CommandLine_GetWindow()
to obtain the Mac OS window pointer and then use the
IsWindowVisible() system call.

(3.0)
*/
Boolean
CommandLine_IsVisible	(void)
{
	return gIsShowing;
}// IsVisible


/*!
To remove the keyboard focus from the command line
windoid, use this method.

(3.0)
*/
void
CommandLine_LoseFocus	(void)
{
	(OSErr)ClearKeyboardFocus(CommandLine_GetWindow());
	DrawOneControl(gCommandLineEditText);
}// LoseFocus


/*!
To determine if the specified point, in global
coordinates, is in the specified region of the
command line windoid, call this method.

The possible region codes are as follows:

"kCommandLineRegionCodeEditText"
	This is the region of the editable text
	control where the user can type commands
	on the command line.

"kCommandLineRegionCodeWindowContent"
	A point is inside this region if it is
	within the content area of the window
	and not within any region for which a
	more specific region code applies.

(3.0)
*/
Boolean
CommandLine_PtInRgn		(Point const						inGlobalPoint,
						 CommandLineRegionCode const		inRegion)
{
	Point		localPoint = inGlobalPoint;
	Boolean		isInContentRegion = false;
	Boolean		result = false;
	
	
	GlobalToLocal(&localPoint);
	{
		RgnHandle		contentRegion = Memory_NewRegion();
		
		
		if (contentRegion != null)
		{
			(OSStatus)GetWindowRegion(gCommandLineWindow, kWindowContentRgn, contentRegion);
			isInContentRegion = PtInRgn(inGlobalPoint, contentRegion);
			Memory_DisposeRegion(&contentRegion);
		}
	}
	
	switch (inRegion)
	{
		case kCommandLineRegionCodeWindowContent:
		case kCommandLineRegionCodeEditText:
			result = isInContentRegion;
			if (result)
			{
				Rect		textRect;
				
				
				GetControlBounds(gCommandLineEditText, &textRect);
				if (inRegion == kCommandLineRegionCodeEditText)
				{
					result = PtInRect(localPoint, &textRect);
				}
				else result = !PtInRect(localPoint, &textRect);
			}
			break;
		
		default:
			break;
	}
	
	return result;
}// PtInRgn


/*!
To activate or deactivate the controls in the
command line window, call this method.

(3.0)
*/
void
CommandLine_SetActive		(Boolean const		inIsActive)
{
	if (inIsActive) ActivateControl(gCommandLineRoot);
	else DeactivateControl(gCommandLineRoot);
	
	//(OSErr)Embedding_OffscreenSetRootControlActive(gCommandLineWindow, inIsActive);
}// SetActive


/*!
To set the suspended state of the command line,
call this method.  If the command line is suspended,
it is invisible no matter what state it is in.  If
the command line is not suspended, its previous
state is restored.  For example, if the window is
visible and suspended, it will be hidden.  If
the window is invisible and suspended, it will
be hidden as well.  But if the window is then set
to a non-suspended state, in the former case the
window will reappear, whereas in the latter case
the window remains invisible.

(3.0)
*/
void
CommandLine_SetSuspended	(Boolean	inIsSuspended)
{
	WindowRef		commandLine = CommandLine_GetWindow();
	
	
	if (inIsSuspended) HideWindow(commandLine);
	else
	{
		if (CommandLine_IsVisible()) ShowWindow(commandLine);
		else HideWindow(commandLine);
	}
}// SetSuspended


/*!
To set the text displayed in the message window
editable text field, use this method.  You might
use this to handle a text drag-and-drop into this
windoid.

(3.0)
*/
void
CommandLine_SetText		(void*			inDataPtr,
						 Size 			inSize)
{
	(OSErr)SetControlData(gCommandLineEditText, kControlEditTextPart, kControlEditTextTextTag, inSize, inDataPtr);
	DrawOneControl(gCommandLineEditText);
}// SetText


/*!
To show or hide the command line windoid,
use this method.

(3.0)
*/
void
CommandLine_SetVisible	(Boolean	inIsVisible)
{
	Boolean			oldVisibility = gIsShowing;
	
	
	gIsShowing = inIsVisible;
	if (oldVisibility != gIsShowing)
	{
		RgnHandle		structureRegion = null;
		WindowRef		commandLine = CommandLine_GetWindow();
		
		
		// show or hide the window, and configure the notification procedures flags appropriately
		if (inIsVisible)
		{
			ShowWindow(commandLine);
			(OSErr)ClearKeyboardFocus(GetRealFrontWindow());
			(OSErr)SetKeyboardFocus(gCommandLineWindow, gCommandLineEditText, kControlEditTextPart);
		}
		else HideWindow(commandLine);
	}
}// SetVisible


/*!
To show or hide the command line based on whether or
not the command line is currently visible, call this
method.  You generally use this to handle a
Show/Hide Command Line menu item, where you called
the CommandLine_SetDependentMenuItem() method to
specify which item handles the windoids visible
state.

(3.0)
*/
void
CommandLine_ToggleVisible	(void)
{
	CommandLine_SetVisible(!CommandLine_IsVisible());
}// ToggleVisible


/*!
To update the command line window, call this method.
This should be done when update events are received
for the command line window.

A memory-related error may result.

(3.0)
*/
OSErr
CommandLine_Update	(GrafPtr		inDrawingPort)
{
	OSErr		result = noErr;
	
	
	if (inDrawingPort != null)
	{
		RgnHandle		updateRegion = Memory_NewRegion();
		
		
		// get the update region
		if (updateRegion != null)
		{
			GetPortClipRegion(inDrawingPort, updateRegion);
			UpdateControls(gCommandLineWindow, updateRegion);
			Memory_DisposeRegion(&updateRegion);
		}
		else
		{
			DrawControls(gCommandLineWindow);
			result = memFullErr;
		}
	}
	else result = memPCErr;
	
	return result;
}// Update


/*!
To set the command line windoid size to either the user
state or the standard state, call this method with,
respectively, window parts of "inZoomIn" or "inZoomOut".
The state of modifier keys, in standard Mac OS Event
Manager format, is provided for special cases, and it
is presumed that any specified modifiers were held down
while the indicated window part was clicked.

(3.0)
*/
void
CommandLine_Zoom	(short		inWindowPart,
					 short		UNUSED_ARGUMENT(inModifiers))
{
	switch (inWindowPart)
	{
		case inZoomOut:
		case inZoomIn:
			// If the Option key is held down, maximize the
			// Command Line windows width; otherwise, move and
			// resize the window to the appropriate boundaries.
			{
				Rect		windowPortRect,
							newRect;
				
				
				GetPortBounds(GetWindowPort(gCommandLineWindow), &windowPortRect);
				EraseRect(&windowPortRect);
				ZoomWindow(gCommandLineWindow, inWindowPart, false);
				GetPortBounds(GetWindowPort(gCommandLineWindow), &newRect);
				(Boolean)handleNewSize(gCommandLineWindow, (SInt32)(newRect.right - windowPortRect.right),
										(SInt32)(newRect.bottom - windowPortRect.bottom), 0L);
			}
			break;
		
		default:
			BEEP;
			break;
	}
}// Zoom


//
// internal methods
//

/*!
Adds the specified command line to the history buffer,
shifting all commands back one (and therefore destroying
the oldest one) and resetting the history pointer.

(3.0)
*/
static void
addToHistory	(Handle		inText)
{
	register SInt16		i = 0,
						last = kCommandLineCommandHistorySize - 1;
	
	
	if (gCommandHistory[last] != null) Memory_DisposeHandle(&gCommandHistory[last]);
	for (i = last; i > 0; i--) gCommandHistory[i] = gCommandHistory[i - 1];
	gCommandHistory[0] = inText;
	gCommandHistoryCurrentIndex = -1;
	if (!IsControlActive(gCommandLineArrows)) ActivateControl(gCommandLineArrows);
}// addToHistory


/*!
Filters out key presses originally sent to the command line
text field.  For example, if the user presses Return or Enter,
the key is interpreted as parse this command line instead of
being passed to the control; if the user presses the up or
down arrow keys, he or she can rotate through past commands.

(3.0)
*/
static pascal ControlKeyFilterResult
commandLineKeyFilter	(ControlRef		inControl,
						 SInt16*		inoutKeyCode,
						 SInt16*		UNUSED_ARGUMENT(inoutCharCode),
						 UInt16*		UNUSED_ARGUMENT(inoutModifiers))
{
	ControlKeyFilterResult		result = kControlKeyFilterPassKey;
	Handle						commandLineText;
	
	
	// determine the text currently in the command line
	commandLineText = GetControlTextAsNewHandle(inControl);
	
	if ((commandLineText == null) || (GetHandleSize(commandLineText) == 0))
	{
		// when empty, restore original biline string
		Str255		string;
		
		
		GetIndString(string, rStringsHelpText, siHelpTextCommandLine);
		setBilineText(string);
	}
	
	// Some actions are based on the physical key, usually
	// because it is easier to do so than checking ASCII
	// codes (or there is no ASCII code).  Check for these
	// keys first.
	if (inoutKeyCode != null) switch (*inoutKeyCode)
	{
		case 0x31: // space
		case 0x30: // tab
			// try to describe what will happen when this command is executed
			if ((GetHandleSize(commandLineText) > 0) && (GetHandleSize(commandLineText) < 256))
			{
				UInt16			argc = 0;
				TokenSetRef		argv = null;
				Str255			string;
				
				
				// copy the data into a Pascal string buffer
				string[0] = GetHandleSize(commandLineText);
				BlockMove(*commandLineText, string + 1, string[0]);
				
				// take apart the command line in its present state, creating a set of words
				StringUtilities_PTokenize(string, &argc, &argv);
				{
					CommandLineInterpreterProcPtr	procPtr = null;
					Str255							finalText;
					Boolean							haveDescription = false;
					
					
					// is the command even recognized?
					if (issueMessageToAllInterpreters(kCommandLineInterpreterMessageIsCommandYours, argc, argv, null, &procPtr))
					{
						if (procPtr != null)
						{
							Str255		descriptionText;
							
							
							haveDescription = InvokeCommandLineInterpreterProc(procPtr, argc, argv, descriptionText,
																				kCommandLineInterpreterMessageDescribeLastArgument);
							if (haveDescription)
							{
								Str31		commandName;
								
								
								// then the command interpreter has a description for the argument - use it!
								GetIndString(finalText, rStringsCommandLineCommandDescriptions, siCommandDescriptionTemplate);
								StringUtilities_PGetToken(0, argv, commandName);
								StringUtilities_PMetaSetDefaults();
								StringUtilities_PBuild(finalText, commandName, descriptionText, EMPTY_PSTRING, EMPTY_PSTRING);
							}
							else if (argc > 1) // that is, the missing description is for an argument, not the entire command
							{
								Str31		commandName,
											argumentName;
								
								
								// then the command didnt recognize the argument - warn the user
								GetIndString(descriptionText, rStringsCommandLineCommandDescriptions, siCommandDescriptionForBadArgument);
								GetIndString(finalText, rStringsCommandLineCommandDescriptions, siCommandDescriptionTemplate);
								StringUtilities_PGetToken(0, argv, commandName);
								StringUtilities_PGetToken(argc - 1/* get last argument */, argv, argumentName);
								StringUtilities_PMetaSetDefaults();
								StringUtilities_PBuild(descriptionText, commandName, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
								StringUtilities_PBuild(finalText, argumentName, descriptionText, EMPTY_PSTRING, EMPTY_PSTRING);
							}
						}
					}
					
					if ((!haveDescription) && (argc == 1))
					{
						// warn the user that the entire contents of the command line will be interpreted as search keywords
						GetIndString(finalText, rStringsCommandLineCommandDescriptions, siCommandDescriptionNotAvailable);
					}
					setBilineText(finalText);
				}
				StringUtilities_PDisposeTokens(argc, &argv);
			}
			break;
		
		case 0x24: // Return key
		case 0x4C: // Enter key
			// If the user presses Return or Enter, the
			// command text should be executed.  If the
			// command makes sense, the text field is
			// cleared and it executes; otherwise, the
			// text remains and the system emits a beep.
			if (GetHandleSize(commandLineText) > 0)
			{
				if (parseCommandLine(commandLineText)) // then the command will already have been executed at this point
				{
					Str255		string;
					
					
					// clear the text field
					(OSErr)SetControlData(inControl, kControlEditTextPart, kControlEditTextTextTag, 0, (Ptr)"");
					DrawOneControl(inControl);
					
					// restore original biline string
					GetIndString(string, rStringsHelpText, siHelpTextCommandLine);
					setBilineText(string);
				}
				else BEEP;
			}
			result = kControlKeyFilterBlockKey;
			break;
		
		case 0x7E: // Up arrow
		case 0x3E: // Up arrow on non-extended keyboards
			// If the user presses an up arrow key, set the
			// command line text to the previous command in
			// the history buffer rotation.
			if (!rotateHistory(+1)) BEEP;
			result = kControlKeyFilterBlockKey;
			break;
		
		case 0x7D: // Down arrow
		case 0x3D: // Down arrow on non-extended keyboards
			// If the user presses a down arrow key, set the
			// command line text to the next command in the
			// history buffer rotation.
			if (!rotateHistory(-1)) BEEP;
			result = kControlKeyFilterBlockKey;
			break;
		
		default:
			// uninteresting key was hit
			break;
	}
	
	return result;
}// commandLineKeyFilter


/*!
To rearrange the command line windows controls
to correspond to the amount the window dimensions
have changed, use this method.

(3.0)
*/
static void
deltaWindowSize		(SInt32 const			inDeltaSizeX,
					 SInt32 const			inDeltaSizeY)
{
	DialogAdjust_BeginControlAdjustment(gCommandLineWindow);
	
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gCommandLinePlacard, inDeltaSizeX);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gCommandLinePlacard, inDeltaSizeY);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gCommandLineEditText, inDeltaSizeX);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gCommandLineHelpText, inDeltaSizeX);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentMoveH, gCommandLineArrows, inDeltaSizeX);
	if (gIsHelpButton) DialogAdjust_AddControlAdjustment(kControlAdjustmentMoveH, gCommandLineHelpButton, inDeltaSizeX);
	
	DialogAdjust_EndAdjustment(inDeltaSizeX, inDeltaSizeY);
}// deltaWindowSize


/*!
To calculate the proper height for the window, by
adding up the vertical expanses of the appropriate
window controls, use this method.

(3.0)
*/
static SInt16
getWindowHeight		(void)
{
	Rect		bounds;
	SInt16		result = 0;
	
	
	result += kEditTextInsetV;
	GetControlBounds(gCommandLineEditText, &bounds);
	result += (bounds.bottom - bounds.top);
	result += kEditTextHelpTextSpacingV;
	GetControlBounds(gCommandLineHelpText, &bounds);
	result += (bounds.bottom - bounds.top);
	result += kHelpTextInsetV;
	
	// constrain the window resize limits to the calculated height
	WindowInfo_SetWindowResizeLimits(gCommandLineWindowInfo,
										result /* minimum height */,
										300 /* minimum width, arbitrary */,
										result /* maximum height */,
										SHRT_MAX/* maximum width */);
	
	return result;
}// getWindowHeight


/*!
This method, which corresponds to WindowResizeResponderProcPtr
(see "InterfaceLib.h"), adjusts the contents of the specified
window based on its new dimensions.  The change between the
old dimensions and the new dimensions is given, where
negative values indicate that the size in one or both
directions has decreased.  The specified window must have
an associated WindowInfoRef, because that reference is used
to determine the exact window descriptor (and thus how to
respond to the resize).

Currently, this internal method only handles the command
line windoid.  The file "DialogAdjust.c" handles resizable
dialogs.

(3.0)
*/
static pascal Boolean
handleNewSize	(WindowRef const		inWindowRef,
				 SInt32 const			inDeltaSizeX,
				 SInt32 const			inDeltaSizeY,
				 SInt32 const			UNUSED_ARGUMENT(inData))
{
	WindowInfoRef		windowFeaturesRef = WindowInfo_GetFromWindow(inWindowRef);
	Boolean				result = false;
	
	
	if (windowFeaturesRef != null)
	{
		Cursors_DeferredUseWatch(20); // on slow computers, display a watch cursor after 1/3 a second!
		
		// determine what window this is
		switch (WindowInfo_GetWindowDescriptor(windowFeaturesRef))
		{
			case kTelnetWindowDescriptorCommandLine:
				deltaWindowSize(inDeltaSizeX, inDeltaSizeY);
				result = true;
				break;
			
			default:
				result = false;
				break;
		}
		
		Cursors_UseArrow();
	}
	
	return result;
}// handleNewSize


/*!
To set the initial size and position of the
command line windows controls, use this method.

(3.0)
*/
static void
initWindowSize		(void)
{
	static SInt16	oldContentHeight = 0;
	SInt16			contentHeight = 0;
	SInt16			fieldHeight = 0,
					helpTextHeight = 0;
	UInt16			kArrowsHelpButtonSpacingH = (gIsHelpButton) ? 10 : 0, // number of pixels between the arrows & help button
					kHelpButtonInsetH = (gIsHelpButton) ? 10 : 0; // number of pixels between help button and the window edge
	Rect			controlRect,
					drawingPortRect,
					littleArrowsRect,
					helpButtonRect;
	GrafPtr			oldPort = null;
	WindowRef		commandLineWindow = gCommandLineWindow;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(gCommandLineWindow);
	MacLook_GetWindowRegionBounds(gCommandLineWindow, kWindowContentRgn, &drawingPortRect);
	
	(OSErr)SetControlVisibility(gCommandLineArrows, false/* visibility */, false/* draw */);
	if (gIsHelpButton)
	{
		(OSErr)SetControlVisibility(gCommandLineHelpButton, false/* visibility */, false/* draw */);
	}
	(OSErr)SetControlVisibility(gCommandLineEditText, false/* visibility */, false/* draw */);
	(OSErr)SetControlVisibility(gCommandLineHelpText, false/* visibility */, false/* draw */);
	{
		Str255		string;
		Size		actualSize = 0;
		
		
		// set the command line help text, resizing the control to fit
		GetIndString(string, rStringsHelpText, siHelpTextCommandLine);
		(UInt16)Localization_SetUpSingleLineTextControl(gCommandLineHelpText, string, kThemeSmallSystemFont);
		
		// set the command line edit text, resizing the control to fit
		CommandLine_GetText(string);
		(UInt16)Localization_SetUpSingleLineTextControl(gCommandLineEditText, string, kThemeSystemFont);
		{
			ControlFontStyleRec		styleRecord;
			TerminalPrefs**			termPrefsHandle = null;
			SInt16					fontID = 0;
			
			
			// use the default terminals font for the message window, because the font is fixed-width
			termPrefsHandle = Preferences_GetDefaultTerminal();
			if (termPrefsHandle != null)
			{
			#if TARGET_API_MAC_OS8
				GetFNum((**termPrefsHandle).normalFont, &fontID);
			#else
				fontID = FMGetFontFamilyFromName((**termPrefsHandle).normalFont);
			#endif
				Memory_DisposeHandle((Handle*)&termPrefsHandle);
			}
			else
			{
			#if TARGET_API_MAC_OS8
				GetFNum("\pMonaco", &fontID);
			#else
				fontID = FMGetFontFamilyFromName("\pMonaco");
			#endif
			}
			styleRecord.size = 12;
			styleRecord.flags = kControlUseFontMask | kControlUseSizeMask;
			styleRecord.font = fontID;
			
			(OSErr)SetControlData(gCommandLineEditText, kControlNoPart, kControlFontStyleTag,
									sizeof(styleRecord), (Ptr)&styleRecord);
		}
	}
	
	GetControlBounds(gCommandLineEditText, &controlRect);
	fieldHeight = (controlRect.bottom - controlRect.top);
	GetControlBounds(gCommandLineHelpText, &controlRect);
	helpTextHeight = (controlRect.bottom - controlRect.top);
	contentHeight = kEditTextInsetV + fieldHeight +
					kEditTextHelpTextSpacingV + helpTextHeight + kHelpTextInsetV;
	
	// arrange the controls, centering the little arrows vertically with the edit text control
	MoveControl(gCommandLinePlacard, -1, -1);
	MoveControl(gCommandLineEditText, kEditTextInsetH, kEditTextInsetV);
	MoveControl(gCommandLineHelpText, kHelpTextInsetH, kEditTextInsetV + fieldHeight + kEditTextHelpTextSpacingV);
	GetControlBounds(gCommandLineEditText, &controlRect);
	GetControlBounds(gCommandLineArrows, &littleArrowsRect); // need width and height of the arrows control
	if (gIsHelpButton) GetControlBounds(gCommandLineHelpButton, &helpButtonRect); // need button width and height
	else SetRect(&helpButtonRect, 0, 0, kHelpTextInsetH, 0); // it is invisible, so make it a narrow space instead
	MoveControl(gCommandLineArrows,
				drawingPortRect.right - kHelpButtonInsetH - (helpButtonRect.right - helpButtonRect.left) -
				kArrowsHelpButtonSpacingH - (littleArrowsRect.right - littleArrowsRect.left),
				kEditTextInsetV - ((littleArrowsRect.bottom - littleArrowsRect.top) -
				(controlRect.bottom - controlRect.top)) / 2);
	if (gIsHelpButton)
	{
		MoveControl(gCommandLineHelpButton,
					drawingPortRect.right - kHelpButtonInsetH - (helpButtonRect.right - helpButtonRect.left),
					kEditTextInsetV - ((helpButtonRect.bottom - helpButtonRect.top) -
					(controlRect.bottom - controlRect.top)) / 2);
	}
	
	// resize the controls to span the width of the window
	SizeControl(gCommandLinePlacard, drawingPortRect.right - drawingPortRect.left + 2, contentHeight + 2);
	SizeControl(gCommandLineEditText,
				drawingPortRect.right - drawingPortRect.left - kEditTextInsetH - kHelpButtonInsetH -
				(helpButtonRect.right - helpButtonRect.left) - kArrowsHelpButtonSpacingH -
				(littleArrowsRect.right - littleArrowsRect.left) - kEditTextArrowsSpacingH,
				fieldHeight);
	SizeControl(gCommandLineHelpText, drawingPortRect.right - drawingPortRect.left - (2 * kHelpTextInsetH),
				helpTextHeight);
	
	(OSErr)SetControlVisibility(gCommandLineArrows, true/* visibility */, true/* draw */);
	if (gIsHelpButton)
	{
		(OSErr)SetControlVisibility(gCommandLineHelpButton, true/* visibility */, true/* draw */);
	}
	(OSErr)SetControlVisibility(gCommandLineEditText, true/* visibility */, true/* draw */);
	(OSErr)SetControlVisibility(gCommandLineHelpText, true/* visibility */, true/* draw */);
	
	SetPort(oldPort);
}// initWindowSize


/*!
Iterates the linked list of command interpreters, issuing
each one the given message.  The first one that returns
"true" to that message stops the list traversal, and that
interpreter is returned.  If "true" is ever returned, this
method returns "true"; otherwise, this method returns
"false" to indicate that no interpreters responded to the
given message.

(3.0)
*/
static Boolean
issueMessageToAllInterpreters	(CommandLineInterpreterMessage const		inMessage,
								 UInt32 const								inArgCount, // argc
								 TokenSetRef								inTokens, // argv
								 void*										outDataPtr,
								 CommandLineInterpreterProcPtr*				outProcPtrPtr)
{
	CommandLineInterpreterProcPtr	procPtr = null;
	ChainLinkRef					commandNode = Chain_GetFirstNode(gCommandLineInterpreterList);
	Boolean							result = false,
									done = false;
	
	
	Cursors_UseWatch();
	while ((commandNode != null) && (!done))
	{
		// continue to service network events while each command is examined
		NetEvents_Service();
		
		procPtr = (CommandLineInterpreterProcPtr)Chain_GetNodeDataPtr(gCommandLineInterpreterList, commandNode);
		
		if (procPtr != null) done = InvokeCommandLineInterpreterProc(procPtr, inArgCount, inTokens, outDataPtr, inMessage);
		if (!done)
		{
			commandNode = Chain_GetNextNode(gCommandLineInterpreterList, commandNode);
		}
		else result = true;
	}
	
	if (outProcPtrPtr != null) *outProcPtrPtr = procPtr;
	Cursors_UseArrow();
	
	return result;
}// issueMessageToAllInterpreters


/*!
To parse the given string as a command line, call
this method.  If the command line is meaningful,
it is executed and true is returned; otherwise,
false is returned to indicate that the command is
in some way unrecognizable.

If valid, the command line is added to the
command history list and the oldest command is
discarded.

(3.0)
*/
static Boolean
parseCommandLine	(Handle		inText)
{
	Boolean		result = false;
	
	
	if (inText != null)
	{
		// check for URLs if nothing valid has been parsed yet
		unless (result) if (URL_GetTypeFromDataHandle(inText) != kNotURL)
		{
			URL_SendGetURLEventToSelf(*inText, GetHandleSize(inText));
			result = true;
		}
		
		// if not a URL, check for UNIX commands
		#if 0
		unless (result) if (GetHandleSize(inText) < 256)
		{
			UInt16			argc = 0;
			TokenSetRef		argv = null;
			Str255			string;
			
			
			// make a string first
			string[0] = GetHandleSize(inText);
			BlockMove(inText, string + 1, string[0]);
			
			// tokenize the string and go from there
			StringUtilities_PTokenize(string, &argc, &argv);
			if ((argc > 0) && (argv != null)) result = parseUNIX(argc, argv);
			StringUtilities_PDisposeTokens(argc, &argv);
		}
		#endif
		
		// if all else fails, send the string to the frontmost session window (if any)
		unless (result)
		{
			WindowRef		frontWindow = GetRealFrontWindow();
			
			
			if (ScreenManager_IsWindowAScreen(frontWindow))
			{
				ConnectionDataPtr		connectionDataPtr = ScreenManager_GetWindowConnectionData(frontWindow);
				
				
				Parse_TypeString(connectionDataPtr, *inText, GetHandleSize(inText), true/* record */);
				Parse_TypeString(connectionDataPtr, "\015\012", 2, true/* record */);
				result = true;
			}
		}
	}
	if (result) addToHistory(inText); // only remember valid command lines
	return result;
}// parseCommandLine


/*!
This routine is structured exactly like a standard
main() function would be for a UNIX application,
with the exception of the Boolean return value.
Note that C strings are passed.  If the specified
arguments make sense (ALL of them), perform the
requested operation and return true; otherwise,
return false (and the command line will then emit
a system beep).

If there are exactly two arguments, where the
first is the standard command name and the second
is one of the recognized help arguments (stored
in Telnets resource file), parseUNIX() will not
try to parse the command, but will rather fire the
commands name as the single search keyword to
Telnet Help.

Just to recap, standard UNIX form means that the
value of "argc" is the size of the specified array,
where "argv[0]" is the program name, or in this
case, the command name.  If "argc" is greater than
1 (it will always be at least 1), it means that
the user has passed arguments to the command.

(3.0)
*/
static Boolean
parseUNIX		(UInt16 const		argc,
				 TokenSetRef		argv)
{
	Boolean		result = false,
				caseCompare = false,
				diacriticsCompare = true;
	Str255		cmdName;
	Boolean		handled = false;
	
	
	// break the command line down into words, then check for known commands
	StringUtilities_PGetToken(0, argv, cmdName);
	
	// look for help arguments
	if (argc > 1)
	{
		register UInt16		i = 0;
		Str31				helpArgument,
							currentHelpArgument;
		
		
		StringUtilities_PGetToken(1, argv, helpArgument);
		for (i = 1; ((i <= rStringCountCommandLineUNIXHelpArguments) && (!handled)); i++)
		{
			GetIndString(currentHelpArgument, rStringsCommandLineUNIXHelpArguments, i);
			if (EqualString(currentHelpArgument, helpArgument, caseCompare, diacriticsCompare))
			{
				if (FlagManager_Test(kFlagAppleGuideFiles))
				{
					OSStatus		error = 0L;
					
					
					error = noErr;
					
					// fire the command name as a keyword to the guide file
					error = AppleGuideUtilities_OpenFirstMainGuideInSearch(cmdName);
					result = (error == noErr);
				}
				handled = true;
			}
		}
	}
	
	// Iterate the linked list of command interpreters, asking each one if it can recognize
	// the command given.  The first one that returns "true" to that message gets sent the
	// execute message, and the list traversal stops.
	if (!handled)
	{
		CommandLineInterpreterProcPtr	procPtr = null;
		
		
		if (issueMessageToAllInterpreters(kCommandLineInterpreterMessageIsCommandYours, argc, argv, null, &procPtr))
		{
			if (procPtr != null) result = InvokeCommandLineInterpreterProc(procPtr, argc, argv, null,
																			kCommandLineInterpreterMessageExecute);
		}
	}
	
	return result;
}// parseUNIX


/*!
To adjust the command line history pointer
by so many positions in one direction or
the other, use this method.  If the request
can be fulfilled (that is, there are enough
commands on one side of the buffer pointer
to be able to move it), true is returned
and the command line text field is changed
to reflect the current history command.
Otherwise, false is returned and the text
field is untouched.

A negative number will move the history
pointer towards the present.  A positive
number will move the history pointer back
in time.  The control value of the little
arrows control is the history pointer.

(3.0)
*/
static Boolean
rotateHistory		(SInt16 const		inHowFarWhichWay)
{
	Boolean		result = false;
	
	
	// the requested adjustment must land the history index in the valid range to be acceptable
	result = (((gCommandHistoryCurrentIndex + inHowFarWhichWay) >= -1) &&
				((gCommandHistoryCurrentIndex + inHowFarWhichWay) < kCommandLineCommandHistorySize));
	
	if (result)
	{
		Handle		command = null;
		
		
		gCommandHistoryCurrentIndex += inHowFarWhichWay;
		SetControlValue(gCommandLineArrows, gCommandHistoryCurrentIndex);
		
		if (gCommandHistoryCurrentIndex >= 0) command = gCommandHistory[gCommandHistoryCurrentIndex];
		
		if (command != null) SetControlTextHandle(gCommandLineEditText, command);
		else SetControlText(gCommandLineEditText, EMPTY_PSTRING);
		(OSErr)SetKeyboardFocus(gCommandLineWindow, gCommandLineEditText, kControlEditTextPart);
		DrawOneControl(gCommandLineEditText);
	}
	return result;
}// rotateHistory


/*!
Use this method to specify the text underneath the
editable text field in the command line windoid.

(3.0)
*/
static void
setBilineText	(StringPtr		inContents)
{
	// set the control text to be the copy that was just created
	SetControlText(gCommandLineHelpText, inContents);
	DrawOneControl(gCommandLineHelpText);
}// setBilineText
