/*###############################################################

	CommandLine.h
	
	The message window management system in MacTelnet 3.0.  This
	window can automatically handle data such as URLs and search
	keywords, but is also pluggable with command handlers that
	interpret arguments like a UNIX program does.
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1992 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#ifndef __COMMANDLINE__
#define __COMMANDLINE__

// Mac includes
#ifndef __MACTYPES__
#	include <MacTypes.h>
#endif
#ifndef __MACWINDOWS__
#	include <MacWindows.h>
#endif

// library includes
#ifndef __INTERFACELIB__
#	include "InterfaceLib.h"
#endif



// constants
typedef UInt16 CommandLineRegionCode;
enum
{
	kCommandLineRegionCodeWindowContent = 0,
	kCommandLineRegionCodeEditText = 1
};



// universal procedure pointers
/**
Command Interpreting Method

This defines the main entry point to a command handler.  You are
given a Handle, an argument count, and a message, and requested to
return "true" or "false".  The messages are defined below.

In response to an is command yours message:
	inArgCount:		1
	inTokens:		the first word entered on the command line
	outDataPtr:		not defined; should be ignored and not changed
	
	You are responsible at this stage for examining only the first
	token of the given set, which is the command name.  Return "true"
	only if the command is yours (or one you can handle), which is
	usually done using a case-insensitive, diacritics-sensitive test.

In response to a describe last argument message:
	inArgCount:		if 1, describe the entire command; otherwise,
					describe the last argument of the given set
	inTokens:		the words entered on the command line
	outDataPtr:		output-only; StringPtr type, capacity 255
	
	The command line continually parses what the user types, as he or
	she types it, sending is command yours messages to all command
	interpreters.  If an interpreter returns "true" to that message,
	the command line modifies its display with a text string that
	describes the command that the user is trying to invoke *or* the
	last argument on the command line.  To find out what this text
	string is, a describe last argument message is passed to the
	appropriate interpreter.  Storage space is set up for a "Str255",
	which is pointed to by "outDataPtr".  Your method should copy
	into this space an appropriate descriptive string, stated as a
	sentence fragment *without* mentioning the commands name or the
	argument name.  If the last argument is recognized, return "true"
	and fill in the descriptive text; otherwise, return "false".  If
	your command recognizes generic arguments, such as file names or
	URLs, you may wish to verify that an unrecognized argument starts
	with "+" or "-" before returning "false", because "false" causes
	MacTelnet to warn the user that the argument is not understood by
	the command.  In other words, return "true" for generic arguments
	as well.

In response to an execute message:
	inArgCount:		number of words, including the command name (the
					first word)
	inTokens:		all words entered on the command line
	outDataPtr:		not defined; should be ignored and not changed
	
	At this stage, the command line has granted tokens exclusively to
	you, which means that regardless of the return value, no further
	parsing will be done.  You should determine if *all* of the
	arguments are valid except argument 0 (which was validated via an
	earlier is command yours message), and if they are valid,
	execute the command and return "true".  If "false" is returned,
	the command line emits a system beep and does not add the text in
	the command line to the command history (nor does it clear the
	command line).  Otherwise, the command is assumed to have
	executed properly, so the command line is cleared and the line is
	added to the command history rotation.  Therefore, if you return
	"false", you should not perform *any* action.  Conversely, if you
	do perform *any* kind of action, return "true".
*/
typedef UInt32 CommandLineInterpreterMessage;
enum
{
	kCommandLineInterpreterMessageIsCommandYours = 1,		// outDataPtr == null
	kCommandLineInterpreterMessageDescribeLastArgument = 2,	// outDataPtr == Str255, write-only
	kCommandLineInterpreterMessageExecute = 3				// outDataPtr == null
};
typedef Boolean (*CommandLineInterpreterProcPtr)	(const UInt32					inArgCount,
													 TokenSetRef					inTokens,
													 void*							outDataPtr,
													 CommandLineInterpreterMessage	inMessage);
#define InvokeCommandLineInterpreterProc(userRoutine, inArgCount, inTokens, outDataPtr, inMessage) \
	(*userRoutine)((inArgCount), (inTokens), (outDataPtr), (inMessage))



//
// public methods
//

void
	CommandLine_Init						(void);

void
	CommandLine_Done						(void);

void
	CommandLine_AddInterpreter				(CommandLineInterpreterProcPtr		inProcPtr);

void
	CommandLine_CopyRgn						(const CommandLineRegionCode		inRegion,
											 RgnHandle							outRegion);

void
	CommandLine_GetText						(Str255								outString);

WindowRef
	CommandLine_GetWindow					(void);

void
	CommandLine_HandleContentClick			(EventRecord*						inoutEventPtr);

Boolean
	CommandLine_HandleKeyDown				(EventRecord*						inoutEventPtr);

void
	CommandLine_Idle						(void);

Boolean
	CommandLine_IsFocused					(void);

Boolean
	CommandLine_IsTextEntered				(void);

Boolean
	CommandLine_IsVisible					(void);

void
	CommandLine_LoseFocus					(void);

Boolean
	CommandLine_PtInRgn						(const Point							inGlobalPoint,
											 const CommandLineRegionCode			inRegion);

void
	CommandLine_SetActive					(const Boolean							inIsActive);

void
	CommandLine_SetSuspended				(Boolean								inIsSuspended);

void
	CommandLine_SetText						(void*									inDataPtr,
											 Size 									inSize);

void
	CommandLine_SetVisible					(Boolean								inIsVisible);

void
	CommandLine_ToggleVisible				(void);

OSErr
	CommandLine_Update						(GrafPtr								inDrawingPort);

void
	CommandLine_Zoom						(short									inWindowPart,
											 short									inModifiers);

#endif