/*###############################################################

	Commands.c
	
	Manages Telnet functions based on command IDs.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>
#include <stdio.h>
#include <string.h>

// Mac includes
#include <AEDataModel.h>
#include <AERegistry.h>
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <Folders.h>
#include <Fonts.h>
#include <PLStringFuncs.h>
#include <Printing.h>
#include <Processes.h>
#include <Resources.h>
#include <Script.h>
#include <TextUtils.h>
#include <ToolUtils.h>

// resource includes
#include "MenuResources.h"
#include "StringResources.h"

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// MacTelnet includes
#include "AboutBox.h"
#include "BasicTypesAE.h"
#include "Commands.h"
#include "Clipboard.h"
#include "CommandLine.h"
#include "configure.proto.h"
#include "Connections.h"
#include "Console.h"
#include "CoolStuff.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "FileUtilities.h"
#include "FormatDialog.h"
#include "FTPLog.h"
#include "Inspector.h"
#include "InstallAE.h"
#include "InternalEvents.h"
#include "Keypads.h"
#include "MacLook.h"
#include "MacroSetupDialog.h"
#include "NetEvents.h"
#include "network.proto.h"
#include "NewSessionsDialog.h"
#include "Parse.h"
#include "Preferences.h"
#include "PreferencesDialog.h"
#include "RasterGraphicsScreen.h"
#include "RealScreen.h"
#include "RequiredAE.h"
#include "ResourceFileManager.h"
#include "ScreenManager.h"
#include "sets.proto.h"
#include "SizeDialog.h"
#include "TektronixRealGraphics.h"
#include "TelnetErrors.h"
#include "TelnetMacros.h"
#include "TelnetMain.h"
#include "TelnetPrinting.h"
#include "TelnetMenus.h"
#include "Terminal.h"
#include "Terminology.h"
#include "vsdata.h"
#include "URL.h"



// types
struct UndoDataFontSizeChanges
{
	// context data for the context ID "kUndoableContextIdentifierTerminalFontSizeChanges"
	TelnetScreenID		screenID;		// which screen was reformatted
	Boolean				undoFontSize,	// is this Undo action going to reverse the font size changes?
						undoFont;		// is this Undo action going to reverse the font changes?
	SInt16				fontSize;		// the old font size (ignored if "undoFontSize" is false)
	SInt16				fontID;			// the old font (ignored if "undoFont" is false)
};
#ifndef __cplusplus
typedef struct UndoDataFontSizeChanges		UndoDataFontSizeChanges;
#endif
typedef UndoDataFontSizeChanges*			UndoDataFontSizeChangesPtr;

// internal methods
static void		hideRealScreenWindow			(const TelnetScreenID			inScreenID);
static void		installUndoFontSizeChanges		(const TelnetScreenID			inScreenID,
												 const Boolean					inUndoFont,
												 const Boolean					inUndoFontSize);
static void		reverseFontChanges				(const UndoActionInstruction	inDoWhat,
												 UndoableContextIdentifier		inContextIdentifier,
												 void*							inContextPtr);



//
// public methods
//

/*!
This method handles Telnet commands using the new
command IDs assigned in Telnet 3.0 for OS 8.  If
the given command is valid, true is returned;
otherwise, false is returned.

(3.0)
*/
Boolean
Commands_ExecuteByID	(UInt32		inCommandID,
						 short		inModifiers)
{
	register SInt16			i = 0;
	ConnectionDataPtr		currentConnectionDataPtr = null;
	Boolean					isScreen = ScreenManager_IsWindowAScreen(GetRealFrontWindow()), // safety net - use it!
							isAtLeastOneOpenConnection = (Connections_GetCount() >= 1),
							isConnectionWindow = false;
	Boolean					result = true;
	
	
	if (isScreen)
	{
		currentConnectionDataPtr = ScreenManager_GetWindowConnectionData(GetRealFrontWindow());
		isScreen = (isScreen && (currentConnectionDataPtr != null)); // additional safety net!
		isConnectionWindow = isAtLeastOneOpenConnection;
	}
	switch (inCommandID)
	{
		case kCommandAboutTelnet:
			if ((inModifiers & shiftKey) && (inModifiers & optionKey))
			{
				CoolStuff_QuadDamageToggle();
			}
			else AboutBox_Display();
			break;
		
		case kCommandTelnetHomePage:
		case kCommandTelnetAuthorMail:
		case kCommandTelnetSourceLicense:
		case kCommandTelnetLibraryLicense:
		case kCommandTelnetProjectStatus:
			{
				Str255			theURL;
				SInt16			theURLStringIndex = 0;
				
				
				// Fire off a URL-handling request for the appropriate (web or mail) URL -
				// this will automatically launch the users helper program (such as a
				// web browser or mail application) to further parse the appropriate URL.
				// Note that MacTelnet can now correctly process ANY kind of URL, so there
				// is complete freedom in what Internet Resources commands can be.
				if (inCommandID == kCommandTelnetAuthorMail) theURLStringIndex = siInternetResourceMailToTheAuthor;
				else if (inCommandID == kCommandTelnetSourceLicense) theURLStringIndex = siInternetResourceSourceLicense;
				else if (inCommandID == kCommandTelnetLibraryLicense) theURLStringIndex = siInternetResourceLibraryLicense;
				else if (inCommandID == kCommandTelnetProjectStatus) theURLStringIndex = siInternetResourceProjectStatus;
				else theURLStringIndex = siInternetResourceOfficialHomePage;
				GetIndString(theURL, rStringsInternetResources, theURLStringIndex);
				URL_SendGetURLEventToSelf(theURL + 1, StrLength(theURL));
			}
			break;
		
		case kCommandOpenConnection:
			{
				int	variant = 0; // temp
				
				
				if (inModifiers & optionKey) variant = 1;
				else if (inModifiers & shiftKey) variant = 2;
				else if (inModifiers & controlKey) variant = 3;
				NewSessionsDialog_Init();
				NewSessionsDialog_Display();
				NewSessionsDialog_Done();
			}
			break;
	
		case kCommandLoadSet:
			Sets_Load();
			break;
		
		case kCommandCloseConnection:
			if (!(inModifiers & optionKey))
			{
				CloseAWindow(GetRealFrontWindow());
			}
			else if (Connections_GetLiveCount() > 0)
			{
				// option-close means disconnect all, no warnings
				ConnectionDataPtr		theScreen = null;
				
				
				for (i = Connections_GetCount(); i >= 1; i--) 
				{
					// 3.0 - to reduce redundant code, this command is now implemented by simply posting events
					theScreen = ScreenManager_GetConnectionDataOf(i);
					NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventClose, theScreen->port, 0);
					NetEvents_Service();
				}
			}
			break;
		
		case kCommandSaveSet:
			if (isConnectionWindow) Sets_Save();
			break;
		
		case kCommandImportMacroSet:
			{
				Boolean		doImport = false;
				
				
				// only warn the user if there are any macros in the active set that can be destroyed
				if (!Macros_AllEmpty(Macros_GetActiveSet()))
				{
					InterfaceLibAlertRef		box = null;
					Str255						dialogText;
					ModalFilterUPP				upp = NewModalFilterUPP(StandardAlertEventFilter);
					
					
					GetIndString(dialogText, rStringsCautionAlerts, siImportOverwriteWarning);
					
					box = Alert_New();
					Alert_SetFilterProc(box, upp);
					Alert_SetHelpButton(box, false);
					Alert_SetParamsFor(box, kAlert_StyleOKCancel);
					Alert_SetText(box, dialogText, EMPTY_PSTRING);
					Alert_SetType(box, kAlertCautionAlert);
					BEEP;
					Alert_Display(box);
					if (Alert_ItemHit(box) == kAlertStdAlertOKButton) doImport = true;
					Alert_Dispose(&box);
					DisposeModalFilterUPP(upp), upp = null;
				}
				else doImport = true;
				
				if (doImport)
				{
					MacroSet			temp = Macros_NewSet();
					TelnetMacroMode		mode = kTelnetMacroModeCommandDigit;
					
					
					if (Macros_ImportFromText(temp, (FSSpec*)null, &mode))
					{
						// Then the user did not cancel and no errors occurred,
						// so the macros must have imported successfully.  Tell
						// the user the good news.
						Macros_Copy(temp, Macros_GetActiveSet());
						Macros_SetMode(mode);
						// should issue message to preferences window panel here, to update display;
						// also, the alert should not appear if the macros panel is displayed, because
						// it should then be obvious to the user that the new macros were imported
						// - unimplemented -
						Errors_DisplayNoteMessage(rStringsNoteAlerts, siMacroImportSuccessful, -1);
					}
					Macros_DisposeSet(&temp);
				}
			}
			break;
		
		case kCommandExportCurrentMacroSet:
			// only stop the user if there are no macros in the active set (and therefore no macros to export)
			if (Macros_AllEmpty(Macros_GetActiveSet()))
			{
				Str255		dialogText,
							helpText,
							alertTitle;
				
				
				GetIndString(dialogText, rStringsNoteAlerts, siExportNotPossible);
				GetIndString(helpText, rStringsNoteAlerts, siExportNotPossibleHelpText);
				GetIndString(alertTitle, rStringsAlertTitles, siAlertTitleExportNotPossible);
				
				// send a Display Alert Apple Event
				{
					TelnetAlertParameterBlock	parameters;
					
					
					Memory_WriteZero((Ptr)&parameters, sizeof(parameters));
					parameters.dialog.type = kAlertStopAlert;
					parameters.flags.alertSound = true;
					parameters.text.large = dialogText;
					parameters.text.small = helpText;
					parameters.text.title = alertTitle;
					(SInt16)Errors_DisplayParameterizedAlert(&parameters);
				}
			}
			else
			{
				Macros_ExportToText(Macros_GetActiveSet(), null/* no file; prompt user */, Macros_GetMode());
			}
			break;
		
		case kCommandFind:
			// send a Launch Find Apple Event, to allow scripts that may be recording to pick up the command
			{
				AppleEvent		launchFindEvent = { typeNull, null },
								reply = { typeNull, null };
				
				
				Alert_ReportOSErr(InstallAE_CreateRecordableAppleEvent(kTelnetEventClass, kTelnetEventIDLaunchFind,
																		&launchFindEvent));
				Alert_ReportOSErr(AESend(&launchFindEvent, &reply, kAENoReply | kAENeverInteract, kAENormalPriority,
											kAEDefaultTimeout, null, null));
				AEDisposeDesc(&launchFindEvent);
			}
			break;
		
		case kCommandPrint:
			// print the selection using the print dialog
			Printing_SetMode(kUniversalPrintingModeNormal);
			TelnetPrinting_PrintSelection();
			break;
		
		case kCommandPrintOne:
			// print the selection immediately, assuming
			// only one copy is desired and the previous
			// print parameters will be used
			Printing_SetMode(kUniversalPrintingModeOneCopy);
			TelnetPrinting_PrintSelection();
			break;
		
		case kCommandPrintScreen:
			if ((i = ScreenManager_GetWindowScreenID(GetRealFrontWindow())) > -1)
			{
				RealScreenDataPtr	realScreenDataPtr = ScreenManager_GetRealScreenDataOf(i);
				Point				start,
									end,
									oldStart,
									oldEnd;
				Boolean				isRectangular = false,
									isSelection = RealScreen_IsTextSelected(i);
				
				
				// preserve any existing selection
				if (isSelection)
				{
					oldStart = realScreenDataPtr->anchor;
					oldEnd = realScreenDataPtr->last;
					isRectangular = RealScreen_IsTextSelectionRectangular(i);
				}
				else
				{
					SetPt(&oldStart, 0, 0);
					SetPt(&oldEnd, 0, 0);
				}
				
				// select the entire visible screen and print it
				SetPt(&start, -1, -1);
				SetPt(&end, realScreenDataPtr->width, realScreenDataPtr->height);
				RealScreen_SelectVirtualRange(i, start, end);
				Printing_SetMode(kUniversalPrintingModeOneCopy);
				TelnetPrinting_PrintSelection();
				
				// clear the full screen selection, restoring any previous selection
				RealScreen_MakeSelectionsRectangular(i, isRectangular);
				RealScreen_SelectVirtualRange(i, oldStart, oldEnd);
			}
			break;
		
		case kCommandPageSetup:
			// work with a saved reference to a printing record in the preferences file
			TelnetPrinting_PageSetup(ResourceFileManager_GetResFile(kTelnetResourceFilePreferences));
			break;
		
		case kCommandQuit:
			// send a Quit Apple Event, to allow scripts that may be recording to pick up the command
			{
				AppleEvent		quitEvent = { typeNull, null },
								reply = { typeNull, null };
				OSErr			error = noErr;
				
				
				error = InstallAE_CreateRecordableAppleEvent(kCoreEventClass, kAEQuitApplication, &quitEvent);
				if (error == noErr)
				{
					error = AESend(&quitEvent, &reply, kAENoReply | kAENeverInteract, kAENormalPriority,
									kAEDefaultTimeout, null, null);
				}
				AEDisposeDesc(&quitEvent);
				Alert_ReportOSErr(error);
			}
			break;
		
		case kCommandUndo:
			Undoables_UndoLastAction();
			break;
		
		case kCommandRedo:
			Undoables_RedoLastUndo();
			break;
		
		case kCommandCut:
		case kCommandCopy:
		case kCommandCopyTable:
		case kCommandClear:
			if ((inCommandID == kCommandCut) ||
				(inCommandID == kCommandCopy) ||
				(inCommandID == kCommandCopyTable))
			{
				// all of these operations involving copying to the clipboard
				SInt16		j = 0;
				
				
				j = MacRGfindwind(GetRealFrontWindow()); // ICR window?
				if (j >= 0)
					MacRGcopy(GetRealFrontWindow()); // copy the ICR window
				else
				{
					if (TektronixRealGraphics_IsRealGraphicsWindow(GetRealFrontWindow(), &j))
					{
						Clipboard_GraphicsToScrap(j); // copy graphics
					}
					else
					{
						// copy text by sending a recordable Apple Event back to MacTelnet
						AppleEvent		copyEvent = { typeNull, null },
										reply = { typeNull, null };
						OSErr			error = noErr;
						
						
						error = InstallAE_CreateRecordableAppleEvent(kTelnetEventClass,
																		kTelnetEventIDCopyToClipboard, &copyEvent);
						if (error == noErr)
						{
							// include copy using spaces for tabs option
							{
								Boolean		copyTable = (inCommandID == kCommandCopyTable);
								
								
								(OSErr)AEPutParamPtr(&copyEvent, kTelnetParameterKeyCopyToClipboardTable,
														typeBoolean, &copyTable, sizeof(copyTable));
							}
							
							error = AESend(&copyEvent, &reply, kAENoReply | kAENeverInteract, kAENormalPriority,
											kAEDefaultTimeout, null, null);
						}
						AEDisposeDesc(&copyEvent);
						Alert_ReportOSErr(error);
					}
				}
			}
			if ((inCommandID == kCommandCut) || (inCommandID == kCommandClear))
			{
				// delete selection -- unimplemented, impossible to implement?
				// at least, impossible until MacTelnet supports generic
				// terminal windows (i.e. windows that are not tied to read-only
				// concepts like connections to remote servers)
			}
			break;

		case kCommandPaste:	
			if (isConnectionWindow) Clipboard_TextFromScrap(); // paste if there is a window to paste into
			break;
		
		case kCommandType:
			if (isConnectionWindow) Clipboard_TextType(); // type if there is a window to type into
			break;
		
		case kCommandSelectAll: // text only
			if ((i = ScreenManager_GetWindowScreenID(GetRealFrontWindow())) > -1)
			{
				RealScreenDataPtr	realScreenDataPtr = ScreenManager_GetRealScreenDataOf(i);
				Point				start,
									end;
				
				
				SetPt(&start, -1, -1);
				SetPt(&end, realScreenDataPtr->width, realScreenDataPtr->height);
				RealScreen_SelectVirtualRange(i, start, end);
			}
			break;
		
		case kCommandShowClipboard:
			Clipboard_ToggleVisible();
			break;
		
		case kCommandPreferences:
			PreferencesDialog_Display();
			break;
		
		case kCommandLargeScreen:
		case kCommandSmallScreen:
			if (isScreen)
			{
				SInt16		columns = 0,
							rows = 0;
				
				
				if (inCommandID == kCommandLargeScreen)
				{
					columns = 132;
					rows = 24;
				}
				else
				{
					columns = 80;
					rows = 24;
				}
				
				// make this undoable
				SizeDialog_InstallUndoDimensionChanges(currentConnectionDataPtr->vs);
				
				// change the window size
				ScreenManager_ChangeDimensionsOf(currentConnectionDataPtr->vs, columns, rows, true/* recordable */);
			}
			break;
		
		case kCommandSetScreenSize:
			if (isScreen)
			{
				SizeDialog_Init(currentConnectionDataPtr->vs);
				SizeDialog_Display();
				SizeDialog_Done();
			}
			break;
		
		case kCommandBiggerText:
		case kCommandSmallerText:
			if (isScreen)
			{
				SInt16		fontSize = 0;
				
				
				// determine the new font size
				RSgetfont(currentConnectionDataPtr->vs, null/* font */, &fontSize);
				if (kCommandBiggerText == inCommandID) ++fontSize;
				else if (kCommandSmallerText == inCommandID)
				{
					if (fontSize > 4/* arbitrary */) --fontSize;
				}
				
				// set the window size to fit the new font size optimally
				installUndoFontSizeChanges(currentConnectionDataPtr->vs, false/* undo font */, true/* undo font size */);
				RSchangefont(currentConnectionDataPtr->vs, null/* font */, fontSize);
				
				result = noErr;
			}
			break;
		
		case kCommandFullScreen:
			if (isScreen)
			{
				SInt16				fontID = 0,		// current terminal font
									fontSize = 0;	// current font size (will change)
				RealScreenDataPtr	realScreenPtr = ScreenManager_GetRealScreenDataOf(currentConnectionDataPtr->vs);
				
				
				if (realScreenPtr != null) if (!RSgetfont(currentConnectionDataPtr->vs, &fontID, &fontSize))
				{
					WindowRef		window = GetRealFrontWindow();
					GrafPtr			oldPort = null,
									windowPort = null;
					SInt16			preservedFontID = 0,
									preservedFontSize = 0;
					SInt16			visibleColumns = VSgetcols(currentConnectionDataPtr->vs),
									visibleRows = VSgetlines(currentConnectionDataPtr->vs);
					Style			preservedFontStyle = 0;
					Str255			fontName;
					Rect			screenBounds,
									viewableBounds;
					FontInfo		info;
					
					
					Console_WriteValue("visible columns", visibleColumns);
					Console_WriteValue("visible rows", visibleRows);
					
				#if TARGET_API_MAC_OS8
					GetFontName(fontID, fontName);
				#else
					(OSStatus)FMGetFontFamilyName(fontID, fontName);
				#endif
					MacLook_GetWindowDeviceGrayRect(window, &screenBounds);
					
					GetPort(&oldPort);
					SetPortWindowPort(window);
					GetPort(&windowPort);
					preservedFontID = GetPortTextFont(windowPort);
					preservedFontSize = GetPortTextSize(windowPort);
					preservedFontStyle = GetPortTextFace(windowPort);
					
					// determine the rectangle that *will* be the optimal new terminal screen area
					SetRect(&viewableBounds,
							screenBounds.left,
							screenBounds.top + RealScreen_GetPlacardHeight(currentConnectionDataPtr->vs),
							screenBounds.right - RealScreen_GetScrollBarWidth(),
							screenBounds.bottom - RealScreen_GetScrollBarHeight());
					
					// choose an appropriate font size to best fill the screen; favor maximum
					// width over height, but reduce the font size if the characters overrun
					// the bottom of the screen
					TextFont(fontID);
					while ((CharWidth('A') * visibleColumns) < (viewableBounds.right - viewableBounds.left))
					{
						TextSize(++fontSize);
					}
					GetFontInfo(&info);
					while (((info.ascent + info.descent + INTEGER_DOUBLED(info.leading)) * visibleRows) >=
							(viewableBounds.bottom - viewableBounds.top))
					{
						TextSize(--fontSize);
						GetFontInfo(&info);
					}
					
					// undo the damage...
					TextFont(preservedFontID);
					TextSize(preservedFontSize);
					TextFace(preservedFontStyle);
					
					// resize the window and center it on its monitor
					installUndoFontSizeChanges(currentConnectionDataPtr->vs, false/* undo font */, true/* undo font size */);
					MoveWindow(window, screenBounds.left, screenBounds.top, false/* activate */);
					RSchangefont(currentConnectionDataPtr->vs, fontName, fontSize);
					{
						Rect		contentRect;
						SInt16		h = 0,
									v = 0;
						
						
						MacLook_GetWindowRegionBounds(window, kWindowContentRgn, &contentRect);
						h = INTEGER_HALVED((screenBounds.right - screenBounds.left) -
											(contentRect.right - contentRect.left));
						v = INTEGER_HALVED((screenBounds.bottom - screenBounds.top) -
											(contentRect.bottom - contentRect.top));
						MoveWindow(window, screenBounds.left + h, screenBounds.top + v, true/* activate */);
					}
					SetPort(oldPort);
				}
				else BEEP;
			}
			break;
		
		case kCommandFormat:
			{
				FormatDialogSetupDataPtr	dataPtr = (FormatDialogSetupDataPtr)Memory_NewPtr(sizeof
																				(FormatDialogSetupData));
				
				
				if (dataPtr != null)
				{
					Boolean		changed = false;
					SInt16		index = 0;
					
					
					// initialize structure with data of frontmost session window - incomplete
					{
						SInt16 		fontID = 0;
						
						
						index = kFormatDialogIndexNormalText;
						RSgetfont(currentConnectionDataPtr->vs, &fontID, &dataPtr->format[index].font.size);
					#if TARGET_API_MAC_OS8
						GetFontName(fontID, dataPtr->format[index].font.familyName);
					#else
						(OSStatus)FMGetFontFamilyName(fontID, dataPtr->format[index].font.familyName);
					#endif
						(Boolean)RealScreen_GetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexNormalText,
														&dataPtr->format[index].colors.foreground);
						(Boolean)RealScreen_GetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexNormalBackground,
														&dataPtr->format[index].colors.background);
						dataPtr->format[index].options = 0;
						
						index = kFormatDialogIndexBoldText; // incomplete
						RSgetfont(currentConnectionDataPtr->vs, &fontID, &dataPtr->format[index].font.size);
					#if TARGET_API_MAC_OS8
						GetFontName(fontID, dataPtr->format[index].font.familyName);
					#else
						(OSStatus)FMGetFontFamilyName(fontID, dataPtr->format[index].font.familyName);
					#endif
						(Boolean)RealScreen_GetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexBoldText,
														&dataPtr->format[index].colors.foreground);
						(Boolean)RealScreen_GetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexBoldBackground,
														&dataPtr->format[index].colors.background);
						dataPtr->format[index].options = kFormatDialogFormatOptionDisableFontItems |
															kFormatDialogFormatOptionDisableFontSizeItems;
						
						index = kFormatDialogIndexBlinkingText;
						(Boolean)RealScreen_GetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexBlinkingText,
														&dataPtr->format[index].colors.foreground);
						(Boolean)RealScreen_GetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexBlinkingBackground,
														&dataPtr->format[index].colors.background);
						dataPtr->format[index].options = kFormatDialogFormatOptionDisableFontItems |
															kFormatDialogFormatOptionDisableFontSizeItems;
					}
					dataPtr->generalOptions = 0L; // turn everything off
					
					// display the dialog box, moving changes from the user back to the data structure
					FormatDialog_Init(dataPtr);
					changed = FormatDialog_Display();
					FormatDialog_Done();
					
					if (changed)
					{
						RealScreenDataPtr		realScreenPtr = ScreenManager_GetRealScreenDataOf(currentConnectionDataPtr->vs);
						
						
						// the user clicked OK, so update the window with the new data
						index = kFormatDialogIndexNormalText;
						RealScreen_SetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexNormalText,
											&dataPtr->format[index].colors.foreground);
						RealScreen_SetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexNormalBackground,
											&dataPtr->format[index].colors.background);
						index = kFormatDialogIndexBoldText;
						RealScreen_SetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexBoldText,
											&dataPtr->format[index].colors.foreground);
						RealScreen_SetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexBoldBackground,
											&dataPtr->format[index].colors.background);
						index = kFormatDialogIndexBlinkingText;
						RealScreen_SetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexBlinkingText,
											&dataPtr->format[index].colors.foreground);
						RealScreen_SetColor(currentConnectionDataPtr->vs, kRealScreenColorIndexBlinkingBackground,
											&dataPtr->format[index].colors.background);
						index = kFormatDialogIndexNormalText;
						RSchangefont(currentConnectionDataPtr->vs, dataPtr->format[index].font.familyName,
														dataPtr->format[index].font.size);
					}
				}
				else Alert_ReportOSErr(memFullErr);
			}
			break;
		
		case kCommandCaptureToFile:
			if (isScreen)
			{
				if (Terminal_FileCaptureInProgress(currentConnectionDataPtr->vs))
				{
					Terminal_FileCaptureEnd(currentConnectionDataPtr->vs);
				}
				else
				{
					if (Terminal_FileCaptureBegin(currentConnectionDataPtr->vs, null/* let user choose a file */))
					{
						// then capturing is on
					}
				}
			}
			break;
		
		case kCommandEcho:
			if (isConnectionWindow)
			{
				// toggle local echo, if possible
				if (currentConnectionDataPtr->echo && (currentConnectionDataPtr->kblen > 0))
				{
					// if not empty, send buffer
					Network_SendData(currentConnectionDataPtr->port, currentConnectionDataPtr->kbbuf,
												currentConnectionDataPtr->kblen);
					currentConnectionDataPtr->kblen = 0;
				}
				currentConnectionDataPtr->echo= !currentConnectionDataPtr->echo;
				Parse_SendLocalEchoOption(currentConnectionDataPtr);
			}
			break;
			
		case kCommandWrapMode:
			if (isConnectionWindow)
			{
				currentConnectionDataPtr->wrap = !currentConnectionDataPtr->wrap;
				if (currentConnectionDataPtr->wrap)
				{
					// turn on wrap mode
					VSwrite(currentConnectionDataPtr->vs, "\033[?7h", 5); // kick emulator
				}
				else
				{
					VSwrite(currentConnectionDataPtr->vs, "\033[?7l", 5);
				}
			}
			break;
		
		case kCommandClearScreenSavesLines:
			if (isConnectionWindow)
			{
				currentConnectionDataPtr->ESscroll = !currentConnectionDataPtr->ESscroll;
				Terminal_SavesLinesOnClear(currentConnectionDataPtr->vs, currentConnectionDataPtr->ESscroll);
			}
			break;
		
		case kCommandJumpScrolling:
			if (isConnectionWindow)
			{
				NetEvents_FlushNetworkForConnection(currentConnectionDataPtr);
			}
			break;
		
		case kCommandTEKPageCommand:
			// Tektronix page command
			if (isConnectionWindow)
			{
				Parse_ShowOrInterpretAsCommand(currentConnectionDataPtr, (unsigned char*)"\033\014", 2);
			}
			break;
		
		case kCommandTEKPageClearsScreen:
			// clear on TEK page
			if (isConnectionWindow)
			{
				currentConnectionDataPtr->TEK.pageLocation = !currentConnectionDataPtr->TEK.pageLocation;
			}
			break;
		
		case kCommandSpeechEnabled:
			if (isConnectionWindow)
			{
				RealScreenDataPtr		realScreenPtr = ScreenManager_GetRealScreenDataOf(currentConnectionDataPtr->vs);
				
				
				realScreenPtr->speech.enabled = !realScreenPtr->speech.enabled;
			}
			break;
		
		case kCommandResetTerminal:
			// reset the terminal, restoring wrapping
			if (isConnectionWindow)
			{
				Terminal_Reset(currentConnectionDataPtr->vs);
				if (currentConnectionDataPtr->wrap) VSwrite(currentConnectionDataPtr->vs, "\033[?7h", 5);
				currentConnectionDataPtr->timing = 0;
			}
			break;
		
		case kCommandDeletePressSendsBackspace:
		case kCommandDeletePressSendsDelete:
			if (isConnectionWindow)
			{
				currentConnectionDataPtr->bsdel = (inCommandID == kCommandDeletePressSendsDelete);
			}
			break;
		
		case kCommandEMACSArrowMapping:
			if (isConnectionWindow)
			{
				currentConnectionDataPtr->arrowmap = !currentConnectionDataPtr->arrowmap;
			}
			break;
		
		case kCommandLocalPageUpDown:
			if (isConnectionWindow)
			{
				currentConnectionDataPtr->pgupdwn = !currentConnectionDataPtr->pgupdwn;
			}
			break;
		
		case kCommandSetKeys:
			MacLook_SetUpKeys(currentConnectionDataPtr->vs);
			break;
		
		case kCommandShowNetworkNumbers:
			Network_ShowIPAddress();
			break;
		
		case kCommandSendInternetProtocolNumber:
		case kCommandSendFTPCommand:
			if (isConnectionWindow)
			{
				char	tmpout[30]; // the final command (containing whatever)
				char	ipout[30]; // the IP storage
				
				
				strcpy(tmpout, ""); // 3.0 - clear that damned string to be safe
				if (currentConnectionDataPtr->echo && (currentConnectionDataPtr->kblen > 0))
				{
					Network_SendData(currentConnectionDataPtr->port, currentConnectionDataPtr->kbbuf,
												currentConnectionDataPtr->kblen); // if not empty, send buffer
					currentConnectionDataPtr->kblen = 0;
				}
				if (Network_CurrentIPAddressToCString(ipout)) // 3.0
				{
					if (inCommandID == kCommandSendFTPCommand)
					{
						if ((gFTPServerPrefs->serverMode == kFTPServerModeOnNoProtection) && !(inModifiers & shiftKey))
							sprintf(tmpout, "ftp -n ");
						else
							sprintf(tmpout, "ftp ");
						strcat(tmpout, ipout); // tack on the IP number
						strcat(tmpout, "\015\012"); // tack on a CR
					}
					else strcat(tmpout, ipout);
					
					// dump it to the screen
					Network_SendData(currentConnectionDataPtr->port, tmpout, strlen(tmpout));
					if (currentConnectionDataPtr->echo) VSwrite(currentConnectionDataPtr->vs, tmpout, strlen(tmpout));
				}
				else BEEP; // IP-to-string failed in this case
			}
			break;
		
		case kCommandSendSync:
			Parse_SendRecordableSpecialTransmitEvent(kTelnetEnumeratedClassSpecialCharacterSync,
														true/* execute */);
			break;
		
		case kCommandSendBreak:
			Parse_SendRecordableSpecialTransmitEvent(kTelnetEnumeratedClassSpecialCharacterBreak,
														true/* execute */);
			break;

		case kCommandSendInterruptProcess:
			Parse_SendRecordableSpecialTransmitEvent(kTelnetEnumeratedClassSpecialCharacterInterruptProcess,
														true/* execute */);
			break;
		
		case kCommandSendAbortOutput:
			Parse_SendRecordableSpecialTransmitEvent(kTelnetEnumeratedClassSpecialCharacterAbortOutput,
														true/* execute */);
			break;
		
		case kCommandSendAreYouThere:
			Parse_SendRecordableSpecialTransmitEvent(kTelnetEnumeratedClassSpecialCharacterAreYouThere,
														true/* execute */);
			break;

		case kCommandSendEraseCharacter:
			Parse_SendRecordableSpecialTransmitEvent(kTelnetEnumeratedClassSpecialCharacterEraseCharacter,
														true/* execute */);
			break;

		case kCommandSendEraseLine:
			Parse_SendRecordableSpecialTransmitEvent(kTelnetEnumeratedClassSpecialCharacterEraseLine,
														true/* execute */);
			break;
		
		case kCommandSendEndOfFile:
			Parse_SendRecordableSpecialTransmitEvent(kTelnetEnumeratedClassSpecialCharacterEndOfFile,
														true/* execute */);
			break;
		
		case kCommandWatchForActivity:
			if (isScreen)
			{
				currentConnectionDataPtr->watched = !currentConnectionDataPtr->watched;
				/*
				if (currentConnectionDataPtr->watched)
				{
					InterfaceLibAlertRef	box = null;
					Str255					dialogText,
											helpText;
					ModalFilterUPP			upp = NewModalFilterProc(StandardAlertEventFilter);
					
					
					GetIndString(dialogText, rStringsNoteAlerts, siActivityWatchBegun);
					GetIndString(helpText, rStringsNoteAlerts, siActivityWatchBegunHelpText);
					
					box = Alert_New();
					Alert_SetFilterProc(box, upp);
					Alert_SetHelpButton(box, false);
					Alert_SetParamsFor(box, kAlert_StyleOK);
					Alert_SetText(box, dialogText, helpText);
					Alert_SetType(box, kAlertNoteAlert);
					Alert_Display(box);
					Alert_Dispose(&box);
					DisposeModalFilterUPP(upp), upp = null;
				}
				*/
			}
			break;
			
		case kCommandSuspendNetwork:
			FlagManager_Set(kFlagNetworkSuspended, !FlagManager_Test(kFlagNetworkSuspended));
			break;
		
		case kCommandHideFrontWindow:
			if (isScreen) hideRealScreenWindow(ScreenManager_GetWindowScreenID(GetRealFrontWindow()));
			break;
		
		case kCommandHideOtherWindows:
			{
				SInt16		limit = Connections_GetCount();
				
				
				for (i = 1; i <= limit; i++)
				{
					if (i != currentConnectionDataPtr->vs) hideRealScreenWindow(i);
				}
			}
			break;
		
		case kCommandShowAllHiddenWindows:
			{
				RealScreenDataPtr		realScreenPtr = null;
				SInt16					limit = Connections_GetCount();
				
				
				for (i = 1; i <= limit; i++)
				{
					realScreenPtr = ScreenManager_GetRealScreenDataOf(i);
					if (realScreenPtr != null)
					{
						if ((realScreenPtr->window != null) && (realScreenPtr->obscured))
						{
							realScreenPtr->obscured = false;
							ShowWindow(realScreenPtr->window);
							SetMenuMarkFromScreen(kMenuMarkTypeEnableIcon, i);
						}
					}
				}
				ScreenManager_SetHiddenCount(0);
			}
			break;
		
		case kCommandStackWindows:
			RealScreen_StackWindows();
			break;
		
		case kCommandNextWindow:
			RealScreen_NextWindow();
			break;
		
		case kCommandChangeWindowTitle:
			RealScreen_DisplaySetWTitleDialogFor(GetRealFrontWindow());
			break;
		
		case kCommandMinimizeWindow:
			{
				WindowRef		frontWindow = GetRealFrontWindow();
				
				
				CollapseWindow(frontWindow, !IsWindowCollapsed(frontWindow));
				// send recordable Apple Event back to MacTelnet - unimplemented
			}
			break;
		
		case kCommandShowSessionInspector:
			Inspector_ToggleVisible();
			break;
		
		case kCommandShowFTPLog:
			FTPLog_SetVisible(!FTPLog_IsVisible());
			if (FTPLog_IsVisible()) SelectOverRealFrontWindow(FTPLog_GetWindow());
			break;
		
		case kCommandShowCommandLine:
			CommandLine_ToggleVisible();
			break;
		
		case kCommandShowKeypad:
			Keypads_ToggleVisibility(kKeypadTypeVT220Keys);
			break;
		
		case kCommandShowFunction:
			Keypads_ToggleVisibility(kKeypadTypeFunctionKeys);
			break;
		
		case kCommandShowControlKeys:
			Keypads_ToggleVisibility(kKeypadTypeControlKeys);
			break;
		
		default:
			result = false;
			break;
	}
	
	return result;
}// ExecuteByID


/*!
To run the Find File or Sherlock application,
use this method.  If the application cannot
be found, the user is asked to locate it and
its location is saved in the preferences file.

(3.0)
*/
OSErr
Commands_LaunchFind		(void)
{
	Str31		sherlockName;
	FSSpec		sherlockSpec; // type 'APPL', creator 'fndf'
	Boolean		launchSuccessful = false;
	OSErr		result = noErr;
	
	
	// guess where Sherlock is
	{
		SInt32		dirID = 0L;
		SInt16		vRefNum = 0;
		
		
		if (FlagManager_Test(kFlagOS8_5API))
		{
			GetIndString(sherlockName, rStringsMiscellaneous, siMacOS8_5AndBeyondFindFileName);
		}
		else
		{
			GetIndString(sherlockName, rStringsMiscellaneous, siPreMacOS8_5FindFileName);
		}
		result = FindFolder(kOnSystemDisk, kAppleMenuFolderType, kDontCreateFolder, &vRefNum, &dirID);
		if (result == noErr) result = FSMakeFSSpec(vRefNum, dirID, sherlockName, &sherlockSpec);
	}
	
	// If the systems Find File (or Sherlock) application cannot be found, first
	// check the global preferences for any previous-saved file location: it may
	// be possible to find the alternate location without asking the user (because
	// the user would have been asked before in order for the preferences to get
	// set).
	if (result != noErr)
	{
		Size		actualSize = 0L;
		
		
		result = noErr; // give it another chance
		
		// get the users Find File application preference, if possible
		if (Preferences_GetData(kTelnetPreferenceClassGeneral,
									kTelnetPreferenceTag_FindFileAppSpec,
									sizeof(sherlockSpec), &sherlockSpec,
									&actualSize) == kTelnetPrefsResultCodeSuccess)
		{
			result = FSMakeFSSpec(sherlockSpec.vRefNum, sherlockSpec.parID,
									sherlockSpec.name, &sherlockSpec);
		}
		else
		{
			// ??? - some error occurred
			result = fnfErr;
		}
		
		if (result == noErr)
		{
			// found a preference for an application, and it exists; launch it
			result = FileUtilities_LaunchApplicationFromFSSpec(&sherlockSpec);
			launchSuccessful = (result == noErr);
		}
	}
	
	// If none of the above measures worked, give in and display
	// a query dialog box asking the user where this program is!
	// If the user picks something, save that location - if it's
	// right, it may not be necessary to query the user again
	// the next time around.
	if (!launchSuccessful)
	{
		Boolean		doContinue = false;
		Str255		prompt,
					title;
		
		
		result = noErr; // give it another chance
		GetIndString(prompt, rStringsNavigationServices, siNavPromptLocateApplication);
		GetIndString(title, rStringsNavigationServices, siNavDialogTitleLocateApplication);
		StringUtilities_PMetaSetDefaults();
		StringUtilities_PBuild(prompt, sherlockName, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
		StringUtilities_PBuild(title, sherlockName, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
		doContinue = GetApplicationSpec(prompt, title, &sherlockSpec);
		
		if (!doContinue) result = userCanceledErr;
		else
		{
			// save the users chosen location so it may be possible to avoid the query next time
			(TelnetPrefsResultCode)Preferences_SetData(kTelnetPreferenceClassGeneral,
														kTelnetPreferenceTag_FindFileAppSpec,
														sizeof(sherlockSpec), &sherlockSpec);
			(OSErr)Preferences_SaveData(kTelnetPreferenceClassGeneral);
		}
		if (doContinue)
		{
			result = FileUtilities_LaunchApplicationFromFSSpec(&sherlockSpec);
			launchSuccessful = (result == noErr);
		}
	}
	
	return result;
}// LaunchFind


//
// internal methods
//

/*!
To obscure a real screen (terminal) window, use
this method.  The specified window is marked as
obscured, its menu item icon is disabled, and
the window zooms to the menu bar for visual
feedback to the user.

(3.0)
*/
static void
hideRealScreenWindow	(const TelnetScreenID		inScreenID)
{
	RealScreenDataPtr		realScreenPtr = ScreenManager_GetRealScreenDataOf(inScreenID);
	
	
	if (realScreenPtr != null)
	{
		if (!realScreenPtr->obscured) // dont hide a window twice
		{
			enum
			{
				kWindowMenuTitleWidthApproximation = 55 // pixels
			};
			Rect		bounds,
						windowMenuTitleRect;
			Boolean		simpleTelnetMode = false;
			Size		actualSize = 0L;
			
			
			// get the users Capture File Creator preference, if possible
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
										kTelnetPreferenceTag_SimpleTelnet,
										sizeof(simpleTelnetMode), &simpleTelnetMode,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				simpleTelnetMode = false; // assume non-Simple mode, if a preference cant be found
			}
			
			
			realScreenPtr->obscured = true;
			MacLook_GetWindowRegionBounds(realScreenPtr->window, kWindowStructureRgn, &bounds);
			
			// approximate the rectangle of the Window menu, which is different in Simple Telnet mode
			// (theres gotta be a beautiful way to get this just right, but Im not finding it now)
			windowMenuTitleRect = (**(GetMainDevice())).gdRect;
			windowMenuTitleRect.bottom = GetMBarHeight();
			windowMenuTitleRect.left += (simpleTelnetMode) ? 180 : 315;
			windowMenuTitleRect.right = windowMenuTitleRect.left + kWindowMenuTitleWidthApproximation;
			
			// make the window zoom into the menu title, for visual feedback
			HiliteWindow(realScreenPtr->window, false); // doing this might avoid a bug on re-display?
			HideWindow(realScreenPtr->window);
			(OSErr)ZoomRects(&bounds, &windowMenuTitleRect, 20/* steps, arbitrary */, kZoomDecelerate);
			
			ScreenManager_SetHiddenCount(ScreenManager_GetHiddenCount() + 1);
			SetMenuMarkFromScreen(kMenuMarkTypeDisableIcon, inScreenID);
		}
	}
}// hideRealScreenWindow


/*!
To install an Undo procedure that will revert
the font and/or font size of the specified
screen to its current font and/or font size
when the user chooses Undo, use this method.

(3.0)
*/
static void
installUndoFontSizeChanges		(const TelnetScreenID		inScreenID,
								 const Boolean				inUndoFont,
								 const Boolean				inUndoFontSize)
{
	OSErr							error = noErr;
	Str255							undoName,
									redoName;
	UndoDataFontSizeChangesPtr		dataPtr = (UndoDataFontSizeChangesPtr)Memory_NewPtr
												(sizeof(UndoDataFontSizeChanges)); // disposed in the action method
	
	
	if (dataPtr == null) error = memFullErr;
	else
	{
		// initialize context structure
		dataPtr->screenID = inScreenID;
		dataPtr->undoFontSize = inUndoFontSize;
		dataPtr->undoFont = inUndoFont;
		RSgetfont(inScreenID, &dataPtr->fontID, &dataPtr->fontSize);
	}
	GetIndString(undoName, rStringsMenuItemTextSubstitutions, siMenuItemTextUndoFormatChanges);
	GetIndString(redoName, rStringsMenuItemTextSubstitutions, siMenuItemTextRedoFormatChanges);
	error = Undoables_AddAction(undoName, redoName, reverseFontChanges,
								kUndoableContextIdentifierTerminalFontSizeChanges, dataPtr);
	if (error != noErr) Console_WriteValue("Warning: Could not make font and/or size change undoable, error", error);
}// installUndoFontSizeChanges


/*!
This routine, of standard UndoActionProcPtr form,
can undo or redo changes to the font and/or font
size of a terminal screen.

Note that it will really be nice to get all of
the AppleScript stuff working, so junk like this
does not have to be done using Copy-and-Paste
coding, and can be made into a recordable event.

(3.0)
*/
static void
reverseFontChanges		(const UndoActionInstruction	inDoWhat,
						 UndoableContextIdentifier		inContextIdentifier,
						 void*							inContextPtr)
{
	// this routine only recognizes one kind of context - be absolutely sure thats what was given!
	if (inContextIdentifier == kUndoableContextIdentifierTerminalFontSizeChanges)
	{
		UndoDataFontSizeChangesPtr		dataPtr = inContextPtr;
		
		
		switch (inDoWhat)
		{
			case kUndoActionInstructionDispose:
				// release memory previously allocated when this action was installed
				if (inContextPtr != null) Memory_DisposePtr((Ptr*)&inContextPtr);
				break;
			
			case kUndoActionInstructionRedo:
			case kUndoActionInstructionUndo:
			default:
				{
					SInt16		oldFontSize = 0,
								oldFontID = 0;
					Str255		fontName;
					
					
					// make this reversible by preserving the dimensions
					RSgetfont(dataPtr->screenID, &oldFontID, &oldFontSize);
					
					// change the font and/or size of the window
				#if TARGET_API_MAC_OS8
					GetFontName(dataPtr->fontID, fontName);
				#else
					(OSStatus)FMGetFontFamilyName(dataPtr->fontID, fontName);
				#endif
					RSchangefont(dataPtr->screenID, fontName, dataPtr->fontSize);
					
					// save the preserved dimensions
					dataPtr->fontSize = oldFontSize;
					dataPtr->fontID = oldFontID;
				}
				break;
		}
	}
}// reverseFontChanges