/*###############################################################

	ConnectionData.h
	
	This file defines the window management data structure
	for MacLook.c.  Telnet 3.0 adds sophisticated data types
	that make it *obvious* the acceptable values for a
	particular field of the structure.  (The corresponding code
	throughout the program has been completely updated to use
	these constants correctly.)  Hopefully, in the future, this
	structure, too, will become opaque like many others have.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#ifndef __CONNECTIONDATA__
#define __CONNECTIONDATA__

// Mac includes
#include <MacTypes.h>
#include <OpenTransportProviders.h>
#include <QuickDraw.h>

// MacTelnet includes
#include "network.proto.h"
#include "ProxyServer.h"
#include "tekdefs.h"



// constants
#define LINE_MODE_SLC_MAX		30		// must be identical to SLC_MAX in "Parse.h"
enum
{
	MAXFTP = 128,
	MAXKB = 256
};

#define MHOPTS_BASE		37	// Base option for {my,his}opts (Authenticate)
							//  {my,his} opts should only be used for telnet options
							//  in the range starting at MHOPTS_BASE and limited
							//  by MHOPTS_SIZE. This saves memory.
#define MHOPTS_SIZE		2	// Number of options supported in {my,his}opts

#define PASTE_IN_PROGRESS	0x4000
#define PASTE_DONE			0x0040

typedef SInt16 ConnectionStatus;
enum
{
	kConnectionStatusNotInUse = -2,
	kConnectionStatusDNRWait = -1,
	kConnectionStatusActive = 0,
	kConnectionStatusOpening = 1,
	kConnectionStatusDead = 2
};



// types
struct ConnectionData
{
	SInt16				vs;				// virtual screen number (TelnetScreenID)
	NetworkPortRef		port;			// internal port (used to access network data)
	SInt16				status;			// a "kConnectionStatus..." constant
	
	UInt32				connectedSince;	// when did this connection start?
	
	WindowRef			window;			// the window for this connection
	
	Handle				setupParametersHandle;	// handle to initial setup (needed after DNS lookup)
	
	struct
	{
		char					canonical[80];	// canonical name
		Str63					name;			// name of machine being connected to
	} host;
	
	struct
	{
		SInt16					emulation;		// virtual terminal emulation type
		Str32					answerBack;		// message to send when server sends TERMTYPE option
	} terminal;
	
	struct
	{
		Boolean					enabled;		// is a proxy server being used?
		ProxyServerRef			ref;			// information on the proxy server (if any)
	} proxy;
	
	struct
	{
		TektronixPageLocation	pageLocation;	// where a new page goes, upon execution of PAGE command
		TektronixMode			mode;
		TektronixGraphicID		graphicsID;		// associated TEK drawing
	} TEK;
	
	SInt16				enabled;		// Are we currently enabled for receive? 
				
	SInt16				bsdel,			// backspace or delete is default 
						eightbit,		// eight bit font displayed (false is seven bit display
						national,		// LU/MP: translation table to use for this connection 
						arrowmap,		// MAT: should we allow the arrow keys to be mapped?? 
						showErrors,		// show ALL errors if this is set 
						pgupdwn,		// JMB/MAT: should we have page up/down do local window movement? 
						keypadmap,		// CCP 2.7: should we have numeric keypad operators work like regular operators? 
						metaKey,		// JMB/SMB:	should option key work as EMACS meta key? 
						Xterm,			// JMB/WNR:	should Xterm sequences be recognized? 
						halfdup,		// If true then half-duplex mode 
						forcesave,		// NCSA 2.5: force lines to be saved 
						crmap,			// CR's second byte is ...
						termstate;		// equal to one of the valid "TelnetEmulationType" constants
	
	struct
	{
		char					suspend,		// character for scrolling to stop
								resume,			// character for scrolling to go
								interrupt,		// character for interrupt process
								pad;			// unused
	} controlKey;
	
	SInt16				addedMenuItem,	// was an item added to the Window menu for this connection?
						wrap,			// whether vtwrap is set or not for this screen 
						echo,			// NOTE: this is backwards - I am echoing your chars 
						ESscroll,		// will Clear Screen Save Lines? 
						unused27;
	
	SInt16				naws,			// will negotiate about window size
						Isga,			// I am supressing go ahead 
						Usga,			// You are supressing go ahead 
						Ittype,			// I am sending terminal type negotiations.. 
						telstate,		// Telnet State 
						timing,			// True if waiting for timing mark 
						remote_flow,	// true if they are dictating flow_control 
						allow_flow,  	// true unless prior is true AND the host has told us to turn flow control off 
						restart_any_flow,// true if any character (except XOFF) will restart flow 
						substat,		// Telnet Subnegot. State 
						parseIndex,		// save an index into the subnegotiation parsed data 
						
						unused38,
						kblen,			// Pointer to next char in buffer to be used 
						ftpport,		// FTP port number
						ftpState,		// Telnet session = 0, Ftp session != 0
						ftpnext,		// Next ftpState after waiting for response from "PORT"
						xfer,			// File X-fer status
						capturesize,	// ftp client size of captured data
						clientflags;	// boolean flags for ftp client 
	
	// paste operations
	struct
	{
		SInt16				method;			// quick or block paste method
		SInt32				blockSize;		// the size of paste blocks
		SInt32				inCount,		// count of bytes into this port 
							outCount;		// count of bytes out this port 
		SInt32				outLength;		// length of text remaining to be pasted
		
		char*				nextCharPtr;	// pointer to next character to send
		char**				text;			// text buffer for pasting
	} paste;
	
	struct
	{
		Boolean					allowed,		// is user preference for line mode set?
								active,			// is line mode active?
								nextIsLiteral;	// should the next key be echoed literal?
		SInt16					mode;			// current MODE (such as EDIT or TRAPSIG)
		SInt16					forwardMask,	// should we forward on certain characters
								numForwardKeys; // how many keys to forward on
		unsigned char			slcLevel[LINE_MODE_SLC_MAX + 1], // levelBits for set-local-characters
								forwardKeys[32],// which keys to forward on 
								slc[LINE_MODE_SLC_MAX + 1]; // actual values
	} lineMode;
	
	char				ftpbuf[MAXFTP],	// Outgoing ftp commands buffer
						kbbuf[MAXKB];	// The keyboard buffer (echo mode ) 
		
	Ptr					aedata;		// Kerberos authentication and encryption data
	
	unsigned char		myopts[MHOPTS_SIZE],	// My telnet options. Warning: see MHOPTS_BASE
						hisopts[MHOPTS_SIZE],	// His telnet options. Warning: see MHOPTS_BASE
						parsedat[450];			// save the subnegotiation data here
												//   must be big enough for a Kerberos AP message
	
	Boolean				authenticate,	// true if authenticating connection wanted
						encrypt;		// true if encrypting connection wanted
	
	Boolean				watched;		// true if the window is currently being watched for activity
	
	SInt16				responseCount;	// used for FTP client windows only
	
	SInt32				blockSize; // used by NetEvents
	
	UInt32				openTimeout;	// number of seconds to wait before giving up connection opening
};
#ifndef __cplusplus
typedef struct ConnectionData	ConnectionData;
#endif
typedef ConnectionData*			ConnectionDataPtr;
typedef ConnectionData*			ConnectionDataArray;

#endif