/*###############################################################

	Connections.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstring>
#else
#	include <string.h>
#endif
#include <ctype.h>

// Mac includes
#include <ASRegistry.h>
#include <Fonts.h>
#include <PLStringFuncs.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "MenuResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "Console.h"
#include "DNR.h"
#include "InternetServices.h"
#include "LineMode.h"
#include "ScreenManager.h"
#include "telneterrors.h"
#include "DialogUtilities.h"

#include "ConnectionData.h"
#include "EventLoop.h"

#include "TelnetMenus.h"
#include "RealScreen.h"
#include "vsdata.h"
#include "vskeys.h"
#include "Terminal.h"
#include "TektronixVirtualGraphics.h"
#include "TektronixRealGraphics.h"
#include "RasterGraphicsKernel.h"
#include "RasterGraphicsScreen.h" 
#include "network.proto.h"
#include "InternalEvents.h"
#include "MacLook.h"
#include "Parse.h"
#include "configure.proto.h"
#include "NetEvents.h"
#include "Preferences.h"
#include "ResourceFileManager.h"
#include "Terminology.h"
#include "InstallAE.h"
#include "ProxyServer.h"

#include "Connections.h"

/*	These are all of the variables we need... */
extern	short		nNational;				// number of user-installed translation tables
static UInt16		gConnectionCount = 0,
					gConnectionActiveCount = 0;

// internal methods
static void					killConnection							(ConnectionDataPtr				inConnectionDataPtr);
static ConnectionSetup**	nameToConnectionSetup					(StringPtr						inString,
																	 Boolean						inUseDefaultTerminal);
static void					processURLEscapeCodes					(char*							url,
																	 char**							end);
static void					tooManyOpenConnections					(void);



//
// public methods
//

/*!
When a connection is ready to open, call this
method.  If a value other than "noErr" is given
as "inDNRError", then that error is reported;
otherwise, the connection is initialized with
the given IP and connection name information,
is marked as "opening" in the Window menu, and
is marked as "opening" internally.  Whether and
when the connection actually opens depends on
network events, which are unpredictable.

(2.6)
*/
void
Connections_CompleteOpening		(TelnetScreenID			inScreenID,
								 UInt32					inIPAddress,
								 OSStatus				inDNRError,
								 char*					inConnectionName)
{
	ConnectionSetup**		setupParametersHandle = null;
	ConnectionDataPtr		connectionDataPtr = null;
	Boolean					isError = false;
	
	
	connectionDataPtr = ScreenManager_GetConnectionDataOf(inScreenID);
	if (connectionDataPtr == null)
	{
		Console_WriteLine("could not complete connection opening, data could not be found");
		isError = true;
	}
	else if (connectionDataPtr->status == kConnectionStatusDNRWait)
	{
		setupParametersHandle = (ConnectionSetup**)connectionDataPtr->setupParametersHandle;
		
		if (inDNRError == noErr)
		{
			HLockHi((Handle)(**setupParametersHandle).session);
			
			if ((**(**setupParametersHandle).session).NetBlockSize < 512)	
				(**(**setupParametersHandle).session).NetBlockSize = 512; // less than this can really get messy
			
			if (NetEvents_SetBlockSize((**(**setupParametersHandle).session).NetBlockSize, connectionDataPtr) == 0)
			{
				short		sessionPort = (connectionDataPtr->ftpState != 0)
											? HFTP : (**(**setupParametersHandle).session).port;
				
				
				connectionDataPtr->port = Network_SocketOpen(inIPAddress, sessionPort, 0/* random local port */,
																connectionDataPtr->openTimeout/* seconds of timeout */);
				
				// The canonical hostname is needed for Kerberos.
				// Make a best guess if the DNR did not return a cname.
				if (inConnectionName != null) strncpy(connectionDataPtr->host.canonical, inConnectionName, sizeof(connectionDataPtr->host.canonical));
				else strncpy(connectionDataPtr->host.canonical, (char*)(**(**setupParametersHandle).session).hostname, sizeof(connectionDataPtr->host.canonical));
				connectionDataPtr->host.canonical[sizeof(connectionDataPtr->host.canonical) - 1] = '\0';
				
				Memory_DisposeHandle(&((Handle)(**setupParametersHandle).session));
				Memory_DisposeHandle(&((Handle)(**setupParametersHandle).terminal));
				Memory_DisposeHandle(&((Handle)setupParametersHandle));
				setupParametersHandle = null;
				
				if (connectionDataPtr->port < 0)
				{
					Console_WriteLine("socket open failed!");
					killConnection(connectionDataPtr); // Network_SocketOpen() failed
				}
				
				connectionDataPtr->status = kConnectionStatusOpening;
				SetMenuMarkFromScreen(kMenuMarkTypeConnectionOpening, inScreenID); // sets status mark
			}
			else
			{
				Console_WriteLine("could not complete connection opening, block size unsettable");
				isError = true;
			}
		}
		else // report the real DNR error
		{
			Str255		errorString,
						osErrorCodeString,
						internalErrorCodeString,
						hostName;
			short		messageStringIndex = 0;
			
			
			HLockHi((Handle)(**setupParametersHandle).session);
			BlockMove((**(**setupParametersHandle).session).hostname, hostName, StrLength((**(**setupParametersHandle).session).hostname) + 1);
			
			// Kill the connection and dispose of all associated data
			// structures, which is best done here in case memory is
			// extra tight at the moment.  Then, hopefully, that extra
			// memory will make it all the more likely that an error
			// can be reported to the user.
			Memory_DisposeHandle(&((Handle)(**setupParametersHandle).session));
			Memory_DisposeHandle(&((Handle)(**setupParametersHandle).terminal));
			Memory_DisposeHandle(&((Handle)setupParametersHandle));
			killConnection(connectionDataPtr);
			
			// If one of the following errors is specified,
			// a descriptive message is displayed; otherwise,
			// by default, a general DNR error is reported.
			switch (inDNRError)
			{
				// NOTE: MacTelnet now uses Open Transport for domain name lookups, so no MacTCP error codes are checked for.
				case kOTBadNameErr:
				case kOTNoDataErr: // no data to read - no names were found
					messageStringIndex = siDNRNameDoesNotExist;
					break;
				
				case kOTCanceledErr: // interrupted
				case kESHUTDOWNErr:
					messageStringIndex = siDNRLookupCancelled;
					break;
				
				case kOTOutOfMemoryErr: // out of memory (extremely unlikely in DNR use)
					messageStringIndex = siDNROutOfMemory;
					break;
				
				case kENETDOWNErr: // network unreachable
				case kENETUNREACHErr:
					messageStringIndex = siDNRNetworkUnreachable;
					break;
				
				case kETIMEDOUTErr: // timed out
				case kETIMEErr:
					messageStringIndex = siDNRNetworkTimeout;
					break;
				
				case kOTBadConfigurationErr: // bad network setup
					messageStringIndex = siDNRNetworkConfigurationError;
					break;
				
				default:
					messageStringIndex = siDNRGeneralError;
					break;
			}
			
			GetIndString(errorString, rStringsDNRMessages, messageStringIndex);
			NumToString((long)0L, internalErrorCodeString);
			NumToString((long)inDNRError, osErrorCodeString);
			
			{
				ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
				InterfaceLibAlertRef	box = null;
				Str255					dialogText,
										helpText;
				
				
				// build the error message
				GetIndString(dialogText, rStringsGeneralMessages, siConnectionOpeningFailed);
				StringUtilities_PMetaSetDefaults();
				StringUtilities_PBuild(dialogText, hostName, errorString,
										EMPTY_PSTRING, EMPTY_PSTRING);
				
				// build the help text
				GetIndString(helpText, rStringsGeneralMessages, siOSErrAndInternalErrHelpText);
				StringUtilities_PMetaSetDefaults();
				StringUtilities_PBuild(helpText, osErrorCodeString, internalErrorCodeString,
										EMPTY_PSTRING, EMPTY_PSTRING);
				
				// ...and display it
				box = Alert_New();
				Alert_SetFilterProc(box, upp);
				Alert_SetHelpButton(box, false);
				Alert_SetParamsFor(box, kAlert_StyleOK);
				Alert_SetText(box, dialogText, helpText);
				Alert_SetType(box, kAlertStopAlert);
				Alert_Display(box);
				Alert_Dispose(&box);
				DisposeModalFilterUPP(upp), upp = null;
			}
		}
	}
	else
	{
		Console_WriteLine("could not complete connection opening, not in DNR wait state");
		isError = true;
	}
	
	if (isError) BEEP;
}// CompleteOpening


/*!
Creates a new telnet connection using a
parameters structure.  Returns true if
no errors occur, or false if there was a
problem creating the connection.

(3.0)
*/
Boolean
Connections_CreateOne	(ConnectionSetup**	inConnectionInitParams)
{
	register SInt16		i = 0;
	SessionPrefs*		sessionPreferencesPtr = null;
	TerminalPrefs*		terminalPreferencesPtr = null;
	ConnectionDataPtr	connectionDataPtr = null;
	TelnetScreenID		currentScreenID = -1;
	Boolean				noErrorOccurred = true;
	
	
	Cursors_UseWatch();
	
	// check if we have the maximum number of sessions open
	if (Connections_GetCount() == kMaximumAllowedTelnetConnections)
	{
		tooManyOpenConnections();
		noErrorOccurred = false;
	}
	
	// verify the connection data
	if (inConnectionInitParams == null)
	{
		Console_WriteLine("ERROR!  Attempt to create connection with no parameters!");
		noErrorOccurred = false;
	}
	
	if (noErrorOccurred)
	{
		HLockHi((Handle)inConnectionInitParams);
		HLockHi((Handle)(**inConnectionInitParams).terminal);
		HLockHi((Handle)(**inConnectionInitParams).session);
		sessionPreferencesPtr = *((**inConnectionInitParams).session);
		terminalPreferencesPtr = *((**inConnectionInitParams).terminal);
		
		if ((sessionPreferencesPtr != null) && (terminalPreferencesPtr != null))
		{
			RealScreenFlags		newScreenFlags = kRealScreenFlagWindowHeader;	// attributes of the new window
			
			
			if (terminalPreferencesPtr->usesVTWrap) newScreenFlags |= kRealScreenFlagTextWraps;
			currentScreenID = Terminal_NewWindow(terminalPreferencesPtr->normalFont,
												terminalPreferencesPtr->fontSize,
												newScreenFlags,
												terminalPreferencesPtr->scrollbackBufferSize,
												terminalPreferencesPtr->rowCount,
												terminalPreferencesPtr->columnCount,
												sessionPreferencesPtr->forcesave);
			if (currentScreenID < 0)
			{
				// invalid ID - screen creation failed
				Console_WriteLine("failed to create a terminal window!");
				Errors_OutOfMemory(kInternalTelnetMemoryErrorRealScreenCreation);
				Memory_DisposeHandle(&((Handle)(**inConnectionInitParams).terminal));
				Memory_DisposeHandle(&((Handle)(**inConnectionInitParams).session));
				Memory_DisposeHandle(&((Handle)inConnectionInitParams));
				noErrorOccurred = false;
			}
		}
		else noErrorOccurred = false;
		
		if (noErrorOccurred)
		{
			connectionDataPtr = ScreenManager_GetConnectionDataOf(currentScreenID);
			noErrorOccurred = (connectionDataPtr != null);
		}
		else noErrorOccurred = false;
		
		if (noErrorOccurred)
		{
			// mark as dead, so Connections_CompleteOpening() can tell if failure occurred
			// (i.e. normally, state is waiting on DNR for Connections_CompleteOpening())
			Connections_SetCount(Connections_GetCount() + 1);
			connectionDataPtr->status = kConnectionStatusNotInUse;
			
			if (StrLength((**inConnectionInitParams).windowName) == 0)
			{
				BlockMove((**(**inConnectionInitParams).session).hostname, (**inConnectionInitParams).windowName, 
							StrLength((**(**inConnectionInitParams).session).hostname) + 1);
			}
			
			// now Network_Connect() is called per-connection, instead of at startup (3.0)
			noErrorOccurred = Network_Connect(true/* ask user if not connected */);
		}
		
		if (noErrorOccurred)
		{
			OSStatus		error = noErr;
			
			
			if ((**inConnectionInitParams).proxy != null)
			{
				// 3.0 - look up the connection via a proxy server (since apparently the user doesnt think the host can be found otherwise)
				connectionDataPtr->proxy.ref = (**inConnectionInitParams).proxy;
				connectionDataPtr->proxy.enabled = true;
				error = DNR_LookUpNewConnectionViaProxy(sessionPreferencesPtr->hostname, connectionDataPtr->proxy.ref, currentScreenID);
			}
			else
			{
				// no proxy server - asynchronously look up the IP address of the host, directly
				error = DNR_LookUpNewConnection(sessionPreferencesPtr->hostname, currentScreenID);
			}
			
			// handle errors in lookup
			if (error != noErr)
			{
				Errors_OutOfMemory(kInternalTelnetMemoryErrorDNRLookup);
				Memory_DisposeHandle(&((Handle)(**inConnectionInitParams).terminal));
				Memory_DisposeHandle(&((Handle)(**inConnectionInitParams).session));
				Memory_DisposeHandle(&((Handle)inConnectionInitParams));
				Connections_SetCount(Connections_GetCount() - 1);
				Cursors_UseArrow();
				Console_WriteValuePString("could not look up host", sessionPreferencesPtr->hostname);
				if ((**inConnectionInitParams).proxy != null)
				{
					Console_WriteLine("while using proxy server");
				}
				Console_WriteValue("error", error);
				noErrorOccurred = false;
			}
		}
		
		if (noErrorOccurred)
		{
			// asynchronously look up the IP address of the host
			
			if (noErrorOccurred)
			{
				// set up the menu bar
				AddConnectionsMenuItem(connectionDataPtr, (**((**inConnectionInitParams).session)).hostname,
										(**inConnectionInitParams).windowName);
				Menus_Service();
				
				// initialize screen settings
				connectionDataPtr->authenticate = 0;
				connectionDataPtr->encrypt = 0;
				connectionDataPtr->aedata = null;
				for (i = 0; i < sizeof(connectionDataPtr->myopts); i++)
				{
					connectionDataPtr->myopts[i] = 0;
					connectionDataPtr->hisopts[i] = 0;
				}
				connectionDataPtr->host.canonical[0] = '\0';
				connectionDataPtr->terminal.emulation = terminalPreferencesPtr->emulation;
				connectionDataPtr->eightbit = terminalPreferencesPtr->usesEightBits; // is this necessary?
				connectionDataPtr->metaKey = terminalPreferencesPtr->metaKey;
				connectionDataPtr->Xterm = terminalPreferencesPtr->usesXTermSequences;
				connectionDataPtr->keypadmap = terminalPreferencesPtr->remapKeypad;
				connectionDataPtr->port = -1; // Network_SocketOpen() will take care of this
				connectionDataPtr->lineMode.active = FlagManager_Test(kFlagAppleScriptRecording);
				connectionDataPtr->forcesave = sessionPreferencesPtr->forcesave;
				
				// 3.0 - line mode must be enforced when recording scripts, to make it easy to grab the data the user is sending
				connectionDataPtr->lineMode.allowed = ((sessionPreferencesPtr->linemode) || FlagManager_Test(kFlagAppleScriptRecording));
				if (connectionDataPtr->lineMode.allowed) LineMode_Init(connectionDataPtr);
				
				// set up the open timeout - incomplete
				{
					Size		actualSize = 0L;
					
					
					unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TimeoutOpen,
											sizeof(connectionDataPtr->openTimeout), &connectionDataPtr->openTimeout,
											&actualSize) == kTelnetPrefsResultCodeSuccess)
					{
						connectionDataPtr->openTimeout = 10; // assume a 10-second timeout, if preference cant be found
					}
				}
				
				if (noErrorOccurred)
				{
					Str255		string;
					
					
					RealScreen_SetWTitle(connectionDataPtr->vs, (**inConnectionInitParams).windowName);
					RealScreen_SetANSIColorsEnabled(connectionDataPtr->vs, terminalPreferencesPtr->usesANSIColors);
					connectionDataPtr->window = ScreenManager_GetWindowOf(connectionDataPtr->vs);
					SetWindowKind(connectionDataPtr->window, WIN_CNXN);
					
					// 3.0 - record the time when the connection began
					GetDateTime(&connectionDataPtr->connectedSince);
					
					connectionDataPtr->arrowmap = terminalPreferencesPtr->usesEMACSArrows;
					connectionDataPtr->bsdel = sessionPreferencesPtr->bksp;
					connectionDataPtr->crmap = sessionPreferencesPtr->crmap;
					connectionDataPtr->TEK.pageLocation = sessionPreferencesPtr->tekclear;
					connectionDataPtr->TEK.mode = sessionPreferencesPtr->tektype;
					connectionDataPtr->TEK.graphicsID = -1;
					connectionDataPtr->ESscroll = terminalPreferencesPtr->savesOnClear;
					connectionDataPtr->wrap = terminalPreferencesPtr->usesVTWrap;
					connectionDataPtr->pgupdwn = terminalPreferencesPtr->mapsPageJumpKeys;
					connectionDataPtr->controlKey.suspend = sessionPreferencesPtr->skey;
					connectionDataPtr->controlKey.resume = sessionPreferencesPtr->qkey;
					connectionDataPtr->controlKey.interrupt = sessionPreferencesPtr->ckey;
					BlockMove((Ptr)sessionPreferencesPtr->hostname, connectionDataPtr->host.name, StrLength(sessionPreferencesPtr->hostname) + 1);
					BlockMove(terminalPreferencesPtr->answerBackMessage, connectionDataPtr->terminal.answerBack, 32);
					connectionDataPtr->naws = 0;
					connectionDataPtr->xfer = 0;
					connectionDataPtr->telstate = 0;
					connectionDataPtr->timing = 0;
					connectionDataPtr->ftpport = -1; // no additional port opened yet
					connectionDataPtr->clientflags = 0;
					connectionDataPtr->kblen = 0; // nothing in the buffer
					connectionDataPtr->enabled = 1;
					connectionDataPtr->Ittype = 0;
					connectionDataPtr->Isga = 0; // I suppress go-ahead
					connectionDataPtr->Usga = 0; // you suppress go-ahead
					
					connectionDataPtr->ftpState = (**inConnectionInitParams).ftpState;
					if ((**inConnectionInitParams).ftpState != 0)
					{
						connectionDataPtr->termstate = DUMBTYPE;
						connectionDataPtr->echo = 1;
						connectionDataPtr->halfdup = 1;
					}
					else
					{
						connectionDataPtr->termstate = VTEKTYPE;
						connectionDataPtr->echo = 1;
						connectionDataPtr->halfdup = sessionPreferencesPtr->halfdup;
					}
					
					connectionDataPtr->remote_flow = 0;
					connectionDataPtr->allow_flow = 1; // allow flow control
					connectionDataPtr->restart_any_flow = 0; // initially, only an XON turns flow control back on
					
					// Now see if the desired translation is available; if not, use default translation.
					connectionDataPtr->national = 0; // default to no translation
					for (i = 1; i <= nNational + 1; i++)
					{
						GetMenuItemText(Menus_GetTranslationMenu(), i, string);
						if (EqualString(sessionPreferencesPtr->TranslationTable, string,
										true/* case sensitive */, false/* diacritics sensitive */))
						{
							connectionDataPtr->national = i - 1;
						}
					}
					
					// set up paste-related variables
					connectionDataPtr->paste.inCount = 0;
					connectionDataPtr->paste.outCount = 0;
					connectionDataPtr->paste.nextCharPtr = null;
					connectionDataPtr->paste.text = null;
					connectionDataPtr->paste.outLength = 0;
					connectionDataPtr->paste.method = sessionPreferencesPtr->pastemethod;
					connectionDataPtr->paste.blockSize = sessionPreferencesPtr->pasteblocksize;
					
					(Boolean)RealScreen_SetColor(connectionDataPtr->vs, kRealScreenColorIndexNormalText, &terminalPreferencesPtr->foregroundNormalColor);
					(Boolean)RealScreen_SetColor(connectionDataPtr->vs, kRealScreenColorIndexNormalBackground, &terminalPreferencesPtr->backgroundNormalColor);
					(Boolean)RealScreen_SetColor(connectionDataPtr->vs, kRealScreenColorIndexBlinkingText, &terminalPreferencesPtr->foregroundBlinkingColor);
					(Boolean)RealScreen_SetColor(connectionDataPtr->vs, kRealScreenColorIndexBlinkingBackground, &terminalPreferencesPtr->backgroundBlinkingColor);
					(Boolean)RealScreen_SetColor(connectionDataPtr->vs, kRealScreenColorIndexBoldText, &terminalPreferencesPtr->foregroundBoldColor);
					(Boolean)RealScreen_SetColor(connectionDataPtr->vs, kRealScreenColorIndexBoldBackground, &terminalPreferencesPtr->backgroundBoldColor);
					
					// 3.0 - handle automatic file captures
					if (sessionPreferencesPtr->autoCaptureToFile)
					{
						FSSpec		file;
						Boolean		successful = false;
						
						
						if (Preferences_IsAliasStored(sessionPreferencesPtr->captureFileAliasID))
						{
							TelnetPrefsAliasID		aliasID = kInvalidTelnetPrefsAliasID;
							
							
							// read and resolve the existing capture file preference (re-saving changes if necessary)
							aliasID = Preferences_NewSavedAlias(sessionPreferencesPtr->captureFileAliasID);
							successful = Preferences_ParseAlias(aliasID, &file);
						}
						else successful = false;
						
						if (!successful)
						{
							// no user preference for which file to use; create a default file on the root of the system disk
							Str255		defaultName;
							OSErr		error = noErr;
							
							
							// get the default capture file name
							GetIndString(defaultName, rStringsImportantFileNames, siImportantFileNameCaptureTextDefault);
							
							// initialize this structure, in case FSMakeFSSpec() fails
							PLstrcpy(file.name, defaultName);
							file.vRefNum = -1;
							file.parID = fsRtDirID;
							error = FSMakeFSSpec(file.vRefNum, file.parID, file.name, &file);
							successful = ((error == noErr) || (error == fnfErr));
						}
						
						// if it is possible to do a capture (i.e. the file could be created or located, and written to), then start one
						if (successful) successful = Terminal_FileCaptureBegin(connectionDataPtr->vs, &file);
						
						unless (successful) Console_WriteLine("unable to start automatic file capture");
					}
					
					connectionDataPtr->status = kConnectionStatusDNRWait;
					connectionDataPtr->setupParametersHandle = (Handle)inConnectionInitParams;
					HUnlock((Handle)(**inConnectionInitParams).terminal);
					HUnlock((Handle)(**inConnectionInitParams).session);
					
					Terminal_SavesLinesOnClear(connectionDataPtr->vs, connectionDataPtr->ESscroll);
					noErrorOccurred = true; // we made it!
				}
			}
		}
	}
	Cursors_UseArrow();
	
	return noErrorOccurred;
}// CreateOne


/*!
To open a new connection using a parameter block of
connection parameters, use this method.  The possible
parameters are those accessible through scripting.

Incomplete - eventually, this function will have a
meaningful result.

(3.0)
*/
void
Connections_CreateOneParameterized	(TelnetConnectionParameterBlockPtr		inParameters)
{
	if (inParameters != null)
	{
		// send a New Connection Apple Event
		AppleEvent			connectionOpenEvent,
							reply;
		StringPtr			string = null;
		
		
		Alert_ReportOSErr(InstallAE_CreateRecordableAppleEvent(kTelnetEventClass, kTelnetEventIDConnectionOpen, &connectionOpenEvent));
		
		// window parameters
		if (inParameters->window.title != null)
		{
			// connection windows initial window title
			string = inParameters->window.title;
			(OSErr)AEPutParamPtr(&connectionOpenEvent, kTelnetParameterKeyConnectionOpenWindowTitle,
									cString, string + 1, StrLength(string) * sizeof(unsigned char));
		}
		
		// connection parameters
		{
			// connection method
			FourCharCode	method = inParameters->connection.method;
			
			
			(OSErr)AEPutParamPtr(&connectionOpenEvent, kTelnetParameterKeyConnectionOpenProtocol,
									kTelnetEnumeratedClassConnectionMethod,
									&method, sizeof(FourCharCode));
		}
		{
			// host name, IP address or Session Favorites name
			string = inParameters->connection.host;
			(OSErr)AEPutParamPtr(&connectionOpenEvent, kTelnetParameterKeyConnectionOpenHost,
									cString, string + 1, StrLength(string) * sizeof(unsigned char));
		}
		{
			// port number
			SInt16		port = inParameters->connection.port;
			
			
			(OSErr)AEPutParamPtr(&connectionOpenEvent, kTelnetParameterKeyConnectionOpenPortNumber,
									typeSInt16, &port, sizeof(SInt16));
		}
		{
			// opening timeout
			UInt32		timeout = inParameters->connection.timeout;
			
			
			(OSErr)AEPutParamPtr(&connectionOpenEvent, kTelnetParameterKeyConnectionOpenTimeoutInSeconds,
									typeUInt32, &timeout, sizeof(UInt32));
		}
		
		// open the connection by sending an Apple Event; if it fails, report the error
		Alert_ReportOSErr(AESend(&connectionOpenEvent, &reply,
									kAENoReply | kAENeverInteract,
									kAENormalPriority, kAEDefaultTimeout, null, null));
		AEDisposeDesc(&connectionOpenEvent);
		
		// reply - unimplemented
	}
}// CreateOneParameterized


/*!
Disconnects an active connection, freeing allocated memory
and (if the user has such a preference set) destroying
Kerberos tickets.  The window remains open unless the user
has that preference disabled, in which case this routine
invokes killConnection(), as well.

If a connection fails or is aborted early, it is usually
destroyed with killConnection().  If the connections
window is already on the screen and the connection has not
been previously disconnected, it is usually terminated with
Connections_Disconnect() and then ultimately killed with
killConnection() later.  If you add clean-up code to this
routine, make sure you consider when your code for
initialization is done - it may be more correct to do your
clean-up in killConnection().  When in doubt, do clean-up
in killConnection().

(2.6)
*/
void
Connections_Disconnect		(ConnectionDataPtr	inConnectionDataPtr)
{
	Console_WriteLine("disconnecting session");
	if (inConnectionDataPtr != null)
	{
		Boolean		leaveWindowOpen = false;
		
		
		if (inConnectionDataPtr->status != kConnectionStatusDead) Network_CloseConnection(inConnectionDataPtr->port);
		if (inConnectionDataPtr->status == kConnectionStatusActive)
		{
			// 3.0 - if an active connection is being disconnected, modify the count
			/*if (inConnectionDataPtr->status == kConnectionStatusActive) */gConnectionActiveCount--;
			
			Console_WriteLine("capture close");
			if (Terminal_FileCaptureInProgress(inConnectionDataPtr->vs)) Terminal_FileCaptureEnd(inConnectionDataPtr->vs);
			
			Console_WriteLine("printing file close");
			if (Terminal_PrintingInProgress(inConnectionDataPtr->vs)) Terminal_PrintingEnd(inConnectionDataPtr->vs);
			
			// destroy the window automatically, unless the users preference is to leave it open
			Console_WriteLine("preference lookup");
			{
				Size		actualSize = 0L;
				
				
				unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_DontAutoClose,
												sizeof(leaveWindowOpen), &leaveWindowOpen, &actualSize) == kTelnetPrefsResultCodeSuccess)
				{
					leaveWindowOpen = false; // assume windows automatically close, if preference cant be found
				}
			}
		}
		else leaveWindowOpen = false; // inactive connections are always killed
		
		if (leaveWindowOpen)
		{
			Str255		string;
			
			
			Console_WriteLine("not being killed");
			
			// add a visual indicator to the window title of disconnected windows, in case
			// the user has the window header collapsed and cant see the status icon
			GetWTitle(inConnectionDataPtr->window, string);
			StringUtilities_PInsert(string, "\p~ ");
			PLstrcat(string, "\p ~");
			SetWTitle(inConnectionDataPtr->window, string);
			
			// mark the session as dead, and update the Window menu icon
			SetMenuMarkFromScreen(kMenuMarkTypeConnectionDead, inConnectionDataPtr->vs);
			
			inConnectionDataPtr->port = kInvalidNetworkPortRef;
			inConnectionDataPtr->status = kConnectionStatusDead;
		}
		else killConnection(inConnectionDataPtr);
	}
	else
	{
		BEEP;
		Console_WriteLine("WARNING: Attempt to disconnect a connection with nonexistent data!");
	}
}// Disconnect


/*!
To determine the number of connections that are
active (that is, their windows are showing AND
they are connected), use this method.

(3.0)
*/
UInt16
Connections_GetActiveCount	(void)
{
	return gConnectionActiveCount;
}// GetActiveCount


/*!
To determine the number of connections that are in
use, in any state, use this method.

(3.0)
*/
UInt16
Connections_GetCount		(void)
{
	return gConnectionCount;
}// GetCount


/*!
Sets up a telnet connection consisting of the
default terminal and session preferences.

(2.6)
*/
ConnectionSetup**
Connections_GetDefault	(void)
{
	ConnectionSetup**	result = null;
	Size				actualSize = 0L;
	
	
	result = (ConnectionSetup**)Memory_NewHandle(sizeof(ConnectionSetup));

	(**result).session = Preferences_GetDefaultSession();
	(**(**result).session).ip_address = 0;
	(**result).terminal = Preferences_GetDefaultTerminal();
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TimeoutOpen,
								sizeof((**result).openTimeout), &(**result).openTimeout,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		(**result).openTimeout = 10; // assume a 10-second timeout, if preference cant be found
	}
	(**result).proxy = null;
	
	return result;
}// GetDefault


/*!
Determines how many connections are either active
or opening.

(2.6)
*/
UInt16
Connections_GetLiveCount	(void)
{
	UInt16			result = 0;
	
	
	result = (ScreenManager_GetStatusCount(kConnectionStatusActive) + ScreenManager_GetStatusCount(kConnectionStatusOpening));
	return result;
}// GetLiveCount


/*!
To mark a connection as active, use this method.
This increments the active connection count (as
returned by Connections_GetActiveCount()), sets
the connection status to "kConnectionStatusActive",
and displays the connections terminal window.

(3.0)
*/
void
Connections_MakeActive		(ConnectionDataPtr		inConnectionDataPtr)
{
	gConnectionActiveCount++;
	inConnectionDataPtr->status = kConnectionStatusActive;
	ShowWindow(inConnectionDataPtr->window);
	SelectBehindDialogWindows(inConnectionDataPtr->window); // 3.0 - asynchronous net events; dont burst in front of a dialog!
	SetMenuMarkFromScreen(kMenuMarkTypeConnectionLive, inConnectionDataPtr->vs);
	Menus_Service();
}// MakeActive


/*!
Opens a connection by taking the name of the
specified item in the given menu and deriving
its connection parameters from the name.

(2.6)
*/
void
Connections_OpenFromMenuItem	(MenuHandle				inMenuHandle,
								 const MenuItemIndex	inItemIndex)
{
	ConnectionSetup**	setupParametersHandle = null;
	Str255				string;
	
	
	GetMenuItemText(inMenuHandle, inItemIndex, string);
	
	setupParametersHandle = nameToConnectionSetup(string, true);
	if (setupParametersHandle == null) Errors_OutOfMemory(kInternalTelnetMemoryErrorConnectionOpenFromMenuItem);
	else (Boolean)Connections_CreateOne(setupParametersHandle);
}// OpenFromMenuItem


/*!
Opens a connection by taking its Uniform
Resource Locator information and deriving
the connection parameters.  This is
expected to be called in response to a
'GURL' Apple Event.

(2.6)
*/
void
Connections_OpenFromURL	(char*		host,
						 char*		portstring,
						 char*		user,
						 char*		password)
{
	ConnectionSetup**	setupParametersHandle = null;
	Str255				windowName,
						tempString;
	short				len = 0;
	long				port = 0L;
	
	
	// When Apple Events are received, the do you want to
	// connect to the network warning should not show up,
	// so just turn it off.
	(Boolean)Network_Connect(false);
	
	setupParametersHandle = Connections_GetDefault();
	
	windowName[0] = 0;
	
	// set up window name if user (and password) given
	if (user != null)
	{
		GetIndString(windowName, rStringsMiscellaneous, siUserPrompt);
		len = strlen(user);
		BlockMoveData(user, &windowName[StrLength(windowName) + 1], len);
		windowName[0] += len;
		if (password != null)
		{
			GetIndString(tempString, rStringsMiscellaneous, siPasswordPrompt);
			BlockMoveData(&tempString[1], &windowName[StrLength(windowName) + 1], tempString[0]);
			windowName[0] += tempString[0];
			len = strlen(password);
			BlockMoveData(password, &windowName[StrLength(windowName) + 1], len);
			windowName[0] += len;
		}

		if (windowName[0] != 0)
		{
			BlockMoveData(windowName, (**setupParametersHandle).windowName, StrLength(windowName) + 1);
		}
	}		
	
	StringUtilities_CToP(host, (**(**setupParametersHandle).session).hostname);
	
	if (portstring != null)
	{
		StringUtilities_CToPInPlace(portstring);
		StringToNum((StringPtr)portstring, &port);
		(**(**setupParametersHandle).session).port = port;
	}
		
	(Boolean)Connections_CreateOne(setupParametersHandle);
}// OpenFromURL


/*!
This method parses a URL in string form, and opens a
connection with it if it is valid.

(3.0)
*/
OSErr
Connections_OpenFromURLString	(const char*	inURLStringData,
								 Size			inSize)
{
	OSErr		result = noErr;
	char*		stringData = null;
	char*		beg = null;
	char*		end = null;
	char*		user = null;
	char*		password = null;
	char*		portstring = null;
	char*		host = null;
	
	
	stringData = (char*)Memory_NewPtr((inSize + 1) * sizeof(char));
	BlockMove(inURLStringData, stringData, inSize);
	
	stringData[inSize] = 0; // terminate the C string
	beg = &stringData[0];
	end = &stringData[inSize - 1];
	
	// strip leading spaces
	while ((beg < end) && (*beg == ' ')) beg++;
	
	// strip trailing spaces
	while ((end < beg) && (*end == ' ')) end--;
	
	// look for (and strip) beginning and ending angle brackets
	if (*beg == '<')
	{
		if (*end != '>')
		{
			// leading angle bracket, but no trailing angle bracket
			result = paramErr;
			goto badExit;
		}
		
		// nuke the brackets
		beg++;
		end--;
	}
	
	// Terminate the string we currently have (for strncmp fn's)
	*(end+1) = 0;
	
	// strip any leading "URL:"
	if (!strncmp(beg, "URL:", 4))
	{
		beg += 4;
	}
	
	// strip any leading "telnet://"
	if (strncmp(beg, "telnet://", 9))   // No leading "telnet://"
	{
		if (strncmp(beg, "rlogin://", 9))
		{	
			result = paramErr;
			goto badExit;
		}
	}
	beg += 9;
	
	// drop any ending slash
	if (*end == '/')
	{
		*end = 0;
		end--;
	}
	
	// process any escaped characters
	processURLEscapeCodes(beg, &end);
	
	// set up for URL parsing
	password = null;
	portstring = null;
	host = null;
	
	// assume <user> exists
	user = beg;
	
	// leading : or @ is not allowed
	if ((*beg == ':') || (*beg == '@'))
	{
		result = paramErr;
		goto badExit;
	}
	
	// scan for : or @ or end of string
	while ((beg <= end) && (*beg != ':') && (*beg != '@')) beg++;
	
	// if the end is reached, then only a host was given
	if (beg > end)
	{
		host = user;
		user = null;
		goto goodUrl;
	}

	// if : was found, scan in the password (or port)
	if (*beg == ':')
	{
		*beg++ = 0;
		password = beg;
		
		// scan for : or @ or end of string
		while ((beg <= end) && (*beg != ':') && (*beg != '@')) beg++;
		
		if (*beg == ':')
		{
			// xxxx:yyyy:<anything> is not allowed
			result = paramErr;
			goto badExit;
		}
		
		if (*beg != '@')
		{
			// end of string - must be host:port
			host = user;
			user = null;
			portstring = password;
			password = null;
			goto goodUrl;
		}
		
		// have xxx:yyy@<something>
		*beg++ = 0;	// terminate password string
	}		
	else
	{
		// found @
		*beg++ = 0;	// terminate user string
	}
	
	// now have either xxx:yyy@<something> or xxxx@<something>
	host = beg;
	
	// scan for : or @ or end of string, fixing bad telnets
	while ((beg <= end) && (*beg != ':') && (*beg != '@') && (*beg != '/')) beg++;
	
	if (*beg == '@')
	{
		// xxx[:yyyy]@xxx@ is not allowed
		result = paramErr;
		goto badExit;
	}
	
	if (*beg != ':')
	{
		// end of string, we have xxxx[:yyyy]@zzzz
		*beg = 0;
	}
	else
	{
		*beg++ = 0;	// terminate host string
		portstring = beg;
	}
	
goodUrl:
	Connections_OpenFromURL(host, portstring, user, password);
	result = noErr;
	
badExit:
	Memory_DisposePtr(&stringData);
	
	return result;
}// OpenFromURLString


/*!
Determines the port number within a string
that may contain a host name and optionally a
port number afterwards ( la New Sessions
dialog box).  The port number can be tagged
on either with white space or a colon.  If
the parse succeeds (that is, if the output
port number is valid), "true" is returned;
otherwise, "false" is returned.  On output,
the string given is modified to include only
the host name (minus any port number).

(2.6)
*/
Boolean
Connections_ProcessHostNameString	(StringPtr		inoutString,
									 SInt16*		outPort)
{
	Str255				string;
	register SInt16		i = 0;
	StringPtr			xxxxptr = null,
						yyyyptr = null;
	Boolean				result = false;
	
	
	// copy the string
	BlockMove(inoutString, string, 255 * sizeof(unsigned char));
	
	// remove leading spaces
	for (i = 1; ((i <= StrLength(string)) && (string[i] == ' ')); i++) {}
	
	if (i > StrLength(string))
	{
		inoutString[0] = 0;
		result = false;
	}
	else
	{
		SInt32		portRequested = 0L;
		
		
		xxxxptr = &string[i - 1];
		
		//	Now look for a port number...
		while ((i <= StrLength(string)) && (string[i] != ' ') && (string[i] != ':')) i++;
		
		yyyyptr = &string[i];
		
		if (i < StrLength(string))
		{
			string[i] = StrLength(string) - i;
			StringToNum(&string[i], &portRequested);
			if ((portRequested > 0) && (portRequested < 65535))	result = true;
		}
		
		xxxxptr[0] = yyyyptr - xxxxptr - 1;
		
		// copy parsed host name string back
		BlockMove(xxxxptr, inoutString, (StrLength(xxxxptr) + 1) * sizeof(unsigned char));
		
		*outPort = (SInt16)portRequested;
	}
	
	return result;
}// ProcessHostNameString


/*!
To specify the number of connections that are active
or opening, use this method.  Generally, you should
always specify a value in terms of one less or greater
than the value of Connections_GetCount(), effectively
incrementing or decrementing the count.

(3.0)
*/
void
Connections_SetCount	(const UInt16		inNewCount)
{
	gConnectionCount = inNewCount;
}// SetCount


//
// internal methods
//

/*!
Destroys the window for a connection, cleans up
flags and memory allocations made for it, and
removes its name from the Window menu.

Typically, in response to a standard connection
close, you use Connections_Disconnect(), which
will take into account the users preference as
to whether windows remain open, etc. and clean
up auxiliary tasks initiated by that session,
such as printing or capturing to a file.  That
routine invokes this one if it turns out that a
session window should be destroyed.

(2.6)
*/
static void
killConnection		(ConnectionDataPtr		inConnectionDataPtr)
{
	Console_BeginFunction();
	Console_WriteLine("connection kill");
	if (inConnectionDataPtr != null)
	{
		Handle		h = null;
		short		i = 0;
		Size 		junk = 0;
		
		
		Cursors_UseWatch();
		
		if (inConnectionDataPtr->status == kConnectionStatusDead)
		{
			if (inConnectionDataPtr->TEK.graphicsID > -1) detachGraphics(inConnectionDataPtr->TEK.graphicsID); // detach the TEK screen
			if (inConnectionDataPtr->paste.outLength > 0)
			{
				inConnectionDataPtr->paste.outLength = 0; // kill the remaining send
				HUnlock(inConnectionDataPtr->paste.text); // buffer
				HPurge(inConnectionDataPtr->paste.text);
			}
		}
		
		Console_WriteLine("preparing to dispose authentication data");
		if (inConnectionDataPtr->aedata != null) Memory_DisposePtr(&((Ptr)inConnectionDataPtr->aedata));
		
		Console_WriteLine("preparing to dispose proxy server");
		if (inConnectionDataPtr->proxy.ref != null) ProxyServer_Dispose(&inConnectionDataPtr->proxy.ref);
		
		// remove the connections name from the Window menu
		Console_WriteLine("servicing menus");
		Menus_RemoveConnectionsMenuItem(inConnectionDataPtr);
		Menus_Service();
		
		inConnectionDataPtr->status = kConnectionStatusNotInUse;
		Console_WriteLine("preparing to kill window");
		RealScreen_Dispose(&inConnectionDataPtr->vs);
		Console_WriteLine("preparing to reset keyboard");
		SetDefaultKCHR();
		
		// 3.0 - destroy the connection initialization parameters!
		//Console_WriteLine("destroying initialization parameters");
		//Memory_DisposeHandle(&inConnectionDataPtr->setupParametersHandle);
		
		// remove the connection data pointer from the list of connections
		Console_WriteLine("destroying screen");
		ScreenManager_Remove(inConnectionDataPtr);
		
		Console_WriteLine("done");
		//MaxMem(&junk);
		Cursors_UseArrow();
	}
	else
	{
		BEEP;
		Console_WriteLine("WARNING: Attempt to kill a connection with nonexistent data!");
	}
	Console_EndFunction();
}// killConnection


/*!
This method recognizes the following input strings: "xxxx yyyy"
and "xxxx:yyyy".

If "xxxx" matches a session name, that session record is used.
Otherwise, the default session record is used with "xxxx" as
the DNS hostname.   The "yyyy", if provided, is converted to a
number.  If it is a valid number, it is used as the port to
connect to.

The second argument specifies whether to get the terminal stuff
from the session preferences.  If not, the terminal is set to
null and should be set up properly later.

The proxy is set to null.

A new connection handle is returned.  Dispose of it using the
Memory_DisposeHandle() routine.

WARNING:	Do not pass this routine a blank string.

(2.6)
*/
static ConnectionSetup**
nameToConnectionSetup	(StringPtr		inString,
						 Boolean		inUseDefaultTerminal)
{
	ConnectionSetup**	result = null;
	
	
	result = (ConnectionSetup**)Memory_NewHandleCritical(sizeof(ConnectionSetup));
	if (result != null)
	{
		SessionPrefs**		sessionHandle = null;
		SInt16				portRequested = 0;
		Boolean				foundPort = false,
							wasAlias = false;
		
		
		foundPort = Connections_ProcessHostNameString(inString, &portRequested);
		
		ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
		sessionHandle = (SessionPrefs**)Get1NamedResource(SESSIONPREFS_RESTYPE, inString);
		if (sessionHandle == null) 
		{
			// connect to host xxxx w/default session
			sessionHandle = Preferences_GetDefaultSession();
			DetachResource((Handle)sessionHandle);
			HLock((Handle)sessionHandle);
			BlockMove(inString, (**sessionHandle).hostname, 64/* bytes */);
		}
		else 
		{	
			DetachResource((Handle)sessionHandle);
			HLock((Handle)sessionHandle);
			wasAlias = true;
		}
		
		(**result).session = sessionHandle;
		
		ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
		if ((inUseDefaultTerminal)||(wasAlias))
		{
			TerminalPrefs**		terminalHandle = null;
			
			
			terminalHandle = (TerminalPrefs**)Get1NamedResource(TERMINALPREFS_RESTYPE, (**sessionHandle).TerminalEmulation);
			if (terminalHandle == null) terminalHandle = Preferences_GetDefaultTerminal();
			DetachResource((Handle) terminalHandle);
			(**result).terminal = terminalHandle;
		}
		else (**result).terminal = null;
		
		// swap out space so we can make the new window CCP
		{
			SInt32		junk = 0L;
			
			
			MaxMem(&junk);
		}
		HUnlock((Handle)sessionHandle);
		
		((**result).windowName)[0] = 0;
		(**result).ftpState = ((foundPort) && (portRequested == 21)) ? 1 : 0;
		(**sessionHandle).ip_address = 0;
		
		if (foundPort) (**sessionHandle).port = portRequested;
		
		// set up proxy
		(**result).proxy = null;
	}
	
	return result;
}// nameToConnectionSetup


/*!
Processes "%xx" escape codes in a URL string (replacing them
by the characters they represent).

	Entry:	url = URL with escape codes.
			
	Exit:	url = URL with escape codes replaced by the
					characters they represent.

Copyright  1994, Northwestern University.
Modified 12/94 Jim Browne for NCSA

(2.6)
*/
static void
processURLEscapeCodes	(char*		url,
						 char**		end)
{
	char*		p = null;
	char*		q = null;
	char		c1 = '\0',
				c2 = '\0';
	
	
	p = q = url;
	while (*p != 0)
	{
		if (*p == '%')
		{
			c1 = tolower(*(p+1));
			c2 = tolower(*(p+2));
			if (isxdigit(c1) && isxdigit(c2))
			{
				c1 = isdigit(c1) ? c1 - '0' : c1 - 'a' + 10;
				c2 = isdigit(c2) ? c2 - '0' : c2 - 'a' + 10;
				*q++ = (c1 << 4) + c2;
				p += 3;
			}
			else
			{
				*q++ = *p++;
			}
		}
		else
		{
			*q++ = *p++;
		}
	}
	*q = 0;
	*end = q - 1;
}// processURLEscapeCodes


/*!
Displays an alert telling the user that there are
too many connections open right now.

(3.0)
*/
static void
tooManyOpenConnections	(void)
{
	InterfaceLibAlertRef	box = null;
	Str255					dialogText;
	Str255					helpText;
	
	
	GetIndString(dialogText, rStringsGeneralMessages, siTooManyOpenConnections);
	GetIndString(helpText, rStringsGeneralMessages, siTooManyOpenConnectionsHelpText);
	
	box = Alert_New();
	Alert_SetHelpButton(box, false);
	Alert_SetParamsFor(box, kAlert_StyleCancel);
	Alert_SetText(box, dialogText, helpText);
	Alert_SetType(box, kAlertStopAlert);
	Alert_Display(box);
	Alert_Dispose(&box);
}// tooManyOpenConnections
