/*###############################################################

	Console.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#ifndef NDEBUG

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdio>
#	include <cstring>
#else
#	include <stdio.h>
#	include <string.h>
#endif

// Mac includes
#include "PLStringFuncs.h"

// resource includes
#include "StringResources.h"
#include "TelnetVersion.h"

// MacTelnet includes
#include "ConnectionData.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "FileUtilities.h"
#include "FolderManager.h"
#include "PreferencesData.h"
#include "RealScreen.h"
#include "ResourceFileManager.h"
#include "ScreenManager.h"
#include "TelnetDescriptors.h"
#include "Terminal.h"



// constants
#define DEBUGGING_LOG_FILE_NAME			"\p! "APPLICATION_SHORT" Debugging Log"
#define DEBUGGING_CONSOLE_NAME			"\pStandard Error"
#define DEBUGGING_PREFIX_MAX			200
#define DEBUGGING_SPACES_PER_INDENT		4

// variables
static SInt16				gIndentationLevel = 0;
static char					gIndentPrefix[DEBUGGING_PREFIX_MAX];	// arbitrary size
static WindowRef			gConsoleWindow = null;
static TelnetScreenID		gConsoleID = 0;

// internal methods
static void		writeString		(char*			inString);



//
// public methods
//

/*!
This method sets up the initial configurations for
the debugging console window.  Call this method
if you want other methods from this file to have
any effect.

The console remains invisible; if you want to show
it, use Console_SetVisible().

IMPORTANT:	To turn off debugging completely (and
			render all subsequent Console_XYZ()
			calls ineffective), simply avoid calling
			Console_Init().

(3.0)
*/
void
Console_Init	(void)
{
	// set up indentation
	strcpy(gIndentPrefix, "");
	
	// create the console
	gConsoleID = Terminal_NewWindow("\pMonaco"/* font */, 9/* font size */, kRealScreenFlagTextWraps,
									1000/* scrollback size */, 20/* rows */, 72/* columns */, false/* force save */);
	if (gConsoleID >= 0)
	{
		// set the window title
		RealScreen_SetWTitle(gConsoleID, DEBUGGING_CONSOLE_NAME);
		
		// automatically capture the debugging console data to a debugging log, located inside the
		// folder that looks like it contains MacTelnet (on Mac OS 8, this is the applications
		// folder; on Mac OS X, its the folder containing the application package)
		{
			FSSpec		debuggingLogFile;
			
			
			if (FolderManager_GetFSSpec(kTelnetFolderPackageParent, &debuggingLogFile) == noErr)
			{
				OSErr		error = noErr;
				
				
				// set up a file specification record for the log file; it doesnt have to exist
				error = FSMakeFSSpec(debuggingLogFile.vRefNum, debuggingLogFile.parID, DEBUGGING_LOG_FILE_NAME, &debuggingLogFile);
				if ((error == noErr) || (error == fnfErr))
				{
					(Boolean)Terminal_FileCaptureBegin(gConsoleID, &debuggingLogFile);
				}
			}
		}
		
		gConsoleWindow = ScreenManager_GetWindowOf(gConsoleID);
		SetWindowKind(gConsoleWindow, WIN_CONSOLE);
		MoveWindow(gConsoleWindow, 0, 0, false/* activate */);
		Console_SetVisible(false); // initialize a static variable in this method
		
		// initialize more of the consoles connection settings, many of which dont apply to the console anyway
		VSwrite(gConsoleID, "\033[24;0H", 7); // move to the bottom of the display
	}
	else BEEP;
}// Init


/*!
This method cleans up the file transfer protocol
activity log window by destroying data structures
allocated by Console_Init().  It also resets the
screen ID to an invalid integer so that future
calls to write to the console have no effect.

(3.0)
*/
void
Console_Done	(void)
{
	Terminal_FileCaptureEnd(gConsoleID);
	gConsoleID = -1;
}// Done


/*!
To signal the start of a new function in the
consoles output, use this method.  This lets
you organize printed statements, since the
console increases the indentation level of
everything that is subsequently written to the
console.  To end the output for a function,
use Console_EndFunction(), which decreases the
indentation level of subsequent writes.

You cannot call this routine too often - if
you call it more than a few hundred times, it
will silently have no effect.

(3.0)
*/
void
Console_BeginFunction	(void)
{
	SInt16		i = 0;
	
	
	if ((gIndentationLevel += DEBUGGING_SPACES_PER_INDENT) > DEBUGGING_PREFIX_MAX) gIndentationLevel = DEBUGGING_PREFIX_MAX;
	strcpy(gIndentPrefix, "");
	for (i = 0; i < gIndentationLevel; i++) strcat(gIndentPrefix, " ");
}// BeginFunction


/*!
To signal the start of a new function in the
consoles output, use this method.  This lets
you organize printed statements, since the
console increases the indentation level of
everything that is subsequently written to the
console.  To end the output for a function,
use Console_EndFunction(), which decreases the
indentation level of subsequent writes.

You cannot call this routine too often - if
you call it more than a few hundred times, it
will silently have no effect.

(3.0)
*/
void
Console_EndFunction		(void)
{
	SInt16		i = 0;
	
	
	if ((gIndentationLevel -= DEBUGGING_SPACES_PER_INDENT) < 0) gIndentationLevel = 0;
	strcpy(gIndentPrefix, "");
	for (i = 0; i < gIndentationLevel; i++) strcat(gIndentPrefix, " ");
}// EndFunction


/*!
To obtain the Mac OS window pointer for the
debugging console window, call this method.

(3.0)
*/
WindowRef
Console_GetWindow	(void)
{
	WindowRef		result = null;
	
	
	result = gConsoleWindow;
	return result;
}// GetWindow


/*!
To determine if the log is showing, call this
method.

(3.0)
*/
Boolean
Console_IsVisible	(void)
{
	Boolean		result = false;
	
	
	if (gConsoleWindow != null) result = IsWindowVisible(gConsoleWindow);
	return result;
}// IsVisible


/*!
To show or hide the debugging console,
call this method.

(3.0)
*/
void
Console_SetVisible		(Boolean	inIsVisible)
{
	static SInt16		oldY = 0;
	WindowRef			window = gConsoleWindow;
	
	
	if ((window != null) && (inIsVisible != IsWindowVisible(window)))
	{
		Rect		rect;
		Boolean		giveUp = false;
		
		
		// animate window by scrolling it and attaching it to the menu bar
		(OSErr)GetWindowBounds(window, kWindowContentRgn, &rect);
		if ((!inIsVisible) && FlagManager_Test(kFlagCarbonWindowManager))
		{
			UInt32			junk = 0;
			SInt16			i = 0,
							j = 0,
							y = 0;
			Boolean			done = false;
			
			
			if (!inIsVisible)
			{
				// preserve the vertical position before the hide so subsequent attempts to show will work
				oldY = rect.top;
			}
			j = (inIsVisible) ? oldY/*rect.top*/ : (rect.top - rect.bottom);
			i = (inIsVisible) ? (rect.top - rect.bottom) : rect.top;
			if (inIsVisible)
			{
				MoveWindow(window, rect.left, i, false/* activate */); // arbitrary vertical starting point
				ShowWindow(window);
			}
			
			while (!done)
			{
				(OSErr)YieldToAnyThread();
				
				// make sure the debugging console remains attached to the menu bar
				if (inIsVisible) y = (i > oldY) ? oldY : i;
				else y = i;
				
				MoveWindow(window, rect.left, y, false/* activate */);
				Delay((inIsVisible) ? 2 : 1, &junk); // arbitrary animation pauses, in 60ths of a second
				done = (inIsVisible) ? (i > j) : (i < j);
				i += (inIsVisible) ? 35 : -35;	// arbitrary animation amounts, in pixels down or up, respectively
			}
		}
		
		// guarantee proper window state and alignment, following any animation
		if (inIsVisible)
		{
			MoveWindow(window, rect.left, GetMBarHeight(), false/* activate */);
			ShowWindow(window);
			SelectOverRealFrontWindow(window);
		}
		else HideWindow(window);
	}
}// SetVisible


/*!
To write a standard horizontal line (made of
dashes) to the console, followed by a carriage
return and line feed, call this method.

(3.0)
*/
void
Console_WriteHorizontalRule	(void)
{
	Console_WriteLine("----------------------------------------------------------------");
}// WriteHorizontalRule


/*!
To write data to the console, followed by a
carriage return and line feed, and preceded by
the identifier of the current thread, use this
method.  Pass an empty string to write just a
new-line character.

The specified text is automatically indented by
the appropriate amount, depending on how many
times Console_BeginFunction() has been called
without a balancing Console_EndFunction() call.

(3.0)
*/
void
Console_WriteLine	(char*		inString)
{
	TelnetThreadRef		currentThread = null;
	SInt16				currentThreadDescriptor = 0;
	char				s[32];
	
	
	currentThread = Threads_GetCurrent();
	if (currentThread == null) currentThreadDescriptor = 0;
	else currentThreadDescriptor = Threads_GetDescriptorOf(currentThread);
	
	writeString(gIndentPrefix);
	sprintf(s, "%d|", currentThreadDescriptor);
	writeString(s);
	{
		// write colon-delimited data in chunks, so it all shows up;
		// long lines get skipped otherwise (pathnames are usually
		// the reason for having long lines)
		char*		p = inString;
		
		
		p = strtok(inString, ":");
		writeString(p);
		while (p = strtok(null, ":"))
		{
			writeString(":");
			writeString(p);
		}
	}
	writeString("\015\012");
}// WriteLine


/*!
Writes a line to the console that includes the time
and date.

(3.0)
*/
void
Console_WriteTimeStamp		(char*		inLabel)
{
	char			dateString[40],
					timeString[40];
	unsigned long	date = 0L;
	
	
	GetDateTime(&date);
	DateString(date, longDate, STATIC_CAST(dateString, StringPtr), null/* intlHandle */);
	StringUtilities_PToCInPlace(STATIC_CAST(dateString, StringPtr));
	TimeString(date, true/* want seconds */, STATIC_CAST(timeString, StringPtr), null/* intlHandle */);
	StringUtilities_PToCInPlace(STATIC_CAST(timeString, StringPtr));
	strcat(dateString, " at ");
	strcat(dateString, timeString);
	strcat(dateString, ":");
	Console_WriteLine(dateString);
	Console_WriteLine(inLabel);
}// WriteTimeStamp


/*!
To write the value of a numeric variable,
use this method.  A string of the form
"label = value" is written to the console
(with a new-line).

(3.0)
*/
void
Console_WriteValue		(char*		inLabel,
						 SInt32		inValue)
{
	char		s[255];
	
	
	sprintf(s, "%s = %ld", inLabel, inValue);
	Console_WriteLine(s);
}// WriteValue


/*!
To write the information inside a file
system specification record to the console,
use this method.  A string of the form
"label = {volume, directory, name}" is written
to the console (with a new-line).

(3.0)
*/
void
Console_WriteValueFile		(char*			inLabel,
							 FSSpec const*	inFSSpecPtr)
{
	if (inFSSpecPtr != null)
	{
		char		s[255],
					name[255];
		Str255		fileName;
		OSErr		error = noErr;
		
		
		error = FileUtilities_GetPathnameFromFSSpec(inFSSpecPtr, fileName, false/* is directory - unknown */);
		StringUtilities_PToC(fileName, name);
		if (strlen(name) && (error == noErr))
		{
			sprintf(s, "%s = %s", inLabel, name);
		}
		else
		{
			// cant get the pathname - does the file not exist?
			Console_WriteValue("Console_WriteValueFile: could not find pathname, error", error);
			PLstrcpy(fileName, inFSSpecPtr->name);
			StringUtilities_PToC(fileName, name);
			sprintf(s, "%s = {%d, %ld, %s}", inLabel, inFSSpecPtr->vRefNum, inFSSpecPtr->parID, name);
		}
		Console_WriteLine(s);
	}
	else Console_WriteLine("[Console_WriteValueFile failed!]");
}// WriteValueFile


/*!
To write the value of an "OSType" variable,
use this method.  A string of the form
"label = value" is written to the console
(with a new-line).

(3.0)
*/
void
Console_WriteValueFourChars		(char*			inLabel,
								 FourCharCode	inValue)
{
	char	s[255];
	char	type[5];
	
	
	*((FourCharCode*)type) = inValue;
	type[4] = '\0'; // terminate string
	sprintf(s, "%s = %s", inLabel, type);
	Console_WriteLine(s);
}// WriteValueFourChars


/*!
To write the value of a Pascal string variable,
use this method.  A string of the form
"label = value" is written to the console
(with a new-line).

(3.0)
*/
void
Console_WriteValuePString		(char*			inLabel,
								 StringPtr		inValue)
{
	char		buffer[255],
				s[255];
	
	
	if (inValue != null) StringUtilities_PToC(inValue, buffer);
	else StringUtilities_PToC("\p<null>", buffer);
	sprintf(s, "%s = %s", inLabel, buffer);
	Console_WriteLine(s);
}// WriteValueFourChars


//
// internal methods
//

/*!
To write data to the debugging console,
call this method.  No new-line is added.

(3.0)
*/
void
writeString		(char*		inString)
{
	// safety check - do nothing unless Console_Init() was called
	if ((gConsoleID >= 0) && (Console_GetWindow() != null))
	{
		short		size = 0;
		
		
		size = strlen(inString);
		if ((size >= 1) && (size <= 80)) VSwrite(gConsoleID, inString, size);
	}
}// writeString

#endif /* ifndef NDEBUG */