/*###############################################################

	ContextsLib.h
	
	This file lists all public constants, functions, and other
	data from all modules of the Contexts Library, which is
	useful for determining the current context in a Mac program.
	This is a portable library, meaning it can be used with any
	program.
	
	Before you use any routine in this library, make sure it is
	available by using the ContextsAvailable() macro.  If the
	library is available, invoke the Contexts_Init() routine
	before using any of its modules.  You can determine the
	version of the library by using Contexts_Version().
	
	Contexts Library 1.0
	 1998-2000 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

/*!
	@header Contexts Utility Library
*/

#ifndef __CONTEXTSLIB__
#define __CONTEXTSLIB__

// Mac includes
#ifndef __MACTYPES__
#	include <MacTypes.h>
#endif
#ifndef __MENUS__
#	include <Menus.h>
#endif

// library includes
#ifndef __RELEASES__
#	include "Releases.h"
#endif

#ifdef powerc



/**************************************************************************************************************************/
// Contexts.c
//
// This module defines functions which apply to the library as a whole.  Use ContextsAvailable() to make sure that this
// shared library can be found by your application (otherwise calling any routine from it will cause your application to
// crash).  Call Contexts_Init() before using any other routine in this library.  Use the Contexts_Version() method if
// your application supports a particular version, or multiple versions, of this library.  See "Releases.h" for valid
// return codes from Contexts_Version().

/*###############################################################
	DETERMINING SPECIFIC CONTEXTS LIBRARY INFORMATION
###############################################################*/

ApplicationSharedLibraryVersion
	Contexts_Version	(void);

#ifndef __cplusplus
#define ContextsAvailable() 	(Contexts_Version != (void*)kUnresolvedCFragSymbolAddress)
#else
inline Boolean
ContextsAvailable()
{
	return (Contexts_Version != (void*)kUnresolvedCFragSymbolAddress);
}
#endif



/**************************************************************************************************************************/
// AppleGuideUtilities.c
//
// For most applications, this module is all the Apple Guide support you'll ever need.  These functions are designed to
// simplify the Apple Guide APIs by assuming that the main help file (the only file in the vast majority of applications
// that use Apple Guide) should be addressed for all requests.  You can use this module to open your help system, either
// to its main topics screen or to particular types of views.  An easy way to manage contextual help buttons in dialog
// boxes is to make sure that all of their help topics are listed in your guide file's index.  Then, simply use the method
// AppleGuideUtilities_OpenFirstMainGuideInSearch(), specifying the name of the index term, and instantly your guide file
// will show contextual help to the user for your dialog!
//
// IMPORTANT:	Since Apple Guide is not supported by Mac OS X, these interfaces are not available for versions of this
//				library that are built for Carbon compatibility.

/*###############################################################
	OPENING SEQUENCES FROM APPLE GUIDE FILES
###############################################################*/

// NO-OP ON CARBON
OSStatus AppleGuideUtilities_OpenFirstMainGuide				(void);

// the given string is converted to use the Apple Help interfaces under Carbon
OSStatus AppleGuideUtilities_OpenFirstMainGuideInSearch		(ConstStr255Param	inSearchString);

// NO-OP ON CARBON
OSStatus AppleGuideUtilities_OpenFirstMainGuideInView		(const short		inViewType);

// NO-OP ON CARBON
OSStatus AppleGuideUtilities_OpenFirstMainGuideShowSequence	(const short		inSequenceID);




/**************************************************************************************************************************/
// ContextualMenus.c
//
// This incredible module will simplify handling contextual menus in your application beyond what you thought possible.
// Fill out a ContextualMenuItem structure for every menu item that *can* appear in a menu, define context-check routines
// that can verify whether certain items should appear, and you're done!  The Contextual Menus module creates a Mac OS
// menu that's ready to be passed to the standard system routines for displaying contextual menus.  You can add items
// conditionally or unconditionally (including dividers, by passing "null" for the ContextualMenuItem structure pointer).
// The flexible Context Verifier Method is a function template for your custom context-checking routines, and provides
// your application with an event record and *two* pieces of data defined by you to indicate what command is being
// queried!

enum
{
	kContextualMenuID = 230
};

/*!
Context Verifier Method

This method gets called by ContextualMenus_AddProvisionally() to
see if a particular item should be added based on the given context-
setting parameters.  Return "true" if the item should be added, or
"false" if it should not be.

The event record provided is a snapshot of the most recent event
that set up the context; this information is almost always needed.
The user data is completely custom, and is a mechanism for allowing
the caller of this arbitration method to specify clues as to the
context that the method should consider.  The command information
parameter is useful to implement variation codes, so you can write
a function that discriminates similar contexts and behaves in certain
ways based on the code provided.

Although this callback is currently used for contextual menus, it
is meant to be flexible enough to handle any generic context
specification.  Therefore, if possible, try to write context
discriminators in such a way as to consider menu items as "commands"
which could represent any invocation method.
*/
typedef Boolean (*ContextVerifierProcPtr)	(EventRecord*		inoutEventPtr,
											 Ptr				inoutUserData,
											 const UInt32		inCommandInfo);
#define InvokeContextVerifierProc(userRoutine, inoutEventPtr, inoutUserData, inCommandInfo) \
	(*userRoutine)((inoutEventPtr), (inoutUserData), (inCommandInfo))

// types
struct ContextualMenuItem
{
	UInt32		commandID;
	Str255		commandText;
	UInt8 		iconType;
	Handle 		iconHandle;
};
typedef struct ContextualMenuItem ContextualMenuItem;
typedef ContextualMenuItem* ContextualMenuItemPtr;
typedef ContextualMenuItemPtr* ContextualMenuItemHandle;



/*###############################################################
	CREATING, INITIALIZING AND DESTROYING CONTEXTUAL MENUS
###############################################################*/

MenuHandle
	ContextualMenus_New					(SInt16*						outMenuID);

void
	ContextualMenus_Dispose				(MenuHandle*					inoutDeadMenu);

void
	ContextualMenus_InitItem			(ContextualMenuItemPtr			inoutItem);

/*###############################################################
	ADDING ITEMS TO CONTEXTUAL MENUS
###############################################################*/

OSErr
	ContextualMenus_Add					(MenuHandle						inToWhichMenu,
										 const ContextualMenuItemPtr	inItemInfo);

OSErr
	ContextualMenus_AddProvisionally	(MenuHandle						inToWhichMenu,
										 const ContextualMenuItemPtr	inItemInfo,
										 ContextVerifierProcPtr			inArbiter,
										 EventRecord*					inoutEventPtr,
										 Ptr							inoutUserDataPtr,
										 const UInt32					inCommandInfo);

void
	ContextualMenus_DoneAdding			(MenuHandle						inToWhichMenu);

// USE THIS METHOD EVERY TIME A DIVIDING LINE *MAY* BE NECESSARY (AVOIDS ADDING MORE THAN ONE DIVIDER)
void
	ContextualMenus_NewGroup			(MenuHandle						inForWhichMenu);



/**************************************************************************************************************************/
// Undoables.c
//
// With the power of this module at your fingertips, theres no excuse not to support Undo and Redo in your application!
// Using a simple but very flexible interface, you can register actions as being undoable as soon as they occur.  This
// routine will retain relevant data (a context), as well as a pointer to your action routine that can use your context
// to perform an Undo or Redo on demand.  The code could not be simpler - just use one function call in response to the
// user selecting Undo, and a different function call if the user specifies Redo.  This method handles the rest - you
// can even find out useful information about what the correct command text and enabled state is for the Undo and Redo
// items in the Edit menu (because they are specified at action registration time).  Although this module does not yet
// support multiple Undo operations, the interface is designed to easily accommodate this enhancement in the future
// by requiring you to make only a single change to your code - the parameter to Undoables_Init().

typedef struct OpaqueUndoableAction { unsigned garbage; }*	UndoableActionRef;

typedef SInt32 UndoableContextIdentifier;	// user-defined ID for a generic context pointer

typedef SInt16 UndoActionInstruction;
enum
{
	kUndoActionInstructionUndo = 1,		// requesting procedure to undo the action that is apparently defined by the context
	kUndoActionInstructionRedo = 2,		// requesting procedure to redo the action that is apparently defined by the context
	kUndoActionInstructionDispose = 3	// the action is being destroyed for some reason - deallocate contexts memory, etc.
};

typedef SInt16 UndoHandlingMechanism;
enum
{
	kUndoHandlingMechanismOnlyOneUndo = 0		// currently, only single Undo is supported
};

/*!
Undo Action Method

This routine is called when the top action on the
Undoables stack is told to Undo itself.  Enough
information is passed to the routine that you
should be able to undo any operation.  For any
instruction that does not have a defined return
value, always return 0.

IMPORTANT:	Your routine should not respond to any
			instruction it does not recognize!
*/
typedef void (*UndoActionProcPtr)		(const UndoActionInstruction	inDoWhat,
										 UndoableContextIdentifier		inContextIdentifier,
										 void*							inContextPtr);
#define InvokeUndoableActionProc(userRoutine, inDoWhat, inContextIdentifier, inContextPtr) \
			(*userRoutine)((inDoWhat), (inContextIdentifier), (inContextPtr))



/*###############################################################
	INITIALIZING AND FINISHING WITH UNDOABLES
###############################################################*/

// CALL THIS ROUTINE ONCE, BEFORE ANY OTHER UNDOABLES ROUTINE
void
	Undoables_Init					(const UndoHandlingMechanism	inUndoHandlingMechanism,
									 StringPtr						inDisabledUndoCommandName,
									 StringPtr						inDisabledRedoCommandNameOrNullIfRedoIsNotUsed);

// CALL THIS ROUTINE AFTER YOU ARE PERMANENTLY DONE WITH UNDOABLES
void
	Undoables_Done					(void);

/*###############################################################
	MANAGING THE STACK OF UNDOABLE OPERATIONS
###############################################################*/

OSErr
	Undoables_AddAction				(StringPtr						inUndoCommandName,
									 StringPtr						inRedoCommandNameOrNullIfNotRedoable,
									 UndoActionProcPtr				inHowToUndoAction,
									 UndoableContextIdentifier		inUserDefinedContextIdentifier,
									 void*							inUserDefinedContextPtr);

void
	Undoables_RedoLastUndo			(void);

void
	Undoables_UndoLastAction		(void);

/*###############################################################
	SETTING UP THE STATES OF EDIT MENU ITEMS
###############################################################*/

void
	Undoables_GetRedoCommandInfo	(Str255							outCommandText,
									 Boolean*						outCommandEnabled);

void
	Undoables_GetUndoCommandInfo	(Str255							outCommandText,
									 Boolean*						outCommandEnabled);

#endif /* ifdef powerc */

#endif