/*###############################################################

	ContextualMenuBuilder.c
	
	This module generates context-based pop-up menu items
	appropriately for a particular object (such as a window).
	The brand new callback mechanism introduced in Telnet 3.0
	means that much of the context work is actually done in
	"Menus.c"!  By using GetMenuItemEnabledByCommandID(), you
	automatically invoke a context handler to determine exactly
	what the item state should be.  If the item is enabled, it
	means that it applies to the current window.  All thats
	left to determine is *what* commands you want in the menu.

	IMPORTANT: Do not over-do contextual menus!  When you build
	           contextual menus, it is important to keep the
			   following, *critical* points in mind:
			   
	- If you are adding an item to a contextual menu, is the
	  item available in a regular, menu bar menu as well?
			   
	- How many items are in this contextual menu?  Do NOT simply
	  add every *possible* item to a contextual menu!  Restrict
	  your choices to those items which will most certainly be
	  useful to the user, either because they are frequently
	  used, or because they are naturally suited to a pop-up
	  menu (such as Copy).
			   
	- With the *sole* exception of possibly the Help item, never
	  show disabled menu items in any contextual menu.  Items
	  which are not available just make menus unnecessarily
	  large, and do not help the user in any way.
	
	- How long does it take your context handler to do its work?
	  Remember, all of that processing time is taken up *after*
	  the user Control-clicks, so it delays the display of the
	  menu to the user!
	
	When a contextual menu pops up, it should contain quick,
	frequently-used and nondestructive commands, and in special
	circumstances only, attributes.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#if TARGET_API_MAC_OS8
#include <Balloons.h>
#endif
#include <MacErrors.h>
#include <MacHelp.h>
#include <MacWindows.h>
#include <Menus.h>
#include <PLStringFuncs.h>
#include <Resources.h>
#include <Scrap.h>

// library includes
#include "ContextsLib.h"
#include "InterfaceLib.h"

// resource includes
#include "MenuResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "Clipboard.h"
#include "ContextualMenuBuilder.h"
#include "DialogUtilities.h"
#include "RealScreenData.h"
#include "RealScreen.h"
#include "ScreenManager.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



// variables
static UInt32		gHelpType = kCMHelpItemNoHelp;

// internal methods
static void buildClipboardWindowContextualMenu		(MenuHandle		inMenuHandle,
													 WindowRef		inWhichWindow,
													 EventRecord*	inoutEventPtr);

static void buildEmptyContextualMenu				(MenuHandle		inMenuHandle,
													 WindowRef		inWhichWindow,
													 EventRecord*	inoutEventPtr);

static void buildTerminalWindowContextualMenu		(MenuHandle		inMenuHandle,
													 WindowRef		inWhichWindow,
													 EventRecord*	inoutEventPtr);



//
// public methods
//

/*!
To determine the kind of help item that will be
available in displayed contextual menus, use
this method.  You can specify this information
using ContextualMenuBuilder_SetMenuHelpType().

(3.0)
*/
UInt32
ContextualMenuBuilder_GetMenuHelpType	(void)
{
	return gHelpType;
}// GetMenuHelpType


/*!
This method displays a contextual menu for the
specified window that is appropriate for the
window.  Call this method after you have
determined that a contextual menu display event
has occurred in the frontmost window, and you
know what part of the window was hit (that is,
the result code from FindWindow() is known).

(3.0)
*/
OSStatus
ContextualMenuBuilder_HandleWindow	(WindowRef		inWhichWindow,
									 EventRecord*	inoutEventPtr,
									 SInt16 const	inWindowPart)
{
	WindowInfoRef				windowFeaturesRef = null;
	WindowInfoDescriptor		windowDescriptor = kInvalidWindowInfoDescriptor;
	SInt16						windowKind = GetWindowKind(inWhichWindow);
	MenuHandle					menuHandle = null;
	short						i = 0;
	OSStatus					result = noErr;
	Boolean						failed = false;
	Boolean						isConnectionWindow = false;
	
	
	windowFeaturesRef = WindowInfo_GetFromWindow(inWhichWindow);
	if (windowFeaturesRef != null) windowDescriptor = WindowInfo_GetWindowDescriptor(windowFeaturesRef);
	isConnectionWindow = ScreenManager_IsWindowAScreen(inWhichWindow);
	
	// get a unique menu ID within the allowed range
	menuHandle = ContextualMenus_New((SInt16*)null);
	
	if (menuHandle != null)
	{
		Boolean		addedBalloonItem = false;
		
		
		// add a Balloon Help item for dialogs
		if (GetWindowKind(inWhichWindow) == kDialogWindowKind)
		{
			ContextualMenuItem	itemInfo;
			
			
			ContextualMenus_InitItem(&itemInfo);
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands,
							#if TARGET_API_MAC_CARBON
								HMAreHelpTagsDisplayed()
							#else
								HMGetBalloons()
							#endif
								? siContextHideBalloons : siContextShowBalloons);
			addedBalloonItem = (ContextualMenus_Add(menuHandle, &itemInfo) == noErr);
		}
		
		// determine what should go in the menu
		switch (inWindowPart)
		{
			//case inSysWindow:
			case inContent:
			case inDrag:
			case inGrow:
			case inGoAway:
			case inZoomIn:
			case inZoomOut:
			case inCollapseBox:
			case inProxyIcon:
				switch (windowDescriptor)
				{
					case kTelnetWindowDescriptorClipboard:
						buildClipboardWindowContextualMenu(menuHandle, inWhichWindow, inoutEventPtr);
						break;
					
					default:
						if (isConnectionWindow)
							buildTerminalWindowContextualMenu(menuHandle, inWhichWindow, inoutEventPtr);
						else
							buildEmptyContextualMenu(menuHandle, inWhichWindow, inoutEventPtr);
						break;
				}
				break;
			
			default:
				failed = true;
				break;
		}
		
		if (!failed)
		{
			UInt32		selectionType = 0;
			SInt16		menuID = 0;
			UInt16		itemNumber = 0;
			AEDesc		textDesc;
			AEDescPtr	paramDesc = null;
			Handle		handle = null;
			
			
			// Depending upon the window type, there may be data that can be used by contextual
			// menu plug-ins.  The method of getting that data also depends on the window type.
			// See what kind of window the contextual-menu-display event was for, and send that
			// windows data (usually the current selection) to the plug-ins via a new Apple
			// Event descriptor.
			if (isConnectionWindow)
			{
				// Let the Contextual Menu Manager know what data is selected in the
				// frontmost window, so that Contextual Menu Plug-Ins can parse it.
				TelnetScreenID			id = ScreenManager_GetWindowScreenID(GetRealFrontWindow());
				RealScreenDataPtr		realScreen = null;
				
				
				realScreen = ScreenManager_GetRealScreenDataOf(id);
				if (realScreen->isTextSelected)
				{
					handle = RealScreen_GetSelectedTextAsNewHandle(id, 0, 0/* flags */);
					if (handle != null)
					{
						if (AECreateDesc(typeChar, *handle, GetHandleSize(handle), &textDesc) == noErr)
						{
							paramDesc = &textDesc;
						}
					}
				}
			}
			else if (windowDescriptor == kTelnetWindowDescriptorClipboard)
			{
				// Let the Contextual Menu Manager know what data is on the clipboard,
				// so that any installed Contextual Menu Plug-Ins can parse it.
				if (Clipboard_CreateContentsAEDesc(&textDesc) == noErr) paramDesc = &textDesc;
			}
			
			// now that all items have been added, notify the Contextual Menu modules to clean up
			ContextualMenus_DoneAdding(menuHandle);
			
			// display and track the menu, getting the name of the Help item from the MacTelnet resource file
			{
				Str31		helpString;
				
				
				GetIndString(helpString, rStringsContextualMenuCommands, siContextHelp);
				result = ContextualMenuSelect(menuHandle, inoutEventPtr->where, false/* reserved */, gHelpType,
												helpString, paramDesc, &selectionType, &menuID, &itemNumber);
			}
			
			if (paramDesc != null)
			{
				// dispose of memory that the Apple Event Manager allocated
				(OSErr)AEDisposeDesc(paramDesc);
			}
			
			if (result == noErr)
			{
				if (selectionType != kCMNothingSelected)
				{
					if (selectionType == kCMShowHelpSelected)
					{
						if (ContextsAvailable()) AppleGuideUtilities_OpenFirstMainGuide();
					}
					else if (selectionType == kCMMenuItemSelected)
					{
						UInt32		commandID;
						
						
						if (menuID == kContextualMenuID)
						{
							if ((addedBalloonItem) && (itemNumber == 1)) // the Balloon Help item was selected
							{
								#if TARGET_API_MAC_CARBON
								HMSetHelpTagsDisplayed(!HMAreHelpTagsDisplayed());
								#else
								HMSetBalloons(!HMGetBalloons());
								#endif
							}
							else if (GetMenuItemCommandID(menuHandle, itemNumber, &commandID) == noErr) // otherwise, let the handler take it
							{
								failed = (!Menus_HandleCommandByID(commandID, inoutEventPtr->modifiers));
							}
							else failed = true;
						}
					}
					else failed = true;
				}
			}
			else if (result != userCanceledErr) failed = true;
			
			if (handle != null) Memory_DisposeHandle(&handle);
		}
		ContextualMenus_Dispose(&menuHandle);
	}
	else
	{
		failed = true;
		result = memFullErr;
	}
	
	return result;
}// HandleWindow


/*!
To specify the kind of help item that will be
available in displayed contextual menus, use this
method.  You can determine this information later
using ContextualMenuBuilder_GetMenuHelpType().

(3.0)
*/
void
ContextualMenuBuilder_SetMenuHelpType	(UInt32 const		inNewHelpType)
{
	gHelpType = inNewHelpType;
}// SetMenuHelpType


//
// internal methods
//

/*!
This routine will look at the frontmost window as being
the Clipboard, and will construct a contextual menu
appropriate for it.  The given window must have
Window Info associated with it (see "WindowInfo.c").

(3.0)
*/
static void
buildClipboardWindowContextualMenu	(MenuHandle		inMenuHandle,
									 WindowRef		UNUSED_ARGUMENT(inWhichWindow),
									 EventRecord*	UNUSED_ARGUMENT(inoutEventPtr))
{
	ContextualMenuItem	itemInfo;
	MenuHandle			menuBarMenu = null;
	
	
	Menus_SetUpItemStates();
	
	//
	// Close command
	//
	ContextualMenus_InitItem(&itemInfo);
	itemInfo.commandID = kCommandCloseConnection;
	if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
	{
		GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextCloseConnection);
		(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Close This Window
	}
	
	//
	// Stack Windows command
	//
	ContextualMenus_InitItem(&itemInfo);
	itemInfo.commandID = kCommandStackWindows;
	if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
	{
		GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextStackWindows);
		(OSErr)ContextualMenus_Add(inMenuHandle, null); // add a divider
		(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Stack Windows
	}
}// buildClipboardWindowContextualMenu


/*!
This routine will look at the frontmost window as one that
is not recognized, and will construct a contextual menu
appropriate for it.  The given window must have
Window Info associated with it (see "WindowInfo.c").

(3.0)
*/
static void
buildEmptyContextualMenu	(MenuHandle		UNUSED_ARGUMENT(inMenuHandle),
							 WindowRef		UNUSED_ARGUMENT(inWhichWindow),
							 EventRecord*	UNUSED_ARGUMENT(inoutEventPtr))
{
	//ContextualMenuItem	itemInfo;
	//MenuHandle			menuBarMenu = null;
	
	
	Menus_SetUpItemStates();
}// buildEmptyContextualMenu


/*!
This routine will look at the frontmost window as a
terminal window, determine its characteristics and
construct a contextual menu appropriate for its
current context.  To keep menus short, this routine
builds two types of menus: one for text selections,
and one if nothing is selected.  If text is selected,
this routine assumes that the user will most likely
not want any of the more-general items, such as
Set Window Title.  Otherwise, these general items
are available if applicable.

(3.0)
*/
static void
buildTerminalWindowContextualMenu	(MenuHandle		inMenuHandle,
									 WindowRef		UNUSED_ARGUMENT(inWhichWindow),
									 EventRecord*	UNUSED_ARGUMENT(inoutEventPtr))
{
	ContextualMenuItem	itemInfo;
	MenuHandle			menuBarMenu = null;
	Boolean				isTextSelected = false;
	
	
	Menus_SetUpItemStates();
	
	{
		TelnetScreenID			id = ScreenManager_GetWindowScreenID(GetRealFrontWindow());
		RealScreenDataPtr		realScreen = null;
		
		
		realScreen = ScreenManager_GetRealScreenDataOf(id);
		
		// is any text selected?  If so, all of these commands apply
		isTextSelected = (realScreen->isTextSelected);
	}
	
	if (isTextSelected) // no need to keep checking for context, since the same context applies to all items!
	{
		// clipboard-related text selection commands
		ContextualMenus_NewGroup(inMenuHandle);
		
		ContextualMenus_InitItem(&itemInfo);
		itemInfo.commandID = kCommandCopy;
		if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
		{
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextCopy);
			(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Copy
		}
		itemInfo.commandID = kCommandCopyTable;
		if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
		{
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextCopyTable);
			(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Copy Using Tabs For Spaces
		}
		
		// other text-selection-related commands
		ContextualMenus_NewGroup(inMenuHandle);
		
		itemInfo.commandID = kCommandPrintOne;
		GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextPrintOne);
		(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Print Selection Now
	}
	else
	{
		// text-editing-related menu items
		ContextualMenus_NewGroup(inMenuHandle);
		
		ContextualMenus_InitItem(&itemInfo);
		itemInfo.commandID = kCommandPaste;
		if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
		{
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextPaste);
			(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Paste
		}
		
		// window arrangement menu items
		ContextualMenus_NewGroup(inMenuHandle);
		
		ContextualMenus_InitItem(&itemInfo);
		itemInfo.commandID = kCommandHideFrontWindow;
		if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
		{
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextHideFrontWindow);
			(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Hide This Window
		}
		
		ContextualMenus_InitItem(&itemInfo);
		itemInfo.commandID = kCommandStackWindows;
		if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
		{
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextStackWindows);
			(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Stack Windows
		}
		
		// connection-related menu items
		ContextualMenus_NewGroup(inMenuHandle);
		
		ContextualMenus_InitItem(&itemInfo);
		itemInfo.commandID = kCommandChangeWindowTitle;
		if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
		{
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextChangeWindowTitle);
			(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Set Title
		}
		
		ContextualMenus_InitItem(&itemInfo);
		itemInfo.commandID = kCommandSetScreenSize;
		if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
		{
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextSetScreenSize);
			(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Screen Dimensions
		}
		
		ContextualMenus_InitItem(&itemInfo);
		itemInfo.commandID = kCommandFormat;
		if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
		{
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextFormat);
			(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Format
		}
		
		ContextualMenus_InitItem(&itemInfo);
		itemInfo.commandID = kCommandSetKeys;
		if (GetMenuItemEnabledByCommandID(itemInfo.commandID))
		{
			GetIndString(itemInfo.commandText, rStringsContextualMenuCommands, siContextSetKeys);
			(OSErr)ContextualMenus_Add(inMenuHandle, &itemInfo); // add Set Keys
		}
	}
}// buildTerminalWindowContextualMenu