/*###############################################################

	ContextualMenus.c
	
	Generates context-based pop-up menus when needed.
	
	Contexts Library 1.0
	 1998-2000 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#ifndef __ERRORS__
#	include <Errors.h>
#endif
#ifndef __MENUS__
#	include <Menus.h>
#endif
#ifndef __PLSTRINGFUNCS__
#	include <PLStringFuncs.h>
#endif
#ifndef __RESOURCES__
#	include <Resources.h>
#endif

// library includes
#include "ContextsLib.h"

// variables
static UInt16		itemCountAddedGroup = 0,
					itemCountAddedTotal = 0;

// internal methods
static UInt16		getGroupAddedItemCount		(void);
static UInt16		getTotalAddedItemCount		(void);
static void			setGroupAddedItemCount		(UInt16		inCount);
static void			setTotalAddedItemCount		(UInt16		inCount);



//
// public methods
//

/**
To create a new, unique menu to use as a contextual menu,
call this routine.  The new menu, if successfully created,
will already have been added to the hierarchical menu list
(the required location for special menus like contextual
menus).  If no valid ID can be found, or there is no
memory, or some other problem occurs, null is returned.

The parameter "outMenuID" is optional; if non-null, this
method returns the menu ID (not a resource ID) of the new
contextual menu.

IMPORTANT:	Before displaying the new menu, be sure to
			invoke the ContextualMenus_DoneAdding() method.

(1.0)
*/
MenuHandle
ContextualMenus_New		(SInt16*		outMenuID)
{
	MenuHandle		result = null;
	
	
	// reinitialize globals
	setGroupAddedItemCount(0);
	setTotalAddedItemCount(0);
	
	if (outMenuID != null) *outMenuID = 0;
	
	result = NewMenu(kContextualMenuID, "\p");
	if (result != null)
	{
		Handle		themeDefProc = GetResource('MDEF', kMenuStdMenuProc);
		
		
		#if TARGET_API_MAC_OS8
		// use the theme-savvy menu definition procedure (automatically used in Carbon)
		if (themeDefProc != null)
		{
			SInt8		handleState = 0;
			
			
			handleState = HGetState((Handle)result);
			HLockHi((Handle)result);
			(*result)->menuProc = themeDefProc;
			HSetState((Handle)result, handleState);
		}
		#endif
		InsertMenu(result, kInsertHierarchicalMenu);
		
		if (outMenuID != null) *outMenuID = kContextualMenuID;
	}
	return result;
}// New


/**
To destroy a menu created with ContextualMenus_New(),
call this method.

(1.0)
*/
void
ContextualMenus_Dispose		(MenuHandle*		inoutDeadMenu)
{
	if (inoutDeadMenu != null)
	{
		DeleteMenu(kContextualMenuID);
		DisposeMenu(*inoutDeadMenu), *inoutDeadMenu = null;
	}
}// Dispose


/**
Use this method to have an item added to the
specified menu.

You can specify information about the commands
appearance in the menu (including its text,
icon and command ID) using the parameter of
type ContextualMenuItemPtr.  This structures
data is copied as needed, so you do not need to
create separate structures for every new item
you pass to this method (you can just modify a
single structure as needed).

If you do not specify command information (that
is, "inItemInfo" is null), then this method
assumes that the item in question is a dividing
line.

(1.0)
*/
OSErr
ContextualMenus_Add		(MenuHandle						inToWhichMenu,
						 const ContextualMenuItemPtr	inItemInfo)
{
	return ContextualMenus_AddProvisionally(inToWhichMenu, inItemInfo, null, null, null, 0L);
}// Add


/**
Use this method to have an item added to the
specified menu only if the context verifier
arbitrates in favor of adding the item.

You can specify information about the commands
appearance in the menu (including its text,
icon and command ID) using the parameter of
type ContextualMenuItemPtr.  This structures
data is copied as needed, so you do not need to
create separate structures for every new item
you pass to this method (you can just modify a
single structure as needed).

If you do not specify command information (that
is, "inItemInfo" is null), then this method
assumes that the item in question is a dividing
line, the inclusion of which is still bound by
the context-determining information given.  You
should use ContextualMenus_NewGroup(), however,
to allow the Contextual Menus module to add
the appropriate dividing lines for you.

The last four arguments indicate respectively
the function that is the context arbiter, and
the parameters to pass to it.  If you pass null
values for these, the item is simply inserted
into the menu without regard for any context.

If the menu item is rejected in the current
context by the specified context verifier,
"reqFailed" is returned.

(1.0)
*/
OSErr
ContextualMenus_AddProvisionally	(MenuHandle						inToWhichMenu,
									 const ContextualMenuItemPtr	inItemInfo,
									 ContextVerifierProcPtr			inArbiter,
									 EventRecord*					inoutEventPtr,
									 Ptr							inoutUserDataPtr,
									 const UInt32					inCommandInfo)
{
	enum
	{
		kEndOfContextualMenu = 999
	};
	OSErr		result = noErr;
	Boolean		doInsert = false;
	
	
	if (inToWhichMenu == null) result = memPCErr;
	
	doInsert = (result == noErr);
	if (doInsert)
	{
		if (inArbiter != null)
		{
			doInsert = InvokeContextVerifierProc(inArbiter, inoutEventPtr, inoutUserDataPtr, inCommandInfo);
			if (!doInsert) result = reqFailed;
		}
		else doInsert = true;
		
		if (doInsert)
		{
			if (inItemInfo != null)
			{
				{
					UInt16		groupAddedItemCount = getGroupAddedItemCount(),
								totalAddedItemCount = getTotalAddedItemCount();
					
					
					// update the count of contextual items added to this menu
					if ((totalAddedItemCount) && (!groupAddedItemCount))
					{
						// if any items are preceding this one and its the first in its group, add a divider automatically
						InsertMenuItem(inToWhichMenu, "\p(-", kEndOfContextualMenu);
					}
					
					// increment counters for next time (group counter resets via a call to ContextualMenus_NewGroup())
					setGroupAddedItemCount(groupAddedItemCount + 1);
					setTotalAddedItemCount(totalAddedItemCount + 1);
				}
				
				InsertMenuItem(inToWhichMenu, inItemInfo->commandText, kEndOfContextualMenu);
				result = SetMenuItemCommandID(inToWhichMenu, CountMenuItems(inToWhichMenu), inItemInfo->commandID);
				if (inItemInfo->iconHandle != null)
				{
					result = SetMenuItemIconHandle(inToWhichMenu, CountMenuItems(inToWhichMenu), inItemInfo->iconType, inItemInfo->iconHandle);
				}
			}
			else
			{
				InsertMenuItem(inToWhichMenu, "\p(-", kEndOfContextualMenu);
			}
		}
	}
	return result;
}// AddProvisionally


/**
When you are finished adding items to a menu,
you MUST call this routine to clean up.

(1.0)
*/
void
ContextualMenus_DoneAdding		(MenuHandle			inWhichMenu)
{
	// okay, this routine doesnt do anything - but forcing a completion routine increases flexibility for the future
}// DoneAdding


/**
To initialize a contextual menu item structure,
call this method.  You should always invoke this
routine on any ContextualMenuItem structure you
allocate, to ensure no fields are uninitialized.

(1.0)
*/
void
ContextualMenus_InitItem	(ContextualMenuItemPtr			inoutItem)
{
	inoutItem->commandID = 0L;
	PLstrcpy(inoutItem->commandText, "\p");
	inoutItem->iconType = 0;
	inoutItem->iconHandle = null;
}// InitItem


/**
To mark subsequent calls to ContextualMenus_Add()
or ContextualMenus_AddProvisionally() as belonging
to a new group of related items, use this method.
The Contextual Menus module will automatically add
only as many dividing lines as are necessary.  So,
if no items from a particular group end up getting
added, no dividing line for that group is inserted.
This is the easiest way to manage dividing lines in
your contextual menus.

(1.0)
*/
void
ContextualMenus_NewGroup	(MenuHandle			inForWhichMenu)
{
	setGroupAddedItemCount(0);
}// NewGroup


//
// internal methods
//

/**
To determine the number of items from the
current group added to a contextual menu,
use this method.

(1.0)
*/
static UInt16
getGroupAddedItemCount		(void)
{
	return itemCountAddedGroup;
}// getGroupAddedItemCount


/**
To determine the total number of items
added to a contextual menu, use this
method.

(1.0)
*/
static UInt16
getTotalAddedItemCount		(void)
{
	return itemCountAddedTotal;
}// getTotalAddedItemCount


/**
To specify the number of items from the
current group added to a contextual menu,
use this method.

(1.0)
*/
static void
setGroupAddedItemCount		(UInt16		inCount)
{
	itemCountAddedGroup = inCount;
}// setGroupAddedItemCount


/**
To specify the number of items added to a
contextual menu, use this method.

(1.0)
*/
static void
setTotalAddedItemCount		(UInt16		inCount)
{
	itemCountAddedTotal = inCount;
}// setTotalAddedItemCount