/*###############################################################

	CoolStuff.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <DrawSprocket.h>
#include <MacErrors.h>
#include <Navigation.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "GeneralResources.h"

// MacTelnet includes
#include "CoolStuff.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "ResourceFileManager.h"
#include "TelnetDescriptors.h"
#include "TelnetThreads.h"



// constants
#define GRAY_BITS		6
#define GRAY_RANGE		(1 << GRAY_BITS)
enum
{
	kStarsGrayRange = GRAY_RANGE,		// number of possible shades of stars
	kStarsBitDifference = 4,			// 2^(this) defines how the color index is derived from the Z position
	kStarsAnimationDelayInTicks = 2,	// number of 60ths of a second between animations
	kStarsBackPlane = kStarsGrayRange * (1 << kStarsBitDifference) - 1,
	kStarsDefaultSpeed = 12,			// pixels (in the Z direction) of change per frame; can be changed later
	kStarsDefaultShift = 16				// pixels (in the Y direction) of change per frame; can be changed later
};
// the view plane is hard-coded so efficient bit shifts can be used
#define VIEW_PLANE					(GRAY_RANGE << 1)
#define VIEW_PLANE_DIVIDE(x)		((x) << GRAY_BITS << 1)

// types
struct PulseDeviceLoopParams
{
	ControlRef			control;
	ControlPartCode		newState;
	UInt16				percentage;
	OSErr				result;
};
typedef struct PulseDeviceLoopParams	PulseDeviceLoopParams;
typedef PulseDeviceLoopParams*			PulseDeviceLoopParamsPtr;

struct StarPoint
{
	SInt16		x,
				y;
	SInt8		diameter;
};
typedef struct StarPoint		StarPoint;

struct StarPosition
{
	SInt16		x,
				y,
				z;
};
typedef struct StarPosition		StarPosition;

struct StarData
{
	struct
	{
		struct
		{
			StarPoint*		positions;		// an array of static 2D points, the locations of nonmoving stars
			SInt16			count;			// determined at run time; the size of the static array above
		} field;
		
		struct
		{
			StarPosition*	positions;	// an array of dynamic 3D points, the locations of moving stars
			SInt16			count;
		} dynamic[2];
		
		struct
		{
			IconSuiteRef	large,
							small;
		} icons;
		
		SInt8			diameter;			// typical diameter of stars (they are drawn as ovals)
		RGBColor		grays[kStarsGrayRange];
	} stars;
	
	struct
	{
		SInt16		width,		// the size of the video device on which the stars appear
					height;
	} screenDimensions;
	
	struct
	{
		SInt16		x,			// the vanishing point
					y;
	} center;
	
	SInt16		shift;		// delta Y-plane per frame
	SInt16		speed;		// delta Z-plane per frame
	SInt32		time;
	Boolean		fasterThan300Megahertz;		// processor speed
};
typedef struct StarData		StarData;
typedef StarData*			StarDataPtr;

// variables
static DSpContextReference		gDefaultDrawingContext = null;
static RgnHandle				gPulseNewClipRgn = null;

// internal methods
static void				drawBlendedControlState			(ControlRef			inControl,
														 ControlPartCode	inNewState,
														 const UInt16		inBrightnessBlendingPercentage);
static pascal void		movingStarsUserPaneDraw			(ControlRef			inControl,
														 SInt16				inPart);
static pascal void		movingStarsUserPaneIdle			(ControlRef			inControl);
static pascal void		offscreenDrawBlendedDeviceLoop	(short				inColorDepth,
														 short				inDeviceFlags,
														 GDHandle			inTargetDevice,
														 long				inDataPtr);



//
// public methods
//

/*!
To initialize Apple Game Sprockets and create all
default contexts, use this method.  Be sure to call
this routine before using any other routines in
this module.

(3.0)
*/
OSStatus
CoolStuff_Init		(void)
{
	OSStatus		result = 0L;
	
	
	gPulseNewClipRgn = Memory_NewRegion();
	if (FlagManager_Test(kFlagDrawSprocket))
	{
		DSpContextAttributes	attrs;
		Rect					screenRect;
		
		
		result = noErr;
		Memory_WriteZero((Ptr)&attrs, sizeof(attrs));
		screenRect = (**(GetMainDevice())).gdRect;
		attrs.frequency = 0.0f;
		attrs.displayWidth = screenRect.right - screenRect.left;
		attrs.displayHeight = screenRect.bottom - screenRect.top;
		attrs.colorNeeds = kDSpColorNeeds_DontCare;
		attrs.colorTable = null;
		attrs.contextOptions = 0L;
		attrs.backBufferDepthMask = kDSpDepthMask_All;
		attrs.displayDepthMask = kDSpDepthMask_All;
		attrs.backBufferBestDepth = 1L;
		attrs.displayBestDepth = 1L;
		attrs.pageCount = 0L;
		attrs.gameMustConfirmSwitch = false;
		result = DSpStartup();
		if (result == noErr)
		{
			Boolean		isError = false;
			
			
			result = DSpFindBestContext(&attrs, &gDefaultDrawingContext);
			isError = (result != noErr);
			
			if (!isError) if (API_AVAILABLE(DSpContext_Reserve))
			{
				result = DSpContext_Reserve(gDefaultDrawingContext, &attrs);
				isError = (result != noErr);
			}
			
			if (isError)
			{
				// if any problems occur creating the context, shut down DrawSprocket and dont use it
				(OSStatus)DSpShutdown();
				FlagManager_Set(kFlagDrawSprocket, false);
			}
		}
	}
	else result = kDSpNotInitializedErr;
	
	return result;
}// Init


/*!
To shut down Apple Game Sprockets and destroy all
default contexts, use this method.  Be sure to call
this routine when you are completely finished with
this module.

(3.0)
*/
void
CoolStuff_Done		(void)
{
	if (FlagManager_Test(kFlagDrawSprocket))
	{
		(OSStatus)DSpContext_Release(gDefaultDrawingContext), gDefaultDrawingContext = null;
		(OSStatus)DSpShutdown();
	}
	Memory_DisposeRegion(&gPulseNewClipRgn);
}// Done


/*!
To conveniently blacken a screen (0% brightness,
black color tint), use this method.  When the
gamma is blackened, every change you make to the
screen is technically visible, and yet will not
be seen by the user because the monitor will be
entirely dark.

(3.0)
*/
void
CoolStuff_GammaBlack	(DSpContextReference	inContext)
{
	RGBColor		black;
	
	
	black.red = black.green = black.blue = 0;
	CoolStuff_GammaSet(inContext, &black, 0);
}// GammaBlack


/*!
To execute a controlled gamma fade using one color
as a starting point, use this method.  You have the
option to yield to threads while the fade takes
place, or simply executing a synchronous fade.

The value of "inStartingPercentage" must be less
than that of "inEndingPercentage" if the sign of
"inGammaChangePercentage" is positive.  Similarly,
for a negative sign, "inEndingPercentage" must be
the lesser.

If MacTelnet is suspended (not the frontmost process),
this routine automatically resets the gamma to
normal, and automatically sets it back when MacTelnet
becomes frontmost.

(3.0)
*/
void
CoolStuff_GammaControlledFade	(DSpContextReference	inContext,
								 RGBColor*				inCurrentColorPtr,
								 const UInt8			inStartingPercentage,
								 const UInt8			inEndingPercentage,
								 const SInt8			inGammaChangePercentage,
								 const UInt8			inMinimumTicksBetweenGammaChanges,
								 const Boolean			inYieldToThreads)
{
	// this entire operation requires DrawSprocket
	if (FlagManager_Test(kFlagDrawSprocket))
	{
		register SInt16		i = 0;
		UInt32				ticks = TickCount();
		OSStatus			error = noErr;
		Boolean				wasError = false;
		
		
		for (i = inStartingPercentage; ; )
		{
			error = DSpContext_FadeGamma(inContext, FlagManager_Test(kFlagSuspended) ? 100 : i/* percent */,
											inCurrentColorPtr);
			if (error != noErr) wasError = true;
			if ((TickCount() - ticks) > inMinimumTicksBetweenGammaChanges)
			{
				ticks = TickCount();
				i += inGammaChangePercentage;
				if (inGammaChangePercentage >= 0) if (i >= inEndingPercentage) break;
				if (inGammaChangePercentage < 0) if (i < inEndingPercentage) break;
			}
			if (inYieldToThreads) Threads_YieldToAnother();
		}
		if (wasError) BEEP;
	}
}// GammaControlledFade


/*!
To conveniently execute a gamma fade from a
completely black screen to a completely normal
screen, use this method.

The Quad Damage effect can be really irritating
when interlaced with other fades, so it is
automatically and temporarily disabled while
performing a fade to black.

(3.0)
*/
void
CoolStuff_GammaControlledFadeFromBlack	(DSpContextReference	inContext,
										 const UInt8			inGammaIncreasePercentage,
										 const UInt8			inMinimumTicksBetweenGammaChanges,
										 const Boolean			inYieldToThreads)
{
	RGBColor		black;
	
	
	black.red = black.green = black.blue = 0;
	CoolStuff_GammaControlledFade(inContext, &black, 0, 100, inGammaIncreasePercentage,
									inMinimumTicksBetweenGammaChanges, inYieldToThreads);
}// GammaControlledFadeFromBlack


/*!
To conveniently execute a gamma fade from a
completely normal screen to a completely black
screen, use this method.

(3.0)
*/
void
CoolStuff_GammaControlledFadeToBlack	(DSpContextReference	inContext,
										 const UInt8			inGammaDecreasePercentage,
										 const UInt8			inMinimumTicksBetweenGammaChanges,
										 const Boolean			inYieldToThreads)
{
	RGBColor		black;
	SInt8			delta = inGammaDecreasePercentage;
	
	
	black.red = black.green = black.blue = 0;
	CoolStuff_GammaControlledFade(inContext, &black, 100, 0, -delta,
									inMinimumTicksBetweenGammaChanges, inYieldToThreads);
}// GammaControlledFadeToBlack


/*!
To conveniently execute a gamma fade from a
completely white screen to a completely normal
screen, use this method.

(3.0)
*/
void
CoolStuff_GammaControlledFadeFromWhite	(DSpContextReference	inContext,
										 const UInt8			inGammaIncreasePercentage,
										 const UInt8			inMinimumTicksBetweenGammaChanges,
										 const Boolean			inYieldToThreads)
{
	RGBColor		white;
	
	
	white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
	CoolStuff_GammaControlledFade(inContext, &white, 0, 100, inGammaIncreasePercentage,
									inMinimumTicksBetweenGammaChanges, inYieldToThreads);
}// GammaControlledFadeFromWhite


/*!
To conveniently execute a gamma fade from a
completely normal screen to a completely white
screen, use this method.

(3.0)
*/
void
CoolStuff_GammaControlledFadeToWhite	(DSpContextReference	inContext,
										 const UInt8			inGammaDecreasePercentage,
										 const UInt8			inMinimumTicksBetweenGammaChanges,
										 const Boolean			inYieldToThreads)
{
	RGBColor		white;
	SInt8			delta = inGammaDecreasePercentage;
	
	
	white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
	CoolStuff_GammaControlledFade(inContext, &white, 100, 0, -delta,
									inMinimumTicksBetweenGammaChanges, inYieldToThreads);
}// GammaControlledFadeToWhite


/*!
To conveniently reset the gamma for a screen
to normal (100% brightness, no color tint),
use this method.

(3.0)
*/
void
CoolStuff_GammaNormal	(DSpContextReference	inContext)
{
	RGBColor		black;
	
	
	black.red = black.green = black.blue = 0;
	CoolStuff_GammaSet(inContext, &black, 100);
}// GammaNormal


/*!
To conveniently set the gamma for a screen,
use this method.  The color specifies how the
screen will be tinted, and the percentage is
used to indicate brightness, where 100 is a
normal screen and 0 is complete darkness.

(3.0)
*/
void
CoolStuff_GammaSet		(DSpContextReference	inContext,
						 RGBColor*				inCurrentColorPtr,
						 const UInt8			inGammaPercentage)
{
	// this entire operation requires DrawSprocket
	if (FlagManager_Test(kFlagDrawSprocket))
	{
		OSStatus		error = noErr;
		
		
		error = DSpContext_FadeGamma(inContext, inGammaPercentage, inCurrentColorPtr);
		if (error != noErr) BEEP;
	}
}// GammaSet


/*!
To acquire a reference to a generic Draw Sprocket
context, use this method.  This context is useful
when you are only interested in simple effects,
such as gamma fading, and do not need fine control
over the attributes of a context.

You should not change the characteristics of the
default context!

(3.0)
*/
DSpContextReference
CoolStuff_GetDefaultDrawSprocketContext	(void)
{
	return gDefaultDrawingContext;
}// GetDefaultDrawSprocketContext


/*!
This routine creates a moving stars background
for the specified control, by assigning special
drawing and idle procedures to it.

The stars are set up in three sets; the first
is a static field (it never moves), and the
other two are groups of stars that move in 3D
space at two different speeds.  This is done
so that the screen always looks full; if a
single set of stars were animated, there would
always be a point in the animation where the
space background appears sparsely populated
with stars.

(3.0)
*/
void
CoolStuff_MovingStarsInit		(ControlRef		inUserPaneControl)
{
	StarDataPtr		dataPtr = (StarDataPtr)Memory_NewPtr(sizeof(StarData));
	Boolean			failed = false;
	
	
	if (dataPtr != null)
	{
		Rect				controlRect;
		register SInt16		i = 0;
		
		
		// determine processor speed (some of this cool stuff is pretty strenuous on the
		// old chips, and must be skipped on slower machines!)
		{
			long		gestaltResult = 0L;
			
			
			dataPtr->fasterThan300Megahertz = false;
			if (Gestalt(gestaltProcClkSpeed, &gestaltResult) == noErr)
			{
				dataPtr->fasterThan300Megahertz = (gestaltResult > 300000000L);
			}
		}
		
		// set up the initial layout data, and calculate a reasonable number of stars
		GetControlBounds(inUserPaneControl, &controlRect);
		dataPtr->screenDimensions.width = controlRect.right - controlRect.left;
		dataPtr->screenDimensions.height = controlRect.bottom - controlRect.top;
		dataPtr->center.x = INTEGER_HALVED(dataPtr->screenDimensions.width);
		dataPtr->center.y = INTEGER_HALVED(dataPtr->screenDimensions.height);
		dataPtr->center.y = 3 * INTEGER_HALVED(dataPtr->center.y); // center on lower quarter
		
		// calculate the number of stars dynamically, and create an array for as many
		// stars as are reasonable to render on the runtime system
		if (GetIconSuite(&dataPtr->stars.icons.large, kIconSuiteStarLarge,
							kSelectorAllSmallData) != noErr) dataPtr->stars.icons.large = null;
		if (GetIconSuite(&dataPtr->stars.icons.small, kIconSuiteStarSmall,
							kSelectorAllSmallData) != noErr) dataPtr->stars.icons.small = null;
		dataPtr->stars.diameter = (controlRect.right - controlRect.left) / 400; // rule of thumb
		dataPtr->stars.dynamic[0].count = INTEGER_MAXIMUM((controlRect.right - controlRect.left),
												(controlRect.bottom - controlRect.top)) + 350; // rule of thumb
		dataPtr->stars.dynamic[1].count = INTEGER_HALVED(dataPtr->stars.dynamic[0].count); // arbitrary
		dataPtr->stars.field.count = dataPtr->stars.dynamic[0].count / 32; // arbitrary
		dataPtr->stars.dynamic[0].positions = null;
		dataPtr->stars.dynamic[1].positions = null;
		dataPtr->stars.field.positions = null;
		
		// create dynamic star field
		for (i = 0; i < 2; i++)
		{
			while ((dataPtr->stars.dynamic[i].count > 0) && (dataPtr->stars.dynamic[i].positions == null))
			{
				dataPtr->stars.dynamic[i].positions = (StarPosition*)Memory_NewPtr
														(dataPtr->stars.dynamic[i].count * sizeof(StarPosition));
				if (dataPtr->stars.dynamic[i].positions == null)
				{
					dataPtr->stars.dynamic[i].count -= 40; // no memory?  Try fewer stars
				}
			}
		}
		
		// create static star field
		while ((dataPtr->stars.field.count > 0) && (dataPtr->stars.field.positions == null))
		{
			dataPtr->stars.field.positions = (StarPoint*)Memory_NewPtr(dataPtr->stars.field.count * sizeof(StarPoint));
			if (dataPtr->stars.field.positions == null)
			{
				dataPtr->stars.field.count -= 40; // no memory?  Try fewer stars
			}
		}
		
		// if stars could be created, set up the control (which draws them and changes their positions)
		if ((dataPtr->stars.field.positions != null) &&
			(dataPtr->stars.dynamic[0].positions != null) &&
			(dataPtr->stars.dynamic[1].positions != null))
		{
			if (inUserPaneControl != null)
			{
				// make the data available to the callbacks
				SetControlReference(inUserPaneControl, REINTERPRET_CAST(dataPtr, SInt32));
				{
					ControlUserPaneIdleUPP		idleUPP = NewControlUserPaneIdleUPP(movingStarsUserPaneIdle);
					ControlUserPaneDrawUPP		drawUPP = NewControlUserPaneDrawUPP(movingStarsUserPaneDraw);
					OSErr						error = noErr;
					
					
					error = SetControlData(inUserPaneControl, kControlNoPart, kControlUserPaneIdleProcTag,
											sizeof(ControlUserPaneIdleUPP), (Ptr)&idleUPP);
					error = SetControlData(inUserPaneControl, kControlNoPart, kControlUserPaneDrawProcTag,
											sizeof(ControlUserPaneDrawUPP), (Ptr)&drawUPP);
				}
				
				// initialize the star data structure
				{
					enum
					{
						kColorUnit = RGBCOLOR_INTENSITY_MAX / kStarsGrayRange
					};
					register SInt16		j = 0;
					
					
					// set the timer to zero (used to enforce animation delay)
					dataPtr->time = 0L;
					
					// set the speeds to the defaults (they can be changed later)
					dataPtr->speed = kStarsDefaultSpeed;
					if ((Random() % 2 + 1) == 1) dataPtr->speed = -dataPtr->speed; // reverse star direction!
					dataPtr->shift = kStarsDefaultShift;
					if ((Random() % 2 + 1) == 1) dataPtr->shift = -dataPtr->shift; // reverse star shift!
					
					// create a grayscale palette, with random shades of light blue
					for (i = 0; i < kStarsGrayRange; i++)
					{
						dataPtr->stars.grays[kStarsGrayRange - i - 1].red =
							dataPtr->stars.grays[kStarsGrayRange - i - 1].green =
							dataPtr->stars.grays[kStarsGrayRange - i - 1].blue = i * kColorUnit;
					}
					
					// initialize the 3D data for the dynamic stars
					SetQDGlobalsRandomSeed(TickCount());
					for (j = 0; j < 2; j++) for (i = 0; i < dataPtr->stars.dynamic[j].count; i++)
					{
						dataPtr->stars.dynamic[j].positions[i].x =
							(INTEGER_ABSOLUTE(Random()) %
								INTEGER_QUADRUPLED(dataPtr->screenDimensions.width) + 1) -
								INTEGER_DOUBLED(dataPtr->screenDimensions.width);
						dataPtr->stars.dynamic[j].positions[i].y =
							(INTEGER_ABSOLUTE(Random()) %
								INTEGER_QUADRUPLED(dataPtr->screenDimensions.height) + 1) -
								INTEGER_DOUBLED(dataPtr->screenDimensions.height);
						dataPtr->stars.dynamic[j].positions[i].z =
							INTEGER_ABSOLUTE(Random()) % kStarsBackPlane + 1;
					}
					
					// initialize the 2D data for the static, background star field
					SetQDGlobalsRandomSeed(TickCount());
					for (i = 0; i < dataPtr->stars.field.count; i++)
					{
						dataPtr->stars.field.positions[i].x =
							(INTEGER_ABSOLUTE(Random()) % dataPtr->screenDimensions.width + 1);
						dataPtr->stars.field.positions[i].y =
							(INTEGER_ABSOLUTE(Random()) % dataPtr->screenDimensions.height + 1);
						
						// make the background stars all sizes (within two pixels of the calculated diameter)
						dataPtr->stars.field.positions[i].diameter =
							(INTEGER_ABSOLUTE(Random()) % (dataPtr->stars.diameter + 2) + 1);
					}
					
					// start a thread to animate the thing
					FlagManager_Set(kFlagMovingStars, true);
					Threads_NewCoOp(NewThreadEntryUPP(Threads_MovingStars),
									inUserPaneControl/* parameter */, 0/* extra stack size*/,
									kTelnetThreadDescriptorMovingStarsHandling);
				}
			}
			else failed = true;
		}
		else failed = true;
	}
	else failed = true;
	
	if (failed) BEEP;
}// MovingStarsInit


/*!
This routine removes a moving stars background
from the specified window that was previously
assigned with CoolStuff_MovingStarsInit().

(3.0)
*/
void
CoolStuff_MovingStarsDone		(ControlRef		inUserPaneControl)
{
	FlagManager_Set(kFlagMovingStars, false);
	if (inUserPaneControl != null)
	{
		StarDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inUserPaneControl), StarDataPtr);
		
		
		// destroy the internal array, and the structure itself
		if (dataPtr != null)
		{
			Memory_DisposePtr((Ptr*)&dataPtr->stars.dynamic[0].positions);
			Memory_DisposePtr((Ptr*)&dataPtr->stars.dynamic[1].positions);
			Memory_DisposePtr((Ptr*)&dataPtr->stars.field.positions);
			Memory_DisposePtr((Ptr*)&dataPtr);
		}
	}
}// MovingStarsDone


/*!
To determine the center point of a field of stars,
relative to the origin of its controls windows
graphics port, use this method.  You might do this
if you want to make the center move relative to
its current position.

(3.0)
*/
void
CoolStuff_MovingStarsGetCenter		(ControlRef		inUserPaneControl,
									 Point*			inoutCenterPtr)
{
	if ((inUserPaneControl != null) && (inoutCenterPtr != null))
	{
		StarDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inUserPaneControl), StarDataPtr);
		
		
		// return the vanishing point
		if (dataPtr != null) SetPt(inoutCenterPtr, dataPtr->center.x, dataPtr->center.y);
	}
}// MovingStarsGetCenter


/*!
To determine how fast a field of stars is animating,
use this method.  The larger the number, the faster
the stars are moving.  The value indicates the number
of units per frame, in the Z direction, that the stars
move.

(3.0)
*/
SInt16
CoolStuff_MovingStarsGetSpeed		(ControlRef		inUserPaneControl)
{
	SInt16		result = 0;
	
	
	if (inUserPaneControl != null)
	{
		StarDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inUserPaneControl), StarDataPtr);
		
		
		// return the average speed
		if (dataPtr != null) result = dataPtr->speed;
	}
	
	return result;
}// MovingStarsGetSpeed


/*!
To change the center point of a field of stars,
relative to the origin of its controls windows
graphics port, use this method.  For example, to
make the stars follow the mouse, you could assign
the local mouse coordinates as the center.

(3.0)
*/
void
CoolStuff_MovingStarsSetCenter		(ControlRef		inUserPaneControl,
									 Point			inCenter)
{
	if (inUserPaneControl != null)
	{
		StarDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inUserPaneControl), StarDataPtr);
		
		
		// set the vanishing point
		if (dataPtr != null) dataPtr->center.x = inCenter.h, dataPtr->center.y = inCenter.v;
	}
}// MovingStarsSetCenter


/*!
To set the number of units per frame, in the Z direction,
that the stars move, use this method.  The speed is an
average, because some stars move faster than others, and
some dont move at all.

By reversing the sign of the speed, you can reverse the
direction of the stars animation.

(3.0)
*/
void
CoolStuff_MovingStarsSetSpeed		(ControlRef		inUserPaneControl,
									 SInt16			inAverageZPlaneDelta)
{
	if (inUserPaneControl != null)
	{
		StarDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inUserPaneControl), StarDataPtr);
		
		
		// set the average speed, and put a reasonable limit on it!
		if (dataPtr != null)
		{
			enum
			{
				kSpeedLimit = 80
			};
			
			
			if (inAverageZPlaneDelta >= 0)
			{
				dataPtr->speed = (inAverageZPlaneDelta < kSpeedLimit) ? inAverageZPlaneDelta : kSpeedLimit;
			}
			else
			{
				dataPtr->speed = (inAverageZPlaneDelta > -kSpeedLimit) ? inAverageZPlaneDelta : -kSpeedLimit;
			}
		}
	}
}// MovingStarsSetSpeed


/*!
To shift the default button of the specified
window forward by one cell in the pulse
animation, use this method.  Repeated calls
to this routine will make a button pulse like
Aqua buttons in Mac OS X do.

(3.0)
*/
void
CoolStuff_PulseDefaultButton		(WindowRef		inWindow)
{
	if (inWindow != null)
	{
		if ((GetWindowKind(inWindow) == kDialogWindowKind) && IsWindowVisible(inWindow))
		{
			enum
			{
				kMaximumPulseStages = 12		// number of pulse changes before is highlighting flag is reversed
			};
			static Boolean		gButtonPulseIsHighlighting = true;
			static SInt16		gButtonPulseStage = 0;
			ControlRef			control = null;
			DialogRef			dialog = GetDialogFromWindow(inWindow);
			
			
			Threads_DelayMinimumTicks(2); // enforce a delay to make the pulse reasonably smooth and visible
			if (GetDialogItemAsControl(dialog, GetDialogDefaultItem(dialog), &control) == noErr)
			{
				// draw the buttons highlighting intermixed with its previous
				// state; the exact mixture depends on the animation stage
				drawBlendedControlState(control, (gButtonPulseIsHighlighting) ? kControlButtonPart : kControlNoPart,
										100 / kMaximumPulseStages/* percentage per change in state */);
			}
			
			// change the state of the pulsing; if the pulse completes
			// going one way (for example, if the button has pulsed and
			// is now completely highlighted), reverse the pulse so it
			// gradually fades toward the opposite button state
			if (gButtonPulseIsHighlighting)
			{
				if (gButtonPulseStage++ >= kMaximumPulseStages) gButtonPulseIsHighlighting = false;
			}
			else
			{
				if (gButtonPulseStage-- < 0) gButtonPulseIsHighlighting = true;
			}
		}
	}
}// PulseDefaultButton


/*!
Asynchronously plays a Quad Damage power
sound,  la Quake.  If this method has
been invoked several times since the last
activation of Quad Damage, then the
effect will automatically be turned off.

(3.0)
*/
void
CoolStuff_QuadDamageActivity	(void)
{
	static SInt32		oldTicks = 0L;
	
	
	if ((TickCount() - oldTicks) > 30) // prevent the sound from playing more than 2x a second
	{
		SInt16				oldResFile = CurResFile();
		static SInt16		count = 0;
		
		
		oldTicks = TickCount();
		ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
		
		count++;
		if (count > 28) // after more than 28 shots, turn Quad Damage off
		{
			CoolStuff_QuadDamageEnd();
			count = 0;
		}
		else if (count == 25)
		{
			// play the warning sound that means Quad Damage almost used up
			(OSErr)Sound_PlayAsyncStereo(EASTER_EGG_SOUND_RESOURCE_TYPE,
											kSoundResourceIDQuadDamageFading,
											null/* SoundInfoRef storage */);
		}
		else
		{
			// play the Quad Damage sound
			(OSErr)Sound_PlayAsyncStereo(EASTER_EGG_SOUND_RESOURCE_TYPE,
										kSoundResourceIDQuadDamageShot,
										null/* SoundInfoRef storage */);
		}
		UseResFile(oldResFile);
	}
}// QuadDamageActivity


/*!
Turns Quad Damage on!  This is from the popular
Quake games; in this case, for the next few
seconds, every key you type will play the ravaging
sound that is heard when you use a powered-up
weapon with Quad Damage in Quake.  While this
mode is active, a separate thread continuously
modifies the gamma setting of the main device so
it pulses with blue (like the Quake III Arena
Quad Damage effect, and not as cool, I suppose).

What programmers do when they have free time.

(3.0)
*/
void
CoolStuff_QuadDamageBegin	(void)
{
	SInt16		oldResFile = CurResFile();
	
	
	ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
	(OSErr)Sound_PlayAsyncStereo(EASTER_EGG_SOUND_RESOURCE_TYPE,
									kSoundResourceIDQuadDamageBeginning,
									null/* SoundInfoRef storage */);
	FlagManager_Set(kFlagQuadDamage, true);
	// set up a thread to handle idle processing
	Threads_NewCoOp(NewThreadEntryUPP(Threads_QuadDamage),
					null/* parameter */, 0/* extra stack size*/,
					kTelnetThreadDescriptorQuadDamagePulseHandling);
	UseResFile(oldResFile);
}// QuadDamageBegin


/*!
Turns Quad Damage off.  This sets a flag that
the running thread immediately detects, causing
the thread to terminate.

(3.0)
*/
void
CoolStuff_QuadDamageEnd		(void)
{
	FlagManager_Set(kFlagQuadDamage, false);
}// QuadDamageEnd


/*!
Returns the color used for the Quad Damage
gamma pulse (the monitor tint color).

(3.0)
*/
void
CoolStuff_QuadDamageGetPulseColor	(RGBColor*		outColorPtr)
{
	if (outColorPtr != null)
	{
		outColorPtr->red = 0;
		outColorPtr->green = 0;
		outColorPtr->blue = 9 * (RGBCOLOR_INTENSITY_MAX / 10);
	}
}// QuadDamageGetPulseColor


/*!
Turns Quad Damage on or off, the reverse of its
current state.

(3.0)
*/
void
CoolStuff_QuadDamageToggle	(void)
{
	if (!FlagManager_Test(kFlagQuadDamage)) CoolStuff_QuadDamageBegin();
	else CoolStuff_QuadDamageEnd();
}// QuadDamageToggle


//
// internal methods
//

/*!
To draw the specified control in such a way as
to combine the appearances of two different
control states, use this method.  This routine
is used to make buttons pulse; you typically
provide "kControlNoPart" for the normal state,
and a specific control part for the brightening
state.  The percentage specifies the blending:
for example, if you specify a percentage of 14,
you must call this routine about 7 times before
the control will blend completely from its
previous state to its new state, to add up to
100% blending.

(3.0)
*/
static void
drawBlendedControlState		(ControlRef			inControl,
							 ControlPartCode	inNewState,
							 const UInt16		inBrightnessBlendingPercentage)
{
#if TARGET_API_MAC_OS8
	GrafPtr			oldPort = null;
	
	
	GetPort(&oldPort);
	
	// clip and draw
	if (gPulseNewClipRgn != null)
	{
		GrafPtr					port = null;
		OSStatus				error = noErr;
		PulseDeviceLoopParams	data;
		DeviceLoopDrawingUPP	upp = null;
		
		
		// initialize globals
		SetEmptyRgn(gPulseNewClipRgn);
		
		// initialize data for the DeviceLoop callback
		data.control = inControl;
		data.newState = inNewState;
		data.percentage = inBrightnessBlendingPercentage;
		data.result = noErr;
		
		// first determine the exact area in which to draw (the intersection
		// of the control region and the graphics ports visible region)
		GetPort(&port);
		if (FlagManager_Test(kFlagCarbonControlManager))
		{
			// use this API, when available, to clip drawing to as precise an area as possible!
			error = GetControlRegion(inControl, kControlStructureMetaPart, gPulseNewClipRgn);
		}
		else error = cfragNoSymbolErr;
		
		if (error != noErr)
		{
			Rect		controlBounds;
			
			
			error = noErr;
			GetControlBounds(inControl, &controlBounds);
			SetRectRgn(gPulseNewClipRgn, controlBounds.left, controlBounds.top, controlBounds.right, controlBounds.bottom);
		}
		
		// call DeviceLoop() to make sure the drawing looks right no matter how many monitors it may cross
		upp = NewDeviceLoopDrawingUPP(offscreenDrawBlendedDeviceLoop);
		DeviceLoopClipAndDraw(gPulseNewClipRgn, upp, (long)&data, 0L/* flags */);
		DisposeDeviceLoopDrawingUPP(upp), upp = null;
	}
	
	SetPort(oldPort);
#endif
}// drawBlendedControlState


/*!
This method, of standard ControlUserPaneDrawUPP
form, draws moving stars in a 3D pattern.  The
control reference of the given control must be
of type "StarDataPtr".

IMPORTANT:	This function may seem messy, but
			work was done to make stars draw
			fairly quickly (because delays will
			appear to stall the animation).  For
			example, some code is basically
			copied and pasted instead of
			appearing in a function, so watch
			for this kind of code if you want to
			change how something is done here.
			I hope that the brief comments are
			sufficient to indicate what I am
			trying to do in each step.

(3.0)
*/
static pascal void
movingStarsUserPaneDraw		(ControlRef		inControl,
							 SInt16			UNUSED_ARGUMENT(inPartCode))
{
	StarDataPtr			dataPtr = null;
	Boolean				followMouse = EventLoop_IsOptionKeyDown(),	// Easter egg!
						reverse = EventLoop_IsShiftKeyDown();		// Easter egg!
	
	
	dataPtr = REINTERPRET_CAST(GetControlReference(inControl), StarDataPtr);
	if (dataPtr != null)
	{
		Rect				rect;
		register SInt16		i = 0;
		SInt16				x = 0,
							y = 0,
							z = 0;
		SInt16				diameter = 0;
		SInt16				grayIndex = 0;
		RGBColor			starColor;
		StarPosition*		starArray = null;
		SInt16				setNumber = 0;
		SInt16				speed = 0;
		SInt16				count = 0;
		
		
		// paint the background and set up the text color
		RGBBackColor(&dataPtr->stars.grays[kStarsGrayRange - 1]);
		GetControlBounds(inControl, &rect);
		EraseRect(&rect);
		
		// draw the stars
		if (followMouse)
		{
			Point		localMouse;
			
			
			// adjust the vanishing point to be the mouse coordinates
			GetMouse(&localMouse);
			dataPtr->center.x = localMouse.h;
			dataPtr->center.y = localMouse.v;
		}
		
		if (reverse)
		{
			// make the stars move the other way
			dataPtr->speed = -dataPtr->speed;
			dataPtr->shift = -dataPtr->shift;
		}
		
		// draw a static field of stars behind the moving stars
		// (theres a lot of copy & paste code here, for the
		// sake of maximizing efficiency (minimizing execution
		// of statements unless absolutely necessary)
		for (i = 0; i < dataPtr->stars.field.count; i++)
		{
			// each static star has a pre-defined diameter (to avoid calculating random ones dynamically)
			diameter = dataPtr->stars.field.positions[i].diameter;
			
			// draw the star, using a gray shade within a few indices of a central gray
			x = dataPtr->stars.field.positions[i].x;
			y = dataPtr->stars.field.positions[i].y;
			if (diameter > 2)
			{
				// draw large stars using the large star icon; at random, use
				// a simple oval instead (helps to improve drawing speeds!)
				#if 1
				{
					if (dataPtr->stars.icons.large != null)
					{
						SetRect(&rect, x - 8, y - 8, x + 7, y + 7);
						(OSErr)PlotIconSuite(&rect, kAlignAbsoluteCenter, kTransformNone,
												dataPtr->stars.icons.large);
					}
				}
				#else
				{
					// dont use this code unless you want increased performance and lousy-looking stars
					grayIndex = INTEGER_HALVED(kStarsGrayRange - 1);
					grayIndex += (Random()) % 16 + 1;
					starColor = dataPtr->stars.grays[grayIndex];
					RGBForeColor(&starColor);
					SetRect(&rect, x - INTEGER_HALVED(diameter), y - INTEGER_HALVED(diameter), 0, 0);
					SetRect(&rect, rect.left, rect.top, rect.left + diameter, rect.top + diameter);
					PaintOval(&rect);
				}
				#endif
			}
			else if (diameter == 2)
			{
				SInt16		random = (INTEGER_ABSOLUTE(Random()) % 6);
				
				
				// draw smaller stars using the small star icon; at random,
				// use a simple oval instead (helps to improve drawing speeds!)
				//if (random > 3)
				{
					if (dataPtr->stars.icons.small != null)
					{
						SetRect(&rect, x - 8, y - 8, x + 7, y + 7);
						(OSErr)PlotIconSuite(&rect, kAlignAbsoluteCenter, kTransformNone,
												dataPtr->stars.icons.small);
					}
				}
				#if 0
				else
				{
					// dont use this code unless you want increased performance and lousy-looking stars
					grayIndex = INTEGER_HALVED(kStarsGrayRange - 1);
					grayIndex += (Random()) % 16 + 1;
					starColor = dataPtr->stars.grays[grayIndex];
					RGBForeColor(&starColor);
					SetRect(&rect, x - INTEGER_HALVED(diameter), y - INTEGER_HALVED(diameter), 0, 0);
					SetRect(&rect, rect.left, rect.top, rect.left + diameter, rect.top + diameter);
					PaintOval(&rect);
				}
				#endif
			}
			else
			{
				// ovals seem like overkill (but would still work) for one-pixel stars;
				// instead, just draw them as a single point
				grayIndex = INTEGER_HALVED(kStarsGrayRange - 1);
				grayIndex += (Random()) % 16 + 1;
				starColor = dataPtr->stars.grays[grayIndex];
				RGBForeColor(&starColor);
				MoveTo(x - INTEGER_HALVED(diameter), y - INTEGER_HALVED(diameter));
				Line(0, 3);
			}
			
			if (dataPtr->fasterThan300Megahertz)
			{
				// on really fast machines, it is reasonable to shift the
				// static star (stars in this field have no 3D properties);
				// the extra math would probably cripple a low-end machine, though,
				// so this is only done for faster processors
				//dataPtr->stars.field.positions[i].x += dataPtr->shift;
				dataPtr->stars.field.positions[i].y++;
				if (dataPtr->stars.field.positions[i].y < 0)
				{
					dataPtr->stars.field.positions[i].y += dataPtr->screenDimensions.height;
				}
				else if (dataPtr->stars.field.positions[i].y > dataPtr->screenDimensions.height)
				{
					dataPtr->stars.field.positions[i].y -= dataPtr->screenDimensions.height;
				}
				if (dataPtr->stars.field.positions[i].x < 0)
				{
					dataPtr->stars.field.positions[i].x += dataPtr->screenDimensions.width;
				}
				else if (dataPtr->stars.field.positions[i].x > dataPtr->screenDimensions.width)
				{
					dataPtr->stars.field.positions[i].x -= dataPtr->screenDimensions.width;
				}
			}
		}
		
		// Now render the stars at their positions in 3D space, first offsetting
		// the Z-plane (to move them).  One set of stars moves at normal speed,
		// and the other at an arbitrarily-derived, different speed.  This gives
		// the stars a much more realistic 3D movement effect.
		diameter = dataPtr->stars.diameter;
		for (setNumber = 0; setNumber < 2; setNumber++)
		{
			if (!setNumber) speed = dataPtr->speed;
			else speed = INTEGER_EIGHTHED(dataPtr->speed); // arbitrary; make some stars slower
			
			starArray = dataPtr->stars.dynamic[setNumber].positions;
			count = dataPtr->stars.dynamic[setNumber].count;
			
			for (i = 0; i < count; i++)
			{
				// offset the backplane, effectively repositioning the star
				starArray[i].z -= speed;
				if (starArray[i].z < VIEW_PLANE) starArray[i].z = kStarsBackPlane;
				if (starArray[i].z > kStarsBackPlane) starArray[i].z = VIEW_PLANE;
				
				// apply perspective projection and draw this star in an appropriate shade of gray
				diameter = dataPtr->stars.diameter;
				z = starArray[i].z;
				
				// make nearer stars slightly bigger
				if (z < INTEGER_QUARTERED(kStarsBackPlane)) diameter++;
				
				y = VIEW_PLANE_DIVIDE(starArray[i].y) / z + dataPtr->center.y;
				x = VIEW_PLANE_DIVIDE(starArray[i].x) / z + dataPtr->center.x;
				RGBForeColor(&dataPtr->stars.grays[z >> kStarsBitDifference]);
				if (diameter > 1)
				{
					// large stars are easily drawn as ovals
					SetRect(&rect, x - INTEGER_HALVED(diameter), y - INTEGER_HALVED(diameter), 0, 0);
					SetRect(&rect, rect.left, rect.top, rect.left + diameter, rect.top + diameter);
					PaintOval(&rect);
				}
				else
				{
					// ovals seem like overkill (but would still work) for one-pixel stars
					MoveTo(x - INTEGER_HALVED(diameter), y - INTEGER_HALVED(diameter));
					Line(1, 0);
				}
			}
		}
	}
}// movingStarsUserPaneDraw


/*!
A standard ControlUserPaneIdleUPP, this
method manages state changes for the
moving stars.

The reference constant of the control is
assumed to be of type "StarDataPtr".
That structure is modified as necessary
to reflect any change of state.

(3.0)
*/
static pascal void
movingStarsUserPaneIdle		(ControlRef		inControl)
{
	StarDataPtr		dataPtr = REINTERPRET_CAST(GetControlReference(inControl), StarDataPtr);
	
	
	if (dataPtr != null)
	{
		Boolean		stateChanged = false;
		
		
		// this code enforces the animation delay
		if ((TickCount() - dataPtr->time) > kStarsAnimationDelayInTicks)
		{
			stateChanged = true;
			dataPtr->time = TickCount();
		}
		
		// redraw the control if it changed
		if (stateChanged) DrawOneControl(inControl);
	}
}// movingStarsUserPaneIdle


/*!
This method, of standard DeviceLoopDrawingUPP form,
handles offscreen drawing of blended control states
correctly, no matter what the device characteristics
are.

(3.0)
*/
static pascal void
offscreenDrawBlendedDeviceLoop		(short			inColorDepth,
									 short			UNUSED_ARGUMENT(inDeviceFlags),
									 GDHandle		inTargetDevice,
									 long			inDataPtr)
{
#if TARGET_API_MAC_OS8
	PulseDeviceLoopParamsPtr	dataPtr = (PulseDeviceLoopParamsPtr)inDataPtr;
	GrafPtr						oldPort = null;
	
	
	GetPort(&oldPort);
	if (dataPtr != null)
	{	
		Rect			worldRect,
						controlRectInWindow,
						maskRectInWorld;
		GWorldPtr		world = null;
		GDHandle		oldDevice = null;
		CGrafPtr		colorWindowPort = GetWindowPort(GetControlOwner(dataPtr->control)),
						oldColorPort = null;
		UInt16			controlWidth = 0,
						controlHeight = 0;
		
		
		GetPortBounds(colorWindowPort, &worldRect);
		GetControlBounds(dataPtr->control, &controlRectInWindow);
		
		controlWidth = controlRectInWindow.right - controlRectInWindow.left,
		controlHeight = controlRectInWindow.bottom - controlRectInWindow.top;
		
		// set the mask region
		SetRect(&maskRectInWorld, 0, 0, controlWidth - 1, controlHeight - 1);
		
		// create a graphics world of adequate size
		dataPtr->result = NewGWorld(&world, inColorDepth, &worldRect, (CTabHandle)null,
									(GDHandle)inTargetDevice, 0L/* flags */);
		if (dataPtr->result == noErr)
		{
			PixMapHandle	pixels = null;
			
			
			// set port to the new port
			GetGWorld(&oldColorPort, &oldDevice);
			SetGWorld(world, (GDHandle)null);
			
			// get the offscreen graphics worlds pixel map
			pixels = GetGWorldPixMap(world);
			
			// perform the offscreen control operation, and then do the blitter as quickly and easily as possible
			if (LockPixels(pixels))
			{
				RGBColor		solid;
				
				
				// initialize the offscreen world
				(OSErr)SetUpControlBackground(dataPtr->control, inColorDepth,
												ColorUtilities_IsColorDevice(inTargetDevice));
				EraseRect(&worldRect);
				
				// create the mask, for blending; by drawing gray pixels that take into account the
				// blending percentage, a mask that properly combines the two images can be generated
				solid.red = solid.blue = solid.green = ((RGBCOLOR_INTENSITY_MAX / 100) * dataPtr->percentage);
				RGBForeColor(&solid);
				PaintRect(&maskRectInWorld);
				
				// now draw the new state of the control in the buffer, so it can be blended with the onscreen
				// image; HiliteControl() is avoided, because it automatically redraws the control in its window
				(**(dataPtr->control)).contrlHilite = IsControlActive(dataPtr->control)
														? dataPtr->newState : kControlInactivePart;
				DrawControlInCurrentPort(dataPtr->control);
				
				// set background to white and foreground to black, so CopyDeepMask() doesnt screw anything up
				solid.red = solid.blue = solid.green = 0;
				RGBForeColor(&solid);
				solid.red = solid.blue = solid.green = RGBCOLOR_INTENSITY_MAX;
				RGBBackColor(&solid);
				
				// now, copy the blended version to the window so the user can see it
				{
					RGBColor		oldForeColor,
									oldBackColor;
					//Rect			pixMapBounds;
					
					
					// change to the window port, and save its colors
					SetGWorld(colorWindowPort, null/* device */);
					GetForeColor(&oldForeColor);
					GetBackColor(&oldBackColor);
					
					// set background to white and foreground to black, so CopyBits() doesnt screw anything up
					solid.red = solid.blue = solid.green = 0;
					RGBForeColor(&solid);
					solid.red = solid.blue = solid.green = RGBCOLOR_INTENSITY_MAX;
					RGBBackColor(&solid);
					
					// check for sufficient stack space before using QuickDraw;
					// bit-copy routines typically require extensive stack space,
					// depending on the size of the source pixel map being used
					// (see IM:IWQ page 3-118 for the reasoning behind the stack
					// space checking)
					//GetPixBounds(pixels, &pixMapBounds);
					//if (Threads_GetStackOf(Threads_GetCurrent()) > (5 * ((1 << inColorDepth) *
					//												(pixMapBounds.right - pixMapBounds.left))))
					{
						#if 1
						// debugging - show the damned mask
						CopyBits(GetPortBitMapForCopyBits(world)/* source pixel map */,
									GetPortBitMapForCopyBits(colorWindowPort)/* destination pixel map */,
									&maskRectInWorld/* source rectangle */,
									&controlRectInWindow/* destination rectangle */,
									srcCopy/* drawing mode */, (RgnHandle)null/* mask */);
						dataPtr->result = QDError();
						#else
						// blend the desired control state with whatever image currently represents the control
						CopyDeepMask(GetPortBitMapForCopyBits(world)/* source pixel map */,
									GetPortBitMapForCopyBits(world)/* the gray pixels used to cause the blending */,
									GetPortBitMapForCopyBits(colorWindowPort)/* destination pixel map */,
									&controlRectInWindow/* source rectangle */,
									&maskRectInWorld/* mask rectangle */,
									&controlRectInWindow/* destination rectangle */,
									srcCopy/* drawing mode */, (RgnHandle)null/* mask */);
						dataPtr->result = QDError();
						#endif
					}
					//else dataPtr->result = memFullErr;
					
					// restore colors
					RGBForeColor(&oldForeColor);
					RGBBackColor(&oldBackColor);
				}
				
				// the pixel map is no longer being directly used
				UnlockPixels(pixels);
			}
			else dataPtr->result = memPCErr;
			
			// restore to saved
			SetGWorld(oldColorPort, oldDevice);
			if (world != null) DisposeGWorld(world), world = null;
		}
	}
	
	SetPort(oldPort);
#endif
}// offscreenDrawBlendedDeviceLoop