/*###############################################################

	Cursors.c
	
	Interface Library 1.0
	 1998-2000 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#ifndef __GESTALT__
#	include <Gestalt.h>
#endif
#ifndef __QUICKDRAW__
#	include <QuickDraw.h>
#endif

// library includes
#include "ContextsLib.h"
#include "InterfaceLib.h"



// types
struct CursorLink
{
	CursHandle				bwCursorHandle;
	CCrsrHandle				colorCursorHandle;
	SInt16					cursorID;
	struct CursorLink*		next;
};
typedef struct CursorLink CursorLink;

// variables
static CursorLink*		dynamicCursorListPtr = null;
static SInt16			currentCursor = kCursor_Arrow,
						deferredCursor = kCursor_Watch;
static UInt32			cursorStageCount = 0L;
static UInt32			countAtLastUpdate = 0L,
						countAtDeferral = 0L,
						deferralTicks = 0L;
static Boolean			haveAppearanceCursors = false,
						appearanceCursorInEffect = false,
						animatedCursorInEffect = false;

// internal methods
static void		setCurrentCursorID	(const SInt16		inNewCursor);



//
// public methods
//

/**
Use this method to set up this module.
You must call this routine before using
any other routines in this module, to
be sure that the module will work.

(1.0)
*/
void
Cursors_Init	(void)
{
	//UInt32		appearanceFlags = 0L;
	//UInt32		appearanceVersion = 0L;
	SInt32		sysv = 0L;
	OSErr		error = noErr;
	
	
	//error = Gestalt(gestaltAppearanceAttr, &appearanceFlags);
	//error = Gestalt(gestaltAppearanceVersion, &appearanceVersion);
	//majorRev = Releases_GetMajorRevisionForVersion(appearanceVersion);
	//minorRev = Releases_GetMinorRevisionForVersion(appearanceVersion);
	
	// Appearance 1.1 or later supports theme cursors
	//haveAppearanceCursors = ((appearanceFlags & (1 << gestaltAppearanceExists)) &&
	//							(((majorRev == 0x01) && (minorRev >= 0x01)) || (majorRev > 0x01)));
	
	// problems with Appearance code - to hell with it, just look for OS 8.5 or later
	error = Gestalt(gestaltSystemVersion, &sysv);
	if (error == noErr)
	{
		UInt8		majorRev = Releases_GetMajorRevisionForVersion(sysv),
					minorRev = Releases_GetMinorRevisionForVersion(sysv);
		
		
		// Mac OS 8.5?
		haveAppearanceCursors = (((majorRev == 0x08) && (minorRev >= 0x05)) || (majorRev > 0x08));
	}
}// Init


/**
Use this method when you are done using
cursors to dispose of any memory that
was allocated.

(1.0)
*/
void
Cursors_Done	(void)
{
	// make sure none of the to-be-killed cursors is in use!
	InitCursor();
	
	// destroy the linked list and all cursors
	if (dynamicCursorListPtr != null)
	{
		CursorLink*		ptr = dynamicCursorListPtr;
		Ptr				memoryManagerPtr = null;
		
		
		while (ptr != null)
		{
			memoryManagerPtr = (Ptr)ptr;
			DisposeCCursor(ptr->colorCursorHandle);
			ptr = ptr->next;
			DisposePtr(memoryManagerPtr), memoryManagerPtr = null;
		}
	}
}// Done


/**
Use this method to determine if the installed
version of the Appearance Manager supports
theme cursors.

(1.0)
*/
Boolean
Cursors_AppearanceCursorsAvail	(void)
{
	return haveAppearanceCursors;
}// AppearanceCursorsAvail


/**
To conditionally change the cursor to a
watch, call this method.  In order for
the deferral to work, your application
must be calling Cursors_Idle() quite
regularly (ideally in its own thread).

This method is useful if you don't want
the cursor to flicker on fast machines,
but you are performing an operation
that is potentially slow on certain
computers.  On the faster machines, it
is likely the deferral time will not
elapse so the cursor will not change;
on slower machines, the operation may
take enough time that it is prudent to
display the watch cursor.

As soon as you force the cursor to a
particular value, any cursor deferral
is automatically cancelled.  Thus, you
generally call Cursors_DeferredWatch(),
perform a lengthy operation (with
sufficiently periodic calls to the
Cursors_Idle() method), and finish by
calling another routine, such as
Cursors_UseArrow(), that forces the
cursor to a particular shape and
cancels the deferred watch cursor.

(1.0)
*/
SInt16
Cursors_DeferredUseWatch	(const UInt32		inAfterHowManyTicks)
{
	countAtDeferral = TickCount();
	deferralTicks = inAfterHowManyTicks;
	deferredCursor = kCursor_Watch;
	return Cursors_GetCurrent();
}// DeferredUseWatch


/**
To determine the ID of the cursor that was
most recently set using this module, use this
method.  If you consistently use this module
to change the cursor *everywhere* in your
program, then you can use this method to
effectively get the current cursor, which
has many advantages.

(1.0)
*/
SInt16
Cursors_GetCurrent	(void)
{
	return currentCursor;
}// GetCurrent


/**
Call this method once in your main event loop
to make sure that animated cursors animate.

(1.0)
*/
void
Cursors_Idle	(void)
{
	UInt32		currentTicks = TickCount();
	
	
	// any deferred cursor?
	if (deferralTicks > 0L) if ((currentTicks - countAtDeferral) > deferralTicks) Cursors_UseWatch();
	
	// animate any theme cursor that can be animated, but no more
	// than every so many ticks (to keep animation smooth)
	if (animatedCursorInEffect)
	{
		if (((currentTicks - countAtLastUpdate) > 8) && (appearanceCursorInEffect))
		{
			// then it's time to update the cursor
			(OSStatus)SetAnimatedThemeCursor(Cursors_GetCurrent(), ++cursorStageCount);
			countAtLastUpdate = TickCount();
		}
	}
}// Idle


/**
Use this method to use a cursor with a particular
resource ID.  This routine will first check for a
theme cursor (Mac OS 8.5 or later).  If it cannot
find a theme cursor appropriate for the request or
if Appearance 1.1 or later is not available, this
method then looks for a color cursor ('crsr'
resource) with the specified ID.  If no color
version is available, it looks for a monochrome
('CURS' resource) cursor.  If no appropriate cursor
can be found, the cursor becomes an arrow.

The ID of the cursor most recently specified using
this library is returned.  It is highly recommended
that you use this library for *every* cursor change
you make, so you can keep track of cursors easily.
Regardless of whether a cursor can be found or not,
this method always returns the cursor that was
*supposed* to be in place prior to the change -
that way, your code can reliably determine what
cursor is to be in effect.

To support animated theme cursors, you should call
the method Cursors_Idle() once per event loop, in
modal dialog box event filters, and continuously
throughout a series of procedures that might take
a non-trivial amount of time on a computer that
your application supports.

(1.0)
*/
SInt16
Cursors_Use	(const SInt16		inCursorID)
{
	Boolean		tryToGetResource = true;
	SInt16		result = Cursors_GetCurrent();
	
	
	// cancel any deferred cursor
	countAtDeferral = deferralTicks = 0L;
	
	if (result != inCursorID)
	{
		// initially assume no Appearance cursors are available
		appearanceCursorInEffect = animatedCursorInEffect = false;
		setCurrentCursorID(inCursorID);
		
		if (Cursors_AppearanceCursorsAvail())
		{
			ThemeCursor		cursor = kThemeArrowCursor;
			
			
			// determine the appropriate Appearance Manager
			// analog cursor for the specified resource ID,
			// and use it
			tryToGetResource = false;
			
			switch (inCursorID)
			{
				case kCursor_Arrow:
					cursor = kThemeArrowCursor;
					appearanceCursorInEffect = true;
					break;
				
				case kCursor_ArrowCopy:
					cursor = kThemeCopyArrowCursor;
					appearanceCursorInEffect = true;
					break;
				
				case kCursor_ArrowContextualMenu:
					cursor = kThemeContextualMenuArrowCursor;
					appearanceCursorInEffect = true;
					break;
				
				case kCursor_Crosshairs:
					cursor = kThemeCrossCursor;
					appearanceCursorInEffect = true;
					break;
				
				case kCursor_HandGrab:
					cursor = kThemeClosedHandCursor;
					appearanceCursorInEffect = true;
					break;
				
				case kCursor_HandOpen:
					cursor = kThemeOpenHandCursor;
					appearanceCursorInEffect = true;
					break;
				
				case kCursor_HandPoint:
					cursor = kThemePointingHandCursor;
					appearanceCursorInEffect = true;
					break;
				
				case kCursor_IBeam:
					cursor = kThemeIBeamCursor;
					appearanceCursorInEffect = true;
					break;
				
				case kCursor_Plus:
					cursor = kThemePlusCursor;
					appearanceCursorInEffect = true;
					break;
				
				case kCursor_Watch:
					cursor = kThemeWatchCursor;
					appearanceCursorInEffect = animatedCursorInEffect = true;
					break;
				
				case kCursor_ArrowHelp:
				case kCursor_ArrowWatch:
				case kCursor_TerminalCursor:
					cursor = kThemeArrowCursor;
					appearanceCursorInEffect = true;
					break;
				
				default:
					tryToGetResource = true;
					break;
			}
			
			if (appearanceCursorInEffect) SetThemeCursor(cursor);
			if (animatedCursorInEffect) cursorStageCount = 0L;
		}
		
		if (tryToGetResource)
		{
			CursorLink*		ptr = dynamicCursorListPtr;
			Boolean			createNew = true;
			
			
			if (dynamicCursorListPtr != null)
			{
				ptr = dynamicCursorListPtr;
				if (ptr != null) do
				{
					if (ptr->cursorID == inCursorID)
						createNew = false;
					else if (ptr->next != null)
						ptr = ptr->next;
				} while ((ptr->next != null) && (createNew));
				if (createNew)
				{
					ptr->next = (CursorLink*)Memory_NewPtr(sizeof(CursorLink));
					ptr = ptr->next;
				}
			}
			else
			{
				ptr = dynamicCursorListPtr = (CursorLink*)Memory_NewPtr(sizeof(CursorLink));
			}
			
			if (ptr != null)
			{
				ptr->next->colorCursorHandle = GetCCursor(inCursorID);
				if (ptr->next->colorCursorHandle == null)
				{
					ptr->next->bwCursorHandle = GetCursor(inCursorID);
					if (ptr->next->bwCursorHandle == null)
					{
						InitCursor(); // set to an arrow
					}
					else
					{
						SInt8		state = HGetState((Handle)ptr->next->bwCursorHandle);
						
						
						HLockHi((Handle)ptr->next->bwCursorHandle);
						SetCursor(*(ptr->next->bwCursorHandle)); // use monochrome cursor
						HSetState((Handle)ptr->next->bwCursorHandle, state);
					}
				}
				else SetCCursor(ptr->next->colorCursorHandle); // use color cursor
			}
			else
			{
				InitCursor(); // set to an arrow
			}
		}
	}
	
	return result;
}// Use


/**
Use this convenience method to set the cursor
to the standard arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseArrow	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_Arrow);
	return result;
}// UseArrow


/**
Use this convenience method to set the cursor
to an arrow with a miniature contextual menu
next to it (to signify that a menu will appear).
If there is a color cursor available, it is
automatically chosen instead of any monochrome
version.  If neither a color nor black-and-white
cursor of the proper ID can be found, the cursor
is set to a plain arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseArrowContextualMenu	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_ArrowContextualMenu);
	return result;
}// UseArrowContextualMenu


/**
Use this convenience method to set the cursor
to an arrow with a miniature plus sign next to
it (to signify that copying will occur).  If
there is a color cursor available, it is
automatically chosen instead of any monochrome
version.  If neither a color nor black-and-white
cursor of the proper ID can be found, the cursor
is set to a plain arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseArrowCopy	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_ArrowCopy);
	return result;
}// UseArrowCopy


/**
Use this convenience method to set the cursor
to an arrow with a miniature question-mark
icon next to it (for Balloon Help).  If there
is a color cursor available, it is automatically
chosen instead of any monochrome version.  If
neither a color nor black-and-white cursor of
the proper ID can be found, the cursor is set
to a plain arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseArrowHelp	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_ArrowHelp);
	return result;
}// UseArrowHelp


/**
Use this convenience method to set the cursor
to an arrow with a miniature watch next to it
(to signify that another thread is in progress).
If there is a color cursor available, it is
automatically chosen instead of any monochrome
version.  If neither a color nor black-and-white
cursor of the proper ID can be found, the cursor
is set to a plain arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseArrowWatch	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_ArrowWatch);
	return result;
}// UseArrowWatch


/**
Use this convenience method to set the cursor
to crosshairs (used for drawing).  If there is
a color cursor available, it is automatically
chosen instead of any monochrome version.  If
neither a color nor black-and-white cursor of
the proper ID can be found, the cursor is set
to an arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseCrosshairs	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_Crosshairs);
	return result;
}// UseCrosshairs


/**
Use this convenience method to set the cursor
to a closed hand (for document-moving operations
that are in dragging progress).  If there is a
color cursor available, it is automatically chosen
instead of any monochrome version.  If neither a
color nor black-and-white cursor of the proper ID
can be found, the cursor is set to an arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseHandGrab	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_HandGrab);
	return result;
}// UseHandGrab


/**
Use this convenience method to set the cursor
to an open hand (for document-moving operations).
If there is a color cursor available, it is
automatically chosen instead of any monochrome
version.  If neither a color nor black-and-white
cursor of the proper ID can be found, the cursor
is set to an arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseHandOpen	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_HandOpen);
	return result;
}// UseHandOpen


/**
Use this convenience method to set the cursor
to a pointing hand (for miscellaneous operations,
but often for mouse-overs of "hot text" areas).
If there is a color cursor available, it is
automatically chosen instead of any monochrome
version.  If neither a color nor black-and-white
cursor of the proper ID can be found, the cursor
is set to an arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseHandPoint (void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_HandPoint);
	return result;
}// UseHandPoint


/**
Use this convenience method to set the cursor
to a magnifying glass (for browsing).  If there
is a color cursor available, it is automatically
chosen instead of any monochrome version.  If
neither a color nor black-and-white cursor of
the proper ID can be found, the cursor is set
to an arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

Unimplemented.

(1.0)
*/
SInt16
Cursors_UseMagnifyingGlass	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_UseArrow();
	return result;
}// UseMagnifyingGlass


/**
Use this convenience method to set the cursor
to an I-beam (for text editing).  If there is
a color cursor available, it is automatically
chosen instead of any monochrome version.  If
neither a color nor black-and-white cursor of
the proper ID can be found, the cursor is set
to an arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseIBeam	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_IBeam);
	return result;
}// UseIBeam


/**
Use this convenience method to set the cursor
to a plus (for spreadsheets).  If there is
a color cursor available, it is automatically
chosen instead of any monochrome version.  If
neither a color nor black-and-white cursor of
the proper ID can be found, the cursor is set
to an arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UsePlus		(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_Plus);
	return result;
}// UsePlus


/**
Use this convenience method to set the cursor
to a watch (for time-consuming operations).
If there is a color cursor available, it is
automatically chosen instead of any monochrome
version.  If neither a color nor black-and-white
cursor of the proper ID can be found, the cursor
is set to an arrow.

The ID of the cursor most recently specified
using this library is returned.  It is highly
recommended that you use this library for
*every* cursor change you make, so you can
keep track of cursors easily.

(1.0)
*/
SInt16
Cursors_UseWatch	(void)
{
	SInt16		result = Cursors_GetCurrent();
	
	
	result = Cursors_Use(kCursor_Watch);
	return result;
}// UseWatch


//
// internal methods
//

/**
To specify the ID of the current
cursor, use this method.  This helps
to keep track of what the shape of
the mouse pointer most likely is
(unless it was changed by some other
process).  Use Cursors_GetCurrent()
to determine this value later.

(1.0)
*/
static void
setCurrentCursorID		(const SInt16		inNewCursor)
{
	currentCursor = inNewCursor;
}// setCurrentCursorID