/*###############################################################

	DNR.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstring>
#else
#	include <string.h>
#endif

// Mac includes
#include <MacErrors.h>
#include <OpenTransport.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// MacTelnet includes
#include "Connections.h"
#include "Console.h"
#include "DNR.h"
#include "InternalEvents.h"
#include "InternetServices.h"
#include "NetEvents.h"
#include "ProxyServer.h"



// types
struct ProxyLookupData
{
	// used by the DNR completion routine when looking up
	// new connections via proxy servers
	TelnetScreenID		screenID;			// screen ID of new connection
	ProxyServerRef		proxyServer;		// proxy server information
	InetDomainName		targetDomainName;	// the target host (looked up via proxy server)
};
#ifndef __cplusplus
typedef struct ProxyLookupData		ProxyLookupData;
#endif
typedef ProxyLookupData*			ProxyLookupDataPtr;

// internal methods
static void			beginLookupViaProxyServerConnection		(SInt32 const					inData,
															 DomainNameLookupDataPtr		inResolverPtr);
static void			openNewResolvedConnection				(SInt32 const					inData,
															 DomainNameLookupDataPtr		inResolverPtr);
static void			openNewResolvedProxyServerConnection	(SInt32 const					inData,
															 ProxyDomainNameLookupDataPtr	inResolverPtr);
static pascal void	openTransportDomainNameLookupProc		(void*							inContextPtr,
															 OTEventCode					inEventCode,
															 OTResult						inResult,
															 void*							inCookie);
static void			postDomainNameLookupBeginEvent			(DomainNameLookupDataPtr		inInfoPtr);
static void			postDomainNameLookupEndEvent			(DomainNameLookupDataPtr		inInfoPtr);



//
// public methods
//

/*!
To dispose of memory allocated by a DNR lookup,
use this method.

(3.0)
*/
void
DNR_Dispose		(DomainNameLookupDataPtr*	inoutDataPtr)
{
	if (inoutDataPtr != null)
	{
		DomainNameLookupDataPtr		ptr = *inoutDataPtr;
		
		
		if (ptr != null)
		{
			// dispose of the DNR memory allocations
			DisposeOTNotifyUPP(ptr->uppToDispose);
			Memory_DisposePtr((Ptr*)&ptr->hostInfoPtr);
			Memory_DisposePtr((Ptr*)&ptr);
		}
		*inoutDataPtr = null;
	}
}// Dispose


/*!
Initiates an asynchronous lookup of a host name,
invoking the specified completion method when the
operation completes (successful or not).  The given
completion method data is passed to the method when
it is called.

The domain name resolver has been re-written in
MacTelnet 3.0 to use Open Transport natively.

(3.0)
*/
OSStatus
DNR_LookUpDomainName		(Ptr							inHostNameBuffer,
							 Size const						inHostNameSize,
							 DNRDoneNotificationProcPtr		inCompletionMethod,
							 SInt32 const					inCompletionMethodData)
{
	OSStatus					result = noErr;
	DomainNameLookupDataPtr		resolverPtr = null;
	Boolean						done = false;
	
	
	// Construct and initialize a data structure that contains everything of
	// interest at all levels (this is necessary because Open Transport fires
	// back its feedback through interrupt-level routines, so the information
	// it provides has to go somewhere).  Memory allocations are destroyed by
	// DNR_OpenNewResolvedConnection().
	Console_WriteLine("domain name lookup");
	resolverPtr = (DomainNameLookupDataPtr)Memory_NewPtr(sizeof(DomainNameLookupData));
	if (resolverPtr == null)
	{
		result = memFullErr;
		done = true;
	}
	else
	{
		resolverPtr->dnrCompletionMethod = inCompletionMethod;
		resolverPtr->data = inCompletionMethodData;
		resolverPtr->hostInfoPtr = (InetHostInfo*)Memory_NewPtr(sizeof(InetHostInfo));
		if (resolverPtr->hostInfoPtr == null)
		{
			result = memFullErr;
			done = true;
		}
	}
	
	// if theres enough memory, initiate the lookup
	if (!done)
	{
		InetAddress		address;
		
		
		// create a C-style host name string
		Memory_WriteZero(resolverPtr->hostName, sizeof(resolverPtr->hostName));
		BlockMove(inHostNameBuffer, resolverPtr->hostName, inHostNameSize);
		
		// if the specified string is already an IP address, theres no need to use the DNR!
		if (StringUtilities_BufferToInternetAddress(inHostNameBuffer, inHostNameSize, &address))
		{
			resolverPtr->error = noErr;
			resolverPtr->hostInfoPtr->addrs[0] = address.fHost;
			postDomainNameLookupEndEvent(resolverPtr);
			result = noErr;
			done = true;
		}
		
		// otherwise, call the domain name resolver asynchronously
		if (!done)
		{
			resolverPtr->uppToDispose = NewOTNotifyUPP(openTransportDomainNameLookupProc);
			resolverPtr->error = result = InternetServices_NewTCPIP(resolverPtr->uppToDispose,
																	resolverPtr/* user-defined context */);
		}
	}
	Console_WriteLine("done");
	
	return result;
}// LookUpDomainName


/*!
Initiates an asynchronous lookup of a host name that will
subsequently cause a new session to be initiated with the
given screen ID.  When the lookup completes, the specified
session window may open (if the IP address could be found),
or it may fail to connect.

(3.0)
*/
OSStatus
DNR_LookUpNewConnection		(StringPtr				inHostName,
							 TelnetScreenID const	inScreenID)
{
	OSStatus		result = noErr;
	SInt32			data = 0L;
	
	
	data = inScreenID;
	
	// asynchronously look up the IP address of the remote host
	Console_WriteValuePString("looking up new connection host", inHostName);
	result = DNR_LookUpDomainName((Ptr)(inHostName + 1), StrLength(inHostName), openNewResolvedConnection, data);
	return result;
}// LookUpNewConnection


/*!
Initiates an asynchronous lookup of a host name that will
subsequently cause a new session to be initiated with the
given screen ID.  When the lookup completes, the specified
session window may open (if the IP address could be found),
or it may fail to connect.

(3.0)
*/
OSStatus
DNR_LookUpNewConnectionViaProxy		(StringPtr				inHostName,
									 ProxyServerRef			inProxyServer,
									 TelnetScreenID const	inScreenID)
{
	OSStatus			result = noErr;
	InetDomainName		domainName;
	InetPort			portNumber = 0;
	ProxyLookupDataPtr	dataPtr = null;
	
	
	// set up data needed by proxy server DNR completion routine
	dataPtr = (ProxyLookupDataPtr)Memory_NewPtr(sizeof(ProxyLookupData));
	if (dataPtr == null) result = memFullErr;
	else
	{
		// set up fields
		dataPtr->screenID = inScreenID;
		dataPtr->proxyServer = inProxyServer;
		StringUtilities_PToC(inHostName, dataPtr->targetDomainName);
		
		// asynchronously look up the IP address of the proxy server
		ProxyServer_GetConnectionInfo(inProxyServer, domainName, &portNumber);
		Console_WriteLine(domainName);
		Console_WriteValuePString("above proxy server being used to find host", inHostName);
		result = DNR_LookUpDomainName(domainName, strlen(domainName), beginLookupViaProxyServerConnection,
										REINTERPRET_CAST(dataPtr, SInt32));
	}
	return result;
}// LookUpNewConnectionViaProxy


//
// internal methods
//

/*!
This method, of standard "DNRDoneNotificationProcPtr"
form, cleans up a DNR lookup that was initiated by
DNR_LookUpNewConnectionViaProxy().  *If* the given
data indicates that the proxy server was found and
is responding to the expected protocol, an attempt to
communicate with the proxy server is made, asking the
proxy server to perform its own DNR lookup of the
users target host.  This initiates another asynchronous
operation, ending with openNewResolvedProxyServerConnection().

(3.0)
*/
static void
beginLookupViaProxyServerConnection		(SInt32 const					inProxyLookupDataPtr,
										 DomainNameLookupDataPtr		inResolverPtr)
{
	if (inResolverPtr != null)
	{
		UInt32				proxyIPAddress;
		OSStatus			error;
		ProxyLookupDataPtr	dataPtr = REINTERPRET_CAST(inProxyLookupDataPtr, ProxyLookupDataPtr);
		
		
		Console_WriteLine("Proxy server DNR lookup cleanup");
		
		// grab the returned IP address, or an error (if any)
		proxyIPAddress = inResolverPtr->hostInfoPtr->addrs[0];
		// proxy host name = inResolverPtr->hostInfoPtr->name
		error = inResolverPtr->error;
		
		// communicate with the proxy server and request a DNR lookup of the users target host
		Console_WriteLine("Initiating communication with proxy server");
		{
			InetDomainName		domainName;
			InetPort			portNumber = 0;
			InetAddress			address;
			
			
			ProxyServer_GetConnectionInfo(dataPtr->proxyServer, domainName, &portNumber);
			OTInitInetAddress(&address, portNumber, proxyIPAddress);
			error = ProxyServer_Connect(dataPtr->proxyServer, &address);
			if (ProxyServer_IsConnected(dataPtr->proxyServer))
			{
				// send a request to look up a host
				error = ProxyServer_LookUpDomainName(dataPtr->proxyServer, dataPtr->targetDomainName,
														strlen(dataPtr->targetDomainName),
														openNewResolvedProxyServerConnection,
														STATIC_CAST(dataPtr->screenID, SInt32));
			}
		}
		Console_WriteLine("Done communicating with proxy server");
		
		// dispose of DNR memory allocations
		DNR_Dispose(&inResolverPtr);
	}
}// beginLookupViaProxyServerConnection


/*!
This method, of standard "DNRDoneNotificationProcPtr"
form, cleans up a DNR lookup that was initiated by
DNR_LookUpNewConnection().  A connection is opened
*if* the given structure indicates that the DNR
operation succeeded.

(3.0)
*/
static void
openNewResolvedConnection	(SInt32 const					inData,
							 DomainNameLookupDataPtr		inResolverPtr)
{
	if (inResolverPtr != null)
	{
		UInt32				ipAddress = 0L;
		OSStatus			error = noErr;
		TelnetScreenID		id = (TelnetScreenID)(inData & 0xFFFF);
		
		
		Console_WriteLine("New connection DNR lookup cleanup");
		
		// grab the returned IP address, or an error (if any)
		ipAddress = inResolverPtr->hostInfoPtr->addrs[0];
		error = inResolverPtr->error;
		
		// finally, complete the connection opening (possibly displaying an error message instead)
		Connections_CompleteOpening(id, ipAddress, error, inResolverPtr->hostInfoPtr->name);
		
		// dispose of the DNR memory allocations
		DNR_Dispose(&inResolverPtr);
	}
}// openNewResolvedConnection


/*!
This method, of "ProxyServerDNRDoneNotificationProcPtr"
form, cleans up a DNR lookup that was performed by a
remote proxy server.  A connection is opened *if* the
given structure indicates that the DNR operation
succeeded.

(3.0)
*/
static void
openNewResolvedProxyServerConnection	(SInt32 const						inData,
										 ProxyDomainNameLookupDataPtr		inResolverPtr)
{
	if (inResolverPtr != null)
	{
		UInt32				ipAddress = 0L;
		OSStatus			error = noErr;
		TelnetScreenID		id = (TelnetScreenID)(inData & 0xFFFF);
		
		
		Console_WriteLine("New connection via proxy DNR lookup cleanup");
		
		// grab the returned IP address, or an error (if any)
		ipAddress = inResolverPtr->hostInfoPtr->addrs[0];
		error = inResolverPtr->error;
		
		// finally, complete the connection opening (possibly displaying an error message instead)
		Connections_CompleteOpening(id, ipAddress, error, inResolverPtr->hostInfoPtr->name);
		
		// dispose of the DNR memory allocations
		ProxyServer_LookUpCleanUp(&inResolverPtr);
	}
}// openNewResolvedProxyServerConnection


/*!
This routine, of standard OTNotifyProcPtr form,
manages asynchronous lookup of domain names that
are attached to new connections.  If a domain
name is found, this routine causes the connection
to open - otherwise, it records the error code
and depends on code elsewhere to tell the user
about the problem.

Open Transport calls this routine under several
circumstances:
- at provider-open time, when this routine will
  send a DNR request out, asynchronously
- at DNR completion time, when this method looks
  at the error code (if any)
- at closing time, if the provider gets killed
  before the DNR lookup succeeds

WARNING:	This routine is deferred-task-level.
			You cannot perform tasks such as
			making system calls or allocating
			memory.  If you want anything
			interesting to happen, you have to
			define and post an event to the
			internal event queue (NetEvents.c)
			so that MacTelnet will know what to
			do.

(3.0)
*/
static pascal void
openTransportDomainNameLookupProc	(void*				inContextPtr,
									 OTEventCode		inEventCode,
									 OTResult			inResult,
									 void*				inCookie)
{
	DomainNameLookupDataPtr		resolverPtr = (DomainNameLookupDataPtr)inContextPtr;
	
	
	if (inEventCode == T_OPENCOMPLETE) // the first call of this routine for a particular DNR lookup
	{
		InetSvcRef		servicesRef = (InetSvcRef)inCookie;
		
		
		// Pass the new Internet Services reference in the structure,
		// and post an event that will initiate an Open Transport DNR
		// lookup call *outside* of this interrupt-level routine.
		resolverPtr->error = inResult;
		resolverPtr->servicesRef = servicesRef;
		postDomainNameLookupBeginEvent(resolverPtr);
	}
	else if (inEventCode == T_DNRSTRINGTOADDRCOMPLETE) // the IP address was found for the given name
	{
		InetHostInfo*		hostInfoPtr = (InetHostInfo*)inCookie;
		
		
		// The DNR lookup (successful or not) has ended.  Pass any error
		// code in the structure, and post an event that will clean up
		// memory allocations, etc. *outside* of this interrupt-level
		// routine.
		resolverPtr->error = inResult;
		postDomainNameLookupEndEvent(resolverPtr);
	}
	else if ((inEventCode == kOTProviderWillClose) ||
				(inEventCode == kOTProviderIsClosed)) // then the provider closed before this could finish
	{
		// This particular event is not deferred-task-level, so its safe.
		// If the provider is opened asynchronously (it is by default),
		// it is not possible to close it before the call has completed.
		// This might happen if the user quits the application before a
		// provider has opened.  For this reason, it is safer to open a
		// provider synchronously and then to use OTSetAsynchronous().
		// MacTelnet does not take this safer approach; in theory, a
		// user quitting MacTelnet, as described, could encounter delays.
		resolverPtr->error = inResult;
		(OSStatus)OTSetSynchronous(resolverPtr->servicesRef);
		(OSStatus)OTCloseProvider(resolverPtr->servicesRef);
		(OSStatus)OTSetAsynchronous(resolverPtr->servicesRef);
	}
}// openTransportDomainNameLookupProc


/*!
To post an event that will cause "NetEvents.c" to
initiate an Open Transport domain name lookup,
use this method.  This approach is necessary
because openTransportDomainNameLookupProc() is
interrupt-level, so it cannot directly invoke an
Open Transport call (or memory allocation routines)
to get the job done.

(3.0)
*/
static void
postDomainNameLookupBeginEvent	(DomainNameLookupDataPtr		inInfoPtr)
{
	NetEvents_Post(kTelnetNetworkEventClassUser, kUserEventDomainNameLookupBegin,
					inInfoPtr->data, (SInt32)inInfoPtr);
}// postDomainNameLookupBeginEvent


/*!
To post an event that will cause "NetEvents.c" to
clean up an Open Transport domain name lookup, use
this method.  This approach is necessary because
openTransportDomainNameLookupProc() is implemented
at the interrupt level, so it cannot directly use
Memory Manager routines to dispose of memory, nor
can it call Connections module routines (which
may move memory).

(3.0)
*/
static void
postDomainNameLookupEndEvent		(DomainNameLookupDataPtr		inInfoPtr)
{
	NetEvents_Post(kTelnetNetworkEventClassUser, kUserEventDomainNameLookupEnd,
					inInfoPtr->data, (SInt32)inInfoPtr);
}// postDomainNameLookupEndEvent