/*###############################################################

	DataAccessLib.h
	
	This file lists all public constants, functions, and other
	data from all modules of the Data Access Library, which is
	useful for simplifying access to data on the Mac OS.  This
	is a portable library, meaning it can be used with any
	program.
	
	Data Access Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

/*!
	@header Data Access Utility Library
*/

#ifndef __DATAACCESSLIB__
#define __DATAACCESSLIB__

#ifndef REZ

#ifdef powerc

// Mac includes
#include <MacTypes.h>
#include <OpenTransportProviders.h>
#include <QuickDraw.h>

// Data Access Library includes
#include "Releases.h"



/**************************************************************************************************************************/
// LinkedList.c
//
// This powerful module lets you create lists of anything, on the fly, represented as opaque objects.  You can create lists
// that are linked singly (one-way), doubly (two-way), or not at all (as vectors), depending upon whether the required
// memory space is more important than list access efficiency.  A host of useful routines lets you manipulate your lists
// in a variety of ways, despite their style (although certain operations are more efficient with particular list styles).
// The Linked List module automatically keeps a count of list items, so counting items in a linked list at any time is
// extremely efficient.  There are even index-based commands, allowing you to treat a linked list exactly like a mutable
// array!  By standardizing on the Linked List object, you dont re-write dynamic list code ten times in one program, *and*
// you dont risk having bugs because there is only one implementation that handles all dynamically-resizeable lists.

typedef struct OpaqueChain { unsigned garbage; }**			ChainRef;
typedef struct OpaqueChainLink { unsigned garbage; }*		ChainLinkRef;

typedef SInt32 ChainDescriptor;
enum
{
	// use descriptors to help you figure out what kind of data is
	// in a list, given a reference to the list
	kChainInvalidDescriptor = -1722		// assign values to your application lists descriptors relative to this
};

typedef UInt16 ChainLinkPlacement;
enum
{
	kChainLinkPlacementInsertAfter = 0,
	kChainLinkPlacementInsertBefore = 1,
	kChainLinkPlacementReplace = 2
};

typedef UInt16 ChainStyle;
enum
{
	// a list where each item points to its next one (non-circular); for large lists,
	// it can be expense to perform operations on items near the end of the list;
	// however, the singly-linked list is a very reliable implementation
	kChainStyleSinglyLinked = 0,
	
	// (buggy) a list where each item points both to its next and its previous items;
	// look-ups are more efficient on doubly-linked lists than singly-linked ones
	kChainStyleDoublyLinked = 1,
	
	// (unimplemented) a list that is technically singly-linked, but which behaves
	// like a doubly-linked list because it is reasonably efficient to traverse the
	// list in either direction; if you were to draw a diagram of the links from one
	// node to the next, a spiral forms, hence the name of this type of linked list;
	// the first nodes next node is the last node in the list, the last nodes
	// next node is the second node, and its next node is the 2nd-last node, etc.
	kChainStyleSpirallyLinked = 2,
	
	// (unimplemented) re-allocates contiguous blocks (arrays) every time items are
	// inserted in the list; look-ups are extremely fast with vectors, but since they
	// always require contiguous space, memory may run out if large vectors are used
	// (singly-linked or doubly-linked lists use more space, but not contiguous space,
	// so they tend to hold up better when memory is low, and/or fragmented badly);
	// vectors are recommended for small-to-medium-sized lists (i.e. most lists)
	kChainStyleVector = 3
};



/*!
Chain Link Sorting Method

Routines of this type are used to compare two links,
when sorting links in a chain.  Returns one of the
standard values -1 (link 1 belongs before link 2 in
an ascending order), 0 (both links appear to be the
same), or 1 (link 2 belongs before link 1 in an
ascending order).

Obviously, your function must be designed to handle
data of a known type; for efficiency, this routine
does not pass in extra information about data size,
or even the Chian that the data comes from.
*/
typedef int	 (*ChainLinkCompareProcPtr)		(void const*		inLinkData1,
											 void const*		inLinkData2);
#define InvokeChainLinkCompareProc(userRoutine, inLinkData1, inLinkData2) \
	(*userRoutine)((inLinkData1), (inLinkData2))



/*###############################################################
	CREATING AND DESTROYING CHAINS
###############################################################*/

ChainRef
	Chain_New							(ChainStyle const				inStyle);

// THE ROUTINE BELOW AUTOMATICALLY CALLS Chain_DisposeAllNodes()
void
	Chain_Dispose						(ChainRef*						inoutChainRef);

/*###############################################################
	CREATING AND DESTROYING CHAIN LINKS
###############################################################*/

ChainLinkRef
	Chain_NewNode						(ChainRef const					inChainRef,
										 void*							inNodeDataPtr);

// RETURNS THE NODE DATA POINTER PROVIDED WHEN THE NODE WAS ORIGINALLY CONSTRUCTED
void*
	Chain_DisposeNode					(ChainRef const					inChainRef,
										 ChainLinkRef*					inoutChainLinkRef);

void
	Chain_DisposeAllNodes				(ChainRef const					inChainRef);

/*###############################################################
	ACCESSING DATA IN CHAINS
###############################################################*/

// LESS EFFICIENT WAY OF DETERMINING LIST SIZE; OUGHT TO AGREE WITH THE RESULTS OF Chain_GetNodeCount()
UInt16
	Chain_DebugGetTraversedNodeCount	(ChainRef const					inChainRef);

void*
	Chain_GetAuxiliaryDataPtr			(ChainRef const					inChainRef);

ChainDescriptor
	Chain_GetDescriptor					(ChainRef const					inChainRef);

UInt16
	Chain_GetNodeCount					(ChainRef const					inChainRef);

ChainStyle
	Chain_GetStyle						(ChainRef const					inChainRef);

void
	Chain_SetAuxiliaryDataPtr			(ChainRef						inChainRef,
										 void*							inAuxiliaryDataPtr);

void
	Chain_SetDescriptor					(ChainRef						inChainRef,
										 ChainDescriptor const			inNewChainDescriptor);

/*###############################################################
	ACCESSING DATA IN CHAIN LINKS
###############################################################*/

void*
	Chain_GetNodeDataPtr				(ChainRef const					inChainRef,
										 ChainLinkRef const				inLinkRef);

void
	Chain_SetNodeDataPtr				(ChainRef const					inChainRef,
										 ChainLinkRef					inLinkRef,
										 void*							inAuxiliaryDataPtr);

/*###############################################################
	ACCESSING AND MANIPULATING ITEMS IN A CHAIN
###############################################################*/

ChainLinkRef
	Chain_DeleteFirstNode				(ChainRef const					inChainRef);

ChainLinkRef
	Chain_DeleteIndexedNode				(ChainRef const					inChainRef,
										 UInt16 const					inZeroBasedNodeIndex);

ChainLinkRef
	Chain_DeleteLastNode				(ChainRef const					inChainRef);

ChainLinkRef
	Chain_DeleteNode					(ChainRef const					inChainRef,
										 ChainLinkRef const				inLinkRef);

ChainLinkRef
	Chain_GetIndexedNode				(ChainRef const					inChainRef,
										 UInt16 const					inZeroBasedNodeIndex);

ChainLinkRef
	Chain_GetFirstNode					(ChainRef const					inChainRef);

ChainLinkRef
	Chain_GetLastNode					(ChainRef const					inChainRef);

ChainLinkRef
	Chain_GetNextNode					(ChainRef const					inChainRef,
										 ChainLinkRef const				inCurrentLinkRef);

ChainLinkRef
	Chain_GetPreviousNode				(ChainRef const					inChainRef,
										 ChainLinkRef const				inCurrentLinkRef);

Boolean
	Chain_IsNodeFirst					(ChainRef const					inChainRef,
										 ChainLinkRef const				inLinkRef);

Boolean
	Chain_IsNodeLast					(ChainRef const					inChainRef,
										 ChainLinkRef const				inLinkRef);

// ALTHOUGH THE NAME IS A BIT CONFUSING, YOU ALSO USE THIS ROUTINE TO ADD NEW NODES TO A LIST (AS WELL AS MOVE THEM)
void
	Chain_MoveNode						(ChainRef const					inChainRef,
										 ChainLinkRef const				inNodeToMoveRef,
										 ChainLinkRef const				inNodeAnchorRefOrNull,
										 ChainLinkPlacement const		inPositionHow);

// WARNING: YOU CANNOT QUICKSORT LISTS THAT ARE TOO LARGE TO ALLOCATE TEMPORARY MEMORY FOR - memFullErr MAY BE RETURNED
OSErr
	Chain_QuickSort						(ChainRef const					inChainRef,
										 ChainLinkCompareProcPtr		inComparisonProc);



/**************************************************************************************************************************/
// FlagManager.c
//
// To organize and streamline storage of and access to those inevitable Boolean globals, use this module.  Every modern
// Mac OS application is storing the typical existence flags: Color QuickDraw, Apple Guide, Navigation Services, and even
// Contextual Menus.  Now, you can maximize efficiency in storing such flags by simply assigning each a unique ID number
// reflecting the flag, and using this module to set or test the value.  Assign your application-defined flags relative
// to "kFlagManager_FirstValidFlag".

typedef UInt16 FlagManagerFlag;
enum
{
	kFlagManager_FirstValidFlag = 0
};



/*###############################################################
	INITIALIZING AND FINISHING WITH THE FLAG MANAGER
###############################################################*/

// CALL THIS ROUTINE ONCE, BEFORE ANY OTHER FLAG MANAGER ROUTINE
void
	FlagManager_Init		(void);

// CALL THIS ROUTINE AFTER YOU ARE PERMANENTLY DONE WITH THE FLAG MANAGER
void
	FlagManager_Done		(void);

/*###############################################################
	MANIPULATING FLAG MANAGER FLAGS IN THE GLOBAL SET
###############################################################*/

void
	FlagManager_Set			(FlagManagerFlag				inFlag,
							 Boolean const					inIsSet);

Boolean
	FlagManager_Test		(FlagManagerFlag				inFlag);



/**************************************************************************************************************************/
// Memory.c
//
// This module handles memory management.  It even includes safe arrays, if protected memory is crucial; these routines
// can be used in place of arrays for debugging to catch all illegal index references.

typedef struct OpaqueSafeArray { unsigned garbage; }*	SafeArrayRef;



/*###############################################################
	MEMORY MANAGEMENT SETUP
###############################################################*/

void
	Memory_Init							(Size const			inNumberOfBytesOfHeapPadding,
										 Size const			inNumberOfBytesToReserve);

/*###############################################################
	PROTECTED MEMORY
###############################################################*/

SafeArrayRef
	Memory_NewSafeArray					(UInt32 const		inDescriptor,
										 UInt16 const		inElementCount,
										 Size const			inElementSize);

void
	Memory_DisposeSafeArray				(SafeArrayRef*		inoutArrayPtr);

void*
	Memory_AccessSafeArray				(SafeArrayRef		inArray,
										 SInt16 const		inIndex,
										 void*				inDataPtrOrNull);

UInt16
	Memory_GetSafeArrayLength			(SafeArrayRef		inRef);

/*###############################################################
	INDISPENSABLE MEMORY MANAGEMENT ROUTINES
###############################################################*/

Handle
	Memory_NewHandle					(Size const			inDesiredNumberOfBytes);

Handle
	Memory_NewHandleCritical			(Size const			inDesiredNumberOfBytes);

Handle
	Memory_NewHandleInProperZone		(Size const			inDesiredNumberOfBytes,
										 Boolean const		inOnlyNeededForShortPeriodOfTime);

Ptr
	Memory_NewHandleLockedHigh			(Size const			inDesiredNumberOfBytes,
										 Handle*			outNewHandlePtr);

void
	Memory_DisposeHandle				(Handle*			inoutHandle);

// USE THIS ROUTINE TO CLARIFY YOUR INTENT WHEN DISPOSING OF A HANDLE
#ifdef __cplusplus
inline void
Memory_ReleaseDetachedResource	(Handle*	inoutHandlePtr)		{	Memory_DisposeHandle(inoutHandlePtr);	}
#else
#	define	Memory_ReleaseDetachedResource(inoutHandlePtr)	do { Memory_DisposeHandle(inoutHandlePtr); } while (0)
#endif

Ptr
	Memory_NewPtr						(Size const			inDesiredNumberOfBytes);

Ptr
	Memory_NewPtrCritical				(Size const			inDesiredNumberOfBytes);

void*
	Memory_NewPtrInterruptSafe			(Size const			inDesiredNumberOfBytes);

void
	Memory_DisposePtr					(Ptr*				inoutPtr);

void
	Memory_DisposePtrInterruptSafe		(void**				inoutPtr);

OSErr
	Memory_SetHandleSize				(Handle				inoutHandle,
										 Size const			inNewHandleSizeInBytes);

OSErr
	Memory_SetHandleSizeCritical		(Handle				inoutHandle,
										 Size const			inNewHandleSizeInBytes);

/*###############################################################
	QUICKDRAW MEMORY MANAGEMENT UTILITIES
###############################################################*/

RgnHandle
	Memory_NewRegion					(void);

void
	Memory_DisposeRegion				(RgnHandle*			inoutRegion);

/*###############################################################
	OTHER MEMORY MANAGEMENT UTILITIES
###############################################################*/

Boolean
	Memory_IsHandleLocked				(Handle				inHandle);

void
	Memory_WriteZero					(void*				inoutBufferPtr,
										 Size const			inSize);

/*###############################################################
	MEMORY EMERGENCIES
###############################################################*/

Boolean
	Memory_RecoverReserveIsAnythingLeft	(void);



/**************************************************************************************************************************/
// Releases.c
//
// This package includes utility routines for decoding Gestalt version numbers into more useful forms; specifically, the
// three important version numbers.  You can use this routine, for example, to decode the return value Gestalt gives you
// when you ask it for the current Mac OS version.

/*###############################################################
	EXTRACTING VERSION INFORMATION FROM LONG INTEGERS
###############################################################*/

UInt8
	Releases_GetMajorRevisionForVersion 		(long		inVersion);

UInt8
	Releases_GetMinorRevisionForVersion 		(long		inVersion);

UInt8
	Releases_GetSuperminorRevisionForVersion 	(long		inVersion);



/**************************************************************************************************************************/
// StringUtilities.c
//
// What powerful data access library would be complete without a set of string utilities?  This module contains powerful
// routines that supplement the Pascal String Functions from the Mac OS Universal Interfaces.  The most powerful
// routine is StringUtilities_PBuild(), which puts ParamText() to shame by providing similar functionality but using
// *configurable* metacharacters which can collectively be a length other than two characters wide.  If you *need* to
// substitute more than 4 different things into a string, just put as many different metacharacters as you need into the
// source string, call StringUtilities_PMetaSet() to change the current set of 4, and re-invoke StringUtilities_PBuild()
// as many times as you need to.  Other useful string utilities let you *insert* a string at the beginning of another
// string (the opposite of concatenation), append a number to a string, and even translate a MacTCP or Open Transport
// IP address into string form.

#define EMPTY_PSTRING				"\p"
#define kMetacharacterDefault1		"\p%a"
#define kMetacharacterDefault2		"\p%b"
#define kMetacharacterDefault3		"\p%c"
#define kMetacharacterDefault4		"\p%d"
enum
{
	kLengthMetacharacterDefault = 2
};

typedef UInt16 StringUtilitiesTruncationMethod;
enum
{
	kStringUtilities_TruncateAtEnd = 0,
	kStringUtilities_TruncateAtMiddle = 1,
	kStringUtilities_TruncateAtBeginning = 2
};

typedef struct OpaqueTokenSet { unsigned garbage; }**		TokenSetRef;

/*###############################################################
	STRING TOKENIZING UTILITIES
###############################################################*/

void
	StringUtilities_PDisposeTokens				(UInt16 const	inArgC,
												 TokenSetRef*	inoutArgV);

void
	StringUtilities_PGetToken					(UInt16 const	inZeroBasedTokenNumber,
												 TokenSetRef	inArgV,
												 StringPtr		outString);

void
	StringUtilities_PTokenize					(StringPtr		inString,
												 UInt16*		outArgC,
												 TokenSetRef*	outArgV);

void
	StringUtilities_PTokenizeBy					(StringPtr		inString,
												 StringPtr		inTokenSeparators,
												 UInt16*		outArgC,
												 TokenSetRef*	outArgV);

/*###############################################################
	STRING MANIPULATION UTILITIES
###############################################################*/

void
	StringUtilities_PAppendNumber				(StringPtr		inoutString,
												 long			inNumber);

void
	StringUtilities_PBuild						(StringPtr		inoutString,
												 StringPtr		inSubstitutionString1,
												 StringPtr		inSubstitutionString2,
												 StringPtr		inSubstitutionString3,
												 StringPtr		inSubstitutionString4);

void
	StringUtilities_PClipHead					(StringPtr		inoutString,
												 UInt16 const	inHowManyCharacters);

void
	StringUtilities_PClipTail					(StringPtr		inoutString,
												 UInt16 const	inHowManyCharacters);

void
	StringUtilities_PInsert						(StringPtr		outDestinationString,
												 StringPtr		inSourceString);

void
	StringUtilities_PStrip						(StringPtr		inoutString,
												 StringPtr		inStartingDelimiter,
												 StringPtr		inEndingDelimiter,
												 Boolean		inStripDelimitersToo);

Boolean
	StringUtilities_PTruncate					(StringPtr							inoutString,
												 UInt32								inAcceptableMaximumWidth,
												 StringUtilitiesTruncationMethod	inTruncationMethod);

/*###############################################################
	SETTING UP StringUtilities_PBuild()
###############################################################*/

void
	StringUtilities_PMetaGet					(StringPtr		outMeta1,
												 StringPtr		outMeta2,
												 StringPtr		outMeta3,
												 StringPtr		outMeta4,
												 UInt16*		outLength);

void
	StringUtilities_PMetaSet					(StringPtr		inMeta1,
												 StringPtr		inMeta2,
												 StringPtr		inMeta3,
												 StringPtr		inMeta4,
												 UInt16			inLength);

#define StringUtilities_PMetaSetDefaults()			StringUtilities_PMetaSet(kMetacharacterDefault1,\
																				kMetacharacterDefault2,\
																				kMetacharacterDefault3,\
																				kMetacharacterDefault4,\
																				kLengthMetacharacterDefault)

void
	StringUtilities_PMetaSetFromResources		(SInt16			inStringListResourceID,
												 SInt16			inMeta1StringIndex,
												 SInt16			inMeta2StringIndex,
												 SInt16			inMeta3StringIndex,
												 SInt16			inMeta4StringIndex,
												 UInt16			inLength);

/*###############################################################
	STRING INFORMATION UTILITIES
###############################################################*/

Boolean
	StringUtilities_PStartsWith					(StringPtr		inLongString,
												 StringPtr		inPrefixString);

/*###############################################################
	STRING CONVERSION UTILITIES
###############################################################*/

// RETURNS TRUE IF THE GIVEN BUFFER CONTAINS A VALID STRING REPRESENTATION OF AN IP ADDRESS (THIS ADDRESS IS RETURNED)
Boolean
	StringUtilities_BufferToInternetAddress		(Ptr			inBuffer,
												 Size const		inBufferLength,
												 InetAddress*	outIPAddressPtr);

void
	StringUtilities_CToP						(char*			inString,
												 Str255			outBuffer);

StringPtr
	StringUtilities_CToPInPlace					(char			inoutString[]);

void
	StringUtilities_PExtractFirstLine			(unsigned char*	inSourcePtr,
												 Size			inSize,
												 StringPtr		outLine);

void
	StringUtilities_PToC						(StringPtr		inString,
												 char			outBuffer[]);

char*
	StringUtilities_PToCInPlace					(StringPtr		inoutString);



/**************************************************************************************************************************/
// TextDataFile.c
//
// In preparation for Mac OS X, migrate your settings away from the Resource Manager, using this module to help!  The
// Text Data File can be used to create simple text files with Unicode representations of your settings, as well as any
// comment lines (prefixed by a hash mark, '#').  Although not quite as compact as most resources, this approach is far
// more portable across file systems, and very easy for users to change without supporting dialog boxes.  The former is
// important when preferences are saved on disks that dont even support multiple data forks.  For convenience, this module
// provides support for internally converting a number of standard Mac OS data types into text format, making it easier to
// save and restore data in memory.  Note that if you need to store pathnames to files, you should continue to use the
// Alias Manager to store alias resources in even a Text Data File, because aliases are more flexible (and more reliable)
// than simply writing file pathnames to a text file.  (Apple has not made it clear what form aliases will take in the
// absence of file forks.)  All of your settings take effect in the order you make them - so, if you write a comment prior
// to writing a name-value pair, the comment will appear before the name-value pair in the file.

typedef struct OpaqueTextDataFile { unsigned garbage; }*		TextDataFileRef;

/*###############################################################
	CREATING AND DESTROYING NAME-VALUE PAIR FILES
###############################################################*/

// THE SPECIFIED FILE MAY ALREADY EXIST - IF SO, ALL OF ITS DATA IS READ INTO MEMORY
TextDataFileRef
	TextDataFile_New					(FSSpec const*			inFSSpecPtr,
										 OSType const			inCreator,
										 OSType const			inType,
										 ScriptCode const		inInternationalCode);

void
	TextDataFile_Dispose				(TextDataFileRef*		inoutRefPtr);

/*###############################################################
	MANAGING THE FILE
###############################################################*/

Boolean
	TextDataFile_GetFSSpec				(TextDataFileRef		inRef,
										 FSSpec*				outFSSpecPtr);

// POINTS BACK TO THE TOP OF THE FILE WITHOUT CLOSING THE FILE; AFFECTS TextDataFile_GetNext...() OPERATIONS
Boolean
	TextDataFile_ResetPointer			(TextDataFileRef		inRef);

/*###############################################################
	SAVING AND RESTORING DATA
###############################################################*/

Boolean
	TextDataFile_AddComment				(TextDataFileRef		inRef,
										 char const*			inCommentTextPtr);

#if DATA_ACCESS_LIB_UNICODE_SUPPORT
Boolean
	TextDataFile_AddNameUnicodeValue	(TextDataFileRef		inRef,
										 char const*			inClassNameOrNull,
										 char const*			inPropertyName,
										 UniChar const*			inValue,
                              			 UniCharCount			inValueLength);
#endif

Boolean
	TextDataFile_AddNameValue			(TextDataFileRef		inRef,
										 char const*			inClassNameOrNull,
										 char const*			inPropertyName,
										 char const*			inValue);

Boolean
	TextDataFile_AddNameValueFlag		(TextDataFileRef		inRef,
										 char const*			inClassNameOrNull,
										 char const*			inPropertyName,
										 Boolean const			inValue);

Boolean
	TextDataFile_AddNameValueNumber		(TextDataFileRef		inRef,
										 char const*			inClassNameOrNull,
										 char const*			inPropertyName,
										 SInt32 const			inValue);

Boolean
	TextDataFile_AddNameValuePString	(TextDataFileRef		inRef,
										 char const*			inClassNameOrNull,
										 char const*			inPropertyName,
										 StringPtr				inValue);

Boolean
	TextDataFile_AddNameValueRGBColor	(TextDataFileRef		inRef,
										 char const*			inClassNameOrNull,
										 char const*			inPropertyName,
										 RGBColor const*		inValuePtr);

#if DATA_ACCESS_LIB_UNICODE_SUPPORT
Boolean
	TextDataFile_GetNextNameUnicodeValue(TextDataFileRef		inRef,
										 char*					outNamePtr,
										 SInt16 const			inNameLength,
										 UniChar*				outValuePtr,
                              			 UniCharCount			inValueLength);
#endif

Boolean
	TextDataFile_GetNextNameValue		(TextDataFileRef		inRef,
										 char*					outNamePtr,
										 SInt16 const			inNameLength,
										 char*					outValuePtr,
										 SInt16 const			inValueLength);

/*###############################################################
	DATA AND STRING CONVERSION UTILITIES
###############################################################*/

Boolean
	TextDataFile_StringToFlag			(TextDataFileRef		inRef,
										 char const*			inStringPtr,
										 Boolean*				outValuePtr);

Boolean
	TextDataFile_StringToNumber			(TextDataFileRef		inRef,
										 char const*			inStringPtr,
										 SInt32*				outValuePtr);

Boolean
	TextDataFile_StringToRGBColor		(TextDataFileRef		inRef,
										 char const*			inStringPtr,
										 RGBColor*				outValuePtr);

#endif /* ifdef powerc */

#endif /* ifndef REZ */

#endif