/*###############################################################

	DialogAdjust.c
	
	Interface Library 1.0
	 1998-2000 by Kevin Grant
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <ControlDefinitions.h>
#include <Lists.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"



// constants
#define NUMBER_OF_DIALOG_ITEM_ADJUSTMENT_TYPES	4	/* the number of constants enumerated in "InterfaceLib.h" */

// types
struct FrequentlyUsedData
{
	// This structure defines all of the common variables
	// usually needed by the internal adjustment methods,
	// such as horizontalMoveDialogItem(), etc.  This is
	// used in the interests of good memory management.
	// The adjust...() routines all make several calls to
	// the internal adjustment methods in order to perform
	// their dialog item adjustments.  Each of these calls
	// would normally require allocation of a rectangle, a
	// control handle, and usually the flag as well.  By
	// passing a pointer to a structure that is allocated
	// only once - at the beginning of an adjust...()
	// method - the process of adjusting dialog items is
	// much faster.
	ControlRef		control;
	Rect			itemRect;
	Boolean			doShow;
};
typedef struct FrequentlyUsedData	FrequentlyUsedData;
typedef FrequentlyUsedData*			FrequentlyUsedDataPtr;

struct AdjustmentData
{
	// It is important to correctly rearrange items
	// belonging to an embedding hierarchy.  Therefore,
	// dialog adjustments are first *proposed* by
	// using DialogAdjust_BeginAdjustment() followed by
	// one or more DialogAdjust_AddxxxAdjustment() calls,
	// where each adjustment is described by an instance
	// of AdjustmentData.
	//
	// A logical implementation order is determined by a
	// call to DialogAdjust_EndAdjustment(), which will
	// also perform all adjustments in an efficient way.
	union
	{
		ControlRef			control;
		DialogItemIndex		itemIndex;
	} adjusted;
	SInt32			amount;
};
typedef struct AdjustmentData	AdjustmentData;
typedef AdjustmentData*			AdjustmentDataPtr;

struct AdjustmentDelta
{
	// This allows both delta-X and delta-Y to be
	// passed to the offscreen graphics routine
	// without wasting both "data1" and "data2".
	// See the adjustControlsOperation() method.
	SInt32		deltaX,
				deltaY;
};
typedef struct AdjustmentDelta	AdjustmentDelta;
typedef AdjustmentDelta*		AdjustmentDeltaPtr;

// variables
static WindowRef	gPendingAdjustmentWindowRef = null;
static Boolean		gPendingAdjustmentListContainsDialogItems = true;
static ChainRef		gPendingAdjustmentLists[NUMBER_OF_DIALOG_ITEM_ADJUSTMENT_TYPES];

// internal methods
static pascal Boolean adjustControlsOperation	(const ControlRef			inSpecificControlOrRoot,
												 const SInt16				inColorDepth,
												 const SInt16				inDeviceFlags,
												 GDHandle					inTargetDevice,
												 const SInt32				inData1,
												 const SInt32				inData2);
static void beginAdjustment					(const WindowRef				inWindow,
											 const Boolean					inIsDialogBox);
static void getDialogItemControlRect		(FrequentlyUsedDataPtr			inoutSpacePtr,
											 const DialogPtr				inDialogPtr,
											 const DialogItemIndex			inItemIndex,
											 Rect*							outRect);
static void horizontalMoveDialogItem		(FrequentlyUsedDataPtr			inoutSpacePtr,
											 const DialogPtr				inDialogPtr,
											 const DialogItemIndex			inItemIndex,
											 const SInt32					inLocationX);
static void horizontalResizeDialogItem		(FrequentlyUsedDataPtr			inoutSpacePtr,
											 const DialogPtr				inDialogPtr,
											 const DialogItemIndex			inItemIndex,
											 const SInt32					inDeltaSizeX);
static void horizontalSizeDialogItem		(FrequentlyUsedDataPtr			inoutSpacePtr,
											 const DialogPtr				inDialogPtr,
											 const DialogItemIndex			inItemIndex,
											 const SInt32					inSizeX);
static void offsetDialogItem				(FrequentlyUsedDataPtr			inoutSpacePtr,
											 const DialogPtr				inDialogPtr,
											 const DialogItemIndex			inItemIndex,
											 const SInt32					inDeltaSizeX,
											 const SInt32					inDeltaSizeY);
static void resizeDialogItem				(FrequentlyUsedDataPtr			inoutSpacePtr,
											 const DialogPtr				inDialogPtr,
											 const DialogItemIndex			inItemIndex,
											 const SInt32					inDeltaSizeX,
											 const SInt32					inDeltaSizeY);
static void verticalMoveDialogItem			(FrequentlyUsedDataPtr			inoutSpacePtr,
											 const DialogPtr				inDialogPtr,
											 const DialogItemIndex			inItemIndex,
											 const SInt32					inLocationY);
static void verticalResizeDialogItem		(FrequentlyUsedDataPtr			inoutSpacePtr,
											 const DialogPtr				inDialogPtr,
											 const DialogItemIndex			inItemIndex,
											 const SInt32					inDeltaSizeY);
static void verticalSizeDialogItem			(FrequentlyUsedDataPtr			inoutSpacePtr,
											 const DialogPtr				inDialogPtr,
											 const DialogItemIndex			inItemIndex,
											 const SInt32					inSizeY);



//
// public methods
//

/*!
Call this routine after DialogAdjust_BeginControlAdjustment().
Use this method once for every control adjustment that must be
made.  If you must resize or move a control in two directions
(horizontal and vertical), you must call this routine twice to
register the operations separately - this is required for proper
handling of the changes.

Always invoke DialogAdjust_AddControlAdjustment() in the order
you would use if the window were made *bigger* in *both*
dimensions; that is, adjust embedders before their subcontrols.
The list of proposed adjustments will be invoked in a logical
order by a call to the DialogAdjust_EndAdjustment() routine.

IMPORTANT:	You cannot use this method if you have started any
			adjustments with DialogAdjust_BeginDialogItemAdjustment().
			Always bracket calls to this routine with calls to
			DialogAdjust_BeginControlAdjustment() and
			DialogAdjust_EndAdjustment().

(3.0)
*/
void
DialogAdjust_AddControlAdjustment	(const DialogItemAdjustment		inAdjustment,
									 ControlRef						inControlToAdjust,
									 const SInt32					inAmount)
{
	Boolean		isError = false;
	
	
	if ((inAdjustment >= 0) && (inAdjustment < NUMBER_OF_DIALOG_ITEM_ADJUSTMENT_TYPES))
	{
		ChainRef			list = gPendingAdjustmentLists[inAdjustment];
		AdjustmentDataPtr	dataPtr = (AdjustmentDataPtr)Memory_NewPtr(sizeof(AdjustmentData));
		
		
		if (dataPtr != null)
		{
			ChainLinkRef	node = null;
			
			
			dataPtr->adjusted.control = inControlToAdjust;
			dataPtr->amount = inAmount;
			node = Chain_NewNode(list, dataPtr);
			Chain_MoveNode(list, node, null/* means end of list */, kChainLinkPlacementInsertAfter);
		}
		else isError = true;
	}
	else isError = true;
	
	if (isError) BEEP;
}// AddControlAdjustment


/*!
Call this routine after DialogAdjust_BeginDialogItemAdjustment().
Use this method once for every dialog item adjustment that must
be made.  If you must resize or move an item in two directions
(horizontal and vertical), you must call this routine twice to
register the operations separately - this is required for proper
handling of the changes.

Always invoke DialogAdjust_AddDialogItem() in the order you
would use if the dialog window were made *bigger* in *both*
dimensions; that is, adjust embedders before their subcontrols.
The list of proposed adjustments will be invoked in a logical
order by a call to the DialogAdjust_EndAdjustment() routine.

IMPORTANT:	You cannot use this method if you have started any
			adjustments with DialogAdjust_BeginControlAdjustment().
			Always bracket calls to this routine with calls to
			DialogAdjust_BeginDialogItemAdjustment() and
			DialogAdjust_EndAdjustment().

(3.0)
*/
void
DialogAdjust_AddDialogItem	(const DialogItemAdjustment		inAdjustment,
							 const DialogItemIndex			inItemIndex,
							 const SInt32					inAmount)
{
	Boolean		isError = false;
	
	
	if ((inAdjustment >= 0) && (inAdjustment < NUMBER_OF_DIALOG_ITEM_ADJUSTMENT_TYPES))
	{
		ChainRef			list = gPendingAdjustmentLists[inAdjustment];
		AdjustmentDataPtr	dataPtr = (AdjustmentDataPtr)Memory_NewPtr(sizeof(AdjustmentData));
		
		
		if (dataPtr != null)
		{
			ChainLinkRef	node = null;
			
			
			dataPtr->adjusted.itemIndex = inItemIndex;
			dataPtr->amount = inAmount;
			node = Chain_NewNode(list, dataPtr);
			Chain_MoveNode(list, node, null/* means end of list */, kChainLinkPlacementInsertAfter);
		}
		else isError = true;
	}
	else isError = true;
	
	if (isError) BEEP;
}// AddDialogItemAdjustment


/*!
Call this routine to construct a new list of adjustments
for a particular window.  Once created, use one or more
calls to DialogAdjust_AddControlAdjustment() to make a
list of proposed control adjustments.  When finished,
call the DialogAdjust_EndAdjustment() routine.

(3.0)
*/
void
DialogAdjust_BeginControlAdjustment		(const WindowRef	inWindow)
{
	beginAdjustment(inWindow, false);
}// BeginControlAdjustment


/*!
Call this routine to construct a new list of adjustments
for a particular dialog box.  Once created, use one or
more calls to DialogAdjust_AddDialogItem() to make a list
of proposed item adjustments.  When finished, call the
DialogAdjust_EndAdjustment() routine.

(3.0)
*/
void
DialogAdjust_BeginDialogItemAdjustment		(const DialogPtr	inDialogPtr)
{
	beginAdjustment(GetDialogWindow(inDialogPtr), true);
}// BeginDialogItemAdjustment


/*!
Call this routine to effect changes to a dialog box or
a window that you listed using other item adjustment
routines from this module.

If the dialog is being made horizontally smaller, then
the list of "kDialogItemAdjustmentResizeH" operations is
traversed backwards; otherwise, it is traversed in the
order the adjustments were specified.

If the dialog is being made vertically smaller, then the
list of "kDialogItemAdjustmentResizeV" operations is
traversed backwards; otherwise, it is traversed in the
order the adjustments were specified.

(3.0)
*/
void
DialogAdjust_EndAdjustment		(const SInt32		inDialogDeltaSizeX,
								 const SInt32		inDialogDeltaSizeY)
{
	AdjustmentDelta		delta;
	
	
	delta.deltaX = inDialogDeltaSizeX;
	delta.deltaY = inDialogDeltaSizeY;
	
	Embedding_OffscreenControlOperation(gPendingAdjustmentWindowRef, null/* use root */, adjustControlsOperation/* proc. */,
										(SInt32)&delta/* data 1 */, 0L/* data 2 (reserved) */);
}// EndAdjustment


//
// internal methods
//

/*!
This method, of standard OffscreenOperationProcPtr
form, performs the specified adjustment of controls
or dialog items in the given window.  This allows,
memory permitting, the entire arrangement to take
place in an offscreen graphics world, which would
greatly smooth the updating of windows.

The value of "inData1" is a pointer to a structure
of type "AdjustmentDelta", specifying the change
in width and height of the window.  The value of
"inData2" is reserved.

If the operation cannot be completed, false is
returned; otherwise, true is returned.

(1.0)
*/
static pascal Boolean
adjustControlsOperation		(const ControlRef			inSpecificControlOrRoot,
							 const SInt16				inColorDepth,
							 const SInt16				inDeviceFlags,
							 GDHandle					inTargetDevice,
							 const SInt32				inData1,
							 const SInt32				inData2)
{
	OSErr			error = noErr;
	Boolean			result = pascal_true;
	
	
	if (inSpecificControlOrRoot == null) error = memPCErr;
	if ((Ptr)inData1 == null) error = memPCErr;
	if (error == noErr)
	{
		register SInt16			i = 0;
		AdjustmentDeltaPtr		deltaPtr = (AdjustmentDeltaPtr)inData1;
		ChainRef				currentList = null;
		ChainLinkRef			node = null;
		AdjustmentDataPtr		dataPtr = null;
		ControlRef				control = null;
		Rect					controlRect;
		WindowRef				window = null;
		Boolean					reverseTraversal = false,
								wasControlVisible = false;
		DialogItemAdjustment	currentAdjustment = kDialogItemAdjustmentResizeH;
		DialogItemAdjustment	executionOrder[NUMBER_OF_DIALOG_ITEM_ADJUSTMENT_TYPES] =
								{
									// redefined below, but this guarantees its initialized to something valid
									kDialogItemAdjustmentMoveH,
									kDialogItemAdjustmentMoveV,
									kDialogItemAdjustmentResizeH,
									kDialogItemAdjustmentResizeV
								};
		
		
		// get the window
		window = GetControlOwner(inSpecificControlOrRoot);
		
		// hide the root, so moving controls, etc. cant be seen
		//(OSErr)SetControlVisibility(inSpecificControlOrRoot, false/* visibility */, true/* draw */);
		
		// Decide which operations should be performed first.  Note
		// that under Mac OS X and beyond, this will not matter,
		// because the virtual memory manager should never allow
		// buffering to fail; so, the entire if-else chain that
		// follows this comment could be removed once Telnet fully
		// migrates to be Mac OS X native.
		if ((deltaPtr->deltaX >= 0) && (deltaPtr->deltaY < 0))
		{
			executionOrder[0] = kDialogItemAdjustmentMoveH;
			executionOrder[1] = kDialogItemAdjustmentResizeH;
			executionOrder[2] = kDialogItemAdjustmentResizeV;
			executionOrder[3] = kDialogItemAdjustmentMoveV;
		}
		else if ((deltaPtr->deltaX < 0) && (deltaPtr->deltaY < 0))
		{
			executionOrder[0] = kDialogItemAdjustmentResizeH;
			executionOrder[1] = kDialogItemAdjustmentResizeV;
			executionOrder[2] = kDialogItemAdjustmentMoveH;
			executionOrder[3] = kDialogItemAdjustmentMoveV;
		}
		else if ((deltaPtr->deltaX >= 0) && (deltaPtr->deltaY >= 0))
		{
			executionOrder[0] = kDialogItemAdjustmentMoveH;
			executionOrder[1] = kDialogItemAdjustmentMoveV;
			executionOrder[2] = kDialogItemAdjustmentResizeH;
			executionOrder[3] = kDialogItemAdjustmentResizeV;
		}
		else if ((deltaPtr->deltaX < 0) && (deltaPtr->deltaY >= 0))
		{
			executionOrder[0] = kDialogItemAdjustmentResizeH;
			executionOrder[1] = kDialogItemAdjustmentMoveH;
			executionOrder[2] = kDialogItemAdjustmentMoveV;
			executionOrder[3] = kDialogItemAdjustmentResizeV;
		}
		
		// Decide how to perform the operations, and then do them!
		// As soon as an operation is performed, it is destroyed,
		// because there is no need to keep the information in
		// memory any longer.
		for (i = 0; i < NUMBER_OF_DIALOG_ITEM_ADJUSTMENT_TYPES; i++)
		{
			currentAdjustment = executionOrder[i];
			currentList = gPendingAdjustmentLists[currentAdjustment];
			
			// determine how to perform these operations
			switch (currentAdjustment)
			{
				case kDialogItemAdjustmentResizeH:
					reverseTraversal = (deltaPtr->deltaX < 0); // resize controls inside-out to avoid cutoff drawing glitches
					break;
				
				case kDialogItemAdjustmentResizeV:
					reverseTraversal = (deltaPtr->deltaY < 0); // resize controls inside-out to avoid cutoff drawing glitches
					break;
				
				case kDialogItemAdjustmentMoveH:
				case kDialogItemAdjustmentMoveV:
					reverseTraversal = false; // unimportant for moving controls
					break;
				
				default:
					// ???
					break;
			}
			
			// Now perform the operations.  As soon as an operation
			// is performed, it is destroyed, because the information
			// is not used for anything else.  When all operations
			// are performed, therefore, the list itself is disposed.
			node = (reverseTraversal) ? Chain_GetLastNode(currentList) : Chain_GetFirstNode(currentList);
			do
			{
				dataPtr = Chain_GetNodeDataPtr(currentList, node);
				if (dataPtr != null)
				{
					// perform the operation...
					if (gPendingAdjustmentListContainsDialogItems)
					{
						(OSErr)GetDialogItemAsControl(GetDialogFromWindow(window), dataPtr->adjusted.itemIndex, &control);
					}
					else control = dataPtr->adjusted.control;
					
					wasControlVisible = IsControlVisible(control);
					GetControlBounds(control, &controlRect);
					SetControlVisibility(control, false/* visible */, false/* draw */);
					
					switch (currentAdjustment)
					{
						case kDialogItemAdjustmentResizeH:
						case kDialogItemAdjustmentResizeV:
							{
								SInt16		sizeX = 0,
											sizeY = 0;
								
								
								if (currentAdjustment == kDialogItemAdjustmentResizeH) controlRect.right += dataPtr->amount;
								else if (currentAdjustment == kDialogItemAdjustmentResizeV) controlRect.bottom += dataPtr->amount;
								
								sizeX = controlRect.right - controlRect.left;
								sizeY = controlRect.bottom - controlRect.top;
								
								if (gPendingAdjustmentListContainsDialogItems)
								{
									SizeDialogItem(GetDialogFromWindow(window), dataPtr->adjusted.itemIndex, sizeX, sizeY);
								}
								else
								{
									SizeControl(control, sizeX, sizeY);
								}
							}
							break;
						
						case kDialogItemAdjustmentMoveH:
						case kDialogItemAdjustmentMoveV:
							{
								SInt16		positionX = 0,
											positionY = 0;
								
								
								if (currentAdjustment == kDialogItemAdjustmentMoveH)
								{
									positionX = controlRect.left + dataPtr->amount;
									positionY = controlRect.top;
								}
								else if (currentAdjustment == kDialogItemAdjustmentMoveV)
								{
									positionX = controlRect.left;
									positionY = controlRect.top + dataPtr->amount;
								}
								
								if (gPendingAdjustmentListContainsDialogItems)
								{
									MoveDialogItem(GetDialogFromWindow(window), dataPtr->adjusted.itemIndex,
													positionX, positionY);
								}
								else
								{
									MoveControl(control, positionX, positionY);
								}
							}
							break;
						
						default:
							// ???
							break;
					}
					if (wasControlVisible) SetControlVisibility(control, true/* visible */, false/* draw */);
					
					// ...and now its no longer needed, so get rid of it
					DisposePtr((Ptr)dataPtr), dataPtr = null;
				}
				node = (reverseTraversal) ? Chain_GetPreviousNode(currentList, node)
											: Chain_GetNextNode(currentList, node);
			} while (node != null);
			Chain_Dispose(&currentList); // automatically destroys all nodes
		}
		gPendingAdjustmentWindowRef = null;
		
		// restore the visible state of the root
		//(OSErr)SetControlVisibility(inSpecificControlOrRoot, true/* visibility */, true/* draw */);
		
		Cursors_UseArrow();
	}
	else result = pascal_false;
	
	return result;
}// adjustControlsOperation


/*!
Call this routine to construct a new list of adjustments
for a particular dialog box or window.  This method is
invoked by the two public list-creation methods.  The
"inIsDialogBox" flag determines what kinds of adjustments
are legal to add to the resultant list.

(3.0)
*/
static void
beginAdjustment		(const WindowRef	inWindow,
					 const Boolean		inIsDialogBox)
{
	register SInt16		i = 0;
	
	
	gPendingAdjustmentWindowRef = inWindow;
	gPendingAdjustmentListContainsDialogItems = inIsDialogBox;
	
	// create linked lists to store operations
	for (i = 0; i < NUMBER_OF_DIALOG_ITEM_ADJUSTMENT_TYPES; i++)
	{
		gPendingAdjustmentLists[i] = Chain_New(kChainStyleSinglyLinked);
	}
}// beginAdjustment


/*!
To determine the rectangular boundary of a dialog item
as a control, call this method.  Dialog items for
resource-based controls will return a bounding rectangle
for the control as described in the control resource,
which is not necessarily the dialog item rectangle.

(3.0)
*/
static void
getDialogItemControlRect	(FrequentlyUsedDataPtr	inoutSpacePtr,
							 const DialogPtr		inDialogPtr,
							 const DialogItemIndex	inItemIndex,
							 Rect*					outRect)
{
	(OSErr)GetDialogItemAsControl(inDialogPtr, inItemIndex, &inoutSpacePtr->control);
	GetControlBounds(inoutSpacePtr->control, outRect);
}// getDialogItemControlRect


/*!
To set the left edge of a specific dialog item to a
particular location, call this method.

(3.0)
*/
static void
horizontalMoveDialogItem	(FrequentlyUsedDataPtr	inoutSpacePtr,
							 const DialogPtr		inDialogPtr,
							 const DialogItemIndex	inItemIndex,
							 const SInt32			inLocationX)
{
	(OSErr)GetDialogItemAsControl(inDialogPtr, inItemIndex, &inoutSpacePtr->control);
	GetControlBounds(inoutSpacePtr->control, &inoutSpacePtr->itemRect);
	inoutSpacePtr->itemRect.left = inLocationX;
	inoutSpacePtr->doShow = IsControlVisible(inoutSpacePtr->control);
	SetControlVisibility(inoutSpacePtr->control, false/* visible */, true/* draw */);
	MoveControl(inoutSpacePtr->control, inoutSpacePtr->itemRect.left, inoutSpacePtr->itemRect.top);
	if (inoutSpacePtr->doShow) SetControlVisibility(inoutSpacePtr->control, true/* visible */, true/* draw */);
}// horizontalMoveDialogItem


/*!
To alter the width of a specific dialog item by a
certain number of pixels, call this method.  Pass
a negative value to shrink the width, or a positive
value to increase it.

(3.0)
*/
static void
horizontalResizeDialogItem	(FrequentlyUsedDataPtr	inoutSpacePtr,
							 const DialogPtr		inDialogPtr,
							 const DialogItemIndex	inItemIndex,
							 const SInt32			inDeltaSizeX)
{
	(OSErr)GetDialogItemAsControl(inDialogPtr, inItemIndex, &inoutSpacePtr->control);
	GetControlBounds(inoutSpacePtr->control, &inoutSpacePtr->itemRect);
	inoutSpacePtr->itemRect.right += inDeltaSizeX;
	SizeControl(inoutSpacePtr->control, (inoutSpacePtr->itemRect.right - inoutSpacePtr->itemRect.left),
					(inoutSpacePtr->itemRect.bottom - inoutSpacePtr->itemRect.top));
}// horizontalResizeDialogItem


/*!
To set the width of a specific dialog item to a
certain number of pixels, call this method.

(3.0)
*/
static void
horizontalSizeDialogItem	(FrequentlyUsedDataPtr	inoutSpacePtr,
							 const DialogPtr		inDialogPtr,
							 const DialogItemIndex	inItemIndex,
							 const SInt32			inSizeX)
{
	(OSErr)GetDialogItemAsControl(inDialogPtr, inItemIndex, &inoutSpacePtr->control);
	GetControlBounds(inoutSpacePtr->control, &inoutSpacePtr->itemRect);
	inoutSpacePtr->itemRect.right = inoutSpacePtr->itemRect.left + inSizeX - 1;
	SizeControl(inoutSpacePtr->control, (inoutSpacePtr->itemRect.right - inoutSpacePtr->itemRect.left),
				(inoutSpacePtr->itemRect.bottom - inoutSpacePtr->itemRect.top));
}// horizontalSizeDialogItem


/*!
To move a specific dialog item by a certain number of
pixels horizontally and vertically, call this method.
Pass a negative value to move the control left (X) or
upwards (Y), or a positive value to move it downwards
(Y) or to the right (X).  Passing a zero in a dimension
causes the control to move only straight up or down,
or straight right or left, depending.

(3.0)
*/
static void
offsetDialogItem	(FrequentlyUsedDataPtr	inoutSpacePtr,
					 const DialogPtr		inDialogPtr,
					 const DialogItemIndex	inItemIndex,
					 const SInt32			inDeltaSizeX,
					 const SInt32			inDeltaSizeY)
{
	(OSErr)GetDialogItemAsControl(inDialogPtr, inItemIndex, &inoutSpacePtr->control);
	GetControlBounds(inoutSpacePtr->control, &inoutSpacePtr->itemRect);
	inoutSpacePtr->itemRect.left += inDeltaSizeX;
	inoutSpacePtr->itemRect.right += inDeltaSizeX;
	inoutSpacePtr->itemRect.top += inDeltaSizeY;
	inoutSpacePtr->itemRect.bottom += inDeltaSizeY;
	inoutSpacePtr->doShow = IsControlVisible(inoutSpacePtr->control);
	SetControlVisibility(inoutSpacePtr->control, false/* visible */, true/* draw */);
	MoveControl(inoutSpacePtr->control, inoutSpacePtr->itemRect.left, inoutSpacePtr->itemRect.top);
	if (inoutSpacePtr->doShow) SetControlVisibility(inoutSpacePtr->control, true/* visible */, true/* draw */);
}// offsetDialogItem


/*!
To resize a specific dialog item by a certain number of
pixels horizontally and vertically, call this method.
Pass a negative value to size the control left (X) or
upwards (Y), or a positive value to size it downwards
(Y) or to the right (X).  Passing a zero in a dimension
causes the control to size only straight up or down,
or straight right or left, depending.

(3.0)
*/
static void
resizeDialogItem	(FrequentlyUsedDataPtr	inoutSpacePtr,
					 const DialogPtr		inDialogPtr,
					 const DialogItemIndex	inItemIndex,
					 const SInt32			inDeltaSizeX,
					 const SInt32			inDeltaSizeY)
{
	(OSErr)GetDialogItemAsControl(inDialogPtr, inItemIndex, &inoutSpacePtr->control);
	GetControlBounds(inoutSpacePtr->control, &inoutSpacePtr->itemRect);
	inoutSpacePtr->itemRect.right += inDeltaSizeX;
	inoutSpacePtr->itemRect.bottom += inDeltaSizeY;
	SizeControl(inoutSpacePtr->control, (inoutSpacePtr->itemRect.right - inoutSpacePtr->itemRect.left),
				(inoutSpacePtr->itemRect.bottom - inoutSpacePtr->itemRect.top));
}// resizeDialogItem


/*!
To set the top edge of a specific dialog item to a
particular location, call this method.

(3.0)
*/
static void
verticalMoveDialogItem	(FrequentlyUsedDataPtr	inoutSpacePtr,
						 const DialogPtr		inDialogPtr,
						 const DialogItemIndex	inItemIndex,
						 const SInt32			inLocationY)
{
	(OSErr)GetDialogItemAsControl(inDialogPtr, inItemIndex, &inoutSpacePtr->control);
	GetControlBounds(inoutSpacePtr->control, &inoutSpacePtr->itemRect);
	inoutSpacePtr->itemRect.top = inLocationY;
	inoutSpacePtr->doShow = IsControlVisible(inoutSpacePtr->control);
	SetControlVisibility(inoutSpacePtr->control, false/* visible */, true/* draw */);
	MoveControl(inoutSpacePtr->control, inoutSpacePtr->itemRect.left, inoutSpacePtr->itemRect.top);
	if (inoutSpacePtr->doShow) SetControlVisibility(inoutSpacePtr->control, true/* visible */, true/* draw */);
}// verticalMoveDialogItem


/*!
To alter the height of a specific dialog item by a
certain number of pixels, call this method.  Pass
a negative value to shrink the height, or a positive
value to increase it.

(3.0)
*/
static void
verticalResizeDialogItem	(FrequentlyUsedDataPtr	inoutSpacePtr,
							 const DialogPtr		inDialogPtr,
							 const DialogItemIndex	inItemIndex,
							 const SInt32			inDeltaSizeY)
{
	(OSErr)GetDialogItemAsControl(inDialogPtr, inItemIndex, &inoutSpacePtr->control);
	GetControlBounds(inoutSpacePtr->control, &inoutSpacePtr->itemRect);
	inoutSpacePtr->itemRect.bottom += inDeltaSizeY;
	SizeControl(inoutSpacePtr->control, (inoutSpacePtr->itemRect.right - inoutSpacePtr->itemRect.left),
				(inoutSpacePtr->itemRect.bottom - inoutSpacePtr->itemRect.top));
}// verticalResizeDialogItem


/*!
To set the height of a specific dialog item to a
certain number of pixels, call this method.

(3.0)
*/
static void
verticalSizeDialogItem		(FrequentlyUsedDataPtr	inoutSpacePtr,
							 const DialogPtr		inDialogPtr,
							 const DialogItemIndex	inItemIndex,
							 const SInt32			inSizeY)
{
	(OSErr)GetDialogItemAsControl(inDialogPtr, inItemIndex, &inoutSpacePtr->control);
	GetControlBounds(inoutSpacePtr->control, &inoutSpacePtr->itemRect);
	inoutSpacePtr->itemRect.bottom = inoutSpacePtr->itemRect.top + inSizeY;
	SizeControl(inoutSpacePtr->control, (inoutSpacePtr->itemRect.right - inoutSpacePtr->itemRect.left),
				(inoutSpacePtr->itemRect.bottom - inoutSpacePtr->itemRect.top));
}// verticalSizeDialogItem