/*###############################################################

	DialogUtilities.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>

// Mac includes
#include <Appearance.h>
#include <Balloons.h>
#include <ControlDefinitions.h>
#include <Fonts.h>
#include <LowMem.h>
#include <MacErrors.h>
#include <PLStringFuncs.h>
#include <Resources.h>
#include <Script.h>
#include <StandardFile.h>
#include <TextUtils.h>
#include <ToolUtils.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "GeneralResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "ContextualMenuBuilder.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "MacLook.h"
#include "NetEvents.h"
#include "ResourceFileManager.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



// variables
#if TARGET_API_MAC_OS8
static short			gSetDir = 0;
static DialogRef		gCurrentThemeSavvyStandardFileModal = null;	// used with GetCurrentThemeSavvyStandardFileModal()
static Str255			gThemeSavvyDialogTextPrompt;				// used with ThemeSavvyStandardGetFile()
static FileFilterUPP	gThemeSavvyGetFileFileFilterUPP = null;		// used with ThemeSavvyStandardGetFile()
#endif

// internal method prototypes
static void				dialogClose									(DialogRef const	inDialog);
static void				dialogDisplayEnsureVisibility				(DialogRef const	inDialog);
static void				dialogDisplayPreparation					(DialogRef const	inDialog);
static void				useBestHelpButtonIcon						(ControlRef			inControl);
#if TARGET_API_MAC_OS8
static void				appearanceizeDialog							(DialogRef			inModalDialog);
static void				catColonAndAppName							(StringPtr			inoutStringPtr);
static WindowRef		newWorthlessWindow							(void);
static void				setStandardFileDialogDefaultButton			(DialogRef			inDialog);
static Boolean			themeSavvyStandardFileEventFilter			(DialogRef			inDialog,
																	 EventRecord*		inoutEventPtr,
																	 short*				inoutItemHit,
																	 void*				inData);
static pascal short		themeSavvyStandardGetDirectoryDialogHook	(short				inItem,
																	 DialogRef			inDialog,
																	 void*				inData);
static pascal Boolean	themeSavvyStandardGetDirectoryFileFilter	(CInfoPBPtr			inParamBlockPtr,
																	 void*				inData);
static pascal Boolean	themeSavvyStandardGetDirectoryModalFilter	(DialogRef			inDialog,
																	 EventRecord*		inoutEventPtr,
																	 short*				inoutItemHit,
																	 void*				inData);
static pascal short		themeSavvyStandardGetFileDialogHook			(short				inItem,
																	 DialogRef			inDialog,
																	 void*				inData);
static pascal Boolean	themeSavvyStandardGetFileFileFilter			(CInfoPBPtr			inParamBlockPtr,
																	 void*				inData);
static pascal Boolean	themeSavvyStandardGetFileModalFilter		(DialogRef			inDialog,
																	 EventRecord*		inoutEventPtr,
																	 short*				inoutItemHit,
																	 void*				inData);
static pascal void		themeSavvyStandardPutFileActivate			(DialogRef			inDialog,
																	 short				inItem,
																	 Boolean			inActivating,
																	 void*				inData);
static pascal short		themeSavvyStandardPutFileDialogHook			(short				inItem,
																	 DialogRef			inDialog,
																	 void*				inData);
static pascal Boolean	themeSavvyStandardPutFileModalFilter		(DialogRef			inDialog,
																	 EventRecord*		inoutEventPtr,
																	 short*				inoutItemHit,
																	 void*				inData);
#endif



//
// public methods
//

/*!
Enables (activates) a control, given its dialog
box and dialog item index.

(3.0)
*/
void
ActivateDialogItem	(DialogRef				inDialog,
					 DialogItemIndex const	inItemIndex)
{
	ControlRef		control = null;
	
	
	(OSErr)GetDialogItemAsControl(inDialog, inItemIndex, &control);
	(OSErr)ActivateControl(control);
}// ActivateDialogItem


/*!
This filter treats ANY key event it receives (with
command key or not) as being a default item hit,
and ANY mouse event it receives as being a default
item hit.  You might use this for an about dialog.

(3.0)
*/
pascal Boolean
AnyKeyDefaultDialogEventFilter	(DialogRef			inDialog,
								 EventRecord*		inoutEventPtr,
								 DialogItemIndex*	outItemIndex)
{
	Boolean		result = pascal_false;
	
	
	Threads_YieldToAnother();
	
	if ((inoutEventPtr->what == keyDown) || (inoutEventPtr->what == mouseDown))
	{
		*outItemIndex = GetDialogDefaultItem(inDialog);
		result = pascal_true;
	}
	
	if (!result) result = StandardDialogEventFilter(inDialog, inoutEventPtr, outItemIndex);
	
	return result;
}// AnyKeyDefaultDialogEventFilter


/*!
Disables (dims) a control, given its dialog
box and dialog item index.

(3.0)
*/
void
DeactivateDialogItem	(DialogRef				inDialog,
						 DialogItemIndex const	inItemIndex)
{
	ControlRef		control = null;
	
	
	(OSErr)GetDialogItemAsControl(inDialog, inItemIndex, &control);
	(OSErr)DeactivateControl(control);
}// DeactivateDialogItem


/*!
To deactivate the root control of the frontmost
window and to change the cursor to the standard
watch, call this method.  Usually, you do this
immediately prior to displaying a dialog box on
the screen.

This routine automatically finds the real
front window, deactivating all floating windows
as well.

(3.0)
*/
void
DeactivateFrontmostWindow	(void)
{
	Embedding_DeactivateFrontmostWindow();
}// DeactivateFrontmostWindow


/*!
This method calls FlashDialogItem() with a predefined
delay and flash repeat count.  This is useful for
determining the control that corresponds to a dialog
item number.

(3.0)
*/
void
DebugFlashDialogItem	(DialogRef const		inDialog,
						 DialogItemIndex const	inItemIndex)
{
	FlashDialogItem(inDialog, inItemIndex, 20/* delay in ticks*/, 4/* repeat count*/);
}// DebugFlashDialogItem


/*!
This method sequentially calls FlashDialogItem() with a
predefined delay and flash repeat count, on every
sub-control from the first level of embedding of the
indicated embedder control.  This is useful for determining
the controls that correspond to the sub-control indices for
an embedder control, indices that are needed when using the
GetIndexedSubControl() system call.

Unimplemented.

(3.0)
*/
void
DebugFlashDialogItemSubControls	(DialogRef const		UNUSED_ARGUMENT(inDialog),
								 DialogItemIndex const	UNUSED_ARGUMENT(inItemIndex))
{
}// DebugFlashDialogItemSubControls


/*!
This method briefly draws a rectangle around
the border of the specified dialog item in
the specified dialog box.  The period of time
to delay (or 0 for no delay) in ticks is
specified in the last parameter.  For example,
to guarantee that the dialog item outline is
visible for at least one second, pass "60".

(3.0)
*/
void
DebugOutlineDialogItem	(DialogRef const		inDialog,
						 DialogItemIndex const	inItemIndex,
						 unsigned long const	inDelay)
{
	DrawDialog(inDialog);
	
	{
		GrafPtr			oldPort = null;
		
		
		GetPort(&oldPort);
		SetPortWindowPort(GetDialogWindow(inDialog));
		PenNormal();
		{
			RGBColor		black;
			
			
			black.red = black.green = black.blue = 0;
			RGBForeColor(&black);
		}
		
		{
			ControlRef		control = null;
			Rect			controlRect;
			
			
			GetDialogItemAsControl(inDialog, inItemIndex, &control);
			GetControlBounds(control, &controlRect);
			FrameRect(&controlRect);
		}
		SetPort(oldPort);
	}
	
	{
		unsigned long	dummy = 0L;
		
		
		Delay(inDelay, &dummy);
	}
}// DebugOutlineDialogItem


/*!
To display a standard Save File dialog box and
create a file containing a dump of the control
embedding hierarchy for the specified window,
use this method.

(3.0)
*/
void
DebugSelectControlHierarchyDumpFile		(WindowRef		inForWindow)
{
	FSSpec		file;
	OSErr		error = noErr;
	Str255		prompt,
				title;
	Boolean		good = false;
	
	
	PLstrcpy(prompt, "\pCreate a file to show the windows control hierarchy.");
	PLstrcpy(title, "\pCreate Control Hierarchy Dump File");
	if (FlagManager_Test(kFlagNavigationServices))
	{
		SInt16			hit = 0;
		NavReplyRecord	reply;
		
		
		Alert_ReportOSErr(error = Navigation_PutFile(prompt, title,
														kApplicationCreatorSignature, 'TEXT',
														kTelnetNavPrefKeyGenericSaveFile,
														kNavDefaultNavDlogOptions | kNavDontAddTranslateItems,
														EventLoop_HandleNavigationUpdate, &reply, &file));
		good = ((error == noErr) && (reply.validRecord));
	}
	else
	{
		StandardFileReply	sfr;
		long				junk = 0L;
		
		
		DeactivateFrontmostWindow();
		ThemeSavvyStandardPutFile(title, prompt, "\phierarchy.txt", &sfr);
		RestoreFrontmostWindow();
		good = sfr.sfGood;
		if (good)
		{
			OSErr		error = noErr;
			
			
			error = FSMakeFSSpec(sfr.sfFile.vRefNum, sfr.sfFile.parID, sfr.sfFile.name, &file);
			good = ((error == noErr) || (error == fnfErr));
		}
		
	}
	
	if (good) error = DumpControlHierarchy(inForWindow, &file);
	else error = userCanceledErr;
	
	if (error != noErr) BEEP;
}// DebugSelectControlHierarchyDumpFile


/*!
If you want to draw something using a DeviceLoop()
drawing procedure, use this method.  The graphics
port is automatically clipped to the specified
region, preserving and restoring any previous
clipping regions.  DeviceLoop() is called using
the given drawing procedure, user data, and flags.

IMPORTANT:	On output, the specified clip region
			will be changed to the intersection of
			the region on input with the visible
			region of the current port.

(3.0)
*/
void
DeviceLoopClipAndDraw		(RgnHandle				inoutNewClipRegion,
							 DeviceLoopDrawingUPP	inProc,
							 long					inDeviceLoopUserData,
							 DeviceLoopFlags		inDeviceLoopFlags)
{
	RgnHandle		oldClipRgn = Memory_NewRegion();
	
	
	// get the arrangement of this control
	if ((oldClipRgn != null) && (inoutNewClipRegion != null))
	{
		// first determine the exact area in which to draw (the intersection
		// of the control region and the graphics ports visible region)
		GetClip(oldClipRgn);
		
		// now clip drawing in the port to the exact area where drawing should occur
		SetClip(inoutNewClipRegion);
		
		// call DeviceLoop() to make sure the window looks right no matter how many monitors it may cross
		DeviceLoop(inoutNewClipRegion, inProc, inDeviceLoopUserData, inDeviceLoopFlags);
		SetClip(oldClipRgn);
	}
	if (oldClipRgn != null) Memory_DisposeRegion(&oldClipRgn);
}// DeviceLoopClipAndDraw


/*!
If you want to draw a user pane control using a
DeviceLoop() drawing procedure, use this method.
The graphics port is automatically clipped to
the controls rectangle (or its even more specific
structure region, if an appropriate API from the
Carbon Control Manager is available).  Then, the
DeviceLoop() routine is called using the given
drawing procedure, passing a user data of
"inControl", and flags set to zero.  In other
words, by using this routine you gain the ability
to automatically set up the clipping properly, but
you have no control over the DeviceLoop() flags
and your drawing procedure *must* interpret its
user data as being of type "ControlRef".

(3.0)
*/
void
DeviceLoopClipAndDrawControl	(ControlRef				inControl,
								 DeviceLoopDrawingUPP	inProc)
{
	WindowRef		window = GetControlOwner(inControl);
	
	
	if ((window != null) && IsWindowVisible(window))
	{
		RgnHandle		newClipRgn = Memory_NewRegion();
		
		
		if (newClipRgn != null)
		{
			OSStatus	error = noErr;
			
			
			// first determine the exact area in which to draw (the intersection
			// of the control region and the graphics ports visible region)
			if (FlagManager_Test(kFlagCarbonControlManager))
			{
				// use this API, when available, to clip drawing to as precise an area as possible!
				error = GetControlRegion(inControl, kControlStructureMetaPart, newClipRgn);
			}
			else error = cfragNoSymbolErr;
			
			if (error != noErr)
			{
				Rect		controlBounds;
				
				
				error = noErr;
				GetControlBounds(inControl, &controlBounds);
				SetRectRgn(newClipRgn, controlBounds.left, controlBounds.top,
							controlBounds.right, controlBounds.bottom);
			}
			
			InsetRgn(newClipRgn, -1, -1);
			DeviceLoopClipAndDraw(newClipRgn, inProc, (long)inControl, 0L/* flags */);
			Memory_DisposeRegion(&newClipRgn);
		}
	}
}// DeviceLoopClipAndDrawControl


/*!
This method will read the specified 'hdlg' resource
(which must *not* have a resource ID identical to
any 'DLOG' or 'DITL' resource in your application
resource file) and display a help balloon for the
dialog item that is highest in the embedding
hierarchy beneath the mouse location given (that is,
the first control the mouse would touch if you
clicked there).

If the balloon to show turns out to be the same
balloon as most recently requested via a call to
this routine, this method has no effect and the
existing balloon will remain displayed until the
user moves the cursor outside the dialog items
rectangle.

See also DisplayDialogItemBalloon(), which this
routine calls.

(3.0)
*/
void
DisplayBalloonForDialogItemUnderMouse	(DialogRef const	inDialog,
										 Point const		inGlobalMouse,
										 SInt16 const		inResourceID_hdlg)
{
	GrafPtr				oldPort = null;
	ControlRef			control = null;
	SInt16				controlPart = kControlNoPart;
	Point				localMouse = inGlobalMouse;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(GetDialogWindow(inDialog));
	GlobalToLocal(&localMouse);
	control = FindControlUnderMouse(localMouse, GetDialogWindow(inDialog), &controlPart);
	if (control != null)
	{
		DialogItemIndex				item = FindControlDialogItemIndex(inDialog, control);
		static DialogItemIndex		oldItem = 0;
		static SInt16				oldValue = 0;
		static Boolean				oldActiveState = true;
		Boolean						isAnyBalloonShowing = false;
		
		
#if TARGET_API_MAC_CARBON
		isAnyBalloonShowing = false; // assume that Mac Help takes care of this
#else
		isAnyBalloonShowing = HMIsBalloon();
#endif
		if ((item != oldItem) ||
			(oldValue != GetControlValue(control)) ||
			(oldActiveState != IsControlActive(control)) ||
			(!isAnyBalloonShowing))
		{
			DisplayDialogItemBalloon(inDialog, item, inResourceID_hdlg);
		}
		oldItem = item;
		oldValue = GetControlValue(control);
		oldActiveState = IsControlActive(control);
	}
	SetPort(oldPort);
}// DisplayBalloonForDialogItemUnderMouse


/*!
This method will read the specified 'hdlg' resource
(which must *not* have a resource ID identical to
any 'DLOG' or 'DITL' resource in your application
resource file) and display a help balloon for the
specified dialog item.

This is used to implement help balloons in MacTelnets
resizeable dialog boxes, because the Dialog Manager
is not smart enough to tell the Help Manager to change
the hot rectangle of dialog items when they are
moved or resized.  This method has been designed to
read 'hdlg' resources, so you can still specify
balloon text in the same way as you would for a normal
dialog box.  However, you require a small bit of
additional code to support DisplayDialogItemBalloon().

You normally dont call this routine directly; you
instead use an event filter function to invoke the
DisplayBalloonForDialogItemUnderMouse() method,
described above.

IMPORTANT:	In order for this routine to work, you
			must specify an 'hdlg' resource that
			defines balloons for every dialog item
			in the specified dialog, starting from
			the first item.  The minimum restriction
			is that the first item in the 'hdlg'
			resource define the balloon for the
			dialog item with index 1.

(3.0)
*/
void
DisplayDialogItemBalloon	(DialogRef const		inDialog,
							 DialogItemIndex const	inItemIndex,
							 SInt16 const			inResourceID_hdlg)
{
#if TARGET_API_MAC_OS8
	ControlRef			control = null;
	
	
	if (IsWindowHilited(GetDialogWindow(inDialog))) // balloons should not appear for background dialog items
	{
		if (GetDialogItemAsControl(inDialog, inItemIndex, &control) == noErr)
		{
			HMMessageRecord		hmmr;
			Rect				hotRect;
			BalloonVariant		variant = kTopLeftTipPointsLeftVariant;
			SInt16				procID = 0;
			SInt16				messageState = kHMEnabledItem;
			UInt32				options = 0L;
			Point				tip;
			Boolean				doDisplay = true;
			SInt16				oldResourceFile = CurResFile();
			GrafPtr				oldPort = null;
			
			
			if (GetControlValue(control) == 1) messageState = kHMCheckedItem;
			if (GetControlValue(control) > 1) messageState = kHMOtherItem;
			if (!IsControlActive(control)) messageState = kHMDisabledItem;
			
			ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
			{
				SInt16		junkMessageCount = 0;
				
				
				(OSErr)HMGetIndHelpMsg(kHMDialogResType, inResourceID_hdlg, inItemIndex, messageState,
										&options, &tip, &hotRect, &procID, &variant, &hmmr, &junkMessageCount);
			}
			GetControlBounds(control, &hotRect); // just force the hot rectangle to be the controls current boundaries
			GetPort(&oldPort);
			SetPortWindowPort(GetDialogWindow(inDialog));
			{
				Point			topLeft;
				
				
				if (doDisplay)
				{
					// globalize the hot rectangle
					SetPt(&topLeft, hotRect.left, hotRect.top);
					LocalToGlobal(&topLeft);
					hotRect.right = topLeft.h + (hotRect.right - hotRect.left);
					hotRect.bottom = topLeft.v + (hotRect.bottom - hotRect.top);
					hotRect.left = topLeft.h;
					hotRect.top = topLeft.v;
					
					// the tip in the resource is relative to the dialog items location
					tip.h = hotRect.left + tip.h;
					tip.v = hotRect.top + tip.v;
				}
			}
			if (doDisplay)
			{
				(OSErr)HMShowBalloon(&hmmr, tip, &hotRect, (TipFunctionUPP)null, procID, variant, kHMRegularWindow);
			}
			SetPort(oldPort);
			UseResFile(oldResourceFile);
		}
	}
#else
	// With CARBON, Mac Help must be used instead of any calls from Balloons.h.  Unfortunately,
	// due to the architecture required to support Mac OS 8 and Balloon Help, the routine below
	// gets called every time a help tag is to be displayed, when in reality it only needs to be
	// called once for any dialog (and help tags are displayed automatically).
	(OSStatus)HMSetDialogHelpFromBalloonRsrc(inDialog, inResourceID_hdlg, 1/* starting item */);
#endif
}// DisplayDialogItemBalloon


/*!
To draw a dialog items control completely, use
this method.  Note that when you set an Appearance
controls data, the control is not updated: thus,
this routine may be helpful.

(3.0)
*/
void
DrawOneDialogItem	(DialogRef const		inDialog,
					 DialogItemIndex const	inItemIndex)
{
	ControlRef			control = null;
	
	
	if (GetDialogItemAsControl(inDialog, inItemIndex, &control) == noErr)
	{
		if (control != null) DrawOneControl(control);
	}
}// DrawOneDialogItem


/*!
To find the dialog item index of the item that
corresponds to the specified control, use this
method.  If the item cannot be found, 0 is
returned.

(3.0)
*/
DialogItemIndex
FindControlDialogItemIndex		(DialogRef const	inDialog,
								 ControlRef			inControl)
{
	DialogItemIndex		result = 0;
	ControlRef			test = null;
	register SInt16		i = 0;
	
	
	for (i = 1; i <= CountDITL(inDialog); i++)
	{
		if (GetDialogItemAsControl(inDialog, i, &test) == noErr) if (test == inControl)
		{
			result = i;
			break;
		}
	}
	return result;
}// FindControlDialogItemIndex


/*!
To simulate the clicking of a button (in
response to a key press or list double-click),
use this method.  This routine automatically
does nothing if the specified control is not
active.

In MacTelnet 3.0, this method will now determine
whether the specified item is a special dialog
item (such as the default button or cancel
button for the specified dialog).  If it is, the
proper theme sound is played under Mac OS 8.5 or
later.

If the button was allowed to flash, true is
returned; otherwise, false is returned.
*/
Boolean
FlashButton		(DialogRef			inDialog,
				 DialogItemIndex	inItemIndex)
{
	ControlRef		control = null;
	Boolean			result = false;
	
	
	if (inItemIndex > 0)
	{
		Boolean		isDefault = false,
					isCancel = false;
		
		
		if (GetDialogDefaultItem(inDialog) == inItemIndex) isDefault = true;
		if (GetDialogCancelItem(inDialog) == inItemIndex) isCancel = true;
		GetDialogItemAsControl(inDialog, inItemIndex, &control);
		result = FlashButtonControl(control, isDefault, isCancel);
	}
	return result;
}// FlashButton


/*!
To simulate the clicking of a button (in
response to a key press or list double-click),
use this method.  This routine automatically
does nothing if the specified control is not
active.

You can specify if a control is a default or
cancel button, and then this routine will
play the proper theme sound under Mac OS 8.5
or later.

If the button was allowed to flash, true is
returned; otherwise, false is returned.

WARNING:	Do not pass a control to this
			routine if it does not have any
			button part!
*/
Boolean
FlashButtonControl		(ControlRef			inControl,
						 Boolean			inIsDefaultButton,
						 Boolean			inIsCancelButton)
{
	Boolean			result = false;
	
	
	if (IsControlActive(inControl))
	{
		unsigned long	dummy = 0L;
		
		
		if (FlagManager_Test(kFlagAppearance1_1API))
		{
			// play sound when highlighting button
			if (inIsDefaultButton) PlayThemeSound(kThemeSoundDefaultButtonPress);
			else if (inIsCancelButton) PlayThemeSound(kThemeSoundCancelButtonPress);
			else PlayThemeSound(kThemeSoundButtonPress);
		}
		HiliteControl(inControl, kControlButtonPart);
		Delay(8, &dummy);
		HiliteControl(inControl, kControlNoPart);
		if (FlagManager_Test(kFlagAppearance1_1API))
		{
			// play sound when unhighlighting button
			if (inIsDefaultButton) PlayThemeSound(kThemeSoundDefaultButtonRelease);
			else if (inIsCancelButton) PlayThemeSound(kThemeSoundCancelButtonRelease);
			else PlayThemeSound(kThemeSoundButtonRelease);
		}
		result = true;
	}
	return result;
}// FlashButtonControl


/*!
This method will hide and show a dialog item
repeatedly for the specified number of
iterations, pausing the specified number of
ticks between each visibility change.  The
dialog item is guaranteed to end up visible,
so you shouldnt call this method for a dialog
item thats supposed to be hidden.

(3.0)
*/
void
FlashDialogItem		(DialogRef const		inDialog,
					 DialogItemIndex const	inItemIndex,
					 unsigned long const	inDelay,
					 short const			inCount)
{
	short			i = 0;
	unsigned long	dummy = 0L;
	ControlRef		control = null;
	
	
	DrawDialog(inDialog);	// avoids problems that occur if this method is called
								// as soon as a dialogs window is made visible
	GetDialogItemAsControl(inDialog, inItemIndex, &control);
	for (i = 0; i < inCount; i++)
	{
		HideDialogItem(inDialog, inItemIndex);
		Delay(inDelay, &dummy);
		ShowDialogItem(inDialog, inItemIndex);
		Draw1Control(control); // ensures that the control appears in time
		Delay(inDelay, &dummy);
	}
}// FlashDialogItem


/*!
Checks an unchecked check box dialog item, or
unchecks a checked or mixed-state check box
dialog item.

(2.6)
*/
void
FlipCheckBox	(DialogRef			inDialog,
				 DialogItemIndex	inItemIndex)
{
	SInt16		value = GetDialogItemValue(inDialog, inItemIndex);
	
	
	SetDialogItemValue(inDialog, inItemIndex,
						(value == kControlCheckBoxUncheckedValue)
									? kControlCheckBoxCheckedValue
									: kControlCheckBoxUncheckedValue);
}// FlipCheckBox


/*!
Highlights an unhighlighted radio button dialog item, or
unhighlights a highlighted or mixed-state radio button
dialog item.

(2.6)
*/
void
FlipRadioButton	(DialogRef			inDialog,
				 DialogItemIndex	inItemIndex)
{
	SInt16		value = GetDialogItemValue(inDialog, inItemIndex);
	
	
	SetDialogItemValue(inDialog, inItemIndex,
						(value == kControlRadioButtonUncheckedValue)
									? kControlRadioButtonCheckedValue
									: kControlRadioButtonUncheckedValue);
}// FlipRadioButton


/*!
A standard ControlUserPaneDrawUPP, this
method frames the region occupied by a
user pane, useful for debugging (or maybe
you really need it to draw rectangles for
you).

(3.0)
*/
pascal void
FrameUserPaneArea	(ControlRef			inControl,
					 SInt16				UNUSED_ARGUMENT(inPartCode))
{
	ColorPenState		state;
	RGBColor			solid;
	Rect				rect;
	
	
	ColorUtilities_GetColorAndPenState(&state);
	ColorUtilities_NormalizeColorAndPen();
	GetControlBounds(inControl, &rect);
	solid.red = solid.blue = solid.green = RGBCOLOR_INTENSITY_MAX;
	RGBBackColor(&solid);
	EraseRect(&rect);
	solid.red = solid.blue = solid.green = 0;
	RGBForeColor(&solid);
	FrameRect(&rect);
	ColorUtilities_SetColorAndPenState(&state);
}// FrameUserPaneArea


/*!
To convert the text in a text field or static
text control into a number, use this convenient
method.

(3.0)
*/
void
GetControlNumericalText		(ControlRef			inControl,
							 SInt32*			outNumberPtr)
{
	if (outNumberPtr != null)
	{
		Str255		string;
		
		
		GetControlText(inControl, string);
		StringToNum(string, outNumberPtr);
	}
}// GetControlNumericalText


/*!
To get a static text or editable text controls
text contents as a four-character code, use this
method.

(3.0)
*/
void
GetControlOSTypeText		(ControlRef		inControl,
							 OSType*		outTypePtr)
{
	if (outTypePtr != null)
	{
		Str15		typeString;
		Size		actualSize = 0L;
		
		
		PLstrcpy(typeString, EMPTY_PSTRING);
		GetControlText(inControl, typeString);
		typeString[0] = 4;
		BlockMove(&typeString[1], outTypePtr, sizeof(OSType));
	}
}// GetControlOSTypeText


/*!
To get a copy of the text in a text field control,
use this method.

(3.0)
*/
void
GetControlText	(ControlRef			inControl,
				 Str255				outText)
{
	if (outText != null)
	{
		Size		actualSize = 0;
		
		
		// obtain the data in the generic buffer
		(OSErr)GetControlData(inControl, kControlEditTextPart, kControlEditTextTextTag,
								255 * sizeof(unsigned char), (Ptr)(1 + outText), &actualSize);
		((Ptr)(outText))[0] = (char)actualSize; // mush in the length byte
	}
}// GetControlText


/*!
To get a copy of the text in a text field control,
even if the text is very long (i.e. longer than a
Pascal string), use this method.

You must dispose of the new handle yourself.  If
any problems occur, "null" is returned.

(3.0)
*/
Handle
GetControlTextAsNewHandle	(ControlRef			inControl)
{
	Handle		result = null;
	Size		actualSize = 0;
	
	
	// obtain the data in the generic buffer
	if (GetControlDataSize(inControl, kControlEditTextPart, kControlEditTextTextTag, &actualSize) == noErr)
	{
		result = Memory_NewHandle(actualSize);
		if (result != null)
		{
			(OSErr)GetControlData(inControl, kControlEditTextPart, kControlEditTextTextTag,
									GetHandleSize(result), *result, &actualSize);
		}
	}
	return result;
}// GetControlTextAsNewHandle


#if TARGET_API_MAC_OS8

/*!
To determine the current theme-savvy modal
Standard File dialog box, use this method.
If no dialog is showing, null is returned.

(3.0)
*/
DialogRef
GetCurrentThemeSavvyStandardFileModal	(void)
{
	return gCurrentThemeSavvyStandardFileModal;
}// GetCurrentThemeSavvyStandardFileModal

#endif


/*!
This method uses the GetControlData() API to
obtain the text of an editable text dialog
item that is actually a new Appearance control.

(3.0)
*/
void
GetDialogItemControlText	(DialogRef const			inDialog,
							 DialogItemIndex const		inItemIndex,
							 Str255						outText)
{
	ControlRef			control = null;
	
	
	(OSErr)GetDialogItemAsControl(inDialog, inItemIndex, &control);
	GetControlText(control, outText);
}// GetDialogItemControlText


/*!
Determines the value of a control, given
its dialog item information.

(3.0)
*/
SInt16
GetDialogItemValue	(DialogRef				inDialog,
					 DialogItemIndex		inItemIndex)
{
	ControlRef		control = null;
	SInt16			result = 0;
	OSErr			error = noErr;
	
	
	(OSErr)GetDialogItemAsControl(inDialog, inItemIndex, &control);
	result = GetControlValue(control);
	
	return result;
}// GetDialogItemValue


/*!
Returns inactive versions of the current
foreground and background colors of the
current graphics port.  You can use this
within custom control drawing procedures
to represent inactive states.

You can pass "null" in place of either
parameter if you are not interested in one
of the colors.

(3.0)
*/
void
GetInactiveColors	(RGBColor*		outInactiveForegroundColor,
					 RGBColor*		outInactiveBackgroundColor)
{
	RGBColor	white;
	
	
	if (outInactiveForegroundColor != null)
	{
		// set up foreground color
		white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
		GetForeColor(outInactiveForegroundColor);
		(Boolean)GetGray(GetMainDevice(), &white, outInactiveForegroundColor);
		white = *outInactiveForegroundColor;
		GetForeColor(outInactiveForegroundColor);
		(Boolean)GetGray(GetMainDevice(), &white, outInactiveForegroundColor);
		(Boolean)GetGray(GetMainDevice(), &white, outInactiveForegroundColor);
	}
	
	if (outInactiveBackgroundColor != null)
	{
		// set up background color
		white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
		GetBackColor(outInactiveBackgroundColor);
		(Boolean)GetGray(GetMainDevice(), &white, outInactiveBackgroundColor);
		white = *outInactiveBackgroundColor;
		GetBackColor(outInactiveBackgroundColor);
		(Boolean)GetGray(GetMainDevice(), &white, outInactiveBackgroundColor);
		(Boolean)GetGray(GetMainDevice(), &white, outInactiveBackgroundColor);
	}
}// GetInactiveColors


/*!
Returns lighter versions of the current foreground
and background colors of the current graphics port.
You can pass "null" in place of either parameter if
you are not interested in one of the colors.

You might use this to represent the inactive state of
something.

(3.0)
*/
void
GetLighterColors	(RGBColor*		outLighterForegroundColorOrNull,
					 RGBColor*		outLighterBackgroundColorOrNull)
{
	RGBColor	white;
	
	
	if (outLighterForegroundColorOrNull != null)
	{
		// set up foreground color
		white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
		GetForeColor(outLighterForegroundColorOrNull);
		(Boolean)GetGray(GetMainDevice(), &white, outLighterForegroundColorOrNull);
		white = *outLighterForegroundColorOrNull;
		GetForeColor(outLighterForegroundColorOrNull);
		(Boolean)GetGray(GetMainDevice(), &white, outLighterForegroundColorOrNull);
		(Boolean)GetGray(GetMainDevice(), &white, outLighterForegroundColorOrNull);
	}
	
	if (outLighterBackgroundColorOrNull != null)
	{
		// set up background color
		white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
		GetBackColor(outLighterBackgroundColorOrNull);
		(Boolean)GetGray(GetMainDevice(), &white, outLighterBackgroundColorOrNull);
		white = *outLighterBackgroundColorOrNull;
		GetBackColor(outLighterBackgroundColorOrNull);
		(Boolean)GetGray(GetMainDevice(), &white, outLighterBackgroundColorOrNull);
		(Boolean)GetGray(GetMainDevice(), &white, outLighterBackgroundColorOrNull);
	}
}// GetLighterColors


/*!
To create a new dialog from a 'DLOG', 'DITL'
and 'dlgx' resource having the specified ID,
use this method.  The dialog is returned,
and if a non-null reference to a Window Info
object is provided, a new Window Info object
is automatically created.  Dispose of the
Window Info object when you are finished
with the dialog by invoking the method
WindowInfo_Dispose().

(3.0)
*/
void
GetNewDialogWithWindowInfo		(SInt16 const		inDialogResourceID,
								 DialogRef*			outDialog,
								 WindowInfoRef*		outWindowInfoRefPtr)
{
	if (outDialog != null)
	{
		Cursors_DeferredUseWatch(30); // if it takes more than half a second to initialize, show the watch cursor
		
		*outDialog = GetNewDialog(inDialogResourceID, null, (WindowRef)-1);
		if (outWindowInfoRefPtr != null) *outWindowInfoRefPtr = WindowInfo_New();
		if (*outDialog != null)
		{
			// MacTelnet has a thread that takes care of this, so dont let the Dialog Manager do it
			SetDialogTracksCursor(*outDialog, false);
		}
	}
}// GetNewDialogWithWindowInfo


/*!
To automatically find the real front window,
ignoring all floating windows and dialog boxes,
use this method.  Use this instead of
FrontWindow() to locate a frontmost window that
could be behind a dialog box.

WARNING:	In order for this routine to work,
			ALL windows in the application must
			have reference constants that are
			either zero, or a reference to a
			Window Info object.

(3.0)
*/
WindowRef
GetRealFrontNonDialogWindow	(void)
{
	WindowRef		result = FrontWindow();
	WindowInfoRef	info = null;
	Boolean			done = false;
	
	
	while (result != null)
	{
		info = WindowInfo_GetFromWindow(result);
		if (WindowInfo_IsWindowFloating(info) || (GetWindowKind(result) == kDialogWindowKind))
		{
			result = GetNextWindow(result);
		}
		else break;
	}
	return result;
}// GetRealFrontNonDialogWindow


/*!
To automatically find the real front window,
ignoring all floating windows, use this method.
Use this instead of FrontWindow().

WARNING:	In order for this routine to work
			prior to Mac OS 8.6, ALL windows
			in the application must have
			reference constants that are either
			zero, or a reference to a standard
			Window Info object.

(3.0)
*/
WindowRef
GetRealFrontWindow	(void)
{
	WindowRef		result = null;
	
	
	if (FlagManager_Test(kFlagOS8_6API)) result = FrontNonFloatingWindow();
	else
	{
		WindowInfoRef	info = null;
		
		
		result = FrontWindow();
		while (result != null)
		{
			info = WindowInfo_GetFromWindow(result);
			if (!WindowInfo_IsWindowFloating(info)) break;
			result = GetNextWindow(result);
		}
	}
	return result;
}// GetRealFrontWindow


#if TARGET_API_MAC_OS8
/*!
This is an implementation of what will eventually
be a Carbon API, hence the TARGET_API_MAC_OS8
preprocessor directive.  The MacTelnet version of
this method will only be used when there is none
defined by Appearance with an identical function
signature.  Since this API is present in Carbon,
it is logical to assume that a future Appearance
control panel will allow some of these metrics to
be user-specified, and others to be specified by
themes (but dynamic nonetheless).  To prepare for
the future dynamics of theme metrics without
crippling legacy code, MacTelnet provides the code
for this routine whenever Carbon is not linked.

See "Appearance.h" for valid metric types.

(3.0)
*/
pascal OSStatus
GetThemeMetric		(ThemeMetric 	inMetric,
					 SInt32 *		outMetric)
{
	OSStatus		result = noErr;
	
	
	if (outMetric != null) switch (inMetric)
	{
		case kThemeMetricScrollBarWidth:
			*outMetric = 16; // pixels
			break;
		
		case kThemeMetricSmallScrollBarWidth:
			*outMetric = 13; // pixels
			break;
		
		case kThemeMetricCheckBoxGlyphHeight:
			*outMetric = 12; // pixels
			break;
		
		case kThemeMetricRadioButtonGlyphHeight:
			*outMetric = 12; // pixels
			break;
		
		default:
			break;
	}
	else result = memPCErr;
	
	return result;
}// GetThemeMetric
#endif

/*!
This routine determines whether help is
on, as specified by the user.  If the
user has Mac OS X, this flag determines
if Help Tags are on; otherwise, it
checks to see if Balloon Help is on.

(3.0)
*/
Boolean
IsPopUpHelpEnabled		(void)
{
	Boolean		result = false;
	
	
#if TARGET_API_MAC_CARBON
	result = HMAreHelpTagsDisplayed();
#else
	result = HMGetBalloons();
#endif
	return result;
}// IsPopUpHelpEnabled


/*!
Converts a region to global coordinates
by translating the origin of its
bounding box.

(2.6)
*/
void
LocalToGlobalRegion		(RgnHandle		inoutRegion)
{
	Point	where;
	
	
	SetPt(&where, 0, 0);
	LocalToGlobal(&where);
	OffsetRgn(inoutRegion, where.h, where.v);
}// LocalToGlobalRegion


/*!
This is a standard ControlKeyFilterUPP that will
prevent any non-digit characters from being typed
into a text field.

If there is a text selection in the specified
field, or if a special key (such as an arrow key
or the delete key) is pressed, this method will
not block it.

(3.0)
*/
pascal ControlKeyFilterResult
NumericalLimiter		(ControlRef			UNUSED_ARGUMENT(inControl),
						 SInt16*			inKeyCode,
						 SInt16*			inCharCode,
						 EventModifiers*	inModifiers)
{
	Size							actualSize = 0L;
	ControlKeyFilterResult			result = kControlKeyFilterPassKey;
	
	
	if (!(*inModifiers & cmdKey))
	{
		// make sure that any arrow or delete key press is allowed
		if ((*inKeyCode != 0x3B/* left arrow */) &&
			(*inKeyCode != 0x7B/* left arrow */) &&
			(*inKeyCode != 0x3E/* up arrow */) &&
			(*inKeyCode != 0x7E/* up arrow */) &&
			(*inKeyCode != 0x3C/* right arrow */) &&
			(*inKeyCode != 0x7C/* right arrow */) &&
			(*inKeyCode != 0x3D/* down arrow */) &&
			(*inKeyCode != 0x7D/* down arrow */) &&
			(*inKeyCode != 0x33/* delete */) &&
			(*inKeyCode != 0x75/* del */))
		{
			// if a character key was pressed, block it when there are four or more characters in the field already
			if (!isdigit(*inCharCode)) result = kControlKeyFilterBlockKey;
		}
	}
	else
	{
		// command key is down - scan for a Paste operation
		if ((*inCharCode == 'V') || (*inCharCode == 'v'))
		{
			// pasting isnt allowed because there is no easy way to filter out non-digits
			result = kControlKeyFilterBlockKey;
		}
	}
	return result;
}// NumericalLimiter


/*!
This is a standard ControlKeyFilterUPP that will
prevent more than 4 characters from being typed
into an OSType text field.

If there is a text selection in the specified
field, or if a special key (such as an arrow key
or the delete key) is pressed, this method will
not block it.

(3.0)
*/
pascal ControlKeyFilterResult
OSTypeLengthLimiter		(ControlRef			inControl,
						 SInt16*			inKeyCode,
						 SInt16*			inCharCode,
						 EventModifiers*	inModifiers)
{
	Size							actualSize = 0L;
	ControlKeyFilterResult			result = kControlKeyFilterPassKey;
	ControlEditTextSelectionRec		selectionRangeInfo;
	
	
	// determine if text is selected - if so, dont block any keys
	selectionRangeInfo.selStart = selectionRangeInfo.selEnd = 0;
	(OSErr)GetControlData(inControl, kControlEditTextPart, kControlEditTextSelectionTag,
							sizeof(selectionRangeInfo), &selectionRangeInfo, &actualSize);
	if (selectionRangeInfo.selStart == selectionRangeInfo.selEnd)
	{
		Str255		text;
		
		
		// determine the text currently in the control - if it is 4 characters long, block the key
		GetControlText(inControl, text);
		
		if (!(*inModifiers & cmdKey))
		{
			// make sure that any arrow or delete key press is allowed
			if ((*inKeyCode != 0x3B/* left arrow */) &&
				(*inKeyCode != 0x7B/* left arrow */) &&
				(*inKeyCode != 0x3E/* up arrow */) &&
				(*inKeyCode != 0x7E/* up arrow */) &&
				(*inKeyCode != 0x3C/* right arrow */) &&
				(*inKeyCode != 0x7C/* right arrow */) &&
				(*inKeyCode != 0x3D/* down arrow */) &&
				(*inKeyCode != 0x7D/* down arrow */) &&
				(*inKeyCode != 0x33/* delete */) &&
				(*inKeyCode != 0x75/* del */))
			{
				// if a character key was pressed, block it when there are four or more characters in the field already
				if (StrLength(text) >= 4) result = kControlKeyFilterBlockKey;
			}
		}
		else
		{
			// command key is down - scan for a Paste operation
			if ((*inCharCode == 'V') || (*inCharCode == 'v'))
			{
				// check the size of the TextEdit scrap
				if ((TEGetScrapLength() + StrLength(text) * sizeof(unsigned char)) > (4 * sizeof(unsigned char)))
				{
					// pasting isnt allowed because adding the scrap would make the field too big
					result = kControlKeyFilterBlockKey;
				}
			}
		}
	}
	return result;
}// OSTypeLengthLimiter


/*!
Removes the interior of a region so that
it ends up describing only its outline.

(2.6)
*/
void
OutlineRegion	(RgnHandle		inoutRegion)
{
	RgnHandle	tempRgn = Memory_NewRegion();
	
	
	if (tempRgn != null)
	{
		CopyRgn(inoutRegion, tempRgn);
		InsetRgn(tempRgn, 1, 1);
		DiffRgn(inoutRegion, tempRgn, inoutRegion);
		Memory_DisposeRegion(&tempRgn);
	}
}// OutlineRegion


/*!
To activate the root control of the frontmost
window and to change the cursor to the standard
arrow, call this method.  Usually, you do this
immediately after removing a dialog box from the
screen, especially if you previously used the
DeactivateFrontmostWindow() method.

(3.0)
*/
void
RestoreFrontmostWindow	(void)
{
	Embedding_RestoreFrontmostWindow();
}// RestoreFrontmostWindow


/*!
To select a new window only if the frontmost window
is not a dialog box, use this method.

(3.0)
*/
void
SelectBehindDialogWindows	(WindowRef		inWindow)
{
	if (GetWindowKind(GetRealFrontWindow()) != kDialogWindowKind)
	{
		SelectWindow(inWindow);
	}
}// SelectBehindDialogWindows


/*!
Displays a file selection dialog box that
shows the user only directories, and asks
the user to choose a default directory.

This has been updated in MacTelnet 3.0 to
support the Navigation Services dialog for
choosing directories, and to return a
Boolean value.  If the user selected a
directory, true is returned; otherwise
(if the user canceled or an error occurred)
false is returned.

(2.6)
*/
Boolean
SelectDirectory	(short*			inoutVolumeReferenceNumber,
				 long*			inoutDirectoryID)
{
	Boolean		result = false;
	Str255		prompt,
				title;
	
	
	GetIndString(prompt, rStringsNavigationServices, siNavPromptSelectDefaultDirectory);
	GetIndString(title, rStringsNavigationServices, siNavDialogTitleDefaultDirectory);
	
	// updated for Navigation Services
#if TARGET_API_MAC_OS8
	if (FlagManager_Test(kFlagNavigationServices))
#endif
	{
		FSSpec		file;
		OSErr		theErr = noErr;
		
		
		if (!Alert_ReportOSErr(theErr = Navigation_GetDirectory(prompt, title,
													kTelnetNavPrefKeyGenericChooseFolder,
													kNavDefaultNavDlogOptions,
													EventLoop_HandleNavigationUpdate, &file)))
		{
			// then we got a directory without a problem or a cancellation
			*inoutVolumeReferenceNumber = file.vRefNum;
			*inoutDirectoryID = file.parID;
		}
		result = (theErr == noErr);
	}
#if TARGET_API_MAC_OS8
	else
	{
		Point				where;
		StandardFileReply	sfr;
		
		
		//LMSetSFSaveDisk(*inoutVolumeReferenceNumber);
		//LMSetCurDirStore(*inoutDirectoryID);
		
		{
			Rect		screenRect;
			
			
			GetScreenBounds(&screenRect);
			where.h = screenRect.left + (screenRect.right - screenRect.left - 373) / 2;
			where.v = 120;
		}
		
		gSetDir = 0;					// clear directory found flag
		
		DeactivateFrontmostWindow();
		ThemeSavvyStandardGetDirectory(title, prompt, &sfr);
		RestoreFrontmostWindow();
		
		if (gSetDir)					// we got a directory
		{
			*inoutVolumeReferenceNumber = sfr.sfFile.vRefNum;
			*inoutDirectoryID = sfr.sfFile.parID;
			result = true;
		}
	}
#endif
	
	return result;
}// SelectDirectory


/*!
To select a new window, automatically finding
and unhighlighting the real front window and
ignoring all floating windows, use this method.
Use this instead of SelectWindow().

WARNING:	In order for this routine to work,
			ALL windows in the application must
			have reference constants that are
			either zero, or a reference to a
			Window Info object.

(3.0)
*/
void
SelectOverRealFrontWindow	(WindowRef		inWindow)
{
	// I give up, dammit
	SelectWindow(inWindow);
}// SelectOverRealFrontWindow


/*!
To fill in a text field or static text control
with a number, use this convenient method.
The specified signed number is first converted
to a string, and then the resultant string is
passed to SetControlText().

Use DrawOneControl() to update the control to
reflect the change.

(3.0)
*/
void
SetControlNumericalText		(ControlRef			inControl,
							 SInt32				inNumber)
{
	Str255		string;
	
	
	NumToString(inNumber, string);
	SetControlText(inControl, string);
}// SetControlNumericalText


/*!
To set a static text or editable text controls
text contents to represent the specified four-
character code, use this method.

Use DrawOneControl() to update the control to
reflect the change.

(3.0)
*/
void
SetControlOSTypeText		(ControlRef		inControl,
							 OSType const	inType)
{
	Str15		typeString;
	
	
	PLstrcpy(typeString, EMPTY_PSTRING);
	typeString[0] = 4;
	BlockMove(&inType, &typeString[1], sizeof(OSType));
	(OSErr)SetControlData(inControl, kControlEditTextPart, kControlEditTextTextTag,
							StrLength(typeString) * sizeof(unsigned char), typeString + 1/* length byte is skipped */);
}// SetControlOSTypeText


/*!
This method uses the SetControlData() API to
change the text of an editable text control.

Use DrawOneControl() to update the control
to reflect the change.

(3.0)
*/
void
SetControlText		(ControlRef			inControl,
					 Str255				inText)
{
	Ptr			stringBuffer = null;
	
	
	// create a disposable copy of the given text
	stringBuffer = Memory_NewPtr(StrLength(inText) * sizeof(unsigned char)); // we skip the length byte!
	BlockMove(inText + 1, stringBuffer, GetPtrSize(stringBuffer)); // buffer has no length byte
	
	// set the control text to be the copy that was just created
	(OSErr)SetControlData(inControl, kControlEditTextPart, kControlEditTextTextTag,
							GetPtrSize(stringBuffer), stringBuffer);
	
	// destroy the copy
	Memory_DisposePtr(&stringBuffer);
}// SetControlText


/*!
This method uses the SetControlData() API to
change the text of an editable text control.
Unlike SetControlText(), this method works
for any size of text buffer.

Use DrawOneControl() to update the control
to reflect the change.

(3.0)
*/
void
SetControlTextHandle	(ControlRef			inControl,
						 Handle				inText)
{
	// set the control text to be the copy that was just created
	(OSErr)SetControlData(inControl, kControlEditTextPart, kControlEditTextTextTag, GetHandleSize(inText), *inText);
}// SetControlTextHandle


/*!
To change the cursor based on the type of
control currently beneath the mouses
position at calling time, use this method.
The cursor will only change if the proper
cursor shape is different than the current
cursor shape.  The ID of the previous
cursor is returned.

(3.0)
*/
SInt16
SetCursorForControlUnderMouse		(WindowRef const		inWindow)
{
	SInt16			result = Cursors_GetCurrent();
	GrafPtr			oldPort = null;
	ControlRef		control = null;
	SInt16			controlPart = kControlNoPart;
	Point			localMouse;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(inWindow);
	GetMouse(&localMouse);
	control = FindControlUnderMouse(localMouse, inWindow, &controlPart);
#if TARGET_API_MAC_OS8
	switch (controlPart)
	{
		case kControlEditTextPart:
			result = Cursors_UseIBeam();
			break;
		
		case kControlIndicatorPart:
			result = Cursors_UseHandOpen();
			break;
		
		case kControlUpButtonPart:
		case kControlDownButtonPart:
		case kControlPageUpPart:
		case kControlPageDownPart:
			result = Cursors_UseHandPoint();
			break;
		
		case kControlMenuPart:
		case kControlTrianglePart:
		case kControlPicturePart:
		case kControlIconPart:
		case kControlClockPart:
		case kControlListBoxPart:
		case kControlListBoxDoubleClickPart:
		case kControlImageWellPart:
		case kControlRadioGroupPart:
		case kControlButtonPart: // kControlClockMinuteMonthPart is the same
		case kControlCheckBoxPart: // kControlRadioButtonPart and kControlClockSecondYearPart are the same
		case kControlClockHourDayPart:
		case kControlClockAMPMPart:
		default:
			result = Cursors_UseArrow();
			break;
	}
#else
	// under Carbon, controls can actually respond to cursor messages
	{
		Boolean			wasCursorSet = false;
		
		
		(OSStatus)HandleControlSetCursor(control, localMouse, EventLoop_CurrentModifiers(), &wasCursorSet);
	}
#endif
	SetPort(oldPort);
	
	return result;
}// SetCursorForControlUnderMouse


/*!
To automatically set the text of a static text
dialog item and truncate it to fit its boundaries,
use this method.

(3.0)
*/
void
SetDialogItemCenterTruncatedText	(DialogRef			inDialog,
									 DialogItemIndex	inItemIndex,
									 StringPtr			inString,
									 ThemeFontID		inThemeFontToUse)
{
	Rect				textItemRect;
	Handle				junk = null;
	DialogItemType		unusedType = 0;
	GrafPtr				oldPort = null;
	Str255				fontName;
	SInt16				preservedFontID = 0;
	SInt16				preservedFontSize = 0,
						fontSize = 0;
	Style				preservedFontStyle = 0,
						fontStyle = 0;
	OSErr				error = noErr;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(GetDialogWindow(inDialog));
	
	// preserve font settings, because they need to be changed
	{
		GrafPtr		port = null;
		
		
		GetPort(&port);
		preservedFontID = GetPortTextFont(port);
		preservedFontSize = GetPortTextSize(port);
		preservedFontStyle = GetPortTextFace(port);
	}
	
	// find or guess the font characteristics corresponding to the given theme font
	error = noErr;
	if (FlagManager_Test(kFlagAppearance1_1API))
	{
		if (!API_AVAILABLE(GetThemeFont)) error = cfragNoSymbolErr;
		else
		{
			error = GetThemeFont(inThemeFontToUse, 0/* script code */, fontName, &fontSize, &fontStyle);
		}
	}
	
	if ((!FlagManager_Test(kFlagAppearance1_1API)) || (error != noErr))
	{
		PLstrcpy(fontName, (inThemeFontToUse == kThemeSystemFont) ? "\pCharcoal" : "\pGeneva");
		fontSize = (inThemeFontToUse == kThemeSystemFont) ? 12 : 10;
		fontStyle = (inThemeFontToUse == kThemeSmallEmphasizedSystemFont) ? bold : normal;
	}
	
	// set the port font characteristics, required by StringUtilities_PTruncate()
	{
		SInt16		fontID = 0;
		
		
	#if TARGET_API_MAC_OS8
		GetFNum(fontName, &fontID);
	#else
		fontID = FMGetFontFamilyFromName(fontName);
	#endif
		TextFont(fontID);
	}
	TextSize(fontSize);
	
	// set the text of the specified control to use the truncated text
	GetDialogItem(inDialog, inItemIndex, &unusedType, &junk, &textItemRect);
	StringUtilities_PTruncate(inString, textItemRect.right - textItemRect.left, kStringUtilities_TruncateAtMiddle);
	SetDialogItemControlText(inDialog, inItemIndex, inString);
	
	// restore previous state
	TextFont(preservedFontID);
	TextSize(preservedFontSize);
	TextFace(preservedFontStyle);
	SetPort(oldPort);
}// SetDialogItemCenterTruncatedText


/*!
This method uses the SetControlData() API to
change the text of an editable text dialog
item that is actually a new Appearance control.

Use DrawOneDialogItem() to update the control
to reflect the change.

(3.0)
*/
void
SetDialogItemControlText	(DialogRef const			inDialog,
							 DialogItemIndex const		inItemIndex,
							 StringPtr const			inText)
{
	ControlRef			control = null;
	
	
	GetDialogItemAsControl(inDialog, inItemIndex, &control);
	SetControlText(control, inText);
}// SetDialogItemControlText


/*!
This method uses the SetControlData() API to
change the selected text of an editable text
dialog item that is actually a new Appearance
control.

Use DrawOneDialogItem() to update the control
to reflect the change.

WARNING: This method does not work properly yet.

(3.0)
*/
void
SetDialogItemControlTextSelection	(DialogRef const			inDialog,
									 DialogItemIndex const		inItemIndex,
									 SInt16 const				inStart,
									 SInt16 const				inEnd)
{
	ControlRef						control = null;
	ControlEditTextSelectionRec		selectionRangeInfo;
	
	
	// put the given information into an OS-savvy format
	selectionRangeInfo.selStart = inStart;
	selectionRangeInfo.selEnd = inEnd;
	
	// set the control text selection range
	GetDialogItemAsControl(inDialog, inItemIndex, &control);
	(OSErr)SetControlData(control, kControlEditTextPart, kControlEditTextSelectionTag,
							sizeof(selectionRangeInfo), (Ptr)&selectionRangeInfo);
}// SetDialogItemControlTextSelection


/*!
Changes the value of a control, given its
dialog item information.

(3.0)
*/
void
SetDialogItemValue	(DialogRef				inDialog,
					 DialogItemIndex		inItemIndex,
					 SInt16					inValue)
{
	ControlRef		control = null;
	OSErr			error = noErr;
	
	
	error = GetDialogItemAsControl(inDialog, inItemIndex, &control);
	if (error == noErr) SetControlValue(control, inValue);
	else BEEP;
}// SetDialogItemValue


/*!
Call this routine on the Help button control of
every window.  This routine checks to see if all
necessary components for contextual help are
available, and changes the icon of the button to
be a high-quality icon (if possible).

(3.0)
*/
void
SetUpHelpButtonControl	(ControlRef const		inControl)
{
	useBestHelpButtonIcon(inControl);
	if (!FlagManager_Test(kFlagAppleGuideFiles))
	{
		// then Help is not available, so deactivate the button
		(OSErr)DeactivateControl(inControl);
	}
}// SetUpHelpButtonDialogItem


/*!
Call this routine on the Help button dialog item
of every dialog box.  This routine checks to see
if all necessary components for contextual help
are available, and changes the icon of the button
to be a high-quality icon (if possible).

(3.0)
*/
void
SetUpHelpButtonDialogItem	(DialogRef const		inDialog,
							 DialogItemIndex const	inHelpButtonIndex)
{
	ControlRef		helpButtonControl = null;
	
	
	(OSErr)GetDialogItemAsControl(inDialog, inHelpButtonIndex, &helpButtonControl);
	SetUpHelpButtonControl(helpButtonControl);
}// SetUpHelpButtonDialogItem


/*!
This standard alert box event filter will:
- recognize 'Return', 'Enter' and 'spacebar' keys as
  a click in the default button of the alert
- recognize 'Esc' and '-.' as a click in the button
  in the Cancel position
- recognize 'Help' and '-?' as a click in the button
  in the help position
- recognize 'tab' to change the default button,
  and 'Shift-tab' to reverse the direction of change
- handle update events in MacTelnet windows, so movable
  alerts dont obliterate anything

(3.0)
*/
pascal Boolean
StandardAlertEventFilter	(DialogRef			inDialog,
							 EventRecord*		inoutEventPtr,
							 DialogItemIndex*	outItemIndex)
{
	Boolean		result = pascal_false, // events not handled by default
				isButtonAndShouldFlash = false, // simplifies handling of multiple key-to-button maps
				commandDown = ((inoutEventPtr->modifiers & cmdKey) != 0),
				optionDown = ((inoutEventPtr->modifiers & optionKey) != 0),
				controlDown = ((inoutEventPtr->modifiers & controlKey) != 0),
				shiftDown = ((inoutEventPtr->modifiers & shiftKey) != 0);
	
	
	// allows network events, etc. to be serviced
	NetEvents_Service();
	Threads_YieldToAnother();
	
	// check for tab or Shift-tab events, and switch the dialog default button if possible
	{
		DialogItemIndex		buttonIndices[3];
		
		
		buttonIndices[0] = kAlert_ItemButton1;
		buttonIndices[1] = kAlert_ItemButton2;
		buttonIndices[2] = kAlert_ItemButton3;
		result = SwitchDialogDefaultItem(inDialog, inoutEventPtr, buttonIndices, 3);
	}
	
	if (!result) switch (inoutEventPtr->what)
	{
		case kHighLevelEvent:
			(void)AEProcessAppleEvent(inoutEventPtr);
			break;
		
		case updateEvt:
			result = EventLoop_HandleEvent(inoutEventPtr);
			break;
		
		case keyDown:
			{
				unsigned char		charASCII = '\0',
									charCode = '\0';
				DialogItemIndex		pendingItem = 0;
				Boolean				usePendingItem = false;
				ControlRef			control = null;
				
				
				charASCII = (inoutEventPtr->message & charCodeMask);
				charCode = ((inoutEventPtr->message & keyCodeMask) >> 8);
				
				if (charASCII == kHelpCharCode) // then the Help key was pressed, or some equivalent
				{
					pendingItem = kAlertStdAlertHelpButton;
					(OSErr)GetDialogItemAsControl(inDialog, pendingItem, &control);
					usePendingItem = IsControlVisible(control); // only handle the Help key if theres help!
					result = pascal_true; // dont let the standard filter see this either way!
				}
				else if ((charCode == 0x31) || (charCode == 0x24) || (charCode == 0x4C))
				{
					// A default button click is a modifier-free press
					// of the spacebar, Enter or Return keys.
					pendingItem = GetDialogDefaultItem(inDialog);
					if ((!commandDown) && (!optionDown) && (!controlDown) && (!shiftDown))
					{
						(OSErr)GetDialogItemAsControl(inDialog, pendingItem, &control);
						usePendingItem = IsControlVisible(control); // only handle the default button if there is one!
						result = pascal_true; // dont let the standard filter see this either way!
					}
				}
				else if ((charCode == 0x35) || (charASCII == '.'))
				{
					// A Cancel button click is a modifier-free press
					// of the Escape key or the -. key combination.
					pendingItem = GetDialogCancelItem(inDialog);
					if (((!commandDown) && (!optionDown) && (!controlDown) && (!shiftDown) && (charCode == 0x35)) ||
						((commandDown) && (!optionDown) && (!controlDown) && (!shiftDown) && (charASCII == '.')))
					{
						(OSErr)GetDialogItemAsControl(inDialog, pendingItem, &control);
						usePendingItem = IsControlVisible(control); // only handle the Cancel button if there is one!
						result = pascal_true; // dont let the standard filter see this either way!
					}
				}
				
				if (usePendingItem)
				{
					*outItemIndex = pendingItem;
					isButtonAndShouldFlash = result = pascal_true;
				}
				
				// if the event was not handled at this point, pass control to the standard Mac filter
				if (!result) result = StdFilterProc(inDialog, inoutEventPtr, outItemIndex);
			}
			break;
		
		case mouseDown:
			{
				// check for resize events: some dialogs are resizable
				WindowRef	window = null;
				SInt16		part = FindWindow(inoutEventPtr->where, &window);
				
				
				if (window == GetDialogWindow(inDialog))
				{
					// then its this dialog thats being touched
					if (IS_SHOW_CONTEXTUAL_MENU_EVENT(inoutEventPtr))
					{
						(OSStatus)ContextualMenuBuilder_HandleWindow(GetDialogWindow(inDialog), inoutEventPtr, part);
						result = true;
					}
				}
				else
				{
					if ((part == inDrag) && (inoutEventPtr->modifiers & cmdKey))
					{
						(Boolean)EventLoop_HandleEvent(inoutEventPtr);
						result = true;
					}
				}
			}
			break;
		
		default:
			break;
	}
	if (isButtonAndShouldFlash) FlashButton(inDialog, *outItemIndex);
	return result;
}// StandardAlertEventFilter


/*!
The modernized, standard dialog event filter
for the rewritten dialog boxes.  This method
handles update events as well as the standard
key equivalents for OK- and Cancel-type buttons.

(3.0)
*/
pascal Boolean
StandardDialogEventFilter	(DialogRef			inDialog,
							 EventRecord*		inoutEventPtr,
							 DialogItemIndex*	outItemIndex)
{
	Boolean			result = pascal_false;
	GrafPtr			oldPort = null;
	
	
	// allows network events, etc. to be serviced
	NetEvents_Service();
	Threads_YieldToAnother();
	
	GetPort(&oldPort);
	SetPortWindowPort(GetDialogWindow(inDialog));
	
	// change the cursor if necessary
	SetCursorForControlUnderMouse(GetDialogWindow(inDialog));
	
	// handle Text Services Manager events
	if (FlagManager_Test(kFlagTextServicesManager))
	{
#if TARGET_API_MAC_OS8
		if (TSMEvent(inoutEventPtr)) result = pascal_true;
		else result = pascal_false;
#else
		result = pascal_false; // Carbon handles these events implicitly, so ignore them
#endif
	}
	
	if (!result) switch (inoutEventPtr->what)
	{
		case kHighLevelEvent:
			(void)AEProcessAppleEvent(inoutEventPtr);
			break;
		
		case updateEvt:
			result = EventLoop_HandleEvent(inoutEventPtr);
			break;
		
		case keyDown:
			{
				short		keyCode = 0,
							key = 0;
				
				
				key = inoutEventPtr->message & charCodeMask;
				keyCode = (inoutEventPtr->message >> 8) & 0xff;
				
				// check for key presses: 0x0d == return,
				// 0x03 == enter, 0x35 == escape
				if ((GetDialogDefaultItem(inDialog) > 0) &&
					((key == 0x0d) || (key == 0x03)))
				{
					if (FlashButton(inDialog, GetDialogDefaultItem(inDialog)))
					{
						*outItemIndex = GetDialogDefaultItem(inDialog);
						result = pascal_true;
					}
				}
				else if ((GetDialogCancelItem(inDialog) > 0) &&
					(((key == '.') && (inoutEventPtr->modifiers & cmdKey)) ||
					((key == 0x1b) && (keyCode == 0x35))))
				{
					if (FlashButton(inDialog, GetDialogCancelItem(inDialog)))
					{
						*outItemIndex = GetDialogCancelItem(inDialog);
						result = pascal_true;
					}
				}
			}
			break;
		
		case mouseDown:
			{
				// check for resize events: some dialogs are resizable
				WindowRef	window = null;
				SInt16		part = FindWindow(inoutEventPtr->where, &window);
				
				
				if (window == GetDialogWindow(inDialog))
				{
					// then its this dialog thats being touched
					if (part == inGrow)
					{
						WindowInfoRef		windowFeaturesRef = null;
						
						
						// the auxiliary structure must exist in order to get the sizing information
						if ((windowFeaturesRef = WindowInfo_GetFromDialog(inDialog)) != null)
						{
							WindowInfo_GrowWindow(GetDialogWindow(inDialog), inoutEventPtr);
						}
						else BEEP;
					}
					else if (IS_SHOW_CONTEXTUAL_MENU_EVENT(inoutEventPtr))
					{
						(OSStatus)ContextualMenuBuilder_HandleWindow(GetDialogWindow(inDialog), inoutEventPtr, part);
						result = pascal_true;
					}
				}
				else
				{
					if ((part == inDrag) && (inoutEventPtr->modifiers & cmdKey))
					{
						(Boolean)EventLoop_HandleEvent(inoutEventPtr);
						result = pascal_true;
					}
				}
			}
			break;
		
		default:
			break;
	}
	
	if (!result) result = StdFilterProc(inDialog, inoutEventPtr, outItemIndex);
	
	SetPort(oldPort);
	return result;
}// StandardDialogEventFilter


/*!
This method will take the given event information
for the specified dialog and see if a 'tab' or
'Shift-tab' event occurred.  If one of these
events took place, then the default item for the
dialog may be changed, based on the list of items
given, the number of items, and the visibility and
enabled states of those items.

Since 'tab' is reserved for dialog box focus
switching, this method is only called for alerts.

(3.0)
*/
Boolean
SwitchDialogDefaultItem	(DialogRef const		inDialog,
						 EventRecord const*		inEventPtr,
						 DialogItemIndex const	inDefaultableItemIndices[],
						 short const			inDefaultableItemIndicesArraySize)
{
	Boolean				result = false, // nothing handled initially
						commandDown = ((inEventPtr->modifiers & cmdKey) != 0),
						optionDown = ((inEventPtr->modifiers & optionKey) != 0),
						controlDown = ((inEventPtr->modifiers & controlKey) != 0),
						shiftDown = ((inEventPtr->modifiers & shiftKey) != 0);
	DialogItemIndex		i = 0;
	
	
	if (inDefaultableItemIndicesArraySize > 1) // cant switch items unless there are at least two!
	{
		switch (inEventPtr->what)
		{
			case keyDown:
			case autoKey:
				{
					unsigned char		charASCII = '\0',
										charCode = '\0';
					
					
					charASCII = (inEventPtr->message & charCodeMask);
					charCode = ((inEventPtr->message & keyCodeMask) >> 8);
					
					if (charASCII == kTabCharCode) // then the Tab key was pressed
					{
						ControlRef			control = null;
						DialogItemIndex		defaultItemIndex = GetDialogDefaultItem(inDialog);
						Boolean				done = false;
						short				start = 0;
						
						
						for (i = 0; i < inDefaultableItemIndicesArraySize; i++)
						{
							if (inDefaultableItemIndices[i] == defaultItemIndex)
							{
								start = i;
								break;
							}
						}
						for (i = start; (!done) && (!result); )
						{
							// set the next item based on whether its forward or reverse order
							DialogItemIndex		nextItem = 0;
							short				oldIndex = i;
							
							
							i = (!shiftDown)
								?	(
										((i + 1) < inDefaultableItemIndicesArraySize)
											? (i + 1)
											: 0
									)
								:	(
										((i - 1) >= 0)
											? (i - 1)
											: inDefaultableItemIndicesArraySize - 1
									);
							nextItem = inDefaultableItemIndices[i];
							
							// now change the dialog default item
							(OSErr)GetDialogItemAsControl(inDialog, nextItem, &control);
							if (control != null)
							{
								if (IsControlVisible(control) && IsControlActive(control))
								{
									if (oldIndex != i)
									{
										(OSErr)SetDialogDefaultItem(inDialog, nextItem);
									}
									result = true;
								}
							}
							
							done = (i == start); // back to where we started?
						}
					}
				}
				break;
			
			default:
				break;
		}
	}
	return result;
}// SwitchDialogDefaultItem


/*!
To change the font of the current graphics port
to be the font with the specified name, use this
method.

This routine is present for assisting Carbon
compatibility: Apple does not recommend using
GetFNum(), so the fewer places font IDs are used,
the better.

(3.0)
*/
void
TextFontByName		(StringPtr		inFontName)
{
	SInt16		fontID = 0;
	
	
#if TARGET_API_MAC_OS8
	GetFNum(inFontName, &fontID);
#else
	fontID = FMGetFontFamilyFromName(inFontName);
#endif
	TextFont(fontID);
}// TextFontByName


/*!
To use a theme-savvy version of the Standard File
dialog box when you would ordinarily call
StandardGetFile(), use this routine.  This routine
improves upon StandardGetFile() by allowing you to
specify a damned text prompt!

Note that Navigation Services is the preferred way to
acquire file information from the user: use this routine
only if Navigation Services is not available.

(3.0)
*/
void
ThemeSavvyStandardGetDirectory	(ConstStr255Param		inTitle,
								 ConstStr255Param 		inPrompt,
								 StandardFileReply*		outReplyRecordPtr)
{
#if TARGET_API_MAC_OS8
	WindowRef	uselessWindow = newWorthlessWindow(); // avoids strange endless-looping bug when no windows are open
	Point		where;
	short		tabSelectableItemsInOrder[] = { sfItemFileListUser };
	
	
	{
		Rect	screenRect;
		
		
		GetScreenBounds(&screenRect);
		where.h = screenRect.left + (screenRect.right - screenRect.left - 373) / 2;
		where.v = 110;
	}
	PLstrcpy(gThemeSavvyDialogTextPrompt, inPrompt);
	CustomGetFile(NewFileFilterYDProc(themeSavvyStandardGetDirectoryFileFilter), -1, null,
					outReplyRecordPtr, kThemeSavvyStandardGetDirectoryDLOG, where,
					NewDlgHookYDProc(themeSavvyStandardGetDirectoryDialogHook),
					NewModalFilterYDProc(themeSavvyStandardGetDirectoryModalFilter),
					null,//(ActivationOrderListPtr)tabSelectableItemsInOrder,
					(ActivateYDUPP)null, (void*)inTitle/* used to set the window title in the hook routine */);
	DisposeWindow(uselessWindow);
#else
	// this should never happen
	BEEP;
#endif
}// ThemeSavvyStandardGetDirectory


/*!
To use a theme-savvy version of the Standard File
dialog box when you would ordinarily call
StandardGetFile(), use this routine.  This routine
improves upon StandardGetFile() by allowing you to
specify a damned text prompt!

Note that Navigation Services is the preferred way to
acquire file information from the user: use this routine
only if Navigation Services is not available.

(3.0)
*/
void
ThemeSavvyStandardGetFile	(ConstStr255Param		inTitle,
							 ConstStr255Param 		inPrompt,
							 FileFilterUPP 			inFileFilterUPP,
							 short 					inTypeListElementCount,
							 ConstSFTypeListPtr 	inTypeListPtr,
							 StandardFileReply*		outReplyRecordPtr)
{
#if TARGET_API_MAC_OS8
	WindowRef	uselessWindow = newWorthlessWindow(); // avoids strange endless-looping bug when no windows are open
	Point		where;
	short		tabSelectableItemsInOrder[] = { sfItemFileListUser };
	
	
	{
		Rect	screenRect;
		
		
		GetScreenBounds(&screenRect);
		where.h = screenRect.left + (screenRect.right - screenRect.left - 373) / 2;
		where.v = 110;
	}
	PLstrcpy(gThemeSavvyDialogTextPrompt, inPrompt);
	gThemeSavvyGetFileFileFilterUPP = inFileFilterUPP;
	CustomGetFile(NewFileFilterYDProc(themeSavvyStandardGetFileFileFilter), inTypeListElementCount, inTypeListPtr,
					outReplyRecordPtr, kThemeSavvyStandardGetFileDLOG, where,
					NewDlgHookYDProc(themeSavvyStandardGetFileDialogHook),
					NewModalFilterYDProc(themeSavvyStandardGetFileModalFilter),
					null,//(ActivationOrderListPtr)tabSelectableItemsInOrder,
					NewActivateYDProc(themeSavvyStandardPutFileActivate), // works for GetFile dialogs, too
					(void*)inTitle/* used to set the window title in the hook routine */);
	DisposeWindow(uselessWindow);
#else
	// this should never happen
	BEEP;
#endif
}// ThemeSavvyStandardGetFile


/*!
To use a theme-savvy version of the Standard File
dialog box when you would ordinarily call
StandardPutFile(), use this routine.

Note that Navigation Services is the preferred way to
acquire file information from the user: use this routine
only if Navigation Services is not available.

(3.0)
*/
void
ThemeSavvyStandardPutFile	(ConstStr255Param		inTitle,
							 ConstStr255Param 		inPrompt,
							 ConstStr255Param 		inOriginalName,
							 StandardFileReply*		outReplyRecordPtr)
{
#if TARGET_API_MAC_OS8
	WindowRef	uselessWindow = newWorthlessWindow(); // avoids strange endless-looping bug when no windows are open
	Point		where;
	short		tabSelectableItemsInOrder[] = { sfItemFileNameTextEdit, sfItemFileListUser };
	
	
	{
		Rect	screenRect;
		
		
		GetScreenBounds(&screenRect);
		where.h = screenRect.left + (screenRect.right - screenRect.left - 373) / 2;
		where.v = 110;
	}
	//PLstrcpy(gThemeSavvyDialogTextPrompt, inPrompt);
	CustomPutFile(inPrompt, inOriginalName, outReplyRecordPtr, kThemeSavvyStandardPutFileDLOG,
					where, NewDlgHookYDProc(themeSavvyStandardPutFileDialogHook),
					NewModalFilterYDProc(themeSavvyStandardPutFileModalFilter),
					null,//(ActivationOrderListPtr)tabSelectableItemsInOrder,
					NewActivateYDProc(themeSavvyStandardPutFileActivate),
					(void*)inTitle/* used to set the window title in the hook routine */);
	DisposeWindow(uselessWindow);
#else
	// this should never happen
	BEEP;
#endif
}// ThemeSavvyStandardPutFile


/*!
Use this method to show dialog windows.  For
Mac OS 8.5 users, this routine will display a
window transition and play the appropriate
theme sound.

(3.0)
*/
void
TransitionDialogDisplay		(DialogRef const		inDialog)
{
	dialogDisplayPreparation(inDialog);
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
		(OSStatus)TransitionWindow(GetDialogWindow(inDialog), kWindowZoomTransitionEffect,
									kWindowShowTransitionAction, null);
	}
	dialogDisplayEnsureVisibility(inDialog);
}// TransitionDialogDisplay


/*!
Use this method to show dialog windows that are
children of other windows.  For Mac OS 8.5 users,
this routine will display a window transition that
comes from the center of the control rectangle
of the specified control in its owning window, and
will play the appropriate theme sound.

If you pass "null" for "inFromWhat", this method
invokes TransitionDialogDisplay().

(3.0)
*/
void
TransitionDialogDisplayFromControl		(DialogRef const		inDialog,
										 ControlRef const		inFromWhat)
{
	if (inFromWhat == null) TransitionDialogDisplay(inDialog);
	else
	{
		dialogDisplayPreparation(inDialog);
		if (FlagManager_Test(kFlagCarbonWindowManager))
		{
			Rect		bounds;
			Point		topLeft;
			
			
			GetControlBounds(inFromWhat, &bounds);
			SetPt(&topLeft, bounds.left, bounds.top);
			{
				GrafPtr		oldPort = null;
				
				
				GetPort(&oldPort);
				SetPortWindowPort(GetControlOwner(inFromWhat));
				LocalToGlobal(&topLeft);
				bounds.left = bounds.right = topLeft.h + INTEGER_HALVED(bounds.right - bounds.left);
				bounds.top = bounds.bottom = topLeft.v + INTEGER_HALVED(bounds.bottom - bounds.top);
				SetPort(oldPort);
			}
			
			(OSStatus)TransitionWindow(GetDialogWindow(inDialog), kWindowZoomTransitionEffect,
										kWindowShowTransitionAction, &bounds);
		}
		dialogDisplayEnsureVisibility(inDialog);
	}
}// TransitionDialogDisplayFromControl


/*!
Use this method to show dialog windows that are
children of other windows.  For Mac OS 8.5 users,
this routine will display a window transition that
comes from the specified window region, and will
play the appropriate theme sound.

(3.0)
*/
void
TransitionDialogDisplayFromWindowRegion	(DialogRef const		inDialog,
										 WindowRef const		inParentWindow,
										 WindowRegionCode const	inFromWhere)
{
	dialogDisplayPreparation(inDialog);
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
		Rect		bounds;
		
		
		MacLook_GetWindowRegionBounds(inParentWindow, inFromWhere, &bounds);
		(OSStatus)TransitionWindow(GetDialogWindow(inDialog), kWindowZoomTransitionEffect,
									kWindowShowTransitionAction, &bounds);
	}
	dialogDisplayEnsureVisibility(inDialog);
}// TransitionDialogDisplayFromWindowRegion


/*!
Use this method to destroy dialog windows.  For
Mac OS 8.5 users, this routine will display a
window transition and play the appropriate
theme sound.

(3.0)
*/
void
TransitionDialogClose	(DialogRef const		inDialog)
{
	HideWindow(GetDialogWindow(inDialog));
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
		(OSStatus)TransitionWindow(GetDialogWindow(inDialog), kWindowZoomTransitionEffect,
									kWindowHideTransitionAction, null);
	}
	dialogClose(inDialog);
}// TransitionDialogClose


/*!
Use this method to destroy dialog windows while
displaying a zoom transition from the windows area
to the lower-right-hand corner of the screen.  For
users of Mac OS 8.5, this routine will display a
window transition and play the appropriate theme
sound; otherwise, the window is simply destroyed.

(3.0)
*/
void
TransitionDialogCloseLowerRight	(DialogRef const		inDialog)
{
	HideWindow(GetDialogWindow(inDialog));
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
		Rect		lowerRight,
					screenRect;
		
		
		MacLook_GetWindowDeviceGrayRect(GetDialogWindow(inDialog), &screenRect);
		SetRect(&lowerRight, screenRect.right, screenRect.bottom, screenRect.right, screenRect.bottom);
		(OSStatus)TransitionWindow(GetDialogWindow(inDialog), kWindowZoomTransitionEffect,
									kWindowHideTransitionAction, &lowerRight);
	}
	dialogClose(inDialog);
}// TransitionDialogCloseLowerRight


/*!
Use this method to close dialog windows that are
children of other windows.  For Mac OS 8.5 users,
this routine will display a window transition that
comes to the center of the control rectangle of
the specified control in its owning window, and
will play the appropriate theme sound.

If you pass "null" for "inToWhat", this method
invokes TransitionDialogClose().

(3.0)
*/
void
TransitionDialogCloseToControl		(DialogRef const		inDialog,
									 ControlRef const		inToWhat)
{
	if (inToWhat == null) TransitionDialogClose(inDialog);
	else
	{
		if (FlagManager_Test(kFlagCarbonWindowManager))
		{
			Rect		bounds;
			Point		topLeft;
			
			
			GetControlBounds(inToWhat, &bounds);
			SetPt(&topLeft, bounds.left, bounds.top);
			{
				GrafPtr		oldPort = null;
				
				
				GetPort(&oldPort);
				SetPortWindowPort(GetControlOwner(inToWhat));
				LocalToGlobal(&topLeft);
				bounds.left = bounds.right = topLeft.h + INTEGER_HALVED(bounds.right - bounds.left);
				bounds.top = bounds.bottom = topLeft.v + INTEGER_HALVED(bounds.bottom - bounds.top);
				SetPort(oldPort);
			}
			
			(OSStatus)TransitionWindow(GetDialogWindow(inDialog), kWindowZoomTransitionEffect,
										kWindowHideTransitionAction, &bounds);
		}
		dialogClose(inDialog);
	}
}// TransitionDialogCloseToControl


/*!
Use this method to hide dialog windows that are
children of other windows.  For Mac OS 8.5 users,
this routine will display a window transition and
play the appropriate theme sound.

(3.0)
*/
void
TransitionDialogCloseToWindowRegion		(DialogRef const			inDialog,
										 WindowRef const			inParentWindow,
										 WindowRegionCode const		inToWhere)
{
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
		Rect		bounds;
		OSStatus	status = 0L;
		
		
		MacLook_GetWindowRegionBounds(inParentWindow, inToWhere, &bounds);
		(OSStatus)TransitionWindow(GetDialogWindow(inDialog), kWindowZoomTransitionEffect,
									kWindowHideTransitionAction, &bounds);
	}
	dialogClose(inDialog);
}// TransitionDialogCloseToWindowRegion


/*!
Use this method to destroy dialog windows while
displaying a zoom transition from the windows area
to the upper-left-hand corner of the screen.  For
users of Mac OS 8.5, this routine will display a
window transition and play the appropriate theme
sound; otherwise, the window is simply destroyed.

(3.0)
*/
void
TransitionDialogCloseUpperLeft	(DialogRef const		inDialog)
{
	HideWindow(GetDialogWindow(inDialog));
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
		Rect		upperLeft;
		
		
		SetRect(&upperLeft, 0, 0, 0, 0);
		(OSStatus)TransitionWindow(GetDialogWindow(inDialog), kWindowZoomTransitionEffect,
									kWindowHideTransitionAction, &upperLeft);
	}
	dialogClose(inDialog);
}// TransitionDialogCloseUpperLeft


/*!
Uses GetInactiveColors() to determine the
dimmed versions of the current foreground
and background colors of the current graphics
port, and makes the inactive versions the
current foreground and background colors.

This routine might be used inside a drawing
procedure for a custom control, to represent
an inactive state.

(3.0)
*/
void
UseInactiveColors	(void)
{
	RGBColor		foreground,
					background;
	
	
	GetInactiveColors(&foreground, &background);
	RGBForeColor(&foreground);
	RGBBackColor(&background);
}// UseInactiveColors


/*!
Uses GetLighterColors() to determine the lighter
versions of the current foreground and background
colors of the current graphics port, and makes the
lighter versions the current foreground and
background colors.

Do not use this routine to represent an inactive
state; for that, call UseInactiveColors().

(3.0)
*/
void
UseLighterColors	(void)
{
	RGBColor	foreground,
				background;
	
	
	GetLighterColors(&foreground, &background);
	RGBForeColor(&foreground);
	RGBBackColor(&background);
}// UseLighterColors


//
// obsolete methods (avoid using if at all possible)
//

void	GetTEText	(DialogRef			inDialog,
					 DialogItemIndex	inItemIndex,
				 	 Str255				outString)
{
	DialogItemType		itemType;
	Handle				itemHandle;
	Rect				itemRect;
	
	
	GetDialogItem(inDialog, inItemIndex, &itemType, &itemHandle, &itemRect);
	GetDialogItemText(itemHandle, outString);
}


void
Hilite	(DialogRef			inDialog,
		 DialogItemIndex	inItemIndex,
		 short				inValue)
{
	ControlRef		control = null;
	
	
	GetDialogItemAsControl(inDialog, inItemIndex, &control);
	HiliteControl(control, inValue);
}// Hilite


short	findPopupMenuItem	(MenuHandle		inMenuHandle,
							 Str255			inWithWhatName)
{
	short		id = 0;
	short		numItems = 0;
	Str255		tempName;
	
	
	numItems = CountMenuItems(inMenuHandle);
	for (id = 1; id < numItems; id++)
	{
		GetMenuItemText(inMenuHandle, id, tempName);
		if (EqualString(tempName, inWithWhatName, 0, 0)) break;
	}
	
	return id;
}


pascal void VersionNumber( DialogRef dptr, short itm)
{
	Rect ibox;
	char **ihndl;
	short typ;
	short item=itm;
	VersRecHndl version;
	Str63 versionString;

	if ((version = (VersRecHndl) GetResource('vers',1)) != null)
	{
		BlockMove((*version)->shortVersion+*(*version)->shortVersion+1, versionString, *((*version)->shortVersion+*(*version)->shortVersion+1)+1);
		ReleaseResource((Handle)version);
	}
	else
		*versionString = 0;

	GetDialogItem( dptr, item, &typ, &ihndl, &ibox);
	TextFontByName("\pGeneva");
	TextFont(GetScriptVariable(GetScriptManagerVariable(smKeyScript), smScriptAppFond));
	TextSize(9);
	MoveTo( ibox.left+((ibox.right-ibox.left)/2)-(StringWidth(versionString)/2), ibox.bottom-2);
	DrawString(versionString);
}// VersionNumber


/*!********************************************************************
 * figure out the appropriate leading for a font
 **********************************************************************/
short GetLeading(short fontID,short fontSize)
{
	FMInput fInInfo;
	FMOutput *fOutInfo;
	
	/* set up the font input struct */
	fInInfo.family = fontID;
	fInInfo.size = fontSize;
	fInInfo.face = 0;
	fInInfo.needBits = false;
	fInInfo.device = 0;
	fInInfo.numer.h = fInInfo.numer.v = 1;
	fInInfo.denom.h = fInInfo.denom.v = 1;
	
	/* get the actual info */
	fOutInfo = FMSwapFont(&fInInfo);
	
	return(((fOutInfo->leading + fOutInfo->ascent + fOutInfo->descent)*fOutInfo->numer.v)/fOutInfo->denom.v);
}

/*!********************************************************************
 * find ascent font
 **********************************************************************/
short GetAscent(short fontID,short fontSize)
{
	FMInput fInInfo;
	FMOutput *fOutInfo;
	
	/*	set up the font input struct */
	fInInfo.family = fontID;
	fInInfo.size = fontSize;
	fInInfo.face = 0;
	fInInfo.needBits = false;
	fInInfo.device = 0;
	fInInfo.numer.h = fInInfo.numer.v = 1;
	fInInfo.denom.h = fInInfo.denom.v = 1;
	
	/*	get the actual info */
	fOutInfo = FMSwapFont(&fInInfo);
	
	return((fOutInfo->ascent * fOutInfo->numer.v)/fOutInfo->denom.v);
}

/*!**********************************************************************
 * GetMaxDepth - find the best monitor depth capable on the main device
 ************************************************************************/
short GetMaxDepth (void)
{
	short		maxDepth = 0;
	short		testDepth = 1;
	
	
	// new theme brushes require device info, so get the maximum depth available
	while (HasDepth(GetMainDevice(), gdDevType, testDepth, 1)) testDepth = testDepth * 2;
	maxDepth = testDepth * 0.5;
	
	return maxDepth;
}

/*!**********************************************************************
 * CenterRectIn - center one rect in another
 ************************************************************************/
void CenterRectIn(Rect *inner,Rect *outer)
{
	OffsetRect(inner,(outer->left+outer->right-inner->left-inner->right)/2, 
									 (outer->top+outer->bottom-inner->top-inner->bottom)/2);																	
}


//
// internal methods
//
#pragma mark -

/*!
To close dialog boxes in a consistent
manner, call this method.  This routine
undoes any damage caused by a call to
dialogDisplayPreparation().

(3.0)
*/
static void
dialogClose	(DialogRef const	inDialog)
{
	HideWindow(GetDialogWindow(inDialog));
	if (GetWindowKind(FrontWindow()) != kDialogWindowKind) EventLoop_SetFloatersHilited(true);
	RestoreFrontmostWindow();
}// dialogClose


/*!
To ensure a dialog box is displayed, use
this method.  The documentation on what
TransitionWindow() does is unclear, so
this makes sure dialog windows are showing.

(3.0)
*/
static void
dialogDisplayEnsureVisibility	(DialogRef const	inDialog)
{
	WindowRef		window = GetDialogWindow(inDialog);
	
	
	if (window != null)
	{
		unless (IsWindowVisible(window)) ShowWindow(window);
		SelectOverRealFrontWindow(window);
	}
	Cursors_UseArrow();
	Menus_EnableEditMenu();
	DrawMenuBar();
}// dialogDisplayEnsureVisibility


/*!
To prepare for a dialogs display by dimming
the frontmost window and deactivating all
floating windows, use this method.  Later,
use dialogClose() to close the dialog, and
all effects of invoking this method will be
reversed.

(3.0)
*/
static void
dialogDisplayPreparation	(DialogRef const	UNUSED_ARGUMENT(inDialog))
{
	EventLoop_SetFloatersHilited(false);
	DeactivateFrontmostWindow();
}// dialogDisplayPreparation


/*!
To automatically use the best (32-bit) Help button icon
available for the specified Help bevel button control,
use this method.  If Icon Services is not available, a
system icon resource is automatically used.

(3.0)
*/
static void
useBestHelpButtonIcon	(ControlRef		inControl)
{
	IconManagerIconRef		icon = IconManager_NewIcon();
	
	
	IconManager_MakeIconSuite(icon, kHelpIconResource, kSelectorAllSmallData);
	(OSErr)IconManager_SetButtonIcon(inControl, icon);
	IconManager_DisposeIcon(&icon);
}// useBestHelpButtonIcon


#if TARGET_API_MAC_OS8

/*!
To attempt to make an existing dialog
box as Appearance-savvy as possible,
use this method.  You typically use
this routine on dialogs that already
have controls and therefore cant use
an embedding hierarchy properly.

One side effect of calling this routine
is that the default button property is
lost.  You need to reset it.

This routine is used to modify the
Standard File nonmovable modal dialogs,
such as the new folder dialog, so
that they use theme characteristics.
It is not Carbon-compliant, but it is
great for old-style dialogs like the
Standard File ones (and Carbon doesnt
even use Standard File).

(3.0)
*/
static void
appearanceizeDialog		(DialogRef		inDialog)
{
	// This whacky code attempts to Appearance-ize a dialog box
	// without knowing anything about its items!  It directly
	// accesses the dialog record, copies the item list, then
	// destroys the items, creates a root control, and adds the
	// items back again.
	{
		Handle		copyOfItemList = null,
					originalItemList = ((DialogPeek)inDialog)->items;
		
		
		copyOfItemList = Memory_NewHandle(GetHandleSize(originalItemList));
		if (copyOfItemList != null)
		{
			ControlRef	root = null;
			SInt8		state = HGetState(originalItemList);
			
			
			HLock(copyOfItemList);
			HLock(originalItemList);
			BlockMove(*originalItemList, *copyOfItemList, GetHandleSize(copyOfItemList));
			HSetState(originalItemList, state);
			
			ShortenDITL(inDialog, CountDITL(inDialog));
			Alert_ReportOSErr(CreateRootControl(inDialog, &root));
			AppendDITL(inDialog, copyOfItemList, overlayDITL);
			Memory_DisposeHandle(&copyOfItemList);
		}
		else BEEP;
	}
	
	// auto-size the dialog to neatly fit its text items
	(OSErr)AutoSizeDialog(inDialog);
	
	// use a theme background
	(OSStatus)SetThemeWindowBackground(GetDialogWindow(inDialog),
										kThemeBrushDialogBackgroundActive, true/* update */);
}// appearanceizeDialog


/*!
To concatenate a colon and the name of the
MacTelnet process to a string, use this method.
This is used by the ThemeSavvyXXX() Standard
File routines.

(3.0)
*/
static void
catColonAndAppName	(StringPtr		inoutString)
{
	if (inoutString != null)
	{
		PLstrcat(inoutString, "\p: ");
		PLstrcat(inoutString, LMGetCurApName());
	}
}// catColonAndAppName


/*!
To create a very small window that is
positioned offscreen so the user cant
see it, use this method.

This routine exists to combat a very odd
Standard File problem that occurs if any
ThemeSavvyxxx() dialog is presented when
there are no windows open.  Specifically,
if no windows are open, some kind of
unending loop basically stalls MacTelnet
and requires a force-quit.  By creating
a useless window first, then there is a
window open (that the user cant see),
and the loop is avoided.

(3.0)
*/
static WindowRef
newWorthlessWindow		(void)
{
	WindowRef		result = null;
	Rect			offscreenBounds;
	
	
	SetRect(&offscreenBounds, -24, -24, -16, -16);
	result = NewCWindow(null, &offscreenBounds, EMPTY_PSTRING, true, kWindowDocumentProc, kInFront, false, 0L);
	return result;
}// newWorthlessWindow


/*!
One of the side effects of using the new
method appearanceizeDialog() on a Standard
File dialog is that the default button is
not set properly.  This routine analyzes
the reference constant of the specified
Standard File dialog (new folder, etc.)
and chooses the appropriate default item
index.

(3.0)
*/
static void
setStandardFileDialogDefaultButton		(DialogRef		inDialog)
{
	switch (GetWRefCon(GetDialogWindow(inDialog)))
	{
		case sfMainDialogRefCon:
			// nothing needs to be done for this one
			break;
		
		case sfNewFolderDialogRefCon:
			// the new folder dialogs default item is 1
			SetDialogDefaultItem(inDialog, 1);
			break;
		
		case sfReplaceDialogRefCon:
			// the Replace dialogs default item is 1
			SetDialogDefaultItem(inDialog, 1);
			break;
		
		case sfStatWarnDialogRefCon:
		case sfLockWarnDialogRefCon:
		case sfErrorDialogRefCon:
			// when in doubt, 1 is a good guess
			SetDialogDefaultItem(inDialog, 1);
			break;
		
		default:
			// unrecognized window
			break;
	}
}// setStandardFileDialogDefaultButton


/*!
This routine performs standard operations applicable
to all variants of the theme-savvy standard file
dialogs when it comes to event handling.  It is not a
real filter; it is invoked by all filter routines.

(3.0)
*/
static Boolean
themeSavvyStandardFileEventFilter	(DialogRef		inDialog,
									 EventRecord*	inoutEventPtr,
									 short*			UNUSED_ARGUMENT(inoutItemHit),
									 void*			UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_false;
	
	
	switch (inoutEventPtr->what)
	{
		case updateEvt:
			if (((WindowRef)inoutEventPtr->message) == GetDialogWindow(inDialog))
			{
				// draw pop-up menu to fix Standard File glitch
				DrawOneDialogItem(inDialog, sfItemPopUpMenuUser);
			}
			else result = EventLoop_HandleEvent(inoutEventPtr);
			break;
		
		default:
			break;
	}
	return result;
}// themeSavvyStandardFileEventFilter


/*!
This routine is called by the system while a
dialog displayed by ThemeSavvyStandardGetDirectory()
is visible.  The dialog box items are initialized
properly in response to "sfHookFirstCall", and most
importantly, clicks in the help button and the
Choose button (custom items) are responded to.

(3.0)
*/
static pascal short
themeSavvyStandardGetDirectoryDialogHook	(short			inItem,
											 DialogRef		inDialog,
											 void*			inData)
{
	// intercept hits on the non-standard dialog items
	short		result = inItem;
	
	
	if (GetWRefCon(GetDialogWindow(inDialog)) == sfMainDialogRefCon)
	{
		if (result == sfHookFirstCall)
		{
			// save this dialog, so others can access it for whatever reason
			gCurrentThemeSavvyStandardFileModal = inDialog;
			
			// the user data is actually a string, the window title
			{
				Str255		string;
				
				
				PLstrcpy(string, (StringPtr)inData);
				catColonAndAppName(string);
				SetWTitle(GetDialogWindow(inDialog), string);
			}
			
			// get rid of the help button if it has no purpose
			SetUpHelpButtonDialogItem(inDialog, iThemeSavvyStandardGetDirectoryAppleGuideButton);
			
			// set the dialog prompt text before the dialog appears (sfHookFirstCall only shows up once)
			SetDialogItemControlText(inDialog, iThemeSavvyStandardGetDirectoryPrompt, gThemeSavvyDialogTextPrompt);
			
			// set the default and Cancel buttons
			SetDialogDefaultItem(inDialog, sfItemOpenButton);
			SetDialogCancelItem(inDialog, sfItemCancelButton);
			
			// localize the thing
			{
				DialogItemIndex		indices[3] =
									{
										iThemeSavvyStandardGetDirectoryChooseButton,
										sfItemOpenButton,
										sfItemCancelButton
									};
				
				
				(UInt16)Localization_AdjustDialogButtonArray(inDialog, indices, 3);
			}
			Localization_AdjustHelpButtonItem(inDialog, iThemeSavvyStandardGetDirectoryAppleGuideButton);
		}
		else if (result == iThemeSavvyStandardGetDirectoryAppleGuideButton)
		{
			// open MacTelnet Help if the help button is clicked
			AppleGuideUtilities_OpenFirstMainGuide();
			result = sfHookNullEvent;
		}
		else if (result == iThemeSavvyStandardGetDirectoryChooseButton)
		{
			// then pick the current directory!
			gSetDir = 1;
			result = sfItemCancelButton;
		}
		else if (result == sfHookLastCall)
		{
			// clear this variable, because no Standard File dialog is showing anymore
			gCurrentThemeSavvyStandardFileModal = null;
		}
	}
	else
	{
		// try to Appearance-ize other standard file dialogs, such as the new folder dialog
		if (result == sfHookFirstCall)
		{
			appearanceizeDialog(inDialog);
			setStandardFileDialogDefaultButton(inDialog);
		}
	}
	return result;
}// themeSavvyStandardGetDirectoryDialogHook


/*!
This routine is called by the system whenever the file list
in a dialog displayed by ThemeSavvyStandardGetDirectory()
needs rebuilding.  This routine prevents non-directories
from showing up in the file list.

(3.0)
*/
static pascal Boolean
themeSavvyStandardGetDirectoryFileFilter	(CInfoPBPtr		inParamBlockPtr,
											 void*			UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_false;
	
	
	// only folders are allowed
	if (inParamBlockPtr != null) if (inParamBlockPtr->hFileInfo.ioFlAttrib & kioFlAttribDirMask) result = pascal_false;
	
	return result;
}// themeSavvyStandardGetDirectoryFileFilter


/*!
This routine exists for convenience, if you need to add
any event-responding functionality to a dialog displayed
by ThemeSavvyStandardGetDirectory().

(3.0)
*/
static pascal Boolean
themeSavvyStandardGetDirectoryModalFilter	(DialogRef		inDialog,
											 EventRecord*	inoutEventPtr,
											 short*			inoutItemHit,
											 void*			inData)
{
	Boolean		result = pascal_false;
	
	
	result = themeSavvyStandardFileEventFilter(inDialog, inoutEventPtr, inoutItemHit, inData);
	return result;
}// themeSavvyStandardGetDirectoryModalFilter


/*!
This routine is called by the system while a
dialog displayed by ThemeSavvyStandardGetFile()
is visible.  The dialog box items are initialized
properly in response to "sfHookFirstCall", and
clicks in the help button (a custom item) are
responded to.

(3.0)
*/
static pascal short
themeSavvyStandardGetFileDialogHook		(short			inItem,
										 DialogRef		inDialog,
										 void*			inData)
{
	// intercept hits on the non-standard dialog items
	short		result = inItem;
	
	
	if (GetWRefCon(GetDialogWindow(inDialog)) == sfMainDialogRefCon)
	{
		if (result == sfHookFirstCall)
		{
			// save this dialog, so others can access it for whatever reason
			gCurrentThemeSavvyStandardFileModal = inDialog;
			
			// the user data is actually a string, the window title
			{
				Str255		string;
				
				
				PLstrcpy(string, (StringPtr)inData);
				catColonAndAppName(string);
				SetWTitle(GetDialogWindow(inDialog), string);
			}
			
			// get rid of the help button if it has no purpose
			SetUpHelpButtonDialogItem(inDialog, iThemeSavvyStandardGetFileAppleGuideButton);
			
			// set the dialog prompt text before the dialog appears (sfHookFirstCall only shows up once)
			SetDialogItemControlText(inDialog, iThemeSavvyStandardGetFilePrompt, gThemeSavvyDialogTextPrompt);
			
			// set the default and Cancel buttons
			SetDialogDefaultItem(inDialog, sfItemOpenButton);
			SetDialogCancelItem(inDialog, sfItemCancelButton);
			
			// localize the thing
			(UInt16)Localization_AdjustDialogButtons(inDialog, true/* OK */, true/* Cancel */);
			Localization_AdjustHelpButtonItem(inDialog, iThemeSavvyStandardGetFileAppleGuideButton);
		}
		else if (result == iThemeSavvyStandardGetFileAppleGuideButton)
		{
			// open MacTelnet Help if the help button is clicked
			AppleGuideUtilities_OpenFirstMainGuide();
			result = sfHookNullEvent;
		}
		else if (result == sfHookLastCall)
		{
			// clear this variable, because no Standard File dialog is showing anymore
			gCurrentThemeSavvyStandardFileModal = null;
		}
	}
	else
	{
		// try to Appearance-ize other standard file dialogs, such as the new folder dialog
		if (result == sfHookFirstCall)
		{
			appearanceizeDialog(inDialog);
			setStandardFileDialogDefaultButton(inDialog);
		}
	}
	return result;
}// themeSavvyStandardGetFileDialogHook


/*!
This routine is called by the system whenever the file list
in a dialog displayed by ThemeSavvyStandardGetFile() needs
rebuilding.  All this routine does is map the systems
file filter request to the function (if any) that is passed
to ThemeSavvyStandardGetFile().

(3.0)
*/
static pascal Boolean
themeSavvyStandardGetFileFileFilter		(CInfoPBPtr		inParamBlockPtr,
										 void*			UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_false;
	
	
	// just map the custom file filter to the original one (after all, StandardGetFile() is being emulated here)
	if (gThemeSavvyGetFileFileFilterUPP != null)
	{
		result = CallFileFilterProc(gThemeSavvyGetFileFileFilterUPP, inParamBlockPtr);
	}
	return result;
}// themeSavvyStandardGetFileFileFilter


/*!
This routine exists for convenience, if you need to add
any event-responding functionality to a dialog displayed
by ThemeSavvyStandardGetFile().

(3.0)
*/
static pascal Boolean
themeSavvyStandardGetFileModalFilter	(DialogRef		inDialog,
										 EventRecord*	inoutEventPtr,
										 short*			inoutItemHit,
										 void*			inData)
{
	Boolean		result = pascal_false;
	
	
	result = themeSavvyStandardFileEventFilter(inDialog, inoutEventPtr, inoutItemHit, inData);
	return result;
}// themeSavvyStandardGetFileModalFilter


/*!
This routine is called by the system whenever the
keyboard focus item of a dialog displayed by
ThemeSavvyStandardPutFile() changes.  Although the
ThemeSavvyStandardPutFile() method does not have
any custom focus items, this routine is very
important because it allows old-style Standard File
save dialog boxes to be drawn correctly under the
Mac OS 8 Appearance Manager.

Since the Appearance Manager converts old-style
dialog items into modern Appearance controls, the
transition to themes is nearly 100%.  However, the
Standard File save dialog box has been implemented
in a very strange way, such that using Appearance
controls causing a number of major cosmetic glitches
to appear.  This method fixes all of the bugs, which
involved the use of an old-style list focus ring,
and incorrectly managing the focus ring and text
highlighting of the editable text item.

(3.0)
*/
static pascal void
themeSavvyStandardPutFileActivate		(DialogRef		inDialog,
										 short			inItem,
										 Boolean		inActivating,
										 void*			UNUSED_ARGUMENT(inData))
{
	// use Appearance calls to fix the annoying cosmetic glitches from Appearance-izing a StandardPutFile dialog
	if (GetWRefCon(GetDialogWindow(inDialog)) == sfMainDialogRefCon)
	{
		ControlRef			control = null;
		Rect				bounds;
		
		
		(OSErr)GetDialogItemAsControl(inDialog, inItem, &control);
		GetControlBounds(control, &bounds);
		if ((inItem == sfItemFileNameTextEdit) || (inItem == sfItemFileListUser))
		{
			// fix the damned focus rings!
			ColorPenState		penState;
			
			
			ColorUtilities_GetColorAndPenState(&penState);
			ColorUtilities_NormalizeColorAndPen();
			if (inItem == sfItemFileNameTextEdit)
			{
				if (!inActivating) 
				{
					// erase extraneous paint marks
					SelectDialogItemText(inDialog, inItem, 0, 0);
					(OSStatus)DrawThemeFocusRect(&bounds, false);
				}
				InsetRect(&bounds, -2, -2); // area for focusing / bevelling is 2 pixels outside the control rectangle
				if (inActivating)
				{
					(OSStatus)DrawThemeFocusRect(&bounds, inActivating);
				}
				else
				{
					(OSStatus)DrawThemeGenericWell(&bounds, IsWindowHilited(GetDialogWindow(inDialog))
															? kThemeStateActive : kThemeStateInactive,
													false/* fill center */);
				}
				DrawOneDialogItem(inDialog, inItem);
			}
			else if (inItem == sfItemFileListUser)
			{
				// First erase the rectangular area around the list to the dialog background
				// pattern, to get rid of the old-style focus ring put there by the system.
				GrafPtr		port = null;
				Boolean		isActiveWindow = IsWindowHilited(GetDialogWindow(inDialog));
				
				
				GetPort(&port);
				(OSStatus)SetThemePen((isActiveWindow)	? kThemeBrushDialogBackgroundActive
														: kThemeBrushDialogBackgroundInactive,
										GetMaxDepth(), ColorUtilities_IsColorGrafPort(port));
				InsetRect(&bounds, -1, -1);
				FrameRect(&bounds);
				InsetRect(&bounds, -1, -1);
				FrameRect(&bounds);
				InsetRect(&bounds, -1, -1);
				FrameRect(&bounds);
				InsetRect(&bounds, 4, 4); // area for focusing / bevelling is 1 pixel inside the control rectangle
				if (inActivating)
				{
					(OSStatus)DrawThemeFocusRect(&bounds, inActivating);
				}
				else
				{
					// in some themes, the fill center request is apparently ignored, so this doesnt work
					//(OSStatus)DrawThemeGenericWell(&bounds, isActiveWindow ? kThemeStateActive : kThemeStateInactive,
					//								false/* fill center */);
				}
			}
			ColorUtilities_SetColorAndPenState(&penState);
		}
		
		// this last call ensures that the text highlighting doesnt get garbled in some cases
		(OSErr)SetKeyboardFocus(GetDialogWindow(inDialog), control,
								(inActivating) ? kControlFocusNextPart : kControlFocusNoPart);
	}
}// themeSavvyStandardPutFileActivate


/*!
This routine is called by the system while a
dialog displayed by ThemeSavvyStandardPutFile()
is visible.  The dialog box items are initialized
properly in response to "sfHookFirstCall", and
clicks in the help button (a custom item) are
responded to.

(3.0)
*/
static pascal short
themeSavvyStandardPutFileDialogHook		(short			inItem,
										 DialogRef		inDialog,
										 void*			inData)
{
	// intercept hits on the non-standard dialog items
	short		result = inItem;
	
	
	if (GetWRefCon(GetDialogWindow(inDialog)) == sfMainDialogRefCon)
	{
		if (result == sfHookFirstCall)
		{
			// save this dialog, so others can access it for whatever reason
			gCurrentThemeSavvyStandardFileModal = inDialog;
			
			// the user data is actually a string, the window title
			{
				Str255		string;
				
				
				PLstrcpy(string, (StringPtr)inData);
				catColonAndAppName(string);
				SetWTitle(GetDialogWindow(inDialog), string);
			}
			
			// get rid of the help button if it has no purpose
			SetUpHelpButtonDialogItem(inDialog, iThemeSavvyStandardPutFileAppleGuideButton);
			
			// set the default and Cancel buttons
			SetDialogDefaultItem(inDialog, sfItemOpenButton);
			SetDialogCancelItem(inDialog, sfItemCancelButton);
			
			// localize the thing
			(UInt16)Localization_AdjustDialogButtons(inDialog, true/* OK */, true/* Cancel */);
			Localization_AdjustHelpButtonItem(inDialog, iThemeSavvyStandardPutFileAppleGuideButton);
		}
		else if (result == iThemeSavvyStandardPutFileAppleGuideButton)
		{
			// open MacTelnet Help if the help button is clicked
			AppleGuideUtilities_OpenFirstMainGuide();
			result = sfHookNullEvent;
		}
		else if (result == sfHookLastCall)
		{
			// clear this variable, because no Standard File dialog is showing anymore
			gCurrentThemeSavvyStandardFileModal = null;
		}
	}
	else
	{
		// try to Appearance-ize other standard file dialogs, such as the replace existing file? dialog
		if (result == sfHookFirstCall)
		{
			appearanceizeDialog(inDialog);
			setStandardFileDialogDefaultButton(inDialog);
		}
	}
	return result;
}// themeSavvyStandardPutFileDialogHook


/*!
This routine exists for convenience, if you need to add
any event-responding functionality to a dialog displayed
by ThemeSavvyStandardPutFile().

(3.0)
*/
static pascal Boolean
themeSavvyStandardPutFileModalFilter	(DialogRef		inDialog,
										 EventRecord*	inoutEventPtr,
										 short*			inoutItemHit,
										 void*			inData)
{
	Boolean		result = pascal_false;
	
	
	result = themeSavvyStandardFileEventFilter(inDialog, inoutEventPtr, inoutItemHit, inData);
	return result;
}// themeSavvyStandardPutFileModalFilter

#endif