/*###############################################################

	Drag.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Appearance.h>
#include <CodeFragments.h>
#include <Drag.h>
#include <Gestalt.h>
#include <MacErrors.h>
#include <Scrap.h>

// library includes
#include "InterfaceLib.h"

// MacTelnet includes
#include "Clipboard.h"
#include "CommandLine.h"
#include "Console.h"
#include "ConnectionData.h"
#include "DialogUtilities.h"
#include "EditLogInMessageDialog.h"
#include "MacLook.h"
#include "MacroSetupDialog.h"
#include "Parse.h"
#include "Preferences.h"
#include "RealScreen.h"
#include "ScreenManager.h"
#include "TelnetDescriptors.h"
#include "TelnetDrag.h"
#include "TelnetHeader.h"
#include "TelnetMacros.h"



// variables
static Boolean		gIsDragManagerAvailable = false,
					gDropCanAcceptItems = false,
					gDropCursorInContent = false,
					gDropDestinationCanAcceptItems = false;

// internal methods
static pascal OSErr		receiveDropHandler	(WindowRef				inWindow,
											 void*					inUserContext,
								 			 DragRef				inDragRef);
static pascal OSErr		trackingHandler		(DragTrackingMessage	inMessage,
											 WindowRef				inWindow,
								  			 void*					inUserContext,
											 DragRef				inDragRef);



//
// public methods
//

/*!
Call this method once, at program startup time,
to initialize this module.

(3.0)
*/
void
Drag_Init	(void)
{
	OSErr		error = noErr;
	long		dragMgrAttr = 0L;
	
	
	error = Gestalt(gestaltDragMgrAttr, &dragMgrAttr);
	gIsDragManagerAvailable = ((error == noErr) && ((dragMgrAttr & (1L << gestaltDragMgrPresent)) != 0));
	gIsDragManagerAvailable = gIsDragManagerAvailable && 
								((dragMgrAttr & (1L << gestaltPPCDragLibPresent)) != 0) &&
								API_AVAILABLE(InstallTrackingHandler);
	
	// if the Drag Manager is available, install the tracking handlers
	if (Drag_Available())
	{
		DragTrackingHandlerUPP		trackingProc = NewDragTrackingHandlerUPP(trackingHandler);
		
		
		error = InstallTrackingHandler(trackingProc, null, null);
		
		if (error == noErr)
		{
			// install the receiving handler
			error = InstallReceiveHandler(NewDragReceiveHandlerUPP(receiveDropHandler), null, null);
			if (error != noErr)
			{
				// an error occurred installing the receiving handler - remove the tracking handler, too
				RemoveTrackingHandler(trackingProc, null);
			}
		}
		
		// if any error occurred, turn off Drag Manager support (avoids potential problems)
		if (error != noErr)
		{
			Console_WriteLine("Warning - Unexpected error initializing drag-and-drop services.");
			Console_WriteValue("          Error code", error);
			gIsDragManagerAvailable = false;
		}
	}
}// Drag_Init


/*!
To determine if Macintosh Drag and Drop
is available, call this method.

(3.0)
*/
Boolean
Drag_Available	(void)
{
	return gIsDragManagerAvailable;
}// Available


/*!
This method handles dragging of a text selection.
If the text is dropped (i.e. not cancelled), then
"true" is returned in "outDragWasDropped".

(3.0)
*/
OSErr
Drag_TrackScreenSelection	(EventRecord*		inoutEventPtr,
							 Point				inLocalMousePosition,
							 TelnetScreenID		inScreenID,
							 Boolean*			outDragWasDropped)
{
	OSErr		result = noErr;
	
	
	*outDragWasDropped = false;
	if (Drag_Available())
	{
		DragRef			dragRef;
		Boolean			haveDragRef = false;
		RgnHandle		dragRgn = null;
		Handle			textH = null;
		long			size = 0L;
		
		
		dragRgn = RealScreen_GetSelectedTextAsNewRegion(inScreenID);
		if (dragRgn != null)
		{
			if (PtInRgn(inLocalMousePosition, dragRgn) && WaitMouseMoved(inoutEventPtr->where))
			{
				*outDragWasDropped = true;
				textH = RealScreen_GetSelectedTextAsNewHandle(inScreenID, 0, 0/* flags */);
				
				if ((textH != (char**)-1L) && (textH != null))
				{
					HLock(textH);
					size = GetHandleSize(textH);
					
					result = NewDrag(&dragRef);
					if (result == noErr)
					{
						haveDragRef = true;
						
						result = AddDragItemFlavor(dragRef, 1, kTelnetDragFlavorStandardText,
													*textH, size, 0);
						if (result == noErr)
						{
							LocalToGlobalRegion(dragRgn);
							OutlineRegion(dragRgn);
							
							{
								SInt16		oldCursor = Cursors_UseHandGrab();
								
								
								result = TrackDrag(dragRef, inoutEventPtr, dragRgn);
								Cursors_Use(oldCursor);
							}
						}
					}
				}
				else textH = null;
			}
		}
		
		if (haveDragRef) DisposeDrag(dragRef);
		if (dragRgn != null) Memory_DisposeRegion(&dragRgn);
		if (textH != null) Memory_DisposeHandle(&textH);
	}
	
	Cursors_UseArrow();
	
	return result;
}// TrackScreenSelection



//
// internal methods
//

/*!
This method is called by the Drag Manager when a drop
occurs over one of MacTelnets windows.

(2.6)
*/
static pascal OSErr
receiveDropHandler	(WindowRef		inWindow,
					 void*			UNUSED_ARGUMENT(inUserContext),
					 DragRef		inDragRef)
{	
	OSErr		result = noErr;
	
	
	if ((!gDropCanAcceptItems) || (!gDropCursorInContent)) result = dragNotAcceptedErr;
	else
	{
		OSErr					memError = noErr;
		Handle					dataH = null;
		unsigned short			items = 0,
								index = 0;
		ItemReference			theItem;
		DragAttributes			attributes = 0L;
		Size					textSize = 0;
		short					mouseDownModifiers = 0,
								mouseUpModifiers = 0;
		WindowInfoDescriptor	windowDescriptor = kInvalidWindowInfoDescriptor;
		
		
		SetPortWindowPort(inWindow);
		
		// determine the MacTelnet window descriptor, if possible
		{
			WindowInfoRef		windowFeaturesRef = null;
			
			
			windowFeaturesRef = WindowInfo_GetFromWindow(inWindow);
			if (windowFeaturesRef != null)
			{
				windowDescriptor = WindowInfo_GetWindowDescriptor(windowFeaturesRef);
			}
		}
		
		// no text yet
		dataH = null;
		
		GetDragAttributes(inDragRef, &attributes);
		GetDragModifiers(inDragRef, 0L, &mouseDownModifiers, &mouseUpModifiers);
	
		// loop through all of the drag items contained in this drag
		// and collect the text into the accumulation handle
		CountDragItems(inDragRef, &items);
		for (index = 1; index <= items; index++)
		{
			// get the items reference number, so it can be referred to later
			GetDragItemReferenceNumber(inDragRef, index, &theItem);
	
			// Try to get the flags for a 'TEXT' flavor.  If this returns
			// without an error, then a 'TEXT' flavor exists in the item.
			result = GetFlavorDataSize(inDragRef, theItem, kTelnetDragFlavorStandardText, &textSize);
	
			if (result == noErr)
			{
				if (dataH == null)
				{
					// no data yet, create a new handle for accumulation
					dataH = Memory_NewHandle(textSize);
					if (dataH == null)
					{
						return memFullErr; // exit if there is an error
					}
				}
				else
				{
					// append to existing TEXT data
					memError = Memory_SetHandleSize(dataH, GetHandleSize(dataH) + textSize);
					if (memError != noErr)
					{
						return memFullErr; // exit if there is a problem
					}
				}
				
				// temporarily lock down the accumulation handle and get the drag data
				HLock(dataH);
				GetFlavorData(inDragRef, theItem, kTelnetDragFlavorStandardText, *dataH, &textSize, 0L);
				HUnlock(dataH);
			}
		}
	
		// get the length of all that was accumulated
		textSize = GetHandleSize(dataH);
	
		// if text was actually received, insert it into the destination
		if (textSize > 0)
		{
			if (attributes & kDragHasLeftSenderWindow)
			{
				HideDragHilite(inDragRef);
			}
			
			HLock(dataH);
	
			// send the text to the window
			if (ScreenManager_IsWindowAScreen(inWindow))
			{
				ConnectionDataPtr	connectionDataPtr = ScreenManager_GetWindowConnectionData(inWindow);
				
				
				if (connectionDataPtr != null)
				{
					Size		actualSize = 0L;
					Boolean		cursorMovesPriorToDrops = false;
					
					
					// if the user preference is set, first move the cursor to the drop location
					unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
												kTelnetPreferenceTag_CursorMovesPriorToDrops,
												sizeof(cursorMovesPriorToDrops), &cursorMovesPriorToDrops,
												&actualSize) == kTelnetPrefsResultCodeSuccess)
					{
						cursorMovesPriorToDrops = false; // assume a value, if a preference cant be found
					}
					if (cursorMovesPriorToDrops)
					{
						Point		mouseLocation,
									globalPinnedMouseLocation;
						
						
						if (GetDragMouse(inDragRef, &mouseLocation, &globalPinnedMouseLocation) == noErr)
						{
							RealScreen_MoveCursorWithArrowKeys(connectionDataPtr->vs, mouseLocation);
							// unimplemented!
						}
					}
					
					// type the text
					Parse_TypeString(connectionDataPtr, *dataH, textSize, false/* record */);
				}
			}
			else if (windowDescriptor == kTelnetWindowDescriptorClipboard)
			{
				// put the text on the clipboard
			#if TARGET_API_MAC_OS8
				(OSStatus)ZeroScrap();
				(long)PutScrap(GetHandleSize(dataH), kScrapFlavorTypeText, *dataH);
			#else
				(OSStatus)ClearCurrentScrap();
				{
					ScrapRef		currentScrap = null;
					
					
					if (GetCurrentScrap(&currentScrap) == noErr)
					{
						(OSStatus)PutScrapFlavor(currentScrap, kScrapFlavorTypeText,
													kScrapFlavorMaskNone, GetHandleSize(dataH),
													*dataH);
					}
				}
			#endif
			}
			else if (windowDescriptor == kTelnetWindowDescriptorCommandLine)
			{
				// try to set the text field using the dropped text
				CommandLine_SetText(*dataH, textSize);
			}
			else if (windowDescriptor == kTelnetDialogDescriptorMacroSetup)
			{
				// try to import macros into Macro Setup dialogs text fields
				MacroSetupDialog_ReceiveDrop(inDragRef, STATIC_CAST(*dataH, unsigned char*), textSize);
			}
			else if (windowDescriptor == kTelnetDialogDescriptorEditLogInMessage)
			{
				// try to set the text field using the dropped text
				EditLogInMessageDialog_SetText(*dataH, textSize);
			}
			else if (windowDescriptor == kTelnetDialogDescriptorSetWindowTitle)
			{
				// try to set the text field using the dropped text
				RealScreen_UpdateSetWTitleDialogUsing((unsigned char*)*dataH, textSize,
														GetDialogFromWindow(inWindow));
			}
			HUnlock(dataH);
		}
	
		if (dataH != null)
		{
			Memory_DisposeHandle(&dataH);
		}
	}
	
	return result;
}// receiveDropHandler


/*!
This is the drag tracking handler, which handles things
such as adding or removing drag highlighting from MacTelnet
windows.

(2.6)
*/
pascal OSErr
trackingHandler	(DragTrackingMessage	inMessage,
				 WindowRef				inWindow,
				 void*					UNUSED_ARGUMENT(inUserContext),
				 DragRef				inDragRef)
{
	short					result = 0;
	WindowInfoDescriptor	windowDescriptor = kInvalidWindowInfoDescriptor;
	unsigned short			count = 0,
							index = 0;
	DragAttributes			attributes = 0L;
	ItemReference			theItem;
	Point					theMouse,
							localMouse;
	Rect					dragRect;
	GrafPtr					savePort = null;
	Boolean					inOriginalSelection = false;
	
	
	if ((inMessage != kDragTrackingEnterHandler) && (!gDropCanAcceptItems))
		return(noErr);
	
	if ((inMessage != kDragTrackingEnterHandler) && (inMessage != kDragTrackingEnterWindow) 
			&& (!gDropDestinationCanAcceptItems))
		return(noErr);
	
	// determine the MacTelnet window descriptor, if possible
	{
		WindowInfoRef		windowFeaturesRef = null;
		
		
		windowFeaturesRef = WindowInfo_GetFromWindow(inWindow);
		if (windowFeaturesRef != null)
		{
			windowDescriptor = WindowInfo_GetWindowDescriptor(windowFeaturesRef);
		}
	}
	
	GetDragAttributes(inDragRef, &attributes);
	
	switch (inMessage)
	{
		// The following message shows up the first time that a drag
		// enters ANY MacTelnet window, regardless of whether the window
		// is capable of receiving a drop.
		case kDragTrackingEnterHandler:
			// assume initially that the window can accept a drop
			gDropCanAcceptItems = true;
			
			// check to see if all of the drag items contain
			// TEXT (a drag is only accepted if all of the
			// items in the drag can be accepted)
			CountDragItems(inDragRef, &count);
			
			for (index = 1; index <= count; index++)
			{
				FlavorFlags		flavorFlags = 0L;
				
				
				GetDragItemReferenceNumber(inDragRef, index, &theItem);
				
				result = GetFlavorFlags(inDragRef, theItem, kTelnetDragFlavorStandardText, &flavorFlags);
				
				if (result != noErr)
				{
					gDropCanAcceptItems = false;
					break;
				}
			}
			break;
		
		// The following message shows up each time a drag enters any
		// MacTelnet window.  Certain variables are used for tracking
		// data in the window, and these are reset for each new window
		// (i.e. for each new receipt of the following message).
		case kDragTrackingEnterWindow:
			// assume initially that the window cannot accept a drop
			gDropDestinationCanAcceptItems = false;
			
			// if it is not a connection window or special dialog, it cannot accept a drop
			if ((windowDescriptor == kTelnetWindowDescriptorAnyTerminal) ||
				(windowDescriptor == kTelnetWindowDescriptorClipboard) ||
				(windowDescriptor == kTelnetWindowDescriptorCommandLine) ||
				(windowDescriptor == kTelnetDialogDescriptorMacroSetup) ||
				(windowDescriptor == kTelnetDialogDescriptorEditLogInMessage) ||
				(windowDescriptor == kTelnetDialogDescriptorSetWindowTitle))
			{
				gDropDestinationCanAcceptItems = true;
				
				// initially no drag highlight region
				gDropCursorInContent = false;
			}
			break;
		
		// The following message appears as long as the mouse is in
		// one of MacTelnets windows during a drag.  Window drag
		// highlighting is drawn when these messages appear.
		case kDragTrackingInWindow:
			GetPort(&savePort);
			SetPortWindowPort(inWindow);
			
			GetDragMouse(inDragRef, &theMouse, 0L);
			
			// If still in the sender window, check to see if a drag highlight region should
			// be drawn.  The value of "inOriginalSelection" is true if the mouse is still
			// within the confines of the selected texts original location.  This allows the
			// user to cancel a drop by dropping the text somewhere inside of the original
			// selection.
			inOriginalSelection = false;
			if (attributes & kDragInsideSenderWindow)
			{
				RgnHandle		selectionRegion = null;
				
				
				localMouse = theMouse;
				GlobalToLocal(&localMouse);
				
				if (ScreenManager_IsWindowAScreen(inWindow))
				{
					selectionRegion = RealScreen_GetSelectedTextAsNewRegion
										(ScreenManager_GetWindowScreenID(inWindow));
					if (selectionRegion != null)
					{
						inOriginalSelection = PtInRgn(localMouse, selectionRegion);
						Memory_DisposeRegion(&selectionRegion);
					}
				}
			}

			// Show or hide the window highlighting when the mouse enters or leaves the window region.
			{
				// most windows will highlight their entire content region for drags...
				MacLook_GetWindowRegionBounds(inWindow, kWindowContentRgn, &dragRect);
				
				// reduce the drag rectangle for windows where the highlight area isnt the entire window
				if (ScreenManager_IsWindowAScreen(inWindow))
				{
					// for connection windows, reduce the drag area to be the terminal screen itself
					RealScreenDataPtr	realScreenPtr = ScreenManager_GetWindowRealScreenData(inWindow);
					Rect				screenRect;
					
					
					// subtract out the window controls
					RealScreen_GetRegionBounds(ScreenManager_GetWindowScreenID(inWindow),
												kRealScreenRegionCodeScreenBackground, &screenRect);
					dragRect.top += screenRect.top;
					dragRect.left += screenRect.left;
					dragRect.bottom = dragRect.top + screenRect.bottom - screenRect.top;
					dragRect.right = dragRect.left + screenRect.right - screenRect.left;
				}
				else if (windowDescriptor == kTelnetWindowDescriptorClipboard)
				{
					// assign the highlight rectangle area to be the clipboard view rectangle
					Rect		localViewRect;
					
					
					Clipboard_GetViewRectangle(&localViewRect);
					dragRect.top += localViewRect.top;
					dragRect.left += localViewRect.left;
					dragRect.bottom = dragRect.top + localViewRect.bottom - localViewRect.top;
					dragRect.right = dragRect.left + localViewRect.right - localViewRect.left;
				}
				
				// If the mouse is in the text area and not
				// in the original selection, draw the drag
				// highlight region.
				if (PtInRect(theMouse, &dragRect) && !inOriginalSelection)
				{
					unless (gDropCursorInContent)
					{
						Point		p1,
									p2;
						
						
						p1.h = dragRect.left;
						p1.v = dragRect.top;
						p2.h = dragRect.right;
						p2.v = dragRect.bottom;
						
						// set up the drag highlight region
						GlobalToLocal(&p1);
						GlobalToLocal(&p2);
						dragRect.left = p1.h;
						dragRect.top = p1.v;
						dragRect.right = p2.h;
						dragRect.bottom = p2.v;
						
						// display user feedback that the drop location is valid
						{
							RgnHandle		region = Memory_NewRegion();
							
							
							if (region != null)
							{
								RectRgn(region, &dragRect);
								ShowDragHilite(inDragRef, region, true/* inside */);
								Memory_DisposeRegion(&region);
							}
						}
					}
					
					// this is a valid drop location, for future reference
					gDropCursorInContent = true;
				}
				else
				{
					// moved out of a valid drop region
					if (gDropCursorInContent) HideDragHilite(inDragRef);
					
					// this is no longer a valid drop region
					gDropCursorInContent = false;
				}
			}
			SetPort(savePort);
			break;
		
		case kDragTrackingLeaveWindow:
			HideDragHilite(inDragRef);
			break;
		
		case kDragTrackingLeaveHandler:
			break;
		
		default:
			break;
	}
	return(noErr);
}// trackingHandler

