/*###############################################################

	Embedding.c
	
	Interface Library 1.0
	 1998-2000 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#ifndef __GESTALT__
#	include <Gestalt.h>
#endif
#ifndef __IMAGECOMPRESSION__
#	include <ImageCompression.h>
#endif
#ifndef __PLSTRINGFUNCS__
#	include <PLStringFuncs.h>
#endif
#ifndef __QUICKDRAW__
#	include <QuickDraw.h>
#endif
#ifndef __SOUND__
#	include <Sound.h>
#endif
#ifndef __TEXTUTILS__
#	include <TextUtils.h>
#endif

// library includes
#include "InterfaceLib.h"



// types
struct OffscreenDeviceLoopParams
{
	OffscreenOperationProcPtr	proc;
	WindowRef					window;
	ControlRef					control;
	SInt32						data1,
								data2;
	SInt32						dumpMode;
	OSErr						result;
};
typedef struct OffscreenDeviceLoopParams	OffscreenDeviceLoopParams;
typedef OffscreenDeviceLoopParams*			OffscreenDeviceLoopParamsPtr;

// internal methods
static pascal void		offscreenDumpDeviceLoop			(short					inColorDepth,
														 short					inDeviceFlags,
														 GDHandle				inTargetDevice,
														 long					inDataPtr);
static pascal Boolean	setControlActiveOperation		(const ControlRef		inSpecificControlOrRoot,
														 const SInt16			inColorDepth,
														 const SInt16			inDeviceFlags,
														 GDHandle				inTargetDevice,
														 const SInt32			inIsEnabled,
														 const SInt32			inData2);
static pascal Boolean	swapControlsOperation			(const ControlRef		inControlToHide,
														 const SInt16			inColorDepth,
														 const SInt16			inDeviceFlags,
														 GDHandle				inTargetDevice,
														 const SInt32			inControlToDisplay,
														 const SInt32			inData2);



//
// public methods
//

/**
To deactivate the root control of the frontmost
window and to change the cursor to a wristwatch,
call this method.  Usually, you do this immediately
prior to displaying a dialog box, and then invoke
Embedding_RestoreFrontmostWindow() after removing
the dialog box at a later time.

(1.0)
*/
void
Embedding_DeactivateFrontmostWindow	(void)
{
	WindowRef		window = FrontWindow();
	
	
	if (window != null)
	{
		if (IsWindowHilited(window)) if (GetWindowKind(window) != kDialogWindowKind)
		{
			(OSErr)Embedding_OffscreenSetRootControlActive(window, false);
		}
	}
	Cursors_UseWatch();
}// DeactivateFrontmostWindow


/**
Use this convenience method to invoke the
Embedding_OffscreenControlOperationInMode()
routine with a "srcCopy" (normal) drawing
mode.

(1.0)
*/
OSErr
Embedding_OffscreenControlOperation		(WindowRef					inForWhichWindow,
										 ControlRef					inForWhichControlOrNullForRoot,
										 OffscreenOperationProcPtr	inWhatToDo,
										 SInt32						inData1,
										 SInt32						inData2)
{
	return Embedding_OffscreenControlOperationInMode(inForWhichWindow, inForWhichControlOrNullForRoot, inWhatToDo, inData1, inData2, srcCopy);
}// OffscreenControlOperation


/**
To minimize the flickering caused by the sequential drawing
of many controls in a window, use this routine.  It will
make the given control of a window invisible, and then call
your routine to perform actions on controls in the window.
When your routine returns, the given control is made visible
and the entire operation (which will be in an offscreen
graphics buffer) will then be dumped to the screen.  The
entire port rectangle of the window is validated.

This has the benefit of completely avoiding flickering,
since all changes you make will appear to occur instantly,
not one after another.  However, if there is a problem with
managing an offscreen graphics world (not enough memory, for
instance), the routine you provide is simply invoked without
operating on a buffer (in other words, flickering occurs in
this case).

I believe there is a small memory leak in here someplace,
but Im not sure where.  Heh heh heh heh heh ha (*cough*).

IMPORTANT:	Only the given control is drawn in the offscreen
			buffer, so this routine will only work for
			windows that do ALL of their drawing using
			embedded controls in a Mac OS 8 Appearance
			control hierarchy.  If you have special drawing
			to do, you had better create a user pane control
			and define a custom drawing routine to perform
			your special drawing; that way, you can still
			take advantage of this routine.

(1.0)
*/
OSErr
Embedding_OffscreenControlOperationInMode	(WindowRef					inForWhichWindow,
											 ControlRef					inForWhichControlOrNullForRoot,
											 OffscreenOperationProcPtr	inWhatToDo,
											 SInt32						inData1,
											 SInt32						inData2,
											 SInt32						inDrawingMode)
{
	ControlRef		control = inForWhichControlOrNullForRoot;
	GrafPtr			oldPort = null;
	OSErr			result = noErr;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(inForWhichWindow);
	
	if (control == null) result = GetRootControl(inForWhichWindow, &control);
	if (result == noErr)
	{
		Rect						controlBounds;
		RgnHandle					oldClipRgn = null,
									newClipRgn = null;
		OffscreenDeviceLoopParams	data;
		
		
		data.proc = inWhatToDo;
		data.window = inForWhichWindow;
		data.control = control;
		data.data1 = inData1;
		data.data2 = inData2;
		data.dumpMode = inDrawingMode;
		data.result = noErr;
		
		newClipRgn = Memory_NewRegion();
		oldClipRgn = Memory_NewRegion();
		
		// get the arrangement of this control
		GetControlBounds(control, &controlBounds);
		if (result == noErr)
		{
			if ((newClipRgn != null) && (oldClipRgn != null))
			{
				DeviceLoopDrawingUPP		upp = null;
				
				
				// note that "oldClipRgn" is being used for multiple purposes here, despite its name
				RectRgn(newClipRgn, &controlBounds);
				GetPortVisibleRegion(GetWindowPort(inForWhichWindow), oldClipRgn);
				SectRgn(oldClipRgn, newClipRgn, newClipRgn);
				GetClip(oldClipRgn);
				SetClip(newClipRgn);
				
				// call DeviceLoop() to make sure the offscreen dump looks correct, no matter how many monitors it may cross
				upp = NewDeviceLoopDrawingUPP(offscreenDumpDeviceLoop);
				DeviceLoop(newClipRgn, upp, (long)&data/* user data required by the DeviceLoopDrawingUPP */, 0/* DeviceLoopFlags */);
				DisposeDeviceLoopDrawingUPP(upp), upp = null;
				result = data.result;
				
				SetClip(oldClipRgn);
			}
			else data.result = memFullErr;
			
			// check for failures
			if (data.result != noErr)
			{
				// offscreen dump failed somehow - just invoke the routine manually, with good guesses of unknown parameters
				(OSErr)SetUpControlBackground(control, 8/* color depth */, ColorUtilities_IsColorDevice(GetMainDevice()));
				EraseRect(&controlBounds);
				if (inWhatToDo != null)
				{
					// desperation: invoke the routine to draw the control anyway, providing reasonable default values
					(Boolean)InvokeOffscreenOperationProc(inWhatToDo, control, 8/* default depth */, 0/* default device flags */,
															GetMainDevice(), inData1, inData2);
				}
				DrawOneControl(control);
			}
			
			if ((oldClipRgn != null) && (newClipRgn != null))
			{
				// Note that the following statements really should be conditionalized
				// beyond the preprocessor level, within a statement that checks for the
				// presence of the Mac OS 8.5 Window Manager.  However, since Carbon
				// *does* include the new Window Manager, and since the old routines all
				// work under even Mac OS 8.5 and beyond (but not Carbon), a TARGET_API_
				// conditional is used for simplified code.
				#if TARGET_API_MAC_OS8
				SetPortWindowPort(inForWhichWindow);
				ValidRgn(newClipRgn);
				#else
				ValidWindowRgn(inForWhichWindow, newClipRgn);
				#endif
			}
			else
			{
				#if TARGET_API_MAC_OS8
				SetPortWindowPort(inForWhichWindow);
				ValidRect(&controlBounds);
				#else
				ValidWindowRect(inForWhichWindow, &controlBounds);
				#endif
			}
		}
		if (oldClipRgn != null) Memory_DisposeRegion(&oldClipRgn);
		if (newClipRgn != null) Memory_DisposeRegion(&newClipRgn);
	}
	SetPort(oldPort);
	
	return result;
}// OffscreenControlOperationInMode


/**
To redraw a complex control without any
flickering (memory permitting), use this
method.  This routine can be used instead
of DrawOneControl() in every case: if not
enough memory is available, this routine
will automatically fall back and not use
an offscreen buffer, which is the same as
calling DrawOneControl() (but slightly
slower, due to the logic required to
conclude that there is insufficient
memory).

(1.0)
*/
OSErr
Embedding_OffscreenDrawOneControl		(ControlRef			inControl)
{
	OSErr		result = noErr;
	
	
	result = Embedding_OffscreenControlOperation(GetControlOwner(inControl), inControl, null, 0L, 0L);
	return result;
}// OffscreenDrawOneControl


/**
To minimize the flickering caused by the
sequential activation and deactivation of
many controls in a window, use this routine.
It will set or clear the activated state of
a particular control, and redirect the
control update to an offscreen graphics
buffer.  When all control states have been
drawn to the buffer, the result is then
dumped to the window, eliminating graphics
flickering altogether.  The entire control
region is validated.

(1.0)
*/
OSErr
Embedding_OffscreenSetControlActive		(ControlRef		inControl,
										 Boolean		inIsActive)
{
	OSErr		result = noErr;
	
	
	result = Embedding_OffscreenControlOperation(GetControlOwner(inControl), inControl, setControlActiveOperation, (inIsActive) ? 1L : 0L, 0L);
	return result;
}// OffscreenSetControlActive


/**
To minimize the flickering caused by the
sequential activation and deactivation of
many controls in a window, use this routine.
It will set or clear the activated state of
a particular windows root control, and
redirect the control update to an offscreen
graphics buffer.  When all control states
have been drawn to the buffer, the result
is then dumped to the window, eliminating
graphics flickering altogether.  The entire
window port rectangle is validated.

(1.0)
*/
OSErr
Embedding_OffscreenSetRootControlActive		(WindowRef		inForWhichWindow,
											 Boolean		inIsActive)
{
	OSErr		result = noErr;
	
	
	//result = Embedding_OffscreenControlOperation(inForWhichWindow, null/* use root */, setControlActiveOperation, (inIsActive) ? 1L : 0L, 0L);
	// TEMP - problems using embedding operations here?
	{
		ControlRef	root = null;
		
		
		if (GetRootControl(inForWhichWindow, &root) == noErr)
		{
			if (inIsActive) (OSErr)ActivateControl(root);
			else (OSErr)DeactivateControl(root);
		}
	}
	return result;
}// OffscreenSetRootControlActive


/**
If two controls overlap precisely (like tab
panes), you can use this routine to hide the
first control and display the second control
in one smooth motion (memory permitting).
The result is that the second control is
drawn without flickering caused by erasing
the first control or drawing parts embedded
in the second control.

(1.0)
*/
OSErr
Embedding_OffscreenSwapOverlappingControls	(WindowRef			inForWhichWindow,
											 ControlRef			inControlToHide,
											 ControlRef			inControlToDisplay)
{
	OSErr		result = noErr;
	
	
	result = Embedding_OffscreenControlOperation(inForWhichWindow, inControlToDisplay, swapControlsOperation, (SInt32)inControlToHide, 0L);
	return result;
}// OffscreenSwapOverlappingControls


/**
To activate the root control of the frontmost
window and to change the cursor to the standard
arrow, call this method.  Usually, you do this
immediately after removing a dialog box from the
screen, especially if you previously used the
Embedding_DeactivateFrontmostWindow() method.

(1.0)
*/
void
Embedding_RestoreFrontmostWindow	(void)
{
	WindowRef		window = FrontWindow();
	
	
	if (window != null)
	{
		if (!IsWindowHilited(window)) if (GetWindowKind(window) != kDialogWindowKind)
		{
			(OSErr)Embedding_OffscreenSetRootControlActive(window, true);
		}
	}
	Cursors_UseArrow();
}// RestoreFrontmostWindow


//
// internal methods
//

/**
This method, of standard DeviceLoopDrawingUPP form,
handles offscreen drawing correctly no matter what
the device characteristics are.

(1.0)
*/
static pascal void
offscreenDumpDeviceLoop		(short			inColorDepth,
							 short			inDeviceFlags,
							 GDHandle		inTargetDevice,
							 long			inDataPtr)
{
	OffscreenDeviceLoopParamsPtr	dataPtr = (OffscreenDeviceLoopParamsPtr)inDataPtr;
	GrafPtr							oldPort = null;
	
	
	GetPort(&oldPort);
	if (dataPtr != null) //if (dataPtr->result == noErr) // any residual error from previous device loops should continue the termination
	{	
		Rect			drawingRect,
						controlRect;
		GWorldPtr		world = null;
		GDHandle		oldDevice = null;
		CGrafPtr		colorWindowPort = GetWindowPort(dataPtr->window),
						oldColorPort = null;
		
		
		GetPortBounds(colorWindowPort, &drawingRect);
		GetControlBounds(dataPtr->control, &controlRect);
		
		// save old port
		dataPtr->result = NewGWorld(&world, inColorDepth, &drawingRect, (CTabHandle)null, (GDHandle)inTargetDevice, 0L/* flags */);
		if (dataPtr->result == noErr)
		{
			PixMapHandle	pixels = null;
			
			
			// set port to the new port
			GetGWorld(&oldColorPort, &oldDevice);
			SetGWorld(world, (GDHandle)null);
			
			// get the offscreen graphics worlds pixel map
			pixels = GetGWorldPixMap(world);
			
			// perform the offscreen control operation, and then do the blitter as quickly and easily as possible
			if (LockPixels(pixels))
			{
				(OSErr)SetUpControlBackground(dataPtr->control, inColorDepth, ColorUtilities_IsColorDevice(inTargetDevice));
				EraseRect(&drawingRect);
				
				// change the controls as specified by the user-defined routine
				if (dataPtr->proc != null)
				{
					(Boolean)InvokeOffscreenOperationProc(dataPtr->proc, dataPtr->control, inColorDepth, inDeviceFlags, inTargetDevice,
															dataPtr->data1, dataPtr->data2);
				}
				
				// now dump the offscreen world to the onscreen window in the most efficient way possible
				{
					ControlRef		root = null;
					long			quickTimeVersion = 0L;
					Boolean			isQuickTimeAvailable = (Gestalt(gestaltQuickTimeVersion, &quickTimeVersion) == noErr),
									drawUsingCopyBits = true;
					OSErr			rootError = noErr;
					
					
					rootError = GetRootControl(dataPtr->window, &root);
					
					// determine how to draw
				#if ONLY_LITE_CARBON_LIB
					drawUsingCopyBits = true; // LiteCarbonLib cannot be linked alongside QuickTimeLib (symbolic collisions)
				#else
					// when using QuickTime, it is difficult to constrain drawing to the control alone
					// (at least, I couldnt find an easy way); so, unless the desired control to draw
					// is the root control (which spans the entire window), QuickTime cannot be used
					// even if it is available
					if (rootError == noErr) drawUsingCopyBits = ((!isQuickTimeAvailable) || (root != dataPtr->control));
				#endif
					
					if (!drawUsingCopyBits)
					{
						// Use QuickTimes Image Compression Manager for ultra-quick (and possibly
						// graphics-accelerated) offscreen-to-onscreen copying (not to mention a
						// hell of a lot less code!).  This is the only Carbon-compliant solution
						// (although supposedly Apple is planning to tweak the hell out of the
						// CopyBits() routine because so many gigabytes of existing code uses it,
						// hence such weird things like the GetPortBitMapForCopyBits() accessor).
						ImageDescriptionHandle		image = null;
						
						
						DrawControlInCurrentPort(root);
						SetGWorld(colorWindowPort, null/* device */);
						MakeImageDescriptionForPixMap(pixels, &image);
						dataPtr->result = DecompressImage(GetPixBaseAddr(pixels), image, GetPortPixMap(colorWindowPort),
															null/* source rectangle - null means decompress entire image */,
															&drawingRect/* destination rectangle */,
															dataPtr->dumpMode/* drawing mode */, null/* mask */);
						DisposeHandle((Handle)image), image = null;
					}
					else
					{
						// no QuickTime, so use CopyBits() as usual
						RGBColor		oldForeColor,
										oldBackColor,
										solid;
						
						
						// when using CopyBits(), it is easy to constrain drawing to the control alone
						DrawControlInCurrentPort(dataPtr->control);
						SetGWorld(colorWindowPort, null/* device */);
						
						// set background to white and foreground to black, so CopyBits() doesnt screw anything up
						GetForeColor(&oldForeColor);
						GetBackColor(&oldBackColor);
						
						// dump the offscreen to the onscreen; check for errors, since in such case a simple control draw can be issued
						solid.red = solid.blue = solid.green = 0;
						RGBForeColor(&solid);
						solid.red = solid.blue = solid.green = RGBCOLOR_INTENSITY_MAX;
						RGBBackColor(&solid);
						
						CopyBits(GetPortBitMapForCopyBits(world)/* source pixel map */,
								GetPortBitMapForCopyBits(colorWindowPort)/* destination pixel map */,
								&controlRect/* source rectangle */,
								&controlRect/* destination rectangle */,
								dataPtr->dumpMode/* drawing mode */, (RgnHandle)null/* mask */);
						dataPtr->result = QDError();
						
						RGBForeColor(&oldForeColor);
						RGBBackColor(&oldBackColor);
					}
				}
				
				// the pixel map is no longer being directly used
				UnlockPixels(pixels);
			}
			else dataPtr->result = memPCErr;
			
			// restore to saved
			SetGWorld(oldColorPort, oldDevice);
			if (world != null) DisposeGWorld(world), world = null;
		}
	}
	
#if 0
	if (dataPtr->result != noErr)
	{
		// send an Apple Event to MacTelnet to print the error
		Str255		string;
		Str31		numString;
		
		
		PLstrcpy(string, "\poffscreen dump error = %a");
		NumToString(dataPtr->result, numString);
		StringUtilities_PMetaSetDefaults();
		StringUtilities_PBuild(string, numString, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
		Alert_DebugSendStringToSelf(string, 'KevG', '?WrL');
	}
#endif
	
	SetPort(oldPort);
}// offscreenDumpDeviceLoop


/**
This method is in standard OffscreenOperationProcPtr
form, and will use "inIsEnabled" (data 1) as the
enabled state (true or false) for the given control.
If nonzero, the specified control of the specified
window is activated; otherwise, it is deactivated.
If the desired control state is already in place,
this method has no effect.

If the operation cannot be completed, false is
returned; otherwise, true is returned.

(1.0)
*/
static pascal Boolean
setControlActiveOperation	(const ControlRef		inSpecificControlOrRoot,
							 const SInt16			inColorDepth,
							 const SInt16			inDeviceFlags,
							 GDHandle				inTargetDevice,
							 const SInt32			inIsEnabled,
							 const SInt32			inData2)
{
	Boolean		result = pascal_true;
	
	
	if (inSpecificControlOrRoot != null)
	{
		if (inIsEnabled != IsControlActive(inSpecificControlOrRoot))
		{
			if (inIsEnabled) (OSErr)ActivateControl(inSpecificControlOrRoot);
			else (OSErr)DeactivateControl(inSpecificControlOrRoot);
		}
	}
	else result = pascal_false;
	
	return result;
}// setControlActiveOperation


/**
This method is in standard OffscreenOperationProcPtr
form, and will use "inControlToDisplay" (data 1) as
the control that should be displayed (while the
specified control "inControlToHide" is hidden).

If the operation cannot be completed, false is
returned; otherwise, true is returned.

(1.0)
*/
static pascal Boolean
swapControlsOperation		(const ControlRef		inControlToDisplay,
							 const SInt16			inColorDepth,
							 const SInt16			inDeviceFlags,
							 GDHandle				inTargetDevice,
							 const SInt32			inControlToHide,
							 const SInt32			inData2)
{
	OSErr			error = noErr;
	Boolean			result = pascal_true;
	
	
	if (inControlToHide != null)
	{
		error = SetControlVisibility((ControlRef)inControlToHide, false/* visibility */, true/* draw */);
		if (error != noErr) result = pascal_false;
		error = SetControlVisibility(inControlToDisplay, true/* visibility */, true/* draw */);
		if (error != noErr) result = pascal_false;
	}
	
	return result;
}// swapControlsOperation