/*###############################################################

	Errors.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <AppleEvents.h>
#include <ASRegistry.h>
#include <Dialogs.h>
#include <Scrap.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "Clipboard.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "InstallAE.h"
#include "InternalEvents.h"
#include "TelnetErrors.h"
#include "TelnetMain.h"
#include "Terminology.h"



// variables
static ModalFilterUPP	gAlertEventUPP = null;

// internal methods
static char* getErrorString		(short		inCode,
								 char*		outString);



//
// public methods
//

/*!
Call this routine before calling any other
routines from this module.

(3.0)
*/
void
Errors_Init	(void)
{
	gAlertEventUPP = NewModalFilterUPP(StandardAlertEventFilter);
}// Init


/*!
Call this routine when you are finished
using this module.

(3.0)
*/
void
Errors_Done	(void)
{
	DisposeModalFilterUPP(gAlertEventUPP), gAlertEventUPP = null;
}// Done


/*!
Displays the specified Note alert to the user
with no help text, emitting no system beep.  If
you do not want the alert to have title text,
pass -1 for the last parameter.

IMPORTANT:	Although the string message can
			come from any string list resource,
			the alert title index (if not -1)
			is assumed to come from the MacTelnet
			alert title string list resource.

(3.0)
*/
void
Errors_DisplayNoteMessage	(SInt16 const			inStringListResourceID,
							 UInt16 const			inStringIndex,
							 SInt16 const			inTitleStringIndex)
{
	Str255						dialogText,
								alertTitle;
	
	
	GetIndString(dialogText, inStringListResourceID, inStringIndex);
	if (inTitleStringIndex >= 0)
	{
		GetIndString(alertTitle, rStringsAlertTitles, inTitleStringIndex);
	}
	
	// send a Display Alert Apple Event
	{
		TelnetAlertParameterBlock	parameters;
		
		
		Memory_WriteZero((Ptr)&parameters, sizeof(parameters));
		parameters.dialog.type = kAlertNoteAlert;
		parameters.text.large = dialogText;
		if (inTitleStringIndex >= 0) parameters.text.title = alertTitle;
		(SInt16)Errors_DisplayParameterizedAlert(&parameters);
	}
}// DisplayNoteMessage


/*!
Displays a fatal alert to the user, with the
specified error information, in a standard
OS Protection Mechanism dialog box.  The user
has the choice to copy the given description
sentence to the clipboard.

IMPORTANT:	This routine does not return; it
			calls ExitToShell().

(3.0)
*/
void
Errors_DisplayOSProtectionMessage	(StringPtr		inErrorName,
									 StringPtr		inDescriptionSentence)
{
	InterfaceLibAlertRef		box = null;
	Str255						dialogText,
								helpText;
	Str31						alertTitle;
	
	
	box = Alert_New();
	Alert_SetParamsFor(box, kAlert_StyleCancel);
	{
		Str31		buttonStringQuit,
					buttonStringCopy;
		
		
		GetIndString(buttonStringQuit, rStringsSpecialButtons, siButtonQuit);
		Alert_SetButtonText(box, kAlertStdAlertOKButton, buttonStringQuit);
		GetIndString(buttonStringCopy, rStringsSpecialButtons, siButtonCopy);
		Alert_SetButtonText(box, kAlertStdAlertOtherButton, buttonStringCopy);
	}
	
	// set up miscellaneous alert parameters
	Alert_SetFilterProc(box, gAlertEventUPP);
	Alert_SetHelpButton(box, false);
	GetIndString(alertTitle, rStringsAlertTitles, siAlertTitleFatalErrorOSProtectionMechanism);
	Alert_SetTitle(box, alertTitle);
	Alert_SetType(box, kAlertStopAlert);
	
	// construct the message
	GetIndString(dialogText, rStringsOSProtectionMechanismMessages, siOSProtectionGenericMessageText);
	StringUtilities_PMetaSetDefaults();
	StringUtilities_PBuild(dialogText, inErrorName, inDescriptionSentence, EMPTY_PSTRING, EMPTY_PSTRING);
	GetIndString(helpText, rStringsOSProtectionMechanismMessages, siOSProtectionHelpText);
	Alert_SetText(box, dialogText, helpText);
	Console_WriteValuePString("Fatal", inDescriptionSentence);
	
	// show it
	BEEP;
	Alert_Display(box);
	
	if (Alert_ItemHit(box) == kAlertStdAlertOtherButton)
	{
		// Copy to Clipboard was requested
	#if TARGET_API_MAC_OS8
		(OSStatus)ZeroScrap();
		PutScrap(StrLength(inDescriptionSentence), kScrapFlavorTypeText,
				(StringPtr)(inDescriptionSentence + 1)); // skip length byte
	#else
		(OSStatus)ClearCurrentScrap();
		{
			ScrapRef		currentScrap = null;
			
			
			if (GetCurrentScrap(&currentScrap) == noErr)
			{
				(OSStatus)PutScrapFlavor(currentScrap, kScrapFlavorTypeText,
											kScrapFlavorMaskNone,
											StrLength((StringPtr)inDescriptionSentence),
											(StringPtr)(inDescriptionSentence + 1)/* skip length byte */);
			}
		}
	#endif
	}
	
	Alert_Dispose(&box);
	ExitToShell(); // does not return!
}// DisplayOSProtectionMessage


/*!
To display an alert described in a parameter block by
sending a Display Alert Apple Event back to MacTelnet,
use this method.  The return value is one of the
standard Mac OS "kAlertStdAlert...Button" constants,
or -1 to indicate that the alert timed out with no
specific button being chosen.

Incomplete.

(3.0)
*/
SInt16
Errors_DisplayParameterizedAlert	(TelnetAlertParameterBlockPtr const		inParameters)
{
	SInt16			result = -1; // a "kAlertStdAlert...Button" constant, or -1 for timed out
	
	
	if (inParameters != null)
	{
		// send a Display Alert Apple Event
		AppleEvent		displayAlertEvent,
						reply;
		StringPtr		string = null;
		
		
		Alert_ReportOSErr(InstallAE_CreateRecordableAppleEvent(kTelnetEventClass,
																kTelnetEventIDAlert,
																&displayAlertEvent));
		
		// text parameters
		{
			// dialog text - REQUIRED
			string = inParameters->text.large;
			(OSErr)AEPutParamPtr(&displayAlertEvent, keyDirectObject,
									cString, string + 1, StrLength(string) * sizeof(unsigned char));
		}
		if (inParameters->text.small != null)
		{
			// help text
			string = inParameters->text.small;
			(OSErr)AEPutParamPtr(&displayAlertEvent, kTelnetParameterKeyAlertHelpText,
									cString, string + 1, StrLength(string) * sizeof(unsigned char));
		}
		if (inParameters->text.title != null)
		{
			// alert title
			string = inParameters->text.title;
			(OSErr)AEPutParamPtr(&displayAlertEvent, kTelnetParameterKeyAlertTitle,
									cString, string + 1, StrLength(string) * sizeof(unsigned char));
		}
		
		// button parameters - UNIMPLEMENTED
		{
		}
		
		// flag parameters
		{
			// alert sound
			Boolean		alertBeeps = inParameters->flags.alertSound;
			
			
			(OSErr)AEPutParamPtr(&displayAlertEvent, kTelnetParameterKeyAlertBeeps,
									typeBoolean, &alertBeeps, sizeof(alertBeeps));
		}
		{
			// help button
			Boolean		isHelpButton = inParameters->flags.helpButton;
			
			
			(OSErr)AEPutParamPtr(&displayAlertEvent, kTelnetParameterKeyAlertHasHelpButton,
									typeBoolean, &isHelpButton, sizeof(isHelpButton));
		}
		{
			// network disruption
			Boolean		isNetworkDisruption = inParameters->flags.networkDisruption;
			
			
			(OSErr)AEPutParamPtr(&displayAlertEvent, kTelnetParameterKeyAlertDisruptsNetwork,
									typeBoolean, &isNetworkDisruption, sizeof(isNetworkDisruption));
		}
		{
			// speech
			Boolean		isSpeech = inParameters->flags.speech;
			
			
			(OSErr)AEPutParamPtr(&displayAlertEvent, kTelnetParameterKeyAlertSpeech,
									typeBoolean, &isSpeech, sizeof(isSpeech));
		}
		{
			// title bar
			Boolean		isMovable = !inParameters->flags.notTitled;
			
			
			(OSErr)AEPutParamPtr(&displayAlertEvent, kTelnetParameterKeyAlertHasTitleBar,
									typeBoolean, &isMovable, sizeof(isMovable));
		}
		
		// icon parameters
		{
			// alert icon
			FourCharCode	alertType = kTelnetEnumeratedClassAlertIconTypeNote;
			Boolean			doSet = true;
			
			
			switch (inParameters->dialog.type)
			{
				case kAlertNoteAlert:
					alertType = kTelnetEnumeratedClassAlertIconTypeNote;
					break;
				
				case kAlertStopAlert:
					alertType = kTelnetEnumeratedClassAlertIconTypeStop;
					break;
				
				case kAlertCautionAlert:
					alertType = kTelnetEnumeratedClassAlertIconTypeCaution;
					break;
				
				default:
					doSet = false;
					break;
			}
			if (doSet)
			{
				(OSErr)AEPutParamPtr(&displayAlertEvent, kTelnetParameterKeyAlertIcon,
										typeEnumerated, &alertType, sizeof(alertType));
			}
		}
		
		// display the alert by sending an Apple Event; if it fails, report the error
		Alert_ReportOSErr(AESend(&displayAlertEvent, &reply,
									kAENoReply | kAENeverInteract,// | kAEDontRecord,
									kAENormalPriority, kAEDefaultTimeout, null, null));
		AEDisposeDesc(&displayAlertEvent);
		
		// reply - unimplemented
	}
	
	return result;
}// DisplayParameterizedAlert


/*!
Displays the specified Stop alert to the user
with no help text, emitting a system beep.  If
you do *not* want the alert to have title text,
pass -1 for the last parameter.

IMPORTANT:	Although the string message can
			come from any string list resource,
			the alert title index (if not -1)
			is assumed to come from the Telnet
			alert title string list resource.

(3.0)
*/
void
Errors_DisplayStopMessage	(SInt16 const			inStringListResourceID,
							 UInt16 const			inStringIndex,
							 SInt16 const			inTitleStringIndex)
{
	InterfaceLibAlertRef		box = null;
	Str255						dialogText,
								alertTitle;
	
	
	GetIndString(dialogText, inStringListResourceID, inStringIndex);
	if (inTitleStringIndex >= 0)
	{
		GetIndString(alertTitle, rStringsAlertTitles, inTitleStringIndex);
	}
	
	// send a Display Alert Apple Event
	{
		TelnetAlertParameterBlock	parameters;
		
		
		Memory_WriteZero((Ptr)&parameters, sizeof(parameters));
		parameters.dialog.type = kAlertStopAlert;
		parameters.flags.alertSound = true;
		parameters.text.large = dialogText;
		if (inTitleStringIndex >= 0) parameters.text.title = alertTitle;
		(SInt16)Errors_DisplayParameterizedAlert(&parameters);
	}
}// DisplayStopMessage


/*!
Displays the specified Stop alert to the user
with no help text, emitting a system beep.  If
you do *not* want the alert to have title text,
pass -1 for the last parameter.

When the user clicks the Quit button in the
alert, the program is forced to quit.  Thus,
this routine will not return.

IMPORTANT:	Although the string message can
			come from any string list resource,
			the alert title index (if not -1)
			is assumed to come from the MacTelnet
			alert title string list resource.

WARNING:	Do not implement this routine using
			Apple Events; it may need to be
			called early in the initialization
			code, possibly before Apple Events
			are configured properly.  Display
			this alert the old-fashioned way.

(3.0)
*/
void
Errors_DisplayStopQuitMessage	(SInt16 const			inStringListResourceID,
								 UInt16 const			inStringIndex,
								 SInt16 const			inTitleStringIndex)
{
	InterfaceLibAlertRef		box = null;
	Str255						dialogText,
								alertTitle;
	Str31						buttonString;
	
	
	GetIndString(dialogText, inStringListResourceID, inStringIndex);
	if (inTitleStringIndex >= 0)
	{
		GetIndString(alertTitle, rStringsAlertTitles, inTitleStringIndex);
	}
	GetIndString(buttonString, rStringsSpecialButtons, siButtonQuit);
	box = Alert_New();
	Alert_SetParamsFor(box, kAlert_StyleCancel);
	Alert_SetButtonText(box, kAlertStdAlertOKButton, buttonString);
	Alert_SetFilterProc(box, gAlertEventUPP);
	Alert_SetHelpButton(box, false);
	if (inTitleStringIndex >= 0) Alert_SetTitle(box, alertTitle);
	Alert_SetText(box, dialogText, EMPTY_PSTRING);
	Alert_SetType(box, kAlertStopAlert);
	Console_WriteValuePString("Fatal", dialogText);
	BEEP;
	Alert_Display(box);
	Alert_Dispose(&box);
	
	// shut down MacTelnet
	Main_QuitImmediately();
}// DisplayStopQuitMessage


/*!
Rewritten in MacTelnet 3.0 to use standard
Appearance-savvy alert windows.  The given
message ID must refer to a string index in
the string list resource of ID
"rStringsOperationFailed".

(3.0)
*/
void
OperationFailedAlert	(short		inMessageID,
						 short		inInternalID,
						 short		inOSID)
{
	Str255		dialogText,
				helpText,
				errorString;
	Str31		osErrorString,
				internalErrorString;
	
	
	GetIndString(dialogText, rStringsGeneralMessages, siCommandCannotContinue);
	GetIndString(errorString, rStringsOperationFailed, inMessageID);
	StringUtilities_PMetaSetDefaults();
	StringUtilities_PBuild(dialogText, errorString, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
	GetIndString(helpText, rStringsGeneralMessages, siOSErrAndInternalErrHelpText);
	NumToString((long)inOSID, osErrorString);
	NumToString((long)inInternalID, internalErrorString);
	StringUtilities_PMetaSetDefaults();
	StringUtilities_PBuild(helpText, osErrorString, internalErrorString, EMPTY_PSTRING, EMPTY_PSTRING);
	Console_WriteValuePString("Failed", dialogText);
	Console_WriteValuePString("Reason", helpText);
	
	// send a Display Alert Apple Event
	{
		TelnetAlertParameterBlock	parameters;
		
		
		Memory_WriteZero((Ptr)&parameters, sizeof(parameters));
		parameters.dialog.type = kAlertStopAlert;
		parameters.flags.alertSound = true;
		parameters.text.large = dialogText;
		parameters.text.small = helpText;
		(SInt16)Errors_DisplayParameterizedAlert(&parameters);
	}
}// OperationFailedAlert


/*!
Reports an out of memory error.  The specific
point in the MacTelnet source code where the error
took place is hinted by the internal error code
(see "TelnetDescriptors.h" for valid values).

(3.0)
*/
void
Errors_OutOfMemory	(InternalTelnetMemoryError const		inInternalID)
{
	OperationFailedAlert(siOpFailedOutOfMemory, inInternalID, 0);
}// OutOfMemory