/*###############################################################

	EventLoop.c
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Gestalt.h>
#include <Icons.h>
#include <ImageCompression.h>
#include <Navigation.h>
#include <Notification.h>
#include <PLStringFuncs.h>
#include <Resources.h>
#include <Script.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "GeneralResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "BasicTypesAE.h"
#include "Clipboard.h"
#include "CommandLine.h"
#include "Commands.h"
#include "ConnectionData.h"
#include "Connections.h"
#include "Console.h"
#include "ContextualMenuBuilder.h"
#include "CoolStuff.h"
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "FTPLog.h"
#include "FTPBin.h"
#include "Inspector.h"
#include "InstallAE.h"
#include "InternalEvents.h"
#include "Keypads.h"
#include "LineMode.h"
#include "MacLook.h"
#include "MacroSetupDialog.h"
#include "NetBackground.h"
#include "NetEvents.h"
#include "network.proto.h"
#include "Parse.h"
#include "Preferences.h"
#include "PreferencesDialog.h"
#include "RasterGraphicsScreen.h"
#include "RealScreen.h"
#include "ScreenManager.h"
#include "SizeDialog.h"
#include "TektronixRealGraphics.h"
#include "TelnetDescriptors.h"
#include "TelnetErrors.h"
#include "TelnetMenus.h"
#include "Terminology.h"
#include "Translation.h"
#include "VirtualScreenEmulation.h"
#include "vsdata.h"
#include "TelnetMacros.h"
#include "Terminal.h"
#include "TextTranslation.h"
#include "VSkeys.h"



// constants
typedef UInt32 MouseRegionType;
enum
{
	kMouseRegionTypeNowhere = 0,
	kMouseRegionTypeTerminalTextSelection = 1,
	kMouseRegionTypeTerminalScreen = 2,
	kMouseRegionTypeGraphicsScreen = 3
};

// The index count is to be equal to the cardinality of
// the enumeration below!  WARNING: If you change this
// value, be sure to visit getFloatingWindowInfo() to
// make sure that the code checks whenever arrays are
// given that are too small.
const UInt16 kFloaterIndexCount = 5;
enum
{
	// arrange these indices in the desired front-to-back ordering
	// of floating windows (used by getFloatingWindowInfo())
	kFloaterIndexInspector = 0,
	kFloaterIndexCommandLine = 1,
	kFloaterIndexControlKeys = 2,
	kFloaterIndexFunction = 3,
	kFloaterIndexKeypad = 4
};

// variables
static MouseRegionType	gMouseRegionCursor = kMouseRegionTypeNowhere;
static RgnHandle		gMouseRegion = null;
static SInt16			gHaveInstalledNotification = 0;
static SInt32			gCurrentScript = 0L,
						gDefaultKCHR = 0L,
						gMungeCount = 0L;
static Boolean			gKCHRHasChanged = false;
static NMRec*			gBeepNotificationPtr = null;

// These are translation tables, to translate key codes to virtual screen codes.
// There are two tables, for keys pressed without or with the Shift key.  The
// base is 0x41, the code for the keypad period (.).  The maximum is 0x7E, the
// code for the up arrow.  This includes everything but the main keyboard parts.
static unsigned char kpxlate[2][62] =
{
	{
		// virtual key codes are in comments
		VSKP,	// 0x41
		VSRT,	// 0x42 (Mac Plus)
		VSF4,	// 0x43 (ADB)
		0,		// 0x44
		VSKC,	// 0x45 (ADB std)
		VSLT,	// 0x46 (Mac Plus)
		VSF1,	// 0x47
		VSDN,	// 0x48 (Mac Plus)
		0,		// 0x49
		0,		// 0x4A
		VSF3,	// 0x4B (ADB)
		VSKE,	// 0x4C
		VSUP,	// 0x4D (Mac Plus)
		VSKM,	// 0x4E
		0,		// 0x4F
		0,		// 0x50
		VSF2,	// 0x51 (ADB)
		VSK0,	// 0x52
		VSK1,	// 0x53
		VSK2,	// 0x54
		VSK3,	// 0x55
		VSK4,	// 0x56
		VSK5,	// 0x57
		VSK6,	// 0x58
		VSK7,	// 0x59
		0,		// 0x5A
		VSK8,	// 0x5B
		VSK9,	// 0x5C
		0,		// 0x5D
		0,		// 0x5E
		0,		// 0x5F
		VSF10,	// 0x60	
		VSF11,	// 0x61	
		VSF12,	// 0x62	
		VSF8,	// 0x63
		VSF13,	// 0x64	
		VSF14,	// 0x65	
		0,		// 0x66
		VSF16,	// 0x67	
		0,		// 0x68
		VSF18,	// 0x69	
		0,		// 0x6A
		VSF19,	// 0x6B	
		0,		// 0x6C
		VSF15,	// 0x6D	
		0,		// 0x6E
		VSF17,	// 0x6F	
		0,		// 0x70
		VSF20,	// 0x71	
		VSHELP,	// 0x72	
		VSHOME,	// 0x73	
		VSPGUP,	// 0x74	
		VSDEL,	// 0x75 (ADB ext)
		VSF9,	// 0x76
		VSEND,	// 0x77	
		VSF7,	// 0x78
		VSPGDN,	// 0x79	
		VSF6,	// 0x7A
		VSLT,	// 0x7B
		VSRT,	// 0x7C
		VSDN,	// 0x7D
		VSUP	// 0x7E
	},
	{
		// virtual key codes are in comments
		VSKP,	// 0x41
		VSF4,	// 0x42 (Mac Plus)
		VSF4,	// 0x43 (ADB)
		0,		// 0x44
		VSKC,	// 0x45 (ADB)
		VSKC,	// 0x46 (Mac Plus)
		VSF1,	// 0x47
		VSF2,	// 0x48
		0,		// 0x49
		0,		// 0x4A
		VSF3,	// 0x4B
		VSKE,	// 0x4C
		VSF3,	// 0x4D
		VSKM,	// 0x4E
		0,		// 0x4F
		0,		// 0x50
		VSF2,	// 0x51
		VSK0,	// 0x52
		VSK1,	// 0x53
		VSK2,	// 0x54
		VSK3,	// 0x55
		VSK4,	// 0x56
		VSK5,	// 0x57
		VSK6,	// 0x58
		VSK7,	// 0x59
		0,		// 0x5A
		VSK8,	// 0x5B
		VSK9,	// 0x5C
		0,		// 0x5D
		0,		// 0x5E
		0,		// 0x5F
		VSF10,	// 0x60	
		VSF11,	// 0x61	
		VSF12,	// 0x62	
		VSF8,	// 0x63
		VSF13,	// 0x64	
		VSF14,	// 0x65	
		0,		// 0x66
		VSF16,	// 0x67	
		0,		// 0x68
		VSF18,	// 0x69	
		0,		// 0x6A
		VSF19,	// 0x6B	
		0,		// 0x6C
		VSF15,	// 0x6D	
		0,		// 0x6E
		VSF17,	// 0x6F	
		0,		// 0x70
		VSF20,	// 0x71	
		VSHELP,	// 0x72	
		VSHOME,	// 0x73	
		VSPGUP,	// 0x74	
		VSDEL,	// 0x75 (ADB ext)
		VSF9,	// 0x76
		VSEND,	// 0x77	
		VSF7,	// 0x78
		VSPGDN,	// 0x79	
		VSF6,	// 0x7A
		VSLT,	// 0x7B
		VSRT,	// 0x7C
		VSDN,	// 0x7D
		VSUP	// 0x7E
	}
};
  
// ASCII character set defines
#define	KILLCHAR	0x15	/* the character to kill the local line with */

// internal methods
static void			checkDefaultScriptSettings		(void);
static Boolean		checkPageKeys					(WindowRef const		inWindow,
													 short const			inCode);
static void			diffFloatingRegions				(RgnHandle				inoutRegion);
static void			emptyGlobalMouseRegion			(void);
static void			ensureFloatingWindowsFloat		(void);
static OSErr		getFloatingWindowInfo			(WindowRef*				inoutWindowArray,
													 Boolean*				inoutVisibilityArray,
													 SInt16 const			inArrayCardinality);
static RgnHandle	getMouseRegion					(void);
static Boolean		handleActivate					(EventRecord*			inoutEventPtr,
													 Boolean const			inActivating);
static Boolean		handleDiskEvent					(EventRecord const*		inEventPtr);
static Boolean		handleKeyDown					(EventRecord*			inoutEventPtr);
static void			handleMouseDown					(EventRecord*			inoutEventPtr);
static void			handleScreenKeyDown				(EventRecord*			inoutEventPtr,
													 ConnectionDataPtr		inConnectionDataPtr);
static Boolean		handleSwitch					(EventRecord*			inoutEventPtr);
static Boolean		handleUpdate					(EventRecord*			inoutEventPtr);
static void			localRectToGlobalMouseRegion	(MouseRegionType const	inAsWhatRegion,
													 WindowRef				inWindowRef,
													 Rect const*			inLocalRectPtr);
static void			localRegionToGlobalMouseRegion	(MouseRegionType const	inAsWhatRegion,
													 WindowRef				inWindowRef,
													 RgnHandle				inLocalRgnHandle);
static void			updateCursor					(EventRecord*			inoutEventPtr,
													 Boolean				inForceCursor);
static void			updateWindowBalloonVisibilities	(EventRecord*			inoutEventPtr);



//
// public methods
//

/*!
Call this method at the start of the program,
before it is necessary to track events.  This
method creates the global mouse region, among
other things.

(3.0)
*/
void
EventLoop_Init	(void)
{
	gMouseRegion = Memory_NewRegion();
	
	gCurrentScript = GetScriptManagerVariable(smKeyScript); // get the active script
	gDefaultKCHR = GetScriptVariable(smRoman, smScriptKeys); // get the smRoman default KCHR
	gMungeCount = GetScriptManagerVariable(smMunged);
	gKCHRHasChanged = false;
}// Init


/*!
Call this method at the end of the program,
when other clean-up is being done and it is
no longer necessary to track events.

(3.0)
*/
void
EventLoop_Done	(void)
{
	Memory_DisposeRegion(&gMouseRegion);
}// Done


/*!
To determine the absolutely current state of
the modifier keys and the mouse button, use
this method.  The modifiers are somewhat
incomplete (for example, no checking is done
for the right shift key, etc.), but all
the common modifiers (command, control,
shift, option, caps lock, and the mouse
button) are returned.

(3.0)
*/
EventModifiers
EventLoop_CurrentModifiers	(void)
{
	SInt16			capsLockDown = 0,
					commandDown = 0,
					controlDown = 0,
					optionDown = 0,
					shiftDown = 0;
	KeyMap			keyMap;
	EventModifiers	result = 0;
	
	
	GetKeys(keyMap);
	commandDown = ((unsigned char*)keyMap)[6] & 0x80; // virtual key 0x37 == bit #55 of 128; the 55->48 (7th) octet, mask 10000000
	shiftDown = ((unsigned char*)keyMap)[7] & 0x01; // virtual key 0x38 == bit #56 of 128; the 63->56 (8th) octet, mask 00000001
	capsLockDown = ((unsigned char*)keyMap)[7] & 0x02; // virtual key 0x39 == bit #57 of 128; the 63->56 (8th) octet, mask 00000010
	optionDown = ((unsigned char*)keyMap)[7] & 0x04; // virtual key 0x3A == bit #58 of 128; the 63->56 (8th) octet, mask 00000100
	controlDown = ((unsigned char*)keyMap)[7] & 0x08; // virtual key 0x3B == bit #59 of 128; the 63->56 (8th) octet, mask 00001000
	if (commandDown) result |= cmdKey;
	if (shiftDown) result |= shiftKey;
	if (capsLockDown) result |= alphaLock;
	if (optionDown) result |= optionKey;
	if (controlDown) result |= controlKey;
	if (Button()) result |= btnState;
	
	return result;
}// CurrentModifiers


/*!
To determine the state of the  key when
you do not have access to an event record,
use this method.  If the key is down, true
is returned; otherwise, false is returned.

(1.0)
*/
Boolean
EventLoop_IsCommandKeyDown		(void)
{
	KeyMap		keyMap;
	Boolean		result = false;
	
	
	// The GetKeys() result is an array of four 32-bit
	// unsigned numbers, for a total of 128 bits.  The
	// particular bit corresponding to a key is its
	// virtual key code.  Thus, since the Command key
	// has a virtual key code of 0x37 (55), it is "down"
	// if bit number 55 is a 1.  For a set of virtual
	// key codes, see page 2-43 of the book "Inside
	// Macintosh: Macintosh Toolbox Essentials".
	GetKeys(keyMap);
	result = ((unsigned char*)keyMap)[6] & 0x80; // virtual key 0x37 == bit #55 of 128; the 55->48 (7th) octet, MSB, mask 10000000
	
	return result;
}// EventLoop_IsCommandKeyDown


/*!
This routine gets the next key event from the
event queue and returns true if it is a "cancel"
(command-period) event.  Use this to monitor
the keyboard for user cancellations while a
dialog box is displayed and/or the normal event
loop is not running.

(3.0)
*/
Boolean
EventLoop_IsNextCancel	(void)
{
	Boolean			result = false;
	EventRecord		event;
	
	
	while (WaitNextEvent(keyDownMask | keyUpMask, &event, 4/* ticks wait */, null))
	{
		if ((event.modifiers & cmdKey) && (event.message & 0xFF) == '.') result = true;
	}
	return result;
}// IsNextCancel


/*!
To determine the state of the Option key
when you do not have access to an event
record, use this method.  If the key is
down, true is returned; otherwise, false
is returned.

(3.0)
*/
Boolean
EventLoop_IsOptionKeyDown	(void)
{
	KeyMap		keyMap;
	Boolean		result = false;
	
	
	// The GetKeys() result is an array of four 32-bit
	// unsigned numbers, for a total of 128 bits.  The
	// particular bit corresponding to a key is its
	// virtual key code.  Thus, since the Option key
	// has a virtual key code of 0x3A (58), it is "down"
	// if bit number 58 is a 1.  For a set of virtual
	// key codes, see page 2-43 of the book "Inside
	// Macintosh: Macintosh Toolbox Essentials".
	GetKeys(keyMap);
	result = ((unsigned char*)keyMap)[7] & 0x04; // virtual key 0x3A == bit #58 of 128; the 63->56 (8th) octet, 3rd from LSB, mask 00000100
	
	return result;
}// IsOptionKeyDown


/*!
To determine the state of the Shift key
when you do not have access to an event
record, use this method.  If the key is
down, true is returned; otherwise, false
is returned.

(3.0)
*/
Boolean
EventLoop_IsShiftKeyDown	(void)
{
	KeyMap		keyMap;
	Boolean		result = false;
	
	
	// The GetKeys() result is an array of four 32-bit
	// unsigned numbers, for a total of 128 bits.  The
	// particular bit corresponding to a key is its
	// virtual key code.  Thus, since the Shift key
	// has a virtual key code of 0x38 (56), it is "down"
	// if bit number 56 is a 1.  For a set of virtual
	// key codes, see page 2-43 of the book "Inside
	// Macintosh: Macintosh Toolbox Essentials".
	GetKeys(keyMap);
	result = ((unsigned char*)keyMap)[7] & 0x01; // virtual key 0x38 == bit #56 of 128; the 63->56 (8th) octet, LSB, mask 00000001
	
	return result;
}// IsShiftKeyDown


/*!
Calls WaitNextEvent(), looking for any kind
of event, and handles the next event it
receives.  Also manages rudimentary tasks
like cursor flashing.

(3.0)
*/
void
EventLoop_Service	(void)
{
	Boolean				isEvent = false,
						isScreen = ScreenManager_IsWindowAScreen(GetRealFrontWindow());
	EventRecord			event;
	UInt32				ticksWaitNextEvent = 0L;
	Boolean				cursorBlinks = false;
	Size				actualSize = 0L;
	
	
	// get the users process service preference, if possible
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_TicksWaitNextEvent,
								sizeof(ticksWaitNextEvent), &ticksWaitNextEvent,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		ticksWaitNextEvent = 4; // default ticks if preference cant be found
	}
	
	isEvent = WaitNextEvent(everyEvent, &event, ticksWaitNextEvent,
							getMouseRegion()/* mouse region, to know when to change the cursor */);
	
	// get the users Cursor Flashing preference, if possible
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_CursorBlinks,
								sizeof(cursorBlinks), &cursorBlinks,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		cursorBlinks = false; // assume cursor shouldnt blink, if preference cant be found
	}
	
	// ensure that the cursor blinks normally, but remains visible while the user types
	if (cursorBlinks && isScreen)
	{
		TelnetScreenID		screenID = ScreenManager_GetWindowScreenID(GetRealFrontWindow());
		
		
		if (isEvent)
		{
			// Make sure the cursor is visible as soon as a key is pressed,
			// and make it disappear as soon as the users typing frenzy ends.
			if ((!(event.modifiers & cmdKey)) &&
				((event.what == keyDown) || (event.what == autoKey)))
			{
				RealScreen_SetCursorVisibility(screenID, true/* visible */);
			}
			else if ((!(event.modifiers & cmdKey)) &&
						(event.what == keyUp))
			{
				RealScreen_SetCursorVisibility(screenID, false/* visible */);
			}
		}
	}
	
	if (isEvent) (Boolean)EventLoop_HandleEvent(&event);
	
	// check for changed modifier keys, which will force the mouse cursor to
	// update (obviously, the cursor is only important if Telnet is frontmost)
	if (!FlagManager_Test(kFlagSuspended))
	{
		static short	optWasDown = 0,
						ctlWasDown = 0,
						cmdWasDown = 0;
		short			optDown = 0,
						ctlDown = 0,
						cmdDown = 0;
		KeyMap			keyMap;
		
		
		// The GetKeys() result is an array of four 32-bit
		// unsigned numbers, for a total of 128 bits.  The
		// particular bit corresponding to a key is its
		// virtual key code.  Thus, since the Option key
		// has a virtual key code of 0x3A (58), it is "down"
		// if bit number 58 is a 1.  For a set of virtual
		// key codes, see page 2-43 of the book "Inside
		// Macintosh: Macintosh Toolbox Essentials".
		GetKeys(keyMap);
		cmdDown = ((unsigned char*)keyMap)[6] &0x80; // virtual key 0x37 == bit #55 of 128; the 55->48 (7th) octet, MSB, mask 10000000
		optDown = ((unsigned char*)keyMap)[7] &0x04; // virtual key 0x3A == bit #58 of 128; the 63->56 (8th) octet, 3rd from LSB, mask 00000100
		ctlDown = ((unsigned char*)keyMap)[7] &0x08; // virtual key 0x3B == bit #59 of 128; the 63->56 (8th) octet, 4th from LSB, mask 00001000
		
		if ((optDown != optWasDown) || (ctlDown != ctlWasDown) || (cmdDown != cmdWasDown))
		{
			// modify the event record to show changes
			event.modifiers = 0;
			if (optDown) event.modifiers |= optionKey;
			if (ctlDown) event.modifiers |= controlKey;
			if (cmdDown) event.modifiers |= cmdKey;
			
			// force the cursor to the appropriate shape, based on the latest modifier key states
			updateCursor(&event, true/* force */);
		}
		
		optWasDown = optDown;
		ctlWasDown = ctlDown;
		cmdWasDown = cmdDown;
	}
	
	// Prevent nonfloating and nondialog windows from getting in front
	// of floating windows.  With Mac OS 8.6 and beyond, the system
	// takes care of this (finally!!!).
	if (!FlagManager_Test(kFlagOS8_6API)) ensureFloatingWindowsFloat();
}// Service


/*!
To change the state of all floating windows at
once, use this method.  This is generally only
done to deactivate floating windows prior to
displaying a dialog box.

(3.0)
*/
void
EventLoop_SetFloatersHilited	(Boolean const		inAreHilited)
{
	WindowRef		window = null;
	
	
	window = CommandLine_GetWindow();
	if (window != null)
	{
		HiliteWindow(window, inAreHilited);
		(OSErr)Embedding_OffscreenSetRootControlActive(window, inAreHilited);
	}
	Inspector_SetSuspended(!inAreHilited);
	Keypads_SetSuspended(kKeypadTypeControlKeys, !inAreHilited);
	Keypads_SetSuspended(kKeypadTypeFunctionKeys, !inAreHilited);
	Keypads_SetSuspended(kKeypadTypeVT220Keys, !inAreHilited);
}// SetFloatersHilited


/*!
This method hides or shows *all* floating
windows belonging to MacTelnet.

(3.0)
*/
void
EventLoop_SetFloatersSuspended		(Boolean const		inAreInvisible)
{
	Boolean		suspending = inAreInvisible; // when suspended, windoids are hidden
	
	
	CommandLine_SetSuspended(suspending);
	Inspector_SetSuspended(suspending);
	Keypads_SetSuspended(kKeypadTypeControlKeys, suspending);
	Keypads_SetSuspended(kKeypadTypeFunctionKeys, suspending);
	Keypads_SetSuspended(kKeypadTypeVT220Keys, suspending);
	
	#ifndef NDEBUG
	// hide the debugging console when in the background
	//DebuggingConsole_SetVisible(inAreInvisible);
	#endif
}// SetFloatersSuspended


/*!
Invoke this routine when a new terminal screen has been
activated, to update the Script Manager settings
appropriately for the new screen.

(3.0)
*/
void
EventLoop_SetScriptFor	(ConnectionDataPtr		inConnectionDataPtr)
{
	checkDefaultScriptSettings(); // see if the script has changed
	
	// this next bit takes care of setting the KCHR based on the EMACS hack pref
	if ((gCurrentScript == smRoman) &&
		(inConnectionDataPtr->metaKey == kEMACSMetaKeyOption))
	{
		if (!gKCHRHasChanged) // if we havent already done this
		{
			SetScriptVariable(gCurrentScript, smScriptKeys, 502); // set the KCHR for EMACS
			KeyScript(smRoman);	// make it active
			gKCHRHasChanged = true;
		}
	}
	else if (gKCHRHasChanged) // new active window doesnt want EMACS hack
	{
		SetScriptVariable(gCurrentScript, smScriptKeys, gDefaultKCHR); 
		KeyScript(smRoman);
		gKCHRHasChanged = false;
	}
}// SetScriptFor


void
SetDefaultKCHR	(void)
{
	if (gKCHRHasChanged)
	{
		SetScriptVariable(smRoman, smScriptKeys, gDefaultKCHR);
		gKCHRHasChanged = false;
		KeyScript(smRoman);
	}
}// SetDefaultKCHR


/*!
This method handles Color Picker 2.0+ event
requests in as simple a way as possible: by
responding to update events.  Time is also
given to other running threads.

(3.0)
*/
pascal Boolean
EventLoop_HandleColorPickerUpdate	(EventRecord*		inoutEventPtr)
{
	Boolean		result = pascal_false; // event handled?
	
	
	switch (inoutEventPtr->what)
	{
		case updateEvt:
			result = EventLoop_HandleEvent(inoutEventPtr);
			break;
		
		default:
			break;
	}
	Threads_YieldToAnother();
	return result;
}// HandleColorPickerUpdate


/*!
This method handles Navigation Services event
requests in as simple a way as possible: by
responding to update events.  Time is also
given to other running threads.

(3.0)
*/
pascal void
EventLoop_HandleNavigationUpdate	(NavEventCallbackMessage		callBackSelector, 
									 NavCBRecPtr					callBackParms, 
									 NavCallBackUserData			UNUSED_ARGUMENT(callBackUD))
{
	if (callBackSelector == kNavCBEvent) switch (callBackParms->eventData.eventDataParms.event->what)
	{
		case updateEvt:
			EventLoop_HandleEvent(callBackParms->eventData.eventDataParms.event);
			break;
		
		default:
			break;
	}
	Threads_YieldToAnother();
}// HandleNavigationUpdate


/*!
Handles all window and modeless dialog box events,
and returns "true" if the specified event was handled.

Note that this routine is very careful in judging the
circumstances under which a value of true is returned.
The Mac OS ultimately assumes that an event that is
"handled" will not be needed by anything else, and
removes such an event from the queue.  The most likely
event for which "true" will get returned is for update
events, because this routine knows if it recognizes
the window requiring updating or not and knows that,
once a window is updated, the event can safely be
discarded.

(2.6)
*/
Boolean
EventLoop_HandleEvent		(EventRecord*		inoutEventPtr)
{
	Boolean		result = false;
	Boolean		isHandledByTSM = false;
	
	
	if (FlagManager_Test(kFlagTextServicesManager))
	{
#if TARGET_API_MAC_OS8
		isHandledByTSM = TSMEvent(inoutEventPtr);
#else
		isHandledByTSM = false; // Carbon handles these events implicitly, so ignore them
#endif
	}
	
	if (!isHandledByTSM) switch (inoutEventPtr->what)
	{
		case mouseDown:
			handleMouseDown(inoutEventPtr);
			result = true;
			break;
		
		case updateEvt: // a window needs all or part of it redrawn
			result = handleUpdate(inoutEventPtr);
			break;
		
		case keyDown:
		case autoKey: // a key was pressed
			result = handleKeyDown(inoutEventPtr);
			break;
		
		case diskEvt: // disk inserted?
			result = handleDiskEvent(inoutEventPtr);
			break;
		
		case activateEvt: // a window was moved in front of or behind another window
			result = handleActivate(inoutEventPtr, inoutEventPtr->modifiers & activeFlag);
			break;
		
		case osEvt: // possibly an application-switch event or mouse-moved event
			switch ((inoutEventPtr->message & osEvtMessageMask) >> 24)
			{
				case suspendResumeMessage:
					result = handleSwitch(inoutEventPtr);
					break;
				
				case mouseMovedMessage:
					updateCursor(inoutEventPtr, false);
					updateWindowBalloonVisibilities(inoutEventPtr);
					result = true;
					break;
				
				default:
					// unknown OS event
					break;
			}
			break;
		
		case kHighLevelEvent: // Apple Events - script commands, for example
			(void)AEProcessAppleEvent(inoutEventPtr);
			result = true;
			break;
		
		default: // who knows WHAT the OS is asking
			break;
	}
	
	return result;
}// HandleEvent


//#pragma profile on

void
CloseAWindow	(WindowRef		theWindow)
{
	SInt16			windowKind = GetWindowKind(theWindow);
	WindowInfoRef	windowInfoRef = WindowInfo_GetFromWindow(theWindow);
	Boolean			weirdWindow = false;
	
	
	if (windowInfoRef != null)
	{
		WindowInfoDescriptor		windowDescriptor = WindowInfo_GetWindowDescriptor(windowInfoRef);
		
		
		switch (windowDescriptor)
		{
			case kTelnetWindowDescriptorClipboard:
				Clipboard_SetVisible(false);
				break;
			
			case kTelnetWindowDescriptorInspectorWindoid:
				Inspector_SetVisible(false);
				break;
			
			case kTelnetWindowDescriptorCommandLine:
				CommandLine_SetVisible(false);
				break;
			
			case kTelnetWindowDescriptorControlWindoid:
				Keypads_SetVisibility(kKeypadTypeControlKeys, false);
				break;
			
			case kTelnetWindowDescriptorFunctionWindoid:
				Keypads_SetVisibility(kKeypadTypeFunctionKeys, false);
				break;
			
			case kTelnetWindowDescriptorKeypadWindoid:
				Keypads_SetVisibility(kKeypadTypeVT220Keys, false);
				break;
			
			case kTelnetDialogDescriptorPreferences:
				PreferencesDialog_Remove();
				break;
			
			default:
				weirdWindow = true;
				break;
		}
	}
	else weirdWindow = true;
	
	if (weirdWindow)
	{
		long				junk = 0L;
		TelnetScreenID		screenID = 0;
		
		
		switch (windowKind)
		{
			case WIN_CONSOLE:
				if (theWindow == Console_GetWindow()) Console_SetVisible(false);
				break;
						
			case WIN_LOG:
				if (theWindow == FTPLog_GetWindow()) FTPLog_SetVisible(false);
				break;
						
			case WIN_CNXN:
				// close the connection, which may or may not actually close the window
				// (depending on the users preferences)
				if ((screenID = ScreenManager_GetWindowScreenID(theWindow)) > 0) 
				{
					ConnectionDataPtr		connectionDataPtr = ScreenManager_GetConnectionDataOf(screenID);
					
					
					if (connectionDataPtr != null)
					{
						if (connectionDataPtr->status == kConnectionStatusDead)
						{
							// connection is already broken; close the window and clean up
							Connections_Disconnect(connectionDataPtr);
						}
						else 
						{
							// connection is still open; disconnect first and (if the user
							// has the preference set) close the window automatically;
							// a confirmation message is displayed UNLESS it has been less
							// than a few seconds since the connection opened, in which
							// case MacTelnet assumes the user doesnt care and bypasses
							// the confirmation message (what can I say, Im a nice guy)
							UInt32		timeInSeconds = 0L;
							Boolean		doClose = false;
							
							
							GetDateTime(&timeInSeconds);
							if ((timeInSeconds - connectionDataPtr->connectedSince) >
								kConnectionTimeMaximumForNoWarningClose)
							{
								ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
								InterfaceLibAlertRef	box = null;
								Str31					titleText;
								Str255					dialogText,
														helpText,
														buttonText;
								
								
								GetIndString(dialogText, rStringsCautionAlerts, siConfirmConnectionClose);
								{
									Str255		windowTitle;
									
									
									GetWTitle(theWindow, windowTitle);
									StringUtilities_PMetaSetDefaults();
									StringUtilities_PBuild(dialogText, windowTitle, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
								}
								GetIndString(titleText, rStringsAlertTitles, siAlertTitleClose);
								GetIndString(helpText, rStringsCautionAlerts, siConfirmConnectionCloseHelpText);
								
								box = Alert_New();
								Alert_SetFilterProc(box, upp);
								Alert_SetHelpButton(box, true);
								Alert_SetParamsFor(box, kAlert_StyleOKCancel);
								GetIndString(buttonText, rStringsSpecialButtons, siButtonClose);
								Alert_SetTitle(box, titleText);
								Alert_SetButtonText(box, kAlertStdAlertOKButton, buttonText);
								Alert_SetText(box, dialogText, helpText);
								Alert_SetType(box, kAlertCautionAlert);
								Alert_Display(box);
								if (Alert_ItemHit(box) == kAlertStdAlertOKButton)
								{
									doClose = true;
								}
								else if (Alert_ItemHit(box) == kAlertStdAlertHelpButton)
								{
									Str255		searchString;
									
									
									GetIndString(searchString, rStringsAppleGuideLookup, siAGIndexTermConnections);
									AppleGuideUtilities_OpenFirstMainGuideInSearch(searchString);
								}
								Alert_Dispose(&box);
								DisposeModalFilterUPP(upp), upp = null;
							}
							else doClose = true; // if the connection JUST opened, kill it without confirmation
							
							if (doClose)
							{
								NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventClose,
												connectionDataPtr->port, 0);
								NetEvents_Service();
							}
						}
					}
					else BEEP;
				}
				break;
			
			case WIN_ICRG:
				MacRGdestroy(MacRGfindwind(theWindow));
				MaxMem(&junk);
				break;
				
			case WIN_TEK:
				destroyGraphics(RGgetVG(theWindow));
				MaxMem(&junk);
				break;
			
			default:
				break;
		}
	}
}// CloseAWindow


/*!
This method is called when it is necessary
to post a notification for the user.  If
this program is not in the background, this
routine does nothing.

(2.6)
*/
void
NotifyUser	(void)
{
	OSErr		error = noErr;
	Boolean		notifyOfBeeps = false;
	Size		actualSize = 0L;
	
	
	// get the users background bell-notification preference, if possible
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_NotifyOfBeeps,
								sizeof(notifyOfBeeps), &notifyOfBeeps,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		notifyOfBeeps = false; // assume beeps dont generate notification, if preference cant be found
	}
	
	if (FlagManager_Test(kFlagSuspended) && notifyOfBeeps && (!gHaveInstalledNotification))
	{
		gBeepNotificationPtr = (NMRecPtr)Memory_NewPtr(sizeof(NMRec));
		if (gBeepNotificationPtr != null)
		{
			gBeepNotificationPtr->qType = nmType;
			gBeepNotificationPtr->nmMark = 1;
			error = GetIconSuite(&(gBeepNotificationPtr->nmIcon), kNCSAIconFamilyId, svAllSmallData);
			if (error != noErr) gBeepNotificationPtr->nmIcon = null;
			gBeepNotificationPtr->nmSound = null;
			gBeepNotificationPtr->nmStr = null;
			gBeepNotificationPtr->nmResp = null;
			NMInstall(gBeepNotificationPtr);
			gHaveInstalledNotification = true;
		}
	}
}// NotifyUser


//
// internal methods
//

/*!
Updates global (internal) variables used
to track changes to the keyboard script.

(2.6)
*/
void
checkDefaultScriptSettings	(void)
{
	SInt32			count = 0L;
	
	
	count = GetScriptManagerVariable(smMunged);
	if (count > gMungeCount)
	{
		gCurrentScript = GetScriptManagerVariable(smKeyScript); // get the new active script
		if ((gCurrentScript != smRoman) && gKCHRHasChanged) // Roman has been switched out
		{	
			SetScriptVariable(smRoman, smScriptKeys, gDefaultKCHR);
			gCurrentScript = GetScriptManagerVariable(smKeyScript); // get the new active script
			gKCHRHasChanged = false; // dont mess with non-Roman scripts	
		}
		gMungeCount = count;
	}
}// checkDefaultScriptSettings


/*!
Marc Tamskys code for responding to events on
the Home, End, Page Up or Page Down keys.  In
MacTelnet 3.0, this function has been forced to
use whatever window thats given.  Returns true
if the event was handled, false otherwise.

(2.6)
*/
static Boolean
checkPageKeys	(WindowRef const		inWindow,
				 short const			inCode)
{
	Boolean		result = false;
	
	
	if (ScreenManager_IsWindowAScreen(inWindow))
	{
		TelnetScreenID		virtualScreenID = 0;
		short     			x1 = 0,
							y2 = 0,
							x2 = 0,
							y1 = 0; // coordinates from window
		
		
		virtualScreenID = ScreenManager_GetWindowScreenID(inWindow);
		
		switch (inCode)
		{
			case VSPGUP:
				RealScreen_SetCursorVisibility(virtualScreenID, false/* visible */);
				VSgetrgn(virtualScreenID, &x1, &y1, &x2, &y2);
				Terminal_ScrollRowsTowardBottomEdge(virtualScreenID, y2 - y1);
				result = true;
				break;
			
			case VSPGDN:		                                                             	 
				RealScreen_SetCursorVisibility(virtualScreenID, false/* visible */);
				VSgetrgn(virtualScreenID, &x1, &y1, &x2, &y2);
				Terminal_ScrollRowsTowardTopEdge(virtualScreenID, y2 - y1);
				result = true;
				break;
			
			case VSHOME:
				RealScreen_SetCursorVisibility(virtualScreenID, false/* visible */);
				Terminal_ScrollToBeginning(virtualScreenID);
				result = true;
				break;
			
			case VSEND:
				RealScreen_SetCursorVisibility(virtualScreenID, false/* visible */);
				Terminal_ScrollToEnd(virtualScreenID);
				result = true;
				break;
			
			default:
				// event not handled
				break;
		}
	}
	return result;
}// checkPageKeys


/*!
To subtract out the regions of visible floating
windows from the given global region, use this
method.  Note that since this effect is based
only on present state information, you generally
should not save the resultant region for very
long.  For example, if the user shows a floating
window soon after you call this routine, you may
need to call diffFloatingRegions() again to make
sure your region is up to date.

(3.0)
*/
static void
diffFloatingRegions		(RgnHandle	inoutRegion)
{
	if (inoutRegion != null)
	{
		Handle		windoidsHandle = Memory_NewHandleInProperZone(kFloaterIndexCount * sizeof(WindowRef),
																	true/* only briefly allocated */),
					visibilitiesHandle = Memory_NewHandleInProperZone(kFloaterIndexCount * sizeof(Boolean),
																	true/* only briefly allocated */);
		
		
		if ((windoidsHandle != null) && (visibilitiesHandle != null))
		{
			WindowRef*		windoids = null;
			Boolean*		visibilities = null;
			
			
			HLock(windoidsHandle), windoids = (WindowRef*)(*windoidsHandle);
			HLock(visibilitiesHandle), visibilities = (Boolean*)(*visibilitiesHandle);
			
			if (getFloatingWindowInfo(windoids, visibilities, kFloaterIndexCount) == noErr)
			{
				RgnHandle		structureRegion = Memory_NewRegion();
				
				
				if (structureRegion != null)
				{
					register SInt16		i = 0;
					
					
					for (i = 0; i < kFloaterIndexCount; i++)
					{
						// cut out the regions of all visible floating windows
						if (visibilities[i]) if (GetWindowRegion(windoids[i], kWindowStructureRgn,
																	structureRegion) == noErr)
						{
							DiffRgn(inoutRegion, structureRegion, inoutRegion);
						}
					}
					Memory_DisposeRegion(&structureRegion);
				}
			}
		}
		Memory_DisposeHandle(&windoidsHandle);
		Memory_DisposeHandle(&visibilitiesHandle);
	}
}// diffFloatingRegions


/*!
To modify the global mouse region (for determining
when the cursor shape should change) to be an
empty region nowhere, use this method.  This
means everywhere else, excluding important
mouse regions.

(3.0)
*/
static void
emptyGlobalMouseRegion	(void)
{
	if (gMouseRegion != null)
	{
		SetEmptyRgn(gMouseRegion);
		gMouseRegionCursor = kMouseRegionTypeNowhere;
	}
}// emptyGlobalMouseRegion


/*!
This method makes sure that the floating windows are
not obscured by anything but a dialog box or another
floating window.

(3.0)
*/
static void
ensureFloatingWindowsFloat	(void)
{
	WindowRef*		windoids = (WindowRef*)Memory_NewPtr(kFloaterIndexCount * sizeof(WindowRef));
	Boolean*		visibilities = (Boolean*)Memory_NewPtr(kFloaterIndexCount * sizeof(Boolean));
	
	
	if (getFloatingWindowInfo(windoids, visibilities, kFloaterIndexCount) == noErr)
	{
		// Make sure every floating window stays in its
		// proper front-to-back order, and prevent any
		// floating window from moving in front of a
		// dialog box.
		{
			WindowRef			frontWindow = GetRealFrontWindow(),	// frontmost non-floating window
								nextVisibleWindoid = null;			// this is used to order windows
			register SInt16		i = 0;
			
			
			if (GetWindowKind(frontWindow) != kDialogWindowKind)
			{
				for (i = kFloaterIndexCount - 1; i >= 0; i--)
				{
					if (visibilities[i])
					{
						BringToFront(windoids[i]);
						if (!IsWindowHilited(windoids[i])) HiliteWindow(windoids[i], true);
						ClipAbove(frontWindow);
						if (i > 0) if (visibilities[i - 1]) nextVisibleWindoid = windoids[i - 1];
					}
				}
			}
		}
	}
	Memory_DisposePtr((Ptr*)&windoids);
	Memory_DisposePtr((Ptr*)&visibilities);
}// ensureFloatingWindowsFloat


/*!
To obtain window pointers and, for convenience,
visibility information, for all of Telnets own
floating windows, use this method.  If any
problems occur (such as null pointers), a
result code is returned; otherwise, "noErr" is
returned.

The value of "inArrayCardinality" must be the
number of elements in the two arrays YOU give
(which must exist and must have the same size).
This allows code using getFloatingWindowInfo()
to antiquate safely, in the event that more
information for more floating windows is made
available in the future.

(3.0)
*/
static OSErr
getFloatingWindowInfo	(WindowRef*		inoutWindowArray,
						 Boolean*		inoutVisibilityArray,
						 SInt16 const	UNUSED_ARGUMENT(inArrayCardinality))
{
	OSErr		result = noErr;
	
	
	if ((inoutWindowArray != null) &&
		(inoutVisibilityArray != null))
	{
		WindowRef*		windoids = inoutWindowArray;
		Boolean*		visibilities = inoutVisibilityArray;
		
		
		// initialize for all windoids; ideally, the given
		// array cardinality should be used as a safety net,
		// if you expand the list of floating windows in the
		// future (Telnet currently ignores the value!!!)
		windoids[kFloaterIndexInspector] = Inspector_GetWindow();
		windoids[kFloaterIndexCommandLine] = CommandLine_GetWindow();
		windoids[kFloaterIndexControlKeys] = Keypads_GetWindow(kKeypadTypeControlKeys);
		windoids[kFloaterIndexFunction] = Keypads_GetWindow(kKeypadTypeFunctionKeys);
		windoids[kFloaterIndexKeypad] = Keypads_GetWindow(kKeypadTypeVT220Keys);
		visibilities[kFloaterIndexInspector] = Inspector_IsVisible();
		visibilities[kFloaterIndexCommandLine] = CommandLine_IsVisible();
		visibilities[kFloaterIndexControlKeys] = Keypads_IsVisible(kKeypadTypeControlKeys);
		visibilities[kFloaterIndexFunction] = Keypads_IsVisible(kKeypadTypeFunctionKeys);
		visibilities[kFloaterIndexKeypad] = Keypads_IsVisible(kKeypadTypeVT220Keys);
	}
	else result = memPCErr;
	
	return result;
}// getFloatingWindowInfo


/*!
To obtain a reference to the region that defines
the boundaries justifying the current cursor
shape, use this method.  If the mouse moves
outside this region, the cursor shape should be
re-evaluated and this region should be modified.

(3.0)
*/
static RgnHandle
getMouseRegion		(void)
{
	return gMouseRegion;
}// getMouseRegion


/*!
Handles events issued when windows are activated
or deactivated.  Returns true if the event was
completely handled, false otherwise.

(3.0)
*/
static Boolean
handleActivate		(EventRecord*		inoutEventPtr,
					 Boolean const		inActivating)
{
	WindowRef					window = null;
	WindowInfoRef				windowInfoRef = null;
	WindowInfoDescriptor		windowDescriptor = kInvalidWindowInfoDescriptor;
	Boolean						result = false;
	
	
	window = (WindowRef)inoutEventPtr->message;
	
	if (inActivating)
	{
		// assume that if the user activates a collapsed window, that he or she wants it uncollapsed
		if (IsWindowCollapsed(window)) (OSStatus)CollapseWindow(window, false);
	}
	
	windowInfoRef = WindowInfo_GetFromWindow(window);
	if (windowInfoRef != null)
	{
		windowDescriptor = WindowInfo_GetWindowDescriptor(windowInfoRef);
		switch (windowDescriptor)
		{
			case kTelnetWindowDescriptorAnyTerminal:
				// terminal window
				//RealScreen_SetActive(window, inActivating);
				break;
			
			case kTelnetWindowDescriptorInspectorWindoid:
			case kTelnetWindowDescriptorControlWindoid:
			case kTelnetWindowDescriptorFunctionWindoid:
			case kTelnetWindowDescriptorKeypadWindoid:
				// there is no concept of activate events for most floating windows;
				// all of their states change at once
				result = true;
				break;
			
			case kTelnetWindowDescriptorCommandLine:
				// the command line uses activate events because it can receive keyboard focus
				CommandLine_SetActive(true);
				result = true;
				break;
			
			default:
				// not a floating window
				break;
		}
	}
	
	unless (result)
	{
		// Appearance windows with root controls can be handled easily
		(OSErr)Embedding_OffscreenSetRootControlActive(window, inActivating);
		result = true;
	}
	
	return result;
}// handleActivate


/*!
Responds to disk insertion events.  May
display the standard OS dialog asking
the user what to do with the disk, if
it cant be mounted.  Returns true if
the event was completely handled, false
otherwise.

(3.0)
*/
static Boolean
handleDiskEvent	(EventRecord const*		inEventPtr)
{
	Boolean		result = false;
#if TARGET_API_MAC_OS8
	Point		dialogUpperLeftHandCorner;
	
	
	SetPt(&dialogUpperLeftHandCorner, 100, 120);
	if (noErr != ((inEventPtr->message >> 16) & 0xFFFF))
	{
		DILoad();
		DIBadMount(dialogUpperLeftHandCorner, inEventPtr->message);
		DIUnload();
		result = true;
	}
#else
	// Carbon handles disk initialization itself (about freakin time the Mac OS could do this itself!)
#endif
	
	return result;
}// handleDiskEvent


/*!
Responds to key presses in any window.

(3.0)
*/
static Boolean
handleKeyDown	(EventRecord*		inoutEventPtr)
{
	Boolean		result = false;
	
	
	// possibly a menu key event
	Menus_SetUpItemStates();
	{
		long	menuEquiv = MenuEvent(inoutEventPtr);
		short	theMenu = 0,
				theItem = 0;
		
		
		theMenu = menuEquiv >> 16;
		theItem = menuEquiv & 0xFFFF;
		if (theMenu != 0)
		{
			MenuHandle		handle = GetMenu(theMenu);
			UInt32			commandID = 0;
			
			
			// shouldnt use this
			result = Menus_HandleCommand(menuEquiv, inoutEventPtr->modifiers);
			
			// use instead:
			//GetMenuItemCommandID(handle, theItem, &commandID);
			//result = Menus_HandleCommandByID(commandID, inoutEventPtr->modifiers);
		}
	}
	
	if (!result)
	{
		WindowRef				whichWindowReal = GetRealFrontWindow(),
								whichWindowAbsolute = FrontWindow();
		WindowInfoRef			windowInfoRefReal = null,
								windowInfoRefAbsolute = null;
		WindowInfoDescriptor	windowDescriptorReal = kInvalidWindowInfoDescriptor,
								windowDescriptorAbsolute = kInvalidWindowInfoDescriptor;
		unsigned char			keyChar = inoutEventPtr->message & charCodeMask,
								keyCode = ((inoutEventPtr->message & keyCodeMask) >> 8);
		
		
		windowInfoRefReal = WindowInfo_GetFromWindow(whichWindowReal);
		if (windowInfoRefReal != null)
		{
			windowDescriptorReal = WindowInfo_GetWindowDescriptor(windowInfoRefReal);
		}
		windowInfoRefAbsolute = WindowInfo_GetFromWindow(whichWindowAbsolute);
		if (windowInfoRefAbsolute != null)
		{
			windowDescriptorAbsolute = WindowInfo_GetWindowDescriptor(windowInfoRefAbsolute);
		}
		
		// typing into a terminal window will have an effect even if the window
		// is collapsed; to ensure that the user can see this, force a collapsed,
		// *frontmost* window to uncollapse as a key is pressed
		if (IsWindowCollapsed(whichWindowReal))
		{
			(OSStatus)CollapseWindow(whichWindowReal, false);
		}
		
		// handle floating windows first
		if (windowDescriptorAbsolute != kInvalidWindowInfoDescriptor) switch (windowDescriptorAbsolute)
		{
			case kTelnetWindowDescriptorCommandLine:
				result = CommandLine_HandleKeyDown(inoutEventPtr);
				break;
			
			default:
				break;
		}
		
		// then, if none of them handled the event, pass it to a real window
		if (!result) if (windowDescriptorReal != kInvalidWindowInfoDescriptor) switch (windowDescriptorReal)
		{
			case kTelnetDialogDescriptorPreferences:
				result = PreferencesDialog_HandleKeyDown(inoutEventPtr);
				break;
			
			default:
				break;
		}
		
		// if its still not handled, assume its a connection window event
		if (!result) if (ScreenManager_IsWindowAScreen(whichWindowReal))
		{
			handleScreenKeyDown(inoutEventPtr, ScreenManager_GetWindowConnectionData(whichWindowReal));
		}
	}
	
	return result;
}// handleKeyDown


/*!
Handles mouse-down events in windows.

(2.6)
*/
void
handleMouseDown		(EventRecord*	inoutEventPtr)
{
	WindowRef					whichWindow = null;
	GrafPtr						whichPort = null;
	WindowInfoRef				windowInfoRef = null;
	WindowInfoDescriptor		windowDescriptor = kInvalidWindowInfoDescriptor;
	short						windowPart = 0,
								myRGMnum = 0;
	short 						i = 0;
	long						mResult = 0;
	UInt32						commandID = 0;
	short						theMenu = 0,
								theItem = 0;
	MenuHandle					handle = null;
	
	
	windowPart = FindWindow(inoutEventPtr->where, &whichWindow);
	whichPort = (GrafPtr)GetWindowPort(whichWindow);
	windowInfoRef = WindowInfo_GetFromWindow(whichWindow);
	if (windowInfoRef != null)
	{
		windowDescriptor = WindowInfo_GetWindowDescriptor(windowInfoRef);
	}
	
	if (windowPart != inNoWindow)
	{
		if (IS_SHOW_CONTEXTUAL_MENU_EVENT(inoutEventPtr))
		{
			Boolean		canContextualMenuClick = true;
			
			
			if (windowInfoRef != null)
			{
				canContextualMenuClick = (!WindowInfo_IsWindowFloating(windowInfoRef));
			}
			if (canContextualMenuClick)
			{
				if (whichWindow != GetRealFrontWindow())
				{
					SelectOverRealFrontWindow(whichWindow); // Human Interface Guidelines
				}
				(OSStatus)ContextualMenuBuilder_HandleWindow(whichWindow, inoutEventPtr, windowPart);
			}
		}
		else
		{
			switch (windowPart)
			{
				case inMenuBar:
					{
						long		menuResult = 0L;
						Boolean		isHandledByTSM = false;
						
						
						// change certain menu items as needed based on modifier key states
						// (or change them back if a particular key is not down)
						Menus_SetUpItemStates();
						
						// track menus and respond to selections
						menuResult = MenuSelect(inoutEventPtr->where);
						if (FlagManager_Test(kFlagTextServicesManager))
						{
						#if TARGET_API_MAC_OS8
							isHandledByTSM = TSMMenuSelect(menuResult);
						#else
							// Carbon handles these events implicitly, so ignore them
							isHandledByTSM = false;
						#endif
						}
						if (!isHandledByTSM)
						{
							if ((menuResult & 0xFFFF0000) != 0)
							{
								(Boolean)Menus_HandleCommand(menuResult, inoutEventPtr->modifiers);
							}
						}
					}
					break;
				
				case inSysWindow:
					// obsolete check
					//SystemClick(inoutEventPtr, whichWindow); // CARBON does not support desk accessories
					break;
				
				case inGoAway:
					if (TrackGoAway(whichWindow, inoutEventPtr->where))
					{
						SInt16		junk = 0;
						
						
						if ((WindowInfo_IsWindowFloating(windowInfoRef)) ||
							(TektronixRealGraphics_IsRealGraphicsWindow(whichWindow, &junk)))
						{
							CloseAWindow(whichWindow);
						}
						else
						{
							(Boolean)Commands_ExecuteByID(kCommandCloseConnection,
															inoutEventPtr->modifiers);
						}
					}
					break;
				
				case inZoomIn:
				case inZoomOut:
					if (TrackBox(whichWindow, inoutEventPtr->where, windowPart))
					{
						Boolean		isValidWindow = true;
						
						
						if (windowDescriptor != kInvalidWindowInfoDescriptor)
						{
							switch (windowDescriptor)
							{
								case kTelnetWindowDescriptorClipboard:
									Clipboard_Zoom(windowPart, inoutEventPtr->modifiers);
									break;
								
								case kTelnetWindowDescriptorCommandLine:
									CommandLine_Zoom(windowPart, inoutEventPtr->modifiers);
									break;
								
								default:
									if (ScreenManager_IsWindowAScreen(whichWindow))
									{
										if (inoutEventPtr->modifiers & optionKey)
										{
											// shortcut for displaying the Set Size window
											SizeDialog_Init(ScreenManager_GetWindowScreenID(whichWindow));
											SizeDialog_Display();
											SizeDialog_Done();
										}
										else
										{
											// zoom a session window or the FTP activity log
											RealScreen_ZoomWindow(whichWindow, windowPart, true/* activate */);
										}
									}
									else isValidWindow = false;
									break;
							}
						}
						
						unless (isValidWindow) BEEP; // unknown window???
					}
					break;
				
				case inGrow:
					/*if (windowDescriptor == kTelnetWindowDescriptorAnyTerminal)
					{
						// temporary - the Window Info version doesnt work yet
						RealScreen_SizeWindow(whichWindow, inoutEventPtr->where, inoutEventPtr->modifiers);
					}
					else */if (windowDescriptor != kInvalidWindowInfoDescriptor)
					{
						// with Window Info, ANY supporting window can be resized abstractly!
						WindowInfo_GrowWindow(whichWindow, inoutEventPtr);
					}
					else
					{
						// is it a TEK window click?
						myRGMnum = RGfindbywind(whichWindow);
						if (myRGMnum != -1)
						{
							RGMgrowme((short)myRGMnum, whichWindow, (long*)&inoutEventPtr->where, inoutEventPtr->modifiers);
						}
					}
					break;
				
				case inContent:
				case inDrag:
					{
						Boolean		frontClick = false;
						
						
						// make sure clicks in floating windows are serviced
						if (windowInfoRef != null)
						{
							frontClick = WindowInfo_IsWindowFloating(windowInfoRef);
						}
						if (!frontClick) frontClick = (whichWindow == GetRealFrontWindow());
						
						// Background windows can be dragged without activating
						// them first if the command key is being held down.  They
						// can be activated by any content click as long as the
						// window to activate is not a floating window (it is
						// undesirable to have a document window unhighlight its
						// frame when the user clicks in a floating window).
						if ((!frontClick) &&
							((!(inoutEventPtr->modifiers & cmdKey)) || (windowPart == inContent)))
						{
							CommandLine_LoseFocus(); // de-focus windoids
							if (inoutEventPtr->modifiers & optionKey)
							{
								// option-clicking one window hides the other
								(Boolean)Commands_ExecuteByID(kCommandHideFrontWindow, 0/* modifiers */);
							}
							SelectOverRealFrontWindow(whichWindow);
						}
						else if (windowPart == inContent)
						{
							DialogItemIndex				dialogItemIndex = 0;
							Boolean						isConnectionWindow = false;
							
							
							if (windowDescriptor != kInvalidWindowInfoDescriptor)
							{
								switch (windowDescriptor)
								{
									case kTelnetDialogDescriptorPreferences:
										CommandLine_LoseFocus(); // de-focus windoids
										PreferencesDialog_HandleContentClick(inoutEventPtr);
										break;
									
									case kTelnetWindowDescriptorInspectorWindoid:
										Inspector_HandleContentClick(inoutEventPtr);
										break;
									
									case kTelnetWindowDescriptorCommandLine:
										CommandLine_HandleContentClick(inoutEventPtr);
										break;
									
									case kTelnetWindowDescriptorControlWindoid:
										Keypads_HandleContentClick(kKeypadTypeControlKeys, inoutEventPtr);
										break;
									
									case kTelnetWindowDescriptorFunctionWindoid:
										Keypads_HandleContentClick(kKeypadTypeFunctionKeys, inoutEventPtr);
										break;
									
									case kTelnetWindowDescriptorKeypadWindoid:
										Keypads_HandleContentClick(kKeypadTypeVT220Keys, inoutEventPtr);
										break;
									
									default:
										// where did this window come from?  Assume its a connection
										isConnectionWindow = true;
										break;
								}
							}
							else isConnectionWindow = true;
							
							if (isConnectionWindow)
							{
								CommandLine_LoseFocus(); // de-focus windoids
								if (RealScreen_HandleContentClick(whichWindow, inoutEventPtr) < 0)
								{
									GrafPtr		oldPort = null;
									
									
									// then its a graphics window
									GetPort(&oldPort);
									SetPort(whichPort);
									GlobalToLocal(&inoutEventPtr->where);
									RGmousedown(whichWindow, &inoutEventPtr->where);
									SetPort(oldPort);
								}
							}
						}
						
						// At this point, allow the user to move the window if
						// the frame of the frontmost window was clicked, or if
						// the frame of a background window was command-clicked.
						// If a screen window contains a background picture, be
						// sure to maximize its rendering speed by aligning the
						// window location on the screen, using QuickTime.
						if (windowPart == inDrag)
						{
							enum
							{
								kDragSlopInsets = 4		// pixels from gray region edges
							};
							Rect		dragRect,
										viewRect,	// only used for window alignment
										screenRect;
							Boolean		alignWindow = false;
							
							
							MacLook_GetWindowDeviceGrayRect(whichWindow, &screenRect);
							SetRect(&dragRect, kDragSlopInsets, GetMBarHeight() + kDragSlopInsets,
									screenRect.right - kDragSlopInsets, screenRect.bottom - kDragSlopInsets);
							
							// align the window, if possible, to its image (only
							// important if the window has a background picture)
							if (FlagManager_Test(kFlagQuickTime))
							{
								if (ScreenManager_IsWindowAScreen(whichWindow))
								{
									TelnetScreenID		screenID = ScreenManager_GetWindowScreenID(whichWindow);
									RealScreenDataPtr	realScreenPtr = null;
									
									
									realScreenPtr = ScreenManager_GetRealScreenDataOf(screenID);
									if (realScreenPtr->customTextBackground)
									{
										GetControlBounds(realScreenPtr->contentUserPane, &viewRect);
										alignWindow = true;
									}
								}
								else
								{
									MacLook_GetWindowRegionBounds(whichWindow, kWindowContentRgn, &viewRect);
									alignWindow = true;
								}
							}
							
							if (alignWindow) DragAlignedWindow(whichWindow, inoutEventPtr->where,
																&dragRect, &viewRect, null);
							else DragWindow(whichWindow, inoutEventPtr->where, &dragRect);
							
							// make the drag recordable by sending events back to MacTelnet
							// (however, dont execute the events, because their effects have
							// already taken place, thanks to DragWindow())
							{
								OSErr		error = noErr;
								AEDesc		setDataEvent,
											replyDescriptor;
								Point		newPosition;
								
								
								MacLook_GetWindowRegionBounds(whichWindow, kWindowContentRgn, &viewRect);
								SetPt(&newPosition, viewRect.left, viewRect.top);
								InstallAE_CreateRecordableAppleEvent(kAECoreSuite, kAESetData, &setDataEvent);
								if (error == noErr)
								{
									AEDesc			containerDesc = { typeNull, null },
													keyDesc = { typeNull, null },
													objectDesc = { typeNull, null },
													propertyDesc = { typeNull, null },
													valueDesc = { typeNull, null };
									
									
									// send a request for "window <index>", which resides in a null container;
									// then, request the "position" property, and finally, issue an event
									// that will change its value
									(OSErr)BasicTypesAE_CreateSInt32Desc(1/* temp - inaccurate! */, &keyDesc);
									error = CreateObjSpecifier(kTelnetClassWindow, &containerDesc,
																formAbsolutePosition, &keyDesc, true/* dispose inputs */,
																&objectDesc);
									if (error != noErr) Console_WriteValue("window access error", error);
									(OSErr)BasicTypesAE_CreatePropertyIDDesc(kTelnetPropertyWindowPosition, &keyDesc);
									error = CreateObjSpecifier(cProperty, &objectDesc,
																formPropertyID, &keyDesc, true/* dispose inputs */,
																&propertyDesc);
									if (error != noErr) Console_WriteValue("property access error", error);
									if (error == noErr)
									{
										// reference to the property to be set - REQUIRED
										(OSErr)AEPutParamDesc(&setDataEvent, keyDirectObject, &propertyDesc);
										
										// the propertys new value - REQUIRED
										(OSErr)BasicTypesAE_CreatePointDesc(newPosition, &valueDesc);
										(OSErr)AEPutParamDesc(&setDataEvent, kTelnetPartialCoreParameterKeySetDataTo, &valueDesc);
										
										// send the event, which will record it into any open script;
										// this event is not executed because DragWindow() already
										// performs its effects; it is simply for convenience to the
										// user, who will be able to achieve the same effect later by
										// running his or her script
										(OSErr)AESend(&setDataEvent, &replyDescriptor,
														kAENoReply | kAECanInteract | kAEDontExecute,
														kAENormalPriority, kAEDefaultTimeout, null, null);
									}
									AEDisposeDesc(&propertyDesc);
									AEDisposeDesc(&valueDesc);
								}
								AEDisposeDesc(&setDataEvent);
							}
						}
					}
					break;
				
				default:
					break;
			}
		}
	}
}// handleMouseDown


/*!
This method handles key events in connection windows.

Note to self: re-write this god-awful function, which
was apparently written by drunk chimpanzies.

Note to inexperienced programmers: writing a function
that is this long is BAD; writing a function with this
many "return" and "goto" statements is BAD; writing a
function with sparse comments is BAD.  Guess what -
this function has all three of these, which means the
execution paths are completely confusing.

(2.6)
*/
static void
handleScreenKeyDown	(EventRecord*		inoutEventPtr,
					 ConnectionDataPtr	inConnectionDataPtr)
{
	if ((inoutEventPtr != null) && (inConnectionDataPtr != null))
	{
		// IMPORTANT:	Keypresses occur so frequently that these memory
		//				allocations will occur frequently as well.  Making
		//				them static restricts their scope, but also keeps
		//				MacTelnet from potentially shuffling stack space
		//				while the user tries to type.
		static unsigned char	ascii = '\0',
								code = '\0';
		static unsigned char	characterToSend = '\0';
		static Boolean			commandDown = false,
								optionDown  = false,
								controlDown = false,
								shiftDown = false;
		static Boolean			remapBackquoteToEscape = false;
		static Size				actualSize = 0L;
		
		
		// get the users process service preference, if possible
		unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
									kTelnetPreferenceTag_MapBackquote,
									sizeof(remapBackquoteToEscape), &remapBackquoteToEscape,
									&actualSize) == kTelnetPrefsResultCodeSuccess)
		{
			remapBackquoteToEscape = false; // assume default if preference cant be found
		}
		
		// 3.0 - Easter egg!
		if (FlagManager_Test(kFlagQuadDamage)) CoolStuff_QuadDamageActivity();
		
		ObscureCursor();
		
		ascii = inoutEventPtr->message & charCodeMask;
		code = ((inoutEventPtr->message & keyCodeMask) >> 8);
		commandDown = ((inoutEventPtr->modifiers & cmdKey) != 0);
		optionDown = ((inoutEventPtr->modifiers & optionKey) != 0);
		controlDown = ((inoutEventPtr->modifiers & controlKey) != 0);
		shiftDown = ((inoutEventPtr->modifiers & shiftKey) != 0);
		
		if ((inConnectionDataPtr->metaKey == kEMACSMetaKeyOption) &&
			(optionDown)) goto emacsHack;  // ha ha hack hack
		
		if ((code == 0x34) && (ascii == 3)) //fix for PowerBook 540  bad KCHR
			ascii = 13; 				//(map control-c to return)
		else if ((controlDown)&&(shiftDown)&&(ascii == '2'))
			ascii = 0;//fix bad KCHR control-@
		else if ((controlDown)&&(shiftDown)&&(ascii == '6'))
			ascii = 0x1E;//fix bad KCHR control-^
		
		if (commandDown)  		
		{ 
			Boolean		menuKeyEquivalents = true;
			Size		actualSize = 0L;
			
			
			// get the users process service preference, if possible
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
										kTelnetPreferenceTag_MenuItemKeys,
										sizeof(menuKeyEquivalents), &menuKeyEquivalents,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				menuKeyEquivalents = true; // assume menus have key equivalents, if preference cant be found
			}
			
			
			if (menuKeyEquivalents)
			{
				// if optioned, retranslate so we can do menu commands
				if (optionDown)
				{
					short			virtualCode = 0;
					Ptr				KCHRPtr = null;
					long			newStuff = 0L;
					unsigned long	state = 0L;
					short			modifiersCopy = 0;
					
					
					modifiersCopy = inoutEventPtr->modifiers;
					virtualCode = (short)(code);
					
					inoutEventPtr->modifiers &= (shiftKey); //turn off option
					virtualCode |= inoutEventPtr->modifiers;  
					KCHRPtr  = (Ptr)GetScriptManagerVariable(smKCHRCache);
					newStuff = KeyTranslate(KCHRPtr,virtualCode,&state);
					newStuff &= 0xFF; //only look at bottom byte
					ascii = (unsigned char) newStuff;
					inoutEventPtr->modifiers = modifiersCopy; //reset option state
				}
				if ((Connections_GetCount() < 1) || (inConnectionDataPtr->status != kConnectionStatusActive)) 
					return;
				// check for EMACS meta key
				if ((inConnectionDataPtr->metaKey == kEMACSMetaKeyControlCommand) && (controlDown))
				{		
					unsigned char temp[2];
					if (ascii <= 32) //control changed the ascii value
						ascii |= 0x40; //move back to a non-control
					if ((shiftDown)||(ascii == 0x5F)) //so we can get meta -
					{
						short			virtualCode = 0;
						Ptr				KCHRPtr;
						long			newStuff = 0L;
						unsigned long	state = 0L;
						
						
						virtualCode = (short)(code);
						inoutEventPtr->modifiers &= (shiftKey); //turn of command
						virtualCode |= inoutEventPtr->modifiers;  
						KCHRPtr  = (Ptr)GetScriptManagerVariable(smKCHRCache);
						newStuff = KeyTranslate(KCHRPtr,virtualCode,&state);
						newStuff &= 0xFF; //only look at bottom byte
						ascii = (unsigned char) newStuff;
					}
		emacsHack: // if the option key KCHR is installed, we will get the right ascii value
					if ((inConnectionDataPtr->clientflags & PASTE_IN_PROGRESS) &&
						(inConnectionDataPtr->paste.method == kTelnetPasteMethodBlock)) // queue this
					{
						inConnectionDataPtr->kbbuf[inConnectionDataPtr->kblen++] = ESC;
						inConnectionDataPtr->kbbuf[inConnectionDataPtr->kblen++] = ascii;
						return;
					}
					if (inConnectionDataPtr->kblen > 0)
					{
						Network_SendData( inConnectionDataPtr->port,inConnectionDataPtr->kbbuf,inConnectionDataPtr->kblen);										
						inConnectionDataPtr->kblen=0;												
					}
						
					temp[0] = ESC;
					temp[1] = ascii;
					if (inConnectionDataPtr->echo && inConnectionDataPtr->halfdup)
					{
						Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, temp, 2);
					}
					Network_SendFlushData(inConnectionDataPtr->port,temp,2);	
					return;
				}
				else if (ascii >= '0' && ascii <= '9')  // now look for macros
				{
					if (Macros_GetMode() == kTelnetMacroModeCommandDigit)
					{
						Macros_Send(inConnectionDataPtr, ascii - '0');
						return;
					}
				}
				else if (ascii >= '0' && ascii <= '9')  // now look for macros
				{
					if (Macros_GetMode() == kTelnetMacroModeCommandDigit)
					{
						Macros_Send(inConnectionDataPtr, ascii - '0');
						return;
					}
				}
				else if (ascii == '=')  // look for special macros
				{
					if (Macros_GetMode() == kTelnetMacroModeCommandDigit)
					{
						Macros_Send(inConnectionDataPtr, 10/* zero-based macro index */);
						return;
					}
				}
				else if (ascii == '/')  // look for special macros
				{
					if (Macros_GetMode() == kTelnetMacroModeCommandDigit)
					{
						Macros_Send(inConnectionDataPtr, 11/* zero-based macro index */);
						return;
					}
				}
				else if (!((ascii == '`' && remapBackquoteToEscape) || (code == BScode)))
					return;
			}
			else // no command key menus
			{
				return;
			}
		} else if ((Connections_GetCount() < 1) || (inConnectionDataPtr->status != kConnectionStatusActive)) 
			return;
		
		// check for Print Screen (F13) and handle this command
		if (code == 0x69)
		{
			Commands_ExecuteByID(kCommandPrintScreen, inoutEventPtr->modifiers);
			return;
		}
		
		// look for F-key macros (see above code for command-digit-based ones)
		if (Macros_GetMode() == kTelnetMacroModeFunctionKeys)
		{
			short		zeroBasedMacroNumber = -1;
			
			
			switch (code)
			{
				case 0x7A: // F1
					zeroBasedMacroNumber = 0;
					break;
				
				case 0x78: // F2
					zeroBasedMacroNumber = 1;
					break;
				
				case 0x63: // F3
					zeroBasedMacroNumber = 2;
					break;
				
				case 0x76: // F4
					zeroBasedMacroNumber = 3;
					break;
				
				case 0x60: // F5
					zeroBasedMacroNumber = 4;
					break;
				
				case 0x61: // F6
					zeroBasedMacroNumber = 5;
					break;
				
				case 0x62: // F7
					zeroBasedMacroNumber = 6;
					break;
				
				case 0x64: // F8
					zeroBasedMacroNumber = 7;
					break;
				
				case 0x65: // F9
					zeroBasedMacroNumber = 8;
					break;
				
				case 0x6D: // F10
					zeroBasedMacroNumber = 9;
					break;
				
				case 0x67: // F11
					zeroBasedMacroNumber = 10;
					break;
				
				case 0x6F: // F12
					zeroBasedMacroNumber = 11;
					break;
				
				default: // ?
					break;
			}
			if (zeroBasedMacroNumber >= 0)
			{
				Macros_Send(inConnectionDataPtr, zeroBasedMacroNumber);
				return;
			}
			// otherwise, a key other than F1 through F12 was probably pressed
		}
		
		if (((ascii == '@') || (ascii == 32)) && controlDown) ascii = null;	// this, along with the fixed KCHR that maps a ^-@ to
																			// a @, takes care of Apple not posting null key values
		
		//	map '`' to ESC if needed
		if (ascii == '`' && remapBackquoteToEscape && !(commandDown)) ascii = ESC;
		
		if (code == BScode) // handle mapping BS to DEL, flipping on option-delete or command-delete
		{
			if (inConnectionDataPtr->bsdel)
			{
				if ((optionDown) || (commandDown)) ascii = BS;
				else ascii = DEL;
			}
			else
			{
				if ((optionDown) || (commandDown)) ascii = DEL;
				else ascii = BS;
			}
		}
			
		if (!inConnectionDataPtr->enabled)
		{
			// if suspended, and restart_any has been negotiated with the host,
			// enable the screen on anything but an XOFF.  If this turns out to
			// be XON, the character is absorbed later. (RFC 1372)
			if ((inConnectionDataPtr->restart_any_flow) && (ascii != inConnectionDataPtr->controlKey.suspend))
			{
				inConnectionDataPtr->enabled = 1;
			}
		}
	
		/* Remap PgUp,PgDown,Home,End if the user wants it that way */
		if ((inConnectionDataPtr->pgupdwn && (inConnectionDataPtr->ftpState == 0) && (code >= KPlowest))||(inConnectionDataPtr->terminal.emulation == kTerminalEmulationTypeVT100)) //do page up/down on vt100
			if (checkPageKeys(GetRealFrontWindow(), code)) return;		
		
		if ((inConnectionDataPtr->ftpState == 0) && (code >= KPlowest))
		{
			long		keyboardType = 0L;
			
			
			/*OSErr=*/Gestalt(gestaltKeyboardType, &keyboardType);
			if (keyboardType == gestaltStdADBKbd)	//standard MacII keyboard has keypad +,- 
			{												// codes switched	
				if (code == 0x45) code = 0x4E;
				else if (code == 0x4E) code = 0x45;
			}
	
			if ((inConnectionDataPtr->terminal.emulation == kTerminalEmulationTypeVT220)||(code >= 0x7B)||(code <= 0x60)) //fkeys dont work in vt100
			{
				if ((!inConnectionDataPtr->keypadmap)||(code == 0x4C)||(code > 0x51)||(code < 0x43)) //dont remap operators
				{
					if ((inConnectionDataPtr->clientflags & PASTE_IN_PROGRESS) &&
						(inConnectionDataPtr->paste.method == kTelnetPasteMethodBlock)) // queue this
					{
						inConnectionDataPtr->kbbuf[inConnectionDataPtr->kblen++] = 0;
						inConnectionDataPtr->kbbuf[inConnectionDataPtr->kblen++] = kpxlate[shiftDown][code - KPlowest];
						return;
					}
		
					if (inConnectionDataPtr->kblen > 0)
					{
						Network_SendData(inConnectionDataPtr->port, inConnectionDataPtr->kbbuf,
											inConnectionDataPtr->kblen);										
						inConnectionDataPtr->kblen=0;												
					}
		
					ascii = kpxlate[shiftDown][code - KPlowest];
					// Should we check here for ascii being zero?
					VSkbsend(inConnectionDataPtr->vs, (unsigned char) ascii, inConnectionDataPtr->echo);
					return;
				}
			}
			else // we dont handle function keys in vt100
				return;
		}
		
		// handle whatever mapping is needed
		TextTranslation_ConvertCharacterFromMac(&ascii, inConnectionDataPtr->national); // translate character
		
		if ((inConnectionDataPtr->clientflags & PASTE_IN_PROGRESS) &&
			(inConnectionDataPtr->paste.method == kTelnetPasteMethodBlock)) // queue this
		{
			inConnectionDataPtr->kbbuf[inConnectionDataPtr->kblen++] = ascii;
			return;
		}
	
		if (inConnectionDataPtr->lineMode.active)
		{
			LineMode_HandleKeyDown(ascii, inConnectionDataPtr);
			return;
		}
		
		// BSD-like mapping.... if we dont want this, set chars to zero and it wont work
		//CCP 2.7: this is now AFTER the linemode stuff, so that linemode can handle it differently 
		if (ascii == inConnectionDataPtr->controlKey.suspend)
		{
			if (inConnectionDataPtr->allow_flow) //remote flow control can turn this off
			{
				inConnectionDataPtr->enabled = 0;
				return;
			}
		}
	
		if (ascii == inConnectionDataPtr->controlKey.resume) 
		{
			if (inConnectionDataPtr->allow_flow) //remote flow control can turn this off
			{
				inConnectionDataPtr->enabled = 1;
				return;
			}
		}		
		if (ascii == inConnectionDataPtr->controlKey.interrupt)  
		{
			Str255		tellUser;
			
			
			// 3.0 - get a string that says [Interrupt Process]
			GetIndString(tellUser, rStringsMiscellaneous, siInterruptProcessTerminalFeedback);
			StringUtilities_PToCInPlace(tellUser);
			
			Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)tellUser, 23);
			Network_SendFlushData(inConnectionDataPtr->port, "\377\364", 2); // IAC IP
			Network_SendFlushData(inConnectionDataPtr->port, "\377\375\006", 3); // Do TM
			inConnectionDataPtr->timing = 1; // set emulate to TMwait
			return;
		}
		
		if (inConnectionDataPtr->ftpState)
		{
			// FTP client
			if (ascii>31 && ascii != 127 && ascii < 255 && code<KPlowest)
			{
				if (inConnectionDataPtr->kblen < (MAXKB -1)) // add to buffer if not full
				{
					inConnectionDataPtr->kbbuf[inConnectionDataPtr->kblen++] = ascii;
					characterToSend = ascii;
					Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, &characterToSend, 1);
				}
				else BEEP; // buffer full
			}
			else // not a printable character
			{
				if (code == BScode)
				{
					if (inConnectionDataPtr->kblen > 0)
					{
						inConnectionDataPtr->kblen--;
						Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)"\010 \010", 3);
					}
				}
				else if (ascii == CR)
				{
					Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)"\015\012", 2);
					inConnectionDataPtr->kbbuf[inConnectionDataPtr->kblen++] = 0;
					ftppi(inConnectionDataPtr->kbbuf); // FTP client
					inConnectionDataPtr->kblen = 0;
				}
				else if (ascii == KILLCHAR)
				{
					while (inConnectionDataPtr->kblen > 0)
					{
						Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)"\010 \010", 3);
						inConnectionDataPtr->kblen--;
					}
				}
			}
			return;
		}
		else if (inConnectionDataPtr->echo && !inConnectionDataPtr->halfdup)	// Handle klude-linemode
		{
			if (ascii > 31 && ascii < 127 && code < KPlowest) // printable ASCII
			{
				if (inConnectionDataPtr->kblen < (MAXKB - 1))
				{
					// add to buffer if not full
					inConnectionDataPtr->kbbuf[inConnectionDataPtr->kblen++] = ascii;
				}
				else 
				{
					// if full, send buffer
					Network_SendData(inConnectionDataPtr->port, inConnectionDataPtr->kbbuf,inConnectionDataPtr->kblen);	
					inConnectionDataPtr->kbbuf[0] = ascii;
					inConnectionDataPtr->kblen = 1;
				}
				
				characterToSend = ascii;
				Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, &characterToSend, 1);
				return;
			}
			else // non-printable ASCII
			{
				if (code == BScode)
				{
					if (inConnectionDataPtr->kblen > 0)
					{
						inConnectionDataPtr->kblen--;
						Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)"\010 \010", 3);
					}
					return;
				}
				else if (ascii == KILLCHAR)
				{
					while (inConnectionDataPtr->kblen > 0)
					{
						Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)"\010 \010", 3);
						inConnectionDataPtr->kblen--;
					}
					return;
				}
				else if (code < KPlowest)
				{
					// if full, send buffer
					Network_SendData(inConnectionDataPtr->port, inConnectionDataPtr->kbbuf, inConnectionDataPtr->kblen);
					inConnectionDataPtr->kblen = 0;
					if (ascii != CR)
					{
						characterToSend = '@' + ascii;
						Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)"^", 1);
						Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, &characterToSend, 1);
					}
				}
			}
		}//end if klude-linemode
		
		if (ascii == '\015'/* CR */)
		{
			// if CR map is on, send CR-null instead of CR-LF
			Network_SendFlushData(inConnectionDataPtr->port, (inConnectionDataPtr->crmap) ? "\015\0" : "\015\012", 2);
			if (inConnectionDataPtr->echo)
			{
				Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)"\012\015", 2);
			}
			return;
		}
		
		if (inConnectionDataPtr->echo && inConnectionDataPtr->halfdup) Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, &ascii, 1);
		
		if (ascii != 255) Network_SendData(inConnectionDataPtr->port, &ascii, 1);
		else Network_SendData(inConnectionDataPtr->port, "\377\377", 2);
	}
}// handleScreenKeyDown


/*!
Handles operating system events, such as an
application-switch event.  Returns true if
the event was handled, false otherwise.

(3.0)
*/
static Boolean
handleSwitch	(EventRecord*		inoutEventPtr)
{
	WindowRef			window = null;
	TelnetScreenID		screenID = 0;
	Boolean				resuming = (inoutEventPtr->message & resumeFlag),
						mustConvertScrap = (inoutEventPtr->message & convertClipboardFlag);
	Boolean				result = false;
	
	
	if (mustConvertScrap)
	{
		// MacTelnet does not do clipboard conversion
	}
	
	FlagManager_Set(kFlagSuspended, !resuming);
	Menus_SetSuspended(!resuming);
	Alert_SetIsBackgrounded(!resuming); // automatically removes any posted notifications from alerts
	Clipboard_SetSuspended(!resuming);
	EventLoop_SetFloatersSuspended(!resuming);
	
	if (resuming) if (gHaveInstalledNotification)
	{
		// this removes any flashing icons, etc. from the process menu, if a notification was posted
		NMRemove(gBeepNotificationPtr);
		Memory_DisposeHandle(&(gBeepNotificationPtr->nmIcon));
		gHaveInstalledNotification = false;
	}
	
	// 3.0 - service a pending alert
	if (resuming) Alert_ServiceNotification();
	
	// a switch event also encompasses an activate event, but
	// the event message must then become the actual window
	inoutEventPtr->message = (unsigned long)GetRealFrontWindow();
	handleActivate(inoutEventPtr, resuming/* activating */);
	
	// 3.0 - restore color look-up table for the benefit of applications
	// that dont use the Palette Manager
	RestoreDeviceClut(GetMainDevice());
	
	return result;
}// handleSwitch


/*!
Responds to update events for windows.  Returns
true if the event was handled, false otherwise.

(3.0)
*/
static Boolean
handleUpdate	(EventRecord*		inoutEventPtr)
{
	GrafPtr			oldPort = null,
					drawingPort = null;
	WindowRef		windowToUpdate = (WindowRef)inoutEventPtr->message;
	RgnHandle		visibleRegion = Memory_NewRegion();
	Boolean			result = false;
	
	
	GetPort(&oldPort);
	
	// set the port to the events intended target
	SetPortWindowPort(windowToUpdate);
	GetPort(&drawingPort);
	if (visibleRegion != null) GetPortVisibleRegion(drawingPort, visibleRegion);
	
	// do not waste time calculating and drawing windows that cannot be seen
	// (although receiving update events for invisible ports is very rare...)
	if (visibleRegion != null) unless (EmptyRgn(visibleRegion))
	{
		WindowInfoRef	windowInfoRef = null;
		
		
		windowInfoRef = WindowInfo_GetFromWindow(windowToUpdate);
		if (windowInfoRef != null)
		{
			WindowInfoDescriptor		windowDescriptor = WindowInfo_GetWindowDescriptor(windowInfoRef);
			
			
			switch (windowDescriptor)
			{
				case kTelnetWindowDescriptorAbout:
					{
						ControlRef		control = null;
						Rect			rect;
						
						
						// redraw the paged text pane offscreen, to ensure smooth redrawing
						(OSErr)GetDialogItemAsControl(GetDialogFromWindow(windowToUpdate),
														iAboutPagedTextPane, &control);
						Embedding_OffscreenDrawOneControl(control);
						GetControlBounds(control, &rect);
						MacLook_ValidateRectangle(windowToUpdate, &rect);
					}
					BeginUpdate(windowToUpdate);
					UpdateDialog(GetDialogFromWindow(windowToUpdate), visibleRegion);
					EndUpdate(windowToUpdate);
					result = true;
					break;
				
				case kTelnetWindowDescriptorClipboard:
					BeginUpdate(windowToUpdate);
					(OSErr)Clipboard_Update(drawingPort);
					EndUpdate(windowToUpdate);
					result = true;
					break;
				
				case kTelnetDialogDescriptorPreferences:
				case kTelnetWindowDescriptorInspectorWindoid:
				case kTelnetWindowDescriptorCommandLine:
				case kTelnetWindowDescriptorControlWindoid:
				case kTelnetWindowDescriptorFunctionWindoid:
				case kTelnetWindowDescriptorKeypadWindoid:
					BeginUpdate(windowToUpdate);
					UpdateControls(windowToUpdate, visibleRegion);
					EndUpdate(windowToUpdate);
					result = true;
					break;
				
				case kTelnetDialogDescriptorConfigurations:
				case kTelnetDialogDescriptorEditANSIColors:
				case kTelnetDialogDescriptorEditLogInMessage:
				case kTelnetDialogDescriptorFormat:
				case kTelnetDialogDescriptorMacroSetup:
				case kTelnetDialogDescriptorNewSessions:
				case kTelnetDialogDescriptorScreenSize:
				case kTelnetDialogDescriptorSetWindowTitle:
				case kTelnetDialogDescriptorNameNewFavorite:
				case kTelnetDialogDescriptorProxyServerSetup:
				case kTelnetDialogDescriptorSessionEditor:
					// IMPORTANT - it is only safe to call DrawDialog() on a known MacTelnet dialog
					BeginUpdate(windowToUpdate);
					UpdateDialog(GetDialogFromWindow(windowToUpdate), visibleRegion);
					EndUpdate(windowToUpdate);
					result = true;
					break;
				
				default:
					break;
			}
		}
		
		unless (result)
		{
			switch (GetWindowKind(windowToUpdate))
			{
				case WIN_CONSOLE: // the console window
				case WIN_LOG: // the FTP log
				case WIN_CNXN: // a connection window
					BeginUpdate(windowToUpdate);
					UpdateControls(windowToUpdate, visibleRegion);
					EndUpdate(windowToUpdate);
					result = true;
					break;
				
				case WIN_ICRG: // an Interactive Color Raster Graphics window
					if (MacRGupdate(windowToUpdate)) {/* error */}
					result = true;
					break;
					
				case WIN_TEK: // a Tektronix display
					if (RGupdate(windowToUpdate) == 0) TekDisable(RGgetVG(windowToUpdate));
					else {/* error */}
					result = true;
					break;
				
				default:
					// bad window kind
					break;
			}
		}
	}
	
	Memory_DisposeRegion(&visibleRegion);
	SetPort(oldPort);
	
	return result;
}// handleUpdate


/*!
To modify the global mouse region (for determining
when the cursor shape should change) to be the
specified rectangle in the specified window, use
this method.  The new mouse region will be labelled
using the specified region code, and the region
will automatically disclude floating windows and
any area outside the content region of the given
window.

Incomplete.

(3.0)
*/
static void
localRectToGlobalMouseRegion	(MouseRegionType const	inAsWhatRegion,
								 WindowRef				inWindowRef,
								 Rect const*			inLocalRectPtr)
{
	if (gMouseRegion != null)
	{
		GrafPtr		oldPort = null;
		RgnHandle	contentRegion = Memory_NewRegion();
		
		
		GetPort(&oldPort);
		SetPortWindowPort(inWindowRef);
		
		// obtain a region in global coordinates representing the specified local rectangle
		RectRgn(gMouseRegion, inLocalRectPtr);
		LocalToGlobalRegion(gMouseRegion);
		if (contentRegion != null)
		{
			// try to clip the region to within the windows content region
			if (GetWindowRegion(inWindowRef, kWindowContentRgn, contentRegion) == noErr)
			{
				DiffRgn(gMouseRegion, contentRegion, gMouseRegion);
			}
			Memory_DisposeRegion(&contentRegion);
		}
		
		// subtract out floating window structure regions
		diffFloatingRegions(gMouseRegion);
		
		// label the region
		gMouseRegionCursor = inAsWhatRegion;
		
		SetPort(oldPort);
	}
}// localRectToGlobalMouseRegion


/*!
To modify the global mouse region (for determining
when the cursor shape should change) to be the
specified rectangle in the specified window, use
this method.  The new mouse region will be labelled
using the specified region code, and the region
will automatically disclude floating windows and
any area outside the content region of the given
window.

Incomplete.

(3.0)
*/
static void
localRegionToGlobalMouseRegion	(MouseRegionType const	inAsWhatRegion,
								 WindowRef				inWindowRef,
								 RgnHandle				inLocalRegionHandle)
{
	if (gMouseRegion != null)
	{
		GrafPtr		oldPort = null;
		RgnHandle	contentRegion = Memory_NewRegion();
		
		
		GetPort(&oldPort);
		SetPortWindowPort(inWindowRef);
		
		// obtain a region in global coordinates representing the specified local rectangle
		CopyRgn(inLocalRegionHandle, gMouseRegion);
		LocalToGlobalRegion(gMouseRegion);
		if (contentRegion != null)
		{
			// try to clip the region to within the windows content region
			if (GetWindowRegion(inWindowRef, kWindowContentRgn, contentRegion) == noErr)
			{
				DiffRgn(gMouseRegion, contentRegion, gMouseRegion);
			}
			Memory_DisposeRegion(&contentRegion);
		}
		
		// subtract out floating window structure regions
		diffFloatingRegions(gMouseRegion);
		
		// label the region
		gMouseRegionCursor = inAsWhatRegion;
		
		SetPort(oldPort);
	}
}// localRegionToGlobalMouseRegion


/*!
To change the shape of the cursor in response
to user mouse movements, or to force the
cursors shape based on modifier keys, call
this method.

This method updates the global region that
defines what might cause the next mouse-
moved event, unless you set "inForceCursor"
to true.  You may want to do this if the
user has simply changed the current state
of modifier keys that affect the cursor
shape without necessarily moving the mouse.

If you do not force the cursor, then the
mouse location stored in the given event
record is used to determine what region the
mouse is in now.  The global region code
and region are then calculated accordingly.
This is the usual mode, invoked in response
to real mouse-moved events from the OS.

(2.6)
*/
static void
updateCursor	(EventRecord*		inoutEventPtr,
				 Boolean			inForceCursor)
{
	SInt16		newCursor = Cursors_GetCurrent();
	
	
	if (!inForceCursor)
	{
		// The cursor will not be forced, which means the current
		// mouse-moved event region needs recalculating.
		GrafPtr				oldPort = null;
		WindowRef			window = null;
		Point				localMouse = inoutEventPtr->where;
		Boolean				isHandledByTSM = false;
		TelnetScreenID		screenID = -1;
		
		
		window = GetRealFrontWindow(); // frontmost non-floating window
		screenID = ScreenManager_GetWindowScreenID(window);
		GetPort(&oldPort);
		SetPortWindowPort(window);
		GlobalToLocal(&localMouse);
		
		// 3.0 - support the Text Services Manager!
		if (FlagManager_Test(kFlagTextServicesManager))
		{
		#if TARGET_API_MAC_OS8
			isHandledByTSM = SetTSMCursor(inoutEventPtr->where);
		#else
			isHandledByTSM = false; // Carbon handles these events implicitly, so ignore them
		#endif
		}
		
		if (!isHandledByTSM)
		{
			switch (GetWindowKind(window))
			{
				case WIN_CONSOLE:
				case WIN_LOG:
				case WIN_CNXN:
					if (RealScreen_PtInRgn(screenID, inoutEventPtr->where, kRealScreenRegionCodeScreenArea)) 
					{
						if (RealScreen_PtInSelection(screenID, localMouse))
						{
							// the cursor is inside the screen, but also within the text selection
							RgnHandle		selectionRegion = null;
							
							
							selectionRegion = RealScreen_GetSelectedTextAsNewRegion(screenID);
							if (selectionRegion != null)
							{
								localRegionToGlobalMouseRegion(kMouseRegionTypeTerminalTextSelection,
																window, selectionRegion);
								Memory_DisposeRegion(&selectionRegion);
							}
						}
						else
						{
							// the cursor is outside the text selection, but within the screen itself
							RealScreenDataPtr		realScreenPtr = ScreenManager_GetRealScreenDataOf
																	(screenID);
							Rect					rect;
							
							
							if (realScreenPtr != null) GetControlBounds(realScreenPtr->contentUserPane, &rect);
							else SetRect(&rect, 0, 0, 0, 0);
							localRectToGlobalMouseRegion(kMouseRegionTypeTerminalScreen, window, &rect);
						}
					}
					else emptyGlobalMouseRegion();
					break;
				
				case WIN_TEK:
					{
						RgnHandle		contentRegion = Memory_NewRegion();
						
						
						GetWindowRegion(window, kWindowContentRgn, contentRegion);
						if (PtInRgn(inoutEventPtr->where, contentRegion))
						{
							// in graphics screen
							localRegionToGlobalMouseRegion(kMouseRegionTypeGraphicsScreen,
															window, contentRegion);
						}
						else emptyGlobalMouseRegion();
						Memory_DisposeRegion(&contentRegion);
					}
					break;
				
				case WIN_ICRG:
				case NO_WINDOW:
				default:
					emptyGlobalMouseRegion();
					break;
			}
			
			// now choose a cursor based on the mouse region
			switch (gMouseRegionCursor)
			{
				case kMouseRegionTypeTerminalScreen:
					newCursor = kCursor_IBeam;
					break;
				
				case kMouseRegionTypeTerminalTextSelection:
					newCursor = kCursor_HandOpen;
					break;
				
				case kMouseRegionTypeGraphicsScreen:
					newCursor = kCursor_Crosshairs;
					break;
				
				case kMouseRegionTypeNowhere:
				default:
					newCursor = kCursor_Arrow;
					break;
			}
		}
		
		SetPort(oldPort);
	}
	else
	{
		// Forced cursor: change the cursor, but not the mouse region.
		switch (gMouseRegionCursor)
		{
			case kMouseRegionTypeTerminalTextSelection:
				// While the cursor is in a terminals text selection, there are
				// certain modifiers which can change the cursor shape:
				//
				//	Modifier Key:			Cursor:
				//		control					contextual menu arrow
				//		option					terminal cursor (positioning)
				//		command					pointing hand (URL recognition)
				//		<others>				open hand (dragging)
				if (inoutEventPtr->modifiers & controlKey) newCursor = kCursor_ArrowContextualMenu;
				else if (inoutEventPtr->modifiers & optionKey) newCursor = kCursor_TerminalCursor;
				else if (inoutEventPtr->modifiers & cmdKey) newCursor = kCursor_HandPoint;
				else newCursor = kCursor_HandOpen;
				break;
			
			case kMouseRegionTypeTerminalScreen:
				// While the cursor is in a terminals screen area, there are
				// certain modifiers which can change the cursor shape:
				//
				//	Modifier Key:			Cursor:
				//		control					contextual menu arrow
				//		option					terminal cursor (positioning)
				//		<others>				I-beam
				if (inoutEventPtr->modifiers & controlKey) newCursor = kCursor_ArrowContextualMenu;
				else if (inoutEventPtr->modifiers & optionKey) newCursor = kCursor_TerminalCursor;
				else newCursor = kCursor_IBeam;
				break;
			
			case kMouseRegionTypeGraphicsScreen:
			case kMouseRegionTypeNowhere:
			default:
				if (inoutEventPtr->modifiers & controlKey) newCursor = kCursor_ArrowContextualMenu;
				break;
		}
	}
	
	// change the cursor, if necessary
	Cursors_Use(newCursor);
}// updateCursor


/*!
To show or hide a balloon for a window that
is not capable of automatically handling
help balloons (such as a terminal window),
use this method.  This routine is called
in response to mouse-moved events.

(3.0)
*/
static void
updateWindowBalloonVisibilities		(EventRecord*		inoutEventPtr)
{
#if TARGET_API_MAC_OS8
	GrafPtr			oldPort = null;
	WindowRef		window = null;
	Point			localMouse = inoutEventPtr->where;
	TelnetScreenID	screenID = -1;
	static Boolean	thisRoutineDisplayedTheBalloon = false; // avoids accidental removal of other balloons
	
	
	window = GetRealFrontWindow(); // frontmost non-floating window
	screenID = ScreenManager_GetWindowScreenID(window);
	GetPort(&oldPort);
	SetPortWindowPort(window);
	GlobalToLocal(&localMouse);
	
	switch (GetWindowKind(window))
	{
		case WIN_CNXN:
			// only connection windows have dynamic balloons
			{
				
			}
			break;
		
		case WIN_CONSOLE:
		case WIN_LOG:
		case WIN_TEK:
		case WIN_ICRG:
		case NO_WINDOW:
		default:
			emptyGlobalMouseRegion();
			if (thisRoutineDisplayedTheBalloon) (OSErr)HMRemoveBalloon();
			break;
	}
	
	SetPort(oldPort);
#endif
}// updateWindowBalloonVisibilities