/*###############################################################

	FinderList.c
	
	Interface Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

// the Mac OS X Aqua Human Interface Guidelines embeds the sort order icon into the
// column button of the current sort column, instead of having a separate button;
// to accommodate this, the code in this file supports both sort reversal alternatives
#if TARGET_API_MAC_OS8
#	define USE_SORT_ORDER_REVERSAL_BUTTON	1
#else
#	define USE_SORT_ORDER_REVERSAL_BUTTON	0
#endif

#include "UniversalDefines.h"

// Mac includes
#include <Appearance.h>
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Gestalt.h>
#include <QuickDraw.h>
#include <PLStringFuncs.h>
#include <Script.h>
#include <Threads.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "SpacingConstants.r"



// constants
enum
{
	kHeaderTextInsetV = 4,	// pixels vertically inward that header text is from the edges of its header
	kRowTextInsetV = 3		// pixels vertically inward that row text is from the edges of its row
};

// types
struct FinderListColumn
{
	FinderListColumnFlags			flags;					// attributes for this column (and possible cell defaults)
	FinderListColumnDataType		dataType;				// cell data type
	FinderListCellCompareProcPtr	compareProcPtr;			// compares two cells from this column, for sorting purposes
	ControlRef						headerButton;			// the bevel button or placard representing the columns title
	UInt16							width;					// convenient, redundant storage of width of header button
	ThemeBrush						backgroundBrush;		// convenient, redundant storage of brush to use for painting
};
#ifndef __cplusplus
typedef struct FinderListColumn		FinderListColumn;
#endif
typedef FinderListColumn*			FinderListColumnPtr;

struct FinderListRow
{
	FinderListRowFlags			flags;					// attributes for a particular row (and possible cell defaults)
	ChainRef					cellList;				// contains cell structure pointers (FinderListCellPtr)
};
#ifndef __cplusplus
typedef struct FinderListRow		FinderListRow;
#endif
typedef FinderListRow*				FinderListRowPtr;

struct FinderListCell
{
	Ptr		dataPtr;		// cell data, whose type depends on the FinderListColumn its in
};
#ifndef __cplusplus
typedef struct FinderListCell	FinderListCell;
#endif
typedef FinderListCell*			FinderListCellPtr;

struct FinderList
{
	FinderListFlags			flags;						// attributes applicable to the entire list
	Boolean					haveAppearance1_1,			// Appearance 1.1 or later available?
							haveCarbonControls,			// Carbon Control Manager available?
							sortDescending;				// reverse current sort?
	
	UInt16					selectedColumnIndex;		// zero-based column index of current sort column; -1 if none
	UInt16					sortableColumnsCount;		// used to initialize sort columns properly
	UInt16					rowHeight;					// the height of each row, in pixels
	ThemeFontID				cellThemeFontID,			// which font to use for cell text
							headerThemeFontID;			// which font to use for heading buttons
	
	struct
	{
		ControlRef				encompassingUserPane,		// an invisible control, used to track hits and draw focus ring
								columnHeaderPlacard,		// a placard that sits above the list content area
							#if USE_SORT_ORDER_REVERSAL_BUTTON
								sortOrderReversalButton,	// if desired, a sort order reversal button
							#endif
								contentUserPane,			// used to render the list content area
								scrollBarV;					// used to scroll the list up or down
	} controls;
	
	struct
	{
		struct
		{
			ControlUserPaneActivateUPP		activationUPP;
			ControlUserPaneDrawUPP			drawingUPP;
			ControlUserPaneIdleUPP			idleUPP;
			ControlUserPaneKeyDownUPP		keyDownUPP;
		} contentUserPane;
		
		struct
		{
			ControlUserPaneDrawUPP			drawingUPP;
			ControlUserPaneFocusUPP			focusUPP;
			ControlUserPaneHitTestUPP		hitTestUPP;
			ControlUserPaneTrackingUPP		trackingUPP;
		} encompassingUserPane;
		
		struct
		{
			ControlActionUPP			actionUPP;
		} scrollBarV;
	} procs;
	
	ChainRef					columnList,			// contains column structure pointers (FinderListColumnPtr)
								rowList,			// contains row structure pointers (FinderListRowPtr)
								selectionList;		// contains zero-based row indices (UInt16*); all selected rows
	
	ControlRef					subControlHit;		// which sub-control was hit, by the most recent hit test invocation
	
	WindowRef					owningWindow;
};
#ifndef __cplusplus
typedef struct FinderList	FinderList;
#endif
typedef FinderList*			FinderListPtr;

// variables
static ControlRef*			gDragDataHeaderButtons = null;
static SInt16				gDragDataHeaderButtonCount = 0;
static Rect					gDragDataInsertionRect = { -20, -20, -20, -20 },
							gDragDataContentRect = { 0, 0, 0, 0 };
static SInt16				gSortColumnIndex = 0; // used ONLY during sorting

// internal methods
static void						calculateViewSize					(FinderListPtr			inListPtr);
static void						columnDrag							(FinderListPtr			inListPtr,
																	 SInt16 const			inDraggingWhichColumnIndex);
static void						columnDragInsertionDraw				(Rect const*			inRectPtr,
																	 Boolean const			inDraw);
static pascal void				columnDragProgress					(void);
static int						compareTextCellData					(void const*			inCellData1,
																	 void const*			inCellData2);
static pascal void				contentUserPaneActivate				(ControlRef				inControl,
																	 Boolean				inActivating);
static pascal void				contentUserPaneDraw					(ControlRef				inControl,
																	 SInt16					inPart);
static pascal void				contentUserPaneDeviceLoop			(short					inColorDepth,
																	 short					inDeviceFlags,
																	 GDHandle				inTargetDevice,
																	 long					inControl);
static pascal void				contentUserPaneIdle					(ControlRef				inControl);
static void						drawCell							(FinderListPtr			inListPtr,
																	 FinderListColumnPtr	inColumnPtr,
																	 FinderListRowPtr		inRowPtr,
																	 FinderListCellPtr		inCellPtr,
																	 short					inColorDepth,
																	 GDHandle				inTargetDevice,
							 										 Rect const*			inBoundsPtr);
static pascal void				encompassingUserPaneDraw			(ControlRef				inControl,
																	 SInt16					inPart);
static pascal ControlPartCode	encompassingUserPaneFocus			(ControlRef				inControl,
																	 ControlFocusPart		inAction);
static pascal ControlPartCode	encompassingUserPaneHitTest			(ControlRef				inControl,
																	 Point					inLocalMouse);
static pascal ControlPartCode	encompassingUserPaneKeyDown			(ControlRef				inControl,
																	 SInt16					inKeyCode,
																	 SInt16					inCharCode,
																	 SInt16					inModifiers);
static pascal ControlPartCode	encompassingUserPaneTrackingProc	(ControlRef				inControl,
																	 Point					inGlobalMouse,
																	 ControlActionUPP		inIgnoredActionUPP);
static FinderListCellPtr		findCell							(FinderListPtr			inListPtr,
																	 SInt16 const			inZeroBasedRowIndex,
																	 SInt16 const			inZeroBasedColumnIndex);
static SInt16					getFirstVisibleRow					(FinderListPtr			inListPtr);
static OSStatus					getThemeMetric						(ThemeMetric 			inMetric,
																	 SInt32*				outMetric);
static SInt16					getVisibleRowCount					(FinderListPtr			inListPtr);
static Boolean					isCommandKeyDown					(void);
static Boolean					isOptionKeyDown						(void);
static Boolean					isShiftKeyDown						(void);
static SInt16					newHeadingButtonsArray				(FinderListPtr			inListPtr,
																	 ControlRef**			outArrayPtr);
static FinderListPtr			refAcquireLock						(FinderListRef			inRef);
static void						refReleaseLock						(FinderListRef			inRef,
																	 FinderListPtr*			inoutPtr);
static void						reverseSort							(FinderListPtr			inListPtr);
static int						rowComparer							(void const*			inFinderListRowPtr1,
																	 void const*			inFinderListRowPtr2);
static pascal void				scrollBarActionProc					(ControlRef				inControl,
																	 ControlPartCode		inPartCode);
static void						scrollDownward						(FinderListPtr			inListPtr,
																	 UInt16 const			inRowsToScroll,
																	 Boolean const			inUpdateList);
static void						scrollToRow							(FinderListPtr			inListPtr,
																	 SInt16 const			inZeroBasedRowIndex,
																	 Boolean const			inVerticallyCenterInView);
static void						scrollUpward						(FinderListPtr			inListPtr,
																	 UInt16 const			inRowsToScroll,
																	 Boolean const			inUpdateList);
static void						selectionRectangleDrag				(FinderListPtr			inListPtr);
static void						selectionRectangleDraw				(FinderListPtr			inListPtr,
																	 Rect const*			inRectPtr,
																	 Boolean const			inDraw);
static void						setSortColumn						(FinderListPtr			inPtr,
																	 SInt16 const			inZeroBasedColumnIndex);
static void						sortRows							(FinderListPtr			inListPtr);
static void						useBestSortAscendingButtonIcon		(ControlRef				inControl);
static void						useBestSortDescendingButtonIcon		(ControlRef				inControl);



//
// public methods
//

/*!
Creates a new Finder List control that will appear
in the specified window.

If any problems occur, null is returned; otherwise,
a reference to the new Finder List is returned.

Once you have a new Finder List, you need to give
it content by inserting rows and columns.

(1.0)
*/
FinderListRef
FinderList_New		(WindowRef const			inOwningWindow,
					 FinderListFlags const		inFlags)
{
	FinderListRef		result = (FinderListRef)Memory_NewPtr(sizeof(FinderList));
	
	
	if (result != null)
	{
		FinderListPtr		ptr = refAcquireLock(result);
		ControlRef			control = null;
		
		
		ptr->flags = inFlags;
		{
			long		gestaltResult = 0L;
			UInt8		majorRev = 0,
						minorRev = 0;
			
			
			if (Gestalt(gestaltAppearanceVersion, &gestaltResult) == noErr)
			{
				majorRev = Releases_GetMajorRevisionForVersion(gestaltResult);
				minorRev = Releases_GetMinorRevisionForVersion(gestaltResult);
				
				// any advanced APIs available?
				ptr->haveAppearance1_1 = ((((majorRev == 0x01) && (minorRev >= 0x01)) || (majorRev > 0x01)));
			}
			ptr->haveCarbonControls = (Gestalt(gestaltControlMgrAttr, &gestaltResult) == noErr);
		}
		ptr->sortDescending = false;
		ptr->selectedColumnIndex = -1;
		ptr->sortableColumnsCount = 0;
		ptr->rowHeight = 0;
		ptr->cellThemeFontID = kThemeViewsFont;
		ptr->headerThemeFontID = (ptr->flags & kFinderListFlagsHeaderButtonsUseSmallFont)
									? kThemeSmallSystemFont : kThemeViewsFont;
		ptr->owningWindow = inOwningWindow;
		ptr->columnList = Chain_New(kChainStyleSinglyLinked/*kChainStyleDoublyLinked*/);
		ptr->rowList = Chain_New(kChainStyleSinglyLinked);
		ptr->selectionList = Chain_New(kChainStyleSinglyLinked);
		
		// create controls and set up an embedding hierarchy; the custom
		// (user pane) controls have their control reference set to the
		// "FinderListRef" that they belong to, so that the callback
		// routines have access to the proper object reference; also, the
		// controls belonging to a Finder List are all supervised by
		// the encompassing user pane, so that it may handle all events
		ptr->controls.encompassingUserPane = control = GetNewControl(kInterfaceLibControlFinderListEncompassingUserPane,
																		ptr->owningWindow);
		SetControlReference(control, REINTERPRET_CAST(result, SInt32)); // IMPORTANT - all controls with callbacks need this!
		
		ptr->controls.contentUserPane = control = GetNewControl(kInterfaceLibControlFinderListContentUserPane,
																ptr->owningWindow);
		SetControlReference(control, REINTERPRET_CAST(result, SInt32)); // IMPORTANT - all controls with callbacks need this!
		(OSErr)EmbedControl(control, ptr->controls.encompassingUserPane);
		(OSErr)SetControlSupervisor(control, ptr->controls.encompassingUserPane);
		
		ptr->controls.columnHeaderPlacard = control = GetNewControl(kInterfaceLibControlFinderListHeadingPlacard,
																	ptr->owningWindow);
		(OSErr)EmbedControl(control, ptr->controls.encompassingUserPane);
		(OSErr)SetControlSupervisor(control, ptr->controls.encompassingUserPane);
		
	#if USE_SORT_ORDER_REVERSAL_BUTTON
		// under Aqua, there is no sort order reversal button, because
		// each column button reflects the sort direction
		ptr->controls.sortOrderReversalButton = control = GetNewControl
															(kInterfaceLibControlFinderListHeadingSortOrderReversalButton,
															ptr->owningWindow);
		useBestSortAscendingButtonIcon(control);
		(OSErr)EmbedControl(control, ptr->controls.encompassingUserPane);
		(OSErr)SetControlSupervisor(control, ptr->controls.encompassingUserPane);
	#endif
		
		ptr->controls.scrollBarV = control = GetNewControl(kInterfaceLibControlFinderListScrollBarV,
																ptr->owningWindow);
		SetControlReference(control, REINTERPRET_CAST(result, SInt32)); // IMPORTANT - all controls with callbacks need this!
		SetControlMinimum(control, 0);
		SetControlMaximum(control, -1);
		SetControlValue(control, -1);
		(OSErr)EmbedControl(control, ptr->controls.encompassingUserPane);
		// the scroll bar is not supervised by the list because SetControlAction() is used
		//(OSErr)SetControlSupervisor(control, ptr->controls.encompassingUserPane);
		
		// create control user pane callbacks - note that it is probably not
		// necessary to create new UPPs for each individual Finder List (it
		// is probably sufficient to use global variables); however, this
		// would require FinderList_Init() and FinderList_Done() functions to
		// allocate and dispose of the UPP structures properly...so, copies
		// of the UPPs are used for each Finder List
		ptr->procs.contentUserPane.activationUPP = NewControlUserPaneActivateUPP(contentUserPaneActivate);
		ptr->procs.contentUserPane.drawingUPP = NewControlUserPaneDrawUPP(contentUserPaneDraw);
		ptr->procs.contentUserPane.idleUPP = NewControlUserPaneIdleUPP(contentUserPaneIdle);
		ptr->procs.contentUserPane.keyDownUPP = NewControlUserPaneKeyDownUPP(encompassingUserPaneKeyDown);
		ptr->procs.encompassingUserPane.drawingUPP = NewControlUserPaneDrawUPP(encompassingUserPaneDraw);
		ptr->procs.encompassingUserPane.focusUPP = NewControlUserPaneFocusUPP(encompassingUserPaneFocus);
		ptr->procs.encompassingUserPane.hitTestUPP = NewControlUserPaneHitTestUPP(encompassingUserPaneHitTest);
		ptr->procs.encompassingUserPane.trackingUPP = NewControlUserPaneTrackingUPP(encompassingUserPaneTrackingProc);
		ptr->procs.scrollBarV.actionUPP = NewControlActionUPP(scrollBarActionProc);
		(OSErr)SetControlData(ptr->controls.contentUserPane, kControlNoPart, kControlUserPaneActivateProcTag,
									sizeof(ControlUserPaneActivateUPP), (Ptr)&ptr->procs.contentUserPane.activationUPP);
		(OSErr)SetControlData(ptr->controls.contentUserPane, kControlNoPart, kControlUserPaneDrawProcTag,
									sizeof(ControlUserPaneDrawUPP), (Ptr)&ptr->procs.contentUserPane.drawingUPP);
		(OSErr)SetControlData(ptr->controls.contentUserPane, kControlNoPart, kControlUserPaneIdleProcTag,
									sizeof(ControlUserPaneIdleUPP), (Ptr)&ptr->procs.contentUserPane.idleUPP);
		(OSErr)SetControlData(ptr->controls.encompassingUserPane, kControlNoPart, kControlUserPaneKeyDownProcTag,
									sizeof(ControlUserPaneKeyDownUPP), (Ptr)&ptr->procs.contentUserPane.keyDownUPP);
		(OSErr)SetControlData(ptr->controls.encompassingUserPane, kControlNoPart, kControlUserPaneDrawProcTag,
									sizeof(ControlUserPaneDrawUPP), (Ptr)&ptr->procs.encompassingUserPane.drawingUPP);
		(OSErr)SetControlData(ptr->controls.encompassingUserPane, kControlNoPart, kControlUserPaneFocusProcTag,
									sizeof(ControlUserPaneFocusUPP), (Ptr)&ptr->procs.encompassingUserPane.focusUPP);
		(OSErr)SetControlData(ptr->controls.encompassingUserPane, kControlNoPart, kControlUserPaneHitTestProcTag,
									sizeof(ControlUserPaneHitTestUPP), (Ptr)&ptr->procs.encompassingUserPane.hitTestUPP);
		(OSErr)SetControlData(ptr->controls.encompassingUserPane, kControlNoPart, kControlUserPaneTrackingProcTag,
									sizeof(ControlUserPaneTrackingUPP), (Ptr)&ptr->procs.encompassingUserPane.trackingUPP);
		SetControlAction(ptr->controls.scrollBarV, ptr->procs.scrollBarV.actionUPP);
		
		// release lock
		refReleaseLock(result, &ptr);
	}
	return result;
}// New


/*!
This method cleans up a Finder List by destroying all
of the data associated with it.  On output, your copy
of the given reference will be set to null.

(1.0)
*/
void
FinderList_Dispose		(FinderListRef*		inoutRefPtr)
{
	if (inoutRefPtr != null)
	{
		FinderListPtr		ptr = refAcquireLock(*inoutRefPtr);
		
		
		if (ptr != null)
		{
			// dispose of callbacks
			DisposeControlUserPaneActivateUPP(ptr->procs.contentUserPane.activationUPP);
			DisposeControlUserPaneDrawUPP(ptr->procs.contentUserPane.drawingUPP);
			DisposeControlUserPaneIdleUPP(ptr->procs.contentUserPane.idleUPP);
			DisposeControlUserPaneKeyDownUPP(ptr->procs.contentUserPane.keyDownUPP);
			DisposeControlUserPaneDrawUPP(ptr->procs.encompassingUserPane.drawingUPP);
			DisposeControlUserPaneFocusUPP(ptr->procs.encompassingUserPane.focusUPP);
			DisposeControlUserPaneHitTestUPP(ptr->procs.encompassingUserPane.hitTestUPP);
			DisposeControlUserPaneTrackingUPP(ptr->procs.encompassingUserPane.trackingUPP);
			DisposeControlActionUPP(ptr->procs.scrollBarV.actionUPP);
			
			// dispose of controls
			DisposeControl(ptr->controls.encompassingUserPane); // this destroys all of the Finder Lists controls
			
			// free all rows and columns
			{
				ChainLinkRef		columnDataNode = null,
									rowDataNode = null,
									cellDataNode = null;
				FinderListColumnPtr	columnDataPtr = null;
				FinderListRowPtr	rowDataPtr = null;
				FinderListCellPtr	cellDataPtr = null;
				SInt16				columnCount = 0,
									rowCount = 0,
									cellCount = 0;
				register SInt16		i = 0;
				
				
				// iterate through rows
				rowCount = Chain_GetNodeCount(ptr->rowList);
				rowDataNode = Chain_GetFirstNode(ptr->rowList);
				for (i = 0; i < rowCount; i++)
				{
					rowDataPtr = (FinderListRowPtr)Chain_GetNodeDataPtr(ptr->rowList, rowDataNode);
					if (rowDataPtr != null)
					{
						// iterate through cells, destroying each one
						cellCount = Chain_GetNodeCount(rowDataPtr->cellList);
						cellDataNode = Chain_GetFirstNode(rowDataPtr->cellList);
						for (i = 0; i < rowCount; i++)
						{
							cellDataPtr = (FinderListCellPtr)Chain_GetNodeDataPtr(rowDataPtr->cellList, cellDataNode);
							if (cellDataPtr != null)
							{
								if (cellDataPtr->dataPtr != null)
								{
									DisposePtr((Ptr)cellDataPtr->dataPtr), cellDataPtr->dataPtr = null;
								}
								DisposePtr((Ptr)cellDataPtr), cellDataPtr = null;
							}
							cellDataNode = Chain_GetNextNode(rowDataPtr->cellList, cellDataNode);
						}
						
						// destroy the row data itself
						DisposePtr((Ptr)rowDataPtr), rowDataPtr = null;
					}
					rowDataNode = Chain_GetNextNode(ptr->rowList, rowDataNode);
				}
				
				// iterate through columns
				columnCount = Chain_GetNodeCount(ptr->columnList);
				columnDataNode = Chain_GetFirstNode(ptr->columnList);
				for (i = 0; i < columnCount; i++)
				{
					columnDataPtr = (FinderListColumnPtr)Chain_GetNodeDataPtr(ptr->columnList, columnDataNode);
					if (columnDataPtr != null)
					{
						DisposePtr((Ptr)columnDataPtr), columnDataPtr = null;
					}
					columnDataNode = Chain_GetNextNode(ptr->columnList, columnDataNode);
				}
			}
			
			// delete lists
			Chain_Dispose(&ptr->columnList);
			Chain_Dispose(&ptr->rowList);
			Chain_Dispose(&ptr->selectionList);
			
			// delete data structure
			DisposePtr(*((Ptr*)inoutRefPtr)), ptr = null, *inoutRefPtr = null;
		}
	}
}// Dispose


/*!
Changes the size of a Finder List by adjusting
appropriate controls by the specified horizontal
and vertical deltas.

(1.0)
*/
void
FinderList_DeltaSizeControl		(FinderListRef			inRef,
								 SInt16 const			inChangeInWidth,
								 SInt16 const			inChangeInHeight)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ControlRef		control = null;
		Rect			controlRect;
		
		
		// adjust controls based on the specified deltas
		control = ptr->controls.encompassingUserPane;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidth,
					controlRect.bottom - controlRect.top + inChangeInHeight);
		if (inChangeInWidth)
		{
			control = ptr->controls.columnHeaderPlacard;
			GetControlBounds(control, &controlRect);
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidth,
						controlRect.bottom - controlRect.top);
		#if USE_SORT_ORDER_REVERSAL_BUTTON
			control = ptr->controls.sortOrderReversalButton;
			GetControlBounds(control, &controlRect);
			MoveControl(control, controlRect.left + inChangeInWidth, controlRect.top);
		#endif
			{
				// for each installed column, resize each button appropriately
				ChainLinkRef			node = null;
				FinderListColumnPtr		columnDataPtr = null;
				Rect					placardRect,
										buttonRect;
				SInt16					leftEdge = 0,
										topEdge = 0,
										buttonWidth = 0,
										buttonDeltaWidth = inChangeInWidth / Chain_GetNodeCount(ptr->columnList);
				
				
				GetControlBounds(ptr->controls.columnHeaderPlacard, &placardRect);
				leftEdge = placardRect.left;
				topEdge = placardRect.top;
				node = Chain_GetFirstNode(ptr->columnList);
				do
				{
					columnDataPtr = Chain_GetNodeDataPtr(ptr->columnList, node);
					if (columnDataPtr != null)
					{
						control = columnDataPtr->headerButton;
						GetControlBounds(control, &buttonRect);
						buttonWidth = buttonRect.right - buttonRect.left;
						MoveControl(control, leftEdge, topEdge);
						SizeControl(control, buttonWidth + buttonDeltaWidth, placardRect.bottom - placardRect.top);
						leftEdge += (buttonWidth + buttonDeltaWidth);
						columnDataPtr->width += buttonDeltaWidth;
					}
					node = Chain_GetNextNode(ptr->columnList, node);
				} while (node != null);
			}
		}
		control = ptr->controls.scrollBarV;
		GetControlBounds(control, &controlRect);
		MoveControl(control, controlRect.left + inChangeInWidth, controlRect.top);
		SizeControl(control, controlRect.right - controlRect.left,
					controlRect.bottom - controlRect.top + inChangeInHeight);
		control = ptr->controls.contentUserPane;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidth,
					controlRect.bottom - controlRect.top + inChangeInHeight);
		
		// update scroll bars
		calculateViewSize(ptr);
		
		refReleaseLock(inRef, &ptr);
	}
}// DeltaSizeControl


/*!
To get a COPY of the data in a cell, use this method.
Generally, prior to using this routine, you should
call FinderList_GetCellDataSize(), to make sure that
your buffer space is sufficient.

Either way, if the specified buffer is not large
enough (or if any other error occurs), the returned
size will be 0.

(1.0)
*/
void
FinderList_GetCellData	(FinderListRef			inRef,
						 SInt16 const			inZeroBasedRowIndex,
						 SInt16 const			inZeroBasedColumnIndex,
						 void*					inoutDataPtr,
						 Size const				inBufferSpace,
						 Size*					outActualSizePtr)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		FinderListCellPtr	cellPtr = findCell(ptr, inZeroBasedRowIndex, inZeroBasedColumnIndex);
		
		
		if (cellPtr != null)
		{
			Size		actualSize = GetPtrSize(cellPtr->dataPtr);
			
			
			if (outActualSizePtr != null) *outActualSizePtr = 0L;
			if (actualSize <= inBufferSpace)
			{
				if (outActualSizePtr != null) *outActualSizePtr = actualSize;
				if (inoutDataPtr != null)
				{
					// success - copy the data to the buffer
					BlockMove(cellPtr->dataPtr, inoutDataPtr, actualSize);
				}
			}
		}
		refReleaseLock(inRef, &ptr);
	}
}// GetCellData


/*!
To determine the size of the data in a cell, use
this method.  Generally, prior to using the
FinderList_GetCellData() method, you would invoke
this routine to determine the amount of space
required.

(1.0)
*/
Size
FinderList_GetCellDataSize	(FinderListRef		inRef,
							 SInt16 const		inZeroBasedRowIndex,
							 SInt16 const		inZeroBasedColumnIndex)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	Size				result = 0L;
	
	
	if (ptr != null)
	{
		FinderListCellPtr	cellPtr = findCell(ptr, inZeroBasedRowIndex, inZeroBasedColumnIndex);
		
		
		if (cellPtr != null)
		{
			result = GetPtrSize(cellPtr->dataPtr);
		}
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetCellDataSize


/*!
Returns the number of columns defined for the list.

(1.0)
*/
UInt16
FinderList_GetColumnCount		(FinderListRef			inRef)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	UInt16				result = 0;
	
	
	if (ptr != null)
	{
		result = Chain_GetNodeCount(ptr->columnList);
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetColumnCount


/*!
To acquire a handle to a control that encompasses
the entire Finder List (so you can insert or
remove all of the controls contents in a window),
use this method.  The boundaries of the returned
control are the same as those of the Finder List
itself, with the exceptions of extraneous regions
such as the focus rectangle or outer bevel.

(1.0)
*/
ControlRef
FinderList_GetContainerControl		(FinderListRef		inRef)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	ControlRef			result = null;
	
	
	if (ptr != null)
	{
		result = ptr->controls.encompassingUserPane;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetContainerControl


/*!
Returns a reference to the window that contains
the specified Finder List.

(1.0)
*/
WindowRef
FinderList_GetOwningWindow		(FinderListRef		inRef)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	WindowRef			result = null;
	
	
	if (ptr != null)
	{
		result = ptr->owningWindow;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetOwningWindow


/*!
Returns the index of the row that is at the top of
the viewable region of the list.  If the list is
empty, -1 is returned.

(1.0)
*/
SInt16
FinderList_GetFirstVisibleRow		(FinderListRef			inRef)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	SInt16				result = -1;
	
	
	if (ptr != null)
	{
		result = getFirstVisibleRow(ptr);
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetFirstVisibleRow


/*!
Returns a new region describing, in local coordinates,
the requested component of a Finder List.  You must
dispose of this region yourself when you are finished
with it.

You might use this routine to help you implement
special external functionality, such as drag and drop.

IMPORTANT:	These regions are dynamic; do not retain a
			region for any length of time if it is at
			all possible that the event loop will run,
			because the user could modify a list so as
			to relocate components, making your region
			handle the improper size.

(1.0)
*/
RgnHandle
FinderList_GetNewComponentRegion	(FinderListRef						inRef,
									 FinderListComponentType const		inWhatComponent)
{
	RgnHandle		result = null;
	FinderListPtr	ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		result = Memory_NewRegion();
		if ((QDError() == noErr) && (result != null)) switch (inWhatComponent)
		{
			case kFinderListComponentTypeListContents:
			default:
				{
					Rect		bounds;
					
					
					GetControlBounds(ptr->controls.contentUserPane, &bounds);
					RectRgn(result, &bounds);
				}
				break;
		}
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetNewComponentRegion


/*!
Returns the number of columns defined for the list.

(1.0)
*/
UInt16
FinderList_GetRowCount		(FinderListRef			inRef)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	UInt16				result = 0;
	
	
	if (ptr != null)
	{
		result = Chain_GetNodeCount(ptr->rowList);
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetRowCount


/*!
Returns the index of the row that is at the top of
the viewable region of the list.  If the list is
empty, -1 is returned.

(1.0)
*/
SInt16
FinderList_GetVisibleRowCount		(FinderListRef			inRef)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	SInt16				result = -1;
	
	
	if (ptr != null)
	{
		result = getVisibleRowCount(ptr);
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetVisibleRowCount


/*!
Inserts a new column into the list.  Always
insert all of your desired columns before
inserting rows - otherwise, inserting a
column will enter nullified data for the
cells in existing rows that intersect the
new column.

The sizes of the lists sub-controls are
re-calculated to ensure that the new column
fits in the list.

The number of columns in the list, following
the insertion, is returned.

Currently, Finder Lists only support cells of
the type "kFinderListColumnDataTypePlainText"
(text); any other type of data will be
rendered only if it can be coerced into text.

You must redraw the list yourself.

(1.0)
*/
UInt16
FinderList_InsertColumn		(FinderListRef						inRef,
							 SInt16 const						inInsertBeforeWhichZeroBasedIndexOrMinus1,
							 FinderListColumnDataType const		inDataType,
							 FinderListColumnFlags const		inFlags,
							 StringPtr							inTitle)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	UInt16				result = 0;
	
	
	if (ptr != null)
	{
		FinderListColumnPtr		columnDataPtr = (FinderListColumnPtr)Memory_NewPtr(sizeof(FinderListColumn));
		FinderListRowPtr		rowDataPtr = null;
		FinderListCellPtr		cellDataPtr = null;
		ChainLinkRef			node = null,
								rowDataNode = null;
		ControlRef				control = null;
		Rect					controlRect;
		register SInt16			i = 0,
								index = -1,
								rowCount = Chain_GetNodeCount(ptr->rowList);
		
		
		// set the index of this column
		if (inInsertBeforeWhichZeroBasedIndexOrMinus1 < 0) index = Chain_GetNodeCount(ptr->columnList);
		else index = inInsertBeforeWhichZeroBasedIndexOrMinus1;
		
		// increment number of sortable columns, if necessary
		unless (inFlags & kFinderListColumnFlagsNotSortable) ptr->sortableColumnsCount++;
		
		// initialize the new column, creating empty cells for each row in the list
		columnDataPtr->flags = inFlags;
		columnDataPtr->dataType = inDataType;
		columnDataPtr->compareProcPtr = compareTextCellData;
		columnDataPtr->backgroundBrush = (ptr->sortableColumnsCount == 1) ? kThemeBrushListViewSortColumnBackground
																			: kThemeBrushListViewBackground;
		columnDataPtr->headerButton = control = GetNewControl
			((inFlags & kFinderListColumnFlagsNotSortable) ? kInterfaceLibControlFinderListHeadingPlacard
															: kInterfaceLibControlFinderListHeadingBevelButton,
																ptr->owningWindow);
		SetBevelButtonTextAlignment(control, kControlBevelButtonAlignTextFlushLeft, kHeaderTextInsetV/* offset */);
		SetBevelButtonGraphicAlignment(control, kControlBevelButtonAlignRight, 0/* offset horiz. */, 0/* offset vert. */);
		SetControlTitle(control, inTitle);
		(OSErr)EmbedControl(control, ptr->controls.columnHeaderPlacard);
		(OSErr)SetControlSupervisor(control, ptr->controls.encompassingUserPane);
		rowDataNode = Chain_GetFirstNode(ptr->rowList);
		for (i = 0; i < rowCount; i++)
		{
			rowDataPtr = (FinderListRowPtr)Chain_GetNodeDataPtr(ptr->rowList, rowDataNode);
			if (rowDataPtr != null)
			{
				cellDataPtr = (FinderListCellPtr)Memory_NewPtr(sizeof(FinderListCell));
				if (cellDataPtr != null)
				{
					// initialize new cell
					cellDataPtr->dataPtr = null;
				}
				node = Chain_NewNode(rowDataPtr->cellList, cellDataPtr);
				Chain_MoveNode(rowDataPtr->cellList, node,
									(inInsertBeforeWhichZeroBasedIndexOrMinus1 < 0)
										? null
										: Chain_GetIndexedNode(rowDataPtr->cellList, inInsertBeforeWhichZeroBasedIndexOrMinus1),
									kChainLinkPlacementInsertBefore);
			}
			rowDataNode = Chain_GetNextNode(ptr->rowList, rowDataNode);
		}
		
		// add the data to the list of columns
		node = Chain_NewNode(ptr->columnList, columnDataPtr);
		Chain_MoveNode(ptr->columnList, node,
							(inInsertBeforeWhichZeroBasedIndexOrMinus1 < 0)
								? null
								: Chain_GetIndexedNode(ptr->columnList, inInsertBeforeWhichZeroBasedIndexOrMinus1),
							kChainLinkPlacementInsertAfter);
		
		// if this is the first column, make it the default sort column
		if (ptr->sortableColumnsCount == 1)
		{
			SetControlValue(control, kControlCheckBoxCheckedValue);
			ptr->selectedColumnIndex = index;
		}
		
		// by resizing the control to its current size, all sub-controls are
		// automatically resized properly to account for the new column
		GetControlBounds(control, &controlRect);
		FinderList_SizeControl(inRef, controlRect.right - controlRect.left, controlRect.bottom - controlRect.top);
		
		// return the number of items now in the list
		result = Chain_GetNodeCount(ptr->columnList);
		
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// InsertColumn


/*!
Inserts a new row into the list.  You can
populate cells in the new row (or any row)
using FinderList_SetCellData().

The number of rows in the list, following
the insertion, is returned.

You must redraw the list yourself.

(1.0)
*/
UInt16
FinderList_InsertRow		(FinderListRef					inRef,
							 SInt16 const					inInsertBeforeWhichZeroBasedIndexOrMinus1,
							 FinderListRowFlags const		inFlags)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	UInt16				result = 0;
	
	
	if (ptr != null)
	{
		FinderListRowPtr		rowDataPtr = (FinderListRowPtr)Memory_NewPtr(sizeof(FinderListRow));
		
		
		SetControlMaximum(ptr->controls.scrollBarV, GetControlMaximum(ptr->controls.scrollBarV) + 1);
		if (rowDataPtr != null)
		{
			ChainLinkRef		node = null;
			FinderListCellPtr	cellDataPtr = null;
			register SInt16		i = 0,
								cellCount = Chain_GetNodeCount(ptr->columnList);
			
			
			// initialize the new row, creating empty cells for each column in the list
			rowDataPtr->flags = inFlags;
			rowDataPtr->cellList = Chain_New(kChainStyleSinglyLinked/*kChainStyleVector*/);
			for (i = 0; i < cellCount; i++)
			{
				cellDataPtr = (FinderListCellPtr)Memory_NewPtr(sizeof(FinderListCell));
				if (cellDataPtr != null)
				{
					// initialize new cell
					cellDataPtr->dataPtr = null;
				}
				node = Chain_NewNode(rowDataPtr->cellList, cellDataPtr);
				Chain_MoveNode(rowDataPtr->cellList, node, null/* means end of list */,
									kChainLinkPlacementInsertAfter);
			}
			
			// add the data to the list of rows
			node = Chain_NewNode(ptr->rowList, rowDataPtr);
			Chain_MoveNode(ptr->rowList, node,
								(inInsertBeforeWhichZeroBasedIndexOrMinus1 < 0)
									? null
									: Chain_GetIndexedNode(ptr->rowList, inInsertBeforeWhichZeroBasedIndexOrMinus1),
								kChainLinkPlacementInsertBefore);
			
			// return the number of items now in the list
			result = Chain_GetNodeCount(ptr->rowList);
		}
		
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// InsertRow


/*!
Changes the view so the specified row is showing.
If you specify "inVerticallyCenterInView" to be
true, then the specified row is positioned
approximately in the middle (compared to the top
and bottom edges of the view area of the list);
otherwise, the specified row appears at the top
of the view area.

You must redraw the list yourself.

(1.0)
*/
void
FinderList_ScrollToRow		(FinderListRef			inRef,
							 SInt16 const			inZeroBasedRowIndex,
							 Boolean const			inVerticallyCenterInView)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		scrollToRow(ptr, inZeroBasedRowIndex, inVerticallyCenterInView);
		refReleaseLock(inRef, &ptr);
	}
}// ScrollToRow


/*!
To change the data in a cell after it has been
created, use this method.  If the data was
successfully changed, "true" is returned;
otherwise (if an error occurs) "false" is
returns.

You must redraw the list yourself.

(1.0)
*/
Boolean
FinderList_SetCellData		(FinderListRef		inRef,
							 SInt16 const		inZeroBasedRowIndex,
							 SInt16 const		inZeroBasedColumnIndex,
							 void const*		inCellDataPtr,
							 Size const			inDataSize)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	Boolean				result = false;
	
	
	if (ptr != null)
	{
		FinderListCellPtr	cellPtr = findCell(ptr, inZeroBasedRowIndex, inZeroBasedColumnIndex);
		
		
		if ((cellPtr != null) && (inCellDataPtr != null))
		{
			// COPY the given data into the cell
			if (cellPtr->dataPtr != null) DisposePtr(cellPtr->dataPtr);
			cellPtr->dataPtr = Memory_NewPtr(inDataSize);
			if ((MemError() == noErr) && (cellPtr->dataPtr != null))
			{
				BlockMove(inCellDataPtr, cellPtr->dataPtr, GetPtrSize(cellPtr->dataPtr));
				result = true;
			}
		}
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// SetCellData


/*!
To change the title text for the button or
placard representing a columns heading, use
this method.

You must redraw the list yourself.

(1.0)
*/
void
FinderList_SetColumnTitle	(FinderListRef		inRef,
							 SInt16 const		inZeroBasedColumnIndex,
							 StringPtr			inTitle)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ChainLinkRef			node = null;
		FinderListColumnPtr		columnDataPtr = null;
		
		
		node = Chain_GetIndexedNode(ptr->columnList, inZeroBasedColumnIndex);
		columnDataPtr = (FinderListColumnPtr)Chain_GetNodeDataPtr(ptr->columnList, node);
		if (columnDataPtr != null)
		{
			// set the control text to be the copy that was just created
			(OSErr)SetControlData(columnDataPtr->headerButton, kControlEditTextPart, kControlStaticTextTextTag,
									StrLength(inTitle) * sizeof(unsigned char), inTitle + 1/* skip length byte */);
		}
		refReleaseLock(inRef, &ptr);
	}
}// SetColumnTitle


/*!
To change the title text offset, in pixels, for
the button or placard representing a columns
heading, use this method.  You typically do this
to align the title text horizontally with the
contents of the column (for example, the Finder
offsets the Name title of a list view to make
room for a disclosure triangle and Finder icon).

You must redraw the list yourself.

(1.0)
*/
void
FinderList_SetColumnTitleOffset		(FinderListRef		inRef,
									 SInt16 const		inZeroBasedColumnIndex,
									 SInt16 const		inTitleOffsetInPixels)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ChainLinkRef			node = null;
		FinderListColumnPtr		columnDataPtr = null;
		
		
		node = Chain_GetIndexedNode(ptr->columnList, inZeroBasedColumnIndex);
		columnDataPtr = (FinderListColumnPtr)Chain_GetNodeDataPtr(ptr->columnList, node);
		if (columnDataPtr != null)
		{
			// set the control title offset
			(OSErr)SetControlData(columnDataPtr->headerButton, kControlButtonPart, kControlBevelButtonTextOffsetTag,
									sizeof(SInt16), (Ptr)&inTitleOffsetInPixels);
		}
		refReleaseLock(inRef, &ptr);
	}
}// SetColumnTitleOffset


/*!
To change the width of a column, in pixels, use
this method.  Note that a column is initially
created to be just wide enough for its title,
so you only need to use this routine if you want
to override the default width.

If column widths add up to be greater than the
width of the list, the list truncates the last
columns to fit.

You must redraw the list yourself.

(1.0)
*/
void
FinderList_SetColumnWidthInPixels	(FinderListRef		inRef,
									 SInt16 const		inZeroBasedColumnIndex,
									 UInt16 const		inWidthInPixels)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ChainLinkRef			node = null;
		FinderListColumnPtr		columnDataPtr = null;
		
		
		node = Chain_GetIndexedNode(ptr->columnList, inZeroBasedColumnIndex);
		columnDataPtr = (FinderListColumnPtr)Chain_GetNodeDataPtr(ptr->columnList, node);
		if (columnDataPtr != null)
		{
			columnDataPtr->width = inWidthInPixels;
		}
		refReleaseLock(inRef, &ptr);
	}
}// SetColumnWidthInPixels


/*!
To change the currently selected sort column,
use this method.  If you try to select any
column with "kFinderListColumnFlagsNotSortable"
set, this routine silently fails.

The rows are automatically sorted appropriately.

You must redraw the list yourself.

(1.0)
*/
void
FinderList_SetSortColumn	(FinderListRef		inRef,
							 SInt16 const		inZeroBasedColumnIndex)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		setSortColumn(ptr, inZeroBasedColumnIndex);
		refReleaseLock(inRef, &ptr);
	}
}// SetSortColumn


/*!
Changes the size of a Finder List by adjusting
appropriate controls so the entire control has
the specified absolute dimensions.

(1.0)
*/
void
FinderList_SizeControl		(FinderListRef			inRef,
							 SInt16 const			inWidth,
							 SInt16 const			inHeight)
{
	FinderListPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ControlRef		control = null;
		Rect			paneRect,
						placardRect;
		SInt32			scrollBarWidth = 0L;
		UInt16			rowTextHeight = 0;
		
		
		// calculate some metrics
		(OSStatus)getThemeMetric(kThemeMetricScrollBarWidth, &scrollBarWidth);
		rowTextHeight = Localization_GetSingleLineTextHeight(ptr->cellThemeFontID);
		
		// set the height of each row in the list (different rows with different heights are not currently supported)
		ptr->rowHeight = rowTextHeight + INTEGER_DOUBLED(kRowTextInsetV);
		
		// recalculate the entire controls encompassing rectangle, and its header rectangle
		control = ptr->controls.encompassingUserPane;
		SizeControl(control, inWidth, inHeight);
		GetControlBounds(ptr->controls.encompassingUserPane, &paneRect);
		
		control = ptr->controls.columnHeaderPlacard;
	#if USE_SORT_ORDER_REVERSAL_BUTTON
		SizeControl(control, paneRect.right - paneRect.left - scrollBarWidth + 1,
					rowTextHeight + INTEGER_DOUBLED(kHeaderTextInsetV));
	#else
		SizeControl(control, paneRect.right - paneRect.left,
					rowTextHeight + INTEGER_DOUBLED(kHeaderTextInsetV));
	#endif
		GetControlBounds(ptr->controls.columnHeaderPlacard, &placardRect);
		
		// adjust controls based on the specified size
	#if USE_SORT_ORDER_REVERSAL_BUTTON
		control = ptr->controls.sortOrderReversalButton;
		MoveControl(control, placardRect.right - 1, placardRect.top);
		SizeControl(control, scrollBarWidth, placardRect.bottom - placardRect.top);
	#endif
		control = ptr->controls.scrollBarV;
		MoveControl(control, paneRect.right - scrollBarWidth, placardRect.bottom - 1);
		if (ptr->flags & kFinderListFlagsSizeBoxSpace)
		{
			SizeControl(control, scrollBarWidth, paneRect.bottom - placardRect.bottom + 1 - scrollBarWidth + 1);
		}
		else
		{
			SizeControl(control, scrollBarWidth, paneRect.bottom - placardRect.bottom + 1);
		}
		
		// for each installed column, size each button appropriately
		{
			enum
			{
			#if USE_SORT_ORDER_REVERSAL_BUTTON
				kHorizontalTitlePadding = 16			// pixels of extra space in addition to title text width
			#else
				kHorizontalTitlePadding = 8 + 16 + 8	// pixels of extra space in addition to title text width;
														// 8 pixels on each side, plus room for a sort direction icon
			#endif
			};
			ChainLinkRef			node = null;
			FinderListColumnPtr		columnDataPtr = null;
			Rect					buttonRect;
			UInt16					leftEdge = 0,
									topEdge = 0,
									buttonSumWidth = 0,
									index = 0;
			Str255					buttonTitle;
			Boolean					lastButton = false;		// last column (and therefore it may be clipped)?
			
			
			// traverse once to determine the widths for each column, if necessary
			buttonSumWidth = 0;
			node = Chain_GetFirstNode(ptr->columnList);
			do
			{
				columnDataPtr = Chain_GetNodeDataPtr(ptr->columnList, node);
				
				// if the column width is zero, it should be calculated automatically
				// to be wide enough for the title text; otherwise, its left alone
				if (columnDataPtr != null)
				{
					if (columnDataPtr->width == 0)
					{
						control = columnDataPtr->headerButton;
						Localization_SetControlThemeFontInfo(control, ptr->headerThemeFontID);
						GetControlBounds(control, &buttonRect);
						GetControlTitle(control, buttonTitle);
						columnDataPtr->width = kHorizontalTitlePadding +
												Localization_GetSingleLineTextWidth(buttonTitle, ptr->headerThemeFontID);
					}
					buttonSumWidth += columnDataPtr->width;
				}
				node = Chain_GetNextNode(ptr->columnList, node);
			} while (node != null);
			
			// traverse again to check, and actually change, the button sizes and locations;
			// without support for horizontal scrolling, overflow must result in truncation
			leftEdge = placardRect.left;
			topEdge = placardRect.top;
			index = 0;
			node = Chain_GetFirstNode(ptr->columnList);
			do
			{
				columnDataPtr = Chain_GetNodeDataPtr(ptr->columnList, node);
				if (columnDataPtr != null)
				{
					control = columnDataPtr->headerButton;
					
					// if the list remains over-full, or if this is the last button, optimize this buttons width
					if ((buttonSumWidth > (placardRect.right - placardRect.left)) ||
						(index >= Chain_GetNodeCount(ptr->columnList)))
					{
						// unimplemented
					}
					MoveControl(control, leftEdge, topEdge);
					SizeControl(control, columnDataPtr->width, placardRect.bottom - placardRect.top);
					leftEdge += columnDataPtr->width;
				}
				node = Chain_GetNextNode(ptr->columnList, node);
				index++;
			} while (node != null);
		}
		control = ptr->controls.contentUserPane;
		MoveControl(control, paneRect.left + 1, placardRect.bottom);
		SizeControl(control, paneRect.right - paneRect.left - scrollBarWidth - 1, paneRect.bottom - placardRect.bottom - 1);
		
		// update scroll bars
		calculateViewSize(ptr);
		
		refReleaseLock(inRef, &ptr);
	}
}// SizeControl


//
// internal methods
//

/*!
If you ever change the dimensions of the list, use this
method to recalculate the view sizes horizontally and
vertically.  Under Mac OS 8.5 and beyond, this will
update the proportional scroll bars properly.

(1.0)
*/
static void
calculateViewSize	(FinderListPtr			inListPtr)
{
	if (inListPtr->haveCarbonControls)
	{
		Rect		contentBounds;
		
		
		GetControlBounds(inListPtr->controls.contentUserPane, &contentBounds);
		SetControlViewSize(inListPtr->controls.scrollBarV,
							(contentBounds.bottom - contentBounds.top) / inListPtr->rowHeight);
	}
}// calculateViewSize


/*!
This method begins a drag of a column in a Finder List.
The users mouse movements are tracked continuously, and
the heading buttons move out of the way as the user drags
to make room at valid drop-points.  If the user decides
to move the column, it is automatically relocated before
this method returns.

(1.0)
*/
static void
columnDrag	(FinderListPtr			inListPtr,
			 SInt16 const			inDraggingWhichColumnIndex)
{
	RgnHandle		grayRgn = null;
	
	
	// there must be enough memory for a drag region
	grayRgn = Memory_NewRegion();
	if ((MemError() == noErr) && (grayRgn != null))
	{
		enum
		{
			// arbitrary distance in each dimension; if the user moves
			// the mouse further than this distance from the drag limit
			// rectangle, the drag region is reset
			kSlopH = 32,
			kSlopV = 32
		};
		DragGrayRgnUPP			upp = null;
		FinderListColumnPtr		columnDataPtr = (FinderListColumnPtr)Chain_GetNodeDataPtr
																	(inListPtr->columnList, Chain_GetIndexedNode
																	(inListPtr->columnList, inDraggingWhichColumnIndex));
		SInt16					oldCursor = Cursors_UseHandGrab();
		Rect					columnRect,
								buttonRect,
								limitRect,
								slopRect;
		Point					localMouse;
		UInt32					dragResult = 0L;
		
		
		// define the drag outline
		GetControlBounds(inListPtr->controls.contentUserPane, &limitRect);
		gDragDataContentRect = limitRect;
		GetControlBounds(columnDataPtr->headerButton, &buttonRect);
		limitRect.top = buttonRect.top;
		columnRect = limitRect;
		columnRect.left = buttonRect.left;
		columnRect.right = buttonRect.right;
		slopRect = limitRect;
		InsetRect(&slopRect, -kSlopH, -kSlopV);
		OpenRgn();
		FrameRect(&columnRect);
		InsetRect(&buttonRect, 2, 2);
		FrameRect(&buttonRect);
		CloseRgn(grayRgn);
		
		// drag an outline of the column while the user holds the mouse button down
		if (inListPtr->haveAppearance1_1) (OSStatus)BeginThemeDragSound(kThemeDragSoundDragging);
		GetMouse(&localMouse);
		gDragDataHeaderButtonCount = newHeadingButtonsArray(inListPtr, &gDragDataHeaderButtons);
		upp = NewDragGrayRgnUPP(columnDragProgress);
		dragResult = DragGrayRgn(grayRgn, localMouse, &limitRect, &slopRect, hAxisOnly, upp);
		columnDragInsertionDraw(&gDragDataInsertionRect, false);
		SetRect(&gDragDataInsertionRect, -20, -20, -20, -20);
		DisposePtr((Ptr)gDragDataHeaderButtons), gDragDataHeaderButtons = null;
		Memory_DisposeRegion(&grayRgn);
		DisposeDragGrayRgnUPP(upp), upp = null;
		if (inListPtr->haveAppearance1_1) (OSStatus)EndThemeDragSound();
		Cursors_Use(oldCursor);
		
		unless (dragResult == kMouseUpOutOfSlop)
		{
			// if the user moved the column, update the Finder List
			SInt16		deltaX = dragResult & 0x00FF,
						deltaY = (dragResult >> 16) & 0x00FF;
			
			
			// incomplete
			if (inListPtr->haveAppearance1_1) (OSStatus)PlayThemeSound(kThemeSoundDragTargetDrop);
		}
	}
	else BEEP;
}// columnDrag


/*!
To draw or erase a shape representing where a column
that is being dragged by the user will be inserted,
use this method.  The current implementation draws
a two-pixel-thick border in a checkerboard gray pattern.

(1.0)
*/
static void
columnDragInsertionDraw		(Rect const*		inRectPtr,
							 Boolean const		UNUSED_ARGUMENT(inDraw))
{
	PenState		state;
	
	
	GetPenState(&state);
	PenNormal();
	PenSize(2, 2);
	PenMode(srcXor); // XOR will either draw or erase, so "inDraw" is currently ignored
	ColorUtilities_SetGrayPenPattern();
	FrameRect(inRectPtr);
	SetPenState(&state);
}// columnDragInsertionDraw


/*!
This method is invoked continuously as the user
drags a column in a Finder List.  It provides
visual feedback whenever the user has dragged a
column over a valid drop point.

This routine relies on global variables being
set up from within the columnDrag() function.

Incomplete.

(1.0)
*/
static pascal void
columnDragProgress	(void)
{
	if (gDragDataHeaderButtonCount > 1)
	{
		enum
		{
			kInsertSlopOutsetInPixels = 5,			// arbitrary
			kInsertHighlightInsetInPixels = 6		// arbitrary - should be larger than the above
		};
		Rect				buttonRect;
		Point				localMouse;
		register SInt16		i = 0;
		static SInt16		gOldInsertionIndex = -1;
		
		
		// determine if the mouse is approximately between any two buttons
		GetMouse(&localMouse);
		for (i = 0; i < gDragDataHeaderButtonCount; i++)
		{
			GetControlBounds(gDragDataHeaderButtons[i], &buttonRect);
			
			// create a small rectangle representing the insertion area between buttons
			SetRect(&buttonRect,
					buttonRect.left - kInsertSlopOutsetInPixels, buttonRect.top - kInsertSlopOutsetInPixels,
					buttonRect.left + kInsertSlopOutsetInPixels, gDragDataContentRect.bottom + kInsertSlopOutsetInPixels);
			if (PtInRect(localMouse, &buttonRect)) if (i != gOldInsertionIndex)
			{
				gOldInsertionIndex = i;
				
				// draw a highlight to show where the button will be inserted (erasing any previous highlight)
				columnDragInsertionDraw(&gDragDataInsertionRect, false);
				columnDragInsertionDraw(&buttonRect, false);
				gDragDataInsertionRect = buttonRect;
				break;
			}
		}
	}
}// columnDragProgress


/*!
A standard FinderListCellCompareProcPtr, this
method compares two cells that contain character
data, and returns -1 if the first cells data is
less than the second, 1 if the reverse is true,
and 0 if the two cells appear to contain the same
data.

IMPORTANT:	The two pointers must be Mac OS
			pointers, so that this routine can
			use GetPtrSize() to find out how
			large the data is.

(1.0)
*/
static int
compareTextCellData		(void const*		inCellData1,
						 void const*		inCellData2)
{
	int		result = 0;
	
	
	{
		StringPtr	text1 = (StringPtr)Memory_NewPtr(GetPtrSize((Ptr)inCellData1)),
					text2 = (StringPtr)Memory_NewPtr(GetPtrSize((Ptr)inCellData2));
		
		
		text1[0] = GetPtrSize((Ptr)text1);
		text2[0] = GetPtrSize((Ptr)text2);
		BlockMove(inCellData1, &text1[1], text1[0] * sizeof(char));
		BlockMove(inCellData2, &text2[1], text2[0] * sizeof(char));
		result = PLstrcmp(text1, text2);
	}
	return result;
}// compareTextCellData


/*!
A standard ControlUserPaneActivateUPP, this
method activates or deactivates an editable
text area if a list item is being edited.

Unimplemented.

(1.0)
*/
static pascal void
contentUserPaneActivate		(ControlRef		UNUSED_ARGUMENT(inControl),
							 Boolean		UNUSED_ARGUMENT(inActivating))
{
	//if (something is being edited)
	//{
	//	if (inActivating) TEActivate(ptr->textEditHandle);
	//	else TEDeactivate(ptr->textEditHandle);
	//}
	//if (something is selected)
	//{
	//	change selection to outline
	//}
}// contentUserPaneActivate


/*!
A standard ControlUserPaneDrawUPP, this
method draws the currently visible part
of the list.

(1.0)
*/
static pascal void
contentUserPaneDraw		(ControlRef			inControl,
						 SInt16				UNUSED_ARGUMENT(inPartCode))
{
	DeviceLoopDrawingUPP		upp = NewDeviceLoopDrawingUPP(contentUserPaneDeviceLoop);
	RgnHandle					oldClipRgn = Memory_NewRegion(),
								newClipRgn = Memory_NewRegion();
	
	
	if (oldClipRgn != null) GetClip(oldClipRgn);
	if (newClipRgn != null)
	{
		{
			Rect		contentRect;
			
			
			GetControlBounds(inControl, &contentRect);
			RectRgn(newClipRgn, &contentRect);
			SetClip(newClipRgn); // ??? is this necessary? ???
		}
		DeviceLoop(newClipRgn, upp, (long)inControl/* user data required by the DeviceLoopDrawingUPP */,
					0/* DeviceLoopFlags */);
		Memory_DisposeRegion(&newClipRgn);
	}
	DisposeDeviceLoopDrawingUPP(upp), upp = null;
	if (oldClipRgn != null)
	{
		SetClip(oldClipRgn);
		Memory_DisposeRegion(&oldClipRgn);
	}
}// contentUserPaneDraw


/*!
This method, of standard DeviceLoopDrawingUPP form,
handles drawing of Finder Lists correctly no matter
what the device characteristics are.

Incomplete.

(1.0)
*/
static pascal void
contentUserPaneDeviceLoop	(short				inColorDepth,
							 short				UNUSED_ARGUMENT(inDeviceFlags),
							 GDHandle			inTargetDevice,
							 long				inControl)
{
	FinderListPtr		ptr = refAcquireLock((FinderListRef)GetControlReference((ControlRef)inControl));
	
	
	if (ptr != null)
	{
		WindowRef		window = GetControlOwner((ControlRef)inControl);
		ColorPenState	state;
		Rect			controlRect;
		Boolean			isColorDevice = ColorUtilities_IsColorDevice(inTargetDevice);
		
		
		// save state
		ColorUtilities_GetColorAndPenState(&state);
		ColorUtilities_NormalizeColorAndPen();
		
		// set up the control background and foreground colors
		(OSErr)SetUpControlBackground((ControlRef)inControl, inColorDepth, isColorDevice);
		
		// draw list background
		GetControlBounds((ControlRef)inControl, &controlRect);
		(OSStatus)SetThemeBackground(kThemeBrushListViewBackground, inColorDepth, isColorDevice);
		EraseRect(&controlRect);
		if (ptr->selectedColumnIndex >= 0)
		{
			ChainLinkRef			node = null;
			FinderListColumnPtr		columnDataPtr = null;
			
			
			node = Chain_GetIndexedNode(ptr->columnList, ptr->selectedColumnIndex);
			columnDataPtr = (FinderListColumnPtr)Chain_GetNodeDataPtr(ptr->columnList, node);
			if (columnDataPtr != null)
			{
				Rect		buttonRect,
							columnRect;
				
				
				(OSStatus)SetThemeBackground(columnDataPtr->backgroundBrush, inColorDepth, isColorDevice);
				GetControlBounds(columnDataPtr->headerButton, &buttonRect);
				if (buttonRect.left <= controlRect.left) buttonRect.left++; // dont draw on the bounding box edges
				if (buttonRect.right >= controlRect.right) buttonRect.right--; // dont draw on the bounding box edges
				SetRect(&columnRect, buttonRect.left, controlRect.top, buttonRect.right, controlRect.bottom);
				EraseRect(&columnRect);
			}
		}
		
		// draw items
		//if (Chain_GetNodeCount(ptr->rowList) > 0)
		{
			enum
			{
				kRowSeparatorHeight = 1		// number of pixels high the list row separator is
			};
			ChainLinkRef			columnNode = null,
									rowNode = null,
									cellNode = null;
			FinderListColumnPtr		columnDataPtr = null;
			FinderListRowPtr		rowDataPtr = null;
			FinderListCellPtr		cellDataPtr = null;
			Rect					cellRect;
			SInt16					cellLeftEdge = controlRect.left,
									cellTopEdge = controlRect.top + INTEGER_HALVED(kRowTextInsetV);
			SInt16					rowIndex = 0,
									rowIndexLimit = 0;
			Boolean					done = false;
			
			
			// draw only those cells that are actually going to be visible!
			rowIndex = getFirstVisibleRow(ptr);
			if (ptr->sortDescending) rowIndexLimit = rowIndex - getVisibleRowCount(ptr);
			else rowIndexLimit = rowIndex + getVisibleRowCount(ptr);
			
			rowNode = Chain_GetIndexedNode(ptr->rowList, rowIndex);
			do
			{
				rowDataPtr = Chain_GetNodeDataPtr(ptr->rowList, rowNode);
				if (rowDataPtr != null)
				{
					cellNode = Chain_GetFirstNode(rowDataPtr->cellList);
					columnNode = Chain_GetFirstNode(ptr->columnList);
					do
					{
						cellDataPtr = Chain_GetNodeDataPtr(rowDataPtr->cellList, cellNode);
						columnDataPtr = Chain_GetNodeDataPtr(ptr->columnList, columnNode);
						if ((cellDataPtr != null) && (columnDataPtr != null))
						{
							// calculate the cell rectangle and draw its contents
							SetRect(&cellRect, cellLeftEdge, cellTopEdge,
												cellLeftEdge + columnDataPtr->width - 1, cellTopEdge + ptr->rowHeight - 1);
							drawCell(ptr, columnDataPtr, rowDataPtr, cellDataPtr, inColorDepth, inTargetDevice, &cellRect);
							
							// prepare for next cell
							cellLeftEdge += columnDataPtr->width;
						}
						cellNode = Chain_GetNextNode(rowDataPtr->cellList, cellNode);
						columnNode = Chain_GetNextNode(ptr->columnList, columnNode);
					} while (columnNode != null);
				}
				
				// draw row separator
				cellTopEdge += (ptr->rowHeight + 1);
				SetRect(&controlRect, controlRect.left, cellTopEdge, controlRect.right, cellTopEdge + kRowSeparatorHeight);
				cellTopEdge += kRowSeparatorHeight;
				ColorUtilities_NormalizeColorAndPen();
				(OSStatus)SetThemeBackground(kThemeBrushListViewSeparator, inColorDepth, isColorDevice);
				EraseRect(&controlRect);
				
				// determine if all visible rows have been drawn; if any rows
				// still need to be drawn, reset the left edge and get next row
				if (ptr->sortDescending)
				{
					rowIndex--;
					done = (rowIndex < rowIndexLimit);
					if (!done) rowNode = Chain_GetPreviousNode(ptr->rowList, rowNode);
					if (rowNode == null) done = true;
				}
				else
				{
					rowIndex++;
					done = (rowIndex > rowIndexLimit);
					if (!done) rowNode = Chain_GetNextNode(ptr->rowList, rowNode);
					if (rowNode == null) done = true;
				}
				cellLeftEdge = controlRect.left;
			} while (!done);
		}
		
		// restore state
		ColorUtilities_SetColorAndPenState(&state);
		refReleaseLock((FinderListRef)GetControlReference((ControlRef)inControl), &ptr);
	}
}// contentUserPaneDeviceLoop


/*!
This method, of standard ControlUserPaneIdleUPP form,
handles cursor flashing if a Finder List item is
being edited.

Unimplemented.

(1.0)
*/
static pascal void
contentUserPaneIdle			(ControlRef			inControl)
{
	FinderListPtr		ptr = refAcquireLock((FinderListRef)GetControlReference(inControl));
	
	
	if (ptr != null)
	{
		//if (something is being edited)
		//{
		//	TEIdle(ptr->textEditHandle);
		//}
		refReleaseLock((FinderListRef)GetControlReference(inControl), &ptr);
	}
}// contentUserPaneIdle


/*!
Paints the specified list cell in the current
graphics port, within the specified local
rectangle.  The cell is painted in its proper
state.

Incomplete.

(1.0)
*/
static void
drawCell		(FinderListPtr			inListPtr,
				 FinderListColumnPtr	inColumnPtr,
				 FinderListRowPtr		inRowPtr,
				 FinderListCellPtr		inCellPtr,
				 short					inColorDepth,
				 GDHandle				inTargetDevice,
				 Rect const*			inBoundsPtr)
{
	// only text data is currently supported
	if (inColumnPtr->dataType == kFinderListColumnDataTypePlainText)
	{
		GrafPortFontState	fontState;
		ColorPenState		colorPenState;
		Point				preservedPenLocation;
		Rect				textRect;
		FontInfo			theInfo;
		Boolean				dimInactiveItems = (inListPtr->flags & kFinderListFlagsInactiveItemsDimmed),
							cellInactive = ((inRowPtr->flags & kFinderListRowFlagsDisabled) ||
											(!IsControlActive(inListPtr->controls.encompassingUserPane))),
							cellSelected = ((inRowPtr->flags & kFinderListRowFlagsSelected) &&
											(!(inColumnPtr->flags & kFinderListColumnFlagsNoHighlighting)));
		
		
		// save port settings
		GetPen(&preservedPenLocation);
		Localization_GetPortFontState(&fontState);
		ColorUtilities_GetColorAndPenState(&colorPenState);
		
		// set up font
		{
			Str255		fontName;
			SInt16		fontSize = 0;
			Style		fontStyle = normal;
			
			
			Localization_UseThemeFont(inListPtr->cellThemeFontID, fontName, &fontSize, &fontStyle);
		}
		GetFontInfo(&theInfo);
		
		// specify the smallest rectangle containing the text itself
		SetRect(&textRect, inBoundsPtr->left + kHeaderTextInsetV, inBoundsPtr->top + kRowTextInsetV, 0, 0);
		SetRect(&textRect, textRect.left, textRect.top,
				textRect.left + TextWidth(inCellPtr->dataPtr, 0/* first byte */, GetPtrSize(inCellPtr->dataPtr)) - 1,
				textRect.top + INTEGER_DOUBLED(theInfo.leading) + theInfo.ascent + theInfo.descent - 1);
		
		// determine proper background color and paint cell background
		ColorUtilities_NormalizeColorAndPen();
		(OSStatus)SetThemeBackground(inColumnPtr->backgroundBrush, inColorDepth,
										ColorUtilities_IsColorDevice(inTargetDevice));
		EraseRect(inBoundsPtr);
		
		// set up font color and paint cell background
		if (cellSelected)
		{
			RGBColor		color;
			
			
			unless (cellInactive)
			{
				// Finder-style highlighting; the text color for selected icons is
				// white-on-black, in absence of a better theme color to use
				ColorUtilities_NormalizeColorAndPen();
				color.red = color.green = color.blue = 0;
				RGBBackColor(&color);
				EraseRect(&textRect);
			}
			
			// outline the text; for active-window highlighted text, this makes the
			// highlighting look better; for inactive-window text, this causes ONLY
			// the outline to be drawn, which is consistent with Mac OS highlighting
			color.red = color.green = color.blue = 0;
			RGBForeColor(&color);
			InsetRect(&textRect, -1, -1);
			FrameRect(&textRect);
			InsetRect(&textRect, 1, 1);
			
			unless (cellInactive)
			{
				// Finder-style highlighting; the text color for selected icons is
				// white-on-black, in absence of a better theme color to use
				color.red = color.green = color.blue = RGBCOLOR_INTENSITY_MAX;
				RGBForeColor(&color);
			}
		}
		else 
		{
			if ((cellInactive) && (dimInactiveItems))
			{
				// the text color to use for this isnt clear, but modeless dialog inactive text seems closest
				(OSStatus)SetThemeTextColor(kThemeTextColorModelessDialogInactive,
											inColorDepth, ColorUtilities_IsColorDevice(inTargetDevice));
			}
			else
			{
				// the text color for active list views is obvious, its a theme color (use it!)
				(OSStatus)SetThemeTextColor(kThemeTextColorListView,
											inColorDepth, ColorUtilities_IsColorDevice(inTargetDevice));
			}
			TextMode(transparent);
		}
		
		// draw the text on top of its background
		MoveTo(textRect.left, textRect.top + theInfo.leading + theInfo.ascent);
		DrawText(inCellPtr->dataPtr, 0/* first byte */, GetPtrSize(inCellPtr->dataPtr));
		
		// restore port settings
		ColorUtilities_SetColorAndPenState(&colorPenState);
		Localization_SetPortFontState(&fontState);
		MoveTo(preservedPenLocation.h, preservedPenLocation.v);
	}
}// drawCell


/*!
A standard ControlUserPaneDrawUPP, this
method draws a Finder Lists frame
(transparently, so the content isnt
covered).

(1.0)
*/
static pascal void
encompassingUserPaneDraw	(ControlRef			inControl,
							 SInt16				UNUSED_ARGUMENT(inPartCode))
{
	ControlRef			focusControl = null;
	ColorPenState		state;
	Rect				controlRect;
	FinderListPtr		ptr = refAcquireLock((FinderListRef)GetControlReference(inControl));
	
	
	ColorUtilities_GetColorAndPenState(&state);
	ColorUtilities_NormalizeColorAndPen();
	
	GetControlBounds((ControlRef)inControl, &controlRect);
	
	// frame the list
	{
		RGBColor		black;
		GrafPtr			port = null;
		
		
		GetPort(&port);
		black.red = black.green = black.blue = 0;
		RGBForeColor(&black);
		(OSStatus)SetThemeTextColor(IsControlActive(inControl) ?
								kThemeTextColorModelessDialogActive : kThemeTextColorModelessDialogInactive,
								ColorUtilities_GetCurrentDepth((CGrafPtr)port),
								ColorUtilities_IsColorDevice(GetMainDevice()));
		FrameRect(&controlRect);
	}
	
	if (ptr != null)
	{
		unless (ptr->flags & kFinderListFlagsNoFocusRing)
		{
			// now draw the appropriate outer edges: a bevel or a focus ring
			(OSErr)GetKeyboardFocus(GetControlOwner(inControl), &focusControl);
			if ((inControl == focusControl) && (IsControlActive(inControl)))
			{
				// the Finder List is focused, so draw a focus ring instead of a frame
				InsetRect(&controlRect, 1, 1);
				(OSStatus)DrawThemeFocusRect(&controlRect, true/* focused */);
			}
			else
			{
				// draw a standard bevel instead of a focus ring
				(OSStatus)DrawThemeFocusRect(&controlRect, false/* focused */);
				(OSStatus)DrawThemeGenericWell(&controlRect, IsControlHilited(inControl)
															? kThemeStateActive
															: kThemeStateInactive, false/* fill center */);
			}
		}
		refReleaseLock((FinderListRef)GetControlReference(inControl), &ptr);
	}
	
	ColorUtilities_SetColorAndPenState(&state);
}// encompassingUserPaneDraw


/*!
This method, of standard ControlUserPaneFocusUPP form,
adds or removes a focus ring from a Finder List control.

(1.0)
*/
static pascal ControlPartCode
encompassingUserPaneFocus	(ControlRef			inControl,
							 ControlFocusPart	inAction)
{
	ControlPartCode		result = (inAction == kControlFocusNoPart) ? kControlNoPart : kControlListBoxPart;
	FinderListPtr		ptr = refAcquireLock((FinderListRef)GetControlReference(inControl));
	
	
	if (ptr != null)
	{
		Rect		controlRect;
		
		
		// draw a focus ring when focused
		unless (ptr->flags & kFinderListFlagsNoFocusRing)
		{
			GetControlBounds(inControl, &controlRect);
			InsetRect(&controlRect, 1, 1);
			(OSStatus)DrawThemeFocusRect(&controlRect, (inAction != kControlFocusNoPart));
		}
		refReleaseLock((FinderListRef)GetControlReference(inControl), &ptr);
	}
	return result;
}// encompassingUserPaneFocus


/*!
This method, of standard ControlUserPaneHitTestUPP form,
handles clicks anywhere inside a Finder List.

(1.0)
*/
static pascal ControlPartCode
encompassingUserPaneHitTest		(ControlRef		inControl,
								 Point			inLocalMouse)
{
	ControlPartCode		result = kControlNoPart;
	FinderListPtr		ptr = refAcquireLock((FinderListRef)GetControlReference(inControl));
	
	
	if (ptr != null)
	{
		ControlRef		control = null;
		
		
		// re-initialize to null at each new hit test
		ptr->subControlHit = null;
		
		// check for hits in the content area of the control
		if (ptr->subControlHit == null)
		{
			Rect		controlRect;
			
			
			control = ptr->controls.contentUserPane;
			GetControlBounds(control, &controlRect);
			if (PtInRect(inLocalMouse, &controlRect))
			{
				result = kControlInactivePart;
				ptr->subControlHit = control;
			}
		}
		
		// check to see if any heading buttons were hit
		if (ptr->subControlHit == null)
		{
			ControlRef*			headingButtons = null;
			register SInt16		i = 0,
								headingButtonsArraySize = 0;
			
			
			headingButtonsArraySize = newHeadingButtonsArray(ptr, &headingButtons);
			for (i = 0; ((i < headingButtonsArraySize) && (ptr->subControlHit == null)); i++)
			{
				control = headingButtons[i];
				result = TestControl(control, inLocalMouse);
				if (result == kControlButtonPart)
				{
					ptr->subControlHit = control;
					if (ptr->haveAppearance1_1) (OSStatus)PlayThemeSound(kThemeSoundBevelPress);
				}
			}
			DisposePtr((Ptr)headingButtons), headingButtons = null;
		}
		
		// check to see if the sort order reversal button was hit
	#if USE_SORT_ORDER_REVERSAL_BUTTON
		if (ptr->subControlHit == null)
		{
			control = ptr->controls.sortOrderReversalButton;
			result = TestControl(control, inLocalMouse);
			if (result == kControlButtonPart)
			{
				ptr->subControlHit = control;
				if (ptr->haveAppearance1_1) (OSStatus)PlayThemeSound(kThemeSoundBevelPress);
			}
		}
	#endif
		
		refReleaseLock((FinderListRef)GetControlReference(inControl), &ptr);
	}
	
	return result;
}// encompassingUserPaneHitTest


/*!
This method, of standard ControlUserPaneKeyDownUPP form,
handles key presses.  If a list item is being edited,
the keys will be typed; otherwise, they will be used
to affect the current list selection.

Incomplete.

(1.0)
*/
static pascal ControlPartCode
encompassingUserPaneKeyDown		(ControlRef		inControl,
								 SInt16			inKeyCode,
								 SInt16			UNUSED_ARGUMENT(inCharCode),
								 SInt16			UNUSED_ARGUMENT(inModifiers))
{
	ControlPartCode		result = kControlNoPart;
	FinderListPtr		ptr = refAcquireLock((FinderListRef)GetControlReference(inControl));
	
	
	if (ptr != null)
	{
		//if (something is being edited)
		//{
		//	TEKey(inCharCode, ptr->textEditHandle);
		//}
		//else
		//{
			switch (inKeyCode)
			{
				case 0x7E: // Up arrow
				case 0x3E: // Up arrow on non-extended keyboards
					// If the user presses an up arrow key, select the previous enabled
					// item (in editable lists) or scroll downward (in read-only lists).
					scrollDownward(ptr, 1/* number of rows */, true/* update list */);
					break;
				
				case 0x7D: // Down arrow
				case 0x3D: // Down arrow on non-extended keyboards
					// If the user presses a down arrow key, select the next enabled
					// item (in editable lists) or scroll upward (in read-only lists).
					scrollUpward(ptr, 1/* number of rows */, true/* update list */);
					break;
				
				case 0x74: // Page Up (extended keyboards)
					// If the user presses the Page Up arrow key, scroll downward
					// by the number of visible rows.
					scrollDownward(ptr, getVisibleRowCount(ptr)/* number of rows */, true/* update list */);
					break;
				
				case 0x79: // Page Down (extended keyboards)
					// If the user presses the Page Down arrow key, scroll upward
					// by the number of visible rows.
					scrollUpward(ptr, getVisibleRowCount(ptr)/* number of rows */, true/* update list */);
					break;
				
				default:
					break;
			}
		//}
		refReleaseLock((FinderListRef)GetControlReference(inControl), &ptr);
	}
	return result;
}// encompassingUserPaneKeyDown


/*!
Determines the proper control part corresponding
to the specified mouse location, for a Finder List.
This routine continues to loop until the mouse is
released.  Since a Finder List maintains all of its
own controls, this routine always returns
"kControlNoPart".

The control action routine is ignored (but for
future reference, valid values are null (0), -1,
or a valid function pointer!).

(1.0)
*/
static pascal ControlPartCode
encompassingUserPaneTrackingProc	(ControlRef			inControl,
									 Point				UNUSED_ARGUMENT(inGlobalMouse),
									 ControlActionUPP	UNUSED_ARGUMENT(inIgnoredActionUPP))
{
	ControlPartCode		result = kControlNoPart;
	FinderListRef		ref = (FinderListRef)GetControlReference(inControl);
	FinderListPtr		ptr = null;
	
	
	ptr = refAcquireLock(ref);
	if (ptr != null)
	{
		enum
		{
			kDragDelay = 30,	// arbitrary; in ticks, length of time user must pause on mouse-down
								//   before a click on a heading button turns into a column drag, or
								//   before a click on a list item turns into an icon drag
			kRepeatDelay = 10	// arbitrary; in ticks, length of time user must remain in a scroll
								//   arrow button before the control value will be changed again
		};
		ControlPartCode		part = kControlNoPart,
							previousPart = kControlNoPart;
		Point				localMouse;
		Boolean				hitHeadingButton = false,
							redrawContent = false,
							done = false,
							firstTime = true; // used to prevent control entered sounds from playing the first time
		UInt32				clickTime = TickCount();
		register SInt16		i = 0,
							headingButtonsArraySize = 0;
		ControlRef*			headingButtons = null;
		Rect				bounds;
		
		
		// traverse the list of column buttons once, and create an array for efficiency in the loop
		headingButtonsArraySize = newHeadingButtonsArray(ptr, &headingButtons);
		
		while ((!done) && StillDown())
		{
			(OSErr)YieldToAnyThread();
			
			GetMouse(&localMouse);
			
			// dont waste time if the mouse is not even in a useful control part
			GetControlBounds(ptr->controls.contentUserPane, &bounds);
			if (PtInRect(localMouse, &bounds))
			{
				// hit in the content area of the control
				//if (ptr->haveAppearance1_1) (OSStatus)PlayThemeSound(kThemeSoundSelectItem);
				
				//if (isCommandKeyDown())
				//{
				//	scroll in place with hand cursor
				//}
				//if (inside item)
				//{
				//	// if tracking lasts for more than a second, start a drag (unimplemented)
					#if 0
					if ((TickCount() - clickTime) > kDragDelay)
					{
						// begin drag, if possible - unimplemented
					}
					#endif
				//	else
				//	{
				//		change selection (use modifiers to see how it changes)
				//	}
				//}
				//else
				//{
					// drag multiple-selection rectangle
					//BEEP;
					selectionRectangleDrag(ptr);
				//}
				part = kControlInactivePart;
			}
			else if (TestControl(ptr->subControlHit, localMouse) != kControlNoPart)
			{
			#if USE_SORT_ORDER_REVERSAL_BUTTON
				if (ptr->subControlHit == ptr->controls.sortOrderReversalButton)
				{
					if (part != kControlButtonPart)
					{
						unless (firstTime) if (ptr->haveAppearance1_1) (OSStatus)PlayThemeSound(kThemeSoundBevelEnter);
					}
					part = kControlButtonPart;
				}
				else
			#endif
				{
					Boolean		found = false;
					
					
					// a heading button was hit
					for (i = 0; ((i < headingButtonsArraySize) && (!found)); i++)
					{
						if (ptr->subControlHit == headingButtons[i])
						{
							found = true;
							if (part != kControlButtonPart)
							{
								unless (firstTime)
								{
									if (ptr->haveAppearance1_1) (OSStatus)PlayThemeSound(kThemeSoundBevelEnter);
								}
							}
							part = kControlButtonPart;
							hitHeadingButton = true;
							
							// if tracking lasts for more than a second, start a heading column drag (unimplemented)
							if ((TickCount() - clickTime) > kDragDelay)
							{
								done = true; // stop after beginning the drag
								columnDrag(ptr, i);
							}
						}
						else hitHeadingButton = false;
					}
				}
			}
			else
			{
				// play the bevel-button-exit sound if the last entered control was a button
				if (part == kControlButtonPart)
				{
					if (ptr->haveAppearance1_1) (OSStatus)PlayThemeSound(kThemeSoundBevelExit);
				}
				part = kControlNoPart;
				
				// reset the click time - the drag delay is accounted for
				// only while the mouse is located in the same region
				clickTime = 0L;
			}
			
			// redraw control if necessary, to show its new state
			if ((previousPart != part) && (ptr->subControlHit != ptr->controls.contentUserPane))
			{
				HiliteControl(ptr->subControlHit, part);
			}
			
			previousPart = part;
			firstTime = false;
		}
		
		// if the inside of some control part was hit, do clean-up tasks
		// and otherwise respond to the control click where appropriate
		if (part != kControlNoPart)
		{
			if (hitHeadingButton)
			{
				// play the bevel-button-release sound
				if (ptr->haveAppearance1_1) (OSStatus)PlayThemeSound(kThemeSoundBevelRelease);
				
				// ensure that only one heading column is highlighted
				for (i = 0; i < headingButtonsArraySize; i++)
				{
					if (headingButtons[i] == ptr->subControlHit)
					{
						if (i == ptr->selectedColumnIndex)
						{
						#if !USE_SORT_ORDER_REVERSAL_BUTTON
							// this buttons already selected; reverse the sort instead
							reverseSort(ptr);
						#endif
						}
						else
						{
							setSortColumn(ptr, i);
						}
						redrawContent = true;
						break;
					}
				}
			}
		#if USE_SORT_ORDER_REVERSAL_BUTTON
			else if (ptr->subControlHit == ptr->controls.sortOrderReversalButton)
			{
				// play the bevel-button-release sound
				if (ptr->haveAppearance1_1) (OSStatus)PlayThemeSound(kThemeSoundBevelRelease);
				
				// react appropriately to the button press
				reverseSort(ptr);
				redrawContent = true;
			}
		#endif
		}
		
		DisposePtr((Ptr)headingButtons), headingButtons = null;
		
		// if anything except the list content region was clicked, remove highlighting (which redraws the control)
		if (ptr->subControlHit != ptr->controls.contentUserPane) HiliteControl(ptr->subControlHit, kControlNoPart);
		if (redrawContent) Embedding_OffscreenDrawOneControl(ptr->controls.contentUserPane);
		
		refReleaseLock(ref, &ptr);
	}
	
	return result;
}// encompassingUserPaneTrackingProc


/*!
To obtain a pointer to the internal cell structure
corresponding to the specified row and column index
(zero-based) of a Finder List, use this method.

(1.0)
*/
static FinderListCellPtr
findCell		(FinderListPtr		inListPtr,
				 SInt16 const		inZeroBasedRowIndex,
				 SInt16 const		inZeroBasedColumnIndex)
{
	FinderListCellPtr		result = null;
	
	
	if (inListPtr != null)
	{
		ChainLinkRef		node = Chain_GetIndexedNode(inListPtr->rowList, inZeroBasedRowIndex);
		
		
		if (node != null)
		{
			FinderListRowPtr		rowDataPtr = (FinderListRowPtr)Chain_GetNodeDataPtr(inListPtr->rowList, node);
			
			
			if (rowDataPtr != null)
			{
				ChainLinkRef		cellNode = Chain_GetIndexedNode(rowDataPtr->cellList, inZeroBasedColumnIndex);
				
				
				result = (FinderListCellPtr)Chain_GetNodeDataPtr(rowDataPtr->cellList, cellNode);
			}
		}
	}
	return result;
}// findCell


/*!
Returns the index of the topmost row that can
be seen within the current view rectangle of
the specified Finder List.  If the list is
empty, -1 is returned.

(1.0)
*/
static SInt16
getFirstVisibleRow		(FinderListPtr		inListPtr)
{
	SInt16		result = -1;
	UInt16		rowCount = Chain_GetNodeCount(inListPtr->rowList);
	
	
	if (rowCount > 0)
	{
		result = GetControlValue(inListPtr->controls.scrollBarV);
		if (inListPtr->sortDescending) result = (rowCount - 1 - result);
	}
	
	return result;
}// getFirstVisibleRow


/*!
This is an implementation of what will eventually
be a Carbon API, hence the TARGET_API_MAC_OS8
preprocessor directive.  The internal version of
this method will only be used when there is none
defined by Appearance with an identical function
signature.  Since this API is present in Carbon,
it is logical to assume that a future Appearance
control panel will allow some of these metrics to
be user-specified, and others to be specified by
themes (but dynamic nonetheless).  To prepare for
the future dynamics of theme metrics without
crippling legacy code, this provides the code
for this routine whenever Carbon is not linked.

See "Appearance.h" for valid metric types.

(1.0)
*/
static OSStatus
getThemeMetric		(ThemeMetric 	inMetric,
					 SInt32*		outMetric)
{
	OSStatus		result = noErr;
	
	
#if TARGET_API_MAC_OS8
	if (outMetric != null) switch (inMetric)
	{
		case kThemeMetricScrollBarWidth:
			*outMetric = 16; // pixels
			break;
		
		case kThemeMetricSmallScrollBarWidth:
			*outMetric = 13; // pixels
			break;
		
		case kThemeMetricCheckBoxGlyphHeight:
			*outMetric = 12; // pixels
			break;
		
		case kThemeMetricRadioButtonGlyphHeight:
			*outMetric = 12; // pixels
			break;
		
		default:
			break;
	}
	else result = memPCErr;
#else
	result = GetThemeMetric(inMetric, outMetric);
#endif
	
	return result;
}// getThemeMetric


/*!
Returns the number of rows that can be seen
within the current view rectangle of the
specified Finder List.

(1.0)
*/
static SInt16
getVisibleRowCount		(FinderListPtr		inListPtr)
{
	SInt16		result = Chain_GetNodeCount(inListPtr->rowList);
	Rect		contentBounds;
	
	
	if (result > 0)
	{
		GetControlBounds(inListPtr->controls.contentUserPane, &contentBounds);
		result = 1 + (contentBounds.bottom - contentBounds.top) / inListPtr->rowHeight;
	}
	else result = 0;
	
	return result;
}// getVisibleRowCount


/*!
To determine the state of the  key when
you do not have access to an event record,
use this method.  If the key is down, true
is returned; otherwise, false is returned.

(1.0)
*/
Boolean
isCommandKeyDown		(void)
{
	KeyMap		keyMap;
	Boolean		result = false;
	
	
	// The GetKeys() result is an array of four 32-bit
	// unsigned numbers, for a total of 128 bits.  The
	// particular bit corresponding to a key is its
	// virtual key code.  Thus, since the Command key
	// has a virtual key code of 0x37 (55), it is "down"
	// if bit number 55 is a 1.  For a set of virtual
	// key codes, see page 2-43 of the book "Inside
	// Macintosh: Macintosh Toolbox Essentials".
	GetKeys(keyMap);
	result = ((unsigned char*)keyMap)[6] & 0x80; // virtual key 0x37 == bit #55 of 128; the 55->48 (7th) octet, MSB, mask 10000000
	
	return result;
}// isCommandKeyDown


/*!
To determine the state of the Option key
when you do not have access to an event
record, use this method.  If the key is
down, true is returned; otherwise, false
is returned.

(1.0)
*/
Boolean
isOptionKeyDown		(void)
{
	KeyMap		keyMap;
	Boolean		result = false;
	
	
	// The GetKeys() result is an array of four 32-bit
	// unsigned numbers, for a total of 128 bits.  The
	// particular bit corresponding to a key is its
	// virtual key code.  Thus, since the Option key
	// has a virtual key code of 0x3A (58), it is "down"
	// if bit number 58 is a 1.  For a set of virtual
	// key codes, see page 2-43 of the book "Inside
	// Macintosh: Macintosh Toolbox Essentials".
	GetKeys(keyMap);
	result = ((unsigned char*)keyMap)[7] & 0x04; // virtual key 0x3A == bit #58 of 128; the 63->56 (8th) octet, 3rd from LSB, mask 00000100
	
	return result;
}// isOptionKeyDown


/*!
To determine the state of the Shift key
when you do not have access to an event
record, use this method.  If the key is
down, true is returned; otherwise, false
is returned.

(1.0)
*/
Boolean
isShiftKeyDown		(void)
{
	KeyMap		keyMap;
	Boolean		result = false;
	
	
	// The GetKeys() result is an array of four 32-bit
	// unsigned numbers, for a total of 128 bits.  The
	// particular bit corresponding to a key is its
	// virtual key code.  Thus, since the Shift key
	// has a virtual key code of 0x38 (56), it is "down"
	// if bit number 56 is a 1.  For a set of virtual
	// key codes, see page 2-43 of the book "Inside
	// Macintosh: Macintosh Toolbox Essentials".
	GetKeys(keyMap);
	result = ((unsigned char*)keyMap)[7] & 0x01; // virtual key 0x3A == bit #56 of 128; the 63->56 (8th) octet, LSB, mask 00000001
	
	return result;
}// isShiftKeyDown


/*!
To create an array of references to the
*current* heading button controls, use
this method.  This is for iterators, since
arrays are more efficient than the internal
linked list and the number of columns is
very few.  The number of items in the
resultant array is returned (0 or less
implies an error).

WARNING:	The returned array must be
			destroyed with DisposePtr()
			when you are finished with it.

(1.0)
*/
static SInt16
newHeadingButtonsArray	(FinderListPtr	inListPtr,
						 ControlRef**	outArrayPtr)
{
	SInt16		result = 0;
	
	
	if ((inListPtr != null) && (outArrayPtr != null))
	{
		ChainLinkRef			node = null;
		FinderListColumnPtr		columnDataPtr = null;
		register SInt16			i = 0;
		
		
		*outArrayPtr = (ControlRef*)Memory_NewPtr(Chain_GetNodeCount(inListPtr->columnList) * sizeof(ControlRef));
		node = Chain_GetFirstNode(inListPtr->columnList);
		do
		{
			columnDataPtr = Chain_GetNodeDataPtr(inListPtr->columnList, node);
			if (columnDataPtr != null) (*outArrayPtr)[i++] = columnDataPtr->headerButton;
			node = Chain_GetNextNode(inListPtr->columnList, node);
		} while (node != null);
		result = Chain_GetNodeCount(inListPtr->columnList);
	}
	return result;
}// newHeadingButtonsArray


/*!
To obtain a pointer to the internal structure,
given a reference to it, use this method.

(1.0)
*/
static FinderListPtr
refAcquireLock	(FinderListRef	inRef)
{
	return ((FinderListPtr)inRef);
}// refAcquireLock


/*!
To obtain a pointer to the internal structure,
given a reference to it, use this method.

(1.0)
*/
static void
refReleaseLock	(FinderListRef		UNUSED_ARGUMENT(inRef),
				 FinderListPtr*		inoutPtr)
{
	// the memory is not relocatable, so nothing needs to be done here
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock


/*!
Reverses the current sort of the specified
Finder List, updating the sort order reversal
buttons icon to match.

(1.0)
*/
static void
reverseSort		(FinderListPtr		inListPtr)
{
	inListPtr->sortDescending = !inListPtr->sortDescending;
#if USE_SORT_ORDER_REVERSAL_BUTTON
	if (inListPtr->sortDescending) useBestSortDescendingButtonIcon(inListPtr->controls.sortOrderReversalButton);
	else useBestSortAscendingButtonIcon(inListPtr->controls.sortOrderReversalButton);
#endif
	setSortColumn(inListPtr, inListPtr->selectedColumnIndex); // updates sort direction icon and list contents
}// reverseSort


/*!
A standard comparison function that expects
both of its operands to be of type
"FinderListRowPtr".  If the data of the
currently selected column intersecting the
first given row seems to indicate that the
row belongs ahead of the second given row,
in an ascending order, then 1 is returned.
If the reverse is true, -1 is returned.
Otherwise, the rows seem to be equal, so 0
is returned.

(1.0)
*/
static int
rowComparer			(void const*		inFinderListRowPtr1,
					 void const*		inFinderListRowPtr2)
{
	FinderListRowPtr	dataPtr1 = (FinderListRowPtr)inFinderListRowPtr1,
						dataPtr2 = (FinderListRowPtr)inFinderListRowPtr2;
	int					result = 0;
	
	
	if ((dataPtr1 != null) && (dataPtr2 != null))
	{
		ChainLinkRef		cell1 = null,
							cell2 = null;
		FinderListCellPtr	cellDataPtr1 = null,
							cellDataPtr2 = null;
		
		
		cell1 = Chain_GetIndexedNode(dataPtr1->cellList, gSortColumnIndex);
		cell2 = Chain_GetIndexedNode(dataPtr2->cellList, gSortColumnIndex);
		cellDataPtr1 = (FinderListCellPtr)Chain_GetNodeDataPtr(dataPtr1->cellList, cell1);
		cellDataPtr2 = (FinderListCellPtr)Chain_GetNodeDataPtr(dataPtr2->cellList, cell2);
		if ((cellDataPtr1 != null) && (cellDataPtr2 != null))
		{
			result = compareTextCellData(cellDataPtr1->dataPtr, cellDataPtr2->dataPtr);
		}
	}
	return result;
}// rowComparer


/*!
Allows a Finder List to scroll dynamically, as
the user drags the scroll bar thumb.

(1.0)
*/
static pascal void
scrollBarActionProc		(ControlRef				inControl,
						 ControlPartCode		inPartCode)
{
	FinderListRef		ref = (FinderListRef)GetControlReference(inControl);
	FinderListPtr		ptr = null;
	
	
	ptr = refAcquireLock(ref);
	if (ptr != null)
	{
		switch (inPartCode)
		{
			case kControlUpButtonPart:
				scrollDownward(ptr, 1/* number of rows */, true/* update list */);
				break;
			
			case kControlDownButtonPart:
				scrollUpward(ptr, 1/* number of rows */, true/* update list */);
				break;
			
			case kControlPageUpPart:
				scrollDownward(ptr, getVisibleRowCount(ptr)/* number of rows */, true/* update list */);
				break;
			
			case kControlPageDownPart:
				scrollUpward(ptr, getVisibleRowCount(ptr)/* number of rows */, true/* update list */);
				break;
			
			case kControlIndicatorPart:
				// live scrolling - update contents
				Embedding_OffscreenDrawOneControl(ptr->controls.contentUserPane);
				break;
			
			default:
				break;
		}
		
		refReleaseLock(ref, &ptr);
	}
}// scrollBarActionProc


/*!
Scrolls the CONTENTS of the list downward, which
corresponds to the user moving the scroll bar
thumb upwards or clicking an up arrow.  The list
and scroll bar are only updated if "inUpdateList"
is set to "true".

(1.0)
*/
static void
scrollDownward		(FinderListPtr		inListPtr,
					 UInt16 const		inRowsToScroll,
					 Boolean const		inUpdateList)
{
	// this is safe, because the control value automatically does not exceed its maximum
	SetControlValue(inListPtr->controls.scrollBarV, GetControlValue(inListPtr->controls.scrollBarV) - inRowsToScroll);
	if (inUpdateList)
	{
		Embedding_OffscreenDrawOneControl(inListPtr->controls.scrollBarV);
		Embedding_OffscreenDrawOneControl(inListPtr->controls.contentUserPane);
	}
}// scrollDownward


/*!
Changes the view so the specified row is showing.
If you specify "inVerticallyCenterInView" to be
true, then the specified row is positioned
approximately in the middle (compared to the top
and bottom edges of the view area of the list);
otherwise, the specified row appears at the top
of the view area.

You must redraw the list yourself.

(1.0)
*/
static void
scrollToRow		(FinderListPtr			inListPtr,
				 SInt16 const			inZeroBasedRowIndex,
				 Boolean const			inVerticallyCenterInView)
{
	UInt16		rowCount = Chain_GetNodeCount(inListPtr->rowList);
	
	
	if (rowCount > 0)
	{
		// set the vertical scroll bar value; upon a redraw, this will cause the desired behavior
		UInt16		newValue = (inListPtr->sortDescending) ? (rowCount - inZeroBasedRowIndex) : inZeroBasedRowIndex;
		
		
		if (inVerticallyCenterInView) newValue += INTEGER_HALVED(getVisibleRowCount(inListPtr));
		SetControlValue(inListPtr->controls.scrollBarV, newValue);
	}
}// scrollToRow


/*!
Scrolls the CONTENTS of the list upward, which
corresponds to the user moving the scroll bar
thumb downwards or clicking a down arrow.  The list
and scroll bar are only updated if "inUpdateList"
is set to "true".

(1.0)
*/
static void
scrollUpward		(FinderListPtr		inListPtr,
					 UInt16 const		inRowsToScroll,
					 Boolean const		inUpdateList)
{
	// this is safe, because the control value automatically does not exceed its minimum
	SetControlValue(inListPtr->controls.scrollBarV, GetControlValue(inListPtr->controls.scrollBarV) + inRowsToScroll);
	if (inUpdateList)
	{
		Embedding_OffscreenDrawOneControl(inListPtr->controls.scrollBarV);
		Embedding_OffscreenDrawOneControl(inListPtr->controls.contentUserPane);
	}
}// scrollUpward


/*!
While the user holds the mouse down, this routine
continuously draws a rectangular outline and
selects exactly those rows containing one or more
cells intersecting the selection rectangle.  If
the user moves the mouse beyond an edge of the
content region, this routine analyzes the mouse
acceleration and automatically scrolls the list
at an appropriate speed (the faster the mouse
moves, the faster scrolling is done).

Incomplete.

(1.0)
*/
static void
selectionRectangleDrag		(FinderListPtr		inListPtr)
{
	enum
	{
		kAutoScrollDelayInTicks = 4,	// minimum time mouse must drag outside the content area before auto-scrolling
		kSlowScrollDelayInTicks = 16,	// minimum time between scrolls once slow scrolling has begun
		kNormalScrollDelayInTicks = 4,	// minimum time between scrolls once normal scrolling has begun
		kFastScrollDelayInTicks = 0,	// minimum time between scrolls once fast scrolling has begun
		kSlowScrollRangeInPixels = 20,	// maximum pixels outside the content region the mouse must be to scroll slowly
		kNormalScrollRangeInPixels = 30	// maximum pixels outside the slow area the mouse must be to scroll normally
	};
	Rect		selectionRect,
				contentRect;
	Point		localMouse,
				oldMouse,
				anchor;
	RgnHandle	oldClip = null;
	UInt32		autoScrollTimer = 0L;
	UInt16		scrollDelay = 0;
	
	
	oldClip = Memory_NewRegion();
	
	if (oldClip != null) GetClip(oldClip);
	
	GetControlBounds(inListPtr->controls.contentUserPane, &contentRect);
	ClipRect(&contentRect);
	
	GetMouse(&anchor);
	localMouse = oldMouse = anchor;
	Pt2Rect(localMouse, anchor, &selectionRect);
	selectionRectangleDraw(inListPtr, &selectionRect, true);
	
	autoScrollTimer = TickCount();
	scrollDelay = kAutoScrollDelayInTicks;
	while (StillDown())
	{
		(OSErr)YieldToAnyThread();
		
		// draw new selection rectangle
		GetMouse(&localMouse);
		
		// erase old rectangle, if necessary
		unless (EqualPt(localMouse, oldMouse)) selectionRectangleDraw(inListPtr, &selectionRect, false);
		
		// if the mouse has been outside the content region for a certain period of time,
		// then automatic scrolling should begin; figure out the direction for scrolling
		if (PtInRect(localMouse, &contentRect))
		{
			autoScrollTimer = TickCount();
			scrollDelay = kAutoScrollDelayInTicks;
		}
		else if ((TickCount() - autoScrollTimer) > scrollDelay)
		{
			ControlRef		scrollBarControl = inListPtr->controls.scrollBarV;
			
			
			autoScrollTimer = TickCount();
			selectionRectangleDraw(inListPtr, &selectionRect, false);
			if (localMouse.v >= contentRect.bottom)
			{
				// ensure scrolling at a reasonable rate by analyzing how far away from the list the mouse is...
				if ((localMouse.v - contentRect.bottom) > (kSlowScrollRangeInPixels + kNormalScrollRangeInPixels))
				{
					scrollDelay = kFastScrollDelayInTicks;
				}
				else if ((localMouse.v - contentRect.bottom) < kSlowScrollRangeInPixels)
				{
					scrollDelay = kSlowScrollDelayInTicks;
				}
				else scrollDelay = kNormalScrollDelayInTicks;
				
				// update the scroll bar and the list, if necessary
				if (GetControlValue(scrollBarControl) < GetControlMaximum(scrollBarControl))
				{
					anchor.v -= inListPtr->rowHeight;
					scrollUpward(inListPtr, 1/* number of rows */, true/* update list */);
				}
			}
			else if (localMouse.v <= contentRect.top)
			{
				// ensure scrolling at a reasonable rate by analyzing how far away from the list the mouse is...
				if ((contentRect.top - localMouse.v) > (kSlowScrollRangeInPixels + kNormalScrollRangeInPixels))
				{
					scrollDelay = kFastScrollDelayInTicks;
				}
				else if ((contentRect.top - localMouse.v) < kSlowScrollRangeInPixels)
				{
					scrollDelay = kSlowScrollDelayInTicks;
				}
				else scrollDelay = kNormalScrollDelayInTicks;
				
				// update the scroll bar and the list, if necessary
				if (GetControlValue(scrollBarControl) > GetControlMinimum(scrollBarControl))
				{
					anchor.v += inListPtr->rowHeight;
					scrollDownward(inListPtr, 1/* number of rows */, true/* update list */);
				}
			}
			selectionRectangleDraw(inListPtr, &selectionRect, true);
		}
		
		// figure out where the new rectangle should be, and draw it
		unless (EqualPt(localMouse, oldMouse))
		{
			Pt2Rect(localMouse, anchor, &selectionRect);
			selectionRectangleDraw(inListPtr, &selectionRect, true);
		}
		
		// change selected rows, if necessary - incomplete
		
		oldMouse = localMouse;
	}
	selectionRectangleDraw(inListPtr, &selectionRect, false);
	if (oldClip != null)
	{
		SetClip(oldClip);
		Memory_DisposeRegion(&oldClip);
	}
}// selectionRectangleDrag


/*!
To draw or erase a rectangular outline for selecting
rows of the given list, use this method.  The current
implementation draws a two-pixel-thick border in a
checkerboard gray pattern, like the Mac OS 9 Finder.

(1.0)
*/
static void
selectionRectangleDraw		(FinderListPtr		UNUSED_ARGUMENT(inListPtr),
							 Rect const*		inRectPtr,
							 Boolean const		UNUSED_ARGUMENT(inDraw))
{
	PenState		state;
	
	
	GetPenState(&state);
	PenNormal();
	PenSize(2, 2);
	PenMode(srcXor); // XOR will either draw or erase, so "inDraw" is currently ignored
	ColorUtilities_SetGrayPenPattern();
	FrameRect(inRectPtr);
	SetPenState(&state);
}// selectionRectangleDraw


/*!
Changes the currently selected sort column.
The rows of the list are automatically
sorted appropriately.

If you try to select any column with
"kFinderListColumnFlagsNotSortable" set,
this routine silently fails.

You must redraw the list yourself.

(1.0)
*/
void
setSortColumn		(FinderListPtr		inPtr,
					 SInt16 const		inZeroBasedColumnIndex)
{
	//if (inPtr->selectedColumnIndex != inZeroBasedColumnIndex)
	{
		ChainLinkRef				node = null;
		FinderListColumnPtr			columnDataPtr = null;
		register SInt16				i = 0,
									headingButtonsArraySize = 0;
		ControlRef*					headingButtons = null;
		ControlButtonContentInfo	contentInfo;
		
		
		// set up a content information structure that can be used to remove button icons
		contentInfo.contentType = kControlContentTextOnly;
		
		// traverse the list of column buttons once, and create an array for efficiency in the loop
		headingButtonsArraySize = newHeadingButtonsArray(inPtr, &headingButtons);
		node = Chain_GetFirstNode(inPtr->columnList);
		for (i = 0; i < headingButtonsArraySize; i++)
		{
			columnDataPtr = (FinderListColumnPtr)Chain_GetNodeDataPtr(inPtr->columnList, node);
			if (columnDataPtr != null)
			{
				if (i == inZeroBasedColumnIndex)
				{
					// make this column the new sort column
					SetControlValue(headingButtons[i], kControlCheckBoxCheckedValue);
					columnDataPtr->backgroundBrush = kThemeBrushListViewSortColumnBackground;
				#if !USE_SORT_ORDER_REVERSAL_BUTTON
					if (inPtr->sortDescending) useBestSortDescendingButtonIcon(headingButtons[i]);
					else useBestSortAscendingButtonIcon(headingButtons[i]);
				#endif
				}
				else
				{
					// ensure this column is not selected
					SetControlValue(headingButtons[i], kControlCheckBoxUncheckedValue);
					columnDataPtr->backgroundBrush = kThemeBrushListViewBackground;
					
					// remove the direction icon from the button
					(OSErr)SetControlData(headingButtons[i], kControlNoPart, kControlBevelButtonContentTag,
											sizeof(contentInfo), &contentInfo);
				}
				Embedding_OffscreenDrawOneControl(headingButtons[i]);
			}
			node = Chain_GetNextNode(inPtr->columnList, node);
		}
		DisposePtr((Ptr)headingButtons), headingButtons = null;
		
		inPtr->selectedColumnIndex = inZeroBasedColumnIndex;
		sortRows(inPtr);
	}
}// setSortColumn


/*!
Sorts all of the rows in the list based on the
cell in each row corresponding to the current sort
column.  If none of the columns is sortable, this
routine has no effect.

The quick sort algorithm is used.

(1.0)
*/
static void
sortRows	(FinderListPtr		inListPtr)
{
	if (inListPtr != null) if (inListPtr->selectedColumnIndex >= 0)
	{
		OSErr		error = noErr;
		
		
		gSortColumnIndex = inListPtr->selectedColumnIndex;
		error = Chain_QuickSort(inListPtr->rowList, rowComparer);
		if (error != noErr) BEEP;
	}
}// sortRows


/*!
To automatically use the best (32-bit) sort-ascending
button icon available for the specified bevel button
control, use this method.  If Icon Services is not
available, an Interface Library icon resource is
automatically used.

(1.0)
*/
static void
useBestSortAscendingButtonIcon		(ControlRef		inControl)
{
	IconManagerIconRef		icon = IconManager_NewIcon();
	
	
	IconManager_MakeIconSuite(icon, kInterfaceLibIconIDSortAscending, kSelectorAllSmallData);
	(OSErr)IconManager_SetButtonIcon(inControl, icon);
	IconManager_DisposeIcon(&icon);
}// useBestSortAscendingButtonIcon


/*!
To automatically use the best (32-bit) sort-descending
button icon available for the specified bevel button
control, use this method.  If Icon Services is not
available, an Interface Library icon resource is
automatically used.

(1.0)
*/
static void
useBestSortDescendingButtonIcon		(ControlRef		inControl)
{
	IconManagerIconRef		icon = IconManager_NewIcon();
	
	
	IconManager_MakeIconSuite(icon, kInterfaceLibIconIDSortDescending, kSelectorAllSmallData);
	(OSErr)IconManager_SetButtonIcon(inControl, icon);
	IconManager_DisposeIcon(&icon);
}// useBestSortDescendingButtonIcon