/*###############################################################

	FolderManager.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Folders.h>
#include <PLStringFuncs.h>

// library includes
#include "DataAccessLib.h"

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "FileUtilities.h"
#include "FolderManager.h"



// internal methods
static void		transformFolderFSSpec		(FSSpec*				inoutFolderFSSpecPtr);



//
// public methods
//

/*!
To fill in a file system specification record with
information for a particular folder, call this
routine.  If the folder doesnt exist, it is created.

The name information for the resultant file
specification record is left blank, so that you may
fill in any name you wish.  That way, you can either
use FSMakeFSSpec() to obtain a file specification
for a file contained in the requested folder, or you
can leave the name blank, and FSMakeFSSpec() will
fill in the name of the folder and adjust the parent
directory ID appropriately.

If no problems occur, "noErr" is returned.  If the
given folder type is not recognized as one of the
valid types, "invalidFolderTypeErr" is returned.

(3.0)
*/
OSErr
FolderManager_GetFSSpec		(TelnetFolder const		inFolderType,
							 FSSpec*				outFolderFSSpecPtr)
{
	OSErr		result = noErr;
	
	
	if (outFolderFSSpecPtr != null) switch (inFolderType)
	{
		case kTelnetFolderApplication: // the folder that ALWAYS contains MacTelnets application (no matter where it appears to be)
			{
				ProcessSerialNumber		telnetItself;
				
				
				result = GetCurrentProcess(&telnetItself);
				if (result == noErr)
				{
					ProcessInfoRec		info;
					FSSpec				applicationProgramFile;
					
					
					// clear the buffer to zeroes
					Memory_WriteZero((Ptr)&info, sizeof(info));
					
					// GetProcessInformation() requires that the following
					// three fields be initialized ahead of time.
					info.processInfoLength = sizeof(info);
					info.processName = null; // dont get the name
					info.processAppSpec = &applicationProgramFile;
					
					if (GetProcessInformation(&telnetItself, &info) == noErr)
					{
						result = FSMakeFSSpec(info.processAppSpec->vRefNum, info.processAppSpec->parID, EMPTY_PSTRING,
												outFolderFSSpecPtr);
						
						// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
						// (this makes the FSSpec much more useful, because it can then be used to place files *inside* a folder)
						transformFolderFSSpec(outFolderFSSpecPtr);
					}
				}
			}
			break;
		
		case kTelnetFolderFavorites: // the Favorites folder inside the MacTelnet preferences folder
			result = FolderManager_GetFSSpec(kTelnetFolderPreferences, outFolderFSSpecPtr);
			if (result == noErr)
			{
				Str255		favoritesFolderName;
				long		unusedDirID = 0L;
				
				
				GetIndString(favoritesFolderName, rStringsImportantFolderNames, siImportantFolderNameApplicationFavorites);
				result = FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										favoritesFolderName, outFolderFSSpecPtr);
				
				// if no Favorites folder exists in MacTelnets preferences folder, create it
				(OSErr)FSpDirCreate(outFolderFSSpecPtr, GetScriptManagerVariable(smSysScript), &unusedDirID);
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			break;
		
		case kTelnetFolderPackageParent: // the folder that looks like it contains MacTelnets application
		#if TARGET_API_MAC_OS8
			// in Mac OS 8 and 9, the package parent is the same as the application folder
			// because the application is the package
			result = FolderManager_GetFSSpec(kTelnetFolderApplication, outFolderFSSpecPtr);
		#else
			// in Mac OS X, the package looks like the application, but from a file
			// system standpoint, the application program file is actually two levels
			// down (the application is in the Contents folder, which in turn is in
			// the MacTelnet.app folder, the parent of which is the package parent)
			result = FolderManager_GetFSSpec(kTelnetFolderApplication, outFolderFSSpecPtr);
			
			// My mistake - apparently, in Mac OS X, the Process Manager *IS*
			// smart enough to return the specification of the package icon
			// and not the application program file itself; great!  In case
			// this changes, the code below would return the folder that
			// contains the package folder itself.
			#if 0
			if (result == noErr)
			{
				(OSErr)FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										outFolderFSSpecPtr->name, outFolderFSSpecPtr); // this gives the parent ID of the directory...
				(OSErr)FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										EMPTY_PSTRING, outFolderFSSpecPtr); // ...therefore, this gives the ID of the parents parent...
				(OSErr)FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										EMPTY_PSTRING, outFolderFSSpecPtr); // ...and this gives the parent itself 
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			#endif
		#endif
			break;
		
		case kTelnetFolderPreferences: // the MacTelnet Preferences folder inside the system Preferences folder
			result = FolderManager_GetFSSpec(kTelnetFolderMacPreferences, outFolderFSSpecPtr);
			if (result == noErr)
			{
				Str255		macTelnetPreferencesFolderName;
				long		unusedDirID = 0L;
				
				
				GetIndString(macTelnetPreferencesFolderName, rStringsImportantFolderNames, siImportantFolderNameApplicationPreferences);
				result = FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										macTelnetPreferencesFolderName, outFolderFSSpecPtr);
				
				// if no MacTelnet Preferences folder exists in the current users Preferences folder, create it
				(OSErr)FSpDirCreate(outFolderFSSpecPtr, GetScriptManagerVariable(smSysScript), &unusedDirID);
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			break;
		
		case kTelnetFolderScriptsMenuItems: // the MacTelnet Scripts Menu Items folder, in MacTelnets folder
			result = FolderManager_GetFSSpec(kTelnetFolderPackageParent, outFolderFSSpecPtr);
			if (result == noErr)
			{
				Str255		scriptsFolderName;
				
				
				GetIndString(scriptsFolderName, rStringsImportantFolderNames, siImportantFolderNameApplicationScriptsMenuItems);
				result = FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										scriptsFolderName, outFolderFSSpecPtr);
				
				// unlike most folders, if this folder cant be found, dont create it
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			break;
		
		case kTelnetFolderStartupItems: // the MacTelnet Startup Items folder, in MacTelnets preferences folder
			result = FolderManager_GetFSSpec(kTelnetFolderPreferences, outFolderFSSpecPtr);
			if (result == noErr)
			{
				Str255		startupItemsFolderName;
				long		unusedDirID = 0L;
				
				
				GetIndString(startupItemsFolderName, rStringsImportantFolderNames, siImportantFolderNameApplicationStartupItems);
				result = FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										startupItemsFolderName, outFolderFSSpecPtr);
				
				// if no Startup Items folder exists in the MacTelnet Preferences folder, create it
				(OSErr)FSpDirCreate(outFolderFSSpecPtr, GetScriptManagerVariable(smSysScript), &unusedDirID);
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			break;
		
		case kTelnetFolderUserFTPIDFavorites: // the FTP Users folder inside MacTelnets Favorites folder (in preferences)
			result = FolderManager_GetFSSpec(kTelnetFolderFavorites, outFolderFSSpecPtr);
			if (result == noErr)
			{
				Str255		ftpUsersFolderName;
				long		unusedDirID = 0L;
				
				
				GetIndString(ftpUsersFolderName, rStringsImportantFolderNames, siImportantFolderNameApplicationFavoritesFTPUsers);
				result = FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										ftpUsersFolderName, outFolderFSSpecPtr);
				
				// if no folder exists in the Favorites folder, create it
				(OSErr)FSpDirCreate(outFolderFSSpecPtr, GetScriptManagerVariable(smSysScript), &unusedDirID);
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			break;
		
		case kTelnetFolderUserProxyFavorites: // the Proxies folder inside MacTelnets Favorites folder (in preferences)
			result = FolderManager_GetFSSpec(kTelnetFolderFavorites, outFolderFSSpecPtr);
			if (result == noErr)
			{
				Str255		proxiesFolderName;
				long		unusedDirID = 0L;
				
				
				GetIndString(proxiesFolderName, rStringsImportantFolderNames, siImportantFolderNameApplicationFavoritesProxies);
				result = FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										proxiesFolderName, outFolderFSSpecPtr);
				
				// if no folder exists in the Favorites folder, create it
				(OSErr)FSpDirCreate(outFolderFSSpecPtr, GetScriptManagerVariable(smSysScript), &unusedDirID);
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			break;
		
		case kTelnetFolderUserSessionFavorites: // the Sessions folder inside MacTelnets Favorites folder (in preferences)
			result = FolderManager_GetFSSpec(kTelnetFolderFavorites, outFolderFSSpecPtr);
			if (result == noErr)
			{
				Str255		sessionsFolderName;
				long		unusedDirID = 0L;
				
				
				GetIndString(sessionsFolderName, rStringsImportantFolderNames, siImportantFolderNameApplicationFavoritesSessions);
				result = FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										sessionsFolderName, outFolderFSSpecPtr);
				
				// if no folder exists in the Favorites folder, create it
				(OSErr)FSpDirCreate(outFolderFSSpecPtr, GetScriptManagerVariable(smSysScript), &unusedDirID);
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			break;
		
		case kTelnetFolderUserTerminalFavorites: // the Terminals folder inside MacTelnets Favorites folder (in preferences)
			result = FolderManager_GetFSSpec(kTelnetFolderFavorites, outFolderFSSpecPtr);
			if (result == noErr)
			{
				Str255		terminalsFolderName;
				long		unusedDirID = 0L;
				
				
				GetIndString(terminalsFolderName, rStringsImportantFolderNames, siImportantFolderNameApplicationFavoritesTerminals);
				result = FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										terminalsFolderName, outFolderFSSpecPtr);
				
				// if no folder exists in the Favorites folder, create it
				(OSErr)FSpDirCreate(outFolderFSSpecPtr, GetScriptManagerVariable(smSysScript), &unusedDirID);
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			break;
		
		case kTelnetFolderRecentSessions: // the Recent Sessions folder inside MacTelnets preferences folder
			result = FolderManager_GetFSSpec(kTelnetFolderPreferences, outFolderFSSpecPtr);
			if (result == noErr)
			{
				Str255		recentSessionsFolderName;
				long		unusedDirID = 0L;
				
				
				GetIndString(recentSessionsFolderName, rStringsImportantFolderNames, siImportantFolderNameApplicationRecentSessions);
				result = FSMakeFSSpec(outFolderFSSpecPtr->vRefNum, outFolderFSSpecPtr->parID,
										recentSessionsFolderName, outFolderFSSpecPtr);
				
				// if no folder exists in the MacTelnet Preferences folder, create it
				(OSErr)FSpDirCreate(outFolderFSSpecPtr, GetScriptManagerVariable(smSysScript), &unusedDirID);
				
				// now reconstruct the FSSpec so the parent directory is the directory itself, and the name is blank
				// (this makes the FSSpec much more useful, because it can then be used to place files *in* a folder)
				transformFolderFSSpec(outFolderFSSpecPtr);
			}
			break;
		
		case kTelnetFolderMacDesktop: // the Mac OS Desktop folder
			result = FindFolder(kOnSystemDisk, kDesktopFolderType, kCreateFolder,
								&(outFolderFSSpecPtr->vRefNum), &(outFolderFSSpecPtr->parID));
			outFolderFSSpecPtr->name[0] = 0;
			break;
		
		case kTelnetFolderMacFavorites: // the Mac OS Favorites folder
			result = FindFolder(kOnSystemDisk, kFavoritesFolderType, kCreateFolder,
								&(outFolderFSSpecPtr->vRefNum), &(outFolderFSSpecPtr->parID));
			outFolderFSSpecPtr->name[0] = 0;
			break;
		
		case kTelnetFolderMacHelp: // the Mac OS Help folder
			result = FindFolder(kOnSystemDisk, kHelpFolderType, kCreateFolder,
								&(outFolderFSSpecPtr->vRefNum), &(outFolderFSSpecPtr->parID));
			outFolderFSSpecPtr->name[0] = 0;
			break;
		
		case kTelnetFolderMacPreferences: // the Mac OS Preferences folder
			result = FindFolder(kOnSystemDisk, kPreferencesFolderType, kCreateFolder,
								&(outFolderFSSpecPtr->vRefNum), &(outFolderFSSpecPtr->parID));
			outFolderFSSpecPtr->name[0] = 0;
			break;
		
		case kTelnetFolderMacSystem: // the Mac OS System Folder folder
			result = FindFolder(kOnSystemDisk, kSystemFolderType, kCreateFolder,
								&(outFolderFSSpecPtr->vRefNum), &(outFolderFSSpecPtr->parID));
			outFolderFSSpecPtr->name[0] = 0;
			break;
		
		case kTelnetFolderMacTemporaryItems: // the invisible Mac OS Temporary Items folder
			result = FindFolder(kOnSystemDisk, kTemporaryFolderType, kCreateFolder,
								&(outFolderFSSpecPtr->vRefNum), &(outFolderFSSpecPtr->parID));
			outFolderFSSpecPtr->name[0] = 0;
			break;
		
		case kTelnetFolderMacTrash: // the Trash
			result = FindFolder(kOnSystemDisk, kTrashFolderType, kCreateFolder,
								&(outFolderFSSpecPtr->vRefNum), &(outFolderFSSpecPtr->parID));
			outFolderFSSpecPtr->name[0] = 0;
			break;
		
		default:
			// ???
			result = invalidFolderTypeErr;
			break;
	}
	else result = memPCErr;
	
	return result;
}// GetFSSpec


//
// internal methods
//

/*!
Transforms a folder specification so that the
file name of the folder is empty, and the
parent ID is the ID of the folder itself.
This is very useful, since it then becomes
easy to create file specifications for files
*inside* the folder (and yet the folder
specification itself remains a valid way to
refer to the folder).

(3.0)
*/
static void
transformFolderFSSpec		(FSSpec*	inoutFolderFSSpecPtr)
{
	inoutFolderFSSpecPtr->parID = FileUtilities_GetDirectoryIDFromFSSpec(inoutFolderFSSpecPtr);
	PLstrcpy(inoutFolderFSSpecPtr->name, EMPTY_PSTRING);
}// transformFolderFSSpec