/*###############################################################

	FolderManager.h
	
	The Folder Manager has been created in MacTelnet 3.0 so that
	MacTelnet manages folder information dynamically and without
	global variables.  Unlike Telnet 2.6, folders can always be
	found in version 3.0, even if the user moves or renames a
	folder while MacTelnet is running.  Also, using these methods
	guarantees consistent results, since you can request a folder
	in terms of a context, instead of using hard-coded file or
	path information.  You should use this whenever possible!
	
	For convenience, FolderManager_GetFSSpec() returns file
	specifications with empty file names.  This means that the
	volume information is filled in, the file name is an empty
	string, and the parent ID is actually the ID of the
	directory itself.  This makes it easy to construct file
	specifications for any item inside a directory, since you
	need only fill in the file name appropriately and call
	FSMakeFSSpec().  Alternately, you can call FSMakeFSSpec()
	immediately on the returned FSSpec, and the File Manager
	will reconstruct the file specification so that the
	file name is the name of the directory and the parent ID
	is the ID of the parent folder.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#ifndef __FOLDERMANAGER__
#define __FOLDERMANAGER__

// Mac includes
#include <Files.h>



// constants
typedef SInt16 TelnetFolder;
enum
{
	// folders defined by MacTelnet
	kTelnetFolderApplication = 0,				// The folder containing the application program FILE; this
												//	is where mandatory files (such as shared libraries and
												//	the MacTelnet Help launcher) should exist.  This is NOT
												//  where user files (such as the Scripts Menu Items folder)
												//  go, because on Mac OS X the application folder is not
												//  accessible - it is hidden by the package (which looks
												//  like a file).  In general, to look for user files that
												//  should go in the application folder, use the constant
												//  "kTelnetFolderPackageParent", which returns the correct
												//  folder on both Mac OS 8/9 and Mac OS X.
	
	kTelnetFolderFavorites = 1,					// MacTelnets Favorites folder, in the preferences folder.
	
	kTelnetFolderPackageParent = 3,				// The folder that LOOKS LIKE it contains MacTelnets
												//   application program file.  On Mac OS 8, this is the same
												//   as "kTelnetFolderApplication"; however, on Mac OS X,
												//   this refers to the folder that contains the application
												//   program package (which looks like a normal application
												//   icon in Mac OS X).  This folder is user-accessible at
												//   all times, whereas "kTelnetFolderApplication" is invisible
												//   for Mac OS X users.  If you want the user to places files
												//   (such as Scripts Menu Items) into a place relative to the
												//   location of MacTelnet, use this constant to describe the
												//   folder instead of "kTelnetFolderApplication".
	
	kTelnetFolderPreferences = 4,				// The folder MacTelnet Preferences, in the preferences
												//   folder of the user currently logged in.
	
	kTelnetFolderScriptsMenuItems = 5,			// MacTelnets Scripts Menu Items folder.
	
	kTelnetFolderStartupItems = 6,				// MacTelnets Startup Items folder, in MacTelnets
												//   preferences folder.
	
	kTelnetFolderUserFTPIDFavorites = 7,		// FTP Users folder, in the Favorites folder.
	
	kTelnetFolderUserProxyFavorites = 8,		// Proxies folder, in the Favorites folder.
	
	kTelnetFolderUserSessionFavorites = 9,		// Sessions folder, in the Favorites folder.
	
	kTelnetFolderUserTerminalFavorites = 10,	// Terminals folder, in the Favorites folder.
	
	kTelnetFolderRecentSessions = 11,			// Recent Sessions folder, in MacTelnets preferences folder.
	
	// folders defined by the Mac OS
	kTelnetFolderMacPreferences = -1,			// the Preferences folder for the user currently logged in
	
	kTelnetFolderMacSystem = -2,				// the System Folder
	
	kTelnetFolderMacTemporaryItems = -3,		// the (invisible) Temporary Items folder
	
	kTelnetFolderMacFavorites = -4,				// the Favorites folder for the user currently logged in
	
	kTelnetFolderMacDesktop = -5,				// the Desktop for the user currently logged in
	
	kTelnetFolderMacHelp = -6,					// the system-wide Help folder
	
	kTelnetFolderMacTrash = -7					// the Trash
};



//
// public methods
//

// RETURNS A FOLDER, CREATING IT IF IT DOESNT EXIST (THE FILE NAME IS ALWAYS AN EMPTY STRING)
OSErr
	FolderManager_GetFSSpec				(TelnetFolder const		inFolderType,
										 FSSpec*				outFolderFSSpecPtr);

// RETURNS A FOLDER, BUT GENERATES fnfErr, INSTEAD OF CREATING THE FOLDER, IF IT DOESNT EXIST
OSErr
	FolderManager_GetFSSpecDoNotCreate	(TelnetFolder const		inFolderType,
										 FSSpec*				outFolderFSSpecPtr);

#endif