/*###############################################################

	FormatDialog.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <ControlDefinitions.h>
#include <MacErrors.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "StringResources.h"
#include "TelnetFormatDialog.r"

// MacTelnet includes
#include "ColorBox.h"
#include "configure.proto.h"
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "FolderManager.h"
#include "FormatDialog.h"
#include "InterfaceLib.h"
#include "Preferences.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



// types
struct UndoDataFormatChanges
{
	// context data for the context ID "kUndoableContextIdentifierTerminalFormatChanges"
	TelnetScreenID			screenID;	// which screen was reformatted
	FormatDialogSetupData	data;		// the collection of formatting information
};
#ifndef __cplusplus
typedef struct UndoDataFormatChanges		UndoDataFormatChanges;
#endif
typedef UndoDataFormatChanges*				UndoDataFormatChangesPtr;

// variables
static SInt16						gOldPanelIndex = kFormatDialogIndexNormalText;
static Boolean						gIsCustomSizeMenuItemAdded = false;
static FormatDialogSetupDataPtr		gFormatDialogDataPtr = null;
static MenuHandle					gFormatTypeMenuHandle = null;
static DialogPtr					gFormatDialog = null; // the Mac OS dialog pointer
static WindowInfoRef				gFormatDialogWindowInfo;

// internal methods
static void				checkSizeMenuSize			(MenuHandle						inStandardSizeMenu,
													 const SInt16					inSize);
static Boolean			customSizeMenuSize			(MenuHandle						inStandardSizeMenu,
													 const SInt16					inCustomSizeOrZero);
static void				disableItemGroupsAsNeeded	(FormatDialogFormatOptions		inBasedOnTheseOptions);
static void				fontUpdate					(void);
static pascal Boolean	formatDialogEventFilter		(DialogPtr						inDialogPtr,
													 EventRecord*					inoutEventPtr,
													 short*							outItemIndex);
static MenuHandle		getFontMenu					(void);
static MenuHandle		getSizeMenu					(void);
static pascal Boolean	handleNewSize				(const WindowRef				inWindow,
													 const SInt32					inDeltaSizeX,
													 const SInt32					inDeltaSizeY,
													 const SInt32					inData);
static void				setFormatFromItems			(FormatDialogSetupDataPtr		inDataPtr,
													 const UInt16					inFormatIndexToUse);
static void				setItemsForFormat			(FormatDialogSetupDataPtr		inDataPtr,
													 const UInt16					inFormatIndexToUse);
static void				setSizeField				(const SInt16					inNewSizeOrZero);
static void				setUpSizeMenuForFont		(MenuHandle						inStandardSizeMenu,
													 StringPtr						inFontName);



//
// public methods
//

/*!
This method is used to initialize the Format dialog
box.  It creates the dialog box invisibly, and uses
the specified data to set up the corresponding
controls in the dialog box.  Do not pass null for
"inSetupDataPtr".

The extended dialog resource for the preferences
dialog causes the Appearance Manager to automatically
construct an embedding hierarchy, so the dialogs
root control and container controls are supposed to
be set up once GetNewDialog() returns.

WARNING:	The pointer provided must remain valid
			until you call FormatDialog_Done().

(3.0)
*/
void
FormatDialog_Init	(FormatDialogSetupDataPtr		inSetupDataPtr)
{
	if (inSetupDataPtr != null)
	{
		Cursors_DeferredUseWatch(30); // if it takes more than half a second to initialize, show the watch cursor
		
		gFormatDialogDataPtr = inSetupDataPtr;
		gOldPanelIndex = kFormatDialogIndexNormalText;
		gIsCustomSizeMenuItemAdded = false;
		
		GetNewDialogWithWindowInfo(kDialogIDFormat, &gFormatDialog, &gFormatDialogWindowInfo);
		
		if (gFormatDialog != null)
		{
			ControlRef		control = null;
			
			
			// determine if the Help button should be disabled...
			SetUpHelpButtonDialogItem(gFormatDialog, iFormatAppleGuideButton);
			
			// adjust the OK and Cancel buttons so they are big enough for their titles
			SetDialogDefaultItem(gFormatDialog, iFormatOKButton);
			SetDialogCancelItem(gFormatDialog, iFormatCancelButton);
			(UInt16)Localization_AdjustDialogButtons(gFormatDialog, true/* OK */, true/* Cancel */);
			Localization_AdjustHelpButtonItem(gFormatDialog, iFormatAppleGuideButton);
			
			// set the static text controls (auto-sizing them) and the bottom checkbox (not changing its size)
			{
				ControlRef		control = null;
				Str255			text;
				
				
				GetDialogItemAsControl(gFormatDialog, iFormatEditSizeLabel, &control);
				GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextFormatSizeLabel);
				(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
				
				GetDialogItemAsControl(gFormatDialog, iFormatForegroundColorLabel, &control);
				GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextFormatForegroundColor);
				(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
				
				GetDialogItemAsControl(gFormatDialog, iFormatBackgroundColorLabel, &control);
				GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextFormatBackgroundColor);
				(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
				
				GetDialogItemAsControl(gFormatDialog, iFormatUpdateDefaultTerminalCheckBox, &control);
				GetIndString(text, rStringsMiscellaneousControlTitles, siControlTitleUpdateDefaultTerminal);
				SetControlTitle(control, text);
			}
			
			// create the two color boxes
			ColorBox_AttachToUserPaneDialogItem(gFormatDialog, iFormatForegroundColorBox, null/* initial color */);
			ColorBox_AttachToUserPaneDialogItem(gFormatDialog, iFormatBackgroundColorBox, null/* initial color */);
			
			// prevent the user from typing non-digits into the Font Size field
			{
				ControlRef				sizeField = null;
				ControlKeyFilterUPP		upp = NewControlKeyFilterUPP(NumericalLimiter);
				
				
				GetDialogItemAsControl(gFormatDialog, iFormatEditSizeField, &sizeField);
				(OSErr)SetControlData(sizeField, kControlNoPart, kControlKeyFilterTag,
										sizeof(ControlKeyFilterUPP), (Ptr)&upp);
			}
			
			// set up the little arrows control so it can span the font size range
			{
				ControlRef		littleArrows = null;
				
				
				GetDialogItemAsControl(gFormatDialog, iFormatEditSizeArrows, &littleArrows);
				SetControlMaximum(littleArrows, SHRT_MAX);
				SetControlMinimum(littleArrows, 0);
			}
			
			#if 0
			// temp - what the hell is going on with the user panes?
			{
				ControlUserPaneDrawUPP		upp = NewControlUserPaneDrawProc(FrameUserPaneArea);
				ControlRef					control = null;
				
				
				GetDialogItemAsControl(gFormatDialog, iFormatEditMenuFontItemsPane, &control);
				(OSErr)SetControlData(control, kControlNoPart, kControlUserPaneDrawProcTag,
										sizeof(ControlUserPaneDrawUPP), (Ptr)&upp);
				GetDialogItemAsControl(gFormatDialog, iFormatEditMenuFontSizeItemsPane, &control);
				(OSErr)SetControlData(control, kControlNoPart, kControlUserPaneDrawProcTag,
										sizeof(ControlUserPaneDrawUPP), (Ptr)&upp);
				GetDialogItemAsControl(gFormatDialog, iFormatEditMenuForeColorItemsPane, &control);
				(OSErr)SetControlData(control, kControlNoPart, kControlUserPaneDrawProcTag,
										sizeof(ControlUserPaneDrawUPP), (Ptr)&upp);
				GetDialogItemAsControl(gFormatDialog, iFormatEditMenuBackColorItemsPane, &control);
				(OSErr)SetControlData(control, kControlNoPart, kControlUserPaneDrawProcTag,
										sizeof(ControlUserPaneDrawUPP), (Ptr)&upp);
			}
			#endif
			
			// initialize the controls based on the default set
			setItemsForFormat(gFormatDialogDataPtr, 0);
			fontUpdate();
			
			// if the option was set to hide the Update Default Terminal Configuration checkbox, then hide it!
			if (gFormatDialogDataPtr->generalOptions & kFormatDialogGeneralOptionHideUpdateDefaultTerminalCheckBox)
			{
				HideDialogItem(gFormatDialog, iFormatUpdateDefaultTerminalCheckBox);
			}
			
			// set up the Window Info information
			{
				WindowResizeResponderProcPtr	resizeProc = handleNewSize;
				
				
				WindowInfo_SetWindowDescriptor(gFormatDialogWindowInfo, kTelnetDialogDescriptorFormat);
				WindowInfo_SetWindowResizeLimits(gFormatDialogWindowInfo,
														FORMAT_DIALOG_HT/* minimum height */,
														FORMAT_DIALOG_WD/* minimum width */,
														FORMAT_DIALOG_HT/* maximum height */,
														FORMAT_DIALOG_WD + 50/* maximum width */);
				WindowInfo_SetWindowResizeResponderUPP(gFormatDialogWindowInfo, resizeProc, 0L);
			}
			WindowInfo_SetForDialog(gFormatDialog, gFormatDialogWindowInfo);
		}
	}
}// Init


/*!
Call this method to destroy the formatting dialog
box, and to fill in the data structure provided to
FormatDialog_Init() to contain the users changes.

(3.0)
*/
void
FormatDialog_Done	(void)
{
	if (gFormatTypeMenuHandle != null) DisposeMenu(gFormatTypeMenuHandle), gFormatTypeMenuHandle = null;
	WindowInfo_Dispose(gFormatDialogWindowInfo);
	
	// clean up UPPs
	ColorBox_DetachFromUserPaneDialogItem(gFormatDialog, iFormatForegroundColorBox);
	ColorBox_DetachFromUserPaneDialogItem(gFormatDialog, iFormatBackgroundColorBox);
	
	DisposeDialog(gFormatDialog), gFormatDialog = null;
}// Done


/*!
This method displays and handles events in the
Format dialog box.  If the user clicks OK,
"true" is returned; otherwise, "false" is
returned.

(3.0)
*/
Boolean
FormatDialog_Display	(void)
{
	Boolean		result = false;
	
	
	if (gFormatDialog != null)
	{
		register DialogItemIndex	i = 0;
		DialogItemIndex				itemIndex = 0;
		ModalFilterUPP				filterProc = NewModalFilterUPP(formatDialogEventFilter);
		
		
		// define this only for debugging; this lets you dump the dialogs
		// control embedding hierarchy to a file, before it is displayed
		#if 0
		DebugSelectControlHierarchyDumpFile(GetDialogWindow(gFormatDialog));
		#endif
		
		// display the dialog
		TransitionDialogDisplay(gFormatDialog);
		
		// handle events
		Menus_EnableEditMenu();
		itemIndex = 0;
		do
		{
			ModalDialog(filterProc, &itemIndex);
			switch (itemIndex)
			{
				case iFormatAppleGuideButton:
					{
						Str255		searchString;
						
						
						GetIndString(searchString, rStringsAppleGuideLookup, siAGIndexTermFormatting);
						AppleGuideUtilities_OpenFirstMainGuideInSearch(searchString);
					}
					break;
				
				case iFormatEditMenuPrimaryGroup:
					{
						SInt16		value = GetDialogItemValue(gFormatDialog, iFormatEditMenuPrimaryGroup);
						
						
						switch (--value) // indices are zero-based, item numbers are one-based
						{
							case kFormatDialogIndexNormalText:
							case kFormatDialogIndexBoldText:
							case kFormatDialogIndexBlinkingText:
								if (value != gOldPanelIndex)
								{
									setFormatFromItems(gFormatDialogDataPtr, gOldPanelIndex); // save old
									setItemsForFormat(gFormatDialogDataPtr, value); // draw new
									gOldPanelIndex = value;
								}
								break;
							
							default:
								// ???
								break;
						}
					}
					break;
				
				case iFormatEditFontPopUpMenu:
					fontUpdate();
					break;
				
				case iFormatEditSizeArrows:
					// necessarily handled in the event filter
					break;
				
				case iFormatEditSizePopUpMenu:
					{
						SInt16		value = GetDialogItemValue(gFormatDialog, iFormatEditSizePopUpMenu);
						
						
						switch (value)
						{
							default:
								// simply copy the menu item text to the text field
								{
									Str255			itemText;
									MenuHandle		sizeMenu = null;
									ControlRef		sizeMenuButton = null;
									Size			actualSize = 0L;
									
									
									GetDialogItemAsControl(gFormatDialog, iFormatEditSizePopUpMenu, &sizeMenuButton);
									if (GetControlData(sizeMenuButton, kControlMenuPart,
														kControlPopupButtonMenuHandleTag,
														sizeof(sizeMenu), (Ptr)&sizeMenu, &actualSize) == noErr)
									{
										SInt32		size = 0L;
										
										
										GetMenuItemText(sizeMenu, value, itemText);
										StringToNum(itemText, &size);
										setSizeField(size);
									}
								}
								break;
						}
					}
					break;
				
				case iFormatForegroundColorBox:
				case iFormatBackgroundColorBox:
					if (gFormatDialogDataPtr != null)
					{
						ControlRef		control = null;
						
						
						(OSErr)GetDialogItemAsControl(gFormatDialog, itemIndex, &control);
						ColorBox_UserSetColor(control);
					}
					break;
				
				case iFormatUpdateDefaultTerminalCheckBox:
					FlipCheckBox(gFormatDialog, iFormatUpdateDefaultTerminalCheckBox);
					break;
				
				default:
					break;
			}
		} while ((itemIndex != iFormatOKButton) && (itemIndex != iFormatCancelButton));
		DisposeModalFilterUPP(filterProc), filterProc = null;
		
		// undo any damage to this menu
		(Boolean)customSizeMenuSize(getSizeMenu(), 0);
		
		// handle OK or Cancel click
		if (itemIndex != iFormatCancelButton)
		{
			// kill the dialog
			setFormatFromItems(gFormatDialogDataPtr, gOldPanelIndex); // save current
			TransitionDialogClose(gFormatDialog);
			result = true;
		}
		else
		{
			// user cancelled - close the dialog with an appropriate transition for cancelling
			TransitionDialogCloseCancel(gFormatDialog);
			result = false;
		}
	}
	else Alert_ReportOSErr(memFullErr);
	
	return result;
}// Display


//
// internal methods
//

/*!
To find the item in the Size pop-up menu that
corresponds to the given size and automatically
make that the current value of the pop-up menu
(checking that size in the menu), use this
method.

(3.0)
*/
static void
checkSizeMenuSize		(MenuHandle			inStandardSizeMenu,
						 const SInt16		inSize)
{
	register SInt16 	i = 0;
	SInt32				itemFontSize = 0L;
	Str255 				itemString;
	
	
	for (i = 1; i <= CountMenuItems(inStandardSizeMenu); i++)
	{
		GetMenuItemText(inStandardSizeMenu, i, itemString);
		StringToNum(itemString, &itemFontSize);
		if (itemFontSize == inSize)
		{
			SetDialogItemValue(gFormatDialog, iFormatEditSizePopUpMenu, i);
			break;
		}
	}
}// checkSizeMenuSize


/*!
To add or remove a custom size item from a Size
menu, use this method.  Pass 0 as the size to
remove any custom item (has no effect if a
custom item is not there).  If a custom item is
already in place, the new size you specify will
replace the custom size.  This routine will
automatically check the numerical values in the
Size menu you provide, and if any of them
matches the given size, no custom item is added
and any existing custom item is automatically
removed.

If a custom item is in the menu when this
routine is finished, "true" is returned; if not,
"false" is returned.

IMPORTANT:	This method may not produce the
			expected results if you change your
			Size menu outside of the control of
			this routine.

(3.0)
*/
static Boolean
customSizeMenuSize		(MenuHandle			inStandardSizeMenu,
						 const SInt16		inCustomSizeOrZero)
{
	SInt16 			i = 0;
	SInt32			itemFontSize = 0L;
	Str255 			itemString;
	ControlRef		sizeMenuButton = null;
	Boolean			result = false;
	
	
	// get the size pop-up menu button, because its maximum must change when items are added or removed
	(OSErr)GetDialogItemAsControl(gFormatDialog, iFormatEditSizePopUpMenu, &sizeMenuButton);
	
	if (gIsCustomSizeMenuItemAdded)
	{
		// delete last two appended items (the divider and custom size)
		DeleteMenuItem(inStandardSizeMenu, CountMenuItems(inStandardSizeMenu));
		DeleteMenuItem(inStandardSizeMenu, CountMenuItems(inStandardSizeMenu));
		SetControlMaximum(sizeMenuButton, GetControlMaximum(sizeMenuButton) - 2);
		gIsCustomSizeMenuItemAdded = false;
	}
	
	if (inCustomSizeOrZero)
	{
		// check Size menu for a number identical to the given size
		gIsCustomSizeMenuItemAdded = true;
		for (i = 1; i <= CountMenuItems(inStandardSizeMenu); i++)
		{
			GetMenuItemText(inStandardSizeMenu, i, itemString);
			StringToNum(itemString, &itemFontSize);
			if (itemFontSize == inCustomSizeOrZero) gIsCustomSizeMenuItemAdded = false;
		}
	}
	else gIsCustomSizeMenuItemAdded = false;
	
	if (gIsCustomSizeMenuItemAdded)
	{
		NumToString(inCustomSizeOrZero, itemString);
		AppendMenu(inStandardSizeMenu, "\p(-");
		AppendMenu(inStandardSizeMenu, itemString);
		SetControlMaximum(sizeMenuButton, GetControlMaximum(sizeMenuButton) + 2);
	}
	result = gIsCustomSizeMenuItemAdded;
	
	return result;
}// customSizeMenuSize


/*!
To disable portions of the formatting controls based on format
dialog options, use this method.  You can individually disable
groups of controls related to setting the font, the font size,
the foreground color, and the background color.

IMPORTANT:	Any group that you do not explicitly identify as
			being disabled will become enabled after calling
			this routine.  To enable all item groups, mask off
			"kFormatDialogFormatOptionMaskDisableAllItems" in
			the given format options.

(3.0)
*/
static void
disableItemGroupsAsNeeded		(FormatDialogFormatOptions		inBasedOnTheseOptions)
{
	// initially enable everything
	ActivateDialogItem(gFormatDialog, iFormatEditMenuFontItemsPane);
	ActivateDialogItem(gFormatDialog, iFormatEditMenuFontSizeItemsPane);
	ActivateDialogItem(gFormatDialog, iFormatEditMenuForeColorItemsPane);
	ActivateDialogItem(gFormatDialog, iFormatEditMenuBackColorItemsPane);
	
	// now disable whatever was requested by the options
	if (inBasedOnTheseOptions & kFormatDialogFormatOptionDisableFontItems)
	{
		DeactivateDialogItem(gFormatDialog, iFormatEditMenuFontItemsPane);
	}
	if (inBasedOnTheseOptions & kFormatDialogFormatOptionDisableFontSizeItems)
	{
		DeactivateDialogItem(gFormatDialog, iFormatEditMenuFontSizeItemsPane);
	}
	if (inBasedOnTheseOptions & kFormatDialogFormatOptionDisableForeColorItems)
	{
		DeactivateDialogItem(gFormatDialog, iFormatEditMenuForeColorItemsPane);
	}
	if (inBasedOnTheseOptions & kFormatDialogFormatOptionDisableBackColorItems)
	{
		DeactivateDialogItem(gFormatDialog, iFormatEditMenuBackColorItemsPane);
	}
}// disableItemGroupsAsNeeded


/*!
To make the Font pop-up menu button use the font
of its current selection, use this method.  As a
convenience, this routine also invokes the
getSizeMenu() method, to make sure it is up to
date.

(3.0)
*/
static void
fontUpdate		(void)
{
	ControlFontStyleRec		styleRecord;
	ControlRef				fontMenuButton = null;
	MenuHandle				fontMenu = getFontMenu();
	Size					actualSize = 0L;
	SInt16					fontID = 0;
	SInt16					value = 0;
	Str255					itemText;
	
	
	(OSErr)GetDialogItemAsControl(gFormatDialog, iFormatEditFontPopUpMenu, &fontMenuButton);
	value = GetControlValue(fontMenuButton);
	
	GetMenuItemText(fontMenu, value, itemText);
#if TARGET_API_MAC_OS8
	GetFNum(itemText, &fontID);
#else
	fontID = FMGetFontFamilyFromName(itemText);
#endif
	styleRecord.flags = kControlUseFontMask;
	styleRecord.font = fontID;
	(OSErr)SetControlData(fontMenuButton, kControlNoPart, kControlFontStyleTag,
							sizeof(styleRecord), (Ptr)&styleRecord);
	DrawOneControl(fontMenuButton);
	
	setUpSizeMenuForFont(getSizeMenu(), itemText);
}// fontUpdate


/*!
Handles special key events in the Format
dialog box.

(3.0)
*/
static pascal Boolean
formatDialogEventFilter		(DialogPtr		inDialogPtr,
							 EventRecord*	inoutEventPtr,
							 short*			outItemIndex)
{
	Boolean		result = pascal_false,
				isButtonAndShouldFlash = false; // simplifies handling of multiple key-to-button maps
	
	
	// first, make sure help balloons show up properly for the cursor location, if Balloon Help is on
	if (IsPopUpHelpEnabled()) DisplayBalloonForDialogItemUnderMouse(inDialogPtr, inoutEventPtr->where,
																		kDialogBalloonsIDFormat);
	
	// this rigamorole is necessary to distinguish between up-arrow and down-arrow clicks!
	if (inoutEventPtr->what == mouseDown)
	{
		GrafPtr				oldPort = null;
		Point				localMouse = inoutEventPtr->where;
		ControlRef			controlHit = null,
							arrows = null,
							sizeMenuControl = null;
		ControlPartCode		part = kControlNoPart;
		
		
		GetPort(&oldPort);
		SetPortWindowPort(GetDialogWindow(inDialogPtr));
		GlobalToLocal(&localMouse);
		controlHit = FindControlUnderMouse(localMouse, GetDialogWindow(inDialogPtr), &part);
		GetDialogItemAsControl(inDialogPtr, iFormatEditSizeArrows, &arrows);
		GetDialogItemAsControl(inDialogPtr, iFormatEditSizePopUpMenu, &sizeMenuControl);
		if (controlHit == sizeMenuControl)
		{
			// ensure menu is updated prior to display, but dont otherwise intercept the event
			setSizeField(0);
		}
		else if (controlHit == arrows)
		{
			// ensure arrow value is updated prior to use
			setSizeField(0);
			
			// determine which arrow was clicked, and update controls
			part = HandleControlClick(controlHit, localMouse, inoutEventPtr->modifiers, (ControlActionUPP)null);
			switch (part)
			{
				case kControlUpButtonPart:
				case kControlDownButtonPart:
					{
						SInt16		fontSize = 0;
						
						
						// change arrows value...
						fontSize = GetControlValue(arrows);
						if ((part == kControlUpButtonPart) && (fontSize < GetControlMaximum(arrows)))
						{
							fontSize++;
						}
						else if ((part == kControlDownButtonPart) && (fontSize > GetControlMinimum(arrows)))
						{
							fontSize--;
						}
						setSizeField(fontSize);
					}
					result = pascal_true;
					break;
				
				default:
					// ???
					break;
			}
		}
		SetPort(oldPort);
	}
	
	if (!result) result = (StandardDialogEventFilter(inDialogPtr, inoutEventPtr, outItemIndex));
	else if (isButtonAndShouldFlash) FlashButton(inDialogPtr, *outItemIndex);
	
	return result;
}// formatDialogEventFilter


/*!
To acquire a handle to the menu from the Font
pop-up menu button, use this convenient method.

(3.0)
*/
static MenuHandle
getFontMenu		(void)
{
	MenuHandle		result = null;
	ControlRef		fontMenuButton = null;
	Size			actualSize = 0L;
	
	
	GetDialogItemAsControl(gFormatDialog, iFormatEditFontPopUpMenu, &fontMenuButton);
	(OSErr)GetControlData(fontMenuButton, kControlMenuPart,
							kControlPopupButtonMenuHandleTag,
							sizeof(result), (Ptr)&result, &actualSize);
	return result;
}// getFontMenu


/*!
To acquire a handle to the menu from the Size
pop-up menu button, use this convenient method.

(3.0)
*/
static MenuHandle
getSizeMenu		(void)
{
	MenuHandle		result = null;
	ControlRef		sizeMenuButton = null;
	Size			actualSize = 0L;
	
	
	GetDialogItemAsControl(gFormatDialog, iFormatEditSizePopUpMenu, &sizeMenuButton);
	(OSErr)GetControlData(sizeMenuButton, kControlMenuPart,
							kControlPopupButtonMenuHandleTag,
							sizeof(result), (Ptr)&result, &actualSize);
	return result;
}// getSizeMenu


/*!
This method moves and resizes controls in response to
a resize of the Format movable modal dialog box.

(3.0)
*/
static pascal Boolean
handleNewSize		(const WindowRef		inWindow,
					 const SInt32			inDeltaSizeX,
					 const SInt32			inDeltaSizeY,
					 const SInt32			UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_true;
	
	
	DialogAdjust_BeginDialogItemAdjustment(GetDialogFromWindow(inWindow));
	{
		SInt32		amountH = 0L,
					amountV = 0L;
		
		
		if (inDeltaSizeX)
		{
			// controls which are moved horizontally
			if (Localization_IsLeftToRight())
			{
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iFormatOKButton, inDeltaSizeX);
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iFormatCancelButton, inDeltaSizeX);
			}
			else
			{
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iFormatAppleGuideButton, inDeltaSizeX);
			}
			amountH = INTEGER_HALVED(inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iFormatEditSizePopUpMenu, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iFormatEditSizeArrows, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iFormatForegroundColorLabel, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iFormatBackgroundColorLabel, amountH);
			
			// controls which are resized horizontally
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatUpdateDefaultTerminalCheckBox, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatEditMenuPrimaryGroup, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatEditMenuFontItemsPane, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatEditMenuFontSizeItemsPane, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatEditMenuForeColorItemsPane, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatEditMenuBackColorItemsPane, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatEditFontPopUpMenu, inDeltaSizeX);
			amountH = INTEGER_HALVED(inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatEditSizeField, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatForegroundColorBox, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatForegroundColorLabel, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatBackgroundColorBox, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iFormatBackgroundColorLabel, amountH);
		}
	}
	DialogAdjust_EndAdjustment(inDeltaSizeX, inDeltaSizeY); // moves and resizes controls properly
	
	return result;
}// handleNewSize


/*!
To set up the specified format using the values of
appropriate controls in the dialog box, use this
method.

(3.0)
*/
static void
setFormatFromItems		(FormatDialogSetupDataPtr		inDataPtr,
						 const UInt16					inFormatIndexToUse)
{
	Boolean		didSet = true;
	
	
	if (inDataPtr != null)
	{
		if (inFormatIndexToUse < kFormatDialogNumberOfFormats)
		{
			RGBColor*		colorPtr = null;
			ControlRef		control = null;
			
			
			// read colors
			colorPtr = &inDataPtr->format[inFormatIndexToUse].colors.foreground;
			GetDialogItemAsControl(gFormatDialog, iFormatForegroundColorBox, &control);
			ColorBox_GetColor(control, colorPtr);
			
			colorPtr = &inDataPtr->format[inFormatIndexToUse].colors.background;
			GetDialogItemAsControl(gFormatDialog, iFormatBackgroundColorBox, &control);
			ColorBox_GetColor(control, colorPtr);
			
			// read font name
			GetDialogItemAsControl(gFormatDialog, iFormatEditFontPopUpMenu, &control);
			GetMenuItemText(getFontMenu(), GetControlValue(control),
							inDataPtr->format[inFormatIndexToUse].font.familyName);
			
			// read font size
			{
				SInt32			number = 0L;
				
				
				GetDialogItemAsControl(gFormatDialog, iFormatEditSizeField, &control);
				GetControlNumericalText(control, &number);
				inDataPtr->format[inFormatIndexToUse].font.size = number;
			}
		}
		else didSet = false;
	}
	else didSet = false;
	
	if (!didSet) BEEP;
}// setFormatFromItems


/*!
To set up the values of all controls in the group box
to use the information in the specified format, use
this method.

(3.0)
*/
static void
setItemsForFormat		(FormatDialogSetupDataPtr		inDataPtr,
						 const UInt16					inFormatIndexToUse)
{
	Boolean		didSet = true;
	
	
	if (inDataPtr != null)
	{
		if (inFormatIndexToUse < kFormatDialogNumberOfFormats)
		{
			ControlRef		control = null;
			RGBColor*		foreColorPtr = &inDataPtr->format[inFormatIndexToUse].colors.foreground;
			RGBColor*		backColorPtr = &inDataPtr->format[inFormatIndexToUse].colors.background;
			StringPtr		fontName = inDataPtr->format[inFormatIndexToUse].font.familyName;
			SInt16			fontSize = inDataPtr->format[inFormatIndexToUse].font.size;
			
			
			// update foreground color
			GetDialogItemAsControl(gFormatDialog, iFormatForegroundColorBox, &control);
			ColorBox_SetColor(control, foreColorPtr);
			DrawOneDialogItem(gFormatDialog, iFormatForegroundColorBox);
			
			// update background color
			GetDialogItemAsControl(gFormatDialog, iFormatBackgroundColorBox, &control);
			ColorBox_SetColor(control, backColorPtr);
			DrawOneDialogItem(gFormatDialog, iFormatBackgroundColorBox);
			
			setSizeField(fontSize); // update font size
			
			{
				// update font menu
				MenuItemIndex		index = Menus_GetItemIndexByText(getFontMenu(), fontName);
				
				
				SetDialogItemValue(gFormatDialog, iFormatEditFontPopUpMenu, index);
				DrawOneDialogItem(gFormatDialog, iFormatEditFontPopUpMenu);
			}
			
			disableItemGroupsAsNeeded(inDataPtr->format[inFormatIndexToUse].options);
			fontUpdate();
		}
		else didSet = false;
	}
	else didSet = false;
	
	if (!didSet) BEEP;
}// setItemsForFormat


/*!
Call this method to update all controls related
to the font size (such as the size menu and the
arrows).  Pass 0 for "inNewSize" if you want
the text field to be left alone (presumably
because you are responding to the user typing
something); this causes the current value of
the text field to be used to update all other
controls.

(3.0)
*/
static void
setSizeField		(const SInt16		inNewSizeOrZero)
{
	ControlRef		control = null;
	SInt32			size = inNewSizeOrZero;
	
	
	GetDialogItemAsControl(gFormatDialog, iFormatEditSizeField, &control);
	if (size != 0)
	{
		// set the field to the given size
		SetControlNumericalText(control, size);
		DrawOneControl(control);
	}
	else
	{
		// 0 means read the current field value
		GetControlNumericalText(control, &size);
	}
	GetDialogItemAsControl(gFormatDialog, iFormatEditSizeArrows, &control);
	SetControlValue(control, size);
	customSizeMenuSize(getSizeMenu(), size);
	checkSizeMenuSize(getSizeMenu(), size); // synchronizes the pop-up menu value
}// setSizeField


/*!
This method outlines the appropriate menu items
in the Size menu for the font indicated.

(3.0)
*/
static void
setUpSizeMenuForFont	(MenuHandle		inStandardSizeMenu,
						 StringPtr		inFontName)
{
	SInt16 			i = 0,
					fontID = 0;
	SInt32			itemFontSize = 0L;
	Str255 			itemString;
	
	
#if TARGET_API_MAC_OS8
	GetFNum(inFontName, &fontID);
#else
	fontID = FMGetFontFamilyFromName(inFontName);
#endif
	
	// set up Size menu
	for (i = 1; i <= CountMenuItems(inStandardSizeMenu); i++)
	{
		GetMenuItemText(inStandardSizeMenu, i, itemString);
		StringToNum(itemString, &itemFontSize);
		
		if (RealFont(fontID, (short)itemFontSize)) SetItemStyle(inStandardSizeMenu, i, outline);
		else SetItemStyle(inStandardSizeMenu, i, normal);
	}
}// setUpSizeMenuForFont