/*###############################################################

	IconManager.c
	
	Interface Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Appearance.h>
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Dialogs.h>
#include <Folders.h>
#include <Gestalt.h>
#include <Icons.h>
#include <MacTypes.h>
#include <Resources.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "GeneralResources.h"


// constants
typedef UInt16		IconManagerIconType;
enum
{
	kIconManagerIconTypeOS8 = 0,
	kIconManagerIconTypeOSX = 1
};

// types
struct IconManagerIcon
{
	IconManagerIconType		type;
	union
	{
		IconSuiteRef	OS8;
		IconRef			OSX;
	} data;
};
typedef struct IconManagerIcon		IconManagerIcon;
typedef IconManagerIcon*			IconManagerIconPtr;

// variables
static Boolean		gHaveIconServices = false;

// internal methods
static IconManagerIconPtr	refAcquireLock	(IconManagerIconRef		inRef);
static void					refReleaseLock	(IconManagerIconRef		inRef,
											 IconManagerIconPtr*	inoutPtr);
static void					releaseIcons	(IconManagerIconPtr		inPtr);



//
// public methods
//

/*!
To create a new abstract icon, use this method.
The icon is uninitialized; use one of the
IconManager_MakeIcon...() routines to set up an
icon using specific data.

(3.0)
*/
IconManagerIconRef
IconManager_NewIcon		(void)
{
	IconManagerIconRef		result = (IconManagerIconRef)Memory_NewPtr(sizeof(IconManagerIcon));
	
	
	if (result != null)
	{
		IconManagerIconPtr		ptr = refAcquireLock(result);
		
		
		ptr->type = kIconManagerIconTypeOS8;
		ptr->data.OS8 = null;
		ptr->data.OSX = null;
		refReleaseLock(result, &ptr);
	}
	
	// without an Init routine, this test has to be done every time an icon is created
	{
		OSErr		error = noErr;
		SInt32		gestaltResult = 0L;
		
		
		error = Gestalt(gestaltIconUtilitiesAttr, &gestaltResult);
		if (error == noErr)
		{
			// Icon Services?
			gHaveIconServices = (gestaltResult & (1 << gestaltIconUtilitiesHasIconServices));
		}
	}
	
	return result;
}// NewIcon


/*!
To destroy an abstract icon when you are finished
using it, use this method.

(3.0)
*/
void
IconManager_DisposeIcon		(IconManagerIconRef*		inoutRefPtr)
{
	if (inoutRefPtr != null)
	{
		IconManagerIconPtr		ptr = refAcquireLock(*inoutRefPtr);
		
		
		releaseIcons(ptr);
		DisposePtr((Ptr)ptr), ptr = null, *inoutRefPtr = null;
	}
}// DisposeIcon


/*!
To determine if an abstract icon reference defines an
icon based on modern Mac OS icon data, use this method.
If the application is presently running Mac OS 8.0 or
8.1, this routine will never return true.

(3.0)
*/
Boolean
IconManager_IsIconServices	(IconManagerIconRef			inRef)
{
	Boolean					result = false;
	IconManagerIconPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		result = ((ptr->type == kIconManagerIconTypeOSX) && gHaveIconServices);
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// IsIconServices


/*!
To determine if an abstract icon reference defines an
icon based on legacy Mac OS icon data, use this method.

(3.0)
*/
Boolean
IconManager_IsIconSuite		(IconManagerIconRef			inRef)
{
	Boolean					result = false;
	IconManagerIconPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		result = (ptr->type == kIconManagerIconTypeOS8);
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// IsIconSuite


/*!
To force an abstract icon to render itself based on data
in an Icon Services set, use this method.  The specified
icon reference must already exist.

If Mac OS 8.0 or 8.1 is in use, Icon Services will not be
available.  To avoid problems, this routine automatically
looks for an equivalent icon suite resource and uses it
instead.  If no icon of any kind can be found, the result
is "resNotFound".

(3.0)
*/
OSErr
IconManager_MakeIconRef		(IconManagerIconRef		inRef,
							 const SInt16			inIconServicesVolumeNumber,
							 const OSType			inIconServicesCreator,
							 const OSType			inIconServicesDescription)
{
	IconManagerIconPtr		ptr = refAcquireLock(inRef);
	OSErr					result = noErr;
	
	
	if (ptr != null)
	{
		// get rid of any previous icon data
		releaseIcons(ptr);
		
		// verify that Icon Services can be used; if it cant be, look for an equivalent icon suite
		if (gHaveIconServices)
		{
			// Icon Services is available; use it!
			ptr->type = kIconManagerIconTypeOSX;
			result = GetIconRef(inIconServicesVolumeNumber, inIconServicesCreator,
									inIconServicesDescription, &ptr->data.OSX);
		}
		else
		{
			SInt16		suiteID = 0;
			
			
			if ((inIconServicesVolumeNumber == kOnSystemDisk) &&
				(inIconServicesCreator == kSystemIconsCreator))
			{
				// NOTE:	This switch statement is infinitely expandable; Ive only
				//			included generic icons that I need.  The list should be
				//			made more complete as necessary, to include mappings for
				//			other icon suites.
				switch (inIconServicesDescription)
				{
					case kGenericApplicationIcon:
						suiteID = kGenericApplicationIconResource;
						break;
					
					case kGenericDocumentIcon:
						suiteID = kGenericDocumentIconResource;
						break;
					
					case kGenericFolderIcon:
						suiteID = kGenericFolderIconResource;
						break;
					
					case kHelpIcon:
						suiteID = kHelpIconResource;
						break;
					
					case kSortAscendingIcon:
						suiteID = kInterfaceLibIconIDSortAscending;
						break;
					
					case kSortDescendingIcon:
						suiteID = kInterfaceLibIconIDSortDescending;
						break;
					
					default:
						break;
				}
			}
			
			if (suiteID)
			{
				ptr->type = kIconManagerIconTypeOS8;
				result = GetIconSuite(&ptr->data.OS8, suiteID, kSelectorAllAvailableData);
			}
			else result = resNotFound;
		}
		refReleaseLock(inRef, &ptr);
	}
	else result = memPCErr;
	
	return result;
}// MakeIconRef


/*!
To force an abstract icon to render itself based on data
in an icon suite, use this method.  The specified icon
reference must already exist.

Since Icon Services data is more desirable, this routine
will automatically create an Icon Services icon if the
specified icon suite resource ID corresponds to an icon
that has an Icon Services equivalent, and Mac OS 8.5 or
later is in use.

(3.0)
*/
OSErr
IconManager_MakeIconSuite	(IconManagerIconRef			inRef,
							 const SInt16				inSuiteResourceID,
							 IconSelectorValue			inWhichIcons)
{
	IconManagerIconPtr		ptr = refAcquireLock(inRef);
	OSErr					result = noErr;
	
	
	if (ptr != null)
	{
		Boolean		canGetRef = false;
		
		
		// get rid of any previous icon data
		releaseIcons(ptr);
		
		// check for an Icon Services equivalent
		if (gHaveIconServices)
		{
			// first look for a system icon
			{
				SInt16		volume = kOnSystemDisk;				// most icons use this
				OSType		iconCreator = kSystemIconsCreator;	// most icons use this
				OSType		iconType = '????';
				
				
				switch (inSuiteResourceID) // look for a system icon, then
				{
					// NOTE:	This switch statement is infinitely expandable; Ive only
					//			included generic icons that I need.  The list should be
					//			made more complete as necessary, to include mappings for
					//			other Icon Services icons.
					case kGenericApplicationIconResource:
						iconType = kGenericApplicationIcon;
						break;
					
					case kGenericDocumentIconResource:
						iconType = kGenericDocumentIcon;
						break;
					
					case kGenericFolderIconResource:
						iconType = kGenericFolderIcon;
						break;
					
					case kHelpIconResource:
						iconType = kHelpIcon;
						break;
					
					case kInterfaceLibIconIDSortAscending:
						iconType = kSortAscendingIcon;
						break;
					
					case kInterfaceLibIconIDSortDescending:
						iconType = kSortDescendingIcon;
						break;
					
					default:
						break;
				}
				
				if (iconType != '????')
				{
					ptr->type = kIconManagerIconTypeOSX;
					canGetRef = (GetIconRef(volume, iconCreator, iconType, &ptr->data.OSX) == noErr);
				}
				else canGetRef = false;
			}
			
			// if that fails, look for an application icon
			if (!canGetRef)
			{
				ProcessSerialNumber		telnetItself;
				
				
				if (GetCurrentProcess(&telnetItself) == noErr)
				{
					ProcessInfoRec		info;
					FSSpec				telnetApplicationProgramFile;
					
					
					// clear the buffer to zeroes
					{
						Ptr		pointer = (Ptr)&info;
						long	size = sizeof(info);
						
						
						while (size--) *pointer++ = 0;
					}
					
					// GetProcessInformation() requires that the following
					// three fields be initialized ahead of time.
					info.processInfoLength = sizeof(info);
					info.processName = null; // dont get the name
					info.processAppSpec = &telnetApplicationProgramFile;
					
					if (GetProcessInformation(&telnetItself, &info) == noErr)
					{
						// Use the process FSSpec of the application to get at its resource
						// file.  Since every Icon Manager icon is unique, the easiest way
						// to ensure that separate references are created is to coerce the
						// resource ID into an icon type four-character code.  If a
						// static value was chosen, then all icons would be the same!!!
						if (RegisterIconRefFromResource(kApplicationCreatorSignature,
															(OSType)inSuiteResourceID,
															info.processAppSpec, inSuiteResourceID,
															&ptr->data.OSX) == noErr)
						{
							ptr->type = kIconManagerIconTypeOSX;
							canGetRef = (ptr->data.OSX != null);
						}
					}
				}
			}
		}
		
		// if no Icon Services icon can be found, use the specified icon suite resource
		if (!canGetRef)
		{
			ptr->type = kIconManagerIconTypeOS8;
			result = GetIconSuite(&ptr->data.OS8, inSuiteResourceID, inWhichIcons);
		}
		
		refReleaseLock(inRef, &ptr);
	}
	else result = memPCErr;
	
	return result;
}// MakeIconSuite


/*!
To draw an icon properly no matter how it is internally
defined, use this convenient method.  If an error occurs,
the result will depend on the nature of the icon being
drawn.

(3.0)
*/
OSErr
IconManager_PlotIcon		(IconManagerIconRef			inRef,
							 const Rect*				inBounds,
							 IconAlignmentType			inAlignment,
							 IconTransformType			inTransform)
{
	IconManagerIconPtr		ptr = refAcquireLock(inRef);
	OSErr					result = noErr;
	
	
	if (ptr != null)
	{
		switch (ptr->type)
		{
			case kIconManagerIconTypeOS8:	// treat it as an icon suite
				if (ptr->data.OS8 != null)
				{
					result = PlotIconSuite(inBounds, inAlignment, inTransform, ptr->data.OS8);
				}
				else result = memPCErr;
				break;
			
			case kIconManagerIconTypeOSX:	// use Icon Services
				if (gHaveIconServices)
				{
					if (ptr->data.OS8 != null)
					{
						result = PlotIconRef(inBounds, inAlignment, inTransform,
												kIconServicesNormalUsageFlag, ptr->data.OSX);
					}
					else result = memPCErr;
				}
				else result = cfragNoSymbolErr;
				break;
			
			default:
				break;
		}
		refReleaseLock(inRef, &ptr);
	}
	
	return result;
}// PlotIcon


/*!
To conveniently set up any Appearance control that
can take standard content types such as icons,
pictures, etc. so it uses the data defined by an
abstract Icon Manager Icon, use this method.  You
can use this, for example, to set the icon for a
bevel button.

Automatically, if a 32-bit Icon Services icon is
available (and Icon Services is available), the
better icon will be used (and the content type
for the button will be set appropriately).  If
not, an Icon Suite will be used instead.

(3.0)
*/
OSErr
IconManager_SetButtonIcon	(ControlRef				inControl,
							 IconManagerIconRef		inRef)
{
	IconManagerIconPtr		ptr = refAcquireLock(inRef);
	OSErr					result = noErr;
	
	
	if (ptr != null)
	{
		ControlButtonContentInfo		contentInfo;
		
		
		if (IconManager_IsIconServices(inRef))
		{
			contentInfo.contentType = kControlContentIconRef;
			contentInfo.u.iconRef = ptr->data.OSX;
		}
		else
		{
			contentInfo.contentType = kControlContentIconSuiteHandle;
			contentInfo.u.iconSuite = ptr->data.OS8;
		}
		result = SetControlData(inControl, kControlNoPart, kControlBevelButtonContentTag,
								sizeof(contentInfo), &contentInfo);
		refReleaseLock(inRef, &ptr);
	}
	else result = memPCErr;
	
	return result;
}// SetButtonIcon


//
// internal methods
//

/*!
To acquire a pointer to the internal structure,
given a reference to it, use this method.

(3.0)
*/
static IconManagerIconPtr
refAcquireLock	(IconManagerIconRef		inRef)
{
	return ((IconManagerIconPtr)inRef);
}// refAcquireLock


/*!
To release a pointer to the internal structure,
possessing a reference to it, use this method.
On return, your copy of the pointer is set to
"null".

(3.0)
*/
static void
refReleaseLock	(IconManagerIconRef		inRef,
				 IconManagerIconPtr*	inoutPtr)
{
	// the memory is not relocatable, so do nothing with the reference
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock


/*!
To dispose of icon data in an abstract icon,
use this method.  You generally do this just
prior to replacing the data with something
else.

(3.0)
*/
static void
releaseIcons	(IconManagerIconPtr		inPtr)
{
	if (inPtr != null)
	{
		if (inPtr->type == kIconManagerIconTypeOS8)
		{
			if (inPtr->data.OS8 != null)
			{
				(OSErr)DisposeIconSuite(inPtr->data.OS8, true/* dispose data too */);
			}
		}
		else if (inPtr->type == kIconManagerIconTypeOSX)
		{
			if (inPtr->data.OSX != null)
			{
				(OSErr)ReleaseIconRef(inPtr->data.OSX);
			}
		}
	}
}// releaseIcons