/*###############################################################

	Init.c
	
	Initialization code, called only by main().
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdio>
#	include <cstring>
#else
#	include <stdio.h>
#	include <string.h>
#endif

// Mac includes
#include <Appearance.h>
#include <AppleEvents.h>
#include <AppleGuide.h>
#include <ASRegistry.h>
#include <DateTimeUtils.h>
#include <DrawSprocket.h>
#include <Folders.h>
#include <Gestalt.h>
#include <Navigation.h>
#include <OpenTransport.h>
#include <Resources.h>
#include <Script.h>
#include <TextUtils.h>
#include <Threads.h>

// library includes
#include "ContextsLib.h"
#include "InterfaceLib.h"
#include "PrintingLib.h"

// resource includes
#include "GeneralResources.h"
#include "StringResources.h"
#include "DialogResources.h"
#include "MenuResources.h"
#include "TelnetVersion.h"

// MacTelnet includes
#include "Clipboard.h"
#include "CommandLine.h"
#include "ConnectionData.h"
#include "Console.h"
#include "ContextualMenuBuilder.h"
#include "CoolStuff.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "FileUtilities.h"
#include "FolderManager.h"
#include "FTPLog.h"
#include "InstallAE.h"
#include "InternetConfig.h"
#include "InternetServices.h"
#include "InternalEvents.h"
#include "Keypads.h"
#include "MacBinary.h"
#include "NetBackground.h"
#include "NetEvents.h"
#include "Preferences.h"
#include "RasterGraphicsKernel.h"
#include "RasterGraphicsScreen.h"
#include "ResourceFileManager.h"
#include "TektronixVirtualGraphics.h"
#include "TelnetDrag.h"
#include "TelnetErrors.h"
#include "TelnetMacros.h"
#include "TelnetMenus.h"
#include "TelnetThreads.h"
#include "Terminal.h"
#include "Terminology.h"
#include "TextTranslation.h"
#include "UNIXCommandInterpreters.h"
#include "URL.h"
#include "vsdata.h"



// public methods
void Init_StartUpTelnet	(void);

// internal methods
static void		checkMacOSVersion			(void);
static void		drawSplashScreen			(DialogRef		inSplashScreenDialog);
static void		getEnvironment				(void);
static void		initApplicationCore			(void);
static void		initMacOSToolbox			(void);
static void		networkInit					(void);



//
// public methods
//

/*!
This routine initializes everything in
MacTelnet.  See "Main.c" for the quitting
routine, which basically undoes all of this.

(3.0)
*/
void
Init_StartUpTelnet	(void)
{
	DialogRef 		splashScreenDialog = null; // splash screen dialog
	
	
#if TARGET_API_MAC_OS8
	MaxApplZone(); // change this to a stack-space specification in the 'cfrg' resource
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();
	MoreMasters();
#else
	// under Carbon, MaxApplZone() is preempted by a stack-space specification in the 'cfrg' resource, and
	// MoreMasterPointers() takes a long integer parameter specifying the number of master pointers to allocate
	MoreMasterPointers(32);
#endif
	
	FlushEvents(everyEvent, 0);
	
	initMacOSToolbox();
	getEnvironment();
	Threads_Init(); // create threads while memory is relatively unfragmented, and AFTER MaxApplZone()
	initApplicationCore();
	
	// display initialization progress
	DrawMenuBar();
	splashScreenDialog = GetNewDialog(FirstDLOG, null, kInFront); // splash screen
	ShowWindow(GetDialogWindow(splashScreenDialog));
	drawSplashScreen(splashScreenDialog);
	Cursors_UseWatch();
	
	ScreenManager_Init(); // must be before RealScreen_Init()
	Keypads_Init();
	Menus_Init(); // will also build the font-list (if deemed necessary) and the translation table list
	
	if (Terminal_Init(ScreenManager_GetMaximumScreenCount())) // initialize the VS (which also inits RS stuff)
	{
		// initialization failed - there is not enough memory
		Errors_DisplayStopQuitMessage(rStringsMemoryErrors, siMemErrorFatalOutOfMemory, -1); // does not return
	}
	Clipboard_Init();
	CommandLine_Init();
	CommandLine_AddInterpreter(UNIXCommandInterpreters_help);
	CommandLine_AddInterpreter(UNIXCommandInterpreters_telnet);
	
	// Initialize FTP log.  WARNING!!!  The FTP log has a
	// defined screen ID of 0, which allows all other
	// session windows to have 1-based screen IDs.  For
	// debugging purposes, the console window steals the
	// ID of 0, and the FTP Log therefore CANNOT be used
	// when the debugging console is active.  For this
	// reason, the preprocessor directive below exists.
	//#ifdef NDEBUG
	#if 1
	{
		Str255		logName;
		
		
		FTPLog_Init();
		GetWTitle(FTPLog_GetWindow(), logName);
		StringUtilities_PToCInPlace(logName);
		FTPLog_WriteString(APPLICATION_FULL" (Mac) -- ");
		FTPLog_WriteLine(STATIC_CAST(logName, char*));
		FTPLog_WriteLine("");
	}
	#else
	{
		char			dateString[40],
						timeString[40];
		unsigned long	date = 0L;
		
		
		// initialize the console and write initial information (the
		// date and time, the console name, the web site URL, etc.)
		Console_Init();
		GetDateTime(&date);
		DateString(date, longDate, STATIC_CAST(dateString, StringPtr), null/* intlHandle */);
		StringUtilities_PToCInPlace(STATIC_CAST(dateString, StringPtr));
		TimeString(date, true/* want seconds */, STATIC_CAST(timeString, StringPtr), null/* intlHandle */);
		StringUtilities_PToCInPlace(STATIC_CAST(timeString, StringPtr));
		strcat(dateString, " at ");
		strcat(dateString, timeString);
		Console_WriteLine(dateString);
		Console_WriteLine(APPLICATION_FULL" Debugging Console");
		Console_WriteLine("download updates from http://www.mactelnet.com/");
	}
	#endif
	
	TektronixVirtualGraphics_Init(); // TEK setup
	RasterGraphicsKernel_Init(); // ICR setup
	
	init_mb_files(); // initialize files to a closed state
	networkInit(); // initialize networking (networkInit() is an internal method in this file)
	Drag_Init(); // set up Drag & Drop
	{
		Size 		junk = 0L;
		
		
		MaxMem(&junk);
	}
	
	#ifndef NDEBUG
	// write an initial header to the console that describes the users runtime environment
	{
		char		s[255];
		long		gestaltResult = 0L;
		
		
		// useful values
		if (Gestalt(gestaltSystemVersion, &gestaltResult) != noErr)
		{
			strcpy(s, "Could not find Mac OS version!");
		}
		else
		{
			UInt8		major = 0;
			
			
			major = Releases_GetMajorRevisionForVersion(gestaltResult);
			if (major <= 9)
			{
				sprintf(s, "This computer is running Mac OS %d.%d.%d.",
						major, Releases_GetMinorRevisionForVersion(gestaltResult),
						Releases_GetSuperminorRevisionForVersion(gestaltResult));
			}
			else
			{
				strcpy(s, "This computer is running Mac OS X.");
			}
		}
		Console_WriteLine(s);
		
		// state the runtime environment MacTelnet was compiled for
		sprintf(s, "This version is intended for the %s Mac OS runtime environment.",
		#if TARGET_API_MAC_OS8
			"Classic"
		#elif TARGET_API_MAC_CARBON
			"Carbon"
		#else
			// some day, far in the future...
			"Cocoa"
		#endif
			);
		Console_WriteLine(s);
		
		// check for important technologies
		{
			OSErr		error = noErr;
			NumVersion	versionInfo;
			
			
			// Open Transport
			error = Gestalt(gestaltOpenTpt, &gestaltResult);
			if (error == noErr)
			{
				error = Gestalt(gestaltOpenTptVersions, (long*)&versionInfo);
				if (error == noErr)
				{
					sprintf(s, "Open Transport %d.%d.%d:", versionInfo.majorRev,
							(versionInfo.minorAndBugRev & 0xF0) >> 4,
							(versionInfo.minorAndBugRev & 0x0F));
				}
				else sprintf(s, "Open Transport 1.0:");
				Console_WriteLine(s);
				
				sprintf(s, "Open Transport Internet is%s installed.",
						((error == noErr) && (gestaltResult & gestaltOpenTptTCPPresentMask)) ? "" : " not");
				Console_WriteLine(s);
				
				sprintf(s, "Open Transport Internet is%s loaded.",
						((error == noErr) && (gestaltResult & gestaltOpenTptTCPLoadedMask)) ? "" : " not");
				Console_WriteLine(s);
			}
			else
			{
				sprintf(s, "Open Transport is not installed at all?!?!?");
				Console_WriteLine(s);
			}
			
			// AppleScript
			error = Gestalt(gestaltAppleScriptVersion, &gestaltResult);
			if (error == noErr)
			{
				sprintf(s, "AppleScript %d.%d.%d.",
						Releases_GetMajorRevisionForVersion(gestaltResult),
						Releases_GetMinorRevisionForVersion(gestaltResult),
						Releases_GetSuperminorRevisionForVersion(gestaltResult));
			}
			else sprintf(s, "AppleScript (unknown version!):");
			Console_WriteLine(s);
		}
		sprintf(s, "Thread Manager          is%s installed.",
				FlagManager_Test(kFlagThreadManager) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "Text Services Manager   is%s installed.",
				FlagManager_Test(kFlagTextServicesManager) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "Appearance 1.1 or later is%s installed.",
				FlagManager_Test(kFlagAppearance1_1API) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "Icon Services           is%s installed.",
				FlagManager_Test(kFlagIconServices) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "Navigation Services     is%s installed.",
				FlagManager_Test(kFlagNavigationServices) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "Contextual Menu Manager is%s installed.",
				FlagManager_Test(kFlagContextualMenus) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "QuickTime               is%s installed.",
				FlagManager_Test(kFlagQuickTime) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "Draw Sprocket           is%s installed.",
				FlagManager_Test(kFlagDrawSprocket) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "MacTelnet Help          is%s available from the Help menu.",
				FlagManager_Test(kFlagAppleGuideFiles) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "Carbon Control Manager  is%s installed.",
				FlagManager_Test(kFlagCarbonControlManager) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "Carbon Menu Manager     is%s installed.",
				FlagManager_Test(kFlagCarbonMenuManager) ? "" : " not");
		Console_WriteLine(s);
		sprintf(s, "Carbon Window Manager   is%s installed.",
				FlagManager_Test(kFlagCarbonWindowManager) ? "" : " not");
		Console_WriteLine(s);
	}
	//Console_SetVisible(true); // when debugging, this is convenient
	#endif
	
	// automatically install MacTelnet Help, if necessary and if possible
	{
		FSSpec		applicationFolder,
					systemHelpFolder;
		OSErr		error = noErr;
		
		
		// The Folder Manager can provide a useful file specification consisting
		// only of the volume number and directory ID of the applications folder,
		// without the name of the folder.  This makes it easy to construct a
		// subsequent file specification for the About file, which is INSIDE the
		// applications folder and requires the applications folder ID.
		Console_WriteLine("valid application and Help folders?"); // tmp
		if ((FolderManager_GetFSSpec(kTelnetFolderApplication, &applicationFolder) == noErr) &&
			(FolderManager_GetFSSpec(kTelnetFolderMacHelp, &systemHelpFolder) == noErr))
		{
			FSSpec		helpFolderInApplicationFolder;
			Str255		applicationHelpFolderName;
			
			
			// set up specification for MacTelnet Help folder in MacTelnets folder
			// (remember, the Folder Manager returns an FSSpec with a parent ID
			// equal to the folder and an empty name string, so the code below works)
			GetIndString(applicationHelpFolderName, rStringsImportantFolderNames,
							siImportantFolderNameApplicationHelpSystem);
			
			// if necessary, install MacTelnet Help; if the folder cant be found in
			// MacTelnets folder, assume it was installed before; so, basically, if
			// any kind of error (including "fnfErr") is returned, stop immediately
			Console_WriteLine("installation of MacTelnet Help needed?"); // tmp
			error = FSMakeFSSpec(applicationFolder.vRefNum, applicationFolder.parID,
									applicationHelpFolderName, &helpFolderInApplicationFolder);
			if (error == noErr)
			{
				FSSpec		helpFolderInSystemHelpFolder;
				
				
				error = FSMakeFSSpec(systemHelpFolder.vRefNum, systemHelpFolder.parID,
										applicationHelpFolderName, &helpFolderInSystemHelpFolder);
				if (error == noErr)
				{
					FSSpec		trashFolder;
					
					
					// the folder exists both in the Help folder and in MacTelnets folder;
					// assume that the existing MacTelnet Help folder in the system-wide
					// Help folder is out of date (delete it)
					Console_WriteValueFile("deleting", &helpFolderInSystemHelpFolder); // tmp
					error = FolderManager_GetFSSpec(kTelnetFolderMacTrash, &trashFolder);
					if (error == noErr)
					{
						error = FSpCatMove(&helpFolderInSystemHelpFolder, &trashFolder);
					}
					
					if (error != noErr)
					{
						// tried to be nice, didnt work; do an immediate delete instead
						// (but even this could fail...)
						error = FSpDelete(&helpFolderInSystemHelpFolder);
					}
				}
				if ((error == noErr) || (error == fnfErr))
				{
					// move the MacTelnet Help folder from MacTelnets installation into the
					// Help folder; the next time MacTelnet runs, therefore, the folder will be
					// missing, and MacTelnet will skip this entire installation process; this
					// is a minor process, so the error code is completely ignored
					error = FSpCatMove(&helpFolderInApplicationFolder, &systemHelpFolder);
					Console_WriteValueFile("moved", &helpFolderInApplicationFolder); // tmp
					Console_WriteValueFile("to", &systemHelpFolder); // tmp
					Console_WriteValue("...or did we? error", error); // tmp
				}
			}
		}
	}
	
	// restore the visible states of certain MacTelnet windows, states which were implicitly saved at last Quit
	{
		Boolean		windowIsVisible = false;
		Size		actualSize = 0L;
		
		
		// get visibility preference for the Clipboard
		unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
									kTelnetPreferenceTag_WasClipboardShowing,
									sizeof(windowIsVisible), &windowIsVisible,
									&actualSize) == kTelnetPrefsResultCodeSuccess)
		{
			windowIsVisible = false; // assume invisible if the preference cant be found
		}
		Clipboard_SetVisible(windowIsVisible);
		
		// get visibility preference for the Message Window
		unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
									kTelnetPreferenceTag_WasCommandLineShowing,
									sizeof(windowIsVisible), &windowIsVisible,
									&actualSize) == kTelnetPrefsResultCodeSuccess)
		{
			windowIsVisible = false; // assume invisible if the preference cant be found
		}
		CommandLine_SetVisible(windowIsVisible);
	}
	
	// set up a thread to handle idle processing (has no effect if Thread Manager isnt installed - see "Main.c")
	Threads_NewCoOp(NewThreadEntryUPP(Threads_StandardIdle),
					null/* parameter */, INTEGER_KILOBYTES(40)/* extra stack size*/,
					kTelnetThreadDescriptorGeneralIdling);
	
	// Open every file in the Startup Items folder, by sending Apple Events back
	// to MacTelnet for each one.
	{
		enum
		{
			kMaximumStartupItems = 10		// no more than this many startup files will be openable
		};
		Str255		startupItemsFolderName;
		FSSpec		startupItemsFolder;
		OSErr		error = noErr;
		
		
		GetIndString(startupItemsFolderName, rStringsImportantFolderNames,
						siImportantFolderNameApplicationStartupItems);
		
		if (FolderManager_GetFSSpec(kTelnetFolderStartupItems, &startupItemsFolder) == noErr)
		{
		#define NOTHING_IS_SCREWED_UP_WITH_FOLDER_SCANS 1
		#if NOTHING_IS_SCREWED_UP_WITH_FOLDER_SCANS
			FSSpec*				array = (FSSpec*)Memory_NewPtr(kMaximumStartupItems * sizeof(FSSpec));
			UInt32				arrayLength = kMaximumStartupItems;
			register SInt32		i = 0L;
			FileInfo			info;
			Boolean				useFile = true;
			
			
			Console_WriteValueFile("startup items folder", &startupItemsFolder);
			error = FileUtilities_GetAllFilesInDirectory(&startupItemsFolder, array, &arrayLength);
			//if (error != eofErr) Alert_ReportOSErr(error); // debugging
			
			// Iterate through all found files, opening each via Apple Events.
			if (error == noErr) for (i = 0; i < arrayLength; i++)
			{
				useFile = true;
				if (FSpGetFInfo(&array[i], (FInfo*)&info) == noErr)
				{
					// skip invisible files
					useFile = (!(info.finderFlags & kIsInvisible));
				}
				if (useFile) (OSErr)FileUtilities_OpenDocument(&array[i]);
			}
		#endif
		}
	}
	
	Cursors_UseArrow();
	
	DisposeDialog(splashScreenDialog); // remove the splash screen
	
	// tests - temporary
	//FlagManager_Set(kFlagCarbonWindowManager, false);
	//FlagManager_Set(kFlagCarbonControlManager, false);
	//FlagManager_Set(kFlagCarbonMenuManager, false);
}// StartUpTelnet


//
// internal methods
//

/*!
At least Mac OS 8.1 is required.  This routine
makes sure that the user has the minimum system
software version installed - if not, an error
message is displayed and this routine does not
return.

(3.0)
*/
static void
checkMacOSVersion	(void)
{
	OSErr		error = noErr;
	long		sysv = 0L;
	
	
	error = Gestalt(gestaltSystemVersion, &sysv);
	if (error == envVersTooBig)
	{
		Errors_DisplayStopQuitMessage(rStringsStartupErrors, siWrongSysEnvironsVersion, -1); // does not return
	}
	else
	{
		UInt8		majorRev = 0,
					minorRev = 0;
		
		
		majorRev = Releases_GetMajorRevisionForVersion(sysv);
		minorRev = Releases_GetMinorRevisionForVersion(sysv);
		
		if ((majorRev < 0x08) || ((majorRev == 0x08) && (minorRev < 0x01)))
		{
			Errors_DisplayStopQuitMessage(rStringsStartupErrors, siMacOSVersionTooOld, -1); // does not return
		}
	}
}// checkMacOSVersion


/*!
To draw the contents of the splash screen window,
use this method.  MacTelnet draws a random message
string from its resource file on top of the
picture backdrop.

(3.0)
*/
static void
drawSplashScreen	(DialogRef		inSplashScreenDialog)
{
	ColorPenState		state;
	GrafPtr				oldPort = null,
						currentPort = null;
	CGrafPtr			currentColorPort = null;
	SInt16				preservedFontID = 0;
	SInt16				preservedFontSize = 0;
	Style				preservedFontStyle = 0;
	Rect				portRect;
	
	
	BeginUpdate(GetDialogWindow(inSplashScreenDialog));
	DrawDialog(inSplashScreenDialog);
	EndUpdate(GetDialogWindow(inSplashScreenDialog));
	
	// use window port
	GetPort(&oldPort);
	currentColorPort = GetWindowPort(GetDialogWindow(inSplashScreenDialog));
	SetPortWindowPort(GetDialogWindow(inSplashScreenDialog));
	GetPort(&currentPort);
	GetPortBounds(GetWindowPort(GetDialogWindow(inSplashScreenDialog)), &portRect);
	
	// preserve state
	preservedFontID = GetPortTextFont(currentPort);
	preservedFontSize = GetPortTextSize(currentPort);
	preservedFontStyle = GetPortTextFace(currentPort);
	ColorUtilities_GetColorAndPenState(&state);
	ColorUtilities_NormalizeColorAndPen();
	
	// draw text biline
	{
		enum
		{
			kStartupTextFontSize = 26
		};
		Str255			text;
		SInt32			stringCount = 0L;
		UInt16			randomIndex = 1;
		UInt16			stringWidth = 0;
		
		
		// Read the number of strings from the resource - this makes ResEdit hackers happy,
		// since they can add as many random strings as they like to the 'STR#' resource.
		GetIndString(text, rStringsRandomStartupMessages, siRandomStartupMessagesCountOfRemainingStrings);
		StringToNum(text, &stringCount);
		
		// pick a random value; note that modulo (%) only works well for a reasonably small string count
		SetQDGlobalsRandomSeed(TickCount());
		{
			SInt32		random = Random();
			
			
			// skip the string indices used for special purposes (the first two)
			randomIndex = INTEGER_ABSOLUTE(random) % stringCount + 3;
		}
		{
			Str255		fontName;
			
			
			GetIndString(fontName, rStringsRandomStartupMessages, siRandomStartupMessagesFontName);
			TextFontByName(fontName);
		}
		TextSize(kStartupTextFontSize);
		TextFace(normal);
		GetIndString(text, rStringsRandomStartupMessages, randomIndex);
		stringWidth = StringWidth(text);
		if ((portRect.right - portRect.left) > stringWidth) // avoid crashing bug in QuickDraw
		{
			SInt16		minimumFontSize = 0;
			Point		initialPenLoc;
			
			
			// draw the shadow in black
			if (FlagManager_Test(kFlagAppearance1_1API))
			{
				(OSStatus)SetThemeTextColor(kThemeTextColorBlack,
											ColorUtilities_GetCurrentDepth(currentColorPort),
											ColorUtilities_IsColorDevice(GetMainDevice()));
			}
			else
			{
				RGBColor	black;
				
				
				black.red = black.green = black.blue = 0;
				RGBForeColor(&black);
			}
			
			// draw in an open space that doesnt overrun the title text itself
			SetPt(&initialPenLoc, INTEGER_HALVED(portRect.right - portRect.left - stringWidth), 155);
			MoveTo(initialPenLoc.h, initialPenLoc.v);
			DrawString(text);
			
			// draw the text in white
			if (FlagManager_Test(kFlagAppearance1_1API))
			{
				(OSStatus)SetThemeTextColor(kThemeTextColorWhite,
											ColorUtilities_GetCurrentDepth(currentColorPort),
											ColorUtilities_IsColorDevice(GetMainDevice()));
			}
			else
			{
				RGBColor	white;
				
				
				white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
				RGBForeColor(&white);
			}
			MoveTo(initialPenLoc.h, initialPenLoc.v);
			Move(-1, -1); // offset to create the shadow
			DrawString(text);
		}
	}
	
	// restore state
	TextFont(preservedFontID);
	TextSize(preservedFontSize);
	TextFace(preservedFontStyle);
	SetPort(oldPort);
	ColorUtilities_SetColorAndPenState(&state);
}// drawSplashScreen


/*!
This method sets up some flags based on the users
runtime environment (OS version, etc.).

(3.0)
*/
static void
getEnvironment	(void)
{
	OSErr		error = noErr;
	Boolean 	HFSflag = false;
	long		response = 0L;
	
	
	// determines if the user has the required system software, and quits with an error if not
	checkMacOSVersion();
	
	// QuickTime?
	{
		long		qtv = 0L;
		//UInt8		majorRev = 0,
		//			minorRev = 0;
		
		
		error = Gestalt(gestaltQuickTimeVersion, &qtv);
		FlagManager_Set(kFlagQuickTime, (error == noErr));
		
		//majorRev = Releases_GetMajorRevisionForVersion(qtv);
		//minorRev = Releases_GetMinorRevisionForVersion(qtv);
		
		// (if specific QuickTime version checks are needed, do them here, using "majorRev" and "minorRev")
	}
	
	// contextual menus?
	FlagManager_Set(kFlagContextualMenus, (ContextsAvailable() &&
											(InitContextualMenus != (void*)kUnresolvedCFragSymbolAddress)));
	
	// Navigation Services (any version)?
#if TARGET_API_MAC_OS8
	FlagManager_Set(kFlagNavigationServices, NavServicesAvailable());
#else
	FlagManager_Set(kFlagNavigationServices, true); // Carbon definitely has Navigation Services!
#endif
	if (FlagManager_Test(kFlagNavigationServices)) NavLoad();
	
	// Mac OS X in Classic mode?
	{
		long		gestaltResult = 0L;
		
		
		error = Gestalt(gestaltMacOSCompatibilityBoxAttr, &gestaltResult);
		FlagManager_Set(kFlagOSXClassic,
						(error == noErr) && ((gestaltResult & (1 << gestaltMacOSCompatibilityBoxPresent)) != 0));
	}
}// getEnvironment


/*!
This method initializes key modules (both
internally and from MacTelnets libraries),
installs Apple Event handlers for the
Apple Events that MacTelnet supports and
requires, and reads user preferences.

If errors occur at this stage, this method
may not return, displaying an error message
to the user.

(3.0)
*/
static void
initApplicationCore		(void)
{
	// suck out events from the queue at startup time (if the user was
	// clicking the mouse madly when the splash screen was up, MacTelnet
	// can hardly do anything useful with those events)
	EventLoop_Init();
	{
		EventRecord			discardedEvent;
		register SInt16 	i = 0;
		
		
		for (i = 1; i < 12/* arbitrary */; i++) EventAvail(everyEvent, &discardedEvent);
	}
	
	// set up the Localization module, and define where all user interface resources should come from
	{
		UInt32		flags = 0L;
		
		
		if (GetScriptVariable(GetScriptManagerVariable(smSysScript), smScriptRight))
		{
			// then the text reads from the right side of a page to the left (the opposite of North America)
			flags |= kLocalization_FlagReadTextRightToLeft;
		}
		if (false/* I cant think of an appropriate Script Manager variable for determining this */)
		{
			// then the text reads from the bottom of a page to the top (the opposite of North America)
			flags |= kLocalization_FlagReadTextBottomToTop;
		}
		Localization_Init(flags);
	}
	
	// start up DrawSprocket, if its available
	FlagManager_Set(kFlagDrawSprocket, API_AVAILABLE(DSpStartup)); // is DrawSprocket available?
	(OSStatus)CoolStuff_Init();
	Sound_Init();
	
	// determine help system attributes
	#if TARGET_API_MAC_OS8
	{
		Boolean		available = ContextsAvailable(); // the ContextsUtilityLib is required for Apple Guide access
		
		
		if (available)
		{
			long		gestaltResult = 0L;
			
			
			// check if Apple Guide is installed
			if (Gestalt(gestaltHelpMgrAttr, &gestaltResult) == noErr)
			{
				available = ((gestaltResult & (1 << gestaltAppleGuidePresent)) == (1 << gestaltAppleGuidePresent));
			}
			else
			{
				available = false;
			}
			
			if (available)
			{
				// check if guide files are installed
				//UInt32		databaseFlags = 0L;
				short		vRefNum = -LMGetSFSaveDisk();
				long		dirID = LMGetCurDirStore();
				
				
				available = (AGFileGetDBCount(vRefNum, dirID, kAGFileDBTypeAny, false/* want Mixin */) > 0);
				//databaseFlags = AGGetAvailableDBTypes();
				//available = ((databaseFlags & kAGDBTypeBitHelp) != 0);
			}
		}
		FlagManager_Set(kFlagAppleGuideFiles, available);
		ContextualMenuBuilder_SetMenuHelpType(FlagManager_Test(kFlagAppleGuideFiles)
												? kCMHelpItemAppleGuide
												: kCMHelpItemNoHelp);
	}
	#else
	FlagManager_Set(kFlagAppleGuideFiles, false); // Carbon does not support Apple Guide
	ContextualMenuBuilder_SetMenuHelpType(kCMHelpItemAppleGuide);
	#endif
	
	// initialize cursors, and set up a cursor animation thread (has no effect if Thread Manager isnt available)
	Cursors_Init();
	Threads_NewCoOp(NewThreadEntryUPP(Threads_InterfaceIdle),
					null/* parameter */, 0/* extra stack size*/,
					kTelnetThreadDescriptorCursorAnimation);
	
	// get preferences, or create defaults
	Preferences_Init();
	{
		TelnetPrefsResultCode		prefsStatus = Preferences_CreateOrFindFile();
		
		
		switch (prefsStatus)
		{
			case kTelnetPrefsResultCodeBadVersionDataNewer:
				// warn user that some settings in the preferences file may have been lost - unimplemented
				break;
			
			case kTelnetPrefsResultCodeBadVersionDataOlder:
				// warn user that default values will be used for some settings - unimplemented
				break;
			
			case kTelnetPrefsResultCodeSuccess:
			default:
				// file successfully created or read; do nothing
				break;
		}
		Macros_Init(); // set up macro preferences
	}
	
	// initialize some other flags
	FlagManager_Set(kFlagSuspended, false); // initially, the application is active
	FlagManager_Set(kFlagMacBinaryIIEnabled, false); // alternative allows changing state w/o changing preferences
	FlagManager_Set(kFlagNetworkSuspended, false); // is Scroll Lock on?
	
	// look for a terminal bell sound file, and if it exists, open it and save its file number;
	// currently, MacTelnet does this by looking for a particular sound file or alias in its
	// preferences directory
	{
		FSSpec		bellFolder,
					soundFile;
		OSErr		error = noErr;
		SInt16		resourceFileID = -1;
		
		
		if (FolderManager_GetFSSpec(kTelnetFolderPreferences, &bellFolder) == noErr)
		{
			Str255		soundFileName;
			
			
			// create a file specification for the Bell file, which should be in the MacTelnet Preferences folder
			GetIndString(soundFileName, rStringsImportantFileNames, siImportantFileNameTerminalBellSound);
			error = FSMakeFSSpec(bellFolder.vRefNum, bellFolder.parID, soundFileName, &soundFile);
			if (error == noErr)
			{
				error = Navigation_EnsureOriginalFile(&soundFile); // resolve any aliases
				if (error == noErr) resourceFileID = FSpOpenResFile(&soundFile, fsCurPerm);
			}
		}
		ResourceFileManager_SetResFile(kTelnetResourceFileTerminalBellSound, resourceFileID);
	}
}// initApplicationCore


/*!
This method starts up all of the Mac OS tool sets
needed by MacTelnet 3.0, and allocates a memory
reserve.

If errors occur at this stage, this method may
not return.  If this routine actually returns,
then the users computer is at least minimally
capable of running MacTelnet 3.0.  However, more
tests are done in initApplicationCore()...

(3.0)
*/
static void
initMacOSToolbox	(void)
{
	FlagManager_Init();
	
	// register as an Appearance client, and determine if a more advanced API is available
	{
		long		appv = 0L;
		UInt8		majorRev = 0,
					minorRev = 0;
		
		
		if (Gestalt(gestaltAppearanceVersion, &appv) == noErr)
		{
			RegisterAppearanceClient();
			
			majorRev = Releases_GetMajorRevisionForVersion(appv);
			minorRev = Releases_GetMinorRevisionForVersion(appv);
			
			// any advanced APIs available?
			FlagManager_Set(kFlagAppearance1_1API, (((majorRev == 0x01) && (minorRev >= 0x01)) || (majorRev > 0x01)));
		}
		else
		{
			// MacTelnet requires Appearance
			BEEP;
			ExitToShell();
		}
	}
	
#if TARGET_API_MAC_OS8
	if (ContextsAvailable() && (InitContextualMenus != (void*)kUnresolvedCFragSymbolAddress))
	{
		InitContextualMenus();
	}
	InitGraf(GetQDGlobalsThePort()); // CARBON does not require this at all
#endif
	
	// At least Mac OS 8 is now required.  Is Mac OS 8.5 or
	// later being used?  If so, use the cool new stuff
	// where possible.
	{
		long		gestaltResult = 0L;
		UInt8		majorRev = 0,
					minorRev = 0;
		
		
		(OSErr)Gestalt(gestaltSystemVersion, &gestaltResult);
		majorRev = Releases_GetMajorRevisionForVersion(gestaltResult);
		minorRev = Releases_GetMinorRevisionForVersion(gestaltResult);
		
		// any advanced APIs available?
		FlagManager_Set(kFlagOS8_5API, (((majorRev == 0x08) && (minorRev >= 0x05)) || (majorRev > 0x08)));
		FlagManager_Set(kFlagOS8_6API, (((majorRev == 0x08) && (minorRev >= 0x06)) || (majorRev > 0x08)));
		FlagManager_Set(kFlagOS9_0API, (((majorRev == 0x09) && (minorRev >= 0x00)) || (majorRev > 0x09)));
		FlagManager_Set(kFlagOS9_1API, (((majorRev == 0x09) && (minorRev >= 0x01)) || (majorRev > 0x09)));
		FlagManager_Set(kFlagCarbonControlManager, (Gestalt(gestaltControlMgrAttr, &gestaltResult) == noErr));
		FlagManager_Set(kFlagCarbonMenuManager, (Gestalt(gestaltMenuMgrAttr, &gestaltResult) == noErr));
		FlagManager_Set(kFlagCarbonWindowManager, (Gestalt(gestaltWindowMgrAttr, &gestaltResult) == noErr));
	}
	
#if TARGET_API_MAC_OS8
	// CARBON does not require these at all
	InitFonts();
	{
		long		gestaltResult = 0L;
		
		
		if (Gestalt(gestaltWindowMgrAttr, &gestaltResult) != noErr) gestaltResult = 0L;
		if (gestaltResult & gestaltHasFloatingWindowsMask)
		{
			InitFloatingWindows();
		}
		else
		{
			InitWindows();
		}
	}
	InitMenus();
	TEInit();
	InitDialogs(null);
	
	// initialize the Text Services Manager
	FlagManager_Set(kFlagTextServicesManager, (InitTSMAwareApplication() == noErr));
#endif
	InitCursor();
	
	// determine if Icon Services is installed
	{
		long		gestaltResult = 0L;
		
		
		(OSErr)Gestalt(gestaltIconUtilitiesAttr, &gestaltResult);
		FlagManager_Set(kFlagIconServices, gestaltResult & (1 << gestaltIconUtilitiesHasIconServices));
	}
	
	// initialization of the Alert module must be done here, otherwise startup error alerts cant be displayed
	ResourceFileManager_SetResFile(kTelnetResourceFilePreferences, -1);
	ResourceFileManager_SetResFile(kTelnetResourceFileApplication, CurResFile());
	Alert_Init(CurResFile());
	Errors_Init();
	
	// initialize the Undoables module
	{
		Str255		undoName,
					redoName;
		
		
		GetIndString(undoName, rStringsMenuItemTextSubstitutions, siMenuItemTextUndoDefault);
		GetIndString(redoName, rStringsMenuItemTextSubstitutions, siMenuItemTextRedoDefault);
		Undoables_Init(kUndoHandlingMechanismOnlyOneUndo, undoName, redoName);
	}
	
	InternetServices_Init(); // may not return if an error occurs
	
	// initialize MacTelnets memory module, providing a grow zone for emergency situations
	Memory_Init(INTEGER_KILOBYTES(8)/* cushion*/, INTEGER_KILOBYTES(40)/* reserve*/);
	
	// without Apple Events, MacTelnet is hosed - make sure they initialize correctly
	unless (InstallAE_Init())
	{
		// this method does not return
		Errors_DisplayStopQuitMessage(rStringsStartupErrors, siUnableToInstallAppleEventHandlers, -1);
	}
}// initMacOSToolbox


/*!
This method is responsible for initializing
network data, such as variables, procedure
pointers and FTP.  As of MacTelnet 3.0, no attempt
is made to open a connection here unless the FTP
server is supposed to start up.  Instead,
connections are attempted as they are needed.
This decision positively impacts dial-up users,
because it is then possible to use MacTelnet
without necessarily being connected to the net.

(2.6)
*/
static void
networkInit		(void)
{
	Size 		junk = 0L;
	
	
	InternetConfig_Init();
	MaxMem(&junk); // keeps DNR locked high in memory
	NetEvents_Init(); // initializes for error messages to count
	Network_Init(); // initialize Open Transport stuff
	NewFTPServerMode(kFTPServerModeOff/* old mode */, gFTPServerPrefs->serverMode/* new mode */);
}// networkInit