/*###############################################################

	Inspector.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <Appearance.h>
#include <ControlDefinitions.h>
#include <Menus.h>
#include <Palettes.h>
#include <PLStringFuncs.h>
#include <Resources.h>
#include <QuickDraw.h>
#include <TextUtils.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "Inspector.h"
#include "EventLoop.h"
#include "MacLook.h"
#include "ResourceFileManager.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



// types
struct Inspector
{
	WindowRef			window;
	WindowInfoRef		windowInfo;
	ControlRef			root,
						placard;
	Boolean				isSuspended,
						isShowing;
};
typedef struct Inspector	Inspector;
typedef Inspector*			InspectorPtr;

// variables
static InspectorPtr		gSessionInspectorPtr = null;

// internal methods
static WindowRef				getWindow				(InspectorPtr			inPtr);
static pascal Boolean			handleNewSize			(WindowRef const		inWindow,
														 SInt32 const			inDeltaSizeX,
														 SInt32 const			inDeltaSizeY,
														 SInt32 const			inData);
static Boolean					isVisible				(InspectorPtr			inPtr);
static void						setUpControls			(InspectorPtr			inPtr);
static void						setVisible				(InspectorPtr			inPtr,
														 Boolean				inIsVisible);



//
// public methods
//

/*!
Use this method to create a new Inspector windoid.
It is your responsibility to set a valid window
descriptor by using Inspector_GetWindowInfo()
(if desired), and your responsibility to set the
window location properly.

(3.0)
*/
void
Inspector_Init	(void)
{
	gSessionInspectorPtr = (InspectorPtr)Memory_NewPtr(sizeof(Inspector));
	if (gSessionInspectorPtr != null)
	{
		InspectorPtr	ptr = gSessionInspectorPtr;
		Str255			windowName;
		Rect			bounds;
		
		
		SetRect(&bounds, 0, 0, 0, 0);
		
		// create the windoid
		GetIndString(windowName, rStringsMiscellaneous, siWindoidNameSessionInspector);
		if (FlagManager_Test(kFlagCarbonWindowManager))
		{
			OSErr		error = noErr;
			
			
			error = CreateNewWindow(kFloatingWindowClass, kWindowStandardFloatingAttributes | kWindowResizableAttribute,
									&bounds, &ptr->window);
			if (error == noErr) SetWTitle(ptr->window, windowName);
		}
		else
		{
			ptr->window = NewCWindow(null, &bounds, windowName,
									false/* visible */,
									kWindowFloatGrowProc,
									kInFront,
									true/* close  box */,
									0L/* reference constant - MUST be 0 for Window Info */);
		}
		
		// create the window's controls
		{
			ControlRef		control = null;
			
			
			(OSErr)CreateRootControl(ptr->window, &ptr->root);
			ptr->placard = control = GetNewControl(kIDInspectorPlacard, ptr->window);
			(OSErr)EmbedControl(control, ptr->root);
		}
		
		// temp?
		MoveWindow(ptr->window, 430, 50, false/* activate */);
		SizeWindow(ptr->window, 200, 300, false/* update */);
		
		// set up the Window Info information
		ptr->windowInfo = WindowInfo_New();
		WindowInfo_SetWindowDescriptor(ptr->windowInfo, kTelnetWindowDescriptorInspectorWindoid);
		{
			WindowResizeResponderProcPtr	resizeProc = handleNewSize;
			
			
			WindowInfo_SetWindowResizeLimits(ptr->windowInfo,
													300/* minimum height */,
													200/* minimum width */,
													SHRT_MAX/* maximum height */,
													SHRT_MAX/* maximum width */);
			WindowInfo_SetWindowResizeResponderUPP(ptr->windowInfo, resizeProc, 0L);
		}
		WindowInfo_SetAuxiliaryDataPtr(ptr->windowInfo, (void*)ptr);
		WindowInfo_SetWindowFloating(ptr->windowInfo, true);
		WindowInfo_SetForWindow(ptr->window, ptr->windowInfo);
		
		// size and arrange controls
		setUpControls(ptr);
	}
}// Init


/*!
This method cleans up an Inspector window by
destroying data structures allocated by
Inspector_Init().

(3.0)
*/
void
Inspector_Done	(void)
{
	DisposeWindow(gSessionInspectorPtr->window); // disposes of the windoid *and* all of its controls
	gSessionInspectorPtr->window = null;
	WindowInfo_Dispose(gSessionInspectorPtr->windowInfo);
	Memory_DisposePtr((Ptr*)&gSessionInspectorPtr);
	gSessionInspectorPtr = null;
}// Dispose


/*!
To obtain a reference to the Window Info structure
associated with the Inspector windoid, use this method.

The auxiliary data pointer for the Window Info object
is reserved.

(3.0)
*/
WindowInfoRef
Inspector_GetWindowInfo		(void)
{
	WindowInfoRef		result = gSessionInspectorPtr->windowInfo;
	
	
	return result;
}// GetWindowInfo


/*!
To obtain the Mac OS window pointer for the
Inspector windoid, call this routine.

(3.0)
*/
WindowRef
Inspector_GetWindow		(void)
{
	WindowRef		result = gSessionInspectorPtr->window;
	
	
	return result;
}// GetWindow


/*!
When a mouse-down event occurs in the content region
of the Inspector windoid, call this method.

(3.0)
*/
void
Inspector_HandleContentClick	(EventRecord*		inoutEventPtr)
{
	WindowRef			window = getWindow(gSessionInspectorPtr);
	
	
	if (window != null)
	{
		SInt16		windowPart = 0;
		
		
		windowPart = FindWindow(inoutEventPtr->where, &window);
		if (windowPart == inContent)
		{
			BEEP; // TEMP
		}
	}
}// HandleContentClick


/*!
Returns "true" only if an Inspector windoid is showing.
This state is not exactly the window state; specifically,
if the windoid is suspended but set to visible, the
window will be invisible (because the application is
suspended) and yet this method will return "true".

If you need to find the definite visible state of the
windoid, use Inspector_GetWindow() to obtain the Mac OS
window pointer and then use the IsWindowVisible() system
call.

(3.0)
*/
Boolean
Inspector_IsVisible		(void)
{
	Boolean		result = isVisible(gSessionInspectorPtr);
	
	
	return result;
}// IsVisible


/*!
To activate or deactivate the Inspector window,
call this method.

(3.0)
*/
void
Inspector_SetActive		(Boolean const		inIsActive)
{
	HiliteWindow(getWindow(gSessionInspectorPtr), inIsActive);
}// SetActive


/*!
To set the suspended state of the Inspector windoid,
call this method.  If the windoid is suspended, it
is invisible no matter what state it is in.  If
the windoid is not suspended, its previous state
is restored.  For example, if the window is
"visible" and suspended, it will be hidden.  If
the window is "invisible" and suspended, it will
be hidden as well.  But if the window is then set
to a non-suspended state, in the former case the
window will reappear, whereas in the latter case
the window remains invisible.

(3.0)
*/
void
Inspector_SetSuspended		(Boolean			inIsSuspended)
{
	WindowRef		window = getWindow(gSessionInspectorPtr);
	
	
	if (window != null)
	{
		gSessionInspectorPtr->isSuspended = inIsSuspended;
		if (inIsSuspended) HideWindow(window);
		else
		{
			if (isVisible(gSessionInspectorPtr)) ShowWindow(window);
			else HideWindow(window);
		}
	}
}// SetSuspended


/*!
Shows or hides the Inspector windoid.  You can also use
Inspector_ToggleVisible().

(3.0)
*/
void
Inspector_SetVisible	(Boolean			inIsVisible)
{
	setVisible(gSessionInspectorPtr, inIsVisible);
}// SetVisible


/*!
Shows or hides the Inspector windoid based on whether
or not it is currently visible.

(3.0)
*/
void
Inspector_ToggleVisible		(void)
{
	setVisible(gSessionInspectorPtr, !isVisible(gSessionInspectorPtr));
}// ToggleVisible


/*!
To update the Inspector windoid, call this method.  This
should be done when update events are received for the
Inspector window.  This method redraws the contents of
the windoid using the absolutely current windoid state
information.

(3.0)
*/
OSErr
Inspector_Update	(GrafPtr		inDrawingPort)
{
	OSErr		result = noErr;
	GrafPtr		oldPort = null;
	
	
	GetPort(&oldPort);
	SetPort(inDrawingPort);
	{
		RgnHandle		visibleRegion = Memory_NewRegion();
		
		
		if (visibleRegion != null)
		{
			GetPortVisibleRegion(inDrawingPort, visibleRegion);
			UpdateControls(getWindow(gSessionInspectorPtr), visibleRegion);
			Memory_DisposeRegion(&visibleRegion);
		}
	}
	SetPort(oldPort);
	
	return result;
}// Update


//
// internal methods
//

/*!
Returns the Mac OS window reference for the
specified Inspector windoid.

(3.0)
*/
static WindowRef
getWindow		(InspectorPtr		inPtr)
{
	return inPtr->window;
}// getWindow


/*!
This method moves and resizes controls in response to
a resize of an Inspector window.

(3.0)
*/
static pascal Boolean
handleNewSize		(WindowRef const		inWindow,
					 SInt32 const			inDeltaSizeX,
					 SInt32 const			inDeltaSizeY,
					 SInt32 const			UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_true;
	
	
	DialogAdjust_BeginControlAdjustment(inWindow);
	{
		DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, gSessionInspectorPtr->placard, inDeltaSizeX);
		DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeV, gSessionInspectorPtr->placard, inDeltaSizeY);
	}
	DialogAdjust_EndAdjustment(inDeltaSizeX, inDeltaSizeY); // moves and resizes controls properly
	
	return result;
}// handleNewSize


/*!
Returns "true" only if an Inspector windoid is showing.
This state is not exactly the window state; specifically,
if the windoid is suspended but set to visible, the
window will be invisible (because the application is
suspended) and yet this method will return "true".

If you need to find the definite visible state of the
windoid, use getWindow() to obtain the Mac OS window
pointer and then use the IsWindowVisible() system call.

(3.0)
*/
static Boolean
isVisible		(InspectorPtr		inPtr)
{
	return inPtr->isShowing;
}// isVisible


/*!
This routine sizes and arranges controls in the
Inspector window.

(3.0)
*/
static void
setUpControls		(InspectorPtr		inPtr)
{
	ControlRef			control = null;
	Rect				windowRect;
	
	
	MacLook_GetWindowRegionBounds(inPtr->window, kWindowContentRgn, &windowRect);
	
	// size controls
	control = inPtr->placard;
	SizeControl(control, windowRect.right - windowRect.left + 2/* window edges */,
							windowRect.bottom - windowRect.top + 2/* window edges */);
	
	// arrange controls
	control = inPtr->placard;
	MoveControl(control, -1, -1);
}// setUpControls


/*!
Shows or hides an Inspector windoid.

(3.0)
*/
static void
setVisible		(InspectorPtr		inPtr,
				 Boolean			inIsVisible)
{
	Boolean		oldVisibility = inPtr->isShowing;
	
	
	inPtr->isShowing = inIsVisible;
	if (oldVisibility != inPtr->isShowing)
	{
		WindowRef		window = getWindow(inPtr);
		
		
		// show or hide the window
		if (inIsVisible) ShowWindow(window);
		else HideWindow(window);
	}
}// setVisible