/*###############################################################

	LED.c
	
	Interface Library 1.0
	 1998-2000 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#ifndef __MACTYPES__
#	include <MacTypes.h>
#endif
#ifndef __EVENTS__
#	include <Events.h>
#endif
#ifndef __OSUTILS__
#	include <OSUtils.h>
#endif
#ifndef __DESKBUS__
#	include <DeskBus.h>
#endif
#ifndef __MIXEDMODE__
#	include <MixedMode.h>
#endif

// library includes
#include "InterfaceLib.h"


#if TARGET_API_MAC_OS8

// keyboard ADB constants
enum
{
	kTalkCommand = 12, // 8+4         
	kListenCommand = 8,
	kLEDRegister = 2
};

static long     				gCompletionFlag = 0L;
static Byte						gRotationPosition = kLEDNumLock;
static ADBServiceRoutineUPP 	completionProc = null;

// internal method prototypes
static pascal void			completionOfADBOperation	(void);
static void					disposeCompletionProc		(void);
static void					setCompletionProc			(ADBServiceRoutineUPP	inNewProc);
static ADBServiceRoutineUPP	standardCompletionProc		(void);
static void					kb_SetMacNumLockLED			(Byte					LEDPattern);

#endif


//
// public methods
//

/**
Turns on all three keyboard LEDs for one
second, and then turns them all off.

(1.0)
*/
void
LED_Flash	(void)
{
#if TARGET_API_MAC_OS8
	unsigned long	dummy = 0L;
	
	
	if (standardCompletionProc() != null)
	{
		kb_SetMacNumLockLED(kLEDAll);
		Delay(60, &dummy);
		kb_SetMacNumLockLED(kLEDNone);
		Delay(30, &dummy);
		disposeCompletionProc();
	}
#else
	// Carbon does not support this code
#endif
}// Flash


/**
Turns on all three keyboard LEDs for less
than a tenth of a second, and then turns
them all off.

(1.0)
*/
void
LED_FlashQuick	(void)
{
#if TARGET_API_MAC_OS8
	unsigned long	dummy = 0L;
	
	
	if (standardCompletionProc() != null)
	{
		kb_SetMacNumLockLED(kLEDAll);
		Delay(5, &dummy);
		kb_SetMacNumLockLED(kLEDNone);
		Delay(5, &dummy);
		disposeCompletionProc();
	}
#else
	// Carbon does not support this code
#endif
}// FlashQuick


/**
To restore the state of the LEDs based on
the keyboard state, such as whether or not
the Caps Lock key is down, call this
method.  You should always invoke this
routine at least once (before your program
quits) so the user's keyboard lights don't
get screwed up permanently.

(1.0)
*/
void
LED_Restore		(void)
{
#if TARGET_API_MAC_OS8
	KeyMap		keyMap;
	Boolean		capsDown = false;
	
	
	// The GetKeys() result is an array of four 32-bit
	// unsigned numbers, for a total of 128 bits.  The
	// particular bit corresponding to a key is its
	// virtual key code.  Thus, since the Caps Lock key
	// has a virtual key code of 0x39 (57), it is "down"
	// if bit number 57 is a 1.  For a set of virtual
	// key codes, see page 2-43 of the book "Inside
	// Macintosh: Macintosh Toolbox Essentials".
	GetKeys(keyMap);
	capsDown = ((unsigned char*)keyMap)[7] & 0x02; // virtual key 0x39 == bit #57 of 128; the 63->56 (8th) octet, bit #1, mask 00000010
	if (capsDown) LED_Set(kLEDCapsLock);
	else LED_Set(kLEDNone);
#else
	// Carbon does not support this code
#endif
}// Restore


/**
To light up the LED left of the current
rotation position, dimming all other LEDs,
call this method.  By calling this routine
successively with delays in between, you
can create an animation of moving lights
from right to left.

(1.0)
*/
void
LED_RotateLeft	(void)
{
#if TARGET_API_MAC_OS8
	switch (gRotationPosition)
	{
		case kLEDNumLock:
			gRotationPosition = kLEDScrollLock;
			break;
		
		case kLEDScrollLock:
			gRotationPosition = kLEDCapsLock;
			break;
		
		case kLEDCapsLock:
		default:
			gRotationPosition = kLEDNumLock;
			break;
	}
	LED_Set(gRotationPosition);
#else
	// Carbon does not support this code
#endif
}// RotateLeft


/**
To light up the LED right of the current
rotation position, dimming all other LEDs,
call this method.  By calling this routine
successively with delays in between, you
can create an animation of moving lights
from left to right.

(1.0)
*/
void
LED_RotateRight	(void)
{
#if TARGET_API_MAC_OS8
	switch (gRotationPosition)
	{
		case kLEDNumLock:
			gRotationPosition = kLEDCapsLock;
			break;
		
		case kLEDCapsLock:
			gRotationPosition = kLEDScrollLock;
			break;
		
		case kLEDScrollLock:
		default:
			gRotationPosition = kLEDNumLock;
			break;
	}
	LED_Set(gRotationPosition);
#else
	// Carbon does not support this code
#endif
}// RotateRight


/**
Turns on the specified LEDs, and turns
off the rest.

(1.0)
*/
void
LED_Set		(LEDPattern		inPattern)
{
#if TARGET_API_MAC_OS8
	if (standardCompletionProc() != null)
	{
		kb_SetMacNumLockLED(inPattern);
		disposeCompletionProc();
	}
#else
	// Carbon does not support this code
#endif
}// Set


//
// internal methods
//

#if TARGET_API_MAC_OS8

/**
The standard ADB completion routine that
sets a global variable.

(1.0)
*/
static pascal void
completionOfADBOperation	(void)
{
	gCompletionFlag = 1L;
}// completionOfADBOperation


/**
To specify the completion routine to
call when ADB operations are performed,
call this method.

(1.0)
*/
static void
disposeCompletionProc	(void)
{
	DisposeRoutineDescriptor(completionProc);
}// disposeCompletionProc


/**
To specify the completion routine to
call when ADB operations are performed,
call this method.  When finished with
the completion routine, be sure to call
disposeCompletionProc().

(1.0)
*/
static void
setCompletionProc	(ADBServiceRoutineUPP	inNewProc)
{
	completionProc = inNewProc;
}// setCompletionProc


/**
To specify that completionOfADBOperation()
should be the completion routine of ADB
operations from now on, call this method.
When finished with the completion routine,
be sure to call disposeCompletionProc().

(1.0)
*/
static ADBServiceRoutineUPP
standardCompletionProc	(void)
{
	ADBServiceRoutineUPP	result = NewADBServiceRoutineProc(completionOfADBOperation);
	
	
	setCompletionProc(result);
	return result;
}// standardCompletionProc


/**
Turns the keyboard LEDs on or off.

(1.0)
*/
static void
kb_SetMacNumLockLED		(Byte		LEDPattern)
{
	OSErr			retCode = noErr;
	ADBAddress		anADBadd;		// address of each device
	Byte			regBuff[9];		// buffer for ADBOp commands
	Byte			oldReg = 0;
	
	
	// Get an address for an ADB device - here we just talk to whatever keyboard
	// is at ADB Address 2.  A more complete program might check the type of
	// keyboard present to ensure that it has an LED present.
	anADBadd = 2;
	
	regBuff[0] = (Byte)0; // initial data buffer length
	regBuff[1] = 0;
	regBuff[2] = 0;
	
	// Initialize the completion flag so that we will know
	// when the ADBOp call has completed.
	gCompletionFlag = 0, retCode = ADBOp((Ptr)&gCompletionFlag, completionProc, (Ptr)&regBuff,
											kTalkCommand + kLEDRegister + 16 * anADBadd);
	
	if (retCode != noErr)         // queue was full, try again
	{
		gCompletionFlag = 0, retCode = ADBOp((Ptr)&gCompletionFlag, completionProc, (Ptr)&regBuff,
											kTalkCommand + kLEDRegister + 16 * anADBadd);
	}
    
	if (retCode == noErr) // queue still has room?
	{
		// do nothing until completion routine has run
		do {} while (!gCompletionFlag);
				
		regBuff[0] = (Byte)2; // initial data buffer length
		
		// extended keyboard has a word of data, LEDs are low 3 bits
		oldReg = regBuff[2];
		
		// set the specified bits; note that a clear bit indicates a lit LED
		regBuff[2] = (oldReg & 255 - 7) | (7 - LEDPattern);
		
		// Initialize the completion flag so that we will know
		// when the ADBOp call has completed.
		gCompletionFlag = 0, retCode = ADBOp((Ptr)&gCompletionFlag, completionProc, (Ptr)&regBuff,
											kListenCommand + kLEDRegister + 16 * anADBadd);
		
		if (retCode != noErr) {} // error on ADBOp for listen
		else do {} while (!gCompletionFlag);
	}
}// kb_SetMacNumLockLED

#else
	// Carbon does not support any of these internal routines
#endif