/*###############################################################

	LineMode.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdio>
#else
#	include <stdio.h>
#endif

// MacTelnet includes
#include "ConnectionData.h"
#include "LineMode.h"
#include "network.proto.h"
#include "Parse.h"
#include "vsdata.h"
#include "Terminal.h"



// variables
static unsigned char gLineModeSequence[] = { TEL_IAC, TEL_SB, N_LINEMODE, 0/* set as needed */, 0/* set as needed */, TEL_IAC, TEL_SE };

static unsigned char default_slc[] = // default for set local characters
{
	255, 255, 255, DEFAULT_SLC_IP, DEFAULT_SLC_AO, 0, 0, DEFAULT_SLC_ABORT,
	DEFAULT_SLC_EOF, DEFAULT_SLC_SUSP, DEFAULT_SLC_EC, DEFAULT_SLC_EL, DEFAULT_SLC_EW, DEFAULT_SLC_RP, DEFAULT_SLC_LNEXT, DEFAULT_SLC_XON,
	DEFAULT_SLC_XOFF, DEFAULT_SLC_FORW1, DEFAULT_SLC_FORW2, 255, 255, 255, 255, 255,
	255, 255, 255, 255, 255, 255, 255
};

// internal methods
static void		negotiateSLC	(ConnectionDataPtr		tw);
static void		respondSLC		(short					optionNumber,
								 short*					alreadySentSB,
								 ConnectionDataPtr		tw,
								 Boolean				AWK);
static void		setForwardMask	(ConnectionDataPtr		tw);



/*!
Initializes line mode.

(2.6)
*/
void
LineMode_Init	(ConnectionDataPtr		inConnectionDataPtr)
{
	short	i = 0;
	
	
	for (i = 0; i <= SLC_MAX; i++)
	{
		inConnectionDataPtr->lineMode.slc[i] = default_slc[i];
		if (default_slc[i] == 255) inConnectionDataPtr->lineMode.slcLevel[i] = SLC_NOSUPPORT;
		else inConnectionDataPtr->lineMode.slcLevel[i] = SLC_VALUE;
	}
	inConnectionDataPtr->lineMode.mode = 0;
	inConnectionDataPtr->lineMode.nextIsLiteral = 0;
	inConnectionDataPtr->lineMode.active = 0;
}// Init


/*!
Manages keypresses for connections using line mode.

(2.6)
*/
void
LineMode_HandleKeyDown		(unsigned char		ascii,
							 ConnectionDataPtr	tw)
{
	if (tw->lineMode.nextIsLiteral) //do no processing on next key
	{
		tw->lineMode.nextIsLiteral = false;
		if (tw->kblen < (MAXKB -1)) 	/* Add to buffer if not full */
			tw->kbbuf[tw->kblen++] = ascii;
		else 
		{				
			Network_SendFlushData( tw->port, tw->kbbuf, tw->kblen);	/* if full send buffer */
			tw->kbbuf[0]=ascii;
			tw->kblen=1;
		}
		if (tw->echo)
		{
			if (ascii>31 && ascii <127)	/* add these chars to buffer */
			{
				Parse_ShowOrInterpretAsCommand(tw, &ascii, 1);
				return;
			}
			else
				return;
		}
	}

	if (tw->lineMode.mode & 2) // TRAPSIG mode active
	{		
		unsigned char toSend[2] = {IAC,0};
		short whichSignal = 0;
		if (ascii == tw->lineMode.slc[SLC_IP])
		{
			whichSignal = SLC_IP;
			toSend[1] = TEL_SUSP;
		}
		else if (ascii == tw->lineMode.slc[SLC_ABORT])
		{
			whichSignal = SLC_ABORT;
			toSend[1] = TEL_ABORT;
		}
		if (toSend[1]) //if we have a signal to catch
		{
			if (tw->echo) Parse_ShowOrInterpretAsCommand(tw, &ascii, 1); // echo if we should
			
			tw->kblen = 0; // zero out the buffer
			Network_SendFlushData(tw->port, toSend, 2); // send IAC whatever
			
			if (tw->lineMode.slcLevel[whichSignal] & SLC_FLUSHIN)
			{
				unsigned char dm[2] = {IAC,TEL_DM};
				Network_SendFlush(tw->port);
				Network_MarkNextTransmissionAsUrgent();
				Network_SendData(tw->port,dm,2);//send IAC DM
			}
			if (tw->lineMode.slcLevel[whichSignal] & SLC_FLUSHOUT)
			{
				unsigned char tm[3] = {IAC,TEL_DOTEL,N_TIMING};
				tw->timing = 1;	//tell ourselves to wait for WILL TIMING
				Network_SendFlushData(tw->port,tm,3);//send DO TIMING
			}
			return;
		}
	}
	
	if ((tw->lineMode.mode & L_SOFT_TAB)&&(ascii == 0x09)) // SOFT_TAB mode active; expand tab into spaces
	{
		short numSpaces = VSIgetNextTabDistance();
		while (numSpaces > 0)
		{
			while ((numSpaces > 0)&&(tw->kblen < (MAXKB -1)))
			{
				tw->kbbuf[tw->kblen++] = 0x20; //space
				numSpaces--;
			}
			if (tw->kblen == (MAXKB -1))
			{
				Network_SendFlushData( tw->port, tw->kbbuf, tw->kblen);	/* if full send buffer */
				tw->kblen=0;
			}
		}
		if (tw->echo) Parse_ShowOrInterpretAsCommand(tw, &ascii, 1);
		return;
	}

	if (tw->lineMode.mode & L_EDIT) //handle editing functions
	{
		if (ascii == '\015'/* CR */)
		{ //since we are in edit, send the buffer and CR-LF
			if (tw->kblen > 0) Network_SendData(tw->port, tw->kbbuf, tw->kblen);
			Network_SendFlushData(tw->port,"\015\012",2);
			tw->kblen = 0;
			if (tw->echo) Parse_ShowOrInterpretAsCommand(tw, (unsigned char*)"\012\015", 2);
			return;
		}
		
		if (ascii == tw->lineMode.slc[SLC_EC]) //kill the character
		{
			if (tw->echo) Parse_ShowOrInterpretAsCommand(tw, (unsigned char*)"\010 \010", 3);	
			tw->kblen--;
			return;
		}
		else if (ascii == tw->lineMode.slc[SLC_AO]) //kill the line
		{
			while (tw->kblen >0) 
			{
				if (tw->echo) Parse_ShowOrInterpretAsCommand(tw, (unsigned char*)"\010 \010", 3);
				tw->kblen--;
			}
			return;
		}
		else if (ascii == tw->lineMode.slc[SLC_EL]) //kill the line
		{
			while (tw->kblen >0) 
			{
				if (tw->echo) Parse_ShowOrInterpretAsCommand(tw, (unsigned char*)"\010 \010", 3);
				tw->kblen--;
			}
			return;
		}
		else if (ascii == tw->lineMode.slc[SLC_EOF])
		{ //push the buffer, send IAC EOF
			char eofString[2] = { IAC, TEL_EOF };
			if (tw->kblen > 0) Network_SendData(tw->port, tw->kbbuf, tw->kblen);
			tw->kbbuf[0]=ascii;
			tw->kblen=1;
			Network_SendFlushData(tw->port,eofString, 2);
			return;
		}
		else if (ascii == tw->lineMode.slc[SLC_SUSP])
		{
			char eofString[2] = { IAC, TEL_SUSP };
			if (tw->kblen > 0) Network_SendData(tw->port, tw->kbbuf, tw->kblen);
			tw->kblen = 0;
			Network_SendFlushData(tw->port,eofString, 2);
			return;
		}
		else if (ascii == tw->lineMode.slc[SLC_EW])
		{
			while ((tw->kbbuf[tw->kblen-1] != 0x20)&&(tw->kblen >= 0)) //while its not a space
			{
				if (tw->echo) Parse_ShowOrInterpretAsCommand(tw, (unsigned char*)"\010 \010", 3);
				tw->kblen--;
			}
		}
		else if (ascii == tw->lineMode.slc[SLC_RP])
		{
			VSredrawLine(tw->vs);
			return;
		}
		else if (ascii == tw->lineMode.slc[SLC_LNEXT])
		{
			tw->lineMode.nextIsLiteral = true;
			return;
		}
		else if (ascii == tw->lineMode.slc[SLC_XON])
		{
			if (tw->allow_flow) //remote flow control can turn this off
				tw->enabled = 0;
			return;
		}	
		else if (ascii == tw->lineMode.slc[SLC_XOFF])
		{
			if (tw->allow_flow) //remote flow control can turn this off
				tw->enabled = 0;
			return;
		}
		else if ((ascii == tw->lineMode.slc[SLC_FORW1])||(ascii == tw->lineMode.slc[SLC_FORW1]))
		{
			if (tw->kblen > 0) Network_SendData(tw->port, tw->kbbuf, tw->kblen);
			Network_SendFlushData(tw->port,&ascii,1);
			tw->kblen = 0;
			return;

		}
		//ok, at this point, we are past all local editing functions.  Now, add the character to the buffer.
		else
		{
			if (tw->kblen < (MAXKB -1)) 	// Add to buffer if not full
				tw->kbbuf[tw->kblen++] = ascii;
			else 
			{				
				Network_SendFlushData( tw->port, tw->kbbuf, tw->kblen);	// if full send buffer 
				tw->kbbuf[0]=ascii;
				tw->kblen=1;
			}
		}

	}
	else if (ascii == '\015') //CR; map this to CR-LF
	{
		unsigned char toSend[2] = {0x0D,0x00};
		Network_SendFlushData(tw->port,toSend,2);
		if (tw->echo) Parse_ShowOrInterpretAsCommand(tw, (unsigned char*)"\012\015", 2);
		return;	
	}
	else
	{
		// not editing; send it
		Network_SendFlushData(tw->port, &ascii, 1);	// if full, send buffer
	}
	
	// handle local echoes
	if (tw->echo)
	{
		if (ascii > 31 && ascii < 127)
		{
			// add to buffer
			Parse_ShowOrInterpretAsCommand(tw, &ascii, 1);
		}
		else
		{
			// non-printable character; create a two-character printable representation
			if (!(tw->lineMode.mode & L_LIT_ECHO)) // don't echo if this is set
			{
				ascii = '@' + ascii;
				Parse_ShowOrInterpretAsCommand(tw, (unsigned char*)"^", 1);
				Parse_ShowOrInterpretAsCommand(tw, &ascii, 1);
			}
		}
	}
}// HandleKeyDown


/*!
Sends a sub-option for line mode, used at telnet
connection time.

(2.6)
*/
void
LineMode_SendSubOption		(ConnectionDataPtr		inConnectionDataPtr)
{
	switch (inConnectionDataPtr->parsedat[1])
	{
		char	s[80];
		
		
		case L_MODE:	/* change mode */
			if (inConnectionDataPtr->lineMode.allowed)
			{
				// RFC 1184 says a client should ignore MODE negotiations with the MODE_ACK bit set,
				// and should not generate a response if the negotiated MODE matches the current MODE.
				if (!((inConnectionDataPtr->parsedat[2] & L_MODE_ACK) ||
					((inConnectionDataPtr->parsedat[2] & L_MODE_MASK) == inConnectionDataPtr->lineMode.mode))) 
				{
					inConnectionDataPtr->lineMode.mode = inConnectionDataPtr->parsedat[2]; // accept mode
					inConnectionDataPtr->parsedat[2] |= L_MODE_ACK; // set the MODE_ACK bit
					
					sprintf(s, "%c%c%c%c", IAC, TEL_SB, N_LINEMODE, L_MODE);
					Network_SendData(inConnectionDataPtr->port, s, 4);
					sprintf(s, "%c%c%c", inConnectionDataPtr->parsedat[2], IAC, TEL_SE);
					Network_SendData(inConnectionDataPtr->port, s, 3);
				}
			}
			break;
		
		case TEL_DOTEL:	
			inConnectionDataPtr->lineMode.forwardMask = true;
			setForwardMask(inConnectionDataPtr);
			gLineModeSequence[3] = TEL_WILLTEL;
			gLineModeSequence[4] = L_FORWARDMASK;
			Network_SendFlushData(inConnectionDataPtr->port, gLineModeSequence, sizeof(gLineModeSequence));
			break;

		case TEL_DONTTEL:	
			inConnectionDataPtr->lineMode.forwardMask = false;
			gLineModeSequence[3] = TEL_WONTTEL;
			gLineModeSequence[4] = L_FORWARDMASK;
			Network_SendFlushData(inConnectionDataPtr->port, gLineModeSequence, sizeof(gLineModeSequence));
			break;

		case L_SLC:
			negotiateSLC(inConnectionDataPtr);
			break;
		
		default:
			break;
	}
}// SendSubOption


/*!
The user is done with the line; send it.

(2.6)
*/
void
LineMode_SendData		(ConnectionDataPtr		inConnectionDataPtr)
{
	short			i = 0;
	unsigned char	subBeginSeq[4] = { IAC, TEL_SB, N_LINEMODE, L_SLC };
	unsigned char	subEndSeq[2] = { IAC, TEL_SE };
	unsigned char	toSend[3];
	
	
	inConnectionDataPtr->lineMode.mode = 0;
	inConnectionDataPtr->lineMode.active = true;
	inConnectionDataPtr->lineMode.nextIsLiteral = false;

	Network_SendData(inConnectionDataPtr->port, subBeginSeq, 4);
	
	for (i = 1; i <= SLC_MAX; i++)
	{
		if (inConnectionDataPtr->lineMode.slc[i] == 255) 
		{
			toSend[0] = i;
			toSend[1] = SLC_NOSUPPORT;
			toSend[2] = 0;
		}
		else
		{
			toSend[0] = i;
			toSend[1] = SLC_VALUE;
			toSend[2] = inConnectionDataPtr->lineMode.slc[i];
		}
		Network_SendData(inConnectionDataPtr->port, toSend, 3);
	}
	Network_SendFlushData(inConnectionDataPtr->port, subEndSeq, 2); 
}// SendData


//
// internal methods
//

void
negotiateSLC	(ConnectionDataPtr	tw)
{
	short	lmslcflag = 0;
	unsigned char *ourValues = tw->lineMode.slc;
	unsigned char *ourLevels = tw->lineMode.slcLevel;
	short i;


	for (i=2, lmslcflag=0; i <= tw->parseIndex - 3; i+=3) 
	{
		short optionNumber = tw->parsedat[i];
		Boolean awked = (tw->parsedat[i+1] & SLC_AWK);
		unsigned char requestedValue = tw->parsedat[i+2];
		unsigned char requestedLevel = tw->parsedat[i+1] & SLC_LEVELBITS;
		Boolean flushin = tw->parsedat[i+1] & SLC_FLUSHIN;
		Boolean flushout = tw->parsedat[i+1] & SLC_FLUSHOUT;

		if ((ourValues[optionNumber] != requestedValue)||(ourLevels[optionNumber] != tw->parsedat[i+1])) //if we are sent what we already have, ignore it
		{
			if (requestedLevel == SLC_NOSUPPORT)
			{
				if (ourLevels[optionNumber] & SLC_LEVELBITS != SLC_NOSUPPORT)
				{
				 	ourValues[optionNumber] = 255;
					ourLevels[optionNumber] = SLC_NOSUPPORT;//ok, we wont support this
					if (!awked)
						respondSLC(optionNumber,&lmslcflag, tw, true);//awk only if he didnt awk
				}
			}
			else if (requestedLevel == SLC_DEFAULT)
			{
				ourValues[optionNumber] = default_slc[optionNumber]; //get our default
				if (ourValues[optionNumber] == 255)
					ourLevels[optionNumber] = SLC_NOSUPPORT;
				else
					ourLevels[optionNumber] = SLC_VALUE;
				respondSLC(optionNumber, &lmslcflag, tw, false);	//tell him about our choice
			}
			else if (requestedLevel > ourLevels[optionNumber])
				respondSLC(optionNumber, &lmslcflag, tw, false);  //keep lower setting
			else
			{
				ourValues[optionNumber] = requestedValue;
				ourLevels[optionNumber] = requestedLevel;
				if (flushin)
					ourLevels[optionNumber] |= SLC_FLUSHIN;
				if (flushout)
					ourLevels[optionNumber] |= SLC_FLUSHOUT;
				if (!awked)
					respondSLC(optionNumber, &lmslcflag, tw, true); //agree to his new setting
			}
		}		
	}//end for on parseindex
	if (lmslcflag) 
	{
		unsigned char toSend[2] = {IAC,TEL_SE};
		Network_SendFlushData(tw->port,toSend,2);	
	}
}// negotiateSLC


static void
respondSLC		(short				optionNumber,
				 short*				alreadySentSB,
				 ConnectionDataPtr	tw,
				 Boolean			AWK)
{
	unsigned char stuffToSend[3];
	unsigned char subBeginSeq[4] = {IAC,TEL_SB,N_LINEMODE,L_SLC};
	stuffToSend[0] = optionNumber;
	stuffToSend[1] = tw->lineMode.slcLevel[optionNumber];
	if (AWK) 
		stuffToSend[1] |= SLC_AWK;
	stuffToSend[2] = tw->lineMode.slc[optionNumber];	
	
	if (!(*alreadySentSB))
	{
		*alreadySentSB = true;
		Network_SendData(tw->port, subBeginSeq, 4);
	}
	if (stuffToSend[2] == IAC)
	{
		Network_SendData(tw->port, stuffToSend,2);
		Network_SendData(tw->port, subBeginSeq,1); //double the IAC
		Network_SendData(tw->port, subBeginSeq,1);
	}
	else Network_SendData(tw->port, stuffToSend, 3);
}// respondSLC


static void
setForwardMask		(ConnectionDataPtr		tw) // set the keys on which we forward
{
	short j, i = 3;
	Boolean anotherByte = true;
	tw->lineMode.numForwardKeys = 1;
	while((tw->parsedat[i] != 0)&&(i < 35))
	{
		if (tw->parsedat[i] == 255)
		{
			i++; //could be doubled IAC
			if(tw->parsedat[i+1] == TEL_SE)
			{
				anotherByte = false; //but its not
				i--;
			}
		}
		if (anotherByte)
			for(j = 0; j < 8; j++)
			{
				if (tw->parsedat[i] & (j*2))
				{
					tw->lineMode.forwardKeys[tw->lineMode.numForwardKeys-1] = (i-3)*8 + j;
					tw->lineMode.numForwardKeys++;
				}
			}
		else
			break;
		i++; // next byte
	}		
	for (i = tw->lineMode.numForwardKeys; i < 32; i++)  //zero the rest of the array
		tw->lineMode.forwardKeys[i] = 0;
}// setForwardMask