/*###############################################################

	ListUtilities.c
	
	Interface Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <ControlDefinitions.h>
#include <MacTypes.h>
#include <Lists.h>
#include <PLStringFuncs.h>

// library includes
#include "InterfaceLib.h"



//
// public methods
//

/*!
To get a standard ListHandle representation
of an Appearance list box control, use this
method.  If any problems occur, the output
list will be null.

(1.0)
*/
void
ListUtilities_GetControlListHandle		(ControlRef			inListBoxControl,
										 ListHandle*		outListHandle)
{
	Size	actualSize = 0;
	
	
	if (GetControlData(inListBoxControl, kControlListBoxPart, kControlListBoxListHandleTag, sizeof(ListHandle),
							(Ptr)outListHandle, &actualSize) != noErr) *outListHandle = null;
}// GetControlListHandle


/*!
To get a standard ListHandle representation
of an Appearance list box control that is a
dialog box item, use this method.  If any
problems occur, the output list will be null.

(1.0)
*/
void
ListUtilities_GetDialogItemListHandle		(const DialogPtr			inDialogPtr,
											 const DialogItemIndex		inItemIndex,
											 ListHandle*				outListHandle)
{
	ControlRef		listControl = null;
	
	
	if (GetDialogItemAsControl(inDialogPtr, inItemIndex, &listControl) == noErr)
	{
		ListUtilities_GetControlListHandle(listControl, outListHandle);
	}
	else *outListHandle = null;
}// GetDialogItemListHandle


/*!
To determine the total number of rows in a
list, use this method.

(1.0)
*/
UInt16
ListUtilities_GetItemCount	(ListHandle		inListHandle)
{
	UInt16		result = 0;
	
	
	if (inListHandle != null)
	{
		Cell		junk;
		Boolean		searchRows = true,
					searchColumns = false;
		
		
		SetPt(&junk, 0, -1);
		while (LNextCell(searchColumns, searchRows, &junk, inListHandle)) result++;
	}
	
	return result;
}// GetItemCount


/*!
If a list contains Pascal string items (as
standard lists do), you can use this method
to obtain the list data as a string.  The
first column is assumed.

(3.0)
*/
void
ListUtilities_GetListItemText		(ListHandle			inListHandle,
									 const UInt16		inZeroBasedItemIndex,
									 Str255				outItemText)
{
	Cell		cell;
	SInt16		length = 254; // on input, maximum possible length
	
	
	SetPt(&cell, 0, inZeroBasedItemIndex);
	LGetCell(outItemText + 1, &length, cell, inListHandle);
	outItemText[0] = (char)length;
}// GetListItemText


/*!
To determine the total number of selected
cells in a list, use this method.

(1.0)
*/
UInt16
ListUtilities_GetSelectedItemCount	(ListHandle		inListHandle)
{
	UInt16		result = 0;
	
	
	if (inListHandle != null)
	{
		Cell		junk;
		UInt16		total = ListUtilities_GetItemCount(inListHandle);
		
		
		SetPt(&junk, 0, 0);
		while (LGetSelect(true/* next */, &junk, inListHandle) && (result <= total)) result++, junk.v++;
	}
	
	return result;
}// GetSelectedItemCount


/*!
To obtain a unique variation of the specified
item text, guaranteeing that it will not be
the identical text of any item in the specified
list, use this method.

WARNING:	Do not use this method unless your
			list contains Pascal string data
			in its items (standard lists do).

(3.0)
*/
void
ListUtilities_GetUniqueItemText		(ListHandle		inListHandle,
									 Str255			inoutItemText)
{
	UInt16		itemCount = ListUtilities_GetItemCount(inListHandle),
				i = 0;
	
	
	// only change the specified item text if its already taken by an item in the menu
	while (!ListUtilities_IsItemUnique(inListHandle, inoutItemText))
	{
		if ((inoutItemText[inoutItemText[0]] > '9') ||
			(inoutItemText[inoutItemText[0]] < '0')) // add a number
		{
			inoutItemText[++inoutItemText[0]] = ' ';
			inoutItemText[++inoutItemText[0]] = '1';
		}
		else if (inoutItemText[inoutItemText[0]] == '9') // add another digit
		{
			inoutItemText[inoutItemText[0]] = '-';
			inoutItemText[++inoutItemText[0]] = '1';
		}
		else inoutItemText[inoutItemText[0]]++; // increment the number
	}
}// GetUniqueItemText


/*!
To determine if a list contains no item with
text identical to the specified item text,
use this method.

WARNING:	Do not use this method unless your
			list contains Pascal string data
			in its items (standard lists do).

(3.0)
*/
Boolean
ListUtilities_IsItemUnique		(ListHandle		inListHandle,
								 Str255			inItemText)
{
	Boolean		result = true;
	UInt16		itemCount = ListUtilities_GetItemCount(inListHandle),
				i = 0;
	Str255		itemString;
	
	
	// look for an identical item
	for (i = 0; ((result) && (i < itemCount)); i++)
	{
		ListUtilities_GetListItemText(inListHandle, i, itemString);
		if (!PLstrcmp(itemString, inItemText)) result = false;
	}
	
	return result;
}// IsItemUnique


/*!
If a list contains Pascal string items (as
standard lists do), you can use this method
to specify the list data as a string.  The
first column is assumed.

(3.0)
*/
void
ListUtilities_SetListItemText		(ListHandle			inListHandle,
									 const UInt16		inZeroBasedItemIndex,
									 StringPtr			inItemText)
{
	Cell		cell;
	SInt16		length = 0;
	
	
	SetPt(&cell, 0, inZeroBasedItemIndex);
	LSetCell((Ptr)(inItemText + 1), StrLength(inItemText), cell, inListHandle);
}// SetListItemText