/*###############################################################

	ListenerModel.h
	
	An implementation of the listener pattern.  When an object
	can perform operations of potential interest to other things
	(and especially if the other things are not likely to be
	known), it can use a listener model to collect a set of
	callbacks that should be invoked in response to events.
	Then, when a particular type of operation is performed, the
	listener model can be used to automatically notify all of
	the listeners for that operation, by invoking their
	callback functions.
	
	A Listener Model saves you the trouble of figuring out when
	some kind of state has changed, making it really easy to
	synchronize things.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#ifndef __LISTENERMODEL__
#define __LISTENERMODEL__

// Mac includes
#include <MacTypes.h>



// types
typedef SInt32							ListenerModelDescriptor; // values are arbitrary and defined by the model user
typedef FourCharCode					ListenerModelEventType; // values are arbitrary and defined by the model user
typedef struct { unsigned garbage; }*	ListenerModelRef;



// universal procedure pointers
/*!
Listener Callback

These routines are notifiers that are attached to
listener models.  When a particular kind of event
occurs, all callbacks attached to that event in
the listener model are invoked.
*/
typedef void (*ListenerCallbackProcPtr)		(ListenerModelEventType const		inEvent,
											 void*								inContextPtr);
#define InvokeListenerCallbackProc(userRoutine, inEvent, inContextPtr) \
			(*userRoutine)((inEvent), (inContextPtr))



//
// public methods
//

/*###############################################################
	CREATING AND DESTROYING LISTENER MODELS
###############################################################*/

ListenerModelRef
	ListenerModel_New					(ListenerModelDescriptor	inDescriptor);

void
	ListenerModel_Dispose				(ListenerModelRef*			inoutRefPtr);

/*###############################################################
	USING LISTENER MODELS
###############################################################*/

void
	ListenerModel_AddListener			(ListenerModelRef			inToWhichModel,
										 ListenerModelEventType		inForWhichEvent,
										 ListenerCallbackProcPtr	inCallback);

void
	ListenerModel_RemoveListener		(ListenerModelRef			inFromWhichModel,
										 ListenerModelEventType		inForWhichEvent,
										 ListenerCallbackProcPtr	inCallback);

void
	ListenerModel_NotifyListeners		(ListenerModelRef			inModel,
										 ListenerModelEventType		inEvent);

#endif