/*###############################################################

	Localization.c
	
	Use the powerful methods in this module to perform all kinds
	of useful operations on dialog box controls, often a must
	when doing localization work.
	
	Interface Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>

// Mac includes
#include <ControlDefinitions.h>
#include <Gestalt.h>
#include <PLStringFuncs.h>
#include <QuickDraw.h>
#include <Resources.h>
#include <Script.h>
#include <Sound.h>
#include <TextUtils.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "SpacingConstants.r"



// some (bad) magic numbers that are really pretty good estimates
enum
{
	kDialogTextMaxCharWidth = 7		// Width in pixels of typical-to-widest character in standard system font.
									// You can *guarantee* that *any* string you throw at this routine will not
									// get clipped if you set this to the width of the widest possible character
									// in the system font.  However, this will also make the dialog box much too
									// big in the average case.  Instead, I choose a good value that really is
									// sufficient for any string youll ever use.
};

// variables
static Boolean			gLeftToRight = true,		// text reads left-to-right?
						gTopToBottom = true;		// text reads top-to-bottom?

// internal methods
static OSStatus		getThemeFontInfo	(ThemeFontID		inThemeFontToUse,
										 StringPtr			inString,
										 Str255				outFontName,
										 SInt16*			outFontSizePtr,
										 Style*				outFontStylePtr,
										 UInt16*			outStringWidthPtr,
										 UInt16*			outFontHeightPtr);
static OSStatus		setControlFontInfo	(ControlRef			inControl,
										 StringPtr			inFontName,
										 SInt16 const		inFontSize,
										 Style const		inFontStyle);



//
// public methods
//

/*!
This routine should be called once, before any
other Localization Utilities routine.  It allows
you to set flags that affect operation of all of
the modules functions, and (optionally) to set
the file descriptor of a localization resource
file.  You only need to set the second parameter
if you plan on using Localization_UseResourceFile().

(1.0)
*/
void
Localization_Init	(UInt32 const		inFlags)
{
	gLeftToRight = !(inFlags & kLocalization_FlagReadTextRightToLeft);
	gTopToBottom = !(inFlags & kLocalization_FlagReadTextBottomToTop);
}// Init


/*!
To auto-size and auto-arrange a series of action
buttons for a dialog, use this method.  Dialog
buttons are assumed to be positioned on the same
line horizontally, and located at the bottom-right
corner of a dialog box (or the bottom-left corner,
if the localization is right-to-left).  Thus,
buttons are moved away from their assumed corner,
and spaced exactly 12 pixels apart (as specified
in " SpacingConstants.r").  The first button in
the given array of item indices is expected to be
the default button (if any), and buttons are
positioned in the order they appear in the array.

This routine operates on generic dialogs; alerts
automatically invoke it.  For convenience, the
"common size" chosen for the buttons is returned,
as a number in units of horizontal pixels.  The
buttons will not acquire a width smaller than the
standard minimum button width (as specified in
" SpacingConstants.r").

Invoking this method can cause buttons to move
and resize, causing onscreen flickering.  You
should invoke this method while drawing to an
offscreen graphics world or while the specified
dialog is invisible.

(1.0)
*/
UInt16
Localization_AdjustDialogButtonArray		(DialogRef const			inDialog,
											 DialogItemIndex const*		inDialogItemIndices,
											 UInt16 const				inDialogItemCount)
{
	UInt16				result = BUTTON_WD_MINIMUM;
	register SInt16		i = 0;
	SInt16				positionH = 0,
						positionV = 0;
	Rect				windowRect;
	
	
	GetPortBounds(GetWindowPort(GetDialogWindow(inDialog)), &windowRect);
	
	// determine the width of the widest button
	for (i = 0; i < inDialogItemCount; i++)
	{
		result = Localization_AutoSizeButtonItem(inDialog, inDialogItemIndices[i], result);
	}
	
	// resize all of the buttons to that width
	for (i = 0; i < inDialogItemCount; i++)
	{
		(UInt16)Localization_AutoSizeButtonItem(inDialog, inDialogItemIndices[i], result);
	}
	
	// now move the buttons
	{
		SInt16		deltaPositionH = ((Localization_IsLeftToRight()) ? -1 : 1) * (HSP_BUTTONS + result);
		
		
		positionV = windowRect.bottom - VSP_BUTTON_AND_DIALOG - BUTTON_HT;
		positionH = (Localization_IsLeftToRight())
						? (windowRect.right - HSP_BUTTON_AND_DIALOG - result)
						: (HSP_BUTTON_AND_DIALOG);
		for (i = 0; i < inDialogItemCount; i++, positionH += deltaPositionH)
		{
			MoveDialogItem(inDialog, inDialogItemIndices[i], positionH, positionV);
		}
	}
	
	return result;
}// AdjustDialogButtonArray


/*!
To automatically move and resize a dialog boxs
default and/or cancel buttons so that they are
the same size and just large enough to fit their
button titles, call this method.  Dialog buttons
are assumed to be positioned on the same line
horizontally, and located at the bottom-right
corner of a dialog box (or the bottom-left
corner, if the localization is right-to-left).
Thus, buttons are moved away from their assumed
corner, and spaced exactly 12 pixels apart (as
specified in " SpacingConstants.r").

This routine operates on generic dialogs; alerts
automatically invoke it.  For convenience, the
"common size" chosen for both buttons is returned,
as a number in units of horizontal pixels.  The
buttons will not acquire a width smaller than the
standard minimum button width.

Invoking this method can cause buttons to move
and resize, causing onscreen flickering.  You
should invoke this method while drawing to an
offscreen graphics world or while the specified
dialog is invisible.

(1.0)
*/
UInt16
Localization_AdjustDialogButtons	(DialogRef const		inDialog,
									 Boolean const			inAdjustDefaultButton,
									 Boolean const			inAdjustCancelButton)
{
	UInt16				result = 0;
	SInt16				itemCount = 0;
	DialogItemIndex		indices[2];
	
	
	if (inAdjustDefaultButton) indices[itemCount++] = GetDialogDefaultItem(inDialog);
	if (inAdjustCancelButton) indices[itemCount++] = GetDialogCancelItem(inDialog);
	
	result = Localization_AdjustDialogButtonArray(inDialog, indices, itemCount);
	
	return result;
}// AdjustDialogButtons


/*!
To auto-arrange a windows help button to occupy
the lower-left (for left-to-right localization) or
lower-right (for right-to-left localization) corner
of a window, use this method.

(1.0)
*/
void
Localization_AdjustHelpButtonControl	(ControlRef		inControl)
{
	Rect		windowRect,
				controlRect;
	
	
	GetPortBounds(GetWindowPort(GetControlOwner(inControl)), &windowRect);
	GetControlBounds(inControl, &controlRect);
	
	// now move the button
	{
		UInt16		positionH = 0,
					positionV = 0,
					buttonWidth = (controlRect.right - controlRect.left),
					buttonHeight = (controlRect.bottom - controlRect.top);
		
		
		positionV = windowRect.bottom - VSP_BUTTON_AND_DIALOG - buttonHeight;
		positionH = (Localization_IsLeftToRight())
						? (HSP_BUTTON_AND_DIALOG)
						: (windowRect.right - HSP_BUTTON_AND_DIALOG - buttonWidth);
		MoveControl(inControl, positionH, positionV);
	}
}// AdjustHelpButtonControl


/*!
To auto-arrange a dialog boxs help button to
occupy the lower-left (for left-to-right
localization) or lower-right (for right-to-left
localization) corner of a dialog, use this method.

This routine operates on generic dialogs; alerts
automatically invoke it.

(1.0)
*/
void
Localization_AdjustHelpButtonItem	(DialogRef const			inDialog,
									 DialogItemIndex const		inHelpButtonIndex)
{
	ControlRef		control = null;
	Rect			windowRect,
					controlRect;
	
	
	(OSErr)GetDialogItemAsControl(inDialog, inHelpButtonIndex, &control);
	GetPortBounds(GetWindowPort(GetControlOwner(control)), &windowRect);
	GetControlBounds(control, &controlRect);
	
	// now move the button
	{
		UInt16		positionH = 0,
					positionV = 0,
					buttonWidth = (controlRect.right - controlRect.left),
					buttonHeight = (controlRect.bottom - controlRect.top);
		
		
		positionV = windowRect.bottom - VSP_BUTTON_AND_DIALOG - buttonHeight;
		positionH = (Localization_IsLeftToRight())
						? (HSP_BUTTON_AND_DIALOG)
						: (windowRect.right - HSP_BUTTON_AND_DIALOG - buttonWidth);
		MoveDialogItem(inDialog, inHelpButtonIndex, positionH, positionV);
	}
}// AdjustHelpButtonItem


/*!
To automatically set the width of a button
control so it is wide enough to comfortably
fit its current title, but not less wide
than the specified minimum, call this method.
The chosen width for the button is returned.

(1.0)
*/
UInt16
Localization_AutoSizeButtonControl	(ControlRef			inControl,
									 UInt16 const		inMinimumWidth)
{
	Str255			buttonString;
	UInt16			result = inMinimumWidth;
	
	
	if (inControl != null)
	{
		GetControlTitle(inControl, buttonString);
		if (StrLength(buttonString) > 0)
		{
			// The quick, dirty, and usually adequate way - this saves a LOT
			// of fooling around with fonts (but maybe you would like to write
			// a better algorithm that resets the graphics port, extracts the
			// theme font information, does a little dance and calculates  to
			// 26 digits before returning the ideal width of the button?).
			result = INTEGER_DOUBLED(MINIMUM_BUTTON_TITLE_CUSHION) +
						StrLength(buttonString) * kDialogTextMaxCharWidth;
		}
		if (result < inMinimumWidth) result = inMinimumWidth;
		SizeControl(inControl, result, BUTTON_HT);
	}
	return result;
}// AutoSizeButtonControl


/*!
To automatically set the width of a dialog
box button control so it is wide enough to
comfortably fit its current title, but not
less wide than the specified minimum, call
this method.  The chosen width for the
button is returned.

(1.0)
*/
UInt16
Localization_AutoSizeButtonItem		(DialogRef const		inDialog,
									 DialogItemIndex const	inItemIndex,
									 UInt16 const			inMinimumWidth)
{
	ControlRef		control = null;
	UInt16			result = inMinimumWidth;
	
	
	if (GetDialogItemAsControl(inDialog, inItemIndex, &control) == noErr)
	{
		result = Localization_AutoSizeButtonControl(control, inMinimumWidth);
		SizeDialogItem(inDialog, inItemIndex, result, BUTTON_HT); // update Dialog Manager view of control
	}
	return result;
}// AutoSizeButtonItem


/*!
Returns the text encoding base for the specified
font.  This can be used to create a new text
encoding converter.

(3.0)
*/
OSStatus
Localization_GetFontTextEncoding	(StringPtr			inFontName,
									 TextEncoding*		outTextEncoding)
{
	OSStatus		result = noErr;
	
	
	if (outTextEncoding == null) result = memPCErr;
	else if (API_AVAILABLE(FontToScript) && API_AVAILABLE(UpgradeScriptInfoToTextEncoding))
	{
		//LangCode		textLanguageID = GetScriptVariable(smSystemScript, smScriptLang);
		SInt16			fontID = 0;
		ScriptCode		scriptCode = smRoman;
		TextEncoding	encoding = kTextEncodingMacRoman;
		
		
		GetFNum(inFontName, &fontID);
		scriptCode = FontToScript(fontID);
		
		// The Mac OS uses a giant map to determine the proper encoding for
		// special cases (such as Icelandic) that use the Roman script;
		// in these cases, the script code alone isnt enough to choose the
		// right encoding, so the system localization itself is used!  To
		// get this behavior, the language and region must be dont care.
		result = UpgradeScriptInfoToTextEncoding(scriptCode, kTextLanguageDontCare, kTextRegionDontCare,
													inFontName, &encoding);
	}
	else
	{
		*outTextEncoding = kTextEncodingMacRoman; // assume this when an error occurs
		result = cfragNoSymbolErr;
	}
	
	return result;
}// GetFontTextEncoding


/*!
Saves the current font, font size, font style,
and text mode of the current graphics port.

(1.0)
*/
void
Localization_GetPortFontState	(GrafPortFontState*		outState)
{
	if (outState != null)
	{
		GrafPtr		port;
		
		
		GetPort(&port);
		outState->fontID = GetPortTextFont(port);
		outState->fontSize = GetPortTextSize(port);
		outState->fontStyle = GetPortTextFace(port);
		outState->textMode = GetPortTextMode(port);
	}
}// GetPortFontState


/*!
Determines the height required to fit a single
line of text in the specified theme font.  If
no API exists (i.e. earlier than Appearance 1.1)
to calculate the font exactly, the font constant
is used to determine the correct value for a
reasonable font (if you ask for the system
font, you get a 12-point system font height, for
most others you get 10-point Geneva, etc.).

(1.0)
*/
UInt16
Localization_GetSingleLineTextHeight	(ThemeFontID		inThemeFontToUse)
{
	UInt16		result = 0;
	SInt16		fontSize = 0;
	Style		fontStyle = normal;
	Str255		fontName;
	
	
	if (getThemeFontInfo(inThemeFontToUse, null/* string to calculate width of - unused */,
							fontName, &fontSize, &fontStyle, null/* string width - unused */,
							&result/* font height */) != noErr) result = 0;
	return result;
}// GetSingleLineTextHeight


/*!
Determines the width required to fit the given
line of text in the specified theme font.  If
no API exists (i.e. earlier than Appearance 1.1)
to calculate the font exactly, the font constant
is used to determine the correct value for a
reasonable font (if you ask for the system
font, you get a 12-point system font height, for
most others you get 10-point Geneva, etc.).

(1.0)
*/
UInt16
Localization_GetSingleLineTextWidth		(StringPtr			inString,
										 ThemeFontID		inThemeFontToUse)
{
	UInt16		result = 0;
	SInt16		fontSize = 0;
	Style		fontStyle = normal;
	Str255		fontName;
	
	
	if (getThemeFontInfo(inThemeFontToUse, inString/* string to calculate width of */,
							fontName, &fontSize, &fontStyle, &result/* string width */,
							null/* font height - unused */) != noErr) result = 0;
	return result;
}// GetSingleLineTextWidth


/*!
To conditionally rearrange two controls
horizontally so that the left and right
edges of the smallest rectangle bounding
both controls remains the same after they
trade places, use this method.

This method assumes that you have laid out
controls in the left-to-right arrangement
by default, so IF YOU ARE in left-to-right
localization, calling this routine has no
effect.  The first control is assumed to
be the leftmost one, and the second is
assumed to be the rightmost one of the two.

(1.0)
*/
void
Localization_HorizontallyPlaceControls	(ControlRef		inControl1,
										 ControlRef		inControl2)
{
	if (!Localization_IsLeftToRight())
	{
		ControlRef		control = null;
		Rect			controlRect,
						combinedRect;
		UInt16			width = 0;
		
		
		// find a rectangle whose horizontal expanse is exactly large enough to touch the edges of both controls
		GetControlBounds(inControl1, &combinedRect);
		GetControlBounds(inControl2, &controlRect);
		if (combinedRect.right < controlRect.right) combinedRect.right = controlRect.right;
		if (combinedRect.left > controlRect.left) combinedRect.left = controlRect.left;
		
		// switch the controls inside the smallest space that they both occupy
		GetControlBounds(inControl1, &controlRect);
		MoveControl(inControl1, (combinedRect.right - (controlRect.right - controlRect.left)), controlRect.top);
		GetControlBounds(inControl2, &controlRect);
		MoveControl(inControl2, combinedRect.left, controlRect.top);
	}
}// HorizontallyPlaceControls


/*!
To conditionally rearrange two dialog items
horizontally so that the left and right
edges of the smallest rectangle bounding
both controls remains the same after they
trade places, use this method.

This method assumes that you have laid out
controls in the left-to-right arrangement
by default, so IF YOU ARE in left-to-right
localization, calling this routine has no
effect.

(1.0)
*/
void
Localization_HorizontallyPlaceItems		(DialogRef const			inDialog,
										 DialogItemIndex const		inItemIndex1,
										 DialogItemIndex const		inItemIndex2)
{
	if (!Localization_IsLeftToRight())
	{
		ControlRef		control1 = null,
						control2 = null;
		Rect			controlRect,
						combinedRect;
		UInt16			width = 0;
		
		
		(OSErr)GetDialogItemAsControl(inDialog, inItemIndex1, &control1);
		(OSErr)GetDialogItemAsControl(inDialog, inItemIndex2, &control2);
		
		// find a rectangle whose horizontal expanse is exactly large enough to touch the edges of both controls
		GetControlBounds(control1, &combinedRect);
		GetControlBounds(control2, &controlRect);
		if (combinedRect.right < controlRect.right) combinedRect.right = controlRect.right;
		if (combinedRect.left > controlRect.left) combinedRect.left = controlRect.left;
		
		// switch the controls inside the smallest space that they both occupy
		GetControlBounds(control1, &controlRect);
		MoveDialogItem(inDialog, inItemIndex1,
						(combinedRect.right - (controlRect.right - controlRect.left)), controlRect.top);
		GetControlBounds(control2, &controlRect);
		MoveDialogItem(inDialog, inItemIndex2, combinedRect.left, controlRect.top);
	}
}// HorizontallyPlaceItems


/*!
To determine if the Localization module was
initialized such that text reads left to
right (such as in North America), use this
method.

(1.0)
*/
Boolean
Localization_IsLeftToRight		(void)
{
	Boolean		result = gLeftToRight;
	
	
	return result;
}// IsLeftToRight


/*!
To determine if the Localization module was
initialized such that text reads top to
bottom (such as in North America), use this
method.

(1.0)
*/
Boolean
Localization_IsTopToBottom		(void)
{
	Boolean		result = gTopToBottom;
	
	
	return result;
}// IsTopToBottom


/*!
To set the specified controls font, font
size, and font style to match the indicated
theme font, use this method.  If no API is
available to determine the exact font, a
reasonable approximation is made, usually
resulting in a choice of either the 12-point
system font or 10-point Geneva.

(1.0)
*/
OSStatus
Localization_SetControlThemeFontInfo	(ControlRef			inControl,
										 ThemeFontID		inThemeFontToUse)
{
	SInt16		fontSize = 0;
	Style		fontStyle = normal;
	Str255		fontName;
	UInt16		fontHeight = 0;
	OSStatus	result = noErr;
	
	
	result = getThemeFontInfo(inThemeFontToUse, null/* string to calculate width of - unused */,
								fontName, &fontSize, &fontStyle, null/* string width - unused */,
								&fontHeight);
	if (result == noErr)
	{
		result = setControlFontInfo(inControl, fontName, fontSize, fontStyle);
	}
	return result;
}// SetControlThemeFontInfo


/*!
Restores the font, font size, font style, and
text mode of the current graphics port.

(1.0)
*/
void
Localization_SetPortFontState	(GrafPortFontState const*		inState)
{
	if (inState != null)
	{
		TextFont(inState->fontID);
		TextSize(inState->fontSize);
		TextFace(inState->fontStyle);
		TextMode(inState->textMode);
	}
}// SetPortFontState


/*!
Works like Localization_SetUpSingleLineTextControlMax(),
except there is no limit on the width the resultant
control can have (it remains one line, as wide as the
text happens to be).

The chosen width for the control, in pixels, is
returned.

(1.0)
*/
UInt16
Localization_SetUpSingleLineTextControl		(ControlRef			inControl,
											 StringPtr			inTextContents,
											 ThemeFontID		inThemeFontToUse)
{
	return Localization_SetUpSingleLineTextControlMax(inControl, inTextContents, inThemeFontToUse,
														0/* means no limits */,
														kStringUtilities_TruncateAtEnd);
}// SetUpSingleLineTextControl


/*!
Automatically sets the contents, font family,
size, and style, width, and height of a text
control (editable or static) to accommodate the
given theme font.  The width of the control is
made sufficient to hold the specified text,
within the specified width limit (pass 0 to
have no width limits).  If the control text is
too wide, it is automatically changed to include
a trailing ellipsis (...) and truncated to fit.

Pass the special font code "-1" to indicate
that you want to set up a control to use the
Aqua-like huge system font, i.e. the system
font with size 14.

This routine is an easy way to set up a series
of commonly-needed text control settings in a
single step.  Even if you dont want all of the
settings of a control changed, it is usually
easier to call this routine first, and then use
the Mac OS Toolbox to repair damage to the
control properties.

The chosen width for the control, in pixels, is
returned.

(1.0)
*/
UInt16
Localization_SetUpSingleLineTextControlMax		(ControlRef							inControl,
												 StringPtr							inTextContents,
												 ThemeFontID						inThemeFontToUse,
												 UInt16 const						inMaximumAllowedWidth,
												 StringUtilitiesTruncationMethod	inTruncationMethod)
{
	enum
	{
		kStringPaddingH = 8		// arbitrary horizontal extra space, in pixels, just to be safe
	};
	UInt16		result = 0;
	SInt16		fontSize = 0;
	Style		fontStyle = normal;
	Str255		fontName;
	UInt16		fontHeight = 0;
	GrafPtr		oldPort = null;
	Str255		buffer;
	
	
	PLstrcpy(buffer, inTextContents);
	
	GetPort(&oldPort);
	SetPortWindowPort(GetControlOwner(inControl));
	
	// obtain a wealth of information about the requested theme font
	(OSStatus)getThemeFontInfo(inThemeFontToUse, buffer, fontName, &fontSize, &fontStyle,
								&result/* string width */, &fontHeight);
	
	// set the font characteristics to be the font derived above
	(OSStatus)setControlFontInfo(inControl, fontName, fontSize, fontStyle);
	
	// size the control to be just wide enough for the text
	result += kStringPaddingH; // add some (arbitrary) additional space to be sure
	if (inMaximumAllowedWidth > 0)
	{
		// a maximum was given; be sure the text isnt any longer
		(Boolean)StringUtilities_PTruncate(buffer, inMaximumAllowedWidth, inTruncationMethod);
	}
	SizeControl(inControl, result, fontHeight);
	
	// set the control text to be the specified text
	(OSStatus)SetControlData(inControl, kControlEditTextPart, kControlStaticTextTextTag,
								StrLength(buffer) * sizeof(char), (Ptr)(buffer + 1));
	
	SetPort(oldPort);
	
	return result;
}// SetUpSingleLineTextControlMax


/*!
To automatically set the contents, font family,
size, and style, and height of a text control
(editable or static) to accommodate the given
theme font, use this method.  The width of the
control is not changed.

This routine is an easy way to set up a series
of commonly-needed text control settings in a
single step.  Even if you dont want all of the
settings of a control changed, it is usually
easier to call this routine first, and then use
the Mac OS Toolbox to repair damage to the
control properties.

The chosen height for the control, in pixels,
is returned.

(1.0)
*/
UInt16
Localization_SetUpMultiLineTextControl		(ControlRef			inControl,
											 StringPtr			inTextContents,
											 ThemeFontID		inThemeFontToUse)
{
	UInt16		result = 0;
	UInt16		textExpanseH = 0;
	SInt16		fontSize = 0;
	Style		fontStyle = normal;
	Str255		fontName;
	UInt16		fontHeight = 0,
				charWidth = 0;
	GrafPtr		oldPort = null;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(GetControlOwner(inControl));
	
	// obtain a wealth of information about the requested theme font
	// (in the quick and dirty calculation, approximate character
	// widths are used with a small margin for error; if no such
	// short-cut is taken, the error is less but the calculation is
	// significantly slower)
	(OSStatus)getThemeFontInfo(inThemeFontToUse,
								(inThemeFontToUse == kThemeSystemFont)
								? "\pA"
								: "\pA"/* string to calculate width of */, fontName, &fontSize,
								&fontStyle, &charWidth/* string width */, &fontHeight);
	
	// set the font characteristics to be the font derived above
	(OSStatus)setControlFontInfo(inControl, fontName, fontSize, fontStyle);
	
	// size the control to be just high enough for the text
	{
		GrafPortFontState	fontState;
		Rect				controlRect;
		SInt16				fontID = 0;
		
		
		// save port settings
		Localization_GetPortFontState(&fontState);
		
		// set port font so CharWidth() can be used
	#if TARGET_API_MAC_OS8
		GetFNum(fontName, &fontID);
	#else
		fontID = FMGetFontFamilyFromName(fontName);
	#endif
		TextFont(fontID);
		TextSize(fontSize);
		TextFace(fontStyle);
		
		GetControlBounds(inControl, &controlRect);
		textExpanseH = controlRect.right - controlRect.left;
		#define CHEAP_TEXT_HEIGHT_CALCULATION 0
		#if CHEAP_TEXT_HEIGHT_CALCULATION
		//result = controlRect.bottom - controlRect.top; // initially...
		result = (StrLength(inTextContents) * charWidth / textExpanseH + 1) * fontHeight;
		#else
		{
			unsigned char*		ptr = inTextContents;
			SInt16				length = 0, // count of characters
								width = 0;  // pixel width of characters so far on one line
			
			
			result = 0;
			while (length < StrLength(inTextContents))
			{
				// find most words that will fit in one line
				// (this algorithm doesnt work well for fonts that kern significantly)
				while ((width < textExpanseH) && (ptr[0] != '\n') && (length < StrLength(inTextContents)))
				{
					width += CharWidth(ptr[0]);
					ptr++, length++;
				}
				
				// if the line wasnt broken at a word, back up the pointer before continuing
				// (the Mac OS will wrap the entire word when rendering it anyway)
				while (((isalpha((char)ptr[0])) || (isdigit((char)ptr[0]))) && (length > 0))
				{
					ptr--, length--;
				}
				width = 0;
				result += fontHeight;
			}
		}
		#endif
		{
			UInt16				extraBlanks = 0;
			register UInt16		i = 0;
			
			
			for (i = 1; i < StrLength(inTextContents); i++)
			{
				// make room for blank lines, if the text contains new-line characters
				if (inTextContents[i] == (unsigned char)'\n') extraBlanks++;
			}
			result += (extraBlanks * fontHeight);
		}
		SizeControl(inControl, textExpanseH, result);
		
		// restore port settings
		Localization_SetPortFontState(&fontState);
	}
	
	// set the control text to be the specified text
	(OSStatus)SetControlData(inControl, kControlEditTextPart, kControlStaticTextTextTag,
								StrLength(inTextContents) * sizeof(char), (Ptr)(inTextContents + 1));
	
	SetPort(oldPort);
	
	return result;
}// SetUpMultiLineTextControl


/*!
To set the font, size, and style for text drawn
in the current graphics port to correspond to
the specified theme font (or the closest match,
in absence of the appropriate APIs), use this
method.  Useful font information about the new
font is returned (all pointer parameters MUST
be non-null).

(1.0)
*/
void
Localization_UseThemeFont		(ThemeFontID		inThemeFontToUse,
								 StringPtr			outFontName,
								 SInt16*			outFontSizePtr,
								 Style*				outFontStylePtr)
{
	OSStatus		error = noErr;
	Boolean			haveAppearance1_1 = false;
	
	
	// Appearance 1.1 or later?
	{
		long		appv = 0L;
		
		
		error = Gestalt(gestaltAppearanceVersion, &appv);
		if (error == noErr)
		{
			UInt8		majorRev = Releases_GetMajorRevisionForVersion(appv),
						minorRev = Releases_GetMinorRevisionForVersion(appv);
			
			
			haveAppearance1_1 = (((majorRev == 0x01) && (minorRev >= 0x01)) || (majorRev > 0x01));
		}
	}
	
	error = noErr;
	if (haveAppearance1_1)
	{
		error = GetThemeFont((inThemeFontToUse != -1) ? inThemeFontToUse : kThemeSystemFont,
								GetScriptManagerVariable(smSysScript)/* script code */,
								outFontName, outFontSizePtr, outFontStylePtr);
	}
	
	// without the appropriate Appearance Manager APIs, use the Script Manager
	if ((!haveAppearance1_1) || (error != noErr))
	{
		SInt16		fontID = GetScriptVariable(GetScriptManagerVariable(smSysScript),
										((inThemeFontToUse == kThemeSystemFont) || (inThemeFontToUse == -1))
										? smScriptSysFond
										: smScriptAppFond);
		
		
	#if TARGET_API_MAC_OS8
		GetFontName(fontID, outFontName);
	#else
		(OSStatus)FMGetFontFamilyName(fontID, outFontName);
	#endif
		*outFontSizePtr = (inThemeFontToUse == kThemeSystemFont) ? 12 : 10;
		*outFontStylePtr = (inThemeFontToUse == kThemeSmallEmphasizedSystemFont) ? bold : normal;
	}
	
	// special case for Aqua-like dialogs with huge title text
	if (inThemeFontToUse == -1) *outFontSizePtr = 14;
	
	// change the font settings of the current graphics port to match the results
	{
		SInt16		fontID = 0;
		
		
	#if TARGET_API_MAC_OS8
		GetFNum(outFontName, &fontID);
	#else
		fontID = FMGetFontFamilyFromName(outFontName);
	#endif
		TextFont(fontID);
	}
	TextSize(*outFontSizePtr);
	TextFace(*outFontStylePtr);
}// UseThemeFont


/*!
To conditionally rearrange two controls
vertically so that the top and bottom edges
of the smallest rectangle bounding both
controls remains the same after they trade
places, use this method.

This method assumes that you have laid out
controls in the top-to-bottom arrangement
by default, so IF YOU ARE in top-to-bottom
localization, calling this routine has no
effect.  The first control is assumed to be
the topmost one, and the second is assumed
to be the bottommost one of the two.

(1.0)
*/
void
Localization_VerticallyPlaceControls	(ControlRef		inControl1,
										 ControlRef		inControl2)
{
	if (!Localization_IsTopToBottom())
	{
		ControlRef		control = null;
		Rect			controlRect,
						combinedRect;
		UInt16			width = 0;
		
		
		// find a rectangle whose vertical expanse is exactly large enough to touch the edges of both controls
		GetControlBounds(inControl1, &combinedRect);
		GetControlBounds(inControl2, &controlRect);
		if (combinedRect.bottom < controlRect.bottom) combinedRect.bottom = controlRect.bottom;
		if (combinedRect.top > controlRect.top) combinedRect.top = controlRect.top;
		
		// switch the controls inside the smallest space that they both occupy
		GetControlBounds(inControl1, &controlRect);
		MoveControl(inControl1, controlRect.left, (combinedRect.bottom - (controlRect.bottom - controlRect.top)));
		GetControlBounds(inControl2, &controlRect);
		MoveControl(inControl2, controlRect.left, combinedRect.top);
	}
}// VerticallyPlaceControls


/*!
To conditionally rearrange two dialog items
vertically so that the top and bottom edges
of the smallest rectangle bounding both
controls remains the same after they trade
places, use this method.

This method assumes that you have laid out
controls in the top-to-bottom arrangement
by default, so IF YOU ARE in top-to-bottom
localization, calling this routine has no
effect.

(1.0)
*/
void
Localization_VerticallyPlaceItems		(DialogRef const			inDialog,
										 DialogItemIndex const		inItemIndex1,
										 DialogItemIndex const		inItemIndex2)
{
	if (!Localization_IsTopToBottom())
	{
		ControlRef		control1 = null,
						control2 = null;
		Rect			controlRect,
						combinedRect;
		UInt16			width = 0;
		
		
		// find a rectangle whose vertical expanse is exactly large enough to touch the edges of both controls
		GetControlBounds(control1, &combinedRect);
		GetControlBounds(control2, &controlRect);
		if (combinedRect.bottom < controlRect.bottom) combinedRect.bottom = controlRect.bottom;
		if (combinedRect.top > controlRect.top) combinedRect.top = controlRect.top;
		
		// switch the controls inside the smallest space that they both occupy
		GetControlBounds(control1, &controlRect);
		MoveDialogItem(inDialog, inItemIndex1,
						controlRect.left, (combinedRect.bottom - (controlRect.bottom - controlRect.top)));
		GetControlBounds(control2, &controlRect);
		MoveDialogItem(inDialog, inItemIndex2, controlRect.left, combinedRect.top);
	}
}// VerticallyPlaceItems


//
// internal methods
//

/*!
Provides a wealth of information about
the specified theme font (or about an
appropriate substitute font, if the
exact theme font cannot be determined
because Appearance 1.1 is not installed).

The font name, size, and style of the
indicated theme font are returned.  The
string width parameter is optional (if
you provide a string and string width
storage, the width of that string in
pixels is returned).  The font height
parameter is also optional.

(1.0)
*/
static OSStatus
getThemeFontInfo	(ThemeFontID		inThemeFontToUse,
					 StringPtr			inString,
					 Str255				outFontName,
					 SInt16*			outFontSizePtr,
					 Style*				outFontStylePtr,
					 UInt16*			outStringWidthPtr,
					 UInt16*			outFontHeightPtr)
{
	GrafPortFontState	fontState;
	OSStatus			result = noErr;
	
	
	// save port settings
	Localization_GetPortFontState(&fontState);
	
	// set the font, and get additional metrics information if possible
	Localization_UseThemeFont(inThemeFontToUse, outFontName, outFontSizePtr, outFontStylePtr);
	if ((outStringWidthPtr != null) && (inString != null))
	{
		*outStringWidthPtr = StringWidth(inString);
	}
	if (outFontHeightPtr != null)
	{
		FontInfo		theInfo;
		
		
		GetFontInfo(&theInfo);
		*outFontHeightPtr = theInfo.ascent + theInfo.descent + INTEGER_DOUBLED(theInfo.leading);
	}
	
	// restore port settings
	Localization_SetPortFontState(&fontState);
	
	return result;
}// getThemeFontInfo


/*!
To change the text appearance of a control
displaying static or editable text, use
this convenient method.

(1.0)
*/
static OSStatus
setControlFontInfo	(ControlRef			inControl,
					 StringPtr			inFontName,
					 SInt16 const		inFontSize,
					 Style const		inFontStyle)
{
	ControlFontStyleRec		styleRecord;
	SInt16					fontID = 0;
	OSErr					result = noErr;
	
	
#if TARGET_API_MAC_OS8
	GetFNum(inFontName, &fontID);
#else
	fontID = FMGetFontFamilyFromName(inFontName);
#endif
	styleRecord.flags = kControlUseFontMask | kControlUseFaceMask | kControlUseSizeMask;
	styleRecord.font = fontID;
	styleRecord.size = inFontSize;
	styleRecord.style = inFontStyle;
	
	result = SetControlData(inControl, kControlNoPart, kControlFontStyleTag,
							sizeof(styleRecord), (Ptr)&styleRecord);
	return result;
}// setControlFontInfo