/*###############################################################

	MacLook.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#define PREFMASTER

// standard-C includes
#ifdef __cplusplus
#	include <cstdio>
#else
#	include <stdio.h>
#endif
#include <ctype.h>

// Mac includes
#include <Events.h>
#include <OpenTransportProviders.h>
#include <Scrap.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "Clipboard.h"
#include "PreferencesData.h"
#include "NetEvents.h"
#include "ConnectionData.h"
#include "EventLoop.h"
#include "FileUtilities.h"
#include "Connections.h"
#include "RealScreen.h"
#include "vsdata.h"
#include "vskeys.h"
#include "Terminal.h"
#include "TektronixVirtualGraphics.h"
#include "TektronixRealGraphics.h"
#include "network.proto.h"
#include "TelnetMenus.h"
#include "Preferences.h"
#include "MacLook.h"
#include "DialogUtilities.h"
#include "telneterrors.h"
#include "network.proto.h"



// variables
static short	gGraphCurs = 0;
static short	gNumGraphs = 0;
static long		gOldGraphs[kMaximumAllowedGraphicsWindows];



//
// public methods
//

/*!
This powerful routine iterates through the
device list, determines the monitor that
contains the largest chunk of the given
windows structure region, and then returns
the gray region of that device.  Usually,
this will be the entire chunk of the desktop
that crosses the device, unless its the
main device, in which case the region will
have the menu bar height subtracted from it.

This routine is very useful for implementing
operations such as zooming windows, since
you can automatically use the rectangle of
the monitor that contains the window for
zoomed-state calculations.  You should use
this routine in place of the routine
GetScreenBounds() whenever possible.

(3.0)
*/
void
MacLook_GetWindowDeviceGrayRect	(const WindowRef		inWindow,
								 Rect*					outRectPtr)
{
	if (outRectPtr != null)
	{
		GDHandle		device = null,
						windowDevice = null;
		Rect			windowRect,
						areaRect;
		SInt32			area = 0L,
						areaMax = 0L;
		
		
		// obtain global structure rectangle of the window
		MacLook_GetWindowRegionBounds(inWindow, kWindowStructureRgn, &windowRect);
		
		// check for monitors in use - find the one containing the largest chunk of the window
		for (device = GetDeviceList(); (device != null); device = GetNextDevice(device))
		{
			if (TestDeviceAttribute(device, screenDevice)) if (TestDeviceAttribute(device, screenActive))
			{
				(Boolean)SectRect(&(**device).gdRect, &windowRect, &areaRect);
				
				area = (SInt32)(areaRect.bottom - areaRect.top) * (areaRect.right - areaRect.left);
				if (area > areaMax)
				{
					areaMax = area;
					windowDevice = device;
				}
			}
		}
		*outRectPtr = (**windowDevice).gdRect;
		if (TestDeviceAttribute(windowDevice, mainScreen)) outRectPtr->top += GetMBarHeight();
	}
}// GetWindowDeviceGrayRect


/*!
Use this method in place of any call to
GetWindowBounds() or, in many cases,
GetWindowRegion().  It automatically uses
the most recent API (namely GetWindowBounds()
if it is available), for maximum Carbon
compatibility.  For Mac OS 8.0 and 8.1, the
window rectangle is determined by finding
the region first, which is computationally
more expensive.

(3.0)
*/
void
MacLook_GetWindowRegionBounds	(const WindowRef	inWindow,
								 WindowRegionCode	inWindowRegionCode,
								 Rect*				outRectPtr)
{
	if (outRectPtr != null)
	{
#if TARGET_API_MAC_OS8
		if (FlagManager_Test(kFlagCarbonWindowManager))
		{
#endif
			GetWindowBounds(inWindow, inWindowRegionCode, outRectPtr);
#if TARGET_API_MAC_OS8
		}
		else
		{
			RgnHandle		region = Memory_NewRegion();
			
			
			if (region != null)
			{
				if (GetWindowRegion(inWindow, inWindowRegionCode, region) == noErr)
				{
					GetRegionBounds(region, outRectPtr);
				}
				Memory_DisposeRegion(&region);
			}
		}
#endif
	}
}// GetWindowRegionBounds


/*!
Invalidates the entire area of a control.  This
routine calls MacLook_InvalidateRectangle().

(3.0)
*/
void
MacLook_InvalidateControlRectangle		(const ControlRef	inControl)
{
	Rect		bounds;
	
	
	GetControlBounds(inControl, &bounds);
	MacLook_InvalidateRectangle(GetControlOwner(inControl), &bounds);
}// InvalidateControlRectangle


/*!
Invalidates the entire area of a windows
graphics port.  This routine calls
MacLook_InvalidateRectangle().

(3.0)
*/
void
MacLook_InvalidatePortRectangle	(const WindowRef	inWindow)
{
	Rect		windowPortRect;
	
	
	GetPortBounds(GetWindowPort(inWindow), &windowPortRect);
	MacLook_InvalidateRectangle(inWindow, &windowPortRect);
}// InvalidatePortRectangle


/*!
Use this method in place of any call to
InvalRect().  It automatically uses the
most recent API (namely InvalWindowRect()
if it is available), for maximum Carbon
compatibility.

(3.0)
*/
void
MacLook_InvalidateRectangle		(const WindowRef	inWindow,
								 const Rect*		inRectPtr)
{
#if TARGET_API_MAC_OS8
	GrafPtr		oldPort = null;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(inWindow);
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
#endif
		InvalWindowRect(inWindow, inRectPtr);
#if TARGET_API_MAC_OS8
	}
	else
	{
		InvalRect(inRectPtr);
	}
	SetPort(oldPort);
#endif
}// InvalidateRectangle


/*!
Use this method in place of any call to
InvalRgn().  It automatically uses the
most recent API (namely InvalWindowRgn()
if it is available), for maximum Carbon
compatibility.

(3.0)
*/
void
MacLook_InvalidateRegion		(const WindowRef	inWindow,
								 const RgnHandle	inRegion)
{
#if TARGET_API_MAC_OS8
	GrafPtr		oldPort = null;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(inWindow);
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
#endif
		InvalWindowRgn(inWindow, inRegion);
#if TARGET_API_MAC_OS8
	}
	else
	{
		InvalRgn(inRegion);
	}
	SetPort(oldPort);
#endif
}// InvalidateRegion


/*!
Prompts the user for which keys to use for kill,
stop and start.

(2.6)
*/
void
MacLook_SetUpKeys	(const TelnetScreenID		inScreenID)
{
	ConnectionDataPtr	connectionDataPtr = ScreenManager_GetConnectionDataOf(inScreenID);
	DialogPtr			dtemp = null;
	ModalFilterUPP		upp = null;
	Rect				dBox;
	short				dItem = 0,
						kItem = 0,
						sItem = 0,
						eItem = 0;
	Handle				kbox = null,
						sbox = null,
						ebox = null;
	Str31				keyString;
	char*				tempspot = null;
	
	
	tempspot = (char*)Memory_NewPtr(256);
	if (tempspot != null)
	{
		dtemp = GetNewDialog(SetupDLOG, null, kInFront);
		
		SetDialogDefaultItem(dtemp, DLOGOk);
		SetDialogCancelItem(dtemp, DLOGCancel);
		(UInt16)Localization_AdjustDialogButtons(dtemp, true/* OK */, true/* Cancel */);
		
		GetDialogItem(dtemp, killbox, &kItem, &kbox, &dBox);	
		GetDialogItem(dtemp, stopbox, &eItem, &ebox, &dBox);	
		GetDialogItem(dtemp, startbox, &sItem, &sbox, &dBox);	
		
		*tempspot = 0;
		if (connectionDataPtr->controlKey.suspend > 0)
		{
			keyString[0] = 2;
			keyString[1] = '^';
			keyString[2] = STATIC_CAST(connectionDataPtr->controlKey.suspend ^ 0x40, char);
			SetDialogItemText(ebox, keyString);
			SelectDialogItemText(dtemp, stopbox, 0, 32767);
		}
		if (connectionDataPtr->controlKey.resume > 0)
		{
			keyString[0] = 2;
			keyString[1] = '^';
			keyString[2] = STATIC_CAST(connectionDataPtr->controlKey.resume ^ 0x40, char);
			SetDialogItemText(sbox, keyString);
			SelectDialogItemText(dtemp, startbox, 0, 32767);
		}
		if (connectionDataPtr->controlKey.interrupt > 0)
		{
			keyString[0] = 2;
			keyString[1] = '^';
			keyString[2] = STATIC_CAST(connectionDataPtr->controlKey.interrupt ^ 0x40, char);
			SetDialogItemText(kbox, keyString);
			SelectDialogItemText(dtemp, killbox, 0, 32767);
		}
		
		TransitionDialogDisplay(dtemp);
		
		upp = NewModalFilterUPP(StandardDialogEventFilter);
		dItem = 0;
		do
		{
			ModalDialog(upp, &dItem);
			
			// Intermediate check.  If the user hits a key,
			// put its number in the box.
			GetDialogItemText(kbox, (StringPtr)tempspot);
			StringUtilities_PToCInPlace((StringPtr)tempspot);
			if (*tempspot < 32 && *tempspot > 0)
			{
				sprintf(tempspot, "^%c", *tempspot ^ 64);
				StringUtilities_CToPInPlace(tempspot);
				SetDialogItemText(kbox, (StringPtr)tempspot);
				SelectDialogItemText(dtemp, killbox, 0, 32767);
			}
			GetDialogItemText(ebox, (StringPtr)tempspot);
			StringUtilities_PToCInPlace((StringPtr)tempspot);
			if (*tempspot < 32 && *tempspot > 0)
			{
				sprintf(tempspot, "^%c", *tempspot ^ 64);
				StringUtilities_CToPInPlace(tempspot);
				SetDialogItemText(ebox, (StringPtr)tempspot);
				SelectDialogItemText(dtemp, stopbox, 0, 32767);
			}
			GetDialogItemText(sbox, (StringPtr)tempspot);
			StringUtilities_PToCInPlace((StringPtr)tempspot);
			if (*tempspot < 32 && *tempspot > 0)
			{
				sprintf(tempspot, "^%c", *tempspot ^ 64);
				StringUtilities_CToPInPlace(tempspot);
				SetDialogItemText(sbox, (StringPtr)tempspot);
				SelectDialogItemText(dtemp, startbox, 0, 32767);
			}	
		} while ((dItem != DLOGOk) && (dItem != DLOGCancel));
		DisposeModalFilterUPP(upp), upp = null;
		
		if (dItem != DLOGCancel)
		{
			GetDialogItemText(kbox, (StringPtr)tempspot);
			StringUtilities_PToCInPlace((StringPtr)tempspot);
			if (*tempspot != '^') connectionDataPtr->controlKey.interrupt = -1;
			else connectionDataPtr->controlKey.interrupt = toupper(*(tempspot + 1)) ^ 0x40;
			
			GetDialogItemText(ebox, (StringPtr)tempspot);
			StringUtilities_PToCInPlace((StringPtr)tempspot);
			if (*tempspot != '^') connectionDataPtr->controlKey.suspend = -1;
			else connectionDataPtr->controlKey.suspend = toupper(*(tempspot + 1)) ^ 0x40;
			
			GetDialogItemText(sbox, (StringPtr)tempspot);
			StringUtilities_PToCInPlace((StringPtr)tempspot);
			if (*tempspot != '^') connectionDataPtr->controlKey.resume = -1;
			else connectionDataPtr->controlKey.resume = toupper(*(tempspot + 1)) ^ 0x40;
			
			TransitionDialogClose(dtemp);
		}
		else TransitionDialogCloseCancel(dtemp);
	}
	Memory_DisposePtr(&((Ptr)tempspot));
	DisposeDialog(dtemp), dtemp = null;
}// SetUpKeys


/*!
Validates the entire area of a control.  This
routine calls MacLook_ValidateRectangle().

(3.0)
*/
void
MacLook_ValidateControlRectangle	(const ControlRef	inControl)
{
	Rect		bounds;
	
	
	GetControlBounds(inControl, &bounds);
	MacLook_ValidateRectangle(GetControlOwner(inControl), &bounds);
}// ValidateControlRectangle


/*!
Validates the entire area of a windows
graphics port.  This routine calls
MacLook_ValidateRectangle().

(3.0)
*/
void
MacLook_ValidatePortRectangle	(const WindowRef	inWindow)
{
	Rect		windowPortRect;
	
	
	GetPortBounds(GetWindowPort(inWindow), &windowPortRect);
	MacLook_ValidateRectangle(inWindow, &windowPortRect);
}// ValidatePortRectangle


/*!
Use this method in place of any call to
ValidRect().  It automatically uses the
most recent API (namely ValidWindowRect()
if it is available), for maximum Carbon
compatibility.

(3.0)
*/
void
MacLook_ValidateRectangle		(const WindowRef	inWindow,
								 const Rect*		inRectPtr)
{
#if TARGET_API_MAC_OS8
	GrafPtr		oldPort = null;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(inWindow);
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
#endif
		ValidWindowRect(inWindow, inRectPtr);
#if TARGET_API_MAC_OS8
	}
	else
	{
		ValidRect(inRectPtr);
	}
	SetPort(oldPort);
#endif
}// ValidateRectangle


/*!
Use this method in place of any call to
ValidRgn().  It automatically uses the
most recent API (namely ValidWindowRgn()
if it is available), for maximum Carbon
compatibility.

(3.0)
*/
void
MacLook_ValidateRegion			(const WindowRef	inWindow,
								 const RgnHandle	inRegion)
{
#if TARGET_API_MAC_OS8
	GrafPtr		oldPort = null;
	
	
	GetPort(&oldPort);
	SetPortWindowPort(inWindow);
	if (FlagManager_Test(kFlagCarbonWindowManager))
	{
#endif
		ValidWindowRgn(inWindow, inRegion);
#if TARGET_API_MAC_OS8
	}
	else
	{
		ValidRgn(inRegion);
	}
	SetPort(oldPort);
#endif
}// ValidateRegion


void
destroyGraphics	(short		dnum)
{
	short	i = 1;
	
	
	while ((i <= Connections_GetCount()) && (dnum != ScreenManager_GetConnectionDataOf(i)->TEK.graphicsID)) i++;
	if (i <= Connections_GetCount())
	{
		ConnectionDataPtr		theScreen = ScreenManager_GetConnectionDataOf(i);
		
		
		theScreen->TEK.graphicsID = -1;
		theScreen->termstate = VTEKTYPE;
		theScreen->enabled = 1;
	}
	else
	{
		i = 0;
		while ((i < gNumGraphs) && (dnum != gOldGraphs[i])) i++;
		while (i < (kMaximumAllowedGraphicsWindows - 1)) gOldGraphs[i] = gOldGraphs[++i]; // close the gap
		gNumGraphs--;
	}
	VGclose(dnum);
}// destroyGraphics


short
detachGraphics	(short		dnum)
{
	short	result = -1;
	short	i = 1;
	
	
	while ((i <= Connections_GetCount()) && (dnum != ScreenManager_GetConnectionDataOf(i)->TEK.graphicsID)) i++;
	
	if (i > Connections_GetCount()) result = -1;
	else
	{
		ConnectionDataPtr		theScreen = ScreenManager_GetConnectionDataOf(i);
		
		
		gOldGraphs[gNumGraphs++] = dnum;
		theScreen->TEK.graphicsID = -1;
		theScreen->termstate = VTEKTYPE;
		theScreen->enabled = 1;
		RGdetach(dnum);
		result = 0;
	}
	return result;
}// detachGraphics


short
getgraphcurs	(void)
{
	return gGraphCurs;
}// getgraphcurs


void
setgraphcurs	(void)
{
	gGraphCurs = 1;
}// setgraphcurs


void
TekDisable	(short		vg)
{
	short	change = FindByVG(vg);
	
	
	if (change >= 0) ScreenManager_GetConnectionDataOf(change)->enabled = 0;
}// TekDisable


void
TekEnable	(short		vg)
{
	short	change = FindByVG(vg);
	
	
	if (change >= 0) ScreenManager_GetConnectionDataOf(change)->enabled = 1;
}// TekEnable


void
unsetgraphcurs	(void)
{
	gGraphCurs = 0;
}// unsetgraphcurs


short
VGalive	(short		dnum)
{
	short	result = 0;
	short	i = 1;
	
	
	while ((i <= Connections_GetCount()) && (dnum != ScreenManager_GetConnectionDataOf(i)->TEK.graphicsID)) i++;
	
	if (i <= Connections_GetCount())
		result = 1;
	else
	{
		i = 0;
		while ((i < gNumGraphs) && (dnum != gOldGraphs[i])) i++;
		if (i < gNumGraphs) result = 1;
	}
	
	return result;
}// VGalive