/*###############################################################

	MacroSetupDialog.c
	
	MacTelnet
		 1998-2003 by Kevin Grant.
		 2001-2003 by Ian Anderson.
		 1986-1994 University of Illinois Board of Trustees
		(see About box for full list of U of I contributors).
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.
	
	You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <ControlDefinitions.h>
#include <Errors.h>
#include <Navigation.h>
#include <PLStringFuncs.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "TelnetMacrosDialog.r"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "BasicTypesAE.h"
#include "Console.h"
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "DialogResources.h"
#include "FolderManager.h"
#include "InstallAE.h"
#include "MacroSetupDialog.h"
#include "Preferences.h"
#include "ResourceFileManager.h"
#include "StringResources.h"
#include "TelnetDescriptors.h"
#include "TelnetErrors.h"
#include "TelnetMacros.h"
#include "TelnetMenus.h"
#include "Terminology.h"



//
// variables
//

static DialogRef			gMacroSetupDialog = null; // the Mac OS dialog pointer
static WindowInfoRef		gMacroSetupDialogWindowInfo;
static TelnetMacroMode		gPendingMacroMode = kTelnetMacroModeCommandDigit;
static UInt16				gChosenActiveSet = 1,
							gDisplayedSet = 1; // one-based!
static MacroSet				gTempSets[MACRO_SET_COUNT] = { null, null, null, null, null };
static ControlKeyFilterUPP	gMacroFieldKeyFilterUPP = null;
static DialogItemIndex		gTabFocusedFields[MACRO_SET_COUNT] =
							{
								// used to remember the text field from each pane that was focused
								iMacroSetupFirstMacroTextField,
								iMacroSetupFirstMacroTextField,
								iMacroSetupFirstMacroTextField,
								iMacroSetupFirstMacroTextField,
								iMacroSetupFirstMacroTextField
							};

//
// internal methods
//

static void				chooseSet					(UInt16					inMacroSetNumber,
													 Boolean				inSaveOld);

static void				focusField					(DialogItemIndex		inItemIndex);

static pascal Boolean	handleNewSize				(WindowRef				inWindow,
													 SInt32					inDeltaSizeX,
													 SInt32					inDeltaSizeY,
													 SInt32					inData);

static Boolean			isDisplayedSetEmptyTellUser	(void);

static pascal ControlKeyFilterResult	macroFieldKeyFilter		(ControlRef		inControl,
													 SInt16*				inoutKeyCode,
													 SInt16*				inoutCharCode,
													 UInt16*				inoutModifiers);

static pascal Boolean	macroSetupDialogEventFilter	(DialogPtr				inDialogPtr,
													 EventRecord*			inoutEventPtr,
													 short*					outItemIndex);

static void				rememberFocus				(UInt16					inWhichSet);

static void				restoreFocus				(UInt16					inWhichSet);

static void				saveActiveSetNumber			(void);

static pascal Boolean	setButtonsByMode			(ControlRef				inForWhichControl,
													 SInt16					inColorDepth,
													 SInt16					inDeviceFlags,
													 GDHandle				inTargetDevice,
													 SInt32					inMode,
													 SInt32					inUnused);

static void				setButtonsByModeOffscreen	(TelnetMacroMode		inMode);

static void				setTempActive				(UInt16					inMacroSetNumber);



//
// public methods
//

/*!
This method is used to initialize the Macro Setup
dialog box.  It creates the dialog box invisibly,
and uses the global macros data to set up the
corresponding controls in the dialog box panes.

The extended dialog resource for the preferences
dialog causes the Appearance Manager to automatically
construct an embedding hierarchy, so the dialogs
root control and container controls are supposed to
be set up once GetNewDialog() returns.

(3.0)
*/
void
MacroSetupDialog_Init	(void)
{
	Cursors_DeferredUseWatch(30); // if it takes more than half a second to initialize, show the watch cursor
	
	// copy the global macros into temporary macro space, for mucking with by the dialog
	gChosenActiveSet = Macros_GetActiveSetNumber();
	{
		register UInt16		i = 0;
		
		
		for (i = 0; i < MACRO_SET_COUNT; ++i)
		{
			gTempSets[i] = Macros_NewSet();
			Macros_SetActiveSetNumber(i + 1); // its one-based
			Macros_Copy(Macros_GetActiveSet(), gTempSets[i]);
		}
	}
	Macros_SetActiveSetNumber(gChosenActiveSet);
	gPendingMacroMode = Macros_GetMode();
	
	GetNewDialogWithWindowInfo(kDialogIDMacroSetup, &gMacroSetupDialog, &gMacroSetupDialogWindowInfo);
	
	if (gMacroSetupDialog != null)
	{
		ControlRef		control = null;
		
		
		// determine if the Help button should be disabled...
		SetUpHelpButtonDialogItem(gMacroSetupDialog, iMacroSetupAppleGuideButton);
		
		// always open to the active macro set
		(OSStatus)GetDialogItemAsControl(gMacroSetupDialog, iMacroSetupMacroTextEquivalentsGroupBox, &control);
		SetControlMaximum(control, MACRO_SET_COUNT);
		SetControlMinimum(control, 1);
		SetControlValue(control, gChosenActiveSet);
		chooseSet(gChosenActiveSet, false/* save old */);
		
		// install a key filter to prevent user errors
		if (FlagManager_Test(kFlagCarbonControlManager))
		{
			// key filters do *not* seem to work prior to Mac OS 8.5 (they cause a crash when typing)
			ControlRef			field = null;
			register SInt16		i = 0;
			
			
			gMacroFieldKeyFilterUPP = NewControlKeyFilterUPP(macroFieldKeyFilter);
			for (i = 0; i < MACRO_COUNT; ++i)
			{
				(OSStatus)GetDialogItemAsControl(gMacroSetupDialog, iMacroSetupFirstMacroTextField + i, &field);
				(OSStatus)SetControlData(field, kControlEditTextPart, kControlEditTextKeyFilterTag,
										sizeof(gMacroFieldKeyFilterUPP), (Ptr)&gMacroFieldKeyFilterUPP);
			}
		}
		
		// initialize controls
		setTempActive(gChosenActiveSet);
		setButtonsByModeOffscreen(Macros_GetMode());
		
		// adjust the OK and Cancel buttons so they are big enough for their titles
		SetDialogDefaultItem(gMacroSetupDialog, iMacroSetupOKButton);
		SetDialogCancelItem(gMacroSetupDialog, iMacroSetupCancelButton);
		(UInt16)Localization_AdjustDialogButtons(gMacroSetupDialog, true/* OK */, true/* Cancel */);
		Localization_AdjustHelpButtonItem(gMacroSetupDialog, iMacroSetupAppleGuideButton);
		
		// adjust other important buttons
		(UInt16)Localization_AutoSizeButtonItem(gMacroSetupDialog, iMacroSetupSetCurrentActiveButton, 0);
		
		// set up the Window Info information
		{
			WindowResizeResponderProcPtr	resizeProc = handleNewSize;
			
			
			WindowInfo_SetWindowDescriptor(gMacroSetupDialogWindowInfo, kTelnetDialogDescriptorMacroSetup);
			WindowInfo_SetWindowResizeLimits(gMacroSetupDialogWindowInfo,
													MACROSETUP_DIALOG_HT/* minimum height */,
													MACROSETUP_DIALOG_WD/* minimum width */,
													MACROSETUP_DIALOG_HT/* maximum height */,
													SHRT_MAX/* maximum width */);
			WindowInfo_SetWindowResizeResponderUPP(gMacroSetupDialogWindowInfo, resizeProc, 0L);
		}
		WindowInfo_SetForDialog(gMacroSetupDialog, gMacroSetupDialogWindowInfo);
		
		// use the preferred rectangle, if any
		{
			Point	deltaSize;
			
			
			SetPt(&deltaSize, MACROSETUP_DIALOG_WD, MACROSETUP_DIALOG_HT); // initially...
			deltaSize = Preferences_ArrangeWindow(GetDialogWindow(gMacroSetupDialog), deltaSize,
													kTelnetWindowPreferenceTag_MacroSetup);
			
			(SInt16)handleNewSize(GetDialogWindow(gMacroSetupDialog),
									(SInt32)deltaSize.h, (SInt32)deltaSize.v, 0L);
		}
	}
}// Init


/*!
Call this method in the exit method of the program
to ensure that this dialogs memory allocations
get destroyed.  You can also call this method after
you are done using this dialog, provided you call
MacroSetupDialog_Init() before you use this
dialog again.

(3.0)
*/
void
MacroSetupDialog_Done	(void)
{
	{
		register UInt16		i = 0;
		
		
		for (i = 0; i < MACRO_SET_COUNT; ++i) Macros_DisposeSet(&gTempSets[i]);
	}
	WindowInfo_Dispose(gMacroSetupDialogWindowInfo);
	DisposeDialog(gMacroSetupDialog), gMacroSetupDialog = null;
	DisposeControlKeyFilterUPP(gMacroFieldKeyFilterUPP), gMacroFieldKeyFilterUPP = null;
}// Done


/*!
This method displays and handles events in the
macro setup dialog box.  If the user saved
changes, true is returned; otherwise, false is
returned.

(3.0)
*/
Boolean
MacroSetupDialog_Display	(void)
{
	Boolean		result = false;
	
	
	if (gMacroSetupDialog == null) Alert_ReportOSStatus(memFullErr);
	else
	{
		register DialogItemIndex	i = 0;
		DialogItemIndex				itemIndex = 0;
		Str255						temp;
		ModalFilterUPP				filterProc = NewModalFilterUPP(macroSetupDialogEventFilter);
		
		
		// display the dialog
		TransitionDialogDisplay(gMacroSetupDialog);
		
		MacroSetupDialog_UpdateTextFields();
		restoreFocus(gChosenActiveSet);
		
		// handle events
		Menus_EnableEditMenu();
		itemIndex = 0;
		do
		{
			ModalDialog(filterProc, &itemIndex);
			switch (itemIndex)
			{
				case iMacroSetupMacroTextEquivalentsGroupBox:
					{
						ControlRef			menuControl = null;
						register UInt16		editableSet = 1;
						static UInt16		oldMenuItemSelection = 1;
						static Boolean		firstTimeIn = true;
						
						
						if (firstTimeIn) oldMenuItemSelection = gChosenActiveSet, firstTimeIn = false;
						
						(OSStatus)GetDialogItemAsControl(gMacroSetupDialog, itemIndex, &menuControl);
						editableSet = GetControlValue(menuControl);
						chooseSet(editableSet, true/* save old */);
						rememberFocus(oldMenuItemSelection);
						restoreFocus(editableSet);
						oldMenuItemSelection = editableSet;
					}
					break;
				
				case iMacroSetupExportButton:
					if (!isDisplayedSetEmptyTellUser())
					{
						for (i = 0; i < MACRO_COUNT; ++i)
						{
							GetDialogItemControlText(gMacroSetupDialog, i + iMacroSetupFirstMacroTextField, temp);
							StringUtilities_PToCInPlace(temp);
							Macros_Set(MacroSetupDialog_GetDisplayedSet(), i, (char*)&temp);
						}
						Macros_ExportToText(MacroSetupDialog_GetDisplayedSet(), null/* no file; prompt user */,
											gPendingMacroMode);
					}
					break;
				
				case iMacroSetupImportButton:
					{
						TelnetMacroMode		mode = kTelnetMacroModeCommandDigit;
						
						
						if (Macros_ImportFromText(MacroSetupDialog_GetDisplayedSet(), (FSSpec*)null, &mode))
						{
							// then the macros imported successfully to the displayed set - update the fields
							MacroSetupDialog_SetDisplayedMode(mode);
							MacroSetupDialog_UpdateTextFields();
						}
					}
					break;
				
				case iMacroSetupAppleGuideButton:
					{
						Str255		searchString;
						
						
						GetIndString(searchString, rStringsAppleGuideLookup, siAGIndexTermMacros);
						AppleGuideUtilities_OpenFirstMainGuideInSearch(searchString);
					}
					break;
				
				case iMacroSetupMacroKeysCommandKeysRadioButton:
					if (gPendingMacroMode != kTelnetMacroModeCommandDigit)
					{
						MacroSetupDialog_SetDisplayedMode(kTelnetMacroModeCommandDigit);
					}
					break;
				
				case iMacroSetupMacroKeysFunctionKeysRadioButton:
					if (gPendingMacroMode != kTelnetMacroModeFunctionKeys)
					{
						MacroSetupDialog_SetDisplayedMode(kTelnetMacroModeFunctionKeys);
					}
					break;
				
				case iMacroSetupSet1ActiveRadioButton:
					setTempActive(1);
					break;
				
				case iMacroSetupSet2ActiveRadioButton:
					setTempActive(2);
					break;
				
				case iMacroSetupSet3ActiveRadioButton:
					setTempActive(3);
					break;
				
				case iMacroSetupSet4ActiveRadioButton:
					setTempActive(4);
					break;
				
				case iMacroSetupSet5ActiveRadioButton:
					setTempActive(5);
					break;
				
				case iMacroSetupSetCurrentActiveButton:
					setTempActive(0);
					break;
				
				default:
					break;
			}
		} while ((itemIndex != iMacroSetupOKButton) && (itemIndex != iMacroSetupCancelButton));
		DisposeModalFilterUPP(filterProc), filterProc = null;
		
		// save window size and location in preferences
		Preferences_SaveWindowArrangement(GetDialogWindow(gMacroSetupDialog),
											kTelnetWindowPreferenceTag_MacroSetup);
		
		// handle OK or Cancel click
		if (itemIndex == iMacroSetupCancelButton)
		{
			// user cancelled - close the dialog with an appropriate transition for cancelling
			TransitionDialogCloseCancel(gMacroSetupDialog);
		}
		else
		{
			// user okayed - save macro set changes
			ControlRef			menuControl = null;
			register UInt16		editableSet = 1;
			
			
			// save current text field information first...
			(OSStatus)GetDialogItemAsControl(gMacroSetupDialog, iMacroSetupMacroTextEquivalentsGroupBox,
												&menuControl);
			editableSet = GetControlValue(menuControl);
			chooseSet(editableSet, true/* save old */);
			
			// copy entire temporary space into "real" space
			{
				register UInt16		i = 0;
				
				
				for (i = 0; i < MACRO_SET_COUNT; ++i)
				{
					Macros_SetActiveSetNumber(i + 1); // its one-based
					Macros_Copy(gTempSets[i], Macros_GetActiveSet());
				}
				Macros_SetActiveSetNumber(gChosenActiveSet);
			}
			rememberFocus(editableSet);
			
			#if 0
			Macros_SetMode(gPendingMacroMode);
			#else
			// test a recordable implementation - this shows why Apple Events
			// are such a pain, and yet extremely cool at the same time...
			{
				OSStatus		error = noErr;
				AppleEvent		setDataEvent,
								reply;
				FourCharCode	newKeys = (gPendingMacroMode == kTelnetMacroModeFunctionKeys)
											? kTelnetEnumeratedClassMacroSetKeyEquivalentsFunctionKeys
											: kTelnetEnumeratedClassMacroSetKeyEquivalentsCommandDigit;
				
				
				error = InstallAE_CreateRecordableAppleEvent(kAECoreSuite, kAESetData, &setDataEvent);
				if (error == noErr)
				{
					AEDesc			containerDesc, // never changes, in this case
									keyDesc,
									objectDesc,
									propertyDesc,
									valueDesc;
					
					
					(OSStatus)AppleEventUtilities_InitAEDesc(&containerDesc);
					(OSStatus)AppleEventUtilities_InitAEDesc(&keyDesc);
					(OSStatus)AppleEventUtilities_InitAEDesc(&objectDesc);
					(OSStatus)AppleEventUtilities_InitAEDesc(&propertyDesc);
					(OSStatus)AppleEventUtilities_InitAEDesc(&valueDesc);
					
					// send a request for "macro set <index>", which resides in a null container;
					// then, request the "key equivalents" property, and finally, issue an event
					// that will change its value
					(OSErr)BasicTypesAE_CreateSInt32Desc(Macros_GetActiveSetNumber(), &keyDesc);
					error = CreateObjSpecifier(kTelnetClassMacroSet, &containerDesc,
												formAbsolutePosition, &keyDesc, true/* dispose inputs */,
												&objectDesc);
					if (error != noErr) Console_WriteValue("macro set access error", error);
					(OSErr)BasicTypesAE_CreatePropertyIDDesc(kTelnetPropertyMacroSetKeyEquivalents, &keyDesc);
					error = CreateObjSpecifier(cProperty, &objectDesc,
												formPropertyID, &keyDesc, true/* dispose inputs */,
												&propertyDesc);
					if (error != noErr) Console_WriteValue("property access error", error);
					if (error == noErr)
					{
						// reference to the property to be set - REQUIRED
						(OSStatus)AEPutParamDesc(&setDataEvent, keyDirectObject, &propertyDesc);
						
						// the propertys new value - REQUIRED
						(OSErr)BasicTypesAE_CreateEnumerationDesc(newKeys, &valueDesc);
						(OSErr)AEPutParamDesc(&setDataEvent, kTelnetPartialCoreParameterKeySetDataTo, &valueDesc);
						
						// send the event, which will record it into any open script and actually change the value
						(OSStatus)AESend(&setDataEvent, &reply, kAENoReply | kAENeverInteract,
											kAENormalPriority, kAEDefaultTimeout, null, null);
					}
					AEDisposeDesc(&propertyDesc);
					AEDisposeDesc(&valueDesc);
				}
				AEDisposeDesc(&setDataEvent);
			}
			#endif
			
			// kill the dialog
			TransitionDialogClose(gMacroSetupDialog);
			
			// the user saved changes
			result = true;
		}
	}
	
	return result;
}// Display


/*!
To gain access to the memory being used to
store the temporary macros currently displayed
in the Macro Setup dialog box, use this method.
This set should reflect what is displayed in
the text fields of the dialog box.  You normally
require access to this set only in unusual
circumstances.

(3.0)
*/
MacroSet
MacroSetupDialog_GetDisplayedSet	(void)
{
	return gTempSets[gDisplayedSet - 1];
}// GetDisplayedSet


/*!
Handles a drag-and-drop of text by parsing the
given buffer for macros, and updating the dialog
boxs text fields appropriately if valid data is
given.

The given data is text, and is for convenience
only; the drag reference could be used to find
the text data with a bit of effort (or any other
data for that matter).  The drag reference is
given so that future versions of this function
can use other Drag Manager information such as
modifier keys, etc.

(3.0)
*/
void
MacroSetupDialog_ReceiveDrop		(DragReference		UNUSED_ARGUMENT(inDragRef),
									 unsigned char*		inData,
									 Size				inDataSize)
{
	if (inData != null)
	{
		TelnetMacroMode		mode = kTelnetMacroModeCommandDigit;
		
		
		Macros_ParseTextBuffer(MacroSetupDialog_GetDisplayedSet(), inData, inDataSize, &mode);
		MacroSetupDialog_UpdateTextFields();
		MacroSetupDialog_SetDisplayedMode(mode);
	}
}// ReceiveDrop


/*!
To change the temporary radio buttons for the
macro key equivalents, use this method.  This is
done, for instance, when text is dragged into
the dialog box, because the dialog box must be
updated to reflect potentially changed macro
key equivalents from the imported macros.

(3.0)
*/
void
MacroSetupDialog_SetDisplayedMode		(TelnetMacroMode		inMacroKeyEquivalents)
{
	setButtonsByModeOffscreen(gPendingMacroMode = inMacroKeyEquivalents);
}// SetDisplayedMode


/*!
To update a specific text field (representing
a macros text) to display a particular C
string, call this method.  In order to display
the string, it is translated to a Pascal string.

If the dialog box is not in memory (so its not
on the screen either), this method does nothing.

(3.0)
*/
void
MacroSetupDialog_UpdateTextField	(unsigned char*		inoutMacroText,
									 int				inZeroBasedMacroNumber)
{
	if (gMacroSetupDialog != null)
	{
		StringUtilities_CToPInPlace((char*)inoutMacroText);
		SetDialogItemControlText(gMacroSetupDialog, inZeroBasedMacroNumber + iMacroSetupFirstMacroTextField,
									inoutMacroText);
		{
			ControlRef		control = null;
			
			
			(OSStatus)GetDialogItemAsControl(gMacroSetupDialog,
												inZeroBasedMacroNumber + iMacroSetupFirstMacroTextField,
												&control);
			Draw1Control(control);
		}
	}
}// UpdateTextField


/*!
To make sure that the text fields in the
Macro Setup dialog box reflect exactly the
values of the macros in memory, call this
method.

If the dialog box is not in memory (so
its not on the screen either), this method
does nothing.

(3.0)
*/
void
MacroSetupDialog_UpdateTextFields	(void)
{
	if (gMacroSetupDialog != null)
	{
		unsigned char		newMacro[256];
		DialogItemIndex		i = 0;
		
		
		for (i = 0; i < MACRO_COUNT; ++i)
		{
			Macros_Get(MacroSetupDialog_GetDisplayedSet(), i, (char*)&newMacro, 256);
			MacroSetupDialog_UpdateTextField(newMacro, i);
		}
	}
}// UpdateTextFields


//
// internal methods
//
#pragma mark -

/*!
To select a particular macro set, call this method.

Calling this method causes all of the text fields
in the dialog box to be updated with the appropriate
macro sets macros.  However, the tab control is not
updated because it is assumed that a tab was clicked
to cause this change.

You can choose not to save the information currently
in the text fields by setting "inSaveOld" to false.
This is done only when the dialog first opens.

(3.0)
*/
static void
chooseSet	(UInt16		inMacroSetNumber,
			 Boolean	inSaveOld)
{
	if ((inMacroSetNumber >= 1) && (inMacroSetNumber <= MACRO_SET_COUNT))
	{
		Str255				temp;
		register UInt16		i = 0;
		
		
		// save current text field information first...
		if (inSaveOld) for (i = 0; i < MACRO_COUNT; ++i)
		{
			GetDialogItemControlText(gMacroSetupDialog, i + iMacroSetupFirstMacroTextField, temp);
			StringUtilities_PToCInPlace(temp);
			Macros_Set(MacroSetupDialog_GetDisplayedSet(), i, (char*)&temp);
		}
		
		// change temporary index
		gDisplayedSet = inMacroSetNumber;
		
		// ...and obliterate the text fields, replacing whatever is there with macros from the new set
		MacroSetupDialog_UpdateTextFields();
	}
	else BEEP;
}// chooseSet


/*!
To move the keyboard focus ring around a specific
text field in the dialog, call this method.

(3.0)
*/
static void
focusField	(DialogItemIndex	inWhichField)
{
	ControlRef		fieldControl = null;
	
	
	(OSStatus)GetDialogItemAsControl(gMacroSetupDialog, inWhichField, &fieldControl);
	(OSStatus)SetKeyboardFocus(GetDialogWindow(gMacroSetupDialog), fieldControl, kControlFocusNextPart);
}// focusField


/*!
This method moves and resizes controls in response to
a resize of the Macro Setup movable modal dialog box.

(3.0)
*/
static pascal Boolean
handleNewSize	(WindowRef	inWindow,
				 SInt32		inDeltaSizeX,
				 SInt32		inDeltaSizeY,
				 SInt32		UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_true;
	
	
	DialogAdjust_BeginDialogItemAdjustment(GetDialogFromWindow(inWindow));
	{
		SInt32				amountH = 0L,
							amountV = 0L;
		register SInt16		i = 0;
		
		
		// controls which are moved in both directions
		if (Localization_IsLeftToRight())
		{
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iMacroSetupOKButton, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iMacroSetupOKButton, inDeltaSizeY);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iMacroSetupCancelButton, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iMacroSetupCancelButton, inDeltaSizeY);
		}
		else
		{
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iMacroSetupAppleGuideButton, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iMacroSetupAppleGuideButton, inDeltaSizeY);
		}
		
		if (inDeltaSizeX)
		{
			// controls which are moved horizontally
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iMacroSetupImportButton, inDeltaSizeX / 6);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iMacroSetupSetCurrentActiveButton, INTEGER_HALVED(inDeltaSizeX));
			amountH = INTEGER_HALVED(inDeltaSizeX);
			for (i = (iMacroSetupFirstMacroTextField + INTEGER_HALVED(MACRO_COUNT));
				i < (iMacroSetupFirstMacroTextField + MACRO_COUNT); ++i)
			{
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, i, amountH);
			}
			for (i = (iMacroSetupFirstMacroButton + INTEGER_HALVED(MACRO_COUNT));
				i < (iMacroSetupFirstMacroButton + MACRO_COUNT); ++i)
			{
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, i, amountH);
			}
			
			// controls which are resized horizontally
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iMacroSetupMacroTextEquivalentsGroupBox, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iMacroSetupMacroTextEquivalentsDivider, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iMacroSetupActiveSetGroupBox, inDeltaSizeX);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iMacroSetupSetCurrentActiveButton, INTEGER_HALVED(inDeltaSizeX));
			amountH = inDeltaSizeX / 6;
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iMacroSetupImportButton, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iMacroSetupExportButton, amountH);
			
			amountH = INTEGER_HALVED(inDeltaSizeX);
			for (i = iMacroSetupFirstMacroTextField; i < (iMacroSetupFirstMacroTextField + MACRO_COUNT); ++i)
			{
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, i, amountH);
			}
		}
	}
	DialogAdjust_EndAdjustment(inDeltaSizeX, inDeltaSizeY); // moves and resizes controls properly
	
	return result;
}// handleNewSize


/*!
To display an alert if all of the macro fields
are empty, use this method.  If any text is in
any of the fields, nothing happens.

If all 10 fields are empty, "true" is returned;
otherwise, "false" is returned.

(3.0)
*/
static Boolean
isDisplayedSetEmptyTellUser	(void)
{
	Boolean				result = true;
	register SInt16		i = 0;
	Str255				temp;
	
	
	// only stop the user if there are no macros in the text fields
	for (i = 0; i < MACRO_COUNT; ++i)
	{
		GetDialogItemControlText(gMacroSetupDialog, i + iMacroSetupFirstMacroTextField, temp);
		if (StrLength(temp) > 0) result = false;
	}
	if (result)
	{
		Errors_DisplayStopMessage(rStringsNoteAlerts, siExportNotPossible, siAlertTitleExportNotPossible);
	}
	
	return result;
}// isDisplayedSetEmptyTellUser


/*!
Filters out key presses originally sent to a macro text field.
Certain characters are not allowed in macros without escaping,
so this routine checks for such errors.

(3.0)
*/
static pascal ControlKeyFilterResult
macroFieldKeyFilter		(ControlRef		inControl,
						 SInt16*		UNUSED_ARGUMENT(inoutKeyCode),
						 SInt16*		inoutCharCode,
						 UInt16*		UNUSED_ARGUMENT(inoutModifiers))
{
	ControlKeyFilterResult		result = kControlKeyFilterPassKey;
	Str255						macroText;
	
	
	// determine the text currently in the command line
	GetControlText(inControl, macroText);
	
	if (inoutCharCode != null) switch (*inoutCharCode)
	{
		case '"':
			// the quotation mark must be escaped
			if (macroText[PLstrlen(macroText)] != '\\')
			{
				BEEP;
				result = kControlKeyFilterBlockKey;
			}
			break;
		
		default:
			// uninteresting key was hit
			break;
	}
	
	return result;
}// macroFieldKeyFilter


/*!
Handles special key events in the Macro Setup
dialog box.  I clicks the Import button, and
E clicks the Export button.  F chooses the
F1-F10 keys for macros, and any 0 to 9
combination chooses 0-9 for macros.

(3.0)
*/
static pascal Boolean
macroSetupDialogEventFilter		(DialogRef		inDialog,
								 EventRecord*	inoutEventPtr,
								 short*			outItemIndex)
{
	Boolean		result = pascal_false,
				isButtonAndShouldFlash = false; // simplifies handling of multiple key-to-button maps
	
	
	// first, make sure help balloons show up properly for the cursor location, if Balloon Help is on
	if (IsPopUpHelpEnabled()) DisplayBalloonForDialogItemUnderMouse(inDialog, inoutEventPtr->where,
																		kDialogBalloonsIDMacroSetup);
	
	// now handle events
	if (inoutEventPtr->what == keyDown)
	{
		SInt16		key = 0;
		
		
		key = inoutEventPtr->message & charCodeMask;
		if (((key == 'I') || (key == 'i')) && (inoutEventPtr->modifiers & cmdKey))
		{
			ControlRef		button = null;
			
			
			(OSStatus)GetDialogItemAsControl(inDialog, iMacroSetupImportButton, &button);
			if (IsControlActive(button))
			{
				*outItemIndex = iMacroSetupImportButton;
				isButtonAndShouldFlash = true;
				result = pascal_true;
			}
		}
		else if (((key == 'E') || (key == 'e')) && (inoutEventPtr->modifiers & cmdKey))
		{
			ControlRef		button = null;
			
			
			(OSStatus)GetDialogItemAsControl(inDialog, iMacroSetupExportButton, &button);
			if (IsControlActive(button))
			{
				*outItemIndex = iMacroSetupExportButton;
				isButtonAndShouldFlash = true;
				result = pascal_true;
			}
		}
		else if (((key == 'F') || (key == 'f')) && (inoutEventPtr->modifiers & cmdKey))
		{
			ControlRef		button = null;
			
			
			(OSStatus)GetDialogItemAsControl(inDialog, iMacroSetupMacroKeysFunctionKeysRadioButton, &button);
			if (IsControlActive(button))
			{
				*outItemIndex = iMacroSetupMacroKeysFunctionKeysRadioButton;
				result = pascal_true;
			}
		}
		else if (((key >= '0') && (key <= '9')) && (inoutEventPtr->modifiers & cmdKey))
		{
			ControlRef		button = null;
			
			
			(OSStatus)GetDialogItemAsControl(inDialog, iMacroSetupMacroKeysCommandKeysRadioButton, &button);
			if (IsControlActive(button))
			{
				*outItemIndex = iMacroSetupMacroKeysCommandKeysRadioButton;
				result = pascal_true;
			}
		}
		else if (((key == 'U') || (key == 'u')) && (inoutEventPtr->modifiers & cmdKey))
		{
			ControlRef		button = null;
			
			
			(OSStatus)GetDialogItemAsControl(inDialog, iMacroSetupSetCurrentActiveButton, &button);
			if (IsControlActive(button))
			{
				*outItemIndex = iMacroSetupSetCurrentActiveButton;
				isButtonAndShouldFlash = true;
				result = pascal_true;
			}
		}
	}
	
	if (!result) result = (StandardDialogEventFilter(inDialog, inoutEventPtr, outItemIndex));
	else if (isButtonAndShouldFlash) FlashButton(inDialog, *outItemIndex);
	
	return result;
}// macroSetupDialogEventFilter


/*!
To save the currently-focused text field control
as being associated with the specified tab, use
this method.  Use restoreFocus() to restore the
focus for the tab later.

(3.0)
*/
static void
rememberFocus	(UInt16		inWhichSet)
{
	if ((inWhichSet >= 1) && (inWhichSet <= MACRO_SET_COUNT))
	{
		ControlRef			control = null,
							thisField = null;
		register UInt16		i = 0;
		
		
		(OSStatus)GetKeyboardFocus(GetDialogWindow(gMacroSetupDialog), &control);
		for (i = 0, gTabFocusedFields[inWhichSet - 1] = iMacroSetupFirstMacroTextField; i < MACRO_COUNT; ++i)
		{
			(OSStatus)GetDialogItemAsControl(gMacroSetupDialog, iMacroSetupFirstMacroTextField + i, &thisField);
			if (thisField == control)
			{
				gTabFocusedFields[inWhichSet - 1] = iMacroSetupFirstMacroTextField + i;
				break;
			}
		}
	}
}// rememberFocus


/*!
To restore the focused text field control
to the one last saved for the specified set,
use this method.  If rememberFocus() has not
yet been called, the first field is simply
focused.

(3.0)
*/
static void
restoreFocus	(UInt16		inWhichSet)
{
	focusField(gTabFocusedFields[inWhichSet - 1]);
}// restoreFocus


/*!
To make the most recent value set with setTempActive()
the "official" value, call this method.

(3.0)
*/
static void
saveActiveSetNumber	(void)
{
	Macros_SetActiveSetNumber(gChosenActiveSet);
}// saveActiveSetNumber


/*!
To label the ten macro buttons based on the current
macro mode (function key or command-digit, for
example), call this method.  This method is in
standard OffscreenOperationProcPtr form.

(3.0)
*/
static pascal Boolean
setButtonsByMode	(ControlRef		UNUSED_ARGUMENT(inForWhichControl),
					 SInt16			UNUSED_ARGUMENT(inColorDepth),
					 SInt16			UNUSED_ARGUMENT(inDeviceFlags),
					 GDHandle		UNUSED_ARGUMENT(inTargetDevice),
					 SInt32			inMode,
					 SInt32			UNUSED_ARGUMENT(inUnused))
{
	Str255				titleText;
	ControlRef			control = null;
	register UInt16		i = 0,
						j = 0;
	Boolean				result = pascal_true;
	
	
	// unmark all radio buttons first...
	SetDialogItemValue(gMacroSetupDialog, iMacroSetupMacroKeysCommandKeysRadioButton,
						kControlRadioButtonUncheckedValue);
	SetDialogItemValue(gMacroSetupDialog, iMacroSetupMacroKeysFunctionKeysRadioButton,
						kControlRadioButtonUncheckedValue);
	
	// now mark the radio button for the desired mode
	if (inMode == kTelnetMacroModeCommandDigit)
	{
		SetDialogItemValue(gMacroSetupDialog, iMacroSetupMacroKeysCommandKeysRadioButton,
							kControlRadioButtonCheckedValue);
	}
	if (inMode == kTelnetMacroModeFunctionKeys)
	{
		SetDialogItemValue(gMacroSetupDialog, iMacroSetupMacroKeysFunctionKeysRadioButton,
							kControlRadioButtonCheckedValue);
	}
	
	// change the field labels to match the mode
	for (i = iMacroSetupFirstMacroButton, j = 0; i <= iMacroSetupLastMacroButton; ++i, ++j)
	{
		PLstrcpy(titleText, EMPTY_PSTRING);
		if (inMode == kTelnetMacroModeCommandDigit)
		{
			PLstrcpy(titleText, "\p");
			if (i < (iMacroSetupFirstMacroButton + 10)) StringUtilities_PAppendNumber(titleText, (long)j);
			else if (i == (iMacroSetupFirstMacroButton + 10)) PLstrcat(titleText, "\p="); // special macro
			else if (i == (iMacroSetupFirstMacroButton + 11)) PLstrcat(titleText, "\p/"); // special macro
		}
		else if (inMode == kTelnetMacroModeFunctionKeys)
		{
			PLstrcpy(titleText, "\pF");
			StringUtilities_PAppendNumber(titleText, (long)(j + 1));
		}
		(OSErr)GetDialogItemAsControl(gMacroSetupDialog, i, &control);
		(UInt16)Localization_SetUpSingleLineTextControl(control, titleText, kThemeSystemFont);
		
		DrawOneControl(control);
	}
	return result;
}// setButtonsByMode


/*!
To label the ten macro buttons based on the current
macro mode (function key or command-digit, for
example) using an offscreen graphics buffer, call
this method.

(3.0)
*/
static void
setButtonsByModeOffscreen	(TelnetMacroMode	inMode)
{
	OffscreenOperationProcPtr		proc = setButtonsByMode;
	ControlRef						groupBox = null;
	
	
	(OSStatus)GetDialogItemAsControl(gMacroSetupDialog, iMacroSetupMacroTextEquivalentsGroupBox, &groupBox);
	(Boolean)Embedding_OffscreenControlOperation(GetDialogWindow(gMacroSetupDialog), groupBox,
													proc, (SInt32)inMode, 0L);
	
	// hack - apparently the offscreen thing causes the following controls to reflect incorrect states
	DrawOneDialogItem(gMacroSetupDialog, iMacroSetupMacroKeysCommandKeysRadioButton);
	DrawOneDialogItem(gMacroSetupDialog, iMacroSetupMacroKeysFunctionKeysRadioButton);
}// setButtonsByModeOffscreen


/*!
To mark a particular set of macros as the active
set temporarily, call this method.  The temporary
setting should only be made real if the user okays
the dialog box.

To set the active set based on the current value
of the menu control, pass 0.

The radio buttons in the "Active Set" group box
are automatically updated to match.

(3.0)
*/
static void
setTempActive	(UInt16		inMacroSetNumber)
{
	if (inMacroSetNumber != 0)
		gChosenActiveSet = inMacroSetNumber;
	else
	{
		ControlRef		menuControl = null;
		
		
		// the value "0" means use the menu value
		(OSStatus)GetDialogItemAsControl(gMacroSetupDialog, iMacroSetupMacroTextEquivalentsGroupBox,
											&menuControl);
		gChosenActiveSet = GetDialogItemValue(gMacroSetupDialog, iMacroSetupMacroTextEquivalentsGroupBox);
	}
	SetDialogItemValue(gMacroSetupDialog, iMacroSetupSet5ActiveRadioButton, kControlRadioButtonUncheckedValue);
	SetDialogItemValue(gMacroSetupDialog, iMacroSetupSet4ActiveRadioButton, kControlRadioButtonUncheckedValue);
	SetDialogItemValue(gMacroSetupDialog, iMacroSetupSet3ActiveRadioButton, kControlRadioButtonUncheckedValue);
	SetDialogItemValue(gMacroSetupDialog, iMacroSetupSet2ActiveRadioButton, kControlRadioButtonUncheckedValue);
	SetDialogItemValue(gMacroSetupDialog, iMacroSetupSet1ActiveRadioButton, kControlRadioButtonUncheckedValue);
	{
		register short		radioIndex = iMacroSetupSet1ActiveRadioButton;
		
		
		switch (gChosenActiveSet)
		{
			case 5:
				radioIndex = iMacroSetupSet5ActiveRadioButton;
				break;
			
			case 4:
				radioIndex = iMacroSetupSet4ActiveRadioButton;
				break;
			
			case 3:
				radioIndex = iMacroSetupSet3ActiveRadioButton;
				break;
			
			case 2:
				radioIndex = iMacroSetupSet2ActiveRadioButton;
				break;
			
			case 1:
			default:
				radioIndex = iMacroSetupSet1ActiveRadioButton;
				break;
		}
		SetDialogItemValue(gMacroSetupDialog, radioIndex, kControlRadioButtonCheckedValue);
	}
}// setTempActive