/*###############################################################

	Macros.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdio>
#	include <cstring>
#else
#	include <stdio.h>
#	include <string.h>
#endif
#include <ctype.h>

// Mac includes
#include <MacErrors.h>
#include <Navigation.h>
#include <Script.h>
#include <StandardFile.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "GeneralResources.h"
#include "StringResources.h"
#include "TelnetVersion.h"

// MacTelnet includes
#include "Connections.h"
#include "DialogUtilities.h"
#include "FolderManager.h"
#include "FTPBin.h"
#include "Parse.h"
#include "ScreenManager.h"
#include "Sets.proto.h" // for CStringToFile
#include "EventLoop.h"
#include "TelnetDescriptors.h"
#include "TelnetMacros.h"
#include "Terminal.h"



//
// constants
//

#define MACRO_IP		0xFF	// metacharacter (for "\i" in source string); substituted with IP address
#define MACRO_LINES		0xFE	// metacharacter (for "\#" in source string); substituted with screen row count

enum
{
	MACRO_MAX_LEN = 256		// maximum number of characters allowed in any macro
};

//
// variables
//

static Handle			gMacros[MACRO_SET_COUNT][MACRO_COUNT] =
						{
							// IMPORTANT - initialize all macros to empty!
							{ null, null, null, null, null, null, null, null, null, null },
							{ null, null, null, null, null, null, null, null, null, null },
							{ null, null, null, null, null, null, null, null, null, null },
							{ null, null, null, null, null, null, null, null, null, null },
							{ null, null, null, null, null, null, null, null, null, null }
						};
static UInt16			gActiveSet = 1;
static Boolean			gInited = false;
static TelnetMacroMode	gMacroMode = kTelnetMacroModeCommandDigit;

//
// internal methods
//

static Boolean	allEmpty					(MacroSet			inSet);
static SInt16	getMacro					(MacroSet			inSet,
											 UInt16				inMacroNumber,
											 char*				outValue,
											 SInt16				inRoom);
static void		parseFile					(MacroSet			inSet,
											 SInt16				fileRef,
											 TelnetMacroMode*	outMacroModeOfImportedSet);
static void		setTempMacrosToNull			(MacroSet			inSet);
static void		setMacro					(MacroSet			inoutDestination,
											 UInt16				inMacroNumber,
											 char*				inoutValue);



//
// public methods
//

/*!
Loads the special macro files in the Preferences
folder, and initializes the five macro sets from
those files.  If any or all of the files are
missing, the respective macro sets are simply
left blank.

(3.0)
*/
void
Macros_Init		(void)
{
	FSSpec				parseFolder,
						parseFile;
	Str255				fileName;
	register SInt16		i = 0;
	TelnetMacroMode		mode = kTelnetMacroModeCommandDigit;
	
	
	// there are five files, one per set; read each file, importing only if no errors occur
	for (i = MACRO_SET_COUNT - 1; i >= 0; i--)
	{
		Macros_SetActiveSetNumber(i + 1); // set numbers are one-based
		GetIndString(fileName, rStringsImportantFileNames, siImportantFileNameMacroSet1 + i);
		if (StrLength(fileName) > 0)
		{
			if (FolderManager_GetFSSpec(kTelnetFolderPreferences, &parseFolder) == noErr)
			{
				if (FSMakeFSSpec(parseFolder.vRefNum, parseFolder.parID, fileName, &parseFile) == noErr)
				{
					// if this point is reached, then the requested file exists; import it!
					Macros_ImportFromText(Macros_GetActiveSet(), &parseFile, &mode);
				}
			}
		}
	}
	Macros_SetMode(mode); // the effect of this is that Macro Set 1 determines the persistent macro keys
	gInited = true;
}// Init


/*!
Maintains implicit persistence of the five macro
sets by saving them into special macro files in
the Preferences folder.

Note that errors are basically ignored; a nicer
implementation might advise the user when some
kind of problem occurs that could jeopardize
the saving of macros.  In fact, this should be a
definite priority for the next release.

(3.0)
*/
void
Macros_Done		(void)
{
	if (gInited)
	{
		FSSpec				parseFolder,
							parseFile;
		Str255				fileName;
		register SInt16		i = 0;
		OSErr				error = noErr;
		
		
		// there are five files, one per set; find each file, exporting only if no errors occur
		for (i = MACRO_SET_COUNT - 1; i >= 0; i--)
		{
			Macros_SetActiveSetNumber(i + 1); // set numbers are one-based
			GetIndString(fileName, rStringsImportantFileNames, siImportantFileNameMacroSet1 + i);
			if (StrLength(fileName) > 0)
			{
				if (FolderManager_GetFSSpec(kTelnetFolderPreferences, &parseFolder) == noErr)
				{
					error = FSMakeFSSpec(parseFolder.vRefNum, parseFolder.parID, fileName, &parseFile);
					
					if ((error == noErr) || (error == fnfErr))
					{
						// If this point is reached, then it is possible to specify where on
						// disk the macros go; export them!  Note that due to the nature of
						// the import procedure, the persistent macro mode will really be
						// Macro Set 1s mode.
						Macros_ExportToText(Macros_GetActiveSet(), &parseFile, Macros_GetMode());
						
						// since Telnet depends on the file having a specific name, make sure the
						// user cant change it
						{
							FileInfo		info;
							
							
							FSpGetFInfo(&parseFile, (FInfo*)&info);
							info.finderFlags |= kNameLocked;
							FSpSetFInfo(&parseFile, (FInfo*)&info);
						}
					}
				}
			}
		}
		gInited = false;
	}
}// Done


/*!
To create a temporary storage area for macros,
call this method.  Use Macros_DisposeSet() to
destroy this set when you are finished with it.
If any problems occur, null is returned.

(3.0)
*/
MacroSet
Macros_NewSet	(void)
{
	return (MacroSet)Memory_NewPtr(MACRO_COUNT * sizeof(Handle));
}// NewSet


/*!
To destroy a temporary storage area that you
previously created with Macros_NewSet(), call
this method.  On output, your pointer to the
macro set is automatically set to null.

(3.0)
*/
void
Macros_DisposeSet	(MacroSet*		inoutSetToDestroy)
{
	if (inoutSetToDestroy != null)
	{
		register SInt16		i = 0;
		
		
		for (i = 0; i < MACRO_COUNT; i++)
		{
			if ((*inoutSetToDestroy)[i] != null)
			{
				HUnlock((*inoutSetToDestroy)[i]);
				Memory_DisposeHandle(&(*inoutSetToDestroy)[i]);
			}
		}
		Memory_DisposePtr((Ptr*)inoutSetToDestroy);
	}
}// DisposeSet


/*!
To determine if there is any text in any of the
macros in the specified macro set, call this
method.

(3.0)
*/
Boolean
Macros_AllEmpty		(MacroSet		inSet)
{
	return allEmpty(inSet);
}// AllEmpty


/*!
To duplicate all of the macros in one set in
a second set, use this method.

(3.0)
*/
void
Macros_Copy		(MacroSet		inSource,
				 MacroSet		inDestination)
{
	register SInt16		i = 0,
						j = Macros_GetActiveSetNumber();
	
	
	for (i = 0; i < MACRO_COUNT; i++)
	{
		if (inDestination[i] != null) Memory_DisposeHandle(&inDestination[i]);
		if (inSource[i] != null)
		{
			SInt8		handleState = 0;
			Size		moveSize = GetHandleSize(inSource[i]);
			
			
			inDestination[i] = Memory_NewHandle(moveSize);
			handleState = HGetState(inSource[i]);
			HLockHi(inSource[i]);
			HLockHi(inDestination[i]);
			BlockMoveData(*(inSource[i]), *(inDestination[i]), moveSize);
			HUnlock(inDestination[i]);
			HSetState(inSource[i], handleState);
		}
	}
}// Copy


/*!
Exports macros stored in the specified macro storage area
to a file.  If you pass "null" for the file specification,
the user is prompted to create a file in which to store
the macro set.  The exported file also indicates the keys
to be used for the macro keys (such as function keys,
versus command-digit keys).

To place macros in temporary storage space, you can use
the Macros_Set() or Macros_Copy() methods.

(2.6)
*/
void
Macros_ExportToText		(MacroSet			inSet,
						 FSSpec*			inFileSpecPtrOrNull,
						 TelnetMacroMode	inMacroModeOfExportedSet)
{
	SInt16 		refNum = 0;
	FSSpec		macroFile; // this is ONLY storage, used as needed
	FSSpec*		fileSpecPtr = inFileSpecPtrOrNull; // this refers to the file to write to!
	OSErr		error = noErr;
	Boolean		good = true;
	Str255		prompt,
				title;
	
	
	if (inFileSpecPtrOrNull == null)
	{
		GetIndString(prompt, rStringsNavigationServices, siNavPromptExportMacrosToFile);
		GetIndString(title, rStringsNavigationServices, siNavDialogTitleExportMacros);
		if (FlagManager_Test(kFlagNavigationServices))
		{
			SInt16			hit = 0;
			NavReplyRecord	reply;
			
			
			Alert_ReportOSErr(error = Navigation_PutFile(prompt, title,
									kApplicationCreatorSignature, kApplicationFileTypeMacroSet,
									kTelnetNavPrefKeyMacroStuff,
								kNavDefaultNavDlogOptions | kNavDontAddTranslateItems,
								EventLoop_HandleNavigationUpdate, &reply, &macroFile));
			good = ((error == noErr) && (reply.validRecord));
			if (good) fileSpecPtr = &macroFile;
		}
		else
		{
			StandardFileReply	sfr;
			Str255				tempString;
			long				junk = 0L;
			
			
			GetIndString(tempString, rStringsImportantFileNames, siImportantFileNameMacroSetDefault);
			
			DeactivateFrontmostWindow();
			ThemeSavvyStandardPutFile(title, prompt, tempString, &sfr);
			RestoreFrontmostWindow();
			good = sfr.sfGood;
			if (good)
			{
				OSErr		error = noErr;
				
				
				error = FSMakeFSSpec(sfr.sfFile.vRefNum, sfr.sfFile.parID, sfr.sfFile.name, &macroFile);
				good = ((error == noErr) || (error == fnfErr));
				if (good) fileSpecPtr = &macroFile;
			}
		}
	}
	
	if (good)
	{
		SInt16 		i = 0;
		char		temp[256],
					temp2[256];
		Boolean		exists = false;
		
		
		// try to create the file; if this fails with a duplicate file name error, then the file already exists
		if ((error = FSpCreate(fileSpecPtr, kApplicationCreatorSignature,
								kApplicationFileTypeMacroSet, GetScriptManagerVariable(smSysScript))) == dupFNErr)
		{
			exists = true;
		}
		
		error = FSpOpenDF(fileSpecPtr, fsWrPerm, &refNum);
		
		if (exists) SetEOF(refNum, 0L);
		
		for (i = 0; i < MACRO_COUNT; i++)
		{
			Macros_Get(inSet, i, temp, sizeof(temp));
			
			// IMPORTANT:	The internal method parseFile() determines the key equivalents of an imported
			//				set by checking to see if the first character of each line is a capital "F".  If
			//				you see a need to change the code below, be sure to update parseFile()!
			sprintf(temp2, (inMacroModeOfExportedSet == kTelnetMacroModeFunctionKeys)
							? "F%d = \""
							: "Cmd%d = \"",
							(inMacroModeOfExportedSet == kTelnetMacroModeFunctionKeys)
							? i + 1
							: i);
			CStringToFile(refNum, temp2);
			if (*temp) 
			{									
				CStringToFile(refNum, temp);
			}
			strcpy(temp2, "\"\015");
			CStringToFile(refNum, temp2);	
	
		}
		FSClose(refNum);
	}
}// ExportToText


/*!
To explicitly obtain the value of a particular
macro in a macro space, call this method.  You can
use Macros_Copy() to work between two macro spaces.

The value of "inZeroBasedMacroNumber" must be
between 0 and one less than MACRO_COUNT.

(3.0)
*/
void
Macros_Get		(MacroSet		inFromWhichSet,
				 MacroIndex		inZeroBasedMacroNumber,
				 char*			outValue,
				 SInt16			inRoom)
{
	getMacro(inFromWhichSet, inZeroBasedMacroNumber, outValue, inRoom);
}// Get


/*!
To obtain the active macro set, use this method.

(3.0)
*/
MacroSet
Macros_GetActiveSet	(void)
{
	return gMacros[Macros_GetActiveSetNumber() - 1];
}// GetActiveSet


/*!
To determine which set (from 1 to MACRO_SET_COUNT) is
active, call this method.  The active set is the one
which is referenced by the Macros_Get() and Macros_Set()
methods.

(3.0)
*/
UInt16
Macros_GetActiveSetNumber	(void)
{
	return gActiveSet;
}// GetActiveSetNumber


/*!
To determine the current macro mode, call this
method.  The mode does not affect how macros are
stored or manipulated, it only affects what is
required by the user in order to "type" one.  The
default macro mode is "command-digit", which
means that one of the key combinations from -0
through -9 is required to type a macro.  Other
modes, such as "function key" (F1-F10), are also
available.

(3.0)
*/
TelnetMacroMode
Macros_GetMode	(void)
{
	return gMacroMode;
}// GetMode


/*!
Imports macros.  If you pass "null" for the file
specification, the user is prompted to locate a file
from which to fill in the macro set.  If the user
does not cancel and no errors occur (that is, the
macros actually got imported), true is returned;
otherwise, false is returned.

Sometimes, the key equivalents used for the macros
can be determined from the text file.  If so, the
macro mode as saved in the macro set is imported
as well, and returned in "outMacroModeOfImportedSet".
If not, the value "kTelnetMacroModeCommandDigit" (the
default) is returned in "outMacroModeOfImportedSet".
If you do not need this information, it is safe to
pass "null" for "outMacroModeOfImportedSet".

(3.0)
*/
Boolean
Macros_ImportFromText	(MacroSet			inSet,
						 const FSSpec*		inFileSpecPtrOrNull,
						 TelnetMacroMode*	outMacroModeOfImportedSet)
{
	OSErr 			err = noErr;
	SInt16 			fileRef = 0;
	Boolean			result = true;
	FSSpec			chosenFile; // this is ONLY storage, used as needed
	const FSSpec*	fileSpecPtr = inFileSpecPtrOrNull; // this refers to the file to read from!
	
	
	if (inFileSpecPtrOrNull == null)
	{
		Str255		prompt,
					title;
		
		
		GetIndString(prompt, rStringsNavigationServices, siNavPromptImportMacrosFromFile);
		GetIndString(title, rStringsNavigationServices, siNavDialogTitleImportMacros);
		if (FlagManager_Test(kFlagNavigationServices))
		{
			FileInfo	fileInfo;
			OSType		fileType;
			OSType		types[] = { kApplicationFileTypeMacroSet };
			OSErr		error = noErr;
			SInt16		hit = 0;
			
			
			Alert_ReportOSErr(error = Navigation_GetFile(prompt, title,
									kApplicationCreatorSignature, kTelnetNavPrefKeyMacroStuff,
								kNavDefaultNavDlogOptions | kNavDontAddTranslateItems,
								1, types, EventLoop_HandleNavigationUpdate, &chosenFile,
									&fileType, &fileInfo));
			result = (error == noErr);
		}
		else
		{
			StandardFileReply	sfr;
			SFTypeList			types = { kApplicationFileTypeMacroSet };
			long				junk = 0L;
			
			
			DeactivateFrontmostWindow();
			ThemeSavvyStandardGetFile(title, prompt, (FileFilterUPP)null, 1, types, &sfr);
			RestoreFrontmostWindow();
			result = sfr.sfGood;
			if (result)
			{
				result = (FSMakeFSSpec(sfr.sfFile.vRefNum, sfr.sfFile.parID, sfr.sfFile.name,
										&chosenFile) == noErr);
			}
		}
		if (result) fileSpecPtr = &chosenFile;
	}
	
	if (fileSpecPtr != null) err = FSpOpenDF(fileSpecPtr, fsRdPerm, &fileRef);
	
	if ((err == noErr) && (result))
	{
		parseFile(inSet, fileRef, outMacroModeOfImportedSet);
		FSClose(fileRef);
	}
	
	return result;
}// ImportFromText


/*!
To explicitly change all of the macros in the
specified temporary space to correspond to the text
contained in a buffer of the specified size, call
this method.

IMPORTANT:	The internal parseFile() method does
			the same fundamental thing as this
			routine, but the complexities of
			file access made it easier to simply
			duplicate most of the code.  If you
			ever change anything here, be sure
			to compare your changes with the
			code of parseFile().

(3.0)
*/
void
Macros_ParseTextBuffer	(MacroSet			inSet,
						 unsigned char*		inSourcePtr,
						 Size				inSize,
						 TelnetMacroMode*	outMacroModeOfImportedSet)
{
	unsigned char	buffer[300];
	unsigned char	newMacro[256];
	unsigned char*	bufferPtr = null;
	unsigned char*	newMacroPtr = null;
	unsigned char*	dynamicSourcePtr = inSourcePtr;
	OSErr			error = noErr;
	UInt16			numMacrosRead = 0;
	register SInt16	totalLen = 0,
					i = 0;
	Size			count = 1; // number of bytes copied in total
	
	
	bufferPtr = buffer; 

	setTempMacrosToNull(inSet); // sets all handles in the current temporary macro space to null

	while ((error != eofErr) && (numMacrosRead < MACRO_COUNT))
	{
		*bufferPtr = *dynamicSourcePtr++;
		if (count >= inSize) error = eofErr;
		while ((*bufferPtr != 0x0D) && (error != eofErr)) // while not CR or EOF
		{
			++bufferPtr;
			++count;
			if (count <= inSize) *bufferPtr = *dynamicSourcePtr++;
			else error = eofErr;
		}
		
		// 3.0 - determine the key equivalents, if possible (this only needs to be done once)
		if ((!numMacrosRead) && (outMacroModeOfImportedSet != null))
		{
			// See Macros_ExportToText() for code that explicitly specifies that
			// the key for a function key macro starts with a capital "F".  That
			// code is the basis for the following test, which simply looks at
			// the first character of each macro to see what the key equivalents
			// in the set are.
			if (buffer[0] == 'F') *outMacroModeOfImportedSet = kTelnetMacroModeFunctionKeys;
			else *outMacroModeOfImportedSet = kTelnetMacroModeCommandDigit;
		}
		
		totalLen = bufferPtr - buffer;
		bufferPtr = buffer;
		newMacroPtr = newMacro;
		while ((*bufferPtr++ != '"') && (totalLen != 0)) --totalLen;
		
		while ((*bufferPtr != '"') && (totalLen != 0))
		{
			*newMacroPtr++ = *bufferPtr++;
			--totalLen;
		}
		*newMacroPtr = '\0'; // make this a C string
		
		Macros_Set(inSet, numMacrosRead, (char*)newMacro);
		numMacrosRead++;
		bufferPtr = buffer;
	}
}// ParseTextBuffer


/*!
To explicitly change the value of a particular
macro, call this method.  You can use the
Macros_Copy() method to work between two macro
spaces.

The value of "inZeroBasedMacroNumber" must be
between 0 and one less than MACRO_COUNT.

(3.0)
*/
void
Macros_Set		(MacroSet		inSet,
				 MacroIndex		inZeroBasedMacroNumber,
				 char*			inoutValue)
{
	setMacro(inSet, inZeroBasedMacroNumber, inoutValue);
}// Set


/*!
To specify which set (from 1 to MACRO_SET_COUNT) is
active, call this method.  The active set is the one
which is referenced by the Macros_Get() and Macros_Set()
methods.

(3.0)
*/
void
Macros_SetActiveSetNumber	(UInt16		inMacroSetNumber)
{
	gActiveSet = inMacroSetNumber;
}// SetActiveSetNumber


/*!
To specify the current macro mode, call this
method.  The mode does not affect how macros are
stored or manipulated, it only affects what is
required by the user in order to "type" one.  The
default macro mode is "command-digit", which
means that one of the key combinations from -0
through -9 is required to type a macro.  Other
modes, such as "function key" (F1-F10), are also
available.

(3.0)
*/
void
Macros_SetMode	(TelnetMacroMode	inNewMode)
{
	gMacroMode = inNewMode;
}// SetMode


/*!
Sends a specific macro (from the currently active
macro set) as typed text to the specified screen.

If the given index is valid, 0 is returned; otherwise,
-1 is returned.

(2.6)
*/
SInt16
Macros_Send		(ConnectionDataPtr	inScreen,
				 MacroIndex			inZeroBasedMacroNumber)
{
	char				temp[20];
	unsigned char*		mp = null;
	unsigned char*		first = null;
	unsigned char		myipnum[4];
	register UInt16		zeroBasedActiveSetNumber = Macros_GetActiveSetNumber() - 1;
	SInt16				result = 0;
	
	
	if ((inZeroBasedMacroNumber < 0) || (inZeroBasedMacroNumber > (MACRO_COUNT - 1)))
	{
		// invalid number
		result = -1;
	}
	else if (gMacros[zeroBasedActiveSetNumber][inZeroBasedMacroNumber] != null)
	{
		HLock(gMacros[zeroBasedActiveSetNumber][inZeroBasedMacroNumber]);
		mp = (unsigned char*)*gMacros[zeroBasedActiveSetNumber][inZeroBasedMacroNumber];
		first = mp;
		
		Network_GetFourCharIPAddress(myipnum);
	
		while (*mp)
		{
			if (*mp == MACRO_IP)
			{
				Parse_TypeString(inScreen, (char*)first, mp - first, true/* record */);
				sprintf(temp, "%d.%d.%d.%d", myipnum[0], myipnum[1], myipnum[2], myipnum[3]);
				Parse_TypeString(inScreen, temp, strlen(temp), true/* record */);
				first = mp + 1;
			}
			else if (*mp == MACRO_LINES)
			{
				Parse_TypeString(inScreen, (char*)first, mp - first, true/* record */);
				sprintf(temp, "%d", VSgetlines(inScreen->vs));
				Parse_TypeString(inScreen, temp, strlen(temp), true/* record */);
				first = mp + 1;
			}
			else if (inScreen->ftpState != 0)
			{
				if (*mp == CR)
				{
					Parse_ShowOrInterpretAsCommand(inScreen, (unsigned char*)"\015\012", 2);
					inScreen->kbbuf[inScreen->kblen++] = 0;
					ftppi(inScreen->kbbuf); // FTP client
					inScreen->kblen = 0;
				}
				else
				{
					inScreen->kbbuf[inScreen->kblen++] = *mp;
					Parse_ShowOrInterpretAsCommand(inScreen, mp, 1);
				}
			}
			
			mp++;
		}
	
		if (inScreen->ftpState == 0) Parse_TypeString(inScreen, (char*)first, mp - first, true/* record */);
	
		HUnlock(gMacros[zeroBasedActiveSetNumber][inZeroBasedMacroNumber]);
	}
	
	return result;
}// Send


//
// internal methods
//
#pragma mark -

/*!
Determines if all macros in the specified set
are empty.  This is the case if all strings
in the set have zero length.

(3.0)
*/
static Boolean
allEmpty	(MacroSet		inSet)
{
	Boolean				result = true;
	register SInt16		i = 0;
	Str255				temp;
	
	
	for (i = 0; (i < MACRO_COUNT) && (result); i++)
	{
		getMacro(inSet, i, (char*)&temp, 256);
		if (strlen((char*)&temp) > 0) result = false;
	}
	return result;
}// allEmpty


/*!
To obtain a macro identically from a particular index
into an arbitrary array of macros, call this method.
The macro number is a zero-based index into the
given array of handles, and the desired array of
characters is copied from a handle that is obtained
appropriately from the array of handles.

If the given index is valid, 0 is returned; otherwise,
-1 is returned.

(3.0)
*/
static SInt16
getMacro	(MacroSet		inSet,
			 UInt16			inMacroNumber,
			 char*			outValue,
			 SInt16			inRoom)
{
	SInt16				result = 0;
	unsigned char*		s = null;
	
	
	// invalid number?
	if ((inMacroNumber >= 0) && (inMacroNumber <= (MACRO_COUNT - 1)))
	{
		if (inSet[inMacroNumber] != null)
		{
			s = (unsigned char*)*inSet[inMacroNumber];
			
			while (*s && (inRoom >= 5)) // 5 = (size of \xxx) + (terminating \0)
			{
				switch (*s)
				{
					case MACRO_IP:
						// 0xFF -> "\i"
						*outValue++='\\';
						*outValue++='i';
						inRoom--;
						break;
					
					case MACRO_LINES:
						// 0xFE -> "\#"
						*outValue++='\\';
						*outValue++='#';
						inRoom--;
						break;
					
					case '\\':
						// \ -> "\\"
						*outValue++='\\';
						*outValue++='\\';
						inRoom--;
						break;
					
					case '\033':
						// ESC -> "\e"
						*outValue++='\\';
						*outValue++='e';
						inRoom--;
						break;
					
					case '\015':
						// CR -> "\r"
						*outValue++='\\';
						*outValue++='r';
						inRoom--;
						break;
					
					case '\012':
						// LF -> "\n"
						*outValue++='\\';
						*outValue++='n';
						inRoom--;
						break;
					
					case '\t':
						// tab -> "\t"
						*outValue++='\\';
						*outValue++='t';
						inRoom--;
						break;
					
					default: 
						if (isprint(*s)) *outValue++=*s;
						else
						{
							*outValue++='\\';
							*outValue++= (*s / 64) +'0';
							*outValue++= ((*s % 64) / 8)+'0';
							*outValue++= (*s % 8) +'0';
							inRoom = inRoom - 3;
						}
						break;
				}
				
				inRoom--;
				s++;
			}
		}
		*outValue = 0;
	}
	else result = -1;
	
	return result;
}// getMacro


/*!
To explicitly change all of the macros in the
specified temporary space to correspond to the
text imported from an open file, call this
method.

If specified (that is, non-null), the parameter
"outMacroModeOfImportedSet" can be used to
determine the type of key equivalents that a
files macros were meant to use.  This is not
always possible to determine - in such case, the
default value is "kTelnetMacroModeCommandDigit".

IMPORTANT:	Macros_ParseTextBuffer() does the
			same fundamental thing as this
			routine, but the complexities of
			file access made it easier to simply
			duplicate most of the code.  If you
			ever change anything here, be sure
			to compare your changes with the
			code of Macros_ParseTextBuffer().

(2.6)
*/
static void
parseFile	(MacroSet			inSet,
			 SInt16				inFileRef,
			 TelnetMacroMode*	outMacroModeOfImportedSet)
{
	unsigned char		buffer[300];
	unsigned char		newMacro[256];
	unsigned char*		bufferPtr = null;
	unsigned char*		newMacroPtr = null;
	OSErr				fileErr = noErr;
	UInt16				numMacrosRead = 0;
	register SInt16		totalLen = 0,
						i = 0;
	long				count = 1; // read one byte per file access
	
	
	bufferPtr = buffer; 
	
	setTempMacrosToNull(inSet); // sets all handles in the current temporary macro space to null
	
	while ((fileErr != eofErr) && (numMacrosRead < MACRO_COUNT))
	{
		fileErr = FSRead(inFileRef, &count, bufferPtr);
		while ((*bufferPtr != 0x0D) && (fileErr != eofErr)) // while not CR or EOF
		{
			++bufferPtr;
			fileErr = FSRead(inFileRef, &count, bufferPtr);
		}
		
		// 3.0 - determine the key equivalents, if possible (this only needs to be done once)
		if ((!numMacrosRead) && (outMacroModeOfImportedSet != null))
		{
			// See Macros_ExportToText() for code that explicitly specifies that
			// the key for a function key macro starts with a capital "F".  That
			// code is the basis for the following test, which simply looks at
			// the first character to find the key equivalents of the set.
			if (buffer[0] == 'F') *outMacroModeOfImportedSet = kTelnetMacroModeFunctionKeys;
			else *outMacroModeOfImportedSet = kTelnetMacroModeCommandDigit;
		}
		
		totalLen = bufferPtr - buffer;
		bufferPtr = buffer;
		newMacroPtr = newMacro;
		while ((*bufferPtr++ != '"') && (totalLen != 0)) --totalLen;
		
		while ((*bufferPtr != '"') && (totalLen != 0))
		{
			*newMacroPtr++ = *bufferPtr++;
			--totalLen;
		}
		*newMacroPtr = null; // make this a C string
		
		Macros_Set(inSet, numMacrosRead, (char*)newMacro);
		numMacrosRead++;
		bufferPtr = buffer;
	}
}// parseFile


/*!
To nullify all macro handles in the space of the
active temporary set, leaving other temporary sets
untouched, call this method.

(3.0)
*/
static void
setTempMacrosToNull		(MacroSet		inSet)
{
	register SInt16		i = 0;
	
	
	for (i = 0; i < MACRO_COUNT; i++)
	{
		if (inSet[i] != null) Memory_DisposeHandle(&inSet[i]);
		inSet[i] = null;
	}
}// setTempMacrosToNull


/*!
To set a macro identically for a particular index
into an arbitrary array of macros, call this method.
The macro number is a zero-based index into the
given array of handles, and the given array of
characters is copied into a new handle that is
placed appropriately in the array of handles.

If the given index is valid, 0 is returned; otherwise,
-1 is returned.

(3.0)
*/
static void
setMacro	(MacroSet		inoutDestination,
			 UInt16			inMacroNumber,
			 char*			inoutValue)
{
	unsigned char*		p = null;
	SInt16				num = 0,
						pos = 0,
						escape = 0;
	SInt16				len = 0;
	OSErr 				memError = noErr;
	
	
	if ((inMacroNumber >= 0) && (inMacroNumber <= (MACRO_COUNT - 1)))
	{
		// restrict the maximum length of macros to MACRO_MAX_LEN bytes
		len = strlen(inoutValue) + 1;
		if (len > (MACRO_MAX_LEN - 1))
		{
			len = MACRO_MAX_LEN;
			inoutValue[MACRO_MAX_LEN - 1] = 0;
		}
		
		// if this is an empty string, remove whatever storage might have
		// been used previously by this macro
		if (len == 1)
		{
			if (inoutDestination[inMacroNumber] != null)
			{
				Memory_DisposeHandle(&inoutDestination[inMacroNumber]);
			}
			return;
		}
			
		// if necessary, create storage for the macro
		if (inoutDestination[inMacroNumber] == null)
		{
			inoutDestination[inMacroNumber] = Memory_NewHandle(len);
			if (inoutDestination[inMacroNumber] == null)
			{		// Memory error
				return;
			}
		}
	
		// adjust the handle to the proper size (may be making an existing macro longer)
		memError = Memory_SetHandleSize(inoutDestination[inMacroNumber], len);
		if (memError != noErr)
		{
			return;
		}
		
		HLock(inoutDestination[inMacroNumber]);
		p = (unsigned char*)*inoutDestination[inMacroNumber];
	
		num = 0;
		pos = 0;
		escape = 0;
		
		while (*inoutValue)
		{
			if (escape)
			{
				escape = 0;
				switch (*inoutValue)
				{
					case 'i': // current IP address
						if (pos > 0)
						{
							*p++ = num;
							*p++ = *inoutValue;
							pos = 0;
						}
						*p++ = MACRO_IP;
						break;
					
					case '#': // number of rows in terminal screen
						if (pos > 0)
						{
							*p++ = num;
							*p++ = *inoutValue;
							pos = 0;
						}
						*p++ = MACRO_LINES;
						break;
					
					case 'e': // ESC
						if (pos > 0)
						{
							*p++ = num;
							*p++ = *inoutValue;
							pos = 0;
						}
						*p++ = '\033';
						break;
					
					case 'n': // new-line
						if (pos > 0)
						{
							*p++ = num;
							*p++ = *inoutValue;
							pos = 0;
						}
						*p++ = '\012';
						break;
					
					case 'r': // carriage return
						if (pos > 0)
						{
							*p++ = num;
							*p++ = *inoutValue;
							pos = 0;
						}
						*p++ = '\015';
						break;
					
					case 't': // tab
						if (pos > 0)
						{
							*p++ = num;
							*p++ = *inoutValue;
							pos = 0;
						}
						*p++ = '\t';
						break;
					
					case '"': // double-quotes
						if (pos > 0)
						{
							*p++ = num;
							*p++ = *inoutValue;
							pos = 0;
						}
						*p++ = '\"';
						break;
							
					case '\\': // single backslash
						if (pos > 0)
						{
							*p++ = num;
							escape = 1;
							pos = 0;
							num = 0;
						}
						else *p++ = '\\';
						break;
					
					default:
						if (*inoutValue <= '9' && *inoutValue >= '0' && pos < 3)
						{
							num = num * 8 + (*inoutValue - '0');
							pos++;
							escape = 1;
						}
						else
						{
							if (pos == 0 && num == 0)
							{
								*p++ = '\\';
								*p++ = *inoutValue;
							}
							else
							{
								*p++ = num;
								pos = 0;
								inoutValue--; // retreat to previous character
							}
						}
						break;
				}
			}
			else
			{
				if (*inoutValue == '\\')
				{
					num = 0;
					pos = 0;
					escape = 1;
				}
				else *p++ = *inoutValue;
			}
			inoutValue++;
		}
		
		if (pos > 0) *p++ = num;
		
		*p = 0;
		
		// The resultant macro may be shorter than the input string due to escaped characters.
		// So, recalculate the length of the macro and resize the handle if necessary.
		len = strlen(*inoutDestination[inMacroNumber]) + 1;
		
		HUnlock(inoutDestination[inMacroNumber]);
		Memory_SetHandleSize(inoutDestination[inMacroNumber], len);
	}
}// setMacro