/*###############################################################

	Main.c
	
	This module contains the main entry point to the program.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <OpenTransport.h>

// resource includes
#include "DialogResources.h"

// library includes
#include "InterfaceLib.h"

// MacTelnet includes
#include "Clipboard.h"
#include "CommandLine.h"
#include "ConnectionData.h"
#include "Console.h"
#include "CoolStuff.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "InternetConfig.h"
#include "InternetServices.h"
#include "Keypads.h"
#include "MacBinary.h"
#include "NetEvents.h"
#include "network.proto.h"
#include "Preferences.h"
#include "PreferencesDialog.h"
#include "RealScreen.h"
#include "ResourceFileManager.h"
#include "InstallAE.h"
#include "TelnetDescriptors.h"
#include "TelnetErrors.h"
#include "TelnetMacros.h"
#include "TelnetMain.h"
#include "TelnetMenus.h"
#include "TelnetThreads.h"



// external methods
void Init_StartUpTelnet	(void);



// debugging variables
SInt32			DEBUG_STATE;

// QuickDraws global variables
#ifndef __MWERKS__
QDGlobals		qd;
#endif

// internal methods
static void		exitCleanly			(void);



//
// public methods
//

/*!
This is the main entry point to the MacTelnet 3.0 application
program.  This is the first function that gets called, and
(ideally) the program should end only when this function ends.

(2.6)
*/
void
main	(void)
{
	Init_StartUpTelnet();
	{
		FlagManager_Set(kFlagGlobalHaveNotifiedLowMemory, false);
		FlagManager_Set(kFlagGlobalMemoryIsOK, false);
		FlagManager_Set(kFlagGlobalMainLoopDone, false);
		do
		{
			if (FlagManager_Test(kFlagThreadManager)) Threads_YieldToAnother();
			else
			{
				// no threading means that calls have to be made manually
				(voidPtr)Threads_InterfaceIdle(null/* unused parameter */);
				(voidPtr)Threads_StandardIdle(null/* unused parameter */);
			}
			
			// event handling, always done in the main thread
			NetEvents_Service();
			EventLoop_Service();
			
			// low-memory checks
			FlagManager_Set(kFlagGlobalMemoryIsOK, Memory_RecoverReserveIsAnythingLeft());
			if (FlagManager_Test(kFlagGlobalMemoryIsOK))
			{
				FlagManager_Set(kFlagGlobalHaveNotifiedLowMemory, false);
			}
			else if (!FlagManager_Test(kFlagGlobalHaveNotifiedLowMemory)) // only warn the user once
			{
				Alert(MemoryLowAlert, null);
				FlagManager_Set(kFlagGlobalHaveNotifiedLowMemory, true);
			}
		} while (!FlagManager_Test(kFlagGlobalMainLoopDone));
	}
	exitCleanly();
}// main


/*!
Registers a module with the main program, specifying
optional functions for initializing and cleaning up
the module.  The module name is used for debugging.
The initialization function is called almost
immediately, as the program starts up.  The clean-up
function is invoked at the end of the program.

The order in which Main_RegisterModule() is invoked
for the various modules will determine the order in
which initialization functions are called.  At
quitting time, the clean-up functions are called in
exactly the reverse of this order.

Unimplemented.

(3.0)
*/
void
Main_RegisterModule		(StringPtr				inModuleName,
						 ModuleInitProcPtr		UNUSED_ARGUMENT(inInitializationFunctionOrNull),
						 ModuleCleanUpProcPtr	UNUSED_ARGUMENT(inCleanUpFunctionOrNull))
{
	Console_WriteValuePString("Registering module", inModuleName);
	
	// unimplemented
}// RegisterModule


/*!
Call this method to force the program to clean up (if
possible) and to then *definitely* shut down.

(3.0)
*/
void
Main_QuitImmediately	(void)
{
	// WARNING - until I finish adding inited flags to the modules,
	// it is not safe to use quit-signal; if an error occurs at startup
	// time, Main_QuitSignal() may accidentally trigger a "..._Done()"
	// call for a module whose "...Init()" has not been called, and in
	// certain cases this will cause a hang.  I am working to detect
	// and avoid this possibility.  In general, Main_QuitImmediately()
	// is only used in error conditions anyway, so this safeguard makes
	// sure that an error causing MacTelnet to quit doesnt CRASH THE
	// SYSTEM TOO!
	//Main_QuitSignal();
	exitCleanly();
}// QuitImmediately


/*!
Call this method when the program is about to quit,
so things can be cleaned up.  This method sets the
done flag, so the program will shut down virtually
right after this method returns and the next event
loop occurs.

(3.0)
*/
void
Main_QuitSignal		(void)
{
	// saving user preferences is a top priority, since other quitting procedures could cause a crash
	PreferencesDialog_Done();
	Macros_Done();
	
	// set implicit window visibility preferences, to be restored the next time Telnet is run
	{
		Boolean		windowIsVisible = false;
		
		
		windowIsVisible = Clipboard_IsVisible();
		(TelnetPrefsResultCode)Preferences_SetData(kTelnetPreferenceClassGeneral,
													kTelnetPreferenceTag_WasClipboardShowing,
													sizeof(Boolean), &windowIsVisible);
		windowIsVisible = CommandLine_IsVisible();
		(TelnetPrefsResultCode)Preferences_SetData(kTelnetPreferenceClassGeneral,
													kTelnetPreferenceTag_WasCommandLineShowing,
													sizeof(Boolean), &windowIsVisible);
		(OSErr)Preferences_SaveData(kTelnetPreferenceClassGeneral); // write the implicit data to disk
	}
	
	Clipboard_Done();
	CommandLine_Done();
	
	CoolStuff_Done();
	
	Network_Done();
	close_mb_files();
	InternetConfig_Done();
	
	Keypads_Done();
	#ifndef NDEBUG
	Console_Done();
	#else
	FTPLog_Done();
	#endif
	RealScreen_Done();
	ScreenManager_Done();
	Menus_Done();
	Errors_Done();
	Alert_Done();
	
	EventLoop_Done();
	Preferences_Done();
	InstallAE_Done();
	Cursors_Done();
	Sound_Done();
	Threads_Done();
	LED_Restore();
	Undoables_Done();
	InternetServices_Done(); // doing this last avoids CloseOpenTransport()s memory corruption bug
	
	// set the flag that will cause the next main loop to terminate the program
	FlagManager_Set(kFlagGlobalMainLoopDone, true);
}// QuitSignal


//
// internal methods
//

/*!
Quits MacTelnet, freeing any locked-up
resources (this function does not return).

(3.0)
*/
static void
exitCleanly		(void)
{
	// close any resource files that MacTelnet would have opened
	{
		SInt16		resFile = 0;
		
		
		resFile = ResourceFileManager_GetResFile(kTelnetResourceFilePreferences);
		if (resFile >= 0) CloseResFile(resFile);
		resFile = ResourceFileManager_GetResFile(kTelnetResourceFileTerminalBellSound);
		if (resFile >= 0) CloseResFile(resFile);
	}
	FlagManager_Done();
	ExitToShell();
}// exitCleanly