/*###############################################################

	Memory.c
	
	Utility library for more-controlled memory management.
	Consistent use of this module has many benefits, including
	more grace if the program runs out of memory, more ability
	to gauge memory requirements, and increased stability.
	
	Data Access Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <MacErrors.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"



// types
struct SafeArray
{
	SInt16		lockCount;			// how many locks are on the handle
	void*		array;
	UInt32		descriptor;
	struct
	{
		Size		size;
		UInt16		count;
	} element;
};
typedef struct SafeArray	SafeArray;
typedef SafeArray*			SafeArrayPtr;
typedef SafeArray**			SafeArrayHandle;

// variables
static Size			gCushion = 0;				/* size of memory cushion */
static Size			gReserve = 0;				/* size of memory reserve */
static Handle		gReserveBlock = null;		/* reserve block */
static Boolean		gCritical = true;			/* true if current memory request is critical - see internal methods */

// internal methods
static void			beginMemoryManagerSystemCall		(void);
static void			endMemoryManagerSystemCall			(void);
static pascal long	growZone							(Size			howMuch);
static Boolean		memoryAvailable						(Size			bytesNeeded);
static SafeArrayPtr	refAcquireLock						(SafeArrayRef	inRef);
static void			refReleaseLock						(SafeArrayRef	inRef,
														 SafeArrayPtr*	inoutArrayPtr);
static Boolean		waitingOnMemoryManagerSystemCall	(void);



//
// public methods
//

/*!
Initializes the memory management variables used by
the routines in this file, and allocates reserve
memory space.

(1.0)
*/
void
Memory_Init		(Size const		inNumberOfBytesOfHeapPadding,
				 Size const		inNumberOfBytesToReserve)
{
	OSErr					err = noErr;
	static GrowZoneUPP		myGrowZoneUPP = null;
	
	
	gCushion = inNumberOfBytesOfHeapPadding;
	gReserve = inNumberOfBytesToReserve;
	gReserveBlock = NewHandle(inNumberOfBytesToReserve);
	myGrowZoneUPP = NewGrowZoneUPP(growZone/* an internal method */);
	SetGrowZone(myGrowZoneUPP);
}// Init


/*!
All array accesses should use this function, for the
protection of the system and for debugging.  In the event
of an invalid array access, an error message is displayed
and the program is forced to shut down, so this routine
will not return in that case.

If you pass null for the last parameter, then this routine
simply returns a pointer to the beginning of the array item
specified.  If you do not pass null to this routine, the
data pointed to in the last parameter will presumably
point to a buffer of the exact size as was specified for
array elements at array creation time.  In this latter
case, the data will be copied to replace the specified
item in the array, *and* the address will be returned in
the same way as if a simple get operation were performed.

See Memory_MakeArray() for more details on safe arrays.

(1.0)
*/
void*
Memory_AccessSafeArray	(SafeArrayRef		inArray,
						 SInt16 const		inIndex,
						 void*				inDataPtrOrNull)
{
	void*			result = null;
	SafeArrayPtr	ptr = refAcquireLock(inArray);
	Boolean			isError = false;
	
	
	if (ptr != null)
	{
		if ((inIndex < 0) || (inIndex >= ptr->element.count)) isError = true;
		else
		{
			result = (void*)(((Size)ptr->array) + inIndex * ptr->element.size); // return the memory location of the requested array element
			if (inDataPtrOrNull != null)
			{
				// set data
				BlockMove((Ptr)inDataPtrOrNull, (Ptr)result, ptr->element.size);
			}
		}
	}
	else isError = true;
	
	if (isError)
	{
		// THROW EXCEPTION HERE!
	}
	
	refReleaseLock(inArray, &ptr);
	return result;
}// AccessSafeArray


/*!
Destroys the memory allocated by a Mac OS Memory Manager
handle, and then sets the handle to null.  Since one
frequently compares a handle with the value "null" to see
if its valid, it makes sense to ensure that a handle is
null when it actually gets disposed.  Favor this method
over a call to the Mac OS Memory Manager routine
DisposeHandle().

(1.0)
*/
void
Memory_DisposeHandle	(Handle*		inoutHandle)
{
	if (inoutHandle != null)
	{
		DisposeHandle(*inoutHandle);
		*inoutHandle = null;
	}
}// DisposeHandle


/*!
Destroys the memory allocated by a Mac OS Memory Manager
pointer, and then sets the pointer to null.  Since one
frequently compares a pointer with the value "null" to see
if its valid, it makes sense to ensure that a pointer is
null when it actually gets disposed.  Favor this method
over a call to the Mac OS Memory Manager routine
DisposePtr().

(1.0)
*/
void
Memory_DisposePtr		(Ptr*		inoutPtr)
{
	if (inoutPtr != null)
	{
		DisposePtr(*inoutPtr);
		*inoutPtr = null;
	}
}// DisposePtr


/*!
Frees a memory block previously allocated with the
Memory_NewPtrInterruptSafe() routine.  This routine
uses the Open Transport call OTFreeMem(), and is
interrupt-safe.

(1.0)
*/
void
Memory_DisposePtrInterruptSafe		(void**		inoutPtr)
{
	if (inoutPtr != null) OTFreeMem(*inoutPtr), *inoutPtr = null;
}// DisposePtrInterruptSafe


/*!
Destroys a QuickDraw region allocated with Memory_NewRegion().
On output, the region handle is automatically set to null to
prevent you from accidentally using it once it is destroyed.

(1.0)
*/
void
Memory_DisposeRegion	(RgnHandle*		inoutRegion)
{
	if (inoutRegion != null) DisposeRgn(*inoutRegion), *inoutRegion = null;
}// DisposeRegion


/*!
Destroys a safe array previously created with the
Memory_NewSafeArray() method.

(1.0)
*/
void
Memory_DisposeSafeArray		(SafeArrayRef*		inoutArrayPtr)
{
	if (inoutArrayPtr != null)
	{
		SafeArrayPtr		ptr = refAcquireLock(*inoutArrayPtr);
		
		
		// release all memory occupied by the safe array
		Memory_DisposePtr((Ptr*)&ptr->array);
		Memory_DisposeHandle((Handle*)inoutArrayPtr);
	}
}// DisposeSafeArray


/*!
To determine the number of elements in a safe array,
use this method.

(1.0)
*/
UInt16
Memory_GetSafeArraySize		(SafeArrayRef		inRef)
{
	SafeArrayPtr		ptr = refAcquireLock(inRef);
	UInt16				result = 0;
	
	
	if (ptr != null)
	{
		result = ptr->element.count;
	}
	
	refReleaseLock(inRef, &ptr);
	return result;
}// GetSafeArraySize


/*!
Prior to any segment of code that uses HLock(), you need
to call this routine to determine the locked state of the
specified handle.  Do not use HGetState() and HSetState()
when you only wish to save and restore the locked state,
because HSetState() may restore more than the locked
state (the purged state, for example, may get restored).
Instead, use this routine to save the locked state, and
then call HUnlock() when you are done with the handle
*only* if the handle was unlocked when you saved it (that
is, this routine returned false).

(1.0)
*/
Boolean
Memory_IsHandleLocked		(Handle			inHandle)
{
	enum
	{
		kHandleLockMask = 0x80
	};
	Boolean		result = false;
	SInt8		hState = HGetState(inHandle);
	
	
	result = (hState & kHandleLockMask);
	
	return result;
}// IsHandleLocked


/*!
Allocates a new Mac OS Memory Manager memory block and
returns a handle to it.  If there is not enough contiguous
free space in the heap zone to successfully perform the
allocation, a null handle is returned.

IMPORTANT:	You should not normally use Memory_NewHandle();
			instead, use Memory_NewHandleInProperZone(),
			which lets you specify additionally how your
			handle will be used.  Based on whether the new
			handle is going to be needed for a long time or
			a short time, Memory_NewHandleInProperZone()
			automatically calls ReserveMem() or MoveHHi()
			appropriately.

(1.0)
*/
Handle
Memory_NewHandle		(Size const		inDesiredNumberOfBytes)
{
	Handle		result = null;
	
	
	if (memoryAvailable(inDesiredNumberOfBytes))
	{
		beginMemoryManagerSystemCall();
		
		// this check helps to avoid a serious Memory Manager bug that exists on
		// some versions of the Mac OS
		if ((inDesiredNumberOfBytes >= 0xFFFFFFE1) && (inDesiredNumberOfBytes <= 0xFFFFFFFF))
		{
			result = null;
		}
		else result = NewHandleClear(inDesiredNumberOfBytes);
		
		endMemoryManagerSystemCall();
	}
	return result;
}// NewHandle


/*!
Allocates a new Mac OS Memory Manager memory block and
returns a handle to it, taking steps to compact the
heap if this critical memory management operation fails.
If there is not enough contiguous free space in the heap
zone, even after compaction, to successfully perform the
allocation, a null handle is returned.

(1.0)
*/
Handle
Memory_NewHandleCritical		(Size const		inDesiredNumberOfBytes)
{
	Handle		result = null;
	
	
	result = NewHandleClear(inDesiredNumberOfBytes);
	return result;
}// NewHandleCritical


/*!
Allocates a new Mac OS Memory Manager memory block and
returns a handle to it, automatically calling either
MoveHHi() or ReserveMem(), as appropriate, based on
whether you need the handle for a short period of time.
If there is not enough contiguous free space in the
heap zone to successfully perform the allocation, a
null handle is returned.

If you say you are using the handle for a long time, it
is AUTOMATICALLY LOCKED, and should not be unlocked
unless you have a good reason!

In general, you should call this routine for all new
dynamic memory allocations.  This helps to maintain as
much contiguous free space as possible for as long as
possible, reducing the likelihood that future memory
requests will fail.

(3.0)
*/
Handle
Memory_NewHandleInProperZone	(Size const		inDesiredNumberOfBytes,
								 Boolean const	inOnlyNeededForShortPeriodOfTime)
{
	Handle		result = null;
	
	
	// handles sticking around for a long time should have space reserved for them ahead of time
	if (!inOnlyNeededForShortPeriodOfTime) ReserveMem(inDesiredNumberOfBytes);
	
	// allocate space
	result = Memory_NewHandle(inDesiredNumberOfBytes);
	
	// handles sticking around for a short period of time should be high in the heap zone;
	// conversely, those that dont should be locked throughout their entire existence
	if (inOnlyNeededForShortPeriodOfTime) MoveHHi(result);
	else HLock(result);
	
	return result;
}// NewHandleInProperZone


/*!
Allocates a new Mac OS Memory Manager memory block and
immediately moves it high in the heap and locks it.
In this locked state, it is safe to point to the data,
so a pointer to the data is returned for convenience.
Use this routine to allocate blocks that you only need
briefly.

If there is not enough contiguous free space in the
heap zone to successfully perform the allocation, a
null pointer is returned.

In general, you should call this routine for any new
dynamic memory allocation where you might ordinarily
use a pointer.  This makes the block relocatable in
memory emergencies, but nonetheless places it high in
the heap.  Call HUnlock() on the returned new handle
when you want to make the block relocatable again.

(1.0)
*/
Ptr
Memory_NewHandleLockedHigh	(Size const		inDesiredNumberOfBytes,
							 Handle*		outNewHandlePtr)
{
	Ptr			result = null;
	
	
	if (outNewHandlePtr != null)
	{
		*outNewHandlePtr = Memory_NewHandleInProperZone(inDesiredNumberOfBytes, true/* only briefly allocated */);
		if (*outNewHandlePtr != null)
		{
			HLock(*outNewHandlePtr);
			result = (**outNewHandlePtr); // the handle is locked, so it is safe to return a pointer
		}
	}
	return result;
}// NewHandleLockedHigh


/*!
Allocates a new Mac OS Memory Manager memory block and
returns a pointer to it.  If there is not enough contiguous
free space in the heap zone to successfully perform the
allocation, a null pointer is returned.

(1.0)
*/
Ptr
Memory_NewPtr		(Size const		inDesiredNumberOfBytes)
{
	Ptr		result = null;
	
	
	if (memoryAvailable(inDesiredNumberOfBytes))
	{
		beginMemoryManagerSystemCall();
		result = NewPtrClear(inDesiredNumberOfBytes);
		endMemoryManagerSystemCall();
	}
	return result;
}// NewPtr


/*!
Allocates a new Mac OS Memory Manager memory block and
returns a pointer to it, taking steps to compact the
heap if this critical memory management operation fails.
If there is not enough contiguous free space in the heap
zone, even after compaction, to successfully perform the
allocation, a null pointer is returned.

(1.0)
*/
Ptr
Memory_NewPtrCritical	(Size const		inDesiredNumberOfBytes)
{
	Ptr		result = null;
	
	
	result = NewPtrClear(inDesiredNumberOfBytes);
	return result;
}// NewPtrCritical


/*!
Allocates a new memory block and returns a pointer to it.
This routine uses the Open Transport call OTAllocMem(),
and automatically uses the proper context.  It is
interrupt-safe.

(1.0)
*/
void*
Memory_NewPtrInterruptSafe		(Size const		inDesiredNumberOfBytes)
{
	void*		result = null;
	
	
#if TARGET_API_MAC_OS8
	result = OTAllocMem(inDesiredNumberOfBytes);
#else
	result = OTAllocMemInContext(inDesiredNumberOfBytes, null/* client context; null = use application context */);
#endif
	if (result != null) OTMemzero(result, inDesiredNumberOfBytes);
	return result;
}// NewPtrInterruptSafe


/*!
Allocates a new QuickDraw region that can be used
with any of the standard QuickDraw routines that
manipulate regions.  Use Memory_DisposeRegion()
to destroy this region when youre done.

There are certain conditions under which it may be
impossible to detect failed region allocations:
specifically, if QDError() returns "noErr" and
memory conditions are extremely low, a region may
in fact be garbage but not seem like it.  This
routine performs a sanity check on new regions to
make sure there really is enough space in the grow
zone to accommodate the region that QuickDraw
claims to have created, and returns "null" if an
otherwise-valid region is accidentally allocated.

(1.0)
*/
RgnHandle
Memory_NewRegion	(void)
{
	RgnHandle		result = null;
	Size			growSize = 0;
	
	
	(Size)MaxMem(&growSize);
	result = NewRgn();
	if (GetHandleSize((Handle)result) > growSize) result = null; // QuickDraw tried to pull a fast one!  The region is garbage.
	return result;
}// NewRegion


/*!
Use this method to create a safe array, consisting
of the specified number of elements.  The array size
is stored as well.  You define a descriptor so you can
uniquely describe the contents of the array, in the
absence of the usual type protection from the compiler.

Safe arrays are useful when you are dealing with a
difficult bug involving array overflows.  Usually, the
Mac OS will freeze or do something horrible, and it is
hard to debug between restarts.  By replacing your
normal C array with a safe array, you can debug the
fault without actually damaging the contents of memory
when triggering the fault.

If any problems occur, null is returned.

WARNING:	The returned pointer is not an array!  You
			cannot use it as one.  Use the additional
			helper routine Memory_AccessSafeArray() to
			work with items using the returned pointer.

(1.0)
*/
SafeArrayRef
Memory_NewSafeArray		(UInt32 const		inDescriptor,
						 UInt16 const		inElementCount,
						 Size const			inElementSize)
{
	SafeArrayRef		result = (SafeArrayRef)Memory_NewHandle(sizeof(SafeArray));
	Boolean				isError = false;
	
	
	if (result != null)
	{
		SafeArrayPtr	ptr = refAcquireLock(result);
		
		
		ptr->array = Memory_NewPtr(inElementSize * inElementCount);
		ptr->descriptor = inDescriptor;
		if (ptr->array != null)
		{
			ptr->element.size = inElementSize;
			ptr->element.count = inElementCount;
		}
		else
		{
			Memory_DisposePtr((Ptr*)&result);
			isError = true;
		}
		refReleaseLock(result, &ptr);
	}
	else isError = true;
	
	if (isError)
	{
		// OUT OF MEMORY!
	}
	
	return result;
}// NewSafeArray


/*!
Attempts to recover reserve, emergency memory that
would have been freed when a critical memory shortage
occurred.  If the reclaiming succeeds, "true" is
returned; otherwise, there is no free memory, so
"false" is returned.

(1.0)
*/
Boolean
Memory_RecoverReserveIsAnythingLeft		(void)
{
	Boolean		result = false;
	
	
	if (gReserveBlock == null) result = true;
	else
	{
		if (*gReserveBlock != null) result = true;
		else
		{
			ReallocateHandle(gReserveBlock, gReserve);
			if (MemError() != noErr) result = false;
			else result = memoryAvailable(0);
		}
	}
	return result;
}// RecoverReserveIsAnythingLeft


/*!
Resizes a Mac OS Memory Manager memory block to which
you have a handle.  If there is not enough contiguous
free space in the heap zone to successfully perform the
resize operation, "memFullErr" is returned; otherwise,
"noErr" is returned.

(1.0)
*/
OSErr
Memory_SetHandleSize	(Handle			inoutHandle,
						 Size const		inNewHandleSizeInBytes)
{
	Size		oldLen = 0;
	OSErr		err = noErr;
	
	
	oldLen = GetHandleSize(inoutHandle);
	if (oldLen < inNewHandleSizeInBytes)
	{
		if (!memoryAvailable(inNewHandleSizeInBytes - oldLen)) return memFullErr;
	}
	beginMemoryManagerSystemCall();
	SetHandleSize(inoutHandle, inNewHandleSizeInBytes);
	endMemoryManagerSystemCall();
	err = MemError();
	if (err == memFullErr)
	{
		// the Mac OS Memory Manager doesnt do this itself for some reason
		MoveHHi(inoutHandle);
		CompactMem(maxSize);
		beginMemoryManagerSystemCall();
		SetHandleSize(inoutHandle, inNewHandleSizeInBytes);
		endMemoryManagerSystemCall();
		err = MemError();
	}
	if (err != noErr) return err;
	if (oldLen >= inNewHandleSizeInBytes) return noErr;
	if (memoryAvailable(0)) return noErr;
	SetHandleSize(inoutHandle, oldLen);
	return memFullErr;
}// SetHandleSize


/*!
Resizes a Mac OS Memory Manager memory block to which
you have a handle, taking steps to compact the heap if
this critical memory management operation fails.  If
there is not enough contiguous free space in the heap
zone, even after compaction, to successfully perform the
resize operation, "memFullErr" is returned; otherwise,
"noErr" is returned.

(1.0)
*/
OSErr
Memory_SetHandleSizeCritical	(Handle			inoutHandle,
								 Size const		inNewHandleSizeInBytes)
{
	SetHandleSize(inoutHandle, inNewHandleSizeInBytes);
	if (MemError() == memFullErr)
	{
		// the Mac OS Memory Manager doesnt do this itself for some reason
		MoveHHi(inoutHandle);
		CompactMem(maxSize);
		SetHandleSize(inoutHandle, inNewHandleSizeInBytes);
	}
	return MemError();
}// SetHandleSizeCritical


/*!
Clears a memory buffer in the most efficient way
possible for the current architecture.

(1.0)
*/
void
Memory_WriteZero	(void*			inoutBufferPtr,
					 Size const		inSize)
{
#if 0
	memset(inoutBufferPtr, 0, inSize);
#else
	UInt32*		pointer = inoutBufferPtr;
	long		size = inSize;
	
	
	while (size >= sizeof(UInt32))
	{
		size -= sizeof(UInt32);
		*pointer++ = 0L;
	}
	while (size-- > 0)
	{
		*((UInt8*)pointer++) = 0;
	}
#endif
}// WriteZero


//
// internal methods
//

/*!
Call this method before using most Mac OS Memory Manager
system calls in this file, and follow the system call
with a call to endMemoryManagerSystemCall().

(1.0)
*/
static void
beginMemoryManagerSystemCall	(void)
{
	gCritical = false;
}// beginMemoryManagerSystemCall


/*!
You usually need to call this method after using a
Mac OS Memory Manager routine in this file (one that
is preceded by beginMemoryManagerSystemCall()).

(1.0)
*/
static void
endMemoryManagerSystemCall	(void)
{
	gCritical = true;
}// endMemoryManagerSystemCall


/*!
This routine is automatically called by the Mac OS
Memory Manager if any problems are encountered while trying
to fulfil a memory request initiated via a system call
(like NewHandle(), for example).  This method attempts to
free a bit more memory in the heap, after which the Mac OS
will try again to fulfil the failed request.

Always bracket Memory Manager system calls in this file
with calls to the methods beginMemoryManagerSystemCall()
and endMemoryManagerSystemCall().

(1.0)
*/
static pascal long
growZone	(Size		cbNeeded)
{
	long		result = 0L;
	
	
	if (gReserveBlock != null)
	{
		if (waitingOnMemoryManagerSystemCall() && // is this a critical request?
			(*gReserveBlock != null) && // is there a memory reserve to free?
			gReserveBlock != GZSaveHnd()) // is the memory reserve free from protection by the Memory Manager?
		{
			EmptyHandle(gReserveBlock);
			result = gReserve;
		}
	}
	
	return result;
}// growZone


/*!
Determines if there is enough contiguous free space in
the heap zone to accommodate a memory allocation of
the specified number of bytes.

(1.0)
*/
static Boolean
memoryAvailable		(Size		inRequiredNumberOfBytes)
{
	Boolean		result = false,
				done = false;
	
	
	if (gReserveBlock == null || *gReserveBlock == null)
	{
		if (!Memory_RecoverReserveIsAnythingLeft())
		{
			result = false;
			done = true;
		}
	}
	if (!done)
	{
		long	unused = 0L,
				byteCountContiguousFree = 0L;
		
		
		PurgeSpace(&unused/*total free*/, &byteCountContiguousFree);
		result = ((inRequiredNumberOfBytes + gCushion) < byteCountContiguousFree);
	}
	return result;
}// memoryAvailable


/*!
To acquire a pointer to the internal safe array
structure, given a reference to it, use this method.
You must call refReleaseLock() when finished using a
pointer, at which point it will be an error to use
the same pointer before calling refAcquireLock() to
acquire the (possibly changed) pointer again.

(1.0)
*/
static SafeArrayPtr
refAcquireLock	(SafeArrayRef		inRef)
{
	SafeArrayPtr		result = *((SafeArrayHandle)inRef);
	
	
	HLock((Handle)inRef);
	result->lockCount++;
	return result;
}// refAcquireLock


/*!
To tell the Memory module that you are finished using
a safe array pointer, given a reference to it, use
this method.  This essentially makes the pointer
useless - do not use it again without calling the
refAcquireLock() function to obtain its (possibly
changed) value.  On output, your copy of the pointer
will be set to "null".

(1.0)
*/
static void
refReleaseLock	(SafeArrayRef		inRef,
				 SafeArrayPtr*		inoutArrayPtr)
{
	SafeArrayPtr		ptr = *((SafeArrayHandle)inRef);
	
	
	if (--ptr->lockCount <= 0) HUnlock((Handle)inRef);
	if (inoutArrayPtr != null) *inoutArrayPtr = null;
}// refReleaseLock


/*!
Call this method to determine if a critical memory request
was cleared by beginMemoryManagerSystemCall() and has not
yet been asserted with a call to endMemoryManagerSystemCall().

(1.0)
*/
static Boolean
waitingOnMemoryManagerSystemCall	(void)
{
	return gCritical;
}// waitingOnMemoryManagerSystemCall