/*###############################################################

	Menus.c
	
	This file contains menu handling and initialization code,
	and has been completely re-written from the ground up in
	MacTelnet 3.0.  Now, every single item of every menu has a
	callback routine associated with it that determines its
	correct state (item text, enabled state, any marks, icons,
	etc.).  Just before menus are displayed or menu key
	equivalents are parsed, all callbacks are invoked to set
	up menus correctly.  Further, this functionality is used
	for contextual menus, effectively getting the context right
	every single time, because context is defined in exactly
	one place in the entire program.
	
	MacTelnet 3.0 also introduces Mac OS 8 command IDs for menus.
	Menus are now referred to exclusively in terms of their
	command IDs, when it comes to modifying items.  Every single
	reference to menu handle / item index information has been
	removed from MacTelnets source files, consolidating it all
	here.  This means that item indices are now defined as an
	internal enumeration, as youll see at the top of this file.
	This means that it is relatively easy to shuffle commands in
	menus, create new menus, and the like, without modifying
	anything outside of this file and the resource files.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <Balloons.h>
#include <ControlDefinitions.h>
#include <Folders.h>
#include <Fonts.h>
#include <OSA.h>
#include <PLStringFuncs.h>
#include <Script.h>
#include <Resources.h>
#include <Scrap.h>
#include <ToolUtils.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "MenuResources.h"
#include "StringResources.h"
#include "TelnetMenuItemNames.h"
#include "TelnetVersion.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "Inspector.h"
#include "FTPLog.h"
#include "Console.h"
#include "Commands.h"
#include "DialogItemIndices.h"

#include "FileUtilities.h"
#include "FolderManager.h"
#include "ScreenManager.h"

#include "PreferencesData.h"
#include "Preferences.h"
#include "Clipboard.h"
#include "CommandLine.h"
#include "ConnectionData.h"
#include "TelnetMenus.h"
#include "Sets.proto.h"
#include "configure.proto.h"
#include "Parse.h"
#include "InternalEvents.h"
#include "RealScreen.h"
#include "network.proto.h"
#include "MacLook.h"
#include "RasterGraphicsScreen.h"
#include "rgmp.proto.h"
#include "TektronixRealGraphics.h"
#include "TektronixVirtualGraphics.h"
#include "NetEvents.h"
#include "Connections.h"
#include "vsdata.h"
#include "Terminal.h"
#include "CoolStuff.h"
#include "TelnetThreads.h"
#include "EventLoop.h"
#include "DialogUtilities.h"
#include "TelnetErrors.h"
#include "TextTranslation.h"
#include "Keypads.h"
#include "ResourceFileManager.h"
#include "VirtualScreenEmulation.h"



/*!
IMPORTANT

The constants below are INTERNAL to this file and should NOT be
externalized under any circumstances.  The idea is to create an
opaque interface for using menus that is conducive to rearranging,
adding or removing menu commands easily in the future.

NEVER use Menu Manager functions to mess with menus unless it is
ABSOLUTELY necessary.  This file contains several utility functions
for determining and changing item attributes such as enabled state,
marks, and item text, all of which are keyed to work with command
IDs instead of absolute item locations.  You may hate me now, but
youll thank me if you ever have to add, remove or change menu
items in the program.  This approach also helps migrate to Carbon.

Adding menu items to the program:
	1.	Create appropriate data in " TelnetMenuItemNames.h", using
		the naming convention in that file.
	2.	Go to "MenuResources.h" and create a command ID for the
		item in the appropriate place, using the same suffix as
		you did for the item name constant (the same basic stems
		for menu commands are used everywhere data is defined for
		menu items - it should be obvious how it works).
	3.	Go to " TelnetMenus.r" and change the 'MENU' and 'xmnu'
		data in the menu Rez file to include an item with your new
		command ID.  Be careful not to mess up the 'xmnu'.  Now
		make sure you didnt screw up any 'hmnu' balloons.  You
		should go to " TelnetBalloonText.h" to create balloon text
		for the new item, in the appropriate string list.
	5.	Add an item index to the enumeration below that describes
		the item index of the menu item.
	6.	Modify the GetMenusAndMenuItemIndicesByCommandID() method
		to handle the case of your new item index; provide the
		menu position and item index in response to a query for
		your new command ID.
	7.	Set up state trackers for your new or changed items; this
		is done only in this file, in a couple of places: look
		for adjustMenuItemByCommandID() usage, and installation of
		state tracker routines.
*/
enum
{
/* Apple */
	iAboutTelnet = 1,
	iTelnetOnTheWeb = 2,
/* File (normal) */
	iOpenConnection = 1,
	iOpenConnectionFavorite = 2,
	iLoadSet = 3,
	//
	iCloseConnection = 5,
	iSaveSet = 6,
	//
	iImportMacroSet = 8,
	iExportCurrentMacroSet = 9,
	//
	iPrint = 11,
	iPrintOne = 12,
	iPrintScreen = 13,
	iPageSetup = 14,
	//
	iQuit = 16,
/* File (simple) */
	iSimpleOpenConnection = 1,
	iSimpleOpenConnectionFavorite = 2,
	//
	iSimpleCloseConnection = 4,
	//
	iSimplePrint = 6,
	iSimplePageSetup = 7,
	//
	iSimpleQuit = 9,
/* Edit (normal) */
	iUndo = 1,
	iRedo = 2,
	//
	iCut = 4,
	iCopy = 5,
	iCopyTable = 6,
	iType = 7,
	iPaste = 8,
	iClear = 9,
	//
	iSelectAll = 11,
	iShowClipboard = 12,
	//
	iPreferences = 14,
/* Edit (simple) */
	iSimpleCut = 1,
	iSimpleCopy = 2,
	iSimplePaste = 3,
	iSimpleClear = 4,
	//
	iSimplePreferences = 6,
/* View */
	iLargeScreen = 1,
	iSmallScreen = 2,
	iSetScreenSize = 3,
	//
	iBiggerText = 5,
	iSmallerText = 6,
	iFullScreen = 7,
	iFormat = 8,
/* Terminal */
	iCaptureToFile = 1,
	//
	iEcho = 3,
	iWrapMode = 4,
	iClearScreenSavesLines = 5,
	iJumpScrolling = 6,
	iTektronix = 7, // submenu
	//
	iSpeechEnabled = 9,
	//
	iResetTerminal = 11,
/* Keys */
	iDeletePressSendsBackspace = 1,
	iDeletePressSendsDelete = 2,
	//
	iEMACSArrowMapping = 4,
	iLocalPageUpDown = 5,
	//
	iSetKeys = 7,
	//
	iTitleTranslationChoices = 9,
	iFirstTranslationTable = 10,
/* Network */
	iShowNetworkNumbers = 1,
	iSendInternetProtocolNumber = 2,
	iSendFTPCommand = 3,
	//
	iSendAreYouThere = 5,
	iSendAbortOutput = 6,
	iSendBreak = 7,
	iSendInterruptProcess = 8,
	iSendEraseCharacter = 9,
	iSendEraseLine = 10,
	iSendEndOfFile = 11,
	iSendSync = 12,
	//
	iCommandWatchForActivity = 14,
	iSuspendNetwork = 15,
/* Window */
	iHideFrontWindow = 1,
	iHideOtherWindows = 2,
	iShowAllHiddenWindows = 3,
	//
	iStackWindows = 5,
	iNextWindow = 6,
	//
	iChangeWindowTitle = 8,
	iMinimizeWindow = 9,
	//
	iShowSessionInspector = 11,
	iShowFTPLog = 12,
	iShowCommandLine = 13,
	iShowControlKeys = 14,
	iShowFunction = 15,
	iShowKeypad = 16,
	//
	iFirstConnection = 18, // IMPORTANT!  Assumes the presence of the preceding dividing-line item.
/* Size submenu */
	iSize4 = 1,
	iSize9 = 2,
	iSize10 = 3,
	iSize11 = 4,
	iSize12 = 5,
	iSize14 = 6,
	iSize18 = 7,
/* Tektronix submenu */
	iTEKPageCommand = 1,
	iTEKPageClearsScreen = 3,
/* Internet Resources submenu */
	iTelnetHomePage = 1,
	iTelnetAuthorMail = 2
};

// types
struct ScreenTranslationInfo
{
	SInt16		oldTable;		// previous translation table number
	SInt16		newTable;		// number of translation table to be used from now on
};
#ifndef __cplusplus
typedef struct ScreenTranslationInfo	ScreenTranslationInfo;
#endif
typedef ScreenTranslationInfo*			ScreenTranslationInfoPtr;

// variables
extern short			nNational;
static MenuRef			gMenus[kMenuResourceCount];
static MenuItemIndex	helpItemIndexTelnetHelp = 0;
static IconSuiteRef		gScriptsMenuIconSuite = null,
						gApplicationMenuIconSuite = null;
static SInt16			gNumberOfConnectionItemsAdded = 0,
						gNumberOfSpecialScriptMenuItems = 0;
static Boolean			gFontMenusAvailable = false,
						gScriptsMenuAvailable = true; // flags to keep things thread safe

// internal methods
static void				adjustMenuItem						(MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static void				adjustMenuItemByCommandID			(UInt32 const			inCommandID);
static Boolean			areSessionRelatedItemsEnabled		(void);
static Boolean			areTEKRelatedItemsEnabled			(void);
static void 			buildMenuBar						(void);
static void				displayStatus						(TelnetScreenID const	inScreenID);
static void				disposeMainMenus					(void);
static void				executeScriptByMenuEvent			(MenuRef				inMenu,
															 MenuItemIndex			inItemNumber,
															 EventModifiers			inModifiers);
static void				getMainMenus						(void);
static void				getMenuItemAdjustmentProc			(MenuRef							inMenu,
															 MenuItemIndex						inItemNumber,
															 MenuCommandStateTrackerProcPtr*	outProcPtrPtr);
static TelnetScreenID	getScreenIDFromWindowMenuItemIndex	(MenuItemIndex			inIndex);
static MenuItemIndex	getWindowMenuItemIndexByScreenID	(TelnetScreenID			inScreenID);
static void				installDynamicMenuItemStateTrackers	(void);
static void				installMenuItemStateTrackers		(void);
static void				loadMenus							(Boolean const			inMenusHaveKeyEquivalents);
static short			portsOpen							(void);
static void				removeMenuItemModifiers				(MenuRef				inMenu);
static void				setClipboardMenuItemText			(SInt16 const			inFlags);
static void				setCommandLineMenuItemText			(SInt16 const			inFlags);
static void				setMenuEnabled						(MenuRef				inMenu,
															 Boolean const			inEnabled);
static void				setMenuItemEnabled					(MenuRef				inMenu,
															 UInt16 const			inItemNumberOrZero,
															 Boolean const			inEnabled);
static void				setMenuItemKeysFromControlKeyPrefs	(void);
static void				setUpSessionFavoritesSubmenu		(MenuRef				inMenu);
static void				setUpScriptsMenu					(MenuRef				inMenu);
static Boolean			stateTrackerAlwaysDisabled			(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerCheckableItems			(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerGenericSessionItems		(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerKeyboardMenuTranslationChoices	(UInt32 const	inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerNetSendItems			(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerPreferences				(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerPrintingItems			(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerShowHideItems			(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerStandardEditItems		(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerTEKItems				(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			stateTrackerWindowMenuConnections	(UInt32 const			inCommandID,
															 MenuRef				inMenu,
															 MenuItemIndex			inItemNumber);
static Boolean			translateScreenLine					(TelnetScreenID const	inScreenID,
															 char*					inoutLineTextBuffer,
															 UInt16 const			inLineTextBufferLength,
															 UInt16 const			inOneBasedLineNumber,
															 void*					inContextPtr);



//
// public methods
//

/*!
Call this routine once before performing any other
menu operations.

(3.0)
*/
void
Menus_Init	(void)
{
	Boolean		flag = false;
	Size		actualSize = 0L;
	
	
	// get the users menu keys preference, if possible
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_MenuItemKeys,
								sizeof(flag), &flag,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		flag = true; // assume menus have key equivalents, if preference cant be found
	}
	loadMenus(flag);
	
	// get the users simplified-user-interface preference, if possible
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_SimpleTelnet,
								sizeof(flag), &flag,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		flag = false; // assume normal mode, if preference cant be found
	}
	Menus_CheckSimpleTelnet(flag);
	
	// perform initial servicing
	Menus_Service();
	
#if TARGET_API_MAC_CARBON
	// explicitly request the Help menu, which will add this menu to the menu bar
	{
		MenuRef				helpMenu = null;
		MenuItemIndex		firstItemIndex = 0;
		OSStatus			error = noErr;
		
		
		error = HMGetHelpMenu(&helpMenu, &firstItemIndex/* first item index, or null */);
		if (error == noErr)
		{
			// under Carbon, Help items must be added explicitly
			// unimplemented
		}
	}
#endif
}// Init


/*!
Call this routine when you are completely done
with menus (at the end of the program).

(3.0)
*/
void
Menus_Done	(void)
{
	SetMenuMarkFromScreen(kMenuMarkTypeDispose, 0);
	if (gScriptsMenuIconSuite != null) DisposeIconSuite(gScriptsMenuIconSuite, true/* dispose data */);
	if (gApplicationMenuIconSuite != null) DisposeIconSuite(gApplicationMenuIconSuite, true/* dispose data */);
}// Done


/*!
Invoke this routine in response to an application
suspend or resume event, to ensure that menus get
updated appropriately.

(3.0)
*/
void
Menus_SetSuspended	(Boolean const		inIsSuspended)
{
	unless (inIsSuspended) setUpScriptsMenu(gMenus[kMenuPositionScripts]);
}// SetSuspended


/*!
This routine is used by a thread to set a global
variable that the Menu module requires.  The number
of special items refers to the number of items in
the top half of the menu, above any dividing line
(these items appear if any file names in the Scripts
folder start with "!").

(3.0)
*/
void
Menus_SetNumberOfSpecialScriptMenuItems	(UInt16 const		inNumber)
{
	gNumberOfSpecialScriptMenuItems = inNumber;
}// SetNumberOfSpecialScriptMenuItems


/*!
This method behaves exactly like the NewMenu() system
call, except it creates a menu that uses the new
Appearance menu definition procedure.  You should use
this method in place of any call to NewMenu().

(3.0)
*/
MenuRef
Menus_NewMenuThemeSavvy		(short				inMenuID,
							 ConstStr255Param	inMenuTitle)
{
	MenuRef		result = null;
	
	
#if TARGET_API_MAC_OS8
	result = NewMenu(inMenuID, inMenuTitle);
	if (result != null)
	{
		Handle		themeDefProc = GetResource('MDEF', kMenuStdMenuProc);
		
		
		// use the theme-savvy menu definition procedure
		if (themeDefProc != null)
		{
			SInt8		handleState = 0;
			
			
			handleState = HGetState((Handle)result);
			HLockHi((Handle)result);
			(*result)->menuProc = themeDefProc;
			HSetState((Handle)result, handleState);
		}
	}
#else
	(OSStatus)CreateNewMenu(inMenuID, 0/* menu attributes */, &result);
	SetMenuTitle(result, inMenuTitle);
#endif
	return result;
}// NewMenuThemeSavvy


/*!
To determine the item index of a menu item that
has the specified identical text, use this method.
If the item is not in the menu, 0 is returned.

(3.0)
*/
MenuItemIndex
Menus_GetItemIndexByText	(MenuRef		inMenu,
							 Str255			inItemText)
{
	MenuItemIndex	result = 0;
	UInt16			itemCount = CountMenuItems(inMenu),
					i = 0;
	Str255			itemString;
	
	
	// look for an identical item
	for (i = 1; i <= itemCount; i++)
	{
		GetMenuItemText(inMenu, i, itemString);
		if (PLstrlen(itemString) > 0) unless (PLstrcmp(itemString, inItemText))
		{
			result = i;
			break;
		}
	}
	
	return result;
}// GetItemIndexByText


/*!
To acquire a handle to the font menu, which will
have been stripped of all proportional font names
and have the font of each menu item set to the
same font it represents, use this routine.

(3.0)
*/
MenuRef
Menus_GetFontMenu	(void)
{
	return gMenus[kMenuPositionFont];
}// GetFontMenu


/*!
To acquire a handle to the scripts menu, which will
contain the names of scripts in MacTelnets Scripts
folder, use this method.

(3.0)
*/
MenuRef
Menus_GetScriptsMenu	(void)
{
	return gMenus[kMenuPositionScripts];
}// GetScriptsMenu


/*!
To acquire a handle to the translation menu, use
this routine.

(3.0)
*/
MenuRef
Menus_GetTranslationMenu	(void)
{
	return gMenus[kMenuPositionTranslation];
}// GetTranslationMenu


/*!
To obtain a unique variation of the specified
item text, guaranteeing that it will not be
the identical text of any item in the specified
menu, use this method.

(3.0)
*/
void
Menus_GetUniqueItemText		(MenuRef		inMenu,
							 Str255			inoutItemText)
{
	UInt16		itemCount = CountMenuItems(inMenu),
				i = 0;
	
	
	// only change the specified item text if its already taken by an item in the menu
	while (!Menus_IsItemUnique(inMenu, inoutItemText))
	{
		if ((inoutItemText[inoutItemText[0]] > '9') ||
			(inoutItemText[inoutItemText[0]] < '0')) // add a number
		{
			inoutItemText[++inoutItemText[0]] = ' ';
			inoutItemText[++inoutItemText[0]] = '1';
		}
		else if (inoutItemText[inoutItemText[0]] == '9') // add another digit
		{
			inoutItemText[inoutItemText[0]] = '-';
			inoutItemText[++inoutItemText[0]] = '1';
		}
		else inoutItemText[inoutItemText[0]]++; // increment the number
	}
}// GetUniqueItemText


/*!
To determine if a menu contains no item with
text identical to the specified item text,
use this method.

(3.0)
*/
Boolean
Menus_IsItemUnique		(MenuRef		inMenu,
						 Str255			inItemText)
{
	Boolean		result = true;
	UInt16		itemCount = CountMenuItems(inMenu),
				i = 0;
	Str255		itemString;
	
	
	// look for an identical item
	for (i = 1; ((result) && (i <= itemCount)); i++)
	{
		GetMenuItemText(inMenu, i, itemString);
		unless (PLstrcmp(itemString, inItemText)) result = false;
	}
	
	return result;
}// IsItemUnique


/*!
Enables the Edit menu.  This is only used for
dialog boxes with editable text items.

(3.0)
*/
void
Menus_EnableEditMenu	(void)
{
	Boolean		simplifiedUserInterface = false;
	Size		actualSize = 0L;
	
	
	// get the users operating mode preference, if possible
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_SimpleTelnet,
								sizeof(simplifiedUserInterface), &simplifiedUserInterface,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		simplifiedUserInterface = false; // assume normal mode, if preference cant be found
	}
	setMenuEnabled(gMenus[(simplifiedUserInterface) ? kMenuPositionSimpleEdit : kMenuPositionEdit], true);
}// EnableEditMenu


/*!
Fills in a menu item information structure for
the Mac OS Color Picker that describes, among
other things, Telnets Edit menu.  This method
is aware of the Color Picker 2.1 version of
this structure.

(3.0)
*/
void
Menus_FillInColorPickerMenuItemInfo			(PickerMenuItemInfo*			outPickerInfo)
{
	Boolean		simpleTelnetMode = false;
	Size		actualSize = 0L;
	
	
	// get the users operating mode preference, if possible
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_SimpleTelnet,
								sizeof(simpleTelnetMode), &simpleTelnetMode,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		simpleTelnetMode = false; // assume normal mode, if preference cant be found
	}
	outPickerInfo->editMenuID = (simpleTelnetMode) ? kMenuIDSimpleEdit : kMenuIDEdit;
	outPickerInfo->cutItem = (simpleTelnetMode) ? iSimpleCut : iCut;
	outPickerInfo->copyItem = (simpleTelnetMode) ? iSimpleCopy : iCopy;
	outPickerInfo->pasteItem = (simpleTelnetMode) ? iSimplePaste : iPaste;
	outPickerInfo->clearItem = (simpleTelnetMode) ? iSimpleClear : iClear;
	outPickerInfo->undoItem = (simpleTelnetMode) ? 0 : iUndo;
}// FillInColorPickerMenuItemInfo


/*!
This method ensures that everything about the menus
(such as item text, item enabled states, item checked
states and menu enabled states) is correct for the
context of the program at the time this method is
invoked.  It is always called just before menu item
selection proceeds (that is, when the user clicks on
the menu bar or invokes a menu key equivalent).

(2.6)
*/
void
Menus_SetUpItemStates		(void)
{
	// 3.0 - Invoke the menu state tracking procedure of any menu item that has one.  See
	// installMenuItemStateTrackers() for a series of method invocations that assign each
	// command its own state-tracking function.
	
	// Apple
	adjustMenuItemByCommandID(kCommandAboutTelnet);
	adjustMenuItemByCommandID(kCommandTelnetOnTheWeb);
	
	// File
	adjustMenuItemByCommandID(kCommandOpenConnection);
	adjustMenuItemByCommandID(kCommandOpenConnectionFavorite);
	adjustMenuItemByCommandID(kCommandCloseConnection);
	adjustMenuItemByCommandID(kCommandLoadSet);
	adjustMenuItemByCommandID(kCommandSaveSet);
	adjustMenuItemByCommandID(kCommandImportMacroSet);
	adjustMenuItemByCommandID(kCommandExportCurrentMacroSet);
	adjustMenuItemByCommandID(kCommandPrint);
	adjustMenuItemByCommandID(kCommandPrintOne);
	adjustMenuItemByCommandID(kCommandPrintScreen);
	adjustMenuItemByCommandID(kCommandPageSetup);
	adjustMenuItemByCommandID(kCommandQuit);
	
	// Edit
	adjustMenuItemByCommandID(kCommandUndo);
	adjustMenuItemByCommandID(kCommandRedo);
	adjustMenuItemByCommandID(kCommandCut);
	adjustMenuItemByCommandID(kCommandCopy);
	adjustMenuItemByCommandID(kCommandCopyTable);
	adjustMenuItemByCommandID(kCommandType);
	adjustMenuItemByCommandID(kCommandPaste);
	adjustMenuItemByCommandID(kCommandClear);
	adjustMenuItemByCommandID(kCommandSelectAll);
	adjustMenuItemByCommandID(kCommandShowClipboard);
	adjustMenuItemByCommandID(kCommandPreferences);
	
	// View
	adjustMenuItemByCommandID(kCommandLargeScreen);
	adjustMenuItemByCommandID(kCommandSmallScreen);
	adjustMenuItemByCommandID(kCommandSetScreenSize);
	adjustMenuItemByCommandID(kCommandBiggerText);
	adjustMenuItemByCommandID(kCommandSmallerText);
	adjustMenuItemByCommandID(kCommandFullScreen);
	adjustMenuItemByCommandID(kCommandFormat);
	
	// Session
	adjustMenuItemByCommandID(kCommandCaptureToFile);
	adjustMenuItemByCommandID(kCommandEcho);
	adjustMenuItemByCommandID(kCommandWrapMode);
	adjustMenuItemByCommandID(kCommandClearScreenSavesLines);
	adjustMenuItemByCommandID(kCommandJumpScrolling);
	adjustMenuItemByCommandID(kCommandTektronix);
	adjustMenuItemByCommandID(kCommandSpeechEnabled);
	adjustMenuItemByCommandID(kCommandResetTerminal);
	
	// Keyboard
	adjustMenuItemByCommandID(kCommandDeletePressSendsBackspace);
	adjustMenuItemByCommandID(kCommandDeletePressSendsDelete);
	adjustMenuItemByCommandID(kCommandSetKeys);
	adjustMenuItemByCommandID(kCommandEMACSArrowMapping);
	adjustMenuItemByCommandID(kCommandLocalPageUpDown);
	adjustMenuItemByCommandID(kCommandTitleTranslationChoices);
	{
		MenuRef			menu = gMenus[kMenuPositionKeyboard];
		MenuItemIndex	itemIndex = 0;
		SInt16			endMarker = 0;
		
		
		endMarker = iFirstTranslationTable + CountMenuItems(menu);
		for (itemIndex = iFirstTranslationTable; itemIndex < endMarker; itemIndex++)
		{
			adjustMenuItem(menu, itemIndex);
		}
	}
	
	// Network
	adjustMenuItemByCommandID(kCommandShowNetworkNumbers);
	adjustMenuItemByCommandID(kCommandSendInternetProtocolNumber);
	adjustMenuItemByCommandID(kCommandSendFTPCommand);
	adjustMenuItemByCommandID(kCommandSendAreYouThere);
	adjustMenuItemByCommandID(kCommandSendAbortOutput);
	adjustMenuItemByCommandID(kCommandSendBreak);
	adjustMenuItemByCommandID(kCommandSendInterruptProcess);
	adjustMenuItemByCommandID(kCommandSendEraseCharacter);
	adjustMenuItemByCommandID(kCommandSendEraseLine);
	adjustMenuItemByCommandID(kCommandSendEndOfFile);
	adjustMenuItemByCommandID(kCommandSendSync);
	adjustMenuItemByCommandID(kCommandWatchForActivity);
	adjustMenuItemByCommandID(kCommandSuspendNetwork);
	
	// Window
	adjustMenuItemByCommandID(kCommandHideFrontWindow);
	adjustMenuItemByCommandID(kCommandHideOtherWindows);
	adjustMenuItemByCommandID(kCommandShowAllHiddenWindows);
	adjustMenuItemByCommandID(kCommandStackWindows);
	adjustMenuItemByCommandID(kCommandNextWindow);
	adjustMenuItemByCommandID(kCommandChangeWindowTitle);
	adjustMenuItemByCommandID(kCommandMinimizeWindow);
	adjustMenuItemByCommandID(kCommandShowSessionInspector);
	adjustMenuItemByCommandID(kCommandShowFTPLog);
	adjustMenuItemByCommandID(kCommandShowCommandLine);
	adjustMenuItemByCommandID(kCommandShowKeypad);
	adjustMenuItemByCommandID(kCommandShowFunction);
	adjustMenuItemByCommandID(kCommandShowControlKeys);
	
	// TEK
	adjustMenuItemByCommandID(kCommandTEKPageCommand);
	adjustMenuItemByCommandID(kCommandTEKPageClearsScreen);
	
	// Internet Resources
	adjustMenuItemByCommandID(kCommandTelnetHomePage);
	adjustMenuItemByCommandID(kCommandTelnetAuthorMail);
}// SetUpItemStates


/*!
To remove from or restore to menu items their key
equivalents, use this method.  At startup time,
this method behaves slightly differently - therefore,
the additional parameter is set to "true" only then.

(3.0)
*/
void
Menus_SetKeyEquivalents	(Boolean const	inMenusHaveKeyEquivalents,
						 Boolean const	inVeryFirstCall)
{
	register short	i = 0;
	static short	lastMenuSetHadCommandKeys = -1;
	Boolean			simpleTelnetMode = false;
	Size			actualSize = 0L;
	
	
	// get the users operating mode preference, if possible
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_SimpleTelnet,
								sizeof(simpleTelnetMode), &simpleTelnetMode,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		simpleTelnetMode = false; // assume normal mode, if preference cant be found
	}
	
	if ((lastMenuSetHadCommandKeys != inMenusHaveKeyEquivalents) || (inVeryFirstCall))
	{
		lastMenuSetHadCommandKeys = inMenusHaveKeyEquivalents;
		
		// Now (3.0) there is a more intelligent key-equivalents disabler!
		// The command keys are removed manually, if needed.
		if (!inMenusHaveKeyEquivalents)
		{
			// Remove key equivalents from all menus that have them!
			removeMenuItemModifiers(gMenus[kMenuPositionFile]);
			removeMenuItemModifiers(gMenus[kMenuPositionEdit]);
			removeMenuItemModifiers(gMenus[kMenuPositionAppearance]);
			removeMenuItemModifiers(gMenus[kMenuPositionSession]);
			removeMenuItemModifiers(gMenus[kMenuPositionKeyboard]);
			removeMenuItemModifiers(gMenus[kMenuPositionNetwork]);
			removeMenuItemModifiers(gMenus[kMenuPositionConnections]);
			removeMenuItemModifiers(gMenus[kMenuPositionSimpleFile]);
			removeMenuItemModifiers(gMenus[kMenuPositionSimpleEdit]);
		}
		else
		{
			// remove and destroy the old menus, because their key equivalents may be gone
			DeleteMenu((simpleTelnetMode) ? kMenuIDSimpleFile : kMenuIDFile);
			DeleteMenu((simpleTelnetMode) ? kMenuIDSimpleEdit : kMenuIDEdit);
			DeleteMenu(kMenuIDAppearance);
			DeleteMenu(kMenuIDSession);
			DeleteMenu(kMenuIDKeyboard);
			DeleteMenu(kMenuIDNetwork);
			DeleteMenu(kMenuIDConnections);
			disposeMainMenus();
			
			// Obtain fresh copies of the killed menus from the
			// resource file.  The disposeMainMenus() and
			// getMainMenus() routines do *not* mess with any
			// dynamic menus (such as the Window menu), because
			// it is too difficult to reconstruct them.  The
			// worst effect this will have on the user is that
			// any key equivalents in such menus are not possible
			// to restore right away, but will show up the next
			// time MacTelnet is launched.
			getMainMenus();
			installMenuItemStateTrackers();
			buildMenuBar();
		}
		
		unless (inVeryFirstCall)
		{
			Menus_SetUpItemStates();
			Menus_CheckSimpleTelnet(simpleTelnetMode); // need to reset this since menus were reloaded
		}
		DrawMenuBar();
		setUpSessionFavoritesSubmenu(gMenus[kMenuPositionOpen]); // rebuild Open Special submenu
	}
}// SetKeyEquivalents


/*!
To make sure that there are as many items in the
New Connection From Favorites submenu as there
are saved session preferences, use this method.

(2.6)
*/
void
CheckOpSpecSubmenu	(void)
{
	UInt16		numItems = 0;
	SInt16		oldResFile = 0;
	
	
	oldResFile = CurResFile();
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	numItems = Count1Resources(SESSIONPREFS_RESTYPE);
	
	if (CountMenuItems(gMenus[kMenuPositionOpen]) != numItems)
	{
		setUpSessionFavoritesSubmenu(gMenus[kMenuPositionOpen]);
	}
	UseResFile(oldResFile);
}// CheckOpSpecSubmenu


/*!
This sets the menus appropriately based on whether or not were
in simplified-user-interface mode.  In simplified-user-interface
mode, there are fewer menu items and fewer command key
equivalents to make the interface easier for novices to use.

(3.0)
*/
void
Menus_CheckSimpleTelnet		(short		inSimpleTelnetOn)
{
	if (inSimpleTelnetOn != 0)
	{
		// simplify menus
		DeleteMenu(kMenuIDSession);
		DeleteMenu(kMenuIDKeyboard);
		DeleteMenu(kMenuIDNetwork);
		DeleteMenu(kMenuIDFile);
		InsertMenu(gMenus[kMenuPositionSimpleFile], kMenuIDEdit);
		DeleteMenu(kMenuIDEdit);
		InsertMenu(gMenus[kMenuPositionSimpleEdit], kMenuIDAppearance);
		DrawMenuBar();
	}
	else
	{
		// change menus back to normal
		DeleteMenu(kMenuIDSimpleFile);
		InsertMenu(gMenus[kMenuPositionFile], kMenuIDSimpleEdit);
		DeleteMenu(kMenuIDSimpleEdit);
		InsertMenu(gMenus[kMenuPositionEdit], kMenuIDAppearance);
		InsertMenu(gMenus[kMenuPositionSession], kMenuIDConnections);
		InsertMenu(gMenus[kMenuPositionKeyboard], kMenuIDConnections);
		InsertMenu(gMenus[kMenuPositionNetwork], kMenuIDConnections);
		DrawMenuBar();
	}
}// CheckSimpleTelnet


/*!
Handles the specified menu and menu item selection.

(2.6)
*/
Boolean
Menus_HandleCommand		(long				inMenuEventResult,
						 EventModifiers		inModifiers)
{
	register short		i = 0;
	short				theItem = 0,
						theMenu = 0;
	Boolean				doWrap = false;
	UInt32				commandID = 0;
	Boolean				result = true;
	
	
	theMenu = inMenuEventResult >> 16;
	theItem = inMenuEventResult & 0xFFFF;
	GetMenuItemCommandID(GetMenuHandle(theMenu), theItem, &commandID);
	
	switch (theMenu)
	{
		case kMenuIDApple:
			if (theItem < 4)
			{
				// MacTelnet items in the Apple menu are handled by command ID...
				(Boolean)Menus_HandleCommandByID(commandID, inModifiers);
			}
			// Carbon handles its own Apple menu items
		#if TARGET_API_MAC_OS8
			else
			{
				Str255	name;
				
				
				GetMenuItemText(gMenus[0], theItem, name); // Apple Menu Item
				OpenDeskAcc(name);
			}
		#endif
			break;
		
		case kMenuIDKeyboard:
			if (theItem < iFirstTranslationTable)
			{
				(Boolean)Menus_HandleCommandByID(commandID, inModifiers);
				break;
			}
			else
			{
				SInt16		index = (theItem - iFirstTranslationTable); // array index in list of translation tables
				
				
				// change the current translation table, translating all text in the terminal window to match
				if (ScreenManager_IsWindowAScreen(GetRealFrontWindow()))
				{
					ConnectionDataPtr		connectionDataPtr = ScreenManager_GetWindowConnectionData(GetRealFrontWindow());
					ScreenTranslationInfo	translation;
					
					
					// translate screen and scrollback buffer to use new table
					translation.oldTable = connectionDataPtr->national;
					translation.newTable = index;
					if (Terminal_ForEveryLineDo(connectionDataPtr->vs, translateScreenLine, &translation) < 0)
					{
						Console_WriteLine("Warning: Problems completing translation update of screen.");
					}
					
					// redraw the screen
					VSredraw(connectionDataPtr->vs, 0, 0, VSmaxwidth(connectionDataPtr->vs),
								VSgetlines(connectionDataPtr->vs) - 1);
					connectionDataPtr->national = index;
				}
			}
	
		case kMenuIDConnections:
			if (theItem < iFirstConnection)
			{
				(Boolean)Menus_HandleCommandByID(commandID, inModifiers);
				break;
			}
			else
			{
				// This code assumes that connection IDs begin at 1!
				TelnetScreenID		id = theItem - iFirstConnection + 1;
				TelnetScreenID		oldScreenID = 0;
				
				
				if (ScreenManager_IsWindowAScreen(GetRealFrontWindow()))
				{
					// current window is a terminal window
					oldScreenID = ScreenManager_GetWindowScreenID(GetRealFrontWindow());
				}
				if (id <= Connections_GetCount())
				{
					ConnectionDataPtr		connectionDataPtr = ScreenManager_GetConnectionDataOf(id);
					
					
					if (connectionDataPtr->status != kConnectionStatusActive)
					{
						displayStatus(id); // show a message detailing whats up with this connection
						break;
					}
					else
					{
						// 3.0 - first make the window visible if it was obscured
						{
							RealScreenDataPtr		realScreenPtr = null;
							
							
							realScreenPtr = ScreenManager_GetRealScreenDataOf(id);
							if (realScreenPtr != null)
							{
								if ((realScreenPtr->window != null) && (realScreenPtr->obscured))
								{
									realScreenPtr->obscured = false;
									ShowWindow(realScreenPtr->window);
									ScreenManager_SetHiddenCount(ScreenManager_GetHiddenCount() - 1);
									SetMenuMarkFromScreen(kMenuMarkTypeEnableIcon, id);
								}
							}
						}
						
						if (!(inModifiers & optionKey)) SelectOverRealFrontWindow(ScreenManager_GetWindowOf(id));
						else HiliteWindow(ScreenManager_GetWindowOf(id), true);
					}
				}
			}
			break;
			
		case kMenuIDScripts:
			// execute script using name of menu item
			executeScriptByMenuEvent(GetMenuHandle(theMenu), theItem, inModifiers);
			break;
			
		case kSubmenuIDOpen:
			Connections_OpenFromMenuItem(gMenus[kMenuPositionOpen], theItem);
			break;
		
		case kSubmenuIDTektronix:
			(Boolean)Menus_HandleCommandByID(commandID, inModifiers);
			break;
		
		case kSubmenuIDTelnetOnTheWeb:
			(Boolean)Menus_HandleCommandByID(commandID, inModifiers);
			break;
		
		case kMenuIDFile:
		case kMenuIDSimpleFile:
		case kMenuIDEdit:		// assume its a Telnet-handled editing command (to hell with
		case kMenuIDSimpleEdit:	// desk accessories, theyre not CARBON-supported)
		case kMenuIDAppearance:
		case kMenuIDSession:
		case kMenuIDNetwork:
			// In the vast majority of cases, PREFER to use
			// command IDs exclusively.  Hopefully, all other
			// cases will be eliminated in future releases.
			(Boolean)Menus_HandleCommandByID(commandID, inModifiers);
			break;
			
		default:
			result = false;
			break;
	
	}
	
	// if a menu item wasnt handled, make this an obvious bug by leaving the menu title highlighted
	if (result) HiliteMenu(0);
	
	return result;
}// HandleCommand


/*!
This method handles menu commands using the new
command IDs assigned in MacTelnet 3.0 for OS 8.
The command ID to use can be found in the file
"MenuResources.h", and the modifiers come from a
standard event record.

(3.0)
*/
Boolean
Menus_HandleCommandByID		(UInt32				inCommandID,
							 EventModifiers		inModifiers)
{
	Boolean		result = false;
	
	
	result = Commands_ExecuteByID(inCommandID, inModifiers);
	return result;
}// HandleCommandByID


/*!
To obtain the item index (1, 2, 3, etc.) for an item in its main
(menu bar) menu based only on its command ID, call this function.

WARNING:	Usually, you call GetMenuAndMenuItemIndexByCommandID()
			instead of this method, so that you can obtain the menu
			that contains a command as well.  For standard items
			(such as "Cut" or "Copy", which are always in the Edit
			menu), it is OK to assume the menu based on a command:
			however, many application commands should not make this
			assumption in order to guarantee compatibility with
			future updates to Telnet.

(3.0)
*/
short
GetMenuItemIndexByCommandID		(UInt32 const	inCommandID)
{
	MenuRef		dummy = null;
	
	
	return GetMenuAndMenuItemIndexByCommandID(inCommandID, &dummy);
}// GetMenuItemIndexByCommandID


/*!
To obtain the main menu and item index for an item based only on
its command ID, call this function.  If the specified command
cannot be matched with an item index, -1 is returned.  If the
specified command cannot be matched with a menu, null is returned
via the specified menu handle pointer.

For special circumstances where the menu containing a command is
known and is very unlikely to change in the future, you can use
the similar method GetMenuItemIndexByCommandID(), which does not
return a menu handle (but only the menu item number).

IMPORTANT:	There are several utility functions in this file that
			automatically use the power of this routine to get
			and set various attributes of menu items (such as
			item text, item enabled state and checkmarks).  If
			you think this method is useful, be sure to first
			check if there is not an even more useful version that
			directly performs a command-ID-based function that you
			need.

WARNING:	This method will only return the menu handle and item
			number for the command appropriate for the current
			mode (i.e. simplified-user-interface or otherwise).
			To ensure you change all versions of a command, you
			should use the GetMenusAndMenuItemIndicesByCommandID()
			variation of this method.  If you use the convenience
			functions in this module (SetMenuItemTextByCommandID(),
			for example), you avoid this potential problem.

(3.0)
*/
MenuItemIndex
GetMenuAndMenuItemIndexByCommandID	(UInt32 const	inCommandID,
									 MenuRef*		outMenu)
{
	MenuRef			simpleTelnetMenu = null;
	MenuItemIndex	simpleTelnetMenuItemIndex = 0;
	MenuItemIndex	result = 0;
	Boolean			simpleTelnetMode = false;
	Size			actualSize = 0L;
	
	
	// get the users operating mode preference, if possible
	unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_SimpleTelnet,
								sizeof(simpleTelnetMode), &simpleTelnetMode,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
	{
		simpleTelnetMode = false; // assume normal mode, if preference cant be found
	}
	
	GetMenusAndMenuItemIndicesByCommandID(inCommandID, outMenu, &simpleTelnetMenu, &result, &simpleTelnetMenuItemIndex);
	
	// If the user is in simplified-user-interface mode, then its the
	// secondary menus that are potentially important.  If the secondary
	// menu is null, however, then the primary menu is also the menu to
	// use while in simplified-user-interface mode.
	if (simpleTelnetMode && (simpleTelnetMenu != null))
	{
		*outMenu = simpleTelnetMenu;
		result = simpleTelnetMenuItemIndex;
	}
	return result;
}// GetMenuAndMenuItemIndexByCommandID


/*!
To obtain the two menus and item indices for an item based only on
its command ID, call this function.  If the specified command
cannot be matched with an item index, -1 is returned in the item
pointers.  If the specified command cannot be matched with a menu,
null is returned via the specified menu handle pointers.

The primary menu specifications refer to the menu in normal mode,
and the secondary menu specifications refer to the menu in
simplified-user-interface mode.  Some menus do not have versions
for the simplified-user-interface mode, so the primary and secondary
menus are identical in those cases.  Also, sometimes an item present
in a primary menu will not be present in a secondary menu.  -1 is
returned for the item index in this case, and the Mac OS Menu Manager
should safely ignore any attempts to perform menu operations with
this index.

IMPORTANT:	This is the only place in the entire program where menu
			commands are physically tied to menus and menu item
			indices.  DO NOT DEFEAT THE PURPOSE OF THIS POWER by
			placing similar mappings ANY PLACE ELSE.
			
			There are several utility functions in this file that
			automatically use the power of this routine to get and
			set various attributes of menu items (such as item
			text, item enabled state and checkmarks) using only a
			command ID.  If you think this method is useful, be
			sure to first check if there is an even more useful
			version that directly performs a command-ID-based
			function that you need.  In particular, you avoid the
			hassle of working with the primary and secondary menus
			if you use one of the utility functions, because the
			utility functions automatically update both menus.

(3.0)
*/
void
GetMenusAndMenuItemIndicesByCommandID	(UInt32 const		inCommandID,
										 MenuRef*			outPrimaryMenu,
										 MenuRef*			outSecondaryMenu,
										 MenuItemIndex*		outPrimaryItemIndex,
										 MenuItemIndex*		outSecondaryItemIndex)
{
	MenuItemIndex	menuPosition1 = 0,
					menuPosition2 = 0; // not set yet
	SInt16			itemIndex1 = -1,
					itemIndex2 = -1; // not set yet
	
	
	// determine the menu that contains the command, and its item number
	switch (inCommandID)
	{
		// Apple menu commands
		
		case kCommandAboutTelnet:
			menuPosition1 = menuPosition2 = kMenuPositionApple;
			itemIndex1 = itemIndex2 = iAboutTelnet;
			break;
		
		case kCommandTelnetOnTheWeb:
			menuPosition1 = menuPosition2 = kMenuPositionApple;
			itemIndex1 = itemIndex2 = iTelnetOnTheWeb;
			break;
		
		// File menu commands
		
		case kCommandOpenConnection:
			menuPosition1 = kMenuPositionFile;
			menuPosition2 = kMenuPositionSimpleFile;
			itemIndex1 = iOpenConnection;
			itemIndex2 = iSimpleOpenConnection;
			break;
		
		case kCommandOpenConnectionFavorite:
			menuPosition1 = kMenuPositionFile;
			menuPosition2 = kMenuPositionSimpleFile;
			itemIndex1 = iOpenConnectionFavorite;
			itemIndex2 = iSimpleOpenConnectionFavorite;
			break;
		
		case kCommandLoadSet:
			menuPosition1 = menuPosition2 = kMenuPositionFile;
			itemIndex1 = itemIndex2 = iLoadSet;
			break;
		
		case kCommandCloseConnection:
			menuPosition1 = kMenuPositionFile;
			menuPosition2 = kMenuPositionSimpleFile;
			itemIndex1 = iCloseConnection;
			itemIndex2 = iSimpleCloseConnection;
			break;
		
		case kCommandSaveSet:
			menuPosition1 = menuPosition2 = kMenuPositionFile;
			itemIndex1 = itemIndex2 = iSaveSet;
			break;
		
		case kCommandImportMacroSet:
			menuPosition1 = menuPosition2 = kMenuPositionFile;
			itemIndex1 = itemIndex2 = iImportMacroSet;
			break;
		
		case kCommandExportCurrentMacroSet:
			menuPosition1 = menuPosition2 = kMenuPositionFile;
			itemIndex1 = itemIndex2 = iExportCurrentMacroSet;
			break;
		
		case kCommandPrint:
			menuPosition1 = kMenuPositionFile;
			menuPosition2 = kMenuPositionSimpleFile;
			itemIndex1 = iPrint;
			itemIndex2 = iSimplePrint;
			break;
		
		case kCommandPrintOne:
			menuPosition1 = menuPosition2 = kMenuPositionFile;
			itemIndex1 = itemIndex2 = iPrintOne;
			break;
		
		case kCommandPrintScreen:
			menuPosition1 = menuPosition2 = kMenuPositionFile;
			itemIndex1 = itemIndex2 = iPrintScreen;
			break;
		
		case kCommandPageSetup:
			menuPosition1 = kMenuPositionFile;
			menuPosition2 = kMenuPositionSimpleFile;
			itemIndex1 = iPageSetup;
			itemIndex2 = iSimplePageSetup;
			break;
		
		case kCommandQuit:
			menuPosition1 = kMenuPositionFile;
			menuPosition2 = kMenuPositionSimpleFile;
			itemIndex1 = iQuit;
			itemIndex2 = iSimpleQuit;
			break;
		
		// Edit menu commands
		
		case kCommandUndo:
			menuPosition1 = menuPosition2 = kMenuPositionEdit;
			itemIndex1 = iUndo;
			break;
		
		case kCommandRedo:
			menuPosition1 = menuPosition2 = kMenuPositionEdit;
			itemIndex1 = iRedo;
			break;
		
		case kCommandCut:
			menuPosition1 = kMenuPositionEdit;
			menuPosition2 = kMenuPositionSimpleEdit;
			itemIndex1 = iCut;
			itemIndex2 = iSimpleCut;
			break;
		
		case kCommandCopy:
			menuPosition1 = kMenuPositionEdit;
			menuPosition2 = kMenuPositionSimpleEdit;
			itemIndex1 = iCopy;
			itemIndex2 = iSimpleCopy;
			break;
		
		case kCommandCopyTable:
			menuPosition1 = menuPosition2 = kMenuPositionEdit;
			itemIndex1 = iCopyTable;
			break;
		
		case kCommandPaste:
			menuPosition1 = kMenuPositionEdit;
			menuPosition2 = kMenuPositionSimpleEdit;
			itemIndex1 = iPaste;
			itemIndex2 = iSimplePaste;
			break;
		
		case kCommandType:
			menuPosition1 = menuPosition2 = kMenuPositionEdit;
			itemIndex1 = iType;
			break;
		
		case kCommandClear:
			menuPosition1 = kMenuPositionEdit;
			menuPosition2 = kMenuPositionSimpleEdit;
			itemIndex1 = iClear;
			itemIndex2 = iSimpleClear;
			break;
		
		case kCommandSelectAll:
			menuPosition1 = menuPosition2 = kMenuPositionEdit;
			itemIndex1 = iSelectAll;
			break;
		
		case kCommandShowClipboard:
			menuPosition1 = menuPosition2 = kMenuPositionEdit;
			itemIndex1 = iShowClipboard;
			break;
		
		case kCommandPreferences:
			menuPosition1 = kMenuPositionEdit;
			menuPosition2 = kMenuPositionSimpleEdit;
			itemIndex1 = iPreferences;
			itemIndex2 = iSimplePreferences;
			break;
		
		// View menu commands
		
		case kCommandLargeScreen:
			menuPosition1 = menuPosition2 = kMenuPositionAppearance;
			itemIndex1 = itemIndex2 = iLargeScreen;
			break;
		
		case kCommandSmallScreen:
			menuPosition1 = menuPosition2 = kMenuPositionAppearance;
			itemIndex1 = itemIndex2 = iSmallScreen;
			break;
		
		case kCommandSetScreenSize:
			menuPosition1 = menuPosition2 = kMenuPositionAppearance;
			itemIndex1 = itemIndex2 = iSetScreenSize;
			break;
		
		case kCommandBiggerText:
			menuPosition1 = menuPosition2 = kMenuPositionAppearance;
			itemIndex1 = itemIndex2 = iBiggerText;
			break;
		
		case kCommandSmallerText:
			menuPosition1 = menuPosition2 = kMenuPositionAppearance;
			itemIndex1 = itemIndex2 = iSmallerText;
			break;
		
		case kCommandFullScreen:
			menuPosition1 = menuPosition2 = kMenuPositionAppearance;
			itemIndex1 = itemIndex2 = iFullScreen;
			break;
		
		case kCommandFormat:
			menuPosition1 = menuPosition2 = kMenuPositionAppearance;
			itemIndex1 = itemIndex2 = iFormat;
			break;
		
		// Session menu commands
		
		case kCommandCaptureToFile:
			menuPosition1 = menuPosition2 = kMenuPositionSession;
			itemIndex1 = itemIndex2 = iCaptureToFile;
			break;
		
		case kCommandEcho:
			menuPosition1 = menuPosition2 = kMenuPositionSession;
			itemIndex1 = itemIndex2 = iEcho;
			break;
		
		case kCommandWrapMode:
			menuPosition1 = menuPosition2 = kMenuPositionSession;
			itemIndex1 = itemIndex2 = iWrapMode;
			break;
		
		case kCommandClearScreenSavesLines:
			menuPosition1 = menuPosition2 = kMenuPositionSession;
			itemIndex1 = itemIndex2 = iClearScreenSavesLines;
			break;
		
		case kCommandJumpScrolling:
			menuPosition1 = menuPosition2 = kMenuPositionSession;
			itemIndex1 = itemIndex2 = iJumpScrolling;
			break;
		
		case kCommandTektronix:
			menuPosition1 = menuPosition2 = kMenuPositionSession;
			itemIndex1 = itemIndex2 = iTektronix;
			break;
		
		case kCommandSpeechEnabled:
			menuPosition1 = menuPosition2 = kMenuPositionSession;
			itemIndex1 = itemIndex2 = iSpeechEnabled;
			break;
		
		case kCommandResetTerminal:
			menuPosition1 = menuPosition2 = kMenuPositionSession;
			itemIndex1 = itemIndex2 = iResetTerminal;
			break;
		
		// Keyboard menu commands
		
		case kCommandDeletePressSendsBackspace:
			menuPosition1 = menuPosition2 = kMenuPositionKeyboard;
			itemIndex1 = itemIndex2 = iDeletePressSendsBackspace;
			break;
		
		case kCommandDeletePressSendsDelete:
			menuPosition1 = menuPosition2 = kMenuPositionKeyboard;
			itemIndex1 = itemIndex2 = iDeletePressSendsDelete;
			break;
		
		case kCommandEMACSArrowMapping:
			menuPosition1 = menuPosition2 = kMenuPositionKeyboard;
			itemIndex1 = itemIndex2 = iEMACSArrowMapping;
			break;
		
		case kCommandLocalPageUpDown:
			menuPosition1 = menuPosition2 = kMenuPositionKeyboard;
			itemIndex1 = itemIndex2 = iLocalPageUpDown;
			break;
		
		case kCommandSetKeys:
			menuPosition1 = menuPosition2 = kMenuPositionKeyboard;
			itemIndex1 = itemIndex2 = iSetKeys;
			break;
		
		case kCommandTitleTranslationChoices:
			menuPosition1 = menuPosition2 = kMenuPositionKeyboard;
			itemIndex1 = itemIndex2 = iTitleTranslationChoices;
			break;
		
		// Network menu commands
		
		case kCommandShowNetworkNumbers:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iShowNetworkNumbers;
			break;
		
		case kCommandSendInternetProtocolNumber:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendInternetProtocolNumber;
			break;
		
		case kCommandSendFTPCommand:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendFTPCommand;
			break;
		
		case kCommandSendAreYouThere:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendAreYouThere;
			break;
		
		case kCommandSendAbortOutput:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendAbortOutput;
			break;
		
		case kCommandSendBreak:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendBreak;
			break;
		
		case kCommandSendInterruptProcess:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendInterruptProcess;
			break;
		
		case kCommandSendEraseCharacter:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendEraseCharacter;
			break;
		
		case kCommandSendEraseLine:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendEraseLine;
			break;
		
		case kCommandSendEndOfFile:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendEndOfFile;
			break;
		
		case kCommandSendSync:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSendSync;
			break;
		
		case kCommandSuspendNetwork:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iSuspendNetwork;
			break;
		
		case kCommandWatchForActivity:
			menuPosition1 = menuPosition2 = kMenuPositionNetwork;
			itemIndex1 = itemIndex2 = iCommandWatchForActivity;
			break;
		
		// Window menu commands
		
		case kCommandHideFrontWindow:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iHideFrontWindow;
			break;
		
		case kCommandHideOtherWindows:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iHideOtherWindows;
			break;
		
		case kCommandShowAllHiddenWindows:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iShowAllHiddenWindows;
			break;
		
		case kCommandStackWindows:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iStackWindows;
			break;
		
		case kCommandNextWindow:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iNextWindow;
			break;
		
		case kCommandChangeWindowTitle:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iChangeWindowTitle;
			break;
		
		case kCommandMinimizeWindow:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iMinimizeWindow;
			break;
		
		case kCommandShowSessionInspector:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iShowSessionInspector;
			break;
		
		case kCommandShowFTPLog:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iShowFTPLog;
			break;
		
		case kCommandShowCommandLine:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iShowCommandLine;
			break;
		
		case kCommandShowControlKeys:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iShowControlKeys;
			break;
		
		case kCommandShowFunction:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iShowFunction;
			break;
		
		case kCommandShowKeypad:
			menuPosition1 = menuPosition2 = kMenuPositionConnections;
			itemIndex1 = itemIndex2 = iShowKeypad;
			break;
		
		// Tektronix submenu commands
		
		case kCommandTEKPageCommand:
			menuPosition1 = menuPosition2 = kMenuPositionTektronix;
			itemIndex1 = itemIndex2 = iTEKPageCommand;
			break;
		
		case kCommandTEKPageClearsScreen:
			menuPosition1 = menuPosition2 = kMenuPositionTektronix;
			itemIndex1 = itemIndex2 = iTEKPageClearsScreen;
			break;
		
		// Internet Resources submenu commands
		
		case kCommandTelnetHomePage:
			menuPosition1 = menuPosition2 = kMenuPositionTelnetOnTheWeb;
			itemIndex1 = itemIndex2 = iTelnetHomePage;
			break;
		
		case kCommandTelnetAuthorMail:
			menuPosition1 = menuPosition2 = kMenuPositionTelnetOnTheWeb;
			itemIndex1 = itemIndex2 = iTelnetAuthorMail;
			break;
		
		// no match?
		
		default:
			break;
	}
	
	if ((outPrimaryMenu != null) && (menuPosition1 >= 0) && (menuPosition1 < kMenuResourceCount))
	{
		*outPrimaryMenu = gMenus[menuPosition1];
	}
	else if (outPrimaryMenu != null) *outPrimaryMenu = null;
	
	if ((outSecondaryMenu != null) && (menuPosition2 >= 0) && (menuPosition2 < kMenuResourceCount))
	{
		*outSecondaryMenu = gMenus[menuPosition2];
	}
	else if (outSecondaryMenu != null) *outSecondaryMenu = null;
	
	if (outPrimaryItemIndex != null) *outPrimaryItemIndex = itemIndex1;
	if (outSecondaryItemIndex != null) *outSecondaryItemIndex = itemIndex2;
}// GetMenusAndMenuItemIndicesByCommandID


/*!
To determine whether a checkmark is next to a menu item
in a menu bar menu knowing only its command ID, call
this method.  If the command specified cannot be
found in any menu, this method does nothing and returns
a value of "false".

This function has not been tested, and may not do what
it claims to do.

(3.0)
*/
Boolean
GetMenuItemCheckedByCommandID	(UInt32 const		inCommandID)
{
	MenuRef			menuHandle = null;
	SInt16			itemIndex = -1;
	Boolean			result = false;
	CharParameter	mark = '\0';
	
	
	itemIndex = GetMenuAndMenuItemIndexByCommandID(inCommandID, &menuHandle);
	if (menuHandle != null)
	{
		GetItemMark(menuHandle, itemIndex, &mark);
		result = (mark == checkMark);
	}
	
	return result;
}// GetMenuItemCheckedByCommandID


/*!
To determine the enabled or disabled (dimmed) state of a
menu item in a menu bar menu knowing only its command ID,
call this method.  If the command specified cannot be
found in any menu, this method does nothing and returns
a value of "false".

IMPORTANT:	All commands must be associated with a
			MenuCommandStateTrackerProcPtr callback, or
			this routine will return "false" for any
			item that does not have one.  Telnet always
			determines item states dynamically, not
			statically, so this routine will not read
			the current enabled state of the item.
			Rather, it invokes the commands callback to
			determine what the item state *should* be,
			which ultimately is what youre interested
			in, right???

(3.0)
*/
Boolean
GetMenuItemEnabledByCommandID	(UInt32 const		inCommandID)
{
	MenuRef							menu1 = null,
									menu2 = null;
	MenuItemIndex					itemIndex1 = 0,
									itemIndex2 = 0;
	MenuCommandStateTrackerProcPtr	proc = null;
	Boolean							result = false;
	
	
	GetMenusAndMenuItemIndicesByCommandID(inCommandID, &menu1, &menu2, &itemIndex1, &itemIndex2);
	if ((menu1 != null) && (itemIndex1 > 0)) // only look at the primary menu (both should be the same anyway)
	{
		getMenuItemAdjustmentProc(menu1, itemIndex1, &proc);
		if (proc != null) result = InvokeMenuCommandStateTrackerProc(proc, inCommandID, menu1, itemIndex1);
	}
	
	return result;
}// GetMenuItemEnabledByCommandID


/*!
To determine the symbols that appear next to the key
equivalent for a menu item in a menu bar menu knowing
only its command ID, call this method.  The presence
of certain symbols also affects the required modifier
keys to invoke the item (Option or Control, for
example).  If the command specified cannot be found
in any menu, this method does nothing.

This is the best way to determine key equivalent symbols
next to menu items in MacTelnet 3.0.  It is completely
independent of menu implementation, and will always
modify the correct item even if a menu command is
relocated within a menu or moved to a different menu
altogether.  Do not defeat the purpose of the new,
centralized menu management system in MacTelnet 3.0!

(3.0)
*/
UInt8
GetMenuItemModifiersByCommandID	(UInt32 const		inCommandID)
{
	MenuRef		menuHandle = null;
	SInt16		itemIndex = -1;
	UInt8		result = 0;
	
	
	itemIndex = GetMenuAndMenuItemIndexByCommandID(inCommandID, &menuHandle);
	if (menuHandle != null) GetMenuItemModifiers(menuHandle, itemIndex, &result);
	return result;
}// GetMenuItemModifiersByCommandID


/*!
To determine the name of a menu item in a menu bar menu
knowing only its command ID, call this method.  If the
command specified cannot be found in any menu, this method
does nothing and the returned text is invalid.

This is the best way to determine the menu item text for
menu items in MacTelnet 3.0.  It is completely independent
of menu implementation, and will always return the correct
item text even if a menu command is relocated within a menu
or moved to a different menu altogether.  The contextual
menu builder, for example, relies heavily on the power of
routines like this.  Do not defeat the purpose of the new,
centralized menu management system in MacTelnet 3.0!

(3.0)
*/
void
GetMenuItemTextByCommandID	(UInt32 const		inCommandID,
							 Str255				outText)
{
	MenuRef		menuHandle = null;
	SInt16		itemIndex = -1;
	
	
	itemIndex = GetMenuAndMenuItemIndexByCommandID(inCommandID, &menuHandle);
	if (menuHandle != null) GetMenuItemText(menuHandle, itemIndex, outText);
}// GetMenuItemTextByCommandID


/*!
To associate a menu item with a function specifically
designed to report what its proper state should be, use
this method.  The menu items reference constant is used
to store the information.

See also SetMenuItemAdjustmentProcByCommandID().

(3.0)
*/
void
SetMenuItemAdjustmentProc	(MenuRef							inMenu,
							 MenuItemIndex const				inItemIndex,
							 MenuCommandStateTrackerProcPtr		inProc)
{
	if ((inMenu != null) && (inItemIndex > 0)) SetMenuItemRefCon(inMenu, inItemIndex, (UInt32)inProc);
}// SetMenuItemAdjustmentProc


/*!
To associate a menu command with a function specifically
designed to report what its proper state should be, use
this method.  The menu items reference constant is used
to store the information.

This is the best way to set or remove a callback from a
menu item in MacTelnet 3.0.  It is completely independent
of menu implementation, and will always modify the correct
item even if a menu command is relocated within a menu or
moved to a different menu altogether.  Do not defeat the
purpose of the new, centralized menu management system in
MacTelnet 3.0!

(3.0)
*/
void
SetMenuItemAdjustmentProcByCommandID	(UInt32 const						inCommandID,
										 MenuCommandStateTrackerProcPtr		inProc)
{
	MenuRef			menuHandle1 = null,
					menuHandle2 = null;
	MenuItemIndex	itemIndex1 = 0,
					itemIndex2 = 0;
	
	
	GetMenusAndMenuItemIndicesByCommandID(inCommandID, &menuHandle1, &menuHandle2, &itemIndex1, &itemIndex2);
	if ((menuHandle1 != null) && (itemIndex1 > 0)) SetMenuItemRefCon(menuHandle1, itemIndex1, (UInt32)inProc);
	if ((menuHandle2 != null) && (itemIndex2 > 0)) SetMenuItemRefCon(menuHandle2, itemIndex2, (UInt32)inProc);
}// SetMenuItemAdjustmentProcByCommandID


/*!
To place or remove a checkmark next to a menu item
in a menu bar menu knowing only its command ID, call
this method.  If the command specified cannot be
found in any menu, this method does nothing.

This is the best way to place or remove a checkmark
next to menu items in MacTelnet 3.0.  It is completely
independent of menu implementation, and will always
modify the correct item even if a menu command is
relocated within a menu or moved to a different menu
altogether.  Do not defeat the purpose of the new,
centralized menu management system in MacTelnet 3.0!

(3.0)
*/
void
SetMenuItemCheckedByCommandID	(UInt32 const		inCommandID,
								 Boolean const		inIsChecked)
{
	MenuRef			menuHandle1 = null,
					menuHandle2 = null;
	MenuItemIndex	itemIndex1 = 0,
					itemIndex2 = 0;
	
	
	GetMenusAndMenuItemIndicesByCommandID(inCommandID, &menuHandle1, &menuHandle2, &itemIndex1, &itemIndex2);
	if ((menuHandle1 != null) && (itemIndex1 > 0)) CheckMenuItem(menuHandle1, itemIndex1, inIsChecked);
	if ((menuHandle2 != null) && (itemIndex2 > 0)) CheckMenuItem(menuHandle2, itemIndex2, inIsChecked);
}// SetMenuItemCheckedByCommandID


/*!
To enable or disable (dim) a menu item in a menu bar
menu knowing only its command ID, call this method.
If the command specified cannot be found in any menu,
this method does nothing.

This is the best way to set the enabled state of
menu items in MacTelnet 3.0.  It is completely
independent of menu implementation, and will always
modify the correct item even if a menu command is
relocated within a menu or moved to a different menu
altogether.  Do not defeat the purpose of the new,
centralized menu management system in MacTelnet 3.0!

(3.0)
*/
void
SetMenuItemEnabledByCommandID	(UInt32 const		inCommandID,
								 Boolean const		inIsEnabled)
{
	MenuRef			menuHandle1 = null,
					menuHandle2 = null;
	MenuItemIndex	itemIndex1 = 0,
					itemIndex2 = 0;
	
	
	GetMenusAndMenuItemIndicesByCommandID(inCommandID, &menuHandle1, &menuHandle2, &itemIndex1, &itemIndex2);
	if ((menuHandle1 != null) && (itemIndex1 > 0))
	{
		setMenuItemEnabled(menuHandle1, itemIndex1, inIsEnabled);
	}
	if ((menuHandle2 != null) && (itemIndex2 > 0))
	{
		setMenuItemEnabled(menuHandle2, itemIndex2, inIsEnabled);
	}
}// SetMenuItemEnabledByCommandID


/*!
To specify the symbols that appear next to the key
equivalent for a menu item in a menu bar menu knowing
only its command ID, call this method.  The presence
of certain symbols also affects the required modifier
keys to invoke the item (Option or Control, for
example).  If the command specified cannot be found
in any menu, this method does nothing.

This is the best way to modify key equivalent symbols
next to menu items in MacTelnet 3.0.  It is completely
independent of menu implementation, and will always
modify the correct item even if a menu command is
relocated within a menu or moved to a different menu
altogether.  Do not defeat the purpose of the new,
centralized menu management system in MacTelnet 3.0!

(3.0)
*/
void
SetMenuItemModifiersByCommandID	(UInt32 const		inCommandID,
								 UInt8 const		inNewModifiers)
{
	MenuRef			menuHandle1 = null,
					menuHandle2 = null;
	MenuItemIndex	itemIndex1 = 0,
					itemIndex2 = 0;
	
	
	GetMenusAndMenuItemIndicesByCommandID(inCommandID, &menuHandle1, &menuHandle2, &itemIndex1, &itemIndex2);
	if ((menuHandle1 != null) && (itemIndex1 > 0)) SetMenuItemModifiers(menuHandle1, itemIndex1, inNewModifiers);
	if ((menuHandle2 != null) && (itemIndex2 > 0)) SetMenuItemModifiers(menuHandle2, itemIndex2, inNewModifiers);
}// SetMenuItemModifiersByCommandID


/*!
To specify the name of a menu item in a menu bar menu
knowing only its command ID, call this method.  If the
command specified cannot be found in any menu, this method
does nothing.

This is the best way to set the text of menu items in
MacTelnet 3.0.  It is completely independent of menu
implementation, and will always modify the correct item
even if a menu command is relocated within a menu or moved
to a different menu altogether.  Do not defeat the purpose
of the new, centralized menu management system in
MacTelnet 3.0!

(3.0)
*/
void
SetMenuItemTextByCommandID	(UInt32 const		inCommandID,
							 ConstStr255Param	inText)
{
	MenuRef			menuHandle1 = null,
					menuHandle2 = null;
	MenuItemIndex	itemIndex1 = 0,
					itemIndex2 = 0;
	
	
	GetMenusAndMenuItemIndicesByCommandID(inCommandID, &menuHandle1, &menuHandle2, &itemIndex1, &itemIndex2);
	if ((menuHandle1 != null) && (itemIndex1 > 0)) SetMenuItemText(menuHandle1, itemIndex1, inText);
	if ((menuHandle2 != null) && (itemIndex2 > 0)) SetMenuItemText(menuHandle2, itemIndex2, inText);
}// SetMenuItemTextByCommandID


/*!
To remove a connections name from the Window menu
by using its screen ID, use this method.

(2.6)
*/
void
Menus_RemoveConnectionsMenuItem		(ConnectionDataPtr		inConnectionDataPtr)
{
	if (inConnectionDataPtr != null)
	{
		MenuItemIndex		index = getWindowMenuItemIndexByScreenID(inConnectionDataPtr->vs);
		
		
		// make absolutely sure that menus should be screwed with
		if ((index >= iFirstConnection) && (inConnectionDataPtr->addedMenuItem))
		{
			DeleteMenuItem(gMenus[kMenuPositionConnections], index);
			gNumberOfConnectionItemsAdded--;
			
			// remove the divider in the event that all connections are gone
			unless (gNumberOfConnectionItemsAdded) DeleteMenuItem(gMenus[kMenuPositionConnections], iFirstConnection - 1);
			
			installDynamicMenuItemStateTrackers();
			Menus_SetUpItemStates();
		}
	}
}// RemoveScreenNameByID


/*!
Use this method to set the item mark for the menu item
representing the specified screen so the mark reflects
the connection state.  In MacTelnet 3.0, connection
menu items are now marked with icons.  Consequently,
their use requires allocation of memory.  At shutdown
time, call this method with a menu mark equal to
"kMenuMarkTypeDispose" to dispose of any memory that
this routine might allocate.  (This is done automatically
by Menus_Done().)

The status icon in a windows header, if existent, is
automatically updated by this routine as well.

(3.0)
*/
void
SetMenuMarkFromScreen	(TelnetMenuMarkType		inMarkType,
						 TelnetScreenID const	inScreenID)
{
	static IconSuiteRef		liveConnectionIcons = null,
							openingConnectionIcons = null,
							deadConnectionIcons = null;
	OSErr					error = noErr;
	Boolean					doSet = false;
	IconSuiteRef			icons = null;
	MenuRef					windowMenu = gMenus[kMenuPositionConnections];
	MenuItemIndex			windowMenuItemIndex = getWindowMenuItemIndexByScreenID(inScreenID);
	
	
	// set up icons as necessary
	switch (inMarkType)
	{
		case kMenuMarkTypeConnectionLive:
			if (liveConnectionIcons == null)
			{
				error = GetIconSuite(&liveConnectionIcons, kIconSuiteConnectionActive, kSelectorAllSmallData);
				doSet = (error == noErr);
				unless (doSet) liveConnectionIcons = null;
			}
			else doSet = true;
			icons = liveConnectionIcons;
			break;
		
		case kMenuMarkTypeConnectionOpening:
			if (openingConnectionIcons == null)
			{
				error = GetIconSuite(&openingConnectionIcons, kIconSuiteConnectionOpening, kSelectorAllSmallData);
				doSet = (error == noErr);
				unless (doSet) openingConnectionIcons = null;
			}
			else doSet = true;
			icons = openingConnectionIcons;
			break;
		
		case kMenuMarkTypeConnectionDead:
			if (deadConnectionIcons == null)
			{
				error = GetIconSuite(&deadConnectionIcons, kIconSuiteConnectionDead, kSelectorAllSmallData);
				doSet = (error == noErr);
				unless (doSet) deadConnectionIcons = null;
			}
			else doSet = true;
			icons = deadConnectionIcons;
			break;
		
		case kMenuMarkTypeDispose:
		case kMenuMarkTypeDisableIcon:
		case kMenuMarkTypeEnableIcon:
		default:
			break;
	}
	
	if (inMarkType == kMenuMarkTypeDispose)
	{
		// destroy allocated icons
		Console_WriteLine("disposing live-connection icons...");
		if (liveConnectionIcons != null) (OSErr)DisposeIconSuite(liveConnectionIcons, false), liveConnectionIcons = null;
		Console_WriteLine("disposing opening-connection icons...");
		if (openingConnectionIcons != null) (OSErr)DisposeIconSuite(openingConnectionIcons, false), openingConnectionIcons = null;
		Console_WriteLine("disposing dead-connection icons...");
		if (deadConnectionIcons != null) (OSErr)DisposeIconSuite(deadConnectionIcons, false), deadConnectionIcons = null;
		Console_WriteLine("...done");
	}
	else if (inMarkType == kMenuMarkTypeEnableIcon)
	{
		// set the style of the item to normal, undoing the style setting of "kMenuMarkTypeDisableIcon"
		SetItemStyle(windowMenu, windowMenuItemIndex, normal);
		
		// enabling the icon itself is only possible with Mac OS 8.5 or later
		if (FlagManager_Test(kFlagCarbonMenuManager)) EnableMenuItemIcon(windowMenu, windowMenuItemIndex);
	}
	else if (inMarkType == kMenuMarkTypeDisableIcon)
	{
		// set the style of the item to italic, which makes it more obvious that a window is hidden
		SetItemStyle(windowMenu, windowMenuItemIndex, italic);
		
		// disabling the icon itself is only possible with Mac OS 8.5 or later
		if (FlagManager_Test(kFlagCarbonMenuManager)) DisableMenuItemIcon(windowMenu, windowMenuItemIndex);
	}
	else
	{
		if (doSet)
		{
			RealScreenDataPtr		realScreenPtr = ScreenManager_GetRealScreenDataOf(inScreenID);
			ConnectionDataPtr		connectionDataPtr = ScreenManager_GetConnectionDataOf(inScreenID);
			
			
			// set icon of Window menu item
			(OSErr)SetMenuItemIconHandle(windowMenu, windowMenuItemIndex, kMenuIconSuiteType, icons);
			
			// set status icon and header status text of the window
			if (realScreenPtr != null)
			{
				Str255		statusText;
				
				
				StatusIcon_SetNextCell(realScreenPtr->connectionStatusIcon,
										(inMarkType == kMenuMarkTypeConnectionDead) ? 1 : 0);
				
				// set status text
				if (inMarkType == kMenuMarkTypeConnectionDead)
				{
					GetIndString(statusText, rStringsStatus, siStatusNotConnected);
					StringUtilities_PMetaSetDefaults();
					StringUtilities_PBuild(statusText, connectionDataPtr->host.name, EMPTY_PSTRING,
											EMPTY_PSTRING, EMPTY_PSTRING);
				}
				else if (inMarkType == kMenuMarkTypeConnectionLive)
				{
					Str32		timeString,
								dayString;
					UInt32		now = 0L;
					
					
					GetDateTime(&now);
					GetIndString(statusText, rStringsStatus, siStatusConnectedToHostSinceTime);
					TimeString(connectionDataPtr->connectedSince, false/* want seconds */,
								timeString, null/* international data handle */);
					DateString(connectionDataPtr->connectedSince, longDate/* date format */,
								dayString, null/* international data handle */);
					StringUtilities_PMetaSetDefaults();
					StringUtilities_PBuild(statusText, connectionDataPtr->host.name, timeString,
											dayString, EMPTY_PSTRING);
				}
				else
				{
					// ???
					PLstrcpy(statusText, EMPTY_PSTRING);
				}
				TerminalWindowHeader_SetStatusText(realScreenPtr->windowHeader, statusText);
			}
			
			// as additional feedback for dead connections, cease any cursor flashing
			// and make the cursor always visible (temporary - this is a weird place
			// to do something like this, but hey, it works, so for now...)
			//if (inMarkType == kMenuMarkTypeConnectionDead)
			//{
			//	RealScreen_SetCursorVisibility(inScreenID, true/* visible */);
			//}
		}
	}
}// SetMenuMarkFromScreen


/*!
Use this method to set the item text for the menu item
representing the specified screen.

(3.0)
*/
void
SetMenuTextFromScreen	(StringPtr					inText,
						 TelnetScreenID const		inScreenID)
{
	SetMenuItemText(gMenus[kMenuPositionConnections], getWindowMenuItemIndexByScreenID(inScreenID), inText);
}// SetMenuTextFromScreen


/*!
Adds a connections name to the Window menu in a
position determined by the connection parameters.

(3.0)
*/
void
AddConnectionsMenuItem	(ConnectionDataPtr		inConnectionDataPtr,
						 StringPtr				inHostName,
						 StringPtr				inWindowName)
{
	Str255		temp = "\psetup error for connection #%a"; // LOCALIZE THIS
	Str255		numString;
	char		menuItemMark = diamondMark;
	
	
	// set a flag to guarantee menus dont get screwed up later
	inConnectionDataPtr->addedMenuItem = true;
	
	// construct a dummy menu item, whose text gets used if all other attempts to change the text, below, fail
	NumToString((long)inConnectionDataPtr->vs, numString);
	StringUtilities_PMetaSetDefaults();
	StringUtilities_PBuild(temp, numString, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
	unless (gNumberOfConnectionItemsAdded) AppendMenu(gMenus[kMenuPositionConnections], "\p(-"); // add a divider for the first
	InsertMenuItem(gMenus[kMenuPositionConnections], temp, getWindowMenuItemIndexByScreenID(inConnectionDataPtr->vs));
	gNumberOfConnectionItemsAdded++;
	SetMenuMarkFromScreen(kMenuMarkTypeConnectionOpening, inConnectionDataPtr->vs);
	
	// set the item text to be the window title, if one was specified...
	if (StrLength(inWindowName) > 0)
	{
		PLstrcpy(temp, inWindowName);
	}
	else
	{
		// if the title is blank, make a string that includes the host name;
		// otherwise, the remaining string is the initial string, which shows
		// that some kind of connection setup error has occurred
		if (StrLength(inHostName) > 0)
		{
			PLstrcpy(temp, "\pConnection to %a"); // LOCALIZE THIS
			StringUtilities_PMetaSetDefaults();
			StringUtilities_PBuild(temp, inHostName, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
		}
	}
	
	SetMenuItemText(gMenus[kMenuPositionConnections], getWindowMenuItemIndexByScreenID(inConnectionDataPtr->vs), temp);
	installDynamicMenuItemStateTrackers();
}// AddConnectionsMenuItem


/*!
To flag whether or not the font menus are
available, use this method.  This is only
necessary to keep the font list rebuilding
process reasonably thread safe, making
sure the user cant access fonts until it
is done its work.

(3.0)
*/
void
Menus_SetFontMenuAvailable		(Boolean const		inIsAvailable)
{
	gFontMenusAvailable = inIsAvailable;
}// SetFontMenuAvailable


/*!
To flag whether or not the Scripts menu is
available, use this method.  This is only
necessary to keep the script-finding
process reasonably thread safe, making
sure the user cant access scripts until
it is done its work.

(3.0)
*/
void
Menus_SetScriptsMenuAvailable		(Boolean const		inIsAvailable)
{
	gScriptsMenuAvailable = inIsAvailable;
}// SetScriptsMenuAvailable


/*!
Updates the state of all menu titles in the menu bar.

(3.0)
*/
void
Menus_Service	(void)
{
	WindowRef		window = GetRealFrontWindow();
	UInt16			openConnectionCount = portsOpen();
	
	
	if (ScreenManager_IsWindowAScreen(window))
	{
		setMenuEnabled(gMenus[kMenuPositionAppearance], true);
		setMenuEnabled(gMenus[kMenuPositionSession], (openConnectionCount > 0));
		setMenuEnabled(gMenus[kMenuPositionKeyboard], (openConnectionCount > 0));
		//Keypads_SetActive(kKeypadTypeControlKeys, (openConnectionCount > 0));
		//Keypads_SetActive(kKeypadTypeFunctionKeys, (openConnectionCount > 0));
		//Keypads_SetActive(kKeypadTypeVT220Keys, (openConnectionCount > 0));
	}
	else
	{
		setMenuEnabled(gMenus[kMenuPositionAppearance], false);
		setMenuEnabled(gMenus[kMenuPositionSession], false);
		setMenuEnabled(gMenus[kMenuPositionKeyboard], false);
		//Keypads_SetActive(kKeypadTypeControlKeys, false);
		//Keypads_SetActive(kKeypadTypeFunctionKeys, false);
		//Keypads_SetActive(kKeypadTypeVT220Keys, false);
	}
	DrawMenuBar();
}// Service


//
// internal methods
//

/*!
To automatically set the enabled state, and as
appropriate, the mark, item text, etc. for a
menu item that has no command ID, use this
method.

For this method to work, you must have previously
associated a state tracking procedure with the
item using SetMenuItemAdjustmentProc().

This routine is called by Menus_SetUpItemStates(),
and is probably not needed any place else.

(3.0)
*/
static void
adjustMenuItem		(MenuRef			inMenu,
					 MenuItemIndex		inItemNumber)
{
	MenuCommandStateTrackerProcPtr	proc = null;
	
	
	if ((inMenu != null) && (inItemNumber > 0))
	{
		MenuCommand		commandID = 0;
		
		
		(OSErr)GetMenuItemCommandID(inMenu, inItemNumber, &commandID);
		getMenuItemAdjustmentProc(inMenu, inItemNumber, &proc);
		if (proc != null)
		{
			Boolean			isEnabled = InvokeMenuCommandStateTrackerProc(proc, commandID, inMenu, inItemNumber);
			
			
			setMenuItemEnabled(inMenu, inItemNumber, isEnabled);
		}
	}
}// adjustMenuItem


/*!
To automatically set the enabled state, and as
appropriate, the mark, item text, etc. for a
menu item, using only its command ID, use this
method.  Both primary and secondary menus are
updated, which means that menus specific to
simplified-user-interface mode are automatically
adjusted, too.

For this method to work, you must have previously
associated a state tracking procedure with the
item using SetMenuItemAdjustmentProcByCommandID().

This routine is called by Menus_SetUpItemStates(),
and is probably not needed any place else.

(3.0)
*/
static void
adjustMenuItemByCommandID	(UInt32 const		inCommandID)
{
	MenuRef							menu1 = null,
									menu2 = null;
	MenuItemIndex					itemIndex1 = 0,
									itemIndex2 = 0;
	
	
	GetMenusAndMenuItemIndicesByCommandID(inCommandID, &menu1, &menu2, &itemIndex1, &itemIndex2);
	adjustMenuItem(menu1, itemIndex1);
	adjustMenuItem(menu2, itemIndex2);
}// adjustMenuItemByCommandID


/*!
Many menu items are all jointly enabled or
disabled based on the same basic factor -
whether any connection window is frontmost.
Use this method to determine if session-
related menu items should be enabled.

(3.0)
*/
static Boolean
areSessionRelatedItemsEnabled	(void)
{
	Boolean		result = (areTEKRelatedItemsEnabled() && ScreenManager_IsWindowAScreen(GetRealFrontWindow()) &&
							(GetWindowKind(GetRealFrontWindow()) == WIN_CNXN));
	
	
	return result;
}// areSessionRelatedItemsEnabled


/*!
Many menu items are all jointly enabled or
disabled based on the same basic factor -
whether any graphics window is frontmost.
Use this method to determine if TEK-
related menu items should be enabled.

(3.0)
*/
static Boolean
areTEKRelatedItemsEnabled	(void)
{
	Boolean			result = false;
	
	
	if (ScreenManager_IsWindowAScreen(GetRealFrontWindow()))
	{
		result = (portsOpen() >= 1);
	}
	return result;
}// areTEKRelatedItemsEnabled
	
	
/*!
This method sets up the menus and menu items
that are specifically part of Telnet (i.e.
not the Apple or Help menus).  It initializes
the menu bar, inserts all of Telnets menus
not part of the menu bar resource and then
updates dynamic items.  In addition, submenus
are inserted into the submenu list.

This routine is called at startup, but also
after menu command keys are turned on (because
menus are grabbed fresh from the resource file).

(3.0)
*/
static void
buildMenuBar (void)
{
	// build the static part of the menu bar from a 'MBAR' resource
	SetMenuBar(GetNewMBar(kMenuBarIDMain));
	DrawMenuBar();
	
	// add pop-up menus and submenus to the submenu list
	InsertMenu(gMenus[kMenuPositionFont], kInsertHierarchicalMenu);
	InsertMenu(gMenus[kMenuPositionOpen], kInsertHierarchicalMenu);
	InsertMenu(gMenus[kMenuPositionTranslation], kInsertHierarchicalMenu);
	InsertMenu(gMenus[kMenuPositionTektronix], kInsertHierarchicalMenu);
	InsertMenu(gMenus[kMenuPositionTelnetOnTheWeb], kInsertHierarchicalMenu);
	
	// build the dynamic part of the menu bar
	InsertMenu(gMenus[kMenuPositionConnections], 0);
	InsertMenu(gMenus[kMenuPositionScripts], 0);
}// buildMenuBar


/*!
Displays the status of the indicated connection.

(2.6)
*/
static void
displayStatus	(TelnetScreenID const		inScreenID)
{
	short				hit = 0;
	ConnectionDataPtr	connectionDataPtr = null;
	Str255				alertString,
						junkString,
						statusString;
	
	
	Cursors_UseArrow();
	connectionDataPtr = ScreenManager_GetConnectionDataOf(inScreenID);
	
	// construct a message showing the status of the connection
	GetIndString(alertString, rStringsStatus, siStatusTheConnection);
	switch (connectionDataPtr->status)
	{
		case kConnectionStatusDead:
			GetWTitle(connectionDataPtr->window, junkString);
			GetIndString(statusString, rStringsMiscellaneous, siAwaitingDismissal);
			break;
		
		case kConnectionStatusOpening:
			PLstrcpy(junkString, connectionDataPtr->host.name);
			GetIndString(statusString, rStringsMiscellaneous, siBeingOpened);
			break;
		
		default:
			PLstrcpy(junkString, connectionDataPtr->host.name);
			GetIndString(statusString, rStringsMiscellaneous, siBeingLookedUp);
			break;
	}
	StringUtilities_PMetaSetDefaults();
	StringUtilities_PBuild(alertString, junkString, statusString, EMPTY_PSTRING, EMPTY_PSTRING);
	
	// now display the message, giving the user the chance to kill the connection
	{
		ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
		InterfaceLibAlertRef	box = null;
		Str255					buttonStringAbort,
								buttonStringContinue,
								helpString;
		
		
		GetIndString(buttonStringContinue, rStringsSpecialButtons, siButtonContinue);
		GetIndString(buttonStringAbort, rStringsSpecialButtons, siButtonAbort);
		GetIndString(helpString, rStringsStatus, siStatusHelpText);
		StringUtilities_PMetaSetDefaults();
		StringUtilities_PBuild(helpString, buttonStringAbort, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
		
		box = Alert_New();
		Alert_SetFilterProc(box, upp);
		Alert_SetHelpButton(box, false);
		Alert_SetParamsFor(box, kAlert_StyleOKCancel);
		Alert_SetButtonText(box, kAlertStdAlertOKButton, buttonStringContinue);
		Alert_SetButtonText(box, kAlertStdAlertCancelButton, buttonStringAbort);
		Alert_SetText(box, alertString, helpString);
		Alert_SetType(box, kAlertNoteAlert);
		Alert_Display(box);
		hit = Alert_ItemHit(box);
		Alert_Dispose(&box);
		DisposeModalFilterUPP(upp), upp = null;
	}
	
	if (hit == kAlertStdAlertCancelButton) // the Abort button
	{
		Connections_Disconnect(connectionDataPtr);
	}
}// displayStatus


/*!
To dispose of menus previously created with getMainMenus(),
use this method.

(3.0)
*/
static void
disposeMainMenus	(void)
{
	DisposeMenu(gMenus[kMenuPositionFile]), gMenus[kMenuPositionFile] = null;
	DisposeMenu(gMenus[kMenuPositionEdit]), gMenus[kMenuPositionEdit] = null;
	DisposeMenu(gMenus[kMenuPositionAppearance]), gMenus[kMenuPositionAppearance] = null;
	DisposeMenu(gMenus[kMenuPositionSession]), gMenus[kMenuPositionSession] = null;
	DisposeMenu(gMenus[kMenuPositionKeyboard]), gMenus[kMenuPositionKeyboard] = null;
	DisposeMenu(gMenus[kMenuPositionNetwork]), gMenus[kMenuPositionNetwork] = null;
	DisposeMenu(gMenus[kMenuPositionSimpleFile]), gMenus[kMenuPositionSimpleFile] = null;
	DisposeMenu(gMenus[kMenuPositionSimpleEdit]), gMenus[kMenuPositionSimpleEdit] = null;
}// disposeMainMenus


/*!
To run a script, whose file is located in MacTelnets
Scripts Menu Items folder with a name identical to that
of the given item of the given menu, use this method.

(3.0)
*/
static void
executeScriptByMenuEvent	(MenuRef			inMenu,
							 MenuItemIndex		inItemNumber,
							 EventModifiers		UNUSED_ARGUMENT(inModifiers))
{
	FSSpec		scriptFile;
	OSErr		error = noErr;
	
	
	error = FolderManager_GetFSSpec(kTelnetFolderScriptsMenuItems, &scriptFile);
	if (error == noErr)
	{
		GetMenuItemText(inMenu, inItemNumber, scriptFile.name);
		
		// restore the "!" prefix for special files (it was stripped prior to placement in the menu)
		if (inItemNumber <= gNumberOfSpecialScriptMenuItems) StringUtilities_PInsert(scriptFile.name, "\p!");
		
		error = FSMakeFSSpec(scriptFile.vRefNum, scriptFile.parID, scriptFile.name, &scriptFile);
		if (error == noErr) error = AppleEventUtilities_ExecuteScriptFile(&scriptFile, true/* notify user */);
	}
}// executeScriptByMenuEvent


/*!
To get fresh menu handles for the main menus from the resource
file, copying them into the global array, use this method.

This routine automatically checks to see if the Aqua menu
layout is in use, and if so, it removes the divider and Quit
items from the File menu and renames the menu Session.  The
Aqua layout adds an automatic Quit item to the application menu,
so MacTelnet doesnt need one.

This method does NOT get the Apple menu (which never changes)
or the Connections menu (which is built dynamically).

WARNING:	The global handles to the old menus will be lost.
			Make sure you delete any old menus before using
			this method to re-grab new handles from the
			resource file, otherwise a memory leak ensues.

(3.0)
*/
static void
getMainMenus	(void)
{
	gMenus[kMenuPositionFile] = GetMenu(kMenuIDFile);
	gMenus[kMenuPositionEdit] = GetMenu(kMenuIDEdit);
	gMenus[kMenuPositionSession] = GetMenu(kMenuIDSession);
	gMenus[kMenuPositionKeyboard] = GetMenu(kMenuIDKeyboard);
	gMenus[kMenuPositionAppearance] = GetMenu(kMenuIDAppearance);
	gMenus[kMenuPositionNetwork] = GetMenu(kMenuIDNetwork);
	
	// Right now, MacTelnet 3.0 wastefully grabs simple menus
	// along with regular menus, even though the two will never
	// be used simultaneously.  In a future release, this code
	// will be cleaned up to make sure that memory is not used
	// to hold the menu records of menus not applicable to the
	// current mode (i.e. simplified interface or otherwise).
	gMenus[kMenuPositionSimpleFile] = GetMenu(kMenuIDSimpleFile);
	gMenus[kMenuPositionSimpleEdit] = GetMenu(kMenuIDSimpleEdit);
	
#if TARGET_API_MAC_CARBON
	// check to see if a Quit item is needed; if not, remove it (and the divider above it);
	// also, under the Aqua layout, the File menu is renamed to Session
	{
		long		gestaltResult = 0L;
		
		
		(OSErr)Gestalt(gestaltMenuMgrAttr, &gestaltResult);
		if ((gestaltResult & gestaltMenuMgrAquaLayoutMask) != 0)
		{
			// with an Aqua-style layout, the Quit item appears automatically in the Application
			// menu and therefore should not appear in the File menu (the command continues to work
			// because Mac OS X sends a "quit" Apple Event to MacTelnet when the user invokes Quit,
			// so MacTelnet doesnt have to worry about handling the command directly)
			DeleteMenuItem(gMenus[kMenuPositionFile], iQuit);
			DeleteMenuItem(gMenus[kMenuPositionFile], iQuit - 1);
			DeleteMenuItem(gMenus[kMenuPositionSimpleFile], iSimpleQuit);
			DeleteMenuItem(gMenus[kMenuPositionSimpleFile], iSimpleQuit - 1);
			
			// rename the File menu to Session, using the Mac OS X convention of specificity in
			// the first menus name
			{
				Str255		newMenuName;
				
				
				GetIndString(newMenuName, rStringsMenuItemTextSubstitutions, siMenuItemTextFileMenuTitleUnderAqua);
				(OSStatus)SetMenuTitle((gMenus[kMenuPositionFile]), newMenuName);
				(OSStatus)SetMenuTitle((gMenus[kMenuPositionSimpleFile]), newMenuName);
			}
		}
	}
#endif
	
	// add Translation Tables to the Keyboard menu
	{
		Str255			text;
		SInt16			i = 0,
						tableCount = 0;
		
		
		tableCount = CountMenuItems(gMenus[kMenuPositionTranslation]);
		for (i = 1; i <= tableCount; i++)
		{
			GetMenuItemText(gMenus[kMenuPositionTranslation], i, text);
			StringUtilities_PInsert(text, "\p  "); // indent them - it looks better :)
			AppendMenu(gMenus[kMenuPositionKeyboard], text);
		}
	}
}// getMainMenus


/*!
To interpret the reference constant of a menu item as if
it were a command state tracking routine, use this method.
This allows you to directly get access to a routine that
can describe an items state, using only the item
information itself.

(3.0)
*/
static void
getMenuItemAdjustmentProc	(MenuRef							inMenu,
							 MenuItemIndex						inItemNumber,
							 MenuCommandStateTrackerProcPtr*	outProcPtrPtr)
{
	(OSErr)GetMenuItemRefCon(inMenu, inItemNumber, (UInt32*)outProcPtrPtr);
}// getMenuItemAdjustmentProc


/*!
To determine the screen ID given the item index of a
menu item in the Window menu, use this method.

(3.0)
*/
static TelnetScreenID
getScreenIDFromWindowMenuItemIndex	(MenuItemIndex		inIndex)
{
	return (inIndex - iFirstConnection + 1);
}// getScreenIDFromWindowMenuItemIndex


/*!
To determine the item index of an item in the Window menu
corresponding to the specified connections window, use
this method.

(3.0)
*/
static MenuItemIndex
getWindowMenuItemIndexByScreenID	(TelnetScreenID		inScreenID)
{
	return (inScreenID + iFirstConnection - 1);
}// getWindowMenuItemIndexByScreenID


/*!
Installs state trackers for menu items that tend to
change over time (e.g. number of connections open
in the Window menu).  These trackers ensure that
the item states are correct for these items (such
as the presence of a checkmark, the item text, and
the enabled state).

(3.0)
*/
static void
installDynamicMenuItemStateTrackers		(void)
{
	register MenuItemIndex	itemIndex = 0;
	SInt16					itemCount = 0;
	MenuRef					menu = null;
	
	
	// Map menu
	menu = gMenus[kMenuPositionKeyboard];
	itemCount = CountMenuItems(gMenus[kMenuPositionTranslation]);
	for (itemIndex = 0; itemIndex < itemCount; itemIndex++)
	{
		SetMenuItemAdjustmentProc(menu, iFirstTranslationTable + itemIndex, stateTrackerKeyboardMenuTranslationChoices);
	}
	
	// Window menu
	menu = gMenus[kMenuPositionConnections];
	itemCount = Connections_GetCount();
	for (itemIndex = 0; itemIndex < itemCount; itemIndex++)
	{
		SetMenuItemAdjustmentProc(menu, iFirstConnection + itemIndex, stateTrackerWindowMenuConnections);
	}
}// installDynamicMenuItemStateTrackers


/*!
After the global set of menus has been created (probably
by getting menus from resources), use this method to
install item state trackers in each one.  If you ever
re-grab a menu fresh from a resource file, be sure to
call this method to make sure that the new menu is set up
properly, otherwise its items states will not be set by
a call to Menus_SetUpItemStates().

Any menu command whose state (enabled, checked, item text,
etc.) could ever change should have a state tracker.  Only
one state tracker is allowed per command.  The trackers are
invoked by Menus_SetUpItemStates().

For convenience, the procedures are set in the exact order
they appear in menus, and grouped by menu.  Any command for
which no state tracker is used is listed anyway, as a
comment, for completeness.

This routine also invokes installDynamicMenuStateTrackers().

(3.0)
*/
static void
installMenuItemStateTrackers	(void)
{
	// Apple
	SetMenuItemAdjustmentProcByCommandID(kCommandAboutTelnet, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandTelnetOnTheWeb, null/* no state tracker necessary */);
	
	// File
	SetMenuItemAdjustmentProcByCommandID(kCommandOpenConnection, null/* no state tracker necessary */);
	SetMenuItemAdjustmentProcByCommandID(kCommandOpenConnectionFavorite, null/* no state tracker necessary */);
	SetMenuItemAdjustmentProcByCommandID(kCommandCloseConnection, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandLoadSet, null/* no state tracker necessary */);
	SetMenuItemAdjustmentProcByCommandID(kCommandSaveSet, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandImportMacroSet, null/* no state tracker necessary */);
	SetMenuItemAdjustmentProcByCommandID(kCommandExportCurrentMacroSet, null/* no state tracker necessary */);
	SetMenuItemAdjustmentProcByCommandID(kCommandPrint, stateTrackerPrintingItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandPrintOne, stateTrackerPrintingItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandPrintScreen, stateTrackerPrintingItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandPageSetup, stateTrackerPrintingItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandQuit, null/* no state tracker necessary */);
	
	// Edit
	SetMenuItemAdjustmentProcByCommandID(kCommandUndo, stateTrackerStandardEditItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandRedo, stateTrackerStandardEditItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandCut, stateTrackerStandardEditItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandCopy, stateTrackerStandardEditItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandCopyTable, stateTrackerStandardEditItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandPaste, stateTrackerStandardEditItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandType, stateTrackerStandardEditItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandClear, stateTrackerStandardEditItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSelectAll, stateTrackerStandardEditItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandShowClipboard, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandPreferences, stateTrackerPreferences);
	
	// Appearance
	SetMenuItemAdjustmentProcByCommandID(kCommandLargeScreen, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSmallScreen, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSetScreenSize, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandBiggerText, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSmallerText, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandFullScreen, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandFormat, stateTrackerGenericSessionItems);
	
	// Session
	SetMenuItemAdjustmentProcByCommandID(kCommandCaptureToFile, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandEcho, stateTrackerCheckableItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandWrapMode, stateTrackerCheckableItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandClearScreenSavesLines, stateTrackerCheckableItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandJumpScrolling, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandTektronix, stateTrackerTEKItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSpeechEnabled, stateTrackerCheckableItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandResetTerminal, stateTrackerGenericSessionItems);
	
	// Keyboard
	SetMenuItemAdjustmentProcByCommandID(kCommandDeletePressSendsBackspace, stateTrackerCheckableItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandDeletePressSendsDelete, stateTrackerCheckableItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSetKeys, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandEMACSArrowMapping, stateTrackerCheckableItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandLocalPageUpDown, stateTrackerCheckableItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandTitleTranslationChoices, stateTrackerAlwaysDisabled);
	// state trackers for translation table items are installed dynamically
	
	// Network
	SetMenuItemAdjustmentProcByCommandID(kCommandShowNetworkNumbers, null/* no state tracker necessary */);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendInternetProtocolNumber, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendFTPCommand, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendAreYouThere, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendAbortOutput, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendBreak, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendInterruptProcess, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendEraseCharacter, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendEraseLine, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendEndOfFile, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSendSync, stateTrackerNetSendItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandWatchForActivity, stateTrackerCheckableItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandSuspendNetwork, stateTrackerCheckableItems);
	
	// Window
	SetMenuItemAdjustmentProcByCommandID(kCommandHideFrontWindow, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandHideOtherWindows, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandShowAllHiddenWindows, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandStackWindows, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandNextWindow, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandChangeWindowTitle, stateTrackerGenericSessionItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandMinimizeWindow, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandShowSessionInspector, stateTrackerAlwaysDisabled/* tmp *//*stateTrackerShowHideItems*/);
	SetMenuItemAdjustmentProcByCommandID(kCommandShowFTPLog, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandShowCommandLine, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandShowKeypad, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandShowFunction, stateTrackerShowHideItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandShowControlKeys, stateTrackerShowHideItems);
	// state trackers for connection items are installed dynamically
	
	// TEK
	SetMenuItemAdjustmentProcByCommandID(kCommandTEKPageCommand, stateTrackerTEKItems);
	SetMenuItemAdjustmentProcByCommandID(kCommandTEKPageClearsScreen, stateTrackerTEKItems);
	
	// Internet Resources
	SetMenuItemAdjustmentProcByCommandID(kCommandTelnetHomePage, null/* no state tracker necessary */);
	SetMenuItemAdjustmentProcByCommandID(kCommandTelnetAuthorMail, null/* no state tracker necessary */);
	
	installDynamicMenuItemStateTrackers();
}// installMenuItemStateTrackers


/*!
This method grabs all menus from the resource file
not grabbed by getMainMenus(), and initializes them
appropriately.

(2.6)
*/
static void
loadMenus	(Boolean const		inMenusHaveKeyEquivalents)
{
	// dynamic menus and submenus
	gMenus[kMenuPositionFont] = GetMenu(kPopUpMenuIDFont);
	gMenus[kMenuPositionOpen] = GetMenu(kSubmenuIDOpen);
	gMenus[kMenuPositionTektronix] = GetMenu(kSubmenuIDTektronix);
	gMenus[kMenuPositionTelnetOnTheWeb] = GetMenu(kSubmenuIDTelnetOnTheWeb);
	gMenus[kMenuPositionTranslation] = GetMenu(kSubmenuIDTranslation);
	TextTranslation_Init(gMenus[kMenuPositionTranslation]);
	
	// other main menus
	gMenus[kMenuPositionApple] = GetMenu(kMenuIDApple);
#if TARGET_API_MAC_OS8
	AppendResMenu(gMenus[kMenuPositionApple], 'DRVR'); // add Apple Menu Items to the Apple menu
#endif
	gMenus[kMenuPositionConnections] = GetMenu(kMenuIDConnections);
	
	gMenus[kMenuPositionScripts] = GetMenu(kMenuIDScripts);
#if TARGET_API_MAC_OS8
	// make the Scripts menu title iconic; to do this, the first two bytes
	// of its menu handle must be "0x0501" and the remaining data must be
	// an icon suite handle (this hack doesnt work on Mac OS X, so for now
	// Im using a text title under Carbon)
	{
		MenuRef		menu = gMenus[kMenuPositionScripts];
		
		
		if (GetIconSuite(&gScriptsMenuIconSuite, kIconSuiteScriptsMenu, kSelectorAllSmallData) == noErr)
		{
			// for some reason, explicitly setting the length field causes a crash
			//(**menu).menuData[0] = 0x05; // length of data that follows (this is still a Pascal string here)
			(**menu).menuData[1] = 0x01;
			*((UInt32*)&((**menu).menuData[2])) = (UInt32)gScriptsMenuIconSuite;
		}
	}
#endif
	
	// get the rest of the main menus, removing key equivalents if that preference is set
	Menus_SetKeyEquivalents(inMenusHaveKeyEquivalents, true/* first call */); // removes command keys if preference is set
	
	// set up dynamic menu items
	setUpSessionFavoritesSubmenu(gMenus[kMenuPositionOpen]);
	setUpScriptsMenu(gMenus[kMenuPositionScripts]);
	
	if (FlagManager_Test(kFlagThreadManager))
	{
		// try to asynchronously build the font list...
		Console_WriteLine("Starting a thread to check installed fonts.");
		Threads_NewCoOp(NewThreadEntryUPP(Threads_InitializeFontMenu),
						gMenus[kMenuPositionFont]/* parameter */, 0/* extra stack size*/,
						kTelnetThreadDescriptorFontListRebuilding);
	}
	else
	{
		// ...but if the Thread Manager is not available, the font list must be scanned synchronously
		Console_WriteLine("Synchronously checking installed fonts.");
		(voidPtr)Threads_InitializeFontMenu(gMenus[kMenuPositionFont]);
	}
}// loadMenus


/*!
Counts the number of ports that are open (active).
This method has been rewritten in MacTelnet 3.0.
*/
static short
portsOpen (void)
{
	return (ScreenManager_GetStatusCount(kConnectionStatusActive));
}// portsOpen


/*!
To remove the key equivalents from all menu items
in a menu, call this routine.  This routine will
automatically avoid potential disasters involving
specially-encoded command keys (0x1B, etc.) as
documented in Inside Macintosh: Macintosh Toolbox
Essentials, by not touching command key fields
for items with special command key fields.

(3.0)
*/
static void
removeMenuItemModifiers		(MenuRef		inMenu)
{
	CharParameter		c = '\0';
	register int		i = 0;
	
	
	for (i = 1; i <= CountMenuItems(inMenu); i++)
	{
		SetMenuItemModifiers(inMenu, i, kMenuNoCommandModifier);
		GetItemCmd(inMenu, i, &c);
		if ((c != 0x1B) && (c != 0x1C) && (c != 0x1D) && (c != 0x1E)) SetItemCmd(inMenu, i, 0);
	}
}// removeMenuItemModifiers


/*!
To set the enabled state of an entire menu,
use this method.  This routine automatically
uses the most advanced routine possible
(namely the EnableMenuItem() or
DisableMenuItem() commands under OS 8.5 and
beyond, if available).

(3.0)
*/
static void
setMenuEnabled		(MenuRef			inMenu,
					 Boolean const		inEnabled)
{
	setMenuItemEnabled(inMenu, 0, inEnabled);
}// setMenuEnabled


/*!
Changes the key equivalents of certain menu
items to match the users preferences for special
control key combinations (interrupt, suspend,
and resume).  This can change with each session,
so its important that this routine be invoked
whenever the active window changes.

Unimplemented.

(3.0)
*/
static void
setMenuItemKeysFromControlKeyPrefs	(void)
{
	
}// setMenuItemKeysFromControlKeyPrefs


/*!
To set the enabled state of a menu item or an
entire menu, use this method.  This routine
automatically uses the most advanced routine
possible (namely the EnableMenuItem() or
DisableMenuItem() commands under OS 8.5 and
beyond, if available).

(3.0)
*/
static void
setMenuItemEnabled		(MenuRef			inMenu,
						 UInt16 const		inItemNumberOrZero,
						 Boolean const		inEnabled)
{
#if TARGET_API_MAC_OS8
	if (!FlagManager_Test(kFlagCarbonMenuManager))
	{
		if (inEnabled) EnableItem(inMenu, inItemNumberOrZero);
		else DisableItem(inMenu, inItemNumberOrZero);
	}
	else
#endif
	{
		if (inEnabled) EnableMenuItem(inMenu, inItemNumberOrZero);
		else DisableMenuItem(inMenu, inItemNumberOrZero);
	}
}// setMenuItemEnabled


/*!
Destroys and rebuilds the contents of the
New Session From Favorites submenu,
utilizing the most current saved session
preferences for the new menu items.

(2.6)
*/
static void
setUpSessionFavoritesSubmenu	(MenuRef		inMenu)
{
	register UInt16		i = 0;
	SInt16				oldResFile = CurResFile();
	
	
	for (i = CountMenuItems(inMenu); i > 0; i--) DeleteMenuItem(inMenu, i);
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	InsertResMenu(inMenu, SESSIONPREFS_RESTYPE, 999/* item to insert after (last item) */);
	
	UseResFile(oldResFile);
}// setUpSessionFavoritesSubmenu


/*!
To destroy and rebuild the contents of the
Scripts menu, reading all available
scripts and executables from the Scripts
folder in MacTelnets folder, use this
method.

Note that this is not re-entrant with prior
calls: in the extreme case, this routine
could start rebuilding the Scripts menu
from a folder containing a lot of scripts,
during which time the user deletes most of
them and then returns to MacTelnet, causing
another Scripts menu rebuild that would
collide with the previous one.  For this
reason, a flag is used to make sure that
this routine does nothing if a previous
call is in progress.

(3.0)
*/
static void
setUpScriptsMenu	(MenuRef		inMenu)
{
	if (gScriptsMenuAvailable)
	{
		OSErr		error = noErr;
		
		
		gScriptsMenuAvailable = false; // lock down (this is reversed at the end of the threads code)
		if (FlagManager_Test(kFlagThreadManager))
		{
			// try to asynchronously build the Scripts menu...
			error = Threads_NewCoOp(NewThreadEntryUPP(Threads_InitializeScriptsMenu),
									inMenu/* parameter */, INTEGER_KILOBYTES(2)/* extra stack size*/,
									kTelnetThreadDescriptorScriptsMenuRebuilding);
		}
		else error = threadNotFoundErr;
		
		if (error != noErr)
		{
			// ...but if the Thread Manager is not available, or the thread
			// creation otherwise fails, the font list must be scanned synchronously
			(voidPtr)Threads_InitializeScriptsMenu(inMenu);
		}
	}
}// setUpScriptsMenu


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, forces items to be disabled.

(3.0)
*/
static Boolean
stateTrackerAlwaysDisabled		(UInt32 const		UNUSED_ARGUMENT(inCommandID),
								 MenuRef			UNUSED_ARGUMENT(inMenu),
								 MenuItemIndex		UNUSED_ARGUMENT(inItemNumber))
{
	return false;
}// stateTrackerAlwaysDisabled


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines whether the specified item should be
enabled, and automatically sets its checkmark properly.

(3.0)
*/
static Boolean
stateTrackerCheckableItems		(UInt32 const		inCommandID,
								 MenuRef			inMenu,
								 MenuItemIndex		inItemNumber)
{
	ConnectionDataPtr		currentConnectionDataPtr = null;
	RealScreenDataPtr		currentRealScreenDataPtr = null;
	Boolean					connectionCommandResult = areSessionRelatedItemsEnabled(),
							checked = false;
	Boolean					result = false;
	
	
	// set up convenient data pointers
	if (areSessionRelatedItemsEnabled())
	{
		WindowRef		frontWindow = GetRealFrontWindow();
		
		
		if (ScreenManager_IsWindowAScreen(frontWindow))
		{
			currentConnectionDataPtr = ScreenManager_GetWindowConnectionData(frontWindow);
			currentRealScreenDataPtr = ScreenManager_GetWindowRealScreenData(frontWindow);
		}
	}
	
	switch (inCommandID)
	{
		case kCommandSpeechEnabled:
			result = connectionCommandResult;
			if (currentRealScreenDataPtr != null) checked = (result) ? (currentRealScreenDataPtr->speech.enabled) : false;
			break;
		
		case kCommandEcho:
			result = connectionCommandResult;
			if (currentConnectionDataPtr != null) checked = (result) ? (currentConnectionDataPtr->echo) : false;
			break;
		
		case kCommandWrapMode:
			result = connectionCommandResult;
			if (currentConnectionDataPtr != null) checked = (result) ? (currentConnectionDataPtr->wrap) : false;
			break;
		
		case kCommandClearScreenSavesLines:
			result = connectionCommandResult;
			if (currentConnectionDataPtr != null) checked = (result) ? (currentConnectionDataPtr->ESscroll) : false;
			break;
		
		case kCommandDeletePressSendsBackspace:
			result = connectionCommandResult;
			if (currentConnectionDataPtr != null) checked = (result) ? (!currentConnectionDataPtr->bsdel) : false;
			break;
		
		case kCommandDeletePressSendsDelete:
			result = connectionCommandResult;
			if (currentConnectionDataPtr != null) checked = (result) ? (currentConnectionDataPtr->bsdel) : false;
			break;
		
		case kCommandEMACSArrowMapping:
			// this is a VT220-only option, so disable the command otherwise
			result = connectionCommandResult;
			if (result) result = (currentConnectionDataPtr->terminal.emulation == kTerminalEmulationTypeVT220);
			if (currentConnectionDataPtr != null) checked = (result) ? (currentConnectionDataPtr->arrowmap) : false;
			break;
		
		case kCommandLocalPageUpDown:
			// this is a VT220-only option, so disable the command otherwise
			result = connectionCommandResult;
			if (result) result = (currentConnectionDataPtr->terminal.emulation == kTerminalEmulationTypeVT220);
			if (currentConnectionDataPtr != null) checked = (result) ? (currentConnectionDataPtr->pgupdwn) : false;
			break;
		
		case kCommandWatchForActivity:
			result = connectionCommandResult;
			if (currentConnectionDataPtr != null) checked = (result) ? (currentConnectionDataPtr->watched) : false;
			break;
		
		case kCommandSuspendNetwork:
			result = connectionCommandResult;
			checked = (result) ? FlagManager_Test(kFlagNetworkSuspended) : false;
			break;
		
		default:
			// unknown command!
			break;
	}
	
	CheckMenuItem(inMenu, inItemNumber, checked);
	
	return result;
}// stateTrackerCheckableItems


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines whether the specified session-related
commands are supposed to be enabled.  See also the
stateTrackerCheckableItems() method, which handles some
session-related checkmarked items.

(3.0)
*/
static Boolean
stateTrackerGenericSessionItems		(UInt32 const		inCommandID,
									 MenuRef			inMenu,
									 MenuItemIndex		inItemNumber)
{
	WindowRef		frontWindow = GetRealFrontWindow();
	Boolean			result = false;
	
	
	switch (inCommandID)
	{
		case kCommandCloseConnection:
			// this applies to any window with a close box
			result = (frontWindow != null);
			if (result)
			{
			#if TARGET_API_MAC_OS8
				result = GetWindowGoAwayFlag(frontWindow);
			#else
				WindowAttributes	attributes;
				
				
				GetWindowAttributes(frontWindow, &attributes);
				result = ((attributes & kWindowCloseBoxAttribute) != 0);
			#endif
			}
			if (result)
			{
				Str255			menuItemText;
				short			stringIndex = 0;
				UInt32			commandID = kCommandCloseConnection;
				UInt8			itemModifiers = 0;
				Boolean			windowsStickAround = false;
				Size			actualSize = 0L;
				
				
				// get the users operating mode preference, if possible
				unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
											kTelnetPreferenceTag_DontAutoClose,
											sizeof(windowsStickAround), &windowsStickAround,
											&actualSize) == kTelnetPrefsResultCodeSuccess)
				{
					windowsStickAround = false; // assume windows always close, if preference cant be found
				}
				
				// use the most accurate terminology possible; close doesnt make
				// sense when it merely disconnects session windows that remain open
				if ((windowsStickAround) && (ScreenManager_IsWindowAScreen(frontWindow)) &&
					(ScreenManager_GetConnectionStatusOf
						(ScreenManager_GetWindowScreenID(frontWindow)) != kConnectionStatusDead))
				{
					// set the menu item text based on the state of modifier keys
					if (EventLoop_IsOptionKeyDown())
					{
						stringIndex = siMenuItemTextDisconnectAll;
						itemModifiers = kMenuOptionModifier;
					}
					else stringIndex = siMenuItemTextDisconnect;
				}
				else
				{
					// set the menu item text based on the state of modifier keys
					if (EventLoop_IsOptionKeyDown())
					{
						stringIndex = siMenuItemTextCloseAll;
						itemModifiers = kMenuOptionModifier;
					}
					else stringIndex = siMenuItemTextClose;
				}
				GetIndString(menuItemText, rStringsMenuItemTextSubstitutions, stringIndex);
				SetMenuItemTextByCommandID(commandID, menuItemText);
				SetMenuItemModifiersByCommandID(commandID, itemModifiers);
			}
			break;
		
		case kCommandCaptureToFile:
			result = areSessionRelatedItemsEnabled();
			{
				Str255		string;
				SInt16		index = siMenuItemTextCaptureToFileBegin;
				
				
				if (ScreenManager_IsWindowAScreen(GetRealFrontWindow()))
				{
					ConnectionDataPtr	currentConnectionDataPtr = ScreenManager_GetWindowConnectionData(GetRealFrontWindow());
					
					
					if (result && Terminal_FileCaptureInProgress(currentConnectionDataPtr->vs))
					{
						index = siMenuItemTextCaptureToFileEnd;
					}
				}
				GetIndString(string, rStringsMenuItemTextSubstitutions, index);
				SetMenuItemText(inMenu, inItemNumber, string);
			}
			break;
		
		case kCommandSaveSet:
		case kCommandLargeScreen:
		case kCommandSmallScreen:
		case kCommandSetScreenSize:
		case kCommandBiggerText:
		case kCommandSmallerText:
		case kCommandFullScreen:
		case kCommandFormat:
		case kCommandJumpScrolling:
		case kCommandSetKeys:
		case kCommandResetTerminal:
		case kCommandStackWindows:
		case kCommandNextWindow:
		case kCommandChangeWindowTitle:
			result = areSessionRelatedItemsEnabled();
			if (inCommandID == kCommandFormat)
			{
				// this item isnt available if the rebuilding font list thread is still running
				result = (result && gFontMenusAvailable);
			}
			else if (inCommandID == kCommandSetScreenSize)
			{
				// this item is also available for the console window
				result = (ScreenManager_IsWindowAScreen(frontWindow) &&
							((GetWindowKind(frontWindow) == WIN_CONSOLE) ||
							(GetWindowKind(frontWindow) == WIN_CNXN)));
			}
			else if (inCommandID == kCommandNextWindow)
			{
				Str255			menuItemText;
				short			stringIndex = 0;
				UInt8			itemModifiers = 0;
				UInt32			commandID = kCommandNextWindow;
				
				
				// set the menu item text based on the state of modifier keys
				if (EventLoop_IsShiftKeyDown())
				{
					stringIndex = siMenuItemTextPreviousWindow;
					itemModifiers = kMenuShiftModifier;
				}
				else if (EventLoop_IsOptionKeyDown())
				{
					stringIndex = siMenuItemTextNextWindowHidingPreviousOne;
					itemModifiers = kMenuOptionModifier;
				}
				else stringIndex = siMenuItemTextNextWindow;
				GetIndString(menuItemText, rStringsMenuItemTextSubstitutions, stringIndex);
				SetMenuItemTextByCommandID(commandID, menuItemText);
				SetMenuItemModifiersByCommandID(commandID, itemModifiers);
				result = (result && (GetNextWindow(GetRealFrontWindow()) != null));
			}
			else if (inCommandID == kCommandChangeWindowTitle)
			{
				result = (result && ScreenManager_IsWindowAScreen(frontWindow));
			}
			break;
		
		default:
			// unknown command!
			break;
	}
	return result;
}// stateTrackerGenericSessionItems


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines the mark for translation choice items
in the Keyboard menu, and whether or not they should be
enabled.

Although the menu reference is given, this method assumes
you are referring to the Keyboard menu and does not check
that the specified menu really is the Keyboard menu.

(3.0)
*/
static Boolean
stateTrackerKeyboardMenuTranslationChoices	(UInt32 const		UNUSED_ARGUMENT(inCommandID),
											 MenuRef			inMenu,
											 MenuItemIndex		inItemNumber)
{
	WindowRef		frontWindow = GetRealFrontWindow();
	Boolean			checked = false;
	Boolean			result = true;
	
	
	if (ScreenManager_IsWindowAScreen(frontWindow))
	{
		ConnectionDataPtr	connectionDataPtr = ScreenManager_GetWindowConnectionData(frontWindow);
		
		
		checked = (inItemNumber == (iFirstTranslationTable + connectionDataPtr->national));
	}
	else
	{
		result = false;
		checked = false;
	}
	CheckMenuItem(inMenu, inItemNumber, checked);
	
	return result;
}// stateTrackerKeyboardMenuTranslationChoices


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines whether the specified Network menu
data-sending item should be enabled.

(3.0)
*/
static Boolean
stateTrackerNetSendItems	(UInt32 const		inCommandID,
							 MenuRef			UNUSED_ARGUMENT(inMenu),
							 MenuItemIndex		UNUSED_ARGUMENT(inItemNumber))
{
	Boolean			result = false;
	
	
	switch (inCommandID)
	{
		case kCommandSendFTPCommand:
		case kCommandSendInternetProtocolNumber:
		case kCommandSendAreYouThere:
		case kCommandSendAbortOutput:
		case kCommandSendBreak:
		case kCommandSendInterruptProcess:
		case kCommandSendEraseCharacter:
		case kCommandSendEraseLine:
		case kCommandSendEndOfFile:
		case kCommandSendSync:
			result = areSessionRelatedItemsEnabled();
			break;
		
		default:
			// unknown command!
			break;
	}
	return result;
}// stateTrackerNetSendItems


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines whether the Preferences item should be
enabled.

(3.0)
*/
static Boolean
stateTrackerPreferences		(UInt32 const		inCommandID,
							 MenuRef			UNUSED_ARGUMENT(inMenu),
							 MenuItemIndex		UNUSED_ARGUMENT(inItemNumber))
{
	Boolean			result = false;
	
	
	switch (inCommandID)
	{
		case kCommandPreferences:
			// this item is disabled if the Preferences window is in front
			if (GetRealFrontWindow() == null) result = true;
			else
			{
				WindowInfoRef			windowInfoRef = null;
				WindowInfoDescriptor	windowDescriptor = kInvalidWindowInfoDescriptor;
				
				
				windowInfoRef = WindowInfo_GetFromWindow(GetRealFrontWindow());
				if (windowInfoRef != null) windowDescriptor = WindowInfo_GetWindowDescriptor(windowInfoRef);
				result = (windowDescriptor != kTelnetDialogDescriptorPreferences);
			}
			break;
		
		default:
			// unknown command!
			break;
	}
	return result;
}// stateTrackerPreferences


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines whether the specified Network menu
data-sending item should be enabled.

(3.0)
*/
static Boolean
stateTrackerPrintingItems	(UInt32 const		inCommandID,
							 MenuRef			UNUSED_ARGUMENT(inMenu),
							 MenuItemIndex		UNUSED_ARGUMENT(inItemNumber))
{
	Boolean			result = false;
	
	
	switch (inCommandID)
	{
		case kCommandPrint:
		case kCommandPrintOne:
		case kCommandPrintScreen:
			result = ScreenManager_IsWindowAScreen(GetRealFrontWindow());
			if (result)
			{
				// for selection-dependent Print items, make sure text is selected
				if (inCommandID != kCommandPrintScreen)
				{
					RealScreenDataPtr	realScreenPtr = ScreenManager_GetWindowRealScreenData(GetRealFrontWindow());
					
					
					result = realScreenPtr->isTextSelected;
				}
			}
			break;
		
		case kCommandPageSetup:
			result = true;
			break;
		
		default:
			// unknown command!
			break;
	}
	return result;
}// stateTrackerPrintingItems


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines whether the specified show/hide item
should be enabled, and sets its item text appropriately.

(3.0)
*/
static Boolean
stateTrackerShowHideItems	(UInt32 const		inCommandID,
							 MenuRef			inMenu,
							 MenuItemIndex		inItemNumber)
{
	Str255		string;
	Boolean		result = false;
	
	
	switch (inCommandID)
	{
		case kCommandAboutTelnet:
			result = true;
			GetIndString(string, rStringsMenuItemTextSubstitutions,
							(EventLoop_IsOptionKeyDown() && EventLoop_IsShiftKeyDown())
							? siMenuItemTextAboutEgg : siMenuItemTextAboutNormal);
			SetMenuItemText(inMenu, inItemNumber, string);
			break;
		
		case kCommandShowClipboard:
			result = true;
			GetIndString(string, rStringsShowHide, Clipboard_IsVisible() ? siHideClipboard : siShowClipboard);
			SetMenuItemText(inMenu, inItemNumber, string);
			break;
		
		case kCommandHideFrontWindow:
			result = (areSessionRelatedItemsEnabled() && // only session windows can be hidden
						(ScreenManager_GetHiddenCount() < portsOpen()) && // cant hide any windows if they are all hidden!
						(portsOpen() >= 1)); // there must be at least one connection window open!
			break;
		
		case kCommandHideOtherWindows:
			result = (areSessionRelatedItemsEnabled() && // only session windows can be hidden
						(ScreenManager_GetHiddenCount() < portsOpen()) && // cant hide any windows if they are all hidden!
						(portsOpen() > 1)); // there must be multiple connection windows open!
			break;
		
		case kCommandShowAllHiddenWindows:
			result = (ScreenManager_GetHiddenCount() > 0); // there must be at least one hidden window
			break;
		
		case kCommandMinimizeWindow:
			{
				WindowRef		frontWindow = GetRealFrontWindow();
				
				
				result = (frontWindow != null); // there must be at least one window
				if (result) result = IsWindowCollapsable(frontWindow);
				GetIndString(string, rStringsShowHide, IsWindowCollapsed(frontWindow) ? siRestoreWindow : siMinimizeWindow);
				SetMenuItemText(inMenu, inItemNumber, string);
			}
			break;
		
		case kCommandShowSessionInspector:
			result = true;
			GetIndString(string, rStringsShowHide, Inspector_IsVisible() ? siHideSessionInspector : siShowSessionInspector);
			SetMenuItemText(inMenu, inItemNumber, string);
			break;
		
		case kCommandShowFTPLog:
			//#ifdef NDEBUG
			#if 1
			result = true;
			GetIndString(string, rStringsShowHide, FTPLog_IsVisible() ? siHideFTPLog : siShowFTPLog);
			SetMenuItemText(inMenu, inItemNumber, string);
			#else
			// the debugging console takes the place of this log, so disable the feature
			result = false;
			#endif
			break;
		
		case kCommandShowCommandLine:
			result = true;
			GetIndString(string, rStringsShowHide, CommandLine_IsVisible() ? siHideCommandLine : siShowCommandLine);
			SetMenuItemText(inMenu, inItemNumber, string);
			break;
		
		case kCommandShowKeypad:
			result = true;
			GetIndString(string, rStringsShowHide, Keypads_IsVisible(kKeypadTypeVT220Keys) ? siHideKeypad : siShowKeypad);
			SetMenuItemText(inMenu, inItemNumber, string);
			break;
		
		case kCommandShowFunction:
			result = true;
			GetIndString(string, rStringsShowHide, Keypads_IsVisible(kKeypadTypeFunctionKeys) ? siHideFunction : siShowFunction);
			SetMenuItemText(inMenu, inItemNumber, string);
			break;
		
		case kCommandShowControlKeys:
			result = true;
			GetIndString(string, rStringsShowHide, Keypads_IsVisible(kKeypadTypeControlKeys) ? siHideControlKeys : siShowControlKeys);
			SetMenuItemText(inMenu, inItemNumber, string);
			break;
		
		default:
			// unknown command!
			break;
	}
	return result;
}// stateTrackerShowHideItems


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines whether the specified standard Edit menu
item (Undo, Redo, Cut, Copy, Paste, Clear, Select All, or
a variation) should be enabled.

(3.0)
*/
static Boolean
stateTrackerStandardEditItems		(UInt32 const		inCommandID,
									 MenuRef			inMenu,
									 MenuItemIndex		inItemNumber)
{
	RealScreenDataPtr	currentRealScreen = null;
	Boolean				isReadOnly = ((GetWindowKind(GetRealFrontWindow()) == WIN_LOG) ||
										(GetWindowKind(GetRealFrontWindow()) == WIN_CONSOLE)), // is the frontmost window read-only?
						isTerminal = ScreenManager_IsWindowAScreen(GetRealFrontWindow()); // terminal windows include the FTP log, too
	Boolean				result = false;
	
	
	currentRealScreen = (isTerminal) ? ScreenManager_GetWindowRealScreenData(GetRealFrontWindow()) : null;
	switch (inCommandID)
	{
		case kCommandUndo:
			// set up the Edit menu properly to reflect the action
			// (if any) that can be undone
			{
				Str255		commandText;
				
				
				Undoables_GetUndoCommandInfo(commandText, &result);
				SetMenuItemText(inMenu, inItemNumber, commandText);
			}
			break;
		
		case kCommandRedo:
			// set up the Edit menu properly to reflect the action
			// (if any) that can be redone
			{
				Str255		commandText;
				
				
				Undoables_GetRedoCommandInfo(commandText, &result);
				SetMenuItemText(inMenu, inItemNumber, commandText);
			}
			break;
		
		case kCommandCut:
		case kCommandClear:
			result = false; // Telnet does not support these commands at all
			break;
		
		case kCommandCopy:
		case kCommandCopyTable:
		case kCommandType:
			result = isTerminal;
			if (currentRealScreen != null) result = (result && currentRealScreen->isTextSelected);
			if (inCommandID == kCommandCopyTable)
			{
				// Copy Table cant be used for copying graphics
				SInt16		j = 0;
				
				
				if (TektronixRealGraphics_IsRealGraphicsWindow(GetRealFrontWindow(), &j))
				{
					if ((j = RGgetVS(j)) > -1) result = false;
				}
			}
			if (inCommandID == kCommandType) result = (result && (!isReadOnly)); // cant type in a read-only window!
			break;
		
		case kCommandPaste:
			result = areSessionRelatedItemsEnabled() && (!isReadOnly);
			if (result)
			{
			#if TARGET_API_MAC_OS8
				long	offset = 0L;
				
				
				// there must be text on the clipboard in order for Paste to work
				result = (GetScrap(0L, kScrapFlavorTypeText, &offset) > 0L);
			#else
				ScrapRef			currentScrap = null;
				ScrapFlavorFlags	currentScrapFlags = 0L;
				
				
				if (GetCurrentScrap(&currentScrap) == noErr)
				{
					result = (GetScrapFlavorFlags(currentScrap, kScrapFlavorTypeText, &currentScrapFlags) == noErr);
				}
				else result = false;
			#endif
			}
			break;
		
		case kCommandSelectAll:
			result = isTerminal; // available if there is a terminal window in front
			break;
		
		default:
			// unknown command!
			break;
	}
	return result;
}// stateTrackerStandardEditItems


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines whether TEK commands should be enabled.

(3.0)
*/
static Boolean
stateTrackerTEKItems	(UInt32 const		inCommandID,
						 MenuRef			inMenu,
						 MenuItemIndex		inItemNumber)
{
	Boolean				result = false;
	
	
	switch (inCommandID)
	{
		case kCommandTektronix: // the TEK submenu
		case kCommandTEKPageCommand:
		case kCommandTEKPageClearsScreen:
			if (ScreenManager_IsWindowAScreen(GetRealFrontWindow()))
			{
				ConnectionDataPtr		connectionDataPtr = ScreenManager_GetWindowConnectionData(GetRealFrontWindow());
				
				
				result = areTEKRelatedItemsEnabled();
				if (result) result = (connectionDataPtr->TEK.mode != kTektronixModeNotAllowed);
				if (inCommandID == kCommandTEKPageClearsScreen)
				{
					CheckMenuItem(inMenu, inItemNumber,
									(result)
									? (connectionDataPtr->TEK.pageLocation == kTektronixPageLocationSameWindowClear)
									: false);
				}
			}
			break;
		
		default:
			// unknown command!
			break;
	}
	return result;
}// stateTrackerTEKItems


/*!
This routine, of standard MenuCommandStateTrackerProcPtr
form, determines the item icon and mark for connection
items in the Window menu.

Although the menu reference is given, this method assumes
you are referring to the Window menu and does not check
that the specified menu really is the Window menu.

Icons are currently not set using this routine; see
SetMenuMarkFromScreen().

(3.0)
*/
static Boolean
stateTrackerWindowMenuConnections	(UInt32 const		UNUSED_ARGUMENT(inCommandID),
									 MenuRef			inMenu,
									 MenuItemIndex		inItemNumber)
{
	TelnetScreenID		screenID = getScreenIDFromWindowMenuItemIndex(inItemNumber);
	Boolean				result = true;
	
	
	if (ScreenManager_GetWindowOf(screenID) == GetRealFrontWindow())
	{
		CheckMenuItem(inMenu, inItemNumber, true);
	}
	else
	{
		SInt16		zeroBasedIndex = (inItemNumber - iFirstConnection);
		
		
		CheckMenuItem(inMenu, inItemNumber, false);
		if (zeroBasedIndex < 10) SetItemMark(inMenu, inItemNumber, '1' + zeroBasedIndex);
	}
	
	return result;
}// stateTrackerWindowMenuConnections


/*!
A method of standard ScreenLineOperationProcPtr form,
this routine uses the Text Translation module to
convert the text on the specified line of a terminal
screen from one encoding to another.

Since this operation modifies the buffer, "true" is
always returned.
*/
static Boolean
translateScreenLine		(TelnetScreenID const	inScreenID,
						 char*					inoutLineTextBuffer,
						 UInt16 const			inLineTextBufferLength,
						 UInt16 const			UNUSED_ARGUMENT(inOneBasedLineNumber),
						 void*					inContextPtr)
{
	ScreenTranslationInfoPtr	translationInfoPtr = STATIC_CAST(inContextPtr, ScreenTranslationInfoPtr);
	Boolean						result = true;
	
	
	#if 1
	TextTranslation_ConvertTextToMac((unsigned char*)inoutLineTextBuffer, inLineTextBufferLength,
										translationInfoPtr->oldTable);
	TextTranslation_ConvertTextFromMac((unsigned char*)inoutLineTextBuffer, inLineTextBufferLength,
										translationInfoPtr->newTable);
	#else
	// temp - test Text Encoding Converter stuff (doesnt really work yet)
	{
		RealScreenDataPtr	realScreenPtr = ScreenManager_GetRealScreenDataOf(inScreenID);
		
		
		// translate the lines text into a new buffer, then copy it overtop
		if (realScreenPtr != null)
		{
			Handle		h = null;
			OSStatus	error = noErr;
			
			
			error = TextTranslation_ConvertBufferToNewHandle((unsigned char*)inoutLineTextBuffer, inLineTextBufferLength,
																realScreenPtr->textEncodingConverter, &h);
			if (error == noErr)
			{
				// currently, MacTelnet just truncates the text if its translated equivalent is too long
				// (since the maximum line buffer size is 180 bytes, this is VERY unlikely to occur)
				BlockMove(*h, (Ptr)inoutLineTextBuffer, INTEGER_MINIMUM(GetHandleSize(h), inLineTextBufferLength));
			}
			else Console_WriteValue("translation error", error);
			Memory_DisposeHandle(&h);
		}
		else Console_WriteLine("could not find screen");
	}
	#endif
	
	return result;
}// translateScreenLine