/*###############################################################

	Navigation.c
	
	This is a utility library for simplifying the usage of
	Navigation Services.  All methods that display file dialogs
	automatically determine the name of the application program
	and append it to the window title, separated by a colon.
	
	Interface Library 1.0
	 1998-2000 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#ifndef __FILES__
#	include <Files.h>
#endif
#ifndef __LOWMEM__
#	include <LowMem.h>
#endif
#ifndef __NAVIGATION__
#	include <Navigation.h>
#endif
#ifndef __PLSTRINGFUNCS__
#	include <PLStringFuncs.h>
#endif

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"



// internal methods
static OSErr	getAEDescData			(const AEDesc*		desc,
										 DescType*			typeCode,
										 void*				dataBuffer,
										 ByteCount			maximumSize,
										 ByteCount*			actualSize);
static Handle	newOpenHandle			(OSType				applicationSignature,
										 short				numTypes,
										 OSType				typeList[]);
static OSStatus	sendOpenAE				(AEDescList			list);
static void		setUpWTitleProcessName	(NavDialogOptions*	inoutDataPtr,
										 ConstStringPtr		inTitle);



//
// public methods
//

/**
Use this method after messing with a file to
complete the save by doing any necessary
translation and then disposing of the specified
reply record.

(1.0)
*/
OSStatus
Navigation_CompleteSave	(const FSSpec*		inFile,
						 NavReplyRecord*	inoutReplyPtr)
{
	OSStatus	result = noErr;
	
	
	if (inoutReplyPtr->validRecord)
	{
		result = NavCompleteSave(inoutReplyPtr, kNavTranslateInPlace);
	}
	result = NavDisposeReply(inoutReplyPtr);
	
	return result;
}// CompleteSave


/*!
Determines if the specified file is an alias, and if
so, modifies the file specification to point to the
original file (if possible).  It does this by checking
for the is alias Finder Info bit, and if set, by
reading the first alias resource in the file, and if
successful, by resolving the alias record in that
resource.  In other words, it resolves the damned
alias without any pretty stuff.  For some reason, no
Alias Manager API is quite this simple (sigh).

(3.0)
*/
OSErr
Navigation_EnsureOriginalFile	(FSSpec*		inoutFSSpecPtr)
{
	FileInfo		fileInfo;
	OSErr			result = noErr;
	
	
	// if the file is an alias, resolve it first
	result = FSpGetFInfo(inoutFSSpecPtr, (FInfo*)&fileInfo);
	if ((result == noErr) && (fileInfo.finderFlags & kIsAlias))
	{
		AliasHandle		fileAlias = null;
		SInt16			aliasResourceFile = -1;
		
		
		aliasResourceFile = FSpOpenResFile(inoutFSSpecPtr, fsRdPerm);
		result = ResError();
		if (result == noErr)
		{
			UseResFile(aliasResourceFile);
			if (Count1Resources(rAliasType) <= 0) result = resNotFound;
			else
			{
				// get the alias record and resolve it
				fileAlias = (AliasHandle)Get1IndResource(rAliasType, 1);
				result = ResError();
				
				if (result == noErr)
				{
					Boolean		aliasChanged = false;
					
					
					result = ResolveAlias(null/* base */, fileAlias, inoutFSSpecPtr, &aliasChanged);
				}
			}
		}
	}
	return result;
}// EnsureOriginalFile


/**
This routine is useful when you need to ask
the user to select a directory.

(1.0)
*/
OSErr
Navigation_GetDirectory	(ConstStr255Param	inPromptMessage,
						 ConstStr255Param	inDialogTitle,
						 int				inPrefKey,
						 int				inFlags,
						 NavEventProcPtr	inEventProc,
						 FSSpec*			outFileSpecPtr)
{	
	NavReplyRecord		theReply;
	NavDialogOptions	dialogOptions;
	OSErr				result = noErr;
	NavEventUPP			eventUPP = NewNavEventUPP(inEventProc);
	
	
	result = NavGetDefaultDialogOptions(&dialogOptions);
	dialogOptions.dialogOptionFlags = 0;
	dialogOptions.dialogOptionFlags |= inFlags;
	
	PLstrcpy(dialogOptions.message, inPromptMessage);
	setUpWTitleProcessName(&dialogOptions, inDialogTitle);
	dialogOptions.preferenceKey = inPrefKey;
	
	// display choose directory dialog
	Embedding_DeactivateFrontmostWindow();
	result = NavChooseFolder(null, &theReply, &dialogOptions, eventUPP,
								null, (NavCallBackUserData)null);
	Embedding_RestoreFrontmostWindow();
#if TARGET_API_MAC_OS8
	DisposeRoutineDescriptor(eventUPP);
#endif
	
	// NOTE: for some reason, *aliases* to directories result in the PARENT folder...how is this happening?...
	if ((theReply.validRecord) && (result == noErr))
	{
		// grab the target FSSpec from the AEDesc
		AEDesc		resultDesc;
		
		
		if ((result = AECoerceDesc(&(theReply.selection), typeFSS, &resultDesc)) == noErr)
		{
			result = getAEDescData(&resultDesc, null, outFileSpecPtr, sizeof(FSSpec), null);
		}
		AEDisposeDesc(&resultDesc);
		
		if (result == noErr) result = NavDisposeReply(&theReply);
		else (OSErr)NavDisposeReply(&theReply); // preserves previous error
	}
	return result;
}// GetDirectory


/**
This routine is adequate for 99% of open file
dialog needs.

(1.0)
*/
OSStatus
Navigation_GetFile	(ConstStr255Param	inPromptMessage,
					 ConstStr255Param	inDialogTitle,
					 OSType				inApplicationSignature,
					 int				inPrefKey,
					 int				inFlags,
					 short				inNumTypes,
					 OSType				inTypeList[],
					 NavEventProcPtr	inEventProc,
					 FSSpec*			inoutFileSpecPtr, // non-null for single-select
					 OSType*			inoutFileType, // non-null for single-select
					 FileInfo*			outFileInfo)   // non-null for single-select
{
	NavReplyRecord		theReply;
	NavDialogOptions	dialogOptions;
	NavTypeListHandle	openList = null;
	NavEventUPP			eventUPP = NewNavEventUPP(inEventProc);
	OSErr				result = noErr;
	
	
	result = NavGetDefaultDialogOptions(&dialogOptions);
	dialogOptions.dialogOptionFlags = 0;
	dialogOptions.dialogOptionFlags |= inFlags;
	
	// add customized settings
	PLstrcpy(dialogOptions.message, inPromptMessage);
	setUpWTitleProcessName(&dialogOptions, inDialogTitle);
	dialogOptions.preferenceKey = inPrefKey;
	
	// make sure file selection mode is right
	if (inoutFileSpecPtr != null)
	{
		if ((dialogOptions.dialogOptionFlags & kNavAllowMultipleFiles) != 0)
			dialogOptions.dialogOptionFlags -= kNavAllowMultipleFiles;
	}
	else
	{
		if ((dialogOptions.dialogOptionFlags & kNavAllowMultipleFiles) == 0)
			dialogOptions.dialogOptionFlags |= kNavAllowMultipleFiles;
	}
	
	// restrict displayed file types if a list of types is given
	if (inTypeList != null)
	{
		openList = (NavTypeListHandle)newOpenHandle(inApplicationSignature,
												inNumTypes, inTypeList);
		if (openList) HLock((Handle)openList);
	}
	
	Embedding_DeactivateFrontmostWindow();
	result = NavGetFile(null, &theReply, &dialogOptions, eventUPP, null,
						null, openList, null);
	Embedding_RestoreFrontmostWindow();
#if TARGET_API_MAC_OS8
	DisposeRoutineDescriptor(eventUPP);
#endif
	
	if (result == memFullErr) result = userCanceledErr;
	
	if ((result == noErr) && (theReply.validRecord))
	{
		if (inoutFileSpecPtr != null)
		{
			AEDesc 		resultDesc;
			AEKeyword	keyword;
			
			
			// single-file opening
			result = AEGetNthDesc(&theReply.selection, 1, typeFSS, &keyword, &resultDesc);
			if (result == noErr)
			{
			#if TARGET_API_MAC_OS8
				BlockMove(*resultDesc.dataHandle, inoutFileSpecPtr, sizeof(FSSpec));
			#else
				(OSErr)AEGetDescData(&resultDesc, inoutFileSpecPtr, sizeof(FSSpec));
			#endif
			}
			
			result = FSpGetFInfo(inoutFileSpecPtr, (FInfo*)outFileInfo);
			if (result == noErr)
			{
				if (inoutFileType != null) *inoutFileType = outFileInfo->fileType;
			}
		}
		else
		{
			// multiple files opening: use AppleEvents
			result = sendOpenAE(theReply.selection);
		}

		NavDisposeReply(&theReply);
	}
	
	if (openList != null)
	{
		HUnlock((Handle)openList);
		DisposeHandle((Handle)openList);
		openList = null;
	}
	
	return result;
}// GetFile


/**
Use this method to get access to zero or more
file specification records that are embedded
in an Apple Event description.

(1.0)
*/
OSErr
Navigation_GetFSSpecArrayFromAEDesc	(const AEDesc*		inAEDescPtr,
									 void*				outFSSpecArray,
									 long*				inoutFSSpecArrayLengthPtr)
{
	FSSpecArrayPtr	fileArray = (FSSpecArrayPtr)outFSSpecArray;
	OSErr			result = noErr;
	
	
	fileArray = (FSSpecArrayPtr)Memory_NewPtr(sizeof(FSSpec) * (*inoutFSSpecArrayLengthPtr));
	if (MemError() == noErr)
	{
		int		i = 0;
		
		
		*(void**)outFSSpecArray = fileArray;
		
		for (i = 1; i <= (*inoutFSSpecArrayLengthPtr); i++)
		{
			AEDesc	desc;
			FSSpec	spec;
			
			
			result = AEGetNthDesc(inAEDescPtr, i, typeFSS, null, &desc);
			if (result == noErr)
			{
				result = Navigation_GetFSSpecFromAEDesc(&desc, &spec);
				*fileArray++ = spec;
			}
			else
			{
				break;
			}
		}
	}
	return result;
}// GetFSSpecArrayFromAEDesc


/**
Use this method to get access to a file specification
record that is embedded in an Apple Event description,
in a Carbon-compliant manner.

(1.0)
*/
OSErr
Navigation_GetFSSpecFromAEDesc	(const AEDesc*		inAEDescPtr,
								 FSSpec*			outFSSpecPtr)
{
	AEDesc		tempDesc = { typeNull, null };
	OSErr		result = noErr;
	
	
	if (inAEDescPtr->descriptorType == typeFSS)
	{
		#if TARGET_API_MAC_OS8
			BlockMove(*inAEDescPtr->dataHandle, outFSSpecPtr, sizeof(FSSpec));
		#else
			(OSErr)AEGetDescData(inAEDescPtr, outFSSpecPtr, sizeof(FSSpec));
		#endif
	}
	else
	{
		result = AECoerceDesc(inAEDescPtr, typeFSS, &tempDesc);
		if (result == noErr)
		{
		#if TARGET_API_MAC_OS8
			BlockMove(*tempDesc.dataHandle, outFSSpecPtr, sizeof(FSSpec));
		#else
			(OSErr)AEGetDescData(&tempDesc, outFSSpecPtr, sizeof(FSSpec));
		#endif
		}
	}
	
	AEDisposeDesc(&tempDesc);
	return result;
}// GetFSSpecFromAEDesc


/**
This method can be used as a file filter procedure for a
Navigation Services dialog box when no special handling
whatsoever is desired.  A return value of true means
display query item; a value of false means that the
item should not be displayed.

(1.0)
*/
pascal Boolean
Navigation_NothingFilterProc	(AEDesc*				theItem,
								 void*					info,
								 NavCallBackUserData	callBackUD,
								 NavFilterModes			filterMode)
{
	Boolean					result = pascal_true;
	NavFileOrFolderInfo*	theInfo = (NavFileOrFolderInfo*)info;
	
	
	if (theItem->descriptorType == typeFSS)
	{
		if (!theInfo->isFolder)
		{
			// use:
			// 'theInfo->fileAndFolder.fileInfo.finderInfo.fdType'
			// to check for the file type you want to filter.
		}
	}
	return result;
}// NothingFilterProc


/**
This routine is adequate for 99% of save file
dialog needs.

(1.0)
*/
OSStatus
Navigation_PutFile	(ConstStr255Param	inPromptMessage,
					 ConstStr255Param	inDialogTitle,
					 OSType				inApplicationSignature,
					 OSType				inFileSignature,
					 int				inPrefKey,
					 int				inFlags,
					 NavEventProcPtr	inEventProc,
					 NavReplyRecord*	outReplyPtr,
					 FSSpec*			outFileSpecPtr)
{
	NavDialogOptions	dialogOptions;
	NavEventUPP			eventUPP = NewNavEventUPP(inEventProc);
	OSErr				result = noErr;
	
	
	result = NavGetDefaultDialogOptions(&dialogOptions);
	dialogOptions.dialogOptionFlags = 0;
	dialogOptions.dialogOptionFlags |= inFlags;
	
	// add customized settings
	PLstrcpy(dialogOptions.message, inPromptMessage);
	setUpWTitleProcessName(&dialogOptions, inDialogTitle);
	dialogOptions.preferenceKey = inPrefKey;
	
	Embedding_DeactivateFrontmostWindow();
	result = NavPutFile(null, outReplyPtr, &dialogOptions, eventUPP,
						inFileSignature, inApplicationSignature, null);
	Embedding_RestoreFrontmostWindow();
#if TARGET_API_MAC_OS8
	DisposeRoutineDescriptor(eventUPP);
#endif
	
	if (result == memFullErr) result = userCanceledErr;
	
	if ((result == noErr) && (outReplyPtr->validRecord))
	{
		// grab the target FSSpec from the AEDesc for saving:	
		AEDesc 		resultDesc;
		AEKeyword	keyword;
		Boolean		failed = false;
		
		
		result = AEGetNthDesc(&(outReplyPtr->selection), 1, typeFSS, &keyword, &resultDesc);
		if (result == noErr)
		{
		#if TARGET_API_MAC_OS8
			BlockMove(*resultDesc.dataHandle, outFileSpecPtr, sizeof(FSSpec));
		#else
			(OSErr)AEGetDescData(&resultDesc, outFileSpecPtr, sizeof(FSSpec));
		#endif
		}
		
		if (!outReplyPtr->replacing)
		{
			if (FSpCreate(outFileSpecPtr, inApplicationSignature,
								inFileSignature, outReplyPtr->keyScript))
			{
				BEEP;
				failed = true;
			}
		}
		
		if (!failed) AEDisposeDesc(&resultDesc);
	}
	return result;
}// PutFile


//
// internal methods
//

/**
To copy the data from an Apple Event into an arbitrary
handle, call this method.

(1.0)
*/
static OSErr
getAEDescData		(const AEDesc*		desc,
					 DescType*			typeCode,
					 void*				dataBuffer,
					 ByteCount			maximumSize,
					 ByteCount*			actualSize)
{
	Handle		h;
	ByteCount	dataSize;
	
	
	*typeCode = desc->descriptorType;
	h = (Handle)desc->dataHandle;
	dataSize = GetHandleSize(h);
	if (dataSize > maximumSize)
		*actualSize = maximumSize;
	else
		*actualSize = dataSize;
	BlockMoveData(*h, dataBuffer, *actualSize);
	
	return noErr;
}// getAEDescData


/**
This method creates a handle to a Navigation Services
file type list, which is an 'open' resource.

(1.0)
*/
static Handle
newOpenHandle	(OSType		applicationSignature,
				 short		numTypes,
				 OSType		typeList[])
{
	Handle		hdl = null;
	
	
	if (numTypes > 0)
	{
		hdl = Memory_NewHandle(sizeof(NavTypeList) + numTypes * sizeof(OSType));
		
		if (hdl != null)
		{
			NavTypeListHandle	open = (NavTypeListHandle)hdl;
			
			
			(*open)->componentSignature = applicationSignature;
			(*open)->osTypeCount = numTypes;
			BlockMoveData(typeList, (*open)->osType, numTypes * sizeof(OSType));
		}
	}
	
	return hdl;
}// newOpenHandle


/**
This method will send a series of 'odoc' Apple Events
for each event in the given event description list.

(1.0)
*/
static OSStatus
sendOpenAE		(AEDescList		list)
{
	OSStatus		err;
	AEAddressDesc	theAddress;
	AppleEvent		dummyReply;
	AppleEvent		theEvent;
	
	
	theAddress.descriptorType = typeNull;
	theAddress.dataHandle = null;

	do
	{
		ProcessSerialNumber		psn;
		
		
		err = GetCurrentProcess(&psn);
		if (err == noErr)
		{
			err = AECreateDesc(typeProcessSerialNumber, &psn, sizeof(ProcessSerialNumber), &theAddress);
			if (err == noErr)
			{
				dummyReply.descriptorType = typeNull;
				dummyReply.dataHandle = null;
				
				err = AECreateAppleEvent(kCoreEventClass, kAEOpenDocuments,
											&theAddress, kAutoGenerateReturnID,
											kAnyTransactionID, &theEvent);
				if (err == noErr)
				{
					err = AEPutParamDesc(&theEvent, keyDirectObject, &list);
					if (err == noErr)
					{
						err = AESend(&theEvent, &dummyReply, kAEWaitReply,
									kAENormalPriority, kAEDefaultTimeout, null,
									null);
					}
				}
			}
		}
	} while (err == noErr);
	
	return err;
}// sendOpenAE


/**
To automatically set up a Navigation Services
dialog options structure to contain a window
title consisting of the specified title, a
colon, and the name of the current process as
a client name, use this method.

On output, the "windowTitle" and "clientName"
fields of the NavDialogOptions structure are
filled in.

(1.0)
*/
static void
setUpWTitleProcessName		(NavDialogOptions*		inoutDataPtr,
							 ConstStringPtr			inTitle)
{
	if ((inoutDataPtr != null) && (inTitle != null))
	{
		Str255		titleCopy;
		
		
		PLstrcpy(titleCopy, inTitle);
		
		// %a = (given title), %b = (process name)
		PLstrcpy(inoutDataPtr->windowTitle, "\p%a: %b"); // LOCALIZE THIS
		BlockMoveData(LMGetCurApName(), inoutDataPtr->clientName, LMGetCurApName()[0] + 1);
		StringUtilities_PMetaSetDefaults();
		StringUtilities_PBuild(inoutDataPtr->windowTitle,
								titleCopy, inoutDataPtr->clientName, EMPTY_PSTRING, EMPTY_PSTRING);
	}
}// setUpWTitleProcessName