/*###############################################################

	NetEvents.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdio>
#	include <cstring>
#else
#	include <stdio.h>
#	include <string.h>
#endif

// Mac includes
#include <OpenTransport.h>
#include <OpenTransportProviders.h>
#include <Timer.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "Clipboard.h"
#include "ConnectionData.h"
#include "Connections.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "DNR.h"
#include "EventLoop.h"
#include "FTPBin.h"
#include "FTPLog.h"
#include "InternetServices.h"
#include "InternalEvents.h"
#include "MacLook.h"
#include "NetBackground.h"
#include "NetEvents.h"
#include "network.proto.h"
#include "Parse.h"
#include "RasterGraphicsScreen.h"
#include "RealScreen.h"
#include "TektronixRealGraphics.h"
#include "TelnetErrors.h"
#include "TelnetMenus.h"
#include "Terminal.h"
#include "TextTranslation.h"
#include "vsdata.h"
#include "vskeys.h"



#define USE_TIME_MANAGER_TASK_FOR_KEEP_ALIVE_PACKET_GENERATION		0

// constants
enum
{
	PFTP = 1,
	PRCP = 2,
	PDATA = 3,
	NEVENTS = 45
};

enum
{
	kKeepAlivePacketGenerationDelay = 2000/*30000*/ // 5 minutes, arbitrary - this may be configurable in the future
};

// types
struct InternalEvent
{
	// This structure records what happens, especially errors, and keeps them
	// for any routines that poll, looking for what happened.  The data field
	// is handled differently by each event type.
	Ptr							qLink;
	SInt16						qType;
	TelnetNetworkEventClass		eventClass;
	SInt32						event;
	SInt32						data1;
	SInt32						data2;
};
#ifndef __cplusplus
typedef struct InternalEvent	InternalEvent;
#endif
typedef InternalEvent*			InternalEventPtr;

// variables
static QHdr				gEventsQueue,
						gEventsFreeQueue;
#if USE_TIME_MANAGER_TASK_FOR_KEEP_ALIVE_PACKET_GENERATION
static TMTask			gKeepAlivePacketTask;
#endif
static Boolean			gEventQueueLocked = false; // used to control asynchronous network event servicing
static SInt16			gXferOn = 0;
static SInt16			gFileInTransit = 0;
static SInt16			gQueueLength = 0; // used to monitor queue length for debugging purposes
static SInt16			gQueueInUse = 0,
						gQueueError = 0; // set to 1 if the queue fills up
		
static unsigned char*	gReadBufferPtr = null; // main buffer space

// internal methods
static void					changeQueueLength					(SInt16						inDeltaLength);
static void					connectionClosedEvent				(SInt16						inWhichPort);
static void					connectionDataEvent					(SInt16						inWhichPort);
static void					connectionFailedEvent				(SInt16						inWhichPort);
static void					connectionOpenedEvent				(SInt16						inWhichPort);
static void					connectionSendKeepAliveEvent		(SInt16						inWhichPort);
static void					eventLock							(void);
static void					eventUnlock							(void);
static void					ftpPasteText						(ConnectionDataPtr			inoutTargetScreen);
static ConnectionDataPtr	getConnectionDataFromPort			(UInt16						inDesiredPort);
static void					getConnectionDataFromPortScreenOp	(ConnectionDataPtr const	inConnectionDataPtr,
																 RealScreenDataPtr const	inRealScreenDataPtr,
																 SInt32 const				inDesiredPort,
																 SInt32 const				inData2,
																 void*						inoutConnectionDataPtrPtr);
static SInt32				getNextNetEvent						(TelnetNetworkEventClass*	outEventClass,
																 SInt32*					outData1,
																 SInt32*					outData2);
static void					keepAliveScreenOp					(ConnectionDataPtr const	inConnectionDataPtr,
																 RealScreenDataPtr const	inRealScreenDataPtr,
																 SInt32 const				inData1,
																 SInt32 const				inData2,
																 void*						inoutResultPtr);
#if USE_TIME_MANAGER_TASK_FOR_KEEP_ALIVE_PACKET_GENERATION
static pascal void			keepAliveTimeManagerTask			(TMTaskPtr					inTaskPtr);
#endif
static void					setQueueLength						(SInt16						inNewLength);



//
// public methods
//

/*!
Initializes this module.  This procedure is
only called by "Init.c".

(3.0)
*/
void
NetEvents_Init	(void)
{
	gEventsFreeQueue.qHead = 0;
	gEventsFreeQueue.qTail = 0;
	gEventsFreeQueue.qFlags = 0;
	gEventsQueue.qHead = 0;
	gEventsQueue.qTail = 0;
	gEventsQueue.qFlags = 0;
	
	changeQueueLength(NEVENTS);
	setQueueLength(NEVENTS);
	
#if USE_TIME_MANAGER_TASK_FOR_KEEP_ALIVE_PACKET_GENERATION
	// install a deferred task that sends keep-alive packets to all open connections periodically
	gKeepAlivePacketTask.tmCount = 0L; // must be 0 at installation time
	gKeepAlivePacketTask.tmAddr = NewTimerUPP(keepAliveTimeManagerTask);
	(OSErr)InstallTimeTask((QElemPtr)&gKeepAlivePacketTask);
	(OSErr)PrimeTimeTask((QElemPtr)&gKeepAlivePacketTask,
							kKeepAlivePacketGenerationDelay/* a future release may make this configurable */);
#endif
}// Init


/*!
Performs jump scrolling for a particular screen.

(2.6)
*/
void
NetEvents_FlushNetworkForConnection		(ConnectionDataPtr		inConnectionDataPtr)
{
	short		count = 0;
	
	
	RealScreen_SetDrawingEnabled(inConnectionDataPtr->vs, false); // no output
	count = 1; // just needs to be positive to start with
	
	while (count > 0)
	{
		count = Network_GetData(inConnectionDataPtr->port, gReadBufferPtr, inConnectionDataPtr->blockSize);
		Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, gReadBufferPtr, count);
	}
	
	RealScreen_SetDrawingEnabled(inConnectionDataPtr->vs, true); // output now
	MacLook_InvalidatePortRectangle(inConnectionDataPtr->window);
}// FlushNetworkForConnection


/*!
To determine if a file transfer is in progress,
use this method.

(3.0)
*/
Boolean
NetEvents_IsFileTransferInProgress	(void)
{
	return gXferOn;
}// IsFileTransferInProgress


/*!
To add an event to the queue, call this routine.
This routine will probably get the memory for the
entry from the free list.  Returns 0 if there was
room, or 1 if an event was lost.

WARNING:	This routine is often used within
			interrupt-level code, because it is
			convenient to avoid complex code by
			posting events.  The implementation of
			this routine should not do anything
			that would be illegal at interrupt
			time.

(2.6)
*/
short
NetEvents_Post	(short		inEventClass,
				 short		inWhat,
				 short		inData1,
				 long		inData2)
{
	InternalEventPtr		theEvent = null;
	
	
	while ((theEvent = (InternalEventPtr)gEventsFreeQueue.qHead) != null)
	{
		if (Dequeue((QElemPtr)theEvent, &gEventsFreeQueue) == noErr) break;
	}
			
	if (theEvent == null)
	{
		gQueueError = 1; // queue is full!
		return(-1);
	}
	else gQueueInUse++;
	
	theEvent->qType = 0;	
	theEvent->eventClass = inEventClass;
	theEvent->event = inWhat;
	theEvent->data1 = inData1;
	theEvent->data2 = inData2;
	
	Enqueue((QElemPtr)theEvent, &gEventsQueue);
	return(0);
}// Post


/*!
To put a unique event into the queue, call this
method.  A search is done first to see if this
event information is already in the queue
somewhere.

WARNING:	This routine is often used within
			interrupt-level code, because it is
			convenient to avoid complex code by
			posting events.  The implementation of
			this routine should not do anything
			that would be illegal at interrupt
			time.

(2.6)
*/
short	
NetEvents_PostProvisionally	(short		inEventClass,
							 short		inWhat,
							 short		inData1,
							 long		inData2)
{
	InternalEventPtr	theEvent = (InternalEventPtr)gEventsQueue.qHead;
	short				result = 0;
	
	
	while ((theEvent != null) && ((theEvent->eventClass != inEventClass) || (theEvent->event != inWhat)
			|| (theEvent->data1 != inData1) || (theEvent->data2 != inData2)))
	{
		theEvent = (InternalEventPtr)theEvent->qLink;
	}
	if (theEvent == null) result = NetEvents_Post(inEventClass, inWhat, inData1, inData2);
	
	return result;
}// PostProvisionally


void
pasteText	(TelnetScreenID		inScreenID)
{
	ConnectionDataPtr	connectionDataPtr = ScreenManager_GetConnectionDataOf(inScreenID);
	short				amount = 0;
	
	
	if (connectionDataPtr->paste.outLength != 0)
	{
		if (Network_SendFlush(connectionDataPtr->port) != 0)
		{
			NetEvents_Post(kTelnetNetworkEventClassUser, kUserEventPasteText, inScreenID, 0);
			return;
		}
		if (connectionDataPtr->paste.inCount)
		{
			connectionDataPtr->paste.inCount = 0;
			connectionDataPtr->paste.outCount = 0;
			NetEvents_Post(kTelnetNetworkEventClassUser, kUserEventPasteText, inScreenID, 0);
			return;
		}
		if (connectionDataPtr->paste.outCount < 2)
		{
			connectionDataPtr->paste.outCount++;
			NetEvents_Post(kTelnetNetworkEventClassUser, kUserEventPasteText, inScreenID, 0);
			return;
		}
		if (Network_GetNumberOfBytesWaitingOnPort(connectionDataPtr->port) > 0)
		{
			NetEvents_Post(kTelnetNetworkEventClassUser, kUserEventPasteText, inScreenID, 0);
			return;
		}
		
		if (connectionDataPtr->ftpState != 0) ftpPasteText(connectionDataPtr);
		else
		{
			if (connectionDataPtr->paste.method == kTelnetPasteMethodStandard)
			{
				// do it all at once
				amount = Network_SendData(connectionDataPtr->port, connectionDataPtr->paste.nextCharPtr,
											connectionDataPtr->paste.outLength);
			}
			else
			{
				// do it in blocks
				if (connectionDataPtr->paste.blockSize <= connectionDataPtr->paste.outLength)
				{
					amount = connectionDataPtr->paste.blockSize;
				}
				else
				{
					amount = connectionDataPtr->paste.outLength;
				}
				
				amount = Network_SendData(connectionDataPtr->port, connectionDataPtr->paste.nextCharPtr, amount);
			}
		}
			
		if (connectionDataPtr->echo)
		{
			Parse_ShowOrInterpretAsCommand(connectionDataPtr,
											(unsigned char*)connectionDataPtr->paste.nextCharPtr, amount);
		}
		
		connectionDataPtr->paste.outLength -= amount;	
		connectionDataPtr->paste.nextCharPtr += (long)amount;
		
		if (connectionDataPtr->paste.outLength <= 0)
		{
			int		left = 0;
			
			
			connectionDataPtr->clientflags &= ~PASTE_IN_PROGRESS;
			HUnlock(connectionDataPtr->paste.text);
			Memory_DisposeHandle(&(connectionDataPtr->paste.text));
			connectionDataPtr->paste.nextCharPtr = (char*)0L;	
			connectionDataPtr->paste.text = (char**)0L;	
			
			// this next loop is to unload the queue of stuff typed during the paste
			for (left = 0; left < connectionDataPtr->kblen; left++) 
			{
				if ((connectionDataPtr->kbbuf[left]) == 0) // this signals a VSkey
				{
					left++; // send the VS key
					VSkbsend(connectionDataPtr->vs, (unsigned char)connectionDataPtr->kbbuf[left],
								connectionDataPtr->echo);
				}
				else
				{
					Network_SendFlush(connectionDataPtr->port);
					if (connectionDataPtr->kbbuf[left + 1] != 0x15) // not a KILLCHAR
					{	
						if (connectionDataPtr->kbbuf[left] == '\015') // remap CR-> CR LF
						{
							// 3.0 - take into account the "crmap" setting when pasting, as well
							Network_SendFlushData(connectionDataPtr->port,
													(connectionDataPtr->crmap) ? "\015\0" : "\015\012", 2);
							if (connectionDataPtr->echo)
							{
								Parse_ShowOrInterpretAsCommand(connectionDataPtr, (unsigned char*)"\012\015", 2);
							}
							return;
						}
						Network_SendData(connectionDataPtr->port, &(connectionDataPtr->kbbuf[left]), 1);
						if (connectionDataPtr->echo && (connectionDataPtr->kbbuf[left] > 31) &&
							(connectionDataPtr->kbbuf[left] < 127))
						{
							Parse_ShowOrInterpretAsCommand(connectionDataPtr,
															(unsigned char*)&(connectionDataPtr->kbbuf[left]), 1);
						}
						else if (connectionDataPtr->echo)
						{
							connectionDataPtr->kbbuf[left] += '@';
							Parse_ShowOrInterpretAsCommand(connectionDataPtr, (unsigned char*)"^", 1);
							Parse_ShowOrInterpretAsCommand(connectionDataPtr,
															(unsigned char*)&(connectionDataPtr->kbbuf[left]), 1);
						}
					}
					else left++; // user cancelled one keystroke with another
				}
			}
			
			// now remove the progress indicator in the window header
			{
				RealScreenDataPtr	realScreenPtr = ScreenManager_GetRealScreenDataOf(inScreenID);
				
				
				if (realScreenPtr != null)
				{
					if (realScreenPtr->windowHeader != null)
					{
						// remove window header progress indicator
						TerminalWindowHeader_SetChasingArrowsVisible(realScreenPtr->windowHeader, false);
					}
				}
			}
		}
		else NetEvents_Post(kTelnetNetworkEventClassUser, kUserEventPasteText, inScreenID, 0);
	}
}// pasteText


/*!
This routine is called continuously, from
various locations in the program, to make
sure MacTelnet network events are serviced.

(2.6)
*/
void
NetEvents_Service	(void)
{
#if !USE_TIME_MANAGER_TASK_FOR_KEEP_ALIVE_PACKET_GENERATION
	enum
	{
		// a tick is 1/60th of a second; calculate a LONG delay here, on the order of *minutes*,
		// to specify how frequently MacTelnet should fire off a keep-alive packet (a no-op);
		// NOTE: a future release should make this delay configurable in preferences (or for
		// that matter, a flag for whether MacTelnet should even CARE about keep-alive packets)
		kKeepAliveDelayInTicks = 60 * 60 * 5
	};
	static UInt32		oldTicks = 0L;
	
	
	// TEMP - note that this may be inefficient; a Mac OS Time Manager task would be a MUCH
	// better idea here, since 99.999999997% of the time this condition is going to fail!
	if ((TickCount() - oldTicks) > kKeepAliveDelayInTicks)
	{
		// every so many minutes, post a keep alive event for each open connection
		ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagActiveSessions, keepAliveScreenOp,
										0L/* data 1 - unused */, 0L/* data 2 - unused */, null/* result - unused */);
		oldTicks = TickCount();
	}
#endif
	
	unless (gEventQueueLocked)
	{
		SInt32						pos = 0L;
		SInt32						ftpTime = 0L;
		TelnetNetworkEventClass		eventClass = 0;
		SInt32						event = 0;
		SInt32						data1 = 0L,
									data2 = 0L;
		
		
		eventLock();
		if ((event = getNextNetEvent(&eventClass, &data1, &data2)) >= 0)
		{
			if (((TickCount() - ftpTime) > (60 * 2)) && gFileInTransit)
			{
				Console_WriteLine("ftp progress");
				ftpTime = TickCount();
				Sftpstat(&pos); // get transfer status
				// LOCALIZE THIS
				if (pos <= 0) FTPLog_WriteLine("FTP Status: transferring");
				else
				{
					char*	temp = Memory_NewPtr(50); // enough room for this string
					
					
					if (temp != null)
					{
						// LOCALIZE THIS
						if (gFileInTransit + 2) sprintf((char*)temp, "FTP Status: %ld bytes remaining.", pos);
						else sprintf((char*)temp, "FTP Status: %ld bytes transferred.", pos);
						
						FTPLog_WriteLine((char*)temp);
						Memory_DisposePtr(&temp);
					}
				}
			}
			
			switch (eventClass)
			{
				case kTelnetNetworkEventClassSession:
					switch (event)
					{
						case kSessionEventFTPTransferActive:
							Console_WriteLine("Event: FTP transfer active.");
							ftpd(0, data1);
							break;
						
						#if USE_TIME_MANAGER_TASK_FOR_KEEP_ALIVE_PACKET_GENERATION
						case kSessionEventSendKeepAlivePackets:
							Console_WriteLine("Event: Time to send keep-alive packets.");
							ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagActiveSessions,
															keepAliveScreenOp, 0L/* data 1 - unused */,
															0L/* data 2 - unused */, null/* result - unused */);
							break;
						#endif
						
						// these are not needed by the Open Transport implementation
						case kSessionEventCloseFinished: // used in the drivers
						case kSessionEventCloseFinishedWithError: // used in the drivers
							Console_WriteLine("Event: Session closed.");
							Network_CloseConnection(data1);
							break;
						
						default:
							break;
					}
					break;
				
				case kTelnetNetworkEventClassConnection: // connection type event
					switch(Network_GetPortType(data1))
					{
						case kNetworkPortTypeFTPClientSession:
							// FTP client
							userftpd(event, data1);
							break;
						
						case kNetworkPortTypeFTPServerSession:
							// FTP server connection itself
							rftpd(event, data1);
							break;
						
						case kNetworkPortTypeFTPServerData:
							// FTP server data connection (on FTP data port)
							ftpd(event, data1);
							break;
						
						case kNetworkPortTypeStandardClientSession:
						default:
							// telnet or SSH client
							switch (event)
							{
								case kConnectionEventOpen: // connection opened
									Console_WriteTimeStamp("Event: Connection opened.");
									connectionOpenedEvent(data1);
									break;
								
								case kConnectionEventData: // data arrived
									connectionDataEvent(data1);
									break;
								
								case kConnectionEventFailure:
									Console_WriteTimeStamp("Event: Connection failure!");
									connectionFailedEvent(data1);
									break;
								
								case kConnectionEventClose: // connection closed
									Console_WriteTimeStamp("Event: Connection closed.");
									connectionClosedEvent(data1);
									break;
								
								case kConnectionEventKeepAlive:
									// send a keep-alive packet to the connection
									Console_WriteLine("Event: Connection keep-alive packet generation.");
									connectionSendKeepAliveEvent(data1);
									break;
								
								default:
									break;
							}
							break;
					}
					break;
			
				case kTelnetNetworkEventClassUser:
					switch (event)
					{
						case kUserEventDomainNameLookupBegin:
							Console_WriteTimeStamp("Event: DNR lookup initiated.");
							{
								DomainNameLookupDataPtr		resolverPtr = (DomainNameLookupDataPtr)data2;
								
								
								resolverPtr->error = OTInetStringToAddress(resolverPtr->servicesRef,
																			(char*)resolverPtr->hostName,
																			resolverPtr->hostInfoPtr);
								if (resolverPtr->error == kOTStateChangeErr)
								{
									// state changing - need to try again
									Console_WriteLine("retrying, provider state is currently changing");
									NetEvents_Post(kTelnetNetworkEventClassUser, kUserEventDomainNameLookupBegin,
													data1, data2);
								}
								#if 0
								// is this redundant?
								if (resolverPtr->error != kOTNoError)
								{
									NetEvents_Post(kTelnetNetworkEventClassUser, kUserEventDomainNameLookupEnd,
													(TelnetScreenID)resolverPtr->data, (SInt32)resolverPtr);
								}
								#endif
							}
							break;
						
						case kUserEventDomainNameLookupEnd:
							Console_WriteTimeStamp("Event: DNR lookup completed.");
							{
								DomainNameLookupDataPtr		resolverPtr = (DomainNameLookupDataPtr)data2;
								
								
								if (resolverPtr->dnrCompletionMethod != null)
								{
									InvokeDNRDoneNotificationProc(resolverPtr->dnrCompletionMethod,
																	(TelnetScreenID)data1,
																	resolverPtr);
								}
								else Console_WriteLine("Warning: No DNR lookup completion routine (why did this happen?).");
							}
							break;
						
						case kUserEventFTPTransferBegin:
							Console_WriteTimeStamp("Event: FTP transfer initiated.");
							gFileInTransit = data1;
							ftpTime = TickCount();
							break;
						
						case kUserEventFTPTransferEnd:
							Console_WriteTimeStamp("Event: FTP transfer completed.");
							// LOCALIZE THIS
							FTPLog_WriteLine("FTP Transfer Concluding");
							FTPLog_WriteLine("");
							gFileInTransit = 0;
							break;
						
						case kUserEventFTPConnectionOpen:
							Console_WriteTimeStamp("Event: FTP connection opened.");
							{
								ip_addrbytes	ftpinfo;
								char*			temp = null;
								
								
								if (gFTPServerPrefs->resetMacBinary)
								{
									FlagManager_Set(kFlagMacBinaryIIEnabled, gFTPServerPrefs->useMacBinaryII);
								}
								
								gXferOn = 1;
								
								FTPLog_WriteHorizontalRule();
								// LOCALIZE THIS
								FTPLog_WriteString("FTP server initiated from host: ");
								temp = Memory_NewPtr(80);
								ftpinfo.a.addr = Sftphost();
								
								if (temp != null) // then theres enough memory to tell the log about the FTP session
								{
									UInt16		ipOctets[4];
									
									
									ipOctets[0] = ftpinfo.a.byte[0];
									ipOctets[1] = ftpinfo.a.byte[1];
									ipOctets[2] = ftpinfo.a.byte[2];
									ipOctets[3] = ftpinfo.a.byte[3];
									
									sprintf((char*)&temp[4],"%u.%u.%u.%u\015\012",
											(unsigned char)ipOctets[0],
											(unsigned char)ipOctets[1],
											(unsigned char)ipOctets[2],
											(unsigned char)ipOctets[3]);
									FTPLog_WriteString((char*)&temp[4]);
									Memory_DisposePtr(&temp);
								}
							}
							break;
						
						case kUserEventFTPConnectionClose:
							Console_WriteTimeStamp("Event: FTP connection closed.");
							gXferOn = 0;
							if (gFTPServerPrefs->resetMacBinary)
							{
								FlagManager_Set(kFlagMacBinaryIIEnabled, gFTPServerPrefs->useMacBinaryII);
							}
							// LOCALIZE THIS
							FTPLog_WriteLine("FTP server ending session");
							FTPLog_WriteHorizontalRule();
							FTPLog_WriteLine("");
							break;
						
						case kUserEventRealGraphicsRedraw:
							if (VGalive(data1) && RGsupdate(data1)) TekEnable(data1);
							break;
						
						case kUserEventPasteText:
							{
								RealScreenDataPtr	realScreenPtr = ScreenManager_GetRealScreenDataOf(data1/* screen ID */);
								
								
								if (realScreenPtr != null)
								{
									if (realScreenPtr->windowHeader != null)
									{
										// display window header progress indicator (removed in pasteText())
										TerminalWindowHeader_SetChasingArrowsVisible(realScreenPtr->windowHeader, true);
									}
								}
								
								// paste the text (which could take a while)
								pasteText(data1/* screen ID */);
							}
							break;
						
						default:
							break;
					}
					break;
				
				default:
					break;
			}
		}
		eventUnlock();
	}
}// Service


short
NetEvents_SetBlockSize	(SInt16					inNewSize,
						 ConnectionDataPtr		inConnectionDataPtr)
{
	SInt16		currentSize = 0;
	SInt16		result = 0;
	
	
	inConnectionDataPtr->blockSize = (inNewSize < 512) ? 512 : inNewSize;
	
	if (gReadBufferPtr == null)
	{
		gReadBufferPtr = (unsigned char*)Memory_NewPtrCritical(inNewSize + 100); 
	}
	
	if (gReadBufferPtr != null)
	{
		currentSize = (short) GetPtrSize((Ptr)gReadBufferPtr);
	
		if (currentSize < (inNewSize + 100)) // increase buffer size, if possible
		{
			Memory_DisposePtr(&((Ptr)gReadBufferPtr));
			gReadBufferPtr = (unsigned char*)Memory_NewPtrCritical(inNewSize + 100);
			if (gReadBufferPtr == null)
			{
				// try to get previous size
				gReadBufferPtr = (unsigned char*)Memory_NewPtrCritical(currentSize);
				if (gReadBufferPtr == null) result = -1;
				else inConnectionDataPtr->blockSize = currentSize;
			}
		}
	}
	else result = -1;
	
	return result;
}// SetBlockSize


//
// internal methods
//

/*!
Every time a network connection is opened or closed,
a few elements are added to or removed from the queue
to ensure that there will always be some free ones
lying around.  This avoids Dequeue()ing at interrupt
time, which causes nasty problems since the method
getNextNetEvent() walks the queue at non-interrupt
time.

(2.6)
*/
static void
changeQueueLength		(SInt16		inDeltaLength)
{
	InternalEventPtr	theEvent = null;
	short				delta = inDeltaLength;
	
	
	if (delta > 0)
	{
		while (delta != 0)
		{
			theEvent = (InternalEventPtr)Memory_NewPtrCritical(sizeof(InternalEvent));
			Enqueue((QElemPtr)theEvent, &gEventsFreeQueue);
			setQueueLength(gQueueLength + 1);
			delta--;
		}
	}
	else
	{
		while (delta != 0)
		{
			theEvent = (InternalEventPtr)gEventsFreeQueue.qHead;
			Dequeue((QElemPtr)theEvent, &gEventsFreeQueue);
			if (theEvent) Memory_DisposePtr(&((Ptr)theEvent));
			setQueueLength(gQueueLength - 1);
			delta++;
		}
	}
}// changeQueueLength


/*!
Performs appropriate operations in response to
a connection close event.

(2.6)
*/
static void
connectionClosedEvent	(SInt16		inWhichPort)
{
	ConnectionDataPtr	connectionDataPtr = ScreenManager_GetConnectionDataFromPort(inWhichPort);
	
	
	if (connectionDataPtr == null)
	{
		// error - stop here
		Network_CloseConnection(inWhichPort); // close again...
	}
	else
	{
		// close connection appropriately for its port type
		if (Network_GetPortType(inWhichPort) == kNetworkPortTypeFTPClientSession)
		{
			userftpd(kConnectionEventClose, inWhichPort);
		}
		else
		{
			// ensure any buffered bytes get displayed prior to disconnect
			NetEvents_FlushNetworkForConnection(connectionDataPtr);
			
			// close the connection
			Connections_Disconnect(connectionDataPtr);
		}
	}
}// connectionClosedEvent


/*!
Performs appropriate operations in response to
a data arrived event for a session connection.

(2.6)
*/
static void
connectionDataEvent		(SInt16		inWhichPort)
{
	static short		eventCount = 0;
	ConnectionDataPtr	connectionDataPtr = ScreenManager_GetConnectionDataFromPort(inWhichPort);
	
	
	if (connectionDataPtr != null)
	{
		// 3.0 - jump scrolling at regular intervals
		//if ((Connections_GetCount() >= 1) && (++eventCount > 80))
		//{
		//	NetEvents_FlushNetworkForConnection(connectionDataPtr), eventCount = 0;
		//}
		
		// notify the user that data has appeared if this was a watched window (wascawy wabbit)
		// (this code should be changed to use a floating Notification Manager window, probably)
		if (connectionDataPtr->watched)
		{
			connectionDataPtr->watched = false;
			Errors_DisplayNoteMessage(rStringsNoteAlerts, siActivityDetected, -1);
		}
		
		if (FlagManager_Test(kFlagNetworkSuspended) || (!connectionDataPtr->enabled))
		{
			NetEvents_PostProvisionally(kTelnetNetworkEventClassConnection, kConnectionEventData,
										connectionDataPtr->vs, 0);
		}
		else
		{
			// handle FTP client stuff if necessary
			if (Network_GetPortType(inWhichPort) == kNetworkPortTypeFTPClientSession)
			{
				userftpd(kConnectionEventData, inWhichPort);
			}
			else
			{
				SInt16		count = 0;
				
				
				count = Network_GetData(inWhichPort, gReadBufferPtr, connectionDataPtr->blockSize);
				Parse_ShowOrInterpretAsCommand(connectionDataPtr, gReadBufferPtr, count);
				connectionDataPtr->paste.inCount += count; // what does this do?
			}
		}
	}
}// connectionDataEvent


/*!
Performs appropriate operations in response to a
connection failed event for a session connection.

(2.6)
*/
static void
connectionFailedEvent		(SInt16		inWhichPort)
{
	// failure cleanup is only appropriate for normal types of connections
	if (Network_GetPortType(inWhichPort) == kNetworkPortTypeStandardClientSession)
	{
		ConnectionDataPtr		connectionDataPtr = null;
		
		
		// clean up the failed connection
		connectionDataPtr = ScreenManager_GetConnectionDataFromPort(inWhichPort);
		if (connectionDataPtr != null)
		{
			InterfaceLibAlertRef	box = null;
			ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
			Str255					dialogText,
									helpText,
									alertTitle;
			
			
			// display an alert telling the user what went wrong
			GetIndString(dialogText, rStringsNetworkErrors, siNetErrorHostOrGatewayHasStillNotResponded);
			GetIndString(helpText, rStringsNetworkErrors, siNetErrorHostOrGatewayHasStillNotRespondedHelpText);
			GetIndString(alertTitle, rStringsAlertTitles, siAlertTitleConnectionOpeningFailed);
			StringUtilities_PMetaSetDefaults();
			StringUtilities_PBuild(dialogText, connectionDataPtr->host.name, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
			box = Alert_New();
			Alert_SetFilterProc(box, upp);
			Alert_SetHelpButton(box, false);
			Alert_SetParamsFor(box, kAlert_StyleOK);
			Alert_SetTitle(box, alertTitle);
			Alert_SetText(box, dialogText, helpText);
			Alert_SetType(box, kAlertNoteAlert);
			Alert_Display(box);
			Alert_Dispose(&box);
			DisposeModalFilterUPP(upp), upp = null;
			
			Connections_Disconnect(connectionDataPtr);
		}
	}
}// connectionFailedEvent


/*!
Performs appropriate operations in response to
a connection open event.

(2.6)
*/
static void
connectionOpenedEvent		(SInt16		inWhichPort)
{
	ConnectionDataPtr	connectionDataPtr = ScreenManager_GetConnectionDataFromPort(inWhichPort);
	
	
	if (connectionDataPtr != null)
	{
		Connections_MakeActive(connectionDataPtr);
		
		// only send telnet options for ports that could be using the telnet protocol;
		// all ports less than 1024 are reserved, of which only one (23) can be "telnet";
		// this port restriction allows MacTelnet to successfully connect to any port
		// (non-telnet servers would previously get confused by receiving telnet options)
		if ((connectionDataPtr->port == 23) ||
			(connectionDataPtr->port >= 1024)) Parse_SendInitialTelnetOptions(connectionDataPtr);
		
		// handle FTP client stuff if necessary
		if (Network_GetPortType(inWhichPort) == kNetworkPortTypeFTPClientSession)
		{
			userftpd(kConnectionEventOpen, inWhichPort);
		}
	}
	else BEEP;
}// connectionOpenedEvent


/*!
Keeps a connection alive by sending a NOP back to
the server through the given port.

(3.0)
*/
static void
connectionSendKeepAliveEvent	(SInt16		inWhichPort)
{
	Network_SendData(inWhichPort, "\0361", 1); // send a NOP, which can be sent anytime without affecting the protocol
}// connectionSendKeepAliveEvent


/*!
Locks the event loop, rendering future calls to
NetEvents_Service() useless.  This is important
within event handlers.  Use eventUnlock() to
release the lock.

(3.0)
*/
static void
eventLock	(void)
{
	gEventQueueLocked = true;
}// eventLock


/*!
Unlocks the event loop, meaning that future calls
to NetEvents_Service() once again have effect.
Use eventLock() to acquire a lock.

(3.0)
*/
static void
eventUnlock	(void)
{
	gEventQueueLocked = false;
}// eventUnlock


/*!
Handles pasting operations specific to FTP client
sessions.

(2.6)
*/
static void
ftpPasteText	(ConnectionDataPtr		inoutTargetScreen)
{
	char*		ascii = null;
	
	
	ascii = inoutTargetScreen->paste.nextCharPtr;
	while ((inoutTargetScreen->paste.outLength > 0) && (*ascii > 31) && (*ascii < 127))
	{
		Parse_ShowOrInterpretAsCommand(inoutTargetScreen, (unsigned char*)ascii, 1);
		inoutTargetScreen->kbbuf[inoutTargetScreen->kblen++] = *ascii;
		inoutTargetScreen->paste.outLength--;
		ascii++;
	}
	inoutTargetScreen->paste.nextCharPtr = ascii;
	if (*ascii == '\015')
	{
		Parse_ShowOrInterpretAsCommand(inoutTargetScreen, (unsigned char*)"\015\012", 2);
		inoutTargetScreen->kbbuf[inoutTargetScreen->kblen++] = 0;
		ftppi(inoutTargetScreen->kbbuf); // FTP client
		inoutTargetScreen->kblen = 0;
	}
	while ((inoutTargetScreen->paste.outLength > 0) && ((*ascii <= 31) || (*ascii >= 127)))
	{
		inoutTargetScreen->paste.outLength--;
		inoutTargetScreen->paste.nextCharPtr++;
	}
}// ftpPasteText


/*!
This method returns a pointer to the connection
data structure for the connection occupying the
given port number, if any.  If any problems
occur, null is returned.

(3.0)
*/
static ConnectionDataPtr
getConnectionDataFromPort		(UInt16		inDesiredPort)
{
	ConnectionDataPtr	result = null;
	
	
	ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagInUseSessions/* examine active or opening sessions only */,
									getConnectionDataFromPortScreenOp, (SInt32)inDesiredPort, 0L/* undefined */,
									&result);
	return result;
}// getConnectionDataFromPort


/*!
This method is of standard "ScreenOperationProcPtr" form.
It determines whether the specified screens connection
uses the specified port number ("inData1").  The value of
"inData2" is not defined, and should be 0.

You should provide storage for a "ConnectionDataPtr", such
that the actual type of "inoutConnectionDataPtrPtr" is a
pointer to a "ConnectionDataPtr".

(3.0)
*/
static void
getConnectionDataFromPortScreenOp	(ConnectionDataPtr const	inConnectionDataPtr,
									 RealScreenDataPtr const	UNUSED_ARGUMENT(inRealScreenDataPtr),
									 SInt32 const				inDesiredPort,
									 SInt32 const				UNUSED_ARGUMENT(inData2),
									 void*						inoutConnectionDataPtrPtr)
{
	if ((inoutConnectionDataPtrPtr != null) && (inConnectionDataPtr != null))
	{
		if (inConnectionDataPtr->port == inDesiredPort)
		{
			*((ConnectionDataPtr*)inoutConnectionDataPtrPtr) = inConnectionDataPtr;
		}
	}
}// getConnectionDataFromPortScreenOp

			
/*!
Retrieves the next event (and clears it).  Returns
the event number, or -1 if no event is present.
The exact class and the associated integer is
returned as well.

(2.6)
*/
static SInt32
getNextNetEvent	(TelnetNetworkEventClass*		outEventClass,
				 SInt32*						outData1,
				 SInt32*						outData2)
{
	InternalEventPtr	theEvent = null;
	static SInt32		timer = 0L;
	SInt32				result = -1;
	
	
	if (TickCount() - timer > 60*30) timer = TickCount();
	
	// Spin until the end of the queue is hit or if Dequeue() is used on an element that is
	// not already dequeued.  If this check is not done on Dequeue(), someone else could
	// come in and dequeue an element that is about to get dequeued.
	while ((theEvent = (InternalEventPtr)gEventsQueue.qHead) != null)
	{
		if (Dequeue((QElemPtr)theEvent, &gEventsQueue) == noErr) break;
	}
		
//	while ((theEvent != null) && !(theEvent->eventClass & mask))
//		theEvent = (InternalEventPtr)theEvent->qLink;
	
	if (theEvent != null)
	{
//	(void) Dequeue((QElemPtr)theEvent, &gEventsQueue);
		*outEventClass = theEvent->eventClass;
		*outData1 = theEvent->data1;
		*outData2 = theEvent->data2;
		result = theEvent->event;
		
		Enqueue((QElemPtr)theEvent, &gEventsFreeQueue);
		gQueueInUse--;
	}
	
	return result;
}// getNextNetEvent


/*!
This method is of standard "ScreenOperationProcPtr" form.
It issues a keep alive event for the specified screen.

The values of "inData1", "inData2" and "inoutResultPtr"
are reserved.

(3.0)
*/
static void
keepAliveScreenOp		(ConnectionDataPtr const	inConnectionDataPtr,
						 RealScreenDataPtr const	UNUSED_ARGUMENT(inRealScreenDataPtr),
						 SInt32 const				UNUSED_ARGUMENT(inData1),
						 SInt32 const				UNUSED_ARGUMENT(inData2),
						 void*						UNUSED_ARGUMENT(inoutResultPtr))
{
	if (inConnectionDataPtr != null)
	{
		NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventKeepAlive,
						inConnectionDataPtr->vs, 0/* unused */);
	}
}// keepAliveScreenOp


/*!
This is a standard Time Manager task, called
periodically to ensure all connections are sent
keep alive packets.

WARNING:	This routine is executed at interrupt
			time, so there are many things it
			cannot do safely.  Keep it simple!

(3.0)
*/
#if USE_TIME_MANAGER_TASK_FOR_KEEP_ALIVE_PACKET_GENERATION
static pascal void
keepAliveTimeManagerTask	(TMTaskPtr		inTaskPtr)
{
	NetEvents_Post(kTelnetNetworkEventClassSession, kSessionEventSendKeepAlivePackets,
					inConnectionDataPtr->vs, 0/* unused */);
	
	// to keep the task periodic, insert the task back in the queue
	(OSErr)PrimeTimeTask((QElemPtr)inTaskPtr,
							kKeepAlivePacketGenerationDelay/* a future release may make this configurable */);
}// keepAliveTimeManagerTask
#endif


/*!
Sets the network event queue size (for debugging
purposes).

(3.0)
*/
static void
setQueueLength	(SInt16		inNewLength)
{
	gQueueLength = inNewLength;
}// setQueueLength