/*###############################################################

	Network.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#define OPEN_TRANSPORT_NETWORK_IMPLEMENTATION 0

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdlib>
#	include <cstring>
#else
#	include <stdlib.h>
#	include <string.h>
#endif

// Mac includes
#if !OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
#	include <Devices.h>
#	include <MacTCP.h>
#endif
#include <MacTypes.h>
#include <Scrap.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "MenuResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "Commands.h"
#include "ConnectionData.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "InternalEvents.h"
#include "InternetServices.h"
#include "MacLook.h"
#include "NetEvents.h"
#include "network.proto.h"
#include "PreferencesData.h"
#include "TelnetErrors.h"
#include "TelnetMenus.h"



// constants
#if !OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
#define MAX_FDS_ELEMS 32
#define MAX_SWDS_ELEMS 16
#define TCPBUFSIZ INTEGER_KILOBYTES(8)
#endif

// types
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION

struct NetworkPortContext
{
	NetworkPortRef		portNumber;			// the external reference to this particular port; ALWAYS THE ARRAY INDEX
	NetworkPortType		portType;			// type of port
	OSStatus			error;				// on completion, any error that occurred
	InetSvcRef			servicesRef;		// reference to Open Transport Internet Provider
	EndpointRef			endpointRef;		// reference to Open Transport Endpoint
	OTNotifyUPP			notifierUPP;		// the callback routine, how Open Transport communicates with this module
	TEndpointInfo		endpointInfo;		// current endpoint data; can fluctuate until connection is established
	Ptr					authEncrData;		// Kerberos authentication and encryption data
	
	struct
	{
		TCall			call;				// the remote host & port, options, connection ID, etc. of the destination
		OTFlags			flags;				// send flags
		TNetbuf			normalData;			// data ready to be pushed out
		TNetbuf			expeditedData;		// expedited data ready to be pushed out
	} send;
	
	struct
	{
		TCall			call;				// the remote host & port, options, connection ID, etc. of the source
		OTFlags			flags;				// receive flags
		TNetbuf			normalData;			// data ready to be retrieved
		TNetbuf			expeditedData;		// expedited data ready to be retrieved
	} receive;
};
#ifndef __cplusplus
typedef struct NetworkPortContext	NetworkPortContext;
#endif
typedef NetworkPortContext*			NetworkPortContextPtr;

#else

struct exfds
{
	short		inuse;					/* is this being used? */
	wdsEntry	fds;					/* the real data */
};
typedef struct exfds	exfds;

struct StreamRec // 470 bytes
{
	StreamPtr	stream;					/* Apple's lovely Stream Pointer */
	char*		buffer;					/* where the immovable TCP buffer is */
	short		push;					/* "true" if we should push next data block */
	char*		sbuffer;				/* where the send buffer is */
	wdsEntry	fds[MAX_FDS_ELEMS];		/* Free Data Structure list */
	exfds		exFDS[MAX_FDS_ELEMS];	/* exFDS entries */
	short		maxFDSused;				/* max. of the FDS's that have been used */
	short		portType;				/* type of port */
	Ptr			aedata;					/* pointer to encryption data */
};
typedef struct StreamRec	StreamRec;
typedef StreamRec*			StreamRPtr;

struct MyTCPpb // 106 Bytes
{
	TCPiopb		pb;
	long		SavedA5;
};
typedef struct MyTCPpb	MyTCPpb;
typedef MyTCPpb*		MyTCPpbPtr;

#endif

// variables
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
static OTNotifyUPP				gGenericNotifyUPP = null;
static NetworkPortContextPtr	gPortContexts[kNetworkMaximumNumberOfPorts];
#else
static TCPIOCompletionUPP 	OpenCompleteInitUPP = null;
static TCPIOCompletionUPP 	CloseCompleteInitUPP = null;
static TCPIOCompletionUPP 	SendCompleteInitUPP = null;
static TCPNotifyUPP			NotifyUPP = null;
static StreamRPtr			gStreams[kNetworkMaximumNumberOfPorts];
#endif
static Boolean				isTCPRunning = false;	// flag to record whether Network_Connect succeeded
#if !OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
static short				gTCP_IPDriverRefNum = 0; // reference number of TCP driver
#endif
static SInt8				gSendTimeout = 0;
static QHdr					gFreePBQueue;
static QHdr					gFreeSWDSQueue;
Boolean 	nextIsUrgent = false;

// internal methods
static short					abortConnection				(NetworkPortRef				inPortNumber);
static void						ensureTCPIsOpen				(void);
static Boolean					isValidPortNumber			(NetworkPortRef				inPortNumber);

#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
static void						disposeNetworkDataPtr		(NetworkPortContextPtr*		inoutContextPtr);
static Boolean					endpointSupportsTSDUs		(NetworkPortContextPtr		inoutContextPtr,
															 Boolean const				inExpeditedDataUnits);
static OTDataSize				getDataUnitSize				(NetworkPortContextPtr		inoutContextPtr,
															 OTDataSize const			inWantThisByteCount,
															 Boolean const				inExpeditedDataUnits);
static NetworkPortContextPtr	getNetworkPortContextFor	(NetworkPortRef				inPortNumber);
static NetworkPortContextPtr	newNetworkDataPtr			(InetAddress const*			inRemoteSendAddress,
															 NetworkPortRef*			outLocalPortNumberPtrOrNull);
static pascal void	openTransportGenericNotify				(void*						inoutContextPtr,
															 OTEventCode				inEventCode,
															 OTResult					inResult,
															 void*						inCookie);
static void			openTransportHandleDisconnect			(NetworkPortContextPtr		inoutContextPtr,
															 OTEventCode				inEventCode);
static void			openTransportHandleListen				(NetworkPortContextPtr		inoutContextPtr);
static void			openTransportHandleLook					(NetworkPortContextPtr		inoutContextPtr);
#else
static void OpenCompleteInit(TCPiopb *pbp);
static	void	OpenComplete(MyTCPpbPtr pbp);
static void SendCompleteInit(TCPiopb *pbp);
static	void	SendComplete(MyTCPpbPtr pbp);
static void CloseCompleteInit(TCPiopb *pbp);
static	void	CloseComplete(MyTCPpbPtr pbp);

pascal void Notify(StreamPtr streamPtr, unsigned short code, Ptr uptr, 
						unsigned short terminReason, struct ICMPReport *icmpMsg);

static NetworkPortRef	findbystream(StreamPtr streamPtr);
static SInt16		makeStream	(void);
static	wdsEntry	*getSWDS(void);
static	void		returnSWDS(wdsEntry *wds);
static	void		MyPBreturn(MyTCPpbPtr pbp);
static	MyTCPpb		*getPB(short driver,short call, StreamPtr stream, short usernum);
static	void		clearPB(MyTCPpbPtr pbp, short driver, short call, StreamPtr stream);
static	OSErr		xPBControlSync(MyTCPpbPtr pbp);
static	OSErr		xPBControlAsync(MyTCPpbPtr pbp, TCPIOCompletionUPP completionUPP);
static	void		reclaim(StreamRPtr p);
static	short		giveback(StreamRPtr p, wdsEntry *wds);
static	void		compressfds(wdsEntry *fds);

#endif



//
// public methods
//

/*!
Initializes the Network module.  Call Network_Done()
when you are completely done with the Network module.

(2.6)
*/
void
Network_Init		(void)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	register SInt16		i = 0;
	
	
	gGenericNotifyUPP = NewOTNotifyUPP(openTransportGenericNotify);
	for (i = 0; i < kNetworkMaximumNumberOfPorts; i++) gPortContexts[i] = null; // these are allocated on demand
#else
	OpenCompleteInitUPP  = 	NewTCPIOCompletionProc(OpenCompleteInit);
	CloseCompleteInitUPP = 	NewTCPIOCompletionProc(CloseCompleteInit);
	SendCompleteInitUPP  = 	NewTCPIOCompletionProc(SendCompleteInit);
	NotifyUPP        	 =  NewTCPNotifyProc(Notify);
#endif
}// Init


/*!
When you are completely done with the Network
module, invoke this method to abort all of the
connections (which may or may not close them
properly).

(2.6)
*/
void
Network_Done	(void)
{
	Commands_ExecuteByID(kCommandCloseConnection, optionKey);
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	DisposeOTNotifyUPP(gGenericNotifyUPP);
	gGenericNotifyUPP = null;
#endif
}// Done


/*!
To open a network socket for the user, use this
method.  A stream number (nonnegative) is returned
if all goes well; otherwise, -1 is returned if a
TCP socket cannot be opened, or -2 is returned if
a stream cannot be made.

(3.0)
*/
NetworkPortRef
Network_SocketOpen	(InetHost			inIPNumber,
					 InetPort const		inRemotePortNumber,
					 InetPort const		inLocalPortNumber,
					 SInt8 const		inTimeoutInSeconds)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = null;
	OSStatus				error = noErr;
	NetworkPortRef			result = kInvalidNetworkPortRef;
	SInt16					i = 0;
	
	
	Console_WriteLine("socket open");
	{
		InetAddress			address;
		
		
		// this structure allows some flexibility for future expansion to longer IP addresses
		OTInitInetAddress(&address, inRemotePortNumber, inIPNumber);
		
		// create a context that configures a connection to the specified host
		contextPtr = newNetworkDataPtr(&address, &result/* local number */);
	}
	
	// the real success of this will be determined by a call to the notifier
	// function; the asynchronous nature of this call means that errors will
	// most likely not be returned here, but instead via the notifier routine
	error = InternetServices_NewTCPIP(contextPtr->notifierUPP, contextPtr/* context */);
	if (error != noErr)
	{
		Console_WriteValue("error while attempting to create Internet Services instance", error);
		Errors_DisplayStopMessage(rStringsNetworkErrors, siNetErrorCantOpenSocketOpenFailed, -1);
		result = kInvalidNetworkPortRef;
	}
	else if (isValidPortNumber(result))
	{
		Console_WriteLine("valid port");
	}
	else
	{
		Console_WriteLine("what the hell?!?!?");
		result = kInvalidNetworkPortRef;
	}
	return result;
#else
	NetworkPortRef		result = 0;
	NetworkPortRef		streamNumber = 0;
	StreamRPtr			p = null;
	MyTCPpbPtr			pbp = null;
	
	
	streamNumber = makeStream();
	
	//if (streamNumber < 0)
	if (!isValidPortNumber(streamNumber))
	{
		result = -2;
	}
	else if  (null == (p = gStreams[streamNumber]))
	{
		result = -2;
	}
	else
	{
		pbp=getPB( gTCP_IPDriverRefNum, TCPActiveOpen, p->stream, 10); // listen call
		
		pbp->pb.csParam.open.remoteHost = inIPNumber;
		pbp->pb.csParam.open.remotePort = inRemotePortNumber;
		pbp->pb.csParam.open.localPort = inLocalPortNumber;
		pbp->pb.csParam.open.ulpTimeoutValue = inTimeoutInSeconds;
		pbp->pb.csParam.open.ulpTimeoutAction = 1; // 1 = abort
		pbp->pb.csParam.open.validityFlags = timeoutValue | timeoutAction; // timeout and action are valid
		
		if (xPBControlAsync(pbp, OpenCompleteInitUPP) != noErr)
		{
			Errors_DisplayStopMessage(rStringsNetworkErrors, siNetErrorCantOpenSocketOpenFailed, -1);
			result = kInvalidNetworkPortRef;
		}
		else
		{
			Network_SetPortType(streamNumber, kInvalidNetworkPortType); // allocated to the user
			result = streamNumber;
		}
	}
	return result;
#endif
}// SocketOpen


/*!
This method takes appropriate actions to return the connection
state to SCLOSED, enabling its port to be reused.  If the
status is closed, data structures are released; otherwise, a
background close is performed, which generates CLOSEDONE
(reinvoking this method again).

(2.6)
*/
SInt16
Network_CloseConnection		(NetworkPortRef		inPortNumber)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	SInt16					result = 0;
	
	
	if (contextPtr == null) result = -1;
	else if (Network_IsSessionOnPortBroken(inPortNumber))
	{
		// the endpoint is now closed; memory can be freed
		disposeNetworkDataPtr(&contextPtr);
	}
	else
	{
		OSStatus		error = noErr;
		
		
		// initiate an orderly disconnect; everything beyond this is handled via the notifier
		error = OTSndOrderlyDisconnect(contextPtr->endpointRef);
	}
	
	return result;
#else
	StreamRPtr		p = null;
	MyTCPpbPtr		pbp = null;
	OSErr			errorCode = noErr;
	SInt16			status = 0;
	static SInt16	count = 0;
	SInt16			result = 0;
	
	
	errorCode = 0;
	
	if (!isValidPortNumber(inPortNumber))
	{
		// invalid port
		result = -1;
	}
	else if ((p = gStreams[inPortNumber]) == null)
	{
		// nothing there
		result = 1;
	}
	else
	{
		pbp = getPB(gTCP_IPDriverRefNum, TCPStatus, p->stream, 11); // status call
		if ((errorCode = xPBControlSync(pbp)) != noErr) 
		{
			if (errorCode == invalidStreamPtr)
			{
				// 3.0 - invalid stream pointer, but why tell the user this?
				Console_WriteLine("network connection close: invalid stream pointer");
				result = -1;
			}
			else
			{
				status = 0;
				count = 0;
			}
		}
		else 
		{
			status = pbp->pb.csParam.status.connectionState; // connection status
			if (count++ == 10)
			{
				status = 0;
				count = 0;
			}
		}
		
		if (result == 0)
		{
			Boolean		doContinue = true;
			
			
			if ((status < 18) && (status > 2))
			{
				// not closed yet
				clearPB(pbp, gTCP_IPDriverRefNum, TCPClose, p->stream); // make call to close
				if ((errorCode = xPBControlAsync(pbp, CloseCompleteInitUPP)) != noErr) 
				{
					// 3.0 - TCP close failed - WHO CARES!
					Console_WriteLine("network connection close: TCP close failed (1)");
					result = errorCode;
				}
				else result = 0;
				
				doContinue = false; // return with OK
			}
			
			if (doContinue)
			{
				// if the program reaches this point, the state must be closed, so free memory
				clearPB(pbp, gTCP_IPDriverRefNum, TCPRelease, p->stream); // release call
				if (xPBControlSync(pbp) != noErr)
				{
					// 3.0 - TCP close failed - WHO CARES!
					Console_WriteLine("network connection close: TCP close failed (2)");
					result = -1;
				}
				
				if (result == 0)
				{
					// dispose of the receive and send buffers, and the stream structure
					Memory_DisposePtr(&(p->buffer));
					Memory_DisposePtr(&(p->sbuffer));
					Memory_DisposePtr(&((Ptr)p));
					gStreams[inPortNumber] = 0L;
					Network_SetPortKerberosEncryptionState(inPortNumber, 0);
					MyPBreturn(pbp);
				}
			}
		}
	}
	
	return result;
#endif
}// CloseConnection


/*!
Calls all of the various initialization routines that set up queueing
variables, static values, reads configuration files, etc. and tries
to open a TCP connection.  If TCP is not running, the user is warned
that this procedure will require a network connection and provide an
opportunity to cancel, unless you pass "false" as the first parameter.
You usually should warn the user that a network connection will be
required, unless a connection is being opened in response to an
Apple Event.

As of version 3.0, this method returns a Boolean value, "true" if the
connection succeeded or "false" if the user canceled or the connection
failed.  The return value of this method can be obtained later using
Network_IsConnected().

This routine may be obsolete with Open Transport, Im not sure yet.

It is possible that a fatal error caused by this method will cause
the entire program to quit, in which case this method may never return.

(2.6)
*/
Boolean
Network_Connect		(Boolean	UNUSED_ARGUMENT(inAskUserIfNotConnected))
{
	Boolean		result = true;
	
	
	if (!Network_IsConnected())
	{
		short	i = 0;
		OSErr	err = noErr;
		
		
		{
			short		hit = kAlertStdAlertOKButton;
			
			
			/*
			if (inAskUserIfNotConnected)
			{
				// confirm with the user that a network connection should be attempted
				Str255					dialogText,
										helpText;
				ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
				InterfaceLibAlertRef	box = null;
				
				
				GetIndString(dialogText, rStringsCautionAlerts, siThisWillRequireConnection);
				GetIndString(helpText, rStringsCautionAlerts, siThisWillRequireConnectionHelpText);
				
				box = Alert_New();
				Alert_SetFilterProc(box, upp);
				Alert_SetHelpButton(box, false);
				Alert_SetParamsFor(box, kAlert_StyleOKCancel);
				Alert_SetText(box, dialogText, helpText);
				Alert_SetType(box, kAlertCautionAlert);
				Alert_Display(box);
				hit = Alert_ItemHit(box);
				Alert_Dispose(&box);
				DisposeModalFilterUPP(upp), upp = null;
			} else hit = kAlertStdAlertOKButton;
			*/
			
			if (hit == kAlertStdAlertOKButton)
			{
				ensureTCPIsOpen();
				result = Network_IsConnected();
				if (result) // i.e. "is the network connection active?"
				{
				#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
					// since Network_Init() initializes the OT contexts array, and providers are
					// instantiated only as needed, is ANYTHING necessary here?
				#else
					for (i = 0; i < kNetworkMaximumNumberOfPorts; i++)
					{
						gStreams[i] = (StreamRPtr)null /*0*/;
					} 
					
					gFreePBQueue.qHead = null;
					gFreePBQueue.qTail = null;
					gFreePBQueue.qFlags = 0;
					gFreeSWDSQueue.qHead = null;
					gFreeSWDSQueue.qTail = null;
					gFreeSWDSQueue.qFlags = 0;
					
					if ((err = OpenDriver("\p.IPP", &gTCP_IPDriverRefNum)) != noErr) // can't open the IP driver
					{	
						result = false;
						Cursors_UseArrow();
						Console_WriteValue("IP driver open failed, error", err);
						Errors_DisplayStopQuitMessage(rStringsStartupErrors,
														siUnableToOpenMacTCP, -1); // does not return
					}
				#endif
				}
			}
			else
			{
				result = false;
			}
		}
		isTCPRunning = result;
	}
	return result;
}// Connect


/*!
Obtains the first available IP address of the computer running
this program, and converts it to a C string.  You can obtain a
Pascal string easily by calling a conversion function on the
result.

If the IP address converts successfully, "true" is returned;
otherwise, "false" is returned.

(3.0)
*/
Boolean
Network_CurrentIPAddressToCString	(char*		outString)
{
	Boolean			result = false;
	
	
	if (outString != null)
	{
		InetInterfaceInfo		info;
		
		
		if (OTInetGetInterfaceInfo(&info, kDefaultInetInterface) == noErr)
		{
			OTInetHostToString(info.fAddress, outString);
			result = true;
		}
		else
		{
			strcpy(outString, "unknown"); // LOCALIZE THIS
			result = false;
		}
	}
	return result;
}// CurrentIPAddressToCString


/*!
Reads from a connection buffer into a user buffer.
Returns the number of bytes read, or a negative
number if an error occurs.

(2.6)
*/
short
Network_GetData		(NetworkPortRef		inPortNumber,
					 void*				inBuffer,
					 short const		inBufferMaximumByteCount)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	short					result = 0;
	
	
	if (contextPtr == null) result = -2;
	else
	{
		TNetbuf*		netbufPtr = null;
		OTDataSize		dataSize = contextPtr->endpointInfo.tsdu;
		OTResult		receiveResult = 0;
		
		
		// determine the proper buffer
		if (contextPtr->receive.flags & T_EXPEDITED) netbufPtr = &contextPtr->receive.expeditedData;
		else netbufPtr = &contextPtr->receive.normalData;
		
		// determine endpoint information - NOTE that this code tries to
		// achieve some level of transport independence by checking the
		// endpoint information instead of assuming TCP/IP; therefore, a
		// number of things are checked (such as support for expediting
		// data, which TCP/IP obviously allows) to allow this independence
		contextPtr->receive.flags = 0;
		if (dataSize != T_INVALID)
		{
			// the endpoint supports receiving data in chunks; break it up and receive each piece separately
			char const*		p = inBuffer;
			short			size = 0,
							count = inBufferMaximumByteCount;
			
			
			// determine the packet size
			size = INTEGER_MINIMUM(dataSize, inBufferMaximumByteCount);
			
			// send the largest packets possible until most of the data is sent;
			// then, send the remaining bytes (i.e. whatever does not fit into a
			// full packet)
			while (count >= size)
			{
				Threads_YieldToAnother();
				OTMemcpy(netbufPtr->buf, p, size);
				
				if ((count - size) > 0) contextPtr->receive.flags |= T_MORE; // flag more to come if thats the case
				else contextPtr->receive.flags &= ~(T_MORE); // clear the flag for the last packet
				
				receiveResult = OTRcv(contextPtr->endpointRef, netbufPtr->buf, size, &contextPtr->receive.flags);
				if (receiveResult >= 0)
				{
					p += (inBufferMaximumByteCount - receiveResult);
					count -= (inBufferMaximumByteCount - receiveResult);
				}
				else
				{
					// error?
					if (receiveResult == kOTLookErr) openTransportHandleLook(contextPtr);
				}
			}
			while (count > 0)
			{
				// then there are remaining bytes, but too few to fill a packet;
				// send them separately
				size = count;
				OTMemcpy(netbufPtr->buf, p, size);
				
				if ((count - size) > 0) contextPtr->receive.flags |= T_MORE; // flag more to come if thats the case
				else contextPtr->receive.flags &= ~(T_MORE); // clear the flag for the last packet
				
				receiveResult = OTRcv(contextPtr->endpointRef, netbufPtr->buf, size, &contextPtr->receive.flags);
				if (receiveResult >= 0)
				{
					p += (inBufferMaximumByteCount - receiveResult);
					count -= (inBufferMaximumByteCount - receiveResult);
				}
				else
				{
					// error?
					if (receiveResult == kOTLookErr) openTransportHandleLook(contextPtr);
				}
			}
		}
		else
		{
			// the endpoint doesnt know how to receive data in chunks, so just receive it all at once
			receiveResult = OTRcv(contextPtr->endpointRef, netbufPtr->buf, inBufferMaximumByteCount,
									&contextPtr->receive.flags);
			if (receiveResult >= 0) result = (inBufferMaximumByteCount - receiveResult);
		}
	}
	return result;
#else
	short		i;
	StreamRPtr	p;
	MyTCPpbPtr	pbp;
	short		inQ, reqdamt;

	if (!isValidPortNumber(inPortNumber))
		return(-2);

	if (null == (p = gStreams[inPortNumber]))
		return(-2);
		
	pbp=getPB(gTCP_IPDriverRefNum, TCPStatus, p->stream, 1);				/* Make status call */
	if (xPBControlSync(pbp) != noErr) 					/* status call bombed */
	{
		// 3.0 - see what happens if MacTelnet simply does not display an error alert in this case
		//Errors_DisplayStopMessage(rStringsNetworkErrors, siNetErrorUnableToGetTCPStatus, -1);
	}

	if (pbp->pb.csParam.status.connectionState !=8) {
		MyPBreturn(pbp);
		return(-1);					  		/* Connection not established */
		}
	
	inQ = pbp->pb.csParam.status.amtUnreadData;
	reqdamt = inBufferMaximumByteCount >inQ ? inQ : inBufferMaximumByteCount;
	
	clearPB( pbp, gTCP_IPDriverRefNum, TCPRcv, p->stream);
	pbp->pb.csParam.receive.rcvBuff = inBuffer;
	pbp->pb.csParam.receive.rcvBuffLen = reqdamt;
	
	if (reqdamt<1) 
		{								/* Drop out if no data */
		MyPBreturn(pbp);
		return(0);
		}
	
	if ((i = xPBControlSync(pbp)) != noErr) {
		// 3.0 - why tell the user that a TCP receiving error occurred?
		return(-1);
		}
		
	reqdamt = pbp->pb.csParam.receive.rcvBuffLen;
	if (reqdamt<inQ) 
		NetEvents_PostProvisionally(kTelnetNetworkEventClassConnection, kConnectionEventData, inPortNumber, 0);			/* more data to get */
	
	MyPBreturn(pbp);
	return(reqdamt);
#endif
}// GetData


/**************************************************************************/
/*  Network_GetNumberOfBytesWaitingOnPort
*   return the number of bytes waiting to be read from the incoming queue.
*/
SInt32 Network_GetNumberOfBytesWaitingOnPort(NetworkPortRef		inPortNumber)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	SInt32					result = 0;
	
	
	if (contextPtr == null) result = -2;
	else
	{
		// this retrieves a hint as to the bytes remaining; perhaps it is too potentially inaccurate?
		(OTResult)OTCountDataBytes(contextPtr->endpointRef, (UInt32*)&result);
	}
	return result;
#else
	StreamRPtr	p;
	MyTCPpbPtr	pbp;
	short		inQ;

	if (!isValidPortNumber(inPortNumber))
		return(-2);

	if (null == (p = gStreams[inPortNumber]))
		return(-2);

	pbp=getPB( gTCP_IPDriverRefNum, TCPStatus, p->stream, 4);				/* Make status call */
	if (xPBControlSync(pbp) != noErr) 					/* status call failed */
	{
		// 3.0 - see what happens if MacTelnet simply does not display an error alert in this case
		//Errors_DisplayStopMessage(rStringsNetworkErrors, siNetErrorUnableToGetTCPStatus, -1);
	}

	inQ = pbp->pb.csParam.status.amtUnreadData;
	MyPBreturn( pbp);
	
	p->push = 1;
	return(inQ);
#endif
}// GetNumberOfBytesWaitingOnPort


/**************************************************************************/
/*  Network_GetNumberOfBytesOfRoomOnPort()
*	return how much room is available in output buffer for a connection
*/
SInt32 Network_GetNumberOfBytesOfRoomOnPort(NetworkPortRef	inPortNumber)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	SInt32					result = 0;
	
	
	if (contextPtr == null) result = -2;
	else
	{
		// return the free space in the send buffer
		result = contextPtr->send.normalData.maxlen - contextPtr->send.normalData.len;
	}
	return result;
#else
	StreamRPtr	p;
	short		inQ, n;

	if (!isValidPortNumber(inPortNumber))
		return(-2);

	if (null == (p = gStreams[inPortNumber]))
		return(-2);

	reclaim( p);
	compressfds( p->fds);

	#ifdef OLDM
	pbp=getPB( gTCP_IPDriverRefNum, TCPStatus, p->stream, 5);			/* Make status call */
	if (xPBControlSync(pbp) != noErr) 
	{
		// 3.0 - see what happens if MacTelnet simply does not display an error alert in this case
		//Errors_DisplayStopMessage(rStringsNetworkErrors, siNetErrorUnableToGetTCPStatus, -1);
	}

	inQ = pbp->csParam.status.sendWindow -
				pbp->csParam.status.amtUnackedData;
	MyPBreturn( pbp);
	#endif

	inQ = n = 0;
	while (p->fds[n].ptr) {
	
		inQ += p->fds[n].length;				/* add up free list space */
		n++;
		}

	return(inQ);
#endif
}// GetNumberOfBytesOfRoomOnPort


/*!
Iterates over open ports and finds the first port
(if any) that has the specified type.  If there
is no such port, "kInvalidNetworkPortRef" is returned.

(3.0)
*/
NetworkPortRef
Network_GetFirstPortOfType		(NetworkPortType	inPortType)
{
	NetworkPortRef		result = kInvalidNetworkPortRef;
	NetworkPortRef		i = kInvalidNetworkPortRef;
	Boolean				found = false;
	
	
	for (i = 0; ((!found) && (i < kNetworkMaximumNumberOfPorts)); i++)
	{
		if (Network_GetPortType(i) == inPortType)
		{
			// this port is of the correct type
			result = i;
			found = true;
		}
	}
	unless (found) result = kInvalidNetworkPortRef;
	
	return result;
}// GetFirstPortOfType


/**************************************************************************/
/*!
This routine obtains the IP address of the local
computer as four characters.  Thus, the first
character of the given buffer is the first octet
of the IP address, etc. so "a.b.c.d" would be
stored as "abcd" on output.

(2.6)
*/
void
Network_GetFourCharIPAddress	(unsigned char*		outIPAddressAsFourChars)
{
	// added auto-net-connect; returns true if the connection is opened successfully (3.0)
	if (Network_Connect(true/* ask user if not connected */))
	{
		// 3.0 - use Open Transport
		char		addrString[32]; // should be more than large enough
		Boolean		isSuccessful = true;
		
		
		// get the IP address as a C string, and then break it up into numbers
		isSuccessful = Network_CurrentIPAddressToCString(addrString);
		if (isSuccessful)
		{
			char*				token = null;
			register SInt16		i = 0;
			unsigned char		octetValue = 0;
			
			
			for (i = 0; i < 4; i++) // one iteration per octet
			{
				token = strtok((!i) ? addrString : null, ".");
				if (token != null) // should never be the case
				{
					octetValue = atoi(token);
					BlockMove(&octetValue, outIPAddressAsFourChars + i, 1/* byte count */);
				}
			}
		}
		else
		{
			Errors_DisplayStopMessage(rStringsNetworkErrors, siNetErrorUnableToGetIPAddress, -1);
			strcpy((char*)outIPAddressAsFourChars, "\0x7F\0x00\0x00\0x00"); // set to localhost on error
		}
	}
}// GetFourCharIPAddress


/**************************************************************************/
NetworkPortType
Network_GetPortType		(NetworkPortRef		inPortNumber)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	NetworkPortType			result = kInvalidNetworkPortType;
	
	
	if (contextPtr != null) result = contextPtr->portType;
	return result;
#else
	NetworkPortType		result = kInvalidNetworkPortType;
	
	
	if (inPortNumber >= 0 && inPortNumber < kNetworkMaximumNumberOfPorts && (gStreams[inPortNumber] != null)) result = (gStreams[inPortNumber]->portType);
	
	return result;
#endif
}// GetPortType


/*!
To determine if the TCP stack is loaded, and hence
if the network is connected, use this method.  It
is especially important to use this method if the
user has a dial-up connection, such as PPP, since
using TCP without being connected will cause the
computer to pop up a dialog and dial the modem.
It is not always appropriate to cause a network
connection without the users consent.

(3.0)
*/
Boolean
Network_IsConnected	(void)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	// THIS IMPLEMENTATION IS FOR OPEN TRANSPORT, AND WILL BE FAVORED SOON
	InetInterfaceInfo	info;
	
	
	return (isTCPRunning = (OTInetGetInterfaceInfo(&info, 0) == noErr));
#else
	return isTCPRunning;
#endif
}// IsConnected


/**************************************************************************/
/*  Network_IsSessionOnPortBroken
*  is a particular session established yet?
*  Returns 0 if the connection is in the established state.
*/
short
Network_IsSessionOnPortBroken	(NetworkPortRef		inPortNumber)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	short					result = 0;
	
	
	if (contextPtr == null) result = -2;
	else
	{
		OTResult		state = OTGetEndpointState(contextPtr->endpointRef);
		
		
		// the unbound and uninitialized states would indicate that a TCP/IP-based
		// connection (which is connection-oriented and transactionless) is broken
		result = ((state == T_UNINIT) || (state == T_UNBND));
	}
	return result;
#else
	StreamRPtr	p;
	MyTCPpbPtr	pbp;
	short		inQ;

	if (!isValidPortNumber(inPortNumber))
		return(-2);

	if (null == (p = gStreams[inPortNumber]))
		return(-2);

	pbp=getPB( gTCP_IPDriverRefNum, TCPStatus, p->stream, 6);			/* Make status call */
	if (xPBControlSync(pbp) != noErr)
	{
		Errors_DisplayStopMessage(rStringsNetworkErrors, siNetErrorUnableToGetTCPStatus, -1);
		inQ = -1;
	}
	else inQ = pbp->pb.csParam.status.connectionState !=8;
	MyPBreturn( pbp);
	
	return(inQ);
#endif
}// IsSessionOnPortBroken


/**************************************************************************/
/*  Network_ListenToTCPPortAndGetItsPortNumber
*   Listen to a TCP port number and make the connection automatically when
*   the SYN packet comes in.  The TCP layer will notify the higher layers
*   with a kConnectionEventOpen event.  Save the port number returned to
*	refer to this connection.
*
*   usage:   portnum = Network_ListenToTCPPortAndGetItsPortNumber(service);
*            short service;
*
*/

NetworkPortRef
Network_ListenToTCPPortAndGetItsPortNumber	(InetPort		serv)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortRef		result = 0;
	
	
	return result;
#else
	NetworkPortRef		portNumber;
	StreamRPtr			p;
	MyTCPpbPtr			pbp;

	
	Console_WriteLine("making stream");
	portNumber = makeStream();

	if (!isValidPortNumber(portNumber))
		return(-2);

	if (null == (p = gStreams[portNumber]))
		return(-2);
	Console_WriteLine("valid stream");

	pbp=getPB( gTCP_IPDriverRefNum, TCPPassiveOpen, p->stream, 8);			/* Make Listen call */
	
	pbp->pb.csParam.open.localPort = serv;
	
	Console_WriteLine("calling listen");
	if (xPBControlAsync(pbp, OpenCompleteInitUPP) != noErr)
	{
		// 3.0 - error, not possible to listen, but the user does not need to know this
		Console_WriteLine("error - not possible to listen");
	}

	return(portNumber);
#endif
}// ListenToTCPPortAndGetItsPortNumber


/***********************************************************************/
/*  Network_MakeFTPQueriesOnPort
*  Provides the information that ftp needs to open a stream back to the
*  originator of the command connection.  The other side's IP number
*  and the port numbers to be used to calculate the default data connection
*  number.  Returns values in an integer array for convenient use in 
*  PORT commands.
*/
short
Network_MakeFTPQueriesOnPort		(NetworkPortRef		inPortNumber,
									 InetHost*			addr,
									 InetPort*			localPort,
									 InetPort*			remotePort)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	short					result = 0;
	
	
	if (contextPtr == null) result = -2;
	else
	{
		// unimplemented
	}
	return result;
#else
	StreamRPtr p;
	MyTCPpbPtr pbp;
	
	//if (inPortNumber < 0)
	if (!isValidPortNumber(inPortNumber))
		return(-2);

	if (null == (p = gStreams[inPortNumber]))
		return(-2);

	pbp=getPB( gTCP_IPDriverRefNum, TCPStatus, p->stream, 9);			/* Make status call */
	if (xPBControlSync(pbp) != noErr) 
	{
		// 3.0 - do not tell the user!  Just return -1
		return -1;
	}

	*addr = pbp->pb.csParam.status.remoteHost;
	*localPort = pbp->pb.csParam.status.localPort;
	*remotePort = pbp->pb.csParam.status.remotePort;
	MyPBreturn( pbp);

	return 0;
#endif
}// MakeFTPQueriesOnPort


/*!
Causes the next data transmission to be expedited.

(2.6)
*/
void
Network_MarkNextTransmissionAsUrgent	(void)
{
	nextIsUrgent = true;
}// MarkNextTransmissionAsUrgent


/*!
Writes something into the output queue.  The
data is not actually sent until the data
handler is ready to send it.  The result is
a negative error code, or a positive number
indicating the number of bytes left to send
(therefore, 0 indicates that all of the given
data was successfully sent).

(2.6)
*/
short
Network_SendData	(NetworkPortRef		inPortNumber,
					 void*				buffer,
					 short const		inByteCount)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	short					result = 0;
	
	
	if (contextPtr == null) result = -2;
	else if (inByteCount > 0)
	{
		TNetbuf*		netbufPtr = null;
		OTDataSize		dataSize = getDataUnitSize(contextPtr->endpointRef, 512/* arbitrary ideal size */,
													nextIsUrgent/* expedited data? */);
		OTResult		sendResult = 0;
		
		
		// determine endpoint information - NOTE that this code tries to
		// achieve some level of transport independence by checking the
		// endpoint information instead of assuming TCP/IP; therefore, a
		// number of things are checked (such as support for expediting
		// data, which TCP/IP obviously allows) to allow this independence
		contextPtr->send.flags = 0;
		if ((nextIsUrgent) && endpointSupportsTSDUs(contextPtr->endpointRef, true/* for urgent data */))
		{
			// if the user marked the next send as urgent, and the protocol supports it,
			// then modify the send flags to suggest that this send be expedited
			contextPtr->send.flags |= T_EXPEDITED;
			nextIsUrgent = false;
		}
		
		// determine the proper buffer
		netbufPtr = (contextPtr->send.flags & T_EXPEDITED)
					? &contextPtr->send.expeditedData : &contextPtr->send.normalData;
		
		if (dataSize != T_INVALID)
		{
			// the endpoint supports sending data in chunks; break it up and send each piece separately
			char const*		p = buffer;
			short			size = 0,
							count = inByteCount;
			
			
			// determine the packet size
			size = INTEGER_MINIMUM(dataSize, inByteCount);
			
			// send the largest packets possible until most of the data is sent;
			// then, send the remaining bytes (i.e. whatever does not fit into a
			// full packet)
			while (count > 0)
			{
				OTMemcpy(netbufPtr->buf, p, size);
				
				if ((count - size) > 0) contextPtr->send.flags |= T_MORE; // flag more to come if thats the case
				else contextPtr->send.flags &= ~(T_MORE); // clear the flag for the last packet
				
				sendResult = OTSnd(contextPtr->endpointRef, netbufPtr->buf, size, contextPtr->send.flags);
				if (sendResult >= 0)
				{
					p += (inByteCount - sendResult);
					count -= (inByteCount - sendResult);
				}
				else
				{
					// error?
					if (sendResult == kOTLookErr) openTransportHandleLook(contextPtr);
				}
				if (count < size) size = count;
				
				Threads_YieldToAnother();
			}
		}
		else
		{
			// the endpoint doesnt know how to send data in chunks, so just send it all at once; a
			// positive result indicates the number of bytes transmitted, a negative value is an error
			sendResult = OTSnd(contextPtr->endpointRef, buffer, inByteCount, contextPtr->send.flags);
			if (sendResult >= 0) result = (inByteCount - sendResult);
			if (sendResult == kOTLookErr) openTransportHandleLook(contextPtr);
		}
	}
	return result;
#else
	StreamRPtr		p = null;
	short			result = 0;
	
	
	if (!isValidPortNumber(inPortNumber)) result = -2;
	else if ((p = gStreams[inPortNumber]) == null) result = -2;
	else if (!inByteCount) result = 0;
	else
	{
		MyTCPpbPtr		pbp = null;
		wdsEntry*		swds = null;
		short			remaining = 0,
						queued = 0,
						n = 0,
						m = 0;
		OSErr			error = noErr;
		
		
		swds = getSWDS();
		if (swds == null) result = inByteCount;
		else
		{
			reclaim(p);
			compressfds(p->fds);
			n = 0;
			remaining = inByteCount;
			
			while (p->fds[n].ptr != 0 && remaining > 0)
			{
				swds[n].ptr = p->fds[n].ptr;
				if (p->fds[n].length > remaining)
				{
					swds[n].length = remaining;
					p->fds[n].length -= remaining;
					p->fds[n].ptr += remaining;
					remaining=0;
				}
				else 
				{
					swds[n].length =  p->fds[n].length;
					remaining -= p->fds[n].length;
					p->fds[n].length = 0;
					p->fds[n].ptr = 0;
				}
				n++;
			}
		
			if (n > p->maxFDSused) p->maxFDSused = n;
			
			compressfds(p->fds);
			
			queued = inByteCount - remaining;
			for (m = 0; m < n; m++) 
			{
				BlockMove(buffer, swds[m].ptr, swds[m].length); // put data in WDS
				buffer=(void*)((char*)buffer + swds[m].length);
			}
			
			swds[m].ptr = 0L;
			swds[m].length = 0;
			
			pbp = getPB(gTCP_IPDriverRefNum, TCPSend, p->stream, 2); // send data
			pbp->pb.csParam.send.wdsPtr = (Ptr)swds;
			pbp->pb.csParam.send.pushFlag = p->push;
			pbp->pb.csParam.send.ulpTimeoutValue = gSendTimeout;	// timeout value
			pbp->pb.csParam.send.ulpTimeoutAction = 1;				// abort
			pbp->pb.csParam.send.validityFlags = timeoutValue | timeoutAction; // timeout and action are valid
			if (nextIsUrgent)
			{
				nextIsUrgent = false;
				pbp->pb.csParam.send.urgentFlag = true;
			}
		
			p->push = 0;
		
			if ((error = xPBControlAsync(pbp, SendCompleteInitUPP)) != noErr) 
			{
				// 3.0 - error, could not add data to the queue - but why tell the user this?
				Console_WriteLine("WARNING: Could not send data to the queue.");
				result = -1;
			}
			else result = queued;
		}
	}
	return result;
#endif
}// SendData


/*!
Attempts to push the rest of the data from
the send-queue, and then returns the number
of bytes left in the queue.

(2.6)
*/
short
Network_SendFlush	(NetworkPortRef		inPortNumber)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	// not there yet
	return -2;
#else
	StreamRPtr		p = null;
	short			result = 0;
	
	
	if (!isValidPortNumber(inPortNumber)) result = -2;
	else if (null == (p = gStreams[inPortNumber])) result = -2;
	else
	{
		short			queueOctetCount = 0;
		OSErr			error = noErr;
		MyTCPpbPtr		pbp = getPB(gTCP_IPDriverRefNum, TCPStatus, p->stream, 3); // status call
		
		
		if ((error = xPBControlSync(pbp)) != noErr)
		{
			// 3.0 - see what happens if MacTelnet simply does not display an error alert in this case
			//Errors_DisplayStopMessage(rStringsNetworkErrors, siNetErrorUnableToGetTCPStatus, -1);
			Console_WriteLine("WARNING: Unable to get TCP status on flush of send-queue.");
		}
		
		queueOctetCount = pbp->pb.csParam.status.amtUnackedData;
		MyPBreturn(pbp);
		
		p->push = 1;
		result = queueOctetCount;
	}
	return result;
#endif
}// SendFlush


/*!
This convenience routine first invokes the
Network_SendFlush() method in an attempt to
clear out the send-queue, and then calls
Network_SendData() to write the given buffer
to the queue.

(3.0)
*/
short
Network_SendFlushData		(NetworkPortRef		inPortNumber,
							 void*				inBuffer,
							 SInt16 const		inNumberOfBytes)
{
	(short)Network_SendFlush(inPortNumber);
	return Network_SendData(inPortNumber, inBuffer, inNumberOfBytes);
}// SendFlushData


/*!
Assigns the specified encryption state
information to the stream data structure
for the specified port.

(2.6)
*/
SInt16
Network_SetPortKerberosEncryptionState		(NetworkPortRef		inPortNumber,
											 Ptr				inEncryptionData)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	SInt16					result = 0;
	
	
	if (contextPtr == null) result = -2;
	else
	{
		// unimplemented
		//contextPtr->authEncrData = inEncryptionData;
	}
	return result;
#else
	StreamRPtr		p = null;
	SInt16			result = 0;
	
	
	if (!isValidPortNumber(inPortNumber)) result = -2;
	else if ((p = gStreams[inPortNumber]) == null) result = -2;
	else p->aedata = inEncryptionData;
	
	return result;
#endif
}// SetPortKerberosEncryptionState


/**************************************************************************/
void
Network_SetPortType		(NetworkPortRef		inPortNumber,
						 NetworkPortType	inType)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	
	
	if (contextPtr != null) contextPtr->portType = inType;
#else
	if (inPortNumber >= 0 && inPortNumber < kNetworkMaximumNumberOfPorts && (gStreams[inPortNumber] != null))
		gStreams[inPortNumber]->portType = inType;
#endif
}// SetPortType


/*!
To set the number of seconds that can elapse before
a network socket opening gives up, use this method.

(3.0)
*/
void
Network_SetSendTimeoutInSeconds		(SInt8 const		inTimeout)
{
	gSendTimeout = inTimeout;
}// SetSendTimeoutInSeconds


/*!
Displays the IP address of the computer running
this program.  This method has been completely
rewritten in MacTelnet 3.0 to use an alert window,
to provide support for copying to the clipboard,
and to use Open Transport.

(2.6)
*/
void
Network_ShowIPAddress		(void)
{
	char		ipAddressString[30];
	
	
	if (Network_CurrentIPAddressToCString(ipAddressString))
	{
		ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
		InterfaceLibAlertRef	box = null;
		Str255					dialogText,
								buttonString;
		
		
		StringUtilities_CToPInPlace(ipAddressString);
		
		GetIndString(buttonString, rStringsSpecialButtons, siButtonCopy);
		GetIndString(dialogText, rStringsNoteAlerts, siShowIPAddress);
		StringUtilities_PMetaSetDefaults();
		StringUtilities_PBuild(dialogText, (StringPtr)ipAddressString, EMPTY_PSTRING,
								EMPTY_PSTRING, EMPTY_PSTRING);
		
		box = Alert_New();
		Alert_SetFilterProc(box, upp);
		Alert_SetHelpButton(box, false);
		Alert_SetParamsFor(box, kAlert_StyleOKCancel);
		Alert_SetButtonText(box, kAlertStdAlertCancelButton, null);
		Alert_SetButtonText(box, kAlertStdAlertOtherButton, buttonString);
		Alert_SetText(box, dialogText, EMPTY_PSTRING);
		Alert_SetType(box, kAlertNoteAlert);
		Alert_Display(box);
		
		if (Alert_ItemHit(box) == kAlertStdAlertOtherButton) // then Copy to Clipboard was requested
		{
		#if TARGET_API_MAC_OS8
			(OSStatus)ZeroScrap();
			PutScrap(StrLength((StringPtr)ipAddressString), kScrapFlavorTypeText,
						(StringPtr)(ipAddressString + 1)); // skip length byte
		#else
			(OSStatus)ClearCurrentScrap();
			{
				ScrapRef		currentScrap = null;
				
				
				if (GetCurrentScrap(&currentScrap) == noErr)
				{
					(OSStatus)PutScrapFlavor(currentScrap, kScrapFlavorTypeText,
												kScrapFlavorMaskNone, StrLength((StringPtr)ipAddressString),
												ipAddressString + 1/* skip length byte */);
				}
			}
		#endif
		}
		Alert_Dispose(&box);
		DisposeModalFilterUPP(upp), upp = null;
	}
	else
	{
		Errors_DisplayStopMessage(rStringsGeneralMessages, siIPAddressNotAttainable, -1);
	}
}// ShowIPAddress


#if !OPEN_TRANSPORT_NETWORK_IMPLEMENTATION

/**************************************************************************/

static wdsEntry	*getSWDS(void)
{
	short		n=0;
	wdsEntry *	wds;
	
	if (gFreeSWDSQueue.qHead == null)
		wds = (wdsEntry *) Memory_NewPtrCritical( sizeof(wdsEntry) * MAX_SWDS_ELEMS);	
	else
		{
		wds = (wdsEntry *) gFreeSWDSQueue.qHead;
		Dequeue(gFreeSWDSQueue.qHead, &gFreeSWDSQueue);
		}
		
	if (wds == null) {	/* sorry, cant allocate WDS (BIG PROBLEM) */
		// this method does not return
		Errors_DisplayStopQuitMessage(rStringsNetworkErrors, siNetErrorFatalCannotAllocateTCPBuffer, -1);
		}

	return(wds);
}

/**************************************************************************/
static void	returnSWDS(wdsEntry *wds)
{
	Enqueue((QElemPtr)wds, &gFreeSWDSQueue);
}

/**************************************************************************/
static void MyPBreturn(MyTCPpbPtr pbp)
{
	Enqueue((QElemPtr)pbp, &gFreePBQueue);
}

/**************************************************************************/
static MyTCPpb *getPB(short driver,short call, StreamPtr stream, short usernum)
{
	//UNUSED_ARG(usernum)
	MyTCPpb *pbp;

	if (gFreePBQueue.qHead == null)
		pbp = (MyTCPpbPtr) Memory_NewPtrCritical( sizeof(MyTCPpb) );
	else
		{
		pbp = (MyTCPpbPtr) gFreePBQueue.qHead;
		Dequeue(gFreePBQueue.qHead, &gFreePBQueue);
		}
		
	if (pbp == null) {	/* sorry, cant allocate TCP buffer (BIG PROBLEM) */
		// this method does not return
		Errors_DisplayStopQuitMessage(rStringsNetworkErrors, siNetErrorFatalCannotAllocateTCPBuffer, -1);
		}
	
	memset((char *) pbp, '\0', sizeof(MyTCPpb));	
	pbp->pb.ioCRefNum = driver;
	pbp->pb.tcpStream = stream;
	pbp->pb.csCode = call;
	
	return(pbp);
}

/**************************************************************************/
static void clearPB(MyTCPpbPtr pbp, short driver, short call, StreamPtr stream)
{
	Memory_WriteZero((Ptr)pbp, sizeof(MyTCPpb));
	pbp->pb.ioCRefNum = driver;
	pbp->pb.tcpStream = stream;
	pbp->pb.csCode = call;
}

/**************************************************************************/
static OSErr xPBControlSync(MyTCPpbPtr pbp) 
{ 
	pbp->pb.ioCompletion = 0L;
	return PBControl((ParmBlkPtr)(pbp),false);
}

/**************************************************************************/
static OSErr xPBControlAsync(MyTCPpbPtr pbp, TCPIOCompletionUPP completionUPP) 
{ 
	pbp->pb.ioCompletion = completionUPP; 
	return(PBControl((ParmBlkPtr)(pbp),true));		/* async */
}


/**************************************************************************/
/*	reclaim( p) -
		reclaims buffer space to stream (from pointer p) into the FDS list  */
 
static void reclaim(StreamRPtr p)
{
	short	n = 0,
			offset = 0;
	
	
	while (offset < MAX_FDS_ELEMS && p->fds[offset].ptr != 0L) offset++;
	
	if (offset < MAX_FDS_ELEMS) /* offset small enough? */
	{
		for (n=0 ; n<MAX_FDS_ELEMS && offset< MAX_FDS_ELEMS; n++) 
		{
			if (p->exFDS[ n].inuse) 
			{
				p->fds[ offset++]=p->exFDS[ n].fds;
				p->exFDS[ n].inuse = 0;
			}
		}
	}
}

/**************************************************************************
	giveback( p, wds) - gives WDS entries back to the stream by putting them in the 
 		mutually exclusive buffer.
	p -> stream
	wds -> wds array			*/
static short giveback(StreamRPtr p, wdsEntry *wds)
{
	short n=0, m=0;
	
	while ( n< MAX_SWDS_ELEMS && wds[n].ptr !=0L) 
		{
		while (m< MAX_FDS_ELEMS && p->exFDS[ m].inuse) m++;
		if (m >= MAX_FDS_ELEMS)									// This was off by one - JMB 2.6
			return(-1);				/* No room in the RECLAIMation center */
		else 
			{
			p->exFDS[ m].inuse =1;
			p->exFDS[ m].fds = wds[n];
			m++;
			}
		n++;
		}

	return 0;
}

/**************************************************************************/
/*	compressfds( fds)
		compress an fds data structure to make everyone happy */
static void compressfds(wdsEntry *fds)
{
	short n,m,compressed;
	
	compressed = 0;
	
	while ( !compressed) {
		compressed=1;
		for (n=0; n< MAX_FDS_ELEMS; n++) 
			{				/* Slow Forwards */
			if (fds[n].ptr) 
				{								/* Do if N exists */
				for ( m = MAX_FDS_ELEMS -1; m>=0; m--) 
					{	/* Fast Backwards */
					if (fds[m].ptr && (fds[m].ptr+fds[m].length == fds[n].ptr)) {
						fds[n].length+=fds[m].length;
						fds[n].ptr = fds[m].ptr;
						fds[m].ptr=0L;
						fds[m].length=0;
						compressed=0;
						}
#ifdef CHECKBOTHWAYZ
					else 
					if (fds[n].ptr+fds[n].length == fds[m].ptr) 
						{
						fds[m].length+=fds[n].length;
						fds[n].ptr=0L;
						fds[n].length=0;
						compressed=0;
						}
#endif CHECKBOTHWAYZ
					}
				}
			}
		}
	m=0;n=0;
	
	/* Close the gaps */
	
	while (n+m < MAX_FDS_ELEMS) 
		{
		while (fds[n+m].ptr ==0L && n+m< MAX_FDS_ELEMS)
			m++;			/* increase gap to valid entry */
		if (n+m<MAX_FDS_ELEMS)
			fds[n]=fds[n+m];
		n++;
		}
	
	/* Get rid of the empty spaces */
	
	n--;		/* for the next loop */
	while (n < MAX_FDS_ELEMS)
		{
		fds[n].ptr=0;
		fds[n++].length=0;
		}
}


/**************************************************************************/
static short	findbystream(StreamPtr streamPtr)
{
	short	portNumber=0;
	
	while (portNumber<kNetworkMaximumNumberOfPorts &&
			( (gStreams[portNumber] ==(StreamRPtr)0L) || (streamPtr !=gStreams[portNumber]->stream)))
		portNumber++;
	if (portNumber >=kNetworkMaximumNumberOfPorts)
		return(-1);
	else
		return(portNumber);
}


/**************************************************************************/
pascal void Notify(StreamPtr streamPtr, unsigned short code, Ptr uptr, 
						unsigned short terminReason, struct ICMPReport *icmpMsg)
{
	//UNUSED_ARG(uptr)
	//UNUSED_ARG(terminReason)
	//UNUSED_ARG(icmpMsg)
	StreamRPtr	p;
	short		portNumber;
	
	portNumber = findbystream(streamPtr);
	
	if (portNumber < 0 || (p = gStreams[portNumber]) == 0L)
		return;
	
	switch( code) {
		case TCPTerminate:
		case TCPClosing:
			NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventClose, portNumber,0);
			break;
		case TCPULPTimeout:
			NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventFailure, portNumber,0);
			break;
		case TCPDataArrival:
		case TCPUrgent:
			NetEvents_PostProvisionally(kTelnetNetworkEventClassConnection, kConnectionEventData, portNumber,0);
			break;
		case TCPICMPReceived:
		default:
			break;
		}
	return;
}

/**************************************************************************/
static void OpenCompleteInit(TCPiopb *pbp)
{
	MyTCPpbPtr		theMacTCPParamBlockPtr;
	
	
	theMacTCPParamBlockPtr = (MyTCPpbPtr)pbp;
	OpenComplete(theMacTCPParamBlockPtr);
}

/**************************************************************************/
void	OpenComplete(MyTCPpbPtr pbp)
{
	StreamRPtr	p;
	short		portNumber;
	
	
	portNumber= findbystream(pbp->pb.tcpStream);
	
	if (!(portNumber < 0 || (p = gStreams[portNumber]) == 0L))
	{
		if (pbp->pb.ioResult == duplicateSocket)
		{
			NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventFailure, portNumber,pbp->pb.ioResult);	/* Failure ... */
		}
		else
		{
			if (pbp->pb.ioResult !=noErr) 
				NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventFailure, portNumber,pbp->pb.ioResult);	/* Failure ... */
			else 
				NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventOpen, portNumber,0);			/* Success ! */
		}
	}
	MyPBreturn(pbp);
}

/**************************************************************************/
void	SendComplete(MyTCPpbPtr pbp)
{
	StreamRPtr	p;
	short		portNumber;
	wdsEntry	*swds;

	swds = (wdsEntry *) pbp->pb.csParam.send.wdsPtr;
	
	portNumber = findbystream(pbp->pb.tcpStream);
	
	if (portNumber < 0 || (p = gStreams[portNumber]) == 0L) {
		MyPBreturn(pbp);
		return;
		}
		
	MyPBreturn( pbp);
	giveback( p, (wdsEntry *) pbp->pb.csParam.send.wdsPtr);	/* give this back.... NOW */
	returnSWDS( swds);
}


/**************************************************************************/
static void SendCompleteInit(TCPiopb *pbp)
{
	MyTCPpbPtr	myptr = null;
	
	
	myptr = (MyTCPpbPtr)pbp;
	SendComplete(myptr);
}// SendCompleteInit


/**************************************************************************/
void	CloseComplete(MyTCPpbPtr pbp)
{
	StreamRPtr	p;
	short		portNumber;

	portNumber= findbystream(pbp->pb.tcpStream);
	
	if (portNumber < 0 || (p = gStreams[portNumber]) == 0L)
	  {
		NetEvents_Post(kTelnetNetworkEventClassSession, kSessionEventCloseFinishedWithError, portNumber,0);
		MyPBreturn(pbp);
		return;
	  }
		
	if (pbp->pb.ioResult !=noErr) 
		NetEvents_Post(kTelnetNetworkEventClassSession, kSessionEventCloseFinishedWithError, portNumber,0);
	else 
		NetEvents_Post(kTelnetNetworkEventClassSession, kSessionEventCloseFinished, portNumber,0);			/* Success ! */

	MyPBreturn( pbp);
}

/**************************************************************************/
static void CloseCompleteInit(TCPiopb *pbp)
{
	MyTCPpbPtr	myptr = null;
	
	
	myptr = (MyTCPpbPtr)pbp;
	CloseComplete(myptr);
}// CloseCompleteInit

#endif


//
// internal methods
//

/*!
Closes the specified connection IMMEDIATELY.

(2.6)
*/
static short
abortConnection	(short	inPortNumber)
{
#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION
	NetworkPortContextPtr	contextPtr = getNetworkPortContextFor(inPortNumber);
	short					result = 0;
	
	
	if (contextPtr == null) result = -1;
	else
	{
		OSStatus		error = noErr;
		
		
		// initiate a disconnection; everything beyond this goes to the notifier routine
		error = OTSndDisconnect(contextPtr->endpointRef,
								(TCall*)null/* not required; used if data is to be sent before disconnecting */);
	}
	return result;
#else
	StreamRPtr		p = null;
	MyTCPpbPtr		pbp = null;
	OSErr			errorCode = noErr;
	
	
	if (!isValidPortNumber(inPortNumber))
		return(-1);

	if ((p = gStreams[inPortNumber]) != null) 	/* something there */
	{			
		pbp=getPB( gTCP_IPDriverRefNum, TCPAbort, p->stream, 12);			/* Make Close call */
		if ((errorCode=xPBControlSync(pbp)) != noErr)
		{
			Errors_DisplayNoteMessage(rStringsNetworkErrors, siNetErrorCantAbortConnection, -1);
		}
		clearPB( pbp,gTCP_IPDriverRefNum, TCPRelease, p->stream);			/* Make Close call */
		if (xPBControlSync(pbp) != noErr) 
		{
			// 3.0 - TCP close failed, WHO CARES!
			return(-1);
		}
	}
	else
		return(1);
	
	Memory_DisposePtr(&(p->buffer));				/* Free Receive Buffer */
	Memory_DisposePtr(&(p->sbuffer));				/* Free Send Buffer */
	Memory_DisposePtr(&((Ptr)p));					/* Free Stream Structure */
	gStreams[inPortNumber]=0L;

	MyPBreturn(pbp);
	return(0);
#endif
}// abortConnection


/*!
This method makes sure that the TCP/IP domain name
resolver is started up.  The Network_IsConnected()
value is modified accordingly.

(3.0)
*/
static void
ensureTCPIsOpen	(void)
{
	unless (Network_IsConnected())
	{
		OSErr		io = noErr;
		
		
		if (io != noErr)
		{
			InterfaceLibAlertRef	box = null;
			Str255					dialogText;
			Str255					helpText;
			Str255					string;
			short					hit = 0;
			
			
			// construct a message from several string fragments, including the error code
			GetIndString(dialogText, rStringsGeneralMessages, siDomainNameResolverOpenFailedBecause);
			NumToString((long)io, string);
			StringUtilities_PMetaSetDefaults();
			StringUtilities_PBuild(dialogText, string, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
			GetIndString(helpText, rStringsGeneralMessages, siDomainNamesWontWorkHelpText);
			
			box = Alert_New();
			Alert_SetHelpButton(box, false);
			Alert_SetParamsFor(box, kAlert_StyleOK);
			Alert_SetText(box, dialogText, helpText);
			Alert_SetType(box, kAlertCautionAlert);
			Alert_Display(box);
			Alert_Dispose(&box);
		}
		else isTCPRunning = true;
	}
}// ensureTCPIsOpen


/*!
To determine in a standard way if a port number
is in the valid range, use this method.

(3.0)
*/
static Boolean
isValidPortNumber		(NetworkPortRef		inPortNumber)
{
	return ((inPortNumber >= 0) && (inPortNumber < kNetworkMaximumNumberOfPorts));
}// isValidPortNumber


#if !OPEN_TRANSPORT_NETWORK_IMPLEMENTATION

/*!
To get a new, empty stream, use this method.

(2.6)
*/
static SInt16
makeStream	(void)
{
	register SInt16		i = 0;
	StreamRPtr			p = null;
	MyTCPpbPtr			pbp = null;
	SInt16				portNumber = kInvalidNetworkPortRef;
	SInt16				result = 0;
	
	
	for ( portNumber=0; (gStreams[portNumber]!= null) && portNumber<kNetworkMaximumNumberOfPorts; portNumber++) {}
	
	if (portNumber >= kNetworkMaximumNumberOfPorts)
		return(-2);

 	if (!(p = gStreams[portNumber] = (StreamRPtr) Memory_NewPtrCritical(sizeof(StreamRec))))
 		return(kInvalidNetworkPortRef);
  
 	if ((p->buffer = (char *) Memory_NewPtrCritical( TCPBUFSIZ)) == null) {
 		Memory_DisposePtr(&((Ptr)p));
 		gStreams[portNumber] = 0;
  		return(kInvalidNetworkPortRef);
 	}
 	if ((p->sbuffer = (char *) Memory_NewPtrCritical( TCPBUFSIZ)) == null) {
 		Memory_DisposePtr(&((Ptr)p->buffer));
 		Memory_DisposePtr(&((Ptr)p));
 		gStreams[portNumber] = 0;
  		return(kInvalidNetworkPortRef);
 	}	
 
	
	for (i=0; i<MAX_FDS_ELEMS; i++) 
		{
		p->fds[ i].length =0; p->fds[ i].ptr = 0L;
		p->exFDS[ i].inuse=0; p->exFDS[ i].fds.length=0;p->exFDS[ i].fds.ptr=0L;
		}
	p->fds[0].length = TCPBUFSIZ;
	p->fds[0].ptr = p->sbuffer;
	p->maxFDSused=0;
	p->aedata = null;
	
	pbp=getPB(gTCP_IPDriverRefNum, TCPCreate, (long) 0, 7); // important - make create call
	pbp->pb.csParam.create.rcvBuff = p->buffer;
	pbp->pb.csParam.create.rcvBuffLen = TCPBUFSIZ;
	pbp->pb.csParam.create.notifyProc = NotifyUPP;

	if (xPBControlSync(pbp) != noErr)
	{
		// 3.0 - unable to make a stream because a TCP error has occurred.  Dont tell the user, just return -1!
	}

	p->stream = pbp->pb.tcpStream;
	
	/* made new stream */
	MyPBreturn(pbp);
	return(portNumber);
}// makeStream

#endif


#if OPEN_TRANSPORT_NETWORK_IMPLEMENTATION

/*!
Destroys a structure allocated by newNetworkDataPtr().
Since these are allocated for the benefit of new TCP/IP
service providers, you should not destroy this kind of
data until the corresponding provider has been destroyed.

(3.0)
*/
static void
disposeNetworkDataPtr	(NetworkPortContextPtr*		inoutContextPtr)
{
	if (inoutContextPtr != null)
	{
		Console_WriteLine("disposing of network port context");
		DisposeOTNotifyUPP((*inoutContextPtr)->notifierUPP);
		(*inoutContextPtr)->notifierUPP = null;
		Memory_DisposePtr((Ptr*)&(*inoutContextPtr)->send.call.addr.buf);
		Memory_DisposePtr((Ptr*)&(*inoutContextPtr)->receive.call.addr.buf);
		Memory_DisposePtr((Ptr*)&(*inoutContextPtr)->receive.call.opt.buf);
		Memory_DisposePtr((Ptr*)&(*inoutContextPtr)->receive.call.udata.buf);
		Memory_DisposePtr((Ptr*)inoutContextPtr);
	}
}// disposeNetworkDataPtr


/*!
Returns whether or not the specified network ports
endpoint supports transport service data units.  If
an endpoint doesnt support TSDUs, then it is not
easy to optimize transfers (TSDUs indicate the size
of the largest send or receive buffer that the
endpoint will allow, thereby allowing send calls to
transmit as much data as possible in one call).

An endpoint may support TSDUs differently based on
whether or not the data will be sent or received in
an expedited (urgent) manner.

IMPORTANT:	This method may only be invoked reliably
			if the connection is already open.  The
			TSDU information can change during the
			negotiation process, so results from a
			call at bind time might be different
			than results at connection time.

(3.0)
*/
static Boolean
endpointSupportsTSDUs		(NetworkPortContextPtr		inoutContextPtr,
							 Boolean const				inExpeditedDataUnits)
{
	Boolean		result = 0;
	
	
	if (inExpeditedDataUnits)
	{
		// T_INVALID means that TSDUs are not supported for expedited data,
		// and 0 means that ETSDUs are not supported at all by the endpoint
		result = ((inoutContextPtr->endpointInfo.etsdu != T_INVALID) &&
					(inoutContextPtr->endpointInfo.etsdu != 0));
	}
	else
	{
		// T_INVALID means that TSDUs are not supported for normal data,
		// and 0 means that TSDUs are not supported at all by the endpoint
		result = ((inoutContextPtr->endpointInfo.tsdu != T_INVALID) &&
					(inoutContextPtr->endpointInfo.tsdu != 0));
	}
	return result;
}// endpointSupportsTSDUs


/*!
Returns an appropriate buffer size for data units
in the given network port context.  This allows you
to set up sends and receives as effectively as
possible, maximizing data transfer for each call to
a send or receive function.

If the specified endpoint does not support transport
service data units of the specified type (i.e. normal
or expedited), or if it supports units of an infinite
size, then the value of "inWantThisByteCount" is
returned.

If TSDUs are supported for the type of data requested
(normal or expedited), then the maximum supported
size for a transport service data unit of that type
is returned.  By passing buffers of this resultant
size to OTSnd() or OTRcv(), you will maximize the
effectiveness of each call because they will be less
likely to return failures.

(3.0)
*/
static OTDataSize
getDataUnitSize			(NetworkPortContextPtr		inoutContextPtr,
						 OTDataSize const			inWantThisByteCount,
						 Boolean const				inExpeditedDataUnits)
{
	OTDataSize		result = 0;
	
	
	if (endpointSupportsTSDUs(inoutContextPtr, inExpeditedDataUnits))
	{
		// return the transport service data unit size for the requested type of data
		result = (inExpeditedDataUnits) ? inoutContextPtr->endpointInfo.etsdu : inoutContextPtr->endpointInfo.tsdu;
		if (result == T_INFINITE) result = inWantThisByteCount;
	}
	else
	{
		// not supported by the endpoint - just use the suggested buffer size
		result = inWantThisByteCount;
	}
	return result;
}// getDataUnitSize


/*!
Returns the context for the specified internal port
number.  If the specified number is out of range or
the specified context does not otherwise exist,
"null" is returned.

(3.0)
*/
static NetworkPortContextPtr
getNetworkPortContextFor	(NetworkPortRef		inPortNumber)
{
	NetworkPortContextPtr		result = null;
	
	
	if (isValidPortNumber(inPortNumber))
	{
		result = gPortContexts[inPortNumber]; // may still end up being "null"
	}
	return result;
}// getNetworkPortContextFor


/*!
Allocates a new structure used as a context for
an Open Transport TCP/IP service.  When finished
with the structure, invoke disposeNetworkDataPtr().

The specified address structure is large enough to
hold an IP version 6 address, if necessary.

Specify information on the connection that will
ultimately be attempted using the resultant data.

(3.0)
*/
static NetworkPortContextPtr
newNetworkDataPtr	(InetAddress const*		inRemoteSendAddress,
					 NetworkPortRef*		outLocalPortNumberPtrOrNull)
{
	enum
	{
		// Ill figure out a more elegant way to set the sizes later
		kReceiveBufferSizeAddress = sizeof(InetAddress),
		kReceiveBufferSizeOptions = 4,
		kReceiveBufferSizeUnitData = 100
	};
	NetworkPortContextPtr		result = null;
	
	
	Console_WriteLine("creating new network port context");
	if (inRemoteSendAddress != null)
	{
		result = (NetworkPortContextPtr)Memory_NewPtrCritical(sizeof(NetworkPortContext));
		if (result != null)
		{
			register SInt16		i = 0;
			
			
			// store this context in the global array, in the specified position
			result->portNumber = kInvalidNetworkPortRef;
			for (i = 0; i < kNetworkMaximumNumberOfPorts; i++)
			{
				if (gPortContexts[i] == null)
				{
					result->portNumber = i;
					gPortContexts[i] = result;
					if (outLocalPortNumberPtrOrNull != null) *outLocalPortNumberPtrOrNull = i;
				}
			}
			
			// basic field initializations
			result->portType = kInvalidNetworkPortType;
			result->error = noErr;
			result->servicesRef = null; // this is initialized by the notifier when the endpoint is created
			result->endpointRef = null; // this is initialized by the notifier when the endpoint is created
			result->notifierUPP = NewOTNotifyUPP(openTransportGenericNotify);
			
			// Endpoint information initialization - none of these values will stick, this is
			// just to ensure they have known values.  The endpoint data is first filled in at
			// endpoint creation time, and again upon successful connection (because it could
			// be different after connecting).
			result->endpointInfo.addr = 0;
			result->endpointInfo.options = 0;
			result->endpointInfo.tsdu = 0;
			result->endpointInfo.etsdu = 0;
			result->endpointInfo.connect = 0;
			result->endpointInfo.discon = 0;
			result->endpointInfo.servtype = T_COTS_ORD;
			result->endpointInfo.flags = 0L;
			
			// define where the data goes and how data sends occur
			result->send.call.addr.maxlen = sizeof(*inRemoteSendAddress);
			result->send.call.addr.len = sizeof(*inRemoteSendAddress);
			result->send.call.addr.buf = (UInt8*)Memory_NewPtrCritical(sizeof(*inRemoteSendAddress));
			if (result->send.call.addr.buf != null)
			{
				BlockMove(inRemoteSendAddress, result->send.call.addr.buf, sizeof(InetAddress));
			}
			result->send.call.opt.maxlen = 0;
			result->send.call.opt.len = 0;
			result->send.call.opt.buf = null;
			result->send.call.udata.maxlen = 0;
			result->send.call.udata.len = 0;
			result->send.call.udata.buf = null;
			
			// allocate space to retrieve information on where data comes from, and how data receipts occur
			result->receive.call.addr.maxlen = kReceiveBufferSizeAddress;
			result->receive.call.addr.len = 0;
			result->receive.call.addr.buf = (UInt8*)Memory_NewPtrCritical(kReceiveBufferSizeAddress);
			result->receive.call.opt.maxlen = kReceiveBufferSizeOptions;
			result->receive.call.opt.len = 0;
			result->receive.call.opt.buf = (UInt8*)Memory_NewPtrCritical(kReceiveBufferSizeOptions);
			result->receive.call.udata.maxlen = 0;
			result->receive.call.udata.len = 0;
			result->receive.call.udata.buf = null; // no unit data is expected
			
			// set up buffers
			result->send.normalData.maxlen = 0; // the size of the buffer is not yet known
			result->send.normalData.len = 0;
			result->send.normalData.buf = null;
			result->send.expeditedData.maxlen = 0; // the size of the buffer is not yet known
			result->send.expeditedData.len = 0;
			result->send.expeditedData.buf = null;
			result->receive.normalData.maxlen = 0; // the size of the buffer is not yet known
			result->receive.normalData.len = 0;
			result->receive.normalData.buf = null;
			result->receive.expeditedData.maxlen = 0; // the size of the buffer is not yet known
			result->receive.expeditedData.len = 0;
			result->receive.expeditedData.buf = null;
			result->authEncrData = null; // MacTelnet does not yet support Kerberos
		}
	}
	return result;
}// newNetworkDataPtr


/*!
This routine, of standard OTNotifyProcPtr form,
handles network events from Open Transport,
including connection openings, closings, data
transfers, and errors.  TCP/IP is assumed.

Open Transport calls this routine under several
circumstances:
- at provider-open time
- at completion time for various operations
  (such as binding and disconnecting), when this
  method looks at the error code (if any)
- at closing time, if the provider gets killed
  before an event succeeds

WARNING:	This routine is deferred-task-level.
			You cannot perform tasks such as
			making system calls or allocating
			memory.  If you want anything
			interesting to happen, you have to
			define and post an event to the
			internal event queue (NetEvents.c)
			so that MacTelnet will know what to
			do.

(3.0)
*/
static pascal void
openTransportGenericNotify		(void*				inoutContextPtr,
								 OTEventCode		inEventCode,
								 OTResult			inResult,
								 void*				inCookie)
{
	NetworkPortContextPtr		contextPtr = (NetworkPortContextPtr)inoutContextPtr;
	
	
	if (inEventCode == T_OPENCOMPLETE) // the first call of this routine for a particular event
	{
		InetSvcRef		providerRef = STATIC_CAST(inCookie, InetSvcRef);
		
		
		// Pass the new Internet Services reference in the structure,
		// and post an event that will continue the event *outside*
		// of this deferred-task-level routine.  THIS IS THE ENDPOINT,
		// and is elsewhere passed to functions that would expect
		// the type "EndpointRef".
		contextPtr->error = inResult;
		contextPtr->servicesRef = providerRef;
		contextPtr->endpointRef = STATIC_CAST(providerRef, EndpointRef); // for clarity, the endpoint is separated
		
		// get the initial endpoint info (NOTE that this can be re-negotiated
		// between now and after the connection is opened...)
		contextPtr->error = OTGetEndpointInfo(contextPtr->endpointRef, &contextPtr->endpointInfo);
		if (contextPtr->error == kOTNoError)
		{
			contextPtr->error = OTBind(contextPtr->endpointRef,
										null/* input bind info */, null/* updated bind info */);
			if (contextPtr->error == kOTNoAddressErr)
			{
				// no addresses available for binding!
			}
		}
		else
		{
			// failed to get endpoint information
		}
	}
	else if ((inEventCode == kOTProviderWillClose) ||
				(inEventCode == kOTProviderIsClosed)) // then the provider closed before this could finish
	{
		// If the provider is opened asynchronously (it is by default),
		// it is not possible to close it before the call has completed.
		// This might happen if the user quits the application before a
		// provider has opened.  For this reason, it is safer to open a
		// provider synchronously and then to use OTSetAsynchronous().
		// MacTelnet does not take this safer approach; in theory, a
		// user quitting MacTelnet, as described, could encounter delays.
		// This particular event is not deferred-task-level, so its safe
		// to do interesting things in response to this event (perhaps a
		// progress dialog could be displayed here...).
		contextPtr->error = inResult;
		(OSStatus)OTSetSynchronous(contextPtr->servicesRef);
		(OSStatus)OTCloseProvider(contextPtr->servicesRef);
		(OSStatus)OTSetAsynchronous(contextPtr->servicesRef);
		NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventFailure, contextPtr->portNumber, 0);
	}
	else if (inEventCode == T_BINDCOMPLETE) // attempt to bind to a local port has completed
	{
		// attempt a connection, now that binding is complete
		contextPtr->error = inResult;
		if (contextPtr->error == kOTNoError)
		{
			OSStatus		error = noErr;
			
			
			// since the send call information was set up at service
			// construction time, the IP address and port are already
			// available (in the "send.call" element), so this connection
			// knows the exact remote host and remote port to connect to
			contextPtr->error = OTConnect(contextPtr->endpointRef, &contextPtr->send.call,
											(TCall*)null/* unneeded, as this is an asynchronous call */);
			if (contextPtr->error == kOTNoDataErr)
			{
				// waiting for connection (the notifier gets sent an event when the attempt completes)
			}
			else
			{
				// at the time of this writing, no other result was documented;
				// if anything besides "kOTNoDataErr" results, is it a failure?
			}
		}
		else
		{
			// the bind did not succeed
		}
	}
	else if (inEventCode == T_CONNECT) // attempt to connect has completed
	{
		// did the connection succeed?
		contextPtr->error = OTRcvConnect(contextPtr->endpointRef,
											(TCall*)null/* unneeded, as this is an asynchronous call */);
		if (contextPtr->error == kOTNoError)
		{
			// first re-obtain the endpoint info (since this could have been re-negotiated
			// between endpoint creation time and connection opening, it must be updated)
			contextPtr->error = OTGetEndpointInfo(contextPtr->endpointRef, &contextPtr->endpointInfo);
			if (contextPtr->error == kOTNoError)
			{
				// connection was successfully opened - well, open a damned terminal window, already!
				NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventOpen,
								contextPtr->portNumber, 0);
				
				// initialize and allocate buffers using the data size information
				contextPtr->send.normalData.maxlen = contextPtr->endpointInfo.tsdu;
				contextPtr->send.normalData.buf = Memory_NewPtrInterruptSafe
													(contextPtr->send.normalData.maxlen);
				contextPtr->send.expeditedData.maxlen = contextPtr->endpointInfo.etsdu;
				contextPtr->send.expeditedData.buf = Memory_NewPtrInterruptSafe
														(contextPtr->send.expeditedData.maxlen);
				contextPtr->receive.normalData.maxlen = contextPtr->endpointInfo.tsdu;
				contextPtr->receive.normalData.buf = Memory_NewPtrInterruptSafe
														(contextPtr->receive.normalData.maxlen);
				contextPtr->receive.expeditedData.maxlen = contextPtr->endpointInfo.etsdu;
				contextPtr->receive.expeditedData.buf = Memory_NewPtrInterruptSafe
														(contextPtr->receive.expeditedData.maxlen);
			}
			else
			{
				// failed to get endpoint information
			}
		}
		else
		{
			// not connected; the only expected value here is "kOTNoDataErr"
		}
		
		if (contextPtr->error != kOTNoError)
		{
			// connection failed - give the user a post-mortem report
			NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventFailure,
							contextPtr->portNumber, contextPtr->error);
		}
	}
	else if ((inEventCode == T_DISCONNECT) || (inEventCode == T_ORDREL)) // forced disconnect (connection refused)
	{
		// find out why the connection was refused
		openTransportHandleDisconnect(inoutContextPtr, inEventCode);
	}
	else if (inEventCode == T_DISCONNECTCOMPLETE) // active attempt to disconnect has completed
	{
		contextPtr->error = OTUnbind(contextPtr->endpointRef);
		NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventClose,
						contextPtr->portNumber, contextPtr->error);
		if (contextPtr->error != kOTNoError)
		{
			// unbind failed
		}
	}
	else if ((inEventCode == T_DATA) || (inEventCode == T_EXDATA)) // data (or expedited data) has arrived
	{
		// post an event notifying that OTRcv() can now be called successfully
		NetEvents_PostProvisionally(kTelnetNetworkEventClassConnection, kConnectionEventData,
									contextPtr->portNumber, 0);
	}
	else if (inEventCode == T_GODATA) // data can now be sent
	{
		// a previously-interrupted attempt to send normal data is now a go; try again
		contextPtr->error = OTSnd(contextPtr->endpointRef, contextPtr->send.normalData.buf,
									contextPtr->send.normalData.maxlen, contextPtr->send.flags);
	}
	else if (inEventCode == T_GOEXDATA) // expedited data can now be sent
	{
		// a previously-interrupted attempt to send expedited data is now a go; try again
		contextPtr->error = OTSnd(contextPtr->endpointRef, contextPtr->send.expeditedData.buf,
									contextPtr->send.expeditedData.maxlen, contextPtr->send.flags);
	}
	else
	{
		// IGNORE ALL UNKNOWN EVENT CODES
	}
}// openTransportGenericNotify


/*!
This deferred-task-level method responds to the
T_DISCONNECT or T_ORDREL events.  The error
from handling the disconnect is returned in the
error flag of the context.

(3.0)
*/
static void
openTransportHandleDisconnect		(NetworkPortContextPtr		inoutContextPtr,
									 OTEventCode				inEventCode)
{
	if (inEventCode == T_ORDREL) inoutContextPtr->error = OTRcvOrderlyDisconnect(inoutContextPtr->endpointRef);
	else inoutContextPtr->error = OTRcvDisconnect(inoutContextPtr->endpointRef, (TDiscon*)null);
	NetEvents_Post(kTelnetNetworkEventClassConnection, kConnectionEventFailure, inoutContextPtr->portNumber, 0);
}// openTransportHandleDisconnect


/*!
This deferred-task-level method responds to the
T_LISTEN event.

(3.0)
*/
static void
openTransportHandleListen	(NetworkPortContextPtr	inoutContextPtr)
{
	// get information about the incoming connection request
	inoutContextPtr->error = OTListen(inoutContextPtr->endpointRef, &inoutContextPtr->receive.call);
	
	// on return from OTListen, the receive "TCall" structure is set
	// up as follows:
	//		"sequence" is the connection ID of the incoming request
	//		"addr" is where the data is coming from
	if (inoutContextPtr->error == kOTLookErr) openTransportHandleLook(inoutContextPtr);
	else
	{
		// if "kOTNoDataErr", then there are no pending connection requests
	}
}// openTransportHandleListen


/*!
Invoke this routine in response to receiving the
Open Transport result "kOTLookErr".  It is possible
for some operations (such as sending data) to get
interrupted, generating a look and see error, at
which point you must invoke this routine to avoid
deadlock.

This method should expect to run at deferred task
time, and therefore shouldnt do anything too
spectacular.

(3.0)
*/
static void
openTransportHandleLook		(NetworkPortContextPtr	inoutContextPtr)
{
	switch (OTLook(inoutContextPtr->endpointRef))
	{
		case T_DISCONNECT:
			// forced disconnect
			openTransportHandleDisconnect(inoutContextPtr, T_DISCONNECT);
			break;
		
		case T_ORDREL:
			// orderly disconnect
			openTransportHandleDisconnect(inoutContextPtr, T_ORDREL);
			break;
		
		default:
			break;
	}
}// openTransportHandleLook

#endif