/*###############################################################

	NewSessionsDialog.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <ControlDefinitions.h>
#include <Folders.h>
#include <Icons.h>
#include <Lists.h>
#include <MacErrors.h>
#include <Navigation.h>
#include <PLStringFuncs.h>
#include <Resources.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "StringResources.h"
#include "TelnetNewSessionsDialog.r"
#include "TelnetVersion.h"

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "configure.proto.h"
#include "Connections.h"
#include "Console.h"
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "MacLook.h"
#include "NewSessionsDialog.h"
#include "Preferences.h"
#include "ResourceFileManager.h"
#include "TelnetMenus.h"
#include "TelnetDescriptors.h"



// types
struct ListItemData
{
	ConnectionSetup**		tcHandle;
};
typedef struct ListItemData		ListItemData;
typedef ListItemData*			ListItemDataPtr;

// variables
static ChainRef				gConnectionsToOpen = null;
static DialogRef			gNewSessionsDialog = null; // the Mac OS dialog pointer
static WindowInfoRef		gNewSessionsDialogWindowInfo;
static MenuHandle			gProxiesPopUpMenuHandle = null,
							gTerminalsPopUpMenuHandle = null;
static Boolean				gDialogInUse = false,
							gHaveDeactivatedGroupBox = false,
							gHaveDeactivatedRemoveButton = true;
struct
{
	UInt16			numberOfSelections; // this counter guarantees that the original data is never destroyed
	Str255			hostNameFieldText,
					portNumberFieldText,
					windowTitleFieldText;
	SInt16			proxyCheckBoxValue;
	// incomplete
} gPreservedNewSession;

// internal methods
static void					addDataToList					(void);
static pascal Boolean		changeGroupBoxNameOperation		(ControlRef const			inTheNewSessionGroupBox,
															 SInt16 const				inColorDepth,
															 SInt16 const				inDeviceFlags,
															 GDHandle					inTargetDevice,
															 SInt32 const				inData1,
															 SInt32 const				inData2);
static void					connect							(void);
static void					deleteNodeListItemDataPtr		(ChainLinkRef const			inNode);
static void					getConnectionsListHandle		(ListHandle*				outListHandle);
static ListItemDataPtr		getNodeListItemDataPtr			(ChainLinkRef const			inNode);
static pascal Boolean		handleNewSize					(WindowRef const			inWindow,
														 	 SInt32 const				inDeltaSizeX,
															 SInt32 const				inDeltaSizeY,
															 SInt32 const				inData);
static void					makeAllControlsUpToDate			(void);
static ConnectionSetup**	nameToConnectionHandle			(StringPtr					inoutString,
															 StringPtr					inWindowName,
															 Boolean					inUseDefaultTerminal);
static ListItemDataPtr		newNodeListItemDataPtr			(ChainLinkRef const			inNode);
static pascal Boolean		newSessionsDialogEventFilter	(DialogRef					inDialog,
															 EventRecord*				inoutEventPtr,
															 short*						outItemIndex);
static void					preserveNewSessionData			(void);
static void					removeDataFromList				(void);
static Boolean				sessionNameToHostName			(StringPtr					inoutString);
static void					setDisclosureState				(Boolean const				inIsCollapsed);
static void					setNewSessionGroupBoxName		(Boolean const				inAreAnyListItemsSelected);
static void					setRemoveButtonState			(void);
static void					setSessionCheckBoxStates		(void);
static void					setUpProxiesPopUpMenu			(Boolean const				inRebuildMenu);
static void					setUpTerminalsPopUpMenu			(Boolean const				inRebuildMenu);
static void					toggleDisclosureState			(void);
static void					updateClear						(void);
static void					updateRestore					(void);
static void					updateUsingConnectionSetup		(ConnectionSetup**			inData);
static void					updateUsingSessionResource		(StringPtr					inSessionName);
static void					useNumericalLimiter				(ControlRef					inControl);



//
// public methods
//

/*!
This method is used to initialize the New Sessions
dialog box.  It creates the dialog box invisibly, and
sets up dialog controls.

The extended dialog resource for the New Sessions
dialog causes the Appearance Manager to automatically
construct an embedding hierarchy, so the dialogs
root control and container controls are supposed to
be set up once GetNewDialog() returns.

(3.0)
*/
void
NewSessionsDialog_Init	(void)
{
	Cursors_UseWatch();
	
	gDialogInUse = true;
	gHaveDeactivatedGroupBox = false;
	gHaveDeactivatedRemoveButton = true;
	
	GetNewDialogWithWindowInfo(kDialogIDNewSessions, &gNewSessionsDialog, &gNewSessionsDialogWindowInfo);
	
	gConnectionsToOpen = Chain_New(kChainStyleSinglyLinked);
	Chain_SetDescriptor(gConnectionsToOpen, kTelnetChainDescriptorConnectionList);
	
	if (gNewSessionsDialog != null)
	{
		// determine if the Help button should be disabled...
		SetUpHelpButtonDialogItem(gNewSessionsDialog, iNewSessionsAppleGuideButton);
		
		// outline the default button
		SetDialogDefaultItem(gNewSessionsDialog, iNewSessionsConnectButton);
		SetDialogCancelItem(gNewSessionsDialog, iNewSessionsCancelButton);
		
		// adjust the buttons so they are big enough for their titles
		(UInt16)Localization_AdjustDialogButtons(gNewSessionsDialog, true/* adjust default button */,
													true/* adjust Cancel button */);
		Localization_AdjustHelpButtonItem(gNewSessionsDialog, iNewSessionsAppleGuideButton);
		
		// under Appearance 1.1 or later, the key script property is supported; domain names
		// require the Roman script and cant use others, so require that script for the host field
		if (FlagManager_Test(kFlagAppearance1_1API))
		{
			ControlRef					hostField = null;
			ControlKeyScriptBehavior	hostFieldBehavior = kControlKeyScriptBehaviorRequiresRoman;
			
			
			(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsHostNameField, &hostField);
			(OSErr)SetControlData(hostField, kControlEditTextPart, kControlEditTextKeyScriptBehaviorTag,
									sizeof(hostFieldBehavior), &hostFieldBehavior);
		}
		
		// set the controls properly
		setUpProxiesPopUpMenu(true/* rebuild menu */);
		setUpTerminalsPopUpMenu(true/* rebuild menu */);
		{
			Str255		sessionName;
			
			
			StringUtilities_CToP(SESSIONPREFS_DEFAULTNAME, sessionName);
			updateUsingSessionResource(sessionName);
		}
		setNewSessionGroupBoxName(false/* there are items selected */);
		setRemoveButtonState();
		{
			ControlRef		portNumberField = null;
			
			
			(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsPortNumberField, &portNumberField);
			useNumericalLimiter(portNumberField);
		}
		
		// minor addition - remove those annoying delays in displaying bevel button pop-up menus;
		// also, set the Recent Items button to a 32-bit icon, if there is one available
		{
			ControlRef		bevelButtonPopUp = null;
			SInt32			delay = 0L; // in ticks
			
			
			// Favorites button
			if (GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsFavoritesPopUpMenuBevelButton,
										&bevelButtonPopUp) == noErr)
			{
				(OSErr)SetControlData(bevelButtonPopUp, kControlNoPart, kControlBevelButtonMenuDelayTag,
										sizeof(delay), &delay);
			}
			
			// Recent Items button
			if (GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsRecentPopUpMenuBevelButton,
										&bevelButtonPopUp) == noErr)
			{
				(OSErr)SetControlData(bevelButtonPopUp, kControlNoPart, kControlBevelButtonMenuDelayTag,
										sizeof(delay), &delay);
				
				// in addition, use a 32-bit icon, if available
				{
					IconManagerIconRef		icon = IconManager_NewIcon();
					
					
					if (IconManager_MakeIconRef(icon, kOnSystemDisk, kSystemIconsCreator, kRecentItemsIcon) == noErr)
					{
						(OSErr)IconManager_SetButtonIcon(bevelButtonPopUp, icon);
					}
					IconManager_DisposeIcon(&icon);
				}
			}
		}
		
		// in Simplified User Interface mode, the detailed information is irrelevant
		// (chances are very good that novices will not need to set a non-23 port number,
		// and proxy server and terminal stuff is really complicated)
		{
			SInt32		actualSize = 0L;
			Boolean		isSimpleMode = false;
			
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
								kTelnetPreferenceTag_SimpleTelnet,
								sizeof(isSimpleMode), &isSimpleMode,
								&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				isSimpleMode = false; // assume normal mode, if preference cant be found
			}
			if (isSimpleMode)
			{
				HideDialogItem(gNewSessionsDialog, iNewSessionsPortNumberLabel);
				HideDialogItem(gNewSessionsDialog, iNewSessionsPortNumberField);
				HideDialogItem(gNewSessionsDialog, iNewSessionsDisclosureTriangle);
				HideDialogItem(gNewSessionsDialog, iNewSessionsDisclosureTriangleLabel);
			}
		}
		
		// TEMP - this button doesnt do anything at the moment, so disable it
		DeactivateDialogItem(gNewSessionsDialog, iNewSessionsRecentPopUpMenuBevelButton);
		
		// set up the Window Info information
		{
			WindowResizeResponderProcPtr	resizeProc = handleNewSize;
			
			
			WindowInfo_SetWindowDescriptor(gNewSessionsDialogWindowInfo, kTelnetDialogDescriptorNewSessions);
			WindowInfo_SetWindowResizeLimits(gNewSessionsDialogWindowInfo,
													NEWSESSIONS_DIALOG_HT/* minimum height */,
													NEWSESSIONS_DIALOG_WD/* minimum width */,
													SHRT_MAX/* maximum height */,
													SHRT_MAX/* maximum width */);
			WindowInfo_SetWindowResizeResponderUPP(gNewSessionsDialogWindowInfo, resizeProc, 0L);
		}
		WindowInfo_SetForDialog(gNewSessionsDialog, gNewSessionsDialogWindowInfo);
		
		// use the preferred rectangle, if any
		{
			Point		deltaSize;
			
			
			SetPt(&deltaSize, NEWSESSIONS_DIALOG_WD, NEWSESSIONS_DIALOG_HT); // initially...
			deltaSize = Preferences_ArrangeWindow(GetDialogWindow(gNewSessionsDialog), deltaSize,
													kTelnetWindowPreferenceTag_NewSessions);
			
			(Boolean)handleNewSize(GetDialogWindow(gNewSessionsDialog), (SInt32)deltaSize.h, (SInt32)deltaSize.v, 0L);
		}
		
		// TEMP - the below disclosure call does not work properly on all systems
		HideDialogItem(gNewSessionsDialog, iNewSessionsDisclosureTriangle); // TEMP
		HideDialogItem(gNewSessionsDialog, iNewSessionsDisclosureTriangleLabel); // TEMP
		// set up dialog box based on whether or not its initially collapsed
		//setDisclosureState(true/* is collapsed */);
	}
}// Init


/*!
Call this method in the exit method of the program
to ensure that this dialogs memory allocations
get destroyed.  You can also call this method after
you are done using this dialog, provided you call
NewSessionsDialog_Init() before you use this
dialog again.

(3.0)
*/
void
NewSessionsDialog_Done	(void)
{
	gDialogInUse = false;
	WindowInfo_Dispose(gNewSessionsDialogWindowInfo);
	if (gProxiesPopUpMenuHandle != null) DisposeMenu(gProxiesPopUpMenuHandle), gProxiesPopUpMenuHandle = null;
	if (gTerminalsPopUpMenuHandle != null) DisposeMenu(gTerminalsPopUpMenuHandle), gTerminalsPopUpMenuHandle = null;
	
	{
		ChainLinkRef	node = null;
		
		
		node = Chain_GetFirstNode(gConnectionsToOpen);
		do
		{
			deleteNodeListItemDataPtr(node);
			node = Chain_GetNextNode(gConnectionsToOpen, node);
		} while (node != null);
	}
	Chain_Dispose(&gConnectionsToOpen);
	DisposeDialog(gNewSessionsDialog), gNewSessionsDialog = null;
}// Done


/*!
This method displays and handles events in the
New Sessions dialog box.

(3.0)
*/
void
NewSessionsDialog_Display	(void)
{
	// The network must be connected prior to displaying this dialog
	// box.  If the network is not connected and an attempt to open
	// multiple connections is made, the computer may stop responding
	// when it tries to start the connectivity software.
	Boolean		connected = true;//Network_Connect();
	
	
	if (connected) if (gNewSessionsDialog != null)
	{
		ModalFilterUPP				filterProc = NewModalFilterUPP(newSessionsDialogEventFilter);
		register DialogItemIndex	i = 0;
		DialogItemIndex				itemIndex = 0;
		
		
		// display the dialog
		TransitionDialogDisplay(gNewSessionsDialog);
		
		// focus the host name field
		{
			ControlRef		fieldControl = null;
			
			
			(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsHostNameField, &fieldControl);
			(OSErr)SetKeyboardFocus(GetDialogWindow(gNewSessionsDialog), fieldControl, kControlEditTextPart);
		}
		
		// handle events
		itemIndex = 0;
		do
		{
			ModalDialog(filterProc, &itemIndex);
			switch (itemIndex)
			{
				case iNewSessionsAddButton:
					addDataToList();
					SetDialogDefaultItem(gNewSessionsDialog, iNewSessionsAddButton);
					break;
				
				case iNewSessionsRemoveButton:
					removeDataFromList();
					setRemoveButtonState();
					break;
				
				case iNewSessionsAppleGuideButton:
					{
						Str255		searchString;
						
						
						GetIndString(searchString, rStringsAppleGuideLookup, siAGIndexTermConnections);
						AppleGuideUtilities_OpenFirstMainGuideInSearch(searchString);
					}
					break;
				
				case iNewSessionsDisclosureTriangle:
					toggleDisclosureState();
					break;
				
				case iNewSessionsUseProxyServerCheckBox:
					FlipCheckBox(gNewSessionsDialog, itemIndex);
					setSessionCheckBoxStates();
					break;
				
				case iNewSessionsTerminalsPopUpMenu:
					// keep the arrow control up to date
					SetDialogItemValue(gNewSessionsDialog, iNewSessionsTerminalChoiceLittleArrows,
										GetDialogItemValue(gNewSessionsDialog, iNewSessionsTerminalsPopUpMenu));
					break;
				
				case iNewSessionsTerminalChoiceLittleArrows:
					// necessarily handled by the event filter procedure
					break;
				
				case iNewSessionsFavoritesPopUpMenuBevelButton:
					{
						ControlRef		control = null;
						OSErr			error = noErr;
						
						
						error = GetDialogItemAsControl(gNewSessionsDialog,
														iNewSessionsFavoritesPopUpMenuBevelButton, &control);
						if (error == noErr)
						{
							MenuHandle		menuHandle = null;
							
							
							error = GetBevelButtonMenuHandle(control, &menuHandle);
							if (error == noErr)
							{
								Str255		name;
								SInt16		menuValue = 0;
								
								
								error = GetBevelButtonMenuValue(control, &menuValue);
								if (error == noErr)
								{
									SetDialogDefaultItem(gNewSessionsDialog, iNewSessionsConnectButton);
									GetMenuItemText(menuHandle, menuValue, name);
									updateUsingSessionResource(name);
									addDataToList();
								}
							}
						}
						
						if (error != noErr)
						{
							Console_WriteValue("new sessions Favorites pop-up menu error", error);
						}
					}
					break;
				
				case iNewSessionsConnectionsToOpenList:
					// on double-click, remove selected items automatically
					{
						ControlRef		listControl = null;
						
						
						if (GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsConnectionsToOpenList,
													&listControl) == noErr)
						{
							Boolean		isDoubleClick = false;
							Size		actualSize = 0;
							
							
							if (GetControlData(listControl, kControlListBoxDoubleClickPart,
												kControlListBoxDoubleClickTag, sizeof(Boolean), (Ptr)&isDoubleClick,
												&actualSize) == noErr)
							{
								if (isDoubleClick) removeDataFromList();
							}
						}
					}
					setRemoveButtonState();
					break;
				
				default:
					break;
			}
		} while ((itemIndex != iNewSessionsConnectButton) && (itemIndex != iNewSessionsCancelButton));
		DisposeModalFilterUPP(filterProc), filterProc = null;
		
		// save window size and location in preferences
		Preferences_SaveWindowArrangement(GetDialogWindow(gNewSessionsDialog),
											kTelnetWindowPreferenceTag_NewSessions);
		
		if (itemIndex != iNewSessionsCancelButton)
		{
			ListHandle		list = null;
			
			
			// open the connections, and then kill the dialog
			getConnectionsListHandle(&list);
			if (ListUtilities_GetItemCount(list) <= 0) addDataToList();
			SetDialogDefaultItem(gNewSessionsDialog, 0);
			DeactivateFrontmostWindow();
			connect();
			TransitionDialogClose(gNewSessionsDialog);
		}
		else
		{
			// user cancelled - close the dialog with an appropriate transition for cancelling
			TransitionDialogCloseCancel(gNewSessionsDialog);
		}
	}
	else Alert_ReportOSErr(memFullErr);
	
	gDialogInUse = false;
}// Display


/*!
Returns "true" if the dialog box is not in
a completely dead state - this means that
if the dialog has been initialized but is
not visible, or if it is visible, or if it
is invisible but its preferences have not
yet been saved, "true" is returned.  This
routine is used only for scripting.

(3.0)
*/
Boolean
NewSessionsDialog_IsDisplayed	(void)
{
	Boolean		result = gDialogInUse;
	
	
	return result;
}// IsDisplayed


//
// internal methods
//

/*!
To take the data currently in the dialog box,
compose a connection data structure and add it
to the list of connections to open, use this
method.  The data is then cleared from the
dialog in preparation for a new connection,
re-focusing the Host Name field.

If the number of connections already open
plus the number of connections in the list is
equal to the maximum number of connections,
or if memory problems occur, a message is
displayed to the user saying that the new
connection cannot be added to the list.

Incomplete.

(3.0)
*/
static void
addDataToList	(void)
{
	Str255		hostHintString;
	
	
	GetDialogItemControlText(gNewSessionsDialog, iNewSessionsHostNameField, hostHintString);
	
	if (StrLength(hostHintString) <= 0)
	{
		// there has to be some text entered there; let the user
		// know that a blank is unacceptable
		BEEP;
	}
	else
	{
		ListHandle		list = null;
		Str31			portNumberString;
		ControlRef		control = null;
		UInt16			listItemCount = 0;
		Boolean			cantAdd = false;
		
		
		GetDialogItemControlText(gNewSessionsDialog, iNewSessionsPortNumberField, portNumberString);
		
		// construct a <host name>:<port number> string (this is used when processing)
		PLstrcat(hostHintString, "\p:");
		PLstrcat(hostHintString, portNumberString);
		
		getConnectionsListHandle(&list);
		if (list != null)
		{
			listItemCount = ListUtilities_GetItemCount(list);
			cantAdd = ((1 + listItemCount + Connections_GetCount()) > kMaximumAllowedTelnetConnections);
		}
		else cantAdd = true;
		
		unless (cantAdd) // try to add the connection...it may still not be possible, for memory reasons
		{
			Str255				windowTitleString;
			Cell				cell;
			ListItemDataPtr		newLink = null;
			ChainLinkRef		newNode = null;
			long				junk = 0L;
			
			
			// get the data and add it
			GetDialogItemControlText(gNewSessionsDialog, iNewSessionsWindowTitleField, windowTitleString);
			
			MaxMem(&junk);
			
			newNode = Chain_NewNode(gConnectionsToOpen, null);
			newLink = newNodeListItemDataPtr(newNode);
			Chain_MoveNode(gConnectionsToOpen, newNode, null, kChainLinkPlacementInsertBefore);
			newLink->tcHandle = nameToConnectionHandle(hostHintString, windowTitleString,
														false/* use default terminal */);
			if (newLink->tcHandle == null) cantAdd = true;
			
			unless (cantAdd)
			{
				ConnectionSetup**		theConnectionDataHandle = newLink->tcHandle;
				Str255					popUpTerminalString;
				Str15					connectionTypeString;
				
				
				// set up the data to reflect the dialog box control states
				HLock((Handle)theConnectionDataHandle);
				HLock((Handle)(**theConnectionDataHandle).session);
				if ((**theConnectionDataHandle).terminal == null)
				{
					UInt16		selectedPopUpTerminalItem = 0;
					
					
					selectedPopUpTerminalItem = GetDialogItemValue(gNewSessionsDialog,
																	iNewSessionsTerminalsPopUpMenu);
					GetMenuItemText(gTerminalsPopUpMenuHandle, selectedPopUpTerminalItem, popUpTerminalString);
					
					// get the specified terminal and make the handle accessible for modification
					{
						SInt16		oldResFile = CurResFile();
						
						
						ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
						(**theConnectionDataHandle).terminal = (TerminalPrefs**)Get1NamedResource
																	(TERMINALPREFS_RESTYPE, popUpTerminalString);
						UseResFile(oldResFile);
					}
					DetachResource((Handle)(**theConnectionDataHandle).terminal);
				}
				
				(**theConnectionDataHandle).ftpState = ((**((**theConnectionDataHandle).session)).port == 21) ? 1 : 0;
				PLstrcpy(connectionTypeString, "\pTelnet");
				
				// Kerberos is unsupported - clear authentication and encryption flags
				(**(**theConnectionDataHandle).session).authenticate = 0;
				(**(**theConnectionDataHandle).session).encrypt = 0;
				
				// grab the window title
				GetDialogItemControlText(gNewSessionsDialog, iNewSessionsWindowTitleField,
										(*theConnectionDataHandle)->windowName);
				
				HUnlock((Handle)(**theConnectionDataHandle).session);
				HUnlock((Handle)theConnectionDataHandle);
				
				// add the item to the top of the list
				(SInt16)LAddRow(1, 0, list);
				
				// set the column data
				SetPt(&cell, 0, 0); // connection type
				//LSetCell((Ptr)(1 + connectionTypeString), StrLength(connectionTypeString), cell, list);
				//SetPt(&cell, 1, 0); // host
				PLstrcat(hostHintString, "\p:");
				PLstrcat(hostHintString, portNumberString);
				LSetCell((Ptr)(1 + hostHintString), StrLength(hostHintString), cell, list);
				//SetPt(&cell, 2, 0); // window title
				//LSetCell((Ptr)(1 + (*theConnectionDataHandle)->windowName),
				//			StrLength((*theConnectionDataHandle)->windowName) * sizeof(unsigned char), cell, list);
				//SetPt(&cell, 3, 0); // terminal
				//LSetCell((Ptr)(1 + popUpTerminalString), StrLength(popUpTerminalString), cell, list);
				//SetPt(&cell, 4, 0); // proxy
				//LSetCell((Ptr)(1 + popUpTerminalString), StrLength(popUpTerminalString), cell, list);
				// unimplemented
				
				// update the list
				DrawOneDialogItem(gNewSessionsDialog, iNewSessionsConnectionsToOpenList);
				
				// clear data entry area and the internal new session data buffer
				updateClear();
				preserveNewSessionData(); // reads the now-cleared controls, effectively clearing the internal buffers
				
				// theme sounds!
				if (FlagManager_Test(kFlagAppearance1_1API))
				{
					(OSStatus)PlayThemeSound(kThemeSoundNewItem);
				}
			}
		}
		
		if (cantAdd)
		{			
			ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
			InterfaceLibAlertRef	box = null;
			Str255					dialogText;
			
			
			GetIndString(dialogText, rStringsGeneralMessages, siWillNotBeAbleToOpenThatManyConnections);
			box = Alert_New();
			Alert_SetFilterProc(box, upp);
			Alert_SetHelpButton(box, false);
			Alert_SetParamsFor(box, kAlert_StyleOK);
			Alert_SetText(box, dialogText, EMPTY_PSTRING);
			Alert_SetType(box, kAlertStopAlert);
			BEEP;
			Alert_Display(box);
			Alert_Dispose(&box);
			DisposeModalFilterUPP(upp), upp = null;
		}
	}
}// addDataToList


/*!
This method is in standard OffscreenOperationProcPtr
form, and will use "inData1" as a string pointer
(StringPtr) for the new name of the specified text
group box control ("inSpecificControlOrRoot"), and
"inData2" as the flag are any list items selected?.
This operation is only intended for the New Session
group box; it is not extendable to other types of
controls directly.

If the operation cannot be completed, false is
returned; otherwise, true is returned.

(1.0)
*/
static pascal Boolean
changeGroupBoxNameOperation		(ControlRef const		inTheNewSessionGroupBox,
								 SInt16 const			UNUSED_ARGUMENT(inColorDepth),
								 SInt16 const			UNUSED_ARGUMENT(inDeviceFlags),
								 GDHandle				UNUSED_ARGUMENT(inTargetDevice),
								 SInt32 const			inData1,
								 SInt32 const			UNUSED_ARGUMENT(inData2))
{
	OSErr			error = noErr;
	Boolean			result = pascal_true;
	
	
	if (error == noErr)
	{
		SetControlTitle(inTheNewSessionGroupBox, (StringPtr)inData1);
	}
	else result = pascal_false;
	
	return result;
}// changeGroupBoxNameOperation


/*!
To open connections for all of the items
in the list "gConnectionsToOpen", use
this method.  An excellent opportunity
for multithreading, when the time comes...
just use Chain_GetIndexedNode() instead
of Chain_GetNextNode().

(3.0)
*/
static void
connect		(void)
{
	ChainLinkRef		item = Chain_GetFirstNode(gConnectionsToOpen);
	ListItemDataPtr		dataPtr = null;
	ListHandle			list = null;
	Boolean				successful = false;
	
	
	Cursors_UseWatch();
	getConnectionsListHandle(&list);
	while (item != null)
	{
		dataPtr = getNodeListItemDataPtr(item);
		if (dataPtr->tcHandle != null) successful = Connections_CreateOne(dataPtr->tcHandle);
		item = Chain_GetNextNode(gConnectionsToOpen, item);
		if (list != null)
		{
			ChainLinkRef		deletedNode = null;
			
			
			LDelRow(1, 0, list); // delete one row, the first row (0)
			deletedNode = Chain_DeleteFirstNode(gConnectionsToOpen);
			deleteNodeListItemDataPtr(Chain_DisposeNode(gConnectionsToOpen, &deletedNode));
			LSetDrawingMode(true, list);
			DrawOneDialogItem(gNewSessionsDialog, iNewSessionsConnectionsToOpenList);
		}
	}
	Cursors_UseArrow();
}// connect


/*!
To destroy memory for a ListItemData
structure previously allocated using the
newNodeListItemDataPtr() method, use this
routine.  The node data pointer is set to
null.

(3.0)
*/
static void
deleteNodeListItemDataPtr		(ChainLinkRef const		inNode)
{
	ListItemDataPtr		data = null;
	
	
	data = (ListItemDataPtr)Chain_GetNodeDataPtr(gConnectionsToOpen, inNode);
	Memory_DisposePtr((Ptr*)&data);
	Chain_SetNodeDataPtr(gConnectionsToOpen, inNode, null);
}// deleteNodeListItemDataPtr


/*!
To get a standard ListHandle representation
of the connections to open list, use this
method.  If any problems occur, the output
list will be null.

(3.0)
*/
static void
getConnectionsListHandle		(ListHandle*	outListHandle)
{
	ListUtilities_GetDialogItemListHandle(gNewSessionsDialog, iNewSessionsConnectionsToOpenList, outListHandle);
}// getConnectionsListHandle


/*!
To obtain a reference to the internal
connection data structure, given a linked
list node reference, use this method.

(3.0)
*/
static ListItemDataPtr
getNodeListItemDataPtr		(ChainLinkRef const		inNode)
{
	return ((ListItemDataPtr)Chain_GetNodeDataPtr(gConnectionsToOpen, inNode));
}// getNodeListItemDataPtr


/*!
This method moves the buttons in response to a resize of
the New Sessions movable modal dialog box.

(3.0)
*/
static pascal Boolean
handleNewSize		(WindowRef const		inWindow,
				 	 SInt32 const			inDeltaSizeX,
					 SInt32 const			inDeltaSizeY,
					 SInt32 const			UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_false;
	
	
	DialogAdjust_BeginDialogItemAdjustment(GetDialogFromWindow(inWindow));
	{
		// install proposed adjustments into appropriate lists
		SInt32		amountH = 0L,
					amountV = 0L;
		
		
		// controls which are moved in both directions
		amountH = inDeltaSizeX;
		amountV = inDeltaSizeY;
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsConnectButton, amountV);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsCancelButton, amountV);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsAppleGuideButton, amountV);
		if (Localization_IsLeftToRight())
		{
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsConnectButton, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsCancelButton, amountH);
		}
		else
		{
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsAppleGuideButton, amountH);
		}
		
		// controls which are resized in both directions
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsConnectionsToOpenList, amountH);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeV, iNewSessionsConnectionsToOpenList, amountV);
		
		if (inDeltaSizeX)
		{
			// controls which are moved horizontally
			amountH = inDeltaSizeX;
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsTerminalChoiceLittleArrows, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsProxyChoiceLittleArrows, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsAddButton, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsRecentPopUpMenuBevelButton, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsFavoritesPopUpMenuBevelButton, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsRemoveButton, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsPortNumberLabel, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iNewSessionsPortNumberField, amountH);
			
			// controls which are resized horizontally
			amountH = inDeltaSizeX;
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsBottomSeparator, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsSessionInfoPrimaryGroupBox, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsDisclosureTriangleLabel, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsDisclosedContentPane, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsSessionInfoSeparator, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsWindowTitleField, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsTerminalsPopUpMenu, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsProxiesPopUpMenu, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iNewSessionsHostNameField, amountH);
		}
		
		if (inDeltaSizeY)
		{
			// controls which are moved vertically
			amountV = inDeltaSizeY;
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsBottomSeparator, amountV);
		}
	}
	DialogAdjust_EndAdjustment(inDeltaSizeX, inDeltaSizeY); // moves and resizes controls properly
	
	// now make sure the cell width of the list is correct
	{
		ListHandle		list = null;
		
		
		ListUtilities_GetDialogItemListHandle(GetDialogFromWindow(inWindow), iNewSessionsConnectionsToOpenList, &list);
		{
			Point			newCellDimensions;
			ControlRef		listControl = null;
			Rect			listBounds;
			
			
			(OSErr)GetDialogItemAsControl(GetDialogFromWindow(inWindow), iNewSessionsConnectionsToOpenList, &listControl);
			GetControlBounds(listControl, &listBounds);
			SetPt(&newCellDimensions, listBounds.right - listBounds.left, 18);
			LSetDrawingMode(false, list);
			LCellSize(newCellDimensions, list);
			LSetDrawingMode(true, list);
		}
	}
	
	return result;
}// handleNewSize


/*!
To analyze the host name field, checkboxes, and
other controls and make sure all dependent
control values and states are current, use this
method.

(3.0)
*/
static void
makeAllControlsUpToDate	(void)
{
	setRemoveButtonState();
	setSessionCheckBoxStates();
	setUpProxiesPopUpMenu(false/* rebuild menu */);
	setUpTerminalsPopUpMenu(false/* rebuild menu */);
}// makeAllControlsUpToDate


/*!
This method recognizes the following input string: xxxx yyyy.
If xxxx matches a session name, that session record is used.
Otherwise, the default session record is used with xxxx as
the DNS hostname.   The yyyy, if provided, is converted to a
number.  If it is a valid number, it is used as the port to
connect to.  The second argument specifies whether to get the
terminal stuff from the session preferences.  If not, the
terminal is set to null.

A new connection handle is returned.  Dispose of it using the
Memory_DisposeHandle() routine.

WARNING:	Do not pass this routine an empty name string.

(2.6)
*/
static ConnectionSetup**
nameToConnectionHandle	(StringPtr		inoutString,
						 StringPtr		inWindowName,
						 Boolean		inUseDefaultTerminal)
{
	ConnectionSetup**	result = null;
	SessionPrefs**		sessionHandle = null;
	TerminalPrefs**		terminalHandle = null;
	short				portRequested = 23; // by default, the Telnet port
	Boolean				foundPort = false,
						wasAlias = false;
	
	
	result = (ConnectionSetup**)Memory_NewHandleCritical(sizeof(ConnectionSetup));
	if (result != null)
	{
		foundPort = Connections_ProcessHostNameString(inoutString, &portRequested);
		
		ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
		sessionHandle = (SessionPrefs**)Get1NamedResource(SESSIONPREFS_RESTYPE, inoutString);
		if (sessionHandle == null) 
		{				// connect to host xxxx w/default session
			sessionHandle = Preferences_GetDefaultSession();
			DetachResource((Handle)sessionHandle);
			HLock((Handle)sessionHandle);
			BlockMove(inoutString, (**sessionHandle).hostname, 64);
		}
		else 
		{	
			DetachResource((Handle) sessionHandle);
			HLock((Handle)sessionHandle);
			wasAlias = true;
		}
		
		(**result).session = sessionHandle;
		
		ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
		if ((inUseDefaultTerminal) || (wasAlias))
		{	
			terminalHandle = (TerminalPrefs**)Get1NamedResource
												(TERMINALPREFS_RESTYPE, (**sessionHandle).TerminalEmulation);
			if (terminalHandle == null) terminalHandle = Preferences_GetDefaultTerminal();
			DetachResource((Handle)terminalHandle);
			(**result).terminal = terminalHandle;
		}
		else (**result).terminal = null;
		
		// set up open timeout
		{
			Size		actualSize = 0L;
			
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TimeoutOpen,
									sizeof((**result).openTimeout), &(**result).openTimeout,
									&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				(**result).openTimeout = 10; // assume a 10-second timeout, if preference cant be found
			}
		}
		
		// swap out space
		{
			long 		junk = 0L;
			
			
			MaxMem(&junk);
		}
		HUnlock((Handle)sessionHandle);
		
		{
			UInt16		length = StrLength(inWindowName);
			
			
			if (length > 0) BlockMove(inWindowName, (**result).windowName, (length > 63) ? 64 : (length + 1));
			else ((**result).windowName)[0] = 0;
		}
		(**result).ftpState = ((**((**result).session)).port == 21) ? 1 : 0;
		(**sessionHandle).ip_address = 0;
		
		if (foundPort) (**sessionHandle).port = portRequested;
	}
	return result;
}// nameToConnectionHandle


/*!
To allocate memory for a new ListItemData
structure and associate its pointer with
the given linked list node, use this method.
Use deleteNodeListItemDataPtr() to destroy
this allocated memory later.

A pointer to the new structure is returned,
or null if any problems occur (out of
memory).

(3.0)
*/
static ListItemDataPtr
newNodeListItemDataPtr		(ChainLinkRef const		inNode)
{
	ListItemDataPtr		result = (ListItemDataPtr)Memory_NewPtr(sizeof(ListItemData));
	
	
	result->tcHandle = null;
	Chain_SetNodeDataPtr(gConnectionsToOpen, inNode, result);
	return result;
}// newNodeListItemDataPtr


/*!
Handles special key events in the New Sessions
dialog box.  F chooses the FTP option from the
Connection Method pop-up menu, and T chooses the
Telnet option.  A checks Authentication.  +
clicks the Add button, and - clicks the Remove
button.

(3.0)
*/
static pascal Boolean
newSessionsDialogEventFilter	(DialogRef		inDialog,
								 EventRecord*	inoutEventPtr,
								 short*			outItemIndex)
{
	Boolean		result = pascal_false,
				isButtonAndShouldFlash = false; // simplifies handling of multiple key-to-button maps
	
	
	// first, make sure help balloons show up properly for the cursor location, if Balloon Help is on
	if (IsPopUpHelpEnabled()) DisplayBalloonForDialogItemUnderMouse(inDialog, inoutEventPtr->where,
																		kDialogBalloonsIDNewSessions);
	
	// now handle events
	if (inoutEventPtr->what == mouseDown)
	{
		// this rigamorole is necessary to distinguish between up-arrow and down-arrow clicks!
		GrafPtr				oldPort = null;
		Point				localMouse = inoutEventPtr->where;
		ControlRef			controlHit = null,
							arrows = null;
		ControlPartCode		part = kControlNoPart;
		
		
		GetPort(&oldPort);
		SetPortWindowPort(GetDialogWindow(inDialog));
		GlobalToLocal(&localMouse);
		controlHit = FindControlUnderMouse(localMouse, GetDialogWindow(inDialog), &part);
		GetDialogItemAsControl(inDialog, iNewSessionsTerminalChoiceLittleArrows, &arrows);
		if ((controlHit == arrows) && (IsControlActive(arrows)))
		{
			part = HandleControlClick(controlHit, localMouse, inoutEventPtr->modifiers, (ControlActionUPP)null);
			switch (part)
			{
				case kControlUpButtonPart:
				case kControlDownButtonPart:
					{
						ControlRef		control = null;
						SInt16			value = GetControlValue(arrows);
						Boolean			changed = false;
						
						
						// change arrows value...this is reversed, due to the effect the user would expect
						// each arrow to have on the next item of the Terminals pop-up menu control
						if ((part == kControlUpButtonPart) && (value > GetControlMinimum(arrows)))
						{
							value--;
							changed = true;
						}
						else if ((part == kControlDownButtonPart) && (value < GetControlMaximum(arrows)))
						{
							value++;
							changed = true;
						}
						if (changed)
						{
							SetControlValue(arrows, value);
							SetDialogItemValue(inDialog, iNewSessionsTerminalsPopUpMenu, value);
						}
					}
					result = pascal_true;
					break;
				
				default:
					// ???
					break;
			}
		}
		SetPort(oldPort);
	}
	else if (inoutEventPtr->what == keyDown)
	{
		short		key = 0;
		
		
		key = inoutEventPtr->message & charCodeMask;
		if (((key == '+') || (key == '=') || (key == 'A') || (key == 'a')) && (inoutEventPtr->modifiers & cmdKey))
		{
			ControlRef		button = null;
			
			
			(OSErr)GetDialogItemAsControl(inDialog, iNewSessionsAddButton, &button);
			if (IsControlActive(button))
			{
				*outItemIndex = iNewSessionsAddButton;
				isButtonAndShouldFlash = result = pascal_true;
			}
		}
		else if (((key == '-') || (key == '_') || (key == 'R') || (key == 'r')) && (inoutEventPtr->modifiers & cmdKey))
		{
			ControlRef		button = null;
			
			
			(OSErr)GetDialogItemAsControl(inDialog, iNewSessionsRemoveButton, &button);
			if (IsControlActive(button))
			{
				*outItemIndex = iNewSessionsRemoveButton;
				isButtonAndShouldFlash = result = pascal_true;
			}
		}
	#if 0
		else if (((key == 'D') || (key == 'd')) && (inoutEventPtr->modifiers & cmdKey))
		{
			ControlRef		triangle = null;
			
			
			(OSErr)GetDialogItemAsControl(inDialog, iNewSessionsDisclosureTriangle, &triangle);
			if (IsControlActive(triangle))
			{
				SInt16			newValue = kControlCheckBoxUncheckedValue;
				unsigned long	dummy = 0L;
				
				
				newValue = (GetControlValue(triangle) == kControlCheckBoxCheckedValue)
							? kControlCheckBoxUncheckedValue : kControlCheckBoxCheckedValue;
				if (FlagManager_Test(kFlagAppearance1_1API))
				{
					// play sound when highlighting button
					(OSStatus)PlayThemeSound(kThemeSoundDisclosurePress);
				}
				HiliteControl(triangle, kControlTrianglePart);
				Delay(8, &dummy);
				HiliteControl(triangle, kControlNoPart);
				if (FlagManager_Test(kFlagAppearance1_1API))
				{
					// play sound when unhighlighting button
					(OSStatus)PlayThemeSound(kThemeSoundDisclosureRelease);
				}
				SetDisclosureTriangleLastValue(triangle, newValue);
				DrawOneControl(triangle);
				SetControlValue(triangle, newValue);
				//DrawOneControl(triangle);
				*outItemIndex = iNewSessionsDisclosureTriangle;
				result = pascal_true;
			}
		}
	#endif
	}
	
	if (isButtonAndShouldFlash) FlashButton(inDialog, *outItemIndex);
	if (!result) result = (StandardDialogEventFilter(inDialog, inoutEventPtr, outItemIndex));
	
	return result;
}// newSessionsDialogEventFilter


/*!
Use this method to restore all of the controls in the
New Session primary group box to contain the global
session variables.  This allows the user to type in
information for a new session, NOT add it to the list,
highlight a list item, edit that list item, deselect
all list items, and have the un-added session
information reappear.

Use updateRestore() to restore this preserved data at
a later time.

Incomplete.

(3.0)
*/
static void
preserveNewSessionData		(void)
{
	unless (gPreservedNewSession.numberOfSelections++) // only actually save data the first time; otherwise, count requests
	{
		ControlRef		control = null;
		
		
		// restore text fields
		GetDialogItemControlText(gNewSessionsDialog, iNewSessionsHostNameField, gPreservedNewSession.hostNameFieldText);
		GetDialogItemControlText(gNewSessionsDialog, iNewSessionsPortNumberField, gPreservedNewSession.portNumberFieldText);
		GetDialogItemControlText(gNewSessionsDialog, iNewSessionsWindowTitleField, gPreservedNewSession.windowTitleFieldText);
		
		// restore checkboxes
		gPreservedNewSession.proxyCheckBoxValue = GetDialogItemValue(gNewSessionsDialog, iNewSessionsUseProxyServerCheckBox);
		
		// incomplete
	}
}// preserveNewSessionData


/*!
To destroy the data currently selected in the
list of connections to open, use this method.

If the number of selected items is large, a
warning message appears to the user that allows
he or she to cancel the removal.

Incomplete.

(3.0)
*/
static void
removeDataFromList	(void)
{
	ControlRef		control = null;
	ListHandle		list = null;
	UInt16			listItemCount = 0,
					listSelectedItemCount = 0;
	Boolean			proceed = true,
					warn = false;
	
	
	getConnectionsListHandle(&list);
	if (list != null)
	{
		listSelectedItemCount = ListUtilities_GetSelectedItemCount(list);
		
		// the user can bypass the warning if the number of selected items is less than some arbitrary value
		if (!EventLoop_IsOptionKeyDown()) warn = (listSelectedItemCount > 1);
		else warn = false;
		
		if (warn)
		{			
			ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
			InterfaceLibAlertRef	box = null;
			Str255					dialogText,
									helpText,
									countText;
			
			
			GetIndString(dialogText, rStringsCautionAlerts, siConfirmRemoveManyListItems);
			NumToString(listSelectedItemCount, countText);
			StringUtilities_PMetaSetDefaults();
			StringUtilities_PBuild(dialogText, countText, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
			GetIndString(helpText, rStringsCautionAlerts, siProcedureNotUndoableHelpText);
			box = Alert_New();
			Alert_SetFilterProc(box, upp);
			Alert_SetHelpButton(box, false);
			Alert_SetParamsFor(box, kAlert_StyleOKCancel);
			Alert_SetText(box, dialogText, helpText);
			Alert_SetType(box, kAlertCautionAlert);
			Alert_Display(box);
			proceed = (Alert_ItemHit(box) == kAlertStdAlertOKButton);
			Alert_Dispose(&box);
			DisposeModalFilterUPP(upp), upp = null;
		}
		
		if (proceed)
		{
			Cell			cellToDelete;
			ChainLinkRef	deletedNode = null;
			
			
			SetPt(&cellToDelete, 0, 0);
			LSetDrawingMode(false, list);
			while (LGetSelect(true/* next */, &cellToDelete, list))
			{
				LDelRow(1, cellToDelete.v, list);
				deletedNode = Chain_DeleteIndexedNode(gConnectionsToOpen, cellToDelete.v); // zero-based N
				deleteNodeListItemDataPtr(Chain_DisposeNode(gConnectionsToOpen, &deletedNode));
			}
			LSetDrawingMode(true, list);
			DrawOneDialogItem(gNewSessionsDialog, iNewSessionsConnectionsToOpenList);
		}
	}
}// removeDataFromList


/*!
Use this method to find the host name for
a session name.  If that is not possible,
false is returned.

(3.0)
*/
static Boolean
sessionNameToHostName	(StringPtr		inoutString)
{
	SessionPrefs**	sessionPrefsHandle = null;
	OSErr			error = noErr;
	short			oldResFile = CurResFile();
	Boolean			result = true;
	
	
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	sessionPrefsHandle = (SessionPrefs**)Get1NamedResource(SESSIONPREFS_RESTYPE, inoutString);
	error = ResError();
	if (error == noErr)
	{
		HLock((Handle)sessionPrefsHandle);
		
		BlockMove((**sessionPrefsHandle).hostname, inoutString, StrLength((**sessionPrefsHandle).hostname) + 1);
		
		ReleaseResource((Handle)sessionPrefsHandle), sessionPrefsHandle = null;
	}
	else result = false;
	
	UseResFile(oldResFile);
	return result;
}// sessionNameToHostName


/*!
Specifies whether or not the extra content controls are hidden.
If "inIsCollapsed" is "true", the dialog box is vertically shrunk
and the controls in the disclosure pane are hidden.  Otherwise,
the dialog box is vertically expanded and the controls are
displayed.

(3.0)
*/
static void
setDisclosureState	(Boolean const		inIsCollapsed)
{
	SInt32			deltaSizeX = 0,
					deltaSizeY = 0;
	
	
	// determine how much the height of the dialog will change
	{
		ControlRef		disclosurePane = null;
		Rect			paneBounds;
		
		
		(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsDisclosedContentPane, &disclosurePane);
		GetControlBounds(disclosurePane, &paneBounds);
		deltaSizeY = paneBounds.bottom - paneBounds.top;
		if (inIsCollapsed) deltaSizeY = -deltaSizeY;
	}
	
	// when collapsing, hide the controls first, then resize the dialog
	if (inIsCollapsed) HideDialogItem(gNewSessionsDialog, iNewSessionsDisclosedContentPane);
	
	// change the text describing the collapsed state
	{
		Str255		string;
		
		
		GetIndString(string, rStringsMiscellaneous,
						(inIsCollapsed) ? siNewSessionsShowDetails : siNewSessionsHideDetails);
		SetDialogItemControlText(gNewSessionsDialog, iNewSessionsDisclosureTriangleLabel, string);
	}
	
	// change the dialog size, and update the minimum size so window resizing works properly
	{
		WindowInfoRef	windowInfoRef = null;
		Rect			rect;
		Rect*			limitsPtr = null;
		
		
		MacLook_GetWindowRegionBounds(GetDialogWindow(gNewSessionsDialog), kWindowContentRgn, &rect);
		SizeWindow(GetDialogWindow(gNewSessionsDialog),
					rect.right - rect.left + deltaSizeX, rect.bottom - rect.top + deltaSizeY, false/* update */);
		windowInfoRef = WindowInfo_GetFromDialog(gNewSessionsDialog);
		limitsPtr = WindowInfo_GetWindowResizeLimits(windowInfoRef);
		if (limitsPtr != null)
		{
			WindowInfo_SetWindowResizeLimits(windowInfoRef, limitsPtr->top + deltaSizeY, limitsPtr->left,
												limitsPtr->bottom, limitsPtr->right);
		}
	}
	
	// now move or resize the appropriate controls to accommodate the change
	{
		DialogAdjust_BeginDialogItemAdjustment(gNewSessionsDialog);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsConnectButton, deltaSizeY);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsCancelButton, deltaSizeY);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsAppleGuideButton, deltaSizeY);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsBottomSeparator, deltaSizeY);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsConnectionsToOpenList, deltaSizeY);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveV, iNewSessionsRemoveButton, deltaSizeY);
		DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeV, iNewSessionsSessionInfoPrimaryGroupBox, deltaSizeY);
		DialogAdjust_EndAdjustment(deltaSizeX, deltaSizeY); // moves and resizes controls properly
	}
	
	// Some (WEIRD!) problem that occurs ONLY on CERTAIN Macs causes disclosure to
	// be handled incorrectly; specifically, the disclosed content pane is incorrectly
	// delta-sized to become too large, so the content is shifted downwards and the
	// BUTTONS are invisible!  As a temporary hack for people unfortunate enough to
	// encounter this bug, call an explicitly button arrangement (as opposed to simply
	// moving the buttons by some delta amount), to make sure the buttons show up in
	// the proper place!
	(UInt16)Localization_AdjustDialogButtons(gNewSessionsDialog, true/* adjust default button */,
												true/* adjust Cancel button */);
	Localization_AdjustHelpButtonItem(gNewSessionsDialog, iNewSessionsAppleGuideButton);
	
	// when uncollapsing, resize the dialog first, then show the controls
	unless (inIsCollapsed)
	{
		ShowDialogItem(gNewSessionsDialog, iNewSessionsDisclosedContentPane);
		DrawOneDialogItem(gNewSessionsDialog, iNewSessionsDisclosedContentPane);
	}
}// setDisclosureState


/*!
To change the title of the primary group box
on the left-hand side of the dialog to reflect
whether its information is for a new session
or the session selected in the list, use this
method.

(3.0)
*/
static void
setNewSessionGroupBoxName		(Boolean const		inAreAnyListItemsSelected)
{
	Str255			nameString;
	ControlRef		groupBox = null;
	
	
	GetIndString(nameString, rStringsMiscellaneous,
				(inAreAnyListItemsSelected) ? siNewSessionsGroupBoxNameSelectedSession
											: siNewSessionsGroupBoxNameNewSessions);
	if (GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsSessionInfoPrimaryGroupBox, &groupBox) == noErr)
	{
		//SetControlTitle(groupBox, nameString);
		(OSErr)Embedding_OffscreenControlOperation(GetControlOwner(groupBox), groupBox, changeGroupBoxNameOperation,
													(SInt32)nameString, 0L);
	}
}// setNewSessionGroupBoxName


/*!
Activates or deactivates the Remove button
based on whether any items in the connections
list are selected.  Since this routine already
examines the selected items in the list, it
automatically performs additional tasks such
as enabling or disabling other controls, and
updating the Selected Session group box to
contain data from the selected item in the
list.

If the Remove button can be activated, it is
made the default button as well.  If the
Remove button is deactivated, the button that
was previously the default button is made
default again.  If the Remove button is made
active because more than one item in the list
is selected, the Selected Session group box
is deactivated.

(3.0)
*/
static void
setRemoveButtonState	(void)
{
	ListHandle					list = null;
	static DialogItemIndex		oldDefaultButton = iNewSessionsConnectButton;
	
	
	getConnectionsListHandle(&list);
	if (list != null)
	{
		UInt16				numberOfItemsSelected = ListUtilities_GetSelectedItemCount(list);
		
		
		if ((numberOfItemsSelected > 0) && (gHaveDeactivatedRemoveButton))
		{
			DialogItemIndex		defaultButton = GetDialogDefaultItem(gNewSessionsDialog);
			
			
			setNewSessionGroupBoxName(true/* there are items selected */);
			ActivateDialogItem(gNewSessionsDialog, iNewSessionsRemoveButton);
			if (defaultButton != iNewSessionsRemoveButton) oldDefaultButton = defaultButton;
			SetDialogDefaultItem(gNewSessionsDialog, iNewSessionsRemoveButton);
			DeactivateDialogItem(gNewSessionsDialog, iNewSessionsAddButton);
			gHaveDeactivatedRemoveButton = false;
		}
		else if ((numberOfItemsSelected <= 0) && (!gHaveDeactivatedRemoveButton))
		{
			setNewSessionGroupBoxName(false/* there are items selected */);
			DeactivateDialogItem(gNewSessionsDialog, iNewSessionsRemoveButton);
			SetDialogDefaultItem(gNewSessionsDialog, oldDefaultButton);
			ActivateDialogItem(gNewSessionsDialog, iNewSessionsAddButton);
			gHaveDeactivatedRemoveButton = true;
		}
		
		if (numberOfItemsSelected > 1)
		{
			unless (gHaveDeactivatedGroupBox)
			{
				updateClear();
				DeactivateDialogItem(gNewSessionsDialog, iNewSessionsSessionInfoPrimaryGroupBox);
				gHaveDeactivatedGroupBox = true; // use of this flag prevents flickering
			}
		}
		else
		{
			// update the New Session area to reflect the data of the ONE selected item
			ListItemDataPtr		listDataPtr = null;
			Cell				cell;
			
			
			preserveNewSessionData(); // preserve data, if it has not been preserved already
			SetPt(&cell, 0, 0);
			if (numberOfItemsSelected == 1)
			{
				if (LGetSelect(true/* next */, &cell, list))
				{
					ChainLinkRef		node = null;
					
					
					node = Chain_GetIndexedNode(gConnectionsToOpen, cell.v); // zero-based N
					if (node != null)
					{
						listDataPtr = getNodeListItemDataPtr(node);
						if (listDataPtr != null) updateUsingConnectionSetup(listDataPtr->tcHandle);
						else {}
					}
				}
			}
			else updateRestore();
			if (gHaveDeactivatedGroupBox)
			{
				ActivateDialogItem(gNewSessionsDialog, iNewSessionsSessionInfoPrimaryGroupBox);
				gHaveDeactivatedGroupBox = false;
			}
		}
	}
}// setRemoveButtonState


/*!
Adjusts the enabled and checked states of other
checkboxes and pop-up menu items based on the
state of checkboxes.

(3.0)
*/
static void
setSessionCheckBoxStates	(void)
{
	if (GetDialogItemValue(gNewSessionsDialog, iNewSessionsUseProxyServerCheckBox) !=
		kControlCheckBoxCheckedValue)
	{
		// proxy server choices are not applicable
		DeactivateDialogItem(gNewSessionsDialog, iNewSessionsProxiesPopUpMenu);
	}
	else ActivateDialogItem(gNewSessionsDialog, iNewSessionsProxiesPopUpMenu);
}// setSessionCheckBoxStates


/*!
To read in the available proxy server configuration
resources and update the pop-up menu, use this
method.

(3.0)
*/
static void
setUpProxiesPopUpMenu	(Boolean const	inRebuildMenu)
{
	UInt16			numberOfProxies = 0;
	ControlRef		proxiesPopUpMenu = null,
					arrows = null;
	SInt16			oldResFile = CurResFile();
	
	
	(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsProxiesPopUpMenu, &proxiesPopUpMenu);
	(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsProxyChoiceLittleArrows, &arrows);
	
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	numberOfProxies = Count1Resources(PROXYPREFS_RESTYPE);
	if (inRebuildMenu)
	{
		if (gProxiesPopUpMenuHandle != null) DisposeMenu(gProxiesPopUpMenuHandle), gProxiesPopUpMenuHandle = null;
		gProxiesPopUpMenuHandle = Menus_NewMenuThemeSavvy(kMenuIDNewSessionsProxies, EMPTY_PSTRING);
		InsertResMenu(gProxiesPopUpMenuHandle, PROXYPREFS_RESTYPE, 999/* insert after item */);
		(OSErr)SetControlData(proxiesPopUpMenu, kControlMenuPart, kControlPopupButtonMenuHandleTag,
								sizeof(gProxiesPopUpMenuHandle), (Ptr)&gProxiesPopUpMenuHandle);
	}
	
	// set up control states appropriately
	if (numberOfProxies <= 1)
	{
		DeactivateDialogItem(gNewSessionsDialog, iNewSessionsProxyChoiceLittleArrows);
		SetControlMaximum(arrows, 1);
		SetControlMaximum(proxiesPopUpMenu, 1);
	}
	else
	{
		ActivateDialogItem(gNewSessionsDialog, iNewSessionsProxyChoiceLittleArrows);
		SetControlMaximum(arrows, numberOfProxies);
		SetControlMaximum(proxiesPopUpMenu, numberOfProxies);
	}
	SetControlMinimum(arrows, 1);
	SetControlMinimum(proxiesPopUpMenu, 1);
	SetControlValue(arrows, 1);
	SetControlValue(proxiesPopUpMenu, 1);
	
	UseResFile(oldResFile);
}// setUpProxiesPopUpMenu


/*!
To read in the available terminal configuration
resources and update the pop-up menu, use this
method.

(3.0)
*/
static void
setUpTerminalsPopUpMenu	(Boolean const	inRebuildMenu)
{
	UInt16			numberOfTerminals = 0;
	ControlRef		terminalsPopUpMenu = null,
					arrows = null;
	SInt16			oldResFile = CurResFile();
	
	
	(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsTerminalsPopUpMenu, &terminalsPopUpMenu);
	(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsTerminalChoiceLittleArrows, &arrows);
	
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	numberOfTerminals = Count1Resources(TERMINALPREFS_RESTYPE);
	if (inRebuildMenu)
	{
		if (gTerminalsPopUpMenuHandle != null) DisposeMenu(gTerminalsPopUpMenuHandle), gTerminalsPopUpMenuHandle = null;
		gTerminalsPopUpMenuHandle = Menus_NewMenuThemeSavvy(kMenuIDSessionEditorTerminals, EMPTY_PSTRING);
		InsertResMenu(gTerminalsPopUpMenuHandle, TERMINALPREFS_RESTYPE, 999/* insert after item */);
		(OSErr)SetControlData(terminalsPopUpMenu, kControlMenuPart, kControlPopupButtonMenuHandleTag,
								sizeof(gTerminalsPopUpMenuHandle), (Ptr)&gTerminalsPopUpMenuHandle);
	}
	
	// set up control states appropriately
	if (numberOfTerminals <= 1)
	{
		DeactivateDialogItem(gNewSessionsDialog, iNewSessionsTerminalChoiceLittleArrows);
		SetControlMaximum(arrows, 1);
		SetControlMaximum(terminalsPopUpMenu, 1);
	}
	else
	{
		ActivateDialogItem(gNewSessionsDialog, iNewSessionsTerminalChoiceLittleArrows);
		SetControlMaximum(arrows, numberOfTerminals);
		SetControlMaximum(terminalsPopUpMenu, numberOfTerminals);
	}
	SetControlMinimum(arrows, 1);
	SetControlMinimum(terminalsPopUpMenu, 1);
	SetControlValue(arrows, 1);
	SetControlValue(terminalsPopUpMenu, 1);
	
	UseResFile(oldResFile);
}// setUpTerminalsPopUpMenu


/*!
To respond to clicks in the disclosure triangle, use
this method.  The dialog box is vertically resized
to match the CURRENT triangle value, and the controls
in the disclosure pane are hidden or displayed
appropriately.

(3.0)
*/
static void
toggleDisclosureState	(void)
{
	ControlRef		triangle = null;
	SInt32			deltaSizeX = 0,
					deltaSizeY = 0;
	Boolean			isCollapsing = false;
	
	
	(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsDisclosureTriangle, &triangle);
	isCollapsing = (GetControlValue(triangle) == kControlCheckBoxUncheckedValue);
	
	setDisclosureState(isCollapsing);
}// toggleDisclosureState


/*!
Use this method to empty all of the controls in the
New Session primary group box.

Incomplete.

(3.0)
*/
static void
updateClear		(void)
{
	ControlRef		control = null;
	
	
	// clear text fields
	SetDialogItemControlText(gNewSessionsDialog, iNewSessionsHostNameField, EMPTY_PSTRING);
	DrawOneDialogItem(gNewSessionsDialog, iNewSessionsHostNameField);
	SetDialogItemControlText(gNewSessionsDialog, iNewSessionsPortNumberField, EMPTY_PSTRING);
	DrawOneDialogItem(gNewSessionsDialog, iNewSessionsPortNumberField);
	SetDialogItemControlText(gNewSessionsDialog, iNewSessionsWindowTitleField, EMPTY_PSTRING);
	DrawOneDialogItem(gNewSessionsDialog, iNewSessionsWindowTitleField);
	
	// clear checkboxes
	SetDialogItemValue(gNewSessionsDialog, iNewSessionsUseProxyServerCheckBox, kControlCheckBoxUncheckedValue);
	
	// incomplete
	
	// re-focus host name field
	(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsHostNameField, &control);
	(OSErr)SetKeyboardFocus(GetDialogWindow(gNewSessionsDialog), control, kControlEditTextPart);
}// updateClear


/*!
Use this method to restore all of the controls in the
New Session primary group box to contain the global
session variables previously defined using the
preserveNewSessionData() method.  This allows the user
to type in information for a new session, NOT add it
to the list, highlight a list item, edit that list
item, deselect all list items, and have the un-added
session information reappear.  I know, thats a very
complicated series of actions, but trust me, it makes
the dialog box feel very intelligent, because no
data is every really lost.

Incomplete.

(3.0)
*/
static void
updateRestore		(void)
{
	unless (--gPreservedNewSession.numberOfSelections) // until this counter is zero, the original data is not being restored
	{
		ControlRef		control = null;
		
		
		// restore text fields
		SetDialogItemControlText(gNewSessionsDialog, iNewSessionsHostNameField, gPreservedNewSession.hostNameFieldText);
		DrawOneDialogItem(gNewSessionsDialog, iNewSessionsHostNameField);
		SetDialogItemControlText(gNewSessionsDialog, iNewSessionsPortNumberField, gPreservedNewSession.portNumberFieldText);
		DrawOneDialogItem(gNewSessionsDialog, iNewSessionsPortNumberField);
		SetDialogItemControlText(gNewSessionsDialog, iNewSessionsWindowTitleField, gPreservedNewSession.windowTitleFieldText);
		DrawOneDialogItem(gNewSessionsDialog, iNewSessionsWindowTitleField);
		
		// restore checkboxes
		SetDialogItemValue(gNewSessionsDialog, iNewSessionsUseProxyServerCheckBox, gPreservedNewSession.proxyCheckBoxValue);
		
		// incomplete
		
		// re-focus host name field
		(OSErr)GetDialogItemAsControl(gNewSessionsDialog, iNewSessionsHostNameField, &control);
		(OSErr)SetKeyboardFocus(GetDialogWindow(gNewSessionsDialog), control, kControlEditTextPart);
	}
}// updateRestore


/*!
Use this method to set the values of all of the
New Session primary group box controls using
the standard connection-opening data structure.

Incomplete.

(3.0)
*/
static void
updateUsingConnectionSetup		(ConnectionSetup**		inData)
{
	SessionPrefs**		sessionHandle = null;
	TerminalPrefs**		terminalHandle = null;
	
	
	if (inData != null)
	{
		sessionHandle = (*inData)->session;
		terminalHandle = (*inData)->terminal;
		if ((sessionHandle != null) && (terminalHandle != null))
		{
			SInt8		sessionHandleState = HGetState((Handle)sessionHandle),
						terminalHandleState = HGetState((Handle)terminalHandle);
			
			
			HLock((Handle)sessionHandle);
			HLock((Handle)terminalHandle);
			
			// set the text in the fields for the host name and port number
			{
				Str255		fieldString;
				Str15		portString;
				SInt32		port = (**sessionHandle).port;
				
				
				PLstrcpy(fieldString, (**sessionHandle).hostname);
				SetDialogItemControlText(gNewSessionsDialog, iNewSessionsHostNameField, fieldString);
				DrawOneDialogItem(gNewSessionsDialog, iNewSessionsHostNameField);
				NumToString(port, portString);
				SetDialogItemControlText(gNewSessionsDialog, iNewSessionsPortNumberField, portString);
				DrawOneDialogItem(gNewSessionsDialog, iNewSessionsPortNumberField);
			}
			
			// set the window title text field
			SetDialogItemControlText(gNewSessionsDialog, iNewSessionsWindowTitleField, (*inData)->windowName);
			DrawOneDialogItem(gNewSessionsDialog, iNewSessionsWindowTitleField);
			
			// update Proxies pop-up menu (unimplemented)
			
			// update Terminal pop-up menu (test later)
			SetDialogItemValue(gNewSessionsDialog, iNewSessionsTerminalsPopUpMenu,
								Menus_GetItemIndexByText(gTerminalsPopUpMenuHandle, (**sessionHandle).TerminalEmulation));
			
			HSetState((Handle)sessionHandle, sessionHandleState);
			HSetState((Handle)terminalHandle, terminalHandleState);
		}
		else
		{
			Console_WriteLine("cannot update, connection setup terminal or session is null");
		}
	}
	else
	{
		Console_WriteLine("cannot update, connection setup is null");
	}
}// updateUsingConnectionSetup


/*!
Use this method to acquire a named resource
containing session preferences, and use its
data to set the controls in this dialog.

(3.0)
*/
static void
updateUsingSessionResource	(StringPtr		inSessionName)
{
	SessionPrefs**		sessionHandle = null;
	SInt16				oldResFile = CurResFile();
	
	
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	sessionHandle = (SessionPrefs**)Get1NamedResource(SESSIONPREFS_RESTYPE, inSessionName);
	if ((sessionHandle != null) && (ResError() == noErr))
	{
		HLock((Handle)sessionHandle);
		
		{
			Str255			string;
			ControlRef		control = null;
			SInt32			port = (**sessionHandle).port;
			
			
			PLstrcpy(string, inSessionName);
			(Boolean)sessionNameToHostName(string);
			SetDialogItemControlText(gNewSessionsDialog, iNewSessionsHostNameField, string);
			DrawOneDialogItem(gNewSessionsDialog, iNewSessionsHostNameField);
			NumToString(port, string);
			SetDialogItemControlText(gNewSessionsDialog, iNewSessionsPortNumberField, string);
			DrawOneDialogItem(gNewSessionsDialog, iNewSessionsPortNumberField);
			SetDialogItemControlText(gNewSessionsDialog, iNewSessionsWindowTitleField, EMPTY_PSTRING);
			DrawOneDialogItem(gNewSessionsDialog, iNewSessionsWindowTitleField);
		}
		
		// select sessions default proxy server - unimplemented
		
		// select sessions default terminal
		SetDialogItemValue(gNewSessionsDialog, iNewSessionsTerminalsPopUpMenu,
							Menus_GetItemIndexByText(gTerminalsPopUpMenuHandle, (**sessionHandle).TerminalEmulation));
		
		Memory_DisposeHandle(&((Handle)sessionHandle));
		
		makeAllControlsUpToDate();
	}
	else
	{
		Console_WriteValuePString("unexpectedly could not update for session", inSessionName);
	}
	
	UseResFile(oldResFile);
}// updateUsingSessionResource


/*!
If an editable text control is to represent a
number, use this method to assign a key filter
to it that prevents non-digits from being
entered into such a field.

(3.0)
*/
static void
useNumericalLimiter		(ControlRef		inControl)
{
	ControlKeyFilterUPP		upp = NewControlKeyFilterUPP(NumericalLimiter);
	
	
	(OSErr)SetControlData(inControl, kControlNoPart, kControlKeyFilterTag, sizeof(ControlKeyFilterUPP), (Ptr)&upp);
}// useNumericalLimiter