/*###############################################################

	Panel.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Dialogs.h>
#include <PLStringFuncs.h>
#include <Resources.h>

// library includes
#include "ContextsLib.h"
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "Panel.h"



// types
struct Panel
{
	struct
	{
		void*					auxiliaryDataPtr;	// arbitrary data
		PanelChangedProcPtr		changedProc;		// called when anything changes in the panel
		ControlRef				container;			// the super-control of every panel control
		PanelDescriptor			descriptor;			// an identifier that helps distinguish panels
		Point					minimumSize;		// smallest possible panel dimensions
		
		IconManagerIconRef		icon;
		Str255					name;				// a label for this panel
	} customizerWriteable;
	
	struct
	{
		WindowRef				owningWindow;	// the window associated with panel controls
	} utilizerWriteable;
};
typedef struct Panel	Panel;
typedef Panel*			PanelPtr;

// internal methods
static PanelPtr		refAcquireLock		(PanelRef		inRef);
static void			refReleaseLock		(PanelRef		inRef,
										 PanelPtr*		inoutPtr);



//
// public methods
//

/*!
To create a new Dialog Panel reference,
use this method.  If any problems occur,
null is returned.

Your dialog panel cannot function without
a change notification procedure.  This
routine is the entry point that dialog
boxes use (indirectly, by invoking one or
more methods from the Dialog Panel module)
to communicate with panels in an abstract
way.

(3.0)
*/
PanelRef
Panel_New		(PanelChangedProcPtr		inProc)
{
	PanelRef		result = (PanelRef)Memory_NewPtr(sizeof(Panel));
	
	
	if ((result != null) && (inProc != null))
	{
		PanelPtr		ptr = refAcquireLock(result);
		
		
		ptr->customizerWriteable.changedProc = inProc;
		ptr->customizerWriteable.container = null;
		ptr->customizerWriteable.descriptor = kInvalidPanelDescriptor;
		ptr->customizerWriteable.icon = IconManager_NewIcon();
		PLstrcpy(ptr->customizerWriteable.name, EMPTY_PSTRING);
		refReleaseLock(result, &ptr);
	}
	return result;
}// New


/*!
To destroy a Panel previously created
with Panel_New(), use this method.
Your copy of the reference is
automatically set to null.

(3.0)
*/
void
Panel_Dispose		(PanelRef*		inoutRefPtr)
{
	if (inoutRefPtr != null)
	{
		PanelPtr		ptr = refAcquireLock(*inoutRefPtr);
		
		
		Panel_Changed(*inoutRefPtr, kPanel_MessageDestroyed, Panel_GetAuxiliaryDataPtr(*inoutRefPtr));
		IconManager_DisposeIcon(&ptr->customizerWriteable.icon);
		Memory_DisposePtr((Ptr*)inoutRefPtr);
	}
}// Dispose


//
// methods for dialog access to panels
//

/*!
When a dialog box performs some kind of operation on a
panel, such as changing its size or visible state, it
should always invoke this routine to notify the panel
that a change has occurred.  Some messages generate
defined result codes, and others do not.  See "Panel.h"
to see which messages have results.

You specify the routine that will get this notification
when you create a panel with Panel_New().

The possible messages are as follows:

kPanel_MessageClicked
	This property indicates that the mouse has been
	clicked somewhere in the panels container control.
	The panel responds by using the event information
	given to figure out what the user is doing, and
	react accordingly (tracking a control click, for
	instance).  This message has a defined result code
	(see "Panel.h" for the two possible values).

kPanel_MessageCreateControls
	Dialogs must fire this property change exactly once,
	immediately after the dialog window is created.
	This allows panels to create their controls in the
	right window.

kPanel_MessageDeltaSize
	This property indicates that the dialog has somehow
	altered itself such that the panel now occupies a
	new set of boundaries.  The dialog should *first*
	move and resize the containing control of the panel,
	then notify it that it has changed.

kPanel_MessageDestroyed
	This property indicates that the panel is just about
	to be disposed of.  Your routine should dispose of
	any private data structures that are only used by
	your panel.  This message is issued only when the
	Panel_Dispose() method is called.

kPanel_MessageFocusGained
	This property indicates that the user has shifted
	focus to another control.  The given control is the
	one that has gained focus, and therefore is now
	taking keyboard input.  If the specified control is
	not one you recognize, do not do anything with it!

kPanel_MessageFocusLost
	This property indicates that the user has shifted
	focus to another control.  The given control is the
	one that has lost focus, and therefore is not taking
	keyboard input anymore.  If the specified control is
	not one you recognize, do not do anything with it!

kPanel_MessageInitialSize
	Dialogs must fire this property change exactly once,
	immediately after the dialog window has set the
	initial size for a panel, which should be after it
	fires the owning window property change.  This
	allows panels to arrange their new controls correctly.
	Note that subsequent events (namely the dimensions
	property change) only specify how much the dimensions
	have been changed by, so it is very important to allow
	panels to set up their controls correctly initially.

kPanel_MessageNewActiveState
	This property indicates that the dialog has activated
	or deactivated the panels containing control.  Pass a
	pointer to a Boolean variable that indicates the new
	active state.

kPanel_MessageNewAppearanceTheme
	This property indicates that the application has
	received a theme switch event, and therefore should
	recalculate the size of controls.  The panel generally
	responds by re-setting its minimum size, which the
	dialog can subsequently check.

kPanel_MessageNewVisibility
	This property indicates that the dialog has hidden or
	shown a panel.  For some panels, such as ones that
	provide audible feedback, it may be important to know
	when the panel is no longer showing.  The dialog is
	free to fire this message when the panel may not be
	really invisible, but is obscured in some way (such
	as by another, larger control in the same window).

(3.0)
*/
SInt32
Panel_Changed		(PanelRef				inRef,
					 const PanelMessage		inMessage,
					 void*					inData)
{
	SInt32		result = 0L;
	
	
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			// automatically save this property, for convenience
			if (inMessage == kPanel_MessageCreateControls)
			{
				ptr->utilizerWriteable.owningWindow = *((WindowRef*)inData);
			}
			
			if (ptr->customizerWriteable.changedProc != null)
			{
				// notify the panel of the change
				result = InvokePanelChangedProc(ptr->customizerWriteable.changedProc,
												inRef, inMessage, inData);
			}
			refReleaseLock(inRef, &ptr);
		}
	}
	
	return result;
}// Changed


/*!
When a dialog box needs to manipulate a panel,
its code should invoke this routine.  The code
responsible for customizing the specified panel
should have already used the method
Panel_SetContainerControl() to set the
super-control (which is most likely a user
pane) that embeds every other control in the
panel.

The dialog box can then move the control
provided by this routine, instantly relocating
every control in the panel, or perform other
operations, such as changing the visibility or
active state of an entire panel.

See also Panel_ChangedSize(), which is a
useful follow-up call when you resize container
controls of panels.

(3.0)
*/
void
Panel_GetContainerControl		(PanelRef			inRef,
								 ControlRef*		outControl)
{
	if ((outControl != null) && (inRef != null))
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			*outControl = ptr->customizerWriteable.container;
			refReleaseLock(inRef, &ptr);
		}
	}
}// GetContainerControl


/*!
To determine the minimum size that a
panel prefers, in pixels, use this
method.  Dialogs always have the final
say as to the space a panel occupies
(and the panel should follow suit by
arranging its controls to fit), but it
is generally desirable to try to give
a panel at least its minimum space.  A
good rule of thumb is to create a
dialog flexible enough to simply make
its window large enough to fit a panel
in its minimum size, whenever that
panel is showing.

(3.0)
*/
void
Panel_GetMinimumSize	(PanelRef		inRef,
						 Point*			outSize)
{
	if ((outSize != null) && (inRef != null))
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			*outSize = ptr->customizerWriteable.minimumSize;
			refReleaseLock(inRef, &ptr);
		}
	}
}// GetMinimumSize


/*!
To acquire a label string that can be
used to describe a panel, use this
method.  Generally, even if panel
implementations do not use labels
themselves, they should specify a
label that dialogs can use as desired
to represent a set of panels.  For
example, labels may be used for text
in buttons, menus or tabs that select
amongst active panels.  Since you
should not assume how your panel will
be displayed in a user interface,
provide a label whether you need it or
not.

See also Panel_GetIconSuite(), which
lets you provide a set of icons for
your panel that is often paired with
a label in dialog boxes.

IMPORTANT:	The size of the storage
			space is a hint; names are
			short!

(3.0)
*/
void
Panel_GetName		(PanelRef		inRef,
					 Str31			outName)
{
	if ((outName != null) && (inRef != null))
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			PLstrcpy(outName, ptr->customizerWriteable.name);
			refReleaseLock(inRef, &ptr);
		}
	}
}// GetName


/*!
To conveniently set up any Appearance control that
can take standard content types such as icons,
pictures, etc. so it uses the icon defined for this
panel (if any), use this method.  You can use this,
for example, to set the icon for a bevel button.

(3.0)
*/
OSErr
Panel_SetButtonIcon		(ControlRef		inControl,
						 PanelRef		inRef)
{
	OSErr		result = noErr;
	
	
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			result = IconManager_SetButtonIcon(inControl, ptr->customizerWriteable.icon);
			refReleaseLock(inRef, &ptr);
		}
	}
	return result;
}// SetButtonIcon


//
// panel-customizing methods
//

/*!
To acquire a pointer to auxiliary data you
previously set with Panel_SetAuxiliaryDataPtr(),
use this method.

Auxiliary data is reserved for use by panel customizing
code.  You generally use the panel descriptor to help
identify what the auxiliary data may be.  See also
Panel_GetDescriptor().

(3.0)
*/
void*
Panel_GetAuxiliaryDataPtr		(PanelRef			inRef)
{
	void*		result = null;
	
	
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			result = ptr->customizerWriteable.auxiliaryDataPtr;
			refReleaseLock(inRef, &ptr);
		}
	}
	
	return result;
}// GetAuxiliaryDataPtr


/*!
To help identify a panel, use this method.

(3.0)
*/
PanelDescriptor
Panel_GetDescriptor		(PanelRef			inRef)
{
	PanelDescriptor		result = kInvalidPanelDescriptor;
	
	
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			result = ptr->customizerWriteable.descriptor;
			refReleaseLock(inRef, &ptr);
		}
	}
	
	return result;
}// GetDescriptor


/*!
To determine the window that a panel is in,
use this method.  This property is set once,
automatically, when a dialog box fires the
"kPanel_MessageCreateControls" message.

(3.0)
*/
WindowRef
Panel_GetOwningWindow		(PanelRef			inRef)
{
	WindowRef		result = null;
	
	
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			result = ptr->utilizerWriteable.owningWindow;
			refReleaseLock(inRef, &ptr);
		}
	}
	
	return result;
}// GetOwningWindow


/*!
To associate arbitrary data with a panel,
use this method.  You generally also call
Panel_SetDescriptor(), to help you identify
what this auxiliary data is.  This access
is only allowed by panel customizing code,
not dialogs.

To acquire this data later, use this method
Panel_GetAuxiliaryDataPtr().

(3.0)
*/
void
Panel_SetAuxiliaryDataPtr		(PanelRef		inRef,
								 void*			inAuxiliaryDataPtr)
{
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			ptr->customizerWriteable.auxiliaryDataPtr = inAuxiliaryDataPtr;
			refReleaseLock(inRef, &ptr);
		}
	}
}// SetAuxiliaryDataPtr


/*!
To set the super-control (usually a user
pane) that directly or indirectly embeds
every control in the specified panel,
use this method.

Dialog boxes that use Panel objects
generally require custom panels to have
a container control!  Without one, it is
not possible for a dialog to include the
controls of a panel, much less manipulate
them.  A dialog uses the method
Panel_GetContainerControl() to acquire
the control that is specified here.

(3.0)
*/
void
Panel_SetContainerControl		(PanelRef			inRef,
								 ControlRef			inControl)
{
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			ptr->customizerWriteable.container = inControl;
			refReleaseLock(inRef, &ptr);
		}
	}
}// SetContainerControl


/*!
To change the descriptor of a dialog panel,
use this method.  Your descriptor codes
should be created by adding values to
"kInvalidPanelDescriptor", to avoid
collisions with any reserved descriptors.
Descriptors can be useful in identifying
auxiliary data: see also the method
Panel_SetAuxiliaryDataPtr().

(3.0)
*/
void
Panel_SetDescriptor		(PanelRef					inRef,
						 const PanelDescriptor		inDescriptor)
{
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			ptr->customizerWriteable.descriptor = inDescriptor;
			refReleaseLock(inRef, &ptr);
		}
	}
}// SetDescriptor


/*!
To use an Icon Services icon set to describe
a panel, use this method.  An Icon Services
reference will not take effect unless the
user has at least OS 8.5.  If OS 8.5 or later
is not in use, an attempt is made to find an
equivalent icon suite, which may fail.

See also Panel_SetIconSuite().

(3.0)
*/
void
Panel_SetIconRef	(PanelRef			inRef,
					 const SInt16		inIconServicesVolumeNumber,
					 const OSType		inIconServicesCreator,
					 const OSType		inIconServicesDescription)
{
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			(OSErr)IconManager_MakeIconRef(ptr->customizerWriteable.icon,
											inIconServicesVolumeNumber,
											inIconServicesCreator,
											inIconServicesDescription);
			refReleaseLock(inRef, &ptr);
		}
	}
}// SetIconRef


/*!
To use a traditional icon suite to describe
a panel, use this method.  An Icon Services
reference will automatically take effect if
OS 8.5 or later is in use and an equivalent
icon can be found.  If no equivalent can be
found, the icon suite resource is used.

See also Panel_SetIconRef().

(3.0)
*/
void
Panel_SetIconSuite	(PanelRef			inRef,
					 const SInt16		inSuiteResourceID,
					 IconSelectorValue	inWhichIcons)
{
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			(OSErr)IconManager_MakeIconSuite(ptr->customizerWriteable.icon,
												inSuiteResourceID, inWhichIcons);
			refReleaseLock(inRef, &ptr);
		}
	}
}// SetIconSuite


/*!
To specify the smallest dimensions that your panel
can acquire without cutting off any of its
contents, use this method.  A flexible dialog box
will allow space for your panels minimum size,
but is not required to use it.

IMPORTANT:	You should invoke this routine again
			whenever your change notification
			routine receives the property
			"kPanel_MessageNewAppearanceTheme",
			because the new theme may have larger
			controls that will increase the panels
			minimum size.  Correct dialog box
			implementations will re-check the
			minimum size property as soon as they
			finish firing the appearance theme
			property change to your panel.

(3.0)
*/
void
Panel_SetMinimumSize		(PanelRef			inRef,
							 const Point		inSize)
{
	if (inRef != null)
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			SetPt(&ptr->customizerWriteable.minimumSize, inSize.h, inSize.v);
			refReleaseLock(inRef, &ptr);
		}
	}
}// SetMinimumSize


/*!
To specify a label string that can be
used to describe a panel, use this
method.  Generally, even if panel
implementations do not use labels
themselves, they should specify a
label that dialogs can use as desired
to represent a set of panels.  For
example, labels may be used for text
in buttons, menus or tabs that select
amongst active panels.  Since you
should not assume how your panel will
be displayed in a user interface,
provide a label whether you need it or
not.

See also Panel_SetIconSuite(), which
lets you specify a set of icons for
your panel that is often paired with
a label in dialog boxes.

IMPORTANT:	The size of the storage
			space is a hint; names are
			short!

(3.0)
*/
void
Panel_SetName		(PanelRef			inRef,
					 ConstStr31Param	inName)
{
	if ((inName != null) && (inRef != null))
	{
		PanelPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			PLstrcpy(ptr->customizerWriteable.name, inName);
			refReleaseLock(inRef, &ptr);
		}
	}
}// SetName


//
// internal methods
//

/*!
To acquire a pointer to the internal structure,
given a reference to it, use this method.

(3.0)
*/
static PanelPtr
refAcquireLock		(PanelRef		inRef)
{
	return ((PanelPtr)inRef);
}// refAcquireLock


/*!
To discontinue use of a pointer to the internal
structure, given a reference to it, use this
method.

(3.0)
*/
static void
refReleaseLock		(PanelRef		inRef,
					 PanelPtr*		inoutPtr)
{
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock