/*###############################################################

	Panel.h
	
	An abstract interface that allows panel-based dialog boxes
	to be easily constructed.  Use the methods listed in this
	file to implement a panels interface to its dialog, and
	then it can easily be used in any dialog box that supports
	this same interface to panels.
	
	Generally, your panel customizing code creates a panel as
	an Appearance user pane control, and assigns that control
	to a new Panel object.  This allows any dialog box to
	incorporate a panel, provided that the code for the dialog
	box also implements this interface completely.
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#ifndef __PANEL__
#define __PANEL__

// Mac includes
#include <Appearance.h>
#include <Dialogs.h>
#include <MacTypes.h>

// MacTelnet includes
#include "PreferencesData.h"



// types
typedef struct OpaquePanel { unsigned garbage; }*	PanelRef;

typedef SInt32 PanelDescriptor;
enum
{
	kInvalidPanelDescriptor = -1722
};

typedef SInt16 PanelContainerType;
enum
{
	kPanel_ContainerTypeDialogBox = 1,
	kPanel_ContainerTypeWindow = 2
};

// universal procedure pointers
/**
Panel State Changed Notification Method

This routine type is used to signal a panel when the
dialog box that contains it has changed some aspect of
the panel.  This kind of notification should always be
invoked by dialog box code when the code operates on a
panel.  A flexible dialog box will send every kind of
change notification, even seemingly unimportant ones
(such as changing the active state), so that any kind
of added panel is definitely able to work in the
dialog.  All dialogs must send the one time property
changes in a logical order.
*/
typedef UInt32 PanelMessage;
enum
{
	kPanel_MessageClicked = 'clik',					// data: ControlRef*, the control being hit
		kPanel_ResponseClickHandled = (1U >> 0),		// result code of the above
		kPanel_ResponseClickNotHandled = (0 >> 0),		// result code of the above
	kPanel_MessageCreateControls = '1win',			// data: WindowRef*, the owning window
	kPanel_MessageDeltaSize = 'size',				// data: Point*, CHANGE in dimensions
	kPanel_MessageDestroyed = 'tobe',				// data: void*, the auxiliary data pointer
	kPanel_MessageFocusGained = 'focg',				// data: ControlRef*, the field gaining focus
	kPanel_MessageFocusLost = 'focl',				// data: ControlRef*, the field losing focus
	kPanel_MessageInitialSize = '1siz',				// data: Point*, EXACT dimensions
	kPanel_MessageNewActiveState = 'actv',			// data: Boolean*, is now active?
	kPanel_MessageNewAppearanceTheme = 'athm',		// data: null
	kPanel_MessageNewVisibility = 'visb'			// data: Boolean*, is now visible?
};
typedef SInt32 (*PanelChangedProcPtr)	(PanelRef			inRef,
										 PanelMessage		inMessage,
										 void*				inData);
#define InvokePanelChangedProc(userRoutine, inRef, inMessage, inData) \
			(*userRoutine)((inRef), (inMessage), (inData))



/*###############################################################
	CREATING AND DESTROYING DIALOG PANELS (DONE BY CUSTOMIZERS)
###############################################################*/

PanelRef
	Panel_New					(PanelChangedProcPtr		inProc);

void
	Panel_Dispose				(PanelRef*					inoutRefPtr);

/*###############################################################
	UTILITIES
###############################################################*/

OSErr
	Panel_SetButtonIcon			(ControlRef					inControl,
								 PanelRef					inRef);

/*###############################################################
	METHODS ALLOWING DIALOGS TO MANIPULATE PANELS
###############################################################*/

SInt32
	Panel_Changed				(PanelRef					inRef,
								 PanelMessage const			inMessage,
								 void*						inData);

void
	Panel_GetContainerControl	(PanelRef					inRef,
								 ControlRef*				outControl);

void
	Panel_GetMinimumSize		(PanelRef					inRef,
								 Point*						outMinimumSize);

void
	Panel_GetName				(PanelRef					inRef,
								 Str31						outName);

/*###############################################################
	METHODS ALLOWING PANEL-CUSTOMIZING CODE TO ACCESS PANELS
###############################################################*/

void*
	Panel_GetAuxiliaryDataPtr	(PanelRef					inRef);

PanelDescriptor
	Panel_GetDescriptor			(PanelRef					inRef);

WindowRef
	Panel_GetOwningWindow		(PanelRef					inRef);

void
	Panel_SetAuxiliaryDataPtr	(PanelRef					inRef,
								 void*						inAuxiliaryDataPtr);

void
	Panel_SetContainerControl	(PanelRef					inRef,
								 ControlRef					inControl);

void
	Panel_SetDescriptor			(PanelRef					inRef,
								 PanelDescriptor const		inDescriptor);

void
	Panel_SetIconRef			(PanelRef					inRef,
								 SInt16 const				inIconServicesVolumeNumber,
								 OSType const				inIconServicesCreator,
								 OSType const				inIconServicesDescription);

void
	Panel_SetIconSuite			(PanelRef					inRef,
								 SInt16 const				inSuiteResourceID,
								 IconSelectorValue			inWhichIcons);

void
	Panel_SetMinimumSize		(PanelRef					inRef,
								 Point const				inSize);

void
	Panel_SetName				(PanelRef					inRef,
								 ConstStr31Param			inName);

#endif