/*###############################################################

	Parse.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdio>
#	include <cstring>
#else
#	include <stdio.h>
#	include <string.h>
#endif

// resource includes
#include "MenuResources.h"

// MacTelnet includes
#include "BasicTypesAE.h"
#include "Console.h"
#include "Connections.h"
#include "Parse.h"				// for our #defines
#include "network.proto.h"				// for net functions
#include "LineMode.h"
#include "TektronixVirtualGraphics.h"
#include "vsdata.h"
#include "Terminal.h"
#include "InstallAE.h"
#include "MacLook.h"
#include "RealScreen.h"
#include "TektronixRealGraphics.h"
#include "TelnetMenus.h"
#include "Terminology.h"
#include "TextTranslation.h"
#include "RasterGraphicsKernel.h"



// internal methods
static void		encryptStatechange	(ConnectionDataPtr			inConnectionDataPtr);
static char*	getcname			(ConnectionDataPtr 			inConnectionDataPtr);
static void		process_suboption	(ConnectionDataPtr			inConnectionDataPtr,
									 unsigned char*				inStart,
									 unsigned char*				inEnd);
static void		send_do				(SInt16						inPort,
									 TelnetNegotiableOption		inOption);
static void		send_dont			(SInt16						inPort,
									 TelnetNegotiableOption		inOption);
static void		send_will			(SInt16						inPort,
									 TelnetNegotiableOption		inOption);
static void		send_wont			(SInt16						inPort,
									 TelnetNegotiableOption		inOption);
static void		telnet_do			(ConnectionDataPtr			inConnectionDataPtr,
									 TelnetNegotiableOption		inOption);
static void		telnet_dont			(ConnectionDataPtr			inConnectionDataPtr,
									 TelnetNegotiableOption		inOption);
static void		telnet_will			(ConnectionDataPtr			inConnectionDataPtr,
									 TelnetNegotiableOption		inOption);
static void		telnet_wont			(ConnectionDataPtr			inConnectionDataPtr,
									 TelnetNegotiableOption		inOption);



//
// public methods
//

/*!
To send Negotiate About Window Size information
to a screens connection, use this method.

(2.6)
*/
void
Parse_NegotiateAboutWindowSize	(ConnectionDataPtr		inConnectionDataPtr,
								 SInt16					inWidth,
								 SInt16					inHeight)
{
	if (inConnectionDataPtr != null)
	{
		char			buffer[20];
		unsigned char	height = inHeight & 0xFF,
						width = inWidth & 0xFF;
		
		
		// ugly patch to not send IAC as window size
		if (height == 0xFF) height = 0xFE;
		if (width == 0xFF) width = 0xFE;
		
		// The syntax for a Negotiate About Window Size command is as follows:
		//		IAC SB
		//			NAWS
		//			high byte of the width
		//			low byte of the width
		//			high byte of the height
		//			low byte of the height
		//		IAC SE
		Network_SendFlushData(inConnectionDataPtr->port, "\377\372\037\000", 4);
		sprintf(buffer, "%c\000", width);
		Network_SendData(inConnectionDataPtr->port, buffer, 2);
		sprintf(buffer, "%c\377\360", height);
		Network_SendData(inConnectionDataPtr->port, buffer, 3);
	}
}// NegotiateAboutWindowSize


/*!
When a telnet connection first opens, invoke this
method to send a series of initial protocol options.

(2.6)
*/
void
Parse_SendInitialTelnetOptions		(ConnectionDataPtr		inConnectionDataPtr)
{
	if (inConnectionDataPtr != null)
	{
		send_do(inConnectionDataPtr->port, N_ECHO); // echo
		inConnectionDataPtr->echo = 1;
		
		send_do(inConnectionDataPtr->port, N_SGA); // supress go-ahead
		inConnectionDataPtr->Usga = 1;
		
		send_wont(inConnectionDataPtr->port, N_XDISPLOC);
		
		send_will(inConnectionDataPtr->port, N_NAWS); // negotiate about window size
	}
}// SendInitialTelnetOptions


/*!
To send a do or dont option for local echo
for the specified connection, use this method.

(3.0)
*/
void
Parse_SendLocalEchoOption		(ConnectionDataPtr		inConnectionDataPtr)
{
	if (inConnectionDataPtr != null)
	{
		if (inConnectionDataPtr->echo)
		{
			// local echo
			send_dont(inConnectionDataPtr->port, N_ECHO);
		}
		else
		{
			// remote echo
			send_do(inConnectionDataPtr->port, N_ECHO);
		}
	}
}// SendLocalEchoOption


/*!
Sends Apple Events back to MacTelnet that instruct
it to send the specified data to the frontmost
connection ("connection 1").  If desired, you can
simultaneously execute the send (although it is
usually faster to execute separately, or perhaps
you may not want to execute at all).

(3.0)
*/
void
Parse_SendRecordableDataTransmitEvent	(char*				inStringToType,
										 SInt16				inStringLength,
										 Boolean const		inExecute)
{
	OSErr		error = noErr;
	AEDesc		dataSendEvent,
				replyDescriptor;
	
	
	InstallAE_CreateRecordableAppleEvent(kTelnetEventClass, kTelnetEventIDDataSend, &dataSendEvent);
	if (error == noErr)
	{
		Boolean			sendNewLine = false;
		SInt16			stringLength = inStringLength;
		AEDesc			containerDesc = { typeNull, null },
						keyDesc = { typeNull, null },
						objectDesc = { typeNull, null },
						textDesc = { typeNull, null };
		
		
		// clip newlines and turn them into "with newline" parameters (newlines
		// might be represented by \n, \r, "\012", "\015", or "\015\012"; look
		// for any of these and decrement the length to prevent including them
		// in the recorded string)
		if ((stringLength >= 1) && inStringToType[stringLength - 1] == '\n')
		{
			Console_WriteLine("\\n: removing");
			stringLength--;
			sendNewLine = true;
		}
		if ((stringLength >= 1) && inStringToType[stringLength - 1] == '\r')
		{
			Console_WriteLine("\\r: removing");
			stringLength--;
			sendNewLine = true;
		}
		if ((stringLength >= 4) &&
					(inStringToType[stringLength - 1] == '2') &&
					(inStringToType[stringLength - 2] == '1') &&
					(inStringToType[stringLength - 2] == '0') &&
					(inStringToType[stringLength - 3] == '\\'))
		{
			Console_WriteLine("\\012: removing");
			stringLength -= 4;
			sendNewLine = true;
		}
		if ((stringLength >= 4) &&
					(inStringToType[stringLength - 1] == '5') &&
					(inStringToType[stringLength - 2] == '1') &&
					(inStringToType[stringLength - 2] == '0') &&
					(inStringToType[stringLength - 3] == '\\'))
		{
			Console_WriteLine("\\015: removing");
			stringLength -= 4;
			sendNewLine = true;
		}
		
		// create a descriptor containing the given string
		error = AECreateDesc(cString, inStringToType, stringLength * sizeof(char), &textDesc);
		if (error != noErr) Console_WriteValue("text descriptor creation error", error);
		
		// send a request for "connection <index>", which resides in a null container
		(OSErr)BasicTypesAE_CreateSInt32Desc(1/* frontmost window */, &keyDesc);
		error = CreateObjSpecifier(kTelnetClassConnection, &containerDesc,
									formAbsolutePosition, &keyDesc, true/* dispose inputs */,
									&objectDesc);
		if (error != noErr) Console_WriteValue("connection access error", error);
		
		if (error == noErr)
		{
			// the data to transmit - REQUIRED
			(OSErr)AEPutParamDesc(&dataSendEvent, keyDirectObject, &textDesc);
			
			// reference to the connection where the data will be sent - REQUIRED
			(OSErr)AEPutParamDesc(&dataSendEvent, kTelnetParameterKeyDataSendToConnection, &objectDesc);
			
			if (sendNewLine)
			{
				// add a parameter to send a newline
				(OSErr)AEPutParamPtr(&dataSendEvent, kTelnetParameterKeyDataSendWithNewLine,
										typeBoolean, &sendNewLine, sizeof(sendNewLine));
			}
			
			// send the event, which will record it into any open script;
			// this event is only executed if requested to do so by the caller
			{
				UInt32		flags = kAENoReply | kAECanInteract;
				
				
				unless (inExecute) flags |= kAEDontExecute;
				(OSErr)AESend(&dataSendEvent, &replyDescriptor,
								flags, kAENormalPriority, kAEDefaultTimeout, null, null);
			}
		}
		AEDisposeDesc(&textDesc);
		AEDisposeDesc(&objectDesc);
	}
	AEDisposeDesc(&dataSendEvent);
}// SendRecordableDataTransmitEvent


/*!
Sends Apple Events back to MacTelnet that instruct
it to send the specified special character to the
frontmost connection ("connection 1").  If desired,
you can simultaneously execute the send (although
it is usually faster to execute separately, or
perhaps you may not want to execute at all).

The value of "inSpecialCharacterEnumeration" is an
enumeration (see "Terminology.h") identifying the
special character to send.  This routine directly
sends its value as a parameter to the event, no
error checking is done.

(3.0)
*/
void
Parse_SendRecordableSpecialTransmitEvent	(FourCharCode const		inSpecialCharacterEnumeration,
											 Boolean const			inExecute)
{
	OSErr		error = noErr;
	AEDesc		dataSendEvent,
				replyDescriptor;
	
	
	InstallAE_CreateRecordableAppleEvent(kTelnetEventClass, kTelnetEventIDDataSend, &dataSendEvent);
	if (error == noErr)
	{
		AEDesc			containerDesc = { typeNull, null },
						keyDesc = { typeNull, null },
						objectDesc = { typeNull, null },
						specialCharacterDesc = { typeNull, null };
		
		
		// create a descriptor containing the given enumeration
		error = AECreateDesc(typeEnumerated, &inSpecialCharacterEnumeration,
								sizeof(inSpecialCharacterEnumeration), &specialCharacterDesc);
		if (error != noErr) Console_WriteValue("special character descriptor creation error", error);
		
		// send a request for "connection <index>", which resides in a null container
		(OSErr)BasicTypesAE_CreateSInt32Desc(1/* frontmost window */, &keyDesc);
		error = CreateObjSpecifier(kTelnetClassConnection, &containerDesc,
									formAbsolutePosition, &keyDesc, true/* dispose inputs */,
									&objectDesc);
		if (error != noErr) Console_WriteValue("connection access error", error);
		
		if (error == noErr)
		{
			// the data to transmit - REQUIRED
			(OSErr)AEPutParamDesc(&dataSendEvent, keyDirectObject, &specialCharacterDesc);
			
			// reference to the connection where the data will be sent - REQUIRED
			(OSErr)AEPutParamDesc(&dataSendEvent, kTelnetParameterKeyDataSendToConnection, &objectDesc);
			
			// send the event, which will record it into any open script;
			// this event is only executed if requested to do so by the caller
			{
				UInt32		flags = kAENoReply | kAECanInteract;
				
				
				unless (inExecute) flags |= kAEDontExecute;
				(OSErr)AESend(&dataSendEvent, &replyDescriptor,
								flags, kAENormalPriority, kAEDefaultTimeout, null, null);
			}
		}
		AEDisposeDesc(&specialCharacterDesc);
		AEDisposeDesc(&objectDesc);
	}
	AEDisposeDesc(&dataSendEvent);
}// SendRecordableSpecialTransmitEvent


/*!
Looks at the string which has just come in from outside and
checks for special sequences of interest.  Any routine
string is passed immediately, if possible, waiting only for
special characters ESC and 255 for mode changes.

This routine gets its name from the fact that the given
data will either represent something that shows up on the
terminal screen, or special characters to be interpreted as
a command.

(2.6)
*/
void
Parse_ShowOrInterpretAsCommand	(ConnectionDataPtr		tw,
								 unsigned char*			st,
								 SInt16					cnt)
{
	SInt16			i = 0,
					temptw = 0;
	unsigned char*	mark = st + cnt;	// set to the end of the input string
	unsigned char*	orig = st;			// remember the beginning
	
	
	if (tw->termstate == DUMBTYPE)
	{
		// dumb terminal - raw mode for debugging, pass through escape sequences
		// and other special characters as <27> symbols
		for (i = 0; i < cnt; i++, st++)
		{
			if (*st == 27 || *st > 126 && !(tw->ftpState && gFTPServerPrefs->useISOTranslation))
			{
				sprintf((char*)tw->parsedat, "<%d>", *st);
				VSwrite(tw->vs, (char*)tw->parsedat, strlen((char*)tw->parsedat));
			}
			else VSwrite(tw->vs, (char*)st, 1);
		}
	}
	else
	{
		// traverse the string, looking for special characters indicating a mode change
		while (st < mark) 
		{
			switch (tw->telstate)
			{
				case GS:
				case ESCFOUND:
					if (tw->TEK.mode != kTektronixModeNotAllowed)
					{
						// TEK mode is enabled for this screen, so allow TEK mode changes
						if ((*st == 12) || (tw->telstate == GS))
						{
							// ESC-FF
							if ((*st == 12) && 
								((tw->termstate == VTEKTYPE) ||
								(tw->TEK.pageLocation == kTektronixPageLocationNewWindowClear)))
							{
								if (tw->termstate == VTEKTYPE)
								{
									// entering TEK mode
								}
								else if (tw->TEK.graphicsID > -1)
								{
									detachGraphics(tw->TEK.graphicsID);
								}
								
								if (tw->TEK.graphicsID <= -1)
								{
									// then there is no current TEK window for this screen - make one
									temptw = VGnewwin(TEK_DEVICE_WINDOW, tw->vs);
									
									if (temptw > -1)
									{
										Str255		scratchPstring;
										
										
										tw->TEK.graphicsID = temptw;
										
										VGgiveinfo(temptw);
										GetWTitle(tw->window, scratchPstring);
										StringUtilities_PToCInPlace(scratchPstring);
										RGattach(temptw, tw->vs, (char*)scratchPstring, tw->TEK.mode);
									}
									else tw->telstate = TS_NORM; // cant create TEK window
								}
								if (tw->telstate != TS_NORM)
								{
									Console_WriteLine("termstate->TEKTYPE");
									tw->termstate = TEKTYPE;
								}
							}
							
							if (tw->telstate == GS)
							{
								st++;
								VGwrite(tw->TEK.graphicsID, "\035", 1);
							}
							else if (tw->telstate != TS_NORM)
							{
								VGwrite(tw->TEK.graphicsID, "\037\033\014", 3);
							}
							
							tw->telstate = TS_NORM;
							break;
						}
					}
					
					if (*st == '^')
					{
						// ESC-^
						Console_WriteLine("termstate->RASTYPE");
						tw->termstate = RASTYPE;
						tw->telstate = TS_NORM;
						VRwrite("\033^", 2);
						orig = ++st;
						break;
					}
					
					if (tw->termstate == TEKTYPE) VGwrite(tw->TEK.graphicsID, "\033", 1);
					else if (tw->termstate  == RASTYPE) VRwrite("\033", 1);
					else VSwrite(tw->vs, "\033", 1); // send the missing ESC
					
					tw->telstate = TS_NORM;
					break;
			
				case IACFOUND:
					// This is the start of the telnet option negotiation.  If MacTelnet
					// gets an IAC (interpret as command) character, then negotiation
					// data follows, and is ready to be parsed.
					if (*st == TEL_IAC)
					{
						orig = st;
						st++;
						tw->telstate = TS_NORM;
						break;
					}
					else
					{
						switch (*st++)
						{
							case TEL_EOF:
							case TEL_EL:
							case TEL_EC:
							case TEL_AYT:
							case TEL_AO:
							case TEL_IP:
							case TEL_BREAK:
							case TEL_DM:
							case TEL_NOP:
							case TEL_SE:
							case TEL_GA:
								tw->telstate = TS_NORM;
								orig = st;
								break;
							
							case TEL_DOTEL:
								tw->telstate = TS_DO;
								orig = st;
								break;
							
							case TEL_WILLTEL:
								tw->telstate = TS_WILL;
								orig = st;
								break;
							
							case TEL_DONTTEL:
								tw->telstate = TS_DONT;
								orig = st;
								break;
							
							case TEL_WONTTEL:
								tw->telstate = TS_WONT;
								orig = st;
								break;
							
							case TEL_SB:
								// Subnegotiation - if there is an SB token found, then MacTelnet
								// needs to do some subnegotiating.  The subnegotiation data
								// follows, and needs to be put somewhere safe.  Make sure the
								// state gets updated, so it is clear that subnegotiation is
								// taking place.
								tw->telstate = TS_SB;
								orig = st;
								tw->parseIndex = 0; // no data yet - clear any possible garbage
								break;
							
							default: // unknown IAC <xxx>, where xxx is not known
								orig = st;
								tw->telstate = TS_NORM;						
								break;
						}
					}
					break;
				
				case TS_DO:
					telnet_do(tw, *st++);
					orig = st;
					tw->telstate = TS_NORM;
					break;
				
				case TS_DONT:
					telnet_dont(tw, *st++);
					orig = st;
					tw->telstate = TS_NORM;
					break;
				
				case TS_WILL:
					telnet_will(tw, *st++);
					orig = st;
					tw->telstate = TS_NORM;
					break;
				
				case TS_WONT:
					telnet_wont(tw, *st++);
					orig = st;
					tw->telstate = TS_NORM;
					break;
				
				case TS_SB:
					if (*st == TEL_IAC)
					{
						st++;
						if (*st == TEL_SE) // subnegotiation over 
						{
							st++;
							process_suboption(tw, st, mark);
							orig = st;
							tw->telstate = TS_NORM;
						}
						else
						{
							if (*st == TEL_IAC) // doubled IAC
							{
								tw->parsedat[tw->parseIndex++] = *st++;
							}
							else
							{
								// process this, then go IAC found
								tw->parsedat[tw->parseIndex++] = TEL_IAC;
								tw->parsedat[tw->parseIndex++] = *st;
								process_suboption(tw, st, mark);
								tw->substat = IACFOUND;
							}
						}
					}
					else
					{
						// collect subnegotiation data
						tw->parsedat[tw->parseIndex++] = *st++;
					}
					break;
				
				default:
					tw->telstate = TS_NORM;
					break;
			}
			
			// quick scan of the rest of the string, skipping characters that arent interesting
			if (tw->telstate == TS_NORM)
			{
				if (!tw->eightbit)
				{
					while (st < mark)
					{
						if (*st == IAC) break;												
						else
						{
							*st &= 0x7f; 										
							if (*st == ESC || *st == GS) break;											
							st++;
						}
					}
				}
				else while (st < mark && *st != ESC && *st < 255 && *st != GS) // translation mod
				{
					st++;
				}
				
				// send the string where it belongs
				if (!tw->timing)
				{
					if (tw->termstate == TEKTYPE)
					{
						register SInt16		i = 0;
						
						
						i = VGwrite(tw->TEK.graphicsID, (char*)orig, st - orig);
						if (i < (st - orig))
						{
							detachGraphics(tw->TEK.graphicsID);
							st = orig + i;
						}
					}
					else if (tw->termstate == RASTYPE)
					{
						register SInt16		i = 0;
						
						
						i = VRwrite((char*)orig, st - orig);
						if (i < (st - orig))
						{
							Console_WriteLine("termstate->VTEKTYPE");
							tw->termstate = VTEKTYPE;
							st = orig + i; // next character to parse
						}
					}
					else VSwrite(tw->vs, (char*)orig, st - orig); // send to virtual VT102
				}
				
				orig = st; // forget what has already been sent
				
				if (st < mark) switch (*st)
				{
					case TEL_IAC:
						tw->telstate = IACFOUND;
						st++;
						break;
					
					case GS:
						if (tw->telstate != GS) tw->telstate = GS;
						else tw->telstate = TS_NORM;
						st++;
						break;
					
					case ESC:
						if (st == mark - 1 || *(st + 1) == 12 || *(st + 1) == '^')
						{
							tw->telstate = ESCFOUND;
						}
						st++; // strip or accept the escape character
						break;
					
					default:
						if (*st++ > 127)
						{
							if (st == mark)
							{
								// new addition
								VSwrite(tw->vs, (char*)orig, 1);
	         				}
						}
						break;
				}// switch
			}// if
		}// while
	}
}// ShowOrInterpretAsCommand


/*!
To send a string, properly translated and possibly
echoed, as if it were typed in the specified
connections window, use this method.

(2.6)
*/
void
Parse_TypeString	(ConnectionDataPtr		inConnectionDataPtr,
					 char*					inString,
					 SInt16					inStringLength,
					 Boolean const			inSendToRecordingScripts)
{
	if (inConnectionDataPtr != null)
	{
		TextTranslation_ConvertTextToMac((unsigned char*)inString, inStringLength,
										inConnectionDataPtr->national);
		
		Network_SendFlush(inConnectionDataPtr->port);
		
		if (inConnectionDataPtr->kblen > 0)
		{
			// flush buffer
			Network_SendData(inConnectionDataPtr->port, inConnectionDataPtr->kbbuf,
								inConnectionDataPtr->kblen);
			inConnectionDataPtr->kblen = 0;
		}
		
		Network_SendData(inConnectionDataPtr->port, inString, inStringLength * sizeof(char));
		
		if (inConnectionDataPtr->echo)
		{
			Parse_ShowOrInterpretAsCommand(inConnectionDataPtr, (unsigned char*)inString, inStringLength);
		}
		
		// 3.0 - send to any scripts that may be recording
		if (inSendToRecordingScripts)
		{
			Parse_SendRecordableDataTransmitEvent(inString, inStringLength, false/* execute */);
		}
	}
}// TypeString


//
// internal methods
//

/*
 * encryptStatechange
 * Called when the encryption state changes
 */
static	void encryptStatechange (ConnectionDataPtr tw)
{
#if 0
	if (tw != null)
	{
		tnParams*			ptr = (tnParams*)tw->aedata;
		RealScreenDataPtr	realScreenPtr = ScreenManager_GetRealScreenDataOf(tw->vs);
		
		
		if ((ptr != null) && (realScreenPtr != null))
		{
			StatusIconRef		icon = realScreenPtr->encryptionStatusIcon;
			
			
			if (icon != null)
			{
				if ((ptr->encrypting) && (ptr->decrypting)) // when both are set, it means secure
				{
					// change the status icon to the now encrypted icon
					StatusIcon_SetNextCell(icon, kEncryptionIconIndexSecure);
				}
				else if (ptr->decrypting)
				{
					// change the status icon to the decrypting in progress icon
					StatusIcon_SetNextCell(icon, kEncryptionIconIndexDecrypting);
				}
				else if (ptr->encrypting)
				{
					// change the status icon to the encrypting in progress icon
					StatusIcon_SetNextCell(icon, kEncryptionIconIndexEncrypting);
				}
				else
				{
					// change the status icon to the not encrypted icon
					StatusIcon_SetNextCell(icon, kEncryptionIconIndexInsecure);
				}
			}
		}
	}
#endif
}

	
/*
 * getcname
 * Return a pointer to the canonical host name
 */
static char	*getcname (ConnectionDataPtr tw)
{
	char *cp;
	static char *b, buf[100];

	cp = 0;
	if (tw->host.canonical[0]) cp = tw->host.canonical;

//	Doing the following is bad because we disposed of our init params!
//	else if ((*(*(ConnectionSetup**)(tw->setupParametersHandle))->session)->hostname)
//		cp = (char*)(*(*(ConnectionSetup**)(tw->setupParametersHandle))->session)->hostname;

	/* make a local copy to avoid locking handles */
	if (cp) {
		b = buf;		
		while (*cp)
			*b++ = *cp++;			
		*b++ = '\0';
		return buf;
	}

	return cp;
}// getcname


static void
process_suboption		(ConnectionDataPtr		tw,
						 unsigned char*			UNUSED_ARGUMENT(start),
						 unsigned char*			UNUSED_ARGUMENT(end))
{	
	char	IACSB[] = { TEL_IAC, TEL_SB, 0, 0};
	char	IACSE[] = { TEL_IAC, TEL_SE};
	
	
	switch(tw->parsedat[0]) {
	/*------------------------------------------------------------------------------*
	* 	SUBNegotiate Termtype:  send the term type data now							*
	*------------------------------------------------------------------------------*/
		case N_TERMTYP:			
			if (tw->parsedat[1]==1) {
				char s[60], termtmp[40];
				StringUtilities_PToC(tw->terminal.answerBack, termtmp);
				IACSB[2] = '\030';
				IACSB[3] = 0;
				Network_SendFlushData(tw->port,IACSB,4);
				Network_SendFlush(tw->port);
				sprintf(s,"%s%c%c",termtmp, TEL_IAC, TEL_SE);
				Network_SendData(tw->port, s, strlen(s));
				}
			break;
			
	/*------------------------------------------------------------------------------*
	* 	SUBNegotiate Linemode:  set up local characters, modes, and such			*
	*------------------------------------------------------------------------------*/
		case N_LINEMODE:
			LineMode_SendSubOption(tw);
			break;

	/*------------------------------------------------------------------------------*
	* 	SUBNegotiate REMOTE_FLOW_CONTROL:  determine whether we control flow, and   *
	*									   what restarts flow			 			*
	*------------------------------------------------------------------------------*/
		case N_REMOTEFLOW:
			switch (tw->parsedat[1])
			{

				case FLOW_OFF:
					tw->allow_flow = false;
				break;
				
				case FLOW_ON:
					tw->allow_flow = true;
				break;
				
				case FLOW_RESTART_ANY:
					tw->restart_any_flow = true;
				break;
				
				case FLOW_RESTART_XON:
					tw->restart_any_flow = false;
				break;
				
				default:
				break;	
			}
		
		default: //dont know this subnegotiation!!
			break;
	}
}// process_suboption


/*!
To send a DO option to a connection with the
specified port number (as defined by the
Network module), use this method.

(2.6)
*/
static void
send_do		(SInt16						inPort,
			 TelnetNegotiableOption		inOption)
{
	char	data[] = { IAC, TEL_DOTEL, 0 };
	
	
	data[2] = inOption;
	Network_SendFlushData(inPort, data, 3);
}// send_do


/*!
To send a DONT option to a connection with the
specified port number (as defined by the
Network module), use this method.

(2.6)
*/
static void
send_dont	(SInt16						inPort,
			 TelnetNegotiableOption		inOption)
{
	char	data[] = { IAC, TEL_DONTTEL, 0 };
	
	
	data[2] = inOption;
	Network_SendFlushData(inPort, data, 3);
}// send_dont


/*!
To send a WILL option to a connection with the
specified port number (as defined by the
Network module), use this method.

(2.6)
*/
static void
send_will	(SInt16						inPort,
			 TelnetNegotiableOption		inOption)
{
	char	data[] = { IAC, TEL_WILLTEL, 0 };
	
	
	data[2] = inOption;
	Network_SendFlushData(inPort, data, 3);
}// send_will


/*!
To send a WONT option to a connection with the
specified port number (as defined by the
Network module), use this method.

(2.6)
*/
static	void
send_wont	(SInt16						inPort,
			 TelnetNegotiableOption		inOption)
{
	char	data[] = { IAC, TEL_WONTTEL, 0 };
	
	
	data[2] = inOption;
	Network_SendFlushData(inPort, data, 3);
}// send_wont


/*!
To locally do something specified by an option,
use this method.

(2.6)
*/
static void
telnet_do		(ConnectionDataPtr			inConnectionDataPtr,
				 TelnetNegotiableOption		inOption)
{
	switch (inOption)
	{
		case N_SGA: // suppress go-ahead
			if (!inConnectionDataPtr->Isga)
			{
				inConnectionDataPtr->Isga = 1;
				send_will(inConnectionDataPtr->port, N_SGA);
			}
			break;
		
		case N_TERMTYP: // request for terminal type
			if (!inConnectionDataPtr->Ittype)
			{
				inConnectionDataPtr->Ittype = 1;
				send_will(inConnectionDataPtr->port, N_TERMTYP);
			}
			break;
		
		case N_NAWS: // negotiate about window size
			inConnectionDataPtr->naws = 1;
			send_will(inConnectionDataPtr->port, N_NAWS);
			
			Parse_NegotiateAboutWindowSize(inConnectionDataPtr, VSmaxwidth(inConnectionDataPtr->vs) + 1,
											VSgetlines(inConnectionDataPtr->vs));
			break;
		
		case N_LINEMODE: // line mode
			if (inConnectionDataPtr->lineMode.allowed)
			{
				send_will(inConnectionDataPtr->port, N_LINEMODE);
				LineMode_SendData(inConnectionDataPtr);
			}
			else send_wont(inConnectionDataPtr->port, N_LINEMODE);
			break;
		
		case N_REMOTEFLOW:
			if (!inConnectionDataPtr->remote_flow)
			{
				inConnectionDataPtr->remote_flow = true;
				send_will(inConnectionDataPtr->port, inOption); 
			}
			break;
		
		default:
			// unrecognized, so obviously this terminal wont do it
			send_wont(inConnectionDataPtr->port, inOption);
			break;
	}
}// telnet_do


/*!
To locally not do something specified by an option,
use this method.

(2.6)
*/
static void
telnet_dont		(ConnectionDataPtr			inConnectionDataPtr,
				 TelnetNegotiableOption		inOption)
{
	if (inConnectionDataPtr != null) switch (inOption)
	{
		case N_ENCRYPT:
		case N_AUTHENTICATION:
			inConnectionDataPtr->myopts[inOption - MHOPTS_BASE] = 0;
			send_wont(inConnectionDataPtr->port, inOption);
			break;
		
		case N_LINEMODE: // turn it off, if its on
			if (inConnectionDataPtr->lineMode.active)  
			{
				send_wont(inConnectionDataPtr->port, N_LINEMODE);
				if (inConnectionDataPtr->kblen > 0)
				{
					Network_SendFlushData(inConnectionDataPtr->port, inConnectionDataPtr->kbbuf,
											inConnectionDataPtr->kblen);
				}	
				inConnectionDataPtr->lineMode.active = false;
				inConnectionDataPtr->lineMode.nextIsLiteral = false;
				inConnectionDataPtr->lineMode.mode = 0;
			}
			break;
		
		default:
			// unrecognized option
			break;
	}
}// telnet_dont


/*!
To remotely do something specified by an option,
use this method.

(2.6)
*/
static void
telnet_will		(ConnectionDataPtr			inConnectionDataPtr,
				 TelnetNegotiableOption		inOption)
{
	switch (inOption) 
	{
		case N_ECHO: // echo remotely
			if (inConnectionDataPtr->echo)
			{
				inConnectionDataPtr->echo = 0;
				send_do(inConnectionDataPtr->port, N_ECHO);
			}
			break;
		
		case N_SGA: // remotely suppress go-ahead
			if (!inConnectionDataPtr->Usga)
			{
				inConnectionDataPtr->Usga = 1;
				send_do(inConnectionDataPtr->port, N_SGA);
			}
			break;
		
		case N_TIMING: // no timing mark
			inConnectionDataPtr->timing = 0;
			break;
		
		case N_REMOTEFLOW: // remote flow control
			if (!inConnectionDataPtr->remote_flow)
			{
				inConnectionDataPtr->remote_flow = 1;
				send_do(inConnectionDataPtr->port, N_REMOTEFLOW);
			}
			break; // 3.0 - this break statement was missing
		
		default:
			// unrecognized, so make sure the remote terminal doesnt do it
			send_dont(inConnectionDataPtr->port, inOption);
			break; // 3.0 - this break statement was missing
	}
}// telnet_will


/*!
To remotely not do something specified by an option,
use this method.

(2.6)
*/
static void
telnet_wont		(ConnectionDataPtr			inConnectionDataPtr,
				 TelnetNegotiableOption		inOption)
{
	switch (inOption)
	{
		case N_ECHO: // will not echo remotely
			if (!inConnectionDataPtr->echo)
			{
				inConnectionDataPtr->echo = 1;
				send_dont(inConnectionDataPtr->port, N_ECHO);
			}
			break;
		
		case N_SGA: // will not suppress go-ahead
			if (inConnectionDataPtr->Usga)
			{
				inConnectionDataPtr->Usga = 0;
				send_dont(inConnectionDataPtr->port, N_SGA);
			}
			break;
		
		case N_TIMING: // will not use timing mark
			inConnectionDataPtr->timing = 0;
			break;
		
		case N_ENCRYPT: // will not use Kerberos encryption
		case N_AUTHENTICATION: // will not use Kerberos authentication
			inConnectionDataPtr->hisopts[inOption - MHOPTS_BASE] = 0;
			send_dont(inConnectionDataPtr->port, inOption);
			break;
		
		default:
			// unrecognized option
			break;
	}
}// telnet_wont