/*###############################################################

	Parse.h
	
	Miscellaneous functions for translating data, such as
	sending strings, signals and commands to a connection
	window.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#ifndef __PARSE__
#define __PARSE__

// MacTelnet includes
#include "ConnectionData.h"



//
// constants
//

typedef SInt16 TelnetNegotiableOption;

// the different telnet option negotiation tags
enum
{
	TEL_EOF			= 236,			// End of File							// octal 356
	TEL_SUSP		= 237,			// Suspend								// octal 357
	TEL_ABORT 		= 238,			// for linemode remote flow control		// octal 358
	TEL_SE			= 240,			// Sub-negotiation End					// octal 360
	TEL_NOP			= 241,			// No Operation 						// octal 361
	TEL_DM			= 242,			// Data Mark, accompanied by TCPUrgent	// octal 362
	TEL_BREAK		= 243,			// Break								// octal 363
	TEL_IP			= 244,			// Interrupt Process					// octal 364
	TEL_AO			= 245,			// Abort Output							// octal 365
	TEL_AYT			= 246,			// Are You There?						// octal 366
	TEL_EC			= 247,			// Erase Character						// octal 367
	TEL_EL			= 248,			// Erase Line							// octal 370
	TEL_GA			= 249,			// Telnet Go-Ahead						// octal 371
	TEL_SB			= 250,			// Sub-negotiation Begin				// octal 372
	TEL_WILLTEL 	= 251,			// Negotiate: Will do this option 		// octal 373
	TEL_WONTTEL 	= 252,			// Negotiate: Wont do this option 		// octal 374
	TEL_DOTEL		= 253,			// Negotiate: Do this option 			// octal 375
	TEL_DONTTEL 	= 254,			// Negotiate: Dont do this option 		// octal 376
	TEL_IAC			= 255			// Interpret As Command					// octal 377
};

// the rest of the constants needed for the option negotiations
enum
{
	TS_NORM		= 0,				// normal parsing
	TS_SB		= 1,				// subnegotiation begin (data follows)
	TS_SE		= 2,				// subnegotiation end
	TS_WILL		= 3,
	TS_WONT		= 4,
	TS_DO		= 7,
	TS_DONT		= 8,
	ESCFOUND 			= 5,		// found an ESC sequence
	IACFOUND 			= 6,		// negotiation data follows
	SEXECUTE			= 200,
	IAC					= 255,		// interpret as command, negotiation begin (data follows)
	GS					= 29,		// can drop into TEK from here
	
	// some of the various options that can be negotiated about
	N_BINARY 			= 0,		// binary data transfer
	N_ECHO				= 1,		// local echo
	N_SGA				= 3,		// Suppress Go-Ahead
	N_STATUS			= 5,
	N_TIMING			= 6,		// timing mark
	N_TERMTYP			= 24,		// set the terminal type
	N_NAWS				= 31,		// Negotiate About Window Size
	N_TERMSPEED			= 32,		// request for local terminals maximum speed
	N_REMOTEFLOW		= 33,		// do remote flow control
		FLOW_OFF            = 0,		// dont allow flow control  
		FLOW_ON             = 1,		// allow flow control
		FLOW_RESTART_ANY    = 2,		// restart flow on any character
		FLOW_RESTART_XON    = 3,		// restart flow only on XON
	N_LINEMODE 			= 34,
		L_MODE 				= 1,
		L_FORWARDMASK 		= 2,
		L_SLC 				= 3,
	N_XDISPLOC			= 35,
	N_AUTHENTICATION	= 37,		// Kerberos
	N_ENCRYPT			= 38,		// Kerberos
	N_SUBLIMINAL_MSG	= 257		// display subliminal messages (see RFC 1097)
};

// values for LINEMODE MODE
enum
{
	L_EDIT    	= 1,
	L_TRAPSIG  	= 2,
	L_MODE_ACK	= 4,
	L_SOFT_TAB	= 8,
	L_LIT_ECHO	= 16,
	L_MODE_MASK	= 31
};

// These are the local keys, used for line-mode negotiations.
enum
{
	SLC_NOSUPPORT	= 0,
	SLC_CANTCHANGE	= 1,
	SLC_VALUE		= 2,
	SLC_DEFAULT		= 3,
	SLC_LEVELBITS	= 3,
	SLC_FLUSHOUT	= 32,
	SLC_FLUSHIN		= 64,
	SLC_AWK			= 128,
	
	SLC_MAX			= LINE_MODE_SLC_MAX,
	
	SLC_SYNCH		= 1,
	SLC_BRK			= 2,
	SLC_IP			= 3,
	SLC_AO			= 4,
	SLC_AYT			= 5,
	SLC_EOR			= 6,
	SLC_ABORT		= 7,
	SLC_EOF			= 8,
	SLC_SUSP		= 9,
	SLC_EC			= 10,
	SLC_EL   		= 11,
	SLC_EW   		= 12,
	SLC_RP			= 13,
	SLC_LNEXT		= 14,
	SLC_XON			= 15,
	SLC_XOFF		= 16,
	SLC_FORW1		= 17,
	SLC_FORW2		= 18,
	
	DEFAULT_SLC_AO 		= 21,
	DEFAULT_SLC_IP 		= 3,
	DEFAULT_SLC_EC 		= 127,
	DEFAULT_SLC_EL 		= 21,
	DEFAULT_SLC_EOF 	= 4,
	DEFAULT_SLC_ABORT   = 3,
	DEFAULT_SLC_SUSP 	= 26,
	DEFAULT_SLC_EW 		= 23,
	DEFAULT_SLC_RP 		= 18,
	DEFAULT_SLC_LNEXT 	= 22,
	DEFAULT_SLC_XON 	= 17,
	DEFAULT_SLC_XOFF 	= 19,
	DEFAULT_SLC_FORW1 	= 22,
	DEFAULT_SLC_FORW2 	= 254
};

//
// public methods
//

void
	Parse_NegotiateAboutWindowSize				(ConnectionDataPtr		inScreen,
												 SInt16					inWidth,
												 SInt16					inHeight);

void
	Parse_SendInitialTelnetOptions				(ConnectionDataPtr		inScreen);

void
	Parse_SendLocalEchoOption					(ConnectionDataPtr		inScreen);

void
	Parse_SendRecordableDataTransmitEvent		(char*					inStringToType,
												 SInt16					inStringLength,
												 Boolean const			inExecute);

// PROVIDE AN ENUMERATED VALUE (FROM THE APPLE EVENT TERMINOLOGY) FOR A SPECIAL CHARACTER TO SEND
void
	Parse_SendRecordableSpecialTransmitEvent	(FourCharCode const		inSpecialCharacterEnumeration,
												 Boolean const			inExecute);

void
	Parse_ShowOrInterpretAsCommand				(ConnectionDataPtr		inScreen,
												 unsigned char*			inString,
												 SInt16					inStringLength);

void
	Parse_TypeString							(ConnectionDataPtr		inScreen,
												 char*					inString,
												 SInt16					inStringLength,
												 Boolean const			inSendToRecordingScripts);

#endif