/*###############################################################

	PrefPanelConfigurations.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Dialogs.h>
#include <Folders.h>
#include <Icons.h>
#include <PLStringFuncs.h>
#include <Resources.h>

// library includes
#include "ContextsLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "StringResources.h"
#include "SpacingConstants.r"
#include "TelnetPrefsDialog.r"

// MacTelnet includes
#include "configure.proto.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "MacLook.h"
#include "Panel.h"
#include "Preferences.h"
#include "PreferencesData.h"
#include "PreferencesDialog.h"
#include "PrefPanelConfigurations.h"
#include "EventLoop.h"
#include "ProxyServerDialog.h"
#include "ResourceFileManager.h"
#include "SessionEditorDialog.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



// constants
#define NUMBER_OF_CONFIGURATION_TYPES		4
enum
{
	// must be consecutive and zero-based, in order of tabs
	kConfigurationTypeIndexSessions = 0,
	kConfigurationTabIndexSessions = (kConfigurationTypeIndexSessions + 1),
	kConfigurationTypeIndexTerminals = 1,
	kConfigurationTabIndexTerminals = (kConfigurationTypeIndexTerminals + 1),
	kConfigurationTypeIndexFirewalls = 2,
	kConfigurationTabIndexFirewalls = (kConfigurationTypeIndexFirewalls + 1),
	kConfigurationTypeIndexFTPUsers = 3,
	kConfigurationTabIndexFTPUsers = (kConfigurationTypeIndexFTPUsers + 1)
};

// types
struct ConfigurationsPanelData
{
	struct
	{
		ControlRef		tabs,
						descriptionStaticText,
						listBox,
						buttonContainer,
						newButton,
						newGroupButton,
						removeButton,
						duplicateButton,
						editButton;
	} controls;
	
	// these are used to simplify calls to GetIndString()
	UInt16				descriptionStringIndices[NUMBER_OF_CONFIGURATION_TYPES];
	UInt16				duplicateTitleStringIndices[NUMBER_OF_CONFIGURATION_TYPES];
};
typedef struct ConfigurationsPanelData	ConfigurationsPanelData;
typedef ConfigurationsPanelData*		ConfigurationsPanelDataPtr;

// variables
extern MenuHandle		gListItemsDummyMenu;

// internal methods
static void				createPanelControls				(PanelRef						inPanel,
														 WindowRef						inOwningWindow);
static void				deltaPanelSize					(PanelRef						inPanel,
														 const Point					inChangeInWidthAndHeight);
static void				disposePanel					(PanelRef						inPanel,
														 void*							inDataPtr);
static SInt32			handleContentClick				(PanelRef						inPanel,
														 ControlRef						inControl);
static Boolean			isDefaultString					(ConfigurationsPanelDataPtr		inDataPtr,
														 StringPtr						inResourceName);
static TelnetConfigurationContextRef	newContext		(PanelRef						inPanel);
static SInt32			panelChanged					(PanelRef						inPanel,
														 PanelMessage					inMessage,
														 void*							inData);
static void				setButtonStates					(ConfigurationsPanelDataPtr		inDataPtr);
static void				setUpConfigurationsControls		(PanelRef						inPanel);
static void				setUpConfigurationsListControl	(const ResType					inPreferenceResourceType,
														 ConfigurationsPanelDataPtr		inDataPtr);



//
// public methods
//

/*!
This routine creates a new preference panel for
the Configurations category, initializes it, and
returns a reference to it.  You must destroy the
reference using Panel_Dispose() when you are
done with it.

If any problems occur, null is returned.

(3.0)
*/
PanelRef
PrefPanelConfigurations_New		(void)
{
	PanelRef		result = Panel_New(panelChanged);
	
	
	if (result != null)
	{
		ConfigurationsPanelDataPtr	dataPtr = (ConfigurationsPanelDataPtr)Memory_NewPtr(sizeof(ConfigurationsPanelData));
		Str31						name;
		
		
		Panel_SetDescriptor(result, kTelnetPanelDescriptorPreferencesConfigurations);
		GetIndString(name, rStringsConfigurationData, siConfigurationDataPreferencePanelNameConfigurations);
		Panel_SetName(result, name);
		Panel_SetIconSuite(result, kIconSuitePrefsConfigurations, kSelectorAllLargeData);
		Panel_SetAuxiliaryDataPtr(result, dataPtr);
		
		// initialize the data
		dataPtr->descriptionStringIndices[kConfigurationTypeIndexSessions] = siConfigurationDataPreferencePanelConfigurationsSessionsDescription;
		dataPtr->descriptionStringIndices[kConfigurationTypeIndexTerminals] = siConfigurationDataPreferencePanelConfigurationsTerminalsDescription;
		dataPtr->descriptionStringIndices[kConfigurationTypeIndexFirewalls] = siConfigurationDataPreferencePanelConfigurationsFirewallsDescription;
		dataPtr->descriptionStringIndices[kConfigurationTypeIndexFTPUsers] = siConfigurationDataPreferencePanelConfigurationsFTPUsersDescription;
		dataPtr->duplicateTitleStringIndices[kConfigurationTypeIndexSessions] = siConfigurationDataPreferencePanelConfigurationsSessionsDuplicateTitle;
		dataPtr->duplicateTitleStringIndices[kConfigurationTypeIndexTerminals] = siConfigurationDataPreferencePanelConfigurationsTerminalsDuplicateTitle;
		dataPtr->duplicateTitleStringIndices[kConfigurationTypeIndexFirewalls] = siConfigurationDataPreferencePanelConfigurationsFirewallsDuplicateTitle;
		dataPtr->duplicateTitleStringIndices[kConfigurationTypeIndexFTPUsers] = siConfigurationDataPreferencePanelConfigurationsFTPUsersDuplicateTitle;
	}
	return result;
}// New


/*!
Creates a new context for the specified panel,
which can be passed to a configuration
editing function.

(3.0)
*/
TelnetConfigurationContextRef
newContext		(PanelRef		inPanel)
{
	ConfigurationsPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	TelnetConfigurationContextRef	result = null;
	
	
	if (dataPtr != null)
	{
	}
	return result;
}// newContext


//
// internal methods
//

/*!
This routine creates the controls that belong in the
Configurations panel, and attaches them to the
specified owning window in the proper embedding
hierarchy.  The specified panels descriptor must be
"kTelnetPanelDescriptorPreferencesConfigurations".

The controls are not arranged or sized.

(3.0)
*/
static void
createPanelControls		(PanelRef			inPanel,
						 WindowRef			inOwningWindow)
{
	ControlRef					container = GetNewControl(kIDPanelUserPane, inOwningWindow),
								control = null;
	ConfigurationsPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	Panel_SetContainerControl(inPanel, container);
	SetControlVisibility(container, false/* visible */, false/* draw */);
	
	// tab control
	dataPtr->controls.tabs = control = GetNewControl(kIDPrefConfigurationsTabs, inOwningWindow);
	(OSErr)EmbedControl(control, container);
	
	// group box contents
	dataPtr->controls.descriptionStaticText = control = GetNewControl(kIDStaticText, inOwningWindow);
	(OSErr)EmbedControl(control, dataPtr->controls.tabs);
	dataPtr->controls.listBox = control = GetNewControl(kIDPrefConfigurationsListBox, inOwningWindow);
	{
		ListHandle		configurationList = null;
		
		
		ListUtilities_GetControlListHandle(control, &configurationList);
		SetListFlags(configurationList, lDoVAutoscroll);
		SetListSelectionFlags(configurationList, lOnlyOne | lNoDisjoint | lNoExtend | lUseSense);
	}
	(OSErr)EmbedControl(control, dataPtr->controls.tabs);
	dataPtr->controls.buttonContainer = control = GetNewControl(kIDAnchorUserPane, inOwningWindow);
	(OSErr)EmbedControl(control, dataPtr->controls.tabs);
	dataPtr->controls.newButton = control = GetNewControl(kIDPrefConfigurationsNewButton, inOwningWindow);
	(OSErr)EmbedControl(control, dataPtr->controls.buttonContainer);
	dataPtr->controls.newGroupButton = control = GetNewControl(kIDPrefConfigurationsNewGroupButton, inOwningWindow);
	(OSErr)EmbedControl(control, dataPtr->controls.buttonContainer);
	// temp - disable group button, this feature isnt working yet
	DeactivateControl(control);
	dataPtr->controls.removeButton = control = GetNewControl(kIDPrefConfigurationsRemoveButton, inOwningWindow);
	(OSErr)EmbedControl(control, dataPtr->controls.buttonContainer);
	dataPtr->controls.duplicateButton = control = GetNewControl(kIDPrefConfigurationsDuplicateButton, inOwningWindow);
	(OSErr)EmbedControl(control, dataPtr->controls.buttonContainer);
	dataPtr->controls.editButton = control = GetNewControl(kIDPrefConfigurationsEditButton, inOwningWindow);
	(OSErr)EmbedControl(control, dataPtr->controls.buttonContainer);
}// createPanelControls


/*!
This routine adjusts the controls in preference panels
to match the specified change in dimensions of a panels
container.

(3.0)
*/
static void
deltaPanelSize		(PanelRef			inPanel,
					 const Point		inChangeInWidthAndHeight)
{
	ControlRef					control = null;
	ConfigurationsPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	Rect						controlRect;
	
	
	// move and/or resize controls
	{
		Str255		text;
		
		
		// resize the containing group box
		control = dataPtr->controls.tabs;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
								controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
		
		// recalculate the smallest area that the text can occupy
		control = dataPtr->controls.descriptionStaticText;
		GetControlBounds(control, &controlRect);
		MoveControl(control, controlRect.left, controlRect.top + inChangeInWidthAndHeight.v);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
		GetControlText(control, text);
		(UInt16)Localization_SetUpMultiLineTextControl(control, text, kThemeSmallSystemFont); // auto-sizes text area
		
		// depending on the locale, decide how to resize and/or rearrange the list and buttons
		{
			Boolean		isLeftToRight = Localization_IsLeftToRight(); // text reading direction
			
			
			control = dataPtr->controls.listBox;
			GetControlBounds(control, &controlRect);
			unless (isLeftToRight) MoveControl(control, controlRect.left + inChangeInWidthAndHeight.h, controlRect.top);
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
									controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
			
			control = dataPtr->controls.buttonContainer;
			GetControlBounds(control, &controlRect);
			if (isLeftToRight) MoveControl(control, controlRect.left + inChangeInWidthAndHeight.h, controlRect.top);
			SizeControl(control, controlRect.right - controlRect.left,
									controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
		}
	}
	
	// now make sure the cell width of the list is correct
	{
		ListHandle		list = null;
		
		
		ListUtilities_GetControlListHandle(dataPtr->controls.listBox, &list);
		{
			Point			newCellDimensions;
			Rect			listBounds;
			
			
			GetControlBounds(dataPtr->controls.listBox, &listBounds);
			SetPt(&newCellDimensions, listBounds.right - listBounds.left, 18);
			LSetDrawingMode(false, list);
			LCellSize(newCellDimensions, list);
			LSetDrawingMode(true, list);
		}
	}
}// deltaPanelSize


/*!
This routine responds to a change in the existence
of a panel.  The panel is about to be destroyed, so
this routine disposes of private data structures
associated with the specified panel.

(3.0)
*/
static void
disposePanel	(PanelRef		UNUSED_ARGUMENT(inPanel),
				 void*			inDataPtr)
{
	ConfigurationsPanelDataPtr		dataPtr = inDataPtr;
	
	
	Memory_DisposePtr((Ptr*)&dataPtr);
	DisposeMenu(gListItemsDummyMenu), gListItemsDummyMenu = null;
}// disposePanel


/*!
This method responds to control clicks.

(3.0)
*/
static SInt32
handleContentClick	(PanelRef		inPanel,
					 ControlRef		inControl)
{
	SInt32						result = kPanel_ResponseClickNotHandled;
	ConfigurationsPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	ConfigurationEditProcPtr	procPtr = null;
	ResType						resType = '????';
	
	
	// determine the type of resource being examined
	switch (GetControlValue(dataPtr->controls.tabs))
	{
		case kConfigurationTabIndexSessions: // Session Favorites
			resType = SESSIONPREFS_RESTYPE;
			procPtr = SessionEditorDialog_Display;
			break;
		
		case kConfigurationTabIndexTerminals: // Terminals
			resType = TERMINALPREFS_RESTYPE;
			procPtr = EditTerminal;
			break;
		
		case kConfigurationTabIndexFirewalls: // Proxies
			resType = PROXYPREFS_RESTYPE;
			procPtr = ProxyServerDialog_Display;
			break;
		
		case kConfigurationTabIndexFTPUsers: // FTP Users
			resType = FTPUSERPREFS_RESTYPE;
			procPtr = EditFTPUser;
			break;
		
		default:
			// ???
			procPtr = null; // signal failure
			break;
	}
	
	if (procPtr != null)
	{
		Boolean			handled = false;
		
		
		if (inControl == dataPtr->controls.tabs)
		{
			Str255		descriptionString;
			Rect		textBounds;
			
			
			// change the description string
			GetIndString(descriptionString, rStringsConfigurationData,
							dataPtr->descriptionStringIndices[GetControlValue(dataPtr->controls.tabs) - 1]);
			SetControlText(dataPtr->controls.descriptionStaticText, descriptionString);
			GetControlBounds(dataPtr->controls.descriptionStaticText, &textBounds);
			MacLook_InvalidateRectangle(GetControlOwner(dataPtr->controls.descriptionStaticText), &textBounds);
			
			// update the list to reflect a new set of preferences
			setUpConfigurationsListControl(resType, dataPtr);
			handled = true;
		}
		else if (inControl == dataPtr->controls.newButton)
		{
			// create a resource, and then display the appropriate editor dialog box for editing the selected item
			Str255		nameOfNewResource;
			
			
			nameOfNewResource[0] = 0; // this tells the editor routine to create a new resource
			if ((*procPtr)((StringPtr)&nameOfNewResource)) 
			{
				// 3.0 - theme sounds!
				if (FlagManager_Test(kFlagAppearance1_1API)) (OSStatus)PlayThemeSound(kThemeSoundNewItem);
				setUpConfigurationsListControl(resType, dataPtr);
				CheckOpSpecSubmenu(); // fix New Session From Favorites submenu
			}
			handled = true;
		}
		else if (inControl == dataPtr->controls.newGroupButton)
		{
			// create a new group - unimplemented
			BEEP;
			handled = true;
		}
		else if (inControl == dataPtr->controls.removeButton)
		{
			// destroy the selected items corresponding preferences
			Cell			cell;
			ListHandle		configurationList = null;
			
			
			ListUtilities_GetControlListHandle(dataPtr->controls.listBox, &configurationList);
			SetPt(&cell, 0, 0);
			if (LGetSelect(true/* okay if selection is not exactly the given cell */, &cell, configurationList)) 
			{
				ModalFilterUPP			upp = NewModalFilterUPP(StandardAlertEventFilter);
				InterfaceLibAlertRef	box = null;
				Str255					dialogText,
										helpText,
										itemName;
				SInt16					length = 254;
				
				
				LGetCell(itemName + 1, &length, cell, configurationList);
				itemName[0] = (char)length;
				
				if (isDefaultString(dataPtr, itemName))
				{
					GetIndString(dialogText, rStringsGeneralMessages, siCantRemoveDefaultListItem);
					
					box = Alert_New();
					Alert_SetFilterProc(box, upp);
					Alert_SetHelpButton(box, false);
					Alert_SetParamsFor(box, kAlert_StyleOK);
					Alert_SetText(box, dialogText, EMPTY_PSTRING);
					Alert_SetType(box, kAlertStopAlert);
					BEEP;
					Alert_Display(box);
					Alert_Dispose(&box);
				}
				else
				{
					SInt16		alertItemHit = kAlertStdAlertOKButton;
					
					
					if (!EventLoop_IsOptionKeyDown())
					{
						GetIndString(dialogText, rStringsCautionAlerts, siConfirmRemoveListItem);
						StringUtilities_PMetaSetDefaults();
						StringUtilities_PBuild(dialogText, itemName, EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
						GetIndString(helpText, rStringsCautionAlerts, siProcedureNotUndoableHelpText);
						
						box = Alert_New();
						Alert_SetFilterProc(box, upp);
						Alert_SetHelpButton(box, false);
						Alert_SetParamsFor(box, kAlert_StyleOKCancel);
						Alert_SetText(box, dialogText, helpText);
						Alert_SetType(box, kAlertCautionAlert);
						Alert_Display(box);
						alertItemHit = Alert_ItemHit(box);
						Alert_Dispose(&box);
					}
					else alertItemHit = kAlertStdAlertOKButton; // option key means just delete it, spare the commercial
					
					if (alertItemHit == kAlertStdAlertOKButton)
					{
						Handle		itemResourceHandle = null;
						
						
						DeleteMenuItem(gListItemsDummyMenu, Menus_GetItemIndexByText(gListItemsDummyMenu, itemName));
						LDelRow(1, cell.v, configurationList);
						cell.v--;
						LSetSelect(true, cell, configurationList);
						setButtonStates(dataPtr);
						
						ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
						
						// 3.0 - a kludge to remove supplementary resources special to some types
						itemResourceHandle = Get1NamedResource(resType, itemName);
						DetachResource(itemResourceHandle);
						if (resType == FTPUSERPREFS_RESTYPE)
						{
							FTPUserHandle		handle = (FTPUserHandle)itemResourceHandle;
							TelnetPrefsAliasID	aliasID = kInvalidTelnetPrefsAliasID;
							
							
							HLock((Handle)handle);
							aliasID = Preferences_NewSavedAlias((*handle)->defaultDirAliasID);
							if (aliasID != kInvalidTelnetPrefsAliasID) Preferences_DeleteSavedAlias(aliasID);
							else BEEP;
							Preferences_DisposeAlias(aliasID);
						}
						DisposeHandle(itemResourceHandle), itemResourceHandle = null;
						
						itemResourceHandle = Get1NamedResource(resType, itemName);
						RemoveResource(itemResourceHandle);
						ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences), itemResourceHandle = null;
						CheckOpSpecSubmenu(); // fix New Session From Favorites submenu
					}
				}
				DisposeModalFilterUPP(upp), upp = null;
			}
			handled = true;
		}
		else if (inControl == dataPtr->controls.duplicateButton)
		{
			// create a copy of the selected items preferences, and update the list
			Cell			cell;
			ListHandle		configurationList = null;
			
			
			ListUtilities_GetControlListHandle(dataPtr->controls.listBox, &configurationList);
			
			SetPt(&cell, 0, 0);
			if (LGetSelect(true/* okay if selection is not exactly the given cell */, &cell, configurationList)) 
			{
				Str255		originalName,
							itemName,
							dialogTitle;
				SInt16		length = 254;
				
				
				// extract the selected list cells name
				LGetCell(originalName + 1, &length, cell, configurationList);
				LSetSelect(false, cell, configurationList);
				originalName[0] = (char)length;
				PLstrcpy(itemName, originalName);
				PLstrcat(itemName, "\p copy"); // LOCALIZE THIS
				Menus_GetUniqueItemText(gListItemsDummyMenu, itemName);
				
				switch (GetControlValue(dataPtr->controls.tabs))
				{
					case kConfigurationTabIndexSessions: // Session Favorites
						GetIndString(dialogTitle, rStringsConfigurationData,
										siConfigurationDataPreferencePanelConfigurationsSessionsDuplicateTitle);
						break;
					
					case kConfigurationTabIndexTerminals: // Terminals
						GetIndString(dialogTitle, rStringsConfigurationData,
										siConfigurationDataPreferencePanelConfigurationsTerminalsDuplicateTitle);
						break;
					
					case kConfigurationTabIndexFirewalls: // Proxies
						GetIndString(dialogTitle, rStringsConfigurationData,
										siConfigurationDataPreferencePanelConfigurationsFirewallsDuplicateTitle);
						break;
					
					case kConfigurationTabIndexFTPUsers: // FTP Users
						GetIndString(dialogTitle, rStringsConfigurationData,
										siConfigurationDataPreferencePanelConfigurationsFTPUsersDuplicateTitle);
						break;
					
					default:
						// ???
						PLstrcpy(dialogTitle, EMPTY_PSTRING);
						break;
				}
				
				// display a dialog asking the user what the name of the duplicate should be
				if (Configure_NameNewConfigurationDialogDisplay(dialogTitle, itemName))
				{
					Handle			oldResource = null,
									newResource = null;
					SInt16			resourceID = 0;
					SInt16			index = 0;
					OSErr			error = noErr;
					
					
					// 3.0 - theme sounds!
					if (FlagManager_Test(kFlagAppearance1_1API)) (OSStatus)PlayThemeSound(kThemeSoundNewItem);
					
					// now read the resource with the original name, and construct a copy of it
					ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
					oldResource = GetNamedResource(resType, originalName);
					error = ResError();
					if (error != noErr) Console_WriteValue("duplicate: cannot get named resource, error", error);
					else
					{
						// ensure the users new name doesnt trample on any existing one
						Menus_GetUniqueItemText(gListItemsDummyMenu, itemName);
						
						// now create a duplicate resource
						resourceID = Unique1ID(resType);
						newResource = Memory_NewHandle(GetHandleSize(oldResource));
						if (newResource == null) Console_WriteLine("duplicate: cannot allocate memory!!!");
						else
						{
							BlockMoveData(*oldResource, *newResource, GetHandleSize(newResource));
							AddResource((Handle)newResource, resType, resourceID, itemName);
							error = ResError();
							if (error != noErr) Console_WriteValue("duplicate: cannot add resource, error", error);
							else
							{
								ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
								error = ResError();
								if (error != noErr) Console_WriteValue("duplicate: resource file update failed, error", error);
								AppendMenu(gListItemsDummyMenu, itemName);
								index = CountMenuItems(gListItemsDummyMenu) - 1;
								cell.v = index; 
								LAddRow(1, index, configurationList); // add a new row
								ListUtilities_SetListItemText(configurationList, cell.v, itemName);
								LSetSelect(true, cell, configurationList);
								CheckOpSpecSubmenu(); // fix New Session From Favorites submenu
							}
						}
					}
					if (newResource != null) ReleaseResource((Handle)newResource);
					if (oldResource != null) ReleaseResource((Handle)oldResource);
				}
			}
			handled = true;
		}
		else if ((inControl == dataPtr->controls.editButton) ||
					(inControl == dataPtr->controls.listBox))
		{
			// display the appropriate editor dialog box for editing the new item
			Boolean			doEdit = false;
			Cell			cell;
			ListHandle		configurationList = null;
			
			
			ListUtilities_GetControlListHandle(dataPtr->controls.listBox, &configurationList);
			
			SetPt(&cell, 0, 0);
			
			// on double-click, edit selected item automatically
			if (inControl == dataPtr->controls.listBox)
			{
				Boolean			isDoubleClick = false;
				Size			actualSize = 0;
				
				
				// set button states based on the list selection
				setButtonStates(dataPtr);
				
				// check for list double-clicks...
				if (GetControlData(dataPtr->controls.listBox, kControlListBoxDoubleClickPart,
									kControlListBoxDoubleClickTag, sizeof(Boolean), (Ptr)&isDoubleClick,
									&actualSize) == noErr)
				{
					if ((isDoubleClick) && (LGetSelect(true/* okay if selection is not exactly the given cell */, &cell, configurationList)))
					{
						// only edit the list if there is something selected!
						FlashButtonControl(dataPtr->controls.editButton, false/* is default */, false/* is cancel */);
						doEdit = true;
					}
				}
			}
			else doEdit = true;
			
			if (doEdit)
			{
				Str255			itemName;
				SInt16			length = 254;
				
				
				Embedding_OffscreenDrawOneControl(inControl);
				SetPt(&cell, 0, 0);
				if (LGetSelect(true/* okay if selection is not exactly the given cell */, &cell, configurationList)) 
				{
					Str255		junk;
					
					
					LGetCell(itemName + 1, &length, cell, configurationList);
					itemName[0] = (char)length;
					PLstrcpy(junk, itemName);
					
					if ((*procPtr)((StringPtr)itemName)) 
					{
						if (!(EqualString(itemName, junk, true, false))) // new name
						{
							LDelRow(1, cell.v, configurationList); // delete the old row
							cell.v = -1;
							LAddRow(1, cell.v, configurationList); // add a new row
							LSetCell(&itemName[1], itemName[0], cell, configurationList); // write the name in
							LSetSelect(true, cell, configurationList);
							setButtonStates(dataPtr);
						}
					}
				}
			}
			handled = true;
		}
		if (handled) result = kPanel_ResponseClickHandled;
	}
	else BEEP;
	
	return result;
}// handleContentClick


/*!
To determine if the specified resource conflicts
with the designated default resource name for the
currently selected list (tab), use this method.

(3.0)
*/
static Boolean
isDefaultString		(ConfigurationsPanelDataPtr		inDataPtr,
					 StringPtr						inResourceName)
{
	Str255			defaultName;
	Boolean			result = false,
					callEqualString = true;
	
	
	switch (GetControlValue(inDataPtr->controls.tabs))
	{
		case kConfigurationTabIndexSessions:
			StringUtilities_CToP(SESSIONPREFS_DEFAULTNAME, defaultName);
			break;
		
		case kConfigurationTabIndexTerminals:
			StringUtilities_CToP(TERMINALPREFS_DEFAULTNAME, defaultName);
			break;
		
		case kConfigurationTabIndexFirewalls:
			StringUtilities_CToP(PROXYPREFS_DEFAULTNAME, defaultName);
			break;
		
		case kConfigurationTabIndexFTPUsers:
			result = false;
			callEqualString = false;
			break;
		
		default:
			break;
	}
	if (callEqualString) result = EqualString(inResourceName, defaultName, true/* case sensitive */, false/* diacritics sensitive */);
	return result;
}// isDefaultString


/*!
This routine, of standard PanelChangedProcPtr form,
is invoked by the Panel module whenever a property
of one of the preferences dialogs panels changes.

(3.0)
*/
static SInt32
panelChanged	(PanelRef			inPanel,
				 PanelMessage		inMessage,
				 void*				inData)
{
	SInt32		result = 0L;
	
	
	if (Panel_GetDescriptor(inPanel) == kTelnetPanelDescriptorPreferencesConfigurations) switch (inMessage)
	{
		case kPanel_MessageClicked: // notification of a content click
			{
				ControlRef*			dataPtr = inData;
				
				
				result = handleContentClick(inPanel, *dataPtr);
			}
			break;
		
		case kPanel_MessageCreateControls: // specification of the window containing the panel - create controls using this window
			{
				WindowRef	window = *((WindowRef*)inData);
				
				
				createPanelControls(inPanel, window);
			}
			break;
		
		case kPanel_MessageDeltaSize: // request to alter size and/or position of controls; containing control will already have been changed
			{
				Point			changeInPanelDimensions;
				ControlRef		panelControl = null;
				Boolean			wasVisible = false;
				
				
				SetPt(&changeInPanelDimensions, ((Point*)inData)->h, ((Point*)inData)->v);
				Panel_GetContainerControl(inPanel, &panelControl);
				wasVisible = IsControlVisible(panelControl);
				SetControlVisibility(panelControl, false/* visibility */, wasVisible/* draw */);
				deltaPanelSize(inPanel, changeInPanelDimensions);
				SetControlVisibility(panelControl, wasVisible/* visibility */, wasVisible/* draw */);
			}
			break;
		
		case kPanel_MessageDestroyed: // request to dispose of private data structures
			{
				void*		dataPtr = inData;
				
				
				disposePanel(inPanel, dataPtr);
			}
			break;
		
		case kPanel_MessageFocusGained: // notification that a control is now focused
			{
				ControlRef*		dataPtr = inData;
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageFocusLost: // notification that a control is no longer focused
			{
				ControlRef*		dataPtr = inData;
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageInitialSize: // request for panel to size its controls
			{
				Point		exactPanelDimensions;
				
				
				SetPt(&exactPanelDimensions, ((Point*)inData)->h, ((Point*)inData)->v);
				setUpConfigurationsControls(inPanel);
			}
			break;
		
		case kPanel_MessageNewActiveState: // request to change panel controls to a new active (true) or inactive (false) state
			{
				Boolean		isNowActive = *((Boolean*)inData);
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageNewAppearanceTheme: // notification of theme switch, a request to recalculate control sizes
			{
				// this notification is currently ignored, but shouldnt be...
			}
			break;
		
		case kPanel_MessageNewVisibility: // visible state of the panels container has changed to visible (true) or invisible (false)
			{
				Boolean		isNowVisible = *((Boolean*)inData);
				
				
				// do nothing
			}
			break;
		
		default:
			break;
	}
	
	return result;
}// panelChanged


/*!
To automatically check the list control to see if
any items are selected, activating or deactivating
buttons appropriately, use this method.

(3.0)
*/
static void
setButtonStates		(ConfigurationsPanelDataPtr		inDataPtr)
{
	ListHandle		configurationList = null;
	Cell			cell;
	
	
	// set button states based on the list selection
	ListUtilities_GetControlListHandle(inDataPtr->controls.listBox, &configurationList);
	SetPt(&cell, 0, 0);
	if (LGetSelect(true/* okay if selection is not exactly the given cell */, &cell, configurationList))
	{
		ActivateControl(inDataPtr->controls.removeButton);
		ActivateControl(inDataPtr->controls.duplicateButton);
		ActivateControl(inDataPtr->controls.editButton);
	}
	else
	{
		DeactivateControl(inDataPtr->controls.removeButton);
		DeactivateControl(inDataPtr->controls.duplicateButton);
		DeactivateControl(inDataPtr->controls.editButton);
	}
}// setButtonStates


/*!
This routine sizes and arranges controls in preference
panels to fit the prescribed dimensions of a panels
container.

(3.0)
*/
static void
setUpConfigurationsControls		(PanelRef		inPanel)
{
	ConfigurationsPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	ControlRef						container = null,
									control = null;
	Rect							containerRect,
									controlRect;
	register SInt16					i = 0;
	UInt16							leftEdge = 0, // cumulative, resettable measurement allowing controls to sandwich horizontally
									topEdge = 0; // cumulative, resettable measurement allowing controls to stack vertically
	UInt16							buttonsMaximumWidth = 0,
									interiorWidth = 0,
									listInitialHeight = 0;
	Point							tabPaneTopLeft,
									tabPaneWidthHeight;
	
	
	Panel_GetContainerControl(inPanel, &container);
	GetControlBounds(container, &containerRect);
	
	PreferencesDialog_CalculateTabPane(inPanel, &tabPaneTopLeft, &tabPaneWidthHeight);
	
	// size controls
	control = dataPtr->controls.tabs;
	SizeControl(control, tabPaneWidthHeight.h + INTEGER_DOUBLED(HSP_TABPANE_AND_TAB),
						tabPaneWidthHeight.v + TAB_HT_BIG + INTEGER_DOUBLED(VSP_TABPANE_AND_TAB));
	interiorWidth = tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL);
	control = dataPtr->controls.descriptionStaticText;
	{
		Str255		string;
		UInt16		textHeight = 0;
		
		
		// initialize to the first set (may change if the user selects a different tab)
		GetIndString(string, rStringsConfigurationData, dataPtr->descriptionStringIndices[0]);
		SizeControl(control, interiorWidth, 0);
		textHeight = Localization_SetUpMultiLineTextControl(control, string, kThemeSmallSystemFont); // auto-sizes and sets font
		// TMP - hack in extra space since I cant figure out why this gets calculated incorrectly
		textHeight += 10;
		listInitialHeight = tabPaneWidthHeight.v - textHeight - VSP_CONTROLS;
	}
	{
		// find the widest button, and make the container that wide
		UInt16		buttonWidth = 0;
		
		
		buttonWidth = Localization_AutoSizeButtonControl(dataPtr->controls.newButton, 0);
		buttonWidth = Localization_AutoSizeButtonControl(dataPtr->controls.newGroupButton, buttonWidth);
		buttonWidth = Localization_AutoSizeButtonControl(dataPtr->controls.removeButton, buttonWidth);
		buttonWidth = Localization_AutoSizeButtonControl(dataPtr->controls.duplicateButton, buttonWidth);
		buttonWidth = Localization_AutoSizeButtonControl(dataPtr->controls.editButton, buttonWidth);
		buttonsMaximumWidth = buttonWidth;
		control = dataPtr->controls.newButton;
		GetControlBounds(control, &controlRect);
		SizeControl(control, buttonsMaximumWidth, controlRect.bottom - controlRect.top);
		control = dataPtr->controls.newGroupButton;
		SizeControl(control, buttonsMaximumWidth, controlRect.bottom - controlRect.top);
		control = dataPtr->controls.removeButton;
		SizeControl(control, buttonsMaximumWidth, controlRect.bottom - controlRect.top);
		control = dataPtr->controls.duplicateButton;
		SizeControl(control, buttonsMaximumWidth, controlRect.bottom - controlRect.top);
		control = dataPtr->controls.editButton;
		SizeControl(control, buttonsMaximumWidth, controlRect.bottom - controlRect.top);
	}
	control = dataPtr->controls.buttonContainer;
	SizeControl(control, buttonsMaximumWidth, listInitialHeight);
	control = dataPtr->controls.listBox;
	SizeControl(control, interiorWidth - HSP_BUTTON_AND_CONTROL - buttonsMaximumWidth, listInitialHeight);
	
	// arrange controls
	topEdge = tabPaneTopLeft.v;
	leftEdge = tabPaneTopLeft.h + INTEGER_HALVED(tabPaneWidthHeight.h - interiorWidth);
	{
		UInt16		buttonTopEdge = 0;
		
		
		control = dataPtr->controls.newButton;
		GetControlBounds(control, &controlRect);
		MoveControl(control, 0, buttonTopEdge);
		buttonTopEdge += (controlRect.bottom - controlRect.top + VSP_BUTTONS);
		control = dataPtr->controls.newGroupButton;
		MoveControl(control, 0, buttonTopEdge);
		buttonTopEdge += (controlRect.bottom - controlRect.top + VSP_BUTTONS);
		control = dataPtr->controls.removeButton;
		MoveControl(control, 0, buttonTopEdge);
		buttonTopEdge += (controlRect.bottom - controlRect.top + VSP_BUTTONS);
		control = dataPtr->controls.duplicateButton;
		MoveControl(control, 0, buttonTopEdge);
		buttonTopEdge += (controlRect.bottom - controlRect.top + VSP_BUTTONS);
		control = dataPtr->controls.editButton;
		MoveControl(control, 0, buttonTopEdge);
		buttonTopEdge += (controlRect.bottom - controlRect.top + VSP_BUTTONS);
	}
	control = dataPtr->controls.buttonContainer;
	MoveControl(control, leftEdge + interiorWidth - buttonsMaximumWidth, topEdge);
	control = dataPtr->controls.listBox;
	MoveControl(control, leftEdge, topEdge);
	topEdge += (listInitialHeight + VSP_CONTROLS);
	control = dataPtr->controls.descriptionStaticText;
	MoveControl(control, leftEdge, topEdge);
	topEdge = VSP_TAB_AND_DIALOG;
	leftEdge = 0;
	control = dataPtr->controls.tabs;
	MoveControl(control, leftEdge, topEdge);
	
	// now localize by potentially switching the positions of the buttons and the list
	Localization_HorizontallyPlaceControls(dataPtr->controls.listBox, dataPtr->controls.buttonContainer);
	
	// initialize list contents
	setUpConfigurationsListControl(SESSIONPREFS_RESTYPE, dataPtr);
}// setUpConfigurationsControls


/*!
To erase and rebuild the specified list to
include all resources of the given type in
the preferences file, use this method.

(3.0)
*/
static void
setUpConfigurationsListControl		(const ResType					inPreferenceResourceType,
									 ConfigurationsPanelDataPtr		inDataPtr)
{
	SInt16				oldResFile = CurResFile();
	SInt16				itemCount = 0;
	register SInt16		i = 0;
	ListHandle			list = null;
	
	
	ListUtilities_GetControlListHandle(inDataPtr->controls.listBox, &list);
	if (list == null) Console_WriteLine("WARNING: Configuration list creation apparently failed!");
	
	// first, erase the list
	LDelRow(0, 0, list);
	
	// now re-populate the list using resource information
	{
		Cell		cell;
		Str255		string;
		
		
		if (gListItemsDummyMenu != null) DisposeMenu(gListItemsDummyMenu), gListItemsDummyMenu = null;
		gListItemsDummyMenu = Menus_NewMenuThemeSavvy(kMenuIDPrefsConfigurationsListDummy, EMPTY_PSTRING);
		if (gListItemsDummyMenu == null) Console_WriteLine("WARNING: Mock-menu creation apparently failed!");
		ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
		itemCount = Count1Resources(inPreferenceResourceType);
		InsertResMenu(gListItemsDummyMenu, inPreferenceResourceType, 999/* insert after item */);
		for (i = 1, cell.v = 0, cell.h = 0; i <= itemCount; i++, cell.v++)
		{
			GetMenuItemText(gListItemsDummyMenu, i, string);
			LAddRow(1, -1, list);
			ListUtilities_SetListItemText(list, cell.v, string);
		}
		cell.v = 0;
		cell.h = 0;
		if (itemCount) LSetSelect(1, cell, list);
		setButtonStates(inDataPtr);
	}
	
	// update the list
	Embedding_OffscreenDrawOneControl(inDataPtr->controls.listBox);
	
	UseResFile(oldResFile);
}// setUpConfigurationsListControl