/*###############################################################

	PrefPanelFileTransfers.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Dialogs.h>
#include <Folders.h>
#include <Icons.h>
#include <Resources.h>
#include <StandardFile.h>

// library includes
#include "ContextsLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "GeneralResources.h"
#include "StringResources.h"
#include "SpacingConstants.r"
#include "TelnetPrefsDialog.r"

// MacTelnet includes
#include "configure.proto.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "EditLogInMessageDialog.h"
#include "NetBackground.h"
#include "Panel.h"
#include "Preferences.h"
#include "PreferencesData.h"
#include "PreferencesDialog.h"
#include "PrefPanelFileTransfers.h"
#include "ResourceFileManager.h"
#include "EventLoop.h"
#include "TelnetDescriptors.h"



// constants
enum
{
	kTypeCreatorFieldWidth = 60
};
#define NUMBER_OF_FILETRANSFERS_TABPANES	3
enum
{
	// must reflect 'tab#' information and be one-based
	kTabIndexFileTransfersOptions = 1,
	kTabIndexFileTransfersFileTypes = 2,
	kTabIndexFileTransfersServer = 3
};
#define NUMBER_OF_FILETRANSFERS_OPTIONS		4
enum
{
	// must be consecutive and zero-based
	kFileTransfersOptionIndexShowLogAtStartup = 0,
	kFileTransfersOptionIndexUseMacBinaryII = 1,
	kFileTransfersOptionIndexResetMacBinary = 2,
	kFileTransfersOptionIndexUseISOTranslation = 3
};
#define NUMBER_OF_FILETRANSFERS_SERVERMODES	3
enum
{
	// must be consecutive and zero-based
	kFileTransfersServerModeIndexOff = 0,
	kFileTransfersServerModeIndexOnAnonymous = 1,
	kFileTransfersServerModeIndexOnWithUserLogin = 2
};

// types
struct FileTransfersPanelData
{
	ControlRef		tabControl;
	
	union
	{
		struct
		{
			ControlRef		pane;
			
			struct
			{
				ControlRef		checkBox;
			} options[NUMBER_OF_FILETRANSFERS_OPTIONS]; // access using a kFileTransfersOptionIndex... constant.
		} options;
		
		struct
		{
			ControlRef		pane;
			
			struct
			{
				ControlRef		groupBox,
								typeField,
								creatorField,
								applicationButton;
				StatusIconRef	typeIcon,
								creatorIcon;
			} textFiles;
			
			struct
			{
				ControlRef		groupBox,
								typeField,
								creatorField,
								exampleButton;
				StatusIconRef	typeIcon,
								creatorIcon;
			} binaryFiles;
		} fileTypes;
		
		struct
		{
			ControlRef		pane;
			
			struct
			{
				ControlRef		radioButton;
			} modes[NUMBER_OF_FILETRANSFERS_SERVERMODES]; // access using a kFileTransfersServerModeIndex... constant.
			
			ControlRef		warningIcon,
							anonymousSecurityWarningText,
							userSetupHelpText,
							separator,
							editLogInMessageButton;
		} server;
	} tabs;
};
typedef struct FileTransfersPanelData	FileTransfersPanelData;
typedef FileTransfersPanelData*			FileTransfersPanelDataPtr;

// internal methods
static void				createChosenFileTransfersPane	(PanelRef			inPanel,
														 WindowRef			inOwningWindow);
static void				createPanelControls				(PanelRef			inPanel,
														 WindowRef			inOwningWindow);
static void				deltaPanelSize					(PanelRef			inPanel,
														 const Point		inChangeInWidthAndHeight);
static void				disposeOldTabPane				(PanelRef			inPanel,
														 const UInt16		inTabIndex);
static void				disposePanel					(PanelRef			inPanel,
														 void*				inDataPtr);
static UInt16			getChosenTabIndex				(PanelRef			inPanel);
static SInt32			handleContentClick				(PanelRef			inPanel,
														 ControlRef			inControl);
static void				initPanelSize					(PanelRef			inPanel,
														 const Point		inInitialWidthAndHeight);
static SInt32			panelChanged					(PanelRef			inPanel,
														 PanelMessage		inMessage,
														 void*				inData);
static void				savePreferenceForField			(PanelRef			inPanel,
														 ControlRef			inControl);
static void				setUpChosenFileTransfersPane	(PanelRef			inPanel);
static void				showChosenTabPane				(PanelRef			inPanel);
static Boolean			updateCheckBoxPreference		(PanelRef			inPanel,
														 ControlRef			inCheckBoxClicked);
static void				useOSTypeLengthLimiter			(ControlRef			inControl);



//
// public methods
//

/*!
This routine creates a new preference panel for
the File Transfers category, initializes it, and
returns a reference to it.  You must destroy the
reference using Panel_Dispose() when you are
done with it.

If any problems occur, null is returned.

(3.0)
*/
PanelRef
PrefPanelFileTransfers_New	(void)
{
	PanelRef		result = Panel_New(panelChanged);
	
	
	if (result != null)
	{
		FileTransfersPanelDataPtr	dataPtr = (FileTransfersPanelDataPtr)Memory_NewPtr(sizeof(FileTransfersPanelData));
		Str31						name;
		
		
		Panel_SetDescriptor(result, kTelnetPanelDescriptorPreferencesFileTransfers);
		GetIndString(name, rStringsConfigurationData, siConfigurationDataPreferencePanelNameFileTransfers);
		Panel_SetName(result, name);
		if (FlagManager_Test(kFlagIconServices))
		{
			Panel_SetIconRef(result, kOnSystemDisk, kSystemIconsCreator, kFTPServerIcon);
		}
		else
		{
			Panel_SetIconSuite(result, kIconSuitePrefsFTPServer, kSelectorAllLargeData);
		}
		Panel_SetAuxiliaryDataPtr(result, dataPtr);
	}
	return result;
}// New


//
// internal methods
//

/*!
This routine creates the controls that belong in a
particular tab pane of the File Transfers panel, and
attaches them to the specified owning window in the
proper embedding hierarchy.  The panel descriptor must be
"kTelnetPanelDescriptorPreferencesFileTransfers".

The controls are not arranged or sized.

(3.0)
*/
static void
createChosenFileTransfersPane	(PanelRef			inPanel,
								 WindowRef			inOwningWindow)
{
	ControlRef					container = null,
								control = null;
	Rect						containerRect;
	register SInt16				i = 0;
	FileTransfersPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	Panel_GetContainerControl(inPanel, &container);
	GetControlBounds(container, &containerRect);
	
	if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersOptions)
	{
		// create Options tab pane and controls
		dataPtr->tabs.options.pane = control = GetNewControl(kIDTabUserPane, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabControl);
		{
			Str255		title;
			
			
			for (i = 0; i < NUMBER_OF_FILETRANSFERS_OPTIONS; i++)
			{
				dataPtr->tabs.options.options[i].checkBox = control = GetNewControl(kIDCheckBox, inOwningWindow);
				GetIndString(title, rStringsMiscellaneousControlTitles, siControlTitlePrefsFTPOptionsBase + i);
				SetControlTitle(control, title);
				(OSErr)EmbedControl(control, dataPtr->tabs.options.pane);
			}
			
			// ...initialize checkboxes using user preferences
			SetControlValue(dataPtr->tabs.options.options[kFileTransfersOptionIndexShowLogAtStartup].checkBox,
							BooleanToCheckBoxValue(gFTPServerPrefs->showLog));
			SetControlValue(dataPtr->tabs.options.options[kFileTransfersOptionIndexUseMacBinaryII].checkBox,
							BooleanToCheckBoxValue(gFTPServerPrefs->useMacBinaryII));
			SetControlValue(dataPtr->tabs.options.options[kFileTransfersOptionIndexResetMacBinary].checkBox,
							BooleanToCheckBoxValue(gFTPServerPrefs->resetMacBinary));
			SetControlValue(dataPtr->tabs.options.options[kFileTransfersOptionIndexUseISOTranslation].checkBox,
							BooleanToCheckBoxValue(gFTPServerPrefs->useISOTranslation));
		}
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersFileTypes)
	{
		// create File Types tab pane and controls, initializing using user preferences
		dataPtr->tabs.fileTypes.pane = control = GetNewControl(kIDTabUserPane, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabControl);
		
		dataPtr->tabs.fileTypes.textFiles.groupBox = control = GetNewControl
			(kIDPrefFileTransfersTextFilesGroupBox, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.fileTypes.pane);
		dataPtr->tabs.fileTypes.textFiles.typeField = control = GetNewControl(kIDStaticText, inOwningWindow);
		SetControlOSTypeText(control, 'TEXT');
		(OSErr)EmbedControl(control, dataPtr->tabs.fileTypes.textFiles.groupBox);
		dataPtr->tabs.fileTypes.textFiles.creatorField = control = GetNewControl(kIDEditableText, inOwningWindow);
		SetControlOSTypeText(control, gFTPServerPrefs->textCreator);
		useOSTypeLengthLimiter(control);
		(OSErr)EmbedControl(control, dataPtr->tabs.fileTypes.textFiles.groupBox);
		dataPtr->tabs.fileTypes.textFiles.applicationButton = control = GetNewControl
			(kIDPrefFileTransfersTextFilesFromAppButton, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.fileTypes.textFiles.groupBox);
		
		dataPtr->tabs.fileTypes.binaryFiles.groupBox = control = GetNewControl
			(kIDPrefFileTransfersBinaryFilesGroupBox, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.fileTypes.pane);
		dataPtr->tabs.fileTypes.binaryFiles.typeField = control = GetNewControl(kIDEditableText, inOwningWindow);
		SetControlOSTypeText(control, gFTPServerPrefs->binaryType);
		useOSTypeLengthLimiter(control);
		(OSErr)EmbedControl(control, dataPtr->tabs.fileTypes.binaryFiles.groupBox);
		dataPtr->tabs.fileTypes.binaryFiles.creatorField = control = GetNewControl(kIDEditableText, inOwningWindow);
		SetControlOSTypeText(control, gFTPServerPrefs->binaryCreator);
		useOSTypeLengthLimiter(control);
		(OSErr)EmbedControl(control, dataPtr->tabs.fileTypes.binaryFiles.groupBox);
		dataPtr->tabs.fileTypes.binaryFiles.exampleButton = control = GetNewControl
			(kIDPrefFileTransfersBinaryFilesExampleButton, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.fileTypes.binaryFiles.groupBox);
		
		{
			StatusIconRef		icon = null;
			
			
			// ...create icons to indicate what the text fields are for
			dataPtr->tabs.fileTypes.textFiles.typeIcon = icon = StatusIcon_New(inOwningWindow);
			StatusIcon_AddStage(icon, ResourceFileManager_GetResFile(kTelnetResourceFileApplication), kGenericDocumentIconResource,
								0L/* ticks of delay in animation - N/A */);
			(OSErr)EmbedControl(StatusIcon_GetContainerControl(icon), dataPtr->tabs.fileTypes.textFiles.groupBox);
			
			dataPtr->tabs.fileTypes.textFiles.creatorIcon = icon = StatusIcon_New(inOwningWindow);
			StatusIcon_AddStage(icon, ResourceFileManager_GetResFile(kTelnetResourceFileApplication), kGenericApplicationIconResource,
								0L/* ticks of delay in animation - N/A */);
			(OSErr)EmbedControl(StatusIcon_GetContainerControl(icon), dataPtr->tabs.fileTypes.textFiles.groupBox);
			
			dataPtr->tabs.fileTypes.binaryFiles.typeIcon = icon = StatusIcon_New(inOwningWindow);
			StatusIcon_AddStage(icon, ResourceFileManager_GetResFile(kTelnetResourceFileApplication), kGenericDocumentIconResource,
								0L/* ticks of delay in animation - N/A */);
			(OSErr)EmbedControl(StatusIcon_GetContainerControl(icon), dataPtr->tabs.fileTypes.binaryFiles.groupBox);
			
			dataPtr->tabs.fileTypes.binaryFiles.creatorIcon = icon = StatusIcon_New(inOwningWindow);
			StatusIcon_AddStage(icon, ResourceFileManager_GetResFile(kTelnetResourceFileApplication), kGenericApplicationIconResource,
								0L/* ticks of delay in animation - N/A */);
			(OSErr)EmbedControl(StatusIcon_GetContainerControl(icon), dataPtr->tabs.fileTypes.binaryFiles.groupBox);
		}
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersServer)
	{
		// create Server tab pane and controls - incomplete
		dataPtr->tabs.server.pane = control = GetNewControl(kIDTabUserPane, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabControl);
		{
			Str255		title;
			
			
			for (i = 0; i < NUMBER_OF_FILETRANSFERS_SERVERMODES; i++)
			{
				dataPtr->tabs.server.modes[i].radioButton = control = GetNewControl(kIDRadioButton, inOwningWindow);
				GetIndString(title, rStringsMiscellaneousControlTitles, siControlTitlePrefsFTPModesBase + i);
				SetControlTitle(control, title);
				(OSErr)EmbedControl(control, dataPtr->tabs.server.pane);
			}
			
			// ...initialize radio buttons using user preferences
			SetControlValue(dataPtr->tabs.server.modes[kFileTransfersServerModeIndexOff].radioButton,
							BooleanToRadioButtonValue(gFTPServerPrefs->serverMode == kFTPServerModeOff));
			SetControlValue(dataPtr->tabs.server.modes[kFileTransfersServerModeIndexOnAnonymous].radioButton,
							BooleanToRadioButtonValue(gFTPServerPrefs->serverMode == kFTPServerModeOnNoProtection));
			SetControlValue(dataPtr->tabs.server.modes[kFileTransfersServerModeIndexOnWithUserLogin].radioButton,
							BooleanToRadioButtonValue(gFTPServerPrefs->serverMode == kFTPServerModeOnPasswordProtection));
		}
		dataPtr->tabs.server.warningIcon = control = GetNewControl(kInterfaceLibControlAlertIconCaution, inOwningWindow);
		SizeControl(control, 0, 0); // needed to get around an Appearance bug where the icon gets drawn immediately
		(OSErr)EmbedControl(control, dataPtr->tabs.server.pane);
		dataPtr->tabs.server.anonymousSecurityWarningText = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.server.pane);
		{
			Str255		text;
			
			
			GetIndString(text, rStringsHelpText, siHelpTextPrefsFTPServerAnonymousModeWarning);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSmallSystemFont);
		}
		dataPtr->tabs.server.userSetupHelpText = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.server.pane);
		HideControl(control); // this button is disabled for now, since FTP users are now in the Configurations panel
		dataPtr->tabs.server.separator = control = GetNewControl(kIDSeparator, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.server.pane);
		dataPtr->tabs.server.editLogInMessageButton = control = GetNewControl(kIDPrefFileTransfersServerEditLogInMessageButton, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.server.pane);
	}
}// createChosenFileTransfersPane


/*!
This routine creates the controls that belong in the
File Transfers panel, and attaches them to the
specified owning window in the proper embedding
hierarchy.  The specified panels descriptor must be
"kTelnetPanelDescriptorPreferencesFileTransfers".

The controls are not arranged or sized.

(3.0)
*/
static void
createPanelControls		(PanelRef			inPanel,
						 WindowRef			inOwningWindow)
{
	ControlRef					container = GetNewControl(kIDPanelUserPane, inOwningWindow),
								control = null;
	FileTransfersPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	Panel_SetContainerControl(inPanel, container);
	SetControlVisibility(container, false/* visible */, false/* draw */);
	
	dataPtr->tabControl = control = GetNewControl(kIDPrefFileTransfersTabs, inOwningWindow);
	(OSErr)EmbedControl(control, container);
	
	createChosenFileTransfersPane(inPanel, inOwningWindow);
}// createPanelControls


/*!
This routine adjusts the controls in preference panels
to match the specified change in dimensions of a panels
container.

(3.0)
*/
static void
deltaPanelSize		(PanelRef			inPanel,
					 const Point		inChangeInWidthAndHeight)
{
	FileTransfersPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	ControlRef					control = null;
	Rect						controlRect;
	register SInt16				i = 0;
	
	
	control = dataPtr->tabControl;
	GetControlBounds(control, &controlRect);
	SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
	
	if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersOptions)
	{
		// resize the Options tab pane, and its controls
		control = dataPtr->tabs.options.pane;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
		for (i = 0; i < NUMBER_OF_FILETRANSFERS_OPTIONS; i++)
		{
			control = dataPtr->tabs.options.options[i].checkBox;
			GetControlBounds(control, &controlRect);
			MoveControl(control, controlRect.left, controlRect.top + i * (inChangeInWidthAndHeight.v / NUMBER_OF_FILETRANSFERS_OPTIONS));
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
		}
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersFileTypes)
	{
		// resize the File Types tab pane, and its controls
		control = dataPtr->tabs.fileTypes.pane;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
		control = dataPtr->tabs.fileTypes.textFiles.groupBox;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top);
		control = dataPtr->tabs.fileTypes.binaryFiles.groupBox;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top);
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersServer)
	{
		// resize the Server tab pane, and its controls - incomplete
		control = dataPtr->tabs.server.pane;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
		for (i = 0; i < NUMBER_OF_FILETRANSFERS_SERVERMODES; i++)
		{
			control = dataPtr->tabs.server.modes[i].radioButton;
			GetControlBounds(control, &controlRect);
			MoveControl(control, controlRect.left, controlRect.top + i * (inChangeInWidthAndHeight.v / NUMBER_OF_FILETRANSFERS_SERVERMODES));
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
			if (i == kFileTransfersServerModeIndexOnAnonymous)
			{
				control = dataPtr->tabs.server.warningIcon;
				GetControlBounds(control, &controlRect);
				MoveControl(control, controlRect.left,
									controlRect.top + i * (inChangeInWidthAndHeight.v / NUMBER_OF_FILETRANSFERS_SERVERMODES));
				control = dataPtr->tabs.server.anonymousSecurityWarningText;
				GetControlBounds(control, &controlRect);
				MoveControl(control, controlRect.left,
									controlRect.top + i * (inChangeInWidthAndHeight.v / NUMBER_OF_FILETRANSFERS_SERVERMODES));
			}
		}
		control = dataPtr->tabs.server.userSetupHelpText;
		GetControlBounds(control, &controlRect);
		MoveControl(control, controlRect.left, controlRect.top + inChangeInWidthAndHeight.v);
		control = dataPtr->tabs.server.separator;
		GetControlBounds(control, &controlRect);
		MoveControl(control, controlRect.left, controlRect.top + inChangeInWidthAndHeight.v);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
		control = dataPtr->tabs.server.editLogInMessageButton;
		GetControlBounds(control, &controlRect);
		MoveControl(control, controlRect.left, controlRect.top + inChangeInWidthAndHeight.v);
	}
}// deltaPanelSize


/*!
Use this method to hide all tab panes for the
File Transfers panel except for the pane
corresponding to the active tab in that panel.

(3.0)
*/
static void
disposeOldTabPane		(PanelRef		inPanel,
						 const UInt16	inTabIndex)
{
	FileTransfersPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	ControlRef*					controlHandlePtr = null;
	
	
	switch (inTabIndex)
	{
		case kTabIndexFileTransfersOptions:
			controlHandlePtr = &dataPtr->tabs.options.pane;
			break;
		
		case kTabIndexFileTransfersFileTypes:
			controlHandlePtr = &dataPtr->tabs.fileTypes.pane;
			break;
		
		case kTabIndexFileTransfersServer:
			controlHandlePtr = &dataPtr->tabs.server.pane;
			break;
		
		default:
			// ???
			break;
	}
	
	if (controlHandlePtr != null)
	{
		(OSErr)SetControlVisibility(*controlHandlePtr, false/* visible */, true/* draw */);
		DisposeControl(*controlHandlePtr), *controlHandlePtr = null;
	}
}// disposeOldTabPane


/*!
This routine responds to a change in the existence
of a panel.  The panel is about to be destroyed, so
this routine disposes of private data structures
associated with the specified panel.

(3.0)
*/
static void
disposePanel	(PanelRef			UNUSED_ARGUMENT(inPanel),
				 void*				inDataPtr)
{
	FileTransfersPanelDataPtr		dataPtr = inDataPtr;
	
	
	Memory_DisposePtr((Ptr*)&dataPtr);
}// disposePanel


/*!
Use this method to determine the one-based
tab index of the selected tab for a panel.

(3.0)
*/
static UInt16
getChosenTabIndex	(PanelRef		inPanel)
{
	UInt16						result = 0;
	FileTransfersPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	if (dataPtr != null) result = GetControlValue(dataPtr->tabControl);
	return result;
}// getChosenTabIndex


/*!
This method responds to control clicks.

(3.0)
*/
static SInt32
handleContentClick	(PanelRef		inPanel,
					 ControlRef		inControl)
{
	SInt32						result = kPanel_ResponseClickNotHandled;
	Boolean						handled = false;
	FileTransfersPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	if (inControl == dataPtr->tabControl)
	{
		showChosenTabPane(inPanel);
		handled = true;
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersFileTypes)
	{
		if (inControl == dataPtr->tabs.fileTypes.textFiles.applicationButton)
		{
			OSType		creatorCode = '????';
			
			
			if (GetApplicationType(&creatorCode))
			{
				ControlRef		field = dataPtr->tabs.fileTypes.textFiles.creatorField;
				
				
				SetControlOSTypeText(field, creatorCode);
				DrawOneControl(field);
				savePreferenceForField(inPanel, field);
			}
			handled = true;
		}
		else if (inControl == dataPtr->tabs.fileTypes.binaryFiles.exampleButton)
		{
			FileInfo		fi;
			Boolean			good;
			Str255			prompt,
							title;
			
			
			GetIndString(prompt, rStringsNavigationServices, siNavPromptSelectBinaryFile);
			GetIndString(title, rStringsNavigationServices, siNavDialogTitleExampleBinaryApplication);
			if (FlagManager_Test(kFlagNavigationServices))
			{
				FSSpec		file;
				OSType		fileType;
				OSErr		error = noErr;
				short		hit = 0;
				
				
				Alert_ReportOSErr(error = Navigation_GetFile(prompt, title,
															kApplicationCreatorSignature,
															kTelnetNavPrefKeyGenericOpenFile,
															kNavDefaultNavDlogOptions |
															kNavDontAddTranslateItems |
															kNavNoTypePopup,
															0, 0L, EventLoop_HandleNavigationUpdate,
															&file, &fileType, &fi));
				good = (error == noErr);
			}
			else
			{
				StandardFileReply		sfr;
				
				
				GetIndString(prompt, rStringsNavigationServices, siNavPromptSelectBinaryFile);
				ThemeSavvyStandardGetFile(title, prompt, (FileFilterUPP)null, -1, (ConstSFTypeListPtr)null, &sfr);
				
				good = sfr.sfGood;
				
				if (good) HGetFInfo(sfr.sfFile.vRefNum, sfr.sfFile.parID, sfr.sfFile.name, (FInfo*)&fi);
			}
			
			if (good)
			{
				ControlRef		field = null;
				
				
				field = dataPtr->tabs.fileTypes.binaryFiles.typeField;
				SetControlOSTypeText(field, fi.fileType);
				DrawOneControl(field);
				savePreferenceForField(inPanel, field);
				
				field = dataPtr->tabs.fileTypes.binaryFiles.creatorField;
				SetControlOSTypeText(field, fi.fileCreator);
				DrawOneControl(field);
				savePreferenceForField(inPanel, field);
			}
			
			handled = true;
		}
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersServer)
	{
		if (inControl == dataPtr->tabs.server.editLogInMessageButton)
		{
			// display a dialog to edit the log-in message
			EditLogInMessageDialog_Init();
			EditLogInMessageDialog_Display();
			EditLogInMessageDialog_Done();
			handled = true;
		}
		else
		{
			handled = updateCheckBoxPreference(inPanel, inControl);
		}
	}
	else handled = updateCheckBoxPreference(inPanel, inControl);
	
	if (handled) result = kPanel_ResponseClickHandled;
	
	return result;
}// handleContentClick


/*!
This routine sizes and arranges controls in preference
panels to fit the prescribed dimensions of a panels
container.

(3.0)
*/
static void
initPanelSize		(PanelRef			inPanel,
					 const Point		UNUSED_ARGUMENT(inInitialWidthAndHeight))
{
	setUpChosenFileTransfersPane(inPanel);
}// initPanelSize


/*!
This routine, of standard PanelChangedProcPtr form,
is invoked by the Panel module whenever a property
of one of the preferences dialogs panels changes.

(3.0)
*/
static SInt32
panelChanged	(PanelRef			inPanel,
				 PanelMessage		inMessage,
				 void*				inData)
{
	SInt32		result = 0L;
	
	
	if (Panel_GetDescriptor(inPanel) == kTelnetPanelDescriptorPreferencesFileTransfers) switch (inMessage)
	{
		case kPanel_MessageClicked: // notification of a content click
			{
				ControlRef*		dataPtr = inData;
				
				
				result = handleContentClick(inPanel, *dataPtr);
			}
			break;
		
		case kPanel_MessageCreateControls: // specification of the window containing the panel - create controls using this window
			{
				WindowRef	window = *((WindowRef*)inData);
				
				
				createPanelControls(inPanel, window);
			}
			break;
		
		case kPanel_MessageDeltaSize: // request to alter size and/or position of controls; containing control will already have been changed
			{
				Point			changeInPanelDimensions;
				ControlRef		panelControl = null;
				Boolean			wasVisible = false;
				
				
				SetPt(&changeInPanelDimensions, ((Point*)inData)->h, ((Point*)inData)->v);
				Panel_GetContainerControl(inPanel, &panelControl);
				wasVisible = IsControlVisible(panelControl);
				SetControlVisibility(panelControl, false/* visibility */, wasVisible/* draw */);
				deltaPanelSize(inPanel, changeInPanelDimensions);
				SetControlVisibility(panelControl, wasVisible/* visibility */, wasVisible/* draw */);
			}
			break;
		
		case kPanel_MessageDestroyed: // request to dispose of private data structures
			{
				void*		dataPtr = inData;
				
				
				disposePanel(inPanel, dataPtr);
			}
			break;
		
		case kPanel_MessageFocusGained: // notification that a control is now focused
			{
				ControlRef*		dataPtr = inData;
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageFocusLost: // notification that a control is no longer focused
			{
				ControlRef*		dataPtr = inData;
				
				
				savePreferenceForField(inPanel, *dataPtr);
			}
			break;
		
		case kPanel_MessageInitialSize: // request for panel to size its controls
			{
				Point		exactPanelDimensions;
				
				
				SetPt(&exactPanelDimensions, ((Point*)inData)->h, ((Point*)inData)->v);
				initPanelSize(inPanel, exactPanelDimensions);
			}
			break;
		
		case kPanel_MessageNewActiveState: // request to change panel controls to a new active (true) or inactive (false) state
			{
				Boolean		isNowActive = *((Boolean*)inData);
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageNewAppearanceTheme: // notification of theme switch, a request to recalculate control sizes
			{
				// this notification is currently ignored, but shouldnt be...
			}
			break;
		
		case kPanel_MessageNewVisibility: // visible state of the panels container has changed to visible (true) or invisible (false)
			{
				Boolean		isNowVisible = *((Boolean*)inData);
				
				
				// do nothing
			}
			break;
		
		default:
			break;
	}
	
	return result;
}// panelChanged


/*!
To update the preferences in memory to reflect
the information in the specified text field,
use this method.  This should be done if the
focus of the window changes (by hitting the
tab key or clicking in another text field).

(3.0)
*/
static void
savePreferenceForField		(PanelRef			inPanel,
							 ControlRef			inControl)
{
	FileTransfersPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	if ((inControl != null) && (dataPtr != null))
	{
		if (inControl == dataPtr->tabs.fileTypes.textFiles.creatorField)
		{
			OSType		creator = '????';
			
			
			GetControlOSTypeText(inControl, &creator);
			Preferences_SetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_TextFileCreator, sizeof(OSType), &creator);
		}
		else if (inControl == dataPtr->tabs.fileTypes.binaryFiles.typeField)
		{
			OSType		fileType = '????';
			
			
			GetControlOSTypeText(inControl, &fileType);
			Preferences_SetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_BinaryFileType, sizeof(OSType), &fileType);
		}
		else if (inControl == dataPtr->tabs.fileTypes.binaryFiles.creatorField)
		{
			OSType		creator = '????';
			
			
			GetControlOSTypeText(inControl, &creator);
			Preferences_SetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_BinaryFileCreator, sizeof(OSType), &creator);
		}
	}
}// savePreferenceForField


/*!
This routine sizes and arranges controls in the
currently-selected tab pane of the specified
File Transfers panel to fit the prescribed
dimensions of the panels container.

(3.0)
*/
static void
setUpChosenFileTransfersPane	(PanelRef		inPanel)
{
	enum
	{
		kRadioButtonGlyphWidth = 16
	};
	FileTransfersPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	ControlRef					container = null,
								control = null;
	Rect						containerRect;
	Point						tabPaneTopLeft,
								tabPaneWidthHeight;
	register SInt16				i = 0;
	UInt16						leftEdge = 0, // cumulative, resettable measurement allowing controls to sandwich horizontally
								topEdge = 0; // cumulative, resettable measurement allowing controls to stack vertically
	SInt32						checkBoxHeight = 0L,
								radioButtonHeight = 0L;
	
	
	Panel_GetContainerControl(inPanel, &container);
	GetControlBounds(container, &containerRect);
	
	PreferencesDialog_CalculateTabPane(inPanel, &tabPaneTopLeft, &tabPaneWidthHeight);
	
	//if (GetThemeMetric(kThemeMetricCheckBoxGlyphHeight, &checkBoxHeight) != noErr) checkBoxHeight = CHECKBOX_HT;
	//if (GetThemeMetric(kThemeMetricRadioButtonGlyphHeight, &radioButtonHeight) != noErr) radioButtonHeight = RADIOBUTTON_HT;
	checkBoxHeight = CHECKBOX_HT;
	radioButtonHeight = RADIOBUTTON_HT;
	
	if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersOptions)
	{
		// size Options tab pane and its controls
		control = dataPtr->tabs.options.pane;
		SizeControl(control, tabPaneWidthHeight.h, tabPaneWidthHeight.v);
		for (i = 0; i < NUMBER_OF_FILETRANSFERS_OPTIONS; i++)
		{
			control = dataPtr->tabs.options.options[i].checkBox;
			SizeControl(control, tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), checkBoxHeight);
		}
		
		// arrange Options tab pane and its controls
		topEdge = VSP_PANE_AND_CONTROL;
		for (i = 0; i < NUMBER_OF_FILETRANSFERS_OPTIONS; i++)
		{
			control = dataPtr->tabs.options.options[i].checkBox;
			MoveControl(control, HSP_PANE_AND_CONTROL, topEdge);
			topEdge += (checkBoxHeight + VSP_CHECKBOXES);
		}
		
		control = dataPtr->tabs.options.pane;
		MoveControl(control, tabPaneTopLeft.h, tabPaneTopLeft.v);
		(OSErr)SetControlVisibility(control, true/* visible */, true/* draw */);
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersFileTypes)
	{
		UInt16		boxHeight = (VSP_GROUPBOXTEXT_AND_CONTROL + SPACE_EXTRA_TEXTEDIT + TEXT_HT_BIG + VSP_EDITABLETEXTAREAS +
									TEXT_HT_BIG + SPACE_EXTRA_TEXTEDIT + VSP_GROUPBOX_AND_CONTROL);
		
		
		// size File Types tab pane and its controls
		SizeControl(dataPtr->tabs.fileTypes.textFiles.groupBox,
					tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), boxHeight);
		SizeControl(dataPtr->tabs.fileTypes.textFiles.typeField, kTypeCreatorFieldWidth, TEXT_HT_BIG);
		SizeControl(dataPtr->tabs.fileTypes.textFiles.creatorField, kTypeCreatorFieldWidth, TEXT_HT_BIG);
		(UInt16)Localization_AutoSizeButtonControl(dataPtr->tabs.fileTypes.textFiles.applicationButton, 0);
		SizeControl(dataPtr->tabs.fileTypes.binaryFiles.groupBox,
					tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), boxHeight);
		SizeControl(dataPtr->tabs.fileTypes.binaryFiles.typeField, kTypeCreatorFieldWidth, TEXT_HT_BIG);
		SizeControl(dataPtr->tabs.fileTypes.binaryFiles.creatorField, kTypeCreatorFieldWidth, TEXT_HT_BIG);
		(UInt16)Localization_AutoSizeButtonControl(dataPtr->tabs.fileTypes.binaryFiles.exampleButton, 0);
		SizeControl(dataPtr->tabs.fileTypes.pane, tabPaneWidthHeight.h, tabPaneWidthHeight.v);
		
		// arrange File Types tab pane and its controls
		leftEdge = HSP_GROUPBOX_AND_CONTROL;
		topEdge = VSP_GROUPBOXTEXT_AND_CONTROL + SPACE_EXTRA_TEXTEDIT;
		MoveControl(StatusIcon_GetContainerControl(dataPtr->tabs.fileTypes.textFiles.typeIcon), leftEdge, topEdge);
		leftEdge += (ICON_WD_SMALL + HSP_CONTROLS);
		MoveControl(dataPtr->tabs.fileTypes.textFiles.typeField, leftEdge, topEdge);
		leftEdge = HSP_GROUPBOX_AND_CONTROL;
		topEdge += (TEXT_HT_BIG + VSP_EDITABLETEXTAREAS);
		MoveControl(StatusIcon_GetContainerControl(dataPtr->tabs.fileTypes.textFiles.creatorIcon), leftEdge, topEdge);
		leftEdge += (ICON_WD_SMALL + HSP_CONTROLS);
		MoveControl(dataPtr->tabs.fileTypes.textFiles.creatorField, leftEdge, topEdge);
		leftEdge += (kTypeCreatorFieldWidth + HSP_CONTROLS);
		topEdge -= INTEGER_HALVED(BUTTON_HT - TEXT_HT_BIG);
		MoveControl(dataPtr->tabs.fileTypes.textFiles.applicationButton, leftEdge, topEdge);
		MoveControl(dataPtr->tabs.fileTypes.textFiles.groupBox, HSP_PANE_AND_CONTROL, VSP_PANE_AND_CONTROL);
		
		leftEdge = HSP_GROUPBOX_AND_CONTROL;
		topEdge = VSP_GROUPBOXTEXT_AND_CONTROL + SPACE_EXTRA_TEXTEDIT;
		MoveControl(StatusIcon_GetContainerControl(dataPtr->tabs.fileTypes.binaryFiles.typeIcon), leftEdge, topEdge);
		leftEdge += (ICON_WD_SMALL + HSP_CONTROLS);
		MoveControl(dataPtr->tabs.fileTypes.binaryFiles.typeField, leftEdge, topEdge);
		leftEdge += (kTypeCreatorFieldWidth + HSP_CONTROLS);
		MoveControl(dataPtr->tabs.fileTypes.binaryFiles.exampleButton, leftEdge, topEdge - INTEGER_HALVED(BUTTON_HT - TEXT_HT_BIG));
		leftEdge = HSP_GROUPBOX_AND_CONTROL;
		topEdge += (TEXT_HT_BIG + VSP_EDITABLETEXTAREAS);
		MoveControl(StatusIcon_GetContainerControl(dataPtr->tabs.fileTypes.binaryFiles.creatorIcon), leftEdge, topEdge);
		leftEdge += (ICON_WD_SMALL + HSP_CONTROLS);
		MoveControl(dataPtr->tabs.fileTypes.binaryFiles.creatorField, leftEdge, topEdge);
		MoveControl(dataPtr->tabs.fileTypes.binaryFiles.groupBox, HSP_PANE_AND_CONTROL,
					VSP_PANE_AND_CONTROL + boxHeight + VSP_GROUPBOXES);
		
		control = dataPtr->tabs.fileTypes.pane;
		MoveControl(control, tabPaneTopLeft.h, tabPaneTopLeft.v);
		(OSErr)SetControlVisibility(control, true/* visible */, true/* draw */);
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexFileTransfersServer)
	{
		// size Server tab pane and its controls
		for (i = 0; i < NUMBER_OF_FILETRANSFERS_SERVERMODES; i++)
		{
			control = dataPtr->tabs.server.modes[i].radioButton;
			SizeControl(control, tabPaneWidthHeight.h - INTEGER_DOUBLED((HSP_PANE_AND_CONTROL)), radioButtonHeight);
		}
		control = dataPtr->tabs.server.warningIcon;
		SizeControl(control, ICON_WD_SMALL, ICON_HT_SMALL); // the icon is normally 32x32; scale it down
		control = dataPtr->tabs.server.userSetupHelpText;
		{
			Str255		string;
			UInt16		textHeight = 0;
			
			
			GetIndString(string, rStringsHelpText, siHelpTextPrefsFTPServerHowToDefineUsers);
			SizeControl(control, tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), 0);
			(UInt16)Localization_SetUpMultiLineTextControl(control, string, kThemeSmallSystemFont);
		}
		control = dataPtr->tabs.server.separator;
		SizeControl(control, tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), SEPARATOR_HT);
		control = dataPtr->tabs.server.pane;
		SizeControl(control, tabPaneWidthHeight.h, tabPaneWidthHeight.v);
		
		// arrange Server tab pane and its controls
		topEdge = VSP_PANE_AND_CONTROL;
		for (i = 0; i < NUMBER_OF_FILETRANSFERS_SERVERMODES; i++)
		{
			control = dataPtr->tabs.server.modes[i].radioButton;
			MoveControl(control, HSP_PANE_AND_CONTROL, topEdge);
			topEdge += radioButtonHeight;
			if (i == kFileTransfersServerModeIndexOnAnonymous)
			{
				// special arrangement for the warning icon and message underneath this radio button
				control = dataPtr->tabs.server.warningIcon;
				MoveControl(control, HSP_PANE_AND_CONTROL + kRadioButtonGlyphWidth, topEdge);
				control = dataPtr->tabs.server.anonymousSecurityWarningText;
				MoveControl(control, HSP_PANE_AND_CONTROL + kRadioButtonGlyphWidth + ICON_WD_SMALL + HSP_ICON_AND_TEXT,
									topEdge + INTEGER_HALVED(ICON_HT_SMALL - TEXT_HT_SMALL) + 1);
				topEdge += ICON_HT_SMALL;
			}
			topEdge += VSP_RADIOS;
		}
		control = dataPtr->tabs.server.userSetupHelpText;
		MoveControl(control, HSP_PANE_AND_CONTROL + kRadioButtonGlyphWidth, topEdge);
		{
			Rect		bounds;
			
			
			GetControlBounds(control, &bounds);
			topEdge += (bounds.bottom - bounds.top + VSP_CONTROLS);
		}
		
		control = dataPtr->tabs.server.separator;
		MoveControl(control, HSP_PANE_AND_CONTROL, topEdge);
		topEdge += (SEPARATOR_HT + VSP_GROUPBOX_AND_CONTROL);
		
		control = dataPtr->tabs.server.editLogInMessageButton;
		(UInt16)Localization_AutoSizeButtonControl(control, 0);
		MoveControl(control, HSP_PANE_AND_CONTROL, topEdge);
		topEdge += (BUTTON_HT + VSP_BUTTON_AND_CONTROL);
		
		control = dataPtr->tabs.server.pane;
		MoveControl(control, tabPaneTopLeft.h, tabPaneTopLeft.v);
		(OSErr)SetControlVisibility(control, true/* visible */, true/* draw */);
	}
}// setUpChosenFileTransfersPane


/*!
Use this method to create the controls for
the selected tab of the specified panel.
If the given panel has no tabs, this method
has no effect.

WARNING:	The Preferences dialog box
			disposes of controls that are
			part of invisible tab panes!
			Never reference a control until
			you first use getChosenTabIndex()
			to verify that the current tab
			contains the control you wish to
			modify.  If you attempt to access
			a control on a disposed pane, the
			system will crash.

(3.0)
*/
static void
showChosenTabPane		(PanelRef		inPanel)
{
	UInt16				chosenTab = 0;
	static UInt16		oldChoice = 1;
	
	
	chosenTab = getChosenTabIndex(inPanel);
	if (oldChoice != chosenTab)
	{
		disposeOldTabPane(inPanel, oldChoice);
		oldChoice = chosenTab;
		createChosenFileTransfersPane(inPanel, Panel_GetOwningWindow(inPanel));
		setUpChosenFileTransfersPane(inPanel);
	}
}// showChosenTabPane


/*!
Call this method when a click in a checkbox
or radio button has been detected, but the
control value has *not* yet been toggled.
Upon exit from this routine, the checkbox
state will have been toggled (or, for a
radio button, the group value will change).
In addition, the appropriate preference
will have been updated in memory using the
Preferences_SetData() method.

If the specified control is known, "true"
is returned to indicate that the click was
handled.  Otherwise, "false" is returned.

(3.0)
*/
static Boolean
updateCheckBoxPreference	(PanelRef			inPanel,
							 ControlRef			inCheckBoxClicked)
{
	Boolean						result = false;
	FileTransfersPanelDataPtr	dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	if (dataPtr != null)
	{
		Boolean				checkBoxFlagValue = // IMPORTANT - since the control value hasnt changed yet, the flag assignment is reversed!
								(GetControlValue(inCheckBoxClicked) == kControlCheckBoxCheckedValue) ? false : true;
		register SInt16		i = 0;
		SInt16				chosenOptionIndex = -1;
		UInt16				chosenTabIndex = getChosenTabIndex(inPanel);
		
		
		switch (chosenTabIndex)
		{
			case kTabIndexFileTransfersOptions:
				// determine which Options checkbox was hit
				for (i = 0; i < NUMBER_OF_FILETRANSFERS_OPTIONS; i++)
				{
					if (inCheckBoxClicked == dataPtr->tabs.options.options[i].checkBox)
					{
						chosenOptionIndex = i;
						result = true;
						break;
					}
				}
				// ...NOW toggle the control value...
				SetControlValue(inCheckBoxClicked, BooleanToCheckBoxValue(checkBoxFlagValue));
				
				switch (chosenOptionIndex)
				{
					case kFileTransfersOptionIndexShowLogAtStartup:
						Preferences_SetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_LogInitiallyVisible, sizeof(Boolean),
											&checkBoxFlagValue);
						break;
					
					case kFileTransfersOptionIndexUseMacBinaryII:
						Preferences_SetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_UseMacBinaryII, sizeof(Boolean),
											&checkBoxFlagValue);
						break;
					
					case kFileTransfersOptionIndexResetMacBinary:
						Preferences_SetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_ResetMacBinary, sizeof(Boolean),
											&checkBoxFlagValue);
						break;
					
					case kFileTransfersOptionIndexUseISOTranslation:
						Preferences_SetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_ISOTranslation, sizeof(Boolean),
											&checkBoxFlagValue);
						break;
					
					default:
						// unrecognized checkbox
						break;
				}
				break;
			
			case kTabIndexFileTransfersFileTypes:
				// there are no checkboxes on this tab
				break;
			
			case kTabIndexFileTransfersServer:
				// see if any Server Mode radio buttons were hit
				for (i = 0; i < NUMBER_OF_FILETRANSFERS_SERVERMODES; i++)
				{
					// is this radio button the one that was clicked?
					if (inCheckBoxClicked == dataPtr->tabs.server.modes[i].radioButton)
					{
						chosenOptionIndex = i;
						result = true;
					}
				}
				
				// were any of the radio buttons hit?
				if (chosenOptionIndex >= 0)
				{
					FTPServerMode		oldMode = gFTPServerPrefs->serverMode,
										newMode = kFTPServerModeOff;
					
					
					// set the radio button
					for (i = 0; i < NUMBER_OF_FILETRANSFERS_SERVERMODES; i++)
					{
						SetControlValue(dataPtr->tabs.server.modes[i].radioButton, kControlRadioButtonUncheckedValue);
					}
					SetControlValue(inCheckBoxClicked, kControlRadioButtonCheckedValue);
					
					switch (chosenOptionIndex)
					{
						case kFileTransfersServerModeIndexOff:
							newMode = kFTPServerModeOff;
							break;
						
						case kFileTransfersServerModeIndexOnAnonymous:
							newMode = kFTPServerModeOnNoProtection;
							break;
						
						case kFileTransfersServerModeIndexOnWithUserLogin:
							newMode = kFTPServerModeOnPasswordProtection;
							break;
						
						default:
							// unrecognized radio button
							break;
					}
					
					if (newMode != oldMode)
					{
						Console_WriteLine("registering new FTP server mode");
						Preferences_SetData(kTelnetPreferenceClassFTP, kTelnetPreferenceTag_ServerMode, sizeof(FTPServerMode), &newMode);
						
						// notify the server software of the new mode (this may briefly display a progress dialog)
						// ...should this be done directly in "Preferences.c" whenever this preference is changed???
						NewFTPServerMode(oldMode, newMode);
						Console_WriteLine("done registering new FTP server mode");
					}
				}
				break;
			
			default:
				break;
		}
	}
	
	return result;
}// updateCheckBoxPreference


/*!
If an editable text control is to represent an
OSType (which is four characters), use this
method to assign a key filter to it that prevents
more than four characters from being typed in
such a field.

(3.0)
*/
static void
useOSTypeLengthLimiter	(ControlRef		inControl)
{
	ControlKeyFilterUPP		upp = NewControlKeyFilterUPP(OSTypeLengthLimiter);
	
	
	(OSErr)SetControlData(inControl, kControlNoPart, kControlKeyFilterTag, sizeof(ControlKeyFilterUPP), (Ptr)&upp);
}// useOSTypeLengthLimiter