/*###############################################################

	PrefPanelGeneral.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Dialogs.h>
#include <Folders.h>
#include <Icons.h>
#include <Resources.h>

// library includes
#include "ContextsLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "StringResources.h"
#include "SpacingConstants.r"
#include "TelnetPrefsDialog.r"

// MacTelnet includes
#include "ColorBox.h"
#include "configure.proto.h"
#include "DialogUtilities.h"
#include "MacLook.h"
#include "MacroSetupDialog.h"
#include "Panel.h"
#include "Preferences.h"
#include "PreferencesData.h"
#include "PreferencesDialog.h"
#include "PrefPanelGeneral.h"
#include "RealScreen.h"
#include "ResourceFileManager.h"
#include "EventLoop.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



// constants
enum
{
	kTypeCreatorFieldWidth = 60
};
#define NUMBER_OF_GENERAL_TABPANES			4
enum
{
	// must reflect 'tab#' information and be one-based
	kTabIndexGeneralOptions = 1,
	kTabIndexGeneralSpecial = 2,
	kTabIndexGeneralColors = 3,
	kTabIndexGeneralNotification = 4
};
#define NUMBER_OF_GENERAL_OPTIONS			10
enum
{
	// must be consecutive and zero-based
	kGeneralOptionIndexSimpleTelnet = 0,
	kGeneralOptionIndexDontAutoClose = 1,
	kGeneralOptionIndexDynamicResizing = 2,
	kGeneralOptionIndexHeadersInitiallyCollapsed = 3,
	kGeneralOptionIndexDontDimBackgroundScreens = 4,
	kGeneralOptionIndexPureInverse = 5,
	kGeneralOptionIndexCopySelectedText = 6,
	kGeneralOptionIndexCursorMovesToDropLocation = 7,
	kGeneralOptionIndexMenuItemKeys = 8,
	kGeneralOptionIndexMapBackquote = 9
};
#define NUMBER_OF_GENERAL_NOTIFICATION_METHODS	4
enum
{
	// must be consecutive and zero-based
	kGeneralNotificationMethodIndexNone = 0,
	kGeneralNotificationMethodIndexDiamond = 1,
	kGeneralNotificationMethodIndexFlashIcon = 2,
	kGeneralNotificationMethodIndexDisplayAlert = 3
};

// types
struct GeneralPanelData
{
	ControlRef					tabControl;
	ControlUserPaneDrawUPP		iconDrawingUPP;
	
	union
	{
		// Options tab controls
		struct
		{
			ControlRef		pane;
			
			struct
			{
				ControlRef		checkBox;
			} options[NUMBER_OF_GENERAL_OPTIONS]; // access using a kGeneralOptionIndex... constant.
			
		} options;
		
		// Special tab controls
		struct
		{
			ControlRef		pane;
			
			struct
			{
				ControlRef	groupBox,
							shapeLabel,
							shapeBlockBevelButton,
							shapeVerticalBarBevelButton,
							shapeThickVerticalBarBevelButton,
							shapeUnderlineBevelButton,
							shapeThickUnderlineBevelButton,
							flashingCheckBox;
			} cursor;
			
			struct
			{
				ControlRef	groupBox,
							openLabel,
							openField,
							sendLabel,
							sendField;
			} timeouts;
			
			struct
			{
				ControlRef	groupBox,
							thresholdLabel,
							thresholdField,
							thresholdUnitsLabel;
			} copyTable;
			
			struct
			{
				ControlRef	groupBox,
							creatorField,
							applicationButton;
			} captureFile;
			
		} special;
		
		// Colors tab controls
		struct
		{
			ControlRef		pane,
							headerLabel,
							normalHeaderLabel,
							emphasizedHeaderLabel,
							normalHeaderDivider,
							emphasizedHeaderDivider,
							setToDefaultsButton;
			ControlRef		colorBoxes[kANSIColorCount];
			ControlRef		colorBoxLabels[kANSIColorCount];
		} colors;
		
		// Notification tab controls
		struct
		{
			ControlRef		pane,
							visualBellCheckBox;
			
			struct
			{
				ControlRef	groupBox,
							terminalBeepsCheckBox,
							separator,
							helpText,
							whenPostingText;
				
				struct
				{
					ControlRef		symbolsPane,
									radioButton;
				} methods[NUMBER_OF_GENERAL_NOTIFICATION_METHODS]; // access using a kGeneralNotificationMethodIndex... constant.
			} background;
			
		} notification;
		
	} tabs;
};
typedef struct GeneralPanelData		GeneralPanelData;
typedef GeneralPanelData*			GeneralPanelDataPtr;

// internal methods
static void				changePreferenceUpdateScreen	(const ConnectionDataPtr	inConnectionDataPtr,
														 const RealScreenDataPtr	inRealScreenDataPtr,
														 const SInt32				inDataPreferenceTag,
														 const SInt32				inUnusedData2,
														 void*						inoutUndefinedResultPtr);
static void				colorBoxChangeNotify			(ControlRef					inColorBoxThatChanged,
														 RGBColor const*			inNewColor,
														 void const*				inContextPtr);
static void				createChosenGeneralPane			(PanelRef					inPanel,
														 WindowRef					inOwningWindow);
static void				createPanelControls				(PanelRef					inPanel,
														 WindowRef					inOwningWindow);
static void				deltaPanelSize					(PanelRef					inPanel,
														 const Point				inChangeInWidthAndHeight);
static void				disposeOldTabPane				(PanelRef					inPanel,
														 const UInt16				inTabIndex);
static void				disposePanel					(PanelRef					inPanel,
														 void*						inDataPtr);
static UInt16			getChosenTabIndex				(PanelRef					inPanel);
static SInt32			handleContentClick				(PanelRef					inPanel,
														 ControlRef					inControl);
static pascal void		notificationPictureDeviceLoop	(short						inColorDepth,
														 short						inDeviceFlags,
														 GDHandle					inTargetDevice,
														 long						inControl);
static pascal void		notificationPictureDraw			(ControlRef					inControl,
														 SInt16						inPart);
static SInt32			panelChanged					(PanelRef					inPanel,
														 PanelMessage				inMessage,
														 void*						inData);
static void				savePreferenceForField			(PanelRef					inPanel,
														 ControlRef					inControl);
static void				setUpChosenGeneralPane			(PanelRef					inPanel);
static void				showChosenTabPane				(PanelRef					inPanel);
static Boolean			updateCheckBoxPreference		(PanelRef					inPanel,
														 ControlRef					inCheckBoxClicked);
static void				useNumericalLimiter				(ControlRef					inControl);
static void				useOSTypeLengthLimiter			(ControlRef					inControl);



//
// public methods
//

/*!
Creates a new preference panel for the General
category, initializes it, and returns a reference
to it.  You must destroy the reference using
Panel_Dispose() when you are done with it.

If any problems occur, null is returned.

(3.0)
*/
PanelRef
PrefPanelGeneral_New	(void)
{
	PanelRef		result = Panel_New(panelChanged);
	
	
	if (result != null)
	{
		GeneralPanelDataPtr		dataPtr = (GeneralPanelDataPtr)Memory_NewPtr(sizeof(GeneralPanelData));
		Str31					name;
		
		
		Panel_SetDescriptor(result, kTelnetPanelDescriptorPreferencesGeneral);
		GetIndString(name, rStringsConfigurationData, siConfigurationDataPreferencePanelNameGeneral);
		Panel_SetName(result, name);
		Panel_SetIconSuite(result, kIconSuitePrefsGeneral, kSelectorAllLargeData);
		Panel_SetAuxiliaryDataPtr(result, dataPtr);
	}
	return result;
}// New


//
// internal methods
//

/*!
This method, of ScreenOperationProcPtr form, will
update the specified screen appropriately based
on the value of the preference indicated by the
given General preference tag.

Currently, this method only supports changing of
the use background picture or dont dim
background session windows preferences.  If one
of these tags is given, then the specified screens
port rectangle is invalidated (which will cause the
picture to appear or disappear).  In the case of a
background picture, the per-screen flag is also
updated to reflect the preference value.

(3.0)
*/
static void
changePreferenceUpdateScreen	(const ConnectionDataPtr	inConnectionDataPtr,
								 const RealScreenDataPtr	inRealScreenDataPtr,
								 const SInt32				inDataPreferenceTag,
								 const SInt32				UNUSED_ARGUMENT(inUnusedData2),
								 void*						UNUSED_ARGUMENT(inoutUndefinedResultPtr))
{
	switch (inDataPreferenceTag)
	{
		case kTelnetPreferenceTag_DontDimBackgroundScreens:
			if (inRealScreenDataPtr != null) MacLook_InvalidatePortRectangle(inRealScreenDataPtr->window);
			break;
		
		case kTelnetPreferenceTag_UseBackgroundPicture:
			if (inRealScreenDataPtr != null)
			{
				Boolean		windowsHavePictures = kAlert_NotifyDisplayDiamondMark;
				Size		actualSize = 0L;
				
				
				unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_UseBackgroundPicture,
											sizeof(windowsHavePictures), &windowsHavePictures,
											&actualSize) == kTelnetPrefsResultCodeSuccess)
				{
					windowsHavePictures = false; // assume default, if preference cant be found
				}
				RealScreen_SetSpecialBackground(inConnectionDataPtr->vs, windowsHavePictures);
			}
			break;
		
		default:
			// unrecognized tag
			break;
	}
}// changePreferenceUpdateScreen


/*!
This routine is invoked whenever the color box value
is changed.  The panel responds by updating the color
preferences and updating all open session windows to
use the new color.

(3.0)
*/
static void
colorBoxChangeNotify	(ControlRef			inColorBoxThatChanged,
						 RGBColor const*	inNewColor,
						 void const*		inContextPtr)
{
	GeneralPanelDataPtr		dataPtr = (GeneralPanelDataPtr)inContextPtr;
	register SInt16			i = 0;
	
	
	for (i = 0; i < kANSIColorCount; i++)
	{
		if (inColorBoxThatChanged == dataPtr->tabs.colors.colorBoxes[i])
		{
			Preferences_SetANSIColor(i, inNewColor);
			ScreenManager_UpdatePalettes();
			break;
		}
	}
}// colorBoxChangeNotify


/*!
Creates the controls for the specified tab pane.
The specified panels descriptor must be
"kTelnetPanelDescriptorPreferencesGeneral".

The controls are not arranged or sized.

(3.0)
*/
static void
createChosenGeneralPane		(PanelRef		inPanel,
							 WindowRef		inOwningWindow)
{
	ControlRef				container = null,
							control = null;
	Rect					containerRect;
	register SInt16			i = 0;
	GeneralPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	Size					actualSize = 0L; // used for grabbing preference data
	
	
	Panel_GetContainerControl(inPanel, &container);
	GetControlBounds(container, &containerRect);
	
	if (getChosenTabIndex(inPanel) == kTabIndexGeneralOptions)
	{
		// create Options tab pane and controls
		dataPtr->tabs.options.pane = control = GetNewControl(kIDTabUserPane, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabControl);
		{
			Str255		title;
			Boolean		flag = false; // storage
			
			
			for (i = 0; i < NUMBER_OF_GENERAL_OPTIONS; i++)
			{
				dataPtr->tabs.options.options[i].checkBox = control = GetNewControl(kIDCheckBox, inOwningWindow);
				GetIndString(title, rStringsMiscellaneousControlTitles, siControlTitlePrefsGeneralOptionsBase + i);
				SetControlTitle(control, title);
				(OSErr)EmbedControl(control, dataPtr->tabs.options.pane);
			}
			
			// ...initialize checkboxes using user preferences
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_SimpleTelnet, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = false; // assume normal mode, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexSimpleTelnet].checkBox,
							BooleanToCheckBoxValue(flag));
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_DontAutoClose, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = false; // assume windows automatically close, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexDontAutoClose].checkBox,
							BooleanToCheckBoxValue(flag));
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_DynamicResizing, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = false; // assume normal window resizing, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexDynamicResizing].checkBox,
							BooleanToCheckBoxValue(flag));
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_HeadersCollapsed, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = false; // assume headers arent initially collapsed, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexHeadersInitiallyCollapsed].checkBox,
							BooleanToCheckBoxValue(flag));
			
			//unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_UseBackgroundPicture, sizeof(flag), &flag,
			//							&actualSize) == kTelnetPrefsResultCodeSuccess)
			//{
			//	flag = false; // assume no background pictures, if preference cant be found
			//}
			//SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexUsePicture].checkBox,
			//				BooleanToCheckBoxValue(flag));
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_DontDimBackgroundScreens, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = false; // assume background screens are dimmed, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexDontDimBackgroundScreens].checkBox,
							BooleanToCheckBoxValue(flag));
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_PureInverse, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = false; // assume text is highlighted according to Mac OS standards, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexPureInverse].checkBox,
							BooleanToCheckBoxValue(flag));
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CopySelectedText, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = false; // assume the clipboard isnt affected by text highlighting, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexCopySelectedText].checkBox,
							BooleanToCheckBoxValue(flag));
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CursorMovesPriorToDrops, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = false; // assume the cursor is not affected by drag and drop, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexCursorMovesToDropLocation].checkBox,
							BooleanToCheckBoxValue(flag));
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_MenuItemKeys, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = true; // assume menu items have key equivalents, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexMenuItemKeys].checkBox,
							BooleanToCheckBoxValue(flag));
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_MapBackquote, sizeof(flag), &flag,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				flag = false; // assume the backquote key isnt re-mapped, if preference cant be found
			}
			SetControlValue(dataPtr->tabs.options.options[kGeneralOptionIndexMapBackquote].checkBox,
							BooleanToCheckBoxValue(flag));
		}
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexGeneralSpecial)
	{
		// create Special tab pane and controls
		dataPtr->tabs.special.pane = control = GetNewControl(kIDTabUserPane, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabControl);
		
		// ...create cursor controls and initialize using user preferences
		dataPtr->tabs.special.cursor.groupBox = control = GetNewControl(kIDPrefGeneralCursorGroupBox, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.pane);
		dataPtr->tabs.special.cursor.shapeLabel = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.cursor.groupBox);
		{
			Str255		text;
			
			
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextCursorShape);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
		}
		// set up the cursor shape bevel buttons
		{
			enum
			{
				kNumberOfButtons = 5
			};
			struct CursorButtonInfo
			{
				SInt16				controlResourceID;
				UInt16				cursorType;
				ControlRef*			controlAddress;
			};
			register SInt16				i = 0;
			UInt16						terminalCursorType = kTerminalCursorTypeBlock;
			struct CursorButtonInfo		cursorButtonIDs[kNumberOfButtons] =
										{
											{
												kIDPrefGeneralCursorShapeBlockBevelButton,
												kTerminalCursorTypeBlock,
												null
											},
											{
												kIDPrefGeneralCursorShapeVerticalBarBevelButton,
												kTerminalCursorTypeVerticalLine,
												null
											},
											{
												kIDPrefGeneralCursorShapeThickVerticalBarBevelButton,
												kTerminalCursorTypeThickVerticalLine,
												null
											},
											{
												kIDPrefGeneralCursorShapeUnderlineBevelButton,
												kTerminalCursorTypeUnderscore,
												null
											},
											{
												kIDPrefGeneralCursorShapeThickUnderlineBevelButton,
												kTerminalCursorTypeThickUnderscore,
												null
											}
										};
			
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TerminalCursorType,
										sizeof(terminalCursorType), &terminalCursorType, &actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				terminalCursorType = kTerminalCursorTypeBlock; // assume a block-shaped cursor, if preference cant be found
			}
			
			cursorButtonIDs[0].controlAddress = &dataPtr->tabs.special.cursor.shapeBlockBevelButton;
			cursorButtonIDs[1].controlAddress = &dataPtr->tabs.special.cursor.shapeVerticalBarBevelButton;
			cursorButtonIDs[2].controlAddress = &dataPtr->tabs.special.cursor.shapeThickVerticalBarBevelButton;
			cursorButtonIDs[3].controlAddress = &dataPtr->tabs.special.cursor.shapeUnderlineBevelButton;
			cursorButtonIDs[4].controlAddress = &dataPtr->tabs.special.cursor.shapeThickUnderlineBevelButton;
			for (i = 0; i < kNumberOfButtons; i++)
			{
				*(cursorButtonIDs[i].controlAddress) = control = GetNewControl(cursorButtonIDs[i].controlResourceID, inOwningWindow);
				if (terminalCursorType == cursorButtonIDs[i].cursorType) SetControlValue(control, kControlCheckBoxCheckedValue);
				(OSErr)SetBevelButtonGraphicAlignment(control, kControlBevelButtonAlignBottom, 0/* offset H */, 0/* offset V */);
				(OSErr)EmbedControl(control, dataPtr->tabs.special.cursor.groupBox);
			}
		}
		dataPtr->tabs.special.cursor.flashingCheckBox = control = GetNewControl(kIDCheckBox, inOwningWindow);
		{
			Str255		title;
			Boolean		cursorBlinks = false;
			
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CursorBlinks,
										sizeof(cursorBlinks), &cursorBlinks, &actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				cursorBlinks = false; // assume the cursor doesnt flash, if preference cant be found
			}
			GetIndString(title, rStringsMiscellaneousControlTitles, siControlTitlePrefsCursorFlashing);
			SetControlTitle(control, title);
			SetControlValue(control, BooleanToCheckBoxValue(cursorBlinks));
		}
		(OSErr)EmbedControl(control, dataPtr->tabs.special.cursor.groupBox);
		
		// ...create timeouts controls and initialize using user preferences
		dataPtr->tabs.special.timeouts.groupBox = control = GetNewControl(kIDPrefGeneralTimeoutsGroupBox, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.pane);
		dataPtr->tabs.special.timeouts.openLabel = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.timeouts.groupBox);
		{
			Str255		text;
			
			
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextTimeoutOpen);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
		}
		dataPtr->tabs.special.timeouts.openField = control = GetNewControl(kIDEditableText, inOwningWindow);
		{
			UInt32		timeout = 0L;
			
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TimeoutOpen,
										sizeof(timeout), &timeout, &actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				timeout = 10; // assume a 10-second timeout, if preference cant be found
			}
			SetControlNumericalText(control, timeout);
		}
		useNumericalLimiter(control);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.timeouts.groupBox);
		dataPtr->tabs.special.timeouts.sendLabel = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.timeouts.groupBox);
		{
			Str255		text;
			
			
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextTimeoutSend);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
		}
		dataPtr->tabs.special.timeouts.sendField = control = GetNewControl(kIDEditableText, inOwningWindow);
		{
			UInt32		timeout = 0L;
			
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TimeoutSend,
										sizeof(timeout), &timeout, &actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				timeout = 2; // assume a 2-second timeout, if preference cant be found
			}
			SetControlNumericalText(control, timeout);
		}
		useNumericalLimiter(control);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.timeouts.groupBox);
		
		// ...create Copy Using Spaces For Tabs controls and initialize using user preferences
		dataPtr->tabs.special.copyTable.groupBox = control = GetNewControl(kIDPrefGeneralCopyTableGroupBox, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.pane);
		dataPtr->tabs.special.copyTable.thresholdLabel = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.copyTable.groupBox);
		{
			Str255		text;
			
			
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextCopyTableThreshold);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
		}
		dataPtr->tabs.special.copyTable.thresholdField = control = GetNewControl(kIDEditableText, inOwningWindow);
		{
			UInt16		copyTableThreshold = 0;
			
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CopyTableThreshold,
										sizeof(copyTableThreshold), &copyTableThreshold, &actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				copyTableThreshold = 4; // assume 4 spaces per tab, if preference cant be found
			}
			SetControlNumericalText(control, copyTableThreshold);
		}
		useNumericalLimiter(control);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.copyTable.groupBox);
		dataPtr->tabs.special.copyTable.thresholdUnitsLabel = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.copyTable.groupBox);
		{
			Str255		text;
			
			
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextCopyTableThresholdUnits);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
		}
		
		// ...create capture file controls and initialize using user preferences
		dataPtr->tabs.special.captureFile.groupBox = control = GetNewControl(kIDPrefGeneralCaptureFileGroupBox, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.pane);
		dataPtr->tabs.special.captureFile.creatorField = control = GetNewControl(kIDEditableText, inOwningWindow);
		{
			OSType		captureFileCreator;
			
			
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CaptureFileCreator,
										sizeof(captureFileCreator), &captureFileCreator, &actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				captureFileCreator = 'ttxt'; // assume default, if preference cant be found
			}
			SetControlOSTypeText(control, captureFileCreator);
		}
		useOSTypeLengthLimiter(control);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.captureFile.groupBox);
		dataPtr->tabs.special.captureFile.applicationButton = control = GetNewControl(kIDPrefFileTransfersTextFilesFromAppButton, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.special.captureFile.groupBox);
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexGeneralColors)
	{
		// create Colors tab pane and controls
		RGBColor			color;
		register SInt16		i = 0;
		Str255				text;
		
		
		dataPtr->tabs.colors.pane = control = GetNewControl(kIDTabUserPane, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabControl);
		
		dataPtr->tabs.colors.setToDefaultsButton = control = GetNewControl(kIDPrefGeneralAutoCalcANSIColorsButton, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.colors.pane);
		
		dataPtr->tabs.colors.headerLabel = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.colors.pane);
		GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextANSIColorsHeader);
		(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
		
		dataPtr->tabs.colors.normalHeaderLabel = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.colors.pane);
		GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextANSIColorsNormalHeader);
		(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
		
		dataPtr->tabs.colors.emphasizedHeaderLabel = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.colors.pane);
		GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextANSIColorsEmphasizedHeader);
		(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
		
		dataPtr->tabs.colors.normalHeaderDivider = control = GetNewControl(kIDSeparator, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.colors.pane);
		
		dataPtr->tabs.colors.emphasizedHeaderDivider = control = GetNewControl(kIDSeparator, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.colors.pane);
		
		for (i = 0; i < kANSIColorCount; i++)
		{
			// create a color box for this color
			Preferences_GetANSIColor(i, &color);
			dataPtr->tabs.colors.colorBoxes[i] = control = ColorBox_GetNewControl(inOwningWindow, &color/* initial color */);
			ColorBox_SetColorChangeNotifyProc(control, colorBoxChangeNotify/* notification routine */, dataPtr/* context */);
			(OSErr)EmbedControl(control, dataPtr->tabs.colors.pane);
			
			// create a label for the new color box, auto-sizing it
			dataPtr->tabs.colors.colorBoxLabels[i] = control = GetNewControl(kIDStaticText, inOwningWindow);
			(OSErr)EmbedControl(control, dataPtr->tabs.colors.pane);
			{
				SInt16		index = i;
				
				
				if (i >= 8) index -= 8;
				index += siStaticTextANSIColorLabelBlack; // specify which label string to use
				GetIndString(text, rStringsMiscellaneousStaticText, index);
				(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
			}
		}
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexGeneralNotification)
	{
		Boolean		visualBell = false,
					notifyOfBeeps = false;
		
		
		unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_VisualBell,
								sizeof(visualBell), &visualBell, &actualSize) == kTelnetPrefsResultCodeSuccess)
		{
			visualBell = false; // assume default, if preference cant be found
		}
		unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_NotifyOfBeeps,
								sizeof(notifyOfBeeps), &notifyOfBeeps, &actualSize) == kTelnetPrefsResultCodeSuccess)
		{
			notifyOfBeeps = false; // assume default, if preference cant be found
		}
		
		// create Notification tab pane and controls
		dataPtr->tabs.notification.pane = control = GetNewControl(kIDTabUserPane, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabControl);
		
		dataPtr->tabs.notification.visualBellCheckBox = control = GetNewControl(kIDPrefGeneralNotificationVisualBellCheckBox, inOwningWindow);
		SetControlValue(control, BooleanToCheckBoxValue(visualBell));
		(OSErr)EmbedControl(control, dataPtr->tabs.notification.pane);
		
		dataPtr->tabs.notification.background.groupBox = control = GetNewControl(kIDPrefGeneralBackgroundNotificationGroupBox, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.notification.pane);
		
		dataPtr->tabs.notification.background.terminalBeepsCheckBox = control = GetNewControl
																		(kIDPrefGeneralBackgroundNotificationBeepsCheckBox, inOwningWindow);
		SetControlValue(control, BooleanToCheckBoxValue(notifyOfBeeps));
		(OSErr)EmbedControl(control, dataPtr->tabs.notification.background.groupBox);
		dataPtr->tabs.notification.background.separator = control = GetNewControl(kIDSeparator, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.notification.background.groupBox);
		
		dataPtr->tabs.notification.background.helpText = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.notification.background.groupBox);
		dataPtr->tabs.notification.background.whenPostingText = control = GetNewControl(kIDStaticText, inOwningWindow);
		(OSErr)EmbedControl(control, dataPtr->tabs.notification.background.groupBox);
		{
			Str255		text;
			
			
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextWhenPostingNotifications);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
		}
		{
			Str255		title;
			
			
			for (i = 0; i < NUMBER_OF_GENERAL_NOTIFICATION_METHODS; i++)
			{
				// ...create a radio button for this notification method (getting the proper title string from the string list resource)
				dataPtr->tabs.notification.background.methods[i].radioButton = control = GetNewControl(kIDRadioButton, inOwningWindow);
				GetIndString(title, rStringsMiscellaneousControlTitles, siControlTitlePrefsGeneralNotificationMethodsBase + i);
				SetControlTitle(control, title);
				(OSErr)EmbedControl(control, dataPtr->tabs.notification.background.groupBox);
				
				// ...create a user pane that draws up to three icons next to the radio button
				dataPtr->tabs.notification.background.methods[i].symbolsPane = control = GetNewControl(kIDAnchorUserPane, inOwningWindow);
				(OSErr)SetControlData(control, kControlNoPart, kControlUserPaneDrawProcTag, sizeof(dataPtr->iconDrawingUPP),
										(Ptr)&dataPtr->iconDrawingUPP);
				
				{
					SInt16		notificationType = kAlert_NotifyDoNothing;
					
					
					switch (i)
					{
						case kGeneralNotificationMethodIndexNone:
							notificationType = kAlert_NotifyDoNothing;
							break;
							
						case kGeneralNotificationMethodIndexDiamond:
							notificationType = kAlert_NotifyDisplayDiamondMark;
							break;
							
						case kGeneralNotificationMethodIndexFlashIcon:
							notificationType = kAlert_NotifyDisplayIconAndDiamondMark;
							break;
							
						case kGeneralNotificationMethodIndexDisplayAlert:
							notificationType = kAlert_NotifyAlsoDisplayAlert;
							break;
					}
					SetControlReference(control, notificationType);
				}
				(OSErr)EmbedControl(control, dataPtr->tabs.notification.background.groupBox);
			}
			
			// ...initialize radio buttons using user preferences
			{
				ControlRef		selectedControl = null;
				UInt16			notificationPreferences = kAlert_NotifyDisplayDiamondMark;
				
				
				unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_Notification,
										sizeof(notificationPreferences), &notificationPreferences,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
				{
					notificationPreferences = kAlert_NotifyDisplayDiamondMark; // assume default, if preference cant be found
				}
				SetControlValue(dataPtr->tabs.notification.background.methods[kGeneralNotificationMethodIndexNone].radioButton,
								kControlRadioButtonUncheckedValue);
				SetControlValue(dataPtr->tabs.notification.background.methods[kGeneralNotificationMethodIndexDiamond].radioButton,
								kControlRadioButtonUncheckedValue);
				SetControlValue(dataPtr->tabs.notification.background.methods[kGeneralNotificationMethodIndexFlashIcon].radioButton,
								kControlRadioButtonUncheckedValue);
				SetControlValue(dataPtr->tabs.notification.background.methods[kGeneralNotificationMethodIndexDisplayAlert].radioButton,
								kControlRadioButtonUncheckedValue);
				switch (notificationPreferences)
				{
					case kAlert_NotifyDoNothing:
						selectedControl = dataPtr->tabs.notification.background.methods[kGeneralNotificationMethodIndexNone].radioButton;
						break;
					
					case kAlert_NotifyDisplayDiamondMark:
						selectedControl = dataPtr->tabs.notification.background.methods[kGeneralNotificationMethodIndexDiamond].radioButton;
						break;
					
					case kAlert_NotifyDisplayIconAndDiamondMark:
						selectedControl = dataPtr->tabs.notification.background.methods[kGeneralNotificationMethodIndexFlashIcon].radioButton;
						break;
					
					case kAlert_NotifyAlsoDisplayAlert:
						selectedControl = dataPtr->tabs.notification.background.methods[kGeneralNotificationMethodIndexDisplayAlert].radioButton;
						break;
				}
				
				if (selectedControl != null) SetControlValue(selectedControl, kControlRadioButtonCheckedValue);
			}
		}
	}
}// createChosenGeneralPane


/*!
Creates the controls that belong in the General panel, and
attaches them to the specified owning window in the proper
embedding hierarchy.  The specified panels descriptor must
be "kTelnetPanelDescriptorPreferencesGeneral".

The controls are not arranged or sized.

(3.0)
*/
static void
createPanelControls		(PanelRef			inPanel,
						 WindowRef			inOwningWindow)
{
	ControlRef				container = GetNewControl(kIDPanelUserPane, inOwningWindow),
							control = null;
	GeneralPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	Panel_SetContainerControl(inPanel, container);
	SetControlVisibility(container, false/* visible */, false/* draw */);
	
	dataPtr->tabControl = control = GetNewControl(kIDPrefGeneralTabs, inOwningWindow);
	(OSErr)EmbedControl(control, container);
	
	dataPtr->iconDrawingUPP = NewControlUserPaneDrawUPP(notificationPictureDraw);
	
	createChosenGeneralPane(inPanel, inOwningWindow);
}// createPanelControls


/*!
Adjusts the controls in the General preference panel
to match the specified change in dimensions of its
container.

(3.0)
*/
static void
deltaPanelSize		(PanelRef			inPanel,
					 const Point		inChangeInWidthAndHeight)
{
	GeneralPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	ControlRef				control = null;
	Rect					controlRect;
	register SInt16			i = 0;
	register SInt16			additionalDeltaY = 0;
	
	
	control = dataPtr->tabControl;
	GetControlBounds(control, &controlRect);
	SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
	
	if (getChosenTabIndex(inPanel) == kTabIndexGeneralOptions)
	{
		#if 1
		// resize the Options tab pane, and its controls
		control = dataPtr->tabs.options.pane;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
		for (i = 0; i < NUMBER_OF_GENERAL_OPTIONS; i++)
		{
			control = dataPtr->tabs.options.options[i].checkBox;
			GetControlBounds(control, &controlRect);
			MoveControl(control, controlRect.left, controlRect.top + i * (inChangeInWidthAndHeight.v / NUMBER_OF_GENERAL_OPTIONS));
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
		}
		#else
		// try using the Dialog Adjust module to smooth out resizing this panel
		DialogAdjust_BeginControlAdjustment(Panel_GetOwningWindow(inPanel));
		
		control = dataPtr->tabs.options.pane;
		DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, control, inChangeInWidthAndHeight.h);
		DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeV, control, inChangeInWidthAndHeight.v);
		
		for (i = 0; i < NUMBER_OF_GENERAL_OPTIONS; i++)
		{
			control = dataPtr->tabs.options.options[i].checkBox;
			DialogAdjust_AddControlAdjustment(kControlAdjustmentMoveV, control, i * (inChangeInWidthAndHeight.v / NUMBER_OF_GENERAL_OPTIONS));
			DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, control, inChangeInWidthAndHeight.h);
		}
		
		DialogAdjust_EndAdjustment(inChangeInWidthAndHeight.h, inChangeInWidthAndHeight.v);
		#endif
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexGeneralSpecial)
	{
		// resize the Special tab pane, and its controls
		control = dataPtr->tabs.special.pane;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
		{
			control = dataPtr->tabs.special.cursor.groupBox;
			GetControlBounds(control, &controlRect);
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
								controlRect.bottom - controlRect.top);
			control = dataPtr->tabs.special.timeouts.groupBox;
			GetControlBounds(control, &controlRect);
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
								controlRect.bottom - controlRect.top);
			control = dataPtr->tabs.special.copyTable.groupBox;
			GetControlBounds(control, &controlRect);
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
								controlRect.bottom - controlRect.top);
			control = dataPtr->tabs.special.copyTable.thresholdUnitsLabel;
			GetControlBounds(control, &controlRect);
			MoveControl(control, controlRect.left + INTEGER_HALVED(inChangeInWidthAndHeight.h), controlRect.top);
			SizeControl(control, controlRect.right - controlRect.left + INTEGER_HALVED(inChangeInWidthAndHeight.h),
								controlRect.bottom - controlRect.top);
			control = dataPtr->tabs.special.copyTable.thresholdField;
			GetControlBounds(control, &controlRect);
			MoveControl(control, controlRect.left + INTEGER_HALVED(inChangeInWidthAndHeight.h), controlRect.top);
			control = dataPtr->tabs.special.copyTable.thresholdLabel;
			GetControlBounds(control, &controlRect);
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h / 3,
								controlRect.bottom - controlRect.top);
			control = dataPtr->tabs.special.captureFile.groupBox;
			GetControlBounds(control, &controlRect);
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
								controlRect.bottom - controlRect.top);
		}
		control = dataPtr->tabs.special.cursor.flashingCheckBox;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexGeneralColors)
	{
		// resize the Colors tab pane, and its controls
		control = dataPtr->tabs.colors.pane;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
	}
	else if (getChosenTabIndex(inPanel) == kTabIndexGeneralNotification)
	{
		// resize the Notification tab pane, and its controls
		control = dataPtr->tabs.notification.pane;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
		control = dataPtr->tabs.notification.visualBellCheckBox;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
		{
			Rect		newRect;
			Str255		text;
			Size		actualSize = 0L;
			Point		textDeltaSize;
			
			
			// recalculate the smallest area that the text can occupy
			control = dataPtr->tabs.notification.background.helpText;
			GetControlBounds(control, &controlRect);
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
			GetControlText(control, text);
			(UInt16)Localization_SetUpMultiLineTextControl(control, text, kThemeSmallSystemFont); // auto-sizes text area
			GetControlBounds(control, &newRect);
			SetPt(&textDeltaSize, (newRect.right - newRect.left) - (controlRect.right - controlRect.left),
									(newRect.bottom - newRect.top) - (controlRect.bottom - controlRect.top));
			additionalDeltaY += textDeltaSize.v;
		}
		control = dataPtr->tabs.notification.background.groupBox;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
							controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v + additionalDeltaY);
		control = dataPtr->tabs.notification.background.terminalBeepsCheckBox;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
		control = dataPtr->tabs.notification.background.separator;
		GetControlBounds(control, &controlRect);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
		control = dataPtr->tabs.notification.background.whenPostingText;
		GetControlBounds(control, &controlRect);
		MoveControl(control, controlRect.left, controlRect.top + additionalDeltaY);
		SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
		for (i = 0; i < NUMBER_OF_GENERAL_NOTIFICATION_METHODS; i++)
		{
			control = dataPtr->tabs.notification.background.methods[i].symbolsPane;
			GetControlBounds(control, &controlRect);
			MoveControl(control,
						controlRect.left,
						controlRect.top + i * (inChangeInWidthAndHeight.v / NUMBER_OF_GENERAL_NOTIFICATION_METHODS) + additionalDeltaY);
			control = dataPtr->tabs.notification.background.methods[i].radioButton;
			GetControlBounds(control, &controlRect);
			MoveControl(control,
						controlRect.left,
						controlRect.top + i * (inChangeInWidthAndHeight.v / NUMBER_OF_GENERAL_NOTIFICATION_METHODS) + additionalDeltaY);
			SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h, controlRect.bottom - controlRect.top);
		}
	}
}// deltaPanelSize


/*!
Hides all tab panes for the General panel, except
for the pane corresponding to the active tab in that
panel.

(3.0)
*/
static void
disposeOldTabPane		(PanelRef		inPanel,
						 const UInt16	inTabIndex)
{
	GeneralPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	ControlRef*				controlHandlePtr = null;
	
	
	switch (inTabIndex)
	{
		case kTabIndexGeneralOptions:
			controlHandlePtr = &dataPtr->tabs.options.pane;
			break;
		
		case kTabIndexGeneralSpecial:
			controlHandlePtr = &dataPtr->tabs.special.pane;
			break;
		
		case kTabIndexGeneralColors:
			controlHandlePtr = &dataPtr->tabs.colors.pane;
			break;
		
		case kTabIndexGeneralNotification:
			controlHandlePtr = &dataPtr->tabs.notification.pane;
			break;
		
		default:
			// ???
			break;
	}
	
	if (controlHandlePtr != null)
	{
		(OSErr)SetControlVisibility(*controlHandlePtr, false/* visible */, true/* draw */);
		DisposeControl(*controlHandlePtr), *controlHandlePtr = null;
	}
}// disposeOldTabPane


/*!
Cleans up a panel that is about to be destroyed.

(3.0)
*/
static void
disposePanel	(PanelRef			UNUSED_ARGUMENT(inPanel),
				 void*				inDataPtr)
{
	GeneralPanelDataPtr		dataPtr = inDataPtr;
	
	
	DisposeControlUserPaneDrawUPP(dataPtr->iconDrawingUPP), dataPtr->iconDrawingUPP = null;
	Memory_DisposePtr((Ptr*)&dataPtr);
}// disposePanel


/*!
Returns the one-based tab index of the selected
tab for a panel.

(3.0)
*/
static UInt16
getChosenTabIndex	(PanelRef		inPanel)
{
	UInt16					result = 0;
	GeneralPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	if (dataPtr != null) result = GetControlValue(dataPtr->tabControl);
	return result;
}// getChosenTabIndex


/*!
Responds to control clicks in this panel.

(3.0)
*/
static SInt32
handleContentClick	(PanelRef		inPanel,
					 ControlRef		inControl)
{
	SInt32					result = kPanel_ResponseClickNotHandled;
	Boolean					handled = false;
	GeneralPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	UInt16					chosenTabIndex = getChosenTabIndex(inPanel);
	
	
	if (inControl == dataPtr->tabControl)
	{
		showChosenTabPane(inPanel);
		handled = true;
	}
	else if (chosenTabIndex == kTabIndexGeneralSpecial)
	{
		if ((inControl == dataPtr->tabs.special.cursor.shapeBlockBevelButton) ||
			(inControl == dataPtr->tabs.special.cursor.shapeVerticalBarBevelButton) ||
			(inControl == dataPtr->tabs.special.cursor.shapeThickVerticalBarBevelButton) ||
			(inControl == dataPtr->tabs.special.cursor.shapeUnderlineBevelButton) ||
			(inControl == dataPtr->tabs.special.cursor.shapeThickUnderlineBevelButton))
		{
			UInt16		cursorType = kTerminalCursorTypeBlock;
			
			
			SetControlValue(dataPtr->tabs.special.cursor.shapeBlockBevelButton,
							kControlCheckBoxUncheckedValue);
			SetControlValue(dataPtr->tabs.special.cursor.shapeVerticalBarBevelButton,
							kControlCheckBoxUncheckedValue);
			SetControlValue(dataPtr->tabs.special.cursor.shapeThickVerticalBarBevelButton,
							kControlCheckBoxUncheckedValue);
			SetControlValue(dataPtr->tabs.special.cursor.shapeUnderlineBevelButton,
							kControlCheckBoxUncheckedValue);
			SetControlValue(dataPtr->tabs.special.cursor.shapeThickUnderlineBevelButton,
							kControlCheckBoxUncheckedValue);
			SetControlValue(inControl, kControlCheckBoxCheckedValue);
			if (inControl == dataPtr->tabs.special.cursor.shapeBlockBevelButton) cursorType = kTerminalCursorTypeBlock;
			if (inControl == dataPtr->tabs.special.cursor.shapeVerticalBarBevelButton) cursorType = kTerminalCursorTypeVerticalLine;
			if (inControl == dataPtr->tabs.special.cursor.shapeThickVerticalBarBevelButton) cursorType = kTerminalCursorTypeThickVerticalLine;
			if (inControl == dataPtr->tabs.special.cursor.shapeUnderlineBevelButton) cursorType = kTerminalCursorTypeUnderscore;
			if (inControl == dataPtr->tabs.special.cursor.shapeThickUnderlineBevelButton) cursorType = kTerminalCursorTypeThickUnderscore;
			Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TerminalCursorType,
								sizeof(cursorType), &cursorType);
			handled = true;
		}
		else if (inControl == dataPtr->tabs.special.captureFile.applicationButton)
		{
			OSType		creatorCode = '????';
			
			
			if (GetApplicationType(&creatorCode))
			{
				ControlRef		field = dataPtr->tabs.special.captureFile.creatorField;
				
				
				SetControlOSTypeText(field, creatorCode);
				DrawOneControl(field);
				savePreferenceForField(inPanel, field);
			}
			handled = true;
		}
		else if (inControl == dataPtr->tabs.special.cursor.flashingCheckBox)
		{
			result = updateCheckBoxPreference(inPanel, inControl);
		}
	}
	else if (chosenTabIndex == kTabIndexGeneralColors)
	{
		if (inControl == dataPtr->tabs.colors.setToDefaultsButton)
		{
			PaletteHandle		paletteHandle = null;
			RGBColor			color;
			SInt16				oldResFile = CurResFile();
			register SInt16		i = 0;
			
			
			// WARNING: this code depends on the ordering and values of the ANSI palette
			ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
			paletteHandle = GetNewPalette(ANSICOLORS_DEFAULTID);
			for (i = 0; i < kANSIColorCount; i++)
			{
				GetEntryColor(paletteHandle, i, &color);
				ColorBox_SetColor(dataPtr->tabs.colors.colorBoxes[i], &color);
				DrawOneControl(dataPtr->tabs.colors.colorBoxes[i]);
			}
			DisposePalette(paletteHandle), paletteHandle = null;
			UseResFile(oldResFile);
			handled = true;
		}
		else
		{
			// see if any of the color boxes was hit; if so, display a color chooser
			// and then (if the user accepts a new color) update all open session windows
			register SInt16		i = 0;
			
			
			for (i = 0; i < kANSIColorCount; i++)
			{
				if (inControl == dataPtr->tabs.colors.colorBoxes[i])
				{
					(Boolean)ColorBox_UserSetColor(inControl);
					handled = true;
					break;
				}
			}
		}
	}
	else result = updateCheckBoxPreference(inPanel, inControl);
	
	if (handled) result = kPanel_ResponseClickHandled;
	
	return result;
}// handleContentClick


/*!
This method, of standard DeviceLoopDrawingUPP form,
handles drawing of the notification icons contents
correctly no matter what the device characteristics
are.

Up to three symbols are drawn in the space of the
given control, based on the notification level
set for the control in its reference constant.  The
reference constant, therefore, is one of the
"kAlert_Notify..." constants, as defined in
"InterfaceLib.h".

At the highest level, all three symbols are drawn;
the leftmost symbol is a miniature alert box, the
middle symbol is Telnets notification icon, and
the last symbol is a diamond mark.

Using three user panes that utilize this drawing
procedure, one, two or three of the symbols are
drawn in the Preferences window next to their
respective radio buttons.

(3.0)
*/
static pascal void
notificationPictureDeviceLoop	(short				inColorDepth,
								 short				UNUSED_ARGUMENT(inDeviceFlags),
								 GDHandle			inTargetDevice,
								 long				inControl)
{
	Rect			controlRect;
	ControlRef		control = REINTERPRET_CAST(inControl, ControlRef);
	ColorPenState	state;
	SInt16			notificationType = 0;
	
	
	// figure out which icons should be drawn (if any)
	notificationType = GetControlReference(control);
	
	ColorUtilities_GetColorAndPenState(&state);
	ColorUtilities_NormalizeColorAndPen();
	
	// set up the control background and foreground colors
	(OSErr)SetUpControlBackground(control, inColorDepth, ColorUtilities_IsColorDevice(inTargetDevice));
	(OSStatus)SetThemeTextColor(IsControlActive(control) ? kThemeTextColorModelessDialogActive : kThemeTextColorModelessDialogInactive,
								inColorDepth, ColorUtilities_IsColorDevice(inTargetDevice));
	
	GetControlBounds(control, &controlRect);
	EraseRect(&controlRect);
	
	if (notificationType != kAlert_NotifyDoNothing) // the kAlert_NotifyDoNothing variant is blank!
	{
		OSStatus			error = 0L;
		
		// draw diamond mark
		{
			GrafPortFontState	fontState;
			SInt16				fontSize = 0;
			Style				fontStyle = normal;
			Str255				fontName;
			FontInfo			fontInfo;
			
			
			// save port settings
			Localization_GetPortFontState(&fontState);
			
			Localization_UseThemeFont(kThemeSystemFont, fontName, &fontSize, &fontStyle);
			GetFontInfo(&fontInfo);
			MoveTo(controlRect.right - ICON_WD_SMALL + INTEGER_HALVED(ICON_WD_SMALL - CharWidth('')),
					controlRect.top + INTEGER_HALVED(controlRect.bottom - controlRect.top -
														(fontInfo.ascent + fontInfo.descent + INTEGER_DOUBLED(fontInfo.leading))) +
														(fontInfo.leading + fontInfo.ascent));
			DrawChar('');
			
			// restore port settings
			Localization_SetPortFontState(&fontState);
		}
		
		if (notificationType != kAlert_NotifyDisplayDiamondMark) // only draw subsequent icons for sufficiently-high notification levels
		{
			IconTransformType		transform = (IsControlActive(control) ? kTransformNone : kTransformDisabled);
			IconSuiteRef			icon = null;
			Rect					iconRect;
			OSErr					iconAccessErr = noErr;
			
			
			iconAccessErr = GetIconSuite(&icon, kIconSuiteNotificationFlashIcon, kSelectorAllSmallData);
			if ((iconAccessErr == noErr) && (icon != null))
			{
				SetRect(&iconRect, controlRect.left + INTEGER_HALVED(controlRect.right - controlRect.left - ICON_WD_SMALL), controlRect.top,
									0, controlRect.bottom);
				SetRect(&iconRect, iconRect.left, iconRect.top, iconRect.left + ICON_WD_SMALL, iconRect.bottom);
				(OSErr)PlotIconSuite(&iconRect, kAlignVerticalCenter, transform, icon);
				(OSErr)DisposeIconSuite(icon, true/* dispose data too */), icon = null;
			}
			
			if (notificationType != kAlert_NotifyDisplayIconAndDiamondMark) // only draw the last icon for the highest notification level
			{
				iconAccessErr = GetIconSuite(&icon, kIconSuiteNotificationDisplayAlert, kSelectorAllSmallData);
				if ((iconAccessErr == noErr) && (icon != null))
				{
					SetRect(&iconRect, controlRect.left, controlRect.top, controlRect.left + ICON_WD_SMALL, controlRect.bottom);
					(OSErr)PlotIconSuite(&iconRect, kAlignVerticalCenter, transform, icon);
					(OSErr)DisposeIconSuite(icon, true/* dispose data too */), icon = null;
				}
			}
		}
	}
	
	ColorUtilities_SetColorAndPenState(&state);
}// notificationPictureDeviceLoop


/*!
This method, a standard Mac OS user pane
drawing procedure, draws the icons that
show the different Notification Manager
notification levels (flashing icon,
icon and diamond mark, alert message).

(3.0)
*/
static pascal void
notificationPictureDraw		(ControlRef		inControl,
							 SInt16			UNUSED_ARGUMENT(inPartCode))
{
	DeviceLoopDrawingUPP		upp = NewDeviceLoopDrawingUPP(notificationPictureDeviceLoop);
	
	
	DeviceLoopClipAndDrawControl(inControl, upp);
	DisposeDeviceLoopDrawingUPP(upp), upp = null;
}// notificationPictureDraw


/*!
This routine, of standard PanelChangedProcPtr form,
is invoked by the Panel module whenever a property
of one of the preferences dialogs panels changes.

(3.0)
*/
static SInt32
panelChanged	(PanelRef			inPanel,
				 PanelMessage		inMessage,
				 void*				inData)
{
	SInt32		result = 0L;
	
	
	if (Panel_GetDescriptor(inPanel) == kTelnetPanelDescriptorPreferencesGeneral) switch (inMessage)
	{
		case kPanel_MessageClicked: // notification of a content click
			{
				ControlRef*		dataPtr = inData;
				
				
				result = handleContentClick(inPanel, *dataPtr);
			}
			break;
		
		case kPanel_MessageCreateControls: // specification of the window containing the panel - create controls using this window
			{
				WindowRef	window = *((WindowRef*)inData);
				
				
				createPanelControls(inPanel, window);
			}
			break;
		
		case kPanel_MessageDeltaSize: // request to alter size and/or position of controls; containing control will already have been changed
			{
				Point			changeInPanelDimensions;
				ControlRef		panelControl = null;
				Boolean			wasVisible = false;
				
				
				SetPt(&changeInPanelDimensions, ((Point*)inData)->h, ((Point*)inData)->v);
				Panel_GetContainerControl(inPanel, &panelControl);
				wasVisible = IsControlVisible(panelControl);
				SetControlVisibility(panelControl, false/* visibility */, wasVisible/* draw */);
				deltaPanelSize(inPanel, changeInPanelDimensions);
				SetControlVisibility(panelControl, wasVisible/* visibility */, wasVisible/* draw */);
			}
			break;
		
		case kPanel_MessageDestroyed: // request to dispose of private data structures
			{
				void*		dataPtr = inData;
				
				
				disposePanel(inPanel, dataPtr);
			}
			break;
		
		case kPanel_MessageFocusGained: // notification that a control is now focused
			{
				ControlRef*		dataPtr = inData;
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageFocusLost: // notification that a control is no longer focused
			{
				ControlRef*		dataPtr = inData;
				
				
				savePreferenceForField(inPanel, *dataPtr);
			}
			break;
		
		case kPanel_MessageInitialSize: // request for panel to size its controls
			{
				Point		exactPanelDimensions;
				
				
				SetPt(&exactPanelDimensions, ((Point*)inData)->h, ((Point*)inData)->v);
				setUpChosenGeneralPane(inPanel);
			}
			break;
		
		case kPanel_MessageNewActiveState: // request to change panel controls to a new active (true) or inactive (false) state
			{
				Boolean		isNowActive = *((Boolean*)inData);
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageNewAppearanceTheme: // notification of theme switch, a request to recalculate control sizes
			{
				// this notification is currently ignored, but shouldnt be...
			}
			break;
		
		case kPanel_MessageNewVisibility: // visible state of the panels container has changed to visible (true) or invisible (false)
			{
				Boolean		isNowVisible = *((Boolean*)inData);
				
				
				// do nothing
			}
			break;
		
		default:
			break;
	}
	
	return result;
}// panelChanged


/*!
Updates the preferences in memory to reflect the
information in the specified text field.  This
should be done if the focus of the window changes
(by hitting the tab key or clicking in another
text field).

(3.0)
*/
static void
savePreferenceForField		(PanelRef			inPanel,
							 ControlRef			inControl)
{
	GeneralPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	PanelDescriptor			descriptor = Panel_GetDescriptor(inPanel);
	
	
	if ((inControl != null) && (dataPtr != null))
	{
		if (inControl == dataPtr->tabs.special.captureFile.creatorField)
		{
			OSType		creator = '????';
			
			
			GetControlOSTypeText(inControl, &creator);
			Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CaptureFileCreator, sizeof(OSType), &creator);
		}
		else if (inControl == dataPtr->tabs.special.timeouts.openField)
		{
			SInt32		value = 0L;
			
			
			GetControlNumericalText(inControl, &value);
			Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TimeoutOpen, sizeof(UInt32), &value);
		}
		else if (inControl == dataPtr->tabs.special.timeouts.sendField)
		{
			SInt32		value = 0L;
			
			
			GetControlNumericalText(inControl, &value);
			Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_TimeoutSend, sizeof(UInt32), &value);
		}
		else if (inControl == dataPtr->tabs.special.copyTable.thresholdField)
		{
			SInt32		value = 0L;
			UInt16		threshold = 0;
			
			
			GetControlNumericalText(inControl, &value);
			threshold = value;
			Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CopyTableThreshold, sizeof(UInt16), &threshold);
		}
	}
}// savePreferenceForField


/*!
Sizes and arranges controls in the currently-selected
tab pane of the specified General panel to fit the
dimensions of the panels container.

(3.0)
*/
static void
setUpChosenGeneralPane	(PanelRef		inPanel)
{
	enum
	{
		kNotificationSymbolsWidth = (ICON_WD_SMALL * 3) + INTEGER_DOUBLED(HSP_CONTROLS)
	};
	GeneralPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	ControlRef				container = null,
							control = null;
	Rect					containerRect;
	Point					tabPaneTopLeft,
							tabPaneWidthHeight;
	register SInt16			i = 0;
	UInt16					leftEdge = 0, // cumulative, resettable measurement allowing controls to sandwich horizontally
							topEdge = 0; // cumulative, resettable measurement allowing controls to stack vertically within an embedder
	UInt16					textHeight = 0; // height of help text
	SInt32					checkBoxHeight = 0L,
							radioButtonHeight = 0L;
	UInt16					chosenTabIndex = getChosenTabIndex(inPanel);
	
	
	Panel_GetContainerControl(inPanel, &container);
	GetControlBounds(container, &containerRect);
	
	PreferencesDialog_CalculateTabPane(inPanel, &tabPaneTopLeft, &tabPaneWidthHeight);
	
	//if (GetThemeMetric(kThemeMetricCheckBoxGlyphHeight, &checkBoxHeight) != noErr) checkBoxHeight = CHECKBOX_HT;
	//if (GetThemeMetric(kThemeMetricRadioButtonGlyphHeight, &radioButtonHeight) != noErr) radioButtonHeight = RADIOBUTTON_HT;
	checkBoxHeight = CHECKBOX_HT;
	radioButtonHeight = RADIOBUTTON_HT;
	
	if (chosenTabIndex == kTabIndexGeneralOptions)
	{
		// size Options tab pane and its controls
		control = dataPtr->tabs.options.pane;
		SizeControl(control, tabPaneWidthHeight.h, tabPaneWidthHeight.v);
		for (i = 0; i < NUMBER_OF_GENERAL_OPTIONS; i++)
		{
			control = dataPtr->tabs.options.options[i].checkBox;
			SizeControl(control, tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), checkBoxHeight);
		}
		
		// arrange Options tab pane and its controls
		topEdge = VSP_PANE_AND_CONTROL;
		for (i = 0; i < NUMBER_OF_GENERAL_OPTIONS; i++)
		{
			control = dataPtr->tabs.options.options[i].checkBox;
			MoveControl(control, HSP_PANE_AND_CONTROL, topEdge);
			topEdge += (checkBoxHeight + VSP_CHECKBOXES);
		}
		control = dataPtr->tabs.options.pane;
		MoveControl(control, tabPaneTopLeft.h, tabPaneTopLeft.v);
		(OSErr)SetControlVisibility(control, true/* visible */, true/* draw */);
	}
	else if (chosenTabIndex == kTabIndexGeneralSpecial)
	{
		UInt16		boxHeightCursorBox = (VSP_GROUPBOXTEXT_AND_CONTROL + BUTTON_HT + VSP_GROUPBOX_AND_CONTROL),
					boxHeightOtherBoxes = (VSP_GROUPBOXTEXT_AND_CONTROL + SPACE_EXTRA_TEXTEDIT + TEXT_HT_BIG + SPACE_EXTRA_TEXTEDIT + VSP_GROUPBOX_AND_CONTROL);
		
		
		// size Special tab pane and its controls
		control = dataPtr->tabs.special.pane;
		SizeControl(control, tabPaneWidthHeight.h, tabPaneWidthHeight.v);
		
		SizeControl(dataPtr->tabs.special.cursor.groupBox,
					tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), boxHeightCursorBox);
		SizeControl(dataPtr->tabs.special.cursor.shapeBlockBevelButton, BUTTON_HT, BUTTON_HT);
		SizeControl(dataPtr->tabs.special.cursor.shapeVerticalBarBevelButton, BUTTON_HT, BUTTON_HT);
		SizeControl(dataPtr->tabs.special.cursor.shapeThickVerticalBarBevelButton, BUTTON_HT, BUTTON_HT);
		SizeControl(dataPtr->tabs.special.cursor.shapeUnderlineBevelButton, BUTTON_HT, BUTTON_HT);
		SizeControl(dataPtr->tabs.special.cursor.shapeThickUnderlineBevelButton, BUTTON_HT, BUTTON_HT);
		{
			Rect		cursorGroupBoxRect;
			
			
			GetControlBounds(dataPtr->tabs.special.cursor.groupBox, &cursorGroupBoxRect);
			SizeControl(dataPtr->tabs.special.cursor.flashingCheckBox,
						75/* FIX THIS TO CALCULATE PROPERLY!!! */, checkBoxHeight);
		}
		
		SizeControl(dataPtr->tabs.special.timeouts.groupBox,
					tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), boxHeightOtherBoxes);
		{
			enum
			{
				kTimeoutFieldWidth = 40
			};
			
			
			SizeControl(dataPtr->tabs.special.timeouts.openField, kTimeoutFieldWidth, TEXT_HT_BIG);
			SizeControl(dataPtr->tabs.special.timeouts.sendField, kTimeoutFieldWidth, TEXT_HT_BIG);
		}
		
		SizeControl(dataPtr->tabs.special.copyTable.groupBox,
					tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), boxHeightOtherBoxes);
		{
			enum
			{
				kThresholdFieldWidth = 25
			};
			
			
			SizeControl(dataPtr->tabs.special.copyTable.thresholdField, kThresholdFieldWidth, TEXT_HT_BIG);
		}
		
		SizeControl(dataPtr->tabs.special.captureFile.groupBox,
					tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), boxHeightOtherBoxes);
		SizeControl(dataPtr->tabs.special.captureFile.creatorField, kTypeCreatorFieldWidth, TEXT_HT_BIG);
		(UInt16)Localization_AutoSizeButtonControl(dataPtr->tabs.special.captureFile.applicationButton, 0); // make button sufficiently large
		
		// arrange Special tab pane and its controls
		{
			Rect		controlRect;
			UInt16		buttonTop = 0,
						checkBoxTop = 0;
			
			
			topEdge = VSP_GROUPBOXTEXT_AND_CONTROL + SPACE_EXTRA_TEXTEDIT;
			buttonTop = topEdge + INTEGER_HALVED(TEXT_HT_BIG - BUTTON_HT);
			checkBoxTop = topEdge + INTEGER_HALVED(TEXT_HT_BIG - checkBoxHeight);
			
			// arrange the cursor shape label and bevel buttons
			control = dataPtr->tabs.special.cursor.shapeLabel;
			MoveControl(control, HSP_GROUPBOX_AND_CONTROL, topEdge);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.cursor.shapeBlockBevelButton;
			MoveControl(control, controlRect.right + HSP_CONTROLS, buttonTop);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.cursor.shapeVerticalBarBevelButton;
			MoveControl(control, controlRect.right, buttonTop);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.cursor.shapeThickVerticalBarBevelButton;
			MoveControl(control, controlRect.right, buttonTop);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.cursor.shapeUnderlineBevelButton;
			MoveControl(control, controlRect.right, buttonTop);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.cursor.shapeThickUnderlineBevelButton;
			MoveControl(control, controlRect.right, buttonTop);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.cursor.flashingCheckBox;
			MoveControl(control, controlRect.right + HSP_CONTROLS, checkBoxTop);
			
			// arrange the timeout labels and fields
			topEdge = VSP_GROUPBOXTEXT_AND_CONTROL + SPACE_EXTRA_TEXTEDIT;
			control = dataPtr->tabs.special.timeouts.openLabel;
			MoveControl(control, HSP_GROUPBOX_AND_CONTROL, topEdge);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.timeouts.openField;
			MoveControl(control, controlRect.right + HSP_CONTROLS, topEdge);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.timeouts.sendLabel;
			MoveControl(control, controlRect.right + 3 * HSP_CONTROLS/* give some extra space here to separate the timeouts */, topEdge);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.timeouts.sendField;
			MoveControl(control, controlRect.right + HSP_CONTROLS, topEdge);
			GetControlBounds(control, &controlRect);
			
			// arrange the copy table labels and fields
			topEdge = VSP_GROUPBOXTEXT_AND_CONTROL + SPACE_EXTRA_TEXTEDIT;
			control = dataPtr->tabs.special.copyTable.thresholdLabel;
			MoveControl(control, HSP_GROUPBOX_AND_CONTROL, topEdge);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.copyTable.thresholdField;
			MoveControl(control, controlRect.right + HSP_CONTROLS, topEdge);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.copyTable.thresholdUnitsLabel;
			MoveControl(control, controlRect.right + HSP_CONTROLS, topEdge);
			GetControlBounds(control, &controlRect);
			
			// arrange the capture file labels and fields
			topEdge = VSP_GROUPBOXTEXT_AND_CONTROL + SPACE_EXTRA_TEXTEDIT;
			control = dataPtr->tabs.special.captureFile.creatorField;
			MoveControl(control, HSP_GROUPBOX_AND_CONTROL + SPACE_EXTRA_TEXTEDIT, topEdge);
			GetControlBounds(control, &controlRect);
			control = dataPtr->tabs.special.captureFile.applicationButton;
			MoveControl(control, controlRect.right + HSP_CONTROLS, buttonTop);
			GetControlBounds(control, &controlRect);
		}
		
		topEdge = VSP_PANE_AND_CONTROL;
		MoveControl(dataPtr->tabs.special.cursor.groupBox, HSP_PANE_AND_CONTROL, topEdge);
		topEdge += (boxHeightCursorBox + VSP_GROUPBOXES);
		MoveControl(dataPtr->tabs.special.timeouts.groupBox, HSP_PANE_AND_CONTROL, topEdge);
		topEdge += (boxHeightOtherBoxes + VSP_GROUPBOXES);
		MoveControl(dataPtr->tabs.special.copyTable.groupBox, HSP_PANE_AND_CONTROL, topEdge);
		topEdge += (boxHeightOtherBoxes + VSP_GROUPBOXES);
		MoveControl(dataPtr->tabs.special.captureFile.groupBox, HSP_PANE_AND_CONTROL, topEdge);
		topEdge += (boxHeightOtherBoxes + VSP_BUTTON_AND_CONTROL);
		
		control = dataPtr->tabs.special.pane;
		MoveControl(control, tabPaneTopLeft.h, tabPaneTopLeft.v);
		(OSErr)SetControlVisibility(control, true/* visible */, true/* draw */);
	}
	else if (chosenTabIndex == kTabIndexGeneralColors)
	{
		// size Colors tab pane and its controls
		control = dataPtr->tabs.colors.pane;
		SizeControl(control, tabPaneWidthHeight.h, tabPaneWidthHeight.v);
		(UInt16)Localization_AutoSizeButtonControl(dataPtr->tabs.colors.setToDefaultsButton, 0); // make button sufficiently large
		// dividers are sized below, for convenience
		
		// arrange Colors tab pane and its controls
		topEdge = VSP_PANE_AND_CONTROL;
		{
			UInt16		initialTopEdge = 0,
						leftEdge = HSP_PANE_AND_CONTROL;
			UInt16		widestLabelWidth = 0,
						columnWidth = 0;
			Rect		controlBounds;
			
			
			for (i = 0; i < kANSIColorCount; i++)
			{
				// find the widest label
				control = dataPtr->tabs.colors.colorBoxLabels[i];
				GetControlBounds(control, &controlBounds);
				if ((controlBounds.right - controlBounds.left) > widestLabelWidth)
				{
					widestLabelWidth = controlBounds.right - controlBounds.left;
				}
			}
			columnWidth = COLORBOX_WD + HSP_CONTROLS + widestLabelWidth;
			
			// position headers
			control = dataPtr->tabs.colors.setToDefaultsButton;
			MoveControl(control, leftEdge + columnWidth, topEdge);
			control = dataPtr->tabs.colors.headerLabel;
			MoveControl(control, leftEdge, topEdge + INTEGER_HALVED(BUTTON_HT - TEXT_HT_BIG));
			topEdge += (INTEGER_MAXIMUM(BUTTON_HT, TEXT_HT_BIG) + VSP_CONTROLS);
			control = dataPtr->tabs.colors.normalHeaderLabel;
			MoveControl(control, leftEdge, topEdge);
			control = dataPtr->tabs.colors.emphasizedHeaderLabel;
			MoveControl(control, leftEdge + columnWidth + HSP_CONTROLS, topEdge);
			topEdge += (TEXT_HT_BIG + VSP_CONTROLS);
			control = dataPtr->tabs.colors.normalHeaderDivider;
			MoveControl(control, leftEdge, topEdge);
			control = dataPtr->tabs.colors.emphasizedHeaderDivider;
			MoveControl(control, leftEdge + columnWidth + HSP_CONTROLS, topEdge);
			topEdge += (SEPARATOR_HT + VSP_CONTROLS);
			
			initialTopEdge = topEdge;
			for (i = 0; i < kANSIColorCount; i++)
			{
				// position buttons and labels
				if (i == 8)
				{
					topEdge = initialTopEdge;
					leftEdge += (columnWidth + HSP_CONTROLS);
				}
				control = dataPtr->tabs.colors.colorBoxes[i];
				MoveControl(control, leftEdge, topEdge);
				control = dataPtr->tabs.colors.colorBoxLabels[i];
				MoveControl(control, leftEdge + COLORBOX_WD + HSP_CONTROLS, topEdge + INTEGER_HALVED(COLORBOX_HT - TEXT_HT_BIG));
				topEdge += (COLORBOX_HT + VSP_BUTTONS);
			}
			
			// set divider widths
			control = dataPtr->tabs.colors.normalHeaderDivider;
			SizeControl(control, columnWidth, SEPARATOR_HT);
			control = dataPtr->tabs.colors.emphasizedHeaderDivider;
			SizeControl(control, columnWidth, SEPARATOR_HT);
		}
		control = dataPtr->tabs.colors.pane;
		MoveControl(control, tabPaneTopLeft.h, tabPaneTopLeft.v);
		(OSErr)SetControlVisibility(control, true/* visible */, true/* draw */);
	}
	else if (chosenTabIndex == kTabIndexGeneralNotification)
	{
		// size Notification tab pane and its controls
		control = dataPtr->tabs.notification.pane;
		SizeControl(control, tabPaneWidthHeight.h, tabPaneWidthHeight.v);
		control = dataPtr->tabs.notification.visualBellCheckBox;
		SizeControl(control, tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), checkBoxHeight);
		control = dataPtr->tabs.notification.background.terminalBeepsCheckBox;
		SizeControl(control, tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL + HSP_GROUPBOX_AND_CONTROL), checkBoxHeight);
		control = dataPtr->tabs.notification.background.separator;
		SizeControl(control, tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL + HSP_GROUPBOX_AND_CONTROL), SEPARATOR_HT);
		control = dataPtr->tabs.notification.background.helpText;
		{
			Str255		string;
			
			
			GetIndString(string, rStringsHelpText, siHelpTextPrefsGeneralNotification);
			SizeControl(control, tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL + HSP_GROUPBOX_AND_CONTROL), 0);
			textHeight = Localization_SetUpMultiLineTextControl(control, string, kThemeSmallSystemFont); // auto-sizes and sets font
		}
		{
			UInt16		boxHeight = (VSP_GROUPBOXTEXT_AND_CONTROL + checkBoxHeight + VSP_CONTROLS + SEPARATOR_HT + VSP_CONTROLS +
										textHeight + VSP_CONTROLS + TEXT_HT_BIG + VSP_CONTROLS +
										(NUMBER_OF_GENERAL_NOTIFICATION_METHODS - 1) * (radioButtonHeight + VSP_RADIOS) +
										radioButtonHeight + VSP_GROUPBOX_AND_CONTROL);
			
			
			SizeControl(dataPtr->tabs.notification.background.groupBox,
						tabPaneWidthHeight.h - INTEGER_DOUBLED(HSP_PANE_AND_CONTROL), boxHeight);
		}
		for (i = 0; i < NUMBER_OF_GENERAL_NOTIFICATION_METHODS; i++)
		{
			Rect		boxRect;
			
			
			GetControlBounds(dataPtr->tabs.notification.background.groupBox, &boxRect);
			control = dataPtr->tabs.notification.background.methods[i].symbolsPane;
			SizeControl(control, kNotificationSymbolsWidth, ICON_HT_SMALL);
			control = dataPtr->tabs.notification.background.methods[i].radioButton;
			SizeControl(control, boxRect.right - boxRect.left - INTEGER_DOUBLED(HSP_GROUPBOX_AND_CONTROL) -
									kNotificationSymbolsWidth - HSP_CONTROLS,
								radioButtonHeight);
		}
		
		// arrange Notification tab pane and its controls
		topEdge = VSP_PANE_AND_CONTROL;
		MoveControl(dataPtr->tabs.notification.visualBellCheckBox, HSP_PANE_AND_CONTROL, topEdge);
		topEdge += (checkBoxHeight + VSP_GROUPBOX_AND_CONTROL);
		topEdge = VSP_GROUPBOXTEXT_AND_CONTROL;
		MoveControl(dataPtr->tabs.notification.background.terminalBeepsCheckBox, HSP_GROUPBOX_AND_CONTROL, topEdge);
		topEdge += (checkBoxHeight + VSP_CONTROLS);
		MoveControl(dataPtr->tabs.notification.background.separator, HSP_GROUPBOX_AND_CONTROL, topEdge);
		topEdge += (SEPARATOR_HT + VSP_CONTROLS);
		MoveControl(dataPtr->tabs.notification.background.helpText, HSP_GROUPBOX_AND_CONTROL, topEdge);
		topEdge += (textHeight + VSP_CONTROLS);
		MoveControl(dataPtr->tabs.notification.background.whenPostingText, HSP_GROUPBOX_AND_CONTROL, topEdge);
		topEdge += (TEXT_HT_BIG + VSP_CONTROLS);
		{
			Rect	symbolsRect;
			
			
			for (i = 0; i < NUMBER_OF_GENERAL_NOTIFICATION_METHODS; i++)
			{
				control = dataPtr->tabs.notification.background.methods[i].symbolsPane;
				MoveControl(control, HSP_GROUPBOX_AND_CONTROL, topEdge);
				GetControlBounds(control, &symbolsRect);
				control = dataPtr->tabs.notification.background.methods[i].radioButton;
				MoveControl(control, symbolsRect.right + HSP_CONTROLS, topEdge);
				topEdge += ((ICON_HT_SMALL > radioButtonHeight) ? ICON_HT_SMALL : radioButtonHeight);
				if (i < (NUMBER_OF_GENERAL_NOTIFICATION_METHODS - 1)) topEdge += VSP_RADIOS;
			}
		}
		topEdge += (VSP_GROUPBOX_AND_CONTROL);
		MoveControl(dataPtr->tabs.notification.background.groupBox, HSP_PANE_AND_CONTROL,
						VSP_PANE_AND_CONTROL + checkBoxHeight + VSP_GROUPBOX_AND_CONTROL);
		
		control = dataPtr->tabs.notification.pane;
		MoveControl(control, tabPaneTopLeft.h, tabPaneTopLeft.v);
		(OSErr)SetControlVisibility(control, true/* visible */, true/* draw */);
	}
}// setUpChosenGeneralPane


/*!
Creates the controls for the selected tab of
the specified panel.  If the given panel has
no tabs, this method has no effect.

WARNING:	The Preferences dialog box disposes
			of controls that are part of any
			invisible tab panes!  Never reference
			a control until you first use the
			getChosenTabIndex() routine to verify
			that the current tab contains the
			control you wish to modify.  If you
			attempt to access a control on a
			disposed pane, the system will crash.

(3.0)
*/
static void
showChosenTabPane		(PanelRef		inPanel)
{
	UInt16				chosenTab = 0;
	static UInt16		oldChoice = 1;
	
	
	chosenTab = getChosenTabIndex(inPanel);
	if (oldChoice != chosenTab)
	{
		disposeOldTabPane(inPanel, oldChoice);
		oldChoice = chosenTab;
		createChosenGeneralPane(inPanel, Panel_GetOwningWindow(inPanel));
		setUpChosenGeneralPane(inPanel);
	}
}// showChosenTabPane


/*!
Call this method when a click in a checkbox
or radio button has been detected, but the
control value has *not* yet been toggled.
Upon exit from this routine, the checkbox
state will have been toggled (or, for a
radio button, the group value will change).
In addition, the appropriate preference
will have been updated in memory using the
Preferences_SetData() method.

If the specified control is known, "true"
is returned to indicate that the click was
handled.  Otherwise, "false" is returned.

(3.0)
*/
static Boolean
updateCheckBoxPreference	(PanelRef			inPanel,
							 ControlRef			inCheckBoxClicked)
{
	Boolean					result = false;
	GeneralPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	if (dataPtr != null)
	{
		Boolean					checkBoxFlagValue = // NOTE - the control value hasnt changed yet, so this flag assignment is correct!
									(GetControlValue(inCheckBoxClicked) == kControlCheckBoxCheckedValue) ? false : true;
		register SInt16			i = 0;
		UInt16					chosenTabIndex = getChosenTabIndex(inPanel);
		
		
		switch (chosenTabIndex)
		{
			case kTabIndexGeneralOptions:
				{
					SInt16		chosenOptionIndex = -1;
					
					
					// determine which Options checkbox was hit
					for (i = 0; i < NUMBER_OF_GENERAL_OPTIONS; i++)
					{
						if (inCheckBoxClicked == dataPtr->tabs.options.options[i].checkBox)
						{
							chosenOptionIndex = i;
							result = true;
							break;
						}
					}
					// ...NOW toggle the control value...
					SetControlValue(inCheckBoxClicked, BooleanToCheckBoxValue(checkBoxFlagValue));
					
					switch (chosenOptionIndex)
					{
						case kGeneralOptionIndexSimpleTelnet:
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_SimpleTelnet, sizeof(Boolean),
												&checkBoxFlagValue);
							Menus_CheckSimpleTelnet(checkBoxFlagValue); // enters or leaves Simple Telnet mode
							break;
						
						case kGeneralOptionIndexDontAutoClose:
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_DontAutoClose, sizeof(Boolean),
												&checkBoxFlagValue);
							break;
						
						case kGeneralOptionIndexDynamicResizing:
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_DynamicResizing, sizeof(Boolean),
												&checkBoxFlagValue);
							break;
						
						case kGeneralOptionIndexHeadersInitiallyCollapsed:
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_HeadersCollapsed, sizeof(Boolean),
												&checkBoxFlagValue);
							break;
						
						#if 0
						case kGeneralOptionIndexUsePicture:
							// update the preference value and then update every visible screen to erase/redraw its background
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_UseBackgroundPicture, sizeof(Boolean),
												&checkBoxFlagValue);
							ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagNonSessions |
												kScreenManager_FilterFlagDeadSessions |
												kScreenManager_FilterFlagActiveSessions,
												changePreferenceUpdateScreen,
												(SInt32)kTelnetPreferenceTag_UseBackgroundPicture, 0L/* undefined */,
												null/* undefined */);
							break;
						#endif
						
						case kGeneralOptionIndexDontDimBackgroundScreens:
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_DontDimBackgroundScreens, sizeof(Boolean),
												&checkBoxFlagValue);
							ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagNonSessions |
												kScreenManager_FilterFlagDeadSessions |
												kScreenManager_FilterFlagActiveSessions,
												changePreferenceUpdateScreen,
												(SInt32)kTelnetPreferenceTag_DontDimBackgroundScreens, 0L/* undefined */,
												null/* undefined */);
							break;
						
						case kGeneralOptionIndexPureInverse:
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_PureInverse, sizeof(Boolean),
												&checkBoxFlagValue);
							break;
						
						case kGeneralOptionIndexCopySelectedText:
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CopySelectedText, sizeof(Boolean),
												&checkBoxFlagValue);
							break;
						
						case kGeneralOptionIndexCursorMovesToDropLocation:
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CursorMovesPriorToDrops, sizeof(Boolean),
												&checkBoxFlagValue);
							break;
						
						case kGeneralOptionIndexMenuItemKeys:
							Menus_SetKeyEquivalents(checkBoxFlagValue, false/* first call */);
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_MenuItemKeys, sizeof(Boolean),
												&checkBoxFlagValue);
							break;
						
						case kGeneralOptionIndexMapBackquote:
							Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_MapBackquote, sizeof(Boolean),
												&checkBoxFlagValue);
							break;
						
						default:
							// unrecognized checkbox
							break;
					}
				}
				break;
			
			case kTabIndexGeneralSpecial:
				if (inCheckBoxClicked == dataPtr->tabs.special.cursor.flashingCheckBox)
				{
					// toggle control value and save the new value
					SetControlValue(inCheckBoxClicked, BooleanToCheckBoxValue(checkBoxFlagValue));
					Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CursorBlinks, sizeof(Boolean),
										&checkBoxFlagValue);
					result = true;
				}
				break;
			
			case kTabIndexGeneralNotification:
				{
					SInt16		chosenNotificationTypeIndex = -1;
					
					
					for (i = 0; i < NUMBER_OF_GENERAL_NOTIFICATION_METHODS; i++)
					{
						// is this radio button the one that was clicked?
						if (inCheckBoxClicked == dataPtr->tabs.notification.background.methods[i].radioButton)
						{
							chosenNotificationTypeIndex = i;
							result = true;
						}
					}
					
					// check if any of the notification-method radio buttons was hit; if not, check for other control hits
					if (chosenNotificationTypeIndex >= 0)
					{
						SInt16		newNotificationType = kAlert_NotifyDoNothing;
						
						
						// set the radio button
						for (i = 0; i < NUMBER_OF_GENERAL_NOTIFICATION_METHODS; i++)
						{
							SetControlValue(dataPtr->tabs.notification.background.methods[i].radioButton, kControlRadioButtonUncheckedValue);
						}
						SetControlValue(inCheckBoxClicked, kControlRadioButtonCheckedValue);
						
						switch (chosenNotificationTypeIndex)
						{
							case kGeneralNotificationMethodIndexNone:
								newNotificationType = kAlert_NotifyDoNothing;
								break;
							
							case kGeneralNotificationMethodIndexDiamond:
								newNotificationType = kAlert_NotifyDisplayDiamondMark;
								break;
							
							case kGeneralNotificationMethodIndexFlashIcon:
								newNotificationType = kAlert_NotifyDisplayIconAndDiamondMark;
								break;
							
							case kGeneralNotificationMethodIndexDisplayAlert:
								newNotificationType = kAlert_NotifyAlsoDisplayAlert;
								break;
							
							default:
								// unrecognized radio button
								break;
						}
						Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_Notification, sizeof(SInt16), &newNotificationType);
					}
					else if (inCheckBoxClicked == dataPtr->tabs.notification.visualBellCheckBox)
					{
						// toggle control value and save the new value
						SetControlValue(inCheckBoxClicked, BooleanToCheckBoxValue(checkBoxFlagValue));
						Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_VisualBell, sizeof(Boolean),
											&checkBoxFlagValue);
						result = true;
					}
					else if (inCheckBoxClicked == dataPtr->tabs.notification.background.terminalBeepsCheckBox)
					{
						// toggle control value and save the new value
						SetControlValue(inCheckBoxClicked, BooleanToCheckBoxValue(checkBoxFlagValue));
						Preferences_SetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_NotifyOfBeeps, sizeof(Boolean),
											&checkBoxFlagValue);
						result = true;
					}
				}
				break;
			
			default:
				break;
		}
	}
	
	return result;
}// updateCheckBoxPreference


/*!
If an editable text control is to represent a
number, use this method to assign a key filter
to it that prevents non-digits from being
entered into such a field.

(3.0)
*/
static void
useNumericalLimiter	(ControlRef		inControl)
{
	ControlKeyFilterUPP		upp = NewControlKeyFilterUPP(NumericalLimiter);
	
	
	(OSErr)SetControlData(inControl, kControlNoPart, kControlKeyFilterTag, sizeof(ControlKeyFilterUPP), (Ptr)&upp);
}// useNumericalLimiter


/*!
If an editable text control is to represent an
OSType (which is four characters), use this
method to assign a key filter to it that prevents
more than four characters from being typed in
such a field.

(3.0)
*/
static void
useOSTypeLengthLimiter	(ControlRef		inControl)
{
	ControlKeyFilterUPP		upp = NewControlKeyFilterUPP(OSTypeLengthLimiter);
	
	
	(OSErr)SetControlData(inControl, kControlNoPart, kControlKeyFilterTag, sizeof(ControlKeyFilterUPP), (Ptr)&upp);
}// useOSTypeLengthLimiter