/*###############################################################

	PrefPanelMacros.c
	
	MacTelnet
		 1998-2003 by Kevin Grant.
		 2001-2003 by Ian Anderson.
		 1986-1994 University of Illinois Board of Trustees
		(see About box for full list of U of I contributors).
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.
	
	You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstring>
#else
#	include <string.h>
#endif

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Dialogs.h>
#include <Folders.h>
#include <Icons.h>
#include <Resources.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "StringResources.h"
#include "SpacingConstants.r"
#include "TelnetPrefsDialog.r"

// MacTelnet includes
#include "DialogUtilities.h"
#include "MacLook.h"
#include "MacroSetupDialog.h"
#include "Panel.h"
#include "Preferences.h"
#include "PreferencesData.h"
#include "PreferencesDialog.h"
#include "PrefPanelMacros.h"
#include "TelnetDescriptors.h"
#include "TelnetMacros.h"



// types
struct MacrosPanelData
{
	struct
	{
		ControlRef		tabControl,
						tabPane,
						keysLabel,						// label for the macro keys radio button group
						commandDigitRadioButton,
						functionKeysRadioButton,
						activeSetLabel,					// label for the current macro set
						activeSetText,					// the name of the current macro set
						setActiveSetButton,				// the Use Set Above button
						editButton; // temporary - this panel will eventually be MUCH more sophisticated...
	} controls;
	
	FinderListRef		macrosFinderList;
};
typedef struct MacrosPanelData	MacrosPanelData;
typedef MacrosPanelData*		MacrosPanelDataPtr;

// internal methods
static void				createPanelControls				(PanelRef			inPanel,
														 WindowRef			inOwningWindow);
static void				deltaPanelSize					(PanelRef			inPanel,
														 const Point		inChangeInWidthAndHeight);
static void				disposePanel					(PanelRef			inPanel,
														 void*				inDataPtr);
static SInt32			handleContentClick				(PanelRef			inPanel,
														 ControlRef			inControl);
static SInt32			panelChanged					(PanelRef			inPanel,
														 PanelMessage		inMessage,
														 void*				inData);
static void				rebuildDisplay					(MacrosPanelDataPtr	inDataPtr);
static void				setUpMacrosControls				(PanelRef			inPanel);



//
// public methods
//

/*!
This routine creates a new preference panel for
the Macros category, initializes it, and returns
a reference to it.  You must destroy the reference
using Panel_Dispose() when you are done with it.

If any problems occur, null is returned.

(3.0)
*/
PanelRef
PrefPanelMacros_New		(void)
{
	PanelRef		result = Panel_New(panelChanged);
	
	
	if (result != null)
	{
		MacrosPanelDataPtr		dataPtr = (MacrosPanelDataPtr)Memory_NewPtr(sizeof(MacrosPanelData));
		Str31					name;
		
		
		Panel_SetDescriptor(result, kTelnetPanelDescriptorPreferencesMacros);
		GetIndString(name, rStringsConfigurationData, siConfigurationDataPreferencePanelNameMacros);
		Panel_SetName(result, name);
		Panel_SetIconSuite(result, kIconSuitePrefsMacros, kSelectorAllLargeData);
		Panel_SetAuxiliaryDataPtr(result, dataPtr);
	}
	return result;
}// New


//
// internal methods
//
#pragma mark -

/*!
This routine creates the controls that belong in the
Macros panel, and attaches them to the specified
owning window in the proper embedding hierarchy.  The
specified panels descriptor must be
"kTelnetPanelDescriptorPreferencesMacros".

The controls are not arranged or sized.

(3.0)
*/
static void
createPanelControls		(PanelRef		inPanel,
						 WindowRef		inOwningWindow)
{
	ControlRef				container = GetNewControl(kIDPanelUserPane, inOwningWindow),
							control = null;
	MacrosPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	Panel_SetContainerControl(inPanel, container);
	SetControlVisibility(container, false/* visible */, false/* draw */);
	
	//temp
	dataPtr->controls.editButton = control = GetNewControl(kIDPrefConfigurationsEditButton, inOwningWindow);
	(OSStatus)EmbedControl(control, container);
	
	// tab control
	dataPtr->controls.tabControl = control = GetNewControl(kIDPrefMacrosTabs, inOwningWindow);
	(OSStatus)EmbedControl(control, container);
	(OSStatus)SetControlVisibility(control, true/* visible */, true/* draw */); // necessary?
	SetControlValue(dataPtr->controls.tabControl, Macros_GetActiveSetNumber());
	
	// tab pane
	dataPtr->controls.tabPane = control = GetNewControl(kIDTabUserPane, inOwningWindow);
	(OSStatus)EmbedControl(control, dataPtr->controls.tabControl);
	(OSStatus)SetControlVisibility(control, true/* visible */, true/* draw */);
	
	// create content controls
	dataPtr->controls.commandDigitRadioButton = control = GetNewControl(kIDRadioButton, inOwningWindow);
	(OSStatus)EmbedControl(control, container);
	dataPtr->controls.functionKeysRadioButton = control = GetNewControl(kIDRadioButton, inOwningWindow);
	(OSStatus)EmbedControl(control, container);
	
	// create the macro list; insert appropriate columns
	{
		FinderListRef				list = null;
		FinderListColumnDataType	cellTypes[] =
									{
										kFinderListColumnDataTypePlainText,
										kFinderListColumnDataTypePlainText,
										kFinderListColumnDataTypePlainText
									};
		register SInt16				columnNumber = 0;
		
		
		dataPtr->macrosFinderList = list = FinderList_New(inOwningWindow, kFinderListFlagsInactiveItemsDimmed/* flags */);
		(UInt16)FinderList_InsertColumn(list, -1/* insert at end */,
										cellTypes[columnNumber++]/* column cell data type */,
										kFinderListColumnFlagsActiveSort |
										kFinderListColumnFlagsNoHighlighting/* flags */,
										"\pKey(s)"/*temp(should be localizable)*//* title */);
		(UInt16)FinderList_InsertColumn(list, -1/* insert at end */,
										cellTypes[columnNumber++]/* column cell data type */,
										kFinderListColumnFlagsReadWrite/* flags */,
										"\pMacro Name"/*temp(should be localizable)*//* title */);
		(UInt16)FinderList_InsertColumn(list, -1/* insert at end */,
										cellTypes[columnNumber++]/* column cell data type */,
										kFinderListColumnFlagsReadWrite |
										kFinderListColumnFlagsNoHighlighting/* flags */,
										"\pEquivalent"/*temp(should be localizable)*//* title */);
		//FinderList_SetColumnTitle(list, 1, "\pMacro Name");
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->macrosFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		
		rebuildDisplay(dataPtr);
	}
	(OSErr)EmbedControl(FinderList_GetContainerControl(dataPtr->macrosFinderList), dataPtr->controls.tabPane);
}// createPanelControls


/*!
This routine adjusts the controls in preference panels
to match the specified change in dimensions of a panels
container.

(3.0)
*/
static void
deltaPanelSize		(PanelRef		inPanel,
					 Point			inChangeInWidthAndHeight)
{
	ControlRef				control = null;
	MacrosPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	Rect					controlRect;
	
	
	// move and/or resize controls
	#if 1
	control = dataPtr->controls.tabControl;
	GetControlBounds(control, &controlRect);
	SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
						controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
	control = dataPtr->controls.tabPane;
	GetControlBounds(control, &controlRect);
	SizeControl(control, controlRect.right - controlRect.left + inChangeInWidthAndHeight.h,
						controlRect.bottom - controlRect.top + inChangeInWidthAndHeight.v);
	#else
	// try using the Dialog Adjust module to smooth out resizing this panel
	DialogAdjust_BeginControlAdjustment(Panel_GetOwningWindow(inPanel));
	control = dataPtr->controls.tabControl;
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, control, inChangeInWidthAndHeight.h);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeV, control, inChangeInWidthAndHeight.v);
	control = dataPtr->controls.tabPane;
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, control, inChangeInWidthAndHeight.h);
	DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeV, control, inChangeInWidthAndHeight.v);
	DialogAdjust_EndAdjustment(inChangeInWidthAndHeight.h, inChangeInWidthAndHeight.v);
	#endif
	FinderList_DeltaSizeControl(dataPtr->macrosFinderList, inChangeInWidthAndHeight.h, inChangeInWidthAndHeight.v);
}// deltaPanelSize


/*!
This routine responds to a change in the existence
of a panel.  The panel is about to be destroyed, so
this routine disposes of private data structures
associated with the specified panel.

(3.0)
*/
static void
disposePanel	(PanelRef		UNUSED_ARGUMENT(inPanel),
				 void*			inDataPtr)
{
	MacrosPanelDataPtr		dataPtr = inDataPtr;
	
	
	FinderList_Dispose(&dataPtr->macrosFinderList);
	Memory_DisposePtr((Ptr*)&dataPtr);
}// disposePanel


/*!
This method responds to mouse-down events in the panel.

(3.0)
*/
static SInt32
handleContentClick	(PanelRef		inPanel,
					 ControlRef		inControl)
{
	SInt32					result = kPanel_ResponseClickNotHandled;
	Boolean					handled = false;
	MacrosPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	if (inControl == dataPtr->controls.editButton)
	{
		MacroSetupDialog_Init();
		
		// update the list if the user changed anything
		if (MacroSetupDialog_Display())
		{
			SetControlValue(dataPtr->controls.tabControl, Macros_GetActiveSetNumber());
			rebuildDisplay(dataPtr); // this reads the tab value and draws the proper list items
		}
		
		MacroSetupDialog_Done();
		handled = true;
	}
	else if (inControl == dataPtr->controls.tabControl)
	{
		// update the list to display the users choice of macro set (and change the active set)
		Macros_SetActiveSetNumber(GetControlValue(dataPtr->controls.tabControl));
		rebuildDisplay(dataPtr); // this reads the tab value and draws the proper list items
	}
	
	if (handled) result = kPanel_ResponseClickHandled;
	
	return result;
}// handleContentClick


/*!
This routine, of standard PanelChangedProcPtr form,
is invoked by the Panel module whenever a property
of one of the preferences dialogs panels changes.

(3.0)
*/
static SInt32
panelChanged	(PanelRef			inPanel,
				 PanelMessage		inMessage,
				 void*				inDataPtr)
{
	SInt32		result = 0L;
	
	
	if (Panel_GetDescriptor(inPanel) == kTelnetPanelDescriptorPreferencesMacros) switch (inMessage)
	{
		case kPanel_MessageClicked: // notification of a content click
			{
				ControlRef*			dataPtr = inDataPtr;
				
				
				result = handleContentClick(inPanel, *dataPtr);
			}
			break;
		
		case kPanel_MessageCreateControls: // specification of the window containing the panel - create controls using this window
			{
				WindowRef const*		windowPtr = REINTERPRET_CAST(inDataPtr, WindowRef*);
				
				
				createPanelControls(inPanel, *windowPtr);
			}
			break;
		
		case kPanel_MessageDeltaSize: // request to alter size and/or position of controls; containing control will already have been changed
			{
				Point			changeInPanelDimensions;
				ControlRef		panelControl = null;
				Boolean			wasVisible = false;
				
				
				SetPt(&changeInPanelDimensions, ((Point*)inDataPtr)->h, ((Point*)inDataPtr)->v);
				Panel_GetContainerControl(inPanel, &panelControl);
				wasVisible = IsControlVisible(panelControl);
				SetControlVisibility(panelControl, false/* visibility */, wasVisible/* draw */);
				deltaPanelSize(inPanel, changeInPanelDimensions);
				SetControlVisibility(panelControl, wasVisible/* visibility */, wasVisible/* draw */);
			}
			break;
		
		case kPanel_MessageDestroyed: // request to dispose of private data structures
			{
				void*		panelAuxiliaryDataPtr = inDataPtr;
				
				
				disposePanel(inPanel, panelAuxiliaryDataPtr);
			}
			break;
		
		case kPanel_MessageFocusGained: // notification that a control is now focused
			{
				ControlRef const*	controlPtr = REINTERPRET_CAST(inDataPtr, ControlRef*);
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageFocusLost: // notification that a control is no longer focused
			{
				ControlRef const*	controlPtr = REINTERPRET_CAST(inDataPtr, ControlRef*);
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageInitialSize: // request for panel to size its controls
			{
				Point		exactPanelDimensions;
				
				
				SetPt(&exactPanelDimensions, ((Point*)inDataPtr)->h, ((Point*)inDataPtr)->v);
				setUpMacrosControls(inPanel);
			}
			break;
		
		case kPanel_MessageNewActiveState: // request to change panel controls to a new active (true) or inactive (false) state
			{
				Boolean		isNowActive = *((Boolean*)inDataPtr);
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageNewAppearanceTheme: // notification of theme switch, a request to recalculate control sizes
			{
				MacrosPanelDataPtr		panelDataPtr = REINTERPRET_CAST(Panel_GetAuxiliaryDataPtr(inPanel), MacrosPanelDataPtr);
				Rect					controlRect;
				
				
				GetControlBounds(FinderList_GetContainerControl(panelDataPtr->macrosFinderList), &controlRect);
				FinderList_SizeControl(panelDataPtr->macrosFinderList, controlRect.right - controlRect.left,
										controlRect.bottom - controlRect.top);
			}
			break;
		
		case kPanel_MessageNewVisibility: // visible state of the panels container has changed to visible (true) or invisible (false)
			{
				MacrosPanelDataPtr	panelDataPtr = REINTERPRET_CAST(Panel_GetAuxiliaryDataPtr(inPanel), MacrosPanelDataPtr);
				Boolean				isNowVisible = *((Boolean*)inDataPtr);
				
				
				// hack - on pre-Mac OS 9 systems, the pesky Edit... buttons sticks around for some reason; explicitly show/hide it
				//SetControlVisibility(panelDataPtr->controls.editButton, isNowVisible/* visibility */, true/* draw */);
			}
			break;
		
		default:
			break;
	}
	
	return result;
}// panelChanged


/*!
Rebuilds the Finder List display, copying the current
macro key equivalents, etc. into the list.  Since this
operation is not trivial, and since inline editing is
not yet implemented, this routine is available to tell
the panel when the source of its list data has actually
changed.

WARNING:	Do not pass a null pointer to this routine!

(3.0)
*/
static void
rebuildDisplay		(MacrosPanelDataPtr		inDataPtr)
{
	FinderListRef			list = inDataPtr->macrosFinderList;
	TelnetMacroMode			macroKeys = Macros_GetMode();
	UInt16					oldSetNumber = Macros_GetActiveSetNumber();
	MacroSet				macroSet = null;
	char					data[255];
	register SInt16			rowNumber = 0,
							columnNumber = 0;
	
	
	Macros_SetActiveSetNumber(GetControlValue(inDataPtr->controls.tabControl));
	macroSet = Macros_GetActiveSet();
	
	// temp - test data display
	columnNumber = rowNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F1" : "cmd-0");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 1"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F2" : "cmd-1");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 2"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F3" : "cmd-2");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 3"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F4" : "cmd-3");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 4"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F5" : "cmd-4");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 5"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F6" : "cmd-5");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 6"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F7" : "cmd-6");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 7"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F8" : "cmd-7");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 8"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F9" : "cmd-8");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 9"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F10" : "cmd-9");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 10"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F11" : "cmd-=");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 11"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, (macroKeys == kTelnetMacroModeFunctionKeys) ? "F12" : "cmd-/");
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "macro 12"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	Macros_Get(macroSet, rowNumber, data, sizeof(data));
	FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	// redraw list contents
	{
		RgnHandle		region = FinderList_GetNewComponentRegion(list, kFinderListComponentTypeListContents);
		
		
		if (region != null)
		{
			MacLook_InvalidateRegion(GetControlOwner(inDataPtr->controls.tabControl), region);
			Memory_DisposeRegion(&region);
		}
	}
	
	Macros_SetActiveSetNumber(oldSetNumber);
}// rebuildDisplay


/*!
This routine sizes and arranges controls in preference
panels to fit the prescribed dimensions of a panels
container.

(3.0)
*/
static void
setUpMacrosControls		(PanelRef	inPanel)
{
	enum
	{
		// temp - make room for Edit... button
		kTempButtonOffset = (VSP_BUTTON_AND_DIALOG + (BUTTON_HT + VSP_BUTTON_AND_CONTROL))
	};
	MacrosPanelDataPtr		dataPtr = (MacrosPanelDataPtr)Panel_GetAuxiliaryDataPtr(inPanel);
	ControlRef				container = null,
							control = null;
	Rect					containerRect,
							controlRect;
	Point					tabPaneTopLeft,
							tabPaneWidthHeight;
	register SInt16			i = 0;
	UInt16					leftEdge = 0, // cumulative, resettable measurement allowing controls to sandwich horizontally
							topEdge = 0; // cumulative, resettable measurement allowing controls to stack vertically
	
	
	Panel_GetContainerControl(inPanel, &container);
	GetControlBounds(container, &containerRect);
	
	PreferencesDialog_CalculateTabPane(inPanel, &tabPaneTopLeft, &tabPaneWidthHeight);
	tabPaneWidthHeight.v -= kTempButtonOffset;
	
	// size controls
	control = dataPtr->controls.editButton;
	(UInt16)Localization_AutoSizeButtonControl(control, 0); // make button sufficiently large
	control = dataPtr->controls.tabControl;
	SizeControl(control, containerRect.right - containerRect.left,
						containerRect.bottom - containerRect.top - kTempButtonOffset);
	control = dataPtr->controls.tabPane;
	SizeControl(control, tabPaneWidthHeight.h, tabPaneWidthHeight.v);
	FinderList_SetColumnWidthInPixels(dataPtr->macrosFinderList, 2, 160);
	FinderList_SizeControl(dataPtr->macrosFinderList, tabPaneWidthHeight.h, tabPaneWidthHeight.v);
	
	// arrange controls
	topEdge = VSP_BUTTON_AND_DIALOG;
	control = dataPtr->controls.editButton;
	MoveControl(control, HSP_BUTTON_AND_DIALOG, topEdge);
	GetControlBounds(control, &controlRect);
	topEdge += (controlRect.bottom - controlRect.top + VSP_BUTTON_AND_CONTROL);
	control = FinderList_GetContainerControl(dataPtr->macrosFinderList);
	MoveControl(control, 0, 0);
	control = dataPtr->controls.tabPane;
	MoveControl(control, tabPaneTopLeft.h, tabPaneTopLeft.v);
	control = dataPtr->controls.tabControl;
	MoveControl(control, 0, topEdge);
}// setUpMacrosControls