/*###############################################################

	PrefPanelScripts.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstring>
#else
#	include <string.h>
#endif

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Dialogs.h>
#include <Folders.h>
#include <Icons.h>
#include <Resources.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "StringResources.h"
#include "SpacingConstants.r"
#include "TelnetPrefsDialog.r"

// MacTelnet includes
#include "DialogUtilities.h"
#include "MacLook.h"
#include "Panel.h"
#include "Preferences.h"
#include "PreferencesData.h"
#include "PrefPanelScripts.h"
#include "TelnetDescriptors.h"



// types
struct ScriptsPanelData
{
	ControlRef		tabControl;
	
	union
	{
		struct
		{
			ControlRef		pane;
			
			struct
			{
				ControlRef		groupBox,
								populateLabel,
								setScriptsMenuFolderLocationButton,
								scriptsMenuFolderLocationText,
								useTextMenuTitleCheckBox;
			} scriptsMenu;
		
		} special;
		
		struct
		{
			ControlRef		pane;
			
			struct
			{
				ControlRef		instructionsText,
								descriptionGroupBox,
								descriptionText;
			} controls;
			
			FinderListRef	mappingsFinderList;
		} attachedScripts;
	} tabs;
};
typedef struct ScriptsPanelData	ScriptsPanelData;
typedef ScriptsPanelData*		ScriptsPanelDataPtr;

// internal methods
static void				createPanelControls				(PanelRef				inPanel,
														 WindowRef				inOwningWindow);
static void				deltaPanelSize					(PanelRef				inPanel,
														 const Point			inChangeInWidthAndHeight);
static void				disposePanel					(PanelRef				inPanel,
														 void*					inDataPtr);
static SInt32			handleContentClick				(PanelRef				inPanel,
														 ControlRef				inControl);
static SInt32			panelChanged					(PanelRef				inPanel,
														 PanelMessage			inMessage,
														 void*					inData);
static void				rebuildDisplay					(ScriptsPanelDataPtr	inDataPtr);
static void				setUpPanelControls				(PanelRef				inPanel);



//
// public methods
//

/*!
This routine creates a new preference panel for
the Scripts category, initializes it, and returns
a reference to it.  You must destroy the reference
using Panel_Dispose() when you are done with it.

If any problems occur, null is returned.

(3.0)
*/
PanelRef
PrefPanelScripts_New		(void)
{
	PanelRef		result = Panel_New(panelChanged);
	
	
	if (result != null)
	{
		ScriptsPanelDataPtr		dataPtr = (ScriptsPanelDataPtr)Memory_NewPtr(sizeof(ScriptsPanelData));
		Str31					name;
		
		
		Panel_SetDescriptor(result, kTelnetPanelDescriptorPreferencesScripts);
		GetIndString(name, rStringsConfigurationData, siConfigurationDataPreferencePanelNameScripts);
		Panel_SetName(result, name);
		Panel_SetIconSuite(result, kIconSuitePrefsScripts, kSelectorAllLargeData);
		Panel_SetAuxiliaryDataPtr(result, dataPtr);
	}
	return result;
}// New


//
// internal methods
//

/*!
This routine creates the controls that belong in the
Scripts panel, and attaches them to the specified
owning window in the proper embedding hierarchy.  The
specified panels descriptor must be
"kTelnetPanelDescriptorPreferencesScripts".

The controls are not arranged or sized.

(3.0)
*/
static void
createPanelControls		(PanelRef		inPanel,
						 WindowRef		inOwningWindow)
{
	ControlRef				container = GetNewControl(kIDPanelUserPane, inOwningWindow),
							control = null;
	ScriptsPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	Panel_SetContainerControl(inPanel, container);
	SetControlVisibility(container, false/* visible */, false/* draw */);
	
	// create the scripts list; insert appropriate columns
	{
		FinderListRef				list = null;
		FinderListColumnDataType	cellTypes[] = { kFinderListColumnDataTypePlainText, kFinderListColumnDataTypePlainText };
		register SInt16				columnNumber = 0;
		
		
		dataPtr->tabs.attachedScripts.mappingsFinderList = list = FinderList_New(inOwningWindow, kFinderListFlagsNoFocusRing |
																					kFinderListFlagsSizeBoxSpace |
																					kFinderListFlagsInactiveItemsDimmed/* flags */);
		(UInt16)FinderList_InsertColumn(list, -1/* insert at end */,
										cellTypes[columnNumber++]/* column cell data type */,
										kFinderListColumnFlagsActiveSort |
										kFinderListColumnFlagsNoHighlighting/* flags */,
										"\pEvent"/*temp(should be localizable)*//* title */);
		(UInt16)FinderList_InsertColumn(list, -1/* insert at end */,
										cellTypes[columnNumber++]/* column cell data type */,
										kFinderListColumnFlagsReadWrite |
										kFinderListColumnFlagsNoHighlighting/* flags */,
										"\pScript To Run"/*temp(should be localizable)*//* title */);
		FinderList_InsertRow(dataPtr->tabs.attachedScripts.mappingsFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		FinderList_InsertRow(dataPtr->tabs.attachedScripts.mappingsFinderList, -1/* index; -1 means end of list */, 0L/* flags */);
		
		rebuildDisplay(dataPtr);
	}
	(OSErr)EmbedControl(FinderList_GetContainerControl(dataPtr->tabs.attachedScripts.mappingsFinderList), container);
}// createPanelControls


/*!
This routine adjusts the controls in preference panels
to match the specified change in dimensions of a panels
container.

(3.0)
*/
static void
deltaPanelSize		(PanelRef			inPanel,
					 const Point		inChangeInWidthAndHeight)
{
	ScriptsPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	// move and/or resize controls
	FinderList_DeltaSizeControl(dataPtr->tabs.attachedScripts.mappingsFinderList, inChangeInWidthAndHeight.h, inChangeInWidthAndHeight.v);
}// deltaPanelSize


/*!
This routine responds to a change in the existence
of a panel.  The panel is about to be destroyed, so
this routine disposes of private data structures
associated with the specified panel.

(3.0)
*/
static void
disposePanel	(PanelRef		UNUSED_ARGUMENT(inPanel),
				 void*			inDataPtr)
{
	ScriptsPanelDataPtr		dataPtr = inDataPtr;
	
	
	FinderList_Dispose(&dataPtr->tabs.attachedScripts.mappingsFinderList);
	Memory_DisposePtr((Ptr*)&dataPtr);
}// disposePanel


/*!
This method responds to mouse-down events in the panel.

(3.0)
*/
static SInt32
handleContentClick	(PanelRef		inPanel,
					 ControlRef		UNUSED_ARGUMENT(inControl))
{
	SInt32					result = kPanel_ResponseClickNotHandled;
	Boolean					handled = false;
	ScriptsPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	
	
	if (handled) result = kPanel_ResponseClickHandled;
	
	return result;
}// handleContentClick


/*!
This routine, of standard PanelChangedProcPtr form,
is invoked by the Panel module whenever a property
of one of the preferences dialogs panels changes.

(3.0)
*/
static SInt32
panelChanged	(PanelRef			inPanel,
				 PanelMessage		inMessage,
				 void*				inData)
{
	SInt32		result = 0L;
	
	
	if (Panel_GetDescriptor(inPanel) == kTelnetPanelDescriptorPreferencesScripts) switch (inMessage)
	{
		case kPanel_MessageClicked: // notification of a content click
			{
				ControlRef*			dataPtr = inData;
				
				
				result = handleContentClick(inPanel, *dataPtr);
			}
			break;
		
		case kPanel_MessageCreateControls: // specification of the window containing the panel - create controls using this window
			{
				WindowRef	window = *((WindowRef*)inData);
				
				
				createPanelControls(inPanel, window);
			}
			break;
		
		case kPanel_MessageDeltaSize: // request to alter size and/or position of controls; containing control will already have been changed
			{
				Point			changeInPanelDimensions;
				ControlRef		panelControl = null;
				Boolean			wasVisible = false;
				
				
				SetPt(&changeInPanelDimensions, ((Point*)inData)->h, ((Point*)inData)->v);
				Panel_GetContainerControl(inPanel, &panelControl);
				wasVisible = IsControlVisible(panelControl);
				SetControlVisibility(panelControl, false/* visibility */, wasVisible/* draw */);
				deltaPanelSize(inPanel, changeInPanelDimensions);
				SetControlVisibility(panelControl, wasVisible/* visibility */, wasVisible/* draw */);
			}
			break;
		
		case kPanel_MessageDestroyed: // request to dispose of private data structures
			{
				void*		dataPtr = inData;
				
				
				disposePanel(inPanel, dataPtr);
			}
			break;
		
		case kPanel_MessageFocusGained: // notification that a control is now focused
			{
				ControlRef*		dataPtr = inData;
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageFocusLost: // notification that a control is no longer focused
			{
				ControlRef*		dataPtr = inData;
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageInitialSize: // request for panel to size its controls
			{
				Point		exactPanelDimensions;
				
				
				SetPt(&exactPanelDimensions, ((Point*)inData)->h, ((Point*)inData)->v);
				setUpPanelControls(inPanel);
			}
			break;
		
		case kPanel_MessageNewActiveState: // request to change panel controls to a new active (true) or inactive (false) state
			{
				Boolean		isNowActive = *((Boolean*)inData);
				
				
				// do nothing
			}
			break;
		
		case kPanel_MessageNewAppearanceTheme: // notification of theme switch, a request to recalculate control sizes
			{
				ScriptsPanelDataPtr		panelDataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
				Rect					controlRect;
				
				
				GetControlBounds(FinderList_GetContainerControl(panelDataPtr->tabs.attachedScripts.mappingsFinderList), &controlRect);
				FinderList_SizeControl(panelDataPtr->tabs.attachedScripts.mappingsFinderList, controlRect.right - controlRect.left,
										controlRect.bottom - controlRect.top);
			}
			break;
		
		case kPanel_MessageNewVisibility: // visible state of the panels container has changed to visible (true) or invisible (false)
			{
				Boolean		isNowVisible = *((Boolean*)inData);
				
				
				// do nothing
			}
			break;
		
		default:
			break;
	}
	
	return result;
}// panelChanged


/*!
Rebuilds the Finder List display for attached scripts.

WARNING:	Do not pass a null pointer to this routine!

(3.0)
*/
static void
rebuildDisplay		(ScriptsPanelDataPtr		inDataPtr)
{
	FinderListRef			list = inDataPtr->tabs.attachedScripts.mappingsFinderList;
	char					data[255];
	register SInt16			rowNumber = 0,
							columnNumber = 0;
	
	
	// LOCALIZE THESE
	columnNumber = rowNumber = 0;
	
	strcpy(data, "close connection"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "none"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	strcpy(data, "open connection"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	strcpy(data, "none"), FinderList_SetCellData(list, rowNumber, columnNumber++, data, strlen(data) * sizeof(char));
	rowNumber++, columnNumber = 0;
	
	// redraw list contents
	{
		RgnHandle		region = FinderList_GetNewComponentRegion(list, kFinderListComponentTypeListContents);
		
		
		if (region != null)
		{
			MacLook_InvalidateRegion(GetControlOwner(inDataPtr->tabControl), region);
			Memory_DisposeRegion(&region);
		}
	}
}// rebuildDisplay


/*!
This routine sizes and arranges controls in preference
panels to fit the prescribed dimensions of a panels
container.

(3.0)
*/
static void
setUpPanelControls		(PanelRef		inPanel)
{
	ScriptsPanelDataPtr		dataPtr = Panel_GetAuxiliaryDataPtr(inPanel);
	ControlRef				container = null,
							control = null;
	Rect					containerRect;
	register SInt16			i = 0;
	UInt16					leftEdge = 0, // cumulative, resettable measurement allowing controls to sandwich horizontally
							topEdge = 0; // cumulative, resettable measurement allowing controls to stack vertically
	
	
	Panel_GetContainerControl(inPanel, &container);
	GetControlBounds(container, &containerRect);
	
	// size controls
	FinderList_SetColumnWidthInPixels(dataPtr->tabs.attachedScripts.mappingsFinderList, 0, 150);
	FinderList_SetColumnWidthInPixels(dataPtr->tabs.attachedScripts.mappingsFinderList, 1, 210);
	FinderList_SizeControl(dataPtr->tabs.attachedScripts.mappingsFinderList, containerRect.right - containerRect.left,
							containerRect.bottom - containerRect.top);
	
	// arrange controls
	topEdge = 0;
	control = FinderList_GetContainerControl(dataPtr->tabs.attachedScripts.mappingsFinderList);
	MoveControl(control, 0, topEdge);
}// setUpPanelControls