/*###############################################################

	Preferences.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Balloons.h>
#include <MacTypes.h>
#include <Palettes.h>
#include <PLStringFuncs.h>
#include <QuickDraw.h>
#include <Resources.h>
#include <Script.h>
#include <TextUtils.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "StringResources.h"
#include "GeneralResources.h"
#include "TelnetVersion.h"

// MacTelnet includes
#include "FolderManager.h"
#include "MacLook.h"
#include "Preferences.h"
#include "ResourceFileManager.h"
#include "ScreenManager.h"
#include "TelnetErrors.h"



// constants
enum
{
	OUR_PREFS_VERSION_NUMBER = 3,
	NO_TIMEOUTS_VERSION = 1
};

// types
struct AliasInfo
{
	AliasHandle				alias;
	TelnetPrefsAliasID		resourceID;
	Boolean					wasChanged,
							isResource;
};
typedef struct AliasInfo	AliasInfo;
typedef AliasInfo*			AliasInfoPtr;

struct ApplicationPrefs
{
	// This is the general preferences structure.  Note that, as
	// a preference structure that is dumped to disk as a whole,
	// you CANNOT MOVE and CANNOT CHANGE any parts of this
	// structure that are already defined without becoming
	// backwards-incompatible with older preferences files.  You
	// can only use the padding at the end of the structure (and
	// if you do, you must decrease the padding array to
	// accommodate your additions and keep the structure size
	// EXACTLY the same as it was before).
	UInt16				version;					// The version number of this resource.
	
	UInt16				cursorType;					// A "kTerminalCursorType..." constant.
	
	UInt16				copyTableThreshold;			// Number of spaces that equal one tab.
	
	UInt32				maxTicksWaitNextEvent;		// Timeslice passed to WaitNextEvent.
	
	OSType				captureFileCreator;			// Creator signature for session capture files.
	
	UInt8				windowsDontGoAway,			// Connection windows remain open after a connection closes?
						backgroundNotification,		// Terminal bells cause a Notification Manager posting?
						menusHaveKeyEquivalents,	// Show (and recognize) key shortcuts for menus?
						remapBackquoteToEscape,		// Remap ` to escape?
						cursorBlinks,				// Blink the cursor?
						menusShowKeypadAndFunction,	// no longer used, reserved for backwards compatibility
						visualBell,					// Bell never sounds?
						destroyKerberosTickets,		// Destroy Kerberos tickets when a window closes?
						headersInitiallyCollapsed,	// Connection window headers are initially collapsed?
						dynamicResizing,			// Dragging the size box causes window size to change?
						useBackgroundPicture,		// Terminal windows use 'PICT' 1024?
						dontDimBackgroundScreens,	// Terminal colors never change for background windows?
						newSessionsDialogOnStartup,	// Show New Sessions dialog as soon as MacTelnet starts?
						invertedTextHighlighting,	// Highlight text by inverting, not using the highlight color?
						copySelectedText,			// Automatically copy selected text?
						autoCursorMoveOnDrop;		// Automatically issue cursor movement sequences prior to receiving a drop?
	
	UInt8				unusedFlags[10];			// space for several additional Booleans in the future
	
	UInt8				wasClipboardShowing,		// implicit window visibility state saving
						wasCommandLineShowing,
						wasControlKeypadShowing,
						wasFunctionKeypadShowing,
						wasMacroKeypadShowing,
						wasVT220KeypadShowing;
	
	UInt16				openTimeout,				// Time to Open Connections in seconds
						sendTimeout;				// Time to Send Data (and receive ACK)
	
	UInt8				simpleTelnet,				// Use Simplified User Interface mode?
						directConnect;				// no longer used, reserved for backwards compatibility
	
	SInt16				notificationPrefs;			// A "kAlert_Notify..." constant.
	
	SInt16				sherlockVRefNum;
	SInt32				sherlockDirID;
	Str31				sherlockName;				// used only in alpha releases, so it can be re-used
	
	SInt16				sherlockAliasID;			// this is used if Sherlock or Find File cant be found
	
	SInt16				padding[92];
};
#ifndef __cplusplus
typedef struct ApplicationPrefs	ApplicationPrefs;
#endif
typedef ApplicationPrefs*		ApplicationPrefsPtr;
typedef ApplicationPrefsPtr*	ApplicationPrefsHandle;

struct WindowPrefsWindowData
{
	// generic structure defining the arrangement of a particular window
	struct
	{
		SInt16		x,
					y;
	} location;
	
	struct
	{
		UInt16		width,
					height;
	} size;
};
typedef struct WindowPrefsWindowData WinPref;

struct WindowPrefs
{
	UInt16			version;					// The version number of this resource.
	Rect			screenBounds;				// (No longer used.) The screen boundaries when this resource was saved.
	WinPref			macroSetupDialog;			// Layout preferences for the Macro Setup dialog box.
	WinPref			openConnectionsDialog;		// Layout preferences for the New Sessions dialog box.
	WinPref			commandLineWindow;			// Layout preferences for the Message Window windoid.
	WinPref			keypadWindow;				// Layout preferences for the VT220 Keypad windoid.
	WinPref			functionWindow;				// Layout preferences for the Function windoid.
	WinPref			macrosWindow;				// Unused layout preferences for the (now removed) Macros windoid.
	WinPref			controlKeysWindow;			// Layout preferences for the Control Keys windoid.
	WinPref			preferencesWindow;			// Layout preferences for the Preferences dialog window.
	WinPref			unused[6];					// Space for future windows.
};
typedef struct WindowPrefs		WindowPrefs;
typedef WindowPrefs*			WindowPrefsPtr;
typedef WindowPrefsPtr*			WindowPrefsHandle;

// variables
FTPServerPrefs*				gFTPServerPrefs = null;
static ApplicationPrefs*	gApplicationPrefs = null;
static WindowPrefs*			gWindowPrefs = null;
static Handle				gApplicationPrefsHandle = null,
							gFTPServerPrefsHandle = null,
							gWindowPrefsHandle = null;
static ChainRef				gAliasList = null;
static PaletteHandle		gANSIColorPaletteHandle = null;

//
// internal methods
//

static void						checkPrefsVersion			(void);

static void						deleteAliasData				(AliasInfoPtr*			inoutAliasDataPtr);

static void						deleteAllAliasNodes			(void);

static AliasInfoPtr				findAlias					(TelnetPrefsAliasID		inAliasID);

static Boolean					findAliasOnDisk				(TelnetPrefsAliasID		inAliasID,
															 AliasHandle*			outAliasPtr);

static TelnetPrefsResultCode	getGeneralPreference		(TelnetPreferenceTag	inDataPreferenceTag,
															 Size					inDataSize,
															 void*					outDataPtr,
															 Size*					outActualSizePtr);

static WinPref*					getWinPrefPtrByWindowTag	(TelnetPreferenceTag	inWindowPreferenceTag);

static OSErr					initNewPreferences			(void);

static OSErr					loadAllPreferences			(void);

static OSErr					loadPreferencesStructure	(void*					inStructurePtr,
															 Size					inStructureSize,
															 ResType				inResourceType,
															 SInt16					inResourceID);

static OSErr					savePreferencesStructure	(void*					inStructurePtr,
															 Size					inStructureSize,
															 ResType				inResourceType,
															 SInt16					inResourceID);

static OSErr					setAliasChanged				(AliasInfoPtr			inAliasDataPtr);

static TelnetPrefsResultCode	setFileTransfersPreference	(TelnetPreferenceTag	inDataPreferenceTag,
															 Size					inDataSize,
															 void const*			inDataPtr);

static TelnetPrefsResultCode	setGeneralPreference		(TelnetPreferenceTag	inDataPreferenceTag,
															 Size					inDataSize,
															 void const*			inDataPtr);



//
// public methods
//

/*!
To initialize the global preference structures
(allocate memory for them and place them low in
the application heap), use this method.  Be sure
to call Preferences_Done() when you are finished
with these structures.

(3.0)
*/
void
Preferences_Init	(void)
{
	gANSIColorPaletteHandle = NewPalette(kANSIColorCount, null, pmCourteous/* usage */, 0x1000/* tolerance */);
	
	gApplicationPrefsHandle = Memory_NewHandleInProperZone(sizeof(ApplicationPrefs), false/* only briefly allocated */);
	gApplicationPrefs = (ApplicationPrefs*)*gApplicationPrefsHandle;
	
	gFTPServerPrefsHandle = Memory_NewHandleInProperZone(sizeof(FTPServerPrefs), false/* only briefly allocated */);
	gFTPServerPrefs = (FTPServerPrefs*)*gFTPServerPrefsHandle;
	
	gWindowPrefsHandle = Memory_NewHandleInProperZone(sizeof(WindowPrefs), false/* only briefly allocated */);
	gWindowPrefs = (WindowPrefs*)*gWindowPrefsHandle;
	
	gAliasList = Chain_New(kChainStyleSinglyLinked);
	Chain_SetDescriptor(gAliasList, kTelnetChainDescriptorAliasList);
}// Init


/*!
To destroy the global preference structures,
use this method.

(3.0)
*/
void
Preferences_Done	(void)
{
	// destroy the preferences structures
	Memory_DisposeHandle(&gApplicationPrefsHandle);
	Memory_DisposeHandle(&gFTPServerPrefsHandle);
	Memory_DisposeHandle(&gWindowPrefsHandle);
	
	// destroy the ANSI colors palette
	DisposePalette(gANSIColorPaletteHandle), gANSIColorPaletteHandle = null;
	
	// clean up the list of alias references
	deleteAllAliasNodes();
	Chain_Dispose(&gAliasList);
}// Done


/*!
To automatically size and/or position a
window based on the specified preferences,
use this method.  If the preferences for
the size are all-zero, the window size
will not change.  Similarly, if the
preferences for the location are all-zero,
then the window will not move.

The window will not move to an offscreen
location.

This routine will return the change in
width and height, if any, between the
given initial size and the new, preferred
size.  This is useful for windows with
controls that need to be moved and/or
resized appropriately when the window
changes size.

(3.0)
*/
Point
Preferences_ArrangeWindow	(WindowRef				inWindow,
							 Point					inInitialSize,
							 TelnetPreferenceTag	inWindowPreferenceTag)
{
	Point		result;
	WinPref*	arrangementPrefsPtr = getWinPrefPtrByWindowTag(inWindowPreferenceTag); // find the data, wherever it is
	
	
	SetPt(&result, 0, 0);
	
	if (arrangementPrefsPtr != null)
	{
		if (inWindow != null)
		{
			Point		p;
			Rect		visibleScreenRect;
			
			
			// calculate this area to make sure that windows arent arranged invisibly
			MacLook_GetWindowDeviceGrayRect(inWindow, &visibleScreenRect);
			
			// use the preferred location, if any
			SetPt(&p, visibleScreenRect.left, visibleScreenRect.top);
			if ((arrangementPrefsPtr->location.x != 0) || (arrangementPrefsPtr->location.y != 0)) // zero == no preference
			{
				SetPt(&p, arrangementPrefsPtr->location.x, arrangementPrefsPtr->location.y);
				if (PtInRect(p, &visibleScreenRect))
				{
					MoveWindow(inWindow, p.h, p.v, false/* activate */);
				}
			}
			
			// use the preferred size, if any
			if ((arrangementPrefsPtr->size.width > inInitialSize.h) ||
				(arrangementPrefsPtr->size.height > inInitialSize.v)) // zero means no preference
			{
				Point		allowedSize = inInitialSize;
				
				
				// prevent wild preferred sizes from going off the screen, if possible (unless the minimum size is offscreen)
				allowedSize.h = ((visibleScreenRect.right - p.h) >= arrangementPrefsPtr->size.width)
					? arrangementPrefsPtr->size.width
					: (visibleScreenRect.right - p.h);
				if (allowedSize.h < inInitialSize.h) allowedSize.h = inInitialSize.h;
				allowedSize.v = ((visibleScreenRect.bottom - p.v) >= arrangementPrefsPtr->size.height)
					? arrangementPrefsPtr->size.height
					: (visibleScreenRect.bottom - p.v);
				if (allowedSize.v < inInitialSize.v) allowedSize.v = inInitialSize.v;
				
				SizeWindow(inWindow, allowedSize.h, allowedSize.v, false/* update */);
				SetPt(&result, allowedSize.h - inInitialSize.h, allowedSize.v - inInitialSize.v);
			}
		}
	}
	return result;
}// ArrangeWindow


/*!
Automatically positions a window based on the
specified preferences.  If the preferences for
the location are all-zero, then the window
will not move.

The window will not move to an offscreen location.

Unlike Preferences_ArrangeWindow(), this routine
does not change the window size.  This is useful
for windows that have fixed sizes, but whose saved
sizes may be different than their sizes in the
current version of the application.

Returns "true" only if a preference was found and
enforced.

(3.0)
*/
Boolean
Preferences_ArrangeWindowNoResize	(WindowRef				inWindow,
									 TelnetPreferenceTag	inWindowPreferenceTag)
{
	WinPref*	arrangementPrefsPtr = getWinPrefPtrByWindowTag(inWindowPreferenceTag); // find the data, wherever it is
	Boolean		result = false;
	
	
	if (arrangementPrefsPtr != null)
	{
		if (inWindow != null)
		{
			Point		p;
			Rect		visibleScreenRect;
			
			
			// calculate this area to make sure that windows arent arranged invisibly
			MacLook_GetWindowDeviceGrayRect(inWindow, &visibleScreenRect);
			
			// use the preferred location, if any
			SetPt(&p, visibleScreenRect.left, visibleScreenRect.top);
			if ((arrangementPrefsPtr->location.x != 0) || (arrangementPrefsPtr->location.y != 0)) // zero == no preference
			{
				SetPt(&p, arrangementPrefsPtr->location.x, arrangementPrefsPtr->location.y);
				if (PtInRect(p, &visibleScreenRect))
				{
					MoveWindow(inWindow, p.h, p.v, false/* activate */);
					result = true;
				}
			}
		}
	}
	return result;
}// ArrangeWindowNoResize


/*!
To modify an alias in memory, use this method.
Your changes are not saved to the preferences
file unless you call Preferences_SaveAlias().

(3.0)
*/
void
Preferences_ChangeAlias		(TelnetPrefsAliasID		inAliasID,
							 FSSpec const*			inNewAliasFileSpecificationPtr)
{
	AliasInfoPtr		dataPtr = findAlias(inAliasID);
	
	
	if (dataPtr != null)
	{
		Boolean		wasChanged = false;
		
		
		if (UpdateAlias(null/* base */, inNewAliasFileSpecificationPtr, dataPtr->alias, &wasChanged) == noErr)
		{
			(OSErr)setAliasChanged(dataPtr);
		}
	}
}// ChangeAlias


/*!
To copy the ANSI colors from a palette, use
this method.  The palette must contain at
least "kANSIColorCount" colors.  This is
useful, and is equivalent to extracting the
colors from the palette and using each one
in a call to Preferences_SetANSIColor()
manually.

WARNING:	This method does no checking for
			the validity of the palette you
			provide.

(3.0)
*/
void
Preferences_CopyANSIPaletteFrom		(PaletteHandle			inCopyFromWhichPalette,
									 SInt16 const			inStartAtWhichIndex)
{
	CopyPalette(inCopyFromWhichPalette/* source palette */, gANSIColorPaletteHandle/* destination palette */,
				inStartAtWhichIndex/* source index */, 0/* destination index */, kANSIColorCount/* how many */);
}// CopyANSIPaletteFrom


/*!
To copy the ANSI colors into a palette, use
this method.  The palette must be large
enough to hold "kANSIColorCount" colors.

WARNING:	This method does no checking for
			the validity of the palette you
			provide.

(3.0)
*/
void
Preferences_CopyANSIPaletteTo		(PaletteHandle			outCopyToWhichPalette,
									 SInt16 const			inStartAtWhichIndex)
{
	CopyPalette(gANSIColorPaletteHandle/* source palette */, outCopyToWhichPalette/* destination palette */,
				0/* source index */, inStartAtWhichIndex/* destination index */, kANSIColorCount/* how many */);
}// CopyANSIPaletteTo


/*!
To open or create preferences data on disk, use this
method.  If a previous preferences set is found, this
method simply loads them and returns.  MacTelnet is a
little persistent about this - for example, it will try
to find files whose names match those of older versions
of MacTelnet, and load them into memory (renaming the
old files to the new names).  Note, however, that if the
proper file exists, MacTelnet will ignore older files
even if they exist.

If preferences need to be created, MacTelnet creates a
file in the Preferences folder.  It then initializes the
preferences in the file, using some default resources
from the application program file.

If a serious error prevents this method from completing,
it may not return.

If the disk file is not the ideal version, either
"kTelnetPrefsResultCodeBadVersionDataOlder" or
"kTelnetPrefsResultCodeBadVersionDataNewer" will result.
In the case of the former, you may need to tell the user
that some preferences have assumed default values.  In
the case of the latter, you may need to tell the user
that preferences have been lost.

In the event that the preferences file on disk is the
exact version that this code expects, the result code
"kTelnetPrefsResultCodeSuccess" is returned.

(3.0)
*/
TelnetPrefsResultCode
Preferences_CreateOrFindFile	(void)
{
	TelnetPrefsResultCode		result = kTelnetPrefsResultCodeSuccess;
	
	
	// create the MacTelnet Preferences folder and important sub-folders
	{
		FSSpec		unusedFolderSpec;
		
		
		// create the MacTelnet Preferences folder (no effect if it already exists); the resultant FSSpec is ignored
		FolderManager_GetFSSpec(kTelnetFolderPreferences, &unusedFolderSpec);
		
		// create the Recent Sessions folder (no effect if it already exists); the resultant FSSpec is ignored
		FolderManager_GetFSSpec(kTelnetFolderRecentSessions, &unusedFolderSpec);
		
		// create the Favorites folder (no effect if it already exists); the resultant FSSpec is ignored
		FolderManager_GetFSSpec(kTelnetFolderFavorites, &unusedFolderSpec);
		
		// create the FTP User Favorites folder (no effect if it already exists); the resultant FSSpec is ignored
		FolderManager_GetFSSpec(kTelnetFolderUserFTPIDFavorites, &unusedFolderSpec);
		
		// create the Proxy Favorites folder (no effect if it already exists); the resultant FSSpec is ignored
		FolderManager_GetFSSpec(kTelnetFolderUserProxyFavorites, &unusedFolderSpec);
		
		// create the Session Favorites folder (no effect if it already exists); the resultant FSSpec is ignored
		FolderManager_GetFSSpec(kTelnetFolderUserSessionFavorites, &unusedFolderSpec);
		
		// create the Terminal Favorites folder (no effect if it already exists); the resultant FSSpec is ignored
		FolderManager_GetFSSpec(kTelnetFolderUserTerminalFavorites, &unusedFolderSpec);
	}
	
	if (ResourceFileManager_GetResFile(kTelnetResourceFilePreferences) == -1)
	{
		FSSpec			parseFolder;
		Boolean			fatalStop = false;
		
		
		if (FolderManager_GetFSSpec(kTelnetFolderPreferences, &parseFolder) == noErr)
		{
			OSErr 			error = noErr;
			ScriptCode		prefsScriptCode = GetScriptManagerVariable(smSysScript);
			SInt32			prefsDirID = 0L,
							telnetPrefsDirID = 0L;
			SInt16			volume = 0;
			FSSpec			parseFile;
			Str255			settingsFileName;
			
			
			// the name of the main settings file is stored in a resource;
			// the file will be in the Preferences folder
			GetIndString(settingsFileName, rStringsImportantFileNames, siImportantFileNamePreferences);
			
			volume = parseFolder.vRefNum;
			prefsDirID = parseFolder.parID;
			
			error = FSMakeFSSpec(volume, prefsDirID, settingsFileName, &parseFile);
			if (error == noErr)
			{
				ResourceFileManager_SetResFile(kTelnetResourceFilePreferences, FSpOpenResFile(&parseFile, fsRdWrShPerm));
			}
			else if (error == fnfErr)
			{
				FileInfo	info;
				FSSpec		oldFile;
				
				
				// first check for old files (with different names), since they may
				// contain user preferences that should be converted; right now, these
				// old files are just converted by renaming them and reading them in
				oldFile.vRefNum = parseFile.vRefNum;
				oldFile.parID = parseFile.parID;
				PLstrcpy(oldFile.name, "\pTelnet 3.0 Preferences");
				FSpGetFInfo(&oldFile, (FInfo*)&info);
				info.finderFlags &= (~kNameLocked);
				FSpSetFInfo(&oldFile, (FInfo*)&info);
				
				if (FSpRename(&oldFile, settingsFileName) == noErr)
				{
					// if the file was successfully renamed, then it existed, so just open it;
					// since MacTelnet depends on the preferences file having a specific name,
					// make sure the user cant change it
					ResourceFileManager_SetResFile(kTelnetResourceFilePreferences, FSpOpenResFile(&parseFile, fsRdWrShPerm));
					info.finderFlags |= kNameLocked;
					FSpSetFInfo(&parseFile, (FInfo*)&info);
				}
				else 
				{
					// no previous preferences file (from any version of NCSA Telnet) exists;
					// create a new file and initialize it with default preference values
					if ((error = FSpCreate(&parseFile, kPreferencesFileCreatorSignature,
														kPreferencesFileType, prefsScriptCode)) == noErr)
					{
						// since MacTelnet depends on the preferences file having a specific name,
						// make sure the user cant change it
						FSpGetFInfo(&parseFile, (FInfo*)&info);
						info.finderFlags |= kNameLocked;
						FSpSetFInfo(&parseFile, (FInfo*)&info);
						
						// create a resource fork and initialize it
						FSpCreateResFile(&parseFile, kPreferencesFileCreatorSignature, kPreferencesFileType, prefsScriptCode);
						if (ResError() == noErr)
						{
							ResourceFileManager_SetResFile(kTelnetResourceFilePreferences,
															FSpOpenResFile(&parseFile, fsRdWrShPerm));
							
							if (ResourceFileManager_GetResFile(kTelnetResourceFilePreferences) != -1)
							{
								error = initNewPreferences();
							}
							else fatalStop = true;
						}
						else fatalStop = true;
					}
					else fatalStop = true;
				}
			}
			else fatalStop = true;
			
			// since MacTelnet depends on the preferences file having a specific name,
			// make sure the user cant change it
			unless (fatalStop)
			{
				FileInfo		info;
				
				
				FSpGetFInfo(&parseFile, (FInfo*)&info);
				info.finderFlags |= kNameLocked;
				FSpSetFInfo(&parseFile, (FInfo*)&info);
			}
			
			// load preferences, either from the initialized new file or the existing file
			if ((ResourceFileManager_GetResFile(kTelnetResourceFilePreferences) != -1) && (!fatalStop))
			{
				if (loadAllPreferences() != noErr) fatalStop = true;
			}
			else fatalStop = true;
		}
		else fatalStop = true;
		
		if (fatalStop) Errors_DisplayStopQuitMessage(rStringsStartupErrors,
														siUnableToCreatePreferencesFile, -1); // does not return
	}
	
	return result;
}// CreateOrFindFile


/*!
To remove an alias from the preferences file,
use this method.  This does not affect the
version in memory (the alias remains valid) -
to destroy the memory alias, use the method
Preferences_DisposeAlias().

(3.0)
*/
void
Preferences_DeleteSavedAlias		(TelnetPrefsAliasID		inAliasID)
{
	AliasInfoPtr		dataPtr = findAlias(inAliasID);
	
	
	if (dataPtr != null)
	{
		{
			Handle		temporaryHandle = Memory_NewHandle(GetHandleSize((Handle)dataPtr->alias));
			
			
			if (temporaryHandle != null)
			{
				// copy the alias data, because RemoveResource() creates an unreliable handle
				BlockMove(*((Handle)dataPtr->alias), *temporaryHandle, GetHandleSize(temporaryHandle));
				
				// remove the alias resource, and replace the internal handle with a valid one (the copy)
				RemoveResource((Handle)dataPtr->alias);
				ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
				dataPtr->alias = (AliasHandle)temporaryHandle;
			}
			dataPtr->isResource = false; // signifies that the data in memory no longer represents a resource
		}
		Alert_ReportOSErr(ResError());
	}
}// DeleteSavedAlias


/*!
To destroy the memory version of an alias, use this
method.  To write the contents of an alias in memory to
the preferences file first, use Preferences_SaveAlias(),
and then call this method to dispose of the memory.
After calling this routine, the specified alias ID is
invalid.

(3.0)
*/
void
Preferences_DisposeAlias	(TelnetPrefsAliasID		inAliasID)
{
	ChainLinkRef	node = null;
	AliasInfoPtr	dataPtr = null;
	
	
	// look for the given alias in the linked list
	node = Chain_GetFirstNode(gAliasList);
	do
	{
		dataPtr = Chain_GetNodeDataPtr(gAliasList, node);
		if (dataPtr != null) if (dataPtr->resourceID == inAliasID) break;
		node = Chain_GetNextNode(gAliasList, node);
	} while (node != null);
	
	if (node != null) Chain_DisposeNode(gAliasList, &node);
	deleteAliasData(&dataPtr);
}// DisposeAlias


/*!
To fill in an "RGBColor" structure to match the
specified ANSI color, use this method.  The
resultant color will reflect whatever the user
has defined.

(3.0)
*/
void
Preferences_GetANSIColor	(ANSIColorDescriptor const	inWhichColor,
							 RGBColor*					outColorPtr)
{
	if (outColorPtr != null) GetEntryColor(gANSIColorPaletteHandle, inWhichColor, outColorPtr);
}// GetANSIColor


/*!
To get a preference corresponding to the specified tag, use
this method.  If the specified class or tag is not understood,
this routine returns "kTelnetPrefsResultCodeUnknownTagOrClass".

It is possible that older preferences files may not hold data
from certain classes with certain tags.  If the file on disk
does not contain the requested data, this method will return
"kTelnetPrefsResultCodeBadVersionDataNotAvailable" and the
output data will be garbage, with "outActualSizePtr" set to 0.

If you provide a buffer too small to store the data, the error
code "kTelnetPrefsResultCodeInsufficientBufferSpace" is
returned.  Otherwise, the data is stored, and its exact size
is provided via "outActualSizePtr".

If the data is acquired, "kTelnetPrefsResultCodeSuccess" is
returned.

Incomplete.

(3.0)
*/
TelnetPrefsResultCode
Preferences_GetData		(TelnetPreferenceClass	inDataPreferenceClass,
						 TelnetPreferenceTag	inDataPreferenceTag,
						 Size					inDataStorageSize,
						 void*					outDataStorage,
						 Size*					outActualSizePtr)
{
	TelnetPrefsResultCode		result = kTelnetPrefsResultCodeSuccess;
	
	
	switch (inDataPreferenceClass)
	{
		case kTelnetPreferenceClassGeneral:
			result = getGeneralPreference(inDataPreferenceTag, inDataStorageSize, outDataStorage, outActualSizePtr);
			break;
		
		case kTelnetPreferenceClassFTP:
			switch (inDataPreferenceTag)
			{
				case kTelnetPreferenceTag_ServerLogInMessage:
					if ((outDataStorage != null) && (inDataStorageSize >= sizeof(Handle)))
					{
						SInt16		oldResFile = CurResFile();
						Handle		resHandle = null;
						
						
						ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
						resHandle = Get1Resource('TEXT', FTPSERVERPREFS_LOGINMESSAGEID);
						if ((ResError() != noErr) || (resHandle == null))
						{
							// resource doesnt exist, or couldnt be read
							resHandle = null;
							result = kTelnetPrefsResultCodeBadVersionDataNotAvailable;
						}
						else DetachResource(resHandle);
						UseResFile(oldResFile);
						*((Handle*)outDataStorage) = resHandle;
						if (outActualSizePtr != null) *outActualSizePtr = sizeof(Handle);
					}
					else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
					break;
				
				default:
					result = kTelnetPrefsResultCodeUnknownTagOrClass;
					break;
			}
			break;
		
		case kTelnetPreferenceClassWindow:
		default:
			// unrecognized preference class
			result = kTelnetPrefsResultCodeUnknownTagOrClass;
			break;
	}
	return result;
}// GetData


/*!
Allocates memory for a session preferences
structure, reads the default preferences from
the application resource file, and returns a
handle to memory containing the preferences.

(2.6)
*/
SessionPrefs**
Preferences_GetDefaultSession	(void)
{
	SessionPrefs**		result = null;
	
	
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	result = (SessionPrefs**)Get1Resource(SESSIONPREFS_RESTYPE, SESSIONPREFS_DEFAULTID);
	
	// If there is an error here, restore the default from the
	// master copy in the applications resource fork.
	if (result == null)
	{
		ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
		result = (SessionPrefs**)Get1Resource(SESSIONPREFS_RESTYPE, SESSIONPREFS_DEFAULTID);
		
		if (result != null)
		{
			Str255		sessionDefaultName;
			short		newPrefResourceID = 0;
			
			
			DetachResource((Handle)result);
			ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
			newPrefResourceID = Unique1ID(SESSIONPREFS_RESTYPE);
			StringUtilities_CToP(SESSIONPREFS_DEFAULTNAME, sessionDefaultName);
			AddResource((Handle)result, SESSIONPREFS_RESTYPE, newPrefResourceID, sessionDefaultName);
	
			if (ResError() == noErr) ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
		}
	}
	
	if (result != null) DetachResource((Handle)result);
	return result;
}// GetDefaultSession


/*!
Allocates memory for a terminal preferences
structure, reads the default preferences from
the application resource file, and returns a
handle to memory containing the preferences.

(2.6)
*/
TerminalPrefs**
Preferences_GetDefaultTerminal	(void)
{
	TerminalPrefs**		result = null;
	
	
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	result = (TerminalPrefs**)Get1Resource(TERMINALPREFS_RESTYPE, TERMINALPREFS_DEFAULTID);
	
	// If there is an error here, restore the default from the
	// master copy in the applications resource fork.
	if (result == null)
	{
		ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
		result = (TerminalPrefs**)Get1Resource(TERMINALPREFS_RESTYPE, TERMINALPREFS_DEFAULTID);
		
		if (result != null)
		{
			Str255		terminalDefaultName;
			short		newPrefResourceID = 0;
			
			
			DetachResource((Handle)result);	
			ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
			newPrefResourceID = Unique1ID(TERMINALPREFS_RESTYPE);
			StringUtilities_CToP(TERMINALPREFS_DEFAULTNAME, terminalDefaultName);
			AddResource((Handle)result, TERMINALPREFS_RESTYPE, newPrefResourceID, terminalDefaultName);
	
			if (ResError() == noErr) ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
		}
	}
		
	if (result != null) DetachResource((Handle)result);
	return result;
}// GetDefaultTerminal


/*!
To determine if an alias with the specified ID is in
the preferences file, use this method.

(3.0)
*/
Boolean
Preferences_IsAliasStored	(TelnetPrefsAliasID		inAliasID)
{
	Boolean			result = false;
	AliasHandle		unused = null;
	
	
	result = findAliasOnDisk(inAliasID, &unused);
	ReleaseResource((Handle)unused), unused = null;
	return result;
}// IsAliasStored


/*!
To create a new alias in memory, use this method.  You can
set up pending alias preferences this way, changing the
alias as many times as you need to, but only writing it to
the preferences file if necessary.  This prevents unused
alias resources from appearing in the preferences file.

IMPORTANT:	Always call Preferences_IsAliasStored() before
			creating a new alias.  If you want to update
			an existing alias, you need to create it using
			Preferences_NewSavedAlias().  Otherwise, if
			you use Preferences_NewAlias(), and then save
			the alias to disk, it will never replace any
			existing alias.

(3.0)
*/
TelnetPrefsAliasID
Preferences_NewAlias	(FSSpec const*		inFileSpecificationPtr)
{
	TelnetPrefsAliasID		result = kInvalidTelnetPrefsAliasID;
	AliasInfoPtr			dataPtr = (AliasInfoPtr)Memory_NewPtr(sizeof(AliasInfo));
	
	
	if ((dataPtr != null) && (inFileSpecificationPtr != null))
	{
		ChainLinkRef		node = Chain_NewNode(gAliasList, dataPtr);
		
		
		if (node != null)
		{
			SInt16		oldResFile = CurResFile();
			
			
			ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
			Chain_MoveNode(gAliasList, node, null/* means move to end of list */, kChainLinkPlacementInsertAfter);
			dataPtr->alias = null;
			if (NewAlias(null/* base */, inFileSpecificationPtr, &dataPtr->alias) == noErr)
			{
				dataPtr->resourceID = result = Unique1ID(rAliasType);
				dataPtr->wasChanged = false;
				dataPtr->isResource = false;
			}
			UseResFile(oldResFile);
		}
	}
	return result;
}// NewAlias


/*!
To create a new alias in memory based on an alias already
present in the preferences file, use this method.  You
generally save an alias ID in a preferences structure in
order to refer to a file, and then use that ID and this
method to work with the alias in memory.  To make changes
to an alias in the preferences file, you must first use
this routine to read it into memory, and then use the
Preferences_ChangeAlias() and Preferences_SaveAlias()
methods.

IMPORTANT:	Always call Preferences_IsAliasStored() before
			creating a new alias with this routine.  If an
			alias with the ID of "inAliasID" exists, it
			*will* be overwritten if you subsequently call
			Preferences_SaveAlias() on your new alias.  If
			you want a new alias that will not conflict
			with saved aliases, use Preferences_NewAlias().

WARNING:	Never call Preferences_NewSavedAlias() more than
			once with the same alias ID, unless you always
			call Preferences_DisposeAlias() to destroy prior
			instances with the same ID.  If you create more
			than one alias with the same ID, a memory leak
			will ensue and you will only be able to access
			data for the first alias with the repeated ID.

(3.0)
*/
TelnetPrefsAliasID
Preferences_NewSavedAlias	(TelnetPrefsAliasID		inAliasID)
{
	AliasHandle				alias = null;
	TelnetPrefsAliasID		result = kInvalidTelnetPrefsAliasID;
	
	
	if (findAliasOnDisk(inAliasID, &alias))
	{
		FSSpec		aliasFile;
		Boolean		recordChanged = false;
		
		
		if (ResolveAlias(null/* base */, alias, &aliasFile, &recordChanged) == noErr)
		{
			AliasInfoPtr		dataPtr = findAlias(Preferences_NewAlias(&aliasFile));
			
			
			// set the alias ID to the specified ID, to make sure the stored alias gets overwritten when the alias is saved
			if (dataPtr != null)
			{
				dataPtr->resourceID = result = inAliasID; // changing this field allows the alias to be stored under the given ID
				dataPtr->alias = alias; // set the handle to be a resource handle (see Preferences_SaveAlias())
				dataPtr->isResource = true;
				if (recordChanged) ChangedResource((Handle)alias);
				Alert_ReportOSErr(ResError());
			}
			else LED_Flash();
		}
		else ReleaseResource((Handle)alias), alias = null; // findAliasOnDisk() gets a resource in memory
	}
	return result;
}// NewSavedAlias


/*!
To read the file for an alias, use this method.
If the file does not exist, the result is "false";
otherwise, the result is "true".

(3.0)
*/
Boolean
Preferences_ParseAlias	(TelnetPrefsAliasID		inAliasID,
						 FSSpec*				outAliasFileSpecificationPtr)
{
	Boolean		result = false;
	
	
	if (outAliasFileSpecificationPtr != null)
	{
		AliasInfoPtr		dataPtr = findAlias(inAliasID);
		
		
		if (dataPtr != null)
		{
			Boolean		wasChanged = false;
			
			
			if (ResolveAlias(null/* base */, dataPtr->alias, outAliasFileSpecificationPtr, &wasChanged) == noErr)
			{
				// then the file was found
				result = true;
				if (wasChanged) (OSErr)setAliasChanged(dataPtr);
			}
		}
	}
	return result;
}// ParseAlias


/*!
To save an alias to disk (but only if it has been modified),
use this method.  This is the only way to physically write
an alias record to the preferences file.

The new alias resource will be given the specified name.
If you do not want to use a name, pass an empty string.

(3.0)
*/
void
Preferences_SaveAlias	(TelnetPrefsAliasID		inAliasID,
						 ConstStr255Param		inName)
{
	AliasInfoPtr		dataPtr = findAlias(inAliasID);
	
	
	if (dataPtr != null)
	{
		// only write to disk if the data in memory actually changed at some point
		if (dataPtr->wasChanged)
		{
			SInt16		oldResFile = CurResFile();
			OSErr		error = noErr;
			
			
			ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
			unless (Preferences_IsAliasStored(inAliasID))
			{
				// write a new alias record to the preferences file, using the unique resource ID number
				AddResource((Handle)dataPtr->alias, rAliasType, dataPtr->resourceID, inName);
				error = ResError();
			}
			if (error == noErr) ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
			Alert_ReportOSErr(error);
			UseResFile(oldResFile);
		}
	}
}// SaveAlias


/*!
To write changes youve made using the Preferences_SetData()
routine to disk, use this method.  Only preferences from the
specified class get saved.  You can save preferences from
any class at any time, regardless of which class was altered
at what time.

If a problem occurs, such as a disk-related OS error, an
error code other than "noErr" will result.

(3.0)
*/
OSErr
Preferences_SaveData	(TelnetPreferenceClass		inDataPreferenceClass)
{
	OSErr		result = noErr;
	
	
	switch (inDataPreferenceClass)
	{
		case kTelnetPreferenceClassGeneral:
			result = savePreferencesStructure(gApplicationPrefs, sizeof(ApplicationPrefs), APPLICATIONPREFS_RESTYPE, APPLICATIONPREFS_ID);
			{
				// save ANSI colors
				PaletteHandle		paletteHandle = null;
				
				
				paletteHandle = NewPalette(kANSIColorCount, null, pmCourteous/* usage */, 0x1000/* tolerance */);
				Preferences_CopyANSIPaletteTo(paletteHandle, 0/* start at this index */);
				(TelnetPrefsResultCode)setGeneralPreference(kTelnetPreferenceTag_ANSIColors, sizeof(paletteHandle),
															&paletteHandle);
				DisposePalette(paletteHandle), paletteHandle = null;
			}
			break;
		
		case kTelnetPreferenceClassFTP:
			result = savePreferencesStructure(gFTPServerPrefs, sizeof(FTPServerPrefs), FTPSERVERPREFS_RESTYPE, FTPSERVERPREFS_ID);
			break;
		
		case kTelnetPreferenceClassWindow:
			gWindowPrefs->version = 0;
			SetRect(&gWindowPrefs->screenBounds, 0, 0, 0, 0); // this data is no longer used
			result = savePreferencesStructure(gWindowPrefs, sizeof(WindowPrefs), WINDOWPREFS_RESTYPE, WINDOWPREFS_ID);
			break;
		
		default:
			result = kTelnetPrefsResultCodeUnknownTagOrClass;
			break;
	}
	
	return result;
}// SaveData


/*!
To use an "RGBColor" structure to specify an
ANSI color, use this method.

(3.0)
*/
void
Preferences_SetANSIColor	(ANSIColorDescriptor const	inWhichColor,
							 RGBColor const*			inColorPtr)
{
	if (inColorPtr != null) SetEntryColor(gANSIColorPaletteHandle, inWhichColor, inColorPtr);
}// SetANSIColor


/*!
To set a preference corresponding to the specified tag, use
this method.  If the specified class or tag is not understood,
this routine returns "kTelnetPrefsResultCodeUnknownTagOrClass".
Note that you should use the Preferences_ArrangeWindow() method
for tags of the "kTelnetPreferenceClassWindow" class, so any
attempt to set data with the "kTelnetPreferenceClassWindow"
class will result in "kTelnetPrefsResultCodeUnknownTagOrClass".

If the data is set properly, "kTelnetPrefsResultCodeSuccess" is
returned.

(3.0)
*/
TelnetPrefsResultCode
Preferences_SetData		(TelnetPreferenceClass	inDataPreferenceClass,
						 TelnetPreferenceTag	inDataPreferenceTag,
						 Size					inDataSize,
						 void const*			inDataPtr)
{
	TelnetPrefsResultCode		result = kTelnetPrefsResultCodeSuccess;
	
	
	switch (inDataPreferenceClass)
	{
		case kTelnetPreferenceClassGeneral:
			result = setGeneralPreference(inDataPreferenceTag, inDataSize, inDataPtr);
			break;
		
		case kTelnetPreferenceClassFTP:
			result = setFileTransfersPreference(inDataPreferenceTag, inDataSize, inDataPtr);
			break;
		
		case kTelnetPreferenceClassWindow:
		default:
			result = kTelnetPrefsResultCodeUnknownTagOrClass;
			break;
	}
	return result;
}// SetData


/*!
To automatically size and/or position a window based
on the specified preferences, use this method.  If
the preferences for the size are all-zero, the window
size will not change.  Similarly, if the preferences
for the location are all-zero, then the window will
not move.  If any problems occur, no preference is
saved.

(3.0)
*/
void
Preferences_SaveWindowArrangement	(WindowRef				inWindow,
									 TelnetPreferenceTag	inWindowPreferenceTag)
{
	WinPref*	arrangementPrefsPtr = getWinPrefPtrByWindowTag(inWindowPreferenceTag); // find the data, wherever it is
	
	
	if ((arrangementPrefsPtr != null) && (inWindow != null))
	{
		Rect		rect;
		
		
		MacLook_GetWindowRegionBounds(inWindow, kWindowContentRgn, &rect);
		arrangementPrefsPtr->location.x = rect.left;
		arrangementPrefsPtr->location.y = rect.top;
		arrangementPrefsPtr->size.width = rect.right - rect.left;
		arrangementPrefsPtr->size.height = rect.bottom - rect.top;
		(OSErr)Preferences_SaveData(kTelnetPreferenceClassWindow);
	}
}// SaveWindowArrangement


//
// internal methods
//
#pragma mark -

/*!
Updates older preferences files.

(2.6)
*/
static void
checkPrefsVersion	(void)
{
	if (gApplicationPrefs->version != OUR_PREFS_VERSION_NUMBER) // need to correctly set new values
	{
		// update any out-of-date preferences; N/A
		gApplicationPrefs->version = OUR_PREFS_VERSION_NUMBER;
		(OSErr)Preferences_SaveData(kTelnetPreferenceClassGeneral);
	}
}// checkPrefsVersion


/*!
To acquire a pointer to the internal structure
whose panel field corresponds to the given
panel, use this method.

(3.0)
*/
static void
deleteAliasData		(AliasInfoPtr*		inoutAliasDataPtr)
{
	ChainLinkRef	node = null;
	
	
	// look for the given alias in the linked list
	if (inoutAliasDataPtr != null) if (*inoutAliasDataPtr != null)
	{
		if ((*inoutAliasDataPtr)->alias != null)
		{
			if ((*inoutAliasDataPtr)->isResource)
			{
				ReleaseResource((Handle)(*inoutAliasDataPtr)->alias), (*inoutAliasDataPtr)->alias = null;
			}
			else Memory_DisposeHandle((Handle*)&((*inoutAliasDataPtr)->alias));
		}
		Memory_DisposePtr((Ptr*)inoutAliasDataPtr);
	}
}// deleteAliasData


/*!
One of the primary reasons for using a linked list
is to keep track of alias references.  This routine
will automatically dispose of all outstanding alias
references (in case code utilizing this module fails
to call Preferences_Dispose() for each new alias).

(3.0)
*/
static void
deleteAllAliasNodes		(void)
{
	ChainLinkRef	node = null;
	AliasInfoPtr	dataPtr = null;
	
	
	// look for the given alias in the linked list
	node = Chain_GetFirstNode(gAliasList);
	do
	{
		dataPtr = Chain_DisposeNode(gAliasList, &node); // conveniently returns the node data for the destroyed node
		deleteAliasData(&dataPtr); // dispose of the Preferences module internal data structure associated with the node
		node = Chain_GetNextNode(gAliasList, node);
	} while (node != null);
}// deleteAllAliasNodes


/*!
To acquire a pointer to the data structure for
the specified alias, use this method.  If the
data does not exist, null is returned.

(3.0)
*/
static AliasInfoPtr
findAlias		(TelnetPrefsAliasID		inAliasID)
{
	ChainLinkRef	node = null;
	AliasInfoPtr	result = null;
	
	
	// look for the given alias in the linked list
	node = Chain_GetFirstNode(gAliasList);
	do
	{
		result = Chain_GetNodeDataPtr(gAliasList, node);
		if (result != null) if (result->resourceID == inAliasID) break;
		node = Chain_GetNextNode(gAliasList, node);
	} while (node != null);
	
	return result;
}// findAlias


/*!
To read an alias resource corresponding to
a given MacTelnet alias ID, use this method.
If the resource cannot be read for any
reason, "false" is returned and the output
alias handle should be considered invalid;
otherwise, "true" is returned and a valid
alias record is provided in "outAliasPtr".

You may assume that the "AliasHandle" you
provide will become a resource handle.  As
such, you must eventually free its memory
using ReleaseResource().

(3.0)
*/
static Boolean
findAliasOnDisk		(TelnetPrefsAliasID		inAliasID,
					 AliasHandle*			outAliasPtr)
{
	Boolean		result = false;
	
	
	if (outAliasPtr != null)
	{
		SInt16		oldResFile = CurResFile();
		
		
		// the MacTelnet alias ID is defined as the resource ID of the alias resource
		ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
		*outAliasPtr = (AliasHandle)Get1Resource(rAliasType, inAliasID);
		if ((ResError() == noErr) && (*outAliasPtr != null)) result = true;
		UseResFile(oldResFile);
	}
	return result;
}// findAliasOnDisk


/*!
To get a preference from the General preference
class, use this method.

(3.0)
*/
static TelnetPrefsResultCode	
getGeneralPreference	(TelnetPreferenceTag	inDataPreferenceTag,
						 Size					inDataSize,
						 void*					outDataPtr,
						 Size*					outActualSizePtr)
{
	Size					actualSize = 0L;
	TelnetPrefsResultCode	result = kTelnetPrefsResultCodeSuccess;
	
	
	if (outDataPtr != null) switch (inDataPreferenceTag)
	{
		case kTelnetPreferenceTag_ANSIColors:
			actualSize = sizeof(PaletteHandle);
			if (inDataSize >= actualSize)
			{
				PaletteHandle*		paletteHandlePtr = ((PaletteHandle*)outDataPtr);
				
				
				*paletteHandlePtr = GetNewPalette(ANSICOLORS_ID); // result is not a resource handle
				if (*paletteHandlePtr == null)
				{
					SInt16		oldResFile = CurResFile();
					
					
					// get the applications copy instead
					ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
					*paletteHandlePtr = GetNewPalette(ANSICOLORS_DEFAULTID); // result is not a resource handle
					if (*paletteHandlePtr != null)
					{
						ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
						AddResource((Handle)*paletteHandlePtr, 'pltt', ANSICOLORS_ID, EMPTY_PSTRING);
						ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
					}
					UseResFile(oldResFile);
				}
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CaptureFileCreator:
			actualSize = sizeof(OSType);
			if (inDataSize >= actualSize)
			{
				*((OSType*)outDataPtr) = gApplicationPrefs->captureFileCreator;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CopySelectedText:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->copySelectedText;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CopyTableThreshold:
			actualSize = sizeof(UInt16);
			if (inDataSize >= actualSize)
			{
				*((UInt16*)outDataPtr) = gApplicationPrefs->copyTableThreshold;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CursorBlinks:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->cursorBlinks;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CursorMovesPriorToDrops:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->autoCursorMoveOnDrop;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_DontAutoClose:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->windowsDontGoAway;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_DontDimBackgroundScreens:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->dontDimBackgroundScreens;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_DynamicResizing:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->dynamicResizing;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_ExpireTickets:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->destroyKerberosTickets;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_FindFileAppSpec:
			actualSize = sizeof(FSSpec);
			if (inDataSize >= actualSize)
			{
				((FSSpec*)outDataPtr)->vRefNum = gApplicationPrefs->sherlockVRefNum;
				((FSSpec*)outDataPtr)->parID = gApplicationPrefs->sherlockDirID;
				PLstrcpy(((FSSpec*)outDataPtr)->name, gApplicationPrefs->sherlockName);
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_HeadersCollapsed:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->headersInitiallyCollapsed;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_MapBackquote:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->remapBackquoteToEscape;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_MenuItemKeys:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->menusHaveKeyEquivalents;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_Notification:
			actualSize = sizeof(SInt16);
			if (inDataSize >= actualSize)
			{
				*((SInt16*)outDataPtr) = gApplicationPrefs->notificationPrefs;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_NotifyOfBeeps:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->notificationPrefs;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_PureInverse:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->invertedTextHighlighting;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_ShowNewSessionsDialogAtStartup:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->newSessionsDialogOnStartup;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_SimpleTelnet:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->simpleTelnet;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TerminalCursorType:
			actualSize = sizeof(UInt16);
			if (inDataSize >= actualSize)
			{
				*((UInt16*)outDataPtr) = gApplicationPrefs->cursorType;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TicksWaitNextEvent:
			actualSize = sizeof(UInt32);
			if (inDataSize >= actualSize)
			{
				*((UInt32*)outDataPtr) = gApplicationPrefs->maxTicksWaitNextEvent;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TimeoutOpen:
			actualSize = sizeof(UInt32);
			if (inDataSize >= actualSize)
			{
				*((UInt32*)outDataPtr) = gApplicationPrefs->openTimeout;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TimeoutSend:
			actualSize = sizeof(UInt32);
			if (inDataSize >= actualSize)
			{
				*((UInt32*)outDataPtr) = gApplicationPrefs->sendTimeout;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_UseBackgroundPicture:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->useBackgroundPicture;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_VisualBell:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->visualBell;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasClipboardShowing:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->wasClipboardShowing;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasCommandLineShowing:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->wasCommandLineShowing;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasControlKeypadShowing:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->wasControlKeypadShowing;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasFunctionKeypadShowing:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->wasFunctionKeypadShowing;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasMacroKeypadShowing:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->wasMacroKeypadShowing;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasVT220KeypadShowing:
			actualSize = sizeof(Boolean);
			if (inDataSize >= actualSize)
			{
				*((Boolean*)outDataPtr) = gApplicationPrefs->wasVT220KeypadShowing;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		default:
			// unrecognized tag
			result = kTelnetPrefsResultCodeUnknownTagOrClass;
			break;
	}
	if (outActualSizePtr != null) *outActualSizePtr = actualSize;
	return result;
}// getGeneralPreference


/*!
To acquire a pointer to a "WindowPrefsWindowData"
structure that stores the window sizing and
positioning preferences for the window corresponding
to the given tag, use this method.

(3.0)
*/
static WinPref*
getWinPrefPtrByWindowTag	(TelnetPreferenceTag	inWindowPreferenceTag)
{
	WinPref*		result = null;
	
	
	switch (inWindowPreferenceTag)
	{
		case kTelnetWindowPreferenceTag_CommandLine:
			result = &gWindowPrefs->macroSetupDialog;
			break;
		
		case kTelnetWindowPreferenceTag_ControlKeypad:
			result = &gWindowPrefs->controlKeysWindow;
			break;
		
		case kTelnetWindowPreferenceTag_FunctionKeypad:
			result = &gWindowPrefs->functionWindow;
			break;
		
		case kTelnetWindowPreferenceTag_MacroKeypad:
			result = &gWindowPrefs->macrosWindow;
			break;
		
		case kTelnetWindowPreferenceTag_MacroSetup:
			result = &gWindowPrefs->macroSetupDialog;
			break;
		
		case kTelnetWindowPreferenceTag_NewSessions:
			result = &gWindowPrefs->openConnectionsDialog;
			break;
		
		case kTelnetWindowPreferenceTag_Preferences:
			result = &gWindowPrefs->preferencesWindow;
			break;
		
		case kTelnetWindowPreferenceTag_VT220Keypad:
			result = &gWindowPrefs->keypadWindow;
			break;
		
		default:
			break;
	}
	
	return result;
}// getWinPrefPtrByWindowTag


/*!
Copies default preferences from the application
resource file as a starting point.

Rewritten in MacTelnet 3.0 to preserve the old
resource file, to detect errors better and to
copy both a Finder description string and a
Balloon Help resource to the new file.  The new
resources let the user see a message if he or
she either double-clicks the icon of the
preferences file in the Finder or uses Balloon
Help to find out more about the icon.

If the creation of the items succeeds without
any problems, "noErr" is returned.

(2.6)
*/
static OSErr
initNewPreferences	(void)
{
	OSErr		result = noErr;
	Str255		genericResourceName;
	short		oldResourceFile = 0;
	
	
	oldResourceFile = CurResFile();
	
	ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
	if (result == noErr)
	{
		ApplicationPrefsHandle	prefResHandleApplication = null;
		FTPServerPrefsHandle	prefResHandleFTPServer = null;
		SessionPrefsHandle		prefResHandleSessionDefault = null;
		TerminalPrefsHandle		prefResHandleTerminalDefault = null;
		//ProxyServerPrefsHandle	prefResHandleProxyServerDefault = null;
		WindowPrefsHandle		prefResHandleWindowDefault = null;
		StringHandle			prefResHandleDescriptionString = null;
		Handle					prefResHandleBalloonHelp = null,
								prefResHandleVers1 = null,
								prefResHandleVers2 = null;
		
		
		// get master preferences from the applications resource fork...
		if (result == noErr)
		{
			prefResHandleApplication = (ApplicationPrefsHandle)Get1Resource(APPLICATIONPREFS_RESTYPE, APPLICATIONPREFS_DEFAULTID);
			result = ResError();
			if (prefResHandleApplication == null) result = resNotFound;
		}
		
		if (result == noErr)
		{
			prefResHandleFTPServer = (FTPServerPrefsHandle)Get1Resource(FTPSERVERPREFS_RESTYPE, FTPSERVERPREFS_DEFAULTID);
			result = ResError();
			if (prefResHandleFTPServer == null) result = resNotFound;
		}
		
		if (result == noErr)
		{
			prefResHandleSessionDefault = (SessionPrefsHandle)Get1Resource(SESSIONPREFS_RESTYPE, SESSIONPREFS_DEFAULTID);
			result = ResError();
			if (prefResHandleSessionDefault == null) result = resNotFound;
		}
		
		if (result == noErr)
		{
			prefResHandleTerminalDefault = (TerminalPrefsHandle)Get1Resource(TERMINALPREFS_RESTYPE, TERMINALPREFS_DEFAULTID);
			result = ResError();
			if (prefResHandleTerminalDefault == null) result = resNotFound;
		}
		
		//if (result != noErr)
		//{
		//	prefResHandleProxyServerDefault = (ProxyServerPrefsHandle)Get1Resource(PROXYPREFS_RESTYPE, PROXYPREFS_DEFAULTID);
		//	result = ResError();
		//}
		
		if (result == noErr)
		{
			prefResHandleWindowDefault = (WindowPrefsHandle)Get1Resource(WINDOWPREFS_RESTYPE, WINDOWPREFS_DEFAULTID);
			result = ResError();
			if (prefResHandleWindowDefault == null) result = resNotFound;
		}
		
		// its not important if one of the following cannot be found, so ignore ResError()
		prefResHandleBalloonHelp = Get1Resource(kHMFinderApplResType, PREFERENCES_FILE_INFO_STR_ID);
		prefResHandleDescriptionString = GetString(PREFERENCES_FILE_INFO_STR_ID);
		prefResHandleVers1 = Get1Resource('vers', 1);
		prefResHandleVers2 = Get1Resource('vers', 2);
		
		if (result == noErr)
		{
			DetachResource((Handle)prefResHandleApplication);
			DetachResource((Handle)prefResHandleFTPServer);
			DetachResource((Handle)prefResHandleSessionDefault);
			DetachResource((Handle)prefResHandleTerminalDefault);
			DetachResource((Handle)prefResHandleWindowDefault);
			if (prefResHandleDescriptionString != null) DetachResource((Handle)prefResHandleDescriptionString);
			if (prefResHandleBalloonHelp != null) DetachResource(prefResHandleBalloonHelp);
			if (prefResHandleVers1 != null) DetachResource(prefResHandleVers1);
			if (prefResHandleVers2 != null) DetachResource(prefResHandleVers2);
			ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
			
			// create empty font lists, to be filled in by a font list management thread
			if (result == noErr)
			{
				Handle		theGoodFonts = null,
							theBadFonts = null;
				
				
				theGoodFonts = Memory_NewHandleInProperZone(sizeof(SInt16), true/* only briefly allocated */);
				theBadFonts = Memory_NewHandleInProperZone(sizeof(SInt16), true/* only briefly allocated */);
				if (result == noErr)
				{
					AddResource(theGoodFonts, 'STR#', rStringsGoodFonts, EMPTY_PSTRING);
					result = ResError();
				}
				if (result == noErr)
				{
					AddResource(theBadFonts, 'STR#', rStringsBadFonts, EMPTY_PSTRING);
					result = ResError();
				}
				UpdateResFile(ResourceFileManager_GetResFile(kTelnetResourceFilePreferences));
				Memory_DisposeHandle(&theGoodFonts);
				Memory_DisposeHandle(&theBadFonts);
			}
			
			// This safeguard code prevents errors in the event that
			// the default encoded preferences somehow do not have
			// exactly the same data size as the internal structure.
			// This is mostly useful during coding stages where the
			// contents of the preferences structures are in a state
			// of flux, and the application resource file contains
			// outdated defaults.
			SetHandleSize((Handle)prefResHandleApplication, sizeof(ApplicationPrefs));
			SetHandleSize((Handle)prefResHandleFTPServer, sizeof(FTPServerPrefs));
			SetHandleSize((Handle)prefResHandleSessionDefault, sizeof(SessionPrefs));
			SetHandleSize((Handle)prefResHandleTerminalDefault, sizeof(TerminalPrefs));
			SetHandleSize((Handle)prefResHandleWindowDefault, sizeof(WindowPrefs));
			
			// Create a string resource of ID -16397, which the Finder
			// will display to the user if he or she tries to open the
			// preferences file.
			if (prefResHandleDescriptionString != null) AddResource((Handle)prefResHandleDescriptionString, 'STR ', -16397,
																		EMPTY_PSTRING);
			
			#if TARGET_API_MAC_OS8
			// Create a balloon help resource of ID -5696, which the
			// Finder will display to the user if he or she points to
			// the icon of the preferences file while Balloon Help is
			// on.
			if (prefResHandleBalloonHelp != null) AddResource(prefResHandleBalloonHelp, kHMFinderApplResType, kHMHelpID,
																EMPTY_PSTRING);
			#endif
			
			// Add Telnets version resources to the preferences file.
			// Preferences need version information because they are
			// an integral part of this application.
			if (prefResHandleVers1 != null) AddResource(prefResHandleVers1, 'vers', 1, EMPTY_PSTRING);
			if (prefResHandleVers2 != null) AddResource(prefResHandleVers2, 'vers', 2, EMPTY_PSTRING);
			
			// ...put the master preferences in the Preferences file...
			AddResource((Handle)prefResHandleApplication, APPLICATIONPREFS_RESTYPE, APPLICATIONPREFS_ID, EMPTY_PSTRING);
			result = ResError();
			if (result == noErr)
			{
				AddResource((Handle)prefResHandleFTPServer, FTPSERVERPREFS_RESTYPE, FTPSERVERPREFS_ID, EMPTY_PSTRING);
				result = ResError();
				if (result == noErr)
				{
					StringUtilities_CToP(SESSIONPREFS_DEFAULTNAME, genericResourceName);
					AddResource((Handle)prefResHandleSessionDefault, SESSIONPREFS_RESTYPE, SESSIONPREFS_DEFAULTID, genericResourceName);
					result = ResError();
					if (result == noErr)
					{
						StringUtilities_CToP(TERMINALPREFS_DEFAULTNAME, genericResourceName);
						AddResource((Handle)prefResHandleTerminalDefault, TERMINALPREFS_RESTYPE, TERMINALPREFS_DEFAULTID, genericResourceName);
						result = ResError();
						if (result == noErr)
						{
							AddResource((Handle)prefResHandleWindowDefault, WINDOWPREFS_RESTYPE, WINDOWPREFS_ID, EMPTY_PSTRING);
							result = ResError();
							if (result == noErr)
							{
								// ...update the preferences
								ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
							}
						}
					}
				}
			}
		}
		
		// free critical resources memory referenced by the handles
		if (prefResHandleApplication != null) ReleaseResource((Handle)prefResHandleApplication), prefResHandleApplication = null;
		if (prefResHandleFTPServer != null) ReleaseResource((Handle)prefResHandleFTPServer), prefResHandleFTPServer = null;
		if (prefResHandleSessionDefault != null) ReleaseResource((Handle)prefResHandleSessionDefault),
												prefResHandleSessionDefault = null;
		if (prefResHandleTerminalDefault != null) ReleaseResource((Handle)prefResHandleTerminalDefault),
												prefResHandleTerminalDefault = null;
		if (prefResHandleWindowDefault != null) ReleaseResource((Handle)prefResHandleWindowDefault),
												prefResHandleWindowDefault = null;
		
		// free non-critical resources, if they were ever created
		if (prefResHandleBalloonHelp != null) ReleaseResource(prefResHandleBalloonHelp), prefResHandleBalloonHelp = null;
		if (prefResHandleDescriptionString != null) ReleaseResource((Handle)prefResHandleDescriptionString),
													prefResHandleDescriptionString = null;
		if (prefResHandleVers1 != null) ReleaseResource(prefResHandleVers1), prefResHandleVers1 = null;
		if (prefResHandleVers2 != null) ReleaseResource(prefResHandleVers2), prefResHandleVers2 = null;
	}
	
	UseResFile(oldResourceFile);
	return result;
}// initNewPreferences


/*!
To load preferences into memory, use this method.

(2.6)
*/
OSErr
loadAllPreferences	(void)
{
	ApplicationPrefs**	prefResHandleApplication = null;
	FTPServerPrefs**	prefResHandleFTPServer = null;
	OSErr				result = noErr;
	
	
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	result = ResError();
	
	// load window preferences
	if (result == noErr)
	{
		result = loadPreferencesStructure(gWindowPrefs, sizeof(WindowPrefs),
												WINDOWPREFS_RESTYPE, WINDOWPREFS_ID);
		if (result == resNotFound)
		{
			// copy defaults
			WindowPrefs**		prefResHandleWindowDefault = null;
			
			
			ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
			prefResHandleWindowDefault = (WindowPrefs**)Get1Resource(WINDOWPREFS_RESTYPE, WINDOWPREFS_DEFAULTID);
			result = ResError();
			if (result == noErr)
			{
				DetachResource((Handle)prefResHandleWindowDefault);
				ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
				SetHandleSize((Handle)prefResHandleWindowDefault, sizeof(WindowPrefs));
				AddResource((Handle)prefResHandleWindowDefault, WINDOWPREFS_RESTYPE, WINDOWPREFS_ID, EMPTY_PSTRING);
				result = ResError();
				if (result == noErr)
				{
					// ...update the preferences
					ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
				}
			}
		}
	}
	
	// load general preferences
	if (result == noErr) result = loadPreferencesStructure(gApplicationPrefs, sizeof(ApplicationPrefs),
															APPLICATIONPREFS_RESTYPE, APPLICATIONPREFS_ID);
	
	// load FTP server preferences
	if (result == noErr) result = loadPreferencesStructure(gFTPServerPrefs, sizeof(FTPServerPrefs),
															FTPSERVERPREFS_RESTYPE, FTPSERVERPREFS_ID);
	
	// initialize some stuff
	if (result == noErr)
	{
		Str255		notificationMessage;
		
		
		// the Interface Library Alert module is responsible for handling Notification Manager stuff...
		Alert_SetNotificationPreferences(gApplicationPrefs->notificationPrefs);
		GetIndString(notificationMessage, rStringsNoteAlerts, siNotificationAlert);
		Alert_SetNotificationMessage(notificationMessage);
		
		// the Interface Library Window Info module is responsible for handling dynamic resizing stuff...
		WindowInfo_SetDynamicResizing(null, gApplicationPrefs->dynamicResizing);
		
		// get ANSI colors
		{
			PaletteHandle		paletteHandle = null;
			
			
			if (getGeneralPreference(kTelnetPreferenceTag_ANSIColors, sizeof(paletteHandle), &paletteHandle,
										null/* actual size storage */) == kTelnetPrefsResultCodeSuccess)
			{
				if (paletteHandle != null)
				{
					Preferences_CopyANSIPaletteFrom(paletteHandle, 0/* start at this index */);
					DisposePalette(paletteHandle), paletteHandle = null;
				}
				else BEEP;
			}
		}
	}
	
	if (result == noErr) checkPrefsVersion();
	
	return result;
}// loadAllPreferences


/*!
To fill in a global structure from the preferences
file, use this method.  If the load succeeds,
"noErr" is returned.  Otherwise, any Resource
Manager error may be returned.

(3.0)
*/
static OSErr
loadPreferencesStructure	(void*		inStructurePtr,
							 Size		inStructureSize,
							 ResType	inResourceType,
							 SInt16		inResourceID)
{
	Handle		prefsResourceHandle = null;
	short		oldResourceFile = CurResFile();
	OSErr		result = noErr;
	
	
	ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
	result = ResError();
	if (result == noErr)
	{
		prefsResourceHandle = Get1Resource(inResourceType, inResourceID);
		result = ResError();
		if ((result == noErr) && (prefsResourceHandle != null))
		{
			BlockMove(*prefsResourceHandle, inStructurePtr, inStructureSize);
			ReleaseResource(prefsResourceHandle);
		}
	}
	
	UseResFile(oldResourceFile);
	return result;
}// loadPreferencesStructure


/*!
To mark an alias as having changed, use this
method.  This also marks the alias resource
by calling ChangedResource(), if the alias is
based on a resource.  Any resource error is
returned.

(3.0)
*/
static OSErr
setAliasChanged		(AliasInfoPtr	inAliasDataPtr)
{
	OSErr		result = noErr;
	
	
	if (inAliasDataPtr != null)
	{
		inAliasDataPtr->wasChanged = true;
		if (inAliasDataPtr->isResource)
		{
			ChangedResource((Handle)inAliasDataPtr->alias);
			result = ResError();
		}
	}
	else result = memPCErr;
	
	return result;
}// setAliasChanged


/*!
To write a global structure to the preferences
file, use this method.  If the save succeeds,
"noErr" is returned.  Otherwise, any Resource
Manager error may be returned.

(3.0)
*/
static OSErr
savePreferencesStructure	(void*		inStructurePtr,
							 Size		inStructureSize,
							 ResType	inResourceType,
							 SInt16		inResourceID)
{
	OSErr		result = noErr;
	
	
	if (inStructurePtr != null)
	{
		short		oldResourceFile = CurResFile();
		
		
		ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
		result = ResError();
		if (result == noErr)
		{
			Handle		prefsResourceHandle = Get1Resource(inResourceType, inResourceID);
			
			
			result = ResError();
			if (prefsResourceHandle == null) result = resNotFound;
			if (result == noErr)
			{
				BlockMove(inStructurePtr, *prefsResourceHandle, inStructureSize);
				ChangedResource(prefsResourceHandle);
				result = ResError();
				if (result == noErr)
				{
					WriteResource(prefsResourceHandle);
					result = ResError();
				}
				ReleaseResource(prefsResourceHandle), prefsResourceHandle = null;
			}
		}
		UseResFile(oldResourceFile);
	}
	else result = memPCErr;
	
	return result;
}// savePreferenceStructure


/*!
To set a preference from the File Transfers
preference class, use this method.

(3.0)
*/
static TelnetPrefsResultCode	
setFileTransfersPreference	(TelnetPreferenceTag	inDataPreferenceTag,
							 Size					inDataSize,
							 void const*			inDataPtr)
{
	TelnetPrefsResultCode		result = kTelnetPrefsResultCodeSuccess;
	
	
	switch (inDataPreferenceTag)
	{
		case kTelnetPreferenceTag_BinaryFileCreator:
			if (inDataSize >= sizeof(OSType))
			{
				OSType		data = *((OSType*)inDataPtr);
				
				
				gFTPServerPrefs->binaryCreator = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_BinaryFileType:
			if (inDataSize >= sizeof(OSType))
			{
				OSType		data = *((OSType*)inDataPtr);
				
				
				gFTPServerPrefs->binaryType = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_ServerLogInMessage:
			if (inDataSize >= sizeof(Handle))
			{
				SInt16		oldResFile = CurResFile();
				Handle		data = *((Handle*)inDataPtr),
							resHandle = null;
				
				
				ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
				resHandle = Get1Resource('TEXT', FTPSERVERPREFS_LOGINMESSAGEID);
				if ((ResError() != noErr) || (resHandle == null))
				{
					// resource doesnt exist, or couldnt be read; create a new one
					AddResource(data, 'TEXT', FTPSERVERPREFS_LOGINMESSAGEID, EMPTY_PSTRING);
					if (ResError() == noErr) ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
				}
				else
				{
					// resource does exist; change its contents
					Memory_SetHandleSize(resHandle, GetHandleSize(data));
					BlockMove(*data, *resHandle, GetHandleSize(data));
					ChangedResource(resHandle);
					if (ResError() == noErr) ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
				}
				UseResFile(oldResFile);
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_ServerMode:
			if (inDataSize >= sizeof(FTPServerMode))
			{
				FTPServerMode		data = *((FTPServerMode*)inDataPtr);
				
				
				gFTPServerPrefs->serverMode = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_ISOTranslation:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gFTPServerPrefs->useISOTranslation = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_UseMacBinaryII:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gFTPServerPrefs->useMacBinaryII = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_ResetMacBinary:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gFTPServerPrefs->resetMacBinary = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_LogInitiallyVisible:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gFTPServerPrefs->showLog = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TextFileCreator:
			if (inDataSize >= sizeof(OSType))
			{
				OSType		data = *((OSType*)inDataPtr);
				
				
				gFTPServerPrefs->textCreator = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TextFileType:
			if (inDataSize >= sizeof(OSType))
			{
				OSType		data = *((OSType*)inDataPtr);
				
				
				// this preference doesnt exist on disk (yet)
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_UseDNSNamesInLog:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gFTPServerPrefs->useDNSNamesForLog = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		default:
			// unrecognized tag
			result = kTelnetPrefsResultCodeUnknownTagOrClass;
			break;
	}
	
	return result;
}// setFileTransfersPreference


/*!
To set a preference from the General preference
class, use this method.

(3.0)
*/
static TelnetPrefsResultCode	
setGeneralPreference	(TelnetPreferenceTag	inDataPreferenceTag,
						 Size					inDataSize,
						 void const*			inDataPtr)
{
	TelnetPrefsResultCode		result = kTelnetPrefsResultCodeSuccess;
	
	
	switch (inDataPreferenceTag)
	{
		case kTelnetPreferenceTag_ANSIColors:
			if (inDataSize >= sizeof(PaletteHandle))
			{
				PaletteHandle		data = *((PaletteHandle*)inDataPtr),
									preferencePalette = null;
				SInt16				oldResFile = CurResFile();
				
				
				ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
				preferencePalette = GetNewPalette(ANSICOLORS_ID);
				CopyPalette(data/* source palette */, preferencePalette/* destination palette */,
							0/* source index */, 0/* destination index */, kANSIColorCount/* how many */);
				ChangedResource((Handle)preferencePalette);
				ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
				UseResFile(oldResFile);
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CaptureFileCreator:
			if (inDataSize >= sizeof(OSType))
			{
				OSType		data = *((OSType*)inDataPtr);
				
				
				gApplicationPrefs->captureFileCreator = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CopySelectedText:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->copySelectedText = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CopyTableThreshold:
			if (inDataSize >= sizeof(UInt16))
			{
				UInt16		data = *((UInt16*)inDataPtr);
				
				
				gApplicationPrefs->copyTableThreshold = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CursorBlinks:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->cursorBlinks = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_CursorMovesPriorToDrops:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->autoCursorMoveOnDrop = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_DontAutoClose:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->windowsDontGoAway = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_DontDimBackgroundScreens:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->dontDimBackgroundScreens = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_DynamicResizing:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->dynamicResizing = data;
				WindowInfo_SetDynamicResizing(null, data);
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_ExpireTickets:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->destroyKerberosTickets = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_FindFileAppSpec:
			if (inDataSize >= sizeof(FSSpec))
			{
				FSSpec		data = *((FSSpec*)inDataPtr);
				
				
				gApplicationPrefs->sherlockVRefNum = data.vRefNum;
				gApplicationPrefs->sherlockDirID = data.parID;
				PLstrcpy(gApplicationPrefs->sherlockName, data.name);
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_HeadersCollapsed:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->headersInitiallyCollapsed = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_MapBackquote:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->remapBackquoteToEscape = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_MenuItemKeys:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->menusHaveKeyEquivalents = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_Notification:
			if (inDataSize >= sizeof(SInt16))
			{
				SInt16		data = *((SInt16*)inDataPtr);
				
				
				gApplicationPrefs->notificationPrefs = data;
				Alert_SetNotificationPreferences(gApplicationPrefs->notificationPrefs);
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_NotifyOfBeeps:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->backgroundNotification = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_PureInverse:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->invertedTextHighlighting = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_ShowNewSessionsDialogAtStartup:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->newSessionsDialogOnStartup = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_SimpleTelnet:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->simpleTelnet = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TerminalCursorType:
			if (inDataSize >= sizeof(UInt16))
			{
				UInt16		data = *((UInt16*)inDataPtr);
				
				
				gApplicationPrefs->cursorType = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TicksWaitNextEvent:
			if (inDataSize >= sizeof(UInt32))
			{
				UInt32		data = *((UInt32*)inDataPtr);
				
				
				gApplicationPrefs->maxTicksWaitNextEvent = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TimeoutOpen:
			if (inDataSize >= sizeof(UInt32))
			{
				UInt32		data = *((UInt32*)inDataPtr);
				
				
				gApplicationPrefs->openTimeout = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_TimeoutSend:
			if (inDataSize >= sizeof(UInt32))
			{
				UInt32		data = *((UInt32*)inDataPtr);
				
				
				gApplicationPrefs->sendTimeout = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_UseBackgroundPicture:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->useBackgroundPicture = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_VisualBell:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->visualBell = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasClipboardShowing:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->wasClipboardShowing = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasCommandLineShowing:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->wasCommandLineShowing = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasControlKeypadShowing:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->wasControlKeypadShowing = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasFunctionKeypadShowing:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->wasFunctionKeypadShowing = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasMacroKeypadShowing:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->wasMacroKeypadShowing = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		case kTelnetPreferenceTag_WasVT220KeypadShowing:
			if (inDataSize >= sizeof(Boolean))
			{
				Boolean		data = *((Boolean*)inDataPtr);
				
				
				gApplicationPrefs->wasVT220KeypadShowing = data;
			}
			else result = kTelnetPrefsResultCodeInsufficientBufferSpace;
			break;
		
		default:
			// unrecognized tag
			result = kTelnetPrefsResultCodeUnknownTagOrClass;
			break;
	}
	
	return result;
}// setGeneralPreference