/*###############################################################

	Preferences.h
	
	Defines the brand new preferences management system that
	has been introduced in MacTelnet 3.0.  Now, you work with
	preferences in terms of classes, tags and values, so
	that there is NO direct access to internal preference data
	structures.  In effect, code outside of this module no
	longer has any control over how data is ultimately stored.
	In addition, file preferences are now stored as alias
	records, so they are not only more complete, but they are
	more likely to be salvageable in the event that the user
	changes the location or name of a file, etc.
	
	MacTelnet 3.0,  1998-2003 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.
	
	You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#ifndef __PREFERENCES__
#define __PREFERENCES__

// Mac includes
#include <MacTypes.h>
#include <Palettes.h>

// MacTelnet includes
#include "PreferencesData.h"



//
// constants
//

typedef SInt16 TelnetPrefsAliasID;
enum
{
	kInvalidTelnetPrefsAliasID = 0
};

typedef SInt32 TelnetPrefsResultCode;
enum
{
	kTelnetPrefsResultCodeSuccess = 0,							// no error
	kTelnetPrefsResultCodeUnknownTagOrClass = -1,				// specified preference tag or class tag isnt valid
	kTelnetPrefsResultCodeInsufficientBufferSpace = -2,			// memory space provided isnt large enough to hold data on disk
	kTelnetPrefsResultCodeBadVersionDataNotAvailable = -3,		// preferences file does not contain this information in any form
	kTelnetPrefsResultCodeBadVersionDataNewer = -4,				// preferences file contains more information than necessary
	kTelnetPrefsResultCodeBadVersionDataOlder = -5				// preferences file does not contain all necessary information
};

typedef FourCharCode TelnetPreferenceClass;
enum
{
	// All of the four-character codes for classes are listed
	// alphabetically by code, and must be unique.  ALL of
	// these tags are used with Preferences_SaveData(), but
	// "kTelnetPreferenceClassWindow" does not work with the
	// Preferences_SetData() or Preferences_GetData() methods.
	// Instead, use the special window management routines in
	// this module, Preferences_SaveWindowArrangement() and
	// Preferences_ArrangeWindow().
	kTelnetPreferenceClassGeneral		= FOUR_CHAR_CODE('appl'),
	kTelnetPreferenceClassFTP			= FOUR_CHAR_CODE('ftps'),
	kTelnetPreferenceClassWindow		= FOUR_CHAR_CODE('wind')	// use SaveWindowArrangement() and ArrangeWindow() with this class
};

typedef FourCharCode TelnetNamedPrefClass;
enum
{
	// All of the four-character codes for classes are listed
	// alphabetically by code, and must be unique.  These tags
	// are used only with the methods Preferences_SaveNamedData(),
	// Preferences_SetNamedData() and Preferences_GetNamedData().
	kTelnetNamedPrefClassProxyServer	= FOUR_CHAR_CODE('prox')
};

// All tags from the same preference class must have
// unique values.  The tags are grouped by class.
// When you call Preferences_GetxxxData(), make sure
// the storage space you provide is large enough to
// hold the data type indicated below for the tag you
// specify.  Similarly, with Preferences_SetxxxData(),
// the data pointer you provide should refer to data
// of the type that the tag expects.  In each case,
// the data *points* to storage of the type indicated.
typedef FourCharCode TelnetPreferenceTag;
enum
{
	// tags for kTelnetNamedPrefClassProxyServer
	kTelnetNamedPreferenceTag_ProxyServerObject				= FOUR_CHAR_CODE('psob')	// data: ProxyServerRef
};
enum
{
	// tags for kTelnetPreferenceClassFTP
	kTelnetPreferenceTag_BinaryFileCreator					= FOUR_CHAR_CODE('bnfc'),	// data: OSType
	kTelnetPreferenceTag_BinaryFileType						= FOUR_CHAR_CODE('bnft'),	// data: OSType
	kTelnetPreferenceTag_LogInitiallyVisible				= FOUR_CHAR_CODE('slas'),	// data: Boolean
	kTelnetPreferenceTag_ISOTranslation						= FOUR_CHAR_CODE('isot'),	// data: Boolean
	kTelnetPreferenceTag_ResetMacBinary						= FOUR_CHAR_CODE('rmcb'),	// data: Boolean
	kTelnetPreferenceTag_ServerLogInMessage					= FOUR_CHAR_CODE('slim'),	// data: Handle, a raw character stream
	kTelnetPreferenceTag_ServerMode							= FOUR_CHAR_CODE('ftpm'),	// data: FTPServerMode
	kTelnetPreferenceTag_TextFileCreator					= FOUR_CHAR_CODE('txfc'),	// data: OSType
	kTelnetPreferenceTag_TextFileType						= FOUR_CHAR_CODE('txft'),	// data: OSType
	kTelnetPreferenceTag_UseDNSNamesInLog					= FOUR_CHAR_CODE('udnr'),	// data: Boolean
	kTelnetPreferenceTag_UseMacBinaryII						= FOUR_CHAR_CODE('mcb2'),	// data: Boolean
	// tags for kTelnetPreferenceClassGeneral
	kTelnetPreferenceTag_ANSIColors							= FOUR_CHAR_CODE('ansi'),	// data: PaletteHandle
	kTelnetPreferenceTag_CaptureFileCreator					= FOUR_CHAR_CODE('cpfc'),	// data: OSType
	kTelnetPreferenceTag_CopySelectedText					= FOUR_CHAR_CODE('xtrm'),	// data: Boolean
	kTelnetPreferenceTag_CopyTableThreshold					= FOUR_CHAR_CODE('ctth'),	// data: UInt16, the number of spaces per tab
	kTelnetPreferenceTag_CursorBlinks						= FOUR_CHAR_CODE('curf'),	// data: Boolean
	kTelnetPreferenceTag_CursorMovesPriorToDrops			= FOUR_CHAR_CODE('curm'),	// data: Boolean
	kTelnetPreferenceTag_DontAutoClose						= FOUR_CHAR_CODE('wdga'),	// data: Boolean
	kTelnetPreferenceTag_DontDimBackgroundScreens			= FOUR_CHAR_CODE('wddb'),	// data: Boolean
	kTelnetPreferenceTag_DynamicResizing					= FOUR_CHAR_CODE('dynr'),	// data: Boolean
	kTelnetPreferenceTag_ExpireTickets						= FOUR_CHAR_CODE('dktk'),	// data: Boolean
	kTelnetPreferenceTag_FindFileAppSpec					= FOUR_CHAR_CODE('fndf'),	// data: FSSpec, where Find File or Sherlock is
	kTelnetPreferenceTag_HeadersCollapsed					= FOUR_CHAR_CODE('hdcl'),	// data: Boolean
	kTelnetPreferenceTag_MapBackquote						= FOUR_CHAR_CODE('map`'),	// data: Boolean
	kTelnetPreferenceTag_MenuItemKeys						= FOUR_CHAR_CODE('mkey'),	// data: Boolean
	kTelnetPreferenceTag_Notification						= FOUR_CHAR_CODE('noti'),	// data: a "kAlert_Notify..." constant
	kTelnetPreferenceTag_NotifyOfBeeps						= FOUR_CHAR_CODE('bnot'),	// data: Boolean
	kTelnetPreferenceTag_PureInverse						= FOUR_CHAR_CODE('pinv'),	// data: Boolean
	kTelnetPreferenceTag_ShowNewSessionsDialogAtStartup		= FOUR_CHAR_CODE('nd@s'),	// data: Boolean
	kTelnetPreferenceTag_SimpleTelnet						= FOUR_CHAR_CODE('simp'),	// data: Boolean
	kTelnetPreferenceTag_TerminalCursorType					= FOUR_CHAR_CODE('curs'),	// data: a "kTerminalCursorType..." constant
	kTelnetPreferenceTag_TicksWaitNextEvent					= FOUR_CHAR_CODE('twne'),	// data: UInt32
	kTelnetPreferenceTag_TimeoutOpen						= FOUR_CHAR_CODE('timo'),	// data: UInt32
	kTelnetPreferenceTag_TimeoutSend						= FOUR_CHAR_CODE('tims'),	// data: UInt32
	kTelnetPreferenceTag_UseBackgroundPicture				= FOUR_CHAR_CODE('upic'),	// data: Boolean
	kTelnetPreferenceTag_VisualBell							= FOUR_CHAR_CODE('visb'),	// data: Boolean
	kTelnetPreferenceTag_WasClipboardShowing				= FOUR_CHAR_CODE('wvcl'),	// data: Boolean
	kTelnetPreferenceTag_WasCommandLineShowing				= FOUR_CHAR_CODE('wvcm'),	// data: Boolean
	kTelnetPreferenceTag_WasControlKeypadShowing			= FOUR_CHAR_CODE('wvck'),	// data: Boolean
	kTelnetPreferenceTag_WasFunctionKeypadShowing			= FOUR_CHAR_CODE('wvfk'),	// data: Boolean
	kTelnetPreferenceTag_WasMacroKeypadShowing				= FOUR_CHAR_CODE('wvmk'),	// data: Boolean
	kTelnetPreferenceTag_WasVT220KeypadShowing				= FOUR_CHAR_CODE('wvvk')	// data: Boolean
};
enum
{
	// tags for kTelnetPreferenceClassWindow, only usable with the special window preference routines
	kTelnetWindowPreferenceTag_CommandLine					= FOUR_CHAR_CODE('Wcml'),
	kTelnetWindowPreferenceTag_ControlKeypad				= FOUR_CHAR_CODE('Wctk'),
	kTelnetWindowPreferenceTag_FunctionKeypad				= FOUR_CHAR_CODE('Wfnk'),
	kTelnetWindowPreferenceTag_MacroKeypad					= FOUR_CHAR_CODE('Wmck'),
	kTelnetWindowPreferenceTag_MacroSetup					= FOUR_CHAR_CODE('Wmcs'),
	kTelnetWindowPreferenceTag_NewSessions					= FOUR_CHAR_CODE('Wnws'),
	kTelnetWindowPreferenceTag_Preferences					= FOUR_CHAR_CODE('Wprf'),
	kTelnetWindowPreferenceTag_VT220Keypad					= FOUR_CHAR_CODE('Wvtk')
};

typedef SInt16 ANSIColorDescriptor;
enum
{
	// the values of these are part of the ANSI standard; dont change them!
	kANSIColorDescriptorBlack = 0,
	kANSIColorDescriptorRed = 1,
	kANSIColorDescriptorGreen = 2,
	kANSIColorDescriptorYellow = 3,
	kANSIColorDescriptorBlue = 4,
	kANSIColorDescriptorMagenta = 5,
	kANSIColorDescriptorCyan = 6,
	kANSIColorDescriptorWhite = 7,
	// these values are defined by adding 8 to the above
	kANSIColorDescriptorBlackEmphasized = 8,
	kANSIColorDescriptorRedEmphasized = 9,
	kANSIColorDescriptorGreenEmphasized = 10,
	kANSIColorDescriptorYellowEmphasized = 11,
	kANSIColorDescriptorBlueEmphasized = 12,
	kANSIColorDescriptorMagentaEmphasized = 13,
	kANSIColorDescriptorCyanEmphasized = 14,
	kANSIColorDescriptorWhiteEmphasized = 15,
	kANSIColorCount = 16
};



//
// public methods
//

/*###############################################################
	INITIALIZING AND FINISHING WITH PREFERENCES
###############################################################*/

void
	Preferences_Init					(void);

void
	Preferences_Done					(void);

TelnetPrefsResultCode
	Preferences_CreateOrFindFile		(void);

/*###############################################################
	ACCESSING PREFERENCES IN MEMORY
###############################################################*/

Point
	Preferences_ArrangeWindow			(WindowRef					inWindow,
										 Point						inInitialSize,
										 TelnetPreferenceTag		inWindowPreferenceTag);

Boolean
	Preferences_ArrangeWindowNoResize	(WindowRef					inWindow,
										 TelnetPreferenceTag		inWindowPreferenceTag);

// PROVIDE A PALETTE WITH AT LEAST "kANSIColorCount" COLORS RELATIVE TO THE GIVEN INDEX
void
	Preferences_CopyANSIPaletteFrom		(PaletteHandle				inCopyFromWhichPalette,
										 SInt16						inStartAtWhichIndex);

// PROVIDE A PALETTE WITH ENOUGH SPACE FOR "kANSIColorCount" COLORS RELATIVE TO THE GIVEN INDEX
void
	Preferences_CopyANSIPaletteTo		(PaletteHandle				outCopyToWhichPalette,
										 SInt16						inStartAtWhichIndex);

// NOT FOR RENDERING USE
void
	Preferences_GetANSIColor			(ANSIColorDescriptor		inWhichColor,
										 RGBColor*					outColorPtr);

TelnetPrefsResultCode
	Preferences_GetData					(TelnetPreferenceClass		inDataPreferenceClass,
										 TelnetPreferenceTag		inDataPreferenceTag,
										 Size						inDataStorageSize,
										 void*						outDataStorage,
										 Size*						outActualSizePtr);

TelnetPrefsResultCode
	Preferences_GetNamedData			(ConstStr255Param			inName,
										 TelnetPreferenceClass		inDataPreferenceClass,
										 TelnetPreferenceTag		inDataPreferenceTag,
										 Size						inDataStorageSize,
										 void*						outDataStorage,
										 Size*						outActualSizePtr);

SessionPrefs**
	Preferences_GetDefaultSession		(void);

TerminalPrefs**
	Preferences_GetDefaultTerminal		(void);

void
	Preferences_SetANSIColor			(ANSIColorDescriptor		inWhichColor,
										 RGBColor const*			inColorPtr);

TelnetPrefsResultCode
	Preferences_SetData					(TelnetPreferenceClass		inDataPreferenceClass,
										 TelnetPreferenceTag		inDataPreferenceTag,
										 Size						inDataSize,
										 void const*				inDataPtr);

TelnetPrefsResultCode
	Preferences_SetNamedData			(ConstStr255Param			inName,
										 TelnetNamedPrefClass		inDataPreferenceClass,
										 TelnetPreferenceTag		inDataPreferenceTag,
										 Size						inDataSize,
										 void const*				inDataPtr);

/*###############################################################
	MANAGING PREFERENCES ON DISK
###############################################################*/

OSErr
	Preferences_SaveData				(TelnetPreferenceClass		inDataPreferenceClass);

OSErr
	Preferences_SaveNamedData			(StringPtr					inName,
										 TelnetNamedPrefClass		inDataPreferenceClass);

void
	Preferences_SaveWindowArrangement	(WindowRef					inWindow,
										 TelnetPreferenceTag		inWindowPreferenceTag);

/*###############################################################
	ALIAS MANAGEMENT
###############################################################*/

TelnetPrefsAliasID
	Preferences_NewAlias				(FSSpec const*				inFileSpecificationPtr);

TelnetPrefsAliasID
	Preferences_NewSavedAlias			(TelnetPrefsAliasID			inAliasID);

void
	Preferences_DisposeAlias			(TelnetPrefsAliasID			inAliasID);

void
	Preferences_ChangeAlias				(TelnetPrefsAliasID			inAliasID,
										 FSSpec const*				inNewAliasFileSpecificationPtr);

void
	Preferences_DeleteSavedAlias		(TelnetPrefsAliasID			inAliasID);

Boolean
	Preferences_IsAliasStored			(TelnetPrefsAliasID			inAliasID);

Boolean
	Preferences_ParseAlias				(TelnetPrefsAliasID			inAliasID,
										 FSSpec*					outAliasFileSpecificationPtr);

void
	Preferences_SaveAlias				(TelnetPrefsAliasID			inAliasID,
										 ConstStr255Param			inName);

#endif