/*###############################################################

	PreferencesDialog.c
	
	MacTelnet 3.0,  1998-2003 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.
	
	You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Folders.h>
#include <Icons.h>
#include <LowMem.h>
#include <MacWindows.h>
#include <Resources.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "DialogResources.h"
#include "StringResources.h"
#include "SpacingConstants.r"
#include "TelnetPrefsDialog.r"

// MacTelnet includes
#include "configure.proto.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "MacLook.h"
#include "Panel.h"
#include "Preferences.h"
#include "PreferencesData.h"
#include "PreferencesDialog.h"
#include "PrefPanelConfigurations.h"
#include "PrefPanelFileTransfers.h"
#include "PrefPanelGeneral.h"
#include "PrefPanelMacros.h"
#include "PrefPanelScripts.h"
#include "TelnetDescriptors.h"



//
// types
//

struct PanelData
{
	PanelRef		panel;
	ControlRef		bevelButton;
};
typedef struct PanelData	PanelData;
typedef PanelData*			PanelDataPtr;

//
// variables
//

static PanelRef			gCurrentPanel = null;
static WindowInfoRef	gPreferencesWindowInfo = null;
static WindowRef		gPreferencesWindow = null; // the Mac OS window pointer
static ChainRef			gPanelList = null; // list of PanelData structures
static ControlRef		gControlButtonHelp = null,
						gControlPlacardBevelButtons = null,
						gControlPlacardPanels = null;

//
// internal methods
//

static PanelDataPtr		findPanelData					(PanelRef		inPanel);

static pascal Boolean	handleNewDialogSize				(WindowRef		inWindowRef,
											 			 SInt32			inDeltaSizeX,
														 SInt32			inDeltaSizeY,
														 SInt32			inData);

static void				initializePanel					(PanelRef		inPanel);

static Boolean			isHandledBevelButtonClick		(ControlRef		inControl);

static ControlRef		newBevelButton					(PanelRef		inPanel);



//
// public methods
//

/*!
This method is used to initialize the preferences
dialog box.  It creates the dialog box invisibly,
and uses the global preference data to set up the
corresponding controls in the dialog box panes.

You should call PreferencesDialog_Init() only if
you want the dialog in memory at all times.  It is
better to call PreferencesDialog_Display(), which
will automatically invoke PreferencesDialog_Init()
if necessary.  This allows you to create the dialog
only the first time it is necessary to display it,
and from then on to retain it in memory.

Be sure to call PreferencesDialog_Done() at the end
of the program to dispose of memory that this
module allocates.

(3.0)
*/
void
PreferencesDialog_Init	(void)
{
	Cursors_UseWatch();
	
	gPanelList = Chain_New(kChainStyleSinglyLinked);
	
	// create window and associated data
	gPreferencesWindow = GetNewCWindow(kWindowIDPreferences, null/* storage is not supported in Carbon */, (WindowRef)-1);
	(OSStatus)SetThemeWindowBackground(gPreferencesWindow, kThemeBrushModelessDialogBackgroundActive, false/* update */);
	gPreferencesWindowInfo = WindowInfo_New();
	
	if (gPreferencesWindow != null)
	{
		// concatenate MacTelnets name onto the end of the window title, so its obvious which applications Preferences window it is
		{
			Str255		titleString;
			
			
			GetWTitle(gPreferencesWindow, titleString);
			StringUtilities_PMetaSetDefaults();
			StringUtilities_PBuild(titleString, LMGetCurApName(), EMPTY_PSTRING, EMPTY_PSTRING, EMPTY_PSTRING);
			SetWTitle(gPreferencesWindow, titleString);
		}
		
		// create base controls
		{
			ControlRef		root = null;
			
			
			(OSErr)CreateRootControl(gPreferencesWindow, &root);
		}
		gControlButtonHelp = GetNewControl(kIDAppleGuideButton, gPreferencesWindow);
		Localization_AdjustHelpButtonControl(gControlButtonHelp);
		gControlPlacardBevelButtons = GetNewControl(kIDPrefBevelButtonPlacard, gPreferencesWindow);
		(OSErr)EmbedControl(gControlButtonHelp, gControlPlacardBevelButtons);
		gControlPlacardPanels = GetNewControl(kIDPrefPanelPlacard, gPreferencesWindow);
		
		// determine if the Help button should be disabled...
		SetUpHelpButtonControl(gControlButtonHelp);
		
		// if necessary, rearrange the panels for right-to-left localizations
		Localization_HorizontallyPlaceControls(gControlPlacardBevelButtons, gControlPlacardPanels);
		
		// create category panels - call these routines in the order you want their category buttons to appear
		PreferencesDialog_InstallPanel(PrefPanelGeneral_New());
		PreferencesDialog_InstallPanel(PrefPanelConfigurations_New());
		PreferencesDialog_InstallPanel(PrefPanelMacros_New());
		//PreferencesDialog_InstallPanel(PrefPanelScripts_New());
		PreferencesDialog_InstallPanel(PrefPanelFileTransfers_New());
		
		// use the preferred rectangle, if any
		{
			Point		deltaSize;
			
			
			SetPt(&deltaSize, PREFERENCES_DIALOG_WD, PREFERENCES_DIALOG_HT); // initially...
			deltaSize = Preferences_ArrangeWindow(gPreferencesWindow, deltaSize, kTelnetWindowPreferenceTag_Preferences);
			
			(Boolean)handleNewDialogSize(gPreferencesWindow, (SInt32)deltaSize.h, (SInt32)deltaSize.v, 0L);
		}
	}
}// Init


/*!
Call this method in the exit routine of the program
to ensure that this dialogs memory allocations
get destroyed.  You can also call this method after
you are done using this dialog, but then a call to
PreferencesDialog_Display() will re-initialize the
dialog before redisplaying it.

(3.0)
*/
void
PreferencesDialog_Done	(void)
{
	PreferencesDialog_Remove(); // saves preferences
	DisposeWindow(gPreferencesWindow), gPreferencesWindow = null; // automatically destroys all controls
	if (gPanelList != null)
	{
		ChainLinkRef	node = null;
		PanelDataPtr	dataPtr = null;
		
		
		node = Chain_GetFirstNode(gPanelList);
		do
		{
			dataPtr = Chain_GetNodeDataPtr(gPanelList, node);
			if (dataPtr != null)
			{
				Panel_Dispose(&dataPtr->panel);
				Memory_DisposePtr((Ptr*)&dataPtr);
			}
			node = Chain_GetNextNode(gPanelList, node);
		} while (node != null);
		
		Chain_Dispose(&gPanelList); // automatically destroys all nodes
	}
	WindowInfo_Dispose(gPreferencesWindowInfo);
}// Done


/*!
Use this method to display the preferences dialog box.
If the dialog is not yet in memory, it is created.

(3.0)
*/
void
PreferencesDialog_Display	(void)
{
	if (gPreferencesWindow == null) PreferencesDialog_Init();
	if (gPreferencesWindow != null)
	{
		DialogItemIndex		hit = 0;
		
		
		// define this only for debugging; this lets you dump the dialogs
		// control embedding hierarchy to a file, before it is displayed
		#if 0
		DebugSelectControlHierarchyDumpFile(gPreferencesWindow);
		#endif
		
		// display the dialog and handle events
		unless (IsWindowVisible(gPreferencesWindow)) ShowWindow(gPreferencesWindow);
		SelectOverRealFrontWindow(gPreferencesWindow);
		Cursors_UseArrow();
		
		// set up the Window Info information
		{
			WindowInfo_SetDynamicResizing(gPreferencesWindowInfo, false); // this particular window does not dynamically resize well
			WindowInfo_SetWindowDescriptor(gPreferencesWindowInfo, kTelnetDialogDescriptorPreferences);
			WindowInfo_SetWindowResizeLimits(gPreferencesWindowInfo,
												PREFERENCES_DIALOG_HT/* minimum height */, PREFERENCES_DIALOG_WD/* minimum width */,
												SHRT_MAX/* maximum height */, SHRT_MAX/* maximum width */);
			WindowInfo_SetWindowResizeResponderUPP(gPreferencesWindowInfo, handleNewDialogSize, 0L);
		}
		WindowInfo_SetForWindow(gPreferencesWindow, gPreferencesWindowInfo);
	}
	else Alert_ReportOSErr(memPCErr);
}// Display


/*!
Use this method to hide the preferences dialog box,
which will cause all changes to be saved.

The dialog remains in memory and can be re-displayed
using PreferencesDialog_Display().  To destroy the
dialog, use the method PreferencesDialog_Done().

(3.0)
*/
void
PreferencesDialog_Remove	(void)
{
	if (gPreferencesWindow != null)
	{
		ControlRef		focusControl = null;
		
		
		if (GetKeyboardFocus(gPreferencesWindow, &focusControl) == noErr)
		{
			// if the user is editing a text field, this makes sure the changes stick
			if (focusControl != null) Panel_Changed(gCurrentPanel, kPanel_MessageFocusLost, &focusControl);
		}
	}
	
	// write all the preference data in memory to disk
	Alert_ReportOSErr(Preferences_SaveData(kTelnetPreferenceClassGeneral));
	Alert_ReportOSErr(Preferences_SaveData(kTelnetPreferenceClassFTP));
	
	// save window size and location in preferences
	if (gPreferencesWindow != null)
	{
		Preferences_SaveWindowArrangement(gPreferencesWindow, kTelnetWindowPreferenceTag_Preferences);
		HideWindow(gPreferencesWindow);
	}
}// Remove


/*!
Since all tab panes in the preferences dialog box will
acquire the same size and location, this routine makes
sure the method of calculating that information is the
same for all of them.

(3.0)
*/
void
PreferencesDialog_CalculateTabPane		(PanelRef		inPanel,
										 Point*			outTabPaneTopLeft,
										 Point*			outTabPaneWidthHeight)
{
	if ((outTabPaneTopLeft != null) && (outTabPaneWidthHeight != null))
	{
		ControlRef		container = null;
		Rect			containerRect;
		
		
		Panel_GetContainerControl(inPanel, &container);
		GetControlBounds(container, &containerRect);
		
		SetPt(outTabPaneTopLeft, containerRect.left + HSP_TABPANE_AND_TAB, VSP_TAB_AND_DIALOG + TAB_HT_BIG + VSP_TABPANE_AND_TAB);
		SetPt(outTabPaneWidthHeight, containerRect.right - containerRect.left - INTEGER_DOUBLED(HSP_TABPANE_AND_TAB),
				containerRect.bottom - containerRect.top - VSP_TAB_AND_DIALOG - TAB_HT_BIG - INTEGER_DOUBLED(VSP_TABPANE_AND_TAB));
	}
}// CalculateTabPane


/*!
This method responds to mouse-down events in the
Preferences dialogs window by tracking content
controls and passing events off to the selected
panel.

(3.0)
*/
void
PreferencesDialog_HandleContentClick	(EventRecord*	inoutEventPtr)
{
	SInt16				windowPart = 0;
	ControlPartCode		controlPart = 0;
	WindowRef			window = null;
	ControlRef			controlClicked = null,
						focusControl = null;
	
	
	windowPart = FindWindow(inoutEventPtr->where, &window);
	if (window != null)
	{
		GrafPtr		oldPort = null;
		Point		localMouse = inoutEventPtr->where;
		OSErr		error = noErr;
		
		
		GetPort(&oldPort);
		SetPortWindowPort(window);
		GlobalToLocal(&localMouse);
		controlClicked = FindControlUnderMouse(localMouse, window, &controlPart);
		error = GetKeyboardFocus(window, &focusControl);
		
		// if a focusable control is clicked and not focused, focus it...
		{
			UInt32		featureBits = 0L;
			
			
			(OSErr)GetControlFeatures(controlClicked, &featureBits);
			if ((featureBits & kControlGetsFocusOnClick) && ((focusControl != controlClicked) || (error != noErr)))
			{
				if (focusControl != null) Panel_Changed(gCurrentPanel, kPanel_MessageFocusLost, &focusControl);
				(OSErr)SetKeyboardFocus(window, controlClicked, kControlFocusNextPart);
				Panel_Changed(gCurrentPanel, kPanel_MessageFocusGained, &controlClicked);
			}
		}
		
		if ((controlClicked != null) && (controlPart != kControlNoPart))
		{
			Boolean			handled = false;
			
			
			// ...otherwise, handle mouse-down events in controls
			controlPart = HandleControlClick(controlClicked, localMouse, inoutEventPtr->modifiers, GetControlAction(controlClicked));
			switch (controlPart)
			{
				case kControlNoPart:
					handled = true; // do NOTHING else if the click is meaningless
					break;
				
				case kControlButtonPart:
					// help button?
					if (controlClicked == gControlButtonHelp)
					{
						Str255		searchString;
						
						
						GetIndString(searchString, rStringsAppleGuideLookup, siAGIndexTermPreferences);
						AppleGuideUtilities_OpenFirstMainGuideInSearch(searchString);
						handled = true;
					}
					
					// this special invocation is necessary to allow bevel button clicks to be received last
					// (I will probably re-design the bevel button handler to be smarter, in the future)
					unless (handled)
					{
						handled = (Panel_Changed(gCurrentPanel, kPanel_MessageClicked, &controlClicked) == kPanel_ResponseClickHandled);
					}
					
					// if all else fails, check to see if the user clicked a category button
					unless (handled) handled = isHandledBevelButtonClick(controlClicked);
					break;
				
				default:
					break;
			}
			
			// If the Preferences dialog itself cant handle a click, it must have
			// been in a control owned by a panel; notify the panel of the click.
			unless (handled) Panel_Changed(gCurrentPanel, kPanel_MessageClicked, &controlClicked);
		}
		SetPort(oldPort);
	}
}// HandleContentClick


/*!
This routine responds to key-down events in items in
the Preferences dialog box.

(3.0)
*/
Boolean
PreferencesDialog_HandleKeyDown		(EventRecord*	inoutEventPtr)
{
	Boolean			result = true;
	SInt16			keyCode = ((inoutEventPtr->message & keyCodeMask) >> 8),
					charCode = inoutEventPtr->message & charCodeMask,
					modifiers = inoutEventPtr->modifiers;
	
	
	if (charCode == 0x09)
	{
		ControlRef		focusControl = null;
		OSErr			error = noErr;
		
		
		// tab key was pressed; advance or reverse keyboard focus
		error = GetKeyboardFocus(gPreferencesWindow, &focusControl);
		if (focusControl == null) error = memPCErr;
		if (error == noErr) Panel_Changed(gCurrentPanel, kPanel_MessageFocusLost, &focusControl);
		
		if (modifiers & shiftKey)
		{
			(OSErr)ReverseKeyboardFocus(gPreferencesWindow);
		}
		else
		{
			(OSErr)AdvanceKeyboardFocus(gPreferencesWindow);
		}
		error = GetKeyboardFocus(gPreferencesWindow, &focusControl);
		if (error == noErr) Panel_Changed(gCurrentPanel, kPanel_MessageFocusGained, &focusControl);
	}
	else
	{
		ControlRef		control = null;
		
		
		if (GetKeyboardFocus(gPreferencesWindow, &control) == noErr)
		{
			(SInt16)HandleControlKey(control, keyCode, charCode, modifiers);
		}
	}
	return result;
}// HandleKeyDown


/*!
To add a panel to the preferences dialog from
an abstract Panel object, use this method.  A
preference panel should define a name and an
icon (as recommended by the Panel module anyway),
which is used to render a bevel button for the
panel.  The preferences dialog box automatically
adapts itself for the panel, creating and
positioning a new bevel button to represent it
and firing notifications to the panel whenever
necessary.  If you respond to the standard Panel
interface, the panel will function perfectly in
the preferences dialog box.

There is no way to arrange preference panels
except by invoking PreferencesDialog_InstallPanel()
for each panel in the order you want the bevel
buttons to appear in the dialog.

(3.0)
*/
void
PreferencesDialog_InstallPanel		(PanelRef		inPanel)
{
	ControlRef		button = newBevelButton(inPanel);
	
	
	initializePanel(inPanel);
	
	// if this is the first button, highlight it
	if (Chain_GetNodeCount(gPanelList) == 1)
	{
		(Boolean)isHandledBevelButtonClick(button);
		gCurrentPanel = inPanel;
	}
}// InstallPanel


//
// internal methods
//
#pragma mark -

/*!
To acquire a pointer to the internal structure
whose panel field corresponds to the given
panel, use this method.

(3.0)
*/
static PanelDataPtr
findPanelData		(PanelRef		inPanel)
{
	ChainLinkRef	node = null;
	PanelDataPtr	result = null;
	
	
	// look for the given panel in the linked list
	node = Chain_GetFirstNode(gPanelList);
	do
	{
		result = Chain_GetNodeDataPtr(gPanelList, node);
		if (result != null) if (result->panel == inPanel) break;
		node = Chain_GetNextNode(gPanelList, node);
	} while (node != null);
	
	return result;
}// findPanelData


/*!
This routine is called whenever the window is
resized, to make sure dialog controls match the
new dialog size.

(3.0)
*/
static pascal Boolean
handleNewDialogSize		(const WindowRef	inWindowRef,
				 		 const SInt32		inDeltaSizeX,
						 const SInt32		inDeltaSizeY,
						 const SInt32		UNUSED_ARGUMENT(inData))
{
	SInt16		result = 0; // unused and arbitrary - define it if you need it
	
	
	if (inWindowRef == gPreferencesWindow) // they should be the same!
	{
		ControlRef		control = null;
		Rect			rect;
		Boolean			dynamicResizing = false;
		Size			actualSize = 0L;
		
		
		// get the users Dynamic Resizing preference, if possible
		unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
									kTelnetPreferenceTag_DynamicResizing,
									sizeof(dynamicResizing), &dynamicResizing,
									&actualSize) == kTelnetPrefsResultCodeSuccess)
		{
			dynamicResizing = false; // assume no dynamic resizing if the preference cant be found
		}
		
		unless (dynamicResizing) Cursors_DeferredUseWatch(1); // on slow computers, display a watch cursor after 1/60 a second!
		
		// resize base items
		#if 1
		control = gControlPlacardBevelButtons;
		GetControlBounds(control, &rect);
		unless (Localization_IsLeftToRight()) MoveControl(control, rect.left + inDeltaSizeX, rect.top);
		SizeControl(control, rect.right - rect.left, rect.bottom - rect.top + inDeltaSizeY);
		control = gControlPlacardPanels;
		GetControlBounds(control, &rect);
		SizeControl(control, rect.right - rect.left + inDeltaSizeX, rect.bottom - rect.top + inDeltaSizeY);
		control = gControlButtonHelp;
		GetControlBounds(control, &rect);
		MoveControl(control, rect.left, rect.top + inDeltaSizeY);
		#else
		DialogAdjust_BeginControlAdjustment(inWindowRef);
		
		control = gControlPlacardBevelButtons;
		unless (Localization_IsLeftToRight()) DialogAdjust_AddControlAdjustment(kControlAdjustmentMoveH, control, inDeltaSizeX);
		DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeV, control, inDeltaSizeY);
		
		control = gControlPlacardPanels;
		DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeH, control, inDeltaSizeX);
		DialogAdjust_AddControlAdjustment(kControlAdjustmentResizeV, control, inDeltaSizeY);
		
		control = gControlButtonHelp;
		DialogAdjust_AddControlAdjustment(kControlAdjustmentMoveV, control, inDeltaSizeY);
		
		DialogAdjust_EndAdjustment(inDeltaSizeX, inDeltaSizeY);
		#endif
		
		// resize panels
		{
			Point			deltaSize;
			ChainLinkRef	node = null;
			PanelDataPtr	dataPtr = null;
			
			
			SetPt(&deltaSize, inDeltaSizeX, inDeltaSizeY);
			node = Chain_GetFirstNode(gPanelList);
			do
			{
				dataPtr = Chain_GetNodeDataPtr(gPanelList, node);
				if (dataPtr != null)
				{
					ControlRef		container = null;
					Rect			rect;
					
					
					// adjust the size of this panel in the list
					Panel_GetContainerControl(dataPtr->panel, &container);
					GetControlBounds(container, &rect);
					SizeControl(container, rect.right - rect.left + deltaSize.h, rect.bottom - rect.top + deltaSize.v);
					Panel_Changed(dataPtr->panel, kPanel_MessageDeltaSize, &deltaSize);
				}
				node = Chain_GetNextNode(gPanelList, node);
			} while (node != null);
		}
		
		Cursors_UseArrow();
	}
	
	return result;
}// handleNewDialogSize


/*!
Initializes a panel by passing one-time property
changes to it.

(3.0)
*/
static void
initializePanel		(PanelRef		inPanel)
{
	Rect			initialRect;
	Point			initialDimensions;
	ControlRef		container = null,
					placard = gControlPlacardPanels;
	
	
	GetControlBounds(placard, &initialRect);
	
	// allow the panel to create its controls by providing a pointer to the containing window
	Panel_Changed(inPanel, kPanel_MessageCreateControls, &gPreferencesWindow);
	
	// Allow the panel to size and arrange its controls by setting the initial dimensions of
	// the container.  With Appearance, there is the concept of a control embedding hierarchy.
	// This works in your favor, but only if you know how to use it effectively.  The best order
	// of operations, when working with Appearance controls, is as follows:
	//	1.	Create all of your controls.
	//	2.	Embed each control in its immediate container (its OK at this point if a control is
	//		not visually contained in its embedders).
	//	3.	Set the dimensions of embedders, outside-in.  This allows you to use the dimensions
	//		of other controls (such as group boxes) to help set the size of embedded items.
	//	4.	Set the positions of controls, inside-out.  This allows you to set the position of
	//		all controls to be relative to their containers!  When you move a control that
	//		embeds other controls, the embedded controls will offset automatically.
	SetPt(&initialDimensions, initialRect.right - initialRect.left, initialRect.bottom - initialRect.top);
	Panel_GetContainerControl(inPanel, &container);
	(OSErr)EmbedControl(container, placard);
	SizeControl(container, initialDimensions.h, initialDimensions.v);
	Panel_Changed(inPanel, kPanel_MessageInitialSize, &initialDimensions);
	MoveControl(container, initialRect.left, initialRect.top); // move afterwards, so the embedded controls move too
}// initializePanel


/*!
This routine assumes the specified control is one of
the bevel buttons for an installed preference panel,
and tries to hide/show the appropriate panels.  You
can also use this routine to automatically select and
display a panel if no other panels are selected.

(3.0)
*/
static Boolean
isHandledBevelButtonClick	(ControlRef		inControl)
{
	Boolean			result = false;
	
	
	// dont waste time if the apparently-clicked control is already selected
	unless (IsControlHilited(inControl))
	{
		register SInt16		i = 0;
		ChainLinkRef		node = Chain_GetFirstNode(gPanelList);
		PanelDataPtr		dataPtr = null,
							newPanelDataPtr = null;
		ControlRef			container = null,
							nowInvisibleContainer = null;
		
		
		do
		{
			dataPtr = Chain_GetNodeDataPtr(gPanelList, node);
			if (dataPtr != null)
			{
				Panel_GetContainerControl(dataPtr->panel, &container);
				
				// if necessary, initialize the current panel
				if (gCurrentPanel == null) gCurrentPanel = dataPtr->panel;
				
				// note which panel was displayed before (if any)
				if (gCurrentPanel == dataPtr->panel)
				{
					Boolean			visible = false;
					
					
					// this panels button is currently selected, so this is the panel that will disappear
					nowInvisibleContainer = container;
					Panel_Changed(dataPtr->panel, kPanel_MessageNewVisibility, &visible);
				}
				
				// unhighlight all buttons except the selected button, and remember
				// which panel is to be displayed next (instead of the previous one)
				if (inControl != dataPtr->bevelButton)
				{
					SetControlValue(dataPtr->bevelButton, kControlCheckBoxUncheckedValue);
				}
				else
				{
					// Then this is the category bevel button that was
					// clicked (so its panel will be displayed).  Store
					// this special data pointer.
					newPanelDataPtr = dataPtr;
				}
			}
			node = Chain_GetNextNode(gPanelList, node);
		} while (node != null);
		
		// show the new (chosen) panel
		if (newPanelDataPtr != null)
		{
			ControlRef		nowVisibleContainer = null;
			Boolean			visible = true;
			
			
			result = true;
			Panel_GetContainerControl(newPanelDataPtr->panel, &nowVisibleContainer);
			SetControlValue(newPanelDataPtr->bevelButton, kControlCheckBoxCheckedValue);
			gCurrentPanel = newPanelDataPtr->panel;
			(OSErr)ClearKeyboardFocus(gPreferencesWindow);
			Panel_Changed(dataPtr->panel, kPanel_MessageNewVisibility, &visible);
			
			// swap panels
			Embedding_OffscreenSwapOverlappingControls(gPreferencesWindow, nowInvisibleContainer, nowVisibleContainer);
			
			// fix a minor cosmetic glitch caused by the fact that the panel and the placard overlap by one pixel
			{
				Rect		placardRect;
				
				
				GetControlBounds(gControlPlacardBevelButtons, &placardRect);
				placardRect.left = placardRect.right - 1; // invalidate the sliver (one pixel) of overlap
				MacLook_InvalidateRectangle(gPreferencesWindow, &placardRect); // redraw this area
			}
		}
	}
	
	return result;
}// isHandledBevelButtonClick


/*!
To create a new bevel button control for the preferences
dialog box that represents the specified panel, use this
method.  The button title and icon are acquired from the
specified panel, and the control reference constant is
set to be the specified panel reference.  The button is
returned.

(3.0)
*/
static ControlRef
newBevelButton		(PanelRef		inPanel)
{
	ControlRef		result = GetNewControl(kIDPrefGenericBevelButton, gPreferencesWindow);
	
	
	if (result != null)
	{
		ControlRef			label = GetNewControl(kIDStaticText, gPreferencesWindow),
							control = null; // generic ControlRef, used to represent the current control in a series of actions
		ChainLinkRef		node = null;
		PanelDataPtr		dataPtr = (PanelDataPtr)Memory_NewPtr(sizeof(PanelData));
		SInt16				labelHeight = 16; // I cant think of a reliable way to calculate this dynamically (yet)
		
		
		// store the panel
		dataPtr->bevelButton = result;
		dataPtr->panel = inPanel;
		node = Chain_NewNode(gPanelList, dataPtr);
		Chain_MoveNode(gPanelList, node, null/* means end of list */, kChainLinkPlacementInsertAfter);
		
		// associate the panel with the button
		SetControlReference(result, REINTERPRET_CAST(inPanel, SInt32));
		
		// make the button use the name of the panel, and set up the control font attributes for the button label
		{
			ControlFontStyleRec		style;
			Str31					title;
			
			
			control = label;
			Panel_GetName(inPanel, title);
			style.flags = kControlUseFontMask | kControlUseSizeMask | kControlUseJustMask;
			style.font = kControlFontSmallSystemFont;
			style.size = 10;
			style.just = teCenter;
			(OSErr)SetControlFontStyle(control, &style);
			(OSErr)SetControlData(control, kControlEditTextPart, kControlEditTextTextTag,
									StrLength(title) * sizeof(unsigned char), title + 1/* length byte is skipped */);
			SetControlText(result, title);
		}
		
		// make the button use the icon of the panel
		(OSErr)Panel_SetButtonIcon(result, inPanel);
		
		// embed the bevel button control in its container, and arrange the new button relative to other buttons in the same group
		{
			ControlRef		placard = gControlPlacardBevelButtons, // use a variable just in case it changes in the future (less alterations)
							buttonPane = null,
							panelPane = null;
			UInt16			numberOfButtons = Chain_GetNodeCount(gPanelList);
			UInt16			paneHeight = 0;
			
			
			// determine the height the pane should have
			{
				Rect		buttonRect;
				
				
				GetControlBounds(result, &buttonRect);
				paneHeight = buttonRect.bottom - buttonRect.top + VSP_BUTTON_AND_TEXT_LABEL + labelHeight;
			}
			
			// create an invisible embedder control that will tie together the button and its label
			buttonPane = GetNewControl(kIDAnchorUserPane, gPreferencesWindow);
			(OSErr)EmbedControl(buttonPane, placard);
			(OSErr)EmbedControl(result, buttonPane);
			(OSErr)EmbedControl(label, buttonPane);
			
			// center the button and its label horizontally in the placard
			{
				Rect		paneRect,		// rectangle of area just encompassing the bevel button and its label
							placardRect,	// rectangle of placard in which the buttons sit
							buttonRect;		// rectangle of bevel button (only dimensions are important)
				
				
				GetControlBounds(placard, &placardRect);
				
				// size the controls (the bevel buttons size is defined in the resource file)
				SizeControl(buttonPane, placardRect.right - placardRect.left - INTEGER_DOUBLED(HSP_BUTTON_AND_DIALOG), paneHeight);
				GetControlBounds(buttonPane, &paneRect);
				SizeControl(label, paneRect.right - paneRect.left, labelHeight);
				
				// locations are relative to the embedders origin
				control = result;
				GetControlBounds(result, &buttonRect);
				MoveControl(control, INTEGER_HALVED(paneRect.right - paneRect.left - (buttonRect.right - buttonRect.left)), buttonRect.top);
				MoveControl(label, paneRect.left, paneRect.bottom - labelHeight);
			}
			
			// arrange the button pane, which automatically moves both the bevel button and its label
			Localization_VerticallyPlaceControls(result, label); // switches them around for bottom-to-top localizations
			MoveControl(buttonPane, HSP_BUTTON_AND_DIALOG, VSP_BUTTON_AND_DIALOG + (numberOfButtons - 1) * (paneHeight + VSP_BUTTONS));
		}
	}
	
	return result;
}// newBevelButton