/*###############################################################

	Printing.c
	
	Universal Printing Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#define PM_USE_SESSION_APIS 0

// Mac includes
#include <ControlDefinitions.h>
#if TARGET_API_MAC_OS8
#	include <Printing.h>
#else
#	include <PMApplication.h>
#endif

// library includes
#include "InterfaceLib.h"
#include "PrintingLib.h"
#include "PrintingLib.r"



// types
struct UniversalPrintRecord
{
	OSStatus			lastResult;			// error code from last Print Manager / Carbon Print Manager call
#if TARGET_API_MAC_OS8
	// data required by the Mac OS 8 printing architecture
	THPrint				recordHandle;
	TPPrPort			printingPortPtr;
#else
	// data required by the Carbon Printing Manager
	PMPrintSettings		settings;
	PMPageFormat		pageFormat;
	PMPrintContext		documentContext;
#endif
};
typedef struct UniversalPrintRecord		UniversalPrintRecord;
typedef UniversalPrintRecord*			UniversalPrintRecordPtr;
typedef UniversalPrintRecordPtr*		UniversalPrintRecordHandle;

// variables
static UniversalPrintingMode	gPrintingMode = kUniversalPrintingModeNormal;

// internal methods
#if !TARGET_API_MAC_OS8
static OSStatus						flattenAndCopy		(PMPrintSettings			inPrintSettingsToFlatten,
														 PMPageFormat				inPageFormatToFlatten,
														 Handle						inoutPrintSettingsFlattenedCopy,
														 Handle						inoutPageFormatFlattenedCopy);
#endif
static UniversalPrintRecordPtr		refAcquireLock		(UniversalPrintRecordRef	inRef);
static void							refReleaseLock		(UniversalPrintRecordRef	inRef);



//
// public methods
//

/*!
This method returns the version of this
library.  The high 8 bits represent the
major version number.  The next 11 bits
represent the second decimal version
number.  The following 11 bits represent
the third decimal version number.  The
bottom 2 bits combine to form a number
that is equal to one of the kReleaseKind...
constants, indicating the kind of version
(final, development, alpha or beta).

(1.0)
*/
ApplicationSharedLibraryVersion
Printing_Version	(void)
{
	return Releases_Version(PRINTING_LIBRARY_MAJOR_REVISION, PRINTING_LIBRARY_MINOR_REVISION, 0, kReleaseKindAlpha);
}// Version


/*!
Call this routine to set up the printing
library.  You should call this in the same
place in your code where you would normally
call PrOpen().

(1.0)
*/
void
Printing_Init	(void)
{
#if TARGET_API_MAC_OS8
	PrOpen();
#else
	(OSStatus)PMBegin();
#endif
}// Init


/*!
Call this routine to shut down the printing
library.  You should call this in the same
place in your code where you would normally
call PrClose().

(1.0)
*/
void
Printing_Done	(void)
{
#if TARGET_API_MAC_OS8
	PrClose();
#else
	(OSStatus)PMEnd();
#endif
}// Done


/*!
Returns a constant describing the printing
architecture that this library supports.
The traditional architecture refers to
the Print Manager in place for Mac OS 8.6
and earlier.  The Mac OS X architecture
refers to the Carbon printing architecture
that is in place in Mac OS 9.

Since this library handles printing totally
transparently, you only need to check this
architecture setting to ensure that the
installed printing library will run on the
users system.  A Mac OS 8 version will not
run under Mac OS X, and vice-versa; however,
either version can be used under Mac OS 9.

(1.0)
*/
UniversalPrintingArchitecture
Printing_GetArchitecture		(void)
{
	UniversalPrintingArchitecture		result =
#if TARGET_API_MAC_OS8
					kUniversalPrintingArchitectureTraditional;
#else
					kUniversalPrintingArchitectureMacOSX;
#endif
	
	
	return result;
}// GetArchitecture


/*!
To allocate storage for an architecture-dependent
data structure and acquire an opaque reference to
it, use this method.  By using this routine to
manipulate printing data, you effectively write
the same code whether this library is built for
Mac OS 8 or Mac OS X!

(1.0)
*/
UniversalPrintRecordRef
Printing_NewRecord	(void)
{
	UniversalPrintRecordRef		result = (UniversalPrintRecordRef)NewHandle(sizeof(UniversalPrintRecord));
	
	
	if (result != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(result);
		
		
		ptr->lastResult = noErr;
#if TARGET_API_MAC_OS8
		ptr->recordHandle = (THPrint)NewHandle(sizeof(TPrint));
		ptr->lastResult = MemError();
		if (ptr->lastResult == noErr)
		{
			PrintDefault(ptr->recordHandle);
			ptr->lastResult = PrError();
		}
		ptr->printingPortPtr = null; // this is defined by PrOpenDoc()
#else
		ptr->lastResult = PMNewPrintSettings(&ptr->settings);
		if (ptr->lastResult == noErr) ptr->lastResult = PMDefaultPrintSettings(ptr->settings);
		if (ptr->lastResult == noErr) ptr->lastResult = PMNewPageFormat(&ptr->pageFormat);
		if (ptr->lastResult == noErr) ptr->lastResult = PMDefaultPageFormat(ptr->pageFormat);
		ptr->documentContext = null; // this is defined by PMBeginDocument()
#endif
		refReleaseLock(result);
	}
	return result;
}// NewRecord


/*!
To destroy storage for an architecture-dependent
data structure to which you previously acquired
a reference using Printing_NewRecord(), use this
routine.  On output, your copy of the reference
is set to null.

(1.0)
*/
void
Printing_DisposeRecord	(UniversalPrintRecordRef*		inoutRefPtr)
{
	if (inoutRefPtr != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(*inoutRefPtr);
		
		
#if TARGET_API_MAC_OS8
		DisposeHandle((Handle)ptr->recordHandle), ptr->recordHandle = null;
#else
		(OSStatus)PMDisposePrintSettings(ptr->settings), ptr->settings = null;
		(OSStatus)PMDisposePageFormat(ptr->pageFormat), ptr->pageFormat = null;
#endif
		DisposeHandle((Handle)*inoutRefPtr), *inoutRefPtr = null;
	}
}// DisposeRecord


/*!
To start a new printing document, use this method.
All context information, port information, etc. is
retained via the opaque reference, so you need only
call the corresponding Printing_EndDocument() method
with the same opaque reference to end a document.
The Printing_BeginPage() and Printing_EndPage()
routines automatically know the document as well,
so you can invoke those routines between document
opening and closing and get the expected results.

(1.0)
*/
void
Printing_BeginDocument	(UniversalPrintRecordRef	inRef)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		ptr->printingPortPtr = PrOpenDoc(ptr->recordHandle, null/* port */, null/* I/O buffer */);
		ptr->lastResult = PrError();
#else
		ptr->lastResult = PMBeginDocument(ptr->settings, ptr->pageFormat, &ptr->documentContext);
#endif
		refReleaseLock(inRef);
	}
}// BeginDocument


/*!
To start a new printing page, use this method.  All
context information, port information, etc. is
retained via the opaque reference, so you need only
call the corresponding Printing_EndPage() method
with the same opaque reference to end a page.  The
Printing_BeginDocument() and Printing_EndDocument()
routines should bracket calls to page routines (and
the page automatically refers to such a document).

You can specify "null" for the page boundaries, in
which case default boundaries are used.

(1.0)
*/
void
Printing_BeginPage	(UniversalPrintRecordRef	inRef,
					 const Rect*				inPageBoundsPtr)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		{
			Rect		rect = *inPageBoundsPtr;
			
			
			PrOpenPage(ptr->printingPortPtr, &rect);
			ptr->lastResult = PrError();
		}
#else
		{
			PMRect		preciseBounds;
			PMRect*		boundsPtr = null;
			
			
			if (inPageBoundsPtr != null)
			{
				preciseBounds.top = inPageBoundsPtr->top;
				preciseBounds.left = inPageBoundsPtr->left;
				preciseBounds.bottom = inPageBoundsPtr->bottom;
				preciseBounds.right = inPageBoundsPtr->right;
				boundsPtr = &preciseBounds;
			}
			ptr->lastResult = PMBeginPage(ptr->documentContext, boundsPtr);
		}
#endif
		refReleaseLock(inRef);
	}
}// BeginPage


/*!
Cancels printing in progress.  Under Carbon, this
is not possible, so calling this routine has no
effect.  On the other hand, you generally call this
routine from an idle procedure, and Carbon handles
manual print spooling, etc. anyway.  Since a print
job is likely to be spooled to disk and printed
under Carbons control, and not your applications
explicit loop, you are never responsible for the
cancellation of a print job anyway.  The user can
always drag documents out of a desktop printer to
cancel them (and this behavior is essentially the
case for most drivers under Mac OS 8, as well).

(1.0)
*/
void
Printing_Cancel		(UniversalPrintRecordRef	inRef)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
		ptr->lastResult = noErr;
#if TARGET_API_MAC_OS8
		(*(ptr->recordHandle))->prJob.fFromUsr = true;
#endif
		refReleaseLock(inRef);
	}
}// Cancel


/*!
Use this convenient routine for filling in the
information in an opaque printing record using
architecture-specific data you retrieved (most
likely) from a resource file.  You must specify
a printing architecture ahead of time (by
setting the appropriate field in the specified
structure), and you must create valid handles
that already contain your data.  In the case of
a Mac OS 8 architecture, the structure dictates
that a handle to a filled-in printing record
must be provided.  For the Mac OS X architecture
the structure requires two handles, respectively
to flattened print settings and flattened
page format data.  If the structure contains an
unknown architecture, "cfragArchitectureErr" is
returned by Printing_GetLastResult().

IMPORTANT:	A Pre-Carbon version of this library
			will fail to retrieve data stored in
			a Carbon Printing Manager format,
			whereas the reverse is not true.  If
			the data cant be read, the error
			"cfragArchitectureErr" is returned by
			Printing_GetLastResult().

(1.0)
*/
void
Printing_CopyFromSaved		(UniversalPrintRecordRef		inRef,
							 SavedUniversalPrintRecordPtr	inArchitectureSpecificDataPtr)
{
	if ((inRef != null) && (inArchitectureSpecificDataPtr != null))
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
		switch (inArchitectureSpecificDataPtr->architecture)
		{
			case kUniversalPrintingArchitectureTraditional:
				{
					Handle		printRecHandle = inArchitectureSpecificDataPtr->data.traditional.storageTHPrint;
					
					
					Printing_TakeTraditionalPrintData(inRef, printRecHandle);
				}
				break;
			
			case kUniversalPrintingArchitectureMacOSX:
				#if TARGET_API_MAC_OS8
				// a non-Carbon version of the library cant translate a Mac OS X printing record
				ptr->lastResult = cfragArchitectureErr;
				#else
				{
					Handle			printSettingsHandle =
										inArchitectureSpecificDataPtr->data.osX.storageFlattenedPrintSettings,
									pageFormatHandle =
										inArchitectureSpecificDataPtr->data.osX.storageFlattenedPageFormat;
					PMPrintSettings	unflattenedPrintSettings;
					PMPageFormat	unflattenedPageFormat;
					
					
					// unflatten the given handles, and then fire them off to the sister routines for installation
					ptr->lastResult = PMUnflattenPrintSettings(printSettingsHandle, &unflattenedPrintSettings);
					if (ptr->lastResult == noErr)
					{
						ptr->lastResult = PMUnflattenPageFormat(pageFormatHandle, &unflattenedPageFormat);
						if (ptr->lastResult == noErr)
						{
							Printing_TakeCarbonPrintData(inRef, unflattenedPrintSettings, unflattenedPageFormat);
						}
					}
				}
				#endif
				break;
			
			default:
				// ???
				ptr->lastResult = cfragArchitectureErr;
				break;
		}
		refReleaseLock(inRef);
	}
}// CopyFromSaved


/*!
Use this convenient routine for filling in the
standard architecture-specific data types using
information in an opaque printing record.  This
routine is great for saving information to disk.
You must specify a printing architecture ahead
of time (by setting the appropriate field in the
specified structure), and you must create valid
handles that this method will resize as needed.
If the structure contains an unknown
architecture, "cfragArchitectureErr" is returned
by Printing_GetLastResult().

IMPORTANT:	A Pre-Carbon version of this library
			will fail to store data in a Carbon
			Printing Manager format, whereas the
			reverse is not true.  If the data
			cant be written, the error
			"cfragArchitectureErr" is returned
			by Printing_GetLastResult().

(1.0)
*/
void
Printing_CopyToSaved		(UniversalPrintRecordRef		inRef,
							 SavedUniversalPrintRecordPtr	inoutArchitectureSpecificDataPtr)
{
	if ((inRef != null) && (inoutArchitectureSpecificDataPtr != null))
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
		switch (inoutArchitectureSpecificDataPtr->architecture)
		{
			case kUniversalPrintingArchitectureTraditional:
				{
					Handle			givenHandle = inoutArchitectureSpecificDataPtr->data.traditional.storageTHPrint,
									tempHandle = null;
					
					
					// COPY the data to the given handle, resizing it if necessary; this allows the use
					// of a resource handle, if desired, in the given structure
					Printing_MakeTraditionalPrintData(inRef, &tempHandle);
					if (Printing_GetLastResult(inRef) == noErr)
					{
						if (IsHandleValid(givenHandle))
						{
							SetHandleSize(givenHandle, GetHandleSize(tempHandle));
							ptr->lastResult = MemError();
							if (ptr->lastResult == noErr)
							{
								// note that a simple block-move doesnt require a locked handle
								BlockMove(*((THPrint)tempHandle), *((THPrint)givenHandle), sizeof(TPrint));
							}
						}
						else ptr->lastResult = nilHandleErr;
						DisposeHandle(tempHandle), tempHandle = null;
					}
				}
				break;
			
			case kUniversalPrintingArchitectureMacOSX:
				#if TARGET_API_MAC_OS8
				ptr->lastResult = cfragArchitectureErr;
				#else
				{
					Handle				printSettingsHandle =
											inoutArchitectureSpecificDataPtr->data.osX.storageFlattenedPrintSettings,
										pageFormatHandle =
											inoutArchitectureSpecificDataPtr->data.osX.storageFlattenedPageFormat;
					PMPrintSettings		unflattenedPrintSettings;
					PMPageFormat		unflattenedPageFormat;
					
					
					// grab the Carbon data types first
					Printing_MakeCarbonPrintData(inRef, &unflattenedPrintSettings, &unflattenedPageFormat);
					
					// flatten and COPY data from the opaque types to the handles; that way, the caller
					// can specify a resource handle that will remain a resource handle when this returns
					ptr->lastResult = flattenAndCopy(unflattenedPrintSettings, unflattenedPageFormat,
														printSettingsHandle, pageFormatHandle);
				}
				#endif
				break;
			
			default:
				// ???
				ptr->lastResult = cfragArchitectureErr;
				break;
		}
		refReleaseLock(inRef);
	}
}// CopyToSaved


/*!
To fill in a saved record with default values, as
if you called PrintDefault() or the combined
PMDefaultPrintSettings() and PMDefaultPageFormat()
routines on the internal data, use this method.

IMPORTANT:	This routine may fail if the installed
			version of this library is pre-Carbon
			and the requested data is in Carbon
			format, or vice-versa.

(1.0)
*/
void
Printing_DefaultSaved		(UniversalPrintRecordRef		inRef,
							 SavedUniversalPrintRecordPtr	inoutArchitectureSpecificDataPtr)
{
	if ((inRef != null) && (inoutArchitectureSpecificDataPtr != null))
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
		switch (inoutArchitectureSpecificDataPtr->architecture)
		{
			case kUniversalPrintingArchitectureTraditional:
				// (sigh) life was so much simpler with the traditional printing architecture...
				#if TARGET_API_MAC_OS8
				{
					Handle		printRecHandle = inoutArchitectureSpecificDataPtr->data.traditional.storageTHPrint;
					
					
					PrSetError(noErr);
					PrintDefault((THPrint)printRecHandle);
					ptr->lastResult = PrError();
				}
				#endif
				break;
			
			case kUniversalPrintingArchitectureMacOSX:
				// ...enter Carbon hell, where many steps are needed to get the job done
				#if TARGET_API_MAC_OS8
				ptr->lastResult = cfragArchitectureErr;
				#else
				{
					Handle			printSettingsHandle =
										inoutArchitectureSpecificDataPtr->data.osX.storageFlattenedPrintSettings,
									pageFormatHandle =
										inoutArchitectureSpecificDataPtr->data.osX.storageFlattenedPageFormat;
					PMPrintSettings	unflattenedPrintSettings;
					PMPageFormat	unflattenedPageFormat;
					
					
					// unflatten the given handles, default them, and then re-flatten them
					ptr->lastResult = PMUnflattenPrintSettings(printSettingsHandle, &unflattenedPrintSettings);
					if (ptr->lastResult == noErr)
					{
						ptr->lastResult = PMUnflattenPageFormat(pageFormatHandle, &unflattenedPageFormat);
						if (ptr->lastResult == noErr) ptr->lastResult = PMDefaultPrintSettings
																		(unflattenedPrintSettings);
						if (ptr->lastResult == noErr) ptr->lastResult = PMDefaultPageFormat
																		(unflattenedPageFormat);
						if (ptr->lastResult == noErr) ptr->lastResult = flattenAndCopy(unflattenedPrintSettings,
																						unflattenedPageFormat,
																						printSettingsHandle,
																						pageFormatHandle);
					}
				}
				#endif
				break;
			
			default:
				// ???
				ptr->lastResult = cfragArchitectureErr;
				break;
		}
		refReleaseLock(inRef);
	}
}// DefaultSaved


/*!
To close a printing document, use this method.
All context information, port information, etc. is
retained via the opaque reference, so if you called
Printing_BeginDocument(), Printing_EndDocument()
will automatically know what document you are
referring to.

(1.0)
*/
void
Printing_EndDocument	(UniversalPrintRecordRef	inRef)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		PrCloseDoc(ptr->printingPortPtr);
		ptr->lastResult = PrError();
#else
		ptr->lastResult = PMEndDocument(ptr->documentContext);
#endif
		refReleaseLock(inRef);
	}
}// EndDocument


/*!
To close a printing page, use this method.  All
context information, port information, etc. is
retained via the opaque reference, so if you called
Printing_BeginPage(), Printing_EndPage() will
automatically know what page you are referring to.

(1.0)
*/
void
Printing_EndPage	(UniversalPrintRecordRef	inRef)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		PrClosePage(ptr->printingPortPtr);
		ptr->lastResult = PrError();
#else
		ptr->lastResult = PMEndPage(ptr->documentContext);
#endif
		refReleaseLock(inRef);
	}
}// EndPage


/*!
To acquire the resolution of a printed page, use
this method.  Technically, Carbon allows greater
precision on this measurement, but for compatibility
with Mac OS 8, this routine returns integer numbers.

(1.0)
*/
void
Printing_GetDeviceResolution	(UniversalPrintRecordRef	inRef,
								 SInt16*					outVerticalResolution,
								 SInt16*					outHorizontalResolution)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		ptr->lastResult = noErr;
		*outVerticalResolution = (*(ptr->recordHandle))->prInfo.iVRes;
		*outHorizontalResolution = (*(ptr->recordHandle))->prInfo.iHRes;
#else
		{
			PMResolution		resolution;
			
			
			ptr->lastResult = PMGetResolution(ptr->pageFormat, &resolution);
			*outVerticalResolution = resolution.vRes;
			*outHorizontalResolution = resolution.hRes;
		}
#endif
		refReleaseLock(inRef);
	}
}// GetDeviceResolution


/*!
Returns the error code generated from the last
printing function invoked.  The meaning and
range of values will depend on the architecture
(Mac OS 8 or Mac OS X) of this librarys
implementation.  In the rare event that this
library needs to make multiple printing-related
calls to perform a function, the last result
will actually refer to the error returned from
the first call that fails.

(1.0)
*/
OSStatus
Printing_GetLastResult	(UniversalPrintRecordRef	inRef)
{
	OSErr		result = noErr;
	
	
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
		result = ptr->lastResult;
		refReleaseLock(inRef);
	}
	else result = memPCErr;
	
	return result;
}// GetLastResult


/*!
Returns a constant describing the current
printing mode.  You must explicitly set the
printing mode (using Printing_SetMode()) to
be sure the printing functions will behave
as expected.

(1.0)
*/
UniversalPrintingMode
Printing_GetMode	(void)
{
	return gPrintingMode;
}// GetMode


/*!
To acquire the boundaries of a printed page, use
this method.  Technically, Carbon allows greater
precision on this measurement, but for compatibility
with Mac OS 8, this routine fills in a basic
QuickDraw rectangle.

(1.0)
*/
void
Printing_GetPageBounds		(UniversalPrintRecordRef	inRef,
							 Rect*						outBounds)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		ptr->lastResult = noErr;
		*outBounds = (*(ptr->recordHandle))->prInfo.rPage;
#else
		{
			PMRect		preciseBounds;
			
			
			ptr->lastResult = PMGetPhysicalPageSize(ptr->pageFormat, &preciseBounds);
			SetRect(outBounds, preciseBounds.left, preciseBounds.top, preciseBounds.right, preciseBounds.bottom);
		}
#endif
		refReleaseLock(inRef);
	}
}// GetPageBounds


/*!
To acquire the boundaries of the paper being used for
printing, use this method.  Technically, Carbon allows
greater precision on this measurement, but for
compatibility with Mac OS 8, this routine fills in a
basic QuickDraw rectangle.

(1.0)
*/
void
Printing_GetPaperBounds		(UniversalPrintRecordRef	inRef,
							 Rect*						outBounds)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		ptr->lastResult = noErr;
		*outBounds = (*(ptr->recordHandle))->rPaper;
#else
		{
			PMRect		preciseBounds;
			
			
			ptr->lastResult = PMGetPhysicalPaperSize(ptr->pageFormat, &preciseBounds);
			SetRect(outBounds, preciseBounds.left, preciseBounds.top, preciseBounds.right, preciseBounds.bottom);
		}
#endif
		refReleaseLock(inRef);
	}
}// GetPaperBounds


/*!
To generate Carbon printing data types reflecting the
settings of an opaque print record, use this method.
If this version of the library is pre-Carbon, this
routine will have no effect.

This routine will most likely be widely used in the
future, when Mac OS 8 compatibility is not an issue
and you will be interested in using Carbon data types
exclusively.

(1.0)
*/
#if !TARGET_API_MAC_OS8
void
Printing_MakeCarbonPrintData	(UniversalPrintRecordRef		inRef,
								 PMPrintSettings*				outPrintSettingsPtr,
								 PMPageFormat*					outPageFormatPtr)
{
	if ((inRef != null) && (outPrintSettingsPtr != null) && (outPageFormatPtr != null))
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
		ptr->lastResult = PMCopyPrintSettings(ptr->settings, *outPrintSettingsPtr);
		if (ptr->lastResult == noErr) ptr->lastResult = PMCopyPageFormat(ptr->pageFormat, *outPageFormatPtr);
		refReleaseLock(inRef);
	}
}// MakeCarbonPrintData
#endif


/*!
To create a new handle that is really a traditional
"THPrint" containing settings from an opaque print
record, use this method.

IMPORTANT:	You should generally avoid creating a
			data structure from an obsolete version
			of the Printing Manager; this method is
			for convenience only.

WARNING:	This method allocates memory, and may
			fail.  Check Printing_GetLastResult()
			to be sure it succeeds.

(1.0)
*/
void
Printing_MakeTraditionalPrintData	(UniversalPrintRecordRef		inRef,
									 Handle*						outTHPrintPtr)
{
	if ((inRef != null) && (outTHPrintPtr != null))
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		*outTHPrintPtr = NewHandle(sizeof(TPrint));
		ptr->lastResult = MemError();
		if (*outTHPrintPtr == null) ptr->lastResult = memFullErr;
		if (ptr->lastResult == noErr)
		{
			// note that a simple block-move doesnt require a locked handle
			BlockMove(*(ptr->recordHandle), *((THPrint)*outTHPrintPtr), sizeof(TPrint));
		}
#else
		ptr->lastResult = PMMakeOldPrintRecord(ptr->settings, ptr->pageFormat, outTHPrintPtr);
#endif
		refReleaseLock(inRef);
	}
}// MakeTraditionalPrintData


/*!
To display the standard Page Setup dialog box,
use this method.  Most applications provide a
Page Setup command, and should therefore use
this routine to implement that command.

The dialog box may modify print settings; use
accessor routines in this module to read the
information you need.

If the user saves changes, "true" is returned;
otherwise, "false" is returned.

(1.0)
*/
Boolean
Printing_PageSetupDialogDisplay		(UniversalPrintRecordRef		inRef)
{
	Boolean		result = false;
	
	
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		result = PrStlDialog(ptr->recordHandle);
		ptr->lastResult = PrError();
#else
		ptr->lastResult = PMPageSetupDialog(ptr->pageFormat, &result);
#endif
		refReleaseLock(inRef);
	}
	
	return result;
}// PageSetupDialogDisplay


/*!
To display the standard Print dialog box, use
this method.  Most applications require this
routine for their primary Print command, but
special print commands (such as Print One Copy,
etc.) may not use a print dialog box.

The dialog box may modify print settings; use
accessor routines in this module to read the
information you need.

If the user chooses to go ahead with printing,
"true" is returned; otherwise, the printing job
was cancelled, so "false" is returned.

(1.0)
*/
Boolean
Printing_PrintJobDialogDisplay		(UniversalPrintRecordRef		inRef)
{
	Boolean		result = false;
	
	
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		result = PrJobDialog(ptr->recordHandle);
		ptr->lastResult = PrError();
#else
		ptr->lastResult = PMPrintDialog(ptr->settings, ptr->pageFormat, &result);
#endif
		refReleaseLock(inRef);
	}
	
	return result;
}// PrintJobDialogDisplay


/*!
To install a print idle procedure, use this method.
Under Mac OS 8, the most common reason for doing
this is to poll for command-period (cancellation)
events.

(1.0)
*/
void
Printing_SetIdleProc	(UniversalPrintRecordRef		inRef,
						 UniversalPrintingIdleProcPtr	inProcPtr)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		ptr->lastResult = noErr;
		(*(ptr->recordHandle))->prJob.pIdleProc = NewPrIdleProc((PrIdleProcPtr)inProcPtr);
#else
		ptr->lastResult = PMSetIdleProc((PMIdleUPP)inProcPtr);
#endif
		refReleaseLock(inRef);
	}
}// SetIdleProc


/*!
To set the number of copies to print, use this
method.  Under Mac OS 8, this is usually done
to implement a print now command, bypassing
the print job dialog.  The "inLock" setting is
ignored under Mac OS 8.

(1.0)
*/
void
Printing_SetNumberOfCopies	(UniversalPrintRecordRef	inRef,
							 const UInt16				inNumberOfCopies,
							 const Boolean				inLock)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		ptr->lastResult = noErr;
		(*(ptr->recordHandle))->prJob.iCopies = inNumberOfCopies;
#else
		ptr->lastResult = PMSetCopies(ptr->settings, inNumberOfCopies, inLock);
#endif
		refReleaseLock(inRef);
	}
}// SetNumberOfCopies


/*!
Specifies the printing mode.  The printing
mode affects how the printing functions
behave.  Each printing mode generally
corresponds to a printing option in the File
menu (so there are very few of them).  You
must explicitly set the printing mode to be
sure the printing functions will behave as
expected.

(1.0)
*/
void
Printing_SetMode	(const UniversalPrintingMode	inNewMode)
{
	gPrintingMode = inNewMode;
}// SetMode


/*!
To use data from Carbon printing data types to modify
the settings of an opaque print record, use this
method.  If this version of the library is pre-Carbon,
this routine will have no effect.

(1.0)
*/
#if !TARGET_API_MAC_OS8
void
Printing_TakeCarbonPrintData	(UniversalPrintRecordRef		inRef,
								 PMPrintSettings				inPrintSettingsToUse,
								 PMPageFormat					inPageFormatToUse)
{
	if (inRef != null)
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
		ptr->lastResult = PMCopyPrintSettings(inPrintSettingsToUse, ptr->settings);
		if (ptr->lastResult == noErr) ptr->lastResult = PMCopyPageFormat(inPageFormatToUse, ptr->pageFormat);
		refReleaseLock(inRef);
	}
}// TakeCarbonPrintData
#endif


/*!
To use data from a traditional "THPrint" to modify
the settings of an opaque print record, use this
method.  If this version of the library is pre-Carbon,
this routine will simply copy the handle; otherwise,
it will translate the handle to Carbon data types.
In any case, subsequent calls to accessor functions
using the given opaque printing record reference will
return data that takes into account modifications
from the traditional printing record you provide to
this function.

IMPORTANT:	You should generally avoid using a data
			structure from an obsolete version of the
			Printing Manager; this method is for
			convenience only.  It is generally used
			for updating an opaque record to use data
			stored in a disk file, which might have
			been created by an older version of the
			Mac OS Printing Manager.

(1.0)
*/
void
Printing_TakeTraditionalPrintData	(UniversalPrintRecordRef		inRef,
									 Handle							inTHPrintToUse)
{
	if ((inRef != null) && (IsHandleValid(inTHPrintToUse)))
	{
		UniversalPrintRecordPtr		ptr = refAcquireLock(inRef);
		
		
#if TARGET_API_MAC_OS8
		BlockMove(*((THPrint)inTHPrintToUse), *(ptr->recordHandle), sizeof(TPrint));
		
		// lock the handle in case PrValidate() does something undocumented and stupid with it
		HLock((Handle)ptr->recordHandle);
		(Boolean)PrValidate(ptr->recordHandle);
		ptr->lastResult = PrError();
		HUnlock((Handle)ptr->recordHandle);
#else
		ptr->lastResult = PMConvertOldPrintRecord(inTHPrintToUse, &ptr->settings, &ptr->pageFormat);
#endif
		refReleaseLock(inRef);
	}
}// TakeTraditionalPrintData


//
// internal methods
//

#if !TARGET_API_MAC_OS8
/*!
To flatten opaque Carbon printing types and
COPY the flattened versions to existing
handles, use this method.

(1.0)
*/
static OSStatus
flattenAndCopy		(PMPrintSettings			inPrintSettingsToFlatten,
					 PMPageFormat				inPageFormatToFlatten,
					 Handle						inoutPrintSettingsFlattenedCopy,
					 Handle						inoutPageFormatFlattenedCopy)
{
	OSStatus	result = noErr;
	Handle		tempHandle = null;
	
	
	// flatten the print settings information and COPY it to the appropriate handle;
	// this allows the use of things like resource handles, etc.
	if (result == noErr)
	{
		result = PMFlattenPrintSettings(inPrintSettingsToFlatten, &tempHandle);
		if (result == noErr)
		{
			if (IsHandleValid(inoutPrintSettingsFlattenedCopy))
			{
				SetHandleSize(inoutPrintSettingsFlattenedCopy, GetHandleSize(tempHandle));
				result = MemError();
				if (result == noErr)
				{
					// note that a simple block-move doesnt require a locked handle
					BlockMove(*tempHandle, *inoutPrintSettingsFlattenedCopy, GetHandleSize(tempHandle));
				}
			}
			else result = nilHandleErr;
			DisposeHandle(tempHandle), tempHandle = null;
		}
	}
	
	// flatten the page format information and COPY it to the appropriate handle; this
	// allows the use of things like resource handles, etc.
	if (result == noErr)
	{
		result = PMFlattenPageFormat(inPageFormatToFlatten, &tempHandle);
		if (result == noErr)
		{
			if (IsHandleValid(inoutPageFormatFlattenedCopy))
			{
				SetHandleSize(inoutPageFormatFlattenedCopy, GetHandleSize(tempHandle));
				result = MemError();
				if (result == noErr)
				{
					// note that a simple block-move doesnt require a locked handle
					BlockMove(*tempHandle, *inoutPageFormatFlattenedCopy, GetHandleSize(tempHandle));
				}
			}
			else result = nilHandleErr;
			DisposeHandle(tempHandle), tempHandle = null;
		}
	}
	
	return result;
}// flattenAndCopy
#endif


/*!
Use this method to acquire a pointer to the internal
structure, given a reference to it.  Get in the habit
of calling refReleaseLock() when finished with the
pointer you receive; this allows flexibility in the
future to make dynamically-allocated memory relocatable.

(1.0)
*/
static UniversalPrintRecordPtr
refAcquireLock	(UniversalPrintRecordRef		inRef)
{
	HLock((Handle)inRef);
	return (*((UniversalPrintRecordHandle)inRef));
}// refAcquireLock


/*!
To release the lock on a reference that you previously
locked with refAcquireLock(), use this method.

(1.0)
*/
static void
refReleaseLock	(UniversalPrintRecordRef		inRef)
{
	HUnlock((Handle)inRef);
}// refReleaseLock