/*###############################################################

	PrintingLib.h
	
	Use this library to allow your application to print using
	either the Mac OS 8 or Mac OS X printing managers, depending
	upon which version of this library is installed.  By using
	this librarys interface for your printing needs, you only
	have to write one set of source code (one that interfaces
	with this library), and yet your application can use either
	of the two printing architectures transparently!
	
	This library can be built targeted at Carbon, or not.  Thus,
	two different versions of this library exist, each having
	identical application programming interfaces.  Identical
	interfaces in both means that your application generally
	never needs to worry about which of the two editions is
	actually in place at run time.  Nevertheless, the method
	Printing_GetArchitecture() is provided so you can find out
	which runtime version is being used, when necessary.
	
	Universal Printing Library 1.0
	 1998-2000 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

/*!
	@header Universal Printing Library
*/

#ifndef __PRINTINGLIB__
#define __PRINTINGLIB__

// Mac includes
#ifndef __PMAPPLICATION__
#	include <PMApplication.h>
#endif
#ifndef __PRINTING__
#	include <Printing.h>
#endif

// library includes
#ifndef __RELEASES__
#	include "Releases.h"
#endif

#ifdef powerc



/**************************************************************************************************************************/
// Printing.c
//
// This module lets you implement printing in your application transparently, between the Mac OS 8 and Mac OS X printing
// architectures!  This universal set of APIs is implemented in two different libraries, one that natively calls Mac OS 8
// (traditional) to perform printing related tasks, and one that uses the new Carbon Printing Manager.  If your application
// uses these APIs, it can use identical code to print natively in either environment, provided the native runtime library
// is installed.  For example, end users that have Mac OS X should be allowed to get the version of this library that is
// compiled under the TARGET_API_MAC_CARBON=1 directive, whereas your pre-Mac OS X users should instead be given the one
// compiled under TARGET_API_MAC_OS8=1.  The API for both code sets is similar, and the symbols are identical, so your
// code can link against either library and work with both.

typedef struct { unsigned garbage; }*	UniversalPrintRecordRef;

typedef SInt16 UniversalPrintingArchitecture;
enum
{
	// one of these constants is returned from Printing_GetArchitecture()
	kUniversalPrintingArchitectureTraditional = 0,
	kUniversalPrintingArchitectureMacOSX = 1	// Carbon printing
};

typedef UInt16 UniversalPrintingMode;
enum
{
	kUniversalPrintingModeNormal = 0,		// show job dialog
	kUniversalPrintingModeOneCopy = 1		// dont show job dialog
};

struct SavedUniversalPrintRecord
{
	UniversalPrintingArchitecture		architecture;
	union
	{
		// if the value of field "architecture" is kUniversalPrintingArchitectureTraditional...
		struct
		{
			Handle		storageTHPrint;
		} traditional;
		
		// if the value of field "architecture" is kUniversalPrintingArchitectureMacOSX...
		struct
		{
			Handle		storageFlattenedPrintSettings;
			Handle		storageFlattenedPageFormat;
		} osX;
	} data;
};
#ifndef __cplusplus
typedef struct SavedUniversalPrintRecord	SavedUniversalPrintRecord;
#endif
typedef SavedUniversalPrintRecord*			SavedUniversalPrintRecordPtr;



/**
Printing Idle Method

This routine should, at a minimum, check for -.
keypresses and cancel printing.
*/
typedef void (*UniversalPrintingIdleProcPtr)(void);
#define InvokeUniversalPrintingIdleProc(userRoutine) \
			(*userRoutine)()



/*###############################################################
	INITIALIZING AND FINISHING WITH PRINTING
###############################################################*/

// CALL THIS ROUTINE ONCE, BEFORE ANY OTHER PRINTING ROUTINE; TO GET THE PRINTING DRIVERS
// RESOURCE FILE (MAC OS 8.x) INVOKE CurResFile() IMMEDIATELY AFTER A CALL TO THIS ROUTINE
void
	Printing_Init						(void);

// CALL THIS ROUTINE AFTER YOU ARE DONE WITH PRINTING
void
	Printing_Done						(void);

/*###############################################################
	DETERMINING SPECIFIC PRINTING LIBRARY INFORMATION
###############################################################*/

ApplicationSharedLibraryVersion
	Printing_Version					(void);

#ifndef __cplusplus
#define UniversalPrintingAvailable() 	(Printing_Version != (void*)kUnresolvedCFragSymbolAddress)
#else
inline Boolean UniversalPrintingAvailable()	{	return (Printing_Version != (void*)kUnresolvedCFragSymbolAddress);	}
#endif

UniversalPrintingArchitecture
	Printing_GetArchitecture			(void);

/*###############################################################
	CREATING AND DESTROYING OPAQUE PRINT RECORDS
###############################################################*/

UniversalPrintRecordRef
	Printing_NewRecord					(void);

void
	Printing_DisposeRecord				(UniversalPrintRecordRef*	inoutRefPtr);

/*###############################################################
	STATUS INFORMATION
###############################################################*/

OSStatus
	Printing_GetLastResult				(UniversalPrintRecordRef	inRef);

/*###############################################################
	ACCESSING DATA IN OPAQUE PRINT RECORDS
###############################################################*/

void
	Printing_GetDeviceResolution		(UniversalPrintRecordRef	inRef,
										 SInt16*					outVerticalResolution,
										 SInt16*					outHorizontalResolution);

void
	Printing_GetPageBounds				(UniversalPrintRecordRef	inRef,
										 Rect*						outBounds);

void
	Printing_GetPaperBounds				(UniversalPrintRecordRef	inRef,
										 Rect*						outBounds);

void
	Printing_SetIdleProc				(UniversalPrintRecordRef		inRef,
										 UniversalPrintingIdleProcPtr	inProcPtr);

void
	Printing_SetNumberOfCopies			(UniversalPrintRecordRef	inRef,
										 const UInt16				inNumberOfCopies,
										 const Boolean				inLock);

/*###############################################################
	PRINTING MODE (DIALOG OR NO DIALOG) - FOR CONVENIENCE ONLY
###############################################################*/

UniversalPrintingMode
	Printing_GetMode					(void);

void
	Printing_SetMode					(const UniversalPrintingMode	inNewMode);

/*###############################################################
	PRINTING DIALOGS
###############################################################*/

// CALL THIS ROUTINE INSTEAD OF PrStlDialog()
Boolean
	Printing_PageSetupDialogDisplay		(UniversalPrintRecordRef		inRef);

// CALL THIS ROUTINE INSTEAD OF PrJobDialog()
Boolean
	Printing_PrintJobDialogDisplay		(UniversalPrintRecordRef		inRef);

/*###############################################################
	SAVING AND RESTORING OPAQUE RECORDS
###############################################################*/

// INITIALIZE THE DATA STRUCTURE (SET THE ARCHITECTURE, FILL IN YOUR DATA FROM RESOURCES, ETC.) BEFORE USE
void
	Printing_CopyFromSaved				(UniversalPrintRecordRef		inRef,
										 SavedUniversalPrintRecordPtr	inArchitectureSpecificDataPtr);

// INITIALIZE THE DATA STRUCTURE (SET THE ARCHITECTURE, CREATE ALL HANDLES) BEFORE CALLING THIS ROUTINE
void
	Printing_CopyToSaved				(UniversalPrintRecordRef		inRef,
										 SavedUniversalPrintRecordPtr	inoutArchitectureSpecificDataPtr);

// CALL THIS ROUTINE INSTEAD OF PrintDefault()
void
	Printing_DefaultSaved				(UniversalPrintRecordRef		inRef,
										 SavedUniversalPrintRecordPtr	inoutArchitectureSpecificDataPtr);

/*###############################################################
	SEAMLESS UTILIZATION OF ARCHITECTURE-SPECIFIC DATA
###############################################################*/

// THIS HAS NO EFFECT WITHOUT A CARBON VERSION OF THE LIBRARY; IT IS MOST BENEFICIAL FOR FUTURE USE
#if !TARGET_API_MAC_OS8
void
	Printing_MakeCarbonPrintData		(UniversalPrintRecordRef		inRef,
										 PMPrintSettings*				outPrintSettingsPtr,
										 PMPageFormat*					outPageFormatPtr);
#endif

// IF A MAC OS 8 VERSION OF THIS LIBRARY IS BEING USED, THIS METHOD WORKS MUCH FASTER
void
	Printing_MakeTraditionalPrintData	(UniversalPrintRecordRef		inRef,
										 Handle*						outTHPrintPtr);

// THIS HAS NO EFFECT WITHOUT A CARBON VERSION OF THE LIBRARY
#if !TARGET_API_MAC_OS8
void
	Printing_TakeCarbonPrintData		(UniversalPrintRecordRef		inRef,
										 PMPrintSettings				inPrintSettingsToUse,
										 PMPageFormat					inPageFormatToUse);
#endif

// THIS ROUTINE WORKS UNDER ANY PRINTING ARCHITECTURE (USUALLY USED FOR LEGACY PRINT SETTINGS FROM FILES)
void
	Printing_TakeTraditionalPrintData	(UniversalPrintRecordRef		inRef,
										 Handle							inTHPrintToUse);

/*###############################################################
	PRINTING DOCUMENTS
###############################################################*/

// CALL THIS ROUTINE INSTEAD OF PrOpenDoc()
void
	Printing_BeginDocument				(UniversalPrintRecordRef		inRef);

// CALL THIS ROUTINE INSTEAD OF PrOpenPage()
void
	Printing_BeginPage					(UniversalPrintRecordRef		inRef,
										 const Rect*					inPageBounds);	// can be null

void
	Printing_Cancel						(UniversalPrintRecordRef		inRef);

// CALL THIS ROUTINE INSTEAD OF PrCloseDoc()
void
	Printing_EndDocument				(UniversalPrintRecordRef		inRef);

// CALL THIS ROUTINE INSTEAD OF PrClosePage()
void
	Printing_EndPage					(UniversalPrintRecordRef		inRef);

#endif /* ifdef powerc */

#endif