/*###############################################################

	ProgressDialog.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Dialogs.h>
#include <Resources.h>

// library includes
#include "ContextsLib.h"
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "MacLook.h"
#include "ProgressDialog.h"



// types
struct ProgressDialog
{
	DialogRef		dialog;		// the dialog box representing the progress window
};
typedef struct ProgressDialog	ProgressDialog;
typedef ProgressDialog*			ProgressDialogPtr;

// internal methods
static void						arrangeWindow		(ProgressDialogPtr		inPtr);
static ProgressDialogPtr		refAcquireLock		(ProgressDialogRef		inRef);
static void						refReleaseLock		(ProgressDialogRef		inRef,
													 ProgressDialogPtr*		inoutPtr);

// variables
static Point		gNewDialogOrigin = { 0, 0 };



//
// public methods
//

/*!
To create a new Dialog Panel reference,
use this method.  If any problems occur,
null is returned.

Your dialog panel cannot function without
a change notification procedure.  This
routine is the entry point that dialog
boxes use (indirectly, by invoking one or
more methods from the Dialog Panel module)
to communicate with panels in an abstract
way.

(3.0)
*/
ProgressDialogRef
ProgressDialog_New		(StringPtr			inStatusText,
						 const Boolean		inIsModal)
{
	ProgressDialogRef		result = (ProgressDialogRef)Memory_NewPtr(sizeof(ProgressDialog));
	
	
	if (result != null)
	{
		ProgressDialogPtr		ptr = refAcquireLock(result);
		SInt32					procID = kDialogIDGenericProgress;
		ControlRef				progressBar = null;
		
		
		// modeless windows use a different frame style
		unless (inIsModal) procID = kDialogIDGenericProgressModeless;
		
		// create the dialog and set up its progress bar
		ptr->dialog = GetNewDialog(procID, null, (inIsModal) ? (WindowRef)-1/* in front */ : 0L/* in behind */);
		(OSErr)GetDialogItemAsControl(ptr->dialog, iGenericProgressBar, &progressBar);
		SetControlMaximum(progressBar, 100);
		SetControlMinimum(progressBar, 0);
		SetControlValue(progressBar, 0);
		
		// modeless windows should be offset out of the way (top-right corner),
		// whereas modal windows are best left in the center of the screen
		unless (inIsModal) arrangeWindow(ptr);
		
		ProgressDialog_SetStatus(result, inStatusText);
		ProgressDialog_SetTitle(result, EMPTY_PSTRING);
	}
	return result;
}// New


/*!
To destroy a Panel previously created
with Panel_New(), use this method.
Your copy of the reference is
automatically set to null.

(3.0)
*/
void
ProgressDialog_Dispose		(ProgressDialogRef*		inoutRefPtr)
{
	if (inoutRefPtr != null)
	{
		ProgressDialogPtr		ptr = refAcquireLock(*inoutRefPtr);
		
		
		DisposeDialog(ptr->dialog);
		Memory_DisposePtr((Ptr*)inoutRefPtr);
	}
}// Dispose


/*!
To show a progress window, use this method.
Do not use ProgressDialog_GetWindow() to
help you display a particular progress
dialog, because future implementations may
not place each progress dialog in its own
window.

(3.0)
*/
void
ProgressDialog_Display		(ProgressDialogRef		inRef)
{
	ProgressDialogPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ShowWindow(GetDialogWindow(ptr->dialog));
		DrawDialog(ptr->dialog);
		refReleaseLock(inRef, &ptr);
	}
}// SetProgressPercentFull


/*!
To determine which window contains a progress
dialog, use this method.  If any problems
occur, "null" is returned.

(3.0)
*/
WindowRef
ProgressDialog_GetWindow	(ProgressDialogRef		inRef)
{
	ProgressDialogPtr		ptr = refAcquireLock(inRef);
	WindowRef				result = null;
	
	
	if (ptr != null)
	{
		result = GetDialogWindow(ptr->dialog);
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetWindow


/*!
To specify the title of a progress dialogs
window, use this method.

(3.0)
*/
void
ProgressDialog_SetProgressPercentFull	(ProgressDialogRef		inRef,
										 const SInt8			inPercentage)
{
	ProgressDialogPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ControlRef		progressBar = null;
		
		
		(OSErr)GetDialogItemAsControl(ptr->dialog, iGenericProgressBar, &progressBar);
		SetControlValue(progressBar, inPercentage);
		refReleaseLock(inRef, &ptr);
	}
}// SetProgressPercentFull


/*!
To specify the title of a progress dialogs
window, use this method.

(3.0)
*/
void
ProgressDialog_SetProgressIndicator		(ProgressDialogRef			inRef,
										 TelnetProgressIndicator	inIndicatorType)
{
	ProgressDialogPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		Boolean			isIndeterminate = false;
		ControlRef		progressBar = null;
		
		
		(OSErr)GetDialogItemAsControl(ptr->dialog, iGenericProgressBar, &progressBar);
		switch (inIndicatorType)
		{
			case kTelnetProgressIndicatorIndeterminate:
				isIndeterminate = true;
				break;
			
			default:
				isIndeterminate = false;
				break;
		}
		(OSErr)SetControlData(progressBar, kControlIndicatorPart, kControlProgressBarIndeterminateTag,
								sizeof(Boolean), (Ptr)&isIndeterminate);
		refReleaseLock(inRef, &ptr);
	}
}// SetProgressIndicator


/*!
To specify the status text (prompt) of a
progress dialogs window, use this method.

(3.0)
*/
void
ProgressDialog_SetStatus		(ProgressDialogRef		inRef,
								 StringPtr				inStatusText)
{
	ProgressDialogPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		SetDialogItemControlText(ptr->dialog, iGenericProgressText, inStatusText);
		refReleaseLock(inRef, &ptr);
	}
}// SetStatus


/*!
To specify the title of a progress dialogs
window, use this method.

(3.0)
*/
void
ProgressDialog_SetTitle		(ProgressDialogRef		inRef,
							 StringPtr				inTitleText)
{
	ProgressDialogPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		SetWTitle(GetDialogWindow(ptr->dialog), inTitleText);
		refReleaseLock(inRef, &ptr);
	}
}// SetTitle


//
// internal methods
//

/*!
To arrange the specified progress dialog in an
appropriate stagger position, use this method.

(3.0)
*/
static void
arrangeWindow	(ProgressDialogPtr		inPtr)
{
	if (inPtr != null)
	{
		WindowRef	window = GetDialogWindow(inPtr->dialog);
		
		
		// is the origin already saved?
		if (gNewDialogOrigin.h == 0)
		{
			enum
			{
				kPaddingH = 10,	// pixels of space between window structure and screen edge
				kPaddingV = 10	// pixels of space between window structure and screen edge
			};
			Rect		screenBounds,
						structureRect,
						contentRect;
			
			
			MacLook_GetWindowDeviceGrayRect(window, &screenBounds);
			MacLook_GetWindowRegionBounds(window, kWindowStructureRgn, &structureRect);
			MacLook_GetWindowRegionBounds(window, kWindowContentRgn, &contentRect);
			
			// position the progress window in the top-right corner of its screen (take into
			// account the structure region, but the location is that of the content region)
			SetRect(&contentRect,
					screenBounds.right - kPaddingH - (structureRect.right - structureRect.left) +
						(contentRect.left - structureRect.left),
					screenBounds.top + kPaddingV + (contentRect.top - structureRect.top),
					0, 0);
			SetPt(&gNewDialogOrigin, contentRect.left, contentRect.top);
		}
		else
		{
			// stagger this progress dialog
			gNewDialogOrigin.h += 10;
			gNewDialogOrigin.v += 10;
		}
		
		MoveWindow(window, gNewDialogOrigin.h, gNewDialogOrigin.v, false/* activate */);
	}
}// arrangeWindow


/*!
To acquire a pointer to the internal structure,
given a reference to it, use this method.  Be
sure to release the lock when you are finished,
using refReleaseLock().

(3.0)
*/
static ProgressDialogPtr
refAcquireLock	(ProgressDialogRef		inRef)
{
	return ((ProgressDialogPtr)inRef);
}// refAcquireLock


/*!
To release a pointer to the internal structure,
setting it to "null" and allowing it to be
locked somewhere else, use this method.

(3.0)
*/
static void
refReleaseLock	(ProgressDialogRef		UNUSED_ARGUMENT(inRef),
				 ProgressDialogPtr*		inoutPtr)
{
	// the memory is not relocatable, so this routine doesnt do much
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock