/*###############################################################

	ProxyServer.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstring>
#else
#	include <string.h>
#endif

// Mac includes
#include <OpenTransport.h>
#include <OpenTransportProviders.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// MacTelnet includes
#include "Console.h"
#include "ProxyServer.h"
#include "SOCKS.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



// constants
enum
{
	kProxyServerVersionZero = 0		// represents structure layout in MacTelnet 3.0
};

// types
struct ProxyServer
{
	SInt16					lockCount;				// the number of locks on the handle
	SInt16					version;				// a version number, as a "kProxyServerVersion..." constant
	SInt16					type;					// a TelnetProxyServerType; the protocol identifier for this server
	char					domainName[256];		// the proxy server host name
	UInt16					portNumber;				// the port of the server
	char					userName[256];			// authenticating user name
	Boolean					isConnected,			// is there an open connection to this proxy server?
							unusedFlag1,
							unusedFlag2,
							unusedFlag3;
	SInt16					padding[98];
};
#ifndef __cplusplus
typedef struct ProxyServer	ProxyServer;
#endif
typedef ProxyServer*		ProxyServerPtr;
typedef ProxyServerPtr*		ProxyServerHandle;

// internal methods
static ProxyServerPtr		refAcquireLock	(ProxyServerRef			inRef);
static void					refReleaseLock	(ProxyServerRef			inRef,
											 ProxyServerPtr*		inoutPtr);



//
// public methods
//

/*!
To create a new proxy server, use this method.
You need to initialize it using methods from
the Proxy Server module.

If any problems occur, null is returned;
otherwise, a reference to the new proxy server
is returned.

To destroy the proxy server, use the method
ProxyServer_Dispose().

(3.0)
*/
ProxyServerRef
ProxyServer_New		(void)
{
	ProxyServerRef		result = (ProxyServerRef)Memory_NewHandle(sizeof(ProxyServer));
	
	
	if (result != null)
	{
		ProxyServerPtr		ptr = refAcquireLock(result);
		
		
		// fill in the new data structure
		ptr->version = kProxyServerVersionZero;
		ptr->type = kTelnetProxyServerTypeSOCKS4;
		strcpy(ptr->domainName, "");
		ptr->portNumber = kSOCKSPortNumberDefault;
		strcpy(ptr->userName, "");
		
		refReleaseLock(result, &ptr);
	}
	else Console_WriteLine("ERROR: Proxy server data could NOT be constructed.");
	
	return result;
}// New


/*!
To create a new proxy server from a handle that
you previously received from a call to the
ProxyServer_Flatten() method, use this method.

If any problems occur, null is returned;
otherwise, a reference to the new proxy server
is returned.

To destroy the proxy server, use the method
ProxyServer_Dispose().

(3.0)
*/
ProxyServerRef
ProxyServer_NewFlattened	(Handle			inDataHandle)
{
	ProxyServerRef		result = null;
	
	
	if (inDataHandle != null)
	{
		result = (ProxyServerRef)Memory_NewHandle(GetHandleSize(inDataHandle));
		if (result != null)
		{
			ProxyServerPtr		ptr = refAcquireLock(result);
			
			
			// copy data
			BlockMove(*((ProxyServerHandle)inDataHandle), ptr, GetHandleSize(inDataHandle));
			refReleaseLock(result, &ptr);
		}
	}
	return result;
}// NewFlattened


/*!
This method cleans up a proxy server configuration
by destroying all of the data associated with it.
On output, your copy of the given reference will be
set to null.

(3.0)
*/
void
ProxyServer_Dispose		(ProxyServerRef*		inoutRefPtr)
{
	if (inoutRefPtr != null)
	{
		ProxyServerPtr		ptr = refAcquireLock(*inoutRefPtr);
		
		
		if (ptr != null)
		{
			Memory_DisposeHandle((Handle*)inoutRefPtr), ptr = null;
		}
	}
}// Dispose


/*!
Initiates a connection for the specified proxy
server using the given host information, and
returns an error code ("noErr" if successful).
You typically invoke ProxyServer_GetConnectionInfo()
to determine the host name of a proxy server, and
perform the DNR lookup asynchronously yourself;
once you have the Internet address, pass it to
this routine to initiate a connection.

Once connected, you can issue various requests to
the proxy server, such as looking up a domain name
that is outside the firewall.

To terminate the new connection, invoke the
ProxyServer_Disconnect() method.

Unimplemented.

(3.0)
*/
OSStatus
ProxyServer_Connect		(ProxyServerRef			inRef,
						 InetAddress const*		inAddressPtr)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	OSStatus			result = noErr;
	
	
	if (ptr != null)
	{
		// initiate connection - unimplemented
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// Connect


/*!
Terminates a connection to the specified proxy
server, if any connection is open.

Unimplemented.

(3.0)
*/
void
ProxyServer_Disconnect	(ProxyServerRef			inRef)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		// terminate connection - unimplemented
		refReleaseLock(inRef, &ptr);
	}
}// Connect


/*!
To flatten an opaque proxy server reference into
a handle (in order to write it to a resource on
disk, for example), use this method.  Using the
data in this handle, you can later construct a
new Proxy Server object by using the method
ProxyServer_NewFlattened().

A new locked handle is allocated in reserved
memory space - you must unlock and dispose of
the handle when you are finished with it.

(3.0)
*/
void
ProxyServer_Flatten		(ProxyServerRef		inRef,
						 Handle*			outNewDataHandlePtr)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	
	
	if ((ptr != null) && (outNewDataHandlePtr != null))
	{
		// allocate handle
		*outNewDataHandlePtr = Memory_NewHandle(GetHandleSize((Handle)inRef));
		
		// copy data
		if (*outNewDataHandlePtr != null)
		{
			Console_WriteLine("Preparing to flatten the basic proxy server data.");
			BlockMove(ptr, *(*outNewDataHandlePtr), GetHandleSize((Handle)inRef));
		}
		else Console_WriteLine("ERROR: Proxy server data could NOT be flattened.");
		
		refReleaseLock(inRef, &ptr);
	}
}// Flatten


/*!
Returns the user ID for the proxy server, used
for authentication.

The size of the given character buffer should
be 256 characters, since this is the largest
possible value allowed by protocols.

(3.0)
*/
void
ProxyServer_GetAuthenticatingUserName	(ProxyServerRef		inRef,
										 char*				outUserNamePtr)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		if (outUserNamePtr != null) strcpy(outUserNamePtr, ptr->userName);
		refReleaseLock(inRef, &ptr);
	}
}// GetAuthenticatingUserName


/*!
Returns the length of the user name string in octets
(used for authentication purposes).  This allows you
to allocate sufficient memory prior to invoking the
ProxyServer_GetAuthenticatingUserName() method.

(3.0)
*/
Size
ProxyServer_GetAuthenticatingUserNameSize	(ProxyServerRef		inRef)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	Size				result = 0;
	
	
	if (ptr != null)
	{
		result = strlen(ptr->userName) * sizeof(char);
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetAuthenticatingUserNameSize


/*!
Returns the location of the proxy server
(specified at construction time).  You would
call this routine when you need to describe
the location of the server (such as in a
text field).

If for some reason you want one or the other
but not both, pass "null" for the pointer to
the parameter you do not require.

The size of the given character buffer should
be 256 characters, since this is the largest
possible value allowed by protocols.

(3.0)
*/
void
ProxyServer_GetConnectionInfo		(ProxyServerRef			inRef,
									 InetDomainName			outDomainNamePtr,
									 InetPort*				outPortNumberPtr)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		if (outDomainNamePtr != null) strcpy(outDomainNamePtr, ptr->domainName);
		if (outPortNumberPtr != null) *outPortNumberPtr = ptr->portNumber;
		refReleaseLock(inRef, &ptr);
	}
}// GetConnectionInfo


/*!
Returns the protocol for a proxy server.

(3.0)
*/
TelnetProxyServerType
ProxyServer_GetType		(ProxyServerRef			inRef)
{
	TelnetProxyServerType		result = kTelnetProxyServerTypeSOCKS4;
	ProxyServerPtr				ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		result = ptr->type;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetType


/*!
Returns true only if the specified proxy server
is currently connected to the users machine
(possible via a successful call to the
ProxyServer_Connect() routine).

(3.0)
*/
Boolean
ProxyServer_IsConnected		(ProxyServerRef			inRef)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	Boolean				result = false;
	
	
	if (ptr != null)
	{
		result = ptr->isConnected;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// IsConnected


/*!
Disposes of a "ProxyDomainNameLookupData" data structure,
created by a call to ProxyServer_LookUpDomainName().
You should invoke this method somewhere within any method
of "ProxyServerDNRDoneNotificationProcPtr" form, to make
sure the memory for the DNR data is not leaked.

(3.0)
*/
void
ProxyServer_LookUpCleanUp	(ProxyDomainNameLookupDataPtr*		inoutPtr)
{
	Memory_DisposePtr((Ptr*)inoutPtr);
}// LookUpCleanUp


/*!
Sends a domain name lookup request to the specified
proxy server, using whatever means is appropriate for the
protocol of the server.  This method is asynchronous; when
the server responds (successfully or not), the specified
callback function is invoked, passing "inCompletionMethodData"
as one of the parameters (this data is defined by you).

Unimplemented.

(3.0)
*/
OSStatus
ProxyServer_LookUpDomainName	(ProxyServerRef							inRef,
								 Ptr									inHostNameBuffer,
								 Size const								inHostNameSize,
								 ProxyServerDNRDoneNotificationProcPtr	inCompletionMethod,
								 SInt32 const							inCompletionMethodData)
{
	OSStatus		result = cfragNoSymbolErr;
	
	
	return result;
}// LookUpDomainName


/*!
Specifies the user ID for the proxy server,
for authentication purposes.  You would call
this routine for any type of proxy server
that supports authentication, and not
otherwise.

The given string should be null-terminated.

(3.0)
*/
void
ProxyServer_SetAuthenticatingUserName	(ProxyServerRef		inRef,
										 char*				inUserName)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		if (inUserName != null) strcpy(ptr->userName, inUserName);
		refReleaseLock(inRef, &ptr);
	}
}// SetAuthenticatingUserName


/*!
Sets the location of the proxy server.  You
must call this routine as part of your proxy
server initialization.

(3.0)
*/
void
ProxyServer_SetConnectionInfo		(ProxyServerRef			inRef,
									 InetDomainName			inDomainName,
									 InetPort				inPortNumber)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		if (inDomainName != null) strcpy(ptr->domainName, inDomainName);
		ptr->portNumber = inPortNumber;
		refReleaseLock(inRef, &ptr);
	}
}// SetConnectionInfo


/*!
Specifies the protocol for a proxy server
(e.g. SOCKS version 4).

(3.0)
*/
void
ProxyServer_SetType		(ProxyServerRef				inRef,
						 TelnetProxyServerType		inType)
{
	ProxyServerPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ptr->type = inType;
		refReleaseLock(inRef, &ptr);
	}
}// SetType


//
// internal methods
//

/*!
To obtain a pointer to the internal structure,
given a reference to it, use this method.  You
must call refReleaseLock() to unlock the pointer
(and nullify it) when you are finished with it.

(3.0)
*/
static ProxyServerPtr
refAcquireLock	(ProxyServerRef		inRef)
{
	ProxyServerPtr		result = *((ProxyServerHandle)inRef);
	
	
	HLock((Handle)inRef);
	result->lockCount++;
	return result;
}// refAcquireLock


/*!
To tell the Proxy Server module that you are finished
using a proxy server configuration object pointer,
given a reference to it, use this method.  This
essentially makes the pointer useless - do not use it
again without calling the refAcquireLock() function to
obtain its (possibly changed) value.  On output, your
copy of the pointer will be set to "null".

(3.0)
*/
static void
refReleaseLock	(ProxyServerRef		inRef,
				 ProxyServerPtr*	inoutPtr)
{
	ProxyServerPtr		ptr = *((ProxyServerHandle)inRef);
	
	
	if (--ptr->lockCount <= 0) HUnlock((Handle)inRef);
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock
