/*###############################################################

	ProxyServer.h
	
	This module defines proxy servers in an abstract way.  You
	construct a proxy server of a specific type, and then
	Telnet interfaces with it automatically when creating new
	connections, looking up domain names, etc.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#ifndef __PROXYSERVER__
#define __PROXYSERVER__

// Mac includes
#include <MacTypes.h>
#include <OpenTransport.h>
#include <OpenTransportProviders.h>

// library includes
#include "InterfaceLib.h"



// constants
typedef SInt16 TelnetProxyServerType;
enum
{
	kTelnetProxyServerTypeSOCKS4 = 0,	// NEC SOCKS version 4.0, username-authenticated
	kTelnetProxyServerTypeSOCKS4A = 1,	// NEC SOCKS version 4.0, username-authenticated & proxy is auxiliary DNR
	kTelnetProxyServerTypeSOCKS5 = 2	// NEC SOCKS version 5.0, username- and password-authenticated, encrypted
};

// universal procedure pointers
/*!
Proxy Server DNR Done Notification Method

This routine is invoked when the finished domain name
lookup signal is received from a proxy server connection.
*/
typedef void (*ProxyServerDNRDoneNotificationProcPtr)	(SInt32 const						inData,
														 struct ProxyDomainNameLookupData*	inResolverPtr);
#define InvokeProxyServerDNRDoneNotificationProc(userRoutine, inData, inResolverPtr) \
			(*(userRoutine))((inData), (inResolverPtr))

// types
typedef struct OpaqueProxyServer { unsigned garbage; }**		ProxyServerRef;

struct ProxyDomainNameLookupData
{
	SInt32									data;					// additional information for the completion method
	OSStatus								error;					// on completion, any error that occurred
	InetDomainName							domainName;				// the domain name that is being resolved
	InetHostInfo*							hostInfoPtr;			// on completion, IP addresses for the domain name
	ProxyServerDNRDoneNotificationProcPtr	dnrCompletionMethod;	// the routine to call when the resolving completes
};
#ifndef __cplusplus
typedef struct ProxyDomainNameLookupData	ProxyDomainNameLookupData;
#endif
typedef ProxyDomainNameLookupData*			ProxyDomainNameLookupDataPtr;



//
// public methods
//

/*###############################################################
	CREATING AND DESTROYING PROXY SERVER CONFIGURATIONS
###############################################################*/

ProxyServerRef
	ProxyServer_New								(void);

// WARNING - ONLY PASS IN A HANDLE CONTAINING DATA THAT WAS ORIGINALLY DERIVED FROM A CALL TO ProxyServer_Flatten()!
ProxyServerRef
	ProxyServer_NewFlattened					(Handle							inDataHandle);

void
	ProxyServer_Dispose							(ProxyServerRef*				inoutRefPtr);

/*###############################################################
	INITIALIZING ANY PROXY SERVER
###############################################################*/

void
	ProxyServer_SetConnectionInfo				(ProxyServerRef					inRef,
												 InetDomainName					inDomainName,
												 InetPort						inPortNumber);

void
	ProxyServer_SetType							(ProxyServerRef					inRef,
												 TelnetProxyServerType			inType);

/*###############################################################
	INITIALIZING A PROXY SERVER USING SOCKS 4 AND BEYOND
###############################################################*/

void
	ProxyServer_SetAuthenticatingUserName		(ProxyServerRef					inRef,
												 char*							inUserName);

/*###############################################################
	OPENING COMMUNICATIONS
###############################################################*/

OSStatus
	ProxyServer_Connect							(ProxyServerRef					inRef,
												 InetAddress const*				inInternetAddressPtr);

void
	ProxyServer_Disconnect						(ProxyServerRef					inRef);

Boolean
	ProxyServer_IsConnected						(ProxyServerRef					inRef);

// USED ONLY WITHIN PROXY SERVER DNR-DONE CALLBACKS
void
	ProxyServer_LookUpCleanUp					(ProxyDomainNameLookupDataPtr*	inoutPtr);

// YOU MUST HAVE AN OPEN CONNECTION TO THE PROXY SERVER TO USE THIS ROUTINE
OSStatus
	ProxyServer_LookUpDomainName				(ProxyServerRef							inRef,
												 Ptr									inHostNameBuffer,
												 Size const								inHostNameSize,
												 ProxyServerDNRDoneNotificationProcPtr	inCompletionMethod,
												 SInt32 const							inCompletionMethodData);

/*###############################################################
	ACCESSING ANY PROXY SERVER
###############################################################*/

void
	ProxyServer_GetConnectionInfo				(ProxyServerRef					inRef,
												 InetDomainName					outDomainNamePtr,
												 InetPort*						outPortNumberPtr);

TelnetProxyServerType
	ProxyServer_GetType							(ProxyServerRef					inRef);

/*###############################################################
	ACCESSING A PROXY SERVER USING SOCKS 4 AND BEYOND
###############################################################*/

void
	ProxyServer_GetAuthenticatingUserName		(ProxyServerRef					inRef,
												 char*							outUserNamePtr);

// USE AS A PRECAUTION PRIOR TO CALLING ProxyServer_GetAuthenticatingUserName(), TO ENSURE THE BUFFER YOU PROVIDE IS LARGE ENOUGH
Size
	ProxyServer_GetAuthenticatingUserNameSize	(ProxyServerRef					inRef);

/*###############################################################
	SAVING PROXY SERVER SETTINGS
###############################################################*/

void
	ProxyServer_Flatten							(ProxyServerRef					inRef,
												 Handle*						outNewDataHandle);

#endif