/*###############################################################

	ProxyServerDialog.c
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1992 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <ControlDefinitions.h>
#include <MacErrors.h>
#include <Resources.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "TelnetProxyServerDialog.r"

// MacTelnet includes
#include "Console.h"
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "DialogResources.h"
#include "EventLoop.h"
#include "FolderManager.h"
#include "InterfaceLib.h"
#include "Preferences.h"
#include "ProxyServerDialog.h"
#include "ResourceFileManager.h"
#include "ScreenManager.h"
#include "StringResources.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"



// variables
extern MenuHandle					gListItemsDummyMenu;
static DialogRef					gProxyServerDialog = null; // the Mac OS dialog pointer
static WindowInfoRef				gProxyServerDialogWindowInfo = null;
static ControlRef					gOldFocusedControl = null;
static Boolean						gHaveDeactivatedRemoveButton = true;
static ControlKeyFilterUPP			gNumericalLimiterUPP = null;

// internal methods
static void				done							(void);
static void				getDomainsListHandle			(ListHandle*					outListHandle);
static void				handleCurrentFocus				(void);
static void				init							(StringPtr						inoutPreferencesResourceName);
static void				newDomain						(void);
static pascal Boolean	proxyServerDialogEventFilter	(DialogRef						inDialog,
														 EventRecord*					inoutEventPtr,
														 short*							outItemIndex);
static pascal ProxyServerRef	proxyServerFromDialogItems		(void);
static pascal void		proxyServerToDialogItems		(const ProxyServerRef			inRef);
static void				removeSelectedDomains			(void);
static void				setNewDomainButtonState			(void);
static void				setOKButtonState				(void);
static void				setRemoveButtonState			(void);



//
// public methods
//

/*!
This method displays and handles events in the
Proxy Server Setup dialog box.  If the user
clicks OK, "true" is returned; otherwise,
"false" is returned.

If the specified string is empty, this method
creates new data; otherwise, it looks for data
in the preferences file.

(3.0)
*/
Boolean
ProxyServerDialog_Display	(StringPtr		inoutPreferencesResourceName)
{
	Boolean		result = false;
	
	
	init(inoutPreferencesResourceName);
	
	if (gProxyServerDialog != null)
	{
		register DialogItemIndex	i = 0;
		DialogItemIndex				itemIndex = 0;
		ModalFilterUPP				filterProc = NewModalFilterUPP(proxyServerDialogEventFilter);
		
		
		// define this only for debugging; this lets you dump the dialogs
		// control embedding hierarchy to a file, before it is displayed
		#if 0
		DebugSelectControlHierarchyDumpFile(GetDialogWindow(gProxyServerDialog));
		#endif
		
		// display the dialog and focus the first text field
		TransitionDialogDisplay(gProxyServerDialog);
		{
			ControlRef		fieldControl = null;
			
			
			(OSErr)GetDialogItemAsControl(gProxyServerDialog, iProxyServerNameField, &fieldControl);
			(OSErr)SetKeyboardFocus(GetDialogWindow(gProxyServerDialog), fieldControl, kControlEditTextPart);
		}
		
		// handle events
		Menus_EnableEditMenu();
		itemIndex = 0;
		do
		{
			ModalDialog(filterProc, &itemIndex);
			switch (itemIndex)
			{
				case iProxyServerOKButton:
					// prevent saving preferences without a name, or with a name that is already in use;
					// however, if the name is identical to the given resource name, allow that (otherwise
					// the user would never be able to make later changes to a previously-saved configuration)
					{
						Str255		string;
						
						
						GetDialogItemControlText(gProxyServerDialog, iProxyServerNameField, string);
						unless (EqualString(string, inoutPreferencesResourceName,
											true/* case sensitive */, false/* diacritics sensitive */))
						{
							if ((StrLength(string) <= 0) || (!Menus_IsItemUnique(gListItemsDummyMenu, string)))
							{
								// the user has not entered an acceptable name, but tried to save; disallow this!
								BEEP;
								itemIndex = 0;
							}
						}
					}
					break;
				
				case iProxyServerAppleGuideButton:
					AppleGuideUtilities_OpenFirstMainGuide();
					break;
				
				case iProxyServerTabs:
					switch (GetDialogItemValue(gProxyServerDialog, iProxyServerTabs))
					{
						case 3:
							// show the Direct Domains tab
							HideDialogItem(gProxyServerDialog, iProxyServerInfoTabPane);
							HideDialogItem(gProxyServerDialog, iProxyServerAuthenticationTabPane);
							ShowDialogItem(gProxyServerDialog, iProxyServerDirectDomainsTabPane);
							unless (FlagManager_Test(kFlagOS9_0API))
							{
								// a bug in Mac OS 8.6 (and possibly earlier systems) wont redraw the tab
								DrawOneDialogItem(gProxyServerDialog, iProxyServerDirectDomainsTabPane);
							}
							break;
						
						case 2:
							// show the Authentication tab
							HideDialogItem(gProxyServerDialog, iProxyServerInfoTabPane);
							HideDialogItem(gProxyServerDialog, iProxyServerDirectDomainsTabPane);
							ShowDialogItem(gProxyServerDialog, iProxyServerAuthenticationTabPane);
							unless (FlagManager_Test(kFlagOS9_0API))
							{
								// a bug in Mac OS 8.6 (and possibly earlier systems) wont redraw the tab
								DrawOneDialogItem(gProxyServerDialog, iProxyServerAuthenticationTabPane);
							}
							break;
						
						case 1:
						default:
							// show the Server Info tab
							HideDialogItem(gProxyServerDialog, iProxyServerAuthenticationTabPane);
							HideDialogItem(gProxyServerDialog, iProxyServerDirectDomainsTabPane);
							ShowDialogItem(gProxyServerDialog, iProxyServerInfoTabPane);
							unless (FlagManager_Test(kFlagOS9_0API))
							{
								// a bug in Mac OS 8.6 (and possibly earlier systems) wont redraw the tab
								DrawOneDialogItem(gProxyServerDialog, iProxyServerInfoTabPane);
							}
							break;
					}
					break;
				
				case iProxyServerDirectDomainsList:
					// on double-click, remove selected items automatically
					{
						ControlRef		listControl = null;
						
						
						if (GetDialogItemAsControl(gProxyServerDialog, iProxyServerDirectDomainsList,
													&listControl) == noErr)
						{
							Boolean		isDoubleClick = false;
							Size		actualSize = 0;
							
							
							if (GetControlData(listControl, kControlListBoxDoubleClickPart,
												kControlListBoxDoubleClickTag, sizeof(Boolean), (Ptr)&isDoubleClick,
												&actualSize) == noErr)
							{
								if (isDoubleClick)
								{
									FlashButton(gProxyServerDialog, iProxyServerDirectDomainsRemoveButton);
									removeSelectedDomains();
								}
							}
						}
					}
					break;
				
				case iProxyServerDirectDomainsNewDomainButton:
					newDomain();
					break;
				
				case iProxyServerDirectDomainsRemoveButton:
					removeSelectedDomains();
					break;
				
				default:
					break;
			}
		} while ((itemIndex != iProxyServerOKButton) && (itemIndex != iProxyServerCancelButton));
		DisposeModalFilterUPP(filterProc), filterProc = null;
		
		// handle OK or Cancel click
		if (itemIndex != iProxyServerCancelButton)
		{
			ProxyServerRef			proxyRef = proxyServerFromDialogItems();
			
			
			// determine what the resource should be called
			GetDialogItemControlText(gProxyServerDialog, iProxyServerNameField, inoutPreferencesResourceName);
			
			// save the users settings by creating a Proxy Server object,
			// initializing it properly, flattening it into a resource
			// handle, and then saving the handle using the given name
			{
				Handle		proxyServerDataResourceHandle = null;
				SInt16		oldResFile = CurResFile();
				OSErr		error = noErr;
				
				
				Console_WriteLine("preparing to flatten proxy server data");
				ProxyServer_Flatten(proxyRef, &proxyServerDataResourceHandle);
				ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
				if (error == noErr)
				{
					AddResource(proxyServerDataResourceHandle, PROXYPREFS_RESTYPE,
								Unique1ID(PROXYPREFS_RESTYPE), inoutPreferencesResourceName);
					error = ResError();
				}
				ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
				UseResFile(oldResFile);
				if (proxyServerDataResourceHandle != null) Memory_DisposeHandle(&proxyServerDataResourceHandle);
				
				Console_WriteLine("done flattening proxy server data");
				
				// if any errors occur, report them
				(Boolean)Alert_ReportOSErr(error);
			}
			
			// clean up
			ProxyServer_Dispose(&proxyRef);
			TransitionDialogClose(gProxyServerDialog);
			result = true;
		}
		else
		{
			// user cancelled - close the dialog with an appropriate transition for cancelling
			TransitionDialogCloseCancel(gProxyServerDialog);
			result = false;
		}
	}
	else Alert_ReportOSErr(memFullErr);
	
	done();
	
	return result;
}// Display


//
// internal methods
//

/*!
Call this method to destroy the proxy server
dialog box and its associated data structures.

(3.0)
*/
static void
done	(void)
{
	WindowInfo_Dispose(gProxyServerDialogWindowInfo);
	DisposeControlKeyFilterUPP(gNumericalLimiterUPP), gNumericalLimiterUPP = null;
	DisposeDialog(gProxyServerDialog), gProxyServerDialog = null;
}// done


/*!
To get a standard ListHandle representation
of the direct domains list, use this
method.  If any problems occur, the output
list will be null.

(3.0)
*/
static void
getDomainsListHandle		(ListHandle*	outListHandle)
{
	ListUtilities_GetDialogItemListHandle(gProxyServerDialog, iProxyServerDirectDomainsList, outListHandle);
}// getDomainsListHandle


/*!
Determines the current keyboard focus and
updates other control states appropriately.
This routine does nothing if the current
keyboard focus at call time is no different
than it was the last time this routine was
invoked.

(3.0)
*/
static void
handleCurrentFocus		(void)
{
	ControlRef		focusedControl = null;
	
	
	if (GetKeyboardFocus(GetDialogWindow(gProxyServerDialog), &focusedControl) == noErr)
	{
		ControlRef		otherControl = null;
		
		
		if (gOldFocusedControl != focusedControl)
		{
			gOldFocusedControl = focusedControl;
			
			(OSErr)GetDialogItemAsControl(gProxyServerDialog, iProxyServerDirectDomainsField, &otherControl);
			if (focusedControl == otherControl) setNewDomainButtonState();
			
			(OSErr)GetDialogItemAsControl(gProxyServerDialog, iProxyServerDirectDomainsList, &otherControl);
			if (focusedControl == otherControl) setRemoveButtonState();
			
			(OSErr)GetDialogItemAsControl(gProxyServerDialog, iProxyServerNameField, &otherControl);
			if (focusedControl == otherControl) setOKButtonState();
		}
	}
}// handleCurrentFocus


/*!
This method is used to initialize the Proxy Server Setup
dialog box.  It creates the dialog box invisibly, and uses
the specified screen ID to set up the controls in the
dialog box to use the screens dimensions.

The extended dialog resource for the preferences dialog
causes the Appearance Manager to automatically construct
an embedding hierarchy, so the dialogs root control and
container controls are supposed to be set up once
GetNewDialog() returns.

(3.0)
*/
static void
init	(StringPtr		inoutPreferencesResourceName)
{
	gHaveDeactivatedRemoveButton = true;
	GetNewDialogWithWindowInfo(kDialogIDProxyServerSetup, &gProxyServerDialog, &gProxyServerDialogWindowInfo);
	
	if (gProxyServerDialog != null)
	{
		ProxyServerRef		proxyServerRef = null;
		Handle				proxyServerDataHandle = null;
		Boolean				isNewDataHandle = false;
		OSErr				error = noErr;
		
		
		if (inoutPreferencesResourceName[0] != 0)
		{
			// read the proxy server data from the preferences file
			SInt16		oldResFile = CurResFile();
			
			
			Console_WriteLine("preparing to read proxy server data from disk");
			isNewDataHandle = false;
			ResourceFileManager_UseResFile(kTelnetResourceFilePreferences);
			proxyServerDataHandle = Get1NamedResource(PROXYPREFS_RESTYPE, inoutPreferencesResourceName);
			error = ResError();
			Console_WriteValue("proxy server read resource error", error);
			proxyServerRef = ProxyServer_NewFlattened(proxyServerDataHandle);
			if (proxyServerDataHandle != null) ReleaseResource(proxyServerDataHandle), proxyServerDataHandle = null;
			UseResFile(oldResFile);
			Console_WriteLine("done reading proxy server data from disk");
		}
		else
		{
			// make sure the name is unique
			proxyServerRef = ProxyServer_New();
			isNewDataHandle = true;
			Menus_GetUniqueItemText(gListItemsDummyMenu, inoutPreferencesResourceName);
		}
		
		// display a warning message if any error occurred, ignoring resources that do not exist
		unless (error == resNotFound) Alert_ReportOSErr(error);
		
		GetKeyboardFocus(GetDialogWindow(gProxyServerDialog), &gOldFocusedControl);
		
		// determine if the Help button should be disabled...
		SetUpHelpButtonDialogItem(gProxyServerDialog, iScreenSizeAppleGuideButton);
		
		// adjust the OK and Cancel buttons so they are big enough for their titles
		SetDialogDefaultItem(gProxyServerDialog, iProxyServerOKButton);
		SetDialogCancelItem(gProxyServerDialog, iProxyServerCancelButton);
		(UInt16)Localization_AdjustDialogButtons(gProxyServerDialog, true/* OK */, true/* Cancel */);
		Localization_AdjustHelpButtonItem(gProxyServerDialog, iProxyServerAppleGuideButton);
		
		// make the Set... button sufficiently large
		(UInt16)Localization_AutoSizeButtonItem(gProxyServerDialog, iProxyServerAuthenticationPasswordSetButton, 0);
		
		// until SOCKS 5 is implemented, disable the password info
		DeactivateDialogItem(gProxyServerDialog, iProxyServerAuthenticationPasswordLabel);
		DeactivateDialogItem(gProxyServerDialog, iProxyServerAuthenticationSecurityIcon);
		DeactivateDialogItem(gProxyServerDialog, iProxyServerAuthenticationPasswordSetButton);
		
		// show the first tab, hide the others
		HideDialogItem(gProxyServerDialog, iProxyServerAuthenticationTabPane);
		HideDialogItem(gProxyServerDialog, iProxyServerDirectDomainsTabPane);
		ShowDialogItem(gProxyServerDialog, iProxyServerInfoTabPane);
		
		// size the list cells appropriately
		{
			ListHandle		list = null;
			
			
			getDomainsListHandle(&list);
			{
				Point			newCellDimensions;
				ControlRef		listControl = null;
				Rect			listBounds;
				
				
				(OSErr)GetDialogItemAsControl(gProxyServerDialog, iProxyServerDirectDomainsList, &listControl);
				GetControlBounds(listControl, &listBounds);
				SetPt(&newCellDimensions, listBounds.right - listBounds.left, 18);
				LSetDrawingMode(false, list);
				LCellSize(newCellDimensions, list);
				LSetDrawingMode(true, list);
			}
		}
		
		// set button states
		setOKButtonState();
		setNewDomainButtonState();
		setRemoveButtonState();
		
		// prevent the user from typing non-digits into the Port text field
		{
			ControlRef		field = null;
			
			
			gNumericalLimiterUPP = NewControlKeyFilterUPP(NumericalLimiter);
			GetDialogItemAsControl(gProxyServerDialog, iProxyServerPortField, &field);
			(OSErr)SetControlData(field, kControlNoPart, kControlKeyFilterTag, sizeof(gNumericalLimiterUPP),
									(Ptr)&gNumericalLimiterUPP);
		}
		
		// fill in the data in the dialog box
		SetDialogItemControlText(gProxyServerDialog, iProxyServerNameField, inoutPreferencesResourceName);
		proxyServerToDialogItems(proxyServerRef);
		
		// set up the Window Info information
		WindowInfo_SetWindowDescriptor(gProxyServerDialogWindowInfo, kTelnetDialogDescriptorProxyServerSetup);
		WindowInfo_SetForDialog(gProxyServerDialog, gProxyServerDialogWindowInfo);
	}
}// init


/*!
Call this routine in response to a click
of the New Domain button, to create a
new list item containing whatever text is
entered in the direct domain text field.

(3.0)
*/
static void
newDomain	(void)
{
	ListHandle		list = null;
	
	
	getDomainsListHandle(&list);
	if (list != null)
	{
		Cell		cell;
		Str255		controlText;
		
		
		// get the domain name that the user entered
		GetDialogItemControlText(gProxyServerDialog, iProxyServerDirectDomainsField, controlText);
		
		if (StrLength(controlText) > 0)
		{
			// add the item to the top of the list
			(SInt16)LAddRow(1, 0, list);
			
			// set the column data
			SetPt(&cell, 0, 0);
			LSetCell((Ptr)(1 + controlText), StrLength(controlText), cell, list);
			
			// update the list and clear the text field
			DrawOneDialogItem(gProxyServerDialog, iProxyServerDirectDomainsList);
			SetDialogItemControlText(gProxyServerDialog, iProxyServerDirectDomainsField, EMPTY_PSTRING);
			DrawOneDialogItem(gProxyServerDialog, iProxyServerDirectDomainsField);
		}
		else BEEP; // the user didnt type anything
	}
}// newDomain


/*!
Handles special key events in the
proxy server configuration editor
dialog box.

(3.0)
*/
static pascal Boolean
proxyServerDialogEventFilter	(DialogRef		inDialog,
								 EventRecord*	inoutEventPtr,
								 short*			outItemIndex)
{
	Boolean		result = pascal_false,
				isButtonAndShouldFlash = false; // simplifies handling of multiple key-to-button maps
	
	
	// first, make sure help balloons show up properly for the cursor location, if Balloon Help is on
	if (IsPopUpHelpEnabled()) DisplayBalloonForDialogItemUnderMouse(inDialog, inoutEventPtr->where,
																		kDialogBalloonsIDProxyServerSetup);
	
	// set control states properly
	handleCurrentFocus();
	
	if (inoutEventPtr->what == keyDown)
	{
		short		key = 0;
		
		
		key = inoutEventPtr->message & charCodeMask;
		if (((key == 'I') || (key == 'i')) && (inoutEventPtr->modifiers & cmdKey))
		{
			// selected Server Info (tab #1)
			if (GetDialogItemValue(inDialog, iProxyServerTabs) != 1)
			{
				*outItemIndex = iProxyServerTabs;
				SetDialogItemValue(inDialog, iProxyServerTabs, 1);
				if (FlagManager_Test(kFlagAppearance1_1API)) PlayThemeSound(kThemeSoundTabRelease);
				result = pascal_true;
			}
		}
		else if (((key == 'A') || (key == 'a')) && (inoutEventPtr->modifiers & cmdKey))
		{
			// selected Authentication (tab #2)
			if (GetDialogItemValue(inDialog, iProxyServerTabs) != 2)
			{
				*outItemIndex = iProxyServerTabs;
				SetDialogItemValue(inDialog, iProxyServerTabs, 2);
				if (FlagManager_Test(kFlagAppearance1_1API)) PlayThemeSound(kThemeSoundTabRelease);
				result = pascal_true;
			}
		}
		else if (((key == 'D') || (key == 'd')) && (inoutEventPtr->modifiers & cmdKey))
		{
			// selected Direct Domains (tab #3)
			if (GetDialogItemValue(inDialog, iProxyServerTabs) != 3)
			{
				*outItemIndex = iProxyServerTabs;
				SetDialogItemValue(inDialog, iProxyServerTabs, 3);
				if (FlagManager_Test(kFlagAppearance1_1API)) PlayThemeSound(kThemeSoundTabRelease);
				result = pascal_true;
			}
		}
		else if (((key == 'N') || (key == 'n')) && (inoutEventPtr->modifiers & cmdKey))
		{
			// selected New Domain button
			ControlRef		button = null;
			
			
			(OSErr)GetDialogItemAsControl(inDialog, iProxyServerDirectDomainsNewDomainButton, &button);
			if (IsControlActive(button))
			{
				*outItemIndex = iProxyServerDirectDomainsNewDomainButton;
				isButtonAndShouldFlash = result = pascal_true;
			}
		}
		else if (((key == 'R') || (key == 'r')) && (inoutEventPtr->modifiers & cmdKey))
		{
			// selected Remove button
			ControlRef		button = null;
			
			
			(OSErr)GetDialogItemAsControl(inDialog, iProxyServerDirectDomainsRemoveButton, &button);
			if (IsControlActive(button))
			{
				*outItemIndex = iProxyServerDirectDomainsRemoveButton;
				isButtonAndShouldFlash = result = pascal_true;
			}
		}
	}
	
	if (!result) result = (StandardDialogEventFilter(inDialog, inoutEventPtr, outItemIndex));
	else if (isButtonAndShouldFlash) FlashButton(inDialog, *outItemIndex);
	
	return result;
}// proxyServerDialogEventFilter


/*!
To create a proxy server object using the current
values of the dialog items, use this method.

(3.0)
*/
static pascal ProxyServerRef
proxyServerFromDialogItems		(void)
{
	ProxyServerRef			result = ProxyServer_New();
	TelnetProxyServerType	serverType = kTelnetProxyServerTypeSOCKS4;
	Str255					string;
	Str31					portString;
	SInt32					port = 0;
	
	
	// save the data in the object
	switch (GetDialogItemValue(gProxyServerDialog, iProxyServerProtocolPopUpMenu))
	{
		case 3: // SOCKS 5
			serverType = kTelnetProxyServerTypeSOCKS5;
			break;
		
		case 2: // SOCKS 4A
			serverType = kTelnetProxyServerTypeSOCKS4A;
			break;
		
		case 1: // SOCKS 4
		default:
			serverType = kTelnetProxyServerTypeSOCKS4;
			break;
	}
	GetDialogItemControlText(gProxyServerDialog, iProxyServerHostField, string);
	StringUtilities_PToCInPlace(string);
	GetDialogItemControlText(gProxyServerDialog, iProxyServerPortField, portString);
	StringToNum(portString, &port);
	ProxyServer_SetConnectionInfo(result, (char*)string, port);
	ProxyServer_SetType(result, serverType);
	
	GetDialogItemControlText(gProxyServerDialog, iProxyServerAuthenticationUserNameField, string);
	StringUtilities_PToCInPlace(string);
	ProxyServer_SetAuthenticatingUserName(result, (char*)string);
	
	return result;
}// proxyServerFromDialogItems


/*!
To set up the dialog items to reflect the current
values from a proxy server object, use this method.

(3.0)
*/
static pascal void
proxyServerToDialogItems	(const ProxyServerRef		inRef)
{
	if (inRef != null)
	{
		char					string[256];
		InetPort				port = 0;
		TelnetProxyServerType	serverType = ProxyServer_GetType(inRef);
		
		
		// fill in the Server Info tab
		ProxyServer_GetConnectionInfo(inRef, string, &port);
		StringUtilities_CToPInPlace(string);
		SetDialogItemControlText(gProxyServerDialog, iProxyServerHostField, (StringPtr)string);
		NumToString(port, (StringPtr)string);
		SetDialogItemControlText(gProxyServerDialog, iProxyServerPortField, (StringPtr)string);
		switch (serverType)
		{
			case kTelnetProxyServerTypeSOCKS5:
				SetDialogItemValue(gProxyServerDialog, iProxyServerProtocolPopUpMenu, 3);
				break;
			
			case kTelnetProxyServerTypeSOCKS4A:
				SetDialogItemValue(gProxyServerDialog, iProxyServerProtocolPopUpMenu, 2);
				break;
			
			case kTelnetProxyServerTypeSOCKS4:
			default:
				SetDialogItemValue(gProxyServerDialog, iProxyServerProtocolPopUpMenu, 1);
				break;
		}
		
		// fill in the Authentication tab
		string[0] = ProxyServer_GetAuthenticatingUserNameSize(inRef);
		ProxyServer_GetAuthenticatingUserName(inRef, string + 1);
		SetDialogItemControlText(gProxyServerDialog, iProxyServerAuthenticationUserNameField, (StringPtr)string);
	}
}// proxyServerToDialogItems


/*!
Call this routine in response to a click
of the Remove Domain button, to delete
all selected domain names from the list.

(3.0)
*/
static void
removeSelectedDomains	(void)
{
	ListHandle		list = null;
	
	
	getDomainsListHandle(&list);
	if (list != null)
	{
		Cell		cellToDelete;
		UInt16		listItemCount = 0,
					listSelectedItemCount = 0;
		
		
		// delete all selected items
		listSelectedItemCount = ListUtilities_GetSelectedItemCount(list);
		SetPt(&cellToDelete, 0, 0);
		LSetDrawingMode(false, list);
		while (LGetSelect(true/* next */, &cellToDelete, list))
		{
			LDelRow(1, cellToDelete.v, list);
		}
		LSetDrawingMode(true, list);
		
		// update the list
		DrawOneDialogItem(gProxyServerDialog, iProxyServerDirectDomainsList);
	}
}// removeSelectedDomains


/*!
Makes the New Domains button the default
button if its text field is focused; if not,
the button that was previously the default
button is made default again.

(3.0)
*/
static void
setNewDomainButtonState		(void)
{
	ControlRef				focusedControl = null;
	static DialogItemIndex	oldDefaultButton = iProxyServerOKButton;
	
	
	if (GetKeyboardFocus(GetDialogWindow(gProxyServerDialog), &focusedControl) == noErr)
	{
		ControlRef		newDomainField = null;
		
		
		(OSErr)GetDialogItemAsControl(gProxyServerDialog, iProxyServerDirectDomainsField, &newDomainField);
		if (focusedControl == newDomainField)
		{
			DialogItemIndex		defaultButton = GetDialogDefaultItem(gProxyServerDialog);
			
			
			if (defaultButton != iProxyServerDirectDomainsNewDomainButton) oldDefaultButton = defaultButton;
			SetDialogDefaultItem(gProxyServerDialog, iProxyServerDirectDomainsNewDomainButton);
		}
		else if (oldDefaultButton != GetDialogDefaultItem(gProxyServerDialog))
		{
			SetDialogDefaultItem(gProxyServerDialog, oldDefaultButton);
		}
	}
}// setNewDomainButtonState


/*!
Makes the OK button the default button if
the top text field is focused; if not, the
button that was previously the default button
is made default again.

(3.0)
*/
static void
setOKButtonState	(void)
{
	ControlRef				focusedControl = null;
	static DialogItemIndex	oldDefaultButton = iProxyServerOKButton;
	
	
	if (GetKeyboardFocus(GetDialogWindow(gProxyServerDialog), &focusedControl) == noErr)
	{
		ControlRef		directDomainsList = null,
						newDomainField = null;
		
		
		(OSErr)GetDialogItemAsControl(gProxyServerDialog, iProxyServerDirectDomainsList, &directDomainsList);
		(OSErr)GetDialogItemAsControl(gProxyServerDialog, iProxyServerDirectDomainsField, &newDomainField);
		if ((focusedControl == directDomainsList) && (focusedControl != newDomainField))
		{
			DialogItemIndex		defaultButton = GetDialogDefaultItem(gProxyServerDialog);
			
			
			if (defaultButton != iProxyServerOKButton) oldDefaultButton = defaultButton;
			SetDialogDefaultItem(gProxyServerDialog, iProxyServerOKButton);
		}
		else if (oldDefaultButton != GetDialogDefaultItem(gProxyServerDialog))
		{
			SetDialogDefaultItem(gProxyServerDialog, oldDefaultButton);
		}
	}
}// setOKButtonState


/*!
Activates or deactivates the Remove button
based on whether any items in the direct
domains list are selected.

If the Remove button can be activated, it is
made the default button as well.  If the
Remove button is deactivated, the button that
was previously the default button is made
default again.

(3.0)
*/
static void
setRemoveButtonState	(void)
{
	ListHandle					list = null;
	static DialogItemIndex		oldDefaultButton = iProxyServerOKButton;
	
	
	getDomainsListHandle(&list);
	if (list != null)
	{
		UInt16		numberOfItemsSelected = ListUtilities_GetSelectedItemCount(list);
		
		
		if (numberOfItemsSelected > 0)
		{
			DialogItemIndex		defaultButton = GetDialogDefaultItem(gProxyServerDialog);
			
			
			ActivateDialogItem(gProxyServerDialog, iProxyServerDirectDomainsRemoveButton);
			if (defaultButton != iProxyServerDirectDomainsRemoveButton) oldDefaultButton = defaultButton;
			SetDialogDefaultItem(gProxyServerDialog, iProxyServerDirectDomainsRemoveButton);
			gHaveDeactivatedRemoveButton = false;
		}
		else if (!gHaveDeactivatedRemoveButton)
		{
			DeactivateDialogItem(gProxyServerDialog, iProxyServerDirectDomainsRemoveButton);
			SetDialogDefaultItem(gProxyServerDialog, oldDefaultButton);
			gHaveDeactivatedRemoveButton = true;
		}
	}
}// setRemoveButtonState