/*###############################################################

	RasterGraphicsScreen.c
	
	This file contains the Macintosh real screen calls for
	the NCSA Virtual Graphics Kernel.
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1992 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/
/*
 *      by Gaige B. Paulsen
 *
 *	Following are the Per Device calls:
 *
 *   MacRGraster( p,x1,y1,x2,y2,wid)- Plot raster in rect @(x1,y1,x2,y2) with wid @ p
 *   MacRGcopy( x1,y1,x2,y2,x3,y3,x4,y4)- 
 *   MacRGmap( offset,count,data)   - 
 *
 *
 *  WARNING, WARNING!
 *  Gaige has this cute idea about how to do "subwindows" of real windows by shifting
 *  the window number by 4 bits (MAC_WINDOW_SHIFT).  Then, the remainder is the
 *  sub-window number.  It will probably work, but you MUST keep the shifted and
 *  non-shifted numbers straight.  For example, MacRGdestroy() and MacRGremove() take
 *  different uses of the window number right now.  
 *
 *
 *  Macintosh only Routines:
 *
 *      Version Date    Notes
 *      ------- ------  ---------------------------------------------------
 *		0.5 	880912	Initial Coding -GBP
 *      1.0     890216  Minor fixes for 1.0 - TKK
 */

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstring>
#else
#	include <string.h>
#endif

// Mac includes
#include <Scrap.h>

// resource includes
#include "StringResources.h"

// MacTelnet includes
#define __ALLNU__

#include "Clipboard.h"
#include "Console.h"
#include "MacLook.h"
#include "VirtualDevices.h"
#include "RasterGraphicsKernel.h"
#include "InternalEvents.h"

#define MAX_MAC_RGS			8
#define MAC_WINDOW_SHIFT	4		/* Bits shifted */

#include "RasterGraphicsScreen.h"
#include "TelnetErrors.h"



// types
struct RealGraphicsWindow
{
	VirtualDeviceRef	virtualDevice;		// virtual device to draw in - it has its own colors
	WindowRef 			window;				// window (null if not in use)
	PaletteHandle 		palette;			// the colors
	char				title[256];			// the window title
	Point				size;				// the window dimensions
};
typedef struct RealGraphicsWindow	RealGraphicsWindow;
typedef RealGraphicsWindow*			RealGraphicsWindowPtr;
typedef RealGraphicsWindowPtr		RealGraphicsWindowArray;

// variables
static RealGraphicsWindowArray		MacRGs;
static short						RGwn = 0;		// window number in use

// internal methods
static short	getCurrentICRWindowID		(void);
static void		setCurrentICRWindowID		(const short		inNewRG);



//
// public methods
//

/**
This method creates the global array of real
graphics windows, "MacRGs".

(2.6)
*/
void
MacRGinit	(void)
{
	short	i = 0;
	
	
	MacRGs = (RealGraphicsWindowArray)Memory_NewPtr(MAX_MAC_RGS * sizeof(RealGraphicsWindow));
	for (i = 0; i < MAX_MAC_RGS; i++) MacRGs[i].window = null;
}// MacRGinit


/************************************************************
 * MacRGnewwindow( name, x1, y1, x2, y2)		-			*
 *		make a new subwindow to wn							*
 ************************************************************/
short
MacRGnewwindow		(const char*		inTitle,
					 const short		inX1,
					 const short		inY1,
					 const short		inX2,
					 const short		inY2)
{
	short		result = 0;
	short		x1 = inX1,
				y1 = inY1,
				x2 = inX2,
				y2 = inY2;
	short		w = 0,
				i = 0;
	Rect		wDims;
	
	
	Console_WriteLine("Preparing to create a new ICR graphics window...");
	
	for (w = 0; w < MAX_MAC_RGS && MacRGs[w].window; w++) {}
	
	if (w >= MAX_MAC_RGS) result = -1;
	else
	{
		register SInt16		i = 0;
		
		
		// make sure the width is even
		if ((x2 - x1) & 1) x2++;
		
		SetRect(&wDims, x1 + 80, y1 + 80, x2 + 80, y2 + 80); // arbitrary location
		strcpy(MacRGs[w].title, inTitle);
		
		// create a Mac OS window
		Console_WriteLine("creating Mac OS window...");
		{
			Str255		title;
			
			
			strcpy((char*)title, inTitle);
			StringUtilities_CToPInPlace((char*)title);
			if (FlagManager_Test(kFlagCarbonWindowManager))
			{
				OSStatus		error = 0L;
				
				
				error = CreateNewWindow(kDocumentWindowClass,
										kWindowCloseBoxAttribute | kWindowCollapseBoxAttribute,
										&wDims, &MacRGs[w].window);
				if (error == noErr)
				{
					SetWTitle(MacRGs[w].window, title);
					MoveWindow(MacRGs[w].window, wDims.left, wDims.top, true/* activate */);
					SizeWindow(MacRGs[w].window, wDims.right - wDims.left,
								wDims.bottom - wDims.top, false/* update */);
					ShowWindow(MacRGs[w].window);
					SelectWindow(MacRGs[w].window);
				}
				else MacRGs[w].window = null;
			}
			else
			{
				MacRGs[w].window = NewCWindow(null/* window storage unsupported in Carbon */,
												&wDims, title, true/* visible */, noGrowDocProc,
												kInFront, true/* close box */,
												0L/* MUST be zero for Window Info! */);
			}
			
			if (MacRGs[w].window != null)
			{
				SetWindowKind(MacRGs[w].window, WIN_ICRG);
			}
			else result = -1;
		}
		
		// create a grayscale palette
		if (result >= 0)
		{
			RGBColor	gray;
			
			
			Console_WriteLine("creating a grayscale palette...");
			MacRGs[w].palette = NewPalette(256/* number of entries */,
											(CTabHandle)null/* source colors */,
											pmTolerant, 0/* tolerance */);
			if (MacRGs[w].palette != null)
			{
				// NOTE:	QuickDraw always assumes that white is at location 0 and
				//			black is at location 255, even though this is the inverse
				//			of the actual intensity values.
				for (i = 0; i < 256; i++)
				{
					gray.red = i << 8;
					gray.green = i << 8;
					gray.blue = i << 8;
					SetEntryColor(MacRGs[w].palette, i, &gray);
				}
			}
			else result = -1;
		}
		
		// now create the virtual device
		if (result >= 0)
		{
			VDeviceError	error = kVDeviceErrorNoError;
			
			
			Console_WriteLine("creating virtual device...");
			error = VirtualDevices_New(&MacRGs[w].virtualDevice, &wDims, MacRGs[w].palette);
			
			if ((error != kVDeviceErrorNoError) && (error != kVDeviceErrorEmptyBoundaries))
			{
				result = -1;
			}
		}
		
		if (result >= 0)
		{
			MacRGs[w].size.h = x2 - x1;
			MacRGs[w].size.v = y2 - y1;
			setCurrentICRWindowID(w);
			result = (w << MAC_WINDOW_SHIFT);
		}
	}
	
	// in the event of an error, clean up and tell the user
	if (result < 0)
	{
		if (MacRGs[w].palette != null)
		{
			DisposePalette(MacRGs[w].palette), MacRGs[w].palette = null;
		}
		if (MacRGs[w].window != null)
		{
			DisposeWindow(MacRGs[w].window), MacRGs[w].window = null;
		}
		Errors_DisplayStopMessage(rStringsMemoryErrors,
									siMemErrorCannotCreateRasterGraphicsWindow, -1);
	}
	
	Console_WriteValue("result", result);
	return result;
}// MacRGnewwindow


/****************************************
 * MacRGsetwindow(wn)		-			*
 *		set the drawing window to wn	*
 ****************************************/
void
MacRGsetwindow	(const short		wn)
{
	short		w = wn >> MAC_WINDOW_SHIFT;
	
	
	if (MacRGs[w].window != null)
	{
		SetPortWindowPort(MacRGs[w].window);
		setCurrentICRWindowID(w);
		
		// optionally set the clip region
	}
}// MacRGsetwindow


/**
To dispose of an ICR graphics window,
use this method.

(2.6)
*/
void
MacRGdestroy	(const short		wn)
{
	if (MacRGs[wn].window != null)
	{
		VRdestroybyName((char*)&MacRGs[wn].title);
	}
}// MacRGdestroy

	
/**
To dispose of an ICR graphics subwindow,
use this method.

(2.6)
*/
void
MacRGremove		(const short		wn)
{
	short		w = wn >> MAC_WINDOW_SHIFT;
	
	
	if (MacRGs[w].window != null)
	{
		PixMapHandle	portPixMap = null;
		
		
		VirtualDevices_Dispose(&MacRGs[w].virtualDevice);
		portPixMap = GetPortPixMap(GetWindowPort(MacRGs[w].window));
	#if TARGET_API_MAC_OS8
		(*(*(portPixMap))->pmTable)->ctSeed = GetCTSeed(); // unseed color table
	#else
		// ??? - is the pixmap structure private under Carbon?
		(*(*(portPixMap))->pmTable)->ctSeed = GetCTSeed(); // unseed color table
	#endif
		
		DisposeWindow(MacRGs[w].window), MacRGs[w].window = null;
		if (MacRGs[w].palette != null)
		{
			DisposePalette(MacRGs[w].palette), MacRGs[w].palette = null;
		}
	}
}// MacRGremove


/**
To find the screen ID for an ICR graphics
window, given its Mac OS window pointer,
use this method.

(2.6)
*/
short
MacRGfindwind	(WindowRef		inWindow)
{
	short	result = 0;
	
	
	if (inWindow == null) result = -2;
	else
	{
		while (result < MAX_MAC_RGS && inWindow != MacRGs[result].window) result++;
		if (result == MAX_MAC_RGS) result = -1;
	}
	return result;
}// MacRGfindwind


/************************************************************************************/
/* MacRGcopy
*  Copybits the image window into the clipboard.
*  
*/
void
MacRGcopy	(WindowRef		inWindow)
{
	Rect copysize,copyfrom;
	long len,wn;
	PicHandle picture;
	PixMapHandle hidep;
		
	if (( wn= MacRGfindwind( inWindow)) <0)
		return;						/* Couldn't do it */
 	
	VirtualDevices_GetPixelMap(MacRGs[wn].virtualDevice, &hidep);
	VirtualDevices_GetBounds(MacRGs[wn].virtualDevice, &copyfrom);
	
	SetPortWindowPort(inWindow);
	
	copysize = copyfrom;						/* boundary of drawing area */

	picture = OpenPicture(&copysize);
	{
		BitMap const*	bitMapPtr = null;
		RGBColor		black,
						white;
		
		
		black.red = black.green = black.blue = 0;
		white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
		ClipRect(&copysize);
		RGBForeColor(&black);
		RGBBackColor(&white);
		//ForeColor(blackColor);
		//BackColor(whiteColor);
		LockPixels(hidep);
		VirtualDevices_GetBitMapForCopyBits(MacRGs[wn].virtualDevice, &bitMapPtr);
		CopyBits(bitMapPtr, GetPortBitMapForCopyBits(GetWindowPort(inWindow)),
					&copyfrom, &copysize, srcCopy, null);
		UnlockPixels(hidep);
	}
	ClosePicture();
	
	// copy picture to clipboard
	len = GetHandleSize((Handle)picture);
	HLock((Handle)picture);
#if TARGET_API_MAC_OS8
	(OSStatus)ZeroScrap();
	PutScrap(len, kScrapFlavorTypePicture, (Ptr)*picture);
#else
	(OSStatus)ClearCurrentScrap();
	{
		ScrapRef		currentScrap = null;
		
		
		if (GetCurrentScrap(&currentScrap) == noErr)
		{
			(OSStatus)PutScrapFlavor(currentScrap, kScrapFlavorTypePicture,
										kScrapFlavorMaskNone, GetHandleSize((Handle)picture),
										*picture);
		}
	}
#endif
	HUnlock((Handle)picture);
	KillPicture(picture);
}// MacRGcopy


short MacRGupdate( WindowRef wind)
{
	short		result = 0;
	short		wn = 0;
	GWorldPtr	theWorld = null;
	
	
	Console_WriteLine("updating ICR graphics window...");
	if ((wn = MacRGfindwind(wind)) < 0) result = -1;
	else
	{
		Rect		worldRect,
					windowRect;
		RGBColor	black,
					white;
		
		
		VirtualDevices_GetGraphicsWorld(MacRGs[wn].virtualDevice, &theWorld);
		SetPortWindowPort(wind);
		black.red = black.green = black.blue = 0;
		white.red = white.green = white.blue = RGBCOLOR_INTENSITY_MAX;
		RGBForeColor(&black);
		RGBBackColor(&white);
		//ForeColor(blackColor);
		//BackColor(whiteColor);
		BeginUpdate(wind);
		LockPixels(GetPortPixMap(theWorld));
		GetPortBounds(theWorld, &worldRect);
		GetPortBounds(GetWindowPort(wind), &windowRect);
		CopyBits(GetPortBitMapForCopyBits(theWorld),
					GetPortBitMapForCopyBits(GetWindowPort(wind)),
					&worldRect, &windowRect, srcCopy, null);
		UnlockPixels(GetPortPixMap(theWorld));
		EndUpdate(wind);
	}
	
	Console_WriteValue("result", result);
	return result;
}


/**
To render graphics data in a specific region,
use this method.

(2.6)
*/
short
MacRGraster		(char*				data,
				 const short		x1,
				 const short		y1,
				 const short		x2,
				 const short		y2,
				 const short		rowbytes)
{
	short		result = 0;	// success is 0
	
	
	Console_WriteLine("Preparing to render...");
	if (MacRGs[RGwn].window != null)
	{
		CGrafPtr		oldPort = null;
		GDHandle		oldDevice = null;
		Rect			tr;
		char*			baseOfThisRun = null;
		short			i = 0;
		PixMapHandle	pixMap = null;
		GWorldPtr		theOffscreenWorld = null;
		long			temp = 0L;
		
		
		Console_WriteLine("getting graphics world...");
		VirtualDevices_GetGraphicsWorld(MacRGs[RGwn].virtualDevice, &theOffscreenWorld);
		
		GetGWorld(&oldPort, &oldDevice); // save old settings
		SetGWorld(theOffscreenWorld, null);
		
		Console_WriteLine("getting pixels and locking them...");
		VirtualDevices_GetPixelMap(MacRGs[RGwn].virtualDevice, &pixMap);
		LockPixels(pixMap);
		baseOfThisRun = (char*)GetPixBaseAddr(pixMap);
		temp = (long)((*pixMap)->rowBytes & 0x3FFF);
		baseOfThisRun += temp * y1 + x1;
		
		Console_WriteLine("copying data...");
		{
			char*		p = baseOfThisRun;
			char*		q = data;
			
			
			for (i = 0; i < rowbytes; i++) *p++ = *q++;
		}
		UnlockPixels(pixMap);
		
		SetRect(&tr, x1, y1, x2 + 1, y2 + 1);
		Console_WriteLine("generating update event...");
		MacLook_InvalidateRectangle(MacRGs[RGwn].window, &tr);
		
		SetGWorld(oldPort, oldDevice);
	}
	else result = -1;
	
	Console_WriteValue("result", result);
	return result;
}// MacRGraster


/**
To set up the color palette for the current
graphics window, use this method.

(2.6)
*/
short
MacRGmap	(const short		start,
			 const short		length,
			 char*				data)
{
	short		result = 0; // presently undefined; success is 0
	short		i = 0;
	char*		p = data;
	RGBColor	color;
	
	
	Console_WriteLine("Setting up ICR color palette...");
	for (i = start; i < start + length; i++)
	{
		color.red = (*p++) << 8;
		color.green = (*p++) << 8;
		color.blue = (*p++) << 8;
		SetEntryColor(MacRGs[RGwn].palette, i, &color);
	}
	NSetPalette(MacRGs[RGwn].window, MacRGs[RGwn].palette, pmAllUpdates);
	ActivatePalette(MacRGs[RGwn].window);
	VirtualDevices_SetColorTable(MacRGs[RGwn].virtualDevice, MacRGs[RGwn].palette);
	
	Console_WriteValue("result", result);
	return result;
}// MacRGmap


//
// internal methods
//

/**
To determine the ID of the ICR window
currently in use, use this method.

(3.0)
*/
static short
getCurrentICRWindowID	(void)
{
	short		result = RGwn;
	
	
	return result;
}// getCurrentICRWindowID


/**
To specify the ID of the ICR window to
subsequently work with, use this method.

(3.0)
*/
static void
setCurrentICRWindowID	(const short		inNewRG)
{
	RGwn = inNewRG;
}// setCurrentICRWindowID