/*###############################################################

	RealScreen.h
	
	This is the real screen module, which defines the visual
	component of a terminal screen and tends to refer to screen
	coordinates in pixels.  Compare this to the module virtual
	screen, which works with terminal screens in terms of the
	data in them, and tends to refer to screen coordinates in
	rows and columns.
	
	Generally, you dont work with real screens; instead, you
	modify the data in a window using the virtual screen kernel,
	and then those changes are propagated to this module for
	rendering.
	
	This entire module has received fairly hefty attention in
	MacTelnet 3.0!  To be blunt, it was a disaster of coding,
	and was something that needed re-writing.  Version 3.0
	improves upon this entire module in hundreds of ways, not
	the least of which is the fact that the entire terminal
	screen is now drawn using Appearance controls.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#ifndef __REALSCREEN__
#define __REALSCREEN__

// Mac includes
#include <Controls.h>
#include <Dialogs.h>
#include <Events.h>
#include <MacWindows.h>

// MacTelnet includes
#include "RealScreenData.h"
#include "ScreenManager.h"



// constants
enum
{
	// The values of these indices *must* correspond to
	// the order in which StatusIcon_AddStage() is invoked
	// in "RealScreen.c" for new screen windows.
	kEncryptionIconIndexInsecure = 0,
	kEncryptionIconIndexEncrypting = 1,
	kEncryptionIconIndexSecure = 2,
	kEncryptionIconIndexDecrypting = 3
};

typedef SInt16 RealScreenColorIndex;
enum
{
	// indices into the color palette of a terminal window; the
	// implementation demands that these be consecutive and zero-based
	// (the value for each color constant matches the index, into the
	// palette of a terminal window, corresponding to that color)
	kRealScreenColorIndexNormalText = 0,
	kRealScreenColorIndexNormalBackground = 1,
	kRealScreenColorIndexBlinkingText = 2,
	kRealScreenColorIndexBlinkingBackground = 3,
	kRealScreenColorIndexBoldText = 4,
	kRealScreenColorIndexBoldBackground = 5,
	// the ORDER of these colors must be the same as the numerical order of the
	// ANSI standard values (0 = black, 1 = red, etc., 6 = cyan, 7 = white)
	kRealScreenColorIndexNormalANSIBlack = 6,
	kRealScreenColorIndexNormalANSIRed = 7,
	kRealScreenColorIndexNormalANSIGreen = 8,
	kRealScreenColorIndexNormalANSIYellow = 9,
	kRealScreenColorIndexNormalANSIBlue = 10,
	kRealScreenColorIndexNormalANSIMagenta = 11,
	kRealScreenColorIndexNormalANSICyan = 12,
	kRealScreenColorIndexNormalANSIWhite = 13,
	// the ORDER of these colors must be the same as the numerical order of the
	// ANSI standard values (0 = black, 1 = red, etc., 6 = cyan, 7 = white)
	kRealScreenColorIndexEmphasizedANSIBlack = 14,
	kRealScreenColorIndexEmphasizedANSIRed = 15,
	kRealScreenColorIndexEmphasizedANSIGreen = 16,
	kRealScreenColorIndexEmphasizedANSIYellow = 17,
	kRealScreenColorIndexEmphasizedANSIBlue = 18,
	kRealScreenColorIndexEmphasizedANSIMagenta = 19,
	kRealScreenColorIndexEmphasizedANSICyan = 20,
	kRealScreenColorIndexEmphasizedANSIWhite = 21,
	// counts of the above constants
	kRealScreenColorCountNonANSIColors = 6,
	kRealScreenColorCountNormalANSIColors = 8,
	kRealScreenColorCountEmphasizedANSIColors = 8
};

typedef UInt32 RealScreenFlags;
enum
{
	kRealScreenFlagTextWraps = (1 << 0),		// wrap mode initially enabled?
	kRealScreenFlagWindowHeader = (1 << 1)		// window header initially visible?
};

typedef UInt16 RealScreenTextFlags;
enum
{
	kRealScreenTextFlagInline = (1 << 0)		// replace end-of-line markers with spaces?
};

typedef UInt16 RealScreenRegionCode;
enum
{
	kRealScreenRegionCodeScreenArea = 0,		// the scrollable area
	kRealScreenRegionCodeScreenBackground = 1	// the background area (the scrollable area plus a small border)
};



/*###############################################################
	INITIALIZING THE REAL SCREEN MODULE
###############################################################*/

void
	RealScreen_Init								(void);

void
	RealScreen_Done								(void);

/*###############################################################
	CREATING AND DESTROYING REAL SCREENS
###############################################################*/

TelnetScreenID
	RealScreen_New								(TelnetScreenID const		inIDOfNewScreen,
												 SInt16 const				inColumnCount,
												 SInt16 const				inRowCount,
												 StringPtr					inFontName,
												 SInt16 const				inFontSize,
												 RealScreenFlags const		inFlags);

void
	RealScreen_Dispose							(TelnetScreenID*			inoutScreenIDPtr);

/*###############################################################
	REAL SCREEN ROUTINES
###############################################################*/

void
	RealScreen_AudioEvent						(TelnetScreenID const		inScreenID);

void
	RealScreen_DisplaySetWTitleDialogFor		(WindowRef					inForWhichWindow);

void
	RealScreen_EraseSection						(TelnetScreenID const		inScreenID,
												 SInt16 const				inLeftmostColumnToErase,
												 SInt16 const				inTopmostRowToErase,
												 SInt16 const				inRightmostColumnToErase,
												 SInt16 const				inBottommostRowToErase);
Boolean
	RealScreen_GetColor							(TelnetScreenID const		inScreenID,
												 RealScreenColorIndex const	inColorEntryNumber,
												 RGBColor*					outColorPtr);

// THE BOTTOM-RIGHT INSETS ARE GIVEN AS POSITIVE NUMBERS, MEASURED UP AND TO THE LEFT FROM THE BOTTOM-RIGHT CORNER
void
	RealScreen_GetInsets						(Point*						outTopLeftInsetOrNull,
												 Point*						outBottomRightInsetOrNull);

UInt16
	RealScreen_GetPlacardHeight					(TelnetScreenID const		inScreenID);

void
	RealScreen_GetRegionBounds					(TelnetScreenID const		inScreenID,
												 RealScreenRegionCode const	inRegionCode,
												 Rect*						outBounds);

UInt16
	RealScreen_GetScrollBarHeight				(void);

UInt16
	RealScreen_GetScrollBarWidth				(void);

Handle
	RealScreen_GetSelectedTextAsNewHandle		(TelnetScreenID const		inScreenID,
												 UInt16 const				inNumberOfSpacesToReplaceWithOneTabOrZero,
												 RealScreenTextFlags const	inFlags);

RgnHandle
	RealScreen_GetSelectedTextAsNewRegion		(TelnetScreenID const		inScreenID);

SInt16
	RealScreen_HandleContentClick				(WindowRef					inWindow,
												 EventRecord*				inoutEventPtr);

pascal Boolean
	RealScreen_HandleNewSize					(WindowRef const			inWindowRef,
												 SInt32 const				inDeltaSizeX,
												 SInt32 const				inDeltaSizeY,
												 SInt32 const				inData);

Boolean
	RealScreen_IsDrawingEnabled					(TelnetScreenID const		inScreenID);

Boolean
	RealScreen_IsTextSelected					(TelnetScreenID const		inScreenID);

Boolean
	RealScreen_IsTextSelectionRectangular		(TelnetScreenID const		inScreenID);

void
	RealScreen_LocalToScreen					(TelnetScreenID const		inScreenID,
												 SInt16*					inoutHorizontalPixelOffsetFromPortOrigin,
												 SInt16*					inoutVerticalPixelOffsetFromPortOrigin);

void
	RealScreen_LocalToScreenRect				(TelnetScreenID const		inScreenID,
												 Rect*						inoutPortOriginBounds);

void
	RealScreen_MakeSelectionsRectangular		(TelnetScreenID const		inScreenID,
												 Boolean const				inAreSelectionsNotAttachedToScreenEdges);

void
	RealScreen_MoveCursorWithArrowKeys			(TelnetScreenID const		inScreenID,
												 Point const				inLocalMouse);

void
	RealScreen_NewWidth							(TelnetScreenID const		inScreenID,
												 UInt16 const				inNewNumberOfColumns);

void
	RealScreen_NextWindow						(void);

// IF THE CURSOR IS SET TO FLASH, THIS ROUTINE HIDES THE CURSOR UNTIL THE IDLING ROUTINE REDRAWS IT
void
	RealScreen_ObscureCursor					(TelnetScreenID const		inScreenID);

Boolean
	RealScreen_PtInRgn							(TelnetScreenID const		inScreenID,
												 Point const				inGlobalPoint,
												 RealScreenRegionCode const	inRegionCode);

Boolean
	RealScreen_PtInSelection					(TelnetScreenID const		inScreenID,
												 Point const				inLocalPoint);

void
	RealScreen_ReverseVideo						(TelnetScreenID const		inScreenID,
												 Boolean const				inReverseVideo);

void
	RealScreen_ScreenToLocal					(TelnetScreenID const		inScreenID,
												 SInt16*					inoutHorizontalPixelOffsetFromScreenOrigin,
												 SInt16*					inoutVerticalPixelOffsetFromScreenOrigin);

void
	RealScreen_ScreenToLocalRect				(TelnetScreenID const		inScreenID,
												 Rect*						inoutScreenOriginBounds);

void
	RealScreen_SelectNothing					(TelnetScreenID const		inScreenID);

void
	RealScreen_SelectVirtualRange				(TelnetScreenID const		inScreenID,
												 Point const				inRowColumnSelectionStart,
												 Point const				inRowColumnSelectionEnd);

void
	RealScreen_SetActive						(WindowRef					inWindow,
												 Boolean const				inActivating);

// THIS ALSO SETS THE STATE OF ANSI GRAPHICS MODE
void
	RealScreen_SetANSIColorsEnabled				(TelnetScreenID const		inScreenID,
												 Boolean const				inUseANSIColorSequences);

Boolean
	RealScreen_SetColor							(TelnetScreenID const		inScreenID,
												 RealScreenColorIndex const	inColorEntryNumber,
												 RGBColor const*			inColorPtr);

void
	RealScreen_SetCursorVisibility				(TelnetScreenID const		inScreenID,
												 Boolean const				inIsVisible);

void
	RealScreen_SetDrawingEnabled				(TelnetScreenID const		inScreenID,
												 Boolean const				inIsDrawingEnabled);

SInt16
	RealScreen_SetPortScreenPort				(TelnetScreenID const		inScreenID);

void
	RealScreen_SetSpecialBackground				(TelnetScreenID const		inScreenID,
												 Boolean const				inUsesSpecialBackground);

void
	RealScreen_SetUpCursor						(TelnetScreenID const		inScreenID,
												 SInt16 const				inX,
												 SInt16 const				inY);

void
	RealScreen_SetWTitle						(TelnetScreenID const		inScreenID,
												 Str255						inName);

short
	RealScreen_SizeWindow						(WindowRef					inWindow,
												 Point						inGlobalMousePosition,
												 EventModifiers				inModifiers);

void
	RealScreen_StackWindows						(void);

void
	RealScreen_TextFont							(StringPtr					inFontFamilyName,
												 SInt16 const				inFontSize,
												 StyleParameter const		inFontStyle);
void
	RealScreen_TrackTextSelection				(TelnetScreenID const		inScreenID,
												 Point const				pt,
												 EventRecord*				inoutEventPtr);

void
	RealScreen_UpdateHorizontalScrollBar		(TelnetScreenID const		inScreenID,
												 SInt16 const				inInvisibleCharacterPositionCount,
												 SInt16 const				inLeftmostVisibleColumn);

void
	RealScreen_UpdateVerticalScrollBar			(TelnetScreenID const		inScreenID,
												 UInt16 const				inScrollbackBufferSize,
												 SInt16 const				inTopmostVisibleLineNumber,
												 SInt16 const				inBottommostVisibleLineNumber);

void
	RealScreen_UpdateSetWTitleDialogUsing		(unsigned char*				inSourcePtr,
												 Size						inSize,
												 DialogRef const			inDialog);

void
	RealScreen_ZoomWindow						(WindowRef					inWindow,
												 WindowPartCode const		inWindowPartFromCallToFindWindow,
												 Boolean const				inActivateAfterZooming);



// OBSOLETE METHODS - being phased out
Boolean	ClickInContent		(Point const			inLocalMouseLocation,
							 TelnetScreenID const	inScreenID);
void RSchangefont(TelnetScreenID const inScreenID, StringPtr inFontName, long fsiz);
SInt16
	RSgetfont	(TelnetScreenID const		inScreenID,
				 short*						outFontID,
				 short*						outFontSize);
void	FlashSelection	(TelnetScreenID const		inScreenID);

void	RSdraw		(TelnetScreenID const		inScreenID,
					 SInt16 const				inStartingColumnNumber,
					 SInt16 const				inLineNumberToDraw,
					 SInt16 const				inAttributes,
					 SInt16 const				inCharacterCount,
					 char const*				inTextBufferPtr);
void	RSdelcols	(TelnetScreenID const		inScreenID,
					 SInt16 const				inDeleteHowManyColumns);
void RSdelchars(TelnetScreenID const inScreenID, short x, short y, short n);
void RSdellines(TelnetScreenID const inScreenID, short t, short b, short n, short scrolled);
void RSinslines(TelnetScreenID const inScreenID, short t, short b, short n, short scrolled);
void RSinscols(TelnetScreenID const inScreenID, short n);
void RSinsstring(TelnetScreenID const inScreenID, short x, short y, short a, short len, char *ptr);
void	RSsortAnchors(TelnetScreenID const inScreenID);
void RSsetsize(TelnetScreenID const inScreenID);
void RScheckmaxwind(TelnetScreenID const inScreenID, Rect *origRect, short origW, short origH, short *endW, short *endH);

#endif