/*###############################################################

	IconListDef.c
	
	A custom list definition procedure displaying icons with 
	text labels.  Works with Icon Services, when available.
	Note - the list definition itself doesnt increment or 
	decrement the reference count of IconRefs.  This file has 
	been slightly modified for MacTelnet (see subsequent 
	comments).
	
	Icon List Definition 2.5
		 2001-2002 by Ian Anderson and Kevin Grant.
	
	This module is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
	This module is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.
	
	You should have received a copy of the GNU General Public
	License along with this module; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

// Mac includes
#include <Gestalt.h>
#include <Icons.h>
#include <Lists.h>
#include <MacMemory.h>
#include <MacTypes.h>
#include <Quickdraw.h>
#include <QuickdrawText.h>
#include <TextEdit.h>
#include <ToolUtils.h>

//#include "IconListDef.h"

// MacTelnet includes
// SPECIAL WARNING:  Currently this file is being cross-compiled 
// as a Classic PPC 'LDEF' code resource as well as a regular 
// Carbon C source file.  As long as this is the case, dont use 
// any functions from any non-system, non-shared library, or 
// another source file (or the LDEF will become unreasonably 
// large).
#include "UniversalDefines.h" // defines accessor macros

// constants
enum
{
	kIconWidth = 32,		// pixels
	kIconHeight = 32,		// pixels
	kTextHeight = 14,		// pixels - an estimate, depends on font metrics
	kIconTextSpacingV = 2,	// pixels
	kContentHeight = kIconHeight + kTextHeight + kIconTextSpacingV
};

//
// internal methods - variable names changed to conform with MacTelnet coding style
//

static void	calculateDrawingBounds		(Rect const*					inCellBoundsPtr,
										 Rect*							outIconRect,
										 Rect*							outTextRect);

static void	drawIconListCell			(ListHandle						inList,
										 Rect const*					inCellBoundsPtr,
										 Boolean						inUseIconServices,
										 StandardIconListCellDataPtr	inCellDataPtr,
										 Boolean						inIsCellSelected);

//
// external methods
//

pascal void
	IconListDef		(short			inMessage,
					 Boolean		inIsCellSelected,
					 Rect*			inCellBoundsPtr,
					 Cell			inAffectedCell,
					 short			inDataOffset,
					 short			inDataSizeInBytes,
					 ListHandle		inList);



/*!
The main entry point to the list definition procedure, of 
standard ListDefProc form.

(2.0)
*/
pascal void
IconListDef		(short			inMessage,
				 Boolean		inIsCellSelected,
				 Rect*			inCellBoundsPtr,
				 Cell			UNUSED_ARGUMENT(inAffectedCell),
				 short			inDataOffset,
				 short			inDataSizeInBytes,
				 ListHandle		inList)
{
	static Boolean	useIconServices = false;
	
	
	switch (inMessage)
	{
		case lInitMsg:
			{
				OSStatus	error = noErr;
				long		gestaltResult = 0L;
				
				
				error = Gestalt(gestaltIconUtilitiesAttr, &gestaltResult);
				useIconServices = ((error == noErr) && ((gestaltResult & 
									(1L << gestaltIconUtilitiesHasIconServices)) != 0));
			}
			break;
		
		// Drawing and highlighting are put together because of 
		// text smoothing (anti-aliasing).  Simply using the 
		// QuickDraw hilitetransfermode to paint the cell 
		// doesnt work because the text will be smoothed 
		// according to the background color which would still 
		// be white.  So instead it is necessary to: set the 
		// background color according to whether or not the cell 
		// is highlighted, erase the cell, then draw its 
		// contents for both highlight and draw messages.  An 
		// alternative would be, for lHilite messages, to set 
		// the background color, erase only the region around 
		// the icon, then draw only the text but messing with 
		// all the Regions would probably end up taking more 
		// time so this just erases the entire cell.
		case lDrawMsg:
		case lHiliteMsg:
			// sanity check on buffer size (should always pass)
			if (inDataSizeInBytes == sizeof(StandardIconListCellDataRec))
			{
				StandardIconListCellDataPtr		cellDataPtr = null;
				
				
				// (optimized for MacTelnet)
				// simply point cellDataPtr to the correct byte in the cells DataHandle
				SInt8	oldHState = '\0';
				
				
				// save handle state for later restoration; lock in case
				// a system call that moves memory is invoked (which could
				// invalidate the pointer if the handle wasnt locked)
				oldHState = HGetState((**inList).cells);
				HLock((**inList).cells);
				cellDataPtr = REINTERPRET_CAST(*((**inList).cells) + inDataOffset, 
												StandardIconListCellDataPtr);
				if (cellDataPtr != null)
				{
					drawIconListCell(inList, inCellBoundsPtr, useIconServices, cellDataPtr, 
										inIsCellSelected);
				}
				HSetState((**inList).cells, oldHState);
			}
			break;
	}
}// IconListDef


//
// internal methods
//
#pragma mark -

/*!
Determines the region that the icon and its text label will 
occupy within a particular cells boundaries.  Pass "null" for 
either of the 2nd and 3rd parameters if you are not interested 
in the value.

(2.0)
*/
static void
calculateDrawingBounds	(Rect const*	inCellBoundsPtr,
						 Rect*			outIconRect,
						 Rect*			outTextRect)
{
	short	iconTop;
	
	
	iconTop = inCellBoundsPtr->top + 
				INTEGER_HALVED(inCellBoundsPtr->bottom - inCellBoundsPtr->top - kContentHeight);
	
	if (outIconRect != null)
	{
		short	cellCenterH = INTEGER_HALVED(inCellBoundsPtr->left + inCellBoundsPtr->right);
		
		
		SetRect(outIconRect, cellCenterH - INTEGER_HALVED(kIconWidth), iconTop, 
				cellCenterH + INTEGER_HALVED(kIconWidth), iconTop + kIconHeight);
	}
	
	if (outTextRect != null)
	{
		SetRect(outTextRect, inCellBoundsPtr->left, 
				iconTop + kIconHeight + kIconTextSpacingV, inCellBoundsPtr->right, 
				iconTop + kIconHeight + kIconTextSpacingV + kTextHeight);
	}
}// calculateDrawingBounds


/*!
First checks to see if the specified List Cell is highlighted, 
then sets the background color appropriately, erases the cell (so 
that the new background color is actually drawn), and draws the 
icon and text label.  Because this function correctly sets the 
highlighting, theres no need for a separate function to do so.

(2.0)
*/
static void
drawIconListCell	(ListHandle						inList,
					 Rect const*					inCellBoundsPtr,
					 Boolean						inUseIconServices,
					 StandardIconListCellDataPtr	inCellDataPtr,
					 Boolean						inIsCellSelected)
{
	GrafPtr		savedPort = null;
	CGrafPtr	listPort = null;
	RGBColor	savedColor;
	Boolean		active = false;
	Rect		iconRect,
				textRect;
	
	
	GetPort(&savedPort);
	listPort = GetListPort(inList);
	SetPort((GrafPtr)listPort);
	
	GetBackColor(&savedColor);
	
	if (inIsCellSelected)
	{
		// set the background to the highlight color
		RGBColor	highlightColor;
		
		
		LMGetHiliteRGB(&highlightColor);
		RGBBackColor(&highlightColor);
	}
	
	EraseRect(inCellBoundsPtr);
	
	calculateDrawingBounds(inCellBoundsPtr, &iconRect, &textRect);
	active = GetListActive(inList);
	
	if (inUseIconServices)
	{
		// draw the icon using Icon Services, if available
		(OSErr)PlotIconRef(&iconRect, kAlignNone, active ? kTransformNone : kTransformDisabled, 
							kIconServicesNormalUsageFlag,
							REINTERPRET_CAST(inCellDataPtr->iconHandle, IconRef));
	}
	else
	{	// draw the icon using Icon Utilities if Icon Services isnt available
		(OSErr)PlotIconSuite(&iconRect, kAlignNone, 
								active ? kTransformNone : kTransformDisabled, 
								REINTERPRET_CAST(inCellDataPtr->iconHandle, IconSuiteRef));
	}
	
	// draw label text
	{
		short	savedFont = 0,
				savedSize = 0;
		Style	savedFace = normal;
		
		
		savedFont = GetPortTextFont(listPort);
		savedFace = GetPortTextFace(listPort);
		savedSize = GetPortTextSize(listPort);
		
		TextFont(inCellDataPtr->font);
		TextFace((Style)(inCellDataPtr->face));
		TextSize(inCellDataPtr->size);
		
		TETextBox(&inCellDataPtr->name[1], inCellDataPtr->name[0], &textRect, teCenter);
		
		TextFont(savedFont);
		TextFace(savedFace);
		TextSize(savedSize);
	}
	
	RGBBackColor(&savedColor);
	SetPort(savedPort);
}// drawIconListCell