/*###############################################################

	SOCKS.h
	
	Lists methods and constants applicable to accessing SOCKS
	(versions 4 and 5) proxy servers.  MacTelnet will completely
	support SOCKS versions 4, 4A, and 5, the latter supporting
	both GSS-API and username/password authentication schemes.
	
	IMPORTANT:	This file defines structures for packets; in
				reality, this is an organizational scheme and
				is only used for quick reference when writing
				code to transmit octets, to ensure that the
				data is sent in the right order.  Compilers
				will optimize the alignment of a structure to
				be most efficient for the microprocessor,
				possibly padding bytes.  Do NOT, therefore,
				even THINK about sending an entire structure
				as is along a network connection!  You must
				refer to specific data on an octet-by-octet
				basis, and send it in precisely the order and
				amount defined by the SOCKS protocol.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#ifndef __SOCKS__
#define __SOCKS__

// Mac includes
#include <OpenTransport.h>



// constants
enum
{
	// The default port number for SOCKS; the value of this constant is defined by
	// the SOCKS protocol, and should not be changed.
	kSOCKSPortNumberDefault = 1080
};

// If a BIND command receives this value as a "resultIP", then no IP address was
// provided by the server, so the IP address *of the proxy server* should be used
// instead.  This happens when server hosts are not multi-homed.  The value of
// this constant is defined by the SOCKS protocol, and should not be changed.
#define		kSOCKSResultIPNotAssignedUseSOCKSProxyServerIP \
				((0x00 << 24) | (0x00 << 16) | (0x00 << 8) | (0x00 << 0))

// If available DNS cant resolve a domain name, MacTelnet has the option (with
// SOCKS version 4A, that is) to let the proxy server try to resolve it.  The
// value below should be assigned to the "destinationIP" and sent as a request
// if this occurs.  Any IP address whose first three octets are zero, and last
// octet is not zero, is an invalid IP address, and is therefore an acceptable
// value for this constant.
#define		kSOCKSResultIPNotKnownProxyServerMustResolve \
				((0x00 << 24) | (0x00 << 16) | (0x00 << 8) | (0x01 << 0))

enum
{
	// Packets are null-terminated by an all-zeroes octet; send this value as a single
	// octet at the end of a variable-length packet.  The value of this constant is
	// defined by the SOCKS protocol, and should not be changed.
	kSOCKSNULL = 0x00	// packets are null-terminated by an all-zeroes octet
};

typedef UInt8 SOCKSAddressType;
enum
{
	// the values of these constants are defined by the SOCKS protocol, and should not be changed
	kSOCKSAddressTypeIPAddressVersion4 = 0x01,	// a four-octet IP address
	kSOCKSAddressTypeDomainName = 0x03,			// a domain name, like apple.com, as a Pascal string
	kSOCKSAddressTypeIPAddressVersion6 = 0x04	// a larger IP address
};

typedef UInt8 SOCKSCommand;
enum
{
	// the values of these constants are defined by the SOCKS protocol, and should not be changed
	kSOCKSCommandCONNECT = 1,
	kSOCKSCommandBIND = 2,
	kSOCKSCommandUDP_ASSOCIATE = 3	// SOCKS 5 only
};

typedef UInt8 SOCKSMethod;
enum
{
	// the values of these constants are defined by the SOCKS protocol, and should not be changed
	kSOCKSMethodNoAuthenticationRequired = 0x00,
	kSOCKSMethodGeneralSecurityServicesAPI = 0x01,
	kSOCKSMethodUserNameAndPassword = 0x02,
	// 0x80 to 0xFE are reserved
	kSOCKSMethodUnacceptable = 0xFF
};

typedef UInt8 SOCKSReply;	// SOCKS 5
enum
{
	// the values of these constants are defined by the SOCKS protocol, and should not be changed
	kSOCKSReplySuccess = 0x00,
	kSOCKSReplyGeneralServerFailure = 0x01,
	kSOCKSReplyConnectionNotAllowedByRuleSet = 0x02,
	kSOCKSReplyUnreachableNetwork = 0x03,
	kSOCKSReplyUnreachableHost = 0x04,
	kSOCKSReplyConnectionRefused = 0x05,
	kSOCKSReplyTTLExpired = 0x06,
	kSOCKSReplyUnsupportedCommand = 0x07,
	kSOCKSReplyUnsupportedAddressType = 0x08
};

typedef UInt8 SOCKSResult;	// SOCKS 4, 4A
enum
{
	// the values of these constants are defined by the SOCKS protocol, and should not be changed
	kSOCKSResultRequestGranted = 90,
	kSOCKSResultRequestRejectedOrFailed = 91,	// generic; the following error codes are more specific
	kSOCKSResultRequestRejectedCannotConnectToIdentificationDaemon = 92,
	kSOCKSResultRequestRejectedIdentificationDaemonAndUserDoNotAgree = 93
};

// types
struct SOCKSMethodNegotiationHeader
{
	UInt8		versionNumber;		// VN; 5 (the SOCKS protocol version number)
	union
	{
		struct
		{
			UInt8		methodCount;	// number of methods in the variable-length "methods" array
			Handle		methods;		// variable-length array of between 1 and 255 "kSOCKSMethod..." constants
			UInt8		nullByte;		// set to kSOCKSNULL
		} v5;
	} socks;
};
#ifndef __cplusplus
typedef struct SOCKSMethodNegotiationHeader	SOCKSMethodNegotiationHeader;
#endif
typedef SOCKSMethodNegotiationHeader*		SOCKSMethodNegotiationHeaderPtr;

struct SOCKSRequestPacket
{
	UInt8		versionNumber;		// VN; 4 or 5 (the SOCKS protocol version number)
	union
	{
		struct
		{
			UInt8		commandCode;		// CD; a "kSOCKSCommand..." constant
			SInt16		destinationPort;	// port number; for CONNECT, of the destination; for BIND, of the proxy
			UInt32		destinationIP;		// IP address; for CONNECT, of the destination; for BIND, of the proxy
			Handle		userID;				// identification of the user
			UInt8		nullByte;			// set to kSOCKSNULL
			Handle		addressToResolve;	// SOCKS 4A; IF MacTelnet cant resolve a host name, it sends it here
			UInt8		terminatorByte;		// set to kSOCKSNULL
		} v4;
		
		struct
		{
			UInt8		commandCode;		// CD; a "kSOCKSCommand..." constant
			UInt8		reservedSetToZero;	// a single octet, set to 0x00
			UInt8		addressType;		// a "kSOCKSAddressType..." constant
			Handle		address;			// variable-length array of octets; composition depends on "addressType"
			SInt16		destinationPort;	// port number; for CONNECT, of the destination; for BIND, of the proxy
		} v5;
	} socks;
};
#ifndef __cplusplus
typedef struct SOCKSRequestPacketHeader	SOCKSRequestPacketHeader;
#endif
typedef SOCKSRequestPacketHeader*		SOCKSRequestPacketHeaderPtr;

struct SOCKSReplyPacket
{
	UInt8		versionNumber;		// VN; 0 (the SOCKS reply code version number)
	union
	{
		struct
		{
			UInt8		resultCode;			// CD; a "kSOCKSResult..." constant
			SInt16		resultPort;			// not used for CONNECT; for BIND, this is the connection port number
			UInt32		resultIP;			// not used for CONNECT; for BIND, this is the connection IP address
		} v4;
		
		struct
		{
			UInt8		reply;				// REP; a "kSOCKSReply..." constant
			UInt8		reservedSetToZero;	// a single octet, set to 0x00
			UInt8		addressType;		// a "kSOCKSAddressType..." constant
			Handle		address;			// variable-length array of octets; composition depends on "addressType"
			SInt16		destinationPort;	// port number; for CONNECT, of the destination; for BIND, of the proxy
		} v5;
	} socks;
};
#ifndef __cplusplus
typedef struct SOCKSReplyPacket		SOCKSReplyPacket;
#endif
typedef SOCKSReplyPacket*			SOCKSReplyPacketPtr;

struct SOCKSMethodSelectionPacket
{
	UInt8		versionNumber;		// VN (the SOCKS method selection packet version number)
	union
	{
		struct
		{
			UInt8		method;		// selected method; a "kSOCKSMethod..." constant
		} v5;
	} socks;
};
#ifndef __cplusplus
typedef struct SOCKSMethodSelectionPacket	SOCKSMethodSelectionPacket;
#endif
typedef SOCKSMethodSelectionPacket*			SOCKSMethodSelectionPacketPtr;



//
// public methods
//

#endif