/*###############################################################

	ScreenManager.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "ConnectionData.h"
#include "Connections.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "Inspector.h"
#include "MacLook.h"
#include "Parse.h"
#include "Preferences.h"
#include "PreferencesData.h"
#include "RealScreen.h"
#include "ScreenManager.h"
#include "SizeDialog.h"
#include "TelnetDescriptors.h"
#include "Terminal.h"



// constants
enum
{
	kMaximumRealScreens = kMaximumAllowedTelnetConnections + 1
};

// types
struct ScreenData
{
	ConnectionDataPtr		connectionDataPtr;
	RealScreenDataPtr		realScreenDataPtr;
};
typedef struct ScreenData	ScreenData;
typedef ScreenData*			ScreenDataPtr;

// variables
static UInt16		gScreenHiddenCount = 0;
static ChainRef		gScreenDataList = null;


// internal methods
static void				findHorizontalScrollBarScreenOp		(ConnectionDataPtr const	inConnectionDataPtr,
															 RealScreenDataPtr const	inRealScreenDataPtr,
															 SInt32 const				inControlRefToFind,
															 SInt32 const				inData2,
															 void*						inoutTelnetScreenIDPtr);
static void				findPortScreenOp					(ConnectionDataPtr const	inConnectionDataPtr,
															 RealScreenDataPtr const	inRealScreenDataPtr,
															 SInt32 const				inPortNumberToFind,
															 SInt32 const				inData2,
															 void*						inoutTelnetScreenIDPtr);
static void				findVerticalScrollBarScreenOp		(ConnectionDataPtr const	inConnectionDataPtr,
															 RealScreenDataPtr const	inRealScreenDataPtr,
															 SInt32 const				inControlRefToFind,
															 SInt32 const				inData2,
															 void*						inoutTelnetScreenIDPtr);
static void				findWindowScreenOp					(ConnectionDataPtr const	inConnectionDataPtr,
															 RealScreenDataPtr const	inRealScreenDataPtr,
															 SInt32 const				inWindowRefToFind,
															 SInt32 const				inData2,
															 void*						inoutTelnetScreenIDPtr);
static ScreenDataPtr	getScreenDataOf						(TelnetScreenID const		inScreenID);
static Boolean			validID								(TelnetScreenID const		inScreenID);



//
// public methods
//

/*!
This method sets up the initial configurations for
the screen manager in general.  Call this method
before calling any other methods from this file.

(3.0)
*/
void
ScreenManager_Init	(void)
{
	// create linked lists for important screen data
	gScreenDataList = Chain_New(kChainStyleSinglyLinked);
	
	// create Session Inspector
	Inspector_Init();
}// Init


/*!
This method cleans up the screen manager by
destroying data structures allocated by
ScreenManager_Init().

(3.0)
*/
void
ScreenManager_Done	(void)
{
	ChainLinkRef		node = null;
	ScreenDataPtr		screenDataPtr = null;
	
	
	// destroy connection data, cleaning up connections
	node = Chain_GetFirstNode(gScreenDataList);
	do
	{
		screenDataPtr = Chain_GetNodeDataPtr(gScreenDataList, node); // actually a "ScreenDataPtr"
		if (screenDataPtr != null)
		{
			Memory_DisposePtr((Ptr*)&screenDataPtr->realScreenDataPtr);
			if (screenDataPtr->connectionDataPtr != null)
			{
				Connections_Disconnect(screenDataPtr->connectionDataPtr);
				Memory_DisposePtr((Ptr*)&screenDataPtr->connectionDataPtr);
			}
			Memory_DisposePtr((Ptr*)&screenDataPtr);
		}
		node = Chain_GetNextNode(gScreenDataList, node);
	} while (node != null);
	Chain_Dispose(&gScreenDataList);
	
	// general clean-up
	Inspector_Done();
}// Done


/*!
To resize a terminal screen, use this method.
This routine automatically sends NAWS info
(Negotiate About Window Size) if the specified
connection is an NAWS connection.

(3.0)
*/
void
ScreenManager_ChangeDimensionsOf	(TelnetScreenID const		inScreenID,
									 SInt16 const				inNewColumnCount,
									 SInt16 const				inNewRowCount,
									 Boolean const				inSendToRecordingScripts)
{
	ConnectionDataPtr		connectionDataPtr = ScreenManager_GetConnectionDataOf(inScreenID);
	
	
	if (connectionDataPtr != null)
	{
		if (VSsetlines(inScreenID, inNewRowCount) < 0)
		{
			Console_WriteLine("could not add more rows to the terminal, out of memory");
			//Errors_OutOfMemory(kInternalTelnetMemoryErrorVirtualScreenResize);
		}
		else
		{
			RealScreen_NewWidth(inScreenID, inNewColumnCount);
			if (connectionDataPtr->naws)
			{
				Parse_NegotiateAboutWindowSize(connectionDataPtr, inNewColumnCount, inNewRowCount);
			}
		}
	}
	if (inSendToRecordingScripts) SizeDialog_SendRecordableDimensionChangeEvents(inNewColumnCount, inNewRowCount);
}// ChangeDimensionsOf


/*!
Determines the screen ID corresponding to a scroll bar, and
returns 1 or 2 depending on whether, respectively, a vertical
or horizontal scroll bar is given.  If no screen can be found,
-1 is returned and "outScreenID" is undefined.

(2.6)
*/
SInt16
ScreenManager_FindScrollBar		(ControlRef			inControl,
								 TelnetScreenID*	outScreenID)
{
	SInt16		result = -1;
	
	
	if (outScreenID != null)
	{
		*outScreenID = -1;
		ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagAll, findVerticalScrollBarScreenOp,
										REINTERPRET_CAST(inControl, SInt32), 0L/* undefined */, outScreenID);
		if (validID(*outScreenID)) result = 1;
		else
		{
			ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagAll, findHorizontalScrollBarScreenOp,
											REINTERPRET_CAST(inControl, SInt32), 0L/* undefined */, outScreenID);
			if (validID(*outScreenID)) result = 2;
		}
		unless (validID(*outScreenID)) result = -1;
	}
	
	return result;
}// FindScrollBar


/*!
When you need to perform an operation on a select
subset of all screens, use this method.  Screen IDs
can be used to obtain a wealth of information about
a screen (including pointers to other data
structures), and to manipulate those screens (for
example, hiding a screens window).

The filter flags indicate which screens you wish to
operate upon.  Often, you pass
"kScreenManager_FilterFlagAllConnections" to
indicate that you only wish to work with connection
windows (not, for example, other screens such as the
FTP log).  This simple but powerful specification
allows you to affect specific subsets of the list of
screens, providing flexibility without exposing the
nature of the screen list.

You also provide a routine that operates on each
screen from the subset you indicate.  Your routine
defines what "inoutResultPtr" is, and the meaning of
"inData1" and "inData2".

(3.0)
*/
void
ScreenManager_ForEachScreenDo		(ScreenManagerFilterFlags	inFilterFlags,
									 ScreenOperationProcPtr		inProcPtr,
									 SInt32 const				inData1,
									 SInt32 const				inData2,
									 void*						inoutResultPtr)
{
	ChainLinkRef		node = null;
	ScreenDataPtr		screenDataPtr = null;
	
	
	// traverse the list
	node = Chain_GetFirstNode(gScreenDataList);
	do
	{
		screenDataPtr = Chain_GetNodeDataPtr(gScreenDataList, node);
		if (screenDataPtr != null) if (screenDataPtr->connectionDataPtr != null)
		{
			ConnectionStatus	status = screenDataPtr->connectionDataPtr->status;
			Boolean				doInvoke = false;
			
			
			// Determine if the current screen index meets one or more of the
			// specifications given by the flags.  Most of the specifications
			// require a specific connection status.  Currently, the only non-
			// session window is the FTP log, defined to have an ID of 0.
			doInvoke = /*1*/(((inFilterFlags & kScreenManager_FilterFlagWaitingSessions) == kScreenManager_FilterFlagWaitingSessions) &&
							(status == kConnectionStatusDNRWait)) ||
						/*2*/(((inFilterFlags & kScreenManager_FilterFlagOpeningSessions) == kScreenManager_FilterFlagOpeningSessions) &&
							(status == kConnectionStatusOpening)) ||
						/*3*/(((inFilterFlags & kScreenManager_FilterFlagActiveSessions) == kScreenManager_FilterFlagActiveSessions) &&
							(status == kConnectionStatusActive)) ||
						/*4*/(((inFilterFlags & kScreenManager_FilterFlagDeadSessions) == kScreenManager_FilterFlagDeadSessions) &&
							(status == kConnectionStatusDead)) ||
						/*5*/(((inFilterFlags & kScreenManager_FilterFlagNotInUseSessions) == kScreenManager_FilterFlagNotInUseSessions) &&
							(status == kConnectionStatusNotInUse)) ||
						/*6*/(((inFilterFlags & kScreenManager_FilterFlagNonSessions) == kScreenManager_FilterFlagNonSessions) &&
							(screenDataPtr->connectionDataPtr->vs == 0));
			
			if (doInvoke) InvokeScreenOperationProc(inProcPtr, screenDataPtr->connectionDataPtr, screenDataPtr->realScreenDataPtr,
													inData1, inData2, inoutResultPtr);
		}
		node = Chain_GetNextNode(gScreenDataList, node);
	} while (node != null);
}// ForEachScreenDo


/*!
To obtain a pointer to connection data for the
connection with the given local port number,
call this method.

(3.0)
*/
ConnectionDataPtr
ScreenManager_GetConnectionDataFromPort		(NetworkPortRef const		inPortNumber)
{
	ConnectionDataPtr	result = null;
	TelnetScreenID		screenID = -1;
	
	
	ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagActiveSessions | kScreenManager_FilterFlagOpeningSessions,
									findPortScreenOp, STATIC_CAST(inPortNumber, SInt32), 0L/* undefined */, &screenID);
	result = ScreenManager_GetConnectionDataOf(screenID);
	return result;
}// GetConnectionDataFromPort


/*!
To obtain a pointer to connection data for a
connection with a particular index (ID), call
this method.

The first time connection data is referenced
using a particular ID, it is created and
initialized.

(3.0)
*/
ConnectionDataPtr
ScreenManager_GetConnectionDataOf		(TelnetScreenID const		inScreenID)
{
	ConnectionDataPtr		result = null;
	ScreenDataPtr			screenDataPtr = getScreenDataOf(inScreenID);
	
	
	if (screenDataPtr != null) result = screenDataPtr->connectionDataPtr;
	return result;
}// GetConnectionDataOf


/*!
To determine the connection status for a particular
screen knowing only its ID number, call this method.

A "kConnectionStatus" constant is returned that
describes the state of the connection.

IMPORTANT:	Routines of this nature tend to be
			expensive because they are specialized.
			If you already have access to the
			connection data for the specified ID,
			consider checking the status directly
			instead of using this routine.
			
			Eventually, the code requiring these
			routines will be analyzed and new
			functions that implement *behaviors*
			will be added so that direct access to
			data is not important.

(3.0)
*/
ConnectionStatus
ScreenManager_GetConnectionStatusOf		(TelnetScreenID const		inScreenID)
{
	ConnectionStatus		result = kConnectionStatusDead;
	ConnectionDataPtr		connectionDataPtr = ScreenManager_GetConnectionDataOf(inScreenID);
	
	
	if (connectionDataPtr != null) result = connectionDataPtr->status;
	return result;
}// GetConnectionStatusOf


/*!
To determine the emulation type for a particular
screen knowing only its ID number, call this method.

A "kConnectionEmulationType" constant is returned
that describes the emulation type.  As of version
3.0 of MacTelnet, only VT100 and VT220 are supported.

IMPORTANT:	Routines of this nature tend to be
			expensive because they are specialized.
			If you already have access to the
			connection data for the specified ID,
			consider checking the terminal emulation
			type directly, instead of using this
			routine.
			
			Eventually, the code requiring these
			routines will be analyzed and new
			functions that implement *behaviors*
			will be added so that direct access to
			data is not important.

(3.0)
*/
SInt16
ScreenManager_GetEmulationTypeOf	(TelnetScreenID const		inScreenID)
{
	SInt16					result = 0;
	ConnectionDataPtr		connectionDataPtr = ScreenManager_GetConnectionDataOf(inScreenID);
	
	
	if (connectionDataPtr != null) result = connectionDataPtr->terminal.emulation;
	return result;
}// GetEmulationTypeOf


/*!
To determine the number of screens which are
currently obscured, use this method.  This value
is changed with ScreenManager_SetHiddenCount(),
presumably when the user chooses one of the
Hide items from the Window menu.

(3.0)
*/
UInt16
ScreenManager_GetHiddenCount	(void)
{
	return gScreenHiddenCount;
}// GetHiddenCount


/*!
To determine the number of screens ScreenManager_Init()
was able to allocate, use this method.

(3.0)
*/
UInt16
ScreenManager_GetMaximumScreenCount	(void)
{
	return 20; // this API doesnt really apply anymore
}// GetMaximumScreenCount


/*!
To determine the port number for a particular screen
knowing only its ID number, call this method.

(3.0)
*/
NetworkPortRef
ScreenManager_GetPortNumberOf	(TelnetScreenID const		inScreenID)
{
	SInt16				result = -1;
	ConnectionDataPtr	connectionDataPtr = ScreenManager_GetConnectionDataOf(inScreenID);
	
	
	if (connectionDataPtr != null) result = connectionDataPtr->port;
	return result;
}// GetPortNumberOf


/*!
To gain access to the real screen data of the
specified screen, call this method.

The first time real screen data is referenced
using a particular ID, it is created.

(3.0)
*/
RealScreenDataPtr
ScreenManager_GetRealScreenDataOf		(TelnetScreenID const		inScreenID)
{
	RealScreenDataPtr		result = null;
	ScreenDataPtr			screenDataPtr = getScreenDataOf(inScreenID);
	
	
	if (screenDataPtr != null) result = screenDataPtr->realScreenDataPtr;
	return result;
}// GetRealScreenDataOf


/*!
To determine the FTP client response count for
a screen, call this method.

(3.0)
*/
short
ScreenManager_GetResponseCountOf	(TelnetScreenID const		inNewID)
{
	ConnectionDataPtr		connectionDataPtr = ScreenManager_GetConnectionDataOf(inNewID);
	SInt16					result = 0;
	
	
	if (connectionDataPtr != null) result = connectionDataPtr->responseCount;
	return result;
}// GetResponseCountOf


/*!
To traverse all connections and count the number
of connections with the specified status, use
this method.

(3.0)
*/
UInt16
ScreenManager_GetStatusCount		(ConnectionStatus const		inStatusToCheckFor)
{
	UInt16				result = 0;
	ChainLinkRef		node = null;
	ScreenDataPtr		screenDataPtr = null;
	
	
	node = Chain_GetFirstNode(gScreenDataList);
	do
	{
		screenDataPtr = Chain_GetNodeDataPtr(gScreenDataList, node);
		if (screenDataPtr != null) if (screenDataPtr->connectionDataPtr != null)
		{
			if (screenDataPtr->connectionDataPtr->status == inStatusToCheckFor) result++;
		}
		node = Chain_GetNextNode(gScreenDataList, node);
	} while (node != null);
	
	return result;
}// GetStatusCount


/*!
To acquire the connection data associated with a
Mac OS window, use this method.  If the specified
window is not a screen window, null is returned.

(2.6)
*/
ConnectionDataPtr
ScreenManager_GetWindowConnectionData		(WindowRef		inWindowRef)
{
	ConnectionDataPtr		result = null;
	
	
	if (inWindowRef != null)
	{
		TelnetScreenID			screenID = ScreenManager_GetWindowScreenID(inWindowRef);
		
		
		if (validID(screenID)) result = ScreenManager_GetConnectionDataOf(screenID);
	}
	return result;
}// GetWindowConnectionData


/*!
To determine the Mac OS window pointer that
corresponds to a connection, call this method.

(3.0)
*/
WindowRef
ScreenManager_GetWindowOf	(TelnetScreenID const		inScreenID)
{
	RealScreenDataPtr	realScreenPtr = ScreenManager_GetRealScreenDataOf(inScreenID);
	WindowRef			result = null;
	
	
	if (realScreenPtr != null) result = realScreenPtr->window;
	return result;
}// GetWindowOf


/*!
To acquire the real screen data associated with a
Mac OS window, use this method.  If the specified
window is not a screen window, null is returned.

(2.6)
*/
RealScreenDataPtr
ScreenManager_GetWindowRealScreenData		(WindowRef		inWindowRef)
{
	RealScreenDataPtr		result = null;
	
	
	if (inWindowRef != null)
	{
		TelnetScreenID		screenID = ScreenManager_GetWindowScreenID(inWindowRef);
		
		
		if (validID(screenID)) result = ScreenManager_GetRealScreenDataOf(screenID);
	}
	return result;
}// GetWindowRealScreenData


/*!
To determine the screen ID associated with a
Mac OS window, use this method.  If the specified
window is not a screen window, -1 is returned.

(2.6)
*/
TelnetScreenID
ScreenManager_GetWindowScreenID		(WindowRef		inWindowRef)
{
	TelnetScreenID		result = -1;
	
	
	ScreenManager_ForEachScreenDo(kScreenManager_FilterFlagAll, findWindowScreenOp, REINTERPRET_CAST(inWindowRef, SInt32),
									0L/* undefined */, &result);
	return result;
}// GetWindowScreenID


/*!
To determine if a Mac OS window is a terminal screen,
use this method.  Currently, a terminal screen is any
window that is a session window or the FTP log.

(3.0)
*/
Boolean
ScreenManager_IsWindowAScreen		(WindowRef		inWindowRef)
{
	Boolean		result = false;
	
	
	#if 1
	if (inWindowRef != null)
	{
		TelnetWindowKind	kind = GetWindowKind(inWindowRef);
		
		
		result = ((kind == WIN_CONSOLE) || (kind == WIN_LOG) || (kind == WIN_CNXN));
		if (result) result = (ScreenManager_GetWindowScreenID(inWindowRef) >= 0); // consider windows with invalid IDs to be non-screens
	}
	#else
	{
		WindowInfoRef		windowInfo = WindowInfo_GetFromWindow(inWindowRef);
		
		
		result = (WindowInfo_GetWindowDescriptor(windowInfo) == kTelnetWindowDescriptorAnyTerminal);
	}
	#endif
	
	return result;
}// IsWindowAScreen


/*!
To destroy a screen entirely, including all internal
connection data, use this method.  If any problems
occur, this routine silently fails.

(3.0)
*/
void
ScreenManager_Remove	(ConnectionDataPtr const	inConnectionDataPtr)
{
	ScreenDataPtr	screenDataPtr = null;
	ChainLinkRef	node = null;
	Boolean			updateVS = false;
	
	
	node = Chain_GetFirstNode(gScreenDataList);
	do
	{
		screenDataPtr = Chain_GetNodeDataPtr(gScreenDataList, node);
		if (screenDataPtr != null) if (screenDataPtr->connectionDataPtr != null)
		{
			if (screenDataPtr->connectionDataPtr == inConnectionDataPtr)
			{
				// destroy the screen data completely
				Memory_DisposePtr((Ptr*)&screenDataPtr->connectionDataPtr);
				Memory_DisposePtr((Ptr*)&screenDataPtr->realScreenDataPtr);
				Memory_DisposePtr((Ptr*)&screenDataPtr);
				Chain_DisposeNode(gScreenDataList, &node);
				if (Connections_GetCount() > 0) Connections_SetCount(Connections_GetCount() - 1); // there are now fewer windows
				
				updateVS = true; // for the remaining screens, decrement their virtual screen ID
			}
			else if (updateVS) --(screenDataPtr->connectionDataPtr->vs);
		}
		node = Chain_GetNextNode(gScreenDataList, node);
	} while (node != null);
}// Remove


/*!
Sends some data to the host along the specified screens
connection.

(2.6)
*/
void
ScreenManager_SendData	(TelnetScreenID const		inScreenID,
						 char*						inDataPtr,
						 Size const					inDataSize)
{
	short		port = ScreenManager_GetPortNumberOf(inScreenID);
	
	
	Network_SendFlushData(port, inDataPtr, inDataSize);
}// SendData


/*!
To specify the number of screens which are
currently obscured, use this method.  This
should only be done to implement one of the
Hide or Show items in the Window menu.

(3.0)
*/
void
ScreenManager_SetHiddenCount	(UInt16 const		inNewCount)
{
	gScreenHiddenCount = inNewCount;
}// SetHiddenCount


/*!
To specify the FTP client response count for
a screen, call this method.

(3.0)
*/
void
ScreenManager_SetResponseCountOf	(TelnetScreenID const		inNewID,
									 short const				inNewCount)
{
	ConnectionDataPtr		connectionDataPtr = ScreenManager_GetConnectionDataOf(inNewID);
	
	
	if (connectionDataPtr != null) connectionDataPtr->responseCount = inNewCount;
}// SetResponseCountOf


/*!
When ANSI colors have changed, use this routine to
update each windows palette.  Note that since
MacTelnet 3.0 now uses a modeless window to change
ANSI colors, it might be a better idea to develop
code that changes only a single color entry in
each window (since the user can usually only change
one color at a time); it is overkill to overwrite
all colors when most of them wont have changed.

(2.6)
*/
void
ScreenManager_UpdatePalettes	(void)
{
	GrafPtr				oldPort = null;
	TelnetScreenID		screenIndex = 0;
	
	
	GetPort(&oldPort);
	for (screenIndex = 1; screenIndex <= Connections_GetCount(); screenIndex++)
	{
		if ((ScreenManager_GetConnectionStatusOf(screenIndex) == kConnectionStatusActive) ||
			(ScreenManager_GetConnectionStatusOf(screenIndex) == kConnectionStatusOpening))
		{
			RealScreenDataPtr	realScreenPtr = ScreenManager_GetRealScreenDataOf(screenIndex);
			
			
			if (realScreenPtr->areANSIColorsEnabled)
			{
				register SInt16		i = 0;
				RGBColor			color;
				
				
				SetPortWindowPort(realScreenPtr->window);
				
				// change all window palettes to include the updated ANSI colors
				for (i = 0; i < kANSIColorCount; i++)
				{
					Preferences_GetANSIColor(i, &color);
					SetEntryColor(realScreenPtr->paletteTerminal, i + kRealScreenColorCountNonANSIColors, &color);
					MacLook_InvalidatePortRectangle(realScreenPtr->window);
				}
			}
		}
	}
	SetPort(oldPort);
}// UpdatePalettes


//
// internal methods
//

/*!
This method is of standard "ScreenOperationProcPtr" form.
It determines whether the specified screens horizontal
scroll bar is the same as the control specified by
"inData1".  The value of "inData2" is not defined, and
should be 0.

You should provide storage for a "TelnetScreenID", such
that the actual type of "inoutTelnetScreenIDPtr" is a
pointer to a "TelnetScreenID".

(3.0)
*/
static void
findHorizontalScrollBarScreenOp		(ConnectionDataPtr const	inConnectionDataPtr,
									 RealScreenDataPtr const	inRealScreenDataPtr,
									 SInt32 const				inControlRefToFind,
									 SInt32 const				UNUSED_ARGUMENT(inData2),
									 void*						inoutTelnetScreenIDPtr)
{
	if ((inoutTelnetScreenIDPtr != null) && (inRealScreenDataPtr != null) && (inConnectionDataPtr != null))
	{
		if (inRealScreenDataPtr->left == ((ControlRef)inControlRefToFind))
		{
			*((TelnetScreenID*)inoutTelnetScreenIDPtr) = inConnectionDataPtr->vs;
		}
	}
}// findHorizontalScrollBarScreenOp


/*!
This method is of standard "ScreenOperationProcPtr" form.
It determines whether the specified screens local port
number is the same as the port specified by "inData1".
The value of "inData2" is not defined, and should be 0.

You should provide storage for a "TelnetScreenID", such
that the actual type of "inoutTelnetScreenIDPtr" is a
pointer to a "TelnetScreenID".

(3.0)
*/
static void
findPortScreenOp	(ConnectionDataPtr const	inConnectionDataPtr,
					 RealScreenDataPtr const	UNUSED_ARGUMENT(inRealScreenDataPtr),
					 SInt32 const				inPortNumberToFind,
					 SInt32 const				UNUSED_ARGUMENT(inData2),
					 void*						inoutTelnetScreenIDPtr)
{
	if ((inoutTelnetScreenIDPtr != null) && (inConnectionDataPtr != null))
	{
		if (inConnectionDataPtr->port == inPortNumberToFind)
		{
			*((TelnetScreenID*)inoutTelnetScreenIDPtr) = inConnectionDataPtr->vs;
		}
	}
}// findPortScreenOp


/*!
This method is of standard "ScreenOperationProcPtr" form.
It determines whether the specified screens vertical
scroll bar is the same as the control specified by
"inData1".  The value of "inData2" is not defined, and
should be 0.

You should provide storage for a "TelnetScreenID", such
that the actual type of "inoutTelnetScreenIDPtr" is a
pointer to a "TelnetScreenID".

(3.0)
*/
static void
findVerticalScrollBarScreenOp		(ConnectionDataPtr const	inConnectionDataPtr,
									 RealScreenDataPtr const	inRealScreenDataPtr,
									 SInt32 const				inControlRefToFind,
									 SInt32 const				UNUSED_ARGUMENT(inData2),
									 void*						inoutTelnetScreenIDPtr)
{
	if ((inoutTelnetScreenIDPtr != null) && (inRealScreenDataPtr != null) && (inConnectionDataPtr != null))
	{
		if (inRealScreenDataPtr->scroll == ((ControlRef)inControlRefToFind))
		{
			*((TelnetScreenID*)inoutTelnetScreenIDPtr) = inConnectionDataPtr->vs;
		}
	}
}// findVerticalScrollBarScreenOp


/*!
This method is of standard "ScreenOperationProcPtr" form.
It determines whether the specified screens window is the
same as the window specified by "inData1".  The value of
"inData2" is not defined, and should be 0.

You should provide storage for a "TelnetScreenID", such
that the actual type of "inoutTelnetScreenIDPtr" is a
pointer to a "TelnetScreenID".

(3.0)
*/
static void
findWindowScreenOp		(ConnectionDataPtr const	inConnectionDataPtr,
						 RealScreenDataPtr const	inRealScreenDataPtr,
						 SInt32 const				inWindowRefToFind,
						 SInt32 const				UNUSED_ARGUMENT(inData2),
						 void*						inoutTelnetScreenIDPtr)
{
	if ((inoutTelnetScreenIDPtr != null) && (inRealScreenDataPtr != null) && (inConnectionDataPtr != null))
	{
		if (inRealScreenDataPtr->window == ((WindowRef)inWindowRefToFind))
		{
			*((TelnetScreenID*)inoutTelnetScreenIDPtr) = inConnectionDataPtr->vs;
		}
	}
}// findWindowScreenOp


/*!
To obtain a pointer to screen data for a
connection with a particular index (ID), use
this method.

The first time screen data is referenced
using a particular ID, it is created.

(3.0)
*/
static ScreenDataPtr
getScreenDataOf		(TelnetScreenID const		inScreenID)
{
	ScreenDataPtr	result = null;
	ChainLinkRef	node = null;
	Boolean			exists = false;
	
	
	node = Chain_GetFirstNode(gScreenDataList);
	do
	{
		result = Chain_GetNodeDataPtr(gScreenDataList, node);
		if (result != null) if (result->connectionDataPtr != null)
		{
			if (result->connectionDataPtr->vs == inScreenID) exists = true;
		}
		node = Chain_GetNextNode(gScreenDataList, node);
	} while ((!exists) && (node != null));
	
	unless (exists)
	{
		// this screen ID has never been referenced before - construct data for it
		result = (ScreenDataPtr)Memory_NewPtr(sizeof(ScreenData));
		if (result != null)
		{
			result->connectionDataPtr = (ConnectionDataPtr)Memory_NewPtr(sizeof(ConnectionData));
			if (result->connectionDataPtr != null)
			{
				result->realScreenDataPtr = (RealScreenDataPtr)Memory_NewPtr(sizeof(RealScreenData));
				if (result->realScreenDataPtr != null)
				{
					ChainLinkRef	node = null;
					
					
					// both structures created successfully - initialize everything and add it to the list
					result->connectionDataPtr->vs = inScreenID;
					result->connectionDataPtr->status = kConnectionStatusNotInUse;
					node = Chain_NewNode(gScreenDataList, result);
					Chain_MoveNode(gScreenDataList, node, null/* means end of list */, kChainLinkPlacementInsertAfter);
				}
			}
		}
	}
	return result;
}// getScreenDataOf


/*!
This method is called in order to guarantee that
a specified MacTelnet connection ID corresponds to
a real connection.

(3.0)
*/
static Boolean
validID		(TelnetScreenID const		inScreenID)
{
	Boolean		result = false;
	
	
#if 0
	result = ((inScreenID >= 0) && (inScreenID <= Connections_GetCount()));
#else
	ChainLinkRef		node = null;
	ScreenDataPtr		screenDataPtr = null;
	
	
	node = Chain_GetFirstNode(gScreenDataList);
	do
	{
		screenDataPtr = Chain_GetNodeDataPtr(gScreenDataList, node);
		if (screenDataPtr != null) if (screenDataPtr->connectionDataPtr != null)
		{
			if (screenDataPtr->connectionDataPtr->vs == inScreenID) result = true;
		}
		node = Chain_GetNextNode(gScreenDataList, node);
	} while ((!result) && (node != null));
#endif
	return result;
}// validID