/*###############################################################

	ScreenManager.h
	
	The new real screen management system in Telnet 3.0.  Some
	accessor functions are available only to ease the transition
	to a more opaque data model: these obvious calls (which use
	"ID" values) are short-lived, and you should NOT use them
	unless absolutely necessary.  Also, dont use any function
	that accesses the current connection, if at all possible;
	instead, use a function that obtains data from a window, and
	provide a reference to the real frontmost window (by calling
	GetRealFrontWindow()).
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#ifndef __SCREENMANAGER__
#define __SCREENMANAGER__

// Mac includes
#include <MacTypes.h>
#include <MacWindows.h>

// MacTelnet includes
#include "ConnectionData.h"
#include "RealScreenData.h"



// constants
typedef UInt32 ScreenManagerFilterFlags;
enum
{
	// combine filter flags to specify the screens in a list that an operation should affect
	kScreenManager_FilterFlagNonSessions = (1 << 5), // terminal windows that arent connections (such as the FTP log)
	kScreenManager_FilterFlagDeadSessions = (1 << 4), // disconnected sessions whose windows are still open
	kScreenManager_FilterFlagActiveSessions = (1 << 3), // connected sessions
	kScreenManager_FilterFlagOpeningSessions = (1 << 2), // sessions that are post-lookup but not yet connected
	kScreenManager_FilterFlagWaitingSessions = (1 << 1), // sessions that are awaiting a DNR lookup
	kScreenManager_FilterFlagNotInUseSessions = (1 << 0),
	// use these convenient flags to group the above flags in consistent ways
	kScreenManager_FilterFlagInUseSessions =	kScreenManager_FilterFlagDeadSessions |
												kScreenManager_FilterFlagActiveSessions |
												kScreenManager_FilterFlagOpeningSessions |
												kScreenManager_FilterFlagWaitingSessions,
	kScreenManager_FilterFlagAllSessions =	kScreenManager_FilterFlagNotInUseSessions |
											kScreenManager_FilterFlagInUseSessions,
	kScreenManager_FilterFlagAll =	kScreenManager_FilterFlagNonSessions |
									kScreenManager_FilterFlagAllSessions
};



// types
typedef SInt16 TelnetScreenID;



// universal procedure pointers
/*!
Screen Operation Routine

This procedure type defines a convenient function that,
among other things, can be used as an iterator of all
screens.  A wealth of useful information can be found
using only the provided parameters, shielding code in
the rest of MacTelnet from the actual screen list
implementation.
*/
typedef void (*ScreenOperationProcPtr)	(ConnectionDataPtr const	inConnectionDataPtr,
										 RealScreenDataPtr const	inRealScreenDataPtr,
										 SInt32 const				inData1,
										 SInt32 const				inData2,
										 void*						inoutResultPtr);
#define InvokeScreenOperationProc(userRoutine, inConnectionDataPtr, inRealScreenDataPtr, inData1, inData2, inoutResultPtr) \
			(*userRoutine)((inConnectionDataPtr), (inRealScreenDataPtr), (inData1), (inData2), (inoutResultPtr))



//
// public methods
//

/*###############################################################
	INITIALIZING AND FINISHING WITH THE SCREEN MANAGER
###############################################################*/

void
	ScreenManager_Init						(void);

void
	ScreenManager_Done						(void);

/*###############################################################
	GLOBAL SETTINGS
###############################################################*/

UInt16
	ScreenManager_GetHiddenCount			(void);

UInt16
	ScreenManager_GetMaximumScreenCount		(void);

UInt16
	ScreenManager_GetStatusCount			(ConnectionStatus const		inStatusToCheckFor);

void
	ScreenManager_SetHiddenCount			(UInt16 const				inNewCount);

/*###############################################################
	ITERATING THROUGH SCREENS
###############################################################*/

SInt16
	ScreenManager_FindScrollBar				(ControlRef					inControl,
											 TelnetScreenID*			outScreenID);

void
	ScreenManager_ForEachScreenDo			(ScreenManagerFilterFlags	inFilterFlags,
											 ScreenOperationProcPtr		inProcPtr,
											 SInt32 const				inData1,
											 SInt32 const				inData2,
											 void*						inoutResultPtr);

void
	ScreenManager_UpdatePalettes			(void);

/*###############################################################
	ACCESSING SCREENS BY VIRTUAL SCREEN NUMBER
###############################################################*/

void
	ScreenManager_ChangeDimensionsOf		(TelnetScreenID const		inScreenID,
											 SInt16 const				inNewColumnCount,
											 SInt16 const				inNewRowCount,
											 Boolean const				inSendToRecordingScripts);

ConnectionDataPtr
	ScreenManager_GetConnectionDataOf		(TelnetScreenID const		inScreenID);

ConnectionStatus
	ScreenManager_GetConnectionStatusOf		(TelnetScreenID const		inScreenID);

SInt16
	ScreenManager_GetEmulationTypeOf		(TelnetScreenID const		inScreenID);

NetworkPortRef
	ScreenManager_GetPortNumberOf			(TelnetScreenID const		inScreenID);

RealScreenDataPtr
	ScreenManager_GetRealScreenDataOf		(TelnetScreenID const		inScreenID);

short
	ScreenManager_GetResponseCountOf		(TelnetScreenID const		inScreenID);

WindowRef
	ScreenManager_GetWindowOf				(TelnetScreenID const		inScreenID);

void
	ScreenManager_SetResponseCountOf		(TelnetScreenID const		inScreenID,
											 short const				inNewCount);

/*###############################################################
	ACCESSING SCREENS BY OTHER CRITERIA
###############################################################*/

ConnectionDataPtr
	ScreenManager_GetConnectionDataFromPort	(NetworkPortRef const		inPortNumber);

void
	ScreenManager_Remove					(ConnectionDataPtr const	inConnectionDataPtr);

void
	ScreenManager_SendData					(TelnetScreenID const		inScreenID,
											 char*						inDataPtr,
											 Size const					inDataSize);

/*###############################################################
	SCREEN WINDOW UTILITIES
###############################################################*/

ConnectionDataPtr
	ScreenManager_GetWindowConnectionData	(WindowRef					inWindowRef);

RealScreenDataPtr
	ScreenManager_GetWindowRealScreenData	(WindowRef					inWindowRef);

TelnetScreenID
	ScreenManager_GetWindowScreenID			(WindowRef					inWindowRef);

// IMPORTANT - YOU SHOULD USE THIS FUNCTION TO GUARANTEE A WINDOW CONTAINS A TERMINAL SCREEN, BEFORE MUCKING WITH IT
Boolean
	ScreenManager_IsWindowAScreen			(WindowRef					inWindowRef);

#endif