/*###############################################################

	Sets.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <ctype.h>
#include <stdio.h>
#include <string.h>

// Mac includes
#include <MacErrors.h>
#include <Navigation.h>
#include <StandardFile.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "GeneralResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "Console.h"
#include "DialogUtilities.h"
#include "EventLoop.h"
#include "FileUtilities.h"
#include "MacLook.h"
#include "MacroSetupDialog.h"
#include "ConnectionData.h"
#include "Sets.proto.h"
#include "telneterrors.h"
#include "Preferences.h"
#include "Connections.h"
#include "RealScreen.h"
#include "TelnetMenus.h"
#include "Terminal.h"
#include "vsdata.h"



// variables
ConnectionSetup**	SetParams = null;
SessionPrefs*		SetSessionPtr = null;
TerminalPrefs*		SetTerminalPtr = null;
char*				Cspace = null;
static short		CONFstate = 0;
static short		CONFactive = 0;
static short		gInQuote = 0;
static short		gPosition = 0;



// internal methods
static char		getMacFileChar			(short				inFileID);
static void		nameValuePairToFile		(const char*		inLabel,
										 const char*		inValue,
										 const SInt16		inFileID);



//
// public methods
//

/**
Displays a file selection dialog asking the user to
open a session description file.  In Telnet 3.0,
this method responds by sending an open documents
Apple Event back to Telnet (making it recordable),
instead of directly parsing the file specification
of the chosen file.

(2.6)
*/
void
Sets_Load	(void)
{
	Boolean		good = false;
	FSSpec		set;
	Str255		prompt,
				title;
	
	
	GetIndString(prompt, rStringsNavigationServices, siNavPromptLoadConfigurationSet);
	GetIndString(title, rStringsNavigationServices, siNavDialogTitleLoadSet);
	if (FlagManager_Test(kFlagNavigationServices))
	{
		FileInfo	fileInfo;
		OSType		fileType;
		OSType		types[] = { kApplicationFileTypeSessionDescription };
		OSErr		error = noErr;
		short		hit = 0;
		
		
		Alert_ReportOSErr(error = Navigation_GetFile(prompt, title,
													kApplicationCreatorSignature,
													kTelnetNavPrefKeyGenericOpenFile,
													kNavDontAddTranslateItems,
													1, types, EventLoop_HandleNavigationUpdate,
													&set, &fileType, &fileInfo));
		good = (error == noErr);
	}
	else
	{
		SFTypeList			types = { kApplicationFileTypeSessionDescription };
		StandardFileReply	sfr;
		long				junk = 0L;
		
		
		//DeactivateFrontmostWindow();
		ThemeSavvyStandardGetFile(title, prompt, (FileFilterUPP)null, 1, types, &sfr);
		RestoreFrontmostWindow();
		
		good = sfr.sfGood;
		
		if (good)
		{
			good = (FSMakeFSSpec(sfr.sfFile.vRefNum, sfr.sfFile.parID, sfr.sfFile.name, &set) == noErr);
		}
	}
	if (good) Alert_ReportOSErr(FileUtilities_OpenDocument(&set)); // 3.0 - use Apple Events
}// Load


/**
Reads the saved set file into a data structure in memory.
Everything is handled using keywords that are stored in
resources.

(2.6)
*/
void
Sets_ReadFromFile	(const FSSpec*		inFilePtr)
{
	short		c = 0,
				retval = 0;
	short		fn = 0;
	OSErr		err = noErr;
	
	
	Cspace = Memory_NewPtr(256); // get room for gathering stuff
	
	SetParams = Connections_GetDefault();
	HLockHi((Handle)SetParams);
	HLockHi((Handle)(**SetParams).session);
	SetSessionPtr = *(**SetParams).session;
	HLockHi((Handle)(**SetParams).terminal);
	SetTerminalPtr = *(**SetParams).terminal;
	
	if (null == Cspace) // no memory for the set to load into
	{
		Errors_DisplayStopMessage(rStringsMemoryErrors, siMemErrorCannotLoadSet, -1);
	}
	else
	{
		gPosition = CONFstate = CONFactive = gInQuote = 0;   /* state vars */	
		
		err = HOpen(inFilePtr->vRefNum, inFilePtr->parID, inFilePtr->name, fsRdPerm, &fn);
		
		retval = 0;
		while (!retval)
		{
			c = getMacFileChar(fn);
			if (c == '#' && !gInQuote)
			{
				while (c != EOF && c != '\012' && c != '\015') // skip to EOL
					c = getMacFileChar(fn);
			}
			retval = contoken(c);
		}
		
		FSClose(fn);
		Memory_DisposePtr(&((Ptr)Cspace));
	}
}// ReadFromFile


void
Sets_Save	(void)
{
	short		fn = 0,
				truncate = 0;
	Rect		rect;
	long		junk = 0L;
	char		temp[256],
				temp2[256];
	short		fnum = 0,
				fsiz = 0;
	short		i = 0;
	FSSpec		set;
	OSErr		err = noErr;
	char		scratchString[255];
	Str255		scratchPstring,
				prompt,
				title;
	Boolean		good = false;
	
	
	GetIndString(prompt, rStringsNavigationServices, siNavPromptSaveConfigurationSet);
	GetIndString(title, rStringsNavigationServices, siNavDialogTitleSaveSet);
	if (FlagManager_Test(kFlagNavigationServices))
	{
		NavReplyRecord	reply;
		
		
		Alert_ReportOSErr(err = Navigation_PutFile(prompt, title,
							kApplicationCreatorSignature, kApplicationFileTypeSessionDescription,
							kTelnetNavPrefKeyGenericSaveFile,
							kNavDefaultNavDlogOptions | kNavDontAddTranslateItems,
							EventLoop_HandleNavigationUpdate, &reply, &set));
		good = ((err == noErr) && (reply.validRecord));
	}
	else
	{
		StandardFileReply	sfr;
		
		
		GetIndString(scratchPstring, rStringsImportantFileNames, siImportantFileNameSessionDescriptionDefault);
		DeactivateFrontmostWindow();
		ThemeSavvyStandardPutFile(title, prompt, scratchPstring, &sfr);
		RestoreFrontmostWindow();
		good = sfr.sfGood;
		if (good)
		{
			OSErr	error = noErr;
			
			
			error = FSMakeFSSpec(sfr.sfFile.vRefNum, sfr.sfFile.parID, sfr.sfFile.name, &set);
			good = ((error == noErr) || (error == fnfErr));
		}
		
	}
	
	// not exactly SESE, but simple...
	if (!good) return;
	
	if ((err = HCreate(set.vRefNum, set.parID, set.name, kApplicationCreatorSignature,
			kApplicationFileTypeSessionDescription)) == dupFNErr)
	{
		truncate = 1;
	}
		
	err = HOpen(set.vRefNum, set.parID, set.name, fsWrPerm, &fn);

	if (truncate) SetEOF(fn, 0L);
	
	// write the menus have key equivalents preference
	{
		Boolean		haveKeyEquivalents = false;
		Size		actualSize = 0L;
		
		
		unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
											kTelnetPreferenceTag_MenuItemKeys,
											sizeof(haveKeyEquivalents), &haveKeyEquivalents,
											&actualSize) == kTelnetPrefsResultCodeSuccess)
		{
			// default to true
			haveKeyEquivalents = true;
		}
		nameValuePairToFile("commandkeys", (haveKeyEquivalents) ? "yes" : "no", fn);
	}
	
	for (i = 0; i < 10; i++)
	{
		Macros_Get(Macros_GetActiveSet(), i, temp, sizeof(temp));
		if (*temp)
		{
			sprintf(temp2, "key%d = \"", i);
			CStringToFile(fn,  temp2);
			CStringToFile(fn,  temp);
			strcpy(temp2,"\"\015");
			CStringToFile(fn,  temp2);
		}
	}

#if 0
	for (i = 1; i <= Connections_GetCount(); i++)
	{
		short	j = 0;
		
		
		j = RSgetfont(i, &fnum, &fsiz);
	}
#endif

	for (i = 1; i <= Connections_GetCount(); i++)
	{
		ConnectionDataPtr		theScreen = ScreenManager_GetConnectionDataOf(i);
		RealScreenDataPtr		realScreenPtr = ScreenManager_GetRealScreenDataOf(i);
		short					theVirtualID = i;
		
		
		if ((theScreen == null) || (realScreenPtr == null))
		{
			BEEP;
			Console_WriteLine("a session description save failed because of null data!");
			continue;
		}
		
	  	GetWTitle(theScreen->window, scratchPstring);
		StringUtilities_PToC(scratchPstring, scratchString);
		sprintf(temp2, "\"%s\"", scratchString);
		nameValuePairToFile("name", temp2, fn);
		
		StringUtilities_PToC(theScreen->host.name, scratchString);
		sprintf(temp2, "\"%s\"", scratchString);
		nameValuePairToFile("host", temp2, fn);
		
		if (theScreen->ftpState != 0) nameValuePairToFile("ftp", "yes", fn);
		
		sprintf(temp2, "%d", INTEGER_ABSOLUTE(realScreenPtr->rheight - realScreenPtr->height));
		nameValuePairToFile("scrollback", temp2, fn);
		
		nameValuePairToFile("erase", (theScreen->bsdel) ? "delete" : "backspace", fn);
		
		MacLook_GetWindowRegionBounds(realScreenPtr->window, kWindowContentRgn, &rect);
		sprintf(temp2, "{%d,%d,%d,%d}", rect.top, rect.left, rect.bottom, rect.right);
		nameValuePairToFile("size", temp2, fn);
		
		sprintf(temp2, "%d", VSmaxwidth(theVirtualID) + 1);
		nameValuePairToFile("vtwidth", temp2, fn);

		nameValuePairToFile("tekclear",
							(theScreen->TEK.pageLocation == kTektronixPageLocationSameWindowClear)
								? "yes" : "no", fn);
		
		// save colors
		{
			RGBColor				color;
			register SInt16			j = 0;
			RealScreenColorIndex	indices[] =
									{
										kRealScreenColorIndexNormalText,
										kRealScreenColorIndexNormalBackground,
										kRealScreenColorIndexBlinkingText,
										kRealScreenColorIndexBlinkingBackground,
										kRealScreenColorIndexBoldText,
										kRealScreenColorIndexBoldBackground
									};
			
			
			for (j = 0; j < 6; j++)
			{
				(Boolean)RealScreen_GetColor(theVirtualID, indices[j], &color);
				sprintf(temp2, "rgb%d = {%u,%u,%u}\015", j, color.red, color.green, color.blue);
				CStringToFile(fn, temp2);
			} 
		} 
		
		RSgetfont(theVirtualID, &fnum, &fsiz);
#if TARGET_API_MAC_OS8
		GetFontName(fnum, (StringPtr)temp);
#else
		(OSStatus)FMGetFontFamilyName(fnum, (StringPtr)temp);
#endif
		StringUtilities_PToCInPlace((unsigned char*)temp);
		sprintf(temp2, "\"%s\"", temp);
		nameValuePairToFile("font", temp2, fn);
		sprintf(temp2, "%d", fsiz);
		nameValuePairToFile("fsize", temp2, fn);
		
		sprintf(temp2, "%d", VSgetlines(theVirtualID));
		nameValuePairToFile("nlines", temp2, fn);
		sprintf(temp2, "%d", theScreen->controlKey.suspend);
		nameValuePairToFile("keystop", temp2, fn);
		sprintf(temp2, "%d", theScreen->controlKey.resume);
		nameValuePairToFile("keygo", temp2, fn);
		sprintf(temp2, "%d", theScreen->controlKey.interrupt);
		nameValuePairToFile("keyip", temp2, fn);
		sprintf(temp2, "%d", theScreen->crmap);
		nameValuePairToFile("crmap", temp2, fn);
		sprintf(temp2, "%d", theScreen->TEK.mode);
		nameValuePairToFile("tekem", temp2, fn);
		if (theScreen->national)
		{
			// dont save this if the default translation table is being used
			GetMenuItemText(Menus_GetTranslationMenu(), theScreen->national + 1, scratchPstring);
			StringUtilities_PToC(scratchPstring, scratchString);
			sprintf(temp2, "\"%s\"", scratchString);
			nameValuePairToFile("translation", temp2, fn);
		}
		StringUtilities_PToC(theScreen->terminal.answerBack, scratchString);
		sprintf(temp2, "\"%s\"", scratchString);
		nameValuePairToFile("answerback", temp2, fn);
		
		nameValuePairToFile("authenticate", (theScreen->authenticate) ? "yes" : "no", fn);
		nameValuePairToFile("encrypt", (theScreen->encrypt) ? "yes" : "no", fn);
		nameValuePairToFile("pageup", (theScreen->pgupdwn) ? "yes" : "no", fn);
		nameValuePairToFile("keypad", (theScreen->keypadmap) ? "yes" : "no", fn);
	}// for i
	
	FSClose(fn);						
}// Save


/* affirmative() checks a token to see if it is a valid Affirmation string.
	We now get the affirmative strings from the resource */
	
Boolean	affirmative( char *s)
{
	short	i;
	Str255	AffWords;		/* Telnet 2.6: get string resources */

	for (i = 0; i<gPosition; i++)
				s[i] = tolower(s[i]);

	for (i = 1; i<= rStringCountAffirmativeWords ;i++) 
		{
		GetIndString(AffWords,rStringsAffirmativeWords,i);
		StringUtilities_PToCInPlace(AffWords);
		if (!ncstrcmp((char *)AffWords,s))				
		break;										
		}
		
	if (i <= rStringCountAffirmativeWords)
		return(true);
	else
		return(false);
}


/**************************************************************************/
/*  Sissep
*   is the character a valid separator for the hosts file?
*   separators are white space, special chars and :;=
*
*/
Boolean Sissep( char c)
{
	if (c < 33)
		return(1);
	if (c == ':' || c == ';' || c == '=')
		return(true);
	return(false);
}


/************************************************************************/
/*  ncstrcmp
*   No case string compare.
*   Only returns 0=match, 1=no match, does not compare greater or less
*   There is a tiny bit of overlap with the | 32 trick, but shouldn't be
*   a problem.  It causes some different symbols to match.
*/
short ncstrcmp(char *sa, char *sb)
{
	while (*sa && *sa < 33)		/* don't compare leading spaces */
		sa++;
	while (*sb && *sb < 33)
		sb++;

	while (*sa && *sb)
	{
		if ((*sa != *sb) && ((*sa | 32) != (*sb | 32)))
			return(1);
		sa++; sb++;
	}
	if (!*sa && !*sb)		/* if both at end of string */
		return(0);
	else
		return(1);
}


/* confile() now gets all of its keywords from the string resources, for greater
	ease, flexibility, and overall coolness.  */
short confile( char *s)
{
	short			i = 0;
	Boolean			success = false;
	unsigned int	a = 0,
					b = 0,
					c = 0,
					d = 0;
	int				signedint = 0;
	Str255			Ckeyw;
	char			tempCstring[256];
	
	
	if (!(*s)) return 0;

	switch( CONFstate) {
		case 0:				/* No keyword yet! */
			for (i = 0; i < gPosition; i++) s[i] = tolower(s[i]);
			for (i = 1; i <= rStringCountSaveSetKeywords; i++) 
			{
				GetIndString(Ckeyw, rStringsSaveSetKeywords, i);
				StringUtilities_PToCInPlace(Ckeyw);
				if (!ncstrcmp((char *)Ckeyw, s)) break;										
			}

			if ( i > rStringCountSaveSetKeywords ) 
			{
				OperationFailedAlert(siOpFailedBadSet, 0, 0);
				return(1);
			}

			CONFstate=i;

			if (CONFstate==5)
			{
				SetSessionPtr->bksp = 0;
				CONFstate = 0;
			}
			if (CONFstate==6)
			{
				SetSessionPtr->bksp = 1;
				CONFstate = 0;
			}
			break;

		case 1:				/* NAME */
			{
				char*	p = null;
				
				
				if (null == (p = Memory_NewPtr(40000))) 		/* is there enough memory? */
				{
					// not enough memory for the set
					Errors_DisplayStopMessage(rStringsMemoryErrors, siMemErrorCannotLoadSet, -1);
					return(-1);
				}
				else
					Memory_DisposePtr(&p);
			}
			if (CONFactive)
			{
				success = Connections_CreateOne(SetParams);
				SetParams = null;
				if (!success)
				{
					/* error in opening */
					return(42);
				}
			}

			if (SetParams == null)
			{
				SetParams = Connections_GetDefault();
				HLockHi((Handle)SetParams);
				HLockHi((Handle)(**SetParams).session);
				SetSessionPtr = *(**SetParams).session;
				HLockHi((Handle)(**SetParams).terminal);
				SetTerminalPtr = *(**SetParams).terminal;
			}
				
			strncpy(tempCstring, s, 255);			/* Move name in */
			StringUtilities_CToP(tempCstring, (**SetParams).windowName);
			CONFstate = 0;
			CONFactive = 1;
			break;

		case 2:				/* HOST */
			strncpy(tempCstring, s, 63);			/* Move name in */
			StringUtilities_CToPInPlace(tempCstring);

			//	Process the hostname string.
			{
				short		port = 0;
				
				
				if (Connections_ProcessHostNameString((StringPtr)tempCstring, &port)) SetSessionPtr->port = port;
			}
			
			BlockMove(tempCstring, SetSessionPtr->hostname, tempCstring[0] + 1);

			CONFstate = 0;
			break;

		case 3:				/* SIZE */
			if (4 != sscanf(s, "{%d,%d,%d,%d}", &a, &b, &c,&d))
			{
				/* error in window size */
				return(2);
			}
			
			// 3.0 - this is no longer used
			
			CONFstate = 0;
			break;

		case 4:
			if (1 != sscanf(s,"%d", &a))
			{
				/* scrollback needs parameter */
				return(1);
			}
			SetTerminalPtr->scrollbackBufferSize = a;
			CONFstate = 0;
			break;

		case 5:
			SetSessionPtr->bksp = 0;			
			CONFstate = 0;
			break;

		case 6:
			SetSessionPtr->bksp = 1;
			CONFstate = 0;
			break;

		case 7:
			Macros_Set(Macros_GetActiveSet(), 0, s);
			CONFstate = 0;
			break;
		case 8:
			Macros_Set(Macros_GetActiveSet(), 1, s);
			CONFstate = 0;
			break;
		case 9:
			Macros_Set(Macros_GetActiveSet(), 2, s);
			CONFstate = 0;
			break;
		case 10:
			Macros_Set(Macros_GetActiveSet(), 3, s);
			CONFstate = 0;
			break;
		case 11:
			Macros_Set(Macros_GetActiveSet(), 4, s);
			CONFstate = 0;
			break;
		case 12:
			Macros_Set(Macros_GetActiveSet(), 5, s);
			CONFstate = 0;
			break;
		case 13:
			Macros_Set(Macros_GetActiveSet(), 6, s);
			CONFstate = 0;
			break;
		case 14:
			Macros_Set(Macros_GetActiveSet(), 7, s);
			CONFstate = 0;
			break;
		case 15:
			Macros_Set(Macros_GetActiveSet(), 8, s);
			CONFstate = 0;
			break;
		case 16:
			Macros_Set(Macros_GetActiveSet(), 9, s);
			CONFstate = 0;
			break;
		case 17:
			CONFstate = 0;		// now ignored (was commandkeys)
			break;
		case 18:
			if (!strcmp(s, "backspace"))
				SetSessionPtr->bksp = 0;
			else
				SetSessionPtr->bksp = 1;
			CONFstate = 0;
			break;
		case 19:
		case 21:
			if (1 == sscanf(s, "%d", &a)) 
				SetTerminalPtr->columnCount = a;
			CONFstate = 0;
			break;
		case 20:
			if (affirmative(s))
				SetSessionPtr->tekclear = 1;
			else
				SetSessionPtr->tekclear = 0;
			CONFstate = 0;
			break;
		case 22:
			if (3 != sscanf(s, "{%u,%u,%u}", &a, &b, &c))
			{
				/* bad parameters to RGB */
				return(2);
			}
			SetTerminalPtr->foregroundNormalColor.red = a;
			SetTerminalPtr->foregroundNormalColor.green = b;
			SetTerminalPtr->foregroundNormalColor.blue = c;
			CONFstate = 0;
			break;			
		case 23:
			if (3 != sscanf(s, "{%u,%u,%u}", &a, &b, &c))
			{
				/* bad parameters to RGB */
				return(2);
			}
			SetTerminalPtr->backgroundNormalColor.red = a;
			SetTerminalPtr->backgroundNormalColor.green = b;
			SetTerminalPtr->backgroundNormalColor.blue = c;
			CONFstate = 0;
			break;			
		case 24:
			if (3 != sscanf(s, "{%u,%u,%u}", &a, &b, &c))
			{
				/* bad parameters to RGB */
				return(2);
			}
			SetTerminalPtr->foregroundBlinkingColor.red = a;
			SetTerminalPtr->foregroundBlinkingColor.green = b;
			SetTerminalPtr->foregroundBlinkingColor.blue = c;
			CONFstate = 0;
			break;			
		case 25:
			if (3 != sscanf(s, "{%u,%u,%u}", &a, &b, &c))
			{
				/* bad parameters to RGB */
				return(2);
			}
			SetTerminalPtr->backgroundBlinkingColor.red = a;
			SetTerminalPtr->backgroundBlinkingColor.green = b;
			SetTerminalPtr->backgroundBlinkingColor.blue = c;
			CONFstate = 0;
			break;
		case 26:		/* Font Name */
			strncpy(tempCstring, s, 63);			/* Move name in */
			StringUtilities_CToP(tempCstring, &(SetTerminalPtr->normalFont[0]));
			CONFstate = 0;
			break;
		case 27:		/* Font Size */
			if (1 == sscanf(s, "%d", &a))
				SetTerminalPtr->fontSize = a;
			CONFstate = 0;
			break;
		case 28:		/* number of lines to use for window's editable region */
			if (1 == sscanf(s, "%d", &a))
				SetTerminalPtr->rowCount = a;
			CONFstate = 0;
			break;
		case 29:		/* keystop, XOFF key */
			if (1 == sscanf(s, "%d", &a))
				SetSessionPtr->skey = a;
			CONFstate = 0;
			break;
		case 30:		/* keygo, XON key */
			if (1 == sscanf(s, "%d", &a))
				SetSessionPtr->qkey = a;
			CONFstate = 0;
			break;
		case 31:		/* keyip, kill key */
			if (1 == sscanf(s, "%d", &a))
				SetSessionPtr->ckey = a;
			CONFstate = 0;
			break;
		case 32:		/* cr-map */
			if ((1 == sscanf(s, "%d", &a)) && (a != 0))
				SetSessionPtr->crmap = true;
			else
				SetSessionPtr->crmap = false;
			CONFstate = 0;
			break;
		case 33:					
			if (affirmative(s))		
				SetSessionPtr->linemode = true;	
			else					
				SetSessionPtr->linemode = false;	
			CONFstate = 0;			
			break;					
		case 34:					
			if (affirmative(s))		
				SetTerminalPtr->usesEightBits = true;	
			else					
				SetTerminalPtr->usesEightBits = false;	
			CONFstate = 0;			
			break;					
		case 35:					
			(**SetParams).ftpState = 1;		
			CONFstate = 0;			
			break;					
		case 36:	// ignored
			CONFstate = 0;
			break;
		case 37:	// ignored			
			CONFstate = 0;				
			break;								
		case 38:	// translation
			strncpy((char*)SetSessionPtr->TranslationTable, s, 32);
			StringUtilities_CToPInPlace((char*)SetSessionPtr->TranslationTable);
			CONFstate = 0;
			break;
		case 39:	// tekem
			if (1 == sscanf(s, "%d", &signedint))
				SetSessionPtr->tektype = signedint;
			CONFstate = 0;
			break;
		case 40:	// answerback message
			strncpy((char*)SetTerminalPtr->answerBackMessage, s, 32);
			StringUtilities_CToPInPlace((char*)SetTerminalPtr->answerBackMessage);
			CONFstate = 0;
			break;
		case 41: //authenticate
			if (affirmative(s))		
				SetSessionPtr->authenticate = true;
			else				
				SetSessionPtr->authenticate = false;	
			CONFstate = 0;		
			break;
		case 42: //encrypt
			if (affirmative(s))		
				SetSessionPtr->encrypt = true;	
			else					
				SetSessionPtr->encrypt = false;	
			CONFstate = 0;			
			break;
		
		case 43: //pageup/down map
			if (affirmative(s))
				SetTerminalPtr->mapsPageJumpKeys = true;
			else					
				SetTerminalPtr->mapsPageJumpKeys = false;	
			break;
		case 44: //keypad map
			if (affirmative(s))
				SetTerminalPtr->remapKeypad = true;
			else					
				SetTerminalPtr->remapKeypad = false;	
			break;			
		case 45:
			if (3 != sscanf(s, "{%u,%u,%u}", &a, &b, &c))
			{
				/* bad parameters to RGB */
				return(2);
			}
			SetTerminalPtr->foregroundBoldColor.red = a;
			SetTerminalPtr->foregroundBoldColor.green = b;
			SetTerminalPtr->foregroundBoldColor.blue = c;
			CONFstate = 0;
			break;			
		case 46:
			if (3 != sscanf(s, "{%u,%u,%u}", &a, &b, &c))
			{
				/* bad parameters to RGB */
				return(2);
			}
			SetTerminalPtr->backgroundBoldColor.red = a;
			SetTerminalPtr->backgroundBoldColor.green = b;
			SetTerminalPtr->backgroundBoldColor.blue = c;
			CONFstate = 0;
			break;
		default:
			CONFstate = 0;
		}
	return(0);
  } /* confile */

/************************************************************************/
/* contoken
*  tokenize the strings which get passed to confile.
*  Handles quotes and uses separators:  <33, ;:=
*/ 
short contoken( char c)
{
	short		retval;
	Boolean		success;

	if (c == EOF) {
		Cspace[gPosition++] = '\0';
		/* EOF handler called */
		confile(Cspace);
		if (CONFactive) {
				success = Connections_CreateOne(SetParams);
				if (!success) {
					/* error in opening */
					return(42);
					}
				}
		return(-1);
	}
	
	if (!gPosition && !gInQuote && Sissep(c))
	/*if (!gPosition && Sissep(c))	*/	/* old_skip over junk before keyword */
		return(0);

	if (gInQuote || !Sissep(c)) {

		if (gPosition > 200) {
			/* out of bounds error */
			return(1);
		}
/*
*  check for quotes, a little mixed up here, could be reorganized
*/
		if (c == '"' ) {
			if (!gInQuote) {			/* beginning of quotes */
				gInQuote = 1;
				return(0);
			}
			 Cspace[gPosition++] =c;
			return(0);
		}
		else 
			{						/* include in current string */
			if (c != '\012' && c != '\015')
				{
				Cspace[gPosition++] = c;
				return(0);
				}
			}
				
		}

	if (Cspace[gPosition - 1] == '"') gPosition--;
	Cspace[gPosition++] = '\0';

	retval = confile(Cspace);			/* pass the token along */

	gPosition = 0;
	gInQuote = 0;
	Cspace[0] = '\0';

	return(retval);
}


/**
Writes the contents of a string to a file
that has an ID number in the Macintosh HFS.

(2.6)
*/
void
CStringToFile	(short				inFileID,
				 const char*		inString)
{
	long	count = strlen((char*)inString);
	
	
	(OSErr)FSWrite(inFileID, &count, inString);
}// CStringToFile


//
// internal methods
//

/**
Reads a character from a file that has
an ID in the Macintosh HFS.

(2.6)
*/
static char
getMacFileChar	(short		inFileID)
{
	OSErr			err = noErr;
	long			count = 0L;
	unsigned char	buffer;
	
	
	count = 1;
	if ((err = FSRead(inFileID, &count, &buffer)) == eofErr) buffer = EOF;
	
	return buffer;
}// getMacFileChar


/**
Writes a name-value pair to a file, on a
single line with a carriage return at the
end, and an equal sign (=) separating the
label (name) and its value.

(3.0)
*/
static void
nameValuePairToFile		(const char*		inLabel,
						 const char*		inValue,
						 const SInt16		inFileID)
{
	long	count = 0L;
	
	
	count = strlen(inLabel);
	(short)FSWrite(inFileID, &count, inLabel);
	count = 3;
	(short)FSWrite(inFileID, &count, " = ");
	count = strlen(inValue);
	(short)FSWrite(inFileID, &count, inValue);
	count = 1;
	(short)FSWrite(inFileID, &count, "\015");
}// nameValuePairToFile