/*###############################################################

	SizeDialog.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// Mac includes
#include <ControlDefinitions.h>
#include <MacErrors.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "TelnetSizeDialog.r"

// MacTelnet includes
#include "BasicTypesAE.h"
#include "Console.h"
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "DialogResources.h"
#include "EventLoop.h"
#include "FolderManager.h"
#include "InstallAE.h"
#include "InterfaceLib.h"
#include "Preferences.h"
#include "ScreenManager.h"
#include "SizeDialog.h"
#include "StringResources.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"
#include "Terminal.h"
#include "Terminology.h"



// constants
enum
{
	// the maximum width is dynamic, memory permitting
	kSizeMinimumWidth = 10,
	kSizeMinimumHeight = 10,
	kSizeMaximumHeight = 200
};

// types
struct UndoDataDimensionChanges
{
	// context data for the context ID "kUndoableContextIdentifierTerminalDimensionChanges"
	TelnetScreenID		screenID;		// which screen was resized
	SInt16				columns;		// the number of characters wide
	SInt16				rows;			// the number of lines high
};
#ifndef __cplusplus
typedef struct UndoDataDimensionChanges		UndoDataDimensionChanges;
#endif
typedef UndoDataDimensionChanges*			UndoDataDimensionChangesPtr;

// variables
static TelnetScreenID				gScreenID = 0;
static DialogPtr					gSizeDialog = null; // the Mac OS dialog pointer
static WindowInfoRef				gSizeDialogWindowInfo;

// internal methods
static pascal Boolean	handleNewSize				(const WindowRef				inWindow,
													 const SInt32					inDeltaSizeX,
													 const SInt32					inDeltaSizeY,
													 const SInt32					inData);
static void				reverseDimensionChanges		(const UndoActionInstruction	inDoWhat,
													 UndoableContextIdentifier		inContextIdentifier,
													 void*							inContextPtr);
static pascal Boolean	sizeDialogEventFilter		(DialogPtr						inDialogPtr,
													 EventRecord*					inoutEventPtr,
													 short*							outItemIndex);



//
// public methods
//

/*!
This method is used to initialize the Terminal Screen Size
dialog box.  It creates the dialog box invisibly, and uses
the specified screen ID to set up the controls in the
dialog box to use the screens dimensions.

The extended dialog resource for the preferences dialog
causes the Appearance Manager to automatically construct
an embedding hierarchy, so the dialogs root control and
container controls are supposed to be set up once
GetNewDialog() returns.

(3.0)
*/
void
SizeDialog_Init		(const TelnetScreenID		inScreenID)
{
	Cursors_DeferredUseWatch(30); // if it takes more than half a second to initialize, show the watch cursor
	
	gScreenID = inScreenID;
	GetNewDialogWithWindowInfo(kDialogIDScreenSize, &gSizeDialog, &gSizeDialogWindowInfo);
	
	if (gSizeDialog != null)
	{
		// determine if the Help button should be disabled...
		SetUpHelpButtonDialogItem(gSizeDialog, iScreenSizeAppleGuideButton);
		
		// adjust the OK and Cancel buttons so they are big enough for their titles
		SetDialogDefaultItem(gSizeDialog, iScreenSizeOKButton);
		SetDialogCancelItem(gSizeDialog, iScreenSizeCancelButton);
		(UInt16)Localization_AdjustDialogButtons(gSizeDialog, true/* OK */, true/* Cancel */);
		Localization_AdjustHelpButtonItem(gSizeDialog, iScreenSizeAppleGuideButton);
		
		// set the static text controls (auto-sizing them) and the bottom checkbox
		{
			ControlRef		control = null;
			Str255			text;
			
			
			GetDialogItemAsControl(gSizeDialog, iScreenSizeWidthLabel, &control);
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextScreenSizeWidth);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
			
			GetDialogItemAsControl(gSizeDialog, iScreenSizeWidthUnits, &control);
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextScreenSizeColumns);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
			
			GetDialogItemAsControl(gSizeDialog, iScreenSizeHeightLabel, &control);
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextScreenSizeHeight);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
			
			GetDialogItemAsControl(gSizeDialog, iScreenSizeHeightUnits, &control);
			GetIndString(text, rStringsMiscellaneousStaticText, siStaticTextScreenSizeRows);
			(UInt16)Localization_SetUpSingleLineTextControl(control, text, kThemeSystemFont);
			
			GetDialogItemAsControl(gSizeDialog, iFormatUpdateDefaultTerminalCheckBox, &control);
			GetIndString(text, rStringsMiscellaneousControlTitles, siControlTitleUpdateDefaultTerminal);
			SetControlTitle(control, text);
		}
		
		// prevent the user from typing non-digits into the text fields, and initialize the fields and arrows
		{
			ControlRef				field = null,
									littleArrows = null;
			ControlKeyFilterUPP		upp = NewControlKeyFilterUPP(NumericalLimiter);
			short					dimension = 0;
			
			
			// width controls; arrows are restricted by the valid range of input
			dimension = VSgetcols(inScreenID);
			GetDialogItemAsControl(gSizeDialog, iScreenSizeWidthField, &field);
			SetControlNumericalText(field, dimension);
			(OSErr)SetControlData(field, kControlNoPart, kControlKeyFilterTag, sizeof(upp), (Ptr)&upp);
			GetDialogItemAsControl(gSizeDialog, iScreenSizeWidthArrows, &littleArrows);
			SetControlMaximum(littleArrows, Terminal_GetAllocatedColumnCount());
			SetControlMinimum(littleArrows, kSizeMinimumWidth);
			SetControlValue(littleArrows, dimension);
			
			// height controls; arrows are restricted by the valid range of input
			dimension = VSgetlines(inScreenID);
			GetDialogItemAsControl(gSizeDialog, iScreenSizeHeightField, &field);
			SetControlNumericalText(field, dimension);
			(OSErr)SetControlData(field, kControlNoPart, kControlKeyFilterTag, sizeof(upp), (Ptr)&upp);
			GetDialogItemAsControl(gSizeDialog, iScreenSizeHeightArrows, &littleArrows);
			SetControlMaximum(littleArrows, kSizeMaximumHeight);
			SetControlMinimum(littleArrows, kSizeMinimumHeight);
			SetControlValue(littleArrows, dimension);
		}
		
		// set up the Window Info information
		{
			WindowResizeResponderProcPtr	resizeProc = handleNewSize;
			
			
			WindowInfo_SetWindowDescriptor(gSizeDialogWindowInfo, kTelnetDialogDescriptorScreenSize);
			WindowInfo_SetWindowResizeLimits(gSizeDialogWindowInfo,
													SIZE_DIALOG_HT/* minimum height */,
													SIZE_DIALOG_WD/* minimum width */,
													SIZE_DIALOG_HT/* maximum height */,
													SIZE_DIALOG_WD + 50/* maximum width */);
			WindowInfo_SetWindowResizeResponderUPP(gSizeDialogWindowInfo, resizeProc, 0L);
		}
		WindowInfo_SetForDialog(gSizeDialog, gSizeDialogWindowInfo);
	}
}// Init


/*!
Call this method to destroy the screen size dialog
box and its associated data structures.

(3.0)
*/
void
SizeDialog_Done	(void)
{
	WindowInfo_Dispose(gSizeDialogWindowInfo);
	DisposeDialog(gSizeDialog), gSizeDialog = null;
}// Done


/*!
This method displays and handles events in the
Terminal Screen Size dialog box.  If the user
clicks OK, "true" is returned; otherwise,
"false" is returned.

(3.0)
*/
Boolean
SizeDialog_Display	(void)
{
	Boolean		result = false;
	
	
	if (gSizeDialog != null)
	{
		register DialogItemIndex	i = 0;
		DialogItemIndex				itemIndex = 0;
		ModalFilterUPP				filterProc = NewModalFilterUPP(sizeDialogEventFilter);
		
		
		// define this only for debugging; this lets you dump the dialogs
		// control embedding hierarchy to a file, before it is displayed
		#if 0
		DebugSelectControlHierarchyDumpFile(GetDialogWindow(gSizeDialog));
		#endif
		
		// display the dialog and focus the first text field
		TransitionDialogDisplay(gSizeDialog);
		{
			ControlRef		widthField = null;
			
			
			(OSErr)GetDialogItemAsControl(gSizeDialog, iScreenSizeWidthField, &widthField);
			(OSErr)SetKeyboardFocus(GetDialogWindow(gSizeDialog), widthField, kControlEditTextPart);
		}
		
		// handle events
		Menus_EnableEditMenu();
		itemIndex = 0;
		do
		{
			ModalDialog(filterProc, &itemIndex);
			switch (itemIndex)
			{
				case iScreenSizeAppleGuideButton:
					{
						Str255		searchString;
						
						
						GetIndString(searchString, rStringsAppleGuideLookup, siAGIndexTermScreenSize);
						AppleGuideUtilities_OpenFirstMainGuideInSearch(searchString);
					}
					break;
				
				case iScreenSizeUpdateDefaultTerminalCheckBox:
					FlipCheckBox(gSizeDialog, iScreenSizeUpdateDefaultTerminalCheckBox);
					break;
				
				case iScreenSizeWidthArrows:
				case iScreenSizeHeightArrows:
					// necessarily handled in the event filter
					break;
				
				default:
					break;
			}
		} while ((itemIndex != iScreenSizeOKButton) && (itemIndex != iScreenSizeCancelButton));
		DisposeModalFilterUPP(filterProc), filterProc = null;
		
		// handle OK or Cancel click
		if (itemIndex != iScreenSizeCancelButton)
		{
			ConnectionDataPtr	connectionDataPtr = ScreenManager_GetConnectionDataOf(gScreenID);
			Boolean				notGood = false;
			
			
			if (connectionDataPtr == null) notGood = true;
			else
			{
				ControlRef		control = null;
				SInt32			rows = 0L,
								columns = 0L;
				Boolean			updateDefaultTerminal = false;
				
				
				// get the information from the dialog and then kill it
				(OSErr)GetDialogItemAsControl(gSizeDialog, iScreenSizeWidthField, &control);
				GetControlNumericalText(control, &columns);
				(OSErr)GetDialogItemAsControl(gSizeDialog, iScreenSizeHeightField, &control);
				GetControlNumericalText(control, &rows);
				updateDefaultTerminal = (GetDialogItemValue(gSizeDialog, iScreenSizeUpdateDefaultTerminalCheckBox)
											== kControlCheckBoxCheckedValue);
				TransitionDialogClose(gSizeDialog);
				
				// verify the data, emitting a system beep if the user entered invalid data that needed changing
				if (columns < kSizeMinimumWidth)
				{
					columns = kSizeMinimumWidth;
					notGood = true;
				}
				else if (columns > Terminal_GetAllocatedColumnCount())
				{
					columns = Terminal_GetAllocatedColumnCount();
					notGood = true;
				}
				
				if (rows < kSizeMinimumHeight)
				{
					rows = kSizeMinimumHeight;
					notGood = true;
				}
				else if (rows > kSizeMaximumHeight)
				{
					rows = kSizeMaximumHeight;
					notGood = true;
				}
				
				// save the changes - incomplete
				SizeDialog_InstallUndoDimensionChanges(connectionDataPtr->vs);
				ScreenManager_ChangeDimensionsOf(connectionDataPtr->vs, columns, rows, true/* recordable */);
				if (updateDefaultTerminal)
				{
					// unimplemented
				}
			}
			
			if (notGood) BEEP; // warn the user that his or her specifications were not enforced precisely
			result = true;
		}
		else
		{
			// user cancelled - close the dialog with an appropriate transition for cancelling
			TransitionDialogCloseCancel(gSizeDialog);
			result = false;
		}
	}
	else Alert_ReportOSErr(memFullErr);
	
	return result;
}// Display


/*!
To install an Undo procedure that will revert
the dimensions of the specified screen to its
current dimensions when the user chooses Undo,
use this method.

(3.0)
*/
void
SizeDialog_InstallUndoDimensionChanges		(const TelnetScreenID		inScreenID)
{
	OSErr							error = noErr;
	Str255							undoName,
									redoName;
	UndoDataDimensionChangesPtr		dataPtr = (UndoDataDimensionChangesPtr)Memory_NewPtr
											(sizeof(UndoDataDimensionChanges)); // disposed in the action method
	
	
	if (dataPtr == null) error = memFullErr;
	else
	{
		// initialize context structure
		dataPtr->screenID = inScreenID;
		dataPtr->columns = VSgetcols(dataPtr->screenID);
		dataPtr->rows = VSgetlines(dataPtr->screenID);
	}
	GetIndString(undoName, rStringsMenuItemTextSubstitutions, siMenuItemTextUndoDimensionChanges);
	GetIndString(redoName, rStringsMenuItemTextSubstitutions, siMenuItemTextRedoDimensionChanges);
	error = Undoables_AddAction(undoName, redoName, reverseDimensionChanges,
								kUndoableContextIdentifierTerminalDimensionChanges, dataPtr);
	if (error != noErr) Console_WriteValue("Warning: Could not make dimension change undoable, error", error);
}// InstallUndoDimensionChanges


/*!
Sends Apple Events back to MacTelnet that instruct
it to make the specified changes to the frontmost
window.  THIS DOES NOT ACTUALLY EXECUTE THE EVENTS.
You call this routine simply as a convenience for
the user, in case he or she is recording tasks into
a script; this causes dimension change commands
to appear in the script being recorded.

(3.0)
*/
void
SizeDialog_SendRecordableDimensionChangeEvents		(SInt16 const				inNewColumns,
													 SInt16 const				inNewRows)
{
	OSErr		error = noErr;
	AEDesc		setDataEvent,
				replyDescriptor;
	
	
	InstallAE_CreateRecordableAppleEvent(kAECoreSuite, kAESetData, &setDataEvent);
	if (error == noErr)
	{
		AEDesc			containerDesc = { typeNull, null },
						keyDesc = { typeNull, null },
						objectDesc = { typeNull, null },
						propertyDesc = { typeNull, null },
						valueDesc = { typeNull, null },
						reference1Desc = { typeNull, null },
						reference2Desc = { typeNull, null };
		
		
		// send a request for "window <index>", which resides in a null container;
		// then, request the "columns" property, and finally, issue an event
		// that will change its value
		(OSErr)BasicTypesAE_CreateSInt32Desc(1/* frontmost window */, &keyDesc);
		error = CreateObjSpecifier(kTelnetClassWindow, &containerDesc,
									formAbsolutePosition, &keyDesc, true/* dispose inputs */,
									&objectDesc);
		if (error != noErr) Console_WriteValue("window access error", error);
		(OSErr)BasicTypesAE_CreatePropertyIDDesc(kTelnetPropertyTerminalWindowColumns, &keyDesc);
		error = CreateObjSpecifier(cProperty, &objectDesc,
									formPropertyID, &keyDesc, true/* dispose inputs */,
									&reference1Desc);
		if (error != noErr) Console_WriteValue("property access error", error);
		(OSErr)BasicTypesAE_CreatePropertyIDDesc(kTelnetPropertyTerminalWindowRows, &keyDesc);
		error = CreateObjSpecifier(cProperty, &objectDesc,
									formPropertyID, &keyDesc, true/* dispose inputs */,
									&reference2Desc);
		if (error != noErr) Console_WriteValue("property access error", error);
		
		BasicTypesAE_CreatePairDesc(&reference1Desc, &reference2Desc, &propertyDesc);
		AEDisposeDesc(&reference1Desc);
		AEDisposeDesc(&reference2Desc);
		
		if (error == noErr)
		{
			// reference to the property to be set - REQUIRED
			(OSErr)AEPutParamDesc(&setDataEvent, keyDirectObject, &propertyDesc);
			
			// the propertys new value - REQUIRED
			(OSErr)BasicTypesAE_CreateUInt32PairDesc(inNewColumns, inNewRows, &valueDesc);
			(OSErr)AEPutParamDesc(&setDataEvent, kTelnetPartialCoreParameterKeySetDataTo, &valueDesc);
			
			// send the event, which will record it into any open script;
			// this event is not executed because the screen has already
			// been resized; it is simply for convenience to the user, who
			// will be able to achieve the same effect later by running
			// his or her script
			(OSErr)AESend(&setDataEvent, &replyDescriptor,
							kAENoReply | kAECanInteract | kAEDontExecute,
							kAENormalPriority, kAEDefaultTimeout, null, null);
		}
		AEDisposeDesc(&propertyDesc);
		AEDisposeDesc(&valueDesc);
	}
	AEDisposeDesc(&setDataEvent);
}// SendRecordableDimensionChangeEvents


//
// internal methods
//

/*!
This method moves and resizes controls in response to
a resize of the Terminal Screen Size movable modal
dialog box.

(3.0)
*/
static pascal Boolean
handleNewSize		(const WindowRef		inWindow,
					 const SInt32			inDeltaSizeX,
					 const SInt32			inDeltaSizeY,
					 const SInt32			UNUSED_ARGUMENT(inData))
{
	Boolean		result = pascal_true;
	
	
	DialogAdjust_BeginDialogItemAdjustment(GetDialogFromWindow(inWindow));
	{
		SInt32		amountH = 0L,
					amountV = 0L;
		
		
		if (inDeltaSizeX)
		{
			// controls which are moved horizontally
			if (Localization_IsLeftToRight())
			{
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iScreenSizeOKButton, inDeltaSizeX);
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iScreenSizeCancelButton, inDeltaSizeX);
			}
			else
			{
				DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iScreenSizeAppleGuideButton, inDeltaSizeX);
			}
			amountH = (inDeltaSizeX / 3);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iScreenSizeWidthField, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iScreenSizeHeightField, amountH);
			amountH = (INTEGER_DOUBLED(inDeltaSizeX) / 3);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iScreenSizeWidthArrows, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iScreenSizeHeightArrows, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iScreenSizeWidthUnits, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentMoveH, iScreenSizeHeightUnits, amountH);
			
			// controls which are resized horizontally
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iScreenSizeUpdateDefaultTerminalCheckBox, inDeltaSizeX);
			amountH = (inDeltaSizeX / 3);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iScreenSizeWidthLabel, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iScreenSizeWidthField, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iScreenSizeWidthUnits, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iScreenSizeHeightLabel, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iScreenSizeHeightField, amountH);
			DialogAdjust_AddDialogItem(kDialogItemAdjustmentResizeH, iScreenSizeHeightUnits, amountH);
		}
	}
	DialogAdjust_EndAdjustment(inDeltaSizeX, inDeltaSizeY); // moves and resizes controls properly
	
	return result;
}// handleNewSize


/*!
This routine, of standard UndoActionProcPtr form,
can undo or redo changes to the dimensions of a
terminal screen.

Note that it will really be nice to get all of
the AppleScript stuff working, so junk like this
does not have to be done using Copy-and-Paste
coding, and can be made into a recordable event.

(3.0)
*/
static void
reverseDimensionChanges		(const UndoActionInstruction	inDoWhat,
							 UndoableContextIdentifier		inContextIdentifier,
							 void*							inContextPtr)
{
	// this routine only recognizes one kind of context - be absolutely sure thats what was given!
	Console_BeginFunction();
	Console_WriteValue("Undo callback: reverse dimension changes: instruction", inDoWhat);
	if (inContextIdentifier == kUndoableContextIdentifierTerminalDimensionChanges)
	{
		UndoDataDimensionChangesPtr		dataPtr = inContextPtr;
		
		
		Console_WriteLine("Undo callback: reverse dimension changes: valid data");
		switch (inDoWhat)
		{
			case kUndoActionInstructionDispose:
				// release memory previously allocated when this action was installed
				if (inContextPtr != null) Memory_DisposePtr((Ptr*)&inContextPtr);
				break;
			
			case kUndoActionInstructionRedo:
			case kUndoActionInstructionUndo:
			default:
				Console_WriteValue("Undo callback: reverse dimension changes: screen ID", dataPtr->screenID);
				Console_WriteValue("Undo callback: reverse dimension changes: columns", dataPtr->columns);
				Console_WriteValue("Undo callback: reverse dimension changes: rows", dataPtr->rows);
				{
					ConnectionDataPtr		connectionDataPtr = ScreenManager_GetConnectionDataOf
																(dataPtr->screenID);
					Point					oldDimensions;
					
					
					// make this reversible by preserving the dimensions
					SetPt(&oldDimensions, VSgetcols(dataPtr->screenID), VSgetlines(dataPtr->screenID));
					
					// resize the window
					ScreenManager_ChangeDimensionsOf(dataPtr->screenID, dataPtr->columns, dataPtr->rows,
														true/* recordable */);
					
					// save the preserved dimensions
					dataPtr->columns = oldDimensions.h;
					dataPtr->rows = oldDimensions.v;
				}
				break;
		}
	}
	Console_WriteLine("Undo callback: reverse dimension changes: done");
	Console_EndFunction();
}// reverseDimensionChanges


/*!
Handles special key events in the
Terminal Screen Size dialog box.

(3.0)
*/
static pascal Boolean
sizeDialogEventFilter		(DialogPtr		inDialogPtr,
							 EventRecord*	inoutEventPtr,
							 short*			outItemIndex)
{
	Boolean		result = pascal_false,
				isButtonAndShouldFlash = false; // simplifies handling of multiple key-to-button maps
	
	
	// first, make sure help balloons show up properly for the cursor location, if Balloon Help is on
	if (IsPopUpHelpEnabled()) DisplayBalloonForDialogItemUnderMouse(inDialogPtr, inoutEventPtr->where,
																		kDialogBalloonsIDScreenSize);
	
	// this rigamorole is necessary to distinguish between up-arrow and down-arrow clicks!
	if (inoutEventPtr->what == mouseDown)
	{
		GrafPtr				oldPort = null;
		Point				localMouse = inoutEventPtr->where;
		ControlRef			controlHit = null,
							widthArrows = null,
							heightArrows = null;
		ControlPartCode		part = kControlNoPart;
		
		
		GetPort(&oldPort);
		SetPortWindowPort(GetDialogWindow(inDialogPtr));
		GlobalToLocal(&localMouse);
		controlHit = FindControlUnderMouse(localMouse, GetDialogWindow(inDialogPtr), &part);
		GetDialogItemAsControl(inDialogPtr, iScreenSizeWidthArrows, &widthArrows);
		GetDialogItemAsControl(inDialogPtr, iScreenSizeHeightArrows, &heightArrows);
		if ((controlHit == widthArrows) || (controlHit == heightArrows))
		{
			part = HandleControlClick(controlHit, localMouse, inoutEventPtr->modifiers, (ControlActionUPP)null);
			switch (part)
			{
				case kControlUpButtonPart:
				case kControlDownButtonPart:
					{
						ControlRef		field = null;
						SInt32			newDimension = 0;
						
						
						// Determine which little arrows control was hit, and get the value.
						newDimension = GetControlValue(controlHit);
						if (controlHit == widthArrows)
						{
							(OSErr)GetDialogItemAsControl(gSizeDialog, iScreenSizeWidthField, &field);
						}
						else
						{
							(OSErr)GetDialogItemAsControl(gSizeDialog, iScreenSizeHeightField, &field);
						}
						GetControlNumericalText(field, &newDimension);
						
						// Alter the value, and then change the arrow control (which has the
						// maximum and minimum set up correctly); by setting the text field
						// from the arrow control, the text is automatically prevented from
						// acquiring a value beyond the maximum or minimum of the arrows.
						if (part == kControlUpButtonPart) newDimension++;
						else if (part == kControlDownButtonPart) newDimension--;
						SetControlValue(controlHit, newDimension);
						SetControlNumericalText(field, GetControlValue(controlHit));
						DrawOneControl(field);
					}
					result = pascal_true;
					break;
				
				default:
					// ???
					break;
			}
		}
		SetPort(oldPort);
	}
	
	if (!result) result = (StandardDialogEventFilter(inDialogPtr, inoutEventPtr, outItemIndex));
	else if (isButtonAndShouldFlash) FlashButton(inDialogPtr, *outItemIndex);
	
	return result;
}// sizeDialogEventFilter