/*###############################################################

	Sound.c
	
	Interface Library 1.0
	 1998-2000 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// Mac includes
#ifndef __GESTALT__
#	include <Gestalt.h>
#endif
#ifndef __RESOURCES__
#	include <Resources.h>
#endif
#ifndef __SOUND__
#	include <Sound.h>
#endif



// constants
enum
{
	// The following are possible values for the param1 field
	// of a SndCommand structure, as defined for the internal
	// sound callback routine, soundCallBack().
	kSoundCallBackParam1SoundComplete = 1
};

// types
struct SoundInfo
{
	SndChannelPtr			channelPtr;				// the sound channel this sound is playing on
	Handle					resourceHandle;			// for resource-based sounds, a non-null value; otherwise, null
	Boolean					shouldNowBeDestroyed;	// this allows memory to be disposed of outside of the interrupt-level callback
};
typedef struct SoundInfo		SoundInfo;
typedef SoundInfo*				SoundInfoPtr;
typedef SoundInfoPtr*			SoundInfoHandle;

// variables
static ChainRef		gAsyncPlayingSoundsList = null;		// a list of playing sounds



// internal prototypes
static void				disposeSoundInfoRef		(SoundInfoRef*		inoutRefPtr);
static SoundInfoRef		newSoundInfoRef			(void);
static SoundInfoPtr		refToPtr				(SoundInfoRef		inRef);
static pascal void		soundCallBack			(SndChannelPtr		inChannelPtr,
												 SndCommand*		inCmdPtr);



//
// public methods
//

/**
Call this routine before calling any other
method from this module.  When you are
finished using the Sound module, call the
routine Sound_Done().

(1.0)
*/
void
Sound_Init	(void)
{
	gAsyncPlayingSoundsList = Chain_New(kChainStyleSinglyLinked);
}// Init


/**
Call this routine before calling any other
method from this module.  When you are
finished using the Sound module, call the
routine Sound_Done().

(1.0)
*/
void
Sound_Done	(void)
{
	Chain_Dispose(&gAsyncPlayingSoundsList);
}// Done


/**
If you use this module, you should invoke the
Sound_Idle() routine once in your main event
loop.  This allows asynchronous sounds to be
destroyed when they have finished playing.

(1.0)
*/
void
Sound_Idle	(void)
{
	SoundInfoRef		ref = null;
	SoundInfoPtr		ptr = null;
	ChainLinkRef		item = Chain_GetFirstNode(gAsyncPlayingSoundsList);
	
	
	// service all asynchronous sounds that are still around
	while (item != null)
	{
		ref = Chain_GetNodeDataPtr(gAsyncPlayingSoundsList, item);
		ptr = refToPtr(ref);
		if (ptr != null)
		{
			if (ptr->shouldNowBeDestroyed) disposeSoundInfoRef(&ref), ptr = null;
		}
		item = Chain_GetNextNode(gAsyncPlayingSoundsList, item);
	}
}// Idle


/**
You can use this routine to asynchronously play a
sound stored in a resource in the current resource
file.  If the specified resource type is not
recognized as a different kind of sound, it is
assumed to be a standard sound resource ('snd ').
This flexibility allows this routine to be
expanded in the future to accommodate other sound
resource types seamlessly.

Asynchronous sound play requires that you invoke
the Sound_Idle() routine at least once through your
main event loop.  Otherwise, when an asynchronous
sound finishes playing, it will remain in memory
and not be possible to free!

You usually specify "null" for the last parameter,
unless you want to install a sound completion
routine that requires you to be able to identify
which sound has completed.  If you provide a
storage area for a sound information reference,
this routine will provide a reference to the sound
that begins playing when you call this routine.
Then, if you installed a completion routine, you
will be able to tell which sound has finished
playing.

(1.0)
*/
OSErr
Sound_PlayAsyncStereo	(const OSType			inResourceType,
						 const SInt16			inResourceID,
						 SoundInfoRef*			outRefOrNull)
{
	OSErr			result = noErr;
	SndChannelPtr	channelPtr = null;
	
	
#if ONLY_LITE_CARBON_LIB
	result = soundSupportNotAvailableErr;
#else
	result = SndNewChannel(&channelPtr, sampledSynth, initStereo, NewSndCallBackProc(soundCallBack));
	if ((result == noErr) && (channelPtr != null))
	{
		ChainLinkRef	newNode = null;
		SoundInfoRef	ref = newSoundInfoRef();
		
		
		// add a new sound to the list of asynchronous, playing sounds
		if (ref != null)
		{
			SoundInfoPtr	ptr = refToPtr(ref);
			
			
			ptr->resourceHandle = Get1Resource(inResourceType, inResourceID);
			result = ResError();
			if ((ptr->resourceHandle != null) && (result == noErr))
			{
				// this allows the sound data pointer to be acquired directly in the callback without linked list traversal
				channelPtr->userInfo = (long)ptr;
				
				// add the new sound reference to the linked list
				newNode = Chain_NewNode(gAsyncPlayingSoundsList, null);
				Chain_SetNodeDataPtr(gAsyncPlayingSoundsList, newNode, ref);
				Chain_MoveNode(gAsyncPlayingSoundsList, newNode, null, kChainLinkPlacementInsertBefore);
				
				// start the sound playing
				DetachResource(ptr->resourceHandle);
				HLock(ptr->resourceHandle);
				result = SndPlay(channelPtr, (SndListHandle)ptr->resourceHandle, true/* asynchronous */);
				if (result == noErr)
				{
					SndCommand		commandInfo;
					
					
					commandInfo.cmd = callBackCmd;
					commandInfo.param1 = kSoundCallBackParam1SoundComplete;
					commandInfo.param2 = 0L; // the A5 world, normally copied into param2, is not used by PowerPC-native applications
					result = SndDoCommand(channelPtr, &commandInfo, false);
				}
			}
			else
			{
				// there was a problem grabbing the resource, so free up the acquired sound channel
				(OSErr)SndDisposeChannel(channelPtr, false/* make quiet */);
			}
		}
	}
#endif
	
	return result;
}// PlayAsyncStereo


/**
You can use this routine to synchronously play a
sound stored in a resource in the current resource
file.  If the specified resource type is not
recognized as a different kind of sound, it is
assumed to be a standard sound resource ('snd ').
This flexibility allows this routine to be
expanded in the future to accommodate other sound
resource types seamlessly.

(1.0)
*/
OSErr
Sound_PlaySyncStereo	(const OSType			inResourceType,
						 const SInt16			inResourceID)
{
	OSErr			result = noErr;
	SndChannelPtr	channelPtr = null;
	
	
#if ONLY_LITE_CARBON_LIB
	result = soundSupportNotAvailableErr;
#else
	result = SndNewChannel(&channelPtr, sampledSynth, initStereo, null/* callback */);
	if (result == noErr)
	{
		Handle		soundHandle = Get1Resource(inResourceType, inResourceID);
		
		
		if ((soundHandle != null) && (channelPtr != null))
		{
			DetachResource(soundHandle);
			//channelPtr->userInfo = (long)soundHandle;
			HLock(soundHandle);
			result = SndPlay(channelPtr, (SndListHandle)soundHandle, false/* asynchronous */);
			DisposeHandle(soundHandle), soundHandle = null;
		}
		(OSErr)SndDisposeChannel(channelPtr, false/* make quiet */);
	}
#endif
	
	return result;
}// PlaySyncStereo



//
// internal methods
//

/**
To free memory for a copy of the internal structure
that was created with newSoundInfoRef(), use this
method.

(1.0)
*/
static void
disposeSoundInfoRef		(SoundInfoRef*		inoutRefPtr)
{

	if (inoutRefPtr != null)
	{
	#if (!ONLY_LITE_CARBON_LIB)
		SoundInfoPtr		ptr = refToPtr(*inoutRefPtr);
		
		
		// destroy the associated sound channel, freeing it for use by anything else in the system that needs it
		(OSErr)SndDisposeChannel(ptr->channelPtr, true/* make quiet */), ptr->channelPtr = null;
		
		// the resource was detached, so do not ReleaseResource() here
		if (ptr->resourceHandle != null)
		{
			HUnlock(ptr->resourceHandle);
			HPurge(ptr->resourceHandle);
			ptr->resourceHandle = null;
		}
	#endif
		
		// now destroy the internal storage that was used for it
		DisposePtr((Ptr)*inoutRefPtr), *inoutRefPtr = null;
	}
}// disposeSoundInfoRef


/**
To create a copy of the internal structure and
acquire a reference to it, use this method.

(1.0)
*/
static SoundInfoRef
newSoundInfoRef		(void)
{
	SoundInfoRef		result = null;
	
	
	result = (SoundInfoRef)Memory_NewPtr(sizeof(SoundInfo));
	if (result != null)
	{
		SoundInfoPtr	ptr = refToPtr(result);
		
		
		ptr->channelPtr = null;
		ptr->resourceHandle = null;
		ptr->shouldNowBeDestroyed = false;
	}
	
	return result;
}// newSoundInfoRef


/**
Use this method to acquire a pointer to the internal
structure, given a reference to it.

(1.0)
*/
static SoundInfoPtr
refToPtr	(SoundInfoRef		inRef)
{
	return ((SoundInfoPtr)inRef);
}// refToPtr


/**
This routine searches the linked list of asynchronous
sounds to find the one playing on the given channel,
and destroys that item in the list (along with its
channel and any resource data).

(1.0)
*/
pascal void
soundCallBack	(SndChannelPtr		inChannelPtr,
				 SndCommand*		inCmdPtr)
{
	if ((inCmdPtr != null) && (inChannelPtr != null))
	{
		SoundInfoPtr		ptr = null;
		
		
		// the data pointer is stored in the user data of the sound channel
		ptr = (SoundInfoPtr)inChannelPtr->userInfo;
		
		// now examine the request
		if (ptr != null)
		{
			if (inCmdPtr->param1 == kSoundCallBackParam1SoundComplete)
			{
				// MARK the sound for destruction - WARNING, as an interrupt-level
				// routine, you cant actually dispose memory here!
				ptr->shouldNowBeDestroyed = true;
			}
		}
	}
}// soundCallBack