/*###############################################################

	StatusIcon.c
	
	MacTelnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Appearance.h>
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Folders.h>
#include <Icons.h>
#include <QuickDraw.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"

// MacTelnet includes
#include "DialogUtilities.h"
#include "MacLook.h"
#include "StatusIcon.h"
#include "TelnetDescriptors.h"



// types
struct StatusIcon
{
	UInt16				animationStageIndex;
	ChainRef			iconAnimationStages;
	IconTransformType	iconTransform;
	ControlRef			userPane;
	Boolean				animate;
};
typedef struct StatusIcon	StatusIcon;
typedef StatusIcon*			StatusIconPtr;

struct IconAnimationStage
{
	IconManagerIconRef		icon;
	UInt32					ticks;
};
typedef struct IconAnimationStage	IconAnimationStage;
typedef IconAnimationStage*			IconAnimationStagePtr;

// internal methods
static Boolean					addAnimationStage		(ChainRef					inAnimationStageList,
														 IconAnimationStagePtr		inNewStagePtr);
static void						disposeAnimationStage	(IconAnimationStagePtr*		inoutPtr);
static void						getAnimationStage		(ChainRef					inAnimationStageList,
														 const UInt16				inIndex,
														 IconAnimationStagePtr*		outStagePtr);
static pascal void				iconIdle				(ControlRef					inControl);
static pascal void				iconUserPaneDraw		(ControlRef					inControl,
														 SInt16						inPart);
static IconAnimationStagePtr	newAnimationStage		(const SInt16				inResourceFile,
														 const SInt16				inIconSuiteResourceID,
														 const UInt32				inTicks);
static StatusIconPtr			refAcquireLock			(StatusIconRef				inRef);
static void						refReleaseLock			(StatusIconRef				inRef,
														 StatusIconPtr*				inoutPtr);



//
// public methods
//

/*!
To create a new status icon control that will
appear in the specified window, use this method.

You can use the new reference to manipulate the
status icon using the methods in this module.
The status icon initially contains no icon, so
you need to call StatusIcon_AddStage() to make
at least one animation cell.

If any problems occur, null is returned;
otherwise, a reference to the new status icon
is returned.

(3.0)
*/
StatusIconRef
StatusIcon_New		(const WindowRef	inOwningWindow)
{
	StatusIconRef		result = (StatusIconRef)Memory_NewPtr(sizeof(StatusIcon));
	
	
	if (result != null)
	{
		StatusIconPtr		ptr = refAcquireLock(result);
		OSErr				error = noErr;
		ChainLinkRef		node = null;
		
		
		ptr->animationStageIndex = 0;
		ptr->iconAnimationStages = Chain_New(kChainStyleSinglyLinked);
		
		if (error == noErr)
		{
			ControlUserPaneIdleUPP		idleUPP = NewControlUserPaneIdleUPP(iconIdle);
			ControlUserPaneDrawUPP		drawUPP = NewControlUserPaneDrawUPP(iconUserPaneDraw);
			ControlRef					control = null;
			Rect						rect;
			
			
			SetRect(&rect, 0, 0, 16, 16);
			ptr->userPane = control = GetNewControl(kIDStatusIcon, inOwningWindow);
			SetControlReference(control, REINTERPRET_CAST(result, SInt32)/* encode StatusIconRef into ref. con. */);
			(OSErr)SetControlData(control, kControlNoPart, kControlUserPaneIdleProcTag,
									sizeof(ControlUserPaneIdleUPP), (Ptr)&idleUPP);
			(OSErr)SetControlData(control, kControlNoPart, kControlUserPaneDrawProcTag,
									sizeof(ControlUserPaneDrawUPP), (Ptr)&drawUPP);
		}
		
		if (error != noErr) Memory_DisposePtr((Ptr*)&result); // destroy the new icon if anything went wrong
		refReleaseLock(result, &ptr);
	}
	return result;
}// New


/*!
This method cleans up a status icon by destroying
all of the data associated with it.  On output,
your copy of the given reference will be set to
null.

WARNING:	If you previously used accessor
			methods to acquire physical icon data
			from a status icon, all of those data
			handles will become invalid once you
			call StatusIcon_Dispose().

(3.0)
*/
void
StatusIcon_Dispose		(StatusIconRef*		inoutRefPtr)
{
	if (inoutRefPtr != null)
	{
		StatusIconPtr		ptr = refAcquireLock(*inoutRefPtr);
		
		
		if (ptr != null)
		{
			ChainLinkRef			node = null;
			IconAnimationStagePtr	stagePtr = null;
			
			
			while ((node = Chain_DeleteLastNode(ptr->iconAnimationStages)) != null)
			{
				stagePtr = (IconAnimationStagePtr)Chain_GetNodeDataPtr(ptr->iconAnimationStages, node);
				disposeAnimationStage(&stagePtr);
			}
			Memory_DisposePtr((Ptr*)inoutRefPtr), ptr = null;
		}
	}
}// Dispose


/*!
To add an icon animation stage to a status icon,
use this method.  If you want to create a single
icon that does not animate, simply call this
method only once, specifying an icon and a delay
of zero.  If you call this method multiple times
(to install the various stages of animation),
you must call this routine in the order you want
those stages to be animated.

Under Mac OS 8.5 and beyond, the specified icon
suite ID will be parsed, and if the icon ID
refers to an icon available via Icon Services,
the better icon will be used automatically and
transparently.  In other words, you can specify
an icon as if it were a System 7 icon, and get
Mac OS X functionality when possible without any
extra work on your part.

(3.0)
*/
void
StatusIcon_AddStage		(StatusIconRef		inRef,
						 const SInt16		inResourceFile,
						 const SInt16		inIconSuiteResourceID,
						 const UInt32		inAnimationDelayInTicks)
{
	StatusIconPtr	ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		(Boolean)addAnimationStage(ptr->iconAnimationStages,
									newAnimationStage(inResourceFile, inIconSuiteResourceID, inAnimationDelayInTicks));
		refReleaseLock(inRef, &ptr);
	}
}// AddStage


/*!
To acquire a handle to the control that draws a
status icon, use this method.  You need to call
this method in order to display the icon in a
window!

(3.0)
*/
ControlRef
StatusIcon_GetContainerControl		(StatusIconRef		inRef)
{
	StatusIconPtr	ptr = refAcquireLock(inRef);
	ControlRef		result = false;
	
	
	if (ptr != null)
	{
		result = ptr->userPane;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetContainerControl


/*!
To determine if animation is currently turned on
for a status icon, use this method.

(3.0)
*/
Boolean
StatusIcon_IsAnimating		(StatusIconRef		inRef)
{
	StatusIconPtr	ptr = refAcquireLock(inRef);
	Boolean			result = false;
	
	
	if (ptr != null)
	{
		result = ptr->animate;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// IsAnimating


/*!
To turn icon animation on or off, use this
method.  This flag only has effect if there is
more than one icon animation stage installed;
if animation is on, then the icon will animate
as long as IdleControls() is called
sufficiently often.  Otherwise, the icon will
not change even if IdleControls() is invoked
repeatedly.

You can stop icon animation using this routine,
and then resume it later: to do this, pass
false for "inResetStages".  Or, you can start
the animation over, by passing true.

To step through an animation explicitly,
bracket a call to IdleControls() with calls to
this method, turning animation on just prior to
the idle, and then back off.  If you want to do
this, you should make the inter-frame delay
small (like one tick) to guarantee that the
idle procedure will change the animation cell.
Alternately, you can use StatusIcon_SetNextCell()
and DrawOneControl(), for finer control over the
animation.

(3.0)
*/
void
StatusIcon_SetDoAnimate		(StatusIconRef		inRef,
							 const Boolean		inIconShouldAnimate,
							 const Boolean		inResetStages)
{
	StatusIconPtr	ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ptr->animate = inIconShouldAnimate;
		if (inResetStages) ptr->animationStageIndex = 0;
		refReleaseLock(inRef, &ptr);
	}
}// SetDoAnimate


/*!
To set the next animation cell, use this method.
For icons which do not animate continuously, you
generally call this routine to set which icon in
the set you wish to use, and then call the
IdleControls() routine to update the icon.  For
best results in this latter case, you should
make the inter-frame delay small, like 1 tick.

(3.0)
*/
void
StatusIcon_SetNextCell		(StatusIconRef		inRef,
							 const UInt16		inZeroBasedNextCellIndex)
{
	StatusIconPtr	ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ptr->animationStageIndex = inZeroBasedNextCellIndex;
		refReleaseLock(inRef, &ptr);
	}
}// SetNextCell


/*!
To specify the icon transformation to apply
to this status icon when it is drawn (in any
animation stage), use this method.

(3.0)
*/
void
StatusIcon_SetTransformType		(StatusIconRef		inRef,
								 IconTransformType	inTransform)
{
	StatusIconPtr	ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ptr->iconTransform = inTransform;
		refReleaseLock(inRef, &ptr);
	}
}// SetTransformType


//
// internal methods
//

/*!
To add a new item to an animation stage list,
use this method.  If the stage was added
successfully, true is returned; otherwise,
false is returned.

(3.0)
*/
static Boolean
addAnimationStage		(ChainRef				inAnimationStageList,
						 IconAnimationStagePtr	inNewStagePtr)
{
	Boolean		result = false;
	
	
	if ((inAnimationStageList != null) && (inNewStagePtr != null))
	{
		ChainLinkRef	node = Chain_NewNode(inAnimationStageList, inNewStagePtr);
		
		
		if (node != null)
		{
			Chain_MoveNode(inAnimationStageList, node, null/* means move to end of list */,
							kChainLinkPlacementInsertAfter);
			result = true;
		}
	}
	return result;
}// addAnimationStage


/*!
To destroy an animation stage structure
created with newAnimationStage(), use
this method.

(3.0)
*/
static void
disposeAnimationStage	(IconAnimationStagePtr*		inoutPtr)
{
	if (inoutPtr != null)
	{
		if ((*inoutPtr) != null)
		{
			IconManager_DisposeIcon(&((*inoutPtr)->icon));
			Memory_DisposePtr((Ptr*)inoutPtr);
		}
	}
}// disposeAnimationStage


/*!
To get an item in an animation stage list,
use this method.

(3.0)
*/
static void
getAnimationStage	(ChainRef					inAnimationStageList,
					 const UInt16				inIndex,
					 IconAnimationStagePtr*		outStagePtr)
{
	Boolean		result = false;
	
	
	if ((inAnimationStageList != null) && (outStagePtr != null))
	{
		ChainLinkRef	node = Chain_GetIndexedNode(inAnimationStageList, inIndex);
		
		
		if (node != null) *outStagePtr = (IconAnimationStagePtr)Chain_GetNodeDataPtr(inAnimationStageList, node);
		else *outStagePtr = null;
	}
}// getAnimationStage


/*!
To allow a status icon to perform periodic tasks,
use this method.  You must call this routine often
enough to allow animating icons to animate, which
depends on the intervals you have assigned (that is,
the fastest any frame change can occur).  If
possible, use a thread to invoke this routine.

By regularly calling IdleControls() for application
windows, this routine will be called sufficiently
often.

(3.0)
*/
static pascal void
iconIdle	(ControlRef		inControl)
{
	StatusIconRef		ref = REINTERPRET_CAST(GetControlReference(inControl), StatusIconRef);
	
	
	if (ref != null)
	{
		StatusIconPtr	ptr = refAcquireLock(ref);
		
		
		if (ptr != null)
		{
			if (StatusIcon_IsAnimating(ref)) // only animate if animation is turned on
			{
				UInt16		stageCount = Chain_GetNodeCount(ptr->iconAnimationStages);
				
				
				if (stageCount > 1)
				{
					IconAnimationStagePtr	stagePtr = null;
					
					
					getAnimationStage(ptr->iconAnimationStages, ptr->animationStageIndex, &stagePtr);
					if (stagePtr != null)
					{
						static SInt32		lastTime = 0L;
						SInt32				time = TickCount();
						
						
						if ((time - lastTime) >= stagePtr->ticks)
						{
							Rect		iconBounds;
							
							
							ptr->animationStageIndex++;
							GetControlBounds(ptr->userPane, &iconBounds);
							MacLook_InvalidateRectangle(GetControlOwner(ptr->userPane), &iconBounds);
						}
						if (ptr->animationStageIndex > stageCount) ptr->animationStageIndex = 0;
						lastTime = time;
					}
				}
			}
			refReleaseLock(ref, &ptr);
		}
	}
}// iconIdle


/*!
This method, of standard DeviceLoopDrawingUPP form,
handles drawing of Status Icons correctly no matter
what the device characteristics are.

IMPORTANT:	The control reference constant is to be
			of type StatusIconRef.

(3.0)
*/
static pascal void
iconUserPaneDeviceLoop		(short				inColorDepth,
							 short				UNUSED_ARGUMENT(inDeviceFlags),
							 GDHandle			inTargetDevice,
							 long				inControl)
{
	StatusIconRef		ref = null;
	ControlRef			control = REINTERPRET_CAST(inControl, ControlRef);
	
	
	ref = REINTERPRET_CAST(GetControlReference(control), StatusIconRef);
	if (ref != null)
	{
		StatusIconPtr		ptr = refAcquireLock(ref);
		
		
		if (ptr != null)
		{
			// First, get the current icon animation stage from the stage list
			// of the given status icon.  Then, plot the current stages icon,
			// making it appear disabled if the given control is inactive.
			IconTransformType		additionalTransform = (IsControlActive(control) ? 0 : kTransformDisabled);
			IconAnimationStagePtr	stagePtr = null;
			ChainLinkRef			node = Chain_GetIndexedNode(ptr->iconAnimationStages, ptr->animationStageIndex);
			Rect					controlRect;
			
			
			GetControlBounds(control, &controlRect);
			(OSErr)SetUpControlBackground(control, inColorDepth, ColorUtilities_IsColorDevice(inTargetDevice));
			EraseRect(&controlRect);
			if (node != null)
			{
				stagePtr = (IconAnimationStagePtr)Chain_GetNodeDataPtr(ptr->iconAnimationStages, node);
				if (stagePtr != null)
				{
					(OSErr)IconManager_PlotIcon(stagePtr->icon, &controlRect, kAlignAbsoluteCenter,
												ptr->iconTransform | additionalTransform);
				}
			}
			refReleaseLock(ref, &ptr);
		}
	}
}// iconUserPaneDeviceLoop


/*!
This method, a standard Mac OS user pane
drawing procedure, draws the current status
icon in the control rectangle.

IMPORTANT:	The control reference constant
			is to be of type StatusIconRef.

(3.0)
*/
static pascal void
iconUserPaneDraw		(ControlRef		inControl,
						 SInt16			UNUSED_ARGUMENT(inPartCode))
{
	DeviceLoopClipAndDrawControl(inControl, NewDeviceLoopDrawingUPP(iconUserPaneDeviceLoop));
}// iconUserPaneDraw


/*!
To create a new animation stage structure,
use this method.

(3.0)
*/
static IconAnimationStagePtr
newAnimationStage		(const SInt16		inResourceFile,
						 const SInt16		inIconSuiteResourceID,
						 const UInt32		inTicks)
{
	IconAnimationStagePtr		result = null;
	SInt16						oldResourceFile = CurResFile();
	
	
	UseResFile(inResourceFile);
	if (ResError() == noErr)
	{
		result = (IconAnimationStagePtr)Memory_NewPtr(sizeof(IconAnimationStage));
		if (result != null)
		{
			OSErr		error = noErr;
			
			
			result->ticks = inTicks;
			result->icon = IconManager_NewIcon();
			error = IconManager_MakeIconSuite(result->icon, inIconSuiteResourceID, kSelectorAllSmallData);
			if (error != noErr)
			{
				// destroy the new stage if any problems occur
				Memory_DisposePtr((Ptr*)&result);
			}
		}
	}
	UseResFile(oldResourceFile);
	return result;
}// newAnimationStage


/*!
To obtain a pointer to the internal structure,
given a reference to it, use this method.

(3.0)
*/
static StatusIconPtr
refAcquireLock		(StatusIconRef		inRef)
{
	return ((StatusIconPtr)inRef);
}// refAcquireLock


/*!
To release a pointer to the internal structure,
given a reference to it, use this method.

(3.0)
*/
static void
refReleaseLock		(StatusIconRef		UNUSED_ARGUMENT(inRef),
					 StatusIconPtr*		inoutPtr)
{
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock