/*###############################################################

	StringUtilities.c
	
	This file contains useful methods for string management.
	
	Data Access Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <string.h>

// Mac includes
#include <PLStringFuncs.h>
#include <TextUtils.h>

// library includes
#include "DataAccessLib.h"



// variables
static Str15	gMeta1 = kMetacharacterDefault1,
				gMeta2 = kMetacharacterDefault2,
				gMeta3 = kMetacharacterDefault3,
				gMeta4 = kMetacharacterDefault4;
static UInt16	gMetaLength = 2;

// internal methods
static Boolean		isAmong			(CharParameter const	inCharacter,
									 StringPtr const		inCharacterList);
static Handle		refToHandle		(TokenSetRef			inRef);



//
// public methods
//

/*!
This method can convert a string of bytes into data for an
IP address that is acceptable to the networking software.
The first valid part of the given byte array is used, and
the rest is ignored.  Only groups of 4 octets are allowed;
simpler forms of dot notation are not recognized by this
routine.  This method returns "true" only if the conversion
succeeds; if "false" is returned, the output IP address is
invalid.

For future expandability, this routine uses the structure
"InetAddress", so that eventually a parser that understands
IP version 6.0 addresses, and port numbers, can be written.

(1.0)
*/
Boolean
StringUtilities_BufferToInternetAddress		(Ptr				inBuffer,
											 Size const			inBufferLength,
											 InetAddress*		outIPAddressPtr)
{
	Boolean			result = false;
	Boolean			done = false;
	unsigned long	address = 0;
	short			byteValue = 0;
	Ptr				charPtr;
	short			dotCount = 0;
	
	
	// permit leading space
	for (charPtr = (inBuffer + 1); charPtr <= (inBuffer + inBufferLength); charPtr++)
	{
		if (*charPtr != ' ') break;
	}
	
	// now convert
	for (; charPtr <= (inBuffer + inBufferLength); charPtr++)
	{
		if (*charPtr == '.')
		{
			if (++dotCount > 3) done = true;
			if (!done)
			{
				address <<= 8;
				address |= byteValue;
				byteValue = 0;
			}
		}
		else if (((*charPtr) >= '0') &&
				((*charPtr) <= '9'))	// using "isdigit()" would prevent LiteCarbonLib linkage
		{
			byteValue *= 10;
			byteValue += (*charPtr - '0');
			if (byteValue > 255) done = true;
		}
		else if (*charPtr == ' ') break;
		else done = true;
	}
	
	if (!done)
	{
		// 3 dots would mean four numbers; is it valid?
		if (dotCount == 3)
		{
			address <<= 8;
			address |= byteValue;
			outIPAddressPtr->fAddressType = AF_INET;
			outIPAddressPtr->fPort = 0; // currently not parsed
			outIPAddressPtr->fHost = address;
			result = true;
		}
	}
	
	return result;
}// BufferToInternetAddress


/*!
Performs a normal conversion of a Pascal string to a C string
(that is, it copies and converts a buffer of characters whose
first character is a length byte to a second buffer, where
the first character of the second buffer is real and the
string is terminated by a 0 byte).

(1.0)
*/
void
StringUtilities_CToP	(char*			inString,
						 StringPtr		outBuffer)
{
#if TARGET_API_MAC_OS8
	strcpy(STATIC_CAST(outBuffer, char*), inString);
	C2PStr(STATIC_CAST(outBuffer, char*));
#else
	CopyCStringToPascal(inString, outBuffer);
#endif
}// CToP


/*!
Performs an in-place conversion of a C string to a Pascal
string (that is, it converts a buffer of characters whose first
character is real and last character is a terminating 0 byte to
a buffer whose first character is a length byte).  This is not
usually recommended, since it requires typecasting and can lead
to bugs if constness is casted away incorrectly.

(1.0)
*/
StringPtr
StringUtilities_CToPInPlace		(char		inoutString[])
{
#if TARGET_API_MAC_OS8
	// traditional Mac OS has C2PStr
	return C2PStr(inoutString);
#else
	// Carbon offers a new c2pstrcpy function
	c2pstrcpy(STATIC_CAST(inoutString, StringPtr), inoutString);
	return STATIC_CAST(inoutString, StringPtr);
#endif
}// CToPInPlace


/*!
Appends a string representation of a signed or
unsigned number to the given string.

(1.0)
*/
void
StringUtilities_PAppendNumber	(StringPtr		inoutString,
								 long			inNumber)
{
	Str255		string;
	
	
	NumToString((long)inNumber, string);
	PLstrcat(inoutString, string);
}// PAppendNumber


/*!
To build a string no longer  than 255 characters out of
a source  string  containing  metacharacters and  up to
four substitution strings, call this method.   The four
given substitution strings  are used  in the order that
metacharacters  are  encountered, and  must  be  either
empty or  nonempty strings (that is,  not  null).  This
utility exists  to supplement the functionality  of the
ParamText() system call by supporting four more special
metacharacters (by default %a, %b, %c and %d -- you can
redefine   the   metacharacters   using   the   routine
StringUtilities_PMetaSet()).   If you require more than
four metacharacters in a string, you can use the method
StringUtilities_PMetaSet() to change the metacharacters
used by this routine to a  different set, and then just
call this routine twice.  However, remember, the string
result after substitutions cannot exceed 255 characters
in  length.   The result of  invoking this  method on a
string that has no metacharacters in it is not defined.

Before calling  this routine, you should  *always* call
either StringUtilities_PMetaSetDefaults(),  or the more
flexible method StringUtilities_PMetaSet().  This makes
sure that  the set of four metacharacters  that will be
used for building the string are the metacharacters you
expect!  Once you call StringUtilities_PMetaSet(),  the
settings  you make  will stick for  subsequent calls to
StringUtilities_PBuild()  until you  call  the  routine
StringUtilities_PMetaSet() again to change them.

IMPORTANT:	All the metacharacters should appear in the
			proper order in the input string, otherwise
			the string substitution that gets performed
			is unpredictable.  You can omit one or more
			metacharacters, but the metacharacters must
			still be in order.  For example, if you use
			the string "My name is %a, his name is %d",
			this will work as long as the substitutions
			are provided for strings 1 and 4, not 1 and
			2.  If you used the same two metacharacters
			in the opposite order, only the %a would be
			substituted (%d would be ignored).

(1.0)
*/
void
StringUtilities_PBuild	(StringPtr		inoutString,
						 StringPtr		inSubstitutionString1,
						 StringPtr		inSubstitutionString2,
						 StringPtr		inSubstitutionString3,
						 StringPtr		inSubstitutionString4)
{
	Str255		temp;
	StringPtr	junk = null;
	Ptr			aPtr = PLstrstr(inoutString, gMeta1),
				bPtr = PLstrstr(inoutString, gMeta2),
				cPtr = PLstrstr(inoutString, gMeta3),
				dPtr = PLstrstr(inoutString, gMeta4);
	int			a = (int)((StringPtr)(PLstrstr(inoutString, gMeta1)) - inoutString - 1),
				b = (int)((StringPtr)(PLstrstr(inoutString, gMeta2)) - inoutString - 1),
				c = (int)((StringPtr)(PLstrstr(inoutString, gMeta3)) - inoutString - 1),
				d = (int)((StringPtr)(PLstrstr(inoutString, gMeta4)) - inoutString - 1);
	
	
	if (a < 0) a = -1;
	if (b < 0) b = -1;
	if (c < 0) c = -1;
	if (d < 0) d = -1;
	
	// Here, the metacharacters must be in
	// order for substitution to occur properly.
	// Concatenations to a temporary string are
	// done throughout, where math is the only
	// thing used to perform "substitutions".
	// When 1 is added to the string pointer,
	// the length byte at the beginning of it
	// is being taken into account.  The final
	// step is to copy the temporary string so
	// it overwrites the source string.
	{
		Ptr		pointer = (Ptr)temp;
		long	size = 255 * sizeof(unsigned char);
		
		
		//PLstrcpy(temp, EMPTY_PSTRING);
		while (size--) *pointer++ = 0;
	}
	if (a >= 0)
	{
		PLstrncat(temp, inoutString, a);
		PLstrcat(temp, inSubstitutionString1);
	}
	if (b >= 0)
	{
		// set the length byte (destroys some of the original data, but we have a copy)
		junk = (StringPtr)(aPtr - 1 + ((char)(gMetaLength * sizeof(char))));
		junk[0] = (char)(b - a - gMetaLength);
		PLstrcat(temp, junk);
		PLstrcat(temp, inSubstitutionString2);
	}
	if (c >= 0)
	{
		// set the length byte (destroys some of the original data, but we have a copy)
		junk = (StringPtr)(bPtr - 1 + ((char)(gMetaLength * sizeof(char))));
		junk[0] = (char)(c - b - gMetaLength);
		PLstrcat(temp, junk);
		PLstrcat(temp, inSubstitutionString3);
	}
	if (d >= 0)
	{
		// set the length byte (destroys some of the original data, but we have a copy)
		junk = (StringPtr)(cPtr - 1 + ((char)(gMetaLength * sizeof(char))));
		junk[0] = (char)(d - c - gMetaLength);
		PLstrcat(temp, junk);
		PLstrcat(temp, inSubstitutionString4);
	}
	
	// Find the offset of the last substitution metacharacter,
	// in order to determine where to copy the remaining stuff
	// from the string (if anything).  Set the length byte
	// (destroys some of the original data, but we have a copy).
	{
		Ptr		xPtr = (Ptr)(inoutString + 1);
		int		x = 0;
		
		
		if (d >= 0)
		{
			x = d;
			xPtr = dPtr;
		}
		else if (c >= 0)
		{
			x = c;
			xPtr = cPtr;
		}
		else if (b >= 0)
		{
			x = b;
			xPtr = bPtr;
		}
		else if (a >= 0)
		{
			x = a;
			xPtr = aPtr;
		}
		
		junk = (StringPtr)(xPtr - 1 + ((char)(gMetaLength * sizeof(char))));
		junk[0] = (char)(PLstrlen(inoutString) - x - gMetaLength);
	}
	PLstrcat(temp, junk);
	PLstrcpy(inoutString, temp);
}// PBuild


/*!
To clip a certain number of characters
from the beginning of a string, use
this method.

(1.0)
*/
void
StringUtilities_PClipHead	(StringPtr		inoutString,
							 UInt16 const	inHowManyCharacters)
{
	if (StrLength(inoutString) >= inHowManyCharacters)
	{
		inoutString[inHowManyCharacters] = inoutString[0] - inHowManyCharacters;
		PLstrcpy(inoutString, inoutString + inHowManyCharacters);
	}
}// PClipHead


/*!
To truncate a string, use this method.

(1.0)
*/
void
StringUtilities_PClipTail	(StringPtr		inoutString,
							 UInt16 const	inHowManyCharacters)
{
	inoutString[0] -= inHowManyCharacters;
}// PClipTail


/*!
To dispose of memory allocated by a returned
handle of tokens from StringUtilities_PTokenize(),
use this method.

(1.0)
*/
void
StringUtilities_PDisposeTokens		(UInt16 const	inArgC,
									 TokenSetRef*	inoutArgV)
{
	if (inoutArgV != null)
	{
		register UInt16		i = 0;
		StringHandle		stringHandle = null;
		Handle				handle = refToHandle(*inoutArgV);
		
		
		HLock(handle);
		for (i = 0; i < inArgC; i++)
		{
			stringHandle = ((StringHandle*)*handle)[i];
			DisposeHandle((Handle)stringHandle);
		}
		DisposeHandle(handle), handle = null, *inoutArgV = null;
	}
}// PDisposeTokens


/*!
To extract the first line of characters from the
specified buffer into a string, use this method.
If there are no new-line characters in the first
255 characters of the buffer, then the first 255
characters are considered to be the line (Pascal
string limitations).  The source buffer is not
affected by this call.  The "outLine" string is
the first line of text.  You can clip the
first line of text by incrementing the given
source pointer when this routine returns, using
the length of the resultant string.

This routine is useful for accepting drags of
text chunks, when the destination is a one-line
text field.  It is also useful for iterating
through a text buffer line-by-line.

Note that while the specified source data is
raw data, the resultant string is a Pascal
string (with a leading length byte).

(1.0)
*/
void
StringUtilities_PExtractFirstLine	(unsigned char*		inSourcePtr,
									 Size				inSize,
									 StringPtr			outLine)
{
	if ((outLine != null) && (inSourcePtr != null))
	{
		register short		i = 0;
		
		
		// find the first line
		for (i = 0; i < inSize; i++)
		{
			if (inSourcePtr[i] == '\n') break;
		}
		
		if (i >= inSize) i = inSize; // ...if the end of the buffer was reached, make a minor adjustment
		if (i > 255) i = 255; // ... if the chosen size is bigger than 255 characters, stop it at 255 (Pascal string limitations)
		outLine[0] = i; // set the size byte of this Pascal string buffer
		if (i != 0) BlockMove(inSourcePtr, outLine + 1, i); // copy the data to it
	}
}// PExtractFirstLine


/*!
To acquire a token from a token handle returned
from a call to StringUtilities_PTokenize(), use
this convenience method.

WARNING:	Do not pass a handle that was not
			returned via a call to the method
			StringUtilities_PTokenize().

(1.0)
*/
void
StringUtilities_PGetToken	(UInt16 const	inZeroBasedTokenNumber,
							 TokenSetRef	inArgV,
							 StringPtr		outString)
{
	if ((outString != null) && (inArgV != null))
	{
		StringHandle	stringHandle = null;
		Handle			handle = refToHandle(inArgV);
		
		
		HLock(handle);
		if (GetHandleSize(handle) >= ((inZeroBasedTokenNumber + 1) * sizeof(StringHandle)))
		{
			stringHandle = ((StringHandle*)*handle)[inZeroBasedTokenNumber];
			if (stringHandle != null) PLstrcpy(outString, *stringHandle);
		}
		else BEEP;
		HUnlock(handle);
	}
}// PGetToken


/*!
Inserts the string "inSourceString" at the beginning
of the string "inoutDestinationString".  Unlike the
system call PLstrcat() (which combines two strings
also), this routine modifies the destination string
instead of the source string.

(1.0)
*/
void
StringUtilities_PInsert	(StringPtr		inoutDestinationString,
						 StringPtr		inSourceString)
{
	// make room for new string
	BlockMove(inoutDestinationString + 1, inoutDestinationString + *inSourceString + 1, *inoutDestinationString);
	
	// copy new string in and adjust the length byte
	BlockMove(inSourceString + 1, inoutDestinationString + 1, *inSourceString);
	*inoutDestinationString += *inSourceString;
}// PInsert


/*!
To determine the four metacharacters that
StringUtilities_PBuild() looks for when
performing substitutions, call this method.
The length of all metacharacters is the
same, and is returned as well.

(1.0)
*/
void
StringUtilities_PMetaGet	(StringPtr		outMeta1,
							 StringPtr		outMeta2,
							 StringPtr		outMeta3,
							 StringPtr		outMeta4,
							 UInt16*		outLength)
{
	PLstrcpy(outMeta1, gMeta1);
	PLstrcpy(outMeta2, gMeta2);
	PLstrcpy(outMeta3, gMeta3);
	PLstrcpy(outMeta4, gMeta4);
	*outLength = gMetaLength;
}// PMetaGet


/*!
To specify the four metacharacters that
StringUtilities_PBuild() will look for
when performing substitutions, call this
method.  You must also specify the
length of each metacharacter, which must
be the same for all four.  Usually,
metacharacters are two characters long.
Do not specify metacharacters longer
than 14 characters.

(1.0)
*/
void
StringUtilities_PMetaSet	(StringPtr		inMeta1,
							 StringPtr		inMeta2,
							 StringPtr		inMeta3,
							 StringPtr		inMeta4,
							 UInt16			inLength)
{
	PLstrcpy(gMeta1, inMeta1);
	PLstrcpy(gMeta2, inMeta2);
	PLstrcpy(gMeta3, inMeta3);
	PLstrcpy(gMeta4, inMeta4);
	gMetaLength = inLength;
}// PMetaSet


/*!
To specify strings in a string list resource
(from the current resource file) representing
the four metacharacters that the routine
StringUtilities_PBuild() will look for when
performing substitutions, call this method.
You must also specify the length of each
metacharacter, which must be the same for
all four.  Usually, metacharacters are two
characters long.  Do not specify
metacharacters longer than 14 characters.

(1.0)
*/
void
StringUtilities_PMetaSetFromResources	(short		inStringListResourceID,
										 short		inMeta1StringIndex,
										 short		inMeta2StringIndex,
										 short		inMeta3StringIndex,
										 short		inMeta4StringIndex,
										 UInt16		inLength)
{
	GetIndString(gMeta1, inStringListResourceID, inMeta1StringIndex);
	GetIndString(gMeta1, inStringListResourceID, inMeta2StringIndex);
	GetIndString(gMeta1, inStringListResourceID, inMeta3StringIndex);
	GetIndString(gMeta1, inStringListResourceID, inMeta4StringIndex);
	gMetaLength = inLength;
}// PMetaSetFromResources


/*!
To determine if the *very* first characters of the
specified string (not including the length byte, of
course) are identical to the given prefix, call this
method.

(1.0)
*/
Boolean
StringUtilities_PStartsWith		(StringPtr		inLongString,
								 StringPtr		inPrefixString)
{
	return (!PLstrncmp(inLongString, inPrefixString, StrLength(inPrefixString)));
}// PStartsWith


/*!
To delete *all* substrings in the specified string that
begin with the specified start delimiter string and end
with the specified end delimiter string, use this method.
All characters between the given delimiters are deleted,
but you may choose to delete the delimiters as well, or
leave them untouched.

This routine is not capable of detecting nested delimiters.

Unimplemented.

(1.0)
*/
void
StringUtilities_PStrip		(StringPtr		inoutString,
							 StringPtr		inStartingDelimiter,
							 StringPtr		inEndingDelimiter,
							 Boolean		inStripDelimitersToo)
{
}// PStrip


/*!
Performs a normal conversion of a Pascal string to a C string
(that is, it copies and converts a buffer of characters whose
first character is a length byte to a second buffer, where
the first character of the second buffer is real and the
string is terminated by a 0 byte).

(1.0)
*/
void
StringUtilities_PToC	(StringPtr		inString,
						 char			outBuffer[])
{
#if TARGET_API_MAC_OS8
	PLstrcpy(STATIC_CAST(outBuffer, StringPtr), inString);
	P2CStr(STATIC_CAST(outBuffer, StringPtr));
#else
	CopyPascalStringToC(inString, outBuffer);
#endif
}// PToC


/*!
Performs an in-place conversion of a Pascal string to a
C string (that is, it converts a buffer of characters whose
first character is a length byte to a buffer whose first
character is real and last character is a terminating 0 byte).
This is not usually recommended, since it requires typecasting
and can lead to bugs if constness is casted away incorrectly.

(1.0)
*/
char*
StringUtilities_PToCInPlace		(StringPtr		inoutString)
{
#if TARGET_API_MAC_OS8
	// traditional Mac OS has P2CStr
	return P2CStr(inoutString);
#else
	// Carbon offers a new p2cstrcpy function
	p2cstrcpy(STATIC_CAST(inoutString, char*), inoutString);
	return STATIC_CAST(inoutString, char*);
#endif
}// PToCInPlace


/*!
Conveniently tokenizes a string in the most common
way, by using spaces and tabs (white space) as token
delimiters and invoking StringUtilities_PTokenizeBy().
The resultant token set, and number of tokens in the
set, are returned.

If any problems occur, the output handle will be null.

(1.0)
*/
void
StringUtilities_PTokenize	(StringPtr		inString,
							 UInt16*		outArgC,
							 TokenSetRef*	outArgV)
{
	StringUtilities_PTokenizeBy(inString, "\p \t", outArgC, outArgV);
}// PTokenize


/*!
Takes a Pascal string, removes all characters resident
in the string "inTokenSeparators", and places the
resultant substrings, as StringHandles, into a Token Set.
The number of StringHandles in the set is returned in
"outArgC".

If any problems occur, the output handle will be null.

(1.0)
*/
void
StringUtilities_PTokenizeBy		(StringPtr		inString,
								 StringPtr		inTokenSeparators,
								 UInt16*		outArgC,
								 TokenSetRef*	outArgV)
{
	if ((outArgC != null) && (outArgV != null))
	{
		Str255				buffer,
							token;
		register UInt16		i = 0;
		OSErr				error = noErr;
		
		
		*outArgC = 0;
		*outArgV = (TokenSetRef)Memory_NewHandle(sizeof(StringHandle));
		error = MemError();
		if ((error == noErr) && (*outArgV != null))
		{
			Handle				handle = refToHandle(*outArgV);
			register UInt16		j = 0;
			Boolean				doneLoop = false;
			
			
			// walk the  buffer, creating tokens as theyre found and shortening the buffer...
			PLstrcpy(buffer, inString);
			while (StrLength(buffer) > 0)
			{
				// ...find the next token separator
				while ((!isAmong(buffer[i + 1], inTokenSeparators)) && (i < StrLength(buffer))) i++;
				
				// ...copy the token into a separate string
				token[0] = '\0', PLstrncpy(token, buffer, i);
				
				// ...ignore multiple token separators (such as multiple whitespace characters)
				while (isAmong(buffer[i + 1], inTokenSeparators) && (i < StrLength(buffer))) i++;
				
				// ...remove all the characters parsed thus far from the buffer
				StringUtilities_PClipHead(buffer, i), i = 0;
				
				// ...if a token has been found, place it in the resultant token set
				if (StrLength(token) > 0)
				{
					(*outArgC)++;
					SetHandleSize(handle, (*outArgC) * sizeof(StringHandle));
					if ((error = MemError()) == noErr)
					{
						HLock(handle);
						((StringHandle*)*handle)[(*outArgC) - 1] = NewString(token);
						HUnlock(handle);
					}
					if (error != noErr) break;
				}
			}
			if (error != noErr)
			{
				// returning all the tokens failed - clean up and return zero
				StringUtilities_PDisposeTokens(*outArgC, outArgV);
				*outArgC = 0;
			}
		}
	}
}// PTokenizeBy


/*!
To use the font metrics information of the current
graphics port, along with an acceptable maximum
string width, to automatically truncate the given
string if needed, use this method.  This routine
relies on the StringWidth() and CharWidth() methods
from QuickDraw.

If the string is truncated, it is modified to
include an ellipsis () character.  The position
of the ellipsis depends on the truncation method;
"kStringUtilities_TruncateAtEnd" places it as the
last character in the string, for example.

If truncation was necessary, "true" is returned;
otherwise, "false" is returned.

(1.0)
*/
Boolean
StringUtilities_PTruncate	(StringPtr								inoutString,
							 UInt32									inAcceptableMaximumWidth,
							 StringUtilitiesTruncationMethod		inTruncationMethod)
{
	Boolean				result = (StringWidth(inoutString) > inAcceptableMaximumWidth);
	register UInt16		i = 0;
	
	
	if (result)
	{
		UInt16		max = inAcceptableMaximumWidth - CharWidth('');
		
		
		if (StrLength(inoutString) > 1) switch (inTruncationMethod)
		{
			case kStringUtilities_TruncateAtBeginning:
				{
					Str255		rightSegment;
					
					
					PLstrcpy(rightSegment, inoutString);
					while (StringWidth(rightSegment) > max)
					{
						StringUtilities_PClipHead(rightSegment, 1);
					}
					PLstrcpy(inoutString, rightSegment);
					StringUtilities_PInsert(inoutString, "\p");
				}
				break;
			
			case kStringUtilities_TruncateAtMiddle:
				{
					Str255		leftSegment,
								rightSegment;
					UInt16		leftSize = StrLength(inoutString) / 2;
					UInt16		leftWidth = 0;
					UInt16		halfMax = inAcceptableMaximumWidth / 2;
					
					
					PLstrncpy(leftSegment, inoutString, leftSize);
					PLstrcpy(rightSegment, inoutString);
					StringUtilities_PClipHead(rightSegment, leftSize);
					while (StringWidth(leftSegment) > halfMax)
					{
						leftSegment[--(leftSegment[0])] = '';
					}
					leftWidth = StringWidth(leftSegment);
					while (StringWidth(rightSegment) > (max - leftWidth))
					{
						StringUtilities_PClipHead(rightSegment, 1);
					}
					PLstrcpy(inoutString, leftSegment);
					PLstrcat(inoutString, rightSegment);
				}
				break;
			
			case kStringUtilities_TruncateAtEnd:
			default:
				while (StringWidth(inoutString) > max)
				{
					inoutString[--(inoutString[0])] = '';
				}
				break;
		}
		else if (StrLength(inoutString) == 1) inoutString[1] = '';
	}
	
	return result;
}// PTruncate


//
// internal methods
//

/*!
To determine if the specified character
is one of the characters in the given
string, use this method.

(1.0)
*/
static Boolean
isAmong		(CharParameter const	inCharacter,
			 StringPtr const		inCharacterList)
{
	Boolean		result = false;
	
	
	if (inCharacterList != null)
	{
		register SInt16		i = 0;
		
		
		for (i = 1; ((i <= StrLength(inCharacterList)) && (!result)); i++)
		{
			result = (inCharacter == inCharacterList[i]);
		}
	}
	return result;
}// isAmong


/*!
To acquire a Mac OS Handle representation
of a token set, use this method.

(1.0)
*/
static Handle
refToHandle		(TokenSetRef		inRef)
{
	return ((Handle)inRef);
}// refToHandle