/*###############################################################

	TelnetDescriptors.h
	
	Contains enumerations that uniquely identify MacTelnets
	objects, such as windows, linked lists, arrays and dialogs.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#ifndef __TELNETDESCRIPTORS__
#define __TELNETDESCRIPTORS__

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// MacTelnet includes
#include "Panel.h"
#include "TelnetThreads.h"



/*###############################################################
	WINDOW-RELATED CONSTANTS
###############################################################*/

#define kInFront	((WindowRef)-1)

// MacTelnet window descriptors, used with Window Info.
// These are all to be considered of the type
// "WindowInfoDescriptor", defined in "InterfaceLib.h".
// See handleUpdate(), internal to "EventLoop.c", for
// code which handles updates on a descriptor basis; if
// you add a descriptor here, you should add it to the
// update event handler to ensure that windows get
// repainted automatically for you.
enum
{
	kTelnetWindowDescriptorClipboard		= kInvalidWindowInfoDescriptor + 1,
	kTelnetWindowDescriptorCommandLine		= kInvalidWindowInfoDescriptor + 2,
	kTelnetWindowDescriptorKeypadWindoid	= kInvalidWindowInfoDescriptor + 3,
	kTelnetWindowDescriptorFunctionWindoid	= kInvalidWindowInfoDescriptor + 4,
	kTelnetWindowDescriptorControlWindoid	= kInvalidWindowInfoDescriptor + 5,
	kTelnetWindowDescriptorInspectorWindoid	= kInvalidWindowInfoDescriptor + 6,
	kTelnetWindowDescriptorAbout			= kInvalidWindowInfoDescriptor + 7,
	kTelnetWindowDescriptorAnyTerminal		= kInvalidWindowInfoDescriptor + 8,
	kTelnetWindowDescriptorAssistant		= kInvalidWindowInfoDescriptor + 9,
	kTelnetDialogDescriptorConfigurations	= kInvalidWindowInfoDescriptor + 10,
	kTelnetWindowDescriptorDebuggingConsole	= kInvalidWindowInfoDescriptor + 11,
	kTelnetDialogDescriptorEditANSIColors	= kInvalidWindowInfoDescriptor + 12,
	kTelnetDialogDescriptorEditLogInMessage	= kInvalidWindowInfoDescriptor + 13,
	kTelnetDialogDescriptorFormat			= kInvalidWindowInfoDescriptor + 14,
	kTelnetDialogDescriptorMacroSetup		= kInvalidWindowInfoDescriptor + 15,
	kTelnetDialogDescriptorNameNewFavorite	= kInvalidWindowInfoDescriptor + 16,
	kTelnetDialogDescriptorNewSessions		= kInvalidWindowInfoDescriptor + 17,
	kTelnetDialogDescriptorPreferences		= kInvalidWindowInfoDescriptor + 18,
	kTelnetDialogDescriptorProxyServerSetup	= kInvalidWindowInfoDescriptor + 19,
	kTelnetDialogDescriptorRemapKeyboard	= kInvalidWindowInfoDescriptor + 20,
	kTelnetDialogDescriptorScreenSize		= kInvalidWindowInfoDescriptor + 21,
	kTelnetDialogDescriptorSessionEditor	= kInvalidWindowInfoDescriptor + 22,
	kTelnetDialogDescriptorSetWindowTitle	= kInvalidWindowInfoDescriptor + 23
};

// possible return values from GetWindowKind() on MacTelnet windows
typedef SInt16 TelnetWindowKind;
enum
{
	NO_WINDOW = 0,
	WIN_MODELESS = kApplicationWindowKind + 1,		// Modeless windows
	WIN_CONSOLE = kApplicationWindowKind + 2,		// Console window
	WIN_LOG = kApplicationWindowKind + 3,			// FTP log window
	WIN_CNXN = kApplicationWindowKind + 4,			// Connections
	WIN_TEK = kApplicationWindowKind + 5,			// TEK windows
	WIN_ICRG = kApplicationWindowKind + 6			// Interactive Color Raster Graphics windows
};

/*###############################################################
	UNDOABLE OPERATION CONTEXT IDENTIFIERS
###############################################################*/

// Undoable operation context IDs, used with Undoable objects.
// These are all to be considered of the type
// "UndoableContextIdentifier", defined in "ContextsLib.h".
enum
{
	kUndoableContextIdentifierTerminalDimensionChanges = 1,
	kUndoableContextIdentifierTerminalFormatChanges = 2,
	kUndoableContextIdentifierTerminalFontSizeChanges = 3
};

/*###############################################################
	CHAIN TYPES
###############################################################*/

// Chain descriptors, used with Chain objects.  These are all
// to be considered of the type "ChainDescriptor", defined in
// "InterfaceLib.h".  Using descriptors helps avoid dangerous
// situations where a void* is assumed to be data of a certain
// type; if you assign a descriptor to a chain and test for
// that descriptor, you can be more certain of the type of the
// data in the chain.
enum
{
	kTelnetChainDescriptorConnectionList			= kChainInvalidDescriptor + 1,
	kTelnetChainDescriptorThreadsList				= kChainInvalidDescriptor + 2,
	kTelnetChainDescriptorAliasList					= kChainInvalidDescriptor + 3
};

/*###############################################################
	THREAD TYPES
###############################################################*/

// Thread descriptors, used with MacTelnet Thread objects.
// These are all to be considered of the type
// "TelnetThreadDescriptor", defined in "TelnetThreads.h".
enum
{
	kTelnetThreadDescriptorCursorAnimation			= kTelnetThreadDescriptorMainThread + 1,
	kTelnetThreadDescriptorDelay					= kTelnetThreadDescriptorMainThread + 2,
	kTelnetThreadDescriptorFontListRebuilding		= kTelnetThreadDescriptorMainThread + 3,
	kTelnetThreadDescriptorGeneralIdling			= kTelnetThreadDescriptorMainThread + 4,
	kTelnetThreadDescriptorMovingStarsHandling		= kTelnetThreadDescriptorMainThread + 5,
	kTelnetThreadDescriptorQuadDamagePulseHandling	= kTelnetThreadDescriptorMainThread + 6,
	kTelnetThreadDescriptorScriptsMenuRebuilding	= kTelnetThreadDescriptorMainThread + 7
};

/*###############################################################
	PANEL TYPES
###############################################################*/

// Panel descriptors, which help to identify panels when given
// only an opaque reference to one.  See "Panel.h".
enum
{
	kTelnetPanelDescriptorPreferencesConfigurations	= kInvalidPanelDescriptor + 1,
	kTelnetPanelDescriptorPreferencesFileTransfers	= kInvalidPanelDescriptor + 2,
	kTelnetPanelDescriptorPreferencesGeneral		= kInvalidPanelDescriptor + 3,
	kTelnetPanelDescriptorPreferencesMacros			= kInvalidPanelDescriptor + 4,
	kTelnetPanelDescriptorPreferencesScripts		= kInvalidPanelDescriptor + 5,
	kTelnetPanelDescriptorSetupAssistantIntro		= kInvalidPanelDescriptor + 6
};

/*###############################################################
	SPECIAL GLOBAL FLAGS
###############################################################*/

// pass one of these to FlagManager_Test()
enum
{
	// All of these must be unique.  Note that, thanks to the excellent design
	// of the Flag Manager, there is no real limit as to the number of flags
	// you can define here.  There is also no efficiency concern - it is not
	// necessary to use multiples of 16 or 32 flags, for example.
	kFlagAppearance1_1API		= kFlagManager_FirstValidFlag,			// is Appearance 1.1 or later available?
	kFlagAppleGuideFiles		= (kFlagManager_FirstValidFlag + 1),	// is there a way to open MacTelnet Help directly?
	kFlagAppleScriptRecording	= (kFlagManager_FirstValidFlag + 2),	// is the user recording events into a script?
	kFlagCarbonControlManager	= (kFlagManager_FirstValidFlag + 3),	// is the Mac OS 8.5 Control Manager available?
	kFlagCarbonMenuManager		= (kFlagManager_FirstValidFlag + 4),	// is the Mac OS 8.5 Menu Manager available?
	kFlagCarbonWindowManager	= (kFlagManager_FirstValidFlag + 5),	// is the Mac OS 8.5 Window Manager available?
	kFlagContextualMenus		= (kFlagManager_FirstValidFlag + 6),	// is the Contextual Menu Extension available?
	kFlagDrawSprocket			= (kFlagManager_FirstValidFlag + 7),	// is DrawSprocket available?
	kFlagIconServices			= (kFlagManager_FirstValidFlag + 8),	// is Icon Services available?
	kFlagMacBinaryIIEnabled		= (kFlagManager_FirstValidFlag + 9),	// are MacBinary transfers enabled?
	kFlagMovingStars			= (kFlagManager_FirstValidFlag + 10),	// is a moving stars animation turned on?
	kFlagNavigationServices		= (kFlagManager_FirstValidFlag + 11),	// is Navigation Services (any version) installed?
	kFlagNetworkSuspended		= (kFlagManager_FirstValidFlag + 12),	// is all network activity suspended?
	kFlagOS8_5API				= (kFlagManager_FirstValidFlag + 13),	// is Mac OS 8.5 or later in use?
	kFlagOS8_6API				= (kFlagManager_FirstValidFlag + 14),	// is Mac OS 8.6 or later in use?
	kFlagOS9_0API				= (kFlagManager_FirstValidFlag + 15),	// is Mac OS 9.0 or later in use?
	kFlagOS9_1API				= (kFlagManager_FirstValidFlag + 16),	// is Mac OS 9.1 or later in use?
	kFlagOSXClassic				= (kFlagManager_FirstValidFlag + 17),	// is MacTelnet running on Mac OS X, in Classic?
	kFlagQuadDamage				= (kFlagManager_FirstValidFlag + 18),	// is the Quad Damage feature turned on?
	kFlagQuickTime				= (kFlagManager_FirstValidFlag + 19),	// is QuickTime (any version) installed?
	kFlagSuspended				= (kFlagManager_FirstValidFlag + 20),	// is MacTelnets process in the background?
	kFlagThreadManager			= (kFlagManager_FirstValidFlag + 21),	// is the Thread Manager (any version) installed?
	kFlagTextServicesManager	= (kFlagManager_FirstValidFlag + 22),	// is TSM (any version) properly initialized?
	kLastFlag					= kFlagTextServicesManager,				// for convenience only (see below)
	// The following are variables used in main(); they are here so that they
	// take up as little global memory as possible (dont add "kFlagGlobal..."
	// constants here unless they are used by "Main.c"!!!).
	kFlagGlobalHaveNotifiedLowMemory	= kLastFlag + 1,	// has a low memory warning been shown to the user?
	kFlagGlobalMemoryIsOK				= kLastFlag + 2,	// is there sufficient free space in the heap zone?
	kFlagGlobalMainLoopDone				= kLastFlag + 3		// main event loop termination flag
};

/*###############################################################
	MEMORY ERROR CODES
###############################################################*/

typedef SInt32 InternalTelnetMemoryError;
enum
{
	// these are used to help identify where memory errors take place
	kInternalTelnetMemoryErrorClipboardTextToScrap = 1,
	kInternalTelnetMemoryErrorConnectionOpenFromMenuItem = 2,
	kInternalTelnetMemoryErrorDNRLookup = 3,
	kInternalTelnetMemoryErrorRealScreenCreation = 4,
	kInternalTelnetMemoryErrorSafeArrayCreation = 5,
	kInternalTelnetMemoryErrorVirtualScreenResize = 6
};

/*###############################################################
	MISCELLANEOUS CONSTANTS
###############################################################*/

// Navigation Services preference keys
enum
{
	kTelnetNavPrefKeyGenericOpenFile = 1,
	kTelnetNavPrefKeyGenericSaveFile = 2,
	kTelnetNavPrefKeyGenericChooseFolder = 3,
	kTelnetNavPrefKeyChooseTextEditor = 4,
	kTelnetNavPrefKeyMacroStuff = 5
};

// operation-governing constants (I really hate these, but its gonna be damned hard to avoid em)
enum
{
	kMaximumAllowedTelnetConnections = 20, // the maximum number of sessions
	kMaximumAllowedGraphicsWindows = 30 // the maximum number of graphics
};

#endif