/*###############################################################

	TelnetMenus.h
	
	This file lists methods for menu management.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#ifndef __TELNETMENUS__
#define __TELNETMENUS__

// Mac includes
#include <ColorPicker.h>
#include <Menus.h>
#include <QuickDraw.h>

// library includes
#ifndef __INTERFACELIB__
#	include "InterfaceLib.h"
#endif

// MacTelnet includes
#include "PreferencesData.h"
#include "ScreenManager.h"



// constants
typedef SInt16	TelnetMenuMarkType;
enum
{
	kMenuMarkTypeEnableIcon = -3, // a command, not a mark...
	kMenuMarkTypeDisableIcon = -2, // a command, not a mark...
	kMenuMarkTypeDispose = -1, // a command, not a mark...
	kMenuMarkTypeConnectionLive = 0,
	kMenuMarkTypeConnectionOpening = 1,
	kMenuMarkTypeConnectionDead = 2
};



/**
Menu Command State Tracker Method

Every Telnet menu command has a state tracking method
attached to it as its reference constant.  Each such
method reports on the proper state for one or more
menu items.  Then, prior to the user selecting from
the menu bar or opening a contextual menu, every
commands state tracking method will be invoked to
determine the correct state for each menu item (even
menu item text, checkmarks, etc.).  Any aspect of an
items state that could change, such as its mark,
active state, icon or text, should be updated by an
appropriate routine.  If a particular item never, for
example, changes its item text, there is no point in
explicitly setting the item text each time, however.
Usually, enabled and marked states of items will be
the most common features to change.

As a convenience, and for efficiency, these routines
can report back whether an item is to be enabled, and
the caller will set the enabled state of the item,
automatically setting the correct enabled state for
the entire menu containing the item.  Any other change,
such as the text of a menu item, is not possible to
report back, and should be performed directly by the
state tracking routine.

Note that the command ID should be used by any routine
that can handle multiple item states.  Although the
command ID can be used to perform menu item changes
using the routines listed in this file, it is more
efficient to use the provided menu and menu item
information.
*/
typedef Boolean (*MenuCommandStateTrackerProcPtr)	(UInt32 const		inCommandID,
													 MenuRef			inMenu,
													 MenuItemIndex		inItem);
#define InvokeMenuCommandStateTrackerProc(userRoutine, inCommandID, inMenu, inItem) \
			(*userRoutine)((inCommandID), (inMenu), (inItem))



//
// public methods
//

void
	Menus_Init										(void);

void
	Menus_Done										(void);

void	AddConnectionsMenuItem						(ConnectionDataPtr				inConnectionDataPtr,
													 StringPtr						inHostName,
													 StringPtr						inWindowName);

void	addMonoSpacedFonts							(MenuRef						theMenu);

void	CheckNational								(short							choice);

void	CheckOpSpecSubmenu							(void);

void
	Menus_CheckSimpleTelnet							(short							inSimpleTelnetOn);

void	CloseCaptureFile							(TelnetScreenID const			inScreenID);

void
	Menus_EnableEditMenu							(void);

void
	Menus_FillInColorPickerMenuItemInfo				(PickerMenuItemInfo*			outPickerInfo);

MenuRef
	Menus_GetFontMenu								(void);

MenuItemIndex
	Menus_GetItemIndexByText						(MenuRef						inMenu,
													 Str255							inItemText);

MenuItemIndex	GetMenuAndMenuItemIndexByCommandID	(UInt32 const					inCommandID,
													 MenuRef*						outMenu);

void	GetMenusAndMenuItemIndicesByCommandID		(UInt32 const					inCommandID,
													 MenuRef*						outMenuPrimary,
													 MenuRef*						outMenuSecondary,
													 MenuItemIndex*					outMenuItemIndexPrimary,
													 MenuItemIndex*					outMenuItemIndexSecondary);

Boolean	GetMenuItemCheckedByCommandID				(UInt32 const					inCommandID);

Boolean	GetMenuItemEnabledByCommandID				(UInt32 const					inCommandID);

SInt16	GetMenuItemIndexByCommandID					(UInt32 const					inCommandID);

UInt8	GetMenuItemModifiersByCommandID				(UInt32 const					inCommandID);

void	GetMenuItemTextByCommandID					(UInt32 const					inCommandID,
													 Str255							outText);

MenuRef
	Menus_GetScriptsMenu							(void);

MenuRef
	Menus_GetTranslationMenu						(void);

void
	Menus_GetUniqueItemText							(MenuRef						inMenu,
													 Str255							inoutItemText);

Boolean
	Menus_HandleCommand								(long							inMenuEventResult,
													 EventModifiers					inModifiers);

Boolean
	Menus_HandleCommandByID							(UInt32							inCommandID,
													 EventModifiers					inModifiers);

Boolean
	Menus_IsItemUnique								(MenuRef						inMenu,
													 Str255							inItemText);

MenuRef
	Menus_NewMenuThemeSavvy							(short							inMenuID,
													 ConstStr255Param				inMenuTitle);

void
	Menus_RemoveConnectionsMenuItem					(ConnectionDataPtr				inConnectionDataPtr);

void
	Menus_Service									(void);

void
	Menus_SetFontMenuAvailable						(Boolean const					inIsAvailable);

void
	Menus_SetScriptsMenuAvailable					(Boolean const					inIsAvailable);

void
	Menus_SetUpItemStates							(void);

void
	Menus_SetKeyEquivalents							(Boolean const					inMenusHaveKeyEquivalents,
													 Boolean const					inVeryFirstCall);

void	SetMenuItemAdjustmentProc					(MenuRef						inMenu,
													 MenuItemIndex const			inItemIndex,
													 MenuCommandStateTrackerProcPtr	inProc);

void	SetMenuItemAdjustmentProcByCommandID		(UInt32 const					inCommandID,
													 MenuCommandStateTrackerProcPtr	inProc);

void	SetMenuItemCheckedByCommandID				(UInt32 const					inCommandID,
													 Boolean const					inIsChecked);

void	SetMenuItemEnabledByCommandID				(UInt32 const					inCommandID,
													 Boolean const					inIsEnabled);

void	SetMenuItemModifiersByCommandID				(UInt32 const					inCommandID,
													 UInt8 const					inNewModifiers);

void	SetMenuItemTextByCommandID					(UInt32 const					inCommandID,
													 ConstStr255Param				inText);

void	SetMenuMarkFromScreen						(TelnetMenuMarkType				inMarkType,
													 TelnetScreenID const			inScreenID);

void	SetMenuTextFromScreen						(StringPtr						inText,
													 TelnetScreenID const			inScreenID);

void
	Menus_SetNumberOfSpecialScriptMenuItems			(UInt16 const					inNumber);

long	SetOtherFontSize							(short							currentSize);

void	SetScreenDimensions							(TelnetScreenID const			inScreenID);

void
	Menus_SetSuspended								(Boolean const					inIsSuspended);

#endif