/*###############################################################

	TelnetPrinting.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdio>
#	include <cstring>
#else
#	include <stdio.h>
#	include <string.h>
#endif

// Mac includes
#include <ControlDefinitions.h>
#include <Printing.h>
#include <Resources.h>

// library includes
#include "InterfaceLib.h"
#include "PrintingLib.h"

// resource includes
#include "DialogResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "ConnectionData.h"
#include "Console.h"
#include "DialogUtilities.h"
#include "DialogItemIndices.h"
#include "EventLoop.h"
#include "RealScreen.h"
#include "NetEvents.h"
#include "ProgressDialog.h"
#include "rgmp.proto.h"
#include "ScreenManager.h"
#include "TektronixRealGraphics.h"
#include "TelnetErrors.h"
#include "TelnetHeader.h"
#include "TelnetPrinting.h"
#include "TektronixVirtualGraphics.h"
#include "Terminal.h"
#include "vsdata.h"



// constants
enum
{
	kTHPrintResourceType = FOUR_CHAR_CODE('THPr'),
	kTHPrintResourceID = 128
};
enum
{
	ascLF = 10,
	ascFF = 12,
	ascCR = 13
};

// variables
static UniversalPrintRecordRef		gPrintRecordInUse = null;

// internal methods
static void		printIdle			(void);
static void		printGraph			(SInt16						dnum);
static void		printText			(const TelnetScreenID		vs,
									 Str255						Title);



//
// public methods
//

/*!
To create a new Universal Print Record and
either fill it in with default values or
any saved values from the preferences file,
use this method.  You dispose of the record
later with Printing_DisposeRecord().

Always call this routine prior to starting
a print job, from within a Printing_Init()-
Printing_Done() loop, so that the users
Page Setup and Print settings are used, and
so that the default print idle procedure is
able to let the user cancel the print job.

(3.0)
*/
UniversalPrintRecordRef
TelnetPrinting_GetNewPrintRecord	(void)
{
	UniversalPrintRecordRef			result = null;
	SavedUniversalPrintRecordPtr	printDataPtr = (SavedUniversalPrintRecordPtr)Memory_NewPtr
																				(sizeof(SavedUniversalPrintRecord));
	
	
	if (printDataPtr != null)
	{
		OSStatus		error = noErr;
		
		
		// create a new print record, scanning for a resource
		result = Printing_NewRecord();
		
		printDataPtr->architecture = kUniversalPrintingArchitectureTraditional;
		printDataPtr->data.traditional.storageTHPrint = GetResource(kTHPrintResourceType, kTHPrintResourceID);
		error = ResError();
		if ((printDataPtr->data.traditional.storageTHPrint != null) && (error == noErr))
		{
			// update the opaque data to match the retrieved resource
			Printing_CopyFromSaved(result, printDataPtr);
			ReleaseResource(printDataPtr->data.traditional.storageTHPrint);
		}
		else
		{
			// create a new default print record and update the opaque data to match it
			printDataPtr->data.traditional.storageTHPrint = Memory_NewHandle(sizeof(TPrint));
			Printing_DefaultSaved(result, printDataPtr);
			Printing_CopyFromSaved(result, printDataPtr);
			Memory_DisposeHandle(&printDataPtr->data.traditional.storageTHPrint);
		}
		Memory_DisposePtr((Ptr*)&printDataPtr);
	}
	
	gPrintRecordInUse = result;
	return result;
}// GetNewPrintRecord


/*!
To display the standard Page Setup dialog box
and store any changes in a resource file, use
this method.  The dialogs settings are initially
based on the contents of Telnets printing
resource from the specified source file.  If no
resource can be found, then the users changes
(if any) are stored in the given resource file as
a new resource; otherwise, the original resource
is merely updated.

Could portions of this go in "Preferences.c"?

(3.0)
*/
void
TelnetPrinting_PageSetup		(const SInt16		inResourceFileRefNum)
{
	SInt16		oldResourceFile = CurResFile();
	OSStatus	error = noErr;
	
	
	DeactivateFrontmostWindow();
	Cursors_UseArrow();
	
	{
		SavedUniversalPrintRecordPtr	printDataPtr = (SavedUniversalPrintRecordPtr)Memory_NewPtr
																				(sizeof(SavedUniversalPrintRecord));
		
		
		if (printDataPtr != null)
		{
			UniversalPrintRecordRef		printRecordRef = null;
			SInt16						printingResourceFile = 0;
			Boolean						resourceExists = false;
			
			
			// Telnet currently saves and restores data in the traditional (THPrint) format
			printDataPtr->architecture = kUniversalPrintingArchitectureTraditional;
			
			Printing_Init();
			printingResourceFile = CurResFile();
			printRecordRef = Printing_NewRecord();
			
			UseResFile(inResourceFileRefNum);
			resourceExists = (Count1Resources(kTHPrintResourceType) > 0);
			
			if (resourceExists)
			{
				// retrieve the resource and update the opaque data to match it
				Console_WriteLine("print data exists on disk");
				printDataPtr->data.traditional.storageTHPrint = Get1Resource(kTHPrintResourceType, kTHPrintResourceID);
				error = ResError();
			}
			else
			{
				// create a new default print record and update the opaque data to match it
				Console_WriteLine("print data does not exist on disk - creating new print data");
				printDataPtr->data.traditional.storageTHPrint = Memory_NewHandle(sizeof(TPrint));
				error = MemError();
				if (error == noErr)
				{
					UseResFile(printingResourceFile);
					Printing_DefaultSaved(printRecordRef, printDataPtr); // defaults the print record created above
					error = Printing_GetLastResult(printRecordRef);
				}
			}
			
			// fill in an opaque printing record using whatever data is on disk (this can
			// even create a Carbonized printing record from a legacy version, thanks to
			// the power of UniversalPrintingLib!)
			if (error == noErr)
			{
				Console_WriteLine("reading data (either new or existing) into memory");
				UseResFile(printingResourceFile);
				Printing_CopyFromSaved(printRecordRef, printDataPtr);
				error = Printing_GetLastResult(printRecordRef);
			}
			
			if (error == noErr)
			{
				// display a Page Setup dialog that contains the settings (new or saved)
				UseResFile(printingResourceFile);
				Console_WriteLine("Page Setup dialog display...");
				if (Printing_PageSetupDialogDisplay(printRecordRef))
				{
					// if the user made changes, copy them back
					Console_WriteLine("user made changes: writing data to disk");
					Printing_CopyToSaved(printRecordRef, printDataPtr);
					error = Printing_GetLastResult(printRecordRef);
					if (error == noErr)
					{
						UseResFile(inResourceFileRefNum);
						if (resourceExists)
						{
							Console_WriteLine("resource exists; updating...");
							ChangedResource(printDataPtr->data.traditional.storageTHPrint);
							error = ResError();
							if (error == noErr) UpdateResFile(inResourceFileRefNum);
							DetachResource(printDataPtr->data.traditional.storageTHPrint);
						}
						else
						{
							Console_WriteLine("resource does not exist; creating...");
							AddResource(printDataPtr->data.traditional.storageTHPrint,
										kTHPrintResourceType, kTHPrintResourceID, EMPTY_PSTRING);
							error = ResError();
							if (error == noErr) UpdateResFile(inResourceFileRefNum);
						}
					}
				}
				else error = userCanceledErr;
				
				// since any used resource gets detached, this call works whether a resource was used or not
				Memory_DisposeHandle(&printDataPtr->data.traditional.storageTHPrint);
			}
			
			Memory_DisposePtr((Ptr*)&printDataPtr);
			Printing_DisposeRecord(&printRecordRef);
			Printing_Done();
		}
		else error = memFullErr;
	}
	
	RestoreFrontmostWindow();
	
	Console_WriteValue("result", error);
	
	if (error == fnfErr)
	{
		// no printer configured
		OperationFailedAlert(siOpFailedNoPrinter, 0/* internal code */, error);
	}
	else Alert_ReportOSErr(error); // display a generic alert to the user if anything went wrong
	
	UseResFile(oldResourceFile);
}// PageSetup


/*!
Prints the selection in the frontmost window.
Selections of text or graphics are supported.

(2.6)
*/
void
TelnetPrinting_PrintSelection		(void)
{
	SInt16		i = 0;
	
	
	Printing_Init();
	
	if (TektronixRealGraphics_IsRealGraphicsWindow(GetRealFrontWindow(), &i)) printGraph(i);
	else if ((i = ScreenManager_GetWindowScreenID(GetRealFrontWindow())) > -1)
	{
		Str255		title;
		
		
		GetWTitle(GetRealFrontWindow(), title);
		printText(i, title);
	}
	
	Printing_Done();
}// PrintSelection


/*	printPages(ref,title,columns,charh,refNum) -
 *			Prints <charh> on the printer through opaque <ref> with <title>
 *			using <columns>.
 */
void
printPages		(UniversalPrintRecordRef	inPrintingRef,
				 StringPtr					Title,
				 SInt16						columns,
				 char**						charh,
				 SInt16						refNum)
{
	char*				charp = null;
	char*				start = null;
	long				charlen = 0L,
						count = null;
	SInt16				v = 0,
						h = 0,
						scount = 0,
						lines = 0,
						pgcount = 0,
						pgtmplen = 0,
						maxlines = 0;
	char				pagetemp[20];
	FMetricRec			info;
	SInt16				pFheight = 0,
						pFwidth = 0;
	unsigned char		buf[256];			// to build a line in from the file
	unsigned char		nextchar = '\0';	// next unprocessed char in file
	SInt16				rdy = 0;
	SInt16				indent = 0;			// indent to give reasonable left margin
	OSErr				sts = noErr;
	long				dummyCount = 0L;
	//char				tmp[100];			// only for debugging
	char				stupidarray[160];	// used for finding string widths
	ProgressDialogRef	idleDialog = null;
	Rect				pageBounds;
	
	
	// 3.0 - display a cool printing progress dialog
	{
		Str255			text;
		
		
		GetIndString(text, rStringsStatus, siStatusPrintingInProgress);
		idleDialog = ProgressDialog_New(text, true/* modal */);
	}
	ProgressDialog_Display(idleDialog);
	ProgressDialog_SetProgressIndicator(idleDialog, kTelnetProgressIndicatorIndeterminate);
	IdleControls(ProgressDialog_GetWindow(idleDialog));
	
	// Printer drivers automatically guess the document name by looking
	// at the title of the frontmost window.  Therefore, setting the title
	// of the progress dialog to be the given title will allow drivers to
	// correctly display the title of the printed data.  Since the idle
	// dialog has no title bar, this title will not appear on the screen.
	SetWTitle(ProgressDialog_GetWindow(idleDialog), Title);
	
	for (v=0; v<150; v++) stupidarray[v]='W';	/* Set up the width array */
	
	{
		SInt16		iHRes = 0,
					iVRes = 0;
		
		
		Printing_GetDeviceResolution(inPrintingRef, &iVRes, &iHRes);
		indent = (iHRes * 180) / 254;	/* 1.8 centimeters left margin */
	}
	{
		Rect		paperRect;
		
		
		Printing_GetPaperBounds(inPrintingRef, &paperRect);
		if (-paperRect.left > indent) indent = 0;
		else indent += paperRect.left;
	}
	
	TextFontByName("\pMonaco");
	TextFace(normal);
	v = 12;
	
	Printing_GetPageBounds(inPrintingRef, &pageBounds);
	do {
		TextSize(v);
		FontMetrics( &info );
		pFheight = FixRound( info.ascent + info.descent + info.leading);
		pFwidth  = FixRound( info.widMax);
		//pFwidth  = CharWidth('W');			/* Re-assign to largest char */
		v--;
	} while ( TextWidth(stupidarray,0,columns+1) > (pageBounds.right - pageBounds.left - indent));
	//sprintf (tmp,"Fheight:%d, Fwidth:%d, TextSize:%d",pFheight,pFwidth,v+1); //putln (tmp);

	if (charh!=null) {
		HLock(charh);
		start=charp=*charh;
		charlen= GetHandleSize(charh);
	} else {
		if ((sts=GetFPos(refNum, &charlen)) != noErr)
			{
			//sprintf(tmp,"GetFPos: ERROR %d",(SInt16)sts);
			//putln(tmp);
			}
		charlen-=3;					/* skip last 3 chars, as they are part of ESC seq */
		if ((sts=SetFPos(refNum,(SInt16) fsFromStart,null)) != noErr)	/* rewind to beginning of file */
			{ /*sprintf(tmp,"SetFPos: ERROR %d",sts); putln(tmp);*/ }
		start = (char *)buf;				/* always (the array doesn't move) */
		dummyCount=1;
		if ((sts=FSRead(refNum,&dummyCount,&nextchar)) != noErr)		/* get first char */
			{ /*sprintf(tmp,"FSRead: ERROR %d",sts); putln(tmp);*/ }
	}
	
	h=pageBounds.right - pageBounds.left - indent;	/* Get the width */
	v=pageBounds.bottom- pageBounds.top;			/* Get the height */

	maxlines = v/pFheight-1;
	pgcount = 0;
	while (count<charlen) {
		pgcount++;
		lines = 1;
		Printing_BeginPage(inPrintingRef, (Rect*)null);
		if ((sts=Printing_GetLastResult(inPrintingRef)) == noErr)
		{
			Printing_SetIdleProc(inPrintingRef, printIdle);
			//sprintf (tmp,"printing page:%d",pgcount); //putln(tmp);
	
			if (Title != null) {
				Str255 SessionNameString;
				GetIndString(SessionNameString, rStringsMiscellaneous, siSessionName);
				MoveTo( indent, lines*pFheight);
				DrawString(SessionNameString);
		
				DrawString( Title);
				sprintf(pagetemp, "Page %d", pgcount);
				pgtmplen=strlen(pagetemp);
				MoveTo( pageBounds.right-(pgtmplen * pFwidth)-1, lines++*pFheight);
				DrawText( pagetemp, 0, pgtmplen);
				lines++;					/* one blank line after title line */
			}
	
			if (charh!=null) {									/* print from handle */
				while (lines <= maxlines && count<charlen) {
					scount=0;
					while ((count<charlen) && (*charp++!='\n')) { count++; scount++; }
					MoveTo(indent,lines++*pFheight);
					if (scount>0)
						DrawText(start, 0, scount);
					count++;
					start=charp;
				}
			} else {											/* print from file */
				while (lines <= maxlines && count<charlen) {
					rdy = 0;
					scount = 0;
					while ((count<charlen) && (!rdy)) {
						if (scount > 250)
							nextchar = ascLF;
						switch (nextchar) {
							case ascCR:
								rdy=1;
								break;
							case ascLF:
								rdy=1;
								break;
							case ascFF:
								rdy=1;
								break;
							default:
								buf[scount++]=nextchar;
								count++;
								dummyCount=1;
								if ((sts=FSRead (refNum,&dummyCount,&nextchar)) != noErr)
									{ /*sprintf(tmp,"FSRead: ERROR %d",sts); putln(tmp);*/ }
								break;
						}
					}
					MoveTo(indent,lines*pFheight);
					if (scount>0)
						DrawText(start, 0, scount);
					if (nextchar==ascLF)
						lines++;						/* LF -> new line */
					if (nextchar==ascFF)
						lines = maxlines+1;				/* FF -> new page */
					dummyCount=1;
					if ((sts=FSRead (refNum,&dummyCount,&nextchar)) != noErr)
						{ /*sprintf(tmp,"FSRead: ERROR %d",sts); putln(tmp);*/ }
					count++;
				}
			}
			Printing_EndPage(inPrintingRef);
			if ((sts=Printing_GetLastResult(inPrintingRef)) != noErr) { /*sprintf(tmp,"ClosePage: ERROR %d",sts); putln(tmp);*/ }
		}
		else break;
	}
	if (charh != null) HUnlock(charh);
	
	if (sts != noErr)
	{
		InterfaceLibAlertRef	box = null;
		Str255					dialogText,
								helpText;
		
		
		GetIndString(dialogText, rStringsGeneralMessages, siPrintingFailed);
		GetIndString(helpText, rStringsGeneralMessages, siPrintingFailedHelpText);
		box = Alert_New();
		Alert_SetHelpButton(box, false);
		Alert_SetParamsFor(box, kAlert_StyleOK);
		Alert_SetText(box, dialogText, helpText);
		Alert_SetType(box, kAlertStopAlert);
		Alert_Display(box);
		Alert_Dispose(&box);
	}
	
	ProgressDialog_Dispose(&idleDialog);
	RestoreFrontmostWindow();
	Cursors_UseArrow();
}// printPages


//
// internal methods
//

/*	printGraph -	Print the graphics on the current window
 *		vg - which graphics window to print */

static void printGraph(SInt16 dnum)			/* Which drawing to print */
{
	SInt16 h,v;			/* used for centering (h=horiz. , v=vert.) */
	SInt16 wh,wv;			/* Window horiz and vert */
	Rect prRect;		/* the rectangle to print in */
	SInt16 j;				/* VG identifier for pass-through */
	UniversalPrintRecordRef		printRecordRef = null;
	
	
	Printing_Init();
	printRecordRef = TelnetPrinting_GetNewPrintRecord();
	
	{
		Boolean		printOK = false;
		
		
		if (Printing_GetMode() == kUniversalPrintingModeNormal)
		{
			DeactivateFrontmostWindow();
			Cursors_UseArrow();
			printOK = Printing_PrintJobDialogDisplay(printRecordRef);
			RestoreFrontmostWindow();
		}
		else
		{
			printOK = true;
			Printing_SetNumberOfCopies(printRecordRef, 1, false/* lock */);
		}
		
		if (printOK)
		{			/* Cancel the print if false */
			Printing_BeginDocument(printRecordRef);
			if (Printing_GetLastResult(printRecordRef) == noErr)	{				/* If we can't, then die */
				Printing_BeginPage(printRecordRef, (Rect*)null);				/* Open a page */
				Printing_SetIdleProc(printRecordRef, printIdle);
				
				Printing_GetPageBounds(printRecordRef, &prRect);
				h=prRect.right - prRect.left;	/* Get the width */
				v=prRect.bottom- prRect.top;	/* Get the height */
				if (3*h<4*v) {					/* Center the little bugger */
					wh = h;
					wv = (3 * h)/ 4;
					}
				else {							/* On the page rectangle */
					wv = v;
					wh = (4 * v)/ 3;
					}
		
				prRect.top  = (v- wv) /2;
				prRect.left = (h- wh) /2;
				prRect.bottom = prRect.top + wv;
				prRect.right = prRect.left + wh;

				j=VGnewwin(TEK_DEVICE_PICTURE,VGgetVS(dnum));		/* NCSA 2.5: fixed the print call */
				RGMPsize( &prRect );
				VGzcpy( dnum, j);				/* Love dat zm factr */
				VGredraw(dnum,j);				/* Copy the picture in i to j */
				VGclose(j);						/* OK, we're done, give it to someone else */
				Printing_EndPage(printRecordRef);			/* Page out.... */
				}

			Printing_EndDocument(printRecordRef);					/* Done with the document */
			//putln("Doc is closed... the printing begins....");
		}
	}
	Printing_DisposeRecord(&printRecordRef);
	Cursors_UseArrow();
}// printGraph


/*!
This routine services network events and watches
for user-initiated printing cancellation.  It is
called by the operating system repeatedly while
printing is in progress.

To avoid potentially serious driver problems,
this idle procedure always saves and restores
critical system state information, such as the
current resource file and graphics port.

(3.0)
*/
static void
printIdle	(void)
{
	SInt16		oldResFile = CurResFile();
	GrafPtr		oldPort = null;
	
	
	GetPort(&oldPort);
	
	if (EventLoop_IsNextCancel()) Printing_Cancel(gPrintRecordInUse);
	
	UseResFile(oldResFile); // restore this in case it somehow got changed
	SetPort(oldPort);
}// printIdle


/*	printText -	Print the text selected on the screen 
 *		vs - which vs to print from */
static void
printText	(const TelnetScreenID	vs,			/* Which screen to print */
			 StringPtr				Title)		/* The title of the printout (the session name) */
{
	Handle						charh = null;
	UniversalPrintRecordRef		printRecordRef = null;
	
	
	Printing_Init();
	printRecordRef = TelnetPrinting_GetNewPrintRecord();
	
	charh = RealScreen_GetSelectedTextAsNewHandle(vs, 0, 0/* flags */);				/* Get the characters to print */
	
	if (IsHandleValid(charh))
	{
		Boolean		printOK = false;
		
		
		if (Printing_GetMode() == kUniversalPrintingModeNormal)
		{
			DeactivateFrontmostWindow();
			Cursors_UseArrow();
			printOK = Printing_PrintJobDialogDisplay(printRecordRef);
			RestoreFrontmostWindow();
		}
		else
		{
			printOK = true;
			Printing_SetNumberOfCopies(printRecordRef, 1, false/* lock */);
		}
		
		if (printOK)
		{
			Printing_BeginDocument(printRecordRef);
			if (Printing_GetLastResult(printRecordRef) == noErr)
			{
				printPages(printRecordRef, Title, VSmaxwidth(vs), charh, (SInt16)-1);
				Printing_EndDocument(printRecordRef);
			}
		}
		else
		{
			// print was cancelled...
		}
		HPurge(charh);							/* Kill the printed chars */
		Cursors_UseArrow();
		Printing_DisposeRecord(&printRecordRef);
	}
}// printText