/*###############################################################

	TelnetThreads.h
	
	Telnet 3.0 is now multi-threaded!  This module defines a
	layer of abstraction that will (hopefully) allow migration
	to the preferred (preemptable) Multiprocessing Services API
	without screwing up current multi-threading code for Telnet.
	
	This file also lists the entry points to all Telnet threads.
	It is important to note the following:
	
	- Threads are memory-limited; if you exceed their allocated
	  space, try increasing the amount at thread construction
	  time (you may need to do a content-search on the constant
	  from "TelnetDescriptors.h" corresponding to the thread
	  you are editing).
	
	- After invoking a yielding routine from a cooperative
	  thread, you have to assume that another thread has gone
	  and trampled on your application context (for example,
	  details such as the current resource file and the current
	  graphics port should be set again, to ensure they remain
	  correct).
	
	- If you ever use preemptive threads (Mac OS X only), you
	  need to surround context-critical code within a critical
	  section, because you cannot otherwise be sure whether,
	  or when, the OS will switch you out to another thread
	  that might alter what you would otherwise assume to be
	  constant.  It is probably good thread coding practice
	  to define potentially critical sections anyway, even if
	  you are not using (or even do not intend to use) anything
	  but cooperative threads for your thread code.
	
	Telnet 3.0,  1998-2000 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1992 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#ifndef __TELNETTHREADS__
#define __TELNETTHREADS__

// Mac includes
#ifndef __THREADS__
#	include <Threads.h>
#endif



// constants
enum
{
	kTelnetThreadDescriptorMainThread = 0
};

// types
typedef struct { unsigned garbage; }**		TelnetThreadRef;
typedef SInt32								TelnetThreadDescriptor;



/*###############################################################
	INITIALIZING AND FINISHING WITH THREADS
###############################################################*/

// CALL THIS ROUTINE ONCE, BEFORE ANY OTHER THREAD ROUTINE
void
	Threads_Init					(void);

// CALL THIS ROUTINE AFTER YOU ARE PERMANENTLY DONE WITH THREADS
void
	Threads_Done					(void);

/*###############################################################
	CREATING THREADS
###############################################################*/

OSErr
	Threads_NewCoOp					(ThreadEntryUPP 			inThreadEntry,
									 void*						inThreadParam,
									 Size 						inStackSize,
									 TelnetThreadDescriptor		inThreadDescriptorOfNewThread);

OSErr
	Threads_NewCoOpWithOptions		(ThreadEntryUPP 			inThreadEntry,
									 void*						inThreadParam,
									 Size 						inStackSize,
									 ThreadOptions				inOptions,
									 TelnetThreadDescriptor		inThreadDescriptorOfNewThread);

/*###############################################################
	ACCESSING THREADS AND THREAD INFORMATION
###############################################################*/

TelnetThreadRef
	Threads_GetByDescriptor			(TelnetThreadDescriptor		inThreadDescriptor);

// AVOID IF AT ALL POSSIBLE
TelnetThreadRef
	Threads_GetByThreadManagerID	(ThreadID					inThreadID);

TelnetThreadRef
	Threads_GetCurrent				(void);

TelnetThreadDescriptor
	Threads_GetDescriptorOf			(TelnetThreadRef			inThread);

voidPtr
	Threads_GetResultOf				(TelnetThreadRef			inThread);

UInt32
	Threads_GetStackOf				(TelnetThreadRef			inThread);

// AVOID IF AT ALL POSSIBLE
ThreadID
	Threads_GetThreadManagerID		(TelnetThreadRef			inThread);

/*###############################################################
	THREAD CONTROL
###############################################################*/

// REQUIRED FOR SAFE PREEMPTIVE MULTITASKING; HARMLESS IF USED WITH CODE IN COOPERATIVE THREADS
void
	Threads_CriticalSectionBegin	(void);

// REQUIRED FOR SAFE PREEMPTIVE MULTITASKING; HARMLESS IF USED WITH CODE IN COOPERATIVE THREADS
void
	Threads_CriticalSectionEnd		(void);

void
	Threads_WhenTerminatedInvoke	(TelnetThreadRef			inThread,
									 ThreadTerminationUPP		inProc,
									 void*						inProcArgument);

void
	Threads_YieldTo					(TelnetThreadRef			inThread);

void
	Threads_YieldToAnother			(void);

/*###############################################################
	THREAD UTILITIES
###############################################################*/

Boolean
	Threads_ClickStopDelayTicks		(UInt32 const				inTicksOfInterruptibleDelay);

void
	Threads_DelayMinimumTicks		(UInt32 const				inTicksOfDelay);

/*###############################################################
	THREAD ENTRY POINTS
###############################################################*/

pascal voidPtr
	Threads_InitializeFontMenu		(void*						inMenuRefPosingAsVoidPointer);

pascal voidPtr
	Threads_InitializeScriptsMenu	(void*						inMenuRefPosingAsVoidPointer);

pascal voidPtr
	Threads_InterfaceIdle			(void*						inUnusedThreadParam);

pascal voidPtr
	Threads_MovingStars				(void*						inControlRefPosingAsVoidPointer);

pascal voidPtr
	Threads_QuadDamage				(void*						inUnusedThreadParam);

pascal voidPtr
	Threads_StandardIdle			(void*						inUnusedThreadParam);

#endif