/*###############################################################

	Terminal.c
	
	This file contains the control and interface calls for
	the NCSA Virtual Screen Kernel.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/
/*
 *	  by Gaige B. Paulsen
 *
 *	  VSredraw(w,x1,y1,x2,y2)			- redraw region for window w
 *	  VSwrite(w,ptr,len)				- write text @ptr, length len
 *	  VSkbsend(w,k,echo)				- send keycode k's rep. out window w (w/echo if req.)
 *	  VSgetline(w,y)					- get a ptr to w's line y
 *	  VSsetrgn(w,x1,y1,x2,y2)			- set local display region
 *	  VSgetrgn(w,&x1,&y1,&x2,&y2)		- returns set region
 *	  VSgetlines(w)						- Returns current # of lines
 *	  VSsetlines(w, lines)				- Sets the current # of lines to lines
 */

#define VSMASTER

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstring>
#else
#	include <string.h>
#endif

// Mac includes
#include <MacErrors.h>
#include <Navigation.h>
#include <PLStringFuncs.h>
#include <Script.h>
#include <StandardFile.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "Console.h"
#include "FileUtilities.h"
#include "FolderManager.h"
#include "RealScreen.h"
#include "VirtualScreenEmulation.h"
#include "vsdata.h"
#include "vskeys.h"
#include "VirtualScreenInternal.h"
#include "ConnectionData.h"
#include "DialogUtilities.h"
#include "MacLook.h"
#include "Preferences.h"
#include "TelnetErrors.h"
#include "EventLoop.h"
#include "Terminal.h"



// constants
enum
{
	kVirtualScreenInterfaceAllocatedLineCount = 48,
	kVirtualScreenInterfaceAllocatedColumnCount = 180	// no more than this many columns is allowed
};

// variables
static short		VSmax = 0, // max nr screens allowed
					VSinuse = 0; // nr screens actually in existence
VSscrndata*			VSscreens = null;

// internal methods
static short		setCurrentScreen	(TelnetScreenID const		inScreenID);



//
// public methods
//

/*!
Initializes the virtual screen and real screen modules.
You must call this routine before using terminal screens.

(2.6)
*/
SInt16
Terminal_Init		(SInt16 const		inMaximumNumberOfTerminals)
{
	SInt16		result = 0;
	
	
	RealScreen_Init();
	VSmax = inMaximumNumberOfTerminals;
	VSIwn = 0;
	VSscreens = (VSscrndata*)Memory_NewPtr(inMaximumNumberOfTerminals * sizeof(VSscrndata));
	if (VSscreens == null) result = -2;
	else
	{
		register SInt16		i = 0;
		
		
		for (i = 0; i < inMaximumNumberOfTerminals; i++)
		{
			VSscreens[i].loc = 0L;
			VSscreens[i].stat = 0;
		}
		result = 0;
	}
	return result;
}// Init


/*!
Creates a new virtual screen window and returns a number
identifying the screen.  The index is used throughout
the program, to track all resources attached to a
terminal window (e.g. TEK windows), and even the order
in which a connection shows up in the Window menu.

(2.6)
*/
TelnetScreenID
Terminal_NewWindow		(StringPtr			inFont,
						 SInt16 const		inFontSize,
						 RealScreenFlags	inFlags,
						 SInt16 const		inLineCountScrollBackBuffer,
						 SInt16 const		inLineCountVisibleRows,	// rows
						 SInt16 const		inMaximumColumnCount,
						 Boolean const		inForceLineSaving)
{
	SInt16		numLines = inLineCountVisibleRows - 1, // make this zero-based for internal use
				maxLines = inLineCountScrollBackBuffer;
  	Boolean		screensave = (inLineCountScrollBackBuffer > 0); // scrollback buffer?
	
	
	if (maxLines < kVirtualScreenInterfaceAllocatedLineCount)
		maxLines = kVirtualScreenInterfaceAllocatedLineCount;

	if (VSinuse >= VSmax)
	  /* too many screens in existence */
		return(-1);
	VSIwn = 0;
	while ((VSIwn < VSmax) && (VSscreens[VSIwn].stat == 1))
		VSIwn++;
	if (VSIwn >= VSmax)
	  /* shouldn't occur? */
		return(-1);
	
	/*
	*  Fill initial scrollback buffer and screen storage space.
	*
	*  Memory allocation rules:
	*  line->mem == 0 if not a memory allocation, line->mem == 1 if it is the first
	*     VSline in a block (indeterminate size, may be size == 1)
	*  
	*  attributes array is ALWAYS allocated as one block.  Internally represented and
	*  manipulated as a linked list of lines, but only one of the lines will have 
	*  line->mem == 1.  This list is always supposed to be circular (it is never
	*  extended, as attributes are never scrolled back).
	*
	*  scrollback and screen line buffer space is allocated in large blocks.  Each
	*  block will have line->mem == 1 if the pointer to that VSline is "free"able.
	*  This list will either be circular (which means it has reached its full size),
	*  or it will have a null next field at the end.  During scrolling, the end may
	*  be augmented until VSIw->numlines > VSIw->maxlines or we run out of memory.
	*  Typically allocate memory 100 lines at a time in two blocks, one is the VSline
	*  list, the other is the mem for the character storage.
	*
	*/
	
	/* All memory allocation for this function is done at once, to help damage control in 
	low memory situations */

	if ((VSscreens[VSIwn].loc = VSIw = (VSscrn *) Memory_NewPtr(sizeof(VSscrn))) == 0L)
		return(-2);
	
	VSIw->lines = numLines;
	//VSIw->lines = 23; CCP 2.7 set this from the start

	VSIw->linest = VSInewlinearray(VSIw->lines + 1);
	if (VSIw->linest == null)
 	{
		Memory_DisposePtr(&((Ptr)VSIw));
		VSscreens[VSIwn].loc = VSIw = null;
		return (-2);
	}	
	
	VSIw->attrst = VSInewattrlinearray(VSIw->lines + 1);
	if (VSIw->attrst == null)
 	{
		Memory_DisposePtr(&((Ptr)VSIw->linest));
		Memory_DisposePtr(&((Ptr)VSIw));
		VSscreens[VSIwn].loc = VSIw = null;
		return (-2);
	}	
	
	// 3.0 - now, this is extended to allow more than the previous 132 column limit
	VSIw->tabs = (char*)Memory_NewPtr(Terminal_GetAllocatedColumnCount()); // maximum columns
	if (VSIw->tabs == null)
	{
		Memory_DisposePtr(&((Ptr)VSIw->attrst));
		Memory_DisposePtr(&((Ptr)VSIw->linest));
		Memory_DisposePtr(&((Ptr)VSIw));
		VSscreens[VSIwn].loc = VSIw = null;
		return (-2);
	}
	VSIw->allwidth = (Terminal_GetAllocatedColumnCount() - 1); // always allocate max lines

	if (screensave)
		VSIw->buftop = VSInewlines(VSIw->lines + 1 + kVirtualScreenInterfaceAllocatedLineCount,1); /* screen lines plus some initial preallocated scrollback space */
	else
		VSIw->buftop = VSInewlines(VSIw->lines + 1,1); /* screen lines, no scrollback */
	if (VSIw->buftop == null)
	{
		Memory_DisposePtr(&((Ptr)VSIw->tabs));
		Memory_DisposePtr(&((Ptr)VSIw->attrst));
		Memory_DisposePtr(&((Ptr)VSIw->linest));
		Memory_DisposePtr(&((Ptr)VSIw));
		VSscreens[VSIwn].loc = VSIw = null;
		return(-2);
	}
	VSIw->linest[0] = VSIw->buftop;
	VSIw->attrst[0] = (VSattrlinePtr)VSInewlines(VSIw->lines + 1, 2);		/* new space for attributes (these are never scrolled back) */
	if (VSIw->attrst[0] == null)
	{
		VSIfreelinelist(VSIw->buftop);
		Memory_DisposePtr(&((Ptr)VSIw->tabs));
		Memory_DisposePtr(&((Ptr)VSIw->attrst));
		Memory_DisposePtr(&((Ptr)VSIw->linest));
		Memory_DisposePtr(&((Ptr)VSIw));
		VSscreens[VSIwn].loc = VSIw = null;
		return(-2);
	}

	VSIw->vistop = VSIw->scrntop = VSIw->buftop;	/* initial view = screen */

	VSIlistndx(VSIw->scrntop, VSIw->attrst[0]);	/* Set up screen arrays */
	
	VSIw->attrst[0]->prev = VSIw->attrst[VSIw->lines]; /* make attribute list circular, since it is never extended */
	VSIw->attrst[VSIw->lines]->next = VSIw->attrst[0];
	
	if (!screensave)
	{
		// make text line list circular to indicate no extensions
		VSIw->linest[0]->prev = VSIw->linest[VSIw->lines];
		VSIw->linest[VSIw->lines]->next = VSIw->linest[0];
	}
	
	VSIw->maxlines = maxLines;
	VSIw->numlines = 0;
	VSscreens[VSIwn].captureRN = 0;
	VSIw->id = 'VSCR';
	VSIw->maxwidth = inMaximumColumnCount - 1;
	VSIw->savelines = screensave;
	VSIw->forcesave = inForceLineSaving;		/* NCSA 2.5 */
	VSIw->attrib = 0;
	VSIw->Pattrib = -1; /* initially no saved attribute */
	VSIw->x = 0;
	VSIw->y = 0;
	VSIw->charset = 0;
	VSIw->G0 = 0;
	VSIw->G1 = 1;
	VSIw->DECAWM = 0;
	VSIw->DECCKM = 0;
	VSIw->DECPAM = 0;
	VSIw->DECORG = 0;
	VSIw->IRM = 0;
	VSIw->escflg = 0;
	VSIw->top = 0;
	VSIw->bottom = numLines;
	VSIw->parmptr = 0;
	VSIw->Rtop = 0;
	VSIw->Rleft = 0;
	VSIw->Rright = inMaximumColumnCount - 1;
	VSIw->Rbottom = numLines;
	VSIw->ESscroll = 1;
	VSIw->prredirect = 0;						/* LU */
	VSIw->prbuf = 0;							/* LU */
	VSIw->refNum = -1;							/* LU */
	VSIw->possibleForce = 0;
	VSIclrbuf();
	VSItabinit();
	VSscreens[VSIwn].stat = 1;
	VSinuse++;
	
	// create a new session window
	if (RealScreen_New(VSIwn, inMaximumColumnCount, numLines, inFont, inFontSize, inFlags) < 0)
	{
		// then there is a problem opening up the window
		BEEP;
		VSIwn = -1;
	}
	
	return(VSIwn);
}// New


/*!
Initiates a capture to a file for the specified screen.
If you dont specify a file to use for the capture, the
user is prompted for one.  If the capture begins
successfully, "true" is returned; otherwise, "false" is
returned.

When a capture is in progress, MacTelnet sets the window
proxy icon to be the capture file.

(3.0)
*/
Boolean
Terminal_FileCaptureBegin	(TelnetScreenID const	inScreenID,
							 FSSpec const*			inFileToCaptureToOrNull)
{
	FSSpec			captureFile;
	OSErr			error = noErr;
	Boolean			good = false;
	Boolean			result = false;
	
	
	// solicit a file unless provided with one
	if (inFileToCaptureToOrNull == null) good = Terminal_FileCaptureSaveDialog(&captureFile);
	else
	{
		error = FSMakeFSSpec(inFileToCaptureToOrNull->vRefNum, inFileToCaptureToOrNull->parID, inFileToCaptureToOrNull->name, &captureFile);
		good = ((error == noErr) || (error == fnfErr));
	}
	
	if (good)
	{
		OSType		captureFileCreator;
		Size		actualSize = 0L;
		
		
		// get the users Capture File Creator preference, if possible
		unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
									kTelnetPreferenceTag_CaptureFileCreator,
									sizeof(captureFileCreator), &captureFileCreator,
									&actualSize) == kTelnetPrefsResultCodeSuccess)
		{
			captureFileCreator = 'ttxt'; // default to SimpleText if a preference cant be found
		}
		
		error = FSpCreate(&captureFile, captureFileCreator, 'TEXT', GetScriptManagerVariable(smSysScript));
		if (error == dupFNErr)
		{
			// if a capture file already exists, try to delete it; if the delete fails, use a similar file name
			Console_WriteLine("capture file already exists - attempting to delete");
			if (FSpDelete(&captureFile) == noErr)
			{
				error = FSpCreate(&captureFile, captureFileCreator, 'TEXT', GetScriptManagerVariable(smSysScript));
			}
			else
			{
				Console_WriteLine("capture file deletion failed - using a similar file name");
				error = FileUtilities_PersistentCreate(&captureFile, captureFileCreator, 'TEXT', GetScriptManagerVariable(smSysScript));
			}
		}
			
		if (error != noErr) OperationFailedAlert(siOpFailedCantCreateFile, 500, error);
		else
		{		
			error = FSpOpenDF(&captureFile, fsRdWrPerm, &VSscreens[inScreenID].captureRN);
			if (error != noErr) OperationFailedAlert(siOpFailedCantOpenFile, 501, error);
			else
			{
				SetEOF(VSscreens[inScreenID].captureRN, (long)0);
				
				// on Mac OS 8.5 and beyond, set the window proxy icon appropriately for the file capture
				if (API_AVAILABLE(SetWindowProxyAlias))
				{
					AliasHandle		alias = null;
					
					
					// I cant remember if its a memory leak to forget about an alias or not...warning...
					if (NewAliasMinimal(&captureFile, &alias) == noErr)
					{
						WindowRef		window = ScreenManager_GetWindowOf(inScreenID);
						
						
						if (window != null)
						{
							(OSStatus)SetWindowProxyAlias(window, alias);
							if (API_AVAILABLE(SetWindowProxyCreatorAndType))
							{
								OSType		captureFileCreator;
								
								
								unless (Preferences_GetData(kTelnetPreferenceClassGeneral, kTelnetPreferenceTag_CaptureFileCreator,
															sizeof(captureFileCreator), &captureFileCreator, &actualSize) ==
										kTelnetPrefsResultCodeSuccess)
								{
									captureFileCreator = 'ttxt'; // assume default, if preference cant be found
								}
								(OSStatus)SetWindowProxyCreatorAndType(window, captureFileCreator, 'TEXT', captureFile.vRefNum);
							}
						}
					}
				}
				result = true;
			}
		}
	}
	
	return result;
}// FileCaptureBegin


/*!
Terminates the file capture for the specified screen,
closing the capture file.

(2.6)
*/
void
Terminal_FileCaptureEnd		(TelnetScreenID const		inScreenID)
{
	FSClose(VSscreens[inScreenID].captureRN);
	VSscreens[inScreenID].captureRN = 0;
	if (API_AVAILABLE(RemoveWindowProxy))
	{
		(OSStatus)RemoveWindowProxy(ScreenManager_GetWindowOf(inScreenID));
	}
}// FileCaptureEnd


/*!
Returns "true" only if there is a file capture
in progress for the specified screen.

(2.6)
*/
Boolean
Terminal_FileCaptureInProgress	(TelnetScreenID const		inScreenID)
{
	return (VSscreens[inScreenID].captureRN);
}// FileCaptureInProgress


/*!
Presents a dialog box asking the user to specify
the file where captured data should be sent.  If
the user does not cancel and no error occurs,
"true" is returned, otherwise, "false" is returned.

(3.0)
*/
Boolean
Terminal_FileCaptureSaveDialog		(FSSpec*		outFSSpecPtr)
{
	Boolean		result = false;
	
	
	if (outFSSpecPtr != null)
	{
		Str255			fileName,
						prompt,
						title;
		OSErr			error = noErr;
		Str32			numString;
		static SInt16	captureNumber = 0;
		
		
		NumToString(++captureNumber, numString);
		GetIndString(fileName, rStringsImportantFileNames, siImportantFileNameCaptureTextDefault);
		fileName[++(fileName[0])] = ' ';
		PLstrcat(fileName, numString);
		
		GetIndString(prompt, rStringsNavigationServices, siNavPromptSaveCapturedText);
		GetIndString(title, rStringsNavigationServices, siNavDialogTitleSaveCapturedText);
		if (FlagManager_Test(kFlagNavigationServices))
		{
			NavReplyRecord		reply;
			OSType				captureFileCreator;
			Size				actualSize = 0L;
			
			
			// get the users Capture File Creator preference, if possible
			unless (Preferences_GetData(kTelnetPreferenceClassGeneral,
										kTelnetPreferenceTag_CaptureFileCreator,
										sizeof(captureFileCreator), &captureFileCreator,
										&actualSize) == kTelnetPrefsResultCodeSuccess)
			{
				captureFileCreator = 'ttxt'; // default to SimpleText if a preference cant be found
			}
			
			Alert_ReportOSErr(error = Navigation_PutFile(prompt, title,
								captureFileCreator, 'TEXT', kTelnetNavPrefKeyGenericSaveFile,
								kNavDefaultNavDlogOptions | kNavDontAddTranslateItems,
								EventLoop_HandleNavigationUpdate, &reply, outFSSpecPtr));
			result = ((error == noErr) && (reply.validRecord));
		}
		else
		{
			StandardFileReply	sfr;
			
			
			DeactivateFrontmostWindow();
			ThemeSavvyStandardPutFile(title, prompt, fileName, &sfr);
			RestoreFrontmostWindow();
			result = sfr.sfGood;
			if (result)
			{
				error = FSMakeFSSpec(sfr.sfFile.vRefNum, sfr.sfFile.parID, sfr.sfFile.name, outFSSpecPtr);
				result = ((error == noErr) || (error == fnfErr));
			}
			
		}
	}
	return result;
}// FileCaptureSaveDialog


/*!
Writes the specified text to the capture file
of the current screen.

(2.6)
*/
void
Terminal_FileCaptureWriteData	(unsigned char*		inoutBuffer,
								 SInt32 const		inLength)
{	
	long	i = 0L,
			ln = inLength;
	
	
	if (VSscreens[VSIwn].captureRN)
	{
		unsigned char 	captbuf[512];
		unsigned char*	ptr2 = null;
		unsigned char*	ptr3 = null;
		
		
		ptr2 = inoutBuffer;
		ptr3 = &captbuf[0];
		for (i = 0; i < ln; i++)
		{
			if (*ptr2 >= 32 ||
				*ptr2 == 13 ||
				*ptr2 ==  9)
			{
				*(ptr3++) = *(ptr2++);
			}
			else
			{
				ptr2++;
				ln--;
			}
		}
		if (ln > 0)
		{
			if (FSWrite(VSscreens[VSIwn].captureRN, &ln, captbuf))
			{
				Terminal_FileCaptureEnd(VSIwn);
			}
		}
	}
}// FileCaptureWriteData


/*!
Iterates over the given range of lines on the
specified virtual screen, executing a function
on each of them.  The ranges are one-based; the
value of 0 is reserved for these.  The context
is defined by you, and is passed directly to
the specified function each time it is invoked.

If each line is iterated over successfully, 0
is returned; otherwise, a nonzero value is
returned.

(3.0)
*/
SInt16
Terminal_ForEachLineDo		(TelnetScreenID const		inScreenID,
							 UInt16 const				inOneBasedRangeFirstLine,
							 UInt16 const				inOneBasedRangeLastLine,
							 ScreenLineOperationProcPtr	inDoWhat,
							 void*						inContextPtr)
{
	SInt16		result = 0;
	
	
	if (inDoWhat == null) result = -2;
	else
	{
		VSscrn*			vsscreen = null;
		VSline*			vslin = null;
		VSline*			p = null;
		short			lineNumber = 0,
						maxLineNumber = 0;
		short			first = inOneBasedRangeFirstLine,
						last = inOneBasedRangeLastLine;
		short			maxColumnNumber = 0; // allocated width of window (80/132)
		
		
		vsscreen = VSwhereis(inScreenID);
		vslin = vsscreen->buftop;
		
		maxColumnNumber = VSmaxwidth(inScreenID) + 1;	// VSmaxwidth returns one less than the width
		p = vslin;
		maxLineNumber = vsscreen->numlines + VSgetlines(inScreenID); // VSgetlines returns 24 or whatever
		
		// check range
		if (first < 1) first = 1;
		if (last > maxLineNumber) last = maxLineNumber;
		
		for (lineNumber = first; lineNumber <= last; lineNumber++)
		{
			if (p == null) { result = -1; break; }
			if (p->text == null) { result = -1; break; }
			
			(Boolean)InvokeScreenLineOperationProc(inDoWhat, inScreenID, p->text, maxColumnNumber, lineNumber, inContextPtr);
			p = p->next;
		}
	}
	return result;
}// ForEachLineDo


/*!
Iterates over the entire range of lines on the
specified virtual screen, executing a function
on each of them.  The context is defined by you,
and is passed directly to the specified function
each time it is invoked.

If each line is iterated over successfully, 0
is returned; otherwise, a nonzero value is
returned.

This is a convenience routine that calculates a
range for you (i.e. 1 -> N) as if you invoked
Terminal_ForEachLineDo() with the calculated
range.

(3.0)
*/
SInt16
Terminal_ForEveryLineDo		(TelnetScreenID const		inScreenID,
							 ScreenLineOperationProcPtr	inDoWhat,
							 void*						inContextPtr)
{
	SInt16		result = 0;
	VSscrn*		vsscreen = null;
	short		maxLineNumber = 0;
	
	
	vsscreen = VSwhereis(inScreenID);
	maxLineNumber = vsscreen->numlines + VSgetlines(inScreenID); // scrollback length plus visible screen rows
	result = Terminal_ForEachLineDo(inScreenID, 1/* minimum */, maxLineNumber/* maximum */, inDoWhat, inContextPtr);
	return result;
}// ForEveryLineDo


/*!
To determine the maximum number of columns
allowed (to limit a text field in a dialog
box, for example), use this method.

(3.0)
*/
UInt16
Terminal_GetAllocatedColumnCount	(void)
{
	return kVirtualScreenInterfaceAllocatedColumnCount;
}// GetAllocatedColumnCount


/*!
Terminates a printout for the specified terminal,
closing the temporary file used for print data.

(2.6)
*/
void
Terminal_PrintingEnd	(TelnetScreenID const		inScreenID)
{
	OSErr		error = noErr;
	FSSpec		temporary;
	
	
	// close the printing file...
	error = FSClose((VSscreens[inScreenID].loc)->refNum), VSIw->refNum = -1;
	if (error != noErr) BEEP;
	
	// ...and then destroy it (uses modern File Manager in MacTelnet 3.0)
	error = FolderManager_GetFSSpec(kTelnetFolderMacTemporaryItems, &temporary);
	if (error == noErr)
	{
		error = FSMakeFSSpec(temporary.vRefNum, temporary.parID, (StringPtr)(VSscreens[inScreenID].loc)->fname, &temporary);
		if (error == noErr) error = FSpDelete(&temporary);
	}
	if (error != noErr) BEEP;
	
	(VSscreens[inScreenID].loc)->prredirect = 0; // 3.0
}// PrintingEnd


/*!
Returns "true" only if the specified screens
text is currently being printed.

(2.6)
*/
Boolean
Terminal_PrintingInProgress		(TelnetScreenID const		inScreenID)
{
	return((VSscreens[inScreenID].loc)->prredirect);
}// PrintingInProgress


/*!
Resets the terminal settings to their default states,
clearing the screen and homing the cursor.

(3.0)
*/
void
Terminal_Reset		(TelnetScreenID const		inScreenID)
{
	if (setCurrentScreen(inScreenID) == 0)
	{
		VSIcuroff(inScreenID);
		VSIreset();
		VSIcurson(inScreenID, VSIw->x, VSIw->y, 1); // redisplay cursor at home position
	}
}// Reset


/*!
Specifies whether or not the lines of the terminal
screen are scrolled prior to a clearing of the
visible screen area.

(2.6)
*/
void
Terminal_SavesLinesOnClear	(TelnetScreenID const		inScreenID,
							 Boolean const				inClearScreenSavesLines)
{
	if (setCurrentScreen(inScreenID) == 0) VSIw->ESscroll = inClearScreenSavesLines;
}// SavesLinesOnClear


/*!
Scrolls the contents of the terminal screen as if
the user clicked the right scroll arrow.  You must
specify the number of columns to scroll.

(2.6)
*/
void
Terminal_ScrollColumnsTowardLeftEdge	(TelnetScreenID const		inScreenID,
										 UInt16 const 				inNumberOfColumnsToScroll)
{
	short		sn = 0,
				columnCount = inNumberOfColumnsToScroll,
				lmmax = 0;
	GrafPtr		oldPort = null;
	
	
	if (setCurrentScreen(inScreenID) == 0)
	{
		// limit scroll amount against number of invisible columns
		lmmax = VSIw->maxwidth - (VSIw->Rright - VSIw->Rleft);
		
		if (VSIw->Rleft + columnCount > lmmax) columnCount = lmmax - VSIw->Rleft; // cant scroll any further than this
		if (columnCount != 0)
		{
			VSIcuroff(inScreenID); // temporarily hide cursor
			VSIw->Rleft += columnCount; // update visible region
			VSIw->Rright += columnCount;
			GetPort(&oldPort);
			RealScreen_SetPortScreenPort(inScreenID);
			sn = VSIw->Rbottom - VSIw->Rtop;
			RealScreen_UpdateHorizontalScrollBar(inScreenID, lmmax, VSIw->Rleft);
			RSdelcols(inScreenID, columnCount); // scroll the window contents
			SetPort(oldPort);
			VSIcurson(inScreenID, VSIw->x, VSIw->y, 0); // restore cursor at original position
			VSredraw(inScreenID, (VSIw->Rright - VSIw->Rleft) - columnCount,
						0, (VSIw->Rright - VSIw->Rleft), sn); // redraw newly-revealed portion of screen
		}
	}
}// ScrollColumnsTowardLeftEdge


/*!
Scrolls the contents of the terminal screen as if
the user clicked the left scroll arrow.  You must
specify the number of columns to scroll.

(2.6)
*/
void
Terminal_ScrollColumnsTowardRightEdge		(TelnetScreenID const		inScreenID,
											 UInt16 const 				inNumberOfColumnsToScroll)
{
	short		sn = 0,
				columnCount = inNumberOfColumnsToScroll,
				lmmax = 0;
	GrafPtr		oldPort = null;
	

	if (setCurrentScreen(inScreenID) == 0)
	{
		// limit scroll amount against number of invisible columns
		lmmax = VSIw->maxwidth - (VSIw->Rright - VSIw->Rleft);
		
		if (columnCount > VSIw->Rleft) columnCount = VSIw->Rleft; // cant scroll any further than this
		if (columnCount != 0)
		{
			VSIcuroff(inScreenID); // temporarily hide cursor
			VSIw->Rleft -= columnCount; // update visible region
			VSIw->Rright -= columnCount;
			sn = VSIw->Rbottom - VSIw->Rtop;
			GetPort(&oldPort);
			RealScreen_SetPortScreenPort(inScreenID);
			RealScreen_UpdateHorizontalScrollBar(inScreenID, lmmax, VSIw->Rleft);
			RSinscols(inScreenID, columnCount); // scroll the window contents
			SetPort(oldPort);
			VSIcurson(inScreenID, VSIw->x, VSIw->y, 0); // restore cursor at original position
			VSredraw(inScreenID, 0, 0, columnCount, sn); // redraw newly-revealed portion of screen
		}
	}
}// ScrollColumnsTowardRightEdge


/*!
Scrolls the contents of the terminal screen as if
the user clicked the up scroll arrow.  You must
specify the number of columns to scroll.

(2.6)
*/
short
Terminal_ScrollRowsTowardBottomEdge		(TelnetScreenID const		inScreenID,
										 UInt16 const 				inNumberOfRowsToScroll)
{
	short		sn = 0,
				n = inNumberOfRowsToScroll;
	
	
	if (setCurrentScreen(inScreenID) != 0)
		return(-3);

	if (VSIw->numlines < (n - VSIw->Rtop))
		n = VSIw->Rtop + VSIw->numlines; /* can't scroll back any further than this */
	if (n <= 0)
		return(0);			/* Dont be scrollin' no lines.... */

	VSIcuroff(inScreenID); /* temporarily hide cursor */

	VSIw->Rtop = VSIw->Rtop - n; /* adjust the visible region */
	VSIw->Rbottom = VSIw->Rbottom - n;

  /* find the line list element for the new topmost visible line */
	sn = n;
	while (sn-- > 0)
	  {
		VSIw->vistop = VSIw->vistop->prev;
	  } /* while */

	sn = VSIw->Rbottom - VSIw->Rtop;
	RealScreen_UpdateVerticalScrollBar(inScreenID, VSIw->numlines, VSIw->Rtop, VSIw->Rbottom);

	if (n <= VSIw->lines)
	  {
	  	GrafPtr		oldPort = null;
		
		
		GetPort(&oldPort);
		RealScreen_SetPortScreenPort(inScreenID);
		RSinslines(inScreenID, 0, sn, n, 0);	/* scroll, preserving current selection */
		SetPort(oldPort);
		VSIcurson(inScreenID, VSIw->x, VSIw->y, 0); /* restore cursor at original position */
		VSredraw(inScreenID, 0, 0, VSIw->maxwidth, n - 1); /* redraw newly-revealed portion */
	  }
	else
	  /* currently-visible contents scrolled completely off the screen--
		just redraw everything */
		VSredraw(inScreenID, 0, 0, VSIw->maxwidth, sn);

	return(0);
}// ScrollRowsTowardBottomEdge


/*!
Scrolls the contents of the terminal screen as if
the user clicked the down scroll arrow.  You must
specify the number of columns to scroll.

(2.6)
*/
short
Terminal_ScrollRowsTowardTopEdge	(TelnetScreenID const		inScreenID,
									 UInt16 const 				inNumberOfRowsToScroll)
{
	short	sn = 0,
			n = inNumberOfRowsToScroll;
	
	
	if (setCurrentScreen(inScreenID) != 0)
		return(-3);

	if (n > VSIw->lines - VSIw->Rbottom)
		n = VSIw->lines - VSIw->Rbottom; /* can't scroll any further forward than this */
	if (n <= 0)
		return(0);			/* Dont be scrollin' no lines.... */

	VSIcuroff(inScreenID); /* temporarily hide cursor */

	VSIw->Rtop = n + VSIw->Rtop; /* adjust the visible region */
	VSIw->Rbottom = n + VSIw->Rbottom;

  /* find the line list element for the new topmost visible line */
	sn = n;
	while (sn-- > 0)
	  {
		VSIw->vistop = VSIw->vistop->next;
	  } /* while */

	sn = VSIw->Rbottom - VSIw->Rtop;
	RealScreen_UpdateVerticalScrollBar(inScreenID, VSIw->numlines, VSIw->Rtop, VSIw->Rbottom);

	if (n <= VSIw->lines)
	  {
	  	GrafPtr		oldPort = null;
		
		
		GetPort(&oldPort);
		RealScreen_SetPortScreenPort(inScreenID);
		RSdellines(inScreenID, 0, sn, n, 0);	/* scroll, preserving current selection */
		SetPort(oldPort);
		VSIcurson(inScreenID, VSIw->x, VSIw->y, 0); /* restore cursor at original position */
		VSredraw(inScreenID, 0, (sn + 1) - n, VSIw->maxwidth, sn); /* redraw newly-revealed portion */
	  } /* if */
	else
	  /* currently-visible contents scrolled completely off the screen--
		just redraw everything */
		VSredraw(inScreenID, 0, 0, VSIw->maxwidth, sn);

	return(0);
}// ScrollRowsTowardTopEdge


/*!
Scrolls the contents of the terminal screen as if
the user clicked the home key on an extended
keyboard - conveniently scrolling rows downward
to show the very topmost row in the screen area.

(3.0)
*/
void
Terminal_ScrollToBeginning		(TelnetScreenID const		inScreenID)
{
	if (setCurrentScreen(inScreenID) == 0)
	{
		Terminal_ScrollRowsTowardTopEdge(inScreenID, VSIw->Rtop + VSIw->numlines); // cant scroll any further than this
	}
}// ScrollToBeginning


/*!
Scrolls the contents of the terminal screen as if
the user clicked the end key on an extended
keyboard - conveniently scrolling rows upward to
show the very bottommost row in the screen area.

(3.0)
*/
void
Terminal_ScrollToEnd		(TelnetScreenID const		inScreenID)
{
	if (setCurrentScreen(inScreenID) == 0)
	{
		Terminal_ScrollRowsTowardTopEdge(inScreenID, 32765); // cant scroll any further than this
	}
}// ScrollToEnd


VSscrn *VSwhereis(short i) /* screen number */
/* returns a pointer to the structure for the specified screen. */
{
	setCurrentScreen(i);
	return(VSIw);
} /* VSwhereis */


void VSIclrbuf
  (
	void
  )
  /* clears out the text and attribute buffers for the current screen.
	All text characters are set to blanks, and all attribute bytes
	are set to zero. Doesn't update the display. */
  {
	register short j, i;
	register char *tx;
	register unsigned short *ta;
	for (i = 0; i <= VSIw->lines; i++)
	  {
		ta = &VSIw->attrst[i]->text[0];
		tx = &VSIw->linest[i]->text[0];
		for (j = 0; j <= VSIw->allwidth; j++)
		  {
			*ta++ = 0;
			*tx++ = ' ';
		  } /* for */
	  } /* for */
  } /* VSIclrbuf */

short VSdestroy(short w) /* screen number */
  /* gets rid of a virtual screen. */
  {
	if (setCurrentScreen(w) != 0)
		return(-3);
	VSIfreelines();
	VSIfreelinelist((VSlinePtr)VSIw->attrst[0]);
	if(VSIw->attrst)
		Memory_DisposePtr(&((Ptr)VSIw->attrst));
	if(VSIw->linest)
		Memory_DisposePtr(&((Ptr)VSIw->linest));
	if(VSIw->tabs)
		Memory_DisposePtr(&(VSIw->tabs));
	if(VSIw)
		Memory_DisposePtr(&((Ptr)VSIw));
	VSscreens[w].stat = 0;
	VSIwn = -1;
	VSinuse--;			/* SCA '87 */
	return(0);
  } /* VSdestroy */


void VSredrawLine(short w) //redraws current line
{
	if (setCurrentScreen(w) != 0)
		return;
	VSredraw(w, 0, VSIw->y,VSIw->maxwidth, VSIw->y); 
	VSIcuroff(w);
}


short VSredraw
  (
	short w, 		// window to redraw */
	short x1, 		
	short y1,
	short x2,
	short y2
  )
  /* redisplays the specified portion of a virtual screen. */
  {
	VSlinePtr ypt;
	VSattrlinePtr ypa;
	short y;
	short tx1, tx2, ty1, ty2, tn, offset;
	GrafPtr	oldPort = null;
	
	
	if (setCurrentScreen(w) != 0)
		return(-3);

	x1 += VSIw->Rleft;	// Make local coords global again
	x2 += VSIw->Rleft;
	y1 += VSIw->Rtop;
	y2 += VSIw->Rtop;

	if (x1 < 0) x1 = 0;
	else if (x1 > VSIw->maxwidth) x1 = VSIw->maxwidth;
	if (x2 < 0) x2 = 0;
	else if (x2 > VSIw->maxwidth) x2 = VSIw->maxwidth;
	if (y1 < -VSIw->numlines) y1 = -VSIw->numlines;
	else if (y1 > VSIw->lines) y1 = VSIw->lines;
	if (y2 < -VSIw->numlines) y2 = -VSIw->numlines;	/* limit to amount of scrollback */
	else if (y2 > VSIw->lines) y2 = VSIw->lines;

	tx1 = x1;		// Set up to clip redraw area to visible area
	tx2 = x2;
	ty1 = y1;
	ty2 = y2;
	tn = -1;		// so we include more than 1 line
	
	if (VSIclip(&tx1, &ty1, &tx2, &ty2, &tn, &offset)!=0) return 0;		// test clip region
	
	GetPort(&oldPort);
	RealScreen_SetPortScreenPort(w);
	
	VSIcuroff(w); 						// temporarily hide cursor
	RealScreen_EraseSection(w, tx1, ty1, tx2, ty2);		// Erase the offending area

	// draw visible part of scrollback buffer
	if (y1 < 0) {
	
		tx1 = x1;		// Set up to clip redraw area to visible area of scrollback buffer
		tx2 = x2;
		ty1 = y1;
		ty2 = (y2>=0) ? -1 : y2;
		tn = -1;
		
		if (!VSIclip(&tx1, &ty1, &tx2, &ty2, &tn, &offset)) {
			ypt = VSIw->vistop;
			for(y=VSIw->Rtop; y<y1; y++)
				ypt = ypt->next;		// Get pointer to top line we need
	
			for (y=ty1; y<=ty2; y++) {
				RSdraw(w, tx1, y, 0, tn, ypt->text + VSIw->Rleft +tx1);
				ypt = ypt->next;
			}
		}
		y1 = 0;			// continue with on-screen buffer, if any
	}

	// draw visible part of on-screen buffer, taking account of attributes
	if (y2 >= 0) {
	
		tx1 = x1;		// Set up to clip redraw area to visible area of on-screen buffer
		tx2 = x2;
		ty1 = y1;
		ty2 = y2;
		tn = -1;

		if (!VSIclip(&tx1, &ty1, &tx2, &ty2, &tn, &offset)) {
		
			ypt = VSIw->linest[VSIw->Rtop+ty1];
			ypa = VSIw->attrst[VSIw->Rtop+ty1];
			
			for (y=ty1; y<=ty2; y++) {
				char *pt;
				unsigned short *pa;
				unsigned short lasta;
				short x, lastx;
			
				pt = ypt->text + VSIw->Rleft;
				pa = ypa->text + VSIw->Rleft;
				
				lastx = tx1;
				lasta = pa[tx1];
				for(x=tx1+1; x<=tx2; x++) {
					if (pa[x]!=lasta) {
						RSdraw(w, lastx, y, lasta, x-lastx, pt + lastx);
						lastx = x;
						lasta = pa[x];
					}
				}
				if (lastx<=tx2)
					RSdraw(w, lastx, y, lasta, tx2-lastx+1, pt + lastx);
				
				ypt = ypt->next;
				ypa = ypa->next;
			}
		}
	}

	VSIcurson(w, VSIw->x, VSIw->y, 0); /* restore cursor at original position */

	//tx1 = ty1 = 0;
	//tn = Terminal_GetAllocatedColumnCount();
	
	SetPort(oldPort);
	return(0);
  } /* VSredraw */


/*!
Sends a stream of characters to the specified
window, interpreting escape sequences, etc.
to affect the terminal display.

(2.6)
*/
short
VSwrite		(short			inScreenID,
			 char*			inString,
			 SInt16			inLength)
{
	short		result = 0;
	
	
	if (inLength != 0)
	{
		if (setCurrentScreen(inScreenID) != 0) result = -3;
		else
		{
			VSIcuroff(inScreenID); // hide cursor momentarily
			VirtualScreenEmulation_Parse((unsigned char*)inString, inLength);	// interpret the character stream
			VSIcurson(inScreenID, VSIw->x, VSIw->y, 1/* scroll to find it? */); // restore cursor, force it to be visible
		}
	}
	return result;
}// VSwrite


short VSIgetNextTabDistance(void)
{
	short current;
	
	if (VSIw->x >= VSIw->maxwidth)
		return(0);
	current = VSIw->x + 1;
	while ((VSIw->tabs[current] != 'x')&&(current < VSIw->maxwidth))
		current++;
	return (current - VSIw->x);
}


															// MAT--We need to define how big the Queue is
															// MAT--for sending the cursor characters.
															// MAT--Since MacTCP doesn't like 30+ individual
#define MATSbufSize	254										// MAT--SendChar requests.
void VSpossendEM											// MAT--we can change this to support EMACS
  (															// MAT--style movement commands.
	short w, /* affected screen */												// MAT--
	short x, /* column to move to */												// MAT--
	short y, /* line to move to */												// MAT--
	short UNUSED_ARGUMENT(echo) /* local echo flag */												// MAT--
  )																				// MAT--
  /* sends a stream of VT100 cursor-movement sequences to move the				// MAT--
	cursor on the specified screen to the specified position. */				// MAT--
{																				// MAT--
	UNUSED_ARG(echo)
	static char																	// MAT--
			tt[MATSbufSize] = "";												// MAT--
		char curschar;															// MAT--
		short localCount;															// MAT--
																				// MAT--
																				// MAT--
	if (x < 0 || y < 0 || x > VSIw->maxwidth || y > VSIw->lines)				// MAT--
		return;																	// MAT--
  /* convert x and y to relative distance to move */							// MAT--
	x -= VSIw->x;																// MAT--
	y -= VSIw->y;																// MAT--
																				// MAT--
	curschar = '\002'; /* EMACS cursor left */									// MAT--
	localCount=0;
	while (x < 0 && localCount < MATSbufSize)									// MAT--
	  {																			// MAT--
		tt[localCount] = curschar;												// MAT--
		x++; localCount++;														// MAT--
	  } /* while */																// MAT--
	if(localCount)	ScreenManager_SendData(w, tt, localCount);

	curschar = '\016'; /* EMACS cursor down */									// MAT--
	localCount=0;
	while (y > 0 && localCount < MATSbufSize)									// MAT-- why we check to see if localCount is < MATSbufSize
	  {																			// MAT-- I dont know. But If they had a window > 254 columns
		tt[localCount] = curschar;												// MAT-- maybe it's a good idea.
		y--; localCount++;														// MAT-- but it never hurts to be safe.
	  } /* while */																// MAT--
	if(localCount)	ScreenManager_SendData(w, tt, localCount);

	curschar = '\020'; /* EMACS cursor up */									// MAT--
	localCount=0;
	while (y < 0 && localCount < MATSbufSize)									// MAT--
	  {																			// MAT--
		tt[localCount] = curschar;												// MAT--
		y++; localCount++;														// MAT--
	  } /* while */																// MAT--
	if(localCount)	ScreenManager_SendData(w, tt, localCount);

	curschar = '\006'; /* EMACS cursor right */									// MAT--
	localCount=0;
	while (x > 0 && localCount < MATSbufSize)									// MAT--
	  {																			// MAT--
		tt[localCount] = curschar;												// MAT--
		x--; localCount++;														// MAT--
	  } /* while */																// MAT--
	if(localCount)	ScreenManager_SendData(w, tt, localCount);
																				// MAT--
	if (0)	//(note: supposed to look for "echo" here)							// MAT--
	  {																			// MAT--
		VSIcuroff(w);															// MAT--
		VSIw->x = x;															// MAT--
		VSIw->y = y;															// MAT--
		VSIcurson(w, VSIw->x, VSIw->y, 1); /* Force Move */						// MAT--
	  } /* if */																// MAT--
  } /* VSpossendEM */					// changed comment						// MAT--
 	
void VSpossend
  (
	short w, /* affected screen */
	short x, /* column to move to */
	short y, /* line to move to */
	short echo /* local echo flag */
  )
  /* sends a stream of VT100 cursor-movement sequences to move the
	cursor on the specified screen to the specified position. */
{
	static char
		VSkbax[] = "\033O ",		/* prefix for auxiliary code */
		VSkban[] = "\033[ ";		/* prefix for arrows normal */
	char *vskptr;
	
	
	if (setCurrentScreen(w) != 0)
		return;
	
	
/* NCSA: SB - This would bomb before.  You need to get the screens # from the 
				translation routine before you access the record! */  	
	if (ScreenManager_GetConnectionDataOf(w)->arrowmap) {    /* NCSA: SB - get the CORRECT screens # */
		VSpossendEM(w,x,y,echo);            // MAT--  call our cursor movement routine
		return;                             // MAT--  then exit
	}

	if (VSIw->DECPAM && VSIw->DECCKM)
		vskptr = VSkbax;
	else
		vskptr = VSkban;
	if (x < 0 || y < 0 || x > VSIw->maxwidth || y > VSIw->lines)
		return;
  /* convert x and y to relative distance to move */
	x -= VSIw->x;
	y -= VSIw->y;

	vskptr[2] = 'D'; /* cursor left */
	while (x < 0)
	  {
		x++;
		ScreenManager_SendData(w, vskptr, 3);
	  } /* while */

	vskptr[2] = 'B'; /* cursor down */
	while (y > 0)
	  {
		y--;
		ScreenManager_SendData(w, vskptr, 3);
	  } /* while */
	vskptr[2] = 'A'; /* cursor up */
	while (y < 0)
	  {
		y++;
		ScreenManager_SendData(w, vskptr, 3);
	  } /* while */
	vskptr[2] = 'C'; /* cursor right */
	while (x > 0)
	  {
		x--;
		ScreenManager_SendData(w, vskptr, 3);
	  } /* while */

	if (echo)
	  {
		VSIcuroff(w);
		VSIw->x = x;
		VSIw->y = y;
		VSIcurson(w, VSIw->x, VSIw->y, 1); /* Force Move */
	  } /* if */
  } /* VSpossend */

char VSkbsend
  (
	short w, /* active window */
	unsigned char k, /* special key code if > 128, else ascii code */
	short echo /* local echo flag */
  )
  /* sends the appropriate sequence for the specified key, taking due
	account of terminal mode settings. */
  {
	static char
		VSkbkn[] = "\033O ",		/* prefix for keypad normal */
		VSkbax[] = "\033O ",		/* prefix for auxiliary code*/
		VSkban[] = "\033[ ",		/* prefix for arrows normal */
		VSkbfn[] = "\033O ",		/* prefix for function keys */
		VSk220[] = "\033[  ~";		/* prefix for vt220 keys */
	char *vskptr;
	short vskplen;


	if (setCurrentScreen(w) != 0)
		return(-3);

	if (ScreenManager_GetConnectionDataOf(w)->arrowmap && (k <= VSLT) && (k >= VSUP) )	// MAT--
										// MAT-- important...we need to check this first before
		{								// MAT-- the next if() statement gets its hands on the string.
			switch (k) {				// MAT--  do the mapping from arrowkeys -> EMACS ctrl keys.
				case VSLT:				// MAT-- 
					k = 0x02;			// MAT-- ^B			Question: Is there a way to find out if the option
					break;				// MAT-- 					  key was held down with this character?
				case VSRT:				// MAT-- 			I didn't want to declare myEvent an extern
					k = 0x06;			// MAT-- ^F			(I didn't know if that was a no-no)
					break;				// MAT-- 			If I can.....let me know, I want to make
				case VSUP:				// MAT-- 			option-arrowkey's do useful things too
					k = 0x10;			// MAT-- ^P
					break;				// MAT-- 			checking the keymap would be a kludge here.
				case VSDN:				// MAT-- 
					k = 0x0e;			// MAT-- ^N
					break;				// MAT-- 
			}	/* switch k */			// MAT--
			ScreenManager_SendData(w,(char *)&k,1);
			return(0);					// MAT--
		}								// MAT--


	if (k < VSF10)
	  /* 7-bit ascii code -- send as is */
		ScreenManager_SendData(w,(char *) &k, 1);
	
	/* Keypad (Not Application Mode): 0-9 , - . Enter */	
	if ((k > VSLT) && (k < VSF1) && (!VSIw->DECPAM)) {
		ScreenManager_SendData(w, &VSIkpxlate[0][k - VSUP], 1);
		if (echo)
			VSwrite(w, &VSIkpxlate[0][k - VSUP], 1);
		if (k == VSKE)
			ScreenManager_SendData(w, "\012", 1);
		return(0);
	  } /* if */
	
	  
	if (VSIw->DECPAM && VSIw->DECCKM) {
	  /* aux kp mode */
	  	vskptr = VSkbax;
		vskplen = 3;
		vskptr[2] = VSIkpxlate[1][k - VSUP];
	  }
	else if (k < VSUP) {						
		vskptr = VSk220;						
		vskplen = VSIkplen[k - VSF10];			
		vskptr[2] = VSIkpxlate2[k - VSF10];		
		vskptr[3] = VSIkpxlate3[k - VSF10];		
	} else {									
		vskplen = 3;							
		if (k < VSK0) {	
			vskptr = VSkban;					
			if (VSIw->DECCKM)					
				vskptr[1] = 79;					
			else								
				vskptr[1] = 91;					
		}										
		else if (k < VSF1) 						
			vskptr = VSkbkn;					
		else 									
			vskptr = VSkbfn;					
												
		vskptr[2] = VSIkpxlate[1][k - VSUP];	
	}											

	ScreenManager_SendData(w, vskptr, vskplen);
	if (echo)
		VSwrite(w, vskptr, vskplen);
	return(0);
  } /* VSkbsend */

short VSsetrgn
  (
	short w,
	short x1, /* leftmost column */
	short y1, /* uppermost line */
	short x2, /* rightmost column */
	short y2 /* bottommost line */
  )
  /* sets the visible region for the specified virtual screen
	in its window, scrolling its contents as appropriate. Assumes
	that either the vertical bounds or the height of the region has
	changed, but not both, and similarly that the horizontal bounds
	or the width has changed, but not both. */
  {
	short n;

	if (setCurrentScreen(w) != 0)
		return(-3);

	VSIw->Rbottom = VSIw->Rtop + (y2 - y1); /* make change in height of visible region first */

	if (x2 > VSIw->maxwidth)
	  {
	  /* trying to make columns visible which aren't there--
		adjust the left and right boundaries to avoid this */
		n = x2 - VSIw->maxwidth; /* how far to adjust to the left */
		if (n > x1)
			n = x1; /* but I'd rather have unused columns on the right than on the left */
		x1 -= n;							/* Adjust left	*/
		x2 -= n;							/* Adjust right */
	  } /* if */

	if (VSIw->Rleft != x1)
	  {
	  /* left margin changed -- scroll horizontally */
	  /* (assume width of region hasn't also changed) */
		n = x1 - VSIw->Rleft;
		if (n > 0) Terminal_ScrollColumnsTowardLeftEdge(w, n);
		else Terminal_ScrollColumnsTowardRightEdge(w, -n);
	  }
	else
	  /* just update horizontal scroll bar limits for new width of visible region */
		RealScreen_UpdateHorizontalScrollBar(w, VSIw->maxwidth - (x2 - x1), x1);

	VSIw->Rleft = x1;
	VSIw->Rright = x2;

	if (VSIw->Rbottom > VSIw->lines)
	  /* don't scroll off the bottom of the screen */
		n = VSIw->Rbottom - VSIw->lines;
	else
	  /* scroll to new topmost line as specified */
		n = VSIw->Rtop - y1;

	if (n != 0)
	  /* scroll vertically (assume height of region hasn't also changed) */
		if (n > 0) Terminal_ScrollRowsTowardBottomEdge(w, n);
		else Terminal_ScrollRowsTowardTopEdge(w, -n);
	else
	  /* update vertical scroll bar limits for new height of visible region */
		RealScreen_UpdateVerticalScrollBar(w, VSIw->numlines, VSIw->Rtop, VSIw->Rbottom);
	return(0);
  } /* VSsetrgn */

short VSgetrgn
  (
	short w,
	short *x1,
	short *y1,
	short *x2,
	short *y2
  )
  /* returns the current visible region for a virtual screen. */
  {
	if (setCurrentScreen(w) != 0)
		return(-3);

	*x1 = VSIw->Rleft;
	*y1 = VSIw->Rtop;
	*x2 = VSIw->Rright;
	*y2 = VSIw->Rbottom;
   
	return(0);
  } /* VSgetrgn */


short VSmaxwidth
  (
	short w
  )
  /* returns one less than the number of columns on a virtual screen. */
  {
	if (setCurrentScreen(w) != 0)
		return(-3);
	return(VSIw->maxwidth);
  } /* VSmaxwidth */

VSlinePtr VSIGetLineStart(short w, short y1)
  /* returns a pointer to the specified text line (number may be
	negative for a line in the scrollback buffer). */
  {
	VSlinePtr ptr;
	short n;

	if (setCurrentScreen(w) != 0)
		return((VSlinePtr) -3);

	if (y1 >= 0)
		return(VSIw->linest[y1]);

	n = y1 - VSIw->Rtop;				/* Number of lines from VISTOP to scroll forward */
	ptr = VSIw->vistop;
	while (n > 0)
	  {
		n--;
		ptr = ptr->next;
	  } /* while */
	while (n < 0)
	  {
		n++;
		ptr = ptr->prev;
	  } /* while */
	return(ptr);
  } /* VSIGetLineStart */


char *VSIstrcopy(char *src, short len, char *dest, short table)
  /* copies characters from *src (length len) to *dest, dropping
	trailing blanks. If table is nonzero, then this number (or more) of
	consecutive embedded blanks will be replaced with a tab. Returns a pointer
	to the position one past the last character copied to the *dest buffer. */
  {
	char *p, *tempp;
	short tblck;

	p = src + len - 1;
  /* skip trailing blanks */
	while ((*p == ' ') && (p >= src))
		p--;
	if (p < src)
		return(dest);
	if (!table)
	  /* straight character copy */
		while (src <= p)
			*dest++ = *src++;
	else
	  /* tab-replacement copy */
		while (src <= p)
		  {
			while ((src <= p) && (*src != ' '))
				*dest++ = *src++;
			if (src < p)
			  {
				tempp = dest; /* remember start of run of spaces */
				tblck = 0; /* length of run */
				while ((src <= p) && (*src == ' '))
				  {
					*dest++ = *src++;
					tblck++;
				  } /* while */
				if (tblck >= table)
				  {
					*tempp++ = '\011'; /* replace first space with a tab */
					dest = tempp; /* drop remaining spaces */
				  } /* if */
			  } /* if */
		  } /* while */
	return(dest);
  } /* VSIstrcopy */


/*!
Copies a portion of text from the specified virtual screen
into the "charp" buffer.  The parameter "table", if nonzero,
is the minimum length of runs of spaces to be replaced with
single tabs.  Returns the length of the copied text.  The
value of "max" is supposed to be the maximum length to copy,
but this is currently ignored!  EOLS is the end-of-line
sequence to insert at line boundaries.  This is currently
assumed to be exactly one character long.

NOTE:	You can set the first character of "EOLS" to 0, to
		specify that no character should be used to separate
		lines.  You might do this if you want to copy
		multiple lines of terminal text that really are one
		line.

In MacTelnet 3.0, this method now copies text differently
if the coordinates are to define a rectangular selection.
That is, while normal Macintosh text selection includes all
text in the rows between the start and end anchors, a text
selection that is rectangular will only include text in
the in-between rows that are part of the columns between
the columns of the two anchor points.

(2.6)
*/
long
VSgettext		(short			w,
				 short			x1,
				 short			y1,
				 short			x2,
				 short			y2,
				 char*			charp,
				 long			max,
				 char*			EOLS,
				 short			table,
				 Boolean		inIsRectangular)
{
	UNUSED_ARG(max) /* !! */
	short EOLlen;
	short lx,ly,					/* Upper bounds of selection */
		ux,uy;					/* Lower bounds of selection */
	short maxwid;
	char *origcp;
	VSlinePtr t;
	
	if (setCurrentScreen(w) != 0)
		return(-3);
	EOLlen = strlen(EOLS);
	maxwid = VSIw->maxwidth;
	origcp = charp;

	if (y1 < -VSIw->numlines)
	  {
		y1 = -VSIw->numlines;
		x1 = -1;
	  } /* if */
	if (y1 == y2)
	{
	  	/* copying no more than a single line */
		t = VSIGetLineStart(w, y1);
		if (x1 < x2)	/* Order the lower and upper bounds */
		{
			ux = x1;
			uy = y1;
			lx = x2;
			ly = y2;
		}
		else
		{
			ux = x2;
			uy = y2;
			lx = x1;
			ly = y1;
		}
		if ((long)(lx-ux) < max) charp=VSIstrcopy(&t->text[ux+1], lx-ux, charp, table);
		else charp=VSIstrcopy(&t->text[ux+1], (short)(max - (long)(charp-origcp)), charp, table);
		if ((lx == maxwid) && (EOLlen)) *charp++ = *EOLS; // assumes it's only one character!
	}
	else if (inIsRectangular) // 3.0
	{
		// there are multiple lines of text, all within the same horizontal boundaries
		if (y1 < y2)	/* Order the lower and upper bounds */
		{
			ux = x1;
			uy = y1;
			lx = x2;
			ly = y2;
		}
		else
		{
			ux = x2;
			uy = y2;
			lx = x1;
			ly = y1;
		}
		t = VSIGetLineStart(w, uy);
		while (uy <= ly && uy < VSIw->lines)
		{
			if ((long)(lx-ux) < max) charp = VSIstrcopy(&t->text[ux+1], lx-ux, charp, table);
			else charp = VSIstrcopy(&t->text[ux+1], (short)(max - (long)(charp - origcp)), charp, table);
			if (EOLlen) *charp++ = *EOLS;
			t = t->next; 
			uy++;
		}
	}
	else
	{
	  /* copying more than one line in the standard way */
		if (y1 < y2)	/* Order the lower and upper bounds */
		  {
			ux = x1;
			uy = y1;
			lx = x2;
			ly = y2;
		  }
		else
		  {
			ux = x2;
			uy = y2;
			lx = x1;
			ly = y1;
		  } /* if */
		t = VSIGetLineStart(w, uy);
		if (((long) (maxwid-ux) < max))
			charp=VSIstrcopy(&t->text[ux+1],maxwid-ux,charp,table);
		else
			charp=VSIstrcopy(&t->text[ux+1],(short) (max-(long)(charp-origcp)),charp,table);
		if (EOLlen) *charp++ = *EOLS; /* assumes it's only one character! */
		uy++;
		t = t->next;
		while (uy < ly && uy < VSIw->lines)
		  {
			if ((long)(maxwid+1) < max)
				charp=VSIstrcopy(t->text,maxwid+1,charp, table);
			else
				 charp=VSIstrcopy(t->text,(short)(max - (long) (charp-origcp)),charp, table);
			if (EOLlen) *charp++=*EOLS;
			t=t->next; 
			uy++;
		  } /* while */
		if (ly > VSIw->lines)
			lx = maxwid;

		if ((long) (lx+1) < max)
			charp=VSIstrcopy(t->text,lx+1,charp,table);
		else
			charp=VSIstrcopy(t->text,(short)(max - (long)(charp-origcp)),charp,table);

		if ((lx >= maxwid) && (EOLlen)) *charp++ = *EOLS; /* assumes it's only one character! */
	  } /* if */
	return(charp - origcp);
  } /* VSgettext */

short VSgetlines
  (
	short w
  )
  /* returns the number of lines in a virtual screen. */
  {
	if (setCurrentScreen(w) != 0)
		return(-2);
	return(VSIw->lines + 1);
  } /* VSgetlines */

short VSsetlines
  (
	short w, /* window number */
	short lines /* new number of lines */
  )
  /* sets the number of lines in a virtual screen, reallocating text
	and attribute arrays accordingly. Returns the new number of lines
	on success, or an error otherwise */
  {
	VSlineArray linest;				/* For storage of old ones */
	VSattrlineArray attrst;
	VSlinePtr line;							/* pointer to a line */
	short i, j, oldlines;
	char *temp;
	unsigned short *tempa;
	
	if (setCurrentScreen(w) != 0)
		return(-3000);
	
	lines -= 1;								/* Correct for internal use */
	oldlines = VSIw->lines;
	if (lines == oldlines)	 				/* no change */
	 	return(0);
	
	VSIw->x = 0;
	VSIw->y = 0;
	VSIcurson(w, VSIw->x, VSIw->y, 1); 	/* keeps cursor from pointing outside of window */

	VSIw->vistop = VSIw->scrntop;			/* Force view to the top of the screen */
	
	attrst = VSIw->attrst; /* save old screen arrays */
	linest = VSIw->linest;
 
	VSIw->linest = VSInewlinearray(lines + 1); /* allocate new screen buffers */
	if (!VSIw->linest)
	{
		VSIw->linest = linest;		/* mem problems */
		return (-2000);
	}
	
	VSIw->attrst = VSInewattrlinearray(lines + 1);
	if (!VSIw->attrst)
	{								/* mem problems */
		Memory_DisposePtr(&((Ptr)VSIw->linest));
		VSIw->linest = linest;
		VSIw->attrst = attrst;
		return (-2000);
	}
	
	VSIw->lines = lines; /* set new number of screen lines */
	

	VSIw->linest[0] = VSInewlines(lines + 1,1); /* allocate new text and attribute lines */
	VSIw->attrst[0] = (VSattrlinePtr)VSInewlines(lines + 1,2);
	if (VSIw->linest[0] && VSIw->attrst[0])
	{	/* mem is there */
		
		/* dispose of old attribute lines */
		VSIfreelinelist((VSlinePtr)attrst[0]);
		Memory_DisposePtr(&((Ptr)attrst));

		VSIlistndx(VSIw->linest[0],VSIw->attrst[0]); /* build the new screen arrays */
		if (VSIw->savelines)
		{ /* save previous screen contents in scrollback buffer */
			line = linest[oldlines]->next;				/* save continuation */
			linest[oldlines]->next = VSIw->linest[0];
			VSIw->linest[lines]->next = line;			/* restore continuation */
			VSIw->linest[0]->prev = linest[oldlines];	/* backpointer */
			if (line)									/* if there was a follower */
				line->prev = VSIw->linest[lines];		/* new prev for it */
			VSIw->numlines += oldlines;					/* we made more scrollback */
		}
		else
		{  /* get rid of previous screen contents */
			VSIfreelinelist(linest[0]);
			Memory_DisposePtr(&((Ptr)linest));
		}
	}
	else
	{	/* need more mem - emergency */
		if (VSIw->linest[0])					/*if 1/2 of push for memory died, kill the other */
			VSIfreelinelist(VSIw->linest[0]);
		if (VSIw->attrst[0])
			VSIfreelinelist((VSlinePtr)VSIw->attrst[0]);
		/* Here we should ask if we want to lose the screen buffer!!  CCP */
		VSIfreelines();							  /* release scrollback buffer */
		VSIfreelinelist(linest[0]);				  /* release current visible lines */
		Memory_DisposePtr(&((Ptr)linest));				  
		VSIfreelinelist((VSlinePtr)attrst[0]);				  /* release current visible attrib */
		Memory_DisposePtr(&((Ptr)attrst));
		VSIw->linest[0] = VSInewlines(lines + 1,1); /* allocate new screen arrays */
		VSIw->attrst[0] = (VSattrlinePtr)VSInewlines(lines + 1,2);
		if (!VSIw->linest[0] || !VSIw->attrst[0])
		{  /* still not enough memory;  Try to allocate just enough to go back to original size */
			
			if (VSIw->linest[0])					/* this gets rid of useless */
				VSIfreelinelist(VSIw->linest[0]);	/* memory, since we are giving up */
			if (VSIw->attrst[0])									
				VSIfreelinelist((VSlinePtr)VSIw->attrst[0]);

			VSIw->linest[0] = VSInewlines(oldlines + 1,1); /* try original size */
			VSIw->attrst[0] = (VSattrlinePtr)VSInewlines(oldlines + 1,2);
			
			if (!VSIw->linest[0] || !VSIw->attrst[0])
			/* damage control: */
			{										  /* Nope. Give up, and signal that */
				if (VSIw->linest[0])				  /* caller should kill this screen */
					VSIfreelinelist(VSIw->linest[0]);	
				if (VSIw->attrst[0])				
					VSIfreelinelist((VSlinePtr)VSIw->attrst[0]);
				/* dont destroy everything, as this will screw up VSdestroy later */			
				return(-4000);
			}
			else
			{
				lines = oldlines;
				VSIw->lines = lines;
			}
		}
		VSIw->buftop = VSIw->linest[0];
		VSIw->numlines = 0; /* nothing in scrollback */
	  } /* if */

	VSIw->scrntop = VSIw->linest[0];			/* new top of screen */
	VSIw->vistop = VSIw->scrntop;				/* Force a scroll to the top of the screen */
	VSIlistndx(VSIw->scrntop, VSIw->attrst[0]); /* rebuild screen arrays */	
	VSIw->attrst[0]->prev = VSIw->attrst[lines];	/* Make attribute list circular */
	VSIw->attrst[lines]->next = VSIw->attrst[0];
	if (!VSIw->savelines)
	  {
	  /* make text line list circular to indicate no extensions */
		VSIw->linest[lines]->next = VSIw->linest[0];
		VSIw->linest[0]->prev = VSIw->linest[lines];
	  } /* if */

  /* initialize the new screen lines to blank text and no attributes */
	for (i = 0; i <= lines; i++)
	  {
		tempa = VSIw->attrst[i]->text;
		temp = VSIw->linest[i]->text;
		for (j = 0; j <= VSIw->allwidth; j++)
		  {
			*temp++ = ' ';
			*tempa++ = 0;
		  } /* for */
	  } /* for */
  
  /* reset scrolling region */
	VSIw->top = 0;
	VSIw->bottom = lines;
  
  /* reset visible region */
	VSIw->Rtop = 0;
	VSIw->Rbottom = lines;

	
	VSredraw(w, 0, 0, VSIw->maxwidth, lines); /* draw new blank screen */
	RealScreen_UpdateVerticalScrollBar(w, VSIw->numlines, VSIw->Rtop, VSIw->Rbottom);

	return(VSIw->lines);

} /* VSsetlines */


/*!
Changes the column width of a virtual screen.  This will
not exceed the allocated memory.

(2.6)
*/
short
VSsetcols	(short		inScreenID,
			 short		inNewWidthInColumns)
{
	short		result = 0;
	
	
	if (setCurrentScreen(inScreenID) != 0) result = -1;
	else VSIw->maxwidth = inNewWidthInColumns;
	
	return result;
}// VSsetcols


/*!
Returns the column width of a virtual screen.  As of
MacTelnet 3.0, this routine has been made 100 times
more intelligent by NOT RETURNING A VALUE ONE LESS
THAN THE DAMNED WIDTH.  I have no idea whose bright
idea it was that it should, especially since the
sister routine VSgetlines() returns the actual number
of rows and *not* one less!  All code that calls this
routine has been updated to exclude the extra code to
(get this!) add 1 to the result of calling this
routine.

(2.6)
*/
short
VSgetcols	(TelnetScreenID const		inScreenID)
{
	short		result = -1;
	
	
	if (!setCurrentScreen(inScreenID)) result = (1 + VSIw->maxwidth);
	
	return result;
}// VSgetcols


//
// internal methods
//

/*!
Sets the current screen ID to the specified
screen.  Returns an error code if the given
screen ID is invalid.  Returns 0 if the
specified screen ID is *valid*!!!

(3.0)
*/
static short
setCurrentScreen	(TelnetScreenID const		inScreenID)
{
	short		result = 0;
	
	
	if (VSinuse == 0) result = -5; // no terminal screens are in use!
	else if (VSIwn == inScreenID) result = 0; // the current screen is the specified screen already, so thats OK
	else if ((inScreenID >= VSmax) || (inScreenID < 0)) result = -6; // invalid range!
	else
	{
		VSIwn = inScreenID;
		if (VSscreens[inScreenID].stat != 1) result = -3; // no space allocated
		else
		{
			VSIw = VSscreens[inScreenID].loc;
			if (VSIw == 0L) result = -3; // no space allocated
			else result = 0; // success!
		}
	}
	return result;
}// setCurrentScreen