/*###############################################################

	TerminalWindowHeader.c
	
	MacTelnet 3.0,  1998-2003 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.
	
	You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <Appearance.h>
#include <Controls.h>
#include <ControlDefinitions.h>
#include <DateTimeUtils.h>
#include <Folders.h>
#include <Icons.h>
#include <PLStringFuncs.h>
#include <QuickDraw.h>

// library includes
#include "DataAccessLib.h"
#include "InterfaceLib.h"

// resource includes
#include "ControlResources.h"
#include "SpacingConstants.r"

// MacTelnet includes
#include "DialogUtilities.h"
#include "TelnetMenus.h"
#include "TerminalWindowHeader.h"
#include "TelnetDescriptors.h"



//
// constants
//

enum
{
	kStatusIconInterSpacing = 8,
	kStatusIconWidth = 16,
	kStatusIconHeight = 16,
	kChasingArrowsTotalAreaH = kStatusIconInterSpacing + CHASING_ARROWS_WD
};

//
// types
//

struct TerminalWindowHeader
{
	struct
	{
		ControlRef		placard,
						contentPane,
						displayFrame,
						displayText,
						statusIconAnchorPane,
						chasingArrows,
						triangle;
	} controls;
	
	WindowRef									owningWindow;
	Str255										statusText;		// retained in case displayed version needs truncation
	ChainRef									statusIcons;	// list of StatusIcon structures
	TerminalWindowHeaderStateChangeProcPtr		stateChangeProc;
	Boolean										isCollapsed;
	UInt16										normalWidth,
												normalHeight;
};
typedef struct TerminalWindowHeader	TerminalWindowHeader;
typedef TerminalWindowHeader*		TerminalWindowHeaderPtr;

//
// internal methods
//

static SInt32						getCollapsedHeight		(TerminalWindowHeaderPtr		inPtr);

static SInt32						getCollapsedWidth		(TerminalWindowHeaderPtr		inPtr);

static void							getNthStatusIconBounds	(TerminalWindowHeaderPtr		inPtr,
															 UInt16							inN,	// 1 to 16
															 Rect*							outRect);

static void							getStatusIconsBounds	(TerminalWindowHeaderPtr		inPtr,
															 Rect*							outRect);

static TerminalWindowHeaderPtr		refAcquireLock			(TerminalWindowHeaderRef		inRef);

static void							refReleaseLock			(TerminalWindowHeaderRef		inRef,
															 TerminalWindowHeaderPtr*		inoutPtr);

static void							setUpPlacard			(TerminalWindowHeaderRef		inRef);



//
// public methods
//

/*!
To create a new terminal window header control
that will appear in the specified window, use
this method.

If any problems occur, null is returned;
otherwise, a reference to the new window
header is returned.

(3.0)
*/
TerminalWindowHeaderRef
TerminalWindowHeader_New	(WindowRef	inOwningWindow)
{
	TerminalWindowHeaderRef		result = (TerminalWindowHeaderRef)Memory_NewPtr(sizeof(TerminalWindowHeader));
	
	
	if (result != null)
	{
		enum
		{
			kGroupBoxHeight = (TEXT_HT_SMALL + INTEGER_DOUBLED(VSP_GROUPBOX_AND_HELPTEXT))
		};
		TerminalWindowHeaderPtr		ptr = refAcquireLock(result);
		ControlRef					control = null;
		
		
		ptr->owningWindow = inOwningWindow;
		ptr->statusIcons = Chain_New(kChainStyleSinglyLinked);
		
		// create controls and set up an embedding hierarchy
		ptr->controls.placard = control = GetNewControl(kIDTerminalWindowHeaderPlacard, ptr->owningWindow);
		
		ptr->controls.contentPane = control = GetNewControl(kIDAnchorUserPane, ptr->owningWindow);
		(OSErr)EmbedControl(control, ptr->controls.placard);
		
		ptr->controls.displayFrame = control = GetNewControl(kIDTerminalWindowHeaderDisplayGroupBox, ptr->owningWindow);
		(OSErr)EmbedControl(control, ptr->controls.contentPane);
		
		ptr->controls.displayText = control = GetNewControl(kIDStaticText, ptr->owningWindow);
		(OSErr)EmbedControl(control, ptr->controls.displayFrame);
		
		ptr->controls.statusIconAnchorPane = control = GetNewControl(kIDAnchorUserPane, ptr->owningWindow);
		(OSErr)EmbedControl(control, ptr->controls.contentPane);
		
		ptr->controls.chasingArrows = control = GetNewControl(kIDAsynchronousArrows, ptr->owningWindow);
		(OSErr)EmbedControl(control, ptr->controls.statusIconAnchorPane);
		TerminalWindowHeader_SetChasingArrowsVisible(result, false/* initially make the arrows invisible */);
		
		ptr->controls.triangle = control = GetNewControl(kIDDisclosureTriangleDownReversed, ptr->owningWindow);
		ptr->isCollapsed = false;
		(OSErr)EmbedControl(control, ptr->controls.placard);
		
		// size controls (some are sized elsewhere)
		control = ptr->controls.placard;
		SizeControl(control, 0/* to be set later */,
					ptr->normalHeight = INTEGER_MAXIMUM(kGroupBoxHeight, ICON_HT_SMALL) +
										INTEGER_DOUBLED(MINIMUM_PLACARD_CUSHION) + 2/* bevel width, top & bottom */);
		control = ptr->controls.displayFrame;
		SizeControl(control, 0/* to be set later */, kGroupBoxHeight);
		(UInt16)Localization_SetUpSingleLineTextControl(ptr->controls.displayText, EMPTY_PSTRING, kThemeSmallSystemFont);
		
		// arrange controls (some are arranged elsewhere)
		{
			Rect		frameBounds,
						textBounds;
			
			
			GetControlBounds(ptr->controls.displayFrame, &frameBounds);
			GetControlBounds(ptr->controls.displayText, &textBounds);
			control = ptr->controls.displayText;
			MoveControl(control, HSP_GROUPBOX_AND_HELPTEXT,
						INTEGER_HALVED((frameBounds.bottom - frameBounds.top) - (textBounds.bottom - textBounds.top)) + 1);
		}
		control = ptr->controls.contentPane;
		MoveControl(control, MINIMUM_PLACARD_CUSHION, MINIMUM_PLACARD_CUSHION);
		
		// release lock
		refReleaseLock(result, &ptr);
	}
	return result;
}// New


/*!
This method cleans up a terminal window header by
destroying all of the data associated with it.  On
output, your copy of the given reference will be
set to null.

(3.0)
*/
void
TerminalWindowHeader_Dispose	(TerminalWindowHeaderRef*	inoutRef)
{
	if (inoutRef != null)
	{
		TerminalWindowHeaderPtr		ptr = refAcquireLock(*inoutRef);
		
		
		if (ptr != null)
		{
			DisposeControl(ptr->controls.placard); // this destroys all of the header controls
			Chain_Dispose(&ptr->statusIcons);
			Memory_DisposePtr((Ptr*)inoutRef), ptr = null;
		}
	}
}// Dispose


/*!
To determine the dimensions of a terminal window
header when it is collapsed, use this method.

(3.0)
*/
void
TerminalWindowHeader_GetCollapsedSize	(TerminalWindowHeaderRef	inRef,
										 Point*						outSize)
{
	if (outSize != null)
	{
		TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null) SetPt(outSize, getCollapsedWidth(ptr), getCollapsedHeight(ptr));
		refReleaseLock(inRef, &ptr);
	}
}// GetCollapsedSize


/*!
To acquire a handle to a control that encompasses
the entire window header (so you can insert or
remove all of the headers contents in a window),
use this method.  The boundaries of the returned
control are the same as those of the window header
itself.

(3.0)
*/
ControlRef
TerminalWindowHeader_GetContainerControl	(TerminalWindowHeaderRef	inRef)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	ControlRef					result = null;
	
	
	if (ptr != null)
	{
		result = ptr->controls.placard;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetContainerControl


/*!
To obtain a new region (*local* to the window) defining
the specified component, use this method.  You must
dispose of the region (using Memory_DisposeRegion())
when you are finished with it.  If any problems occur,
"null" is returned.

IMPORTANT:	These regions are dynamic; do not retain a
			region for any length of time if it is at
			all possible that the event loop will run,
			because the user could modify a header so as
			to relocate components, making your region
			handle the improper size.

Incomplete.

(3.0)
*/
RgnHandle
TerminalWindowHeader_GetNewComponentRegion	(TerminalWindowHeaderRef			inRef,
											 TerminalWindowHeaderComponentType	inWhatComponent)
{
	RgnHandle		result = Memory_NewRegion();
	
	
	if (result != null)
	{
		TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null)
		{
			Rect		rect;
			
			
			SetRect(&rect, 0, 0, 0, 0);
			switch (inWhatComponent)
			{
				case kTerminalWindowHeaderComponentTypePanel:
					GetControlBounds(ptr->controls.displayFrame, &rect);
					break;
				
				case kTerminalWindowHeaderComponentTypeCollapsingTriangle:
					GetControlBounds(ptr->controls.triangle, &rect);
					break;
				
				#if 0
				case 1:
				case 2:
				case 3:
				case 4:
				case 5:
				case 6:
				case 7:
				case 8:
				case 9:
				case 10:
				case 11:
				case 12:
				case 13:
				case 14:
				case 15:
				case 16:
					GetControlBounds(ptr->, &rect);
					break;
				#endif
				
				default:
					break;
			}
			RectRgn(result, &rect);
			refReleaseLock(inRef, &ptr);
		}
		else Memory_DisposeRegion(&result);
	}
	return result;
}// GetNewComponentRegion


/*!
To obtain a new region (*local* to the window) for the
component that most precisely intersects the given
point (*local* to the window), use this method.  You
must dispose of the region (using Memory_DisposeRegion())
when you are finished with it.  If any problems occur,
"null" is returned.  A constant defining the identity of
the region is returned, unless you pass "null" in the
"outWhatComponentPtr" parameter.

IMPORTANT:	These regions are dynamic; do not retain a
			region for any length of time if it is at
			all possible that the event loop will run,
			because the user could modify a header so as
			to relocate components, making your region
			handle the improper size.

Unimplemented.

(3.0)
*/
RgnHandle
TerminalWindowHeader_GetNewPointRegion	(TerminalWindowHeaderRef				UNUSED_ARGUMENT(inRef),
										 Point									UNUSED_ARGUMENT(inLocalPoint),
										 TerminalWindowHeaderComponentType*		UNUSED_ARGUMENT(outWhatComponentPtr))
{
	RgnHandle		result = null;
	
	
	return result;
}// GetNewPointRegion


/*!
To determine the dimensions of a terminal window
header when it is uncollapsed, use this method.

(3.0)
*/
void
TerminalWindowHeader_GetNormalSize	(TerminalWindowHeaderRef	inRef,
									 Point*						outSize)
{
	if (outSize != null)
	{
		TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
		
		
		if (ptr != null) SetPt(outSize, ptr->normalWidth, ptr->normalHeight);
		refReleaseLock(inRef, &ptr);
	}
}// GetNormalSize


/*!
To acquire a reference to the window that encompasses
the entire window header (so you can create controls
that are owned by the same window), use this method.

(3.0)
*/
WindowRef
TerminalWindowHeader_GetOwningWindow	(TerminalWindowHeaderRef	inRef)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	WindowRef					result = null;
	
	
	if (ptr != null)
	{
		result = ptr->owningWindow;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetOwningWindow


/*!
When your application receives a mouse-down event
in the content area of the terminal window headers
encompassing control (which you can determine with
TerminalWindowHeader_GetContainerControl()), invoke
this method to handle tracking of the controls in
the terminal window header.

If the user invokes any action that requires further
notification outside this module (such as clicking a
disclosure triangle that changes the size of the
window header), the appropriate notification routine
is called automatically.

(3.0)
*/
void
TerminalWindowHeader_HandleContentClick		(TerminalWindowHeaderRef	inRef,
											 EventRecord*				inoutEventPtr)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		GrafPtr				oldPort = null;
		WindowRef			window = TerminalWindowHeader_GetOwningWindow(inRef);
		ControlRef			headerControl = TerminalWindowHeader_GetContainerControl(inRef),
							controlClicked = null,
							focusControl = null;
		ControlPartCode		controlPart = kControlNoPart;
		Point				localMouse = inoutEventPtr->where;
		OSErr				error = noErr;
		
		
		GetPort(&oldPort);
		SetPortWindowPort(window);
		GlobalToLocal(&localMouse);
		controlClicked = FindControlUnderMouse(localMouse, window, &controlPart);
		error = GetKeyboardFocus(window, &focusControl);
		
		// if a focusable control is clicked and not focused, focus it...
		{
			UInt32		featureBits = 0L;
			
			
			(OSErr)GetControlFeatures(controlClicked, &featureBits);
			if ((featureBits & kControlGetsFocusOnClick) && ((focusControl != controlClicked) || (error != noErr)))
			{
				(OSErr)SetKeyboardFocus(window, controlClicked, kControlFocusNextPart);
			}
		}
		
		if ((controlClicked != null) && (controlPart != kControlNoPart))
		{
			Boolean			handled = false;
			
			
			// ...otherwise, handle mouse-down events in controls
			controlPart = HandleControlClick(controlClicked, localMouse, inoutEventPtr->modifiers, (ControlActionUPP)null);
			switch (controlPart)
			{
				case kControlTrianglePart:
					if (controlClicked == ptr->controls.triangle)
					{
						TerminalWindowHeader_SetCollapsed
							(inRef, (GetControlValue(controlClicked) == kControlCheckBoxUncheckedValue));
						handled = true;
					}
					break;
				
				default:
					break;
			}
		}
		
		SetPort(oldPort);
		refReleaseLock(inRef, &ptr);
	}
}// HandleContentClick


/*!
To install a status icon into the set of icons
that appears at the top-right edge of a terminal
window header, use this method.  Every icon you
add has its idle procedure called if you invoke
IdleControls() once through your event loop.

You need TerminalWindowHeader_GetOwningWindow()
to determine the owning window of a header in
order to create status icons that can be placed
in that header.  If you provide a status icon
that has a different owning window than the
window of this header, the owner of the icon
will be changed.

(3.0)
*/
void
TerminalWindowHeader_InstallStatusIcon	(TerminalWindowHeaderRef	inRef,
										 StatusIconRef				inIcon)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ChainLinkRef	node = Chain_NewNode(ptr->statusIcons, inIcon);
		ControlRef		control = null,
						container = null;
		Rect			paneRect;
		
		
		Chain_MoveNode(ptr->statusIcons, node, null/* means end of list */,
						kChainLinkPlacementInsertAfter);
		
		// Arrange the new icon relative to the icon pane, and
		// embed it in the pane - that way, whenever the status
		// icons need to be moved, they can move all at once.
		container = ptr->controls.statusIconAnchorPane;
		control = StatusIcon_GetContainerControl(inIcon);
		GetControlBounds(container, &paneRect);
		MoveControl(control,
					paneRect.left + kChasingArrowsTotalAreaH + (Chain_GetNodeCount(ptr->statusIcons) - 1) *
																	(kStatusIconWidth + kStatusIconInterSpacing),
					paneRect.top);
		(OSErr)EmbedControl(control, container);
		refReleaseLock(inRef, &ptr);
	}
}// InstallStatusIcon


/*!
To determine whether the placard of the specified
window header is hidden, use this method.  In the
hidden state, only the disclosure triangle is
visible, and the window header has a physical
height of zero.

(3.0)
*/
Boolean
TerminalWindowHeader_IsCollapsed	(TerminalWindowHeaderRef	inRef)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	Boolean						result = false;
	
	
	if (ptr != null)
	{
		result = ptr->isCollapsed;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// IsCollapsed


/*!
To indicate the beginning or ending of a background
operation in a terminal window, use this method.
While the chasing arrows are visible, they animate
in a circular path to indicate progress of something.

(3.0)
*/
void
TerminalWindowHeader_SetChasingArrowsVisible	(TerminalWindowHeaderRef	inRef,
												 Boolean					inShowArrows)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		(OSErr)SetControlVisibility(ptr->controls.chasingArrows, inShowArrows/* visibility */, true/* draw */);
		refReleaseLock(inRef, &ptr);
	}
}// SetChasingArrowsVisible


/*!
To change the collapsed state of a window header,
use this method.  A collapsed window header has
zero height, and consists entirely of a disclosure
triangle in a small square placard.

The size change routine (if any) is invoked, to
allow outside code to respond to a change in the
collapsed state of the header.

(3.0)
*/
void
TerminalWindowHeader_SetCollapsed	(TerminalWindowHeaderRef	inRef,
									 Boolean					inIsCollapsed)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		// Collapse or expand the header.  Since the triangle auto-toggles,
		// the control value is the desired header collapse state, not the
		// current collapsed state.
		ControlRef		headerControl = TerminalWindowHeader_GetContainerControl(inRef);
		
		
		// If the triangle is unchecked, it is in the closed position
		// and therefore the window is uncollapsed and needs collapsing.
		// Otherwise, the reverse is true.
		SizeControl(headerControl,
					(inIsCollapsed) ? getCollapsedWidth(ptr) : ptr->normalWidth,
					(inIsCollapsed) ? getCollapsedHeight(ptr) : ptr->normalHeight + 1/* QUICK (bad) fix to cosmetic glitch!!! */);
		
		// set this flag so it is possible to easily determine whether the placard controls are visible
		ptr->isCollapsed = inIsCollapsed;
		
		// if collapsed, move the placard over to the far side of the window; otherwise, move it to (-1, -1)
		setUpPlacard(inRef);
		
		// notify that the size has changed, if possible
		if (ptr->stateChangeProc != null) InvokeTerminalWindowHeaderStateChangeProc(ptr->stateChangeProc, inRef);
		
		refReleaseLock(inRef, &ptr);
	}
}// SetCollapsed


/*!
To install a routine that gets called whenever
the collapsed state of a terminal window header
changes, use this method.

Your state change notification routine should,
at the very least, alter the boundaries of the
window containing the terminal window header
control to accommodate user modifications.

(3.0)
*/
void
TerminalWindowHeader_SetStateChangeProc		(TerminalWindowHeaderRef				inRef,
											 TerminalWindowHeaderStateChangeProcPtr	inProc)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		ptr->stateChangeProc = inProc;
		refReleaseLock(inRef, &ptr);
	}
}// SetStateChangeProc


/*!
Changes the text display in the window header.
Normally, the display contains connection
status and host information.

(3.0)
*/
void
TerminalWindowHeader_SetStatusText	(TerminalWindowHeaderRef	inRef,
									 StringPtr					inText)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		PLstrcpy(ptr->statusText, inText);
		TerminalWindowHeader_SetWidthInPixels(inRef, ptr->normalWidth); // causes text to be set, truncating if necessary
		refReleaseLock(inRef, &ptr);
	}
}// SetStatusText


/*!
To change the width of the window header, use
this method.  Calling this routine automatically
adjusts the controls of the window for the new
width.

Calling this routine does not cause the size
change routine (if any) to be invoked.

(3.0)
*/
void
TerminalWindowHeader_SetWidthInPixels	(TerminalWindowHeaderRef	inRef,
										 UInt16						inWidth)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		Rect		bounds;
		Point		deltaSize;
		Boolean		wasPlacardVisible = false;
		
		
		ptr->normalWidth = inWidth; // save the new width, but...
		
		// ...dont actually apply any visible change if the header is collapsed
		unless (ptr->isCollapsed)
		{
			GetControlBounds(ptr->controls.placard, &bounds);
			SetPt(&deltaSize, inWidth - (bounds.right - bounds.left), bounds.bottom - bounds.top);
			wasPlacardVisible = IsControlVisible(ptr->controls.placard);
			
			SetControlVisibility(ptr->controls.placard, false/* visibility */, true/* draw */);
			SizeControl(ptr->controls.placard, inWidth, bounds.bottom - bounds.top);
			GetControlBounds(ptr->controls.placard, &bounds); // be sure to update "bounds"
			SizeControl(ptr->controls.contentPane,
							inWidth - INTEGER_DOUBLED(MINIMUM_PLACARD_CUSHION) - DISCLOSURETRIANGLE_WD,
							bounds.bottom - bounds.top - INTEGER_DOUBLED(MINIMUM_PLACARD_CUSHION));
			
			{
				Rect		oldFrameArea,
							statusIconsArea;
				UInt16		textWidth = 0;
				
				
				// size the middle pane based on the size of the status icon pane
				GetControlBounds(ptr->controls.displayFrame, &oldFrameArea);
				getStatusIconsBounds(ptr, &statusIconsArea);
				SizeControl(ptr->controls.displayFrame, statusIconsArea.left - oldFrameArea.left - HSP_CONTROLS,
								oldFrameArea.bottom - oldFrameArea.top);
				GetControlBounds(ptr->controls.displayFrame, &bounds);
				textWidth = bounds.right - bounds.left - INTEGER_DOUBLED(MINIMUM_PLACARD_CUSHION);
				{
					Rect		textBounds;
					
					
					GetControlBounds(ptr->controls.displayText, &textBounds);
					SizeControl(ptr->controls.displayText, textWidth, textBounds.bottom - textBounds.top);
				}
				{
					Str255		text;
					
					
					// make sure the content text fits - if not, truncate and add an ellipsis at the end
					PLstrcpy(text, ptr->statusText);
					StringUtilities_PTruncate(text, textWidth, kStringUtilities_TruncateAtEnd);
					SetControlText(ptr->controls.displayText, text);
				}
				
				// move the status icon pane to its proper location, which will automatically move the embedded controls, too
				MoveControl(ptr->controls.statusIconAnchorPane, statusIconsArea.left, statusIconsArea.top);
				SizeControl(ptr->controls.statusIconAnchorPane, statusIconsArea.right - statusIconsArea.left,
								statusIconsArea.bottom - statusIconsArea.top);
				
				// move the placard (necessary only if collapsed), which will automatically move the embedded controls, too
				setUpPlacard(inRef);
			}
			
			// now show all of the changes at once
			SetControlVisibility(ptr->controls.placard, wasPlacardVisible/* visibility */, wasPlacardVisible/* draw */);
		}
		refReleaseLock(inRef, &ptr);
	}
}// SetWidthInPixels


//
// internal methods
//
#pragma mark -

/*!
To determine how high the header should be when
collapsed, use this method.

(3.0)
*/
static SInt32
getCollapsedHeight	(TerminalWindowHeaderPtr	UNUSED_ARGUMENT(inPtr))
{
	SInt32		result = 16;
	
	
	return result;
}// getCollapsedHeight


/*!
To determine how wide the header should be when
collapsed, use this method.

(3.0)
*/
static SInt32
getCollapsedWidth	(TerminalWindowHeaderPtr	UNUSED_ARGUMENT(inPtr))
{
	SInt32		result = 16;	// typical scroll bar width
	
	
	(OSStatus)GetThemeMetric(kThemeMetricScrollBarWidth, &result);
	
	return result;
}// getCollapsedWidth


/*!
To determine the space occupied by the status icon
whose index is the given "inN" value, use this method.
Since icons are installed right-to-left, the icon of
index 1 is the rightmost status icon.

Unimplemented.

(3.0)
*/
static void
getNthStatusIconBounds	(TerminalWindowHeaderPtr	UNUSED_ARGUMENT(inPtr),
						 UInt16						UNUSED_ARGUMENT(inN),	// 1 to 16
						 Rect*						UNUSED_ARGUMENT(outRect))
{
	
}// getNthStatusIconBounds


/*!
To determine the space occupied by the status icon
area (in the top-right corner of the terminal window
header), use this method.  If no status icons are
installed, the area will be very narrow.

(3.0)
*/
static void
getStatusIconsBounds	(TerminalWindowHeaderPtr	inPtr,
						 Rect*						outRect)
{
	Rect		bounds;
	UInt16		iconCount = Chain_GetNodeCount(inPtr->statusIcons);
	
	
	// use the placard area as a guide
	GetControlBounds(inPtr->controls.placard, &bounds);
	
	// set each part of the rectangle sequentially, so that some parts can be re-used to calculate other parts
	SetRect(outRect, 0, 0, bounds.right - MINIMUM_PLACARD_CUSHION - DISCLOSURETRIANGLE_WD - HSP_CONTROLS, 0);
	SetRect(outRect,
			outRect->right - iconCount * kStatusIconWidth - (iconCount - 1) * kStatusIconInterSpacing -
							kChasingArrowsTotalAreaH, // room for all icons and the chasing arrows
			outRect->top,
			outRect->right,
			outRect->bottom);
	SetRect(outRect, outRect->left,
					INTEGER_HALVED((bounds.bottom - bounds.top) - kStatusIconHeight), // vertically centered in the placard
					outRect->right,
					outRect->bottom);
	outRect->bottom = outRect->top + kStatusIconHeight;
}// getStatusIconsBounds


/*!
To obtain a pointer to the internal structure,
given a reference to it, use this method.

(3.0)
*/
static TerminalWindowHeaderPtr
refAcquireLock	(TerminalWindowHeaderRef	inRef)
{
	return ((TerminalWindowHeaderPtr)inRef);
}// refAcquireLock


/*!
To obtain a pointer to the internal structure,
given a reference to it, use this method.

(3.0)
*/
static void
refReleaseLock	(TerminalWindowHeaderRef	UNUSED_ARGUMENT(inRef),
				 TerminalWindowHeaderPtr*	inoutPtr)
{
	// the memory is not relocatable, so nothing needs to be done here
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock


/*!
To position the placard of a terminal window header
properly based on its collapsed state and normal width,
and to hide or show and arrange its contents properly,
use this method.  A collapsed header has a fixed size,
and is positioned at the far right end of a window.
Conversely, an uncollapsed header has a dynamic size
and occupies the entire window width - therefore, its
position is always (-1, -1).

(3.0)
*/
static void
setUpPlacard	(TerminalWindowHeaderRef	inRef)
{
	TerminalWindowHeaderPtr		ptr = refAcquireLock(inRef);
	
	
	if (ptr != null)
	{
		SInt16		amountToAdd = (ptr->isCollapsed) ? (ptr->normalWidth - getCollapsedWidth(ptr)) : 0;
		Rect		bounds;
		
		
		MoveControl(ptr->controls.placard, -1 + amountToAdd, -1);
		GetControlBounds(ptr->controls.placard, &bounds);
		
		// if the placard is collapsed, most controls should be hidden; otherwise, display them
		(OSErr)SetControlVisibility(ptr->controls.contentPane, !ptr->isCollapsed/* visibility */, true/* draw */);
		
		// move the disclosure triangle
		MoveControl(ptr->controls.triangle, bounds.right - INTEGER_HALVED(getCollapsedWidth(ptr)) -
					INTEGER_HALVED(DISCLOSURETRIANGLE_WD),
					bounds.top + MINIMUM_PLACARD_CUSHION);
		refReleaseLock(inRef, &ptr);
	}
}// setUpPlacard