/*###############################################################

	TextDataFile.c
	
	Data Access Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <stdio.h>
#include <string.h>

// Mac includes
#include <Files.h>
#include <MacErrors.h>
#include <PLStringFuncs.h>
#include <TextUtils.h>

// library includes
#include "DataAccessLib.h"



// types
struct TextDataFile
{
	SInt16		lockCount;				// how many locks are on the handle
	
	struct
	{
		AliasHandle		alias;			// allows the user to move the file without screwing anything up
		SInt16			refNum;			// reference number of open file
		Boolean			isOpen;			// whether the file was closed (happens after a save)
	} file;
};
typedef struct TextDataFile		TextDataFile;
typedef TextDataFile*			TextDataFilePtr;
typedef TextDataFilePtr*		TextDataFileHandle;

// internal methods
static char*			getNameValue				(char*					inoutLine);
static SInt32			readLine					(SInt16 const			inFileRefNum,
													 char*					outLine,
													 SInt16 const			inCapacityLength);
static TextDataFilePtr	refAcquireLock				(TextDataFileRef		inRef);
static void				refReleaseLock				(TextDataFileRef		inRef,
													 TextDataFilePtr*		inoutPtr);



//
// public methods
//

/*!
Creates a new Text Data File object that accesses
the specified file.  If the file does not already
exist, it is created, using the specified type,
creator and script code.  The file must be readable
and writeable, and is opened for access.

An alias to the file is maintained internally, so
the user can move the file around and not affect
this module.

If any problems occur, "null" is returned; otherwise,
a reference to the new object is returned.

(1.0)
*/
TextDataFileRef
TextDataFile_New	(FSSpec const*			inFSSpecPtr,
					 OSType const			inCreator,
					 OSType const			inType,
					 ScriptCode const		inInternationalCode)
{
	TextDataFileRef		result = null;
	
	
	result = (TextDataFileRef)Memory_NewHandle(sizeof(TextDataFile));
	if (result != null)
	{
		TextDataFilePtr		ptr = refAcquireLock(result);
		OSErr				error = noErr;
		
		
		if (ptr != null)
		{
			// find and open the file for reading and writing
			FSSpec		file;
			
			
			error = FSMakeFSSpec(inFSSpecPtr->vRefNum, inFSSpecPtr->parID, inFSSpecPtr->name, &file);
			if (error == fnfErr) error = FSpCreate(&file, inCreator, inType, inInternationalCode);
			if (error == noErr) error = FSpOpenDF(&file, fsRdWrPerm, &ptr->file.refNum);
			ptr->file.isOpen = (error == noErr);
			if (error == noErr) error = NewAlias(null/* no source */, &file/* target*/, &ptr->file.alias);
			
			refReleaseLock(result, &ptr);
		}
		if (error != noErr) TextDataFile_Dispose(&result);
	}
	return result;
}// New


/*!
Deletes the in-memory representation of a Text Data File
that you created with TextDataFile_New(), and closes the
file on disk.

(1.0)
*/
void
TextDataFile_Dispose		(TextDataFileRef*		inoutRefPtr)
{
	if (inoutRefPtr != null)
	{
		TextDataFilePtr		ptr = refAcquireLock(*inoutRefPtr);
		
		
		FSClose(ptr->file.refNum), ptr->file.refNum = -1;
		ptr->file.isOpen = false;
		DisposeHandle((Handle)*inoutRefPtr), *inoutRefPtr = null;
	}
}// Dispose


/*!
Writes a comment to your text data file.  Since the
order of your operations is retained, you can invoke
this routine to write a comment immediately prior to
invoking a routine to set a names value, such as
TextDataFile_AddNameValue(), and the comment will
appear immediately prior to the assignment statement
in the resultant text file.

The specified text is automatically preceded with the
proper comment syntax (a hash mark, '#') so you
shouldnt include anything in "inCommentTextPtr"
except the actual text of the comment.

If successful, "true" is returned.

(1.0)
*/
Boolean
TextDataFile_AddComment		(TextDataFileRef		inRef,
							 char const*			inCommentTextPtr)
{
	TextDataFilePtr		ptr = refAcquireLock(inRef);
	Boolean				result = false;
	
	
	if ((ptr != null) && (inCommentTextPtr != null))
	{
		SInt32		count = 0;
		
		
		// precede the text with a comment hash-mark ('#'), and then write the comment text
		count = 2;
		(OSErr)FSWrite(ptr->file.refNum, &count, "# ");
		count = strlen((char*)inCommentTextPtr);
		(OSErr)FSWrite(ptr->file.refNum, &count, inCommentTextPtr);
		result = true;
		
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// AddComment


/*!
To specify a name-value pair, use this method.  In
the text file, this will appear on a single line in
some format, such as "name = value".

Only alphanumeric characters, i.e. A-Z, a-z and 0-9,
as well as the underscore ('_'), are allowed in
names.

Values may contain most characters, even a carriage-
return (octal 15), as long as no carriage-return
character within the value is followed by a line-feed
(octal 12).  (A line-feed following a carriage-return
marks a Mac OS end-of-line.)

If the name was successfully assigned a value, "true"
is returned; otherwise, "false" is returned.

(1.0)
*/
Boolean
TextDataFile_AddNameValue		(TextDataFileRef		inRef,
								 char const*			inClassNameOrNull,
								 char const*			inPropertyName,
								 char const*			inValue)
{
	TextDataFilePtr		ptr = refAcquireLock(inRef);
	Boolean				result = false;
	
	
	if (ptr != null)
	{
		SInt32		count = 0;
		
		
		if (inClassNameOrNull != null)
		{
			// write class name and a dot (.)
			count = strlen((char*)inClassNameOrNull);
			(OSErr)FSWrite(ptr->file.refNum, &count, inClassNameOrNull);
			count = 1;
			(OSErr)FSWrite(ptr->file.refNum, &count, ".");
		}
		
		// write property name
		count = strlen((char*)inPropertyName);
		(OSErr)FSWrite(ptr->file.refNum, &count, inPropertyName);
		
		// write equal sign
		count = 3;
		(OSErr)FSWrite(ptr->file.refNum, &count, " = ");
		
		// write value string
		count = strlen((char*)inValue);
		(OSErr)FSWrite(ptr->file.refNum, &count, inValue);
		
		// write newline
		count = 2;
		(OSErr)FSWrite(ptr->file.refNum, &count, "\015\012");
		
		result = true;
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// AddNameValue


/*!
To specify a name-value pair that has a true or false
value, use this method.  In the text file, this will
appear on a single line in some format, such as
"name = yes" or "name = no".

If the name was successfully assigned a value, "true"
is returned; otherwise, "false" is returned.

(1.0)
*/
Boolean
TextDataFile_AddNameValueFlag	(TextDataFileRef		inRef,
								 char const*			inClassNameOrNull,
								 char const*			inPropertyName,
								 Boolean const			inValue)
{
	return TextDataFile_AddNameValue(inRef, inClassNameOrNull, inPropertyName, (inValue) ? "yes" : "no");
}// AddNameValueFlag


/*!
To specify a name-value pair that has a numerical
value, use this method.  In the text file, this will
appear on a single line in some format, such as
"name = 1325" or "name = -241".

If the name was successfully assigned a value, "true"
is returned; otherwise, "false" is returned.

(1.0)
*/
Boolean
TextDataFile_AddNameValueNumber		(TextDataFileRef		inRef,
									 char const*			inClassNameOrNull,
									 char const*			inPropertyName,
									 SInt32 const			inValue)
{
	char	buffer[10];
	
	
	sprintf(buffer, "%ld", inValue);
	return TextDataFile_AddNameValue(inRef, inClassNameOrNull, inPropertyName, buffer);
}// AddNameValueNumber


/*!
To specify a name-value pair that has a Pascal string
value, use this method.  In the text file, this will
appear on a single line in some format, such as
"name = value".

If the name was successfully assigned a value, "true"
is returned; otherwise, "false" is returned.

(1.0)
*/
Boolean
TextDataFile_AddNameValuePString	(TextDataFileRef		inRef,
									 char const*			inClassNameOrNull,
									 char const*			inPropertyName,
									 StringPtr				inValue)
{
	char	string[255];
	
	
	StringUtilities_PToC(inValue, string);
	return TextDataFile_AddNameValue(inRef, inClassNameOrNull, inPropertyName, string);
}// AddNameValuePString


/*!
To specify a name-value pair that has an RGB color
value, use this method.  In the text file, this will
appear on a single line in some format, such as
"name = {65535,255,0}".

If the name was successfully assigned a value, "true"
is returned; otherwise, "false" is returned.

(1.0)
*/
Boolean
TextDataFile_AddNameValueRGBColor	(TextDataFileRef		inRef,
									 char const*			inClassNameOrNull,
									 char const*			inPropertyName,
									 RGBColor const*		inValuePtr)
{
	char	buffer[10];
	
	
	sprintf(buffer, "{%ld,%ld,%ld}", inValuePtr->red, inValuePtr->green, inValuePtr->blue);
	return TextDataFile_AddNameValue(inRef, inClassNameOrNull, inPropertyName, buffer);
}// AddNameValueRGBColor


/*!
Determines where a text data file is.  Since an alias
is maintained internally, this routine will find the
file, no matter where the user may have moved it to
since the Text Data File object was first created.

If successful, "true" is returned.

(1.0)
*/
Boolean
TextDataFile_GetFSSpec		(TextDataFileRef		inRef,
							 FSSpec*				outFSSpecPtr)
{
	TextDataFilePtr		ptr = refAcquireLock(inRef);
	Boolean				result = false;
	
	
	if ((ptr != null) && (outFSSpecPtr != null))
	{
		Boolean		unused = false;
		
		
		result = (ResolveAlias(null/* from where */, ptr->file.alias, outFSSpecPtr,
								&unused/* did alias change? */) == noErr);
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetFSSpec


/*!
To obtain the value associated with the specified
name, use this method.  If the value could be
read, "true" is returned and the specified string
will be filled in; otherwise, "false" is returned.

(1.0)
*/
Boolean
TextDataFile_GetNextNameValue	(TextDataFileRef	inRef,
								 char*				outNamePtr,
								 SInt16 const		inNameLength,
								 char*				outValuePtr,
								 SInt16 const		inValueLength)
{
	TextDataFilePtr		ptr = refAcquireLock(inRef);
	Boolean				result = false;
	
	
	if (ptr != null)
	{
		char	buffer[256];
		
		
		if (readLine(ptr->file.refNum, buffer, 256/* length of buffer */) > 0)
		{
			char const*		namePtr = getNameValue(buffer); // on output, "buffer" is only the value
			SInt16			length = 0;
			
			
			// copy the name
			length = strlen(namePtr);
			if (length > inNameLength) length = inNameLength;
			strncpy(outNamePtr, namePtr, length);
			
			// copy the value
			length = strlen(buffer);
			if (length > inValueLength) length = inValueLength;
			strncpy(outValuePtr, buffer, length);
			
			result = true;
		}
		refReleaseLock(inRef, &ptr);
	}
	return result;
}// GetNextNameValue


/*!
Resets the file position of the specified data file
to point to the top of the file.  This affects all
TextDataFile_GetNext...() routines.

If the reset succeeds, "true" is returned; otherwise,
"false" is returned.

(1.0)
*/
Boolean
TextDataFile_ResetPointer	(TextDataFileRef	inRef)
{
	TextDataFilePtr		ptr = refAcquireLock(inRef);
	Boolean				result = false;
	
	
	if (ptr != null)
	{
		if (SetFPos(ptr->file.refNum, fsFromStart/* mode */,
					0/* offset from position indicated by mode */) == noErr)
		{
			// success
			result = true;
		}
	}
	return result;
}// ResetPointer


/*!
Determines a Boolean value using the contents of the
specified string.  If the value could be read, "true"
is returned and the specified flag will be filled in;
otherwise, "false" is returned.

A special heuristic is used to analyze the text and
determine whether it has affirmative or negative
meaning.  However, this routine is guaranteed to
work if you write a flag to the file in the first
place using the TextDataFile_AddNameValueFlag()
method.

(1.0)
*/
Boolean
TextDataFile_StringToFlag	(TextDataFileRef	UNUSED_ARGUMENT(inRef),
							 char const*		inStringPtr,
							 Boolean*			outValuePtr)
{
	Boolean		result = false;
	
	
	if ((outValuePtr != null) && (inStringPtr != null))
	{
		char		buffer[32];
		
		
		// currently a little bit simplistic, but...
		strncpy(buffer, inStringPtr, sizeof(buffer));
		StringUtilities_CToPInPlace(buffer);
		if (EqualString((StringPtr)buffer, "\pyes", false/* case sensitive */, true/* diacritical sensitive */))
		{
			*outValuePtr = true;
		}
		else
		{
			*outValuePtr = false;
		}
		result = true;
	}
	return result;
}// StringToFlag


/*!
Determines a signed, numeric value using the contents
of the specified string.  If the value could be read,
"true" is returned and the specified number will be
filled in; otherwise, "false" is returned.

(1.0)
*/
Boolean
TextDataFile_StringToNumber		(TextDataFileRef	UNUSED_ARGUMENT(inRef),
								 char const*		inStringPtr,
								 SInt32*			outValuePtr)
{
	Boolean		result = false;
	
	
	if ((outValuePtr != null) && (inStringPtr != null))
	{
		char		buffer[32];
		
		
		strncpy(buffer, inStringPtr, sizeof(buffer));
		StringUtilities_CToPInPlace(buffer);
		StringToNum((StringPtr)buffer, outValuePtr);
		result = true;
	}
	return result;
}// StringToNumber


/*!
Determines a red-green-blue color value using the contents
of the specified string.  If the value could be read,
"true" is returned and the specified color data will be
filled in; otherwise, "false" is returned.

(1.0)
*/
Boolean
TextDataFile_StringToRGBColor	(TextDataFileRef	UNUSED_ARGUMENT(inRef),
								 char const*		inStringPtr,
								 RGBColor*			outValuePtr)
{
	Boolean		result = false;
	
	
	if ((outValuePtr != null) && (inStringPtr != null))
	{
		UInt32		r = 0,
					g = 0,
					b = 0;
		
		
		if (3 == sscanf(inStringPtr, "{%u,%u,%u}", &r, &g, &b))
		{
			// three numbers successfully read
			outValuePtr->red = r;
			outValuePtr->green = g;
			outValuePtr->blue = b;
			result = true;
		}
	}
	return result;
}// StringToRGBColor



//
// internal methods
//

/*!
To search for the name-value delimiter within
the given line of text and return a pointer to
only the name, use this method.  On return,
"inoutLine" is reduced to the value.  Any
whitespace surrounding the delimiter is not
stripped, but remains a part of either the
name or the value, depending upon which side
of the delimiter the whitespace was on.

Since this method uses the lines string space
to construct the name and value, if you ever
dispose of the memory occupied by the line of
text, then the name and value strings are also
disposed from memory!  Make copies as needed.

(1.0)
*/
static char*
getNameValue	(char*		inoutLine)
{
	char*		result = null;
	char*		delimiters = "=";
	
	
	result = strtok(inoutLine, delimiters);
	inoutLine = strtok(null, delimiters);
	return result;
}// getNameValue


/*!
To read a single line from a file, use this
method.  The number of characters actually
read is returned.  The new-line characters
at the end of a line are automatically
stripped from the string prior to returning.

(1.0)
*/
static SInt32
readLine	(SInt16 const		inFileRefNum,
			 char*				outLine,
			 SInt16 const		inCapacityLength)
{
	OSErr			error = noErr;
	SInt32			readHowMany = 1,
					count = 0L;
	Boolean			sawCR = false,
					sawNewLine = false;
	char*			ptr = null;
	
	
	for (ptr = outLine; (!sawNewLine) && (*ptr != EOF) && (count < inCapacityLength); ptr++)
	{
		if ((error = FSRead(inFileRefNum, &readHowMany, ptr)) == eofErr) *ptr = EOF;
		else
		{
			if (readHowMany != 1) readHowMany = 1; // reset count (read failed)
			else count++;
			
			if (*ptr == '\015') sawCR = true;
			else if (sawCR)
			{
				if (*ptr == '\012') sawNewLine = true; // CR and LF marks end-of-line
				else unless (*ptr == '\015') sawCR = false; // without LF, it is necessary to seek a new CR
			}
		}
		
		// if a new-line was encountered, strip the CR and LF from the result
		if (sawNewLine)
		{
			// I could use a pretty one-liner here, but I want to make sure the compiler gets it right
			(ptr--)[0] = '\0';
			ptr[0] = '\0';
		}
	}
	
	return count;
}// readLine


/*!
To obtain a pointer to the internal TextDataFile
structure, given a reference to it, use this
method.

(1.0)
*/
static TextDataFilePtr
refAcquireLock	(TextDataFileRef		inRef)
{
	TextDataFilePtr		result = *((TextDataFileHandle)inRef);
	
	
	HLock((Handle)inRef);
	result->lockCount++;
	return result;
}// refAcquireLock


/*!
To release a pointer to the internal TextDataFile
structure, given a reference to it, use this
method.

(1.0)
*/
static void
refReleaseLock	(TextDataFileRef		inRef,
				 TextDataFilePtr*		inoutPtr)
{
	TextDataFilePtr		ptr = *((TextDataFileHandle)inRef);
	
	
	if (--ptr->lockCount <= 0) HUnlock((Handle)inRef);
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock