/*###############################################################

	TextTranslation.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstring>
#else
#	include <string.h>
#endif

// Mac includes
#include <Resources.h>
#include <Script.h>
#include <TextCommon.h>
#include <TextEncodingConverter.h>

// resource includes
#include "GeneralResources.h"
#include "StringResources.h"

// MacTelnet includes
#include "Console.h"
#include "ResourceFileManager.h"
#include "TelnetHeader.h"
#include "TelnetMenus.h"
#include "TextTranslation.h"



// variables
static TECObjectRef		gTextEncodingConverter = null;
static BytePtr			DefaultTable = null;
static BytePtr			FTPinTable = null,
						FTPoutTable = null;
static Handle			transTablesHdl = null;
short					nNational = 0;

// internal methods
static BytePtr		getTableArray				(SInt16 const		inTable,
												 Boolean const		inGetNonMacTable);
static BytePtr		getTranslationResource		(SInt16 const		inResourceID);



/*!
Reads all available translation tables and appends
their names to the given menu.  You must invoke
this routine before any other routines from this
module.

(2.6)
*/
void
TextTranslation_Init		(MenuHandle		inTranslationMenu)
{
	short		i = 0,
				numTables = 0,
				firstTranslationTableIndex = 2;
	short		oldResourceFile = CurResFile();
	Handle		h = null;
	
	
 	nNational = 0;
	
	// The Default table (i.e. no translation) and the two FTP tables are stored in the
	// applications resource fork as resources of type 'TRSL'.  The tables added and
	// removed by the user are stored in the preferences file as resources of type 'taBL'.
	// This loads the default table and the two FTP tables into memory; failures can only
	// be caused by an incorrect application resource fork.
	DefaultTable = getTranslationResource(TRSL_DEFAULT_TABLE);
	FTPinTable = getTranslationResource(TRSL_FTP_TABLE);
	FTPoutTable = FTPinTable + 256;
	
	transTablesHdl = Memory_NewHandle(0);
	ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
	numTables = Count1Resources(USER_TRSL);
	if (numTables > 0)
	{
		Str255		firstItemString;
		
		
		GetIndString(firstItemString, rStringsMiscellaneous, siNone);
		AppendMenu(inTranslationMenu, firstItemString);
		InsertResMenu(inTranslationMenu, USER_TRSL, 999/* insert after item */);
		
		for (i = firstTranslationTableIndex; i <= (firstTranslationTableIndex + numTables - 1); i++)
		{
			Str255		menuItemName;
			
			
			GetMenuItemText(inTranslationMenu, i, menuItemName);
			h = Get1NamedResource(USER_TRSL, menuItemName);
			if (ResError() == noErr && (GetHandleSize(h) == 512))
			{
				nNational++;
				
				// append the tables data to the master array of table data
				HUnlock(transTablesHdl);
				Memory_SetHandleSize(transTablesHdl, (nNational * 512));
				HLockHi(transTablesHdl);
				HLock(h);
				BlockMove(*h, (*transTablesHdl) + ((nNational - 1) * 512), 512);
			}
			ReleaseResource(h);
		}
	}
	
	UseResFile(oldResourceFile);
}// Init


/*!
Persistently translates a buffer using the given
text encoding converter (i.e. if the converter
cannot translate all the text at once, this
method keeps trying until it does).  Since the
encoded text might be longer than the original
(e.g. a single character requiring multiple bytes
to represent using a different encoding), it
cannot be translated in place - therefore, a new
handle is returned containing the translated text.
If there are memory problems, "memFullErr" is the
result, and the output handle is invalid.

(3.0)
*/
OSStatus
TextTranslation_ConvertBufferToNewHandle	(unsigned char*			inText,
											 ByteCount const		inLength,
											 TECObjectRef			inConverter,
											 Handle*				outNewHandleWithTranslatedText)
{
	OSStatus			result = noErr;
	
	
	if ((inText != null) && (outNewHandleWithTranslatedText != null))
	{
		enum
		{
			kOutputBufferSize = 4096	// arbitrary
		};
		Ptr					outputBuffer = null;
		unsigned char*		inputPtr = null;
		ByteCount			inputLength = 0,
							actualInputLength = 0,
							actualOutputLength = 0;
		
		
		// create a handle that will ultimate contain all of the translated text
		*outNewHandleWithTranslatedText = Memory_NewHandleInProperZone(kOutputBufferSize, false/* only need briefly */);
		if (*outNewHandleWithTranslatedText == null) result = memFullErr;
		else
		{
			inputLength = inLength;
			inputPtr = inText;
			outputBuffer = Memory_NewPtr(kOutputBufferSize);
			if (outputBuffer == null) result = memFullErr;
			else
			{
				Ptr			translatedBufferPtr = null;
				Boolean		stillConverting = true;
				
				
				HLock(*outNewHandleWithTranslatedText);
				translatedBufferPtr = *(*outNewHandleWithTranslatedText);
				
				// convert portions of the text until it is all converted
				do
				{
					if (stillConverting)
					{
						// convert a chunk of text
						result = TECConvertText(inConverter, inputPtr, inputLength, &actualInputLength,
												(unsigned char*)outputBuffer, kOutputBufferSize, &actualOutputLength);
						if (result != kTECUsedFallbacksStatus) stillConverting = false;
					}
					else
					{
						// flush remaining text from converters internal buffers
						result = TECFlushText(inConverter, (unsigned char*)outputBuffer,
												kOutputBufferSize, &actualOutputLength);
					}
					
					if (result == noErr)
					{
						// append the newly-translated text to the final buffer; but first, make sure
						// there is enough memory to resize the output buffer to accommodate the text
						result = Memory_SetHandleSize(*outNewHandleWithTranslatedText,
														GetHandleSize(*outNewHandleWithTranslatedText) +
															actualOutputLength);
						if (result == noErr)
						{
							BlockMove(outputBuffer, translatedBufferPtr, actualOutputLength);
							translatedBufferPtr += actualOutputLength;
						}
					}
					inputPtr += actualInputLength;
					inputLength -= actualInputLength;
				} while ((result == kTECUsedFallbacksStatus) || (result == kTECNeedFlushStatus));
				
				Memory_DisposePtr(&outputBuffer);
				HUnlock(*outNewHandleWithTranslatedText);
			}
		}
	}
	
	return result;
}// ConvertBufferToNewHandle


/*!
Converts a single character from the 8-bit
Macintosh encoding to an 8-bit international
encoding.  The new character is returned.

(2.6)
*/
unsigned char
TextTranslation_ConvertCharacterFromMac		(unsigned char*		inoutCharacterPtr,
											 SInt16 const		inWhichTable)
{
	short		b = 0;
	BytePtr		tableArray = getTableArray(inWhichTable, true/* get non-Mac table */);
	
	
	b = (short)*inoutCharacterPtr;
	*inoutCharacterPtr = tableArray[b];
	return (*inoutCharacterPtr);
}// ConvertCharacterFromMac


/*!
Converts a single character from ISO encoding
to the 8-bit Macintosh encoding.  The new
character is returned.

(2.6)
*/
unsigned char
TextTranslation_ConvertISOCharacterToMac		(unsigned char*		inoutCharacterPtr)
{
	short	b = 0;
	
	
	b = (short)*inoutCharacterPtr;
	*inoutCharacterPtr = FTPinTable[b];
	return (*inoutCharacterPtr);
}// ConvertISOCharacterToMac


/*!
Converts a series of characters from ISO encoding
to the 8-bit Macintosh encoding.

(2.6)
*/
void
TextTranslation_ConvertISOTextToMac		(unsigned char*		inoutBuffer,
										 SInt16 const		inLength)
{
	short			i = 0;
	unsigned char	ascii = '\0';
	unsigned char*	p = null;
	
	
	for (i = 0, p = inoutBuffer; i < inLength; i++, p++)
	{
		ascii = *p;
		*p = TextTranslation_ConvertISOCharacterToMac(&ascii);
	}
}// ConvertISOTextToMac


/*!
Converts a single character from the 8-bit
Macintosh encoding to ISO encoding.  The new
character is returned.

(2.6)
*/
unsigned char
TextTranslation_ConvertMacCharacterToISO		(unsigned char*		ascii)
{
	short	b = '\0';
	
	
	b = (short)*ascii;
	*ascii = FTPoutTable[b];
	return (*ascii);
}// ConvertMacCharacterToISO


/*!
Converts a series of characters from the 8-bit
Macintosh encoding to ISO encoding.

(2.6)
*/
void
TextTranslation_ConvertMacTextToISO		(unsigned char*		inoutBuffer,
										 SInt16 const		inLength)
{
	short			i = 0;
	unsigned char	ascii = '\0';
	unsigned char*	p = null;
	
	
	for (i = 0, p = inoutBuffer; i < inLength; i++, p++)
	{
		ascii = *p;
		*p = TextTranslation_ConvertMacCharacterToISO(&ascii);
	}
}// ConvertMacTextToISO


/*!
Converts a series of characters from the 8-bit
Macintosh encoding to an 8-bit international
encoding.

(2.6)
*/
void
TextTranslation_ConvertTextFromMac	(unsigned char*		inoutBuffer,
									 SInt16 const		inLength,
									 SInt16 const		inWhichTable)
{
	short			i = 0;
	unsigned char	ascii = '\0';
	unsigned char*	p = null;
	
	
	for (i = 0, p = inoutBuffer; i < inLength; i++, p++)
	{
		ascii = *p;
		*p = TextTranslation_ConvertCharacterFromMac(&ascii, inWhichTable);
	}
}// ConvertTextFromMac


/*!
Converts a series of characters from an 8-bit
international encoding to the 8-bit Macintosh
encoding.

(2.6)
*/
void
TextTranslation_ConvertTextToMac	(unsigned char*			inoutBuffer,
									 SInt16 const			inLength,
									 SInt16 const			inWhichTable)
{
	short			i = 0;
	unsigned char*	p = null;
	BytePtr			tableArray = null;
	
	
	tableArray = getTableArray(inWhichTable, false/* get non-Mac table */);
	for (i = 0, p = inoutBuffer; i < inLength; i++, p++)
	{
		*p = tableArray[(short)*p];
	}
}// ConvertTextToMac


//
// internal methods
//

/*!
Returns a pointer to a particular table.  Table numbers
1...n correspond to tables in the master array; table 0
is the default table.

(2.6)
*/
static BytePtr
getTableArray	(SInt16 const		inTable,
				 Boolean const		inGetNonMacTable)
{
	if (inTable > nNational || inTable < 1) return(DefaultTable + ((inGetNonMacTable == true) * 256));
	return((BytePtr)(*transTablesHdl + ((inTable - 1) * 512) + ((inGetNonMacTable == true) * 256)));
}// getTableArray


/*!
Loads a translation table resource and returns a
pointer to the new table.

(2.6)
*/
static BytePtr
getTranslationResource		(SInt16 const		inResourceID)
{
	Handle		h = null;
	
	
	h = GetResource(MY_TRSL, inResourceID);
	
	if ((h == null) || (ResError() != noErr))
	{
		Console_WriteLine("Warning: Cannot get translation resource...");
	}
	
	DetachResource(h);
	HLockHi(h);
	return((BytePtr)*h);
}// getTranslationResource