/*###############################################################

	TelnetThreads.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <ControlDefinitions.h>
#include <Controls.h>
#include <Dialogs.h>
#include <PLStringFuncs.h>
#include <Resources.h>
#include <Script.h>
#include <Threads.h>

// library includes
#include "InterfaceLib.h"

// resource includes
#include "DialogResources.h"
#include "MenuResources.h"
#include "StringResources.h"
#include "TelnetVersion.h"

// MacTelnet includes
#include "Clipboard.h"
#include "CoolStuff.h"
#include "CommandLine.h"
#include "Console.h"
#include "DialogItemIndices.h"
#include "DialogUtilities.h"
#include "FileUtilities.h"
#include "FolderManager.h"
#include "ProgressDialog.h"
#include "RealScreenData.h"
#include "RealScreen.h"
#include "ResourceFileManager.h"
#include "ScreenManager.h"
#include "TelnetDescriptors.h"
#include "TelnetMenus.h"
#include "TelnetThreads.h"



// constants
enum
{
	kStandardInactiveTelnetThreadOptions = kNewSuspend | kUsePremadeThread | kCreateIfNeeded,
	kStandardActiveTelnetThreadOptions = kUsePremadeThread | kCreateIfNeeded
};

// types
struct TelnetThread
{
	SInt16						lockCount;			// how many locks are on the handle
	TelnetThreadDescriptor		descriptor;
	ThreadID					threadID;
	voidPtr						result;
};
typedef struct TelnetThread		TelnetThread;
typedef TelnetThread*			TelnetThreadPtr;
typedef TelnetThreadPtr*		TelnetThreadHandle;

// variables
static ChainRef		gThreadList = null;
static Boolean		gInCriticalSection = false;

// internal methods
static void					clearMenu					(MenuRef				inoutMenu);
static void					constructFontMenu			(MenuRef				inoutFontMenu);
static Boolean				isMonospacedFont			(Str255					inFontName);
static TelnetThreadPtr		refAcquireLock				(TelnetThreadRef		inRef);
static void					refReleaseLock				(TelnetThreadRef		inRef,
														 TelnetThreadPtr*		inoutPtr);



//
// public methods
//

/*!
Before calling any other thread routines,
invoke this routine!  The Thread Manager
must be available - if it is not, this
method silently fails and all attempts to
create threads will have no effect.  You
can test the flag "kFlagThreadManager"
to see if this initialization succeeded,
however.

(3.0)
*/
void
Threads_Init		(void)
{
	SInt32		gestaltResult = 0L;
	Boolean		haveThreadManager = false;
	
	
	if (Gestalt(gestaltThreadMgrAttr, &gestaltResult) == noErr)
	{
		Size		stackSize = 0L;
		OSErr		error = noErr;
		
		
		haveThreadManager = (gestaltResult & (1 << gestaltThreadMgrPresent));
		error = GetDefaultThreadStackSize(kCooperativeThread, &stackSize);
		error = CreateThreadPool(kCooperativeThread, 3/* number of threads to create */,
									stackSize + INTEGER_KILOBYTES(1)/* stack size, per */);
		gThreadList = Chain_New(kChainStyleSinglyLinked);
		Chain_SetDescriptor(gThreadList, kTelnetChainDescriptorThreadsList);
	}
	FlagManager_Set(kFlagThreadManager, haveThreadManager);
}// Init


/*!
When you are completely finished with threads,
invoke this method to clean up this module.

(3.0)
*/
void
Threads_Done		(void)
{
	if (FlagManager_Test(kFlagThreadManager))
	{
		ChainLinkRef		node = null;
		TelnetThreadHandle	dataHandle = null;
		
		
		node = Chain_GetFirstNode(gThreadList);
		do
		{
			// destroy the structure pointed to by each linked list element
			dataHandle = Chain_GetNodeDataPtr(gThreadList, node);
			if (dataHandle != null) Memory_DisposeHandle((Handle*)&dataHandle);
			node = Chain_GetNextNode(gThreadList, node);
		} while (node != null);
		Chain_Dispose(&gThreadList);
	}
}// Done


/*!
To create a cooperative thread, using premade
threads if possible and creating a thread if
necessary, use this method.  The amount
"inExtraStackSize" is *added* to the result
of a call to the GetDefaultThreadStackSize()
Mac OS routine.

See also Threads_NewCoOpWithOptions().

IMPORTANT:	The "inThreadDescriptorOfNewThread"
			parameter should be unique for each
			created thread.  The "inThreadEntry"
			value is not considered unique since
			you may want multiple threads using
			the same routine.  As with all other
			descriptors, MacTelnet uses values
			defined in "TelnetDescriptors.h".

(3.0)
*/
OSErr
Threads_NewCoOp		(ThreadEntryUPP 			inThreadEntry,
					 void*						inThreadParam,
					 Size 						inExtraStackSize,
					 TelnetThreadDescriptor		inThreadDescriptorOfNewThread)
{
	return Threads_NewCoOpWithOptions(inThreadEntry, inThreadParam, inExtraStackSize,
										kStandardActiveTelnetThreadOptions, inThreadDescriptorOfNewThread);
}// NewCoOp


/*!
To create a cooperative thread with custom
thread options, use this method.  The amount
"inExtraStackSize" is *added* to the result
of a call to the GetDefaultThreadStackSize()
routine.

See also Threads_NewCoOp().

IMPORTANT:	The "inThreadDescriptorOfNewThread"
			parameter should be unique for each
			created thread.  The "inThreadEntry"
			value is not considered unique since
			you may want multiple threads using
			the same routine.  As with all other
			descriptors, MacTelnet uses values
			defined in "TelnetDescriptors.h".

(3.0)
*/
OSErr
Threads_NewCoOpWithOptions		(ThreadEntryUPP 			inThreadEntry,
								 void*						inThreadParam,
								 Size 						inExtraStackSize,
								 ThreadOptions				inOptions,
								 TelnetThreadDescriptor		inThreadDescriptorOfNewThread)
{
	OSErr		result = noErr;
	
	
	if (FlagManager_Test(kFlagThreadManager))
	{
		Size		defaultSize = 0L;
		
		
	#if TARGET_API_MAC_OS8
		result = GetDefaultThreadStackSize(kCooperativeThread, &defaultSize);
	#else
		// GetDefaultThreadStackSize() always fails on Mac OS X, so always guess a stack value
		result = cfragNoSymbolErr;
	#endif
		if (result != noErr) defaultSize = INTEGER_KILOBYTES(40); // without a real value, take a guess
		
		// create a thread
		result = noErr;
		{
			TelnetThreadHandle		newThread = null;
			
			
			// create global space to store the data and results for this thread
			newThread = (TelnetThreadHandle)Memory_NewHandle(sizeof(TelnetThread));
			if (newThread != null)
			{
				ChainLinkRef		newNode = null;
				TelnetThreadPtr		ptr = refAcquireLock((TelnetThreadRef)newThread);
				
				
				newNode = Chain_NewNode(gThreadList, (TelnetThreadRef)newThread);
				Chain_MoveNode(gThreadList, newNode, null, kChainLinkPlacementInsertBefore);
				
				ptr->descriptor = inThreadDescriptorOfNewThread;
				result = NewThread(kCooperativeThread, inThreadEntry, inThreadParam,
									defaultSize + inExtraStackSize, inOptions,
									&ptr->result, &ptr->threadID);
				if (result != noErr) Console_WriteLine("thread creation failed");
				Threads_YieldTo((TelnetThreadRef)newThread);
				refReleaseLock((TelnetThreadRef)newThread, &ptr);
			}
		}
	}
	else result = threadNotFoundErr;
	
	return result;
}// NewCoOpWithOptions


/*!
To pause for up to a certain number of ticks
(usually accurate within a tick) in a way
that does not block other threads and that is
interruptible by a user mouse click, use this
method.  If the mouse is clicked, the delay
is stopped short and this method returns
immediately with a value "true"; otherwise,
"false" is returned.

(3.0)
*/
Boolean
Threads_ClickStopDelayTicks	(const UInt32		inTicksOfDelay)
{
	UInt32		startTime = TickCount();
	Boolean		result = false;
	
	
	for (; ((TickCount() - startTime) < inTicksOfDelay); )
	{
		Threads_YieldToAnother();
		if (Button())
		{
			result = true;
			break;
		}
	}
	return result;
}// ClickStopDelayTicks


/*!
To mark subsequent code as critical, disabling
any calls to yielding functions or other thread
schedulers, use this method.  Be sure to balance
this call with Threads_CriticalSectionEnd() when
the thread-unsafe code is passed, or no other
threads will get time.

(3.0)
*/
void
Threads_CriticalSectionBegin	(void)
{
	if (FlagManager_Test(kFlagThreadManager))
	{
		Console_WriteLine("Entering a threads critical section.");
		(OSErr)ThreadBeginCritical();
		gInCriticalSection = true;
	}
}// CriticalSectionBegin


/*!
To mark subsequent code as non-critical, enabling
any calls to yielding functions or other thread
schedulers, use this method.  You call this method
after Threads_CriticalSectionBegin(), when you are
finished with thread-unsafe code.

(3.0)
*/
void
Threads_CriticalSectionEnd	(void)
{
	if (FlagManager_Test(kFlagThreadManager))
	{
		Console_WriteLine("Leaving a threads critical section.");
		(OSErr)ThreadEndCritical();
		gInCriticalSection = false;
	}
}// CriticalSectionEnd


/*!
To pause for at *least* a certain number of
ticks (usually accurate within a tick) in a
way that does not block other threads, use
this method.

(3.0)
*/
void
Threads_DelayMinimumTicks	(const UInt32		inTicksOfDelay)
{
	UInt32		startTime = TickCount();
	
	
	for (; ((TickCount() - startTime) < inTicksOfDelay); )
	{
		Threads_YieldToAnother();
	}
}// DelayMinimumTicks


/*!
To acquire a reference to a thread, given the
thread descriptor assigned to it at creation
time, use this method.

(3.0)
*/
TelnetThreadRef
Threads_GetByDescriptor		(TelnetThreadDescriptor		inThreadDescriptor)
{
	TelnetThreadRef		result = null;
	
	
	if (FlagManager_Test(kFlagThreadManager))
	{
		TelnetThreadPtr		threadPtr = null;
		ChainLinkRef		node = null;
		
		
		// look for the given panel in the linked list
		node = Chain_GetFirstNode(gThreadList);
		do
		{
			result = Chain_GetNodeDataPtr(gThreadList, node);
			threadPtr = refAcquireLock(result);
			if (threadPtr != null)
			{
				if (threadPtr->descriptor == inThreadDescriptor) node = null; // got it - stop here
				else node = Chain_GetNextNode(gThreadList, node);
				refReleaseLock(result, &threadPtr);
			}
			else node = null; // problem - stop here
		} while (node != null);
	}
	
	return result;
}// GetByDescriptor


/*!
To acquire a reference to a thread, given the
thread descriptor assigned to it at creation
time, use this method.

(3.0)
*/
TelnetThreadRef
Threads_GetByThreadManagerID		(ThreadID		inThreadID)
{
	TelnetThreadRef		result = null;
	
	
	if (FlagManager_Test(kFlagThreadManager))
	{
		TelnetThreadPtr		threadPtr = null;
		ChainLinkRef		node = null;
		
		
		// look for the given panel in the linked list
		node = Chain_GetFirstNode(gThreadList);
		do
		{
			result = Chain_GetNodeDataPtr(gThreadList, node);
			if (result != null)
			{
				threadPtr = refAcquireLock(result);
				if (threadPtr != null)
				{
					if (threadPtr->threadID == inThreadID) node = null; // got it - stop here
					else node = Chain_GetNextNode(gThreadList, node);
					refReleaseLock(result, &threadPtr);
				}
				else node = null; // problem - stop here
			}
			else node = null; // problem - stop here
		} while (node != null);
	}
	
	return result;
}// GetByThreadManagerID


/*!
To determine the thread currently running, use
this method.  This is the recommended way to
determine the current thread.  If "null" is
returned, an error may have occurred, but you
can assume that the main thread is running.

(3.0)
*/
TelnetThreadRef
Threads_GetCurrent		(void)
{
	TelnetThreadRef		result = null;
	ThreadID			threadID = 0;
	OSErr				error = noErr;
	
	
	error = GetCurrentThread(&threadID);
	if (error == noErr) result = Threads_GetByThreadManagerID(threadID);
	return result;
}// GetCurrent


/*!
To determine the descriptor of a thread, given
a reference to it, use this method.

(3.0)
*/
TelnetThreadDescriptor
Threads_GetDescriptorOf		(TelnetThreadRef		inThread)
{
	TelnetThreadDescriptor		result = kTelnetThreadDescriptorMainThread;
	
	
	if (FlagManager_Test(kFlagThreadManager))
	{
		TelnetThreadPtr		threadPtr = refAcquireLock(inThread);
		
		
		if (threadPtr != null) result = threadPtr->descriptor;
		refReleaseLock(inThread, &threadPtr);
	}
	return result;
}// GetDescriptorOf


/*!
To determine the result of a thread previously
created with Threads_NewCoOp(), use this method.

IMPORTANT:	The result is meaningless until the
			thread is finished executing!  See
			Threads_WhenTerminatedInvoke() for a
			way to install a routine that gets
			called as soon as the result becomes
			valid.

(3.0)
*/
voidPtr
Threads_GetResultOf		(TelnetThreadRef		inThread)
{
	voidPtr				result = null;
	
	
	if (FlagManager_Test(kFlagThreadManager))
	{
		TelnetThreadPtr		threadPtr = refAcquireLock(inThread);
		
		
		if (threadPtr != null) result = threadPtr->result;
		refReleaseLock(inThread, &threadPtr);
	}
	return result;
}// GetResultOf


/*!
Every thread gets its own stack space for making
allocations, space which is also used by the OS
to manage the threads.  If this stack space runs
out, the thread can no longer function properly,
so you should use this method to determine how
much free space is left on a threads stack.

(3.0)
*/
UInt32
Threads_GetStackOf	(TelnetThreadRef		inThread)
{
	UInt32		result = 0L;
	
	
	if (FlagManager_Test(kFlagThreadManager))
	{
		TelnetThreadPtr		threadPtr = refAcquireLock(inThread);
		
		
		if (threadPtr != null) (OSErr)ThreadCurrentStackSpace(threadPtr->threadID, &result);
		refReleaseLock(inThread, &threadPtr);
	}
	return result;
}// GetStackOf


/*!
To determine the Thread Manager ID of a thread
previously created with Threads_New(), use this
method.

WARNING:	If at all possible, use routines from
			this module to access threads.  This
			method is provided only if you wish to
			take advantage of Thread Manager calls
			that do not have equivalents in this
			module.  By accessing a threads ID,
			you risk future incompatibilities if
			this module migrates away from the
			Thread Manager.  Consider adding a new
			routine to this module, if you need
			special functionality not found here.

(3.0)
*/
ThreadID
Threads_GetThreadManagerID	(TelnetThreadRef		inThread)
{
	ThreadID			result = kNoThreadID;
	
	
	if (FlagManager_Test(kFlagThreadManager))
	{
		TelnetThreadPtr		threadPtr = refAcquireLock(inThread);
		
		
		if (threadPtr != null) result = threadPtr->threadID;
		refReleaseLock(inThread, &threadPtr);
	}
	return result;
}// GetThreadManagerID


/*!
Since threads run and complete at irregular times,
it is often necessary to be notified when a thread
completes (if synchronization of multithreaded
operations is required).  Use this method to specify
a routine that should be called as soon as the
given thread terminates (returns a value).  You can
use Threads_GetResult() to access the result of a
thread operation from within a termination method.

(3.0)
*/
void
Threads_WhenTerminatedInvoke	(TelnetThreadRef			inThread,
								 ThreadTerminationUPP		inProc,
								 void*						inProcArgument)
{
	if (FlagManager_Test(kFlagThreadManager))
	{
		TelnetThreadPtr		threadPtr = refAcquireLock(inThread);
		
		
		if (threadPtr != null) (OSErr)SetThreadTerminator(threadPtr->threadID, inProc, inProcArgument);
		refReleaseLock(inThread, &threadPtr);
	}
}// WhenTerminatedInvoke


/*!
To force a thread to suspend and allow a specific
suspended thread to resume executing, use this
method.  If the specified thread cannot execute,
then this method has the same effect as a general
call to Threads_YieldToAnother().

You *must* call a yielding method regularly in an
environment with cooperative threads in order for
threads to get time; however, in a preemptive
environment, this call is optional (but still can
be invoked at strategic times to improve thread
responsiveness).

See also Threads_YieldToAnother().

(3.0)
*/
void
Threads_YieldTo		(TelnetThreadRef		inThread)
{
	unless (gInCriticalSection) if (FlagManager_Test(kFlagThreadManager))
	{
		TelnetThreadPtr		threadPtr = refAcquireLock(inThread);
		
		
		if (threadPtr != null) (OSErr)YieldToThread(threadPtr->threadID);
		refReleaseLock(inThread, &threadPtr);
	}
}// YieldTo


/*!
To force a thread to suspend and allow another
suspended thread to resume executing, use this
method.

You *must* call a yielding method regularly in an
environment with cooperative threads in order for
threads to get time; however, in a preemptive
environment, this call is optional (but still can
be invoked at strategic times to improve thread
responsiveness).

See also Threads_YieldTo().

(3.0)
*/
void
Threads_YieldToAnother	(void)
{
	unless (gInCriticalSection) if (FlagManager_Test(kFlagThreadManager)) (OSErr)YieldToAnyThread();
}// YieldToAnother


//
// thread entry points
//

/*!
This thread reads the font information from the
preferences file, compares it against the fonts
currently installed, and then rebuilds the font
list if necessary.  This is done at startup
time.  Since the font menu is being constructed
here, the Format item in the Edit menu is not
available until this thread unlocks it.

In MacTelnet 3.0, this routine (finally) uses the
partial resource management APIs to properly
construct manageable resource handles.  In
previous versions of MacTelnet, handles would be
expanded continuously, until eventually memory
got trampled (if too many fonts were installed
or there were many fonts with long names).
Now that Mac OS 9 and beyond allow even more
fonts to be installed, this problem is all the
more likely to occur.  MacTelnet 3.0 fixes it.
(And by the way, only in MacTelnet 3.0 was this
process ever multithreaded!  Thank you very
much.)

(3.0)
*/
pascal voidPtr
Threads_InitializeFontMenu		(void*		inMenuRefPosingAsVoidPointer)
{
	enum
	{
		kStringListResourceType = 'STR#'
	};
	OSStatus		error = noErr;
	Boolean			listNeedsRebuilding = false;
	Handle			fontNameStringListResourceHandle = null;
	SInt16			numFontsInList = 0,
					numFontsInMenu = 0;
	MenuRef			theMenu = REINTERPRET_CAST(inMenuRefPosingAsVoidPointer, MenuRef);
    MenuItemIndex	menuIndex = 0;
	Str255			fontListName,
					currentFontName;
	SInt16			oldResFile = CurResFile(),
					preferredResFile = ResourceFileManager_GetResFile(kTelnetResourceFilePreferences);
	
	
	Threads_YieldToAnother();
	UseResFile(preferredResFile = ResourceFileManager_GetResFile(kTelnetResourceFilePreferences));
	
	// delete obsolete resources if they exist in the preferences file
	{
		Handle		obsoleteResourceHandle = null;
		
		
		obsoleteResourceHandle = Get1Resource(kStringListResourceType, rStringsGoodFonts);
		error = ResError();
		if ((error == noErr) && (obsoleteResourceHandle != null))
		{
			RemoveResource(obsoleteResourceHandle);
			error = ResError();
		}
		obsoleteResourceHandle = Get1Resource(kStringListResourceType, rStringsBadFonts);
		error = ResError();
		if ((error == noErr) && (obsoleteResourceHandle != null))
		{
			RemoveResource(obsoleteResourceHandle);
			error = ResError();
		}
		UpdateResFile(CurResFile());
	}
	
	// see if anything was previously saved - retain the resource handle, but leave
	// the data on disk; that way, buffering (with ReadPartialResource()) can be done
	SetResLoad(false);
	fontNameStringListResourceHandle = Get1Resource(kStringListResourceType, rStringsAllFonts);
	error = ResError();
	if ((error != noErr) || (fontNameStringListResourceHandle == null))
	{
		// create a new, empty font list
		Handle		emptyStringListResourceHandle = null;
		
		
		Console_WriteLine("Creating a new font list.");
		emptyStringListResourceHandle = Memory_NewHandleInProperZone(sizeof(SInt16), true/* short lifetime */);
		if (emptyStringListResourceHandle != null)
		{
			**((SInt16**)emptyStringListResourceHandle) = 0; // set length of string list
			AddResource(emptyStringListResourceHandle, 'STR#', rStringsAllFonts, EMPTY_PSTRING);
			error = ResError();
			UpdateResFile(ResourceFileManager_GetResFile(kTelnetResourceFilePreferences));
			error = ResError();
			Memory_DisposeHandle(&emptyStringListResourceHandle);
		}
		
		fontNameStringListResourceHandle = Get1Resource(kStringListResourceType, rStringsAllFonts);
		error = ResError();
		numFontsInList = 0;
		listNeedsRebuilding = true;
	}
	else
	{
		ReadPartialResource(fontNameStringListResourceHandle, 0L, &numFontsInList, sizeof(SInt16));
		Console_WriteValue("Font list found - cardinality", numFontsInList);
	}
	SetResLoad(true);
	
	// start out with all fonts in the menu
	clearMenu(theMenu);
	constructFontMenu(theMenu);
	numFontsInMenu = CountMenuItems(theMenu);
	Console_WriteValue("Font menu created - cardinality", numFontsInMenu);
	
	if ((error == noErr) && (fontNameStringListResourceHandle != null))
	{
		if ((numFontsInList <= 0) || (numFontsInMenu != numFontsInList))
		{
			// the menu and font list clearly disagree, so rebuild the list
			listNeedsRebuilding = true;
		}
		else
		{
			// read the existing list; as soon as fonts disagree, stop and rebuild the list
			Console_WriteLine("Preparing to read existing font list.");
			
			// look at each font; see if it matches the font list
			for (menuIndex = 1; menuIndex <= numFontsInMenu; menuIndex++)
			{
				Threads_YieldToAnother();
				UseResFile(preferredResFile); // just in case another thread screwed up the current resource file
				GetMenuItemText(theMenu, menuIndex, currentFontName);
				GetIndString(fontListName, rStringsAllFonts, menuIndex);
				unless (EqualString(currentFontName, fontListName, true/* case sensitive */,
									false/* diacritics sensitive */))
				{
					listNeedsRebuilding = true;
					break;
				}
			}
		}
	}
	
	if (listNeedsRebuilding)
	{
		// create a new font list
		Size const			kBadFontsPartialBufferSize = INTEGER_KILOBYTES(8); // size of partial-resource buffer
		ProgressDialogRef	progressDialog = null;
		Size				offset = 0L;
		Ptr					fontNameStringListPartialPtr = Memory_NewPtr(kBadFontsPartialBufferSize);
		
		
		Console_WriteLine("Preparing to create a new font list.");
		Cursors_UseWatch();
		
		// clear the list resource
		{
			OSStatus	memError = Memory_SetHandleSize(fontNameStringListResourceHandle, sizeof(SInt16));
			
			
			if (memError != noErr) Console_WriteValue("SetHandleSize error", memError);
			else
			{
				*((SInt16*)*fontNameStringListResourceHandle) = 0;
			}
		}
		
		// reset the menu
		clearMenu(theMenu);
		constructFontMenu(theMenu);
		numFontsInMenu = CountMenuItems(theMenu);
		
		// create a progress dialog
		ResourceFileManager_UseResFile(kTelnetResourceFileApplication);
		{
			Str255		text;
			
			
			GetIndString(text, rStringsStatus, siStatusBuildingFontList);
			progressDialog = ProgressDialog_New(text, false/* modal */);
		}
		
		// 3.0
		ProgressDialog_Display(progressDialog);
		
		// write names of fonts to disk; this is the bottleneck, so display progress to the user in this loop
		{
			SInt16		i = 0;
			Str255		fontName;
			
			
			offset = sizeof(SInt16); // skip the number of strings word
			for (i = 1; i <= numFontsInMenu; i++)
			{
				// 3.0 - read and write a PARTIAL string resource, as this list tends to be long and is a
				// contiguous block (there is limited memory for the thread run, so one must be careful here)
				Size	nameSize = 0L;
				
				
				ProgressDialog_SetProgressPercentFull(progressDialog, 100 * (i - 1) / numFontsInMenu);
				Threads_YieldToAnother();
				UseResFile(preferredResFile); // just in case another thread screwed up the current resource file
				GetMenuItemText(theMenu, i, fontName);
				
				// How big is this font name?  In the *extremely* unlikely event
				// that the buffer is not large enough to hold the name, truncate
				// the buffer (this would actually screw up the font list, but
				// hey, I DID say it was extremely unlikely...).
				nameSize = (1 + PLstrlen(fontName)) * sizeof(UInt8);
				if (nameSize > GetPtrSize(fontNameStringListPartialPtr))
				{
					nameSize = GetPtrSize(fontNameStringListPartialPtr);
				}
				
				BlockMoveData(fontName, fontNameStringListPartialPtr, nameSize);
				SetResourceSize(fontNameStringListResourceHandle, GetResourceSizeOnDisk(fontNameStringListResourceHandle) + nameSize);
				error = ResError();
				if (error != noErr) Console_WriteValue("SetResourceSize error", error);
				WritePartialResource(fontNameStringListResourceHandle, offset,
										fontNameStringListPartialPtr, nameSize);
				error = ResError();
				if (error != noErr) Console_WriteValue("WritePartialResource (1) error", error);
				offset += nameSize;
			}
		}
		
		// when done, make sure the resource is exactly large enough
		SetResourceSize(fontNameStringListResourceHandle, offset);
		(OSStatus)ResError();
		
		Threads_YieldToAnother();
		UseResFile(preferredResFile); // just in case another thread screwed up the current resource file
		
		ProgressDialog_SetProgressPercentFull(progressDialog, 100);
		ReadPartialResource(fontNameStringListResourceHandle, 0L, fontNameStringListPartialPtr, sizeof(SInt16));
		error = ResError();
		if (error != noErr) Console_WriteValue("ReadPartialResource error", error);
		else
		{
			*((SInt16*)fontNameStringListPartialPtr) = numFontsInMenu; // change the string count
			WritePartialResource(fontNameStringListResourceHandle, 0L, fontNameStringListPartialPtr,
									sizeof(SInt16));
			error = ResError();
			if (error != noErr) Console_WriteValue("WritePartialResource (2) error", error);
		}
		
		Memory_DisposePtr(&fontNameStringListPartialPtr); // buffer is finished
		ChangedResource(fontNameStringListResourceHandle);
		
		ProgressDialog_Dispose(&progressDialog);
	}
	
	ReleaseResource(fontNameStringListResourceHandle), fontNameStringListResourceHandle = null;
	Threads_YieldToAnother();
	UseResFile(preferredResFile); // just in case another thread screwed up the current resource file
	ResourceFileManager_UpdateResFile(kTelnetResourceFilePreferences);
	
	// remove proportional fonts from the menu
	{
		SInt16		i = 0;
		Str255		fontName;
		
		
		for (i = 1; i <= numFontsInMenu; i++)
		{
			GetMenuItemText(theMenu, i, fontName);
			unless (isMonospacedFont(fontName))
			{
				DeleteMenuItem(theMenu, i);
				numFontsInMenu--;
				i--; // offset index into menu to account for deleted item
			}
		}
	}
	
	// make the remaining fonts appear in their own typeface
	{
		SInt16		i = 0,
					fontNum = 0;
		Str255		fontName;
		
		
		for (i = 1; i <= numFontsInMenu; i++)
		{
			Threads_YieldToAnother();
			GetMenuItemText(theMenu, i, fontName);
		#if TARGET_API_MAC_OS8
			GetFNum(fontName, &fontNum);
		#else
			fontNum = FMGetFontFamilyFromName(fontName);
		#endif
			SetMenuItemFontID(theMenu, i, fontNum);
		}
	}
	Console_WriteLine("done");
	
	Menus_SetFontMenuAvailable(true); // now that the list rebuilding is finished, it is safe to access font menus
	UseResFile(oldResFile);
	
	Console_WriteLine("Shutting down thread.");
	return null;
}// InitializeFontMenu


/*!
This thread updates the Scripts menu by searching
for valid files (i.e. applications or compiled
scripts) in MacTelnets Scripts folder.

(3.0)
*/
pascal voidPtr
Threads_InitializeScriptsMenu	(void*		inMenuRefPosingAsVoidPointer)
{
	MenuRef		menu = REINTERPRET_CAST(inMenuRefPosingAsVoidPointer, MenuHandle);
	voidPtr		result = null;
	
	
	if (menu != null)
	{
		static unsigned long	lastModDate = 0L;
		unsigned long			date = 0L;
		FSSpec					scriptsFolder;
		OSErr					error = noErr;
		Boolean					canUseScriptsMenu = true; // if not, delete the menu
		
		
		// locate the Scripts folder, and find out when it was last modified
		Threads_YieldToAnother();
		error = FolderManager_GetFSSpec(kTelnetFolderScriptsMenuItems, &scriptsFolder);
		date = FileUtilities_GetDirectoryDateFromFSSpec(&scriptsFolder, kFileUtilitiesDateOfModification);
		
		// only rebuild the Scripts menu if the Scripts Menu Items folder was found
		// and the user has actually changed the folder contents
		if (error != noErr) canUseScriptsMenu = false;
		
		if ((canUseScriptsMenu) && (lastModDate != date))
		{
			enum
			{
				kMaximumScriptsInMenu = 31		// no more than this many script files will be listed in the menu
			};
			FSSpec*				array = (FSSpec*)Memory_NewPtr(kMaximumScriptsInMenu * sizeof(FSSpec));
			UInt32				arrayLength = kMaximumScriptsInMenu;
			register SInt32		i = 0L;
			register SInt16		j = 0;
			FileInfo			info;
			Boolean				useFile = true,
								firstScript = true,		// as soon as a divider is put in the menu, this flag overturns
								anyTopItems = false;	// set if there are any script filenames beginning with "!"
			ProgressDialogRef	progressDialog = null;
			SInt16				numberOfSpecialScriptMenuItems = 0;
			SInt16				oldResFile = CurResFile();
			
			
			// create a progress dialog; if this isnt multithreaded, use a modal frame; otherwise, make it modeless
			{
				Str255			text;
				
				
				GetIndString(text, rStringsStatus, siStatusBuildingScriptsMenu);
				progressDialog = ProgressDialog_New(text, !FlagManager_Test(kFlagThreadManager)/* modal */);
			}
			Threads_YieldToAnother();
			ProgressDialog_SetProgressIndicator(progressDialog, kTelnetProgressIndicatorIndeterminate);
			
			// show the progress window
			ProgressDialog_Display(progressDialog);
			
			// start out by clearing the menu
			while (CountMenuItems(menu) > 0)
			{
				Threads_YieldToAnother();
				DeleteMenuItem(menu, 1);
			}
			
			// sychronously search for scripts
			error = FileUtilities_GetTypedFilesInDirectory(&scriptsFolder, 'osas', 'ToyS', array, &arrayLength);
			
			// Iterate through all found files, adding each one to the menu.
			if (error == noErr) for (numberOfSpecialScriptMenuItems = 0, i = 0, j = 0; i < arrayLength; i++)
			{
				Threads_YieldToAnother();
				
				useFile = true;
				if (FSpGetFInfo(&array[i], (FInfo*)&info) == noErr)
				{
					// skip invisible files
					useFile = (!(info.finderFlags & kIsInvisible));
				}
				Threads_YieldToAnother();
				if (useFile)
				{
					MenuItemIndex	insertAfterIndex = 0;
					
					
					j++;
					if (array[i].name[1] == '!')
					{
						// any file names starting with "!" go in the special, top half of the menu
						insertAfterIndex = numberOfSpecialScriptMenuItems++;
						StringUtilities_PClipHead(array[i].name, 1); // remove the "!"
						anyTopItems = true;
						if (firstScript)
						{
							// add a divider, as well
							j++;
							InsertMenuItem(menu, "\p(-", 0);
							firstScript = false;
						}
					}
					else
					{
						insertAfterIndex = j - 1;	// add to the bottom of the menu
					}
					Threads_YieldToAnother();
					
					// insert some item, then change its text; this avoids Menu Manager pre-processing
					InsertMenuItem(menu, "\p<script>", insertAfterIndex);
					SetMenuItemText(menu, insertAfterIndex + 1, array[i].name);
				}
			}
			Menus_SetNumberOfSpecialScriptMenuItems(numberOfSpecialScriptMenuItems);
			Memory_DisposePtr((Ptr*)&array);
			
			// if there are no scripts, remove the Scripts menu
			if (arrayLength < 1) canUseScriptsMenu = false;
			
			ProgressDialog_Dispose(&progressDialog);
			UseResFile(oldResFile);
		}
		
		// if for some reason the Scripts menu cannot be used, delete it
		unless (canUseScriptsMenu)
		{
			DeleteMenu(kMenuIDScripts);
			DrawMenuBar();
		}
		
		lastModDate = date; // remember this for next time (it could save a lot of work!)
	}
	
	Menus_SetScriptsMenuAvailable(true); // now that the list rebuilding is finished, set this flag
	
	//Console_WriteLine("Shutting down thread.");
	return result;
}// InitializeScriptsMenu


/*!
This thread handles basic UI idle stuff, such
as animating the cursor.

(3.0)
*/
pascal voidPtr
Threads_InterfaceIdle		(void*		UNUSED_ARGUMENT(inUnusedThreadParam))
{
	do
	{
		Cursors_Idle();
		if (GetWindowKind(GetRealFrontWindow()) == kDialogWindowKind)
		{
			// for modal dialog boxes, ensure the cursor is set properly
			SetCursorForControlUnderMouse(GetRealFrontWindow());
		}
		
		Threads_YieldToAnother();
	}
	while (FlagManager_Test(kFlagThreadManager)); // this allows the function to be used if threads arent available
	
	return null;
}// InterfaceIdle


/*!
This thread handles the worthless but fun
programming for moving stars in the About box.

(3.0)
*/
pascal voidPtr
Threads_MovingStars		(void*		inControlRefPosingAsVoidPointer)
{
	WindowRef		window = GetControlOwner((ControlRef)inControlRefPosingAsVoidPointer);
	
	
	for (; FlagManager_Test(kFlagMovingStars); )
	{
		IdleControls(window);
		Threads_YieldToAnother();
	}
	
	return null;
}// MovingStars


/*!
This thread handles the worthless but fun
programming Easter egg for Quad Damage.

(3.0)
*/
pascal voidPtr
Threads_QuadDamage		(void*		UNUSED_ARGUMENT(inUnusedThreadParam))
{
	if (FlagManager_Test(kFlagDrawSprocket)) for (; FlagManager_Test(kFlagQuadDamage); )
	{
		// for users with DrawSprocket who discovered the Quad Damage egg,
		// make the screen gamma pulsate with a blue shimmer while QD is on
		if (FlagManager_Test(kFlagQuadDamage))
		{
			enum
			{
				kFadeMaximumPercent = 80, // 100% = normal, 0% = completely blue screen
				kFadeMinimumPercent = 50,
				kFadeChangePercent = 4 // the larger this number, the faster the blue shade changes
			};
			RGBColor			color;
			UInt32				ticks = TickCount();
			register SInt16		i = 0;
			
			
			CoolStuff_QuadDamageGetPulseColor(&color);
			
			// use a cool partial gamma-fade colorizing effect,
			// but only if MacTelnet is the frontmost application
			CoolStuff_GammaControlledFade(CoolStuff_GetDefaultDrawSprocketContext(), &color,
											kFadeMinimumPercent, kFadeMaximumPercent, kFadeChangePercent,
											4/* minimum ticks */, true/* yield to threads */);
			CoolStuff_GammaControlledFade(CoolStuff_GetDefaultDrawSprocketContext(), &color,
											kFadeMaximumPercent, kFadeMinimumPercent, -kFadeChangePercent,
											4/* minimum ticks */, true/* yield to threads */);
			
			// if Quad Damage gets turned off, restore the gamma so theres no trace of added color
			if (!FlagManager_Test(kFlagQuadDamage))
			{
				(OSStatus)DSpContext_FadeGamma(CoolStuff_GetDefaultDrawSprocketContext(), 100/* percent */, &color);
			}
		}
		Threads_YieldToAnother();
	}
	
	return null;
}// QuadDamage


/*!
This thread handles standard MacTelnet idle
procedures, such as idling windows.

(3.0)
*/
pascal voidPtr
Threads_StandardIdle		(void*		UNUSED_ARGUMENT(inUnusedThreadParam))
{
	do
	{
		Sound_Idle();
		Clipboard_Idle();
		CommandLine_Idle(); // needed because this is a floating window
		IdleControls(GetRealFrontWindow()); // idles the frontmost non-floating window
	#if TARGET_API_MAC_OS8
		//CoolStuff_PulseDefaultButton(GetRealFrontWindow()); // pulses default buttons, like Aqua does (!)
	#endif
		Threads_YieldToAnother();
	}
	while (FlagManager_Test(kFlagThreadManager)); // this allows the function to be used if threads arent available
	
	return null;
}// StandardIdle


//
// internal methods
//
#pragma mark -

/*!
Uses the quickest API available to delete all
items from the specified menu.

(3.0)
*/
static void
clearMenu	(MenuRef	inoutMenu)
{
	while (CountMenuItems(inoutMenu) > 0) DeleteMenuItem(inoutMenu, 1);
}// clearMenu


/*!
Uses the most advanced API available to build
a menu of all available fonts - if possible,
with fonts from the current script, as well
as the system script (if different) and the
most recent key script (if different).

(3.0)
*/
static void
constructFontMenu	(MenuRef	inoutFontMenu)
{
	if (API_AVAILABLE(InsertFontResMenu))
	{
		// insert available fonts from as many different script systems as possible (user convenience)
		ScriptCode		systemScript = GetScriptManagerVariable(smSysScript),	// localized Mac OS script
						currentScript = GetScriptManagerVariable(smKeyScript),	// current keyboard script
						lastScript = GetScriptManagerVariable(smLastScript);	// last keyboard script
		
		
		// get at least the normal fonts
		InsertFontResMenu(inoutFontMenu, 999/* insert after item */, systemScript);
		
		if (systemScript != currentScript) InsertFontResMenu(inoutFontMenu,
																CountMenuItems(inoutFontMenu)/* insert after item */,
																currentScript);
		if ((currentScript != lastScript) &&
			(systemScript != lastScript)) InsertFontResMenu(inoutFontMenu,
																CountMenuItems(inoutFontMenu)/* insert after item */,
																lastScript);
	}
	else
	{
		Console_WriteLine("Missing API - can only use fonts from the current system script.");
		AppendResMenu(inoutFontMenu, 'FONT'); // add all system-script fonts only
	}
}// constructFontMenu

/*!
To determine if a font is monospaced,
use this method.

(2.6)
*/
static Boolean
isMonospacedFont	(Str255		inFontName)
{
	Boolean		result = false;
	
	
	{
		Boolean		haveNonRomanScripts = false,
					doRomanTest = false;
		SInt32		tempLong = 0L;
		
		
		tempLong = GetScriptManagerVariable(smEnabled);	// this returns number of scripts enable
		haveNonRomanScripts = (tempLong > 1); // if there is one, its Roman
		
		if (haveNonRomanScripts)
		{
			SInt32		thisScriptEnabled = 0L;
			ScriptCode	scriptNumber = 0;
			SInt16		fond = 0;
			
			
		#if TARGET_API_MAC_OS8
			GetFNum(inFontName, &fond);
		#else
			fond = FMGetFontFamilyFromName(inFontName);
		#endif
			scriptNumber = FontToScript(fond);
			if (scriptNumber != 0) // if its non-Roman
			{
				thisScriptEnabled = GetScriptVariable(scriptNumber, smScriptEnabled);
				if (thisScriptEnabled)
				{
					// check if this font is the preferred monospaced font for its script
					SInt32		theSizeAndFond = 0L;
					SInt16		thePreferredFond = 0;
					
					
					theSizeAndFond = GetScriptVariable(scriptNumber, smScriptMonoFondSize);
					thePreferredFond = theSizeAndFond >> 16; // high word is fond 
					result = (thePreferredFond == fond);
				}
				else result = false; // this fonts script isnt enabled
			}
			else doRomanTest = true;
		}
		else doRomanTest = true;
			
		if (doRomanTest)
		{
			TextFontByName(inFontName);
			result = (CharWidth('W') == CharWidth('.'));
		}
	}
	
	return result;
}// isMonospacedFont


/*!
To acquire a pointer to the internal structure,
given a reference to it, use this method.

(3.0)
*/
static TelnetThreadPtr
refAcquireLock	(TelnetThreadRef		inRef)
{
	TelnetThreadPtr		result = *((TelnetThreadHandle)inRef);
	
	
	HLock((Handle)inRef);
	result->lockCount++;
	return result;
}// refAcquireLock


/*!
To release a pointer to the internal structure,
given a reference to it, use this method.

(3.0)
*/
static void
refReleaseLock	(TelnetThreadRef		inRef,
				 TelnetThreadPtr*		inoutPtr)
{
	TelnetThreadPtr		ptr = *((TelnetThreadHandle)inRef);
	
	
	if (--ptr->lockCount <= 0) HUnlock((Handle)inRef);
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock