/*###############################################################

	UNIXCommandInterpreters.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <PLStringFuncs.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"

// resource includes
#include "StringResources.h"

// MacTelnet includes
#include "CommandLine.h"
#include "TelnetDescriptors.h"
#include "UNIXCommandInterpreters.h"
#include "URL.h"



//
// public methods
//

/*!
This routine handles specifically the UNIX
"help" command.  Try to support all of the
arguments that the UNIX "help" does.

Telnet interprets a UNIX help command only
if the Telnet Help system is available.

(3.0)
*/
Boolean
UNIXCommandInterpreters_help	(const UInt32					argc,
								 TokenSetRef					argv,
								 void*							outDataPtr,
								 CommandLineInterpreterMessage	inMessage)
{
	Boolean		result = false;
	Boolean		caseCompare = false,
				diacriticsCompare = true;
	
	
	switch (inMessage)
	{
		case kCommandLineInterpreterMessageIsCommandYours:
			{
				Str31		cmdName,
							currentCmdName;
				
				
				StringUtilities_PGetToken(0, argv, cmdName);
				
				// help command?
				GetIndString(currentCmdName, rStringsCommandLineUNIXCommands, siUNIX_help);
				if (EqualString(currentCmdName, cmdName, caseCompare, diacriticsCompare)) result = true;
			}
			break;
		
		case kCommandLineInterpreterMessageDescribeLastArgument:
			if (argc == 1)
			{
				// describe command
				GetIndString((StringPtr)outDataPtr, rStringsCommandLineCommandDescriptions, siCommandDescription_help);
				result = true;
			}
			else
			{
				Str31		argName;
				
				
				StringUtilities_PGetToken(argc - 1, argv, argName);
				
				// describe argument - any "minus" or "plus" argument is invalid
				result = !(StringUtilities_PStartsWith(argName, "\p-") || StringUtilities_PStartsWith(argName, "\p+"));
			}
			break;
		
		case kCommandLineInterpreterMessageExecute:
			// Telnet Help must be around
			if (FlagManager_Test(kFlagAppleGuideFiles))
			{
				OSStatus	error = 0L;
				
				
				error = noErr;
				if (argc == 1)
				{
					// no keywords - just open the main guide file
					error = AppleGuideUtilities_OpenFirstMainGuide();
					result = (error == noErr);
				}
				else
				{
					// fire the keywords to the guide file
					Str255				searchString;
					Str255				keyword;
					register UInt32		i = 0;
					
					
					StringUtilities_PGetToken(1, argv, searchString); // skip argv[0] as always, its just the help command name
					PLstrcat(searchString, "\p ");
					for (i = 2; i < argc; i++)
					{
						StringUtilities_PGetToken(i, argv, keyword);
						PLstrcat(searchString, keyword);
						PLstrcat(searchString, "\p ");
					}
					error = AppleGuideUtilities_OpenFirstMainGuideInSearch(searchString);
					result = (error == noErr);
				}
			}
			break;
		
		default:
			break;
	}
	return result;
}// help


/*!
This routine handles specifically the UNIX
"open", "telnet" and "rlogin" commands.  Try
to support all of the arguments that the UNIX
"telnet" and "rlogin" do.

Argument 0 is the command ("telnet" or
"rlogin").  Argument 1 must be a host name or
IP address.  The optional argument 2, if
provided, must be a port number.

(3.0)
*/
Boolean
UNIXCommandInterpreters_telnet	(const UInt32					argc,
								 TokenSetRef					argv,
								 void*							outDataPtr,
								 CommandLineInterpreterMessage	inMessage)
{
	Boolean			result = false;
	Boolean			caseCompare = false,
					diacriticsCompare = true;
	static Boolean	isTelnet = false,
					isRlogin = false;
	
	
	switch (inMessage)
	{
		case kCommandLineInterpreterMessageIsCommandYours:
			{
				Str31		cmdName,
							currentCmdName;
				
				
				StringUtilities_PGetToken(0, argv, cmdName);
				
				// open, telnet or rlogin command?
				GetIndString(currentCmdName, rStringsCommandLineUNIXCommands, siUNIX_telnet);
				if (EqualString(currentCmdName, cmdName, caseCompare, diacriticsCompare))
				{
					isTelnet = true, isRlogin = false;
					result = true;
				}
				if (!result)
				{
					GetIndString(currentCmdName, rStringsCommandLineUNIXCommands, siUNIX_rlogin);
					if (EqualString(currentCmdName, cmdName, caseCompare, diacriticsCompare))
					{
						isTelnet = false, isRlogin = true;
						result = true;
					}
					if (!result)
					{
						GetIndString(currentCmdName, rStringsCommandLineUNIXCommands, siUNIX_open);
						if (EqualString(currentCmdName, cmdName, caseCompare, diacriticsCompare))
						{
							isTelnet = true, isRlogin = false;
							result = true;
						}
					}
				}
			}
			break;
		
		case kCommandLineInterpreterMessageDescribeLastArgument:
			if (argc == 1)
			{
				// describe command
				GetIndString((StringPtr)outDataPtr, rStringsCommandLineCommandDescriptions, siCommandDescription_telnet);
				result = true;
			}
			else
			{
				Str31		argName;
				
				
				StringUtilities_PGetToken(argc - 1, argv, argName);
				
				// describe argument - any "minus" or "plus" argument is invalid
				result = !(StringUtilities_PStartsWith(argName, "\p-") || StringUtilities_PStartsWith(argName, "\p+"));
			}
			break;
		
		case kCommandLineInterpreterMessageExecute:
			{
				Handle				handle = null;
				Str255				argument;
				register UInt32		i = 0;
				
				
				if ((argc == 2) || (argc == 3))
				{
					StringUtilities_PGetToken(1, argv, argument); // argument 1 is the host, presumably
					
					// Apple Events currently only expect URLs
					if (isTelnet) if (!StringUtilities_PStartsWith(argument, "\ptelnet://")) StringUtilities_PInsert(argument, "\ptelnet://");
					if (isRlogin) if (!StringUtilities_PStartsWith(argument, "\prlogin://")) StringUtilities_PInsert(argument, "\prlogin://");
					
					if (argc == 3)
					{
						Str15		portString;
						
						
						// concatenate the port number in standard URL form (with colon delimiter, that is)
						StringUtilities_PGetToken(2, argv, portString); // argument 2 is the port number, presumably
						PLstrcat(argument, "\p:");
						PLstrcat(argument, portString);
					}
					
					PtrToHand((Ptr)(argument + 1), &handle, StrLength(argument) * sizeof(unsigned char));
					result = URL_ParseDataHandle(handle);
				}
			}
			break;
		
		default:
			break;
	}
	return result;
}// telnet