/*###############################################################

	URL.c
	
	Manages uniform resource locators by firing off Apple events
	interpretable by any Mac OS Internet program.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstring>
#else
#	include <string.h>
#endif

// Mac includes
#include <Files.h>
#include <Folders.h>
#include <MacErrors.h>
#include <Processes.h>

// MacTelnet includes
#include "AppleEventUtilities.h"
#include "Connections.h"
#include "Console.h"
#include "InstallAE.h"
#include "FileUtilities.h"
#include "RealScreen.h"
#include "InternetConfig.h"
#include "Terminal.h"
#include "Terminology.h"
#include "URL.h"



// constants
#define CR			0x0D	/* the carriage return character */

#define IS_WHITE_SPACE_CHARACTER(a)			((a) == ' ' || (a) == '\t')
#define IS_WHITE_SPACE_OR_CR_CHARACTER(a)	(IS_WHITE_SPACE_CHARACTER(a) || (a) == CR)

// variables
static unsigned char gTable[256] =
{
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
	0x60, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
	0x41, 0x41, 0x43, 0x45, 0x4E, 0x4F, 0x55, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x43, 0x45, 0x45,
	0x45, 0x45, 0x49, 0x49, 0x49, 0x49, 0x4E, 0x4F, 0x4F, 0x4F, 0x4F, 0x4F, 0x55, 0x55, 0x55, 0x55,
	0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0x4F,
	0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xAE, 0x4F,
	0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0x22, 0x22, 0xC9, 0x20, 0x41, 0x41, 0x4F, 0xAE, 0xAE,
	0xD0, 0xD1, 0x22, 0x22, 0x27, 0x27, 0xD6, 0xD7, 0x59, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
	0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,
	0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF
};

char* gURLSchemeNames[] =
{
	":",		/* kNotURL */
	"mailto:",	
	"news:",
	"nntp:",
	"ftp:",
	"http:",
	"gopher:",
	"wais:",
	"telnet:",
	"rlogin:",
	"tn3270:",
	"finger:",
	"whois:",
	"rtsp:",
	null // this list must end with a "null"
};

// internal methods
static Boolean		notEqualStringSegments		(char*		s1,
												 char*		s2,
												 short		n);
static Boolean		resizeHandleSyncPointer		(Handle		inHandle,
												 char**		outHandleDereferencedPtr,
												 SInt32		inNewLengthInCharacters);



//
// public methods
//

/*!
Cleans up the specified URL by correcting minor
glitches in its name, such as common typing errors.
This allows MacTelnet to intelligently handle even
imperfect URLs.

(3.0)
*/
void
URL_CleanUp		(Handle		inURLDataHandle)
{
	if ((inURLDataHandle != (char**)-1L) && (inURLDataHandle != null))
	{
		UniformResourceLocatorType	urlKind = kNotURL;
		SInt8						hState = HGetState(inURLDataHandle);
		char*						string = null; // cannot be initialized until HLock() is called
		SInt16						length = 0;
		register SInt16				i = 0,
									j = 0;
		
		
		HLock(inURLDataHandle);
		string = *inURLDataHandle;
		length = GetHandleSize(inURLDataHandle) / sizeof(char);
		
		// first, eliminate the illegal double dot combination (a common typo); replace with one dot
		for (i = length - 1; i >= 0; i--)
		{
			if (string[i] == '.') if (i > 0) if (string[i - 1] == '.')
			{
				for (j = i; j < (length - 1); j++) string[j] = string[j + 1];
				(Boolean)resizeHandleSyncPointer(inURLDataHandle, &string, --length);
			}
		}
		
		// find out what kind of URL this is, and make intelligent corrections appropriately
		urlKind = URL_GetTypeFromDataHandle(inURLDataHandle);
		if (urlKind != kNotURL) switch (urlKind)
		{
			case kHttpURL:
			case kFtpURL:
			case kTelnetURL:
			case kRloginURL:
				// Check for suffix errors; valid slashed suffixes
				// are the following:
				//		.com		.net		.org		.edu
				//		.gov		.mil		.ar			.at
				//		.be			.bg			.ca			.cc
				//		.ch			.cz			.de			.dk
				//		.fi			.fr			.hu			.ie
				//		.il			.is			.it			.li
				//		.lu			.lt			.mc			.mx
				//		.nl			.no			.nu			.pt
				//		.ro			.se			.sk			.tr
				//		.uk			.us
				if (length > 4) // sanity check that makes SURE no assumptions are violated here
				{
					register SInt16		index = 0;
					SInt16				endPoint = length - 1;
					Boolean				isSlashedSuffix = false,
										ok = true;	// helps avoid trampling onward after memory errors
					
					
					// if the URL already ends with a slash, check for typos
					// starting with the preceding character instead of the
					// very last one
					if (string[length - 1] == '/') isSlashedSuffix = true, endPoint--;
					
					// work backwards; if a slash is encountered before a dot,
					// then the URL suffix is supposed to be slashed
					{
						Boolean		foundSlash = false,
									foundDot = false;
						
						
						for (index = endPoint; ((index) && (!foundDot) && (!foundSlash)); index--)
						{
							if (string[index] == '.') foundDot = true;
							else if (string[index] == '/') foundSlash = true;
						}
						if ((foundSlash) && (!foundDot)) isSlashedSuffix = true;
					}
					
					// check for a .co typo, make it .com
					if (ok)
					{
						index = endPoint;
						if ((string[index--] == 'o') &&
							(string[index--] == 'c') &&
							(string[index--] == '.'))
						{
							// enlarge the string by one character, to make room for the correction
							ok = resizeHandleSyncPointer(inURLDataHandle, &string, ++length);
							index = ++endPoint;
							string[index--] = 'm';
							isSlashedSuffix = true;
						}
					}
					
					// check for a .cmo typo, make it .com
					if (ok)
					{
						index = endPoint;
						if ((string[index--] == 'o') &&
							(string[index--] == 'm') &&
							(string[index--] == 'c') &&
							(string[index--] == '.'))
						{
							index = endPoint;
							string[index--] = 'm';
							string[index--] = 'o';
							isSlashedSuffix = true;
						}
					}
				}
				break;
			
			default:
				break;
		}
		HSetState(inURLDataHandle, hState);
	}
}// CleanUp


/*!
Parses the given data handle, assumed to be a
string, and returns what kind of URL it seems
to represent.

(2.6)
*/
UniformResourceLocatorType
URL_GetTypeFromDataHandle	(Handle		inURLDataHandle)
{
	short						size = 0;
	char*						textBegin = null;
	char*						textEnd = null;
	char*						p = null;
	char*						schemeName = null;
	UniformResourceLocatorType	result = kNotURL;
	
	
	size = GetHandleSize(inURLDataHandle);
	textBegin = *inURLDataHandle;
	textEnd = *inURLDataHandle + size - 1;
	
	// strip off leading white space and lagging white space
	while (IS_WHITE_SPACE_OR_CR_CHARACTER(*textBegin) && textBegin < textEnd) textBegin++;
	while (IS_WHITE_SPACE_OR_CR_CHARACTER(*textBegin) && textBegin < textEnd) textEnd--;
		
	// look for < and > on each end, and strip them off too
	if (*textBegin == '<') textBegin++;
	if (*textEnd == '>') textEnd--;
	
	// look for "URL:", strip it
	if ((notEqualStringSegments("URL:", textBegin, 4))) textBegin += 4;
	
	size = textEnd - textBegin + 1;
	
	if (size > 0)
	{
		long	i = 0L;
		
		
		// strip CR inside
		p = textBegin;
		while (p < textEnd)
		{
			if (*p == CR)
			{
				BlockMoveData(p + 1,  p, size - (p - textBegin + 1));
				size--;
				textEnd--;
			}
			else p++;
		}
		
		// clean up and null-terminate
		BlockMoveData(textBegin, *inURLDataHandle, size);
		*(*inURLDataHandle + size) = 0;
		
		// get URL type
		p = *inURLDataHandle;
		while ((*p != ':') && (p < *inURLDataHandle + size - 1)) p++;
		if (p - *inURLDataHandle >= size - 1) // colon not found
		{
			// might be mail address
			//if (isMailAddress(inURLDataHandle)) result = kMailtoURL;
		//	else
				result = kNotURL;
		}
		else for (i = 0; ; i++)
		{
			schemeName = gURLSchemeNames[i];
			if (schemeName == null)
			{
				result = kNotURL;
				break;
			}
			if (notEqualStringSegments(*inURLDataHandle, schemeName, strlen(schemeName)))
			{
				result = (UniformResourceLocatorType)i;
				break;
			}
		}
	}
	return result;
}// GetTypeFromDataHandle


/*!
Examines the currently-selected text of the specified
telnet connection window for a valid URL.  If it finds
one, the text is flashed briefly and then the proper
Apple Events are fired to open the URL.  This method
has been rewritten in MacTelnet 3.0 to call the method
URL_ParseDataHandle().

(2.6)
*/
void
URL_HandleForScreenID	(const TelnetScreenID		inScreenID)
{
	Handle						urlH = RealScreen_GetSelectedTextAsNewHandle(inScreenID, 0, kRealScreenTextFlagInline);
	UniformResourceLocatorType	urlKind = kNotURL;
	OSErr						err = noErr;
	OSType						sig = '????';
	Boolean						isValidHandle = false;
	
	
	isValidHandle = IsHandleValid(urlH);
	if (isValidHandle)
	{
		if (GetHandleSize(urlH) > 0)
		{
			HLock(urlH);
			
			urlKind = URL_GetTypeFromDataHandle(urlH);
			if (urlKind != kNotURL)
			{
				sig = InternetConfig_HelperApplicationForURLType(urlKind);
				
				if (URL_IsTypeTelnetCanHandle(urlKind) || (sig != kInternetConfigUnknownHelper))
				{
					FlashSelection(inScreenID);
				}
				isValidHandle = URL_ParseDataHandle(urlH);
			}
		}
		else isValidHandle = false;
	}
	else isValidHandle = false;
	
	if (!isValidHandle) BEEP;
}// HandleForScreenID


/*!
Examines the data in the specified buffer for a valid URL.
If one is found, the URL is parsed, a connection is
attempted, and "true" is returned; otherwise, nothing
happens and "false" is returned.

(3.0)
*/
Boolean
URL_ParseDataHandle		(Handle		inURLDataHandle)
{
	OSErr			error = noErr;
	Boolean			result = false;
	
	
	if ((inURLDataHandle != (char**)-1L) && (inURLDataHandle != null))
	{
		UniformResourceLocatorType	urlKind = kNotURL;
		SInt8						hState = 0;
		
		
		hState = HGetState(inURLDataHandle);
		URL_CleanUp(inURLDataHandle); // make intelligent corrections to imperfect URLs
		HLock(inURLDataHandle);
		
		urlKind = URL_GetTypeFromDataHandle(inURLDataHandle);
		if (urlKind != kNotURL)
		{
			if (URL_IsTypeTelnetCanHandle(urlKind))
			{
				Connections_OpenFromURLString(*inURLDataHandle, GetHandleSize(inURLDataHandle));
			}
			else
			{
				OSType		helperApplicationSignature = '????';
				
				
				helperApplicationSignature = InternetConfig_HelperApplicationForURLType(urlKind);
				Console_WriteValueFourChars("URL open: helper application", helperApplicationSignature);
				if (helperApplicationSignature != kInternetConfigUnknownHelper)
				{
					Size		stringDataSize = GetHandleSize(inURLDataHandle);
					
					
					// 3.0 - added size information to correct a previous, potentially fatal assumption!
					Console_WriteValue("URL data size", stringDataSize);
					error = URL_OpenHelperForURLString(helperApplicationSignature, *inURLDataHandle,
														stringDataSize);
					result = (error == noErr);
				}
			}
		}
		HSetState(inURLDataHandle, hState);
	}
	return result;
}// ParseDataHandle


/*!
Sends a "handle URL" event back to MacTelnet.
This should be done for most URL handlers, so
that the user can record the action.

Returns "true" if successful, "false" otherwise.

(3.0)
*/
Boolean
URL_SendGetURLEventToSelf	(void const*		inDataPtr,
							 Size const			inDataSize)
{
	AppleEvent		handleURLEvent,
					reply;
	OSErr			error = noErr;
	Boolean			result = false;
	
	
	error = InstallAE_CreateRecordableAppleEvent(kStandardURLEventClass, kStandardURLEventIDGetURL, &handleURLEvent);
	
	// the URL string - REQUIRED
	if (error == noErr) error = AEPutParamPtr(&handleURLEvent, keyDirectObject, cString, inDataPtr, inDataSize);
	
	// handle the URL by sending an Apple Event; if it fails, report the error
	if (error == noErr) error = AESend(&handleURLEvent, &reply,
										kAENoReply | kAENeverInteract,
										kAENormalPriority, kAEDefaultTimeout, null, null);
	AEDisposeDesc(&handleURLEvent);
	
	result = (error == noErr);
	return result;
}// SendGetURLEventToSelf


#define isurlschemechar(c)				(isalnum((c)) || c == '+' || c == '.' || c == '-')


Boolean
URL_SetSelectionToProximalURL	(Point					curr,
								 const TelnetScreenID	w)
{
	/* called by RSSelect when Command Click has occured outside any selection.
   This routine looks for urls around the current point, makes that the selected area if 
   it finds one, and returns true.  Otherwise leaves the current selection area untouched
   and returns false. */
/* Note: the point we receive is already normalized to a character position */
	short columns;	
	Handle block;
	char *original, *start, *end, *textEnd, *p, *q;
	short i, numLines = 1;
	char EOLS = CR;
	short neededLines, endLine = 0, startLine = 0;
	short blockSize;
	Point startPoint, endPoint;
	short  currentLine;
	char *tempBlockPtr;
	short tempCounter, lineLength[20];
	Boolean found = false, allowSpaces=false;
	
	columns = VSgetcols(w);
	
	/* we need enough space to find a 255 char URL on either side of the cursor */
	neededLines = 510/columns; 
	if (255%columns != 0)
		neededLines += 2;
	
	/* want it to be symmetric about this line */
	if (neededLines%2 == 0)
		neededLines += 1;
	
	blockSize = neededLines*(columns + 1); //we need space for the \r
	
	block = Memory_NewHandle((long)blockSize);
	
	textEnd = *block + blockSize;
	
	HLock(block);
	
	/* get the lines we need */
	VSgettext(w, 0, curr.v - INTEGER_HALVED(neededLines - 1), columns - 1,
					curr.v + INTEGER_HALVED(neededLines - 1), *block, blockSize, &EOLS, 0,
					RealScreen_IsTextSelectionRectangular(w)/* rectangular text selection? */); 
	
	original = *block;
	for (i = 0; i < (neededLines-1)/2; i++)
	{
		while (*original != CR)
			original++;
		original++;
	}
	original+= curr.h + 1;

	tempBlockPtr = *block;
	for (i = 0; i < neededLines; i++)
	{
		tempCounter = 0;
		while (*tempBlockPtr++ != CR) tempCounter++;
		lineLength[i] = tempCounter;
	}

	p = original;
	
	q = original - 1;
	currentLine = (neededLines-1)/2 + 1;
	while ((!found)&&(p > *block))
	{
		while (p >= *block && !IS_WHITE_SPACE_OR_CR_CHARACTER(*p) && *p != '<' && *p != '"') p--;
		if (*p == CR)
		{
			if (lineLength[currentLine - 2] == columns)
			{	
				p--; //this is just a wrapped line
				currentLine--;
			}
			else
				found = true;
		}
		else
			found = true;
	}
	
	startLine = currentLine;
	if (*p != '<') 
		p++;
	else
		allowSpaces = true;
	
	
	currentLine = (neededLines-1)/2 + 1;
	
	found = false;
	while (!found)
	{
		while (q < textEnd && (*q != CR) &&(!IS_WHITE_SPACE_CHARACTER(*q) || allowSpaces) 
						&& *q != '>' && *q != '"') q++;
		if (*q == CR)
		{
			if (lineLength[currentLine-1] == columns)
			{	
				q++; //this is just a wrapped line
				currentLine++;
			}
			else
				found = true;
		}
		else
			found = true;
	}
	endLine = currentLine;
	if (*q != '>') q--;
	
	if (p >= q) return false;

	while (p < q && (IS_WHITE_SPACE_OR_CR_CHARACTER(*p))) p++;
	while (p < q && (IS_WHITE_SPACE_OR_CR_CHARACTER(*q))) q--;

	if (*p == '<') 
	{
		if (*q != '>') return false;
	}
	else 
	{
		if (*q == '>') return false;
	} 

	start = p;
	end = q;

	 		
	RealScreen_SelectNothing(w);
	HUnlock(block);
	
	startPoint.v = curr.v - ((neededLines-1)/2 + 1 - startLine);
	endPoint.v   = curr.v + (endLine -(neededLines-1)/2 - 1);
	
	if (startPoint.v == curr.v)
		startPoint.h = curr.h - (original - start);
	else
	{
		long numChars = 0;
		for (i = 0; i < startLine - 1; i++)
			numChars += lineLength[i] + 1;
		startPoint.h = (start - *block - numChars - 1);
	}	
	if (endPoint.v == curr.v)
		endPoint.h = curr.h + (end - original + 1);
	else
	{
		long numChars = 0;
		for (i = 0; i < endLine - 1; i++)
			numChars += lineLength[i] + 1;
		endPoint.h = (end - *block - numChars);
	}	

	RealScreen_SelectVirtualRange(w, startPoint, endPoint);
	
	return true;
}// SetSelectionToProximalURL


/*!
To determine if Telnet supports URLs of the
specified type, use this method.

(3.0)
*/
Boolean
URL_IsTypeTelnetCanHandle	(UniformResourceLocatorType		inType)
{
	Boolean		result = false;
	
	
	result = ((inType == kTelnetURL) || (inType == kRloginURL));
	return result;
}// IsTypeTelnetCanHandle


/*!
Opens a helper program (whose creator code is the
first parameter) and passes it a URL string.  If
no problems occur, "noErr" is returned.

(3.0)
*/
OSErr
URL_OpenHelperForURLString		(OSType		inApplicationSignature,
								 char*		inURLStringBuffer,
								 Size		inStringBufferSize)
{
	OSErr					result = noErr;
	ProcessSerialNumber		psn;
	FSSpec					appSpec;
	Boolean					running = false;
	unsigned short			launchControlFlags = launchContinue | launchNoFileFlags;
	
	
	Console_BeginFunction();
	Console_WriteLine("finding application...");
	result = FindAppFromSig(inApplicationSignature, &appSpec, &running, &psn);
	if (result == noErr)
	{
		Console_WriteLine("launching URL helper application...");
		result = LaunchAppWithEventAndString(running, &appSpec, &psn, kStandardURLEventClass,
												kStandardURLEventIDGetURL, keyDirectObject,
												inURLStringBuffer, inStringBufferSize, 0,
												launchControlFlags);
	}
	if (result != noErr) Console_WriteValue("URL helper launch: error", result);
	Console_EndFunction();
	return result;
}// OpenHelperForURLString



//
// internal methods
//

/*!
Determines if the first "n" characters
of the C string "s1" are identical to
the C string "s2".

(2.6)
*/
static Boolean
notEqualStringSegments	(char*		s1,
						 char*		s2,
						 short		n)
{
	unsigned char*	t = gTable;
	unsigned char*	us1 = (unsigned char*)s1;
	unsigned char*	us2 = (unsigned char*)s2;
	
	
	while (--n >= 0 && t[*us1++] == t[*us2++]) {}
	
	return (n < 0);
}// notEqualStringSegments


/*!
This method is used to automatically
synchronize a pointer that is based on
a handle, resizing the handle to the
specified size and changing the pointer
variable to match.  When this method is
finished, the handle is once again
locked and the pointer is a valid
dereference of the new handle.

If the handle could be resized, "true"
is returned; otherwise, there was not
enough memory to resize the handle, so
"false" is returned.

(3.0)
*/
static Boolean
resizeHandleSyncPointer		(Handle			inHandle,
							 char**			outHandleDereferencedPtr,
							 SInt32			inNewLengthInCharacters)
{
	Boolean		result = false;
	
	
	HUnlock(inHandle), *outHandleDereferencedPtr = null;
	result = (Memory_SetHandleSize(inHandle, inNewLengthInCharacters * sizeof(char)) == noErr);
	HLock(inHandle), *outHandleDereferencedPtr = *inHandle;
	
	return result;
}// resizeHandleSyncPointer