/*###############################################################

	Undoables.c
	
	Contexts Library 1.0
	 1998-2000 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// Mac includes
#include <MacTypes.h>
#include <PLStringFuncs.h>

// library includes
#include "ContextsLib.h"
#include "DataAccessLib.h"

// resource includes
#include "StringResources.h"



// types
struct UndoData
{
	Boolean						isRedoable;
	Str255						undoCommandName,
								redoCommandName;
	UndoActionProcPtr			actionProc;
	UndoableContextIdentifier	contextID;
	void*						contextPtr;
};
#ifndef __cplusplus
typedef struct UndoData		UndoData;
#endif
typedef UndoData*			UndoDataPtr;

// variables
static Str255			gDefaultUndoName,
						gDefaultRedoName;
static UndoDataPtr		gUndoablePtr = null,
						gRedoablePtr = null;

// internal methods
static void			destroyData		(UndoDataPtr*			inoutDataPtr);
static UndoDataPtr	refAcquireLock	(UndoableActionRef		inRef);
static void			refReleaseLock	(UndoableActionRef		inRef,
									 UndoDataPtr*			inoutPtr);



//
// public methods
//

/*!
Call this routine before any other in this module.
Currently, only one Undo handling mechanism is
supported.  If multiple Undo support (an Undoable
action stack) is implemented in the future, it
would be requested using a different parameter to
this initialization routine.

Pass two strings that will be returned by
Undoables_GetUndoCommandInfo() and
Undoables_GetRedoCommandInfo() in the event that no
pending action is installed (and therefore no
command text).  The second item is optional -
specify "null" if your application does not even
have a Redo item in its Edit menu.  Note that it is
better to have a Redo item even if your application
does not support it - keep in mind that if you use
Undoables_GetRedoCommandInfo() and your application
does not support Redo, you will still get back the
correct item text and state for the Redo item!  You
therefore are encouraged to always specify item text
for both Undo and Redo.

IMPORTANT:	I disagree with the pseudo-standard
			taken by many Macintosh programs that
			names inactive Undo Cant Undo.  This
			is wrong, period.  This introduces a
			double-negative, because the disabled
			state of the command obviously implies
			that the most recent action cant be
			done.  In English locales, the parameters
			to this routine should be Pascal strings
			with the terms Undo and Redo,
			respectively.

(1.0)
*/
void
Undoables_Init		(const UndoHandlingMechanism	inUndoHandlingMechanism,
					 StringPtr						inDisabledUndoCommandName,
					 StringPtr						inDisabledRedoCommandNameOrNullIfRedoIsNotUsed)
{
	PLstrcpy(gDefaultUndoName, inDisabledUndoCommandName);
	if (inDisabledRedoCommandNameOrNullIfRedoIsNotUsed != null)
	{
		PLstrcpy(gDefaultRedoName, inDisabledRedoCommandNameOrNullIfRedoIsNotUsed);
	}
	else PLstrcpy(gDefaultRedoName, EMPTY_PSTRING);
}// Init


/*!
Call this routine when you are finished
using this module.

(1.0)
*/
void
Undoables_Done		(void)
{
	// temp - for some reason, using destroyData() here causes a crash!
	//if (gUndoablePtr != null) destroyData(&gUndoablePtr);
	//if (gRedoablePtr != null) destroyData(&gRedoablePtr);
	gUndoablePtr = null;
	gRedoablePtr = null;
}// Done


/*!
As soon as you execute an undoable action
(such as in response to a menu item), you
can register it using this routine.

Specify the name of the menu item to use
for the undoable action (for example,
Undo Cut).  If the action can also be
redone later, provide a menu item name
for its Redo equivalent (Redo Cut, in
this example).  To specify that an action
cannot be redone, simply pass "null" in
the second parameter.

Provide a pointer to a procedure that,
given the additional context identifier
and data pointer, will be able to either
Undo or Redo the action specified.

Your "inHowToUndoAction" routine is also
responsible for disposing of allocated
memory (assuming your context is a chunk
of dynamically allocated memory).

(1.0)
*/
OSErr
Undoables_AddAction		(StringPtr						inUndoCommandName,
						 StringPtr						inRedoCommandNameOrNullIfNotRedoable,
						 UndoActionProcPtr				inHowToUndoAction,
						 UndoableContextIdentifier		inUserDefinedContextIdentifier,
						 void*							inUserDefinedContextPtr)
{
	OSErr		result = noErr;
	
	
	// for single-undo, destroy any previous action before creating a new one
	if (gUndoablePtr != null) destroyData(&gUndoablePtr);
	if (gRedoablePtr != null) destroyData(&gRedoablePtr);
	gUndoablePtr = (UndoDataPtr)Memory_NewPtr(sizeof(UndoData));
	result = MemError();
	
	if ((gUndoablePtr != null) && (result == noErr))
	{
		gUndoablePtr->isRedoable = (inRedoCommandNameOrNullIfNotRedoable != null);
		PLstrcpy(gUndoablePtr->undoCommandName, inUndoCommandName);
		if (inRedoCommandNameOrNullIfNotRedoable != null)
		{
			PLstrcpy(gUndoablePtr->redoCommandName, inRedoCommandNameOrNullIfNotRedoable);
		}
		else PLstrcpy(gUndoablePtr->redoCommandName, EMPTY_PSTRING);
		gUndoablePtr->actionProc = inHowToUndoAction;
		gUndoablePtr->contextID = inUserDefinedContextIdentifier;
		gUndoablePtr->contextPtr = inUserDefinedContextPtr;
	}
	else destroyData(&gUndoablePtr); // give up
	
	return result;
}// AddAction


/*!
To determine the state and correct text
for the Redo item in your application,
use this method.  You really specify
this information yourself, when you use
the Undoables_AddAction() method.  This
routine only keeps track of actions and
reports an accurate state based on
whether *it* knows how to handle Redo
at the moment, and if so, exactly what
Redo command is going to be executed if
Undoables_RedoLastUndo() is called.

(1.0)
*/
void
Undoables_GetRedoCommandInfo	(Str255			outCommandText,
								 Boolean*		outCommandEnabled)
{
	if (gRedoablePtr != null) PLstrcpy(outCommandText, gRedoablePtr->redoCommandName);
	else PLstrcpy(outCommandText, gDefaultRedoName);
	
	if (outCommandEnabled != null) *outCommandEnabled = (gRedoablePtr != null);
}// GetRedoCommandInfo


/*!
To determine the state and correct text
for the Undo item in your application,
use this method.  You really specify
this information yourself, when you use
the Undoables_AddAction() method.  This
routine only keeps track of actions and
reports an accurate state based on
whether *it* knows how to handle Undo
at the moment, and if so, exactly what
Undo command is going to be executed if
Undoables_UndoLastAction() is called.

(1.0)
*/
void
Undoables_GetUndoCommandInfo	(Str255			outCommandText,
								 Boolean*		outCommandEnabled)
{
	if (gUndoablePtr != null) PLstrcpy(outCommandText, gUndoablePtr->undoCommandName);
	else PLstrcpy(outCommandText, gDefaultUndoName);
	
	if (outCommandEnabled != null) *outCommandEnabled = (gUndoablePtr != null);
}// GetUndoCommandInfo


/*!
This should be the only action taken in
response to the user selecting Redo from
your applications Edit menu.

(1.0)
*/
void
Undoables_RedoLastUndo		(void)
{
	if (gRedoablePtr != null)
	{
		// invoke the Redo operation
		InvokeUndoableActionProc(gRedoablePtr->actionProc, kUndoActionInstructionRedo,
									gRedoablePtr->contextID, gRedoablePtr->contextPtr);
		
		// set it up for undoing
		if (gUndoablePtr != null) destroyData(&gUndoablePtr);
		gUndoablePtr = gRedoablePtr;
		gRedoablePtr = null;
	}
}// RedoLastUndo


/*!
This should be the only action taken in
response to the user selecting Undo from
your applications Edit menu.

(1.0)
*/
void
Undoables_UndoLastAction	(void)
{
	if (gUndoablePtr != null)
	{
		// invoke the Undo operation
		InvokeUndoableActionProc(gUndoablePtr->actionProc, kUndoActionInstructionUndo,
									gUndoablePtr->contextID, gUndoablePtr->contextPtr);
		
		// if the action is redoable, set it up for redoing
		if (gUndoablePtr->isRedoable)
		{
			if (gRedoablePtr != null) destroyData(&gRedoablePtr);
			gRedoablePtr = gUndoablePtr;
			gUndoablePtr = null;
		}
	}
}// UndoLastAction


//
// internal methods
//

/*!
To destroy an actions data, automatically
invoking the responder routine of the action
with the dispose instruction, use this
method.  On output, your copy of the pointer
is set to "null".

(1.0)
*/
static void
destroyData		(UndoDataPtr*		inoutDataPtr)
{
	if (inoutDataPtr != null)
	{
		// invoke the responder one last time to dispose of any user-defined data structures
		InvokeUndoableActionProc((*inoutDataPtr)->actionProc, kUndoActionInstructionDispose,
									(*inoutDataPtr)->contextID, (*inoutDataPtr)->contextPtr);
		
		// destroy the action data structure
		DisposePtr((Ptr)*inoutDataPtr), *inoutDataPtr = null;
	}
}// destroyData


/*!
To acquire a pointer to the internal structure,
given a reference to it, use this method.

(1.0)
*/
static UndoDataPtr
refAcquireLock	(UndoableActionRef		inRef)
{
	return ((UndoDataPtr)inRef);
}// refAcquireLock


/*!
To release a pointer to the internal structure,
possessing a reference to it, use this method.
On return, your copy of the pointer is set to
"null".

(1.0)
*/
static void
refReleaseLock	(UndoableActionRef		inRef,
				 UndoDataPtr*			inoutPtr)
{
	// the memory is not relocatable, so do nothing with the reference
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock