/*!	\file UniversalDefines.h
	\brief Definitions with "#define" that can be applied to
	ANY Mac program.
	
	Do not simply put whatever you want in this file: a good
	rule of thumb is, do not place anything in this file that
	is specific to an application or library.  These defines
	should be applicable to any modern Mac program.
	
	There is a section for Carbon helpers, which create
	mock-up macros that mimic the proposed Carbon APIs.  When
	compiling under the Carbon APIs, TARGET_API_MAC_CARBON is
	defined, making the macros disappear and allowing the true
	Carbon APIs to take their place.
*/
/*###############################################################

	Universal Defines
		 1998-2003 by Kevin Grant
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.
	
	You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#ifndef __UNIVERSALDEFINES__
#define __UNIVERSALDEFINES__

// standard-C includes
#ifdef __cplusplus
#	include <cassert>
#	include <climits>
#else
#	include <assert.h>
#	include <limits.h>
#endif

// Mac includes
#include <CodeFragments.h>
#include <MacTypes.h>
#include <Sound.h>



/*###############################################################
	UNUSED ARGUMENTS (PREVENTING COMPILER WARNINGS)
###############################################################*/

//! use to make unused function parameters unnamed;
//! e.g. int f (int arg) would be int f (int UNUSED_ARGUMENT(arg))
#ifdef __MRC__
#	define UNUSED_ARGUMENT(arg) (arg)
#	define UNUSED_ARGUMENT_CARBON(arg) (arg)
#	define UNUSED_ARGUMENT_CLASSIC(arg) (arg)
#else
#	define UNUSED_ARGUMENT(arg)
#	if TARGET_API_MAC_CARBON
#		define UNUSED_ARGUMENT_CARBON(arg)
#		define UNUSED_ARGUMENT_CLASSIC(arg) (arg)
#	else
#		define UNUSED_ARGUMENT_CARBON(arg) (arg)
#		define UNUSED_ARGUMENT_CLASSIC(arg)
#	endif
#endif



/*###############################################################
	C IMPROVED
###############################################################*/

//! C string type, defined as a NULL-terminated character array
typedef char*			CStringPtr;
typedef char const*		ConstCStringPtr;

//!  la Java, this is my preferred null-pointer term
#define null 0L

//!  la Perl, this is my preferred if-not term
#define unless(x)		if (!(x))

//! these return values should be used for "pascal Boolean" functions
enum
{
	pascal_true = (-1),
	pascal_false = 0
};

//! C++ hack (why the hell did they make "this" a pointer?)
#define THIS		(*this)

//! C++ helper to handle compilers that dont define things in the CPP_STD:: namespace
#ifdef __MACH__
#	define CPP_STD
#else
#	define CPP_STD		::std
#endif

//! useful cast operators that are completely unnecessary, but make code
//! more self-documenting (they clarify the programmers intentions)
#ifdef __cplusplus
#define CONST_CAST(ref, intendedType)			const_cast<intendedType>(ref)
#define VOLATILE_CAST(ref, intendedType)		const_cast<intendedType>(ref) // C++ says const_cast overrides volatile too
#define REINTERPRET_CAST(ref, intendedType)		reinterpret_cast<intendedType>(ref)
#define STATIC_CAST(ref, intendedType)			static_cast<intendedType>(ref)
#else
#define CONST_CAST(ref, intendedType)			((intendedType)(ref))	//!< I know I am turning a constant pointer into a non-constant one
#define VOLATILE_CAST(ref, intendedType)		((intendedType)(ref))	//!< I know I am turning a volatile value into a non-volatile one
#define REINTERPRET_CAST(ref, intendedType)		((intendedType)(ref))	//!< I want to look at data as something else (e.g. integer to pointer)
#define STATIC_CAST(ref, intendedType)			((intendedType)(ref))	//!< identical to a cast in the traditional sense (e.g. "char" to "int")
#endif



/*###############################################################
	MAC OS CONSTANTS AND UTILITIES
###############################################################*/

enum
{
	//! for consistency purposes, this is very useful
	RGBCOLOR_INTENSITY_MAX = 65535
};

//! has a value of true if the specified Mac OS application programming interface can be found,
//! or false otherwise; this is Apples supported method for determining the presence of any
//! particular Mac OS system call at runtime
#define API_AVAILABLE(a)	((UInt32)(a) != (UInt32)kUnresolvedCFragSymbolAddress)

//! shortcut for emitting a system beep, useful because the parameter to SysBeep is now ignored
#define BEEP  		SysBeep(128)

//! useful integer routines, for efficiency; optimized for the PowerPC platform
#ifdef __cplusplus
inline int INTEGER_ABSOLUTE		(long a)			{ return ((a >= 0) ? a : -a); }		//!< take absolute value of an integer
inline int INTEGER_MAXIMUM		(long a, long b)	{ return ((a > b) ? a : b); }		//!< return largest of two signed integers
inline int INTEGER_MINIMUM		(long a, long b)	{ return ((a < b) ? a : b); }		//!< return smallest of two signed integers
inline Size INTEGER_MEGABYTES	(long a)			{ return (a << 20); }				//!< return byte count for integer value measured in MB
inline Size INTEGER_KILOBYTES	(long a)			{ return (a << 10); }				//!< return byte count for integer value measured in K
inline int INTEGER_OCTUPLED		(long a)			{ return (a << 3); }				//!< multiply integer by 8
inline int INTEGER_QUADRUPLED	(long a)			{ return (a << 2); }				//!< multiply integer by 4
inline int INTEGER_TRIPLED		(long a)			{ return ((a << 1) + a); }			//!< multiply integer by 3
inline int INTEGER_DOUBLED		(long a)			{ return (a << 1); }				//!< multiply integer by 2
inline int INTEGER_HALVED		(long a)			{ return (a >> 1); }				//!< divide integer by 2
inline int INTEGER_QUARTERED	(long a)			{ return (a >> 2); }				//!< divide integer by 4
inline int INTEGER_EIGHTHED		(long a)			{ return (a >> 3); }				//!< divide integer by 8
#else
#	define INTEGER_ABSOLUTE(a)		(((a) >= 0) ? (a) : (-(a)))
#	define INTEGER_MAXIMUM(a, b)	(((a) > (b)) ? (a) : (b))
#	define INTEGER_MINIMUM(a, b)	(((a) < (b)) ? (a) : (b))
#	define INTEGER_MEGABYTES(a)		((Size)((a) << 20))
#	define INTEGER_KILOBYTES(a)		((Size)((a) << 10))
#	define INTEGER_OCTUPLED(a)		((a) << 3)
#	define INTEGER_QUADRUPLED(a)	((a) << 2)
#	define INTEGER_TRIPLED(a)		(((a) << 1) + (a))
#	define INTEGER_DOUBLED(a)		((a) << 1)
#	define INTEGER_HALVED(a)		((a) >> 1)
#	define INTEGER_QUARTERED(a)		((a) >> 2)
#	define INTEGER_EIGHTHED(a)		((a) >> 3)
#endif

//! debugging variable
extern SInt32		DEBUG_STATE;



/*###############################################################
	CARBON ACCESSOR MACROS - LOW-OVERHEAD CARBON APIS
###############################################################*/

// Feel free to add to this section if there is an accessor you need
// that is not defined.  The idea here is to NOT require CarbonLib
// and yet not #ifdef 50,000 lines of code so that it can compile
// on Mac OS X.  I very much like low-overhead Classic applications,
// and frankly having to link with CarbonLib slows down launch time.
//
// Do be careful, however, that your equivalent API does what it
// is supposed to do.  Carbon documentation is useful. :)



#if TARGET_API_MAC_OS8

	// validators
	#define IsHandleValid(inHandle)							(((inHandle) != null) && ((*(inHandle)) != null))
	#define IsPointerValid(inPtr)							((inPtr) != null)
	
	// AppleEvent-related accessors
	#define AEGetDescDataSize(inAEDescPtr)					(((inAEDescPtr)->dataHandle == null) ? 0L : GetHandleSize((inAEDescPtr)->dataHandle))
	
	// window-related accessors
	#define GetPixBounds(inPixMapHandle, outRectPtr)		(*(outRectPtr) = (*inPixMapHandle)->bounds)
	#define GetPortBackColor(inCGrafPtr, outRGBColorPtr)	(*(outRGBColorPtr) = ((inCGrafPtr)->rgbBkColor))
	#define GetPortBackPixPat(inCGrafPtr, outPattern)		((outPattern) = (inCGrafPtr)->bkPixPat)
	#define GetPortBitMapForCopyBits(inCGrafPtr)			(&((GrafPtr)inCGrafPtr)->portBits)
	#define GetPortBounds(inCGrafPtr, outRectPtr)			(*(outRectPtr) = (inCGrafPtr)->portRect)
	#define GetPortClipRegion(inCGrafPtr, outRegion)		(CopyRgn((inCGrafPtr)->clipRgn, (outRegion)))
	#define GetPortForeColor(inCGrafPtr, outRGBColorPtr)	(*(outRGBColorPtr) = ((inCGrafPtr)->rgbFgColor))
	#define GetPortGrafProcs(inCGrafPtr)					((inCGrafPtr)->grafProcs)
	#define GetPortPenMode(inCGrafPtr)						((inCGrafPtr)->pnMode)
	#define GetPortPenPixPat(inCGrafPtr, outPattern)		((outPattern) = (inCGrafPtr)->pnPixPat)
	#define GetPortPixMap(inCGrafPtr)						((inCGrafPtr)->portPixMap)
	#define GetPortTextFace(inCGrafPtr)						((inCGrafPtr)->txFace)
	#define GetPortTextFont(inCGrafPtr)						((inCGrafPtr)->txFont)
	#define GetPortTextMode(inCGrafPtr)						((inCGrafPtr)->txMode)
	#define GetPortTextSize(inCGrafPtr)						((inCGrafPtr)->txSize)
	#define GetPortVisibleRegion(inCGrafPtr, outRegion)		(CopyRgn((inCGrafPtr)->visRgn, (outRegion)))
	#define GetRegionBounds(inRgnHandle, outRectPtr)		(*(outRectPtr) = (**(inRgnHandle)).rgnBBox)
	#define SetPortBackPixPat(inCGrafPtr, inPattern)		((inCGrafPtr)->bkPixPat = (inPattern))
	#define SetPortGrafProcs(inCGrafPtr, inCQDProcsPtr)		((inCGrafPtr)->grafProcs = (inCQDProcsPtr))
	#define SetPortPenMode(inCGrafPtr, inPenMode)			((inCGrafPtr)->pnMode = (inPenMode))
	#define SetPortPenPixPat(inCGrafPtr, inPattern)			((inCGrafPtr)->pnPixPat = (inPattern))
	
	// control accessors
	#define GetControlBounds(inControlHandle, outRectPtr)	(*(outRectPtr) = (**(inControlHandle)).contrlRect)
	#define GetControlHilite(inControlHandle)				((**(inControlHandle)).contrlHilite)
	#define GetControlOwner(inControlHandle)				((**(inControlHandle)).contrlOwner)
	#define IsControlHilited(inControlHandle)				(GetControlHilite(inControlHandle) > 0)
	#define SetControlBounds(inControlHandle, inRectPtr)	((**(inControlHandle)).contrlRect = *(inRectPtr))
	
	// menu accessors
	#define GetMenuID(inMenuHandle)							((**(inMenuHandle)).menuID)
	#define SetMenuWidth(inMenuHandle, inMenuWidth)			(*(inMenuHandle))->menuWidth = (inMenuWidth)
	#define SetMenuHeight(inMenuHandle, inMenuHeight)		(*(inMenuHandle))->menuHeight = (inMenuHeight)
	
	// list accessors
	#define GetListActive(inListHandle)						((*(inListHandle))->lActive)
	#define GetListDataBounds(inListHandle, outLBPtr)		(*(outLBPtr) = (**(inListHandle)).dataBounds)
	#define GetListHorizontalScrollBar(inListHandle)		((*(inListHandle))->hScroll)
	#define GetListPort(inListHandle)						((CGrafPtr)(*(inListHandle))->port)
	#define GetListVerticalScrollBar(inListHandle)			((*(inListHandle))->vScroll)
	#define SetListFlags(inListHandle, inFlags)				(*(inListHandle))->listFlags = (inFlags)
	#define SetListSelectionFlags(inListHandle, inFlags)	(*(inListHandle))->selFlags = (inFlags)
	
	// QuickDraw globals accessors
	#define GetScreenBounds(outRectPtr)						(*(outRectPtr) = qd.screenBits.bounds)
	#define GetQDGlobalsArrow(inoutCursorPtr)				((*inoutCursorPtr) = qd.arrow)
	#define GetQDGlobalsBlack(inoutPatternPtr)				((*inoutPatternPtr) = qd.black)
	#define GetQDGlobalsGray(inoutPatternPtr)				((*inoutPatternPtr) = qd.gray)
	#define GetQDGlobalsRandomSeed()						(qd.randSeed)
	#define SetQDGlobalsRandomSeed(inSInt32)				(qd.randSeed = inSInt32)
	#define GetQDGlobalsThePort()							(&qd.thePort)
	#define GetQDGlobalsWhite(inoutPatternPtr)				((*inoutPatternPtr) = qd.white)

#else

	// the above macros should be APIs under Carbon

#endif /* if TARGET_API_MAC_OS8 */

#endif