/*###############################################################

	VirtualDevices.c
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// library includes
#include "InterfaceLib.h"

// MacTelnet includes
#include "Console.h"
#include "VirtualDevices.h"



// constants
enum
{
	PIXEL_DEPTH = 8
};

// types
struct VDevice
{
	GWorldPtr			whichWorld;			// GDevice created off-screen
	CTabHandle			colorTable;			// graphics worlds color table
	Rect				bounds;				// boundary rectangle for virtual device
};
typedef struct VDevice		VDevice;
typedef VDevice*			VDevicePtr;

// variables
static CGrafPtr		origPort = null;
static GDHandle		origDev = null;

// internal methods
static Boolean		isOK				(VDeviceError			inError);
static VDevicePtr	refAcquireLock		(VirtualDeviceRef		inRef);
static void			refReleaseLock		(VirtualDeviceRef		inRef,
										 VDevicePtr*			inoutPtr);



//
// public methods
//

/*!
This method creates a new virtual device
with a specific boundary and color table.

Possible return codes:

"kVDeviceErrorInsufficientMemory"
	There is not enough memory to create
	a device.  The output device reference
	is set to null.

"kVDeviceErrorPointerCheck"
	A specified pointer was null.

"kVDeviceErrorEmptyBoundaries"
	Successfully created a new video device,
	but the specified rectangle was empty.
	Therefore, the new device was given a
	default rectangle that may not be the
	desired boundary.

"kVDeviceErrorNoError"
	Successfully created a new video device.

(3.0)
*/
VDeviceError
VirtualDevices_New		(VirtualDeviceRef*		inoutVirtualDeviceRefPtr,
						 const Rect*			inInitialDeviceBounds,
						 PaletteHandle			inInitialPaletteHandle)
{
	VDeviceError		result = kVDeviceErrorNoError;
	
	
	Console_BeginFunction();
	Console_WriteLine("New virtual device...");
	if (inoutVirtualDeviceRefPtr != null)
	{
		*inoutVirtualDeviceRefPtr = (VirtualDeviceRef)Memory_NewPtr(sizeof(VDevice));
		if (*inoutVirtualDeviceRefPtr != null)
		{
			VDevicePtr		ptr = refAcquireLock(*inoutVirtualDeviceRefPtr);
			
			
			GetGWorld(&origPort, &origDev); // save old settings
			
			// copy the given rectangle
			SetRect(&ptr->bounds, 0, 0, 0, 0);
			if (inInitialDeviceBounds != null)
			{
				SetRect(&ptr->bounds, inInitialDeviceBounds->left, inInitialDeviceBounds->top,
							inInitialDeviceBounds->right, inInitialDeviceBounds->bottom);
				if (EmptyRect(&ptr->bounds))
				{
					// empty rectangles are probably indicative of greater
					// problems, but allow them anyway...
					Console_WriteLine("warning: virtual device rectangle is empty");
					result = kVDeviceErrorEmptyBoundaries;
				}
			}
			else result = kVDeviceErrorPointerCheck;
			
			// still fine?
			if (isOK(result))
			{
				// copy "inInitialPaletteHandle" into a new color table handle
				ptr->colorTable = (CTabHandle)Memory_NewHandle(sizeof(ColorTable));
				if (ptr->colorTable != null)
				{
					OSErr		error = noErr;
					
					
					Palette2CTab(inInitialPaletteHandle, ptr->colorTable);
					
					Console_WriteValue("virtual device: proposed width", ptr->bounds.right - ptr->bounds.left);
					Console_WriteValue("virtual device: proposed height", ptr->bounds.bottom - ptr->bounds.top);
					
					// initialize the graphics world
					#if 0
					error = NewGWorld(&ptr->whichWorld, PIXEL_DEPTH, &ptr->bounds,
										ptr->colorTable,
										(GDHandle)null/* device */,
										0/* flags */);
					#else
					// debug - try creating a basic (default) world
					error = NewGWorld(&ptr->whichWorld, PIXEL_DEPTH, &ptr->bounds,
										(CTabHandle)null/* colors */,
										(GDHandle)null/* device */,
										0/* flags */);
					#endif
					
					if ((ptr->whichWorld == null) || (error != noErr))
					{
						Console_WriteValue("virtual device: failed GWorld create: error", error);
						if (ptr->whichWorld == null)
						{
							DisposeGWorld(ptr->whichWorld), ptr->whichWorld = null;
							Memory_DisposeHandle(&((Handle)ptr->colorTable));
						}
						result = kVDeviceErrorInsufficientMemory;
					}
					
					// still fine?
					if (isOK(result))
					{
						PixMapHandle	pixMap = null;
						Rect			rect;
						
						
						SetGWorld(ptr->whichWorld, null);
						pixMap = GetGWorldPixMap(ptr->whichWorld);
						GetPortBounds(ptr->whichWorld, &rect);
						EraseRect(&rect);
						SetGWorld(origPort, origDev);
					}
				}
				else
				{
					Console_WriteLine("virtual device: failed to create color table!");
					result = kVDeviceErrorInsufficientMemory;
				}
			}
			refReleaseLock(*inoutVirtualDeviceRefPtr, &ptr);
		}
		else result = kVDeviceErrorInsufficientMemory;
	}
	else result = kVDeviceErrorPointerCheck;
	
	if (!isOK(result))
	{
		// invalidate the returned reference in the event of a serious error
		if (inoutVirtualDeviceRefPtr != null) *inoutVirtualDeviceRefPtr = null;
	}
	Console_WriteValue("result", result);
	Console_EndFunction();
	
	return result;
}// New


/*!
To dispose of a virtual device previously created
with a call to VirtualDevices_New(), use this method.

(3.0)
*/
void
VirtualDevices_Dispose		(VirtualDeviceRef*		inoutVirtualDeviceRefPtr)
{
	if (inoutVirtualDeviceRefPtr != null)
	{
		VDevicePtr		ptr = refAcquireLock(*inoutVirtualDeviceRefPtr);
		
		
		if (ptr != null)
		{
			if (ptr->colorTable != null) Memory_DisposeHandle(&((Handle)ptr->colorTable));
			if (ptr->whichWorld != null) DisposeGWorld(ptr->whichWorld), ptr->whichWorld = null;
			Memory_DisposePtr((Ptr*)&ptr);
		}
	}
}// Dispose


/*!
If you need to use CopyBits() for a virtual device,
use this routine to access the pixel map in terms
of a bitmap.  DONT use this routine for any other
purpose, since a BitMap is only supported under
Carbon for use with CopyBits().

(3.0)
*/
VDeviceError
VirtualDevices_GetBitMapForCopyBits		(VirtualDeviceRef		inVirtualDeviceRef,
										 BitMap const**			outBitMapPtrPtr)
{
	VDeviceError		result = kVDeviceErrorNoError;
	
	
	if (outBitMapPtrPtr != null)
	{
		VDevicePtr		ptr = refAcquireLock(inVirtualDeviceRef);
		
		
		if (ptr != null)
		{
			*outBitMapPtrPtr = GetPortBitMapForCopyBits(ptr->whichWorld);
		}
		else result = kVDeviceErrorPointerCheck;
		refReleaseLock(inVirtualDeviceRef, &ptr);
	}
	else result = kVDeviceErrorPointerCheck;
	
	return result;
}// GetBitMapForCopyBits


/*!
To access the bounding rectangle of a virtual device,
use this method.  The boundaries of the specified
device are copied into the given rectangle.

(3.0)
*/
VDeviceError
VirtualDevices_GetBounds	(VirtualDeviceRef		inVirtualDeviceRef,
							 Rect*					outRectPtr)
{
	VDeviceError		result = kVDeviceErrorNoError;
	
	
	if (outRectPtr != null)
	{
		VDevicePtr		ptr = refAcquireLock(inVirtualDeviceRef);
		
		
		if (ptr != null)
		{
			SetRect(outRectPtr,
					ptr->bounds.left, ptr->bounds.top, ptr->bounds.right, ptr->bounds.bottom);
		}
		else result = kVDeviceErrorPointerCheck;
		refReleaseLock(inVirtualDeviceRef, &ptr);
	}
	else result = kVDeviceErrorPointerCheck;
	
	return result;
}// GetBounds


/*!
To access the graphics world of a virtual device,
use this method.

(3.0)
*/
VDeviceError
VirtualDevices_GetGraphicsWorld		(VirtualDeviceRef		inVirtualDeviceRef,
									 GWorldPtr*				outGWorldPtrPtr)
{
	VDeviceError		result = kVDeviceErrorNoError;
	
	
	if (outGWorldPtrPtr != null)
	{
		VDevicePtr		ptr = refAcquireLock(inVirtualDeviceRef);
		
		
		if (ptr != null)
		{
			*outGWorldPtrPtr = ptr->whichWorld;
		}
		else result = kVDeviceErrorPointerCheck;
		refReleaseLock(inVirtualDeviceRef, &ptr);
	}
	else result = kVDeviceErrorPointerCheck;
	
	return result;
}// GetGraphicsWorld


/*!
To access the pixel map of the graphics world of a
virtual device, use this method.

(3.0)
*/
VDeviceError
VirtualDevices_GetPixelMap		(VirtualDeviceRef		inVirtualDeviceRef,
								 PixMapHandle*			outPixMapHandle)
{
	VDeviceError		result = kVDeviceErrorNoError;
	
	
	if (outPixMapHandle != null)
	{
		VDevicePtr		ptr = refAcquireLock(inVirtualDeviceRef);
		
		
		if (ptr != null)
		{
			*outPixMapHandle = GetPortPixMap(ptr->whichWorld);
		}
		else result = kVDeviceErrorPointerCheck;
		refReleaseLock(inVirtualDeviceRef, &ptr);
	}
	else result = kVDeviceErrorPointerCheck;
	
	return result;
}// GetPixelMap


/*!
To specify the colors to use for a virtual device,
use this method.

(3.0)
*/
VDeviceError
VirtualDevices_SetColorTable	(VirtualDeviceRef		inVirtualDeviceRef,
								 PaletteHandle			inPaletteToUseForColorTable)
{
	VDeviceError	result = kVDeviceErrorNoError;
	VDevicePtr		ptr = refAcquireLock(inVirtualDeviceRef);
	
	
	if (ptr != null)
	{
		PixMapHandle	pixMap = null;
		
		
		result = VirtualDevices_GetPixelMap(inVirtualDeviceRef, &pixMap);
		if (isOK(result))
		{
			CTabHandle		colorTable = null;
			
			
			colorTable = (*pixMap)->pmTable;
			if (colorTable == null) result = kVDeviceErrorInsufficientMemory;
			else
			{
				Palette2CTab(inPaletteToUseForColorTable, colorTable);
				
				(*colorTable)->ctSeed = GetCTSeed();	// give the table a unique seed
				(*colorTable)->ctFlags = 0x8000;		// high bit of 1 means from device
				
				// make a 3-bit inverse table
				#if 0
				{
					MakeITable(colorTable, (ptr->whichWorld->gdITable, 3);
				}
				#endif
			}
		}
	}
	else result = kVDeviceErrorPointerCheck;
	
	refReleaseLock(inVirtualDeviceRef, &ptr);
	return result;
}// SetColorTable


//
// internal methods
//

/*!
To determine whether a particular error
code is serious, use this method.  This
allows functions to continue even if a
result other than "kVDeviceErrorNoError"
occurs.

(3.0)
*/
static Boolean
isOK			(VDeviceError			inError)
{
	Boolean		result = true;
	
	
	switch (inError)
	{
		case kVDeviceErrorNoError:
		case kVDeviceErrorEmptyBoundaries:
			result = false;
			break;
		
		case kVDeviceErrorInsufficientMemory:
		case kVDeviceErrorPointerCheck:
		default:
			// most errors are serious
			result = true;
			break;
	}
	
	return result;
}// isOK


/*!
Use this method to acquire a pointer to the internal
structure, given a reference to it.  Get in the habit
of calling refReleaseLock() when finished with the
pointer you receive; this allows flexibility in the
future to make dynamically-allocated memory relocatable.

(3.0)
*/
static VDevicePtr
refAcquireLock	(VirtualDeviceRef		inRef)
{
	return ((VDevicePtr)inRef);
}// refAcquireLock


/*!
To release the lock on a reference that you previously
locked with refAcquireLock(), use this method.

(3.0)
*/
static void
refReleaseLock	(VirtualDeviceRef		inRef,
				 VDevicePtr*			inoutPtr)
{
	if (inoutPtr != null) *inoutPtr = null;
}// refReleaseLock