/*###############################################################

	VirtualScreenEmulation.c
	
	This file contains the private emulation calls for the
	NCSA Virtual Screen Kernel, including parsing of escape
	sequences.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#ifdef __cplusplus
#	include <cstdio>
#else
#	include <stdio.h>
#endif

// Mac includes
#include <Files.h>
#include <Script.h>

// resource includes
#include "TelnetVersion.h"

// library includes
#include "PrintingLib.h"

// MacTelnet includes
#include "ConnectionData.h"
#include "Console.h"
#include "EventLoop.h"
#include "FolderManager.h"
#include "MacLook.h"
#include "RealScreen.h"
#include "TelnetMenus.h"
#include "TelnetPrinting.h"
#include "Terminal.h"
#include "TextTranslation.h"
#include "VirtualScreenEmulation.h"
#include "VirtualScreenInternal.h"
#include "vsdata.h"
#include "vskeys.h"



// internal methods
static void		printingBegin		(void);
static void		printingEnd			(void);
static void		printingSpool		(unsigned char**		pc,
									 short*					pctr);
#define VSa(x)			(1 << ((x)-1))
#define VSgraph(x)  	((x) | 0x80)
#define VSnotgraph(x) 	((x) & 0x7F)


//
// public methods
//

/*!
This is the entry point to the terminal emulator.
It places characters on a virtual screen, and
performs interpretation of control characters and
escape sequences.  Some of the code for the
interpretation is in "VirtualScreenInternal.c".

(2.6)
*/
void
VirtualScreenEmulation_Parse	(unsigned char*		inoutString,	// string to parse
								 SInt16 const		inLength)		// length of string
{
	register SInt16		sx = 0;
    register SInt16		escflg = VSIw->escflg; // state of escape sequence interpretation
    SInt16				ctr = inLength; // 3.0 - use a COPY so that the string length parameters original value isnt lost!!!
	SInt16				insert = 0,
						attrib = 0,
						extra = 0;
	Boolean				wrapped = false;
	unsigned char*		c = inoutString;
    char*				current = null;
	char*				start = null;
	unsigned short*		acurrent = null;
	
	
	while (ctr > 0)
	{
		if (VSIw->prredirect) printingSpool(&c, &ctr/* count can decrease, depending on what is printed */);
		
		while ((escflg == 0) && (ctr > 0) && (*c < 32))
		{
			switch (*c)
			{
				case 0x1B: /* esc */
					escflg++;
					break;
				
				case 0x0E: /* shift out */
					if (VSIw->G1) VSIw->attrib = VSgraph(VSIw->attrib);
					else VSIw->attrib = VSnotgraph(VSIw->attrib);
					VSIw->charset = 1;
					break;
				
				case 0x0F: /* shift in */
					if (VSIw->G0) VSIw->attrib = VSgraph(VSIw->attrib);
					else VSIw->attrib = VSnotgraph(VSIw->attrib);
					VSIw->charset = 0;
					break;
				
				case 0x07: /* bell */
					RealScreen_AudioEvent(VSIwn);
					break;
				
				case 0x08: /* backspace */
					VSIw->x--;
					if (VSIw->x < 0)
					  /* hit left margin */
						VSIw->x = 0;
					break;
				
				case 0x0C: /* ff */
					VSIindex();
					break;
				
				case 0x09: /* ht */		/* Later change for versatile tabbing */
					VSItab();
					Terminal_FileCaptureWriteData(c,1);				
					break;
				
				case 0x0A: /* lf */
					VSIindex();
					break;
				
				case 0x0D: /* cr */
					VSIw->x = 0;
					Terminal_FileCaptureWriteData(c,1);				
					break;
				
				case 0x0B: /* vt */
					VSIindex();
					break;
			} 
			c++;
			ctr--;
		} 
		if ((escflg == 0) &&
			(ctr > 0) &&
			(*c & 0x80) &&
			(*c < 0xA0) &&
			(ScreenManager_GetEmulationTypeOf(VSIwn) == kTerminalEmulationTypeVT220))	// VT220 eightbit starts here
		{											
			switch (*c)								
			{									
				case 0x84: /* ind */		//same as ESC D
					VSIindex();		
					goto ShortCut;			
				
				case 0x85: /* nel */		//same as ESC E
					VSIw->x = 0;				
					VSIindex();				
					goto ShortCut;			
				
				case 0x88: /* hts */		//same as ESC H 
					VSIw->tabs[VSIw->x] = 'x';	
					goto ShortCut;	
				
				case 0x8D: /* ri */			//same as ESC M 
					VSIrindex();		
					goto ShortCut;	
				
				case 0x9B: /* csi */		//same as ESC [ 
					VSIapclear();			
					escflg = 2;			
					c++;			//CCP			
					ctr--;					
					break;						
				
				case 0x86: /* ssa */			// - same as ESC F */
				case 0x87: /* esa */			// - same as ESC G */
				case 0x8E: /* ss2 */			// - same as ESC N */
				case 0x8F: /* ss3 */			// - same as ESC O */
				case 0x90: /* dcs */			// - same as ESC P */
				case 0x93: /* sts */			// - same as ESC S */
				case 0x96: /* spa */			// - same as ESC V */
				case 0x97: /* epa */			// - same as ESC W */
				case 0x9D: /* osc */			// - same as ESC ] */
				case 0x9E: /* pm */				// - same as ESC ^ */
				case 0x9F: /* apc */			// - same as ESC _ */
				default:
					goto ShortCut;				
			} 					
		}// end if vt220
		
		while ((escflg == 0) &&
				(ctr > 0) &&
				(*c >= 32) &&
				!((*c & 0x80) &&
				(*c < 0xA0) &&
				(ScreenManager_GetEmulationTypeOf(VSIwn) == kTerminalEmulationTypeVT220)))
		{
			// print lines of text one at a time
			start = &VSIw->linest[VSIw->y]->text[VSIw->x]; /* start of area needing redrawing */
			current = start; /* where to put next char */
			acurrent = &VSIw->attrst[VSIw->y]->text[VSIw->x]; /* where to put corresponding attribute byte */
			attrib = VSIw->attrib; /* current writing attribute */
			insert = VSIw->IRM; /* insert mode (boolean) */
			wrapped = false; /* wrapped to next line (boolean) */
			extra = 0; /* overwriting last character of line  */
			sx = VSIw->x; /* starting column of area needing redrawing */
			
			if (VSIw->x > VSIw->maxwidth)
			{
				if (VSIw->DECAWM)
				{
					// wrap to next line 
					VSIw->x = 0;
					VSIindex();
				}
				else VSIw->x = VSIw->maxwidth; // stay at right margin
				
				current = start = &VSIw->linest[VSIw->y]->text[VSIw->x];
				acurrent = &VSIw->attrst[VSIw->y]->text[VSIw->x];
				sx = VSIw->x;
			}
			
			while ((ctr > 0) &&
					(*c >= 32) &&
					(!wrapped) &&
					!((*c & 0x80) &&
					(*c < 0xA0) &&
					(ScreenManager_GetEmulationTypeOf(VSIwn) == kTerminalEmulationTypeVT220)))
			{
				// write characters on a single line
				TextTranslation_ConvertTextToMac(c, 1, ScreenManager_GetConnectionDataOf(VSIwn)->national);
				if (insert) VSIinschar(1); // make room for the char
				*current = *c;
				*acurrent = attrib;
				c++;
				ctr--;
				if (VSIw->x < VSIw->maxwidth) // advance the cursor position
				{
					acurrent++;
					current++;
					VSIw->x++;
				}
				else // hit right margin
				{
					if (VSIw->DECAWM) // autowrap to start of next line
					{
						VSIw->x++;
						wrapped = true; // terminate inner loop 
					}
					else // stay at right margin
					{
						VSIw->x = VSIw->maxwidth;
						extra = 1; // cursor position doesn't advance 
					} 
				} 
			}
			
			// there is now a full line of text in the virtual screen - update the real screen
			{
				GrafPtr		oldPort = null;
				
				
				GetPort(&oldPort);
				RealScreen_SetPortScreenPort(VSIwn);
				extra += VSIw->x - sx;
				if (insert) RSinsstring(VSIwn, VSIw->x - extra, VSIw->y, VSIw->attrib, extra, start);
				else
				{ 
					short		x2 = 0,
								y2 = 0,
								offset = 0,
								sxCopy = sx,
								yCopy = VSIw->y,
								extraCopy = extra;
					
					
					if (!VSIclip(&sxCopy, &yCopy, &x2, &y2, &extraCopy, &offset))
					{
						RSdraw(VSIwn, sxCopy, yCopy, VSIw->attrib, extraCopy, (char*)(start + offset));
					}
				}
				
				//if (RSisInFront(VSIwn)) // if the frontmost window, validate the screen
				//	RSvalidateRect(VSIwn);
				SetPort(oldPort);
			}
			
			Terminal_FileCaptureWriteData((unsigned char*)start/* data to write */, extra/* buffer length */);
			
			// 3.0 - test speech (this implementation will be greatly enhanced in the near future
			{
				RealScreenDataPtr	realScreenPtr = ScreenManager_GetRealScreenDataOf(VSIwn);
				
				
				if (realScreenPtr != null) if (realScreenPtr->speech.enabled)
				{
					Boolean		doSpeak = false;
					
					
					switch (realScreenPtr->speech.mode)
					{
						case kTelnetSpeechModeSpeakAlways:
							doSpeak = true;
							break;
						
						case kTelnetSpeechModeSpeakWhenActive:
							doSpeak = IsWindowHilited(realScreenPtr->window);
							break;
						
						case kTelnetSpeechModeSpeakWhenInactive:
							doSpeak = !IsWindowHilited(realScreenPtr->window);
							break;
						
						default:
							doSpeak = false;
							break;
					}
					
					if (doSpeak)
					{
						Str255		string;
						
						
						string[0] = extra;
						BlockMove((Ptr)start, (Ptr)(string + 1), extra);
						SpeakString(string);
					}
				}
			}
		}/* while */

		while ((escflg == 1) && (ctr > 0))
		{
			// basic escape sequence processing
			switch (*c)
			{
				case 0x08:
					VSIw->x--;
					if (VSIw->x < 0)
						VSIw->x = 0;
					break;
				
				case '[': /* csi */
					VSIapclear();
					escflg++;
					break;
				
				case '7':
					VSIsave();
					goto ShortCut;				
				
				case '8':
					VSIrestore();
					goto ShortCut;				
				
				case 'c':
					VSIreset();
					break;
				
				case 'D':
					VSIindex();
					goto ShortCut;				
				
				case 'E':
					VSIw->x = 0;
					VSIindex();
					goto ShortCut;				
				
				case 'M':
					VSIrindex();
					goto ShortCut;				
				
				case '>':
					VSIw->DECPAM = 0;
					goto ShortCut;				
				
				case '=':
					VSIw->DECPAM = 1;
					goto ShortCut;				
				
				case 'Z':
					VTsendident();
					goto ShortCut;				
				
				case ' ':						
				case '*':						
				case '#':
					escflg = 3;
					break;
				
				case '(':
					escflg = 4;
					break;
				
				case ')':
					escflg = 5;
					break;
				
				case 'H':
					VSIw->tabs[VSIw->x] = 'x';
					goto ShortCut;				

#ifdef CISB
				case 'I':
					bp_ESC_I();
					break;
#endif CISB
				
				case ']':
					if (ScreenManager_GetConnectionDataOf(VSIwn)->Xterm) escflg = 6;
					break;
					
				default:
					goto ShortCut;				
			}
			c++;
			ctr--;
		}/* while */
		
		while ((escflg == 2) && (ctr > 0))
		{
			// control sequence processing
			switch (*c)
			{
				case 0x08:
					VSIw->x--;
					if (VSIw->x < 0) VSIw->x = 0;
					break;
				
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				  /* parse numeric parameter */
					if (VSIw->parms[VSIw->parmptr] < 0)
						VSIw->parms[VSIw->parmptr] = 0;
						
					VSIw->parms[VSIw->parmptr] = VSIw->parms[VSIw->parmptr] * 10;
					VSIw->parms[VSIw->parmptr] += *c - '0';
					break;
				
				case '?':
				  /* DEC-private control sequence */
					VSIw->parms[VSIw->parmptr++] = -2;
					break;
				
				case ';':
				  /* parameter separator */
					VSIw->parmptr++;
					break;
				
				case 'A': /* cursor up */
#if 1														
					if (VSIw->parms[0]<1) VSIw->y--;		
					else VSIw->y-=VSIw->parms[0];			
					if ( VSIw->y < 0 ) VSIw->y=0;			
					if (VSIw->y < VSIw->top)				/* NCSA: SB */
						VSIw->y = VSIw->top;				/* NCSA: SB */
						
#else														
					if (VSIw->y < VSIw->top)
					  {
					  /* outside scrolling region */
						if (VSIw->parms[0] < 1)
							VSIw->y--;
						else
							VSIw->y -= VSIw->parms[0];
					  }
					else
					  {
					  /* don't leave scrolling region */
						if (VSIw->parms[0] < 1)
							VSIw->y--;
						else
							VSIw->y -= VSIw->parms[0];
						if (VSIw->y < VSIw->top)
							VSIw->y = VSIw->top;
					  }
#endif														
					VSIrange();
					goto ShortCut;				
				
				case 'B': /* cursor down */
#if 1														
					if (VSIw->parms[0]<1) VSIw->y++;		
					else VSIw->y+=VSIw->parms[0];			
					if (VSIw->y > VSIw->bottom)				/* NCSA: SB */
						VSIw->y = VSIw->bottom;				/* NCSA: SB */
						
#else														
					if (VSIw->y > VSIw->bottom)
					  {
					  /* outside scrolling region */
						if (VSIw->parms[0] < 1)
							VSIw->y++;
						else
							VSIw->y += VSIw->parms[0];
					  }
					else
					  {
					  /* don't leave scrolling region */
						if (VSIw->parms[0] < 1)
							VSIw->y++;
						else
							VSIw->y += VSIw->parms[0];
						if (VSIw->y > VSIw->bottom)
							VSIw->y = VSIw->bottom;
					  }
#endif														
					VSIrange();
					goto ShortCut;				
				
				case 'C': /* cursor right */
					if (VSIw->parms[0] < 1)
						VSIw->x++;
					else
						VSIw->x += VSIw->parms[0];
					VSIrange();
					if (VSIw->x > VSIw->maxwidth)
						VSIw->x = VSIw->maxwidth;
					goto ShortCut;				
				
				case 'D': /* cursor left */
					if (VSIw->parms[0] < 1)
						VSIw->x--;
					else
						VSIw->x -= VSIw->parms[0];
					VSIrange();
					goto ShortCut;				
				
				case 'f':
				case 'H':
				  /* absolute cursor positioning */
					VSIw->x = VSIw->parms[1] - 1;
					if (VSIw->DECORG)
					  /* origin mode -- position relative to top of scrolling region */
						VSIw->y = VSIw->parms[0] - 1 + VSIw->top;
					else
						VSIw->y = VSIw->parms[0] - 1;
					/*	Don't use actual VSIrange 'cause it will wrap us to first column if 
						we are past screen edge.  This causes "resize" to break */
					if (VSIw->x < 0)						/* JMB 2.6 */
						VSIw->x = 0;						/* JMB 2.6 */
					if (VSIw->x > (VSIw->maxwidth))			/* JMB 2.6 */
						VSIw->x = VSIw->maxwidth;			/* JMB 2.6 */
					if (VSIw->y < 0)						/* JMB 2.6 */
						VSIw->y = 0;						/* JMB 2.6 */
					if (VSIw->y > VSIw->lines)				/* JMB 2.6 */
						VSIw->y = VSIw->lines;				/* JMB 2.6 */
					if ((VSIw->forcesave)&&(VSIw->y == 0)&&(VSIw->x == 0)) //CCP better FORCESAVE
						VSIw->possibleForce = true;

					goto ShortCut;				
		       
			    case 'i':											/* PR: media copy */
					if (VSIw->parms[VSIw->parmptr] == 5)
					{
						/*c++; ctr--; */							/* PR */			
						printingBegin();	/* PR - set status and open temp file etc */
									/* PR - chars will be redirected at top of loop  */
									/* PR -  in this procedure */
					}				/* PR */
		            escflg = 0;		/* PR */
		            break;			/* PR */
				
				case 'K':
				  /* erase to beginning/end/whole of line */
					switch (VSIw->parms[0])
					  {
						case -1:
						case  0:
							VSIeeol();
							break;
						case  1:
							VSIebol();
							break;
						case  2:
							VSIel(-1);
							break;
						default:
							goto ShortCut;		
					  } /* switch */
					goto ShortCut;				
				
				case 'J':
				  /* erase to beginning/end/whole of screen */
					switch (VSIw->parms[0])
					  {
						case -1:
						case  0:
							VSIeeos();
							break;
						case  1:
							VSIebos();
							break;
						case  2:
							VSIes();
							break;
						default:
							goto ShortCut;		
					  } /* switch */
					goto ShortCut;				
				
				case 'm':
				  /* set/clear attributes */
				  {
					short temp = 0;
					while (temp <= VSIw->parmptr)
					  {
						short p;
						
						if (VSIw->parms[VSIw->parmptr] < 0)
							VSIw->parms[VSIw->parmptr] = 0;
      					
      					p = VSIw->parms[temp];
      					if (p == 0) 
      						VSIw->attrib &= 0x80; //all off
      					else if (p<8) 
      						VSIw->attrib |= VSa(p);//set an attribute
						else if (p>21 && p<28) 
							VSIw->attrib &= ~VSa(p-20); //clear an attribute
              			else if (ScreenManager_GetRealScreenDataOf(VSIwn)->areANSIColorsEnabled) 
              			{
 							if (p>=30 && p<38)
								VSIw->attrib = (VSIw->attrib&~0x0700) | ((p-30)<<8) | 0x0800;
							else if (p>=40 && p<48)
								VSIw->attrib = (VSIw->attrib&~0x7000) | ((p-40)<<12) | 0x8000;
						}
						temp++;
					  } /* while */
				  }
				  goto ShortCut;				
				
				case 'q':
					// flash LEDs
				  	LED_FlashQuick(); // 3.0 - well, now SOMETHING LED-related happens - dont know if its the RIGHT thing...
					goto ShortCut;				
				
				case 'c':
					VTsendident();
					goto ShortCut;				
				
				case 'n':
					switch (VSIw->parms[0])
					  {
						case 5:
							VTsendstat();
							break;
						case 6:
							VTsendpos();
							break;
					  } /* switch */
					goto ShortCut;				
				
				case 'L':
					if (VSIw->parms[0] < 1)
						VSIw->parms[0] = 1;
					VSIinslines(VSIw->parms[0], -1);
					goto ShortCut;				
				
				case 'M':
					if (VSIw->parms[0] < 1)
						VSIw->parms[0] = 1;
					VSIdellines(VSIw->parms[0], -1);
					goto ShortCut;				
				
				case '@':
					if (VSIw->parms[0] < 1)
						VSIw->parms[0] = 1;
					VSIinschar(VSIw->parms[0]);
					VSredrawLine(VSIwn); //redraws current line
					goto ShortCut;
				
				case 'P':
					if (VSIw->parms[0] < 1)
						VSIw->parms[0] = 1;
					VSIdelchars(VSIw->parms[0]);
					goto ShortCut;				
				
				case 'r':
				  /* set scrolling region */
					if (VSIw->parms[0] < 0)
						VSIw->top = 0;
					else
						VSIw->top = VSIw->parms[0] - 1;
					if (VSIw->parms[1] < 0)
						VSIw->bottom = VSIw->lines;
					else
						VSIw->bottom = VSIw->parms[1] - 1;
					if (VSIw->top < 0)
						VSIw->top = 0;
					if (VSIw->top > VSIw->lines - 1)
						VSIw->top = VSIw->lines - 1;
					if (VSIw->bottom < 1)
						VSIw->bottom = VSIw->lines;
					if (VSIw->bottom > VSIw->lines)
						VSIw->bottom = VSIw->lines;

					if (VSIw->top >= VSIw->bottom)
					{
						if (VSIw->bottom >= 1) VSIw->top = VSIw->bottom - 1;
						else VSIw->bottom = VSIw->top + 1;
					}
 				
 					VSIw->x = 0;
					VSIw->y = 0;
					if (VSIw->DECORG) VSIw->y = VSIw->top; // origin mode relative
					goto ShortCut;				
				
				case 'h':
				  /* set options */
					VSIsetoption(1);
					goto ShortCut;				
				
				case 'l':
				  /* reset options */
					VSIsetoption(0);
					goto ShortCut;				
				
				case 'g':
					if (VSIw->parms[0] == 3)
					  /* clear all tabs */
						VSItabclear();
					else if (VSIw->parms[0] <= 0)
					  /* clear tab at current position */
						VSIw->tabs[VSIw->x] = ' ';
					goto ShortCut;				
				
				case '!':
				case '\'':
				case '\"':
					escflg++;					
					break;						
				
				default:			/* Dang blasted strays... */
					goto ShortCut;				
			}
			c++;
			ctr--;
		}/* while */

		while ((escflg == 3) && (ctr > 0))
		{
			// "#" handling
			// (no support for double-width and double-height characters yet)
			switch (*c)
			{
				case 0x08:
					VSIw->x--;
					if (VSIw->x < 0) VSIw->x = 0;
					break;
				
				case '8': // alignment display
					VTalign();
					goto ShortCut;				
				
				default:
					goto ShortCut;				
			}
			c++;
			ctr--;
		}/* while */

		while ((escflg == 4) && (ctr > 0))
		{
			// "(" handling (selection of G0 character set)
			switch (*c)
			{
				case 0x08:
					VSIw->x--;
					if (VSIw->x < 0) VSIw->x = 0;
					break;
				
				case 'A': /* UK */
				case 'B': /* US */
				case '1': /* "soft" */
					VSIw->G0 = 0;
					if (!VSIw->charset)
						VSIw->attrib = VSnotgraph(VSIw->attrib);
					goto ShortCut;				
				
				case '0': /* DEC special graphics */
				case '2': /* "soft" */
					VSIw->G0 = 1;
					if (!VSIw->charset) VSIw->attrib = VSgraph(VSIw->attrib);
					goto ShortCut;				
				
				default:
					goto ShortCut;				
			}
			c++;
			ctr--;
		}/* while */
	
		while ((escflg == 5) && (ctr > 0))
		{
			// ")" handling (selection of G1 character set)
			switch (*c)
			{
				case 0x08:
					VSIw->x--;
					if (VSIw->x < 0) VSIw->x = 0;
					break;
				
				case 'A': /* UK */
				case 'B': /* US */
				case '1': /* "soft" */
					VSIw->G1 = 0;
					if (VSIw->charset) VSIw->attrib = VSnotgraph(VSIw->attrib);
					goto ShortCut;				
				
				case '0': /* DEC special graphics */
				case '2': /* "soft" */
					VSIw->G1 = 1;
					if (VSIw->charset) VSIw->attrib = VSgraph(VSIw->attrib);
					goto ShortCut;				
				
				default:
					goto ShortCut;				
			}
			c++;
			ctr--;
		}/* while */

        // Handle XTerm rename functions, code contributed by Bill Rausch
        // Modified by JMB to handle ESC]2; case as well.
		if ((escflg >= 6) && (ctr > 0))
		{
			static char*	tmp = null;
			static Str255	newname;
			
			
			if ((escflg == 6) && ((*c == '0') || (*c == '2')))
			{
				escflg++;
            	c++;
            	ctr--;
			}
			if ((escflg == 7) && (ctr > 0) && (*c == ';'))
			{
				ctr--;
				c++;
				escflg++;
				newname[0] = 0;
				tmp = (char*)&newname[1];
			}
			
			while ((escflg == 8) && (ctr > 0) && (*c != 07) && (*c != 033))
			{
        		if (*newname < 255)
				{
            	 	*tmp++ = *c;
            		(*newname)++;
            	}
            	c++;
            	ctr--;
			}
        	if ((escflg == 8) && (*c == 07 || *c == 033) && (ctr > 0))
			{
        		RealScreen_SetWTitle(VSIwn, newname);
           		if (*c != 07)
            	{
            		// this will be undone after the "goto ShortCut" below
            		c--;
                	ctr++;
				}
            	goto ShortCut;
			}
		}/* if */
		
#if 0
		while ((escflg == 8) && (ctr > 0) && (*c != 07))
		{
			*tmp++ = *c++;
			ctr--;
			(*newname)++;
		}
		if ((escflg == 8) && (*c == 07) && (ctr > 0))
		{
			RealScreen_SetWTitle(VSIwn, newname);
			goto ShortCut;
		}
#endif
		
		if ((escflg > 2) && (ctr > 0))
		{
ShortCut:
			if ((VSIw->possibleForce) && (*c != 'H')) VSIw->possibleForce = false;
			escflg = 0;
			c++;
			ctr--;
		}
	}/* while (ctr > 0) */
	
	VSIw->escflg = escflg;
}// Parse


//
// internal methods
//

/*!
Begins printer redirection, as requested by a
directive from a remote terminal application.

This routine has been rewritten in Telnet 3.0 to
solve many issues.  One, the new Folder Manager
is now used to find the Temporary Folder, which
eliminates a couple of useless global variables.
Two, the FSpxxxx() routines from the modern Mac
File Manager are now used, instead of outdated
routines from earlier OS versions.  Three, the
function is now single-entry, single-exit.  And
lastly, it is now internal to this module, which
makes the code faster.

(3.0)
*/
static void
printingBegin	(void)
{
	OSErr		error = noErr;
	FSSpec		temporaryItemsFolder,
				temporaryFile;
	Boolean		fileCreated = false;
	
	
	VSIw->prredirect = 1;
	VSIw->prbuf = 0x00000000;
	sprintf(VSIw->fname, APPLICATION_SHORT" tempfile #%d", VSIwn);
	StringUtilities_CToPInPlace(VSIw->fname);
	
	error = FolderManager_GetFSSpec(kTelnetFolderMacTemporaryItems, &temporaryItemsFolder);
	if (error == noErr) error = FSMakeFSSpec(temporaryItemsFolder.vRefNum, temporaryItemsFolder.parID,
												(StringPtr)VSIw->fname, &temporaryFile);
	
	if ((error == noErr) || (error == fnfErr))
	{
		if ((error = FSpCreate(&temporaryFile, '????', 'TEXT', GetScriptManagerVariable(smKeyScript))) != noErr)
		{
			if (error == dupFNErr) error = noErr, fileCreated = true;
		}
		
		if (error == noErr)
		{
			if ((error = FSpOpenDF(&temporaryFile, fsRdWrPerm, &(VSIw->refNum))) == noErr)
			{
				if (SetEOF(VSIw->refNum, 0L)) error = fileBoundsErr;
			}
		}
	}
	
	if (error != noErr)
	{
		// if any problems occur, cancel the printer redirection and delete the temporary file
		BEEP;
		Alert_ReportOSErr(error);
		VSIw->prredirect = 0;
		if (fileCreated) FSpDelete(&temporaryFile);
	}
}// printingBegin


/*!
Ends printer redirection, as requested by a
directive from a remote terminal application.
The contents of the temporary file (which was
capturing data to print) will be read and sent
to the printer.

This routine has been rewritten in Telnet 3.0 to
solve many issues.  One, the new Folder Manager
is now used to find the Temporary Folder, which
eliminates a couple of useless global variables.
Two, the FSpxxxx() routines from the modern Mac
File Manager are now used, instead of outdated
routines from earlier OS versions.  Three, the
function is now single-entry, single-exit.  And
lastly, it is now internal to this module, which
makes the code faster.

(3.0)
*/
static void
printingEnd		(void)
{
	Boolean		failed = false;
	
	
	if (VSIw->prredirect)
	{
		FSSpec		temporaryItemsFolder,
					temporaryFile;
		OSErr		error = noErr;
		
		
		error = FolderManager_GetFSSpec(kTelnetFolderMacTemporaryItems, &temporaryItemsFolder);
		if (error == noErr) error = FSMakeFSSpec(temporaryItemsFolder.vRefNum, temporaryItemsFolder.parID,
													(StringPtr)VSIw->fname, &temporaryFile);
		
		VSIw->prredirect = 0;
		
		// PRINTING BEGIN (ideally, externalize this in TelnetPrinting.c)
		{
			GrafPtr						oldPort = null;
			UniversalPrintRecordRef		printRecordRef = null;
			
			
			GetPort(&oldPort);
			
			Printing_Init();
			printRecordRef = Printing_NewRecord();
			
			Cursors_UseArrow();
			
			if (Printing_PrintJobDialogDisplay(printRecordRef)) // if false, printing was cancelled
			{
				if ((error = Printing_GetLastResult(printRecordRef)) != noErr) {/* error */}
				
				Printing_BeginDocument(printRecordRef);
				if ((error = Printing_GetLastResult(printRecordRef)) != noErr) failed = true;
				else
				{
					if (VSIwn >= 0)
					{
						printPages(printRecordRef, (StringPtr)null/* dont print title */,
									VSmaxwidth(VSIwn), null, VSIw->refNum);
						Printing_EndDocument(printRecordRef);
						if ((error = Printing_GetLastResult(printRecordRef)) != noErr) {/* error */}
					}
					else failed = true;
				}
			}
			
			Printing_DisposeRecord(&printRecordRef);
			Printing_Done();
			
			SetPort(oldPort);
		}
		// PRINTING END (externalize in TelnetPrinting.c)
		
		if ((error = FSClose(VSIw->refNum)) != noErr) failed = true;
		VSIw->refNum = -1;
		
		if ((error = FSpDelete(&temporaryFile)) != noErr) failed = true;
	}
	
	if (failed) BEEP;
}// printingEnd


/*!
Spools the specified data to disk, scanning for
the special end printer redirection character.
If the redirection ends, the spooling is stopped
and printingEnd() is invoked.

(3.0)
*/
static void
printingSpool		(unsigned char**		pc,
					 short*					pctr)
{
#define ENDOFPRT 	'\033[4i'		/* <ESC>[4i   (0x1B5B3469) */
	SInt32		count = 0L;		// the number of characters to spool to disk
	char*		start = null;	// original start of buffer
	OSErr		error = noErr;
	Boolean		ready = false;	// true if <ESC>[4i
	
	
	if (pc != null)
	{
		start = (char*)*pc;
		
		while ((*pctr > 0) && (!ready))
		{
			VSIw->prbuf = (VSIw->prbuf << 8) + **pc;
			if (VSIw->prbuf == ENDOFPRT)
			{
				// printer redirection end signal received
				ready = true;
				count--; // dont count the redirection signal as a printed character
			}
			count++;
			(*pc)++;
			(*pctr)--;
		}
		
		{
			ConnectionDataPtr		screenPtr = ScreenManager_GetConnectionDataOf(VSIwn);
			
			
			if (screenPtr != null)
			{
				TextTranslation_ConvertTextToMac((unsigned char*)start, count, screenPtr->national);
			}
		}
		if ((error = FSWrite(VSIw->refNum, &count, start)) != noErr) BEEP;
		if (ready || (error != noErr)) printingEnd();
	}
#undef ENDOFPRT
}// printingSpool