/*###############################################################

	WindowInfo.c
	
	Interface Library 1.0
	 1998-2001 by Kevin Grant
	
	This library is free software; you can redistribute it or
	modify it under the terms of the GNU Lesser Public License
	as published by the Free Software Foundation; either version
	2.1 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU Lesser Public License for details.

    You should have received a copy of the GNU Lesser Public
	License along with this library; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

// standard-C includes
#include <limits.h>

// library includes
#include "InterfaceLib.h"

// Mac includes
#include <Dialogs.h>
#include <FixMath.h>
#include <Gestalt.h>
#include <ToolUtils.h>



// types
struct WindowInfo
{
	WindowInfoDescriptor			windowDescriptor;
	Boolean							isFloatingWindow;	// is window considered a floating window?
	Boolean							noDynamicResize;	// force window to drag a resize outline?
	union
	{
		struct
		{
			short		minimumHeight,
						minimumWidth,
						maximumHeight,
						maximumWidth;
		} asShorts;
		Rect		asRect; // identical to the above information: convenient for calling GrowWindow()
	} sizeLimitRect;
	
	Point							preferredLocation;
	Point							preferredSize;
	WindowResizeResponderProcPtr	resizeNotifyMethod;
	SInt32							resizeNotifyMethodData;
	void*							auxiliaryDataPtr;
};
typedef struct WindowInfo		WindowInfo;
typedef struct WindowInfo*		WindowInfoPtr;

// variables
static Boolean		gLiveResizing = true;

// internal prototypes
static WindowInfoPtr	refAcquireLock		(WindowInfoRef		inRef);
static void				refReleaseLock		(WindowInfoRef		inRef);



//
// public methods
//

/*!
To create a new Window Info object, use this method.  To
assign the new feature set to a window, call one of the
WindowInfo_SetForDialog() or WindowInfo_SetForWindow()
methods.  To destroy the feature set you created with
this routine, call WindowInfo_Dispose().

(1.0)
*/
WindowInfoRef
WindowInfo_New	(void)
{
	WindowInfoRef		result = null;
	
	
	result = (WindowInfoRef)Memory_NewPtr(sizeof(WindowInfo));
	WindowInfo_Init(result);
	
	return result;
}// New


/*!
To destroy a reference to Window Info
information previously created with the
WindowInfo_New() method, call this method.

WARNING: Calling this method makes the given
         reference completely invalid.  Do not
		 use it again.  Also, if you stored any
		 auxiliary data pointer in the window
		 features record that needs disposing,
		 use WindowInfo_GetAuxiliaryDataPtr()
		 to obtain and destroy that data first,
		 then call this method.

(1.0)
*/
void
WindowInfo_Dispose	(WindowInfoRef	inWindowInfoRef)
{
	if (inWindowInfoRef != null)
	{
		WindowInfoPtr		windowInfoPtr = refAcquireLock(inWindowInfoRef);
		
		
		DisposePtr((Ptr)windowInfoPtr);
		windowInfoPtr = null;
	}
}// Dispose


/*!
To initialize all of the fields of a new window
features structure to which you have a reference,
call this method.  Normally you do not need to
use this method, because WindowInfo_New() calls
it automatically upon creation of every new
reference.

(1.0)
*/
void
WindowInfo_Init			(WindowInfoRef		inoutWindowInfoRef)
{
	WindowInfoPtr		windowInfoPtr = refAcquireLock(inoutWindowInfoRef);
	
	
	windowInfoPtr->windowDescriptor = kInvalidWindowInfoDescriptor;
	windowInfoPtr->isFloatingWindow = false;
	windowInfoPtr->noDynamicResize = false;
	windowInfoPtr->sizeLimitRect.asShorts.minimumHeight =
		windowInfoPtr->sizeLimitRect.asShorts.minimumWidth = 0;
	windowInfoPtr->sizeLimitRect.asShorts.maximumHeight =
		windowInfoPtr->sizeLimitRect.asShorts.maximumWidth = SHRT_MAX;
	windowInfoPtr->preferredLocation.h =
		windowInfoPtr->preferredLocation.v = 0;
	windowInfoPtr->resizeNotifyMethod = null;
	windowInfoPtr->resizeNotifyMethodData = 0L;
	windowInfoPtr->auxiliaryDataPtr = null;
	refReleaseLock(inoutWindowInfoRef);
}// Init


/*!
To call the window resize responder method
from a particular Window Info structure
to which you have a reference, call this
method.  If a window does not have any resize
response procedure in its general window data,
this routine does nothing and returns "false".

For valid return values from a window resize
responder, see "InterfaceLib.h".

WARNING:	Do not invoke this routine if you
			have not associated Window Info
			with the specified Mac OS window
			using WindowInfo_SetForWindow() or
			WindowInfo_SetForDialog().

(1.0)
*/
SInt16
WindowInfo_CallWindowResizeResponderUPP	(WindowRef const		inWindow,
										 SInt32 const			inDeltaSizeX, 
										 SInt32 const			inDeltaSizeY)
{
	WindowInfoRef		windowInfoRef = WindowInfo_GetFromWindow(inWindow);
	WindowInfoPtr		windowInfoPtr = null;
	SInt16				result = 0;
	
	
	if (windowInfoRef != null)
	{
		windowInfoPtr = refAcquireLock(windowInfoRef);
		if (windowInfoPtr->resizeNotifyMethod != null)
		{
			result = (InvokeWindowResizeResponderProc(windowInfoPtr->resizeNotifyMethod,
														inWindow, inDeltaSizeX, inDeltaSizeY,
														windowInfoPtr->resizeNotifyMethodData));
		}
		refReleaseLock(windowInfoRef);
	}
	return result;
}// CallWindowResizeResponderUPP


/*!
To obtain the auxiliary data pointer of a
particular Window Info structure to which
you have a reference, call this method.  Do not
pass a null reference to this method!

(1.0)
*/
void*
WindowInfo_GetAuxiliaryDataPtr	(WindowInfoRef const	inWindowInfoRef)
{
	WindowInfoPtr		windowInfoPtr = refAcquireLock(inWindowInfoRef);
	void*				result = null;
	
	
	result = windowInfoPtr->auxiliaryDataPtr;
	refReleaseLock(inWindowInfoRef);
	return result;
}// GetAuxiliaryDataPtr


/*!
To obtain a reference to the general window data for a
particular dialog box, call this method.

WARNING: The Window Info information is assumed to
         exist in the reference constant field of the
		 dialogs window.  If you provide a dialog whose
		 window has no Window Info but has a non-null
		 reference constant, using the resultant pointer
		 will most likely cause a pointer referencing
		 error and crash the program.

(1.0)
*/
WindowInfoRef
WindowInfo_GetFromDialog	(DialogRef const	inDialog)
{
	return WindowInfo_GetFromWindow(GetDialogWindow(inDialog));
}// GetFromDialog


/*!
To obtain a reference to the general window data for a
particular window, call this method.

WARNING: The Window Info information is assumed to
         exist in the windows reference constant field.
		 If you provide a window that has no window
		 features but has a non-null reference constant,
		 using the resultant pointer will most likely
		 cause a pointer referencing error and crash the
		 program.

(1.0)
*/
WindowInfoRef
WindowInfo_GetFromWindow	(WindowRef const	inWindow)
{
	return ((WindowInfoRef)GetWRefCon(inWindow));
}// GetFromWindow


/*!
To obtain the window descriptor constant from a
particular general window data reference, call
this method.  Do not pass a null reference to
this method!

(1.0)
*/
WindowInfoDescriptor
WindowInfo_GetWindowDescriptor	(WindowInfoRef const	inWindowInfoRef)
{
	WindowInfoDescriptor		result = kInvalidWindowInfoDescriptor;
	WindowInfoPtr				windowInfoPtr = refAcquireLock(inWindowInfoRef);
	
	
	if (windowInfoPtr != null) result = (windowInfoPtr->windowDescriptor);
	refReleaseLock(inWindowInfoRef);
	return result;
}// GetWindowDescriptor


/*!
To determine the largest size that a resizable dialog
box can have, call this method.

(1.0)
*/
void
WindowInfo_GetWindowMaximumDimensions	(WindowInfoRef const		inWindowInfoRef,
										 Point*						outMaximumSizePtr)
{
	if (outMaximumSizePtr != null)
	{
		WindowInfoPtr		windowInfoPtr = refAcquireLock(inWindowInfoRef);
		
		
		SetPt(outMaximumSizePtr, windowInfoPtr->sizeLimitRect.asShorts.maximumWidth,
							windowInfoPtr->sizeLimitRect.asShorts.maximumHeight);
		refReleaseLock(inWindowInfoRef);
	}
}// GetWindowMaximumDimensions


/*!
To determine the smallest size that a resizable dialog
box can have, call this method.

(1.0)
*/
void
WindowInfo_GetWindowMinimumDimensions	(WindowInfoRef const		inWindowInfoRef,
										 Point*						outMinimumSizePtr)
{
	if (outMinimumSizePtr != null)
	{
		WindowInfoPtr		windowInfoPtr = refAcquireLock(inWindowInfoRef);
		
		
		SetPt(outMinimumSizePtr, windowInfoPtr->sizeLimitRect.asShorts.minimumWidth,
							windowInfoPtr->sizeLimitRect.asShorts.minimumHeight);
		refReleaseLock(inWindowInfoRef);
	}
}// GetWindowMinimumDimensions


/*!
To determine the smallest and largest sizes that a
resizable dialog box can have, call this method.
The returned rectangle pointer is specially set up
to be ideal input to the GrowWindow() system call.
Youre welcome.

(1.0)
*/
Rect*
WindowInfo_GetWindowResizeLimits	(WindowInfoRef const		inWindowInfoRef)
{
	WindowInfoPtr		windowInfoPtr = refAcquireLock(inWindowInfoRef);
	Rect*				result = null;
	
	
	result = &windowInfoPtr->sizeLimitRect.asRect;
	refReleaseLock(inWindowInfoRef);
	return result;
}// GetWindowResizeLimits


/*!
To handle a mouse-down event in the size region of a
window that has general window data associated with it,
call this convenient method.  Using window data, this
method automatically determines the resize constraints
on the window, and, if the user resizes the window, the
universal procedure pointer to invoke in order to
handle the new window size.  The Mac OS user state
for the window is automatically updated to reflect the
new size.

WARNING: Do not pass the Mac OS window pointer of a
         window whose reference constant field refers
		 to anything but a reference to general window
		 data as defined by this module.

(1.0)
*/
SInt16
WindowInfo_GrowWindow	(WindowRef			inWindow,
						 EventRecord*		inoutEventPtr)
{
	WindowInfoRef	windowInfoRef = WindowInfo_GetFromWindow(inWindow);
	SInt16			result = 0;
	
	
	if (windowInfoRef != null)
	{
		WindowInfoPtr	ptr = refAcquireLock(windowInfoRef);
		long			growResult = 0L;
		Rect			contentRect;
		Boolean			smallWindow = false;
		
		
		// Obtain the current size of the window, so when it
		// changes it is possible to calculate deltaX and deltaY.
		// Then, change the rectangle so its origin is (0, 0),
		// to make it easier to work with.  Set a flag if the
		// window is small; small windows are dynamically resized
		// one pixel at a time, instead of using a larger grid.
		{
			RgnHandle		contentRegion = Memory_NewRegion();
			
			
			GetWindowRegion(inWindow, kWindowContentRgn, contentRegion);
			GetRegionBounds(contentRegion, &contentRect);
			Memory_DisposeRegion(&contentRegion);
			if (((contentRect.right - contentRect.left) < 100) ||
				((contentRect.bottom - contentRect.top) < 64)) smallWindow = true; // arbitrary rules here
		}
		
		if ((gLiveResizing) && (!ptr->noDynamicResize))
		{
			RgnHandle		region = Memory_NewRegion();
			
			
			if (region != null)
			{
				WindowInfoPtr		ptr = refAcquireLock(windowInfoRef);
				Rect				structureRect,
									pinnedRect;
				Point				oldMouse,
									currentMouse;
				SInt32				deltaX = 0L,
									deltaY = 0L;
				Boolean				done = false,
									haveAppearance1_1 = false; // minimum version for theme drag sounds
				EventRecord			event;
				register UInt16		x = 0;
				
				
				// Appearance 1.1 or later?
				{
					long		sysv = 0L;
					UInt8		majorRev = 0,
								minorRev = 0;
					
					
					(OSErr)Gestalt(gestaltAppearanceVersion, &sysv);
					majorRev = Releases_GetMajorRevisionForVersion(sysv);
					minorRev = Releases_GetMinorRevisionForVersion(sysv);
					
					haveAppearance1_1 = (((majorRev == 0x01) && (minorRev >= 0x01)) || (majorRev > 0x01));
				}
				
				// set up the mouse variables and mouse-moved region prior to starting the loop
				{
					Point		globalMouse;
					
					
					GetMouse(&globalMouse);
					LocalToGlobal(&globalMouse);
					
					// Set the initial mouse region to be the initial mouse location, only.
					// Note that the "region" variable should remain unchanged between here
					// and the WaitNextEvent() at the start of the loop!
					SetEmptyRgn(region);
					SetRectRgn(region, globalMouse.h, globalMouse.v, globalMouse.h, globalMouse.v);
					SetPt(&oldMouse, globalMouse.h, globalMouse.v);
					SetPt(&currentMouse, globalMouse.h, globalMouse.v);
				}
				
				// The slop measurements force the mouse to return to its location
				// before a window can be resized, should the mouse previously move
				// in the opposite direction beyond the resize limitations.  If
				// this wasnt done, it would be possible for the cursor to drift
				// away from the size box, while its movements would still resize
				// the window, which is very frustrating for users.
				//
				// The slop rectangle is defined as the area you get when you map
				// the lower-right corner of the smallest allowed window boundaries
				// to the lower-right corner of the largest allowed boundaries.  If
				// the mouse moves outside of this area, the window size will not
				// change even though the mouse continues to move.
				{
					Point		zero,
								upperBound,
								lowerBound;
					
					
					// this easily determines the global origin of the window, called "zero"
					{
						GrafPtr		oldPort = null;
						
						
						GetPort(&oldPort);
						SetPortWindowPort(inWindow);
						SetPt(&zero, 0, 0);
						LocalToGlobal(&zero);
						SetPort(oldPort);
					}
					
					// Determine the global coordinates of the lower-right corners
					// of the smallest and largest allowed window sizes by adding
					// the coordinates of "zero" to the dimensions.  Careful, the
					// extreme width/height is often set to SHRT_MAX!  Since this
					// integer is as large as it can be, it is incorrect to simply
					// offset this value!  An explicit boundary check is therefore
					// necessary.
					SetPt(&upperBound, ptr->sizeLimitRect.asShorts.minimumWidth,
										ptr->sizeLimitRect.asShorts.minimumHeight);
					if (upperBound.h < (SHRT_MAX - zero.h)) upperBound.h += zero.h;
					if (upperBound.v < (SHRT_MAX - zero.v)) upperBound.v += zero.v;
					SetPt(&lowerBound, ptr->sizeLimitRect.asShorts.maximumWidth,
										ptr->sizeLimitRect.asShorts.maximumHeight);
					if (lowerBound.h < (SHRT_MAX - zero.h)) lowerBound.h += zero.h;
					if (lowerBound.v < (SHRT_MAX - zero.v)) lowerBound.v += zero.v;
					
					SetRect(&pinnedRect, upperBound.h, upperBound.v, lowerBound.h, lowerBound.v);
				}
				
				// Under Mac OS 8.5 or later, use an Appearance drag sound.
				if (haveAppearance1_1)
				{
					(OSStatus)BeginThemeDragSound(kThemeDragSoundGrowWindow);
				}
				
				// In a special event loop, dynamically resize the specified window.
				// The use of an event loop allows the mouse to be tracked accurately,
				// and it allows the system to periodically service update events in
				// *other* applications while the window is being resized!
				while (!done)
				{
					WaitNextEvent(mUpMask | osMask, &event, 4L/* sleep ticks */, region/* mouse-moved region */);
					
					done = (event.what == mouseUp); // stop when the mouse button is released
					if (event.what == osEvt)
					{
						// when the mouse moves, change the window size
						if (((event.message & osEvtMessageMask) >> 24) == mouseMovedMessage)
						{
							// It is possible to cheat by holding the perceived mouse
							// location within the resize limits, thereby making it
							// impossible for the mouse to seem outside this area, no
							// matter where the mouse really is.
							{
								SInt32		pinResult = 0L;
								
								
								pinResult = PinRect(&pinnedRect, event.where/* real mouse location */);
								SetPt(&currentMouse/* perceived location */, LoWord(pinResult), HiWord(pinResult));
							}
							
							deltaX = currentMouse.h - oldMouse.h;
							deltaY = currentMouse.v - oldMouse.v;
							
							SetEmptyRgn(region);
							GetWindowRegion(inWindow, kWindowContentRgn, region);
							GetRegionBounds(region, &contentRect);
							
							// constrain the delta to a multiple of 4 (arbitrary) pixels in each direction
							{
								SInt16		diff = 0;
								
								
								diff = (smallWindow) ? 1 : (deltaX % 4);
								currentMouse.h -= diff;
								deltaX -= diff;
								diff = (smallWindow) ? 1 : (deltaY % 4);
								currentMouse.v -= diff;
								deltaY -= diff;
							}
							
							if ((deltaX) || (deltaY))
							{
								// resize the window frame, and then issue a notification to resize controls, too
								SizeWindow(inWindow, contentRect.right - contentRect.left + deltaX,
													contentRect.bottom - contentRect.top + deltaY, false);
								result = WindowInfo_CallWindowResizeResponderUPP(inWindow, deltaX, deltaY);
								
								// updates are slow, so dont perform one *every* time the mouse moves
								if (!(x % 50/* arbitrary; the larger this number, the rarer the updates */))
								{
									ControlRef		root = null;
									
									
									(OSErr)GetRootControl(inWindow, &root);
									SetEmptyRgn(region);
									#if TARGET_API_MAC_OS8
									GetWindowUpdateRgn(inWindow, region);
									#else
									GetWindowRegion(inWindow, kWindowUpdateRgn, region);
									#endif
									BeginUpdate(inWindow);
									//(OSErr)SetControlVisibility(root, false/* visibility */, true/* draw */);
									//UpdateControls(inWindow, region);
									//(OSErr)SetControlVisibility(root, true/* visibility */, true/* draw */);
									(OSErr)Embedding_OffscreenControlOperation(inWindow, null/* control */,
																				null/* procedure */,
																				0L/* data 1 */, 0L/* data 2 */);
									EndUpdate(inWindow);
								}
							}
							SetPt(&oldMouse, currentMouse.h, currentMouse.v);
							
							// set the mouse-moved region to be the exact point where the mouse is now
							SetEmptyRgn(region);
							SetRectRgn(region, currentMouse.h, currentMouse.v, currentMouse.h, currentMouse.v);
						}
					}
					x++;
					(OSErr)YieldToAnyThread();
				}
				
				// get rid of extra events
				FlushEvents(mUpMask | osMask, 0);
				
				// Under Mac OS 8.5 or later, use an Appearance drag sound.
				if (haveAppearance1_1)
				{
					(OSStatus)EndThemeDragSound();
				}
				
				SetEmptyRgn(region);
				GetWindowRegion(inWindow, kWindowStructureRgn, region);
				GetRegionBounds(region, &structureRect);
				SetWindowUserState(inWindow, &structureRect);
				
				Memory_DisposeRegion(&region);
				
				refReleaseLock(windowInfoRef);
			}
		}
		else
		{
			SetRect(&contentRect, 0, 0, (contentRect.right - contentRect.left),
					(contentRect.bottom - contentRect.top));
			
			growResult = GrowWindow(inWindow, inoutEventPtr->where,
									WindowInfo_GetWindowResizeLimits(windowInfoRef));
			if (growResult != 0)
			{
				SInt32		deltaX = 0L,
							deltaY = 0L;
				
				
				// at this point, contentRect is at a (0, 0) origin, so
				// the "right" and "bottom" components are actually the
				// width and height, respectively
				deltaX = LoWord(growResult) - contentRect.right;
				deltaY = HiWord(growResult) - contentRect.bottom;
				
				// display a cool animation that gradually changes the size when resizing dialogs
				if (GetWindowKind(inWindow) == kDialogWindowKind)
				{
					enum
					{
						// increase this value to animate this effect
						kDeltaDiv = 1	// number of times the window size changes during the animation
					};
					SInt16				width = contentRect.right - contentRect.left,
										height = contentRect.bottom - contentRect.top;
					SInt16				divX = 0,
										divY = 0,
										dx = 0,
										dy = 0;
					register SInt16		i = 0;
					
					
					// how much the window changes per animation stage
					divX = FixRound(FixRatio(deltaX, kDeltaDiv));
					divY = FixRound(FixRatio(deltaY, kDeltaDiv));
					
					// animate!
					for (i = 0; i < kDeltaDiv; i++)
					{
						// calculate the amount to change the window size by, and avoid
						// problems due to limited precision by ensuring the real
						// change in size is ultimately used (otherwise, rounding error
						// would potentially make the window size a few pixels out)
						width += divX;
						dx += divX;
						if (deltaX >= 0) while (dx > deltaX) dx--, divX--, width--;
						else while (dx < deltaX) dx++, divX++, width++;
						height += divY;
						dy += divY;
						if (deltaY >= 0) while (dy > deltaY) dy--, divY--, height--;
						else while (dy < deltaY) dy++, divY++, height++;
						
						// resize the window and its controls
						SizeWindow(inWindow, width, height, false);
						result = WindowInfo_CallWindowResizeResponderUPP(inWindow, divX, divY);
					}
				}
				else
				{
					// resize the window and its controls
					SizeWindow(inWindow, contentRect.right - contentRect.left + deltaX,
								contentRect.bottom - contentRect.top + deltaY, false);
					result = WindowInfo_CallWindowResizeResponderUPP(inWindow, deltaX, deltaY);
				}
				
				// update the zoom box settings
				{
					RgnHandle		region = Memory_NewRegion();
					
					
					if (region != null)
					{
						Rect		structureRect;
						
						
						GetWindowRegion(inWindow, kWindowStructureRgn, region);
						GetRegionBounds(region, &structureRect);
						Memory_DisposeRegion(&region);
						SetWindowUserState(inWindow, &structureRect);
					}
				}
			}
		}
		refReleaseLock(windowInfoRef);
	}
	else BEEP;
	
	return result;
}// GrowWindow


/*!
To determine if a window is a utility window
that can remain active even while another window
is frontmost, use this method.  This simply
returns a flag previously set by a call to
WindowInfo_SetWindowFloating().

(1.0)
*/
Boolean
WindowInfo_IsWindowFloating	(WindowInfoRef const	inWindowInfoRef)
{
	Boolean			result = false;
	WindowInfoPtr	windowInfoPtr = refAcquireLock(inWindowInfoRef);
	
	
	if (windowInfoPtr != null) result = (windowInfoPtr->isFloatingWindow);
	refReleaseLock(inWindowInfoRef);
	return result;
}// IsWindowFloating


/*!
To obtain the auxiliary data pointer of a
particular Window Info structure to which
you have a reference, call this method.

WARNING: If any problems occur, null is returned.

(1.0)
*/
void
WindowInfo_SetAuxiliaryDataPtr	(WindowInfoRef	inWindowInfoRef,
								 void*			inAuxiliaryDataPtr)
{
	WindowInfoPtr		windowInfoPtr = refAcquireLock(inWindowInfoRef);
	
	
	windowInfoPtr->auxiliaryDataPtr = inAuxiliaryDataPtr;
	refReleaseLock(inWindowInfoRef);
}// SetAuxiliaryDataPtr


/*!
To specify whether windows are resized dynamically
(where the window size physically changes as the
user drags, with no drag outline), use this method.

IMPORTANT:	To globally set this flag for all
			windows, pass "null" for the object
			reference.  Generally, you always set
			the global flag, and then pass a
			specific Window Info object to reverse
			the setting for particular windows.

(1.0)
*/
void
WindowInfo_SetDynamicResizing		(WindowInfoRef		inWindowFeaturesRefOrNullToSetGlobalFlag,
									 Boolean const		inLiveResizeEnabled)
{
	if (inWindowFeaturesRefOrNullToSetGlobalFlag != null)
	{
		WindowInfoPtr		ptr = refAcquireLock(inWindowFeaturesRefOrNullToSetGlobalFlag);
		
		
		if (ptr != null)
		{
			ptr->noDynamicResize = !inLiveResizeEnabled;
			refReleaseLock(inWindowFeaturesRefOrNullToSetGlobalFlag);
		}
	}
	else
	{
		gLiveResizing = inLiveResizeEnabled;
	}
}// SetDynamicResizing


/*!
To specify if a window is a utility window that
can remain active even while another window is
frontmost, use this method.  This simply sets
a flag that can subsequently be read by a call
to WindowInfo_IsWindowFloating().

(1.0)
*/
void
WindowInfo_SetWindowFloating	(WindowInfoRef const	inWindowInfoRef,
								 Boolean const			inIsFloating)
{
	WindowInfoPtr	windowInfoPtr = refAcquireLock(inWindowInfoRef);
	
	
	if (windowInfoPtr != null) windowInfoPtr->isFloatingWindow = inIsFloating;
	refReleaseLock(inWindowInfoRef);
}// SetWindowFloating


/*!
To specify a general window data reference for a
particular dialog box, call this method.

WARNING: The Window Info information must be placed
         in the reference constant field of the specified
		 dialogs window.  If you provide a window that
		 needs to use the reference constant field for
		 something else, you cannot use Window Info.

(1.0)
*/
void
WindowInfo_SetForDialog	(DialogRef const			inDialog,
						 WindowInfoRef const		inWindowInfoRef)
{
	WindowInfo_SetForWindow(GetDialogWindow(inDialog), inWindowInfoRef);
}// SetForDialog


/*!
To specify a general window data reference for a
particular window, call this method.

WARNING: The Window Info information must be placed
         in the reference constant field of the specified
		 window.  If you provide a window that needs to use
		 the reference constant field for something else,
		 you cannot use Window Info.

(1.0)
*/
void
WindowInfo_SetForWindow	(WindowRef const			inWindow,
						 WindowInfoRef const		inWindowInfoRef)
{
	SetWRefCon(inWindow, (UInt32)inWindowInfoRef);
}// SetForWindow


/*!
To specify the window descriptor constant from a
particular Window Info structure to which you
have a reference, call this method.

(1.0)
*/
void
WindowInfo_SetWindowDescriptor	(WindowInfoRef					inWindowInfoRef,
								 WindowInfoDescriptor const		inNewWindowDescriptor)
{
	WindowInfoPtr		windowInfoPtr = refAcquireLock(inWindowInfoRef);
	
	
	windowInfoPtr->windowDescriptor = inNewWindowDescriptor;
	refReleaseLock(inWindowInfoRef);
}// SetWindowDescriptor


/*!
To specify the limits on the dimensions of a dialog box
that is associated with Window Info information, call
this method.  A dialog box that does not have a grow box
does not need to define this information in its window
features.

(1.0)
*/
void
WindowInfo_SetWindowResizeLimits	(WindowInfoRef		inWindowInfoRef,
									 short const		inMinimumHeight,
									 short const		inMinimumWidth,
									 short const		inMaximumHeight,
									 short const		inMaximumWidth)
{
	WindowInfoPtr		windowInfoPtr = refAcquireLock(inWindowInfoRef);
	
	
	windowInfoPtr->sizeLimitRect.asShorts.minimumHeight = inMinimumHeight;
	windowInfoPtr->sizeLimitRect.asShorts.minimumWidth = inMinimumWidth;
	windowInfoPtr->sizeLimitRect.asShorts.maximumHeight = inMaximumHeight;
	windowInfoPtr->sizeLimitRect.asShorts.maximumWidth = inMaximumWidth;
	refReleaseLock(inWindowInfoRef);
}// SetWindowResizeLimits


/*!
To specify the method that is invoked following
a size change to a window supporting window
features, call this method.  The resize responder
usually handles rudimentary cleanup tasks, such
as moving and/or resizing content controls and
redrawing any regions that have not been updated.

(1.0)
*/
void
WindowInfo_SetWindowResizeResponderUPP	(WindowInfoRef						inWindowInfoRef,
										 WindowResizeResponderProcPtr const	inNewProc,
										 SInt32 const						inData)
{
	WindowInfoPtr		windowInfoPtr = refAcquireLock(inWindowInfoRef);
	
	
	windowInfoPtr->resizeNotifyMethod = inNewProc;
	windowInfoPtr->resizeNotifyMethodData = inData;
	refReleaseLock(inWindowInfoRef);
}// SetWindowResizeResponderUPP


//
// internal methods
//

/*!
Use this method to acquire a pointer to the internal
structure, given a reference to it.  Get in the habit
of calling refReleaseLock() when finished with the
pointer you receive; this allows flexibility in the
future to make dynamically-allocated memory relocatable.

(1.0)
*/
static WindowInfoPtr
refAcquireLock	(WindowInfoRef		inRef)
{
	return ((WindowInfoPtr)inRef);
}// refAcquireLock


/*!
To release the lock on a reference that you previously
locked with refAcquireLock(), use this method.

(1.0)
*/
static void
refReleaseLock	(WindowInfoRef		inRef)
{
	// the memory is not relocatable, so nothing needs to be done here
}// refReleaseLock

