/*###############################################################

	Network.c
	
	This module consists of low-level network routines that
	abstract the actual method of connection (MacTCP or
	Open Transport).  Use these methods to connect to the
	network, read or write data, and more.
	
	MacTelnet 3.0,  1998-2001 by Kevin Grant.
	Based on "NCSA Telnet",  1986-1994 UofI Board of Trustees.
	
	This program is free software; you can redistribute it or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version
	2 of the License, or (at your option) any later version.
	
    This program is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
	PURPOSE.  See the GNU General Public License for more
	details.

    You should have received a copy of the GNU General Public
	License along with this program; if not, write to:
	
		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA

###############################################################*/

#include "UniversalDefines.h"

#ifndef __NETWORK__
#define __NETWORK__

// Mac includes
#include <OpenTransportProviders.h>



// constants
enum
{
	kNetworkMaximumNumberOfPorts = 30
};

typedef SInt16	NetworkPortType;
enum
{
	// These determine what routine should be called
	// when data arrives on port.
	kInvalidNetworkPortType = -1,
	kNetworkPortTypeStandardClientSession = 0,	// a telnet or SSH client session with its own terminal window
	kNetworkPortTypeFTPClientSession = 1,		// an FTP client session with its own terminal window
	kNetworkPortTypeFTPServerSession = 2,		// an FTP server session
	kNetworkPortTypeFTPServerData = 3			// a connection on the FTP data port initiated for the purpose of transferring a file
};

// types
typedef SInt16	NetworkPortRef;		// this is NOT A TCP PORT NUMBER
enum
{
	kInvalidNetworkPortRef = -1
};



/*###############################################################
	INITIALIZING AND FINISHING WITH THE NETWORK MODULE
###############################################################*/

// CALL THIS METHOD BEFORE ANY OTHER ROUTINE IN THIS MODULE
void
	Network_Init									(void);

// CALL THIS METHOD WHEN YOU ARE COMPLETELY DONE WITH THE NETWORK MODULE
void
	Network_Done									(void);


/*###############################################################
	CREATING AND DESTROYING SOCKETS
###############################################################*/

// RETURNS A PORT NUMBER (NOT A TCP PORT) THAT CAN BE USED FOR ALL SUBSEQUENT REFERENCES TO THE SOCKET
NetworkPortRef
	Network_SocketOpen								(InetHost			inIPNumber,
													 InetPort const		inRemotePortNumber,
													 InetPort const		inLocalPortNumber,
													 SInt8 const		inTimeoutInSeconds);

SInt16
	Network_CloseConnection							(NetworkPortRef		inPort);

/*###############################################################
	WORKING WITH SOCKETS
###############################################################*/

SInt16
	Network_GetData									(NetworkPortRef		inPort,
													 void*				outBuffer,
													 SInt16 const		inNumberOfBytes);

void
	Network_GetFourCharIPAddress					(unsigned char*		outIPAddressAsFourChars);

SInt32
	Network_GetNumberOfBytesOfRoomOnPort			(NetworkPortRef		inPort);

SInt32
	Network_GetNumberOfBytesWaitingOnPort			(NetworkPortRef		inPort);

NetworkPortType
	Network_GetPortType								(NetworkPortRef		inPort);

NetworkPortRef
	Network_GetFirstPortOfType						(NetworkPortType	inPortType);

SInt16
	Network_IsSessionOnPortBroken					(NetworkPortRef		inPort);

NetworkPortRef
	Network_ListenToTCPPortAndGetItsPortNumber		(InetPort			serv);

SInt16
	Network_MakeFTPQueriesOnPort					(NetworkPortRef		inPort,
													 InetHost*			outRemoteHostIPAddress,
													 InetPort*			outLocalPort,
													 InetPort*			outRemotePort);

void
	Network_MarkNextTransmissionAsUrgent			(void);

SInt16
	Network_SendData								(NetworkPortRef		inPort,
													 void*				inBuffer,
													 SInt16 const		inNumberOfBytes);

SInt16
	Network_SendFlush								(NetworkPortRef		inPort);

SInt16
	Network_SendFlushData							(NetworkPortRef		inPort,
													 void*				inBuffer,
													 SInt16 const		inNumberOfBytes);

void
	Network_SetPortForNextOpen						(NetworkPortRef		inPort);

SInt16
	Network_SetPortKerberosEncryptionState			(NetworkPortRef		inPort,
													 Ptr				inEncryptionData);

void
	Network_SetPortType								(NetworkPortRef		inPort,
													 NetworkPortType	inPortType);

void
	Network_SetSendTimeoutInSeconds					(SInt8 const		inTimeout);

/*###############################################################
	UTILITIES
###############################################################*/

Boolean
	Network_Connect									(Boolean			inAskUserIfNotConnected);

Boolean
	Network_CurrentIPAddressToCString				(char*				outString);

Boolean
	Network_IsConnected								(void);

void
	Network_ShowIPAddress							(void);

#endif